#ifndef VASSILIS_AUXILIARIES_H
#define VASSILIS_AUXILIARIES_H

#include "vplatform.h"

#include <assert.h>
#include <cstdarg>
#include <stdio.h>
#include <list>
#include <vector>
#include <math.h>


#include <map>

#include "v_types.h"
#include "general_templates.h"

class class_file;

class vPoint
{
public:
 long row, col;
 vPoint() {}
 vPoint(long in_row, long in_col)
 {
   row = in_row;
   col = in_col;
 }
};

// like printf, but it flushes the output.
int vPrintAndFlush(const char * format, ...);

// like scanf, but it doesn't give up until all fields have been
// successfully matched.
int vScanAndVerify(const char * format, ...);

// like scanf, but it doesn't give up until all fields have been
// successfully matched.
int vVScanAndVerify(const char * format, va_list arguments);

// Like fscanf, but it takes a va_list as an argument (this comes with C++
// in some platforms, called vfscanf)
int vFscan(FILE * fp, const char * format, va_list arguments);

// Get a non-empty line from a file. If the first non-empty line
// found exceeds max_length, only the first max_length characters
// are returned. It returns the length of the string (with the
// newline character removed).
long vFileGetNonEmptyLine(FILE * fp, char * buffer, long max_length);

long vShellGetLine(char * buffer, long max_length);

long vGetNonEmptyLine(char * buffer, long max_length);


class Bitmap
{
private:
 long size;
 long byte_size;
 uchar * data;


 // I use buffer as a continuous memory chunk, that first contains */
 // the value of size and then it contains the data. I thought that */
 // might be useful if I wanted all bitmaps in the HuffmanCodedData */
 // class to be contiguous in memory, so that we could Huffman encode */
 // data that is already Huffman encoded. Eventually I changed my */
 // mind about Huffman encoding implementation, so maybe I should get */
 // rid of buffer and the constructor that takes a buffer as an input */
 uchar * buffer;

 ushort delete_buffer; // Zero if buffer was not allocated by constructor.

public:
 Bitmap(long the_size);
 Bitmap(long the_size, uchar * the_buffer);
 ~Bitmap();
 void Save(const char * filename);
 void Save(FILE * fp);
 static Bitmap * Load(const char * filename);
 static Bitmap * Load(FILE * fp);
 long LoadData(FILE * fp);
 void Write(long index, uchar value);
 uchar Read(long index);
 void Print();
 long ByteSize();
 long BufferSize();
 long Size();
 const uchar * GetData();
 uchar * WritableData();
};


class vBoundingBox
{
public:
 long top, left, bottom, right;
 // An uninitialized bounding box should have invalid values:
 // bottom < top, right < left.
 vBoundingBox()
 {
   top = 0;
   left = 0;
   bottom = -1;
   right = -1;
 }

 vBoundingBox(long in_top, long in_bottom, long in_left, long in_right)
 {
   top = in_top;
   bottom = in_bottom;
   left = in_left;
   right = in_right;
 }

 inline long Height()
 {
   return bottom - top + 1;
 }

 inline long Width()
 {
   return right - left + 1;
 }

 inline long MaxSide()
 {
   return Max(Height(), Width());
 }

 inline long MinSide()
 {
   return Min(Height(), Width());
 }

 inline long CenterRow()
 {
   return (top + bottom) / 2;
 }

 inline long CenterCol()
 {
   return (left + right) / 2;
 }

 inline void Print()
 {
   vPrint("\ncenter = (%.1lf, %.1lf), top = %li, bottom = %li, left = %li, \
            right = %li, height = %li, width = %li\n",
           (double) (top + bottom + 0.0) / 2.0, (double) (left + right + 0.0) / 2.0,
           top, bottom, left, right, bottom - top + 1, right - left + 1);
 }

};


// The class Label was created to label connected components
// of an image.
class Label : public vBoundingBox
{
public:
 long id;
 long count;

 // r, g, b are colors that we assign to the label.
 // this is used to generate an RGB image where
 // each label gets its own color
 uchar r;
 uchar g;
 uchar b;

 // value is the value of the connected component
 // to which the label corresponds
 uchar value;

 Label();
};

typedef Label class_label;

class LabelMore
{
public:
 ushort operator() (const Label & label1, const Label & label2);
};

class LabelBoxMore
{
public:
 ushort operator() (const Label & label1, const Label & label2);
};

ushort vBoxesOverlap(vBoundingBox & b1, vBoundingBox & b2);


class vDate
{
public:
 long year;
 long month;
 long day;
 long hour;
 long minute;
 long second;
 long millisecond;

 ushort Read(FILE * fp);
 void Print(FILE * fp);
};


// Class vStringTable is an inefficient but simple way to create tables
// that map strings to objects. I could have used the STL map class instead,
// hopefully this class will lead to simple code. I intend to use it
// only for tables that will contain a few objects, and where efficiency is
// not really an issue.

class vStringTable
{
private:
 std::vector<char *> keys;
 std::vector<void *> objects;

public:
 vStringTable();
 ~vStringTable();

 // Removes all keys and objects, so that this table becomes empty.
 long Clear();

 // Store object under key. Fail if key is already used.
 long Put(const char * key, void * object);

 // Return object stored under given key
 void * Get(const char * key);

 // Remove object stored under given key, and return it.
 void * Remove(const char * key);

 // Return the number of objects stored.
 long Size();

 // Return the i-th object stored.
 void * GetIndex(long index);

 // Remove the i-th object, and return it.
 void * RemoveIndex(long index);

 // Returns true if the key is being used already, false
 // otherwise.
 long KeyExists(const char * key);

 // Returns the number of keys in the table that are mapped to
 // the given object.
 long CountKeys(void * object);

 // It returns some key that is mapped to the given object. If no
 // such key exists, then it returns 0.
 const char * GetKey(void * object);
};


void vCheckArgs(int arg, int low, int high, const char * error);
short vIsGIF(const char * filename);
short vIsJPEG(const char * filename);
short vIsPPM(const char * filename);
short vIsPGM(const char * filename);

long vFilesize(const char * filename);
unsigned short vCopyFile(const char * source, const char * target);
short vStringEndsIn(const char * string, const char * ending);
short vStringCaseEndsIn(const char * string, const char * ending);

void vAdjustDirectoryName(char * directory);

long vGetNumber(char * directory);

void vIncreaseNumber(char * directory);

void vAddArrays(long * array1, long * array2, long size);

void vSubtractArrays(long * array1, long * array2, long size);

char * vAdjustEnding2(const char * source, const char * ending);

void vAdjustEnding3(const char * source, char * target, const char * ending);

char * vCopyString(const char * source);

char * vCopyString(const char * source, long start, long end);

// Replaces every occurrence of old_char in str with new_char.
long vReplaceCharacter(char * str, char old_char, char new_char);

char * vLongToString(long number);

// this function adds zeros to the beginning of the string, so that
// the length of the result matches the argument "length".
char * function_number_to_string(long number, long length);

char * vMergeStrings2(const char * first, const char * second);

char * vMergeStrings3(const char * first, const char * second,
                      const char * third);

char * vMergeStrings4(const char * first, const char * second,
                      const char * third, const char * fourth);

char * vMergeStrings5(const char * first, const char * second,
                      const char * third, const char * fourth,
                      const char * fifth);

// It returns the shortest ending substring of filename that
// starts with the '.' character.
char * vGetEnding(const char * filename);

// It returns result of removing the filename's ending
// from the filename (see comments before vGetEnding on
// how the "ending" of a file is defined). The difference
// between this function and vFileTitle is that vFileTitle
// discards the path part of the filename.
char * vRemoveEnding(const char * filename);

char * vFileTitle(const char * filename);


// If name ends in old_ending, we return a copy of the name, but
// adjusted so that the old ending is replaced by the new
// ending. If name doesn't end in old_ending, then we just append
// the new ending to it.
char * vReplaceEnding3(const char * name, const char * old_ending,
                      const char * new_ending);

// vReplaceEndings2 finds the ending of the name(defined as the
// smallest ending substring of name that starts with '.') and
// replaces it with new_ending  (which should start with the
// character '.' if we want to have a dot before the ending in
// the result).
char * vReplaceEnding2(const char * name, const char * new_ending);

char * vParentDir(const char * filename, const char * dir_separator);

char * vSimpleName(const char * filename, const char * separator);

ushort vPathComponents(const char * filename, std::vector<voidp> * components);

unsigned short vFileExists(const char *filename);

// This function checks if a directory exists by trying to create
// a file called directory/trash and checking if it succeeds.
// If it doesn't succeed, then we assume that the directory doesn't
// exist. Otherwise we know that it exists.
//
// Things can go wrong if the directory exists but we failed to
// create a file. Here are two examples when that may happen: If
// directory/trash already exists and we don't have write permission to
// it (or to the whole directory). Also, if the disk is full and we
// fail to create the file because of that. I
// can't think of any other cases
ushort vDirectoryExists(const char * filename);

ulong vMaskedValue(ulong number, ulong mask);

uchar vMaskShift(ulong mask);

void vEraseCharpVector(std::vector<voidp> * strings);

void vEraseCharpVector(std::vector<charp> * strings);

static inline void function_erase_string_vector(std::vector<charp> * strings)
{
 vEraseCharpVector(strings);
}

vint4 vHostToPCLong(long number);

ushort vHostToPCUshort(ushort number);

long vPCToHostLong(vint4 number);

long vUnixToHostLong(long number);

ushort vPCToHostUshort(ushort number);

long vReadVint8(FILE * fp, vint8 * a);

long vReadVints8(FILE * fp, vint8 * numbers, long items);

long vWriteVint8(FILE * fp, vint8 a);

long vWriteVints8(FILE * fp, vint8 * numbers, long items);

void vGetVint8Bytes(vint8 a, uchar * buffer);

long vReadPCUshorts(FILE * fp, ushort * numbers, long items);

long vReadUnixUshorts(FILE * fp, ushort * numbers, long items);

long vReadUnixLong(FILE * fp, long * number);

long vReadUnixLongs(FILE * fp, long * numbers, long items);

long vReadPCLong(FILE * fp, long * number);

long vReadPCLong(class_file * fp, long * number);

long vReadPCLongs(FILE * fp, long * numbers, long items);

long vReadPCLongs(class_file * fp, long * numbers, long items);

long vWritePCUshorts(FILE * fp, ushort * numbers, long items);

long vWritePCLong(FILE * fp, long number);

long vWritePCLong(class_file * fp, long number);

long vWritePCLongs(FILE * fp, long * numbers, long items);

long vWritePCLongs(class_file * fp, long * numbers, long items);

long vReadPCDoubles(FILE * fp, double * numbers, long items);

long vWritePCDoubles(FILE * fp, double * numbers, long items);

ushort vFilePatternsMatch(const char * pattern, const char * filename);

void vWaitForEnter();

char * vJoinPaths(const char * path1, const char * path2);

char * vJoinPaths3(const char * path1, const char * path2,
                   const char * path3);

char * vJoinPaths4(const char * path1, const char * path2,
                   const char * path3, const char * path4);

char * vJoinPaths5(const char * path1, const char * path2,
                   const char * path3, const char * path4, const char * path5);

static inline char * vJoinPaths2(const char * path1, const char * path2)
{
 return vJoinPaths(path1, path2);
}

long vTypeToLong(Type type);

Type vLongToType(long number);

long vSizeOfType(Type type);

char * vReadString(FILE * fp);

long vWriteString(FILE * fp, const char * string);

long vWritePCString(FILE * fp, char * string);

char * vReadPCString(FILE * fp);

// Assuming that all the data in the is numbers, the numbers
// are read and stored in the vector "numbers".
long vReadDoubleVectorB(const char * filename,
                       std::vector<double> * numbers);
long vReadDoubleVectorA(FILE * fp, std::vector<double> * numbers);

// True if the file with the given filename can be opened for writing.
// False otherwise (for example, if the directory specified by the
// filename doesn't exist).
unsigned short vWritableFileExists(char * filename);

// Returns the integer that is the nearest to the number.
// x.5 gets rounded to floor(x) + 1.
//inline long vRound2(double number)
//{
//  double lower = floor(number);
//  double decimal_part = number - lower;
//  if (decimal_part < 0.5)
//  {
//    return lower;
//  }
//  else
//  {
//    return lower+1;
//  }
//}


inline long vRound(double number)
{
 if (number >= 0)
 {
   return (long) (number + (double) 0.5);
 }
 else
 {
   return (long) (number - (double) 0.5);
 }
}

inline long vRound(long number)
{
 return number;
}


inline long vRound(float number)
{
 if (number >= 0)
 {
   return (long) (number + (float) 0.5);
 }
 else
 {
   return (long) (number - (float) 0.5);
 }
}


inline long round_number(double number)
{
 if (number >= 0)
 {
   return (long) (number + (double) 0.5);
 }
 else
 {
   return (long) (number - (double) 0.5);
 }
}

inline long round_number(long number)
{
 return number;
}


inline long round_number(float number)
{
 if (number >= 0)
 {
   return (long) (number + (float) 0.5);
 }
 else
 {
   return (long) (number - (float) 0.5);
 }
}


// Length required to write a decimal number with characters. We include
// a character for the minus sign, if the number is negative.
long vDecimalLength(long number);

// Returns a random 32-bit integer
ulong vRandomLong();

// Returns a random number, uniformly distributed between low and high
// (including low and high).
long vRandomLong2(long low, long high);

// Returns a random number, uniformly distributed between low and high
double vRandomDouble(double low, double high);

// exit_error is a handy way to exit when we encounter a condition
// that the code cannot handle. Sometimes that condition is caused by
// bugs, sometimes it is simply stuff I chose not to address to save
// time. Calling this function we get a helpful message. Another very
// useful thing is that, in debug mode, if we set a breakpoint on
// this function, we can easily get to the lines of code that demonstrate
// the problem.
long exit_error(char * format, ...);


// Choose "number" numbers between low and high, and store them
// in result. No repetitions are allowed.
long vPickWithoutRepetitions2(long low, long high, long number,
                              std::vector<long> * result);

// vPair is used so that we can sort an object by some value, but
// still have access to the object itself. This is achieved
// using the function S_PairLess.
class vPair
{
public:
 float value;
 void * object;

 vPair()
 {
 }

 vPair(float in_value, void * in_object)
 {
   value = in_value;
   object = in_object;
 }
};


class vPairLess
{
public:
 ushort operator() (const vPair & object1, const vPair & object2)
 {
   return (object1.value < object2.value);
 }
};


// class_string_couple is used so that we can sort an object by some
// string value, but
// still have access to the object itself. This is achieved
// using the function S_PairLess.
class class_string_couple
{
public:
 char* value;
 void * object;

 class_string_couple ()
 {
 }

 class_string_couple (char* in_value, void * in_object)
 {
   value = in_value;
   object = in_object;
 }
};


class string_couple_less
{
public:
 ushort operator() (const class_string_couple & object1, const
class_string_couple & object2)
 {
   return (strcmp (object1.value, object2.value) < 0);
 }
};


long function_make_directory(const char * directory);




/////////////////////////////////////////////////////////////
// alternative speech-recognition friendly function names
/////////////////////////////////////////////////////////////


static inline long function_file_size(const char * filename)
{
 return vFilesize(filename);
}


static inline unsigned short function_copy_file(const char * source,
const char * target)
{
 return vCopyFile(source, target);
}


static inline short function_check_ending(const char * string, const
char * ending)
{
 return vStringEndsIn(string, ending);
}


static inline short function_check_case_ending(const char * string,
const char * ending)
{
 return vStringCaseEndsIn(string, ending);
}

static inline char * function_adjust_ending_two(const char * source,
const char * ending)
{
 return vAdjustEnding2(source, ending);
}


static inline void function_adjust_ending_three(const char * source,
char * target, const char * ending)
{
	vAdjustEnding3(source, target, ending);
}


static inline char * function_copy_string(const char * source)
{
 return vCopyString(source);
}


static inline char * function_copy_string_three(const char * source,
long start, long end)
{
 return vCopyString(source, start, end);
}


// Replaces every occurrence of old_char in str with new_char.
static inline long function_replace_character(char * str, char
old_char, char new_char)
{
 return vReplaceCharacter(str, old_char, new_char);
}


static inline char * function_number_to_string(long number)
{
 return vLongToString(number);
}


static inline char * function_merge_strings(const char * first, const
char * second)
{
 return vMergeStrings2(first, second);
}


static inline char * function_merge_strings_three(const char * first,
const char * second,
                      const char * third)
{
 return vMergeStrings3(first, second, third);
}


static inline char * function_merge_strings_four(const char * first,
const char * second,
                      const char * third, const char * fourth)
{
 return vMergeStrings4(first, second, third, fourth);
}


static inline char * function_merge_strings_five(const char * first,
const char * second,
                      const char * third, const char * fourth,
                      const char * fifth)
{
 return vMergeStrings5(first, second, third, fourth, fifth);
}


static inline char * function_get_ending(const char * filename)
{
 return vGetEnding(filename);
}


static inline char * function_remove_ending(const char * filename)
{
 return vRemoveEnding(filename);
}


static inline char * function_file_title(const char * filename)
{
 return vFileTitle(filename);
}


static inline char * function_replace_ending_three(const char * name,
const char * old_ending,
                      const char * new_ending)
{
 return vReplaceEnding3(name, old_ending, new_ending);
}


static inline char * function_replace_ending_two(const char * name,
const char * new_ending)
{
 return vReplaceEnding2(name, new_ending);
}


// function_replace_ending finds the ending of the name(defined as the
// smallest ending substring of name that starts with '.') and
// replaces it with new_ending  (which should start with the
// character '.' if we want to have a dot before the ending in
// the result).
static inline char * function_replace_ending(const char * name, const char * new_ending)
{
 return vReplaceEnding2(name, new_ending);
}


static inline char * function_parent_directory(const char * filename,
const char * dir_separator)
{
 return vParentDir(filename, dir_separator);
}


static inline char * function_parent_directory(const char * filename)
{
 return vParentDir(filename, platform_directory_slash);
}


static inline char * function_simple_name(const char * filename, const char * separator)
{
 return vSimpleName(filename, separator);
}

static inline char * function_simple_name(const char * filename)
{
 return vSimpleName(filename, platform_directory_slash);
}


static inline ushort functions_path_components(const char * filename,
std::vector<voidp> * components)
{
 return vPathComponents(filename, components);
}


static inline unsigned short function_file_exists(const char *filename)
{
 return vFileExists(filename);
}


static inline ushort function_directory_exists(const char * filename)
{
 return vDirectoryExists(filename);
}


static inline ulong function_masked_value(ulong number, ulong mask)
{
 return vMaskedValue(number, mask);
}


static inline uchar function_mask_shift(ulong mask)
{
 return vMaskShift(mask);
}


static inline void function_erase_string_vector(std::vector<voidp> * strings)
{
   vEraseCharpVector(strings);
}


static inline long function_Intel_long(long number)
{
 return vHostToPCLong(number);
}


static inline ushort function_Intel_ushort(ushort number)
{
 return vHostToPCUshort(number);
}


static inline vint4 function_Intel_host_long(long number)
{
 return vPCToHostLong(number);
}


static inline long function_UNIX_host_long(long number)
{
 return vUnixToHostLong(number);
}


static inline ushort function_Intel_host_ushort(ushort number)
{
 return vPCToHostUshort(number);
}


static inline long function_read_integer_eight(FILE * fp, vint8 * a)
{
 return vReadVint8(fp, a);
}


static inline long function_read_integers_eight(FILE * fp, vint8 *
numbers, long items)
{
 return vReadVints8(fp, numbers, items);
}


static inline long function_save_integer_eight(FILE * fp, vint8 a)
{
 return vWriteVint8(fp, a);
}


static inline long function_save_integers_eight(FILE * fp, vint8 *
numbers, long items)
{
 return vWriteVints8(fp, numbers, items);
}


static inline void function_get_integer_eight_bytes(vint8 a, uchar * buffer)
{
  vGetVint8Bytes(a, buffer);
}


static inline long function_read_Intel_ushorts(FILE * fp, ushort *
numbers, long items)
{
 return vReadPCUshorts(fp, numbers, items);
}

static inline long function_read_UNIX_ushorts(FILE * fp, ushort *
numbers, long items)
{
 return vReadUnixUshorts(fp, numbers, items);
}


static inline long function_read_UNIX_long(FILE * fp, long * number)
{
 return vReadUnixLong(fp, number);
}


static inline long function_read_UNIX_longs(FILE * fp, long * numbers,
long items)
{
 return vReadUnixLongs(fp, numbers, items);
}


static inline long function_read_Intel_long(FILE * fp, long * number)
{
 return vReadPCLong(fp, number);
}


static inline long function_read_Intel_longs(FILE * fp, long *
numbers, long items)
{
 return vReadPCLongs(fp, numbers, items);
}


static inline long function_save_Intel_ushorts(FILE * fp, ushort *
numbers, long items)
{
 return vWritePCUshorts(fp, numbers, items);
}


static inline long function_save_Intel_long(FILE * fp, long number)
{
 return vWritePCLong(fp, number);
}


static inline long function_save_Intel_longs(FILE * fp, long *
numbers, long items)
{
 return vWritePCLongs(fp, numbers, items);
}


static inline long function_read_Intel_doubles(FILE * fp, double *
numbers, long items)
{
 return vReadPCDoubles(fp, numbers, items);
}


static inline long function_save_Intel_doubles(FILE * fp, double *
numbers, long items)
{
 return vWritePCDoubles(fp, numbers, items);
}


static inline ushort function_file_patterns_match(const char *
pattern, const char * filename)
{
 return vFilePatternsMatch(pattern, filename);
}


static inline void function_wait_for_enter()
{
  vWaitForEnter();
}


static inline char * function_join_paths(const char * path1, const char * path2)
{
 return vJoinPaths(path1, path2);
}


static inline char * function_join_paths_three(const char * path1,
const char * path2,
                   const char * path3)
{
 return vJoinPaths3(path1, path2, path3);
}


static inline char * function_join_paths_four(const char * path1,
const char * path2,
                         const char * path3, const char * path4)
{
 return vJoinPaths4(path1, path2, path3, path4);
}


static inline char * function_join_paths_five(const char * path1,
const char * path2,
                   const char * path3, const char * path4, const char * path5)
{
 return vJoinPaths5(path1, path2, path3, path4, path5);
}


static inline char * function_join_path_two(const char * path1, const
char * path2)
{
 return vJoinPaths2(path1, path2);
}


static inline long function_type_long(Type type)
{
 return vTypeToLong(type);
}


static inline Type function_long_type(long number)
{
 return vLongToType(number);
}


static inline long function_type_size(Type type)
{
 return vSizeOfType(type);
}


static inline char * function_load_string(FILE * fp)
{
 return vReadString(fp);
}


static inline long function_save_string(FILE * fp, const char * string)
{
 return vWriteString(fp, string);
}


static inline long function_write_Intel_string(FILE * fp, char * string)
{
 return vWritePCString(fp, string);
}


static inline char * function_read_Intel_string(FILE * fp)
{
 return vReadPCString(fp);
}


// Assuming that all the data in the is numbers, the numbers
// are read and stored in the vector "numbers".
static inline long function_read_double_vectorB(const char * filename,
                       std::vector<double> * numbers)
{
 return vReadDoubleVectorB(filename, numbers);
}

static inline long function_read_double_vectorA(FILE * fp,
std::vector<double> * numbers)
{
 return vReadDoubleVectorA(fp, numbers);
}


// True if the file with the given filename can be opened for writing.
// False otherwise (for example, if the directory specified by the
// filename doesn't exist).
static inline unsigned short function_writable_file_exists(char * filename)
{
 return vWritableFileExists(filename);
}


// Length required to write a decimal number with characters. We include
// a character for the minus sign, if the number is negative.
static inline long function_decimal_length(long number)
{
 return vDecimalLength(number);
}


// Returns a random 32-bit integer
static inline ulong function_random_long()
{
 return vRandomLong();
}


// Returns a random number, uniformly distributed between low and high
// (including low and high).
static inline long function_random_long2(long low, long high)
{
 return vRandomLong2(low, high);
}


// Returns a random number, uniformly distributed between low and high
static inline double function_random_double(double low, double high)
{
 return vRandomDouble(low, high);
}


long exit_error(char * format, ...);

// function_warning is an alternative to calling exit_error, when we
// encounter a condition that we can handle, but which is strong
// evidence that something is wrong. Again, by putting a breakpoint
// on this function, we can easily spot the cases where that happens.
long function_warning(char * format, ...);


// Choose "number" numbers between low and high, and store them
// in result. No repetitions are allowed.
static inline long function_pick_without_reputations2(long low, long
high, long number,
                              std::vector<long> * result)
{
 return vPickWithoutRepetitions2(low, high, number, result);
}

static inline long function_compare_strings (const char* string1,
const char* string2)
{
 return strcmp (string1, string2);
}

// This class is used so that I can define objects of type
// map<char *, void *, CharpLess>. STL provides a comparator
// object for strings, but not for char * strings.
class CharpLess
{
public:
 ushort operator()(const char *x, const char *y) const
 {
   return (strcmp(x, y) < 0);
 }
};


// this class is aimed to be a substitute for the FILE structure.
// The reason I needed this substitute is that, as of September 30, 2005
// the Linux version that we have on elf and orc cannot handle file sizes
// that are greater than 2 GB.  With existing files that exceeded that limit
// I had to split them using the UNIX command "split", and here I want to have
// a class structure that can transparently handle both regular files
// and split files with exactly the same code, without the user having to
// specify whether a file is regular or split.
// For the time being, I will only handle the case of reading files,
// I think I do need to worry about writing files so far.
class class_file
{
public:
  // split is a flag that is true if we are dealing with a split file.
  long split;
  long current_index;

  char* specifications;

  // file_pointer will simply be the pointer to the file,
  // when the file is not split.  If the file is split,
  // then file_pointer will point to the current chunk of the file.
  FILE* file_pointer;

  // file_size is only used for split files, and stores the
  // length of each chunk.  Naturally, the last chunk may have
  // a smaller length than that.
  long file_size;

  // this is only used for split files, and stores the
  // virtual pathname, i.e., the pathname of each of the split files
  // without the suffixes ended by the UNIX split command.
  char* pathname;

  // these functions add the right suffix to the pathname,
  // for split files.  They should probably be private
  // or protected.
  char* pathname_from_position (long file_position);
  char* pathname_from_index (long index);
  long index_from_position(long position);
  long open_index (long index);

  class_file (const char* filename, const char* specifications);
  ~class_file ();

  // this should be the substitute for fread.
  long read (void* buffer, size_t type_size, size_t items);

  // this should be the substitute for fwrite.  For the time
  // being, this will only work for non-split files.
  long write (void* buffer, size_t type_size, size_t items);

  // this should be the substitute for fseek and my 64-bit
  // equivalent.
  long seek (vint8 position, int origin);

  // they should be the substitute for ftell
  vint8 tell ();
};


// these are overloads of fread, fwrite, fseek, ftell, fclose
// that work with class_file.

long fread (void* buffer, size_t type_size, size_t items, class_file*
            file_pointer);
long fwrite (void* buffer, size_t type_size, size_t items, 
             class_file* file_pointer);
long fseek (class_file* file_pointer, long position, int origin);
long function_seek(class_file* file_pointer, vint8 position, int origin);
long ftell(class_file* file_pointer);
vint8 function_tell (class_file* file_pointer);
long fclose (class_file* file_pointer);

long string_length (const char* argument);



class class_pixel
{
public:
  long vertical;
  long horizontal;

  class_pixel (long in_vertical, long in_horizontal)
  {
    vertical = in_vertical;
    horizontal = in_horizontal;
  }

  long operator == (class_pixel & argument)
  {
    if ((vertical == argument.vertical) && (horizontal == argument.horizontal))
    {
      return 1;
    }
    else
    {
      return 0;
    }
  }
  long operator != (class_pixel & argument)
  {
    if ((vertical == argument.vertical) && (horizontal == argument.horizontal))
    {
      return 0;
    }
    else
    {
      return 1;
    }
  }
};


typedef FILE file_handle;
file_handle * open_read_file(const char * filename);
file_handle * open_save_file(const char * filename);
long close_file(file_handle * file_pointer);

long function_read_floats(file_handle * file_pointer, float * numbers, long size);
long function_save_floats(file_handle * file_pointer, float * numbers, long size);

long function_read_doubles(file_handle * file_pointer, double * numbers, long size);
long function_save_doubles(file_handle * file_pointer, double * numbers, long size);







#endif    //  VASSILIS_AUXILIARIES_H
