#ifndef VASSILIS_GRAPHICAL_MODULE_H
#define VASSILIS_GRAPHICAL_MODULE_H

#include "basics/v_types.h"
#include "basics/video_file.h"
#include "basics/auxiliaries.h"
#include "basics/image_algorithms.h"
#include "basics/video_common.h"
#include "window.h"
#include "basics/base_interpreter.h"

#include <vector>


class graphical_interface_globals
{
public:
  vWindow * window; 
  vWindow * pixel_info_window;
  vWindow * frame_info_window;
  long min_width; // min_width of the main window
  long extra_height; // height in addition to the image, that the window requires
};


class class_sequence_information
{
private:

public:
  // The actual input video file
  vVideoFile * sequence;
  // Current frame in the video file
  long current_frame;
  // The path name from which we opened sequence
  char * path_name;
  // The simple name (without the path part) for the sequence
  char * simple_name;
  // Number of frames the sequence has
  long number_of_frames; 
  // sequence_step determines how many frames we should move forward at each 
  // step. This is used when we just step forward through the sequence using
  // the graphical interfce, but also with things like motion differencing,
  // where it determines which frame should be used as the "previous" frame
  // and which frame should be used as the "next" frame
  long sequence_step;
  // The window row where the top of the frame should be displayed
  long display_top;
  // The window column where the left of the frame should be displayed
  long display_left;

  // Resolution = 0 means we display the frame as we read it. resolution = k-1
  // means we should display the frame at half the resolution that we would use
  // for resolution = k. So, if our frame is 200x100 and resolution is -2 we
  // display a subsampled version of the frame of size 50x25. We can also have
  // positive resolution, in which case we only display a magnified version of
  // the image. In that case, we can't display the whole image anymore, so we
  // have to choose a center row and center col for the region that we will
  // magnify.
  long resolution;

  long center_row, center_col;
  // If we have a resolution > 0, we may want to display the image in a region
  // larger than the original size of the image. The user can set the size
  // of the region, but, by default, it is magnified_rows x magnified_cols.
  //long magnified_rows, magnified_cols;

  // automatic_centering is 1 if center_row, center_col should be set to
  // the center of each image that we show, 0 if they should stay fixed
  // (which is what we want when we set the center manually).
  long automatic_centering; 

  // When we display an image, we also store it in a most-recently-used buffer
  // (also called ring)
  vArray(general_image *) image_buffer;
  long buffer_index;    // Pointing to the current image in the ring
  general_image * current_image; // image_ring[ring_index];
  long buffer_size;

  // Camera input may also be represented as a class_sequence_information object, but in
  // that case we should treat it differently in a few places. For example,
  // it makes it convenient if we don't allow a call to 
  // vInterpreter1::RemoveInput to remove a camera input.
  long is_camera_input;

  class_sequence_information(long in_buffer_size);
  ~class_sequence_information();
  ushort ClearImageBuffer();
  ushort SetImageBufferSize(long in_buffer_size);
  long CurrentFrame();
  ushort SetCurrentFrame(long index);
};


class graphical_module : public class_module
{
public:
  // the next two lines are mandatory for any module
  typedef long (graphical_module::*type_procedure) ();
  std::vector<type_procedure> procedures;

  // A variable of this type is used so that the ProcessMouseClick function knows
  // what it should do when it gets the next mouse click.
  enum MouseClickResponse {pixel_info_response, MoveSequenceResponse, 
                           LineResponse, CenterResponse, CutoffLabelResponse,
                           AddFeaturesResponse, choose_rectangle_response,
                           spline_response};
  MouseClickResponse previous_mouse_response;
  MouseClickResponse mouse_click_response;

  // The main window, can be 0 (then nothing gets displayed)
  vWindow * window; 
  vWindow * pixel_info_window;
  vWindow * frame_info_window;
  long min_width; // min_width of the main window
  long extra_height; // height in addition to the image, that the window requires

  // The default source of input for text commands (as defined in 
  // commands.cpp) is standard input. However, sometimes we may want
  // to execute batch commands, saved in a file. In those cases,
  // input_fp will represent that file.

  FILE * input_fp;

  vector<class_sequence_information *> inputs;
  long default_buffer_size;

  // Data related to current input video files
  long current_input_index;
  long number_of_inputs;
  class_sequence_information * input;

  // Data related to current output video files
  long current_output;
  long number_of_outputs;
  std::vector<void *> output_files;
  std::vector<void *> output_names;

  // Various parameters that are needed to supply default arguments to
  // some functions
  long filter_blur_size; // Used in convolutions.

  // Miscellaneous info about the state of the program
  char * current_directory;

  long histogram_equalization_flag;
  long equalization_window_size;
  long preserve_current_frame;
  long frame_rate;

public:

  //////////////////////////////////////////////////////////
  //                                                      //
  // start of obligatory member functions for any module  //
  //                                                      //
  //////////////////////////////////////////////////////////

  graphical_module(class_interpreter * interpreter,
                 graphical_interface_globals * globals);
  virtual ~graphical_module();
  
  virtual long initialize();
  virtual long clean_up();
  
  long enter_command(char * command, type_procedure procedure);
  virtual long process_command(const char * command);
  long sort_commands ();

  //////////////////////////////////////////////////////////
  //                                                      //
  //  end of obligatory member functions for any module   //
  //                                                      //
  //////////////////////////////////////////////////////////


  //////////////////////////////////////////////////////////
  //                                                      //
  // start of event handling for the graphical interface  //
  //                                                      //
  //////////////////////////////////////////////////////////

  ushort ProcessKey(char command);
  ushort ProcessMouseClick(long button_down, long row, long col);

  //////////////////////////////////////////////////////////
  //                                                      //
  //  end of event handling for the graphical interface   //
  //                                                      //
  //////////////////////////////////////////////////////////


  //////////////////////////////////////////////////////////
  //                                                      //
  // start of core functions for the graphical interface  //
  //                                                      //
  //////////////////////////////////////////////////////////

  long initialize(graphical_interface_globals * globals);

  // Sets up the image as current image of the current input, and
  // calls DrawCurrentImage to display it.
  long ShowImage(general_image * image);
  long store_image(general_image * image);
  inline long show_image(general_image * image)
  {
    return ShowImage(image);
  }
  long ChangeDirectory(const char * filename);
  long ChangeDirectory();
  
  long AddInput();
  long AddInput(const char * filename);
  long AddInputSilent(const char * filename);
  long open_input(const char * filename)
  {
    return AddInput(filename);
  }
  long AddInput(vVideoFile * new_video_file);
  long open_input(class_sequence * video_file)
  {
    AddInput(video_file);
  }

  long default_input0();
  long default_input(long rows, long cols);
  long RemoveInput(long index);
  long ClearInputs();
  vVideoFile * GetInput(long index);
  long PrintInputs();
  // CurrentImage(i) returns the current image of the i-th
  // input file, or 0 if it doesn't exist.
  general_image * CurrentImage(long index);

  long AddOutput();
  long AddOutput(const char * filename);
  long add_output_stream();
  long AddOutput(vVideoFile * output);
  long RemoveOutput(long index);
  long RemoveCurrentOutput();
  long ClearOutputs();
  vVideoFile * GetOutput(long index);

  long write_frame();
  // Write consecutive frames, by calling alternatingly
  // ShowNextImage and write_frame.
  long write_frames0();
  long write_frames(long number_of_frames);

  vWindow * GetWindow();

  long Recenter();
  long Center0();
  long Center(long center_row, long center_col);
  long CenterWithClick();
  long SetMagnifiedSize0();
  long ResetMagnifiedSize();
  long SetMagnifiedSize(long rows, long cols);

  // Draws the current image of the current input, so that it
  // gets displayed on the window.
  long DrawCurrentImage();
  long SetBounds();
  long ResetBounds();
  long SetBounds(long top, long left, 
        	       long rows, long cols);
  long sequence_rectangle(long top, long left, 
                  	      long rows, long cols)
  {
    SetBounds (top, left, rows, cols);
  }
  // If resolution <= 0, we don't need to normalize coordinates.
  // Otherwise, we need to normalize them based on the magnification
  // factor (2 to the power of resolution), and the center_row and
  // center_col of the input.
  long NormalizedCoordinates(long in_row, long in_col,
                             long * out_row, long * out_col);
  long PrintNormalizedCoordinates0();
  long PrintNormalizedCoordinates(long in_row, long in_col);

  long ShowPixelInfo(long row, long col);
  long PrintPixelInfo();
  long PrintPixelInfo(long row, long col);
  long entry_value();
  long ShowFrameInfo();

  // Subtracts one (mod ring size) from the image buffer index.
  long ShowPreviousImage();
  // Adds one (mod ring size) to the image buffer index
  long ShowNextImage();

  //Subtract from or add a given number to the image buffer index.
  long RewindImageBuffer0();
  long ForwardImageBuffer0();

  long RewindImageBuffer(long frames_to_skip);
  long ForwardImageBuffer(long frames_to_skip);

  long DrawInput(long index);
  long SetCurrentInput();
  long SetCurrentInput(long index);
  long PreviousInput();
  long NextInput();

  // these functions set the input without redrawing. Useful when switching
  // inputs and immediately displaying something new, it prevents
  // the "flashing" effect caused by first drawing whatever the current image
  // was for that input and then drawing the new image.
  long SetCurrentInputSilent(long index);
  long PreviousInputSilent();
  long NextInputSilent();

  long RemoveCurrentInput();
  long remove_inputs();
  long MoveSequence();
  long MoveSequenceWithClick();
  long MoveSequence(long top, long left); 
  long move_sequence(long top, long left)
  {
    return MoveSequence(top, left);
  }

  long NextInSequence();
  long AdvanceSequence(long num_frames);

  // Moves all input sequences to the same frame as that of the current input sequence
  long SynchronizeSequences();

  long ShowCurrentFrame();
  long StoreCurrentFrame();
  long JumpToFrame0();
  long JumpToFrame(long frame);
  long make_current_frame(long frame)
  {
    return JumpToFrame(frame);
  }

  general_image * ReadCurrentImage();
  long SetCurrentFrame(long index);
  long PreviousInSequence();

  long BrowseImage0();
  long browse_image(long number);
  long SetSequenceStep();
  long SetSequenceStep(long step);

  long SetCurrentCounter();
  long SetCurrentCounterPercent(double percentage);

  long DecreaseResolution();
  long IncreaseResolution();
  long SetResolution0();
  long SetResolution(long resolution);

  long histogram_equalization_on();
  long histogram_equalization_off();
  long histogram_equalization_size ();
  long histogram_equalization ();
  long histogram_equalization_offset ();
  long histogram_equalization_weights_test ();

  long toggle_preserve_current_frame();
  long current_percentage();
  long match_window_size();
  long set_frame_rate();
  long get_frame_rate();
  general_image * current_image ();
  class_sequence_information * current_sequence ();

  long set_buffer_size();
  long set_default_buffer_size();
  long check_input();
  long CopyInput0();
  long CopyInput(long input_index);



  //////////////////////////////////////////////////////////
  //                                                      //
  //   end of core functions for the graphical interface  //
  //                                                      //
  //////////////////////////////////////////////////////////



  //////////////////////////////////////////////////////////
  //                                                      //
  //      start of generic image processing functions     //
  //                                                      //
  //////////////////////////////////////////////////////////


  long ConnectedComponents();
  long BiggestComponent();
  long SelectComponent();
  long SelectComponents();
  long RedrawWindow();
  long RedrawImages();

  long FrameDifferences();

  // Makes the current image all black
  long ClearImage();
  long SetPixel();
  long SetPixel(long row, long col, uchar value);
  v3dMatrix<double> * ComputeMotion();
  long ComputeMotion0()
  {
    ComputeMotion ();
    return 1;
  }

  long ThresholdCurrentImage();
  long ThresholdCurrentImage(double threshold);
  long InverseThresholdImage();

  long UndirectedEdges();
  long AmplifyImage();
  long MultiplyImage();
  long SaveCurrentImage();
  long ApplyProcedure(vPixelPred proc);
  long HeightAdjustment();
  long Trash();
  long Trash2();
  long Trash3();

  long SaveImage();
  long SavePGM();

  long Blur();
  long AverageRegions();

  long LoadImage();
  long WritePPM();
  long ProcessRegions();

  long NormalizeImage(double distance, double center_row, double center_col);
  long NormalizeLabeledImage(double distance, double center_row, 
                             double center_col);
  long RotateLeft0();
  long RotateLeft(long left_turns);
  long RotateRight0();
  // Rotates the current image by 90 * right_turns degrees
  long RotateRight(long right_turns);
  
  long NewImage();
  long NewImage(long rows, long cols);
  long DrawLine();
  long DrawLineB();
  long DrawLine(long row1, long col1, long row2, long col2, long value);
  long DrawLine2(long row, long col, double angle, long length, long value);
  long DrawRectangle();
  long DrawRectangle(long row1, long col1, long row2, long col2, long value);
  long DrawCircle();
  long DrawCircle(long row, long col, long radius, long value);
  long DrawFilledRectangle();
  long DrawFilledRectangle(long row1, long col1, long row2, long col2, long value);
  long DrawFilledCircle();
  long DrawFilledCircle(long row, long col, long radius, long value);

  long SelectValue0();
  long SelectValue(long value);

  long MaxNeighbors();
  long Histogram0();
  long Histogram(long number_of_bins);
  long Histogram30();
  long Histogram3(long number_of_bins, double lower, double upper);

  long BoundingBox0();
  long Translate0();
  long Translate(long row_diff, long col_diff);
  long ScaleImage0();
  long ScaleImage(double scaling);
  long ResizeImage0();
  long ResizeImage(long new_rows, long new_cols);
  long ScaleBoundingBox0();
  long ScaleBoundingBox0b();
  long ScaleBoundingBox(long new_box_size);
  long ScaleBoundingBox2(long new_box_size, long new_rows, long new_cols);
  long CenterOfMass();

  long TopBottom();
  long Pad();
  long Pad2();

  long make_short();
  long make_long();
  long make_ushort();
  long make_ulong();
  long make_float();
  long make_double();

  // copied from interpreter2.h
  long ImageMax0();
  long ImageMin0();
  long BandMin0();
  long BandMin(long band);
  long BandMax0();
  long BandMax(long band);
  long CopyBand0();
  long CopyBand(long band);

  long CountValue0();
  long CountValue(double value);
  long CountLequal0();
  long CountLequal(double value);
  long CountGequal0();
  long CountGequal(double value);

  long FillRegion0();
  long FillRegion(long row, long col, uchar value);
  long RandomFrame();
  long to_color();
  long to_color_bound();
  long ToGray();
  long ToGrayBound();

  long ConnectedComponents8();
  long EdgeOrientations();
  long EdgeOrientations2();
  long NonMaximaSuppression();
  long IntensityMaxima();
  long GradientNorms();

  long Overlay0();
  long Overlay(long input1, long input2);
  long Randomize0();
  long Randomize1();

  long DilateImage();
  long SelectRange0();
  long SelectRange(long band, double low, double high);
  long CircularDistance0();
  long CountNonZero();
  long TestLinePoints0();
  long TestLinePoints(long row, long col, double angle, long length);

    // SelectPixels uses the current image as a binary mask, and 
  // then applies this mask to the image in the ring buffer that
  // is at the selected index, so that the product of the mask
  // and the buffer image is displayed.
  long SelectPixels0();
  long SelectPixels(long index);

  // SelectBBox uses the current image as a binary mask, finds the
  // bounding box of the non-zero area, and 
  // then selects the corresponding pixels (that belong to that box)
  // from the image in the ring buffer that
  // is at the selected index.
  long SelectBoundingBox0();
  long SelectBoundingBox(long index);

  long FillUpHoles();
  // Rotate image by an arbitrary angle.
  long RotateImage();
  long Binary();

  // copied from interpreter3.h
  long WaitSeconds0();
  long WaitSeconds(long seconds);
  long DoubleSize();
  long SequenceDoubleSize();
  long ReplaceColor0();
  long ReplaceColor(uchar r1, uchar g1, uchar b1,
                    uchar r2, uchar g2, uchar b2);
  long NormalizeIntensity();
  long WriteRange0();
  long WriteRange(long low, long high);

  // The next functions are used to test and debug vScaleImageDown8
  long ScaleImageA0();
  long ScaleImageA(long bbox_size, long image_size);
  long ScaleImageB0();
  long ScaleImageB(double scaling, long center_row, long center_col,
                   long new_center_row, long new_center_col,
                   long new_rows, long new_cols);
  // Selects a color from the current image (assuming it is a color image).
  long SelectColor0();
  long SelectColor(uchar r, uchar g, uchar b);

  // Counts the pixels having the specified color in the current image 
  // (assuming it is a color image).
  long CountColor0();
  long CountColor(uchar r, uchar g, uchar b);
  long CompareInputs0();
  long CompareInputs(long input1, long input2);
  long compare_images();

  // copied from interpreter4.cpp
  long function_file_copy();
  long image_average();
  long nonzero_average();
  long image_median();
  long image_absolute();
  long image_add();
  long image_subtract();
  long image_multiply();
  long image_divide();
  long image_diff();
  long image_total();

  // this function adds all the frames of the current input sequence
  // to the current output sequence, after it processes each frame
  // by running the commands in the given script file

  long save_filtered_frames();
  long save_filtered_frames(const char * script_name, long start, long finish);
  long process_sequence();
  long random_image();
  long resize();
  long add_ax_match(const char * filename);
  long test_canonical_orientation();

  // counts the number of pixels with higher intensity than the specified pixel
  long pixel_rank ();

  // same as pixel_rank, but in addition, this function also shows
  // a thresholded image using the intensity of the specified pixel as a threshold.
  long pixel_threshold_rank ();

  // here the entry at each band is treated as a separate entry
  long entry_rank ();
  long entry_threshold_rank ();
  
  long threshold_entries ();
  long inverse_threshold_entries ();
  long channel_average();
  long channel_total();
  long select_rectangle();
  long load_image();

  long double_threshold();
  long connected_size_threshold();

  long find_smallest();
  long keep_smallest();
  long find_largest();
  long keep_largest();

  long set_range();
  long maximizing_channels();
  long remove_negatives();
  long print_header();

  // this function is used to test the simple_float_matrix class.
  long load_simple_float();

  

  //////////////////////////////////////////////////////////
  //                                                      //
  //      end of generic image processing functions       //
  //                                                      //
  //////////////////////////////////////////////////////////

  
  //////////////////////////////////////////////////////////
  //                                                      //
  //              start of timing functions               //
  //                                                      //
  //////////////////////////////////////////////////////////

  long time_precomputed_square_root();
  long time_square_root();
  long time_precomputed_cosine();
  long time_cosine();
  long time_precomputed_sine();
  long time_sine();
  long time_precomputed_inverse_tangent();
  long time_inverse_tangent();
  long time_canonical_range();

};







#endif // VASSILIS_GRAPHICAL_MODULE_H
