#ifndef VASSILIS_IMAGE_ALGORITHMS_H
#define VASSILIS_IMAGE_ALGORITHMS_H

#include <deque>
#include <vector>
#include <math.h>


#include "basics/image.h"
#include "basics/algebra.h"

#ifdef VASSILIS_WINDOWS
using std::deque;
using std::sort;
using std::vector;
#endif

#ifdef VASSILIS_SGI_PLATFORM
using namespace std;
#endif


typedef void (*vPixelPred)(color_image * the_image, long row, long col,
			    uchar & r, uchar & g, uchar & b);


// The class vCoordinateNormalizer is used for efficiency in normalizing
// an image, so that its center of mass is at the position we want and the
// maximum distance of a point from the center of mass is what we want. In
// order to do that, we need a helper function that, for each pixel coordinate
// in the normalized image, it gives us the corresponding pixels in the source
// image, as well as the weights by which those pixels should be weighted.
// In the function vNormalizeImage we need to normalize coordinates for each 
// pixel. Calling a function that gets passed in all the arguments could be
// inefficient. By creating an object first, that function can have access to 
// that information faster (I hope).

class vCoordinateNormalizer
{
private:
  // These variables are fixed, given the input and output images.
  double current_center_row, current_center_col;
  double desired_center_row, desired_center_col;
  double row_diff, col_diff;
  double current_max_distance;
  double desired_max_distance;
  double distance_ratio;
  // Constrain the coordinates
  long min_row, max_row;
  long min_col, max_col;

public:
  // These variables will change every time we call the Unnormalize function,
  // they are used to store the results.
  long rows[4];
  long cols[4];
  double weights[4];
  long valid_pixels;
  
  vCoordinateNormalizer();
  void SetCurrentCenter(double in_row, double in_col);
  void SetDesiredCenter(double in_row, double in_col);
  void SetCurrentMaxDistance(double in_distance);
  void SetDesiredMaxDistance(double in_distance);
  void SetRowRange(long in_min, long in_max);
  void SetColRange(long in_min, long in_max);
  void Unnormalize(double row, double col);
};


// vCumulativeCalculations takes in a histogram, that is a matrix of
// number_of_bins rows an four columns. We assume that the second column
// is the number of items thrown into that bin. This function puts into
// each row of the third column (col = 2) 
// the sums of the values in the second column
// from row 0 up to that row. It puts into each row of the fourth column the 
// sums of the values in the second column from the last row down to that row.
//
// Return 0 if no entries are in the histogram (all entries in column 1 are 0).
ushort vCumulativeCalculations(vMatrix<double> * histogram);


// Beginning of wrappers around template functions.
//
// Each of the following functions has a name that ends in a 0
// is a wrapper around a template function defined in algo_templates.cpp
// that has the same name and but doesn't end in a 0, and its arguments
// are v3dMatrix<type> objects instead of general_image objects.
// 

v3dMatrix<double> * vComputeMotion0(general_image * the_image1, 
                                       general_image * the_image2,
                                       general_image * the_image3);

vMatrix<double> *
vBoundedRankHistogram0(general_image * input, long number_of_bins,
                        double lower, double upper);

v3dMatrix<double> *
vImageDifference0(general_image * the_image1, general_image * the_image2);

general_image * 
vMaxNeighbors0(general_image * the_image);

vMatrix<double> *
vPixelBoundedHistogram0(general_image * input, long number_of_bins,
                         double lower, double upper);

vMatrix<double> * 
vPixelHistogram0(general_image * input, long number_of_bins);

general_image * vPyramidImage0(general_image * input, long times);

general_image * 
function_translate_image0(general_image * the_image, long translation_rows, 
                  long translation_cols);

general_image * 
vScaleImage0(general_image * the_image, double scaling);

general_image * 
vScaleImage0(general_image * the_image, double scaling, 
              long center_row, long center_col);

general_image * 
vScaleImage0(general_image * the_image, double scaling, 
              long center_row, long center_col,
              long new_rows, long new_cols);

general_image * 
vResizeImage0(general_image * the_image, long new_rows, long new_cols);

vBoundingBox
vBinaryBoundingBox(general_image * input);

general_image * vMagnify0(general_image * the_image, long center_row, 
                         long center_col, long output_rows, long output_cols,
                         long factor);

color_image * vCircularIntensity0(general_image * input, 
                             double low, double high, double ignore);

general_image * vSelectRange0(general_image * input, long band, 
                             double low, double high);

// Returns a binary image that is non-zero at local maxima 
// in the direction of the gradient of the intensity in the image. 
v3dMatrix<double> * vNonMaximaSuppression(general_image * the_image);

// Returns a binary image that is non-zero at local maxima 
// in the direction of the gradient of the intensity in the image. 
v3dMatrix<double> * vIntensityMaxima(general_image * the_image);

// Returns a binary image that is non-zero at local maxima 
// in the direction of the gradient of the intensity in the image. 
float_matrix function_intensity_maxima(float_matrix image);

// Returns a binary image that is non-zero at local maxima 
// in the direction of the normal to the given orientations. 
// pixels with values less than the threshold are ignored
float_matrix function_intensity_maxima2(float_matrix image, float_matrix orientations,
                                        const float threshold);

// Scale image up by a factor of two, and compute pixel values
// based on bilinear interpolation.
v3dMatrix<double> * vDoubleSize(v3dMatrix<double> * the_image);

// Returns an image that has, at every pixel, the norm of the gradient
// of the intensity change at that pixel at color_image. I'm not sure if it
// does the right thing for images that have more than one bands.
v3dMatrix<double> * vGradientNorm0(general_image * the_image);

// Returns an image that has, at every pixel, the norm of the gradient
// of the intensity change at that pixel at color_image. I'm not sure if it
// does the right thing for images that have more than one bands.
v3dMatrix<double> * vGradientNorm(v3dMatrix<double> * the_image);

// here the calculations are non-centered
v3dMatrix<double> * vGradientNorm2b(v3dMatrix<double> * the_image);


// End of wrappers around template functions

// This function puts the locations of non-zero points into row_vector and
// col_vector. 
ushort vGetNonZeroPoints(grayscale_image * binary, vector<long> * rows, 
                          vector<long> * cols);

// This function puts the locations of non-zero points into *rows_address and
// *cols_address, and the number of those points at *number_address. 
ushort vGetNonZeroPoints4(grayscale_image * binary, vArray(long) * rows_address, 
                           vArray(long) * cols_address, long * number_address);

// This function puts the locations of non-zero points into *rows_address and
// *cols_address, and the number of those points at *number_address. 
ushort vGetNonZeroPoints4(grayscale_image * binary, vArray(double) * rows_address, 
                           vArray(double) * cols_address, long * number_address);


color_image * vLabelsToImage(v3dMatrix<long> * labels_image,
                  			      vector<Label> * label_objects);

void vInverseImage(v3dMatrix<uchar> * the_image);

void vFillUpHoles(v3dMatrix<long> * regions, 
		               v3dMatrix<long> * background);

// This function assumes that foreground is a binary image
// (values are 0 or 255). A hole is a connected component
// of value 0 that is not connected to the top left corner
// (which is also assumed to be background). In the
// result, the holes are erased (their values are 255).
grayscale_image * vFillUpHoles1(v3dMatrix<uchar> * foreground);

long vFillUpHoles2(v3dMatrix<uchar> * foreground, 
                    v3dMatrix<uchar> * result);


Label vFindClosestLabel(vector<Label> & labels, Label & label,
			 long start, long end);

Label vLeftmostLabel(vector<Label> & labels, long start, long end);

Label vRightmostLabel(vector<Label> & labels, long start, long end);

double vLabelDistance(Label & label1, Label & label2);

// This function returns a binary image having the desired rows and cols,
// center of mass right at the center of the image and max distance of 
// a non-zero value from the center of mass equal to max_distance.
grayscale_image * vNormalizedBinary(general_image * the_image,
                                 double max_distance, long rows, long cols);


// vScaleBoundingBox2 is an older version of vScaleBoundingBox.
grayscale_image * vScaleBoundingBox2(grayscale_image * binary, long new_box_size,
                                 long new_center_row, long new_center_col);


// Why do we need a separate function to normalize labeled images? Well,
// first of all, what is a labeled image? It is an image where each pixel
// value is a label. We still use numbers for the labels, as we do for 
// intensity, but there is no sense of distance between labels, two labels
// are equally different regardless of whether the numbers they correspond to
// are close together or far apart. The question is, what value should we
// put to a pixel in the target image? For intensity images a reasonable answer
// was to put a weighted average of some pixels in the source image that it
// corresponded to. For labels averages make no sense. Instead, we want
// to put at each target pixel the label stored in the source image pixel which 
// (warped to the output image) has the most overlap with the target pixel.
//
// The problem with this approach is that a pixel with overlap 0.4 and label A
// outweighs 3 pixels with overlap 0.2 and label B, and that is the wrong thing
// to do.
//
// In all other respects, this function has the same behavior as 
// vNormalizedBinary (in terms of arguments and result).
grayscale_image * vNormalizeLabeledImage(grayscale_image * labels, double max_distance,
									                    long rows, long cols);


void vRowEdgePred(color_image * the_image, long row, long col,
		   uchar & r, uchar & g, uchar & b);

color_image * vProcessPixels(color_image * the_image, vPixelPred predicate);


// Find the 8-connected components in the image.
long_matrix vConnectedComponents8(uchar_image * the_image,
                                 vector<Label> * label_objects);


// For each 8-connected component in labels it finds the two points
// in that component that are the most distant from each other (in
// terms of the shortest path within that component that connects them).
// In the result matrix, row i looks like 
// (start_row, start_col, end_row, end_col), where those four numbers
// specify the position of the start and endpoint of segment i.
// Of course, it is arbitrary which endpoint we call start and
// which we call end.
// rows and cols contain the location of foreground pixels. This
// way we don't have to scan the whole image to find them.
//
// IMPORTANT NOTE: This function assumes that the top, bottom,
// left and right sides of the image are background. This
// allows us to skip boundary checks and speeds up the function.
vMatrix<long> v8ComponentEndpoints(vector<long> * rows,
                                     vector<long> * cols,
                                     v3dMatrix<long> * labels);

// This function is an auxiliary function for v8ComponentEndpoints,
// although it can be used by itself as well. It finds the 
// endpoints of a particular component. It is assumed that 
// rows and cols specify the locations of the pixels that belong
// to that component. coordinates is assumed to be an array of
// length 4, and it will contain at the end of the function 
// start_row, start_col, end_row, end_col, in that order.
// If the size of rows is 0, -1 is put in all positions
// in coordinates.
long vFindEndpoints(vector<long> * rows,
                     vector<long> * cols,
                     v3dMatrix<long> * labels,
                     v3dMatrix<long> * scratch,
                     long * coordinates);


long vMergeComponents(v3dMatrix<long> * components,
                       vMatrix<long> * endpoints,
                       vector<long> * rows, vector<long> * cols);


long vMergeComponents(v3dMatrix<long> * components,
                       long label1, long label2,
                       vector<long> * rows, vector<long> * cols);


color_image * vOverlay0(general_image * the_image1, general_image * the_image2);

color_image * vOverlay(grayscale_image * the_image1, grayscale_image * the_image2);

grayscale_image * vDilateImage(grayscale_image * the_image, uchar foreground_low, 
                            uchar foreground_high);

color_image * vNormalizeIntensity(color_image * the_image);

void vPixelGradient(grayscale_image * the_image, long row, long col, 
                      double & horizontal, double & vertical);

// For every pixel, the direction of least change is computed 
// (as an angle between 0 and PI)
v3dMatrix<double> * vEdgeOrientations(grayscale_image * the_image);

// For every pixel, the gradient direction is computed (as
// an angle between 0 and 2_PI).
v3dMatrix<double> * vEdgeOrientations2(grayscale_image * the_image);

// For every pixel, the gradient direction is computed (as
// an angle between 0 and 2_PI).
v3dMatrix<double> * vEdgeOrientations2(v3dMatrix<double> * the_image);

// returns an image that is a copy of the input image except that
// the color specified by old_red, old_green, old_blue 
// is replaced by the color specified by new_red, new_green, new_blue.
color_image * function_copy_replace_color(color_image * input, uchar old_red, 
                                          uchar old_green, uchar old_blue,
                                          uchar new_red, uchar new_green, uchar new_blue);

// modifies the argument image by replacing
// the color specified by old_red, old_green, old_blue 
// by the color specified by new_red, new_green, new_blue.
long function_replace_color(color_image * image, 
                            uchar old_red, uchar old_green, uchar old_blue,
                            uchar new_red, uchar new_green, uchar new_blue);


// this function selects all pixels with intensity greater than
// small_threshold whose 8-connected component contains at least one pixel
// with intensity greater than big_threshold. An example where such a function
// would be useful is for canny edge detection.
float_matrix double_threshold(float_matrix image, float small_threshold, 
                              float big_threshold);

float_matrix function_label_maxima(float_matrix image, long_matrix labels,
                                   std::vector<class_label> * labels_vector);


// selects pixels that belong to a connected component with more pixels than
// the specified threshold.
float_image connected_size_threshold(float_image & image, long threshold);













#endif // VASSILIS_IMAGE_ALGORITHMS_H
