#ifndef VASSILIS_LIBVASSILIS_H
#define VASSILIS_LIBVASSILIS_H

#include <vector>
#include "v_types.h"

//#pragma warning(disable:4786) 

// This class is used so that I can define objects of type
// map<char *, void *, CharpLess>. STL provides a comparator
// object for strings, but not for char * strings.
class CharpLess
{
public:
  ushort operator()(const char *x, const char *y) const;
};



// This class can be used to handle options that the user passes to a program
// from the command line (or from a file). To create an object of this class,
// call the default constructor, and then call ProcessOptions to read
// the options that the user passed from the command line. The option
// -file filename makes the object read key/value pairs from the file called
// filename.
class vProgramOptions
{
private:
  // All the maps that I use will be of the same type, to avoid code-bloating.
  typedef std::map<char *, void *, CharpLess> * table;
  // For every option we have a value. For integers and doubles, the value is
  // a pointer that contains the actual value that we are interested in. This
  // way all values can be represented as void * pointers.
  table values;

  // Also, for every option we have a string that describes that option.
  // For example, such a string can be used for the error message.
  table descriptions;

  // For each option, the type of its value is long, double, char * or
  // vector<void *> *. We need to keep track of the type of each option,
  // so that we know how to deallocate the memory that is used to
  // store the value.
  table types;

  // upper_bounds and lower_bounds contain the highest and lowest permissible
  // values for options whose type is "blong" (which stands for "bounded long")
  table upper_bounds;
  table lower_bounds;

  // Sets member variables to default values.
  void Initialize();

  // Deletes all memory allocated for this object and its member variables
  void CleanUp();

  // The argument can come from the command line, or from a file. This
  // function looks at each entry of the vector, treats it as a string,
  // and records the values for the options that are specified
  ushort ProcessOptionsVector(std::vector<voidp> * options);
  ushort ProcessOptionsFile(const char * filename);

public:
  vProgramOptions();
  ~vProgramOptions();

  // Prints an error message that describes the allowed options.
  void Error();
  void Description(char * option);
  ushort ProcessOptions(int argc, char ** argv);

  // Prints all options, their names, types, values and descriptions
  void Print();

  // Prints all information about a particular option, i.e. its name,
  // value, type and description
  void Print(char * option_name);

  // The "Get" functions set success to 0 if the key is not a valid
  // option, or if the type of that option doesn't match the type
  // that the function assumes for that option.
  const char * GetString(char * key, ushort & success);
  long GetLong(char * key, ushort & success);
  double GetDouble(char * key, ushort & success);
  std::vector<voidp> * GetVector(char * key, ushort & success);

  // The "Set" functions, like the "Get" functions, return 0
  // if the key is not a valid option or if the type of the
  // key is not what the function assumes.
  ushort SetString(char * key, const char * value);
  ushort SetLong(char * key, long value);
  ushort SetBlong(char * key, long value);
  ushort SetDouble(char * key, double value);
  ushort SetVector(char * key, std::vector<voidp> * v);

  ushort EnterString(const char * key, const char * value, 
                     const char * description);
  ushort EnterLong(const char * key, long value, const char * description);
  // blong is like long, but it is only allowed to range between a lowest and
  // a highest value.
  ushort EnterBlong(const char * key, long value, long lowest, long highest,
                    const char * description);

  ushort EnterDouble(const char * key, double value, 
                     const char * description);
  ushort EnterVector(const char * key, const char * description);
};


void * vValue(std::map<char *, void *, CharpLess> * table, char * key);

ushort vSetValue(std::map<char *, void *, CharpLess> * table, 
                  char * key, void * value);


#endif // VASSILIS_LIBVASSILIS_H
