#ifndef VASSILIS__MATMATH_H
#define VASSILIS__MATMATH_H

#include "vplatform.h"

//   matmath.h
//   This is the declaration file for the matmath functions.

//  This is the datatype used for the math and non-type specific ops.  */


// ************     _Vector math routines    ****************

//  _Vector B = _Vector A, of size num_elements   */
template<class m_elem>
void vec_copy( m_elem *src, m_elem *dst, int num_elements );

// _Vector C = _Vector A + _Vector B , both of size n   */
template<class m_elem>
void vec_add( m_elem *A, m_elem *B, m_elem *C, int n );

//  _Vector C = _Vector A - _Vector B , both of size n   */
template<class m_elem>
void vec_sub( m_elem *A, m_elem *B, m_elem *C, int n );

template<class m_elem>
void print_Vector( char *str, m_elem *x, int n );

// ************   Some matrix math routines  **************/

//  matrix B = matrix A, of size num_rows x num_cols   */
template<class m_elem>
void mat_copy( m_elem **A, m_elem **B, int num_rows,
			int num_cols );

// Copy A[top1..bottom1][left1..right1] into B, starting at B[top2..bottom2].
template<class m_elem>
void mat_copy( m_elem **A, m_elem **B, 
	       int top1, int bottom1, int left1, int right1,
	       int top2, int bottom2);

// Copy the rows x cols matrix A into B, starting at B[top2..bottom2].
template<class m_elem>
void mat_copy( m_elem **A, m_elem **B, 
	       int rows, int cols, int top2, int left2);

//  matrix C = matrix A + matrix B , both of size m x n   */
template<class m_elem>
void mat_add( m_elem **A, m_elem **B, m_elem **C, int m, int n );

//  matrix C = matrix A - matrix B , all of size m x n  */
template<class m_elem>
void mat_sub( m_elem **A, m_elem **B, m_elem **C, int m, int n );

//  matrix C = matrix A x matrix B , A(a_rows x a_cols), B(a_cols x b_cols) */
template<class m_elem>
void mat_mult( m_elem **A, m_elem **B, m_elem **C,
		     int a_rows, int a_cols, int b_cols );

//  matrix C = matrix A x _Vector B , A(a_rows x a_cols), B(a_cols x 1) */
template<class m_elem>
void mat_mult_Vector( m_elem **A, m_elem *B, m_elem *C,
			    int a_rows, int a_cols );

//  C = matrix A x trans( matrix B ), A(a_rows x a_cols), B(b_cols x a_cols) */
template<class m_elem>
void mat_mult_transpose( m_elem **A, m_elem **B, m_elem **C,
			   int a_rows, int a_cols, int b_cols );

//  C = trans( matrix A ) x matrix B, A(a_cols x a_rows), B(a_cols x b_cols) */
template<class m_elem>
void mat_transpose_mult( m_elem **A, m_elem **B, m_elem **C,
			   int a_rows, int a_cols, int b_cols );


template<class m_elem>
void print_matrix( char *str, m_elem **A, int m, int n );


template<class m_elem>
void gaussj( m_elem **A, int n, m_elem **B, int m );

template<class m_elem>
void mrqmin( m_elem x[], m_elem y[], m_elem sig[], int ndata,
		   m_elem a[], int ia[], int ma, m_elem **covar,
		   m_elem **alpha, m_elem *chisq,
		   void (*funcs)(m_elem, m_elem [], m_elem *, m_elem [], int),
		   m_elem *alamda);

//  quaternion routines  */

#define QUATERNION_SIZE    4
#define MIN_QUATERNION_MAGNITUDE   0.000001


template<class m_elem>
void quaternion(m_elem ** result);

template<class m_elem>
void quaternion_update( m_elem *quat, m_elem wx,
		       m_elem wy, m_elem wz );
template<class m_elem>
void quaternion_to_rotation( m_elem *quat, m_elem **rot );
template<class m_elem>
void print_quaternion( char *str, m_elem *quat );

//   _Vector allocation routines   */

template<class m_elem>
void _Vector(m_elem ** result, long nl, long nh);

template<class m_elem>
void _Vector(m_elem ** result, long items);

int *i_Vector(long nl, long nh);
unsigned char *c_Vector(long nl, long nh);
unsigned long *l_Vector(long nl, long nh);
float *f_Vector(long nl, long nh);
double *d_Vector(long nl, long nh);

//   Matrix allocation routines    */

template<class m_elem>
void matrix(m_elem *** result, long nrl, long nrh, long ncl, long nch);

template<class m_elem>
void matrix(m_elem *** result, long rows, long cols);

float **fmatrix(long nrl, long nrh, long ncl, long nch);
double **dmatrix(long nrl, long nrh, long ncl, long nch);
int **imatrix(long nrl, long nrh, long ncl, long nch);


template<class m_elem>
m_elem **submatrix(m_elem **a,
		   long oldrl, long oldrh, long oldcl, long oldch,
		   long newrl, long newcl);

template<class m_elem>
m_elem **convert_matrix( m_elem *a, long nrl, long nrh,
			      long ncl, long nch);

//   Deallocation routines   */

template<class m_elem>
void free_Vector(m_elem *v, long nl, long nh);

template<class m_elem>
void free_Vector(m_elem *v, long items);

void free_f_Vector(float *v, long nl, long nh);
void free_i_Vector(int *v, long nl, long nh);
void free_c_Vector(unsigned char *v, long nl, long nh);
void free_l_Vector(unsigned long *v, long nl, long nh);
void free_d_Vector(double *v, long nl, long nh);

template<class m_elem>
void free_matrix(m_elem **m, long nrl, long nrh, long ncl, long nch);

template<class m_elem>
void free_matrix(m_elem **m, long rows, long cols);

void free_fmatrix(float **m, long nrl, long nrh, long ncl, long nch);
void free_dmatrix(double **m, long nrl, long nrh, long ncl, long nch);
void free_imatrix(int **m, long nrl, long nrh, long ncl, long nch);
void free_submatrix(float **b, long nrl, long nrh, long ncl, long nch);
void free_convert_matrix(float **b, long nrl, long nrh, 
				long ncl, long nch);
void nrerror(char error_text[]);   /*  internal error handler */


#ifndef VASSILIS_SGI_PLATFORM
#define VASSILIS_PROCESS_FILE
#include "matmath.cpp"
#endif // VASSILIS_SGI_PLATFORM

#endif // VASSILIS_MATMATH_H

