#ifndef VASSILIS_MATRIX_H
#define VASSILIS_MATRIX_H

#include <vector>

#include "auxiliaries.h"



// v3dMatrix is meant to be a generic 3D_array class, that can be used
// among other things to represent images, with arbitrary numbers of bands, 
// and arbitrary types for the representation of the color of each
// pixel in each  band.

// Note that whether char is the same as signed char or same as
// unsigned char is not specified in C++.
// enum Type {ScharType, ShortType, IntType, LongType, FloatType, DoubleType,
//	   UcharType, UshortType, UintType, UlongType, OtherType};
// In the previous line I keep the original order. If I add more types later, I
// should make sure that the Type values when converted to long give the same
// numbers as before. My code that reads and writes images to and from files
// depends on the assumption that the long number equivalent to a given type
// never changes.

template<class type>
class v3dMatrix;

class color_image;
class grayscale_image;

class general_image : public vUnique
{
protected:
  Type type_name;
  long rows, cols, bands, size, all_band_size;

public:
  virtual ~general_image();
  inline long Bands() const
  {
    return bands;
  }

  inline long channels () const
  {
    return bands;
  }

  inline Type TypeName() const
  {
    return type_name;
  }

  inline long Rows() const
  {
    return rows;
  }

  inline long vertical () const
  {
    return rows;
  }

  inline long Cols() const
  {
    return cols;
  }

  inline long Columns() const
  {
    return cols;
  }

  inline long horizontal () const
  {
    return cols;
  }

  inline long Size() const // Size of one band, in pixels;
  {
    return size;
  }

  inline long length() const // Size of one band, in pixels;
  {
    return size;
  }

  inline long AllBandSize() const
  {
    return all_band_size;
  }

  inline long all_channel_size() const
  {
    return all_band_size;
  }

  static inline long HeaderSize()
  {
    return 4;
  }

  // True if row and col are coordinates of a valid pixel
  // in the image, false if they are out of bounds.
  inline long check_bounds(long row, long col) const
  {
    if (row < 0) return 0;
    if (col < 0) return 0;
    if (row >= rows) return 0;
    if (col >= cols) return 0;
    return 1;
  }

  inline long Index2(long row, long col) const
  {
    return (row * cols + col);
  }

  inline long Index3(long band, long row, long col) const
  {
    return (band * size + Index2(row, col));
  }

  // print some general information about the image: type name,
  // dimensions.
  long print_header() const;

  virtual long SizeOfType() const = 0;
  
  // creates a general image from raw data.  The raw data itself is copied,
  // so it can be deleted after this function has been called. We assume that the data
  // format is interleaved, row-by-row, without any padding.
  static general_image * from_data(void * data, long channels, long vertical, long horizontal,
                                   long bits_per_pixel);
  
  virtual ushort Write(const char * filename) const = 0;
  ushort store(const char * filename) const {  return Write(filename);  }
  virtual ushort Write(FILE * fp) const = 0;
  ushort store(FILE * fp) const {  return Write(fp);  }
  virtual ushort Write(class_file * fp) const = 0;
  ushort store(class_file * fp) const {  return Write(fp);  }
  virtual ushort WriteText(const char * filename) const = 0;
  ushort store_text(const char * filename) const {  return WriteText(filename);  }

  static general_image * Read(const char * filename);
  static general_image * read(const char * filename) {  return Read(filename);  }
  static general_image * Read(FILE * fp);
  static general_image * read(FILE * fp) {  return Read(fp);  }
  static general_image * Read(class_file * fp);
  static general_image * read(class_file * fp) {  return Read(fp);  }
  static general_image * ReadText(const char * filename);
  static general_image * read_text(const char * filename) {  return ReadText(filename);  }

  virtual general_image * SumBands() const = 0;
  general_image * sum_bands() const {  return SumBands();  }
  virtual v3dMatrix<long> * SumBandsLong() const = 0;
  v3dMatrix<long> * sum_bands_long() const {  return SumBandsLong();  }
  virtual v3dMatrix<double> * SumBandsDouble() const = 0;
  v3dMatrix<double> * sum_bands_double() {  SumBandsDouble();  }
  virtual general_image * AverageBands() const = 0;
  general_image * average_bands() {  return AverageBands();  }

  virtual general_image * BoundBands(long lower, long upper) const = 0;
  general_image * bound_bands(long lower, long upper) {  return BoundBands(lower, upper);  }
  virtual general_image * BoundBands(float lower, float upper) const = 0;
  general_image * bound_bands(float lower, float upper) {  return BoundBands(lower, upper);  }
  virtual general_image * BoundBands(double lower, double upper) const = 0;
  general_image * bound_bands(double lower, double upper) {  return BoundBands(lower, upper);  }

  // ToRGB and ToGray should be used when we know that 
  // the values in the image that we 
  // want to convert are already between 0 and 255
  // The functions ToGrayCopy and ToRGBCopy return a new image.
  // The functions ToRGB and ToGray return the object that 
  // they are called from, if it is indeed RGB or gray.
  virtual color_image * ToRGB() const = 0;
  virtual grayscale_image * ToGray() const = 0;
  virtual color_image * ToRGBCopy() const = 0;
  virtual color_image * color_copy() const = 0;
  virtual grayscale_image * ToGrayCopy() const = 0;
  virtual grayscale_image * grayscale_copy() const = 0;

  // ToRGBBound and ToGrayBound shift the values of the 
  // source image so that they are between
  // 0 and 255 before it converts to RGB
  virtual color_image * ToRGBBound() const = 0;
  virtual grayscale_image * ToGrayBound() const = 0;
  virtual grayscale_image * ToGrayBoundCopy() const = 0;
  virtual color_image * ToRGBBoundCopy() const = 0;
  virtual grayscale_image * grayscale_bound_copy() const = 0;
  virtual color_image * color_bound_copy() const = 0;

  // Copy returns a copy of this image
  virtual general_image * Copy() const = 0;
  virtual general_image * Copy(long left, long right, long top, long bottom) const = 0;
  inline general_image * copy() const
  {
    return Copy();
  }

  inline general_image * copy(long left, long right, long top, long bottom) const
  {
    return Copy (left, right, top, bottom);
  }

  virtual void DoubleValues(std::vector<double> & values, long row, long col) const = 0;
  virtual void DoubleValues(std::vector<double> & values, long i) const = 0;
  virtual void LongValues(std::vector<long> & values, long row, long col) const = 0;
  virtual void LongValues(std::vector<long> & values, long i) const = 0;
  virtual double double_value (long channel, long vertical, long horizontal) const = 0;
  virtual long long_value (long channel, long vertical, long horizontal) const = 0;
  virtual double double_value_safe (long channel, long vertical, long horizontal) const = 0;
  virtual long long_value_safe (long channel, long vertical, long horizontal) const = 0;
  virtual double double_value_strict (long channel, long vertical, long horizontal) const = 0;
  virtual long long_value_strict (long channel, long vertical, long horizontal) const = 0;

  // The following functions perform various image operations.
  virtual v3dMatrix<uchar> * threshold(double threshold) const = 0;
  virtual general_image * AverageRegions(long win_rows, long win_cols) const = 0;
  virtual general_image * Resample(long rows2, long cols2) const = 0;
  virtual general_image * Resample(long top, long left, long rows1, long cols1,
				 long rows2, long cols2) const = 0;

  // Resample2 takes as input the subimage defined by (top, left) as its top
  // left corner that has full_rows1 rows and full_cols1 columns. It creates
  // a result image in which the input subimage is subsampled (1 out of
  // every "scale" rows and columns survive). If scale doesn't divide 
  // exactly rows (or cols) the remainder of rows (or cols) of the input
  // image is ignored.
  virtual general_image * Resample2(long top, long left, 
                          long full_rows1, long full_cols1, long scale) const = 0;

  virtual v3dMatrix<long> * ConnectedComponents(std::vector<Label> * label_objects) const = 0;
};


template<class type>
class v3dMatrix : public general_image
{
protected:
  vArray3(type) matrix3; // Indexed by band, row, col.
  vArray2(type) matrix2; // Indexed by band, index 
                          // (index depends on row, col).
  vArray(type) matrix;   // Index depends on band, row, col.
  
  // The next two variables hold the maximum and minimum allowable
  // values for the type. This is useful in template functions where
  // we need to know these values (for example when we amplify an
  // image, and we want to use the whole range of allowable values).
  type max_possible_value;
  type min_possible_value;

  virtual void DeleteUnique();

  vArray2(type) ReorderBands(std::vector<long> * band_order) const;

  // The next function stores in band_order the sequence 0, 1, ..., bands-1
  void DefaultOrder(std::vector<long> & band_order) const;
  void ProcessType();
  void Initialize(long in_rows, long in_cols, long in_bands);
 
public:
  v3dMatrix();
  v3dMatrix(const long in_rows, const long in_cols, const long bands);
  v3dMatrix(const general_image * source);
  v3dMatrix(const general_image * source, long channel);
  ~v3dMatrix();

  static v3dMatrix * from_data(void * data, long channels, long vertical, long horizontal);

  // returns a three-dimensional matrix that provides access to all the data
  // in the image and is indexed by channel, row, column
  vArray3(type) Matrix3() const;

  // returns a two-dimensional matrix that provides access to the data 
  // stored in the i-th band (channel) and is 
  // indexed by row, column
  vArray2(type) Matrix2(long i) const;

  // returns a two-dimensional matrix that provides access to the data of all bands (channels),
  // and is indexed by channel and index (where index is a single number
  // that depends on the row and column.
  vArray2(type) Matrix2() const;

  // returns a one-dimensional matrix that stores the data of channel i,
  // and is indexed by a single number
  vArray(type) Matrix(long i) const;

  // returns a one-dimensional matrix indexed by a single number.
  vArray(type) Matrix() const;

  // these inline commands are simply renamed versions
  // (with names that are easy to dictate using speech recognition)
  // of the previous commands that provide access to the matrix data
  inline three_pointer(type) three_matrix() const
  {
    return Matrix3();
  }

  inline matrix_pointer(type) planar(const long channel) const
  {
    return Matrix2(channel);
  }

  inline matrix_pointer(type) planar_safe(const long channel) const
  {
    if ((channel >= 0) && (channel < bands) &&
        (predicate_zero(matrix3) == 0))
    {
      return Matrix2(channel);
    }
    else
    {
      return function_zero(class_pointer(type));
    }
  }

  inline matrix_pointer(type) planar() const
  {
    return Matrix2();
  }

  inline class_pointer(type) flat(const long channel) const
  {
    return Matrix(channel);
  }

  inline class_pointer(type) flat() const
  {
    return Matrix();
  }

  
  short ReadInterlaced(const char * filename, long offset) const;
  short ReadInterlaced(const char * filename, long offset,
		                   std::vector<long> * band_order) const;

  short ReadInterlaced(FILE * fp) const;
  short ReadInterlaced(class_file * fp) const;
  short ReadInterlaced(FILE * fp, std::vector<long> * band_order) const;

  short ReadInterlaced(class_file * fp, std::vector<long> * band_order) const;

  short ReadBandwise(const char * filename, long offset) const;
  short read_bandwise(const char * filename, long offset) const 
  {  return ReadBandwise(filename, offset);  }
  short ReadBandwise(const char * filename, long offset, 
	                  std::vector<long> * band_order) const;
  short read_bandwise(const char * filename, long offset, 
	                    std::vector<long> * band_order) const
  {  return ReadBandwise(filename, offset, band_order);  }
  short ReadBandwise(FILE * fp) const;
  short read_bandwise(FILE * fp) const {  return ReadBandwise(fp);  }
  short ReadBandwise(class_file * fp) const;
  short read_bandwise(class_file * fp) const {  return ReadBandwise(fp);  }
  short ReadBandwise(FILE * fp, std::vector<long> * band_order) const;
  short ReadBandwise(class_file * fp, std::vector<long> * band_order) const;

  long ReadTextBandwise(const char * filename, long offset) const;
  long ReadTextBandwise(FILE * fp) const;
  long ReadTextBandwise(class_file * fp) const;

  static long ReadHeader(FILE * fp, std::vector<long> * header);
  static long ReadHeader(class_file * fp, std::vector<long> * header);
  static long ReadHeaderText(FILE * fp, std::vector<long> * header);
  static long ReadHeaderText(class_file * fp, std::vector<long> * header);

  ushort WriteInterlaced(FILE * fp) const;
  ushort WriteInterlaced(class_file * fp) const;
  ushort WriteInterlaced(FILE * fp, std::vector<long> * band_order) const;
  ushort WriteInterlaced(class_file * fp, std::vector<long> * band_order) const;
  ushort WriteBandwise(FILE * fp) const;
  ushort store_bandwise(FILE * fp) const {  return WriteBandwise(fp);  }
  ushort WriteBandwise(class_file * fp) const;
  ushort store_bandwise(class_file * fp) const {  return WriteBandwise(fp);  }
  ushort WriteBandwise(FILE * fp, std::vector<long> * band_order) const;
  ushort WriteBandwise(class_file * fp, std::vector<long> * band_order) const;
  long WriteHeader(FILE * fp) const;
  long store_header(FILE * fp) const {  return WriteHeader(fp);  }
  long WriteHeader(class_file * fp) const;
  long store_header(class_file * fp) const {  return WriteHeader(fp);  }
  long WriteHeaderText(FILE * fp) const;
  long WriteHeaderText(class_file * fp) const;
  
  static long WriteHeader(FILE * fp, std::vector<long> * header);
  static long WriteHeader(class_file * fp, std::vector<long> * header);
  static long WriteHeaderText(FILE * fp, std::vector<long> * header);
  static long WriteHeaderText(class_file * fp, std::vector<long> * header);

  type MinValue() const;
  type MaxValue() const;

  virtual long SizeOfType() const;
  general_image * SumBands() const;
  v3dMatrix<long> * SumBandsLong() const;
  v3dMatrix<double> * SumBandsDouble() const;
  general_image * AverageBands() const;

  general_image * BoundBands(long lower, long upper) const;
  general_image * BoundBands(float lower, float upper) const;
  general_image * BoundBands(double lower, double upper) const;
  v3dMatrix * set_range(type lower, type upper);
  long set_range(type lower, type upper, v3dMatrix * target);

  color_image * ToRGB() const;
  grayscale_image * ToGray() const;
  color_image * ToRGBCopy() const;
  grayscale_image * ToGrayCopy() const;
  
  // Returns a copy of this image.
  general_image * Copy() const;
  general_image * Copy(long left, long right, long top, long bottom) const;
  color_image * ToRGBBound() const;
  grayscale_image * ToGrayBound() const;
  color_image * ToRGBBoundCopy() const;
  grayscale_image * ToGrayBoundCopy() const;
  grayscale_image * grayscale_copy() const;
  color_image * color_copy() const;
  grayscale_image * grayscale_bound_copy() const;
  grayscale_image * fast_grayscale_bound_copy() const;
  color_image * color_bound_copy() const;

  void DoubleValues(std::vector<double> & values, long row, long col) const;
  void DoubleValues(std::vector<double> & values, long i) const;
  void LongValues(std::vector<long> & values, long row, long col) const;
  void LongValues(std::vector<long> & values, long i) const;
  virtual double double_value (long channel, long vertical, long horizontal) const;
  virtual long long_value (long channel, long vertical, long horizontal) const;
  virtual double double_value_safe (long channel, long vertical, long horizontal) const;
  virtual long long_value_safe (long channel, long vertical, long horizontal) const;
  virtual double double_value_strict (long channel, long vertical, long horizontal) const;
  virtual long long_value_strict (long channel, long vertical, long horizontal) const;

  // The following functions perform various image operations.
  v3dMatrix<uchar> * threshold(double threshold) const;
  general_image * AverageRegions(long win_rows, long win_cols) const;
  general_image * Resample(long top, long left, long rows1, long cols1,
			 long rows2, long cols2) const;
  general_image * Resample(long rows2, long cols2) const;

  // Resample2 takes as input the subimage defined by (top, left) as its top
  // left corner that has full_rows1 rows and full_cols1 columns. It creates
  // a result image in which the input subimage is subsampled (1 out of
  // every "scale" rows and columns survive). If scale doesn't divide 
  // exactly rows (or cols) the remainder of rows (or cols) of the input
  // image is ignored.
  virtual general_image * Resample2(long top, long left, 
                          long full_rows1, long full_cols1, long scale) const;

  v3dMatrix<long> * ConnectedComponents(std::vector<Label> * label_objects) const;
  
  static v3dMatrix * Read(const char * filename);
  static v3dMatrix * read(const char * filename)
  {
    return Read(filename);
  }

  static v3dMatrix * Read(FILE * fp);
  static v3dMatrix * read(FILE * fp)
  {
    return Read(fp);
  }

  static v3dMatrix * Read(class_file * fp);
  static v3dMatrix * read(class_file * file_pointer)
  {
    return Read(file_pointer);
  }

  static v3dMatrix * ReadText(const char * filename);
  static v3dMatrix * read_text(const char * filename)
  {
    return ReadText(filename);
  }

  static v3dMatrix * ReadText(FILE * fp);
  static v3dMatrix * read_text(FILE * fp)
  {
    return ReadText(fp);
  }

  static v3dMatrix * ReadText(class_file * fp);
  static v3dMatrix * read_text(class_file file_pointer)
  {
    return ReadText(file_pointer);
  }

  ushort Write(const char * filename) const;
  ushort Write(FILE * fp) const;
  ushort Write(class_file * fp) const;
  ushort WriteText(const char * filename) const;
  ushort WriteText(FILE * fp) const;
  ushort WriteText(class_file * fp) const;
  inline ushort store_text(FILE * fp) const
  {
    return WriteText(fp);
  }
  inline ushort store_text(class_file * fp) const
  {
    return WriteText(fp);
  }

  // WriteDebug is useful for saving matrices while debugging, at
  // a pre-specified directory. The first application of this
  // function was in debugging the shape context code: I saved
  // (while debugging) matrices to a file, and then I read them
  // from matlab and compared them to the matlab equivalents.
  long WriteDebug(const char * filename) const;

  // Prints interlaced.
  long Print(const char * text) const;

  // Prints bandwise.
  long Print2(const char * text) const;

  // Referencing items
  inline type & operator () (long band, long row, long col) const
  {
    assert((band >= 0) && (band < bands) &&
           (row >= 0) && (row < rows) &&
           (col >= 0) && (col < cols));
    return matrix3[band][row][col];         
  }

  inline type & operator () (long band, long row, long col)
  {
    assert((band >= 0) && (band < bands) &&
           (row >= 0) && (row < rows) &&
           (col >= 0) && (col < cols));
    return matrix3[band][row][col];         
  }

  inline type & operator () (long band, long index) const
  {
    assert((band >= 0) && (band < bands) &&
           (index >= 0) && (index < size));
    return matrix2[band][index];         
  }
  
  inline type & operator () (long band, long index)
  {
    assert((band >= 0) && (band < bands) &&
           (index >= 0) && (index < size));
    return matrix2[band][index];         
  }
  
  inline type & operator () (long index) const
  {
    assert((index < this->all_band_size) && (index >= 0));
    return this->matrix[index];
  }
  inline type & operator () (long index)
  {
    assert((index < this->all_band_size) && (index >= 0));
    return this->matrix[index];
  }
};
  

ushort predicate_color_image(general_image * base);

ushort predicate_gray_image(general_image * base);


// This class implements a four-dimensional array, with some basic
// functionality: We can create it, delete it, write it to file,
// read it from file. For each dimension, the index range doesn't
// have to start at 0, it can start below zero or above zero.
//
// Note: The function vDebug_4D_Array contains a small interface
// for verifying that the class implementation works correctly.
template<class type>
class v4dArray
{
private:
  // hidden_data4 is an array where every dimension's
  // index range starts at zero.
  vArray4(type) hidden_data4;
  vArray3(type) hidden_data3;
  vArray2(type) hidden_data2;
  vArray(type) hidden_data;

  // data4 is the array from which the data will be read.
  // It points to data in hidden_data, but it adjusts
  // the indices to make up for the index ranges of 
  // all dimensions of the array.
  vArray4(type) data4;
//  vArray3(type) data3;
//  vArray2(type) data2;
//  vArray1(type) data1;

  long dim1, dim2, dim3, dim4;
  long dim1_low, dim1_high, dim2_low, dim2_high;
  long dim3_low, dim3_high, dim4_low, dim4_high;
  long size;

  long ZeroArrays();
  long Initialize();
  long DeleteArrays();

public:
  v4dArray(long in_dim1, long in_dim2, long in_dim3, long in_dim4);
  v4dArray(long in_dim1_low, long in_dim1_high,
             long in_dim2_low, long in_dim2_high,
             long in_dim3_low, long in_dim3_high,
             long in_dim4_low, long in_dim4_high);

  ~v4dArray();

  inline long Dim1()
  {
    return dim1;
  }

  inline long Dim2()
  {
    return dim2;
  }

  inline long Dim3()
  {
    return dim3;
  }

  inline long Dim4()
  {
    return dim4;
  }

  inline long Dim1Low()
  {
    return dim1_low;
  }

  inline long Dim2Low()
  {
    return dim2_low;
  }

  inline long Dim3Low()
  {
    return dim3_low;
  }

  inline long Dim4Low()
  {
    return dim4_low;
  }

  inline long Dim1High()
  {
    return dim1_high;
  }

  inline long Dim2High()
  {
    return dim2_high;
  }

  inline long Dim3High()
  {
    return dim3_high;
  }

  inline long Dim4High()
  {
    return dim4_high;
  }

  inline vArray4(type) Data4()
  {
    return data4;
  }

  inline vArray(type) AllData()
  {
    return hidden_data;
  }

  inline long Size()
  {
    return dim1*dim2*dim3*dim4;
  }

  inline long MemorySize()
  {
    return Size() * sizeof(type);
  }

  long Write(const char * filename);

  long Write(FILE * fp);

  static v4dArray * Read(const char * filename);

  static v4dArray * Read(FILE * fp);

  long SkipHeader(FILE * fp);
  
  long ReadData(FILE * fp);

  // Print is used for debugging
  long Print();

  // Write 0 on all array positions.
  long Zero();

  // Write value on all array positions.
  long WriteValue(type value);
};


Type vSuperType(Type type);


// speech-recognition-friendly names for various classes


typedef v3dMatrix<uchar> standard_image;
typedef v3dMatrix<uchar> uchar_image;
typedef v3dMatrix<ushort> ushort_image;
typedef v3dMatrix<short> short_image;
typedef v3dMatrix<long> long_image;
typedef v3dMatrix<vint4> integer_image;
typedef v3dMatrix<ulong> ulong_image;
typedef v3dMatrix<float> float_image;
typedef v3dMatrix<double> double_image;





#ifndef VASSILIS_SGI_PLATFORM
#define VASSILIS_PROCESS_FILE
#include "matrix.cpp"
#endif // VASSILIS_SGI_PLATFORM


#endif   // VASSILIS_MATRIX_H
