#include "stdafx.h"

#include "vplatform.h"
#include "my_video_format.h"
#include "drawing_temp.h"
#include "wrapper.h"
#include "drawing.h"

#include "definitions.h"


class_image_stream::class_image_stream(const char * in_filename)
{
  filename = vCopyString(in_filename);
  type = vCopyString("class_image_stream");
  fp = fopen(filename, vFOPEN_READ);
  if (fp == 0) return;

  // Read header: number of frames, size of type, type number, rows, cols, bands
  long items;
  items = vReadPCLong(fp, &number_of_frames);
  if (items != 1) return;
  items = vReadPCLong(fp, &size_of_type);
  if (items != 1) return;
  long type_number;
  items = vReadPCLong(fp, &type_number);
  if (items != 1) return;
  image_type = vLongToType(type_number);
  long real_size_of_type = vSizeOfType(image_type);
  // In future platforms maybe long will be 64 bits and other types will
  // also have different lengths than what they have now. This assert
  // will catch such cases, since the code at this point is not able
  // to handle them otherwise.
  assert(real_size_of_type == size_of_type);

  items = vReadPCLong(fp, &rows);
  if (items != 1) return;
  items = vReadPCLong(fp, &cols);
  if (items != 1) return;
  items = vReadPCLong(fp, &bands);
  if (items != 1) return;
  initial_offset = ftell(fp);

  frame_size = rows * cols * bands * size_of_type + 4 * sizeof(long);
  is_valid = 1;
  is_readable = 1;
  current_frame = -1;
}


class_image_stream::class_image_stream(const char * in_filename, long in_rows, long in_cols,
                             long in_bands, ::Type in_type)
{
  filename = vCopyString(in_filename);
  type = vCopyString("class_image_stream");
  rows = in_rows;
  cols = in_cols;
  bands = in_bands;
  image_type = in_type;
  size_of_type = vSizeOfType(image_type);
  frame_size = size_of_type * bands * rows * cols + 4 * sizeof(long);
  number_of_frames = 0;
  fp = fopen(filename, vFOPEN_WRITE);
  if (fp == 0) return;

  long items;
  // Note that number_of_frames is junk at this point, we will write
  // the actual number of frames when we have written the rest of the file.
  items = vWritePCLong(fp, number_of_frames);
  if (items != 1) return;
  items = vWritePCLong(fp, size_of_type);
  if (items != 1) return;
  long type_number = vTypeToLong(image_type);
  items = vWritePCLong(fp, type_number);
  if (items != 1) return;

  items = vWritePCLong(fp, rows);
  if (items != 1) return;
  items = vWritePCLong(fp, cols);
  if (items != 1) return;
  items = vWritePCLong(fp, bands);
  if (items != 1) return;

  // We don't need to iniitalize size_of_frame and initial_offset.
  current_frame = -1;
  is_valid = 1;
  is_writable = 1;
}


class_image_stream::~class_image_stream()
{
  vdelete2(type);
  if (fp != 0) fclose(fp);
}


grayscale_image * class_image_stream::ReadNextFrameGray()
{
  if (!is_readable) return 0;
  general_image * image = ReadNextImage();
  if (image == 0) return 0;

  grayscale_image * gray = image->ToGrayBound();
  if (gray != image)
  {
    vdelete(image);
  }   
 
  return gray;
}


grayscale_image * class_image_stream::ReadFrameGray(long number)
{
  ushort success = GotoFrame(number);
  if (success == 0) return 0;
  return ReadNextFrameGray();
}


color_image * class_image_stream::ReadNextFrame()
{
  if (!is_readable) return 0;
  general_image * image = ReadNextImage();
  if (image == 0) return 0;

  color_image * result = image->ToRGBBound();
  if (result != image)
  {
    vdelete(image);
  }   
 
  return result;
}


color_image * class_image_stream::ReadFrame(long number)
{
  ushort success = GotoFrame(number);
  if (success == 0) return 0;
  return ReadNextFrame();
}


ushort class_image_stream::InitializeWriting()
{
  return 1;
}

// This function refuses to write the image, unless this sequence was 
// designed to save color or gray scale uchar images.
ushort class_image_stream::write_frame(color_image * image)
{
  if (!is_writable) return 0;
  if (image_type != image->TypeName()) return 0;
  if ((bands != 3) && (bands != 1)) return 0;

  long result;
  if (bands == 3)
  {
    result = image->Write(fp);
  }
  else 
  {
    assert(bands == 1);
    grayscale_image * gray = image->ToGray();
    result = gray->Write(fp);
    vdelete(gray);
  }
  if (result != 0) number_of_frames++;
  return (ushort) result;
}


ushort class_image_stream::write_frameGray(grayscale_image * gray)
{
  if (!is_writable) return 0;
  if (image_type != gray->TypeName()) return 0;
  if ((bands != 3) && (bands != 1)) return 0;

  long result;
  if (bands == 1)
  {
    result = gray->Write(fp);
  }
  else 
  {
    assert(bands == 1);
    grayscale_image * image= gray->ToGray();
    result = image->Write(fp);
    vdelete(image);
  }
  if (result != 0) number_of_frames++;
  return (ushort) result;
}


ushort class_image_stream::FinishWriting()
{
  long success = fseek(fp, 0, SEEK_SET);
  if (success != 0) return 0;
  long items;
  items = vWritePCLong(fp, number_of_frames);
  if (items != 1) return 0;
  fclose(fp);
  fp = 0;
  return 1;
}


ushort class_image_stream::GotoFrame(long frame)
{
  if (frame == current_frame + 1) return 1;
  if (frame < 0) return 0;
  if (frame >= number_of_frames) return 0;
  vint8 vint8frame = frame;
  vint8 vint8offset = initial_offset;
  vint8 vint8size = frame_size;
  vint8 position = vint8offset + vint8frame * vint8size;
  long junk = ftell(fp);
  ushort result = function_seek(fp, position, SEEK_SET);
  junk = ftell(fp);
  if (result == 0) return 0;
  else 
  {
    current_frame = frame - 1;
    return 1;
  }
}


ushort class_image_stream::save_image(general_image * image)
{
  if (predicate_color_image(image)) return write_frame((color_image *) image);
  if (predicate_gray_image(image)) return write_frameGray((grayscale_image *) image);

  if (bands != image->Bands()) return 0;
  if (image_type != image->TypeName()) return 0;
  ushort result = image->Write(fp);
  if (result != 0) number_of_frames++;
  return result;
}


general_image * class_image_stream::ReadImage(long number)
{
  ushort success = GotoFrame(number);
  if (success == 0) return 0;
  general_image * the_image = ReadNextImage();
  return the_image;
}


general_image * class_image_stream::ReadNextImage()
{
  general_image * result = general_image::Read(fp);
  if (result != 0) current_frame++;

  if (top != -1)
  {
    long right = left + width - 1;
    long bottom = top + height - 1;
    general_image * result2 =result->Copy(left, right, top, bottom);
    delete result;
    result = result2;
  }

  return result;
}


Type class_image_stream::ImageType()
{
  return image_type;
}


/////////////////////////////////////////////////////////
//
// End of implementation of class class_image_stream
//
/////////////////////////////////////////////////////////



/////////////////////////////////////////////////////////
//
// Start of implementation of class vPPM_Sequence
//
/////////////////////////////////////////////////////////


vPPM_Sequence::vPPM_Sequence(const char * in_filename)
{
  filename = vCopyString(in_filename);
  type = vCopyString("vPPM_Sequence");
  the_image = color_image::load(in_filename);
  if (the_image == 0) 
  {
    is_valid = 0;
    return;
  }
  is_valid = 1;
  is_readable = 1;
  fp = 0;
  rows = the_image->Rows();
  cols = the_image->Cols();
  number_of_frames = 1;
  SetBounds(0, 0, rows, cols);
  current_frame = -1;
}


vPPM_Sequence::vPPM_Sequence(const char * in_filename, long in_rows, long in_cols)
{
  the_image = 0;
  rows = in_rows;
  cols = in_cols;
  current_frame = 0;
  filename = vCopyString(in_filename);
  type = vCopyString("vPPM_Sequence");
  fp = fopen(filename, vFOPEN_WRITE);
  if (fp == 0)
  {
    is_valid = 0;
    return;
  }
  is_valid = 1;
  is_writable = 1;
  current_frame = -1;
  fclose(fp);
}


vPPM_Sequence::~vPPM_Sequence()
{
  vdelete(the_image);
  vdelete2(type);
}


// Functions any vVideoFilesubclass must implement
grayscale_image * vPPM_Sequence::ReadNextFrameGray()
{
  if (the_image == 0) return 0;
  if (current_frame != -1) return 0;
  current_frame = 0;
  return the_image->ToGray();
}


grayscale_image * vPPM_Sequence::ReadFrameGray(long number)
{
  if (number != 0) return 0;
  if (the_image == 0) return 0;
  else return the_image->ToGray();
}


color_image * vPPM_Sequence::ReadFrame(long number)
{
  if (!is_readable) return 0;
  if (number != 0) return 0;
  current_frame = 0;
  return (color_image *) vCopyImage(the_image);
}


color_image * vPPM_Sequence::ReadNextFrame()
{
  return the_image;
}


ushort vPPM_Sequence::InitializeWriting()
{
  return 1;
}


ushort vPPM_Sequence::write_frame(color_image * image)
{
  if (current_frame != -1) return 0;
  return image->save(filename);
}


ushort vPPM_Sequence::write_frameGray(grayscale_image * gray)
{
  if (current_frame != -1) return 0;
  return gray->save(filename);
}


ushort vPPM_Sequence::FinishWriting()
{
  return 1;
}


// End of functions that any vVideoFile subclass must implement

//////////////////////////////////////////////////////////////
//
// End of implementation of class vPPM_Sequence
//
//////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////
//
// Beginning of implementation of class bitmap_sequence
//
//////////////////////////////////////////////////////////////


bitmap_sequence::bitmap_sequence(const char * in_filename)
{
  // Initialize some member variables
  is_valid = 0;
  filename = vAdjustEnding2(in_filename, ".bmpseq");
  type = vCopyString("bitmap_sequence");
  common_part = CommonPart();

  // Read header (number of frames, rows, cols, bytes_per_pixel).
  char * header_file = HeaderFile();
  ushort error_flag = 0;
  fp = fopen(header_file, vFOPEN_READ);
  vdelete2(header_file);
  if (fp == 0) return;
  long items;
  items = vReadPCLongs(fp, &number_of_frames, 1);
  if (items != 1) error_flag = 1;
  items = vReadPCLongs(fp, &rows, 1);
  if (items != 1) error_flag = 1;
  items = vReadPCLongs(fp, &cols, 1);
  if (items != 1) error_flag = 1;
  items = vReadPCLongs(fp, &bytes_per_pixel, 1);
  if (items != 1) error_flag = 1;
  fclose(fp);

  // Initialize more member variables, mark the object as valid
  // if everything is fine.
  if (error_flag != 0) return;
  if ((bytes_per_pixel != 1) && (bytes_per_pixel != 3)) return;
  rate = 30;
  scale = 1;
  current_frame = -1;
  is_readable = 1;
  is_writable = 0;
  is_valid = 1;
}


bitmap_sequence::bitmap_sequence(const char * in_filename, long in_rows, long in_cols,
                               long in_bytes_per_pixel)
{
  is_valid = 0;
  filename = vAdjustEnding2(in_filename, ".bmpseq");
  type = vCopyString("bitmap_sequence");
  common_part = CommonPart();
    
  rows = in_rows;
  cols = in_cols;
  bytes_per_pixel = in_bytes_per_pixel;

  if ((bytes_per_pixel != 1) && (bytes_per_pixel != 3)) return;
  
  // Also, return if a file already exists with the name, because
  // then we won't be able to create the directory
  if (vFileExists(filename)) return;

  // If the directory doesn't exist, we create it.
  long length = strlen(filename);
  char * command = new char[length + 10];
  sprintf(command, "mkdir %s", filename);
  system(command);
  vdelete2(command);

  rate = 30;
  scale = 1;
  current_frame = -1;
  is_readable = 0;
  is_writable = 1;
  is_valid = 1;
}


bitmap_sequence::~bitmap_sequence()
{
  vdelete2(type);
  vdelete2(common_part);
}


// The common part is the part that will be used for the filename
// of every frame, up to the number. For example, a sequence called
// "test.bmpseq" would have "test" as the common part, and frames
// would be saved as test0.bmp, test1.bmp and so on. For sequences
// whose title ends in a number (like "small_2.bmpseq") we add
// an underscore to the end of the common part, so that it is
// "small_2_", abd frames will be saved as small_2_0.bmp, and so on.
char * bitmap_sequence::CommonPart()
{
  char * temp = vFileTitle(filename);
  // Check if the last character of temp is a digit.
  long length = strlen(temp);
  char last = temp[length - 1];
  if ((last >= '0') && (last <= '9'))
  {
    // last character of temp is a digit, we add an underscore.
    char * result = vMergeStrings2(temp, "_");
    vdelete2(temp);
    return result;
  }
  else
  {
    return temp;
  }
}


char * bitmap_sequence::HeaderFile()
{
  char * result = vJoinPaths(filename, "header.txt");
  return result;
}


char * bitmap_sequence::FrameFile(long frame)
{
  char * frame_string = vLongToString(frame);
  char * simple_name = vMergeStrings3(common_part, frame_string, ".bmp");
  char * result = vJoinPaths(filename, simple_name);
  vdelete2(frame_string);
  vdelete(simple_name);
  return result;
}


// Functions any vVideoFilesubclass must implement
grayscale_image * bitmap_sequence::ReadNextFrameGray()
{
  return ReadFrameGray(current_frame + 1);
}


grayscale_image * bitmap_sequence::ReadFrameGray(long number)
{
  if (number >= number_of_frames) return 0;
  char * frame_file = FrameFile(number);
  grayscale_image * gray = 0;
  if (bytes_per_pixel == 1)
  {
    gray = vReadGrayBMP(frame_file);
  }

  if (gray == 0)
  {
    color_image * the_image = vReadBMP(frame_file);
    if (the_image == 0)
    {
      vdelete2(frame_file);
      return 0;
    }
    gray = the_image->ToGray();
  }
  vdelete2(frame_file);
  current_frame = number;
  return gray;
} 


color_image * bitmap_sequence::ReadFrame(long number)
{
  if (number >= number_of_frames) return 0;
  char * frame_file = FrameFile(number);
  color_image * the_image = vReadBMP(frame_file);
  vdelete2(frame_file);
  if (the_image != 0) current_frame = number;
  return the_image;
}


color_image * bitmap_sequence::ReadNextFrame()
{
  return ReadFrame(current_frame + 1);
}


ushort bitmap_sequence::InitializeWriting()
{
  if (is_writable == 0) return 0;
  return 1;
}


ushort bitmap_sequence::write_frame(color_image * image)
{
  if (is_writable == 0) return 0;
  current_frame++;
  char * frame_file = FrameFile(current_frame);
  short success;
  if (bytes_per_pixel == 1)
  {
    grayscale_image * gray = image->ToGray();
    success = vWriteGrayBMP(gray, frame_file);
    vdelete(gray);
  }
  else
  {
    assert(bytes_per_pixel == 3);
    success = vWriteColorBMP(image, frame_file);
  }
  vdelete2(frame_file);
  ushort result = (success >= 1);
  return result;
}


ushort bitmap_sequence::write_frameGray(grayscale_image * gray)
{
  if (is_writable == 0) return 0;
  current_frame++;
  char * frame_file = FrameFile(current_frame);
  short success;
  if (bytes_per_pixel == 3)
  {
    color_image * the_image = gray->ToRGB();
    success = vWriteColorBMP(the_image, frame_file);
    vdelete(the_image);
  }
  else
  {
    assert(bytes_per_pixel == 1);
    success = vWriteGrayBMP(gray, frame_file);
  }
  vdelete2(frame_file);
  ushort result = (success >= 1);
  return result;
}


ushort bitmap_sequence::FinishWriting()
{
  if (is_writable == 0) return 0;
  number_of_frames = current_frame + 1;
  char * header_file = HeaderFile();
  fp = fopen(header_file, vFOPEN_WRITE);
  long items;
  items = vWritePCLongs(fp, &number_of_frames, 1);
  if (items != 1) return 0;
  items = vWritePCLongs(fp, &rows, 1);
  if (items != 1) return 0;
  items = vWritePCLongs(fp, &cols, 1);
  if (items != 1) return 0;
  items = vWritePCLongs(fp, &bytes_per_pixel, 1);
  if (items != 1) return 0;
  fclose(fp);
  vdelete2(header_file);

  // We can't write to the sequence anymore, because
  // that would change the number of frames.
  is_writable = 0;
  return 1;
}


// End of functions that any vVideoFile subclass must implement


///////////////////////////////////////////////////////////////
//
// End of implementation of class bitmap_sequence
//
///////////////////////////////////////////////////////////////
