#ifndef VASSILIS_MY_VIDEO_FORMAT_H
#define VASSILIS_MY_VIDEO_FORMAT_H

#include "video_file.h"
#include "v_types.h"
#include "auxiliaries.h"


// I created class_image_stream for the following reasons:
// - To have a video sequence format that I know everything
// about, and that I can use both to read and to write in any
// platform
// - To have a video sequence format that will allow me to create
// files of any size at all (AVI files can't be larger than 2GB).
// - To have a video sequence format that can handle images
// of any type allowed by the template v3dMatrix class (although
// a single file can't contain any combination of types imaginable).
// 
// class_image_stream basically lets me read and write uncompressed 
// frames.
class class_image_stream : public vVideoFile
{
private:
  // One sequence can have images of different types, as long as each of those
  // types has the same size in bytes, and the images have the same
  // number of bands, rows and cols.
  long size_of_type;
  ::Type image_type;
  long bands;
  long initial_offset;
  long frame_size;

public:
  // Creates a sequence that can be read.
  class_image_stream(const char * filename);

  // Creates a sequence to which we can write.
  class_image_stream(const char * in_filename, long in_rows, long in_cols,
    long in_bands, ::Type in_type);

  ~class_image_stream();

  // Functions any vVideoFilesubclass must implement
  virtual grayscale_image * ReadNextFrameGray();
  virtual grayscale_image * ReadFrameGray(long number);
  virtual color_image * ReadFrame(long number);
  virtual color_image * ReadNextFrame();
  virtual ushort InitializeWriting();
  virtual ushort write_frame(color_image * image);
  virtual ushort write_frameGray(grayscale_image * gray);
  virtual ushort FinishWriting();
  virtual long BytesPerPixel()
  {
    // This function is not really applicable here, since we can have
    // different combinations, but 3 is the least commiting answer.
    return 3;
  }
  // End of functions that any vVideoFile subclass must implement

  // Functions overriding original vVideoFile definitions
  virtual general_image * ReadNextImage();
  virtual general_image * ReadImage(long number);
  inline general_image * read_image(long number)
  {
    return ReadImage(number);
  }

  // These return an image of the same type and number of bands as the
  // image that was originally written.
  ushort GotoFrame(long number);
  ushort save_image(general_image * image);
  ::Type ImageType();
};

// This is just a wrapper class, that simply lets us read and write 
// single frame PPM files. It exists because sometimes it is convenient
// in my code to treat PPM files as one-frame video sequences, so that
// I don't have to treat them any different than, say, AVI files.
class vPPM_Sequence : public vVideoFile
{
private:
  color_image * the_image;

public:
  vPPM_Sequence(const char * filename);
  vPPM_Sequence(const char * filename, long rows, long cols);
  ~vPPM_Sequence();

  // Functions any vVideoFilesubclass must implement
  grayscale_image * ReadNextFrameGray();
  grayscale_image * ReadFrameGray(long number);
  color_image * ReadFrame(long number);
  color_image * ReadNextFrame();
  ushort InitializeWriting();
  ushort write_frame(color_image * image);
  ushort write_frameGray(grayscale_image * gray);
  ushort FinishWriting();
  virtual long BytesPerPixel()
  {
    return 3;
  }
  // End of functions that any vVideoFile subclass must implement
};


// Note that bitmap_sequence is not the BMP equivalent of vPPM_Sequence.
// vPPM_Sequence uses the ppm filename as its filename and is basically
// a one-frame sequence. bitmap_sequence uses as its filename a directory
// name, and reads sequentially numbered BMP files in that directory. So,
// it is a class that lets me treat a set of sequentially numbered files
// as a video sequence. I created this class because I needed a sequence 
// format that I can use for writing, that can be arbitrarily large in 
// file size, and that Cleaner 5 can read and compress. Basically, if I
// have a sequence of BMP files test0.bmp, test1.bmp, test2.bmp, and so on,
// I put them in a directory test.bmpseq, and I create a header.txt file
// in that directory that has the number of frames, number of rows, 
// number of columns and bytes per pixel (1 for colormapped (grayscale
// is a special case of that), 3 for color BMPs). Only uncompressed BMP
// files can be read and written.
//
// See the comments on the CommonPart function for the assumptions on
// how the frames are named.
class bitmap_sequence : public vVideoFile
{
private:
  // Contains the common part of the names of the BMP files (without
  // the path part, which is in the "filename" member variable.
  char * common_part;

  // 1 for colormapped 8-bit images, 3 for uncompressed 24-bit RGB
  // images
  long bytes_per_pixel;

public:
  bitmap_sequence(const char * filename);
  bitmap_sequence(const char * filename, long rows, long cols,
                 long bytes_per_pixel);
  ~bitmap_sequence();

  char * CommonPart();
  char * HeaderFile();
  char * FrameFile(long frame);
  inline long BytesPerPixel()
  {
    return bytes_per_pixel;
  }

  // Functions any vVideoFilesubclass must implement
  grayscale_image * ReadNextFrameGray();
  grayscale_image * ReadFrameGray(long number);
  color_image * ReadFrame(long number);
  color_image * ReadNextFrame();
  ushort InitializeWriting();
  ushort write_frame(color_image * image);
  ushort write_frameGray(grayscale_image * gray);
  ushort FinishWriting();
  // End of functions that any vVideoFile subclass must implement
};



#endif // VASSILIS_MY_VIDEO_FORMAT_H

