#include "stdafx.h"

#include "new_code.h"
#include "basics/algebra.h"
#include "basics/image_algorithms.h"
#include "basics/image_templates.h"

#include "basics/definitions.h"


// returns a matrix with two columns. Each row represents a point,
// column 0 is the row, column 1 is the col of that point. length
// is the length of the longer side of the rectangle, and with
// is the length of the shorter side. Orientation is the direction of 
// the longer side, in radians.
float_matrix rectangle_pixels(float center_vertical, float center_horizontal, 
                              long length, long width, float orientation)
{
  vector<vPoint2> pixels;
  float sine = sin(orientation);
  float cosine = cos(orientation);
  long half_length = (length - 1) / 2;
  long half_width = (width -1) / 2;

  float start_vertical = center_vertical - ((float) half_length) * sine;
  float start_horizontal = center_horizontal - (length / 2.0f) * cosine;

  long length_counter, width_counter;
  for (length_counter = 0; length_counter < length; length_counter++)
  {
    float base_vertical = start_vertical + ((float) length_counter) * sine;
    float base_horizontal = start_horizontal + ((float) length_counter) * cosine;

    for (width_counter = -half_width; width_counter <= half_width; width_counter++)
    {
      float vertical = base_vertical + ((float) width_counter) * cosine;
      float horizontal = base_horizontal + ((float) width_counter) * (-sine);
      vPoint2 pixel(vertical, horizontal);
      pixels.push_back(pixel);
    }
  }

  long number = pixels.size();
  float_matrix result(number, 2);
  long counter;
  for (counter = 0; counter < number; counter++)
  {
    result(counter, 0) = (float) (pixels[counter].row);
    result(counter, 1) = (float) (pixels[counter].col);
  }

  return result;
}


// returns a matrix with two columns. Each row represents a point,
// column 0 is the row, column 1 is the col of that point. length
// is the length of the longer side of the rectangle, and with
// is the length of the shorter side. Orientation is the direction of 
// the longer side, in radians.
long_matrix rectangle_pixels(long center_vertical, long center_horizontal, 
                             long length, long width, float orientation)
{
  // all we have to do is call the floating-point equivalent of this function
  // and convert the results to integers
  float_matrix float_pixels = rectangle_pixels((float) center_vertical, (float) center_horizontal,
                                               length, width, orientation);
  long number = float_pixels.vertical();
  long_matrix result(number, 2);
  long counter;
  for (counter = 0; counter < number; counter++)
  {
    result(counter, 0) = round_number(float_pixels(counter, 0));
    result(counter, 1) = round_number(float_pixels(counter, 1));
  }

  return result;
}


// returns a matrix with three columns. Each row represents a point,
// column 0 is the row, column 1 is the col of that point, column 2
// is the weight of that point. length
// is the length of the longer side of the rectangle, and with
// is the length of the shorter side. Orientation is the direction of 
// the longer side, in radians. The weights sum up to one.
float_matrix rectangle_pixels_factors(long center_vertical, long center_horizontal, 
                                     long length, long width, float orientation)
{
  // this implementation will be inefficient: we just get the pixels for the
  // horizontal orientation and then we rotate the image and go through 
  // the pixels to get the weights. Finally, we normalize the weights.

  const long image_center = length + width;
  const long image_size = 2 * image_center;

  long_matrix horizontal_pixels = rectangle_pixels(image_center, image_center,
                                                   length, width, 0.0f);

  float_matrix horizontal_image(image_size, image_size);
  function_enter_value(& horizontal_image, (float) 0);
  function_enter_values(& horizontal_image, horizontal_pixels, 1.0f);

  float_image * rotated = rotate_image_exact(&horizontal_image, (ulong) image_center, 
                                             (ulong) image_center, (double) orientation);

  long vertical_displacement = center_vertical - image_center;
  long horizontal_displacement = center_horizontal - image_center;
  long number = count_non_zero(rotated);

  float_matrix result(number, 3);
  float total = 0.0f;

  long vertical, horizontal;
  long index = 0;
  for (vertical = 0; vertical < image_size; vertical++)
  {
    for (horizontal = 0; horizontal < image_size; horizontal++)
    {
      float current = (*rotated)(0, vertical, horizontal);
      if (current == 0.0f)
      {
        continue;
      }
      result(index, 0) = (float) (vertical + vertical_displacement);
      result(index, 1) = (float) (horizontal + horizontal_displacement);
      result(index, 2) = current;
      total = total + current;
      index++;
    }
  }

  long counter;
  if (total != 0.0f)
  {
    for (counter = 0; counter < number; counter++)
    {
      result(counter, 2) = result(counter, 2) / total;
    }
  }

  function_delete(rotated);
  return result;
}


precomputed_rectangle_pixels::precomputed_rectangle_pixels(long argument_length, long argument_width, 
                                                           long argument_number)
{
  length = argument_length;
  width = argument_width;
  number_of_orientations = argument_number;
    
  long counter;
  // we do counter <= number_of_orientations to be safe numerically.
  for (counter = 0; counter <= number_of_orientations; counter++)
  {
    float radians = ((float) counter) / ((float) number_of_orientations) * M_PIf;
    float_matrix pixels = rectangle_pixels_factors(0, 0, length, width, radians);
    rectangles.push_back(pixels);
  }
}


precomputed_rectangle_pixels::~precomputed_rectangle_pixels()
{
}


long precomputed_rectangle_pixels::orientation_index(float radians)
{
  float canonical = canonical_radians(radians);
  long index = round_number(canonical / M_PIf * (float) number_of_orientations);
  return index;
}


const float_matrix & precomputed_rectangle_pixels::rectangle(float radians)
{
  long index = orientation_index(radians);
  return rectangles[index];
}


long function_portion(long counter, long steps, long total)
{
  if (steps == 0)
  {
    exit_error("\nerror: steps = 0 in function_portion\n");
  }

  float float_counter = (float) counter;
  float float_total = (float) total;
  float float_steps =(float) steps;
  float step = float_total/float_steps;

  float float_previous_total = float_counter * step;
  float float_current_total = (float_counter +1.0f) * step;
  long previous_total = round_number(float_previous_total);
  long current_total = round_number(float_current_total);
  long result = current_total - previous_total;

  return result;
}

































