// NOTE: The functions in this file can read AVI files produced in our ASL
// lab downstairs. I don't have any documentation about the AVI format, and
// here I just make the following assumptions (that empirically seem to be
// valid in the files that get produced by the computers downstairs):
//
// - The files are uncompressed, 1 or 3 bytes per pixel.
// - The number of bytes per frame is contained in bytes 60-63.
// - The number of frames is contained at bytes 48-51.
// - The number of columns is at bytes 176-179
// - The number of rows is at bytes 180-183
// - Right before each frame data, we get the bytes 48, 48, 100, 98, followed
// by 4 bytes containing the number of bytes in the frame (lower order byte is
// first).
// - Those preceding 8 bytes, for the first frame, come at a position that
// is either 2048 or 4096.


#include "stdafx.h"
#include "vplatform.h"

#ifdef VASSILIS_SGI_PLATFORM
#include <iostream>
#else 
#include <iostream>
#endif

#include "auxiliaries.h"
#include "image.h"
#include "vplatform.h"
#include "read_avi.h"
#include "definitions.h"

const long vReadAVI::frame_header = 1650733104;
const long vReadAVI::LengthLocation = 48;
const long vReadAVI::ScaleLocation = 128;
const long vReadAVI::RateLocation = 132;


long vReadAVI::ReadLong()
{
  long result;
  uchar temp_buff[4];
  fread(temp_buff, 1, 4, fp);
  result = temp_buff[0] + temp_buff[1] * 256;
  result = result + temp_buff[2] * 65536 + temp_buff[3] * 65536 * 256;
  current_position += 4;
  return result;
}

vReadAVI::vReadAVI(const char * in_filename)
{
  filename = vCopyString(in_filename);
  type = vCopyString("vReadAVI");
  fp = fopen(filename, vFOPEN_READ);
  if (fp == 0) return;
  
  // Read number of frames, bytes per frame, rows and columns.
  fseek(fp, LengthLocation, SEEK_SET);
  number_of_frames = ReadLong();
  fseek(fp, 60, SEEK_SET);
  bytes_per_frame = ReadLong();
  fseek(fp, ScaleLocation, SEEK_SET);
  scale = ReadLong();
  rate = ReadLong();
  fseek(fp, 176, SEEK_SET);
  cols = ReadLong();
  rows = ReadLong();
  bytes_per_pixel = bytes_per_frame / (rows * cols);
  fseek(fp, 2048, SEEK_SET);
  long temp_long = ReadLong();
  if (temp_long == frame_header)
  {
    initial_offset = 2048;
  }
  else
  {
    fseek(fp, 4096, SEEK_SET);
    temp_long = ReadLong();
    assert(temp_long == frame_header);
    initial_offset = 4096;
  }
  current_frame = 0;
//  cout << "initial_offset = " << initial_offset << endl;
  
  // Find distance between beginnings of two consecutive frames, by
  // looking for beginning of second frame (first occurence of the frame 
  // header after the end of the data of the first frame).
  fseek(fp, bytes_per_frame + 4, SEEK_CUR);
  long extra_bytes = 0;
  while(1)
  {
    temp_long = ReadLong();
    if (temp_long == frame_header) break;
    fseek(fp, -3, SEEK_CUR);
    extra_bytes++;
  }
  frame_displacement = bytes_per_frame + extra_bytes + 8;
  fseek(fp, initial_offset, SEEK_SET);
  current_position = initial_offset;
//  cout << "frame_displacement = " << frame_displacement << endl;

  // In the default case, we want to read the entire frame.
  SetBounds(0, 0, rows, cols);
  is_valid = 1;
  is_readable = 1;
}


vReadAVI::~vReadAVI()
{
  vdelete2(type);
  if (fp != 0) fclose(fp);
}


grayscale_image * vReadAVI::ReadNextFrameGray()
{
  vArray(uchar) temp_buffer = ReadCurrentFrame();
  grayscale_image * result = new grayscale_image(height, width);
  vArray2(uchar) I = result->GetI2();

  long row, col;
  long index = 0;
  vArray(uchar) base1;
  long r_offset, g_offset, b_offset;
  long frame_row, frame_col;

  if (bytes_per_pixel == 3)
  {
    r_offset = 2;
    g_offset = 1;
    b_offset = 0;
  }
  else assert(bytes_per_pixel == 1);

  for (row = 0; row < height; row++)
  {
    base1 = I[row];
    frame_row = row+top;
    frame_col = left;
    index = ((rows-1-frame_row) * cols + frame_col) * bytes_per_pixel;
    
    if (bytes_per_pixel == 3)
    {
      for (col = 0; col < width; col++)
      {
	long value = temp_buffer[index+r_offset]; //R
	value += temp_buffer[index+g_offset]; // G
	value += temp_buffer[index+b_offset]; // B
	base1[col] = (uchar) (value / 3);
	index += bytes_per_pixel;
      }
    }
    else
    {
      for (col = 0; col < width; col++)
      {
	base1[col] = temp_buffer[index];
	index+= bytes_per_pixel;
      }
    }
  }

  vdelete2(temp_buffer);
  return result;
}


grayscale_image * vReadAVI::ReadFrameGray(long number)
{
  if (number >= number_of_frames) return 0;
  if (number < 0) return 0;
  long frame_start = initial_offset + number * frame_displacement;
  fseek(fp, frame_start - current_position, SEEK_CUR);
  current_position = frame_start;
  current_frame = number;
  return ReadNextFrameGray();
}


color_image * vReadAVI::ReadFrame(long number)
{
  if (number >= number_of_frames) return 0;
  if (number < 0) return 0;
  long frame_start = initial_offset + number * frame_displacement;
  fseek(fp, frame_start - current_position, SEEK_CUR);
  current_position = frame_start;
  current_frame = number;

  vArray(uchar) temp_buffer = ReadCurrentFrame();
  if (vIsZero(temp_buffer)) return 0;
  color_image * result = new color_image(height, width);
  vArray2(uchar) R = result->GetR2();
  vArray2(uchar) G = result->GetG2();
  vArray2(uchar) B = result->GetB2();
  long row, col;
  long index = 0;
  vArray(uchar) base1;
  vArray(uchar) base2;
  vArray(uchar) base3;
  long r_offset, g_offset, b_offset;
  if (bytes_per_pixel == 1)
  {
    r_offset = 0;
    g_offset = 0;
    b_offset = 0;
  }
  else
  {
    assert(bytes_per_pixel == 3);
    r_offset = 2;
    g_offset = 1;
    b_offset = 0;
  }

  for (row = 0; row < height; row++)
  {
    long frame_row, frame_col;
    base1 = R[row];
    base2 = G[row];
    base3 = B[row];
    frame_row = row+top;
    frame_col = left;
    index = ((rows-1-frame_row) * cols + frame_col) * bytes_per_pixel;

    for (col = 0; col < width; col++)
    {
      base1[col] = temp_buffer[index+r_offset]; // R
      base2[col] = temp_buffer[index+g_offset]; // G
      base3[col] = temp_buffer[index+b_offset]; // B
      index += bytes_per_pixel;
    }
  }

  vdelete2(temp_buffer);
  return result;
}


color_image * vReadAVI::ReadNextFrame()
{
  return ReadFrame(current_frame + 1);
}


ushort vReadAVI::InitializeWriting()
{
  exit_error("Error: vReadAVI sequences are not writable.\n");
  return 0;
}


ushort vReadAVI::write_frame(color_image * the_image)
{
  exit_error("Error: vReadAVI sequences are not writable.\n");
  return 0;
}


ushort vReadAVI::write_frameGray(grayscale_image * the_image)
{
  exit_error("Error: vReadAVI sequences are not writable.\n");
  return 0;
}


ushort vReadAVI::FinishWriting()
{
  exit_error("Error: vReadAVI sequences are not writable.\n");
  return 0;
}


ushort vReadAVI::GotoFrame(long frame)
{
  if (frame >= number_of_frames) return 0;
  if (frame < 0) return 0;

  long frame_start = initial_offset + frame * frame_displacement;
  fseek(fp, frame_start - current_position, SEEK_CUR);
  current_position = frame_start;
  current_frame = frame;

  return 1;
}


vArray(uchar) vReadAVI::ReadFrameBuffer(long number)
{
  if (GotoFrame(number) == 0) return vZero(uchar);
  else return ReadCurrentFrame();
}


vArray(uchar) vReadAVI::ReadCurrentFrame()
{
  long temp_header = ReadLong();
  if (temp_header != frame_header) 
  {
    current_position = 0;
    fseek(fp, 0, SEEK_SET);
    return vZero(uchar);
  }
  long temp_bytes_per_frame = ReadLong();
  if(temp_bytes_per_frame != bytes_per_frame) 
  {
    current_position = 0;
    fseek(fp, 0, SEEK_SET);
    return vZero(uchar);
  }

  vArray(uchar) result = vnew(uchar, bytes_per_frame);
  long bytes_read = fread(vData(result), 1, bytes_per_frame, fp);
  if (bytes_read != bytes_per_frame)
  {
    vdelete2(result);
    vSetZero(result);
    current_position = 0;
    fseek(fp, 0, SEEK_SET);
  }
  current_position = current_position + bytes_read;
  return result;
}


vArray(uchar) vReadAVI::GetPreamble()
{
  fseek(fp, 0, SEEK_SET);
  vArray(uchar) result = vnew(uchar, initial_offset);
  long items_read = fread(vData(result), sizeof(uchar), initial_offset, fp);
  if (items_read != initial_offset)
  {
    vdelete2(result);
    vSetZero(result);
  }
  fseek(fp, current_position, SEEK_SET);
  return result;
}


long vReadAVI::InitialOffset()
{
  return initial_offset;
}


long vReadAVI::BytesPerPixel()
{
  return bytes_per_pixel;
}


long vReadAVI::BytesPerFrame()
{
  return bytes_per_frame;
}


vArray(uchar) vReadAVI::ReadTextFrame(long number)
{
  if (number >= number_of_frames) return vZero(uchar);
  if (number < 0) return vZero(uchar);
  long frame_start = initial_offset + number * frame_displacement;
  frame_start = frame_start + bytes_per_frame + 16;
  fseek(fp, frame_start, SEEK_SET);
  vArray(uchar) result = vnew(uchar, 74);
  fread(vData(result), sizeof(uchar), 74, fp);
  current_position = frame_start + 74;
  return result;
}

vAVITimeStamp * vReadAVI::ReadTimeStamp(long number)
{
  printf("Extra bytes = %li\n", frame_displacement - bytes_per_frame - 8);
  printf("frame displ = %li\n", frame_displacement);
  printf("Initial offset = %li\n", initial_offset);
  if (number >= number_of_frames) return 0;
  if (number < 0) return 0;
  long frame_start = initial_offset + number * frame_displacement;
  frame_start = frame_start + bytes_per_frame + 16;
  fseek(fp, frame_start, SEEK_SET);
  current_position = frame_start;
  current_frame = number;

  vAVITimeStamp * result = new vAVITimeStamp;
  fread(result->date, sizeof(char), 19, fp);
  result->date[19] = 0;

  long items_read;
  
  fseek(fp, frame_start + 20, SEEK_SET);
  items_read = fscanf(fp, "%li", &(result->milli_sec_high));
  if (items_read != 1) 
  {
    printf("returning 1\n");
    return result;
  }
  printf("read milli_sec_high = %li\n", result->milli_sec_high);

  fseek(fp, frame_start + 35, SEEK_SET);
  items_read = fscanf(fp, "%li", &(result->milli_sec_low));
  if (items_read != 1) 
  {
    printf("returning 2\n");
    return result;
  }
  printf("read milli_sec_low = %li\n", result->milli_sec_low);
  
  current_position = frame_start + 50;
  fseek(fp, current_position, SEEK_SET);
  fread(&(result->checksum), sizeof(long), 1, fp);
  fread(&(result->seconds), sizeof(long), 1, fp);
  fread(&(result->milli_sec_low), sizeof(long), 1, fp);
  fread(&(result->milli_sec_high), sizeof(long), 1, fp);
  fread(&(result->frame_id_low), sizeof(long), 1, fp);
  fread(&(result->frame_id_high), sizeof(long), 1, fp);
  current_position = frame_start + 74;
  return result;
}


void vAVITimeStampBuffer(vAVITimeStamp * time_stamp,
		       char * buffer)
{
  sprintf(buffer, "%s %-14li %-14li ",
          time_stamp->date, 
          time_stamp->milli_sec_high, 
          time_stamp->milli_sec_low);
  memcpy(&(buffer[50]), &(time_stamp->checksum), sizeof(long));
  memcpy(&(buffer[54]), &(time_stamp->seconds), sizeof(long));
  memcpy(&(buffer[58]), &(time_stamp->milli_sec_low), sizeof(long));
  memcpy(&(buffer[62]), &(time_stamp->milli_sec_high), sizeof(long));
//  memcpy(&(buffer[66]), &(time_stamp->frame_id), sizeof(ULONGLONG));
}


void vAVITimeStamp::Print()
{
  printf("seconds: %li\n", seconds);
  printf("checksum: %li\n", checksum);
  printf("milliseconds: %li, %li\n", milli_sec_high, milli_sec_low);
  printf("date: %s\n", date);
  printf("frame_id: %li, %li\n", frame_id_high, frame_id_low);
  printf("frame_number: %li\n", frame_number);
  printf("\n");
}

/*
unsigned short vGetTimeStamp(FILE * in_fp, vAVITimeStamp * time_stamp)
{
  long items1 = fread(&(time_stamp->checksum),sizeof(long), 1, in_fp);
  if (items1 == 0) return 0;
  long items2 = fread(&(time_stamp->seconds), 
                      sizeof(long), 1, in_fp);
  long items3 = fread(&(time_stamp->milli_sec_low),
                      sizeof(long), 1, in_fp);
  long items4 = fread(&(time_stamp->milli_sec_high),
                      sizeof(long), 1, in_fp);
  long items5 = fread(&(time_stamp->frame_id),
                      sizeof(ULONGLONG), 1, in_fp);
  assert((items1 == 1) && (items2 == 1) && (items3 == 1) &&
         (items4 == 1) && (items5 == 1));

  char * time_string = ctime(&(time_stamp->seconds));
  S_ConvertTimeString(time_string, time_stamp->date);
  S_GetBinaryDigits(time_stamp->frame_id, 
                    time_stamp->frame_id_digits);
  S_GetString(time_stamp->frame_id, time_stamp->frame_id_string);

  return 1;
}
*/
