#include "stdafx.h"

#include "vplatform.h"
#include "rle_encoding.h"
#include "avi_file.h"
#include "drawing_temp.h"
#include "wrapper.h"

#include "definitions.h"

// This should always be smaller than 2^16 - 1, or I should make the 
// "changes" variable of vRLE_Image an array of vectors of longs.
// Now I assume that a "length", a number of consecutive appearance
// of a value in an image row can be represented as a ushort.
static const long COL_LIMIT = 20000;
static const long NUMBER_OF_FRAMES_LIMIT = 2000000; // 2 million

// I need the file size to be less than 2^31, as long as I want to
// be able to run my code on 32-bit architectures. The file offsets
// will not work otherwise (they are signed longs).
static const long FILE_SIZE_LIMIT = 2000000000; // 2 billion

//////////////////////////////////////////
// Implementation of class vRLE_Image
//////////////////////////////////////////


// Default constructor. It should mark the object as invalid, because
// (until we put some stuff in there and mark it as valid) it is utterly 
// useless
vRLE_Image::vRLE_Image()
{
  changes = 0;
  is_valid = 0;
  deleted = 0;
}

// This constructor initializes the "changes" array of vectors. The
// object is still not representing a valid image, though, so
// we still mark it as invalid.
vRLE_Image::vRLE_Image(long in_rows, long in_cols)
{
  is_valid = 0;
	rows = in_rows;
	cols = in_cols;
	changes = new vector<ushort>[rows];
  deleted = 0;
}
 
 
vRLE_Image::~vRLE_Image()
{
	vdelete2(changes);
  // We can't call DeleteSelf from here, because in C++ we are not allowed
  // to call an object's virtual functions after we have entered its destructor.
  // To make sure that there are no memory leaks, we print out an error message
  // if deleted is still 0 (meaning that noone (no subclass destructor) called 
  // DeleteSelf()
  if (deleted == 0)
  {
    vPrint("Memory leak, ~vRLE_Image called before DeleteSelf\n");
  }
}


vRLE_Image * vRLE_Image::Read(FILE * fp)
{
  // We read just enough to get the format (9th byte). Then
  // based on the format's value we call the function of the appropriate
  // subclass.
  long numbers[2]; // rows, cols, format.
  long items_read = vReadPCLongs(fp, numbers, 2);
  if (items_read != 2) return 0;
  // Put the stuff back, so that the specialized function that we will
  // call can read it again.
  uchar format = 0;
  items_read = fread(&format, sizeof(uchar), 1, fp);
  if (items_read != 1) return 0;

  int error = fseek(fp, -9, SEEK_CUR);
  // If an error occurred in fseek, then return 0
  if (error != 0) return 0;
  switch(format)
  {
  case 0:
    return vRLE_Binary::Read(fp);
    
  case 1:
    return vRLE_Gray::Read(fp);

  case 2:
    return vRLE_Mapped::Read(fp);

  default:
    return 0;
  }
}


// Every subclass defines a ToImage function, and here
// we use that fact
ushort vRLE_Image::SavePPM(const char * filename)
{
  color_image * the_image = ToImage();
  ushort success = the_image->save(filename);
  return success;
}

// 1 for binary, 8 for grayscale, 24 for color, I don't
// even know if this function is useful.
long vRLE_Image::BitsPerPixel()
{
  return bits_per_pixel;
}


// 0 for binary, 1 for grayscale, 2 for colormapped
long vRLE_Image::Format()
{
  return format;
}

// Returns the array of vectors "changes"
std::vector<ushort> * vRLE_Image::GetChanges()
{
  return changes;
}


// Returns the sum of lengths of all vectors in the changes array.
// This way we can calculate how much storage we need (in an array
// or a file) to save the values stored in all those vectors.
long vRLE_Image::ChangesSize()
{
	long result = 0;
	long row;
	for (row = 0; row < rows; row++)
	{
		result += changes[row].size();
	}
	return result;
}



///////////////////////////////////////////
// Implementation of class vRLE_Binary
///////////////////////////////////////////

// This function assumes that nothing useful is currently stored
// in the vRLE_Binary object, and does the necessary work
// so that this object is a binary RLE representation of the image
// (note that this is lossy, the image can have 256 distinct values,
// but converting from the RLE binary back to a gray image we 
// will only get an image with 2 distinct values, 0 and 255). All values
// in the gray image that are non-zero are mapped to 255.
ushort vRLE_Binary::FromGray(grayscale_image * image)
{
	rows = image->Rows();
	cols = image->Cols();
	changes = new vector<ushort>[rows];

	vArray2(uchar) I = image->Matrix2(0);
	long row, col;

  // Go through each row and RLE encode it.
	for (row = 0; row < rows; row++)
	{
		vector<ushort> & entries = changes[row];
    // The first value in entries will be the number
    // of value changes in the row, the next line just 
    // reserves space for that.
		entries.push_back(0);
		ushort counter = 0;
    // We assume that the first value occuring in the row is
    // 0. If it is non-zero, it's still OK, we will treat that
    // as the value 0 occuring 0 consecutive times.
		uchar previous = 0; 
		vArray(uchar) base = I[row];
		ushort changes = 0;
		for (col = 0; col < cols; col++)
		{
      // If the current value that we read from the image is not 
      // the same as the previous value, then we store the 
      // number of consecutive occurrences of the previous value that we 
      // counted into the entries vector.
			uchar current;
			if (base[col] == 0) current = 0;
      // All non-zero pixels in the input image are treated as having
      // the value 255.
			else current = 255;
			if (current != previous)
			{
				previous = current;
				entries.push_back(counter);
				counter = 1;
				changes++;
			}
			else
			{
        counter++;
			}
		}
    // Put the total number of changes occuring in the row into the first
    // element of entries
		entries[0] = changes;
	}
  // We successfully finished the conversion
  is_valid = 1;
  return 1;
}


vRLE_Binary::vRLE_Binary(grayscale_image * image)
{
  format = 0;
  bits_per_pixel = 1;
  // FromGray does all the work.
  FromGray(image);
}


// Convert the image into a grayscale image and call
// the FromGray function. It would be a little more
// efficient to do the conversion directly, but at 
// this point I don't think it's worth the effort.
vRLE_Binary::vRLE_Binary(color_image * image)
{
  format = 0;
  bits_per_pixel = 1;
  grayscale_image * gray = image->ToGray();
  FromGray(gray);
  vdelete(gray);
}


vRLE_Binary::vRLE_Binary(long in_rows, long in_cols) :
vRLE_Image(in_rows, in_cols)
{
  format = 0;
}


vRLE_Binary::~vRLE_Binary()
{
  DeleteSelf();
}

// Why we define DeleteSelf and not put all that stuff in the destructor?
// Destructors in C++ can't be virtual. This function is virtual, so we can
// call it whenever we have an object of type vRLE_Image, without having
// to figure out its subclass. Otherwise, to delete that object properly we
// would have to cast it into the right subclass before calling delete.
void vRLE_Binary::DeleteSelf()
{
  // Not much to do for a vRLE_Binary object, the vRLE_Image 
  // destructor deallocates the "changes" array.
  if (deleted != 0) return;
  deleted = 1;
}


grayscale_image * vRLE_Binary::ToGrayImage()
{
	grayscale_image * result = new grayscale_image(rows, cols);
	vArray2(uchar) I = result->Matrix2(0);

	long row;
  // Each row is decoded separately
	for (row = 0; row < rows; row++)
	{
		long index = 1;
		vector<ushort> & entries = changes[row];
		ushort value = 0;
		vArray(uchar) base = I[row];
		long col = 0;
    // For each entry in the entries vector, read it and 
    // use it to write the corresponding value the appropriate 
    // number of consecutive times.
    while(index < (long) entries.size())
		{
			ushort times = entries[index++];
			ushort i;
			for (i = 0; i < times; i++)
			{
				base[col++] = (uchar) value;
			}
      // The value alternates between 0 and 255.
			value = 255 - value;
		}
    // To save space, we don't write the length of the last set
    // of consecutive values, since it can be derived from 
    // the rest.
    while(col < cols)
		{
			base[col++] = (uchar) value;
		}
	}
	return result;
}


color_image * vRLE_Binary::ToImage()
{
  grayscale_image * gray = ToGrayImage();
  color_image * the_image = gray->ToRGB();
  delete gray;
  return the_image;
}

// Write the RLE image into a file (possibly as part 
// of a vRLE_Sequence.
ushort vRLE_Binary::Write(FILE * fp)
{
	ulong size;
	size = ChangesSize();

  // as long as cols < COL_LIMIT we can use 
  // ushorts to represent the lengths. Why do we check 
  // here in particular? I don't remember.
	assert(cols < COL_LIMIT);
  // Write the header (rows, cols, version, size)
	long items;
  items = vWritePCLong(fp, rows);
  if (items != 1) return 0;
  items = vWritePCLong(fp, cols);
  if (items != 1) return 0;

	uchar version = (uchar) format;
	items = fwrite(&version, sizeof(uchar), 1, fp);
  if (items != 1) return 0;
	items = vWritePCLong(fp, size);
  if (items != 1) return 0;

  // Put the contents of all the vectors in "changes" into
  // an array, so that we can write it more efficiently 
  // into the file.
	long row;
	ushort * buffer = new ushort[size];
	long index = 0;
	for (row = 0; row < rows; row++)
	{
		vector<ushort> & entries = changes[row];
		long length = entries.size();
		long i;
		for (i = 0; i < length; i++)
		{
			buffer[index++] = entries[i];
		}
	}

  // Write the whole array intot he file.
	items = vWritePCUshorts(fp, buffer, size);
	delete [] buffer;
  if (items != size) return 0;

  return 1;
}


vRLE_Binary * vRLE_Binary::Read(FILE * fp)
{
  long rows, cols;

	long items = vReadPCLong(fp, &rows);
	if (items != 1) return 0;
	items = vReadPCLong(fp, &cols);
	if (items != 1) return 0;

  // See comments where I define COL_LIMIT
  if (cols >= COL_LIMIT)
  {
    printf("Error, at most %li columns are allowed in an RLE image (%li)\n",
           COL_LIMIT, cols);
    return 0;
  }
  vRLE_Binary * result = new vRLE_Binary(rows, cols);

  // Read in the contents of "changes" and "values".
  ushort success = result->ReadData(fp);
  if (success == 0)
  {
    delete result;
    result = 0;
  }
	return result;
}

// This function gets called after we have read the 
// image rows and cols. The things that are left
// to read are: version (which should equal format),
// size of the run-length encoding and the actual
// encoding.
ushort vRLE_Binary::ReadData(FILE * fp)
{
  long items;
	uchar version;

  // the version should be 0, equal to the format.
	items = fread(&version, sizeof(uchar), 1, fp);
  if (items != 1) return 0;
  if (version != 0) return 0;

	long size;
  items = vReadPCLong(fp, &size);
  if (items != 1) return 0;

  // First read the whole encoding into an array,
  // for efficiency.
	ushort * buffer = new ushort[size];
	items = vReadPCUshorts(fp, buffer, size);
	if (items != size)
  {
    delete [] buffer;
    return 0;
  }

	long index = 0;
	long row;
  // Now move over the data from the buffer into
  // the "changes" array of vectors.
	for (row = 0; row < rows; row++)
	{
		vector<ushort> & entries = changes[row];
    // The next entry in the buffer is the length
    // of the encoding for the current row.
		ushort length = buffer[index++];
		entries.push_back(length);
		long i;
		for (i = 0; i < length; i++)
		{
			entries.push_back(buffer[index++]);
		}
	}
	delete [] buffer;
  return 1;
}

/////////////////////////////////////////
// Implementation of class vRLE_Gray
/////////////////////////////////////////

ushort vRLE_Gray::FromGray(grayscale_image * image)
{
  is_valid = 0;
	rows = image->Rows();
	cols = image->Cols();
  // kind of a trivial check, but there must be at least
  // one row and one column in the image (i.e. at least
  // one pixel. Otherwise we return (and the object is
  // still marked as invalid).
  if (rows < 1) return 0;
  if (cols < 1) return 0;
  
  changes = new vector<ushort>[rows];
  values = new vector<uchar>[rows];

	vArray2(uchar) I = image->Matrix2(0);
	long row, col;
	for (row = 0; row < rows; row++)
	{
		vector<ushort> & lengths = changes[row];
    vector<uchar> & numbers = values[row];
    // At the beginning of each vector we put the number
    // of changes occuring in each row. It's kind of redundant
    // (we could put that number just into the lengths vector)
    // but this way lengths[i] corresponds to numbers[i], which
    // is convenient.
    // Right now we don't know the number of changes, we just
    // put 0 there as a placeholder.
		lengths.push_back(0);
    numbers.push_back(0);

		vArray(uchar) base = I[row];
    // The next line is the reason why the image should have at least 
    // one column.
		uchar previous = base[0];
    // streaks counts the number of distinct streaks in the 
    // row (a streak is consecutive appearances of the same value).
		uchar current;
		ushort streaks = 1;
		ushort counter = 1;
		for (col = 1; col < cols; col++)
		{
      current = base[col];
			if (current != previous)
			{
				lengths.push_back(counter);
        numbers.push_back(previous);
				previous = current;
				counter = 1;
				streaks++;
			}
			else
			{
        counter++;
			}
		}
    lengths.push_back(counter);
    numbers.push_back(current);
		lengths[0] = streaks;
    numbers[0] = (uchar) streaks;
	}

  is_valid = 1;
  return 1;
}


vRLE_Gray::vRLE_Gray()
{
  format = 1;
  bits_per_pixel = 8;
  values = 0;
}


vRLE_Gray::vRLE_Gray(grayscale_image * image)
{
  format = 1;
  bits_per_pixel = 8;
  FromGray(image);
}


vRLE_Gray::vRLE_Gray(color_image * image)
{
  format = 1;
  bits_per_pixel = 8;
  grayscale_image * gray = image->ToGray();
  FromGray(gray);
  vdelete(gray);
}


vRLE_Gray::vRLE_Gray(long in_rows, long in_cols) :
vRLE_Image(in_rows, in_cols)
{
  format = 1;
  bits_per_pixel = 8;
  values = new vector<uchar>[rows];
}


vRLE_Gray::~vRLE_Gray()
{
  DeleteSelf();
}

// See comments on vRLE_Binary::DeleteSelf for why
// this function exists.
void vRLE_Gray::DeleteSelf()
{
  if (deleted != 0) return;
  deleted = 1;
  vdelete2(values);
  // changes is deleted by the vRLE_Image destructor
}


grayscale_image * vRLE_Gray::ToGrayImage()
{
  ushort error = 0; // Will be set to something non-zero
  // if something goes wrong.
	grayscale_image * result = new grayscale_image(rows, cols);
	vArray2(uchar) I = result->Matrix2(0);

	long row;
  // decode each row separately.
  for (row = 0; row < rows; row++)
	{
		vector<ushort> & lengths = changes[row];
    vector<uchar> & numbers = values[row];
		vArray(uchar) base = I[row];
		long col = 0;
		long index = 1;
		while(index < (long) lengths.size())
		{
			ushort length = lengths[index];
      uchar value = numbers[index];
      index++;
			long i;
			for (i = 0; i < length; i++)
			{
				base[col++] = value;
			}
		}
    // We should have filled up all pixels in the row, otherwise
    // the RLE encoding is wrong (which means there is a bug in 
    // the function that does the encoding).
    if (col != cols) 
    {
      error = 1;
      break;
    }
  }
  if (error == 1)
  {
    vdelete(result);
    return 0;
  }
	else return result;
}


color_image * vRLE_Gray::ToImage()
{
  grayscale_image * gray = ToGrayImage();
  color_image * the_image = gray->ToRGB();
  delete gray;
  return the_image;
}


ushort vRLE_Gray::Write(FILE * fp)
{
	ulong size;
  // We don't need to measure the size of the values, it 
  // should be the same as that of changes. The size is
  // defined here as the sum of the lengths of all vectors
  // in the "changes" (or "values") array.
	size = ChangesSize();

  // See comments before definition of COL_LIMIT
	assert(cols < COL_LIMIT);
	long items;
  // Write the header (rows, cols, format = 1, size)
  items = vWritePCLong(fp, rows);
  if (items != 1) return 0;
  items = vWritePCLong(fp, cols);
  if (items != 1) return 0;

  // format = 1 for vRLE_Gray objects.
  uchar format_char = (uchar) format;
	items = fwrite(&format_char, sizeof(uchar), 1, fp);
  if (items != 1) return 0;
	items = vWritePCLong(fp, size);
  if (items != 1) return 0;

  // Copy all the data from "changes" into an array, so
  // that we can write it more efficiently to file. Similarly
  // copy all data from values into an array as well.
	long row;
	ushort * changes_buffer = new ushort[size];
  uchar * values_buffer = new uchar[size];
	long index = 0;
	for (row = 0; row < rows; row++)
	{
		vector<ushort> & lengths = changes[row];
    vector<uchar> & numbers = values[row];
		long length = lengths.size();
    // If the assertion fails, there is a bug somewhere.
    assert(length == numbers.size());
		long i;
		for (i = 0; i < length; i++)
		{
			changes_buffer[index] = lengths[i];
      values_buffer[index] = numbers[i];
      index++;
		}
	}

  // Now write the changes buffer and the values buffer into a file.
	items = vWritePCUshorts(fp, changes_buffer, size);
  long items2 = fwrite(values_buffer, sizeof(uchar), size, fp);
	vdelete2(changes_buffer);
  vdelete2(values_buffer);
  if (items != size) return 0;
  if (items2 != size) return 0;

  return 1;
}


vRLE_Gray * vRLE_Gray::Read(FILE * fp)
{
  long rows, cols;

	long items = vReadPCLong(fp, &rows);
	if (items != 1) return 0;
	items = vReadPCLong(fp, &cols);
	if (items != 1) return 0;

  // See comments where I define COL_LIMIT
  if (cols >= COL_LIMIT)
  {
    printf("Error, at most %li columns are allowed in an RLE image (%li)\n",
           COL_LIMIT, cols);
    return 0;
  }
  vRLE_Gray * result = new vRLE_Gray(rows, cols);
  // This function reads the contents of the "changes" array.
	ushort success = result->ReadData(fp);
  if (success == 0)
  {
    vdelete(result);
    result = 0;
  }
	return result;
}


// This function gets called after we have read the 
// image rows and cols. The things that are left
// to read are: version (which should equal format),
// size of the run-length encoding and the actual
// encoding.
ushort vRLE_Gray::ReadData(FILE * fp)
{
  long items;
	uchar version;

  // the version should be equal to the format. note
  // that the format will be 2 if this object is 
  // an instance of the class vRLE_Mapped (which
  // inherits from vRLE_Gray).
	items = fread(&version, sizeof(uchar), 1, fp);
  if (items != 1) return 0;
  if (version != format) return 0;

	long size;
  items = vReadPCLong(fp, &size);
  if (items != 1) return 0;

  // First read the whole encoding into two arrays, (one
  // for changes and one for values) for efficiency.
	ushort * changes_buffer = new ushort[size];
  uchar * values_buffer = new uchar[size];
	items = vReadPCUshorts(fp, changes_buffer, size);
  long items2 = fread(values_buffer, 1, size, fp);
	if ((items != size) || (items2 != size))
  {
    vdelete2(changes_buffer);
    vdelete2(values_buffer);
    return 0;
  }

	long index = 0;
	long row;
  // Now move over the data from the array buffers into
  // the "changes" and "values" arrays of vectors.
	for (row = 0; row < rows; row++)
	{
		vector<ushort> & lengths = changes[row];
    vector<uchar> & numbers = values[row];
    // The next entry in each buffer is the length
    // of the encoding for the current row.
		ushort length = changes_buffer[index];
    assert(length == values_buffer[index]);
    index++;
		lengths.push_back(length);
    numbers.push_back((uchar) length);
		long i;
		for (i = 0; i < length; i++)
		{
			lengths.push_back(changes_buffer[index]);
      numbers.push_back(values_buffer[index]);
      index++;
		}
	}
  vdelete2(changes_buffer);
  vdelete2(values_buffer);
  return 1;
}


////////////////////////////////////////////////
// Implementation of class vRLE_Mapped
////////////////////////////////////////////////


vRLE_Mapped::vRLE_Mapped(grayscale_image * image) : 
vRLE_Gray(image)
{
  format = 1;
  bits_per_pixel = 24;
  color_map = 0;
  map_save_flag = 1;
  is_valid = 1;
}


vRLE_Mapped::vRLE_Mapped(grayscale_image * the_image, 
                           vColorMap * in_color_map) :
vRLE_Gray(the_image)
{
  format = 2;
  bits_per_pixel = 24;
  color_map = in_color_map->Copy();
  map_save_flag = 1;
}


// This constructor takes in a color image, verifies that it has
// fewer than 256 colors, and builds a color map for that image.
// The program will exit with an error if the image has more than
// 256 colors.
vRLE_Mapped::vRLE_Mapped(color_image * the_image)
{
  format = 2;
  bits_per_pixel = 24;

  // Get the colors and color ids of the pixels.
  v3dMatrix<ulong> * color_ids;
  color_map = new vColorMap(the_image, &color_ids);

  // if there are more than 256 colors, then this function
  // should not have been called, so the program will exit
  // with an error
  if (color_map->number_of_colors > 256)
  {
    exit_error("\nError in vRLE_Mapped constructor, %li colors\n",
                    color_map->number_of_colors);
  }

  // If there are at most 256 colors, then we can convert color_ids
  // into a gray scale image.
  grayscale_image * gray = color_ids->ToGray();
  long success = FromGray(gray);
  map_save_flag = 1;
  vdelete(color_ids);
  vdelete(gray);
}


vRLE_Mapped::vRLE_Mapped(long in_rows, long in_cols) :
vRLE_Gray(in_rows, in_cols)
{
  color_map = 0;
  map_save_flag = 1;
  is_valid = 0;
  format = 2;
}


vRLE_Mapped::~vRLE_Mapped()
{
  DeleteSelf();
}


void vRLE_Mapped::DeleteSelf()
{
  if (deleted != 0) return;
  vdelete(color_map);
  vRLE_Gray::DeleteSelf();
  deleted = 1;
}


// See comments on header file.
grayscale_image * vRLE_Mapped::ToColormappedImage()
{
  return vRLE_Gray::ToGrayImage();
}


// See comments on header file.
grayscale_image * vRLE_Mapped::ToGrayImage()
{
  // If there is no color map, then we just return 
  // the gray scale image that we have RLE-encoded.
  if (color_map == 0) return vRLE_Gray::ToGrayImage();

  color_image * color = ToImage();
  grayscale_image * result = color->ToGray();
  
  vdelete(color);
  return result;
}


color_image * vRLE_Mapped::ToImage()
{
  grayscale_image * gray = ToColormappedImage();
  color_image * the_image;
  if (color_map == 0)
  {
    the_image = gray->ToRGB();
  }
  else
  {
    the_image = color_map->Convert(gray);
  }
  vdelete(gray);
  return the_image;
}

// Writing to file is simple: First we write whatever
// we wrote for the vRLE_Gray class, then we 
// write a flag indicating whether a color map follows, and then, 
// if map_save_flag is set and the color map is not null we
// write the color map.
ushort vRLE_Mapped::Write(FILE * fp)
{
  ushort success = vRLE_Gray::Write(fp);
  if (success == 0) return 0;

  ushort zero = 0;
  ushort one = 1;

  if ((map_save_flag == 0) || (color_map == 0))
  {
    long items = vWritePCUshorts(fp, &zero, 1);
    if (items != 1) return 0;
  }
  else
  {
    long items = vWritePCUshorts(fp, &one, 1);
    if (items != 1) return 0;
    success = color_map->Write(fp);
    if (success == 0) return success;
  }

  return 1;
}


vRLE_Mapped * vRLE_Mapped::Read(FILE * fp)
{
  long rows, cols;

	long items = vReadPCLong(fp, &rows);
	if (items != 1) return 0;
	items = vReadPCLong(fp, &cols);
	if (items != 1) return 0;

  // See comments where I define COL_LIMIT
  if (cols >= COL_LIMIT)
  {
    printf("Error, at most %li columns are allowed in an RLE image (%li)\n",
           COL_LIMIT, cols);
    return 0;
  }

  vRLE_Mapped * result = new vRLE_Mapped(rows, cols);

  // Read in the contents of "changes" and "values".
  ushort success = result->ReadData(fp);
  if (success == 0)
  {
    vdelete(result);
    return 0;
  }

  ushort map_saved = 0;
  items = vReadPCUshorts(fp, &map_saved, 1);
  if (items != 1)
  {
    vdelete(result);
    return 0;
  }
  if (map_saved == 0)
  {
    result->SetColorMap(0);
  }
  else 
  {
    vColorMap * color_map = new vColorMap();
    success = color_map->Read(fp);
    if (success == 0)
    {
      vdelete(result);
      vdelete(color_map);
      return 0;
    }
    result->SetColorMap(color_map);
  }

  return result;
}


vColorMap * vRLE_Mapped::GetColorMap()
{
  return color_map;
}

void vRLE_Mapped::SetColorMap(vColorMap * new_map)
{
  vdelete(color_map);
  color_map = new_map;
  if (color_map == 0) format = 1;
  else format = 2;
}


void vRLE_Mapped::SetMapSaveFlag(ushort new_value)
{
  map_save_flag = new_value;
}


/////////////////////////////////////////////////////////////////////
// Implementation of the class vRLE_Sequence
/////////////////////////////////////////////////////////////////////


vRLE_Sequence::vRLE_Sequence(const char * in_filename)
{
  filename = vCopyString(in_filename);
  type = vCopyString("vRLE_Sequence");
  frame_offsets = vZero(long);
  color_map = 0;
  fp = fopen(filename, vFOPEN_READ);
  if (fp == 0) return;

  long items;
  items = vReadPCLong(fp, &number_of_frames);
  if (items != 1) return;
  items = vReadPCLong(fp, &rows);
  if (items != 1) return;
  items = vReadPCLong(fp, &cols);
  if (items != 1) return;

  long frame_offsets_start;
  items = vReadPCLong(fp, &frame_offsets_start);
  if (items != 1) return;
  long success;
  success = fseek(fp, frame_offsets_start, SEEK_SET);
  if (success != 0) return;

  frame_offsets = vnew(long, number_of_frames);
  items = vReadPCLongs(fp, vData(frame_offsets), number_of_frames);
  if (items != number_of_frames) return;
  success = fseek(fp, frame_offsets[0], SEEK_SET);
  assert(success == 0);

  is_valid = 1;
  is_readable = 1;
  current_frame = -1;
}


vRLE_Sequence::vRLE_Sequence(const char * in_filename, long in_rows, long in_cols)
{
  format = 2;
  filename = vCopyString(in_filename);
  type = vCopyString("vRLE_Sequence");
  rows = in_rows;
  cols = in_cols;
  number_of_frames = 0;
  frame_offsets = vZero(long);
  fp = fopen(filename, vFOPEN_WRITE);
  if (fp == 0) return;

  long items;
  // Note that number_of_frames is junk at this point, we will write
  // the actual number of frames when we have written the rest of the file.
  items = vWritePCLong(fp, number_of_frames);
  if (items != 1) return;
  items = vWritePCLong(fp, rows);
  if (items != 1) return;
  items = vWritePCLong(fp, cols);
  if (items != 1) return;
  
  // Same thing as with number of frames, this is junk right now, it will
  // be written over by the correct value later.
  long frame_offsets_start = 0;
  items = vWritePCLong(fp, frame_offsets_start);
  if (items != 1) return;

  current_frame = -1;
  is_valid = 1;
  is_writable = 1;
  color_map = 0;
}


vRLE_Sequence::~vRLE_Sequence()
{
  vdelete2(type);
  vdelete2(frame_offsets);
  if (fp != 0) fclose(fp);
  vdelete(color_map);
}


grayscale_image * vRLE_Sequence::ReadNextFrameGray()
{
  if (!is_readable) return 0;
  vRLE_Image * rle_image = vRLE_Image::Read(fp);
  if (rle_image == 0) return 0;
  grayscale_image * gray = rle_image->ToGrayImage();
  rle_image->DeleteSelf();
  vdelete(rle_image);
  current_frame++;
 
  // if we want to read only a portion of the image:
  if (top != -1)
  {
    grayscale_image * result;
    long right = left + width - 1;
    long bottom = top + height - 1;
    result = (grayscale_image *) vCopyRectangle(gray, left, right, top, bottom);
    delete gray;
    return result;
  }
  else
  {
    return gray;
  }
}


grayscale_image * vRLE_Sequence::ReadFrameGray(long number)
{
  if (!is_readable) return 0;
  if ((number < 0) || (number >= number_of_frames))
  {
    return 0;
  }
  if (current_frame != number - 1)
  {
    long offset = frame_offsets[number];
    long success = fseek(fp, offset, SEEK_SET);
    if (success != 0) return 0;
  }
  grayscale_image * result = ReadNextFrameGray();
  return result;
}


color_image * vRLE_Sequence::ReadNextFrame()
{
  if (!is_readable) return 0;
  vRLE_Image * rle_image = ReadNextRLEFrame();
  if (rle_image == 0) return 0;
  color_image * the_image = rle_image->ToImage();

  // DeleteSelf is virtual and it will properly deallocate memory
  // depending on the actual subclass that rle_image is an instance of.
  rle_image->DeleteSelf();
  delete rle_image;
  current_frame++;
 
  // if we want to read only a portion of the image:
  if (top != -1)
  {
    color_image * result;
    long right = left + width - 1;
    long bottom = top + height - 1;
    result = (color_image *) vCopyRectangle(the_image, left, right, top, bottom);
    delete the_image;
    return result;
  }
  else
  {
    return the_image;
  }
}


color_image * vRLE_Sequence::ReadFrame(long number)
{
  if (!is_readable) return 0;
  if ((number < 0) || (number >= number_of_frames))
  {
    return 0;
  }
  if (current_frame != number - 1)
  {
    long offset = frame_offsets[number];
    long success = fseek(fp, offset, SEEK_SET);
    if (success != 0) return 0;
  }
  color_image * result = ReadNextFrame();
  return result;
}


ushort vRLE_Sequence::InitializeWriting()
{
  return 1;
}


// The question is, when we pass a color image to an RLE sequence,
// how should it be encoded (as binary, grayscale, colormapped or color?)
// It will be done like this: When we create an RLE video sequence 
// that is writable, we specify a format (0 for binary, 1 for grayscale,
// 2 for colormap). general_image objects are converted as specified by
// the format. write_frame2(image, format) on the other hand allows the
// caller function to specify the format. In general, we are allowed to 
// save images of different formats into one RLE sequence, but it helps
// to always specify a default format. After all, write_frame must take
// only one argument (the function declaration comes from vVideoFile)
ushort vRLE_Sequence::write_frame(color_image * image)
{
  if (!is_writable) return 0;
  return write_frame2(image, format);
}


// See comments on format before the definition of 
// vRLE_Sequence::write_frame
// Note: For color images that we want to save as colormaps, there are
// two options: one option is for the caller
// to already have created the colormap and 
// converted the color image into a grayscale image (which, for every
// pixel, has as value the index of the corresponding color in the 
// color image). In that case, the right function to call is 
// write_frameGray2. The other option is to pass in directly an image,
// and have write_frame2 figure out the color map. In that case,
// write_frame2 with format 2 is the right function to call.
ushort vRLE_Sequence::write_frame2(color_image * image, long format)
{
  vRLE_Image * rle_image = 0;
  switch(format)
  {
  case 0: // Save image as binary RLE
    rle_image = new vRLE_Binary(image);
    break;

  case 1:
    rle_image = new vRLE_Gray(image);
    break;

  case 2:
    rle_image = new vRLE_Mapped(image);
    break;

  default:
    vPrint("vRLE_Sequence:write_frame2: format must be 0, 1, or 2\n");
    return 0;
  }

  ushort result = WriteRLEFrame(rle_image);
  rle_image->DeleteSelf();
  vdelete(rle_image);
  return result;
}


// See comments about the format before the definition of 
// vRLE_Sequence::write_frame(color_image *)
ushort vRLE_Sequence::write_frameGray(grayscale_image * gray)
{
  if (!is_writable) return 0;
  return write_frameGray2(gray, format, 0);
}


// See comments about the format before the definition of 
// vRLE_Sequence::write_frame(color_image *)
ushort vRLE_Sequence::write_frameGray2(grayscale_image * image, long format,
                                       vColorMap * colormap)
{
  vRLE_Image * rle_image;
  switch(format)
  {
  case 0: // Save image as binary RLE
    rle_image = new vRLE_Binary(image);
    break;

  case 1:
  case 2:
    rle_image = new vRLE_Gray(image);
    break;

  default:
    vPrint("vRLE_Sequence:write_frameGray2: Format must be 0, 1, or 2\n");
    return 0;
  }

  ushort result = WriteRLEFrame(rle_image);
  rle_image->DeleteSelf();
  vdelete(rle_image);
  return result;
}


// The only "tricky" thing here is that, if we are reading a 
// colormapped image, we should attach to it our most recently
// read color map, unless it has one of its own, in which case
// we must update our "most recently read color map" variable.
vRLE_Image * vRLE_Sequence::ReadNextRLEFrame()
{
  // First read in the "changes" and (if available) "values" vectors.
  vRLE_Image * result = vRLE_Image::Read(fp);
  current_frame++;
  if (result == 0) return(result);
  // if the frame is not colormapped, we are done
  if (result->Format() != 2) return(result);

  // If the frame is colormapped, check if it already knows its
  // colormap (i.e. if its color map was saved as part of the frame)
  vRLE_Mapped * colormapped = (vRLE_Mapped *) result;
  vColorMap * result_map = colormapped->GetColorMap();
  // If the frame still has no color map, we give it the 
  // most recently read color map.
  if (result_map == 0) 
  {
    vColorMap * new_map = color_map->Copy();
    colormapped->SetColorMap(new_map);
  }
  // If the frame has a color map, then its color map is
  // the most recently read color map, and we update
  // our color_map variable.
  else
  {
    vdelete(color_map);
    color_map = result_map->Copy();
  }
  return result;
}


vRLE_Image * vRLE_Sequence::ReadRLEFrame(long number)
{
  if (!is_readable) return 0;
  if ((number < 0) || (number >= number_of_frames))
  {
    return 0;
  }
  if (current_frame != number - 1)
  {
    long offset = frame_offsets[number];
    long success = fseek(fp, offset, SEEK_SET);
    if (success != 0) return 0;
  }
  vRLE_Image * result = ReadNextRLEFrame();
  return result;

}


ushort vRLE_Sequence::WriteRLEFrame(vRLE_Image * image)
{
  if (!is_writable) return 0;
  current_frame++;
  number_of_frames++;
  if (number_of_frames > NUMBER_OF_FRAMES_LIMIT)
  {
    printf("Error: RLE sequences can't have more than %li frames\n",
           NUMBER_OF_FRAMES_LIMIT);
    return 0;
  }

  long offset = ftell(fp);
  assert(frame_offsets_vector.size() == current_frame);
  frame_offsets_vector.push_back(offset);

  image->Write(fp);

  offset = ftell(fp);
  if (offset > FILE_SIZE_LIMIT)
  {
    printf("Error: RLE sequence files can't have more than %li bytes\n",
           FILE_SIZE_LIMIT);
    number_of_frames--;
    current_frame--;
    return 0;
  }

  return 1;
}


ushort vRLE_Sequence::FinishWriting()
{
  if (!is_writable) return 0;
  long offset = ftell(fp);
  long success = fseek(fp, 0, SEEK_SET);
  if (success != 0) return 0;
  long items;
  items = vWritePCLong(fp, number_of_frames);
  if (items != 1) return 0;
  success = fseek(fp, 12, SEEK_SET);
  if (success != 0) return 0;
  items = vWritePCLong(fp, offset);
  if (items != 1) return 0;
  
  success = fseek(fp, offset, SEEK_SET);
  if (success != 0) return 0;
  long i;
  for (i = 0; i < number_of_frames; i++)
  {
    long frame_offset = frame_offsets_vector[i];
    items = vWritePCLong(fp, frame_offset);
    if (items != 1) return 0;
  }
  fclose(fp);
  fp = 0;
  return 1;
}

  
// Convert an AVI sequence into an RLE sequence. Format must be 0 or 1. At 
// this point the code doesn't create colormapped rle images from color
// images, though it can handle such images if someone else creates them.
void vRLE_Sequence::AVI_To_RLE(const char * source, const char * target, 
                                   long format)
{
	vAVI_Stream * avi = new vAVI_Stream(source);
	assert(avi->IsValid());
  long rows = avi->GetRows();
  long cols = avi->GetCols();

	vRLE_Sequence * rle_sequence = new vRLE_Sequence(target, rows, cols);
  assert(rle_sequence->IsValid());

	while(1)
	{
		ushort success = avi->ReadFrame();
		if (!success) break;
		grayscale_image * gray = (grayscale_image *) avi->ToGrayImage();
    rle_sequence->write_frameGray(gray);
    delete gray;
	}
	delete avi;
  rle_sequence->FinishWriting();
  delete rle_sequence;
}


// Converts an RLE sequence into an AVI sequence. Note that,
void vRLE_Sequence::RLE_To_AVI(const char * source, const char * target, 
									    					   unsigned short color_mode)
{
	long color_bytes;
	if (color_mode) color_bytes = 3;
	else color_bytes = 1;

  vRLE_Sequence * sequence = new vRLE_Sequence(source);
  assert(sequence->IsValid());

  long frames = sequence->NumberOfFrames();
  long rows = sequence->Rows();
  long cols = sequence->Cols();

	vAVI_Stream * avi = new vAVI_Stream(target, 30, 1, rows, cols, color_bytes);
	assert(avi->IsValid());

	long frame;
  for (frame = 0; frame < sequence->number_of_frames; frame++)
	{
		vRLE_Image * rle = sequence->ReadNextRLEFrame();
		assert(rle != 0);
		grayscale_image * gray = rle->ToGrayImage();
		avi->FromGrayImage(gray);
		avi->write_frame();
		delete rle;
		delete gray;
	}
	delete avi;
	delete sequence;
}


/////////////////////////////////////////////////////////////////////
//                                                                 //
//          Implementation of the class vRLE_Sequence2            //
//                                                                 //
/////////////////////////////////////////////////////////////////////


vRLE_Sequence2::vRLE_Sequence2(const char * in_filename)
{
  filename = vCopyString(in_filename);
  type = vCopyString("vRLE_Sequence");
  frame_offsets = vZero(vint8);
  color_map = 0;
  fp = fopen(filename, vFOPEN_READ);
  if (fp == 0) return;

  long items;
  items = vReadPCLong(fp, &version);
  if (items != 1) return;
  // If version is non-negative, it means the file is saved in the 
  // older RLE format that I used, in which no version was specified
  // and the first thing saved was the number of frames.
  if (version >= 0)
  {
    number_of_frames = version;
    version = 0;
  }
  else
  {
    items = vReadPCLong(fp, &number_of_frames);
    if (items != 1) return;
  }
  
  items = vReadPCLong(fp, &rows);
  if (items != 1) return;
  items = vReadPCLong(fp, &cols);
  if (items != 1) return;

  vint8 frame_offsets_start;
  long temp = 0;
  if (version == 0)
  {
    items = vReadPCLong(fp, &temp);
    frame_offsets_start = temp;
  }
  else
  {
    items = vReadVint8(fp, &frame_offsets_start);
  }

  if (items != 1) return;

  long success;
  success = function_seek(fp, frame_offsets_start, SEEK_SET);
  if (success != 0) return;

  frame_offsets = vnew(vint8, number_of_frames);
  if (version == 0)
  {
    vArray(long) temp_longs = vnew(long, number_of_frames);
    items = vReadPCLongs(fp, vData(temp_longs), number_of_frames);
    long i;
    for (i = 0; i < number_of_frames; i++)
    {
      frame_offsets[i] = temp_longs[i];
    }
    vdelete2(temp_longs);
  }
  else
  {
    items = vReadVints8(fp, vData(frame_offsets), number_of_frames);
  }
  if (items != number_of_frames) return;

  success = function_seek(fp, frame_offsets[0], SEEK_SET);
  assert(success == 0);

  SetBounds(0, 0, rows, cols);
  is_valid = 1;
  is_readable = 1;
  current_frame = -1;
}


vRLE_Sequence2::vRLE_Sequence2(const char * in_filename, long in_rows, long in_cols)
{
  format = 0;
  filename = vCopyString(in_filename);
  type = vCopyString("vRLE_Sequence");
  rows = in_rows;
  cols = in_cols;
  number_of_frames = 0;
  frame_offsets = vZero(vint8);
  fp = fopen(filename, vFOPEN_WRITE);
  if (fp == 0) return;

  long items;
  // Note that number_of_frames is junk at this point, we will write
  // the actual number of frames when we have written the rest of the file.
  items = vWritePCLong(fp, number_of_frames);
  if (items != 1) return;
  items = vWritePCLong(fp, rows);
  if (items != 1) return;
  items = vWritePCLong(fp, cols);
  if (items != 1) return;
  
  // Same thing as with number of frames, this is junk right now, it will
  // be written over by the correct value later.
  long frame_offsets_start = 0;
  items = vWritePCLong(fp, frame_offsets_start);
  if (items != 1) return;

  current_frame = -1;
  is_valid = 1;
  is_writable = 1;
  color_map = 0;
}


vRLE_Sequence2::~vRLE_Sequence2()
{
  vdelete2(type);
  vdelete2(frame_offsets);
  if (fp != 0) fclose(fp);
  vdelete(color_map);
}


grayscale_image * vRLE_Sequence2::ReadNextFrameGray()
{
  if (!is_readable) return 0;
  vRLE_Image * rle_image = vRLE_Image::Read(fp);
  if (rle_image == 0) return 0;
  grayscale_image * gray = rle_image->ToGrayImage();
  rle_image->DeleteSelf();
  vdelete(rle_image);
  current_frame++;
 
  // if we want to read only a portion of the image:
  if (top != -1)
  {
    grayscale_image * result;
    long right = left + width - 1;
    long bottom = top + height - 1;
    result = (grayscale_image *) vCopyRectangle(gray, left, right, top, bottom);
    delete gray;
    return result;
  }
  else
  {
    return gray;
  }
}


grayscale_image * vRLE_Sequence2::ReadFrameGray(long number)
{
  if (!is_readable) return 0;
  if ((number < 0) || (number >= number_of_frames))
  {
    return 0;
  }
  if (current_frame != number - 1)
  {
    long offset = (long) frame_offsets[number];
    long success = fseek(fp, offset, SEEK_SET);
    if (success != 0) return 0;
  }
  grayscale_image * result = ReadNextFrameGray();
  return result;
}


color_image * vRLE_Sequence2::ReadNextFrame()
{
  if (!is_readable) return 0;
  vRLE_Image * rle_image = ReadNextRLEFrame();
  if (rle_image == 0) return 0;
  color_image * the_image = rle_image->ToImage();

  // DeleteSelf is virtual and it will properly deallocate memory
  // depending on the actual subclass that rle_image is an instance of.
  rle_image->DeleteSelf();
  delete rle_image;
  current_frame++;
 
  // if we want to read only a portion of the image:
  if (top != -1)
  {
    color_image * result;
    long right = left + width - 1;
    long bottom = top + height - 1;
    result = (color_image *) vCopyRectangle(the_image, left, right, top, bottom);
    delete the_image;
    return result;
  }
  else
  {
    return the_image;
  }
}


color_image * vRLE_Sequence2::ReadFrame(long number)
{
  if (!is_readable) return 0;
  if ((number < 0) || (number >= number_of_frames))
  {
    return 0;
  }
  if (current_frame != number - 1)
  {
    long offset = (long) frame_offsets[number];
    long success = fseek(fp, offset, SEEK_SET);
    if (success != 0) return 0;
  }
  color_image * result = ReadNextFrame();
  return result;
}


ushort vRLE_Sequence2::InitializeWriting()
{
  return 1;
}


// The question is, when we pass a color image to an RLE sequence,
// how should it be encoded (as binary, grayscale, colormapped or color?)
// It will be done like this: When we create an RLE video sequence 
// that is writable, we specify a format (0 for binary, 1 for grayscale,
// 2 for colormap). general_image objects are converted as specified by
// the format. write_frame2(image, format) on the other hand allows the
// caller function to specify the format. In general, we are allowed to 
// save images of different formats into one RLE sequence, but it helps
// to always specify a default format. After all, write_frame must take
// only one argument (the function declaration comes from vVideoFile)
ushort vRLE_Sequence2::write_frame(color_image * image)
{
  if (!is_writable) return 0;
  return write_frame2(image, format);
}


// See comments on format before the definition of 
// vRLE_Sequence2::write_frame
// Note: For color images that we want to save as colormaps, we 
// assume that the caller has already created the colormap and 
// converted the color image into a grayscale image (which, for every
// pixel, has as value the index of the corresponding color in the 
// color image). Therefore, we are not allowed to call this
// function with format 2. On the other hand, note that ReadFrame
// does not have such a limitation, it will convert colormapped images
// into color_image objects (i.e. uncompressed color images).
ushort vRLE_Sequence2::write_frame2(color_image * image, long format)
{
  vRLE_Image * rle_image = 0;
  switch(format)
  {
  case 0: // Save image as binary RLE
    rle_image = new vRLE_Binary(image);
    break;

  case 1:
    rle_image = new vRLE_Gray(image);
    break;

  case 2:
    vPrint("vRLE_Sequence:write_frame2 can't save a color image as a colormap\n");
    return 0;

  default:
    vPrint("vRLE_Sequence:write_frame2: format must be 0 or 1\n");
    return 0;
  }

  ushort result = WriteRLEFrame(rle_image);
  rle_image->DeleteSelf();
  vdelete(rle_image);
  return result;
}


// See comments about the format before the definition of 
// vRLE_Sequence2::write_frame(color_image *)
ushort vRLE_Sequence2::write_frameGray(grayscale_image * gray)
{
  if (!is_writable) return 0;
  return write_frameGray2(gray, format, 0);
}


// See comments about the format before the definition of 
// vRLE_Sequence2::write_frame(color_image *)
ushort vRLE_Sequence2::write_frameGray2(grayscale_image * image, long format,
                                          vColorMap * colormap)
{
  vRLE_Image * rle_image;
  switch(format)
  {
  case 0: // Save image as binary RLE
    rle_image = new vRLE_Binary(image);
    break;

  case 1:
    rle_image = new vRLE_Gray(image);
    break;

  case 2:
    rle_image = new vRLE_Mapped(image, colormap);
    break;

  default:
    vPrint("vRLE_Sequence:write_frameGray2: Format must be 0, 1, or 2\n");
    return 0;
  }

  ushort result = WriteRLEFrame(rle_image);
  rle_image->DeleteSelf();
  vdelete(rle_image);
  return result;
}


// The only "tricky" thing here is that, if we are reading a 
// colormapped image, we should attach to it our most recently
// read color map, unless it has one of its own, in which case
// we must update our "most recently read color map" variable.
vRLE_Image * vRLE_Sequence2::ReadNextRLEFrame()
{
  // First read in the "changes" and (if available) "values" vectors.
  vRLE_Image * result = vRLE_Image::Read(fp);
  current_frame++;
  if (result == 0) return(result);
  // if the frame is not colormapped, we are done
  if (result->Format() != 2) return(result);

  // If the frame is colormapped, check if it already knows its
  // colormap (i.e. if its color map was saved as part of the frame)
  vRLE_Mapped * colormapped = (vRLE_Mapped *) result;
  vColorMap * result_map = colormapped->GetColorMap();
  // If the frame still has no color map, we give it the 
  // most recently read color map.
  if (result_map == 0) 
  {
    vColorMap * new_map = color_map->Copy();
    colormapped->SetColorMap(new_map);
  }
  // If the frame has a color map, then its color map is
  // the most recently read color map, and we update
  // our color_map variable.
  else
  {
    vdelete(color_map);
    color_map = result_map->Copy();
  }
  return result;
}


vRLE_Image * vRLE_Sequence2::ReadRLEFrame(long number)
{
  if (!is_readable) return 0;
  if ((number < 0) || (number >= number_of_frames))
  {
    return 0;
  }
  if (current_frame != number - 1)
  {
    long offset = (long) frame_offsets[number];
    long success = fseek(fp, offset, SEEK_SET);
    if (success != 0) return 0;
  }
  vRLE_Image * result = ReadNextRLEFrame();
  return result;

}


ushort vRLE_Sequence2::WriteRLEFrame(vRLE_Image * image)
{
  if (!is_writable) return 0;
  current_frame++;
  number_of_frames++;
  if (number_of_frames > NUMBER_OF_FRAMES_LIMIT)
  {
    printf("Error: RLE sequences can't have more than %li frames\n",
           NUMBER_OF_FRAMES_LIMIT);
    return 0;
  }

  long offset = ftell(fp);
  assert(frame_offsets_vector.size() == current_frame);
  frame_offsets_vector.push_back(offset);

  image->Write(fp);

  offset = ftell(fp);
  if (offset > FILE_SIZE_LIMIT)
  {
    printf("Error: RLE sequence files can't have more than %li bytes\n",
           FILE_SIZE_LIMIT);
    number_of_frames--;
    current_frame--;
    return 0;
  }

  return 1;
}


ushort vRLE_Sequence2::FinishWriting()
{
  if (!is_writable) return 0;
  long offset = ftell(fp);
  long success = fseek(fp, 0, SEEK_SET);
  if (success != 0) return 0;
  long items;
  items = vWritePCLong(fp, number_of_frames);
  if (items != 1) return 0;
  success = fseek(fp, 12, SEEK_SET);
  if (success != 0) return 0;
  items = vWritePCLong(fp, offset);
  if (items != 1) return 0;
  
  success = fseek(fp, offset, SEEK_SET);
  if (success != 0) return 0;
  long i;
  for (i = 0; i < number_of_frames; i++)
  {
    long frame_offset = (long) frame_offsets_vector[i];
    items = vWritePCLong(fp, frame_offset);
    if (items != 1) return 0;
  }
  fclose(fp);
  fp = 0;
  return 1;
}

  
// Convert an AVI sequence into an RLE sequence. Format must be 0 or 1. At 
// this point the code doesn't create colormapped rle images from color
// images, though it can handle such images if someone else creates them.
void vRLE_Sequence2::AVI_To_RLE(const char * source, const char * target, 
                                   long format)
{
	vAVI_Stream * avi = new vAVI_Stream(source);
	assert(avi->IsValid());
  long rows = avi->GetRows();
  long cols = avi->GetCols();

	vRLE_Sequence * rle_sequence = new vRLE_Sequence(target, rows, cols);
  assert(rle_sequence->IsValid());

	while(1)
	{
		ushort success = avi->ReadFrame();
		if (!success) break;
		grayscale_image * gray = (grayscale_image *) avi->ToGrayImage();
    rle_sequence->write_frameGray(gray);
    delete gray;
	}
	delete avi;
  rle_sequence->FinishWriting();
  delete rle_sequence;
}


// Converts an RLE sequence into an AVI sequence. Note that,
void vRLE_Sequence2::RLE_To_AVI(const char * source, const char * target, 
									    					   unsigned short color_mode)
{
	long color_bytes;
	if (color_mode) color_bytes = 3;
	else color_bytes = 1;

  vRLE_Sequence2 * sequence = new vRLE_Sequence2(source);
  assert(sequence->IsValid());

  long frames = sequence->NumberOfFrames();
  long rows = sequence->Rows();
  long cols = sequence->Cols();

	vAVI_Stream * avi = new vAVI_Stream(target, 30, 1, rows, cols, color_bytes);
	assert(avi->IsValid());

	long frame;
  for (frame = 0; frame < sequence->number_of_frames; frame++)
	{
		vRLE_Image * rle = sequence->ReadNextRLEFrame();
		assert(rle != 0);
		grayscale_image * gray = rle->ToGrayImage();
		avi->FromGrayImage(gray);
		avi->write_frame();
		delete rle;
		delete gray;
	}
	delete avi;
	delete sequence;
}


