#ifndef VASSILIS_RLE_ENCODING_H
#define VASSILIS_RLE_ENCODING_H


#include <vector>

#include "video_file.h"

template<class type>
class v3dMatrix; 

#include "auxiliaries.h"
#include "drawing.h"

// This format stores an image in a way that takes little
// space in memory, using run-length encoding. Run-length encoding
// is beneficial when pixels of the same color tend to appear in 
// consecutive order in the rows of the image. We take care of 
// three cases: Binary images, gray-scale images and colormapped
// images (with at most 256 colors).
// 
// How is it done? For each row, we keep store (in the vector changes[row])
// the lengths of consecutive occurrences of the same value. For binary
// images, that's all we have to do, we assume that the first value is 0
// (if not, its corresponding length is set to 0), the next value is 1, then 0,
// and so on.
// For gray scale images,  and for colormapped images (which are essentially gray
// scale images with a colormap mapping their values to actual RGB colors) we also
// need, for each change recorded in the changes[row] vector, to record the value.
// We do that in the values[row] vector. 
class vRLE_Image
{
protected:
  // Note that vRLE_Image is an abstract class. bits_per_pixel
  // is 1 for vRLE_Binary objects, 8 for vRLE_Gray objects
  // and 24 for Temp_RLE_Color objects.
  long bits_per_pixel;
  
  // 0 for binary images, 1 for grayscale, 2 for color images using
  // a color map, 3 for color images not using a color map.
  long format;
  long rows, cols;
	std::vector<ushort> * changes;
	
  // Since we often create a vRLE_Image by reading data from a file,
  // we need a flag to show if something goes wrong. is_valid is set 
  // by the Read function to true (1) iff it successfully reads all
  // the information from the file.
  ushort is_valid;	

  // DeleteSelf will set deleted to 1 when it gets called, so that if
  // it gets called again it will know that it shouldn't do anything.
  ushort deleted;

  virtual	ushort ReadData(FILE * fp) = 0;

public:
  // Default constructor, included for convenience, it sets data and is_valid
  // to 0.
  vRLE_Image();
  vRLE_Image(long in_rows, long in_cols);
  ~vRLE_Image();

  // The destructor will just call DeleteSelf. This way even if
  // we don't know the subclass of a vRLE_Image object we can
  // just call its destructor and it will know how to delete the object.
  virtual void DeleteSelf() = 0;
  
  virtual grayscale_image * ToGrayImage() = 0;
  virtual color_image * ToImage() = 0;
	virtual ushort Write(FILE * fp) = 0;
	static vRLE_Image * Read(FILE * fp);
	ushort SavePPM(const char * filename);
  long BitsPerPixel();
  long Format();
  std::vector<ushort> * GetChanges();
  // This is the sum of the lengths of all vectors in changes.
  long ChangesSize();

  inline long Rows()
  {
    return rows;
  }

  inline long Cols()
  {
    return cols;
  }
};

// Run-length encoding of binary images (all non-zero values are considered
// to be 255).
class vRLE_Binary : public vRLE_Image
{
private:
  ushort ReadData(FILE * fp);

  // What this function does is used by both constructors
  // that take in images.
  ushort FromGray(grayscale_image * image);

public:
  vRLE_Binary(color_image * the_image);
  vRLE_Binary(grayscale_image * the_image);
  vRLE_Binary(long in_rows, long in_cols);
  ~vRLE_Binary();

  virtual void DeleteSelf();
  virtual grayscale_image * ToGrayImage();
  virtual color_image * ToImage();
	virtual ushort Write(FILE * fp);
	static vRLE_Binary * Read(FILE * fp);
};


// Run-length encoding of grayscale images.
class vRLE_Gray : public vRLE_Image
{
protected:
  // In addition to the "lengths" of consecutive appearances that
  // we have in vRLE_Binary, here we must also keep track of the
  // values that those appearances correspond to.
  std::vector<uchar> * values;

  ushort ReadData(FILE * fp);
  
  // FromGray() is called to do the work of converting a grayscale_image into
  // a vRLE_Gray image, because that work should be done both
  // by the vRLE_Gray constructors and by
  // the vRLE_Mapped constructors
  ushort FromGray(grayscale_image * image);

public:
  vRLE_Gray();
  vRLE_Gray(color_image * the_image);
  vRLE_Gray(grayscale_image * the_image);
  vRLE_Gray(long in_rows, long in_cols);
  ~vRLE_Gray();

  virtual void DeleteSelf();
  virtual grayscale_image * ToGrayImage();
  virtual color_image * ToImage();
	virtual ushort Write(FILE * fp);
	static vRLE_Gray * Read(FILE * fp);
  inline vector<uchar> * GetValues()
  {
    return values;
  }
};


// Run-length encoding of colormapped images. A colormapped image (at this point
// at least) is defined as a gray scale image plus a vColorMap object
// that establishes correspondences between a number between 0 and 255 
// and a color. This class does not create the colormapped images, it
// just takes them as input and encodes them using RLE encoding. The RLE
// part is identical as for the gray scale images.
class vRLE_Mapped : public vRLE_Gray
{
private:
  vColorMap * color_map;

  // It would be a waste of space to save the colormap with each
  // frame in a sequence where all frames have the same color map.
  // On the other hand, we would like to have frames that use
  // different color maps in the same sequence. map_save_flag will
  // be set to true if we want to save the color map with this image.
  // vRLE_Sequence will compare the last color map it saved with
  // the current color map, and it will set the flag to true if
  // they are different.
  ushort map_save_flag;

public:
  // This constructor takes a color map (of at most 256 colors) and 
  // a gray scale images (which uses indices into the color map as
  // pixel values) and produces an RLE representation of that.
  vRLE_Mapped(grayscale_image * the_image, vColorMap * color_map);

  // This constructor takes in a color image, verifies that it has
  // fewer than 256 colors, and builds a color map for that image.
  // The program will exit with an error if the image has more than
  // 256 colors.
  vRLE_Mapped(color_image * the_image);

  // This constructor just RLE-encodes color_image and sets color_map to 0
  // (we don't need a colormap for a gray scale image.
  vRLE_Mapped(grayscale_image * the_image);
  vRLE_Mapped(long in_rows, long in_cols);
  ~vRLE_Mapped();

  virtual void DeleteSelf();

  // ToColormappedImage returns the gray image whose value at each pixel
  // is the index into the appropriate color in the color map. It is simply
  // the result of calling vRLE_Gray::ToGrayImage
  grayscale_image * ToColormappedImage();

  // ToGrayImage calls ToImage (which uses the color 
  // map to create a normal color image) and converts the
  // resulting color image to gray scale.
  virtual grayscale_image * ToGrayImage();

  // ToImage uses the color map to create a regular, non-colormapped 
  // color image.
  virtual color_image * ToImage();
	virtual ushort Write(FILE * fp);
	static vRLE_Mapped * Read(FILE * fp);
  vColorMap * GetColorMap();
  void SetColorMap(vColorMap * new_map);
  void SetMapSaveFlag(ushort new_value);
};


// vRLE_Sequence is a class that can be used to read to and write 
// from video sequence files that follow my own run-length encoding format.
// Each frame can be saved in three ways, as a binary RLE image, as a grayscale
// RLE image or as a colormapped RLE image (i.e. using at most 256 colors). At
// some point I may allow color images with an arbitrary number of colors,
// the code can be enhanced easily.
// An interesting thing to note is that, although a vRLE_Sequence object has
// attributes called "rows" and "cols", an image stored in such a sequence
// can have any number of rows or cols.
class vRLE_Sequence : public vVideoFile
{
private:
  // I am making this code backward-compatible with a previous version
  // of RLE sequences, that worked only for binary images. In that
  // file format, the first thing in the file is a long with the number of 
  // frames. In the newer version it is a long that has a format version
  // number. If it is a format version, I will always make it less than 0,
  // so that the code can tell if it is the number of frames (positive)
  // or a format version (< 0).
  long format;
  vector<long> frame_offsets_vector;
  vArray(long) frame_offsets;

  // The sequence should remember the last color map it has loaded from 
  // a file, so that it can attach it to the next frame that it reads,
  // if the frame doesn't have its own color map. Also, when writing
  // frames to file, the sequence should check the frame's color map 
  // for (for frames of type vRLE_Mapped) to see if it is
  // equal to the previous color map. If not, the frame should save
  // its color map to file.
  vColorMap * color_map;

public:
  // Creates a sequence that can be read.
  vRLE_Sequence(const char * filename);

  // Creates a sequence to which we can write.
  vRLE_Sequence(const char * filename, long rows, long cols);

  ~vRLE_Sequence();

  // Functions any vVideoSequence subclass must implement
  grayscale_image * ReadNextFrameGray();
  grayscale_image * ReadFrameGray(long number);
  color_image * ReadFrame(long number);
  color_image * ReadNextFrame();
  ushort InitializeWriting();
  ushort write_frame(color_image * image);
  ushort write_frameGray(grayscale_image * gray);
  ushort FinishWriting();

  // Functions specific to vRLE_Sequence
  // If we call write_frame, it should know if we want it to save a frame
  // as binary, grayscale or colormapped. It figures that out using the
  // member variable "format". In these functions we specify the format
  // explicitly using frame_format (0 = binary, 1 = gray, 2 = colormapped).
  ushort write_frame2(color_image * image, long frame_format);

  // This function saves the gray image that is used as input as a 
  // binary, grayscale, or colormapped image.
  ushort write_frameGray2(grayscale_image * image, long frame_format, 
                         vColorMap * colormap);

  vRLE_Image * ReadRLEFrame(long number);
  vRLE_Image * ReadNextRLEFrame();
  ushort WriteRLEFrame(vRLE_Image * image);

  inline ushort SetFormatGray()
  {
    format = 1;
    return 1;
  }

  // Converts AVIs to RLE sequences. The format specifies if each frame
  // should be converted to a binary or grayscale image (support for
  // color images is not there yet, I don't know if I will ever work on that).
	static void AVI_To_RLE(const char * source, const char * target, long format);
	static void RLE_To_AVI(const char * target, const char * source, 
												 unsigned short color_mode);
  virtual long BytesPerPixel()
  {
    if (format == 2) return 3;
    else return 1;
  }
};


// vRLE_Sequence2 is a class that can be used to read to and write 
// from video sequence files that follow my own run-length encoding format.
// Each frame can be saved in three ways, as a binary RLE image, as a grayscale
// RLE image or as a colormapped RLE image (i.e. using at most 256 colors). At
// some point I may allow color images with an arbitrary number of colors,
// the code can be enhanced easily.
class vRLE_Sequence2 : public vVideoFile
{
private:
  // I am making this code backward-compatible with a previous version
  // of RLE sequences, that worked only for binary images. In that
  // file format, the first thing in the file is a long with the number of 
  // frames. In the newer version it is a long that describes the version
  // number. If it is a version, I will always make it less than 0,
  // so that the code can tell if it is the number of frames (positive)
  // or a version (< 0).
  long version;

  // format is used in writable sequences, to describe the default format
  // (binary, grayscale or color) in which images passed into 
  // write_frame and write_frameGray must be saved. To save images in a
  // non-default format we just have to call write_frame2 or write_frameGray2,
  // or, if we already have a vRLE_Image object, Write_RLE_Frame.
  long format;
  vector<vint8> frame_offsets_vector;
  vArray(vint8) frame_offsets;

  // The sequence should remember the last color map it has loaded from 
  // a file, so that it can attach it to the next frame that it reads,
  // if the frame doesn't have its own color map. Also, when writing
  // frames to file, the sequence should check the frame's color map 
  // for (for frames of type vRLE_Mapped) to see if it is
  // equal to the previous color map. If not, the frame should save
  // its color map to file.
  vColorMap * color_map;

public:
  // Creates a sequence that can be read.
  vRLE_Sequence2(const char * filename);

  // Creates a sequence to which we can write.
  vRLE_Sequence2(const char * filename, long rows, long cols);

  ~vRLE_Sequence2();

  // Functions any vVideoSequence subclass must implement
  grayscale_image * ReadNextFrameGray();
  grayscale_image * ReadFrameGray(long number);
  color_image * ReadFrame(long number);
  color_image * ReadNextFrame();
  ushort InitializeWriting();
  ushort write_frame(color_image * image);
  ushort write_frameGray(grayscale_image * gray);
  ushort FinishWriting();

  // Functions specific to vRLE_Sequence
  // If we call write_frame, it should know if we want it to save a frame
  // as binary, grayscale or colormapped. It figures that out using the
  // member variable "format". In these functions we specify the format
  // explicitly using frame_format (0 = binary, 1 = gray, 2 = colormapped).
  ushort write_frame2(color_image * image, long frame_format);

  // This function saves the gray image that is used as input as a 
  // binary, grayscale, or colormapped image.
  ushort write_frameGray2(grayscale_image * image, long frame_format, 
                         vColorMap * colormap);

  vRLE_Image * ReadRLEFrame(long number);
  vRLE_Image * ReadNextRLEFrame();
  ushort WriteRLEFrame(vRLE_Image * image);

  inline ushort SetFormatGray()
  {
    format = 1;
    return 1;
  }

  // Converts AVIs to RLE sequences. The format specifies if each frame
  // should be converted to a binary or grayscale image (support for
  // color images is not there yet, I don't know if I will ever work on that).
	static void AVI_To_RLE(const char * source, const char * target, long format);
	static void RLE_To_AVI(const char * target, const char * source, 
												 unsigned short color_mode);
  virtual long BytesPerPixel()
  {
    if (format == 2) return 3;
    else return 1;
  }
};



#endif // VASSILIS_RLE_ENCODING_H

