#ifndef VASSILIS_ALGO_TEMPLATES_H
#define VASSILIS_ALGO_TEMPLATES_H

#include <vector>

#include "basics/matrix.h"
#include "basics/image.h"
#include "vplatform.h"
#include "algorithms.h"
#include "basics/drawing_temp.h"

template<class type>
class vMatrix;

class vSkinTableBase : public vUnique
{
public:
  virtual ushort Write(const char * filename1, const char * filename2) = 0;
  virtual grayscale_image * DetectSkin(color_image * the_image) = 0;
};


template<class type>
class vSkinTable : public vSkinTableBase
{
private: 
  long dim_size;
  v3dMatrix<type> * skin_data;
  v3dMatrix<type> * non_skin_data;

protected:
  virtual void DeleteUnique();

public:
  vSkinTable(const char * filename1, const char * filename2);
  vSkinTable(const char * filename, const char * filename2, long dim_size);
  ~vSkinTable();

  // Writes data so that they can be read by the first constructor.
  ushort Write(const char * filename1, const char * filename2);
  grayscale_image * DetectSkin(color_image * the_image);

  // For debugging, return the value stored at the specified bin.
  type ReadSkinValue(long r_index, long g_index, long b_index);
  type ReadNonSkinValue(long r_index, long g_index, long b_index);
};


// The only use of this class for the time being is in calculating
// motion statistics for a sequence of images (i.e. means and
// covariances of changes of intensity for each pixel).

// This class is used to estimate and store the mean and standard
// deviation of each pixel in a nuber of images. One use of it is to
// build a model of the background that can be used for background
// subtraction. 
template<class type>
class vSequenceStatistics
{
private:
  // Size is the number of pixels in the image.
  long rows, cols, bands, size;
  
  // The sums of values of the pixels in the set of images.
  v3dMatrix<double> * totals;

  // The sum of squared differences of values from means.
  v3dMatrix<double> * total_deviations;

  // The mean values for all pixels.
  v3dMatrix<double> * means;

  // The standard deviations for all pixels.
  v3dMatrix<double> * deviations;

  // True if the values in means are valid (meaning that we haven't
  // called ProcessValues since the last time we called ComputeMeans();
  ushort valid_means;

  // True if the values in deviations are valid (meaning that we
  // haven't called ProcessValues or ProcessDeviations since the last
  // time we called ComputeDeviations()).
  ushort valid_deviations;

  // Number of images whose values are summed up in totals.
  long values_processed;

  // Number of images whose deviations are summed up in total_deviations.
  long deviations_processed;

  // Computes the means by dividing each value in totals by values_processed.
  void ComputeMeans();

  // Computes the deviations by dividing each value in
  // total_deviations by deviations processed minus 1.
  void ComputeDeviations();

public:
  // Constructs an vSequenceStatistics object that can handle images with
  // the specified number of rows, columns, and bands.
  vSequenceStatistics(long in_rows, long in_cols, long in_bands);
  ~vSequenceStatistics();
  
  // Adds the values of the pixels in color_image to the values in totals.
  void ProcessValues(v3dMatrix<type> * the_image);

  // Takes the squared difference betweent the image and the background and
  // adds it to total_deviations.
  void ProcessDeviations(v3dMatrix<type> * the_image);

  // Makes totals and total_deviations zero. Basically, it erases all
  // information processed so far.
  void Reset();

  // Returns the means.
  v3dMatrix<double> * GetMeans();

  // Returns the standard deviation.
  v3dMatrix<double> * GetDeviations();
  long Rows();
  long Cols();
  long Bands();
  long Size();
};

template<class type>
color_image * vFindMaxSpots(v3dMatrix<type> * source, long number);

template<class type>
vSearchGrid * vMakeColorGrid(color_image * source, v3dMatrix<type> * labels,
			      Label & label);

template<class type>
vSearchGrid * vMakeColorGrid2(color_image * source, v3dMatrix<type> * labels,
			       Label & label);



// vSaveMatrixAsUshorts can be used to save disk space when we write a 
// matrix to disk. It is applicable under the following assumptions
// - Suppose low is the min value in source
// - Suppose high is the max value in source
// - Suppose USHORT_MAX is the maximum value attainable by a ushort.
// - Assume that it doesn't make a noticeable difference in our domain 
//   if we replace each value in the matrix with its nearest
//   neighbor in a set of USHORT_MAX uniformly spaced points between low 
//   and high.
// Then, we can save to disk low and high, and for each entry in the matrix
// simply save the index of its nearest neighbor in the set of USHORT_MAX
// points. This information is enough to recover an approximation of the 
// matrix.
template<class type>
long vSaveMatrixAsUshorts(v3dMatrix<type> * source, const char * filename);


// vLoadMatrixAsUshorts returns an approximation of a matrix saved as 
// filename by vSaveMatrixAsUshorts.
template<class type>
v3dMatrix<type> * vLoadMatrixAsUshorts(const char * filename);


// Finds the mean color robustly of the region in color_image that is
// specified with non-zero values in region.
// Stores the mean color in result (the
// result will have as many values as the bands of color_image).
// The result is not necessarily the arithmetic mean of the colors
// in the region. We do some iterations of going through the region,
// finding the mean, rejecting outliers, recalculating the mean,
// re-rejecting outliers, and so on.
template<class type>
long vRegionRobustMean(v3dMatrix<type> * the_image, 
                        v3dMatrix<uchar> * region, 
                        vector<double> * result);

// Returns a binary image, whose non-zero values specify a 
// region in color_image that is a subset of the non-zero pixels
// in candidates, and whose color in color_image is within the
// specfied distance from the color argument.
template<class type>
grayscale_image * vSelectColorRange(v3dMatrix<type> * the_image,
                                 v3dMatrix<uchar> * candidates,
                                 vector<double> * color, double distance);


// This is like vSelectColorRange, but addresses the problem of selecting
// pixels whose colors are much closer to color than to other_color.
// "much closer" is defined as follows: If a pixel's color has a distance
// d1 to color and d2 to other_color, if d2 - d1 > thr, where
// thr is the distance from color to other color minus 2*distance,
// the pixel's color is "much closer" to color than it is to other_color.
template<class type>
grayscale_image * vSelectColorRange5(v3dMatrix<type> * the_image,
                                  v3dMatrix<uchar> * candidates,
                                  vector<double> * color, 
                                  vector<double> * other_color, double distance);


template<class type>
v3dMatrix<type> * normalize_enclosing_circle(v3dMatrix<type> * image, long radius);

template<class type>
v3dMatrix<type> * normalize_enclosing_circle4(v3dMatrix<type> * image, long radius,
                                             long center_row, long center_col);



#ifndef VASSILIS_SGI_PLATFORM
#define VASSILIS_PROCESS_FILE
#include "algo_templates.cpp"
#endif // VASSILIS_SGI_PLATFORM




#endif // VASSILIS_ALGO_TEMPLATES_H

