#ifndef VASSILIS_ALGORITHMS_H
#define VASSILIS_ALGORITHMS_H

#include <deque>
#include <vector>
#include <math.h>


// I had problems when I first included "detect.h". I had to go to Project->Settings
// and specify that it should ignore the library LIBC.LIB
extern "C"
{
#include "basics/detect.h"
}

#include "basics/image.h"
#include "basics/video_common.h"
#include "basics/rle_encoding.h"
#include "basics/my_video_format.h"
#include "basics/algebra.h"

#ifdef VASSILIS_WINDOWS
using std::deque;
using std::sort;
using std::vector;
#endif

#ifdef VASSILIS_SGI_PLATFORM
using namespace std;
#endif



class vLineSegment
{
public:
  double row1, col1, row2, col2;

  inline double CenterRow()
  {
    return (row1 + row2) / 2.0;
  }

  inline double CenterCol()
  {
    return (col1 + col2) / 2.0;
  }

  inline double Orientation()
  {
    return vAngle(col2 - col1, row2 - row1);
  }

  inline double OrientationDegrees()
  {
    static const double factor = 180.0 / M_PI;
    return factor * vAngle(col2 - col1, row2 - row1);
  }

  inline double Length()
  {
    double drow = row1 - row2;
    double dcol = col1 - col2;
    return sqrt(drow * drow + dcol * dcol);
  }
};

class vLineSegmentLong
{
public:
  long row1, col1, row2, col2;
  vLineSegmentLong()
  {
    row1 = -1;
  }

  vLineSegmentLong(vLineSegment & input)
  {
    row1 = vRound(input.row1);
    col1 = vRound(input.col1);
    row2 = vRound(input.row2);
    col2 = vRound(input.col2);
  }

  inline double CenterRow()
  {
    return (row1 + row2) / 2.0;
  }

  inline double CenterCol()
  {
    return (col1 + col2) / 2.0;
  }

  inline double Orientation()
  {
    return vAngle(col2 - col1, row2 - row1);
  }

  inline double OrientationDegrees()
  {
    static const double factor = 180.0 / M_PI;
    return factor * vAngle(col2 - col1, row2 - row1);
  }

  inline double Length()
  {
    double drow = row1 - row2;
    double dcol = col1 - col2;
    return sqrt(drow * drow + dcol * dcol);
  }
};




// Below this line starts the BreadthFirstSearch implementation.


class vNodeSet
{
public:
  virtual ushort IsEmpty() = 0;
  virtual void * Pop() = 0;
};

typedef void (*BreadthFirstProc) (void * node, vNodeSet * to_visit, 
                                  vector<void *> * procedure_args);
			    
void BreadthFirstSearch(vNodeSet & to_visit, BreadthFirstProc procedure, 
			vector<void *> procedure_args);




class vNodeList : public vNodeSet, public std::deque<void *>
{
public:
  vNodeList();
  ~vNodeList();
  ushort IsEmpty();
  void * Pop();
};

// This class is used for the following task: Suppose we are working
// in a 2-dimensional space and we want to do breadth-first search on it.
// The "distances" array of this class is where we would do write the
// results of that search.
//
// The private variables distance_array and temp_array are supersets of 
// distances. The reason we need them is to write special values at the borders
// to keep the search from trying to access elements outside the space.
// We want extra lines at the top, bottom, left and right of the image for those
// values. That extra padding is accessed using distance_array. distances[0][0]
// is distance_array[1][1].
//
// One place where I used this class is hand detection. 
// The grid represents the rg color space (r = R / I, g = G / I), 
// and every entry is the similarity
// of the corresponding rg color to the color of the face. That similarity
// can be defined in different ways, it is the task of the function that
// builds the grid to define it.
class vSearchGrid : public v3dMatrix<long>
{
private:
  vArray2(long) distance_array;
  vArray2(long) temp_array;

public:
  // Note that rows and cols of the general_image superclass represent the
  // rows and cols of distance_array, not of distances.
  long grid_rows;
  long grid_cols;
  vArray2(long) distances;
  vSearchGrid(long in_rows, long in_cols);
  ~vSearchGrid();

  // Returns the number of pixels in distances where the two grids
  // differ. It only looks into the distances array, it doesn't look
  // at the padding (which is accessible via distance_array);
  static long Compare(vSearchGrid * grid1, vSearchGrid * grid2);

  // Writes value to all border entries
  void WriteBorder(long value);

  // Writes value to all interior entries
  void WriteInterior(long value);
  
  // Makes every cell of the grid contain the block metric distance
  // of that cell to the center that is specified here.
  void SetCenter(ushort r, ushort g);
};


class vSearchGridDouble : public v3dMatrix<double>
{
private:
  vArray2(double) distance_array;
  vArray2(double) temp_array;

public:
  // Note that rows and cols of the general_image superclass represent the
  // rows and cols of distance_array, not of distances.
  long grid_rows;
  long grid_cols;
  vArray2(double) distances;
  vSearchGridDouble(long in_rows, long in_cols);
  ~vSearchGridDouble();

  // Returns the number of pixels in distances where the two grids
  // differ. It only looks into the distances array, it doesn't look
  // at the padding (which is accessible via distance_array);
  static long Compare(vSearchGridDouble * grid1, vSearchGridDouble * grid2);

  // Writes value to all border entries
  void WriteBorder(double value);

  // Writes value to all interior entries
  void WriteInterior(double value);
  
  // Makes every cell of the grid contain the block metric distance
  // of that cell to the center that is specified here.
  //void SetCenter(ushort r, ushort g);
};


void GridBreadthFirstProc(void * nodep, vNodeSet & to_visit_set,
			  vector<void *> procedure_args);


v3dMatrix<long> * vGridDistances(color_image * the_image, vSearchGrid * grid);

void vAverageLabelColor(color_image * source, v3dMatrix<long> * labels,
			 Label & label, uchar & r, uchar & g, uchar & b);

v3dMatrix<long> * vColorDistances(color_image * the_image, 
				     uchar r, uchar g, uchar b);

long vFindFaces(color_image * the_image, FaceLocation ** locations);

void vReadFaceLocation(vBoundingBox * box, FaceLocation & loc);



// This function just loads up a specific image file where we 
// encode the adjacency list for an articulated hand, as 
// defined in hand_renderer.cpp. By saving that list as an 
// image file and loading it up here, we avoid the need of 
// having to include all the VirtualHand related files and  
// libraries in a project. This is even more important if we
// consider the fact that the VirtualHand library only works
// on windows, and needs to be installed.

v3dMatrix<ushort> * vLoadFingerAdjacencies();


// vDepthNeighbors is very similar to vMaxNeighbors (applied to
// depth), but it assumes that the pixels belong to specific links,
// as specified by the links argument. For background pixels we write
// 0 to the result. For a pixel in a link A, we consider only neighbors
// belonging also to link A.
v3dMatrix<float> * 
vDepthNeighbors(v3dMatrix<uchar> * links, v3dMatrix<float> * depth,
                 v3dMatrix<ushort> * adjacencies);


// See description of the function in the code_guide directory.
grayscale_image * 
vArticulatedEdges(v3dMatrix<uchar> * object, v3dMatrix<float> * depth,
                   v3dMatrix<ushort> * adjacencies);

// This is a simpler version of vArticulatedEdges, where we don't consider
// depth information at all.
grayscale_image * 
vSimpleArticulatedEdges(v3dMatrix<uchar> * object, 
                         v3dMatrix<ushort> * adjacencies);


// Applies vArticulatedEdges to every frame of links and depths, and
// saves the results to the RLE sequence named by output.
ushort vSequenceArticulatedEdges(vVideoFile * links, class_image_stream * depths,
                                  const char * output);



void vDistanceTransformProcSlow(void * node, vNodeSet * to_visit, 
                                 vector<void *> * procedure_args);


void vDistanceTransformProcFast(void * node, vNodeSet * to_visit, 
                                 vector<void *> * procedure_args);

v3dMatrix<long> * vDistanceTransformSlow(general_image * the_image_base);

v3dMatrix<long> * vDistanceTransform(general_image * the_image_base);

double vSetDistance(vector<long> * rows, vector<long> * cols, 
                     v3dMatrix<float> * distances);

ushort vTranslatePoints(vector<long> * source_rows, vector<long> * source_cols,
                         vector<long> * target_rows, vector<long> * target_cols,
                         long translation_rows, long translation_cols);

ushort vScalePoints(vector<long> * source_rowsp, vector<long> * source_colsp,
                     vector<long> * target_rowsp, vector<long> * target_colsp,
                     double scaling, long center_row, long center_col);



// This function assumes that x and y describe a function from N/x.size()
// to R2. It computes the centered derivative at each point. The last
// point is assumed to be adjacent to the first. The derivative is not computed
// with respect to the next and previous point, but with respect to the points
// whose distance (in terms of order of appearance in x and y) is specified
// by neighborhood. This way small local fluctuations due to discretization
// are ignored.
short vCircularDerivative(std::vector<double> * x, std::vector<double> * y,
                           std::vector<double> * dx, std::vector<double> * dy,
                           long neighborhood);

// At the end of the function, (*out_norm)[i] = norm((*x)[i], (*y)[i]).
short vMagnitudes(std::vector<double> * x, std::vector<double> * y,
                   std::vector<double> * out_norm);

// At the end of the function, (*orientations)[i] is the angle (in radians)
// between the vector specified by (*x)[i] and (*y)[i] and the vector (1, 0)
// (coordinate notation is (x, y)).
short vOrientations(std::vector<double> * x, std::vector<double> * y,
                     std::vector<double> * orientations);


// vLocalMaxima takes in a vector of values, and stores the values of 
// the local maxima of the input into maxima, and the corresponding
// indices for those maxima into indices.
short vLocalMaxima(std::vector<double> * input, 
                    std::vector<double> * maxima, 
                    std::vector<long> * indices);


// vSegmentOverlap returns the number of points 
// belonging to both integer sequences
// (start1, start1+1, ..., end1) and (start2, start2+1, ... end2)
inline long vSegmentOverlap(long start1, long end1, long start2, long end2)
{
  long max_start = Max(start1, start2);
  long min_end = Min(end1, end2);
  long overlap = min_end - max_start + 1;
  if (overlap <= 0) return 0;
  else return overlap;
}


// For the time being, this function assumes that we have
// a periodic space, between 0 and period - 1. start1,
// end1, start2, end2, are in that space. To find the overlap
// between two segments, we need to add period to the ends when
// the ends are numerically smaller than their respective starts.
// After that, calculating the overlap is trivial.
long vCircularSegmentOverlap(long start1, long end1, long start2, long end2,
                              long period);


// Note of 6/17/2002: I should change all these precalculate functions, because
// the way I have it right now, any function that calls them gets a new 
// vMatrix. I should have, for each function, one unique global matrix,
// and when we ask for a higher limit that matrix should change to include
// the calculations for the higher limit. In general, the calculations should
// only be done once and stored at one place.

// The length of the result is square_limit, and
// the i-th entry in result will be the square root of i.
vMatrix<long> * vPrecalculateRoots(long square_limit);

// The length of the result is square_limit, and
// the i-th entry in result will be the square root of i.
vMatrix<double> vPrecalculateRoots2(long square_limit);

// The length of the result is limit, and
// the i-th entry in result will be the log10 of i.
vMatrix<double> vPrecalculateLogs2(long limit);

// The length of the result is limit, and
// the i-th entry in result will be the log base 10 of i.
// The difference between this function and vPrecalculateLogs2
// is that here the function keeps track of the highest
// argument with which it was ever called, and the value
// it returned for that argument. As long as the current
// argument is not larger than the highest argument ever, 
// the value returned for the highest argument is also returned
// here. 
vMatrix<double> vPrecalculateLogs3(long limit);

////////////////////////////////////////////////////////////////
// 
// Start of implementation of class vPCA
//
////////////////////////////////////////////////////////////////


// Class vPCA is the class where appropriate procedures related to
// PCA are implemented. Basically, PCA estimates the mean, eigenvectors,
// and eigenvalues of a set of vectors, and uses them to change bases
// and compute distances.
//
// For the current needs of the system I don't have a reason to implement
// the PCA estimation algorithm, since I can use Matlab to do that. This
// class is initialized by reading results from a textfile.
class vPCA
{
private:
  vMatrix<double> eigenvectors;
  vMatrix<double> eigenvalues;
  vMatrix<double> eigenvalue_roots;
  vMatrix<double> mean;

  // The dimensions of the original vectors, before PCA was done
  long original_dimensions;

  // The number of eigenvectors and eigenvalues that are being used
  // (this can be changed at any moment to any value up to 
  // original_dimensions
  long used_dimensions;
  long is_valid;

  long Initialize();
  long CleanUp();
  long LoadBinary(const char * filename);
  long LoadText(const char * filename);
  long SaveBinary(const char * filename);
  long SaveText(const char * filename);
  long SetEigenvalueRoots();

public:
  vPCA(const char * filename);
  ~vPCA();

  inline vMatrix<double> Eigenvalues()
  {
    return eigenvalues;
  }

  // Converts a column vector v to a new vector in the PCA space.
  // result = eigenvectors' * (v - mean)
  vMatrix<double> PCA_Conversion(vMatrix<double> & v);

  // Calculates the Mahalanobis distance between two vectors.
  // The input vectors are expressed in the PCA basis.
  inline double Mahalanobis(vMatrix<double> & v1, vMatrix<double> & v2)
  {
    if (v1.Size() < used_dimensions) return -1;
    if (v2.Size() < used_dimensions) return -1;
    double result = 0;
    for (long i = 0; i < used_dimensions; i++)
    {
      double diff = v1(i) - v2(i);
      result = result + diff * diff / eigenvalues(i);
    }

    return result;
  }


  long LoadFile(const char * filename);

  long Save(const char * filename);

  inline long IsValid()
  {
    return is_valid;
  }

  inline long OriginalDimensions()
  {
    return original_dimensions;
  }

  inline long UsedDimensions()
  {
    return used_dimensions;
  }

  inline long SetUsedDimensions(long new_used_dimensions)
  {
    if (new_used_dimensions < 1) return 0;
    if (new_used_dimensions > original_dimensions) return 0;
    used_dimensions = new_used_dimensions;
    return 1;
  }
};


////////////////////////////////////////////////////////////////
// 
// End of implementation of class vPCA
//
////////////////////////////////////////////////////////////////

// Class vEdgeHistogram

class vEdgeHistogram : public vMatrix<double>
{
private:
  // Difference between the start and end angles of the interval
  // corresponding to a bin.
  double bin_step;

  long Normalize();
public:
  vEdgeHistogram(long bins);
  vEdgeHistogram(long bins, grayscale_image * original, grayscale_image * edges);
  ~vEdgeHistogram();

  inline long NumberOfBins()
  {
    return cols;
  }

  // Maps from angles (measured in radians) to bins
  inline long BinIndex(double angle)
  {
    long index = (long) (angle / bin_step);
    return index % cols;
  }


  // Reads in original images from the video sequence in original_file
  // and edge images from edge_file. For each frame in original_file,
  // it reads the corresponding edge image from edge_file, calculates
  // the edge histogram and saves it to target.
  static long SaveHistograms(long bins, const char * original_file, 
                             const char * edge_file, const char * target);

  // Each row in the result is a histogram corresponding to one frame.
  static vMatrix<double> * LoadHistograms(const char * filename);

  double MinCumulativeDistance(vEdgeHistogram * histogram2);
  double MeanCumulativeDistance(vEdgeHistogram * histogram2);

  // The distance functions that are static generate 48 different 
  // histograms for each
  // histogram saved in training. Those 48 histograms correspond
  // to different image plane rotations. The order in which they are
  // generated corresponds to the order with which training views get saved
  // by the program save_views. They return a matrix that contains the 
  // distance between the test histogram and each of the generated
  // training histograms (the distance is defined differently in each
  // function).

  static vMatrix<double> MinCumulativeDistance2(vMatrix<double> * training,
                                                 vMatrix<double> * test);

  static vMatrix<double> MeanCumulativeDistance2(vMatrix<double> * training,
                                                  vMatrix<double> * test);
};


// Returns a 1-dimensional matrix with the specified cols,
// whose entries are random.
vMatrix<ulong> vRandomVector(long cols);


// in_rows and in_cols are arrays of length number. out_rows and out_cols
// are arrays with the same length. image_rows, image_cols are the 
// dimensions of the image from which point coordinates have been put to
// in_rows and in_cols. Those point coordinates are rotated by 90, 180,
// 270 or 0 degrees clockwise and the results are put in out_rows and 
// out_cols. The caller can pass the same arrays for in_rows and out_rows,
// or in_cols and out_cols.
long vRotatePointsRight(vArray(long) in_rows, vArray(long) in_cols, 
                         vArray(long) out_rows, vArray(long) out_cols, 
                         long number, long image_rows, long image_cols, 
                         long degress);


// vRefineHandSegmentation should be called whenever we have a clear
// idea of where the hand is in the image, and we have identified
// its connected component as the pixels that are set to 1 in
// first_binary. This function fills up holes of the connected component,
// and makes sure all hand pixels have a value of at least 2 in each
// of their R, G, and B color values (so that they don't get confused
// with their background).
color_image * vRefineHandSegmentation(v3dMatrix<uchar> * first_binary,
                                   color_image * source);



// vSaveEdgeStatistics calculates and saves to a file the following:
// Suppose original_file stores a video sequence, and edge_file
// stores the edge images of the frames of original_file. Then, the
// question is, given an edge pixel A with edge orientation E1, how likely
// is it that there will be an edge pixel B with edge orientation E2
// at location (X,Y). This question can be answered by building a histogram
// where bins are indexed by E1, X, Y, E2, and where each bin contains the
// number of times that the combination E1, X, Y, E2 was found in the 
// sequence.
//
// spatial_subsampling is N if we want each bin to correspond to a NxN
// pixel area (X and Y indices).
// pixel1_bins is the number of discrete values into which values of E1
// will be converted.
// pixel2_bins is the number of discrete values into which values of E2
// will be converted. It is also the number of possible indices 
// corresponding to E2 in the histogram.
// [pixel1_start, ..., pixel1_end] specifies the discrete values of E1
// that will be used as indices in the histogram. If they are set
// to -1, they will acquire default values of 0 and pixel1_bins-1 
// respectively. Sometimes it is enough to keep track of only one of
// all possible discrete values of E1, because then we can use the 
// histogram to figure out the rest of the values. For example, my
// hand database contains 48 image plane rotations for each viewpoint,
// and if we keep edge statistics only for cases where E1 is 0,
// we can calculate based on them the statistics for all other values
// of E1.
long vSaveEdgeStatistics(const char * original_file, const char * edge_file,
                          const char * target_file,
                          long spatial_subsampling, 
                          long pixel1_bins, long pixel2_bins,
                          long pixel1_start, long pixel1_end,
                          long frames_to_process = -1);


// vEdgeStatistics is a class that can store the information calculated
// by vSaveEdgeStatistics
class vEdgeStatistics
{
private:
  v4dArray<double> * array;
  vArray4(double) bins;

  // Frame size of the original video sequence from which the statistics
  // are calculated:
  long image_rows;
  long image_cols;

  long spatial_subsampling;
  long pixel1_bins;
  long pixel2_bins;
  long pixel1_start;
  long pixel1_end;
  
  double bin_step1, bin_step2;
  long dim2_low, dim2_high, dim3_low, dim3_high;

public:
  vEdgeStatistics(long in_image_rows, long in_image_cols,
                   long in_spatial_subsampling, long in_pixel1_bins,
                   long in_pixel2_bins, long in_pixel1_start,
                   long in_pixel1_end);

  ~vEdgeStatistics();

  long ImageRows();
  long ImageCols();
  long SpatialSubsampling();
  long Pixel1Bins();
  long Pixel2Bins();
  long Pixel1Start();
  long Pixel1End();

  long Zero();
  long Save(const char * filename);
  static vEdgeStatistics * Load(const char * filename);
  
  // Add votes to the appropriate bins using the orientation and
  // edge information stored in orientations and edges respectively.
  long ProcessImage(v3dMatrix<double> * orientations, grayscale_image * edges);

  // For each edge pixel in edges get the number of votes 
  // it gets from the edge statistics histogram. The orientation
  // of each edge pixel is stored in orientations.
  vMatrix<double> * Vote(v3dMatrix<double> * orientations, grayscale_image * edges);

  // For each edge pixel in edges get the number of votes 
  // it gets from the edge statistics histogram. The orientation
  // of each edge pixel is stored in orientations. The difference
  // betweent this function and Vote() is that this function is
  // optimized for the database of hand images. In the 
  // edge statistics for that database,
  // the histogram only contains information for e1 = 0 
  // (horizontal orientation). If the orientation of pixel1
  // is e1 and e1 is not 0, then to obtain statistics for that pixel
  // we need to rotate the image so that e1 becomes 0. This can be done
  // by rotating, for every point2, its relative coordinates by an
  // angle of -e1, and by rotating the orientation e2 of point2 also
  // by an angle of -e1
  vMatrix<double> * Vote2(v3dMatrix<double> * orientation_image, 
                           grayscale_image * edge_image);

  // For each pixel get the number of votes it gets from the edge 
  // statistics histogram. The difference between this function
  // and Vote2 is that here we get votes for every pixel in the
  // image, whereas in Vote2 we get votes for every edge pixel.
  vMatrix<double> * Vote4(v3dMatrix<double> * orientation_image, 
                           grayscale_image * edge_image);

  // Bin corresponding to orientation, if this orientation belongs
  // to pixel1
  inline long Bin1(double orientation)
  {
    return ((long) (orientation / bin_step1)) % pixel1_bins;
  }

  // Bin corresponding to orientation, if this orientation belongs
  // to pixel2
  inline long Bin2(double orientation)
  {
    return ((long) (orientation / bin_step2)) % pixel2_bins;
  }

  // Operator () is used for read-only access
  inline double operator () (double e1, long row, long col, double e2)
  {
    long i1 = Bin1(e1);
    long i2 = row / spatial_subsampling;
    long i3 = col / spatial_subsampling;
    long i4 = Bin2(e2);
    if(! ((i1 >= pixel1_start) && (i1 <= pixel1_end) &&
          (i2 >= dim2_low) && (i2 <= dim2_high) &&
          (i3 >= dim3_low) && (i3 <= dim3_high)) )
    {
      return 0;
    }
    return bins[i1][i2][i3][i4];    
  }

  inline double operator () (long e1, long row, long col, long e2)
  {
    long i2 = row / spatial_subsampling;
    long i3 = col / spatial_subsampling;
    assert(e1 >= pixel1_start);
    assert(e1 <= pixel1_end);
    assert(i2 >= dim2_low);
    assert(i2 <= dim2_high);
    assert(i3 >= dim3_low);
    assert(i3 <= dim3_high);
    assert(e2 >= 0);
    assert(e2 < pixel2_bins);

    if (! ((e1 >= pixel1_start) && (e1 <= pixel1_end) &&
           (i2 >= dim2_low) && (i2 <= dim2_high) &&
           (i3 >= dim3_low) && (i3 <= dim3_high) &&
           (e2 >= 0) && (e2 < pixel2_bins)) )
    {
      return 0;
    }
    return bins[e1][i2][i3][e2];
  }

  // Function Data is used for write access to the data
  inline double & Data (double e1, long row, long col, double e2)
  {
    long i1 = Bin1(e1);
    long i2 = row / spatial_subsampling;
    long i3 = col / spatial_subsampling;
    long i4 = Bin2(e2);
    assert(i1 >= pixel1_start);
    assert(i1 <= pixel1_end);
    assert(i2 >= dim2_low);
    assert(i2 <= dim2_high);
    assert(i3 >= dim3_low);
    assert(i3 <= dim3_high);
    return bins[i1][i2][i3][i4];    
  }

  inline double & Data (long e1, long row, long col, long e2)
  {
    long i2 = row / spatial_subsampling;
    long i3 = col / spatial_subsampling;
    assert(e1 >= pixel1_start);
    assert(e1 <= pixel1_end);
    assert(i2 >= dim2_low);
    assert(i2 <= dim2_high);
    assert(i3 >= dim3_low);
    assert(i3 <= dim3_high);
    assert(e2 >= 0);
    assert(e2 < pixel2_bins);
    return bins[e1][i2][i3][e2];
  }

};


// It assumes that the input is a gray-scale binary image (pixels
// have values 0 or 255). It returns
// a binary image whose pixels have the value 255 for the pixels 
// belonging to the largest foreground connected component in binary
// (i.e. the largest connected component whose pixels have the 
// value 255).
grayscale_image * vLargestForegroundComponent(v3dMatrix<uchar> * binary);


// This function assumes that the three arguments are binary images,
// with pixel values either 0 or 255. foreground1 and foreground2
// are two different estimates for the foreground of some image.
// (their pixels are non-zero if they are estimated to be at the 
// foreground). This function returns a binary image that reconciles
// the two estimates, by deciding the fate of pixels for which the 
// two estimates differ. Those pixels are assigned in a way that 
// sets the foreground boundary as close as possible to edge pixels.
grayscale_image *  vPushBoundaryToEdges(v3dMatrix<uchar> * foreground1, 
                                     v3dMatrix<uchar> * foreground2, 
                                     v3dMatrix<uchar> * edges);

// Trim from the foreground all pixels within distance length 
// of the boundary contour. We assume that foreground is a binary
// image (values are 0 or 255).
grayscale_image * vTrimBoundary(v3dMatrix<uchar> * foreground, long length);


// A heuristic function used to find shadows. A pixel's shadow score
// is determined as following: Start from that pixel, and move in
// a straight line towards the direction of greatest decrease in
// intensity. How far can we go in that direction until we actually
// find another increase in intensity? Typically shadows progress smoothly
// from light to dark, and this function should give high scores to 
// pixels where that smooth transition starts.
v3dMatrix<long> * vShadowScore(v3dMatrix<double> * the_image);

// vLocalShadowScores was written for debugging, and so that 
// I could visualize more easily how vShadowScore works. Basically
// it only computes the shadow score for pixel (row, col) and it
// also traces the pixels that it reaches from (row, col) as
// it computes the score.
v3dMatrix<long> * vLocalShadowScore(v3dMatrix<double> * the_image, 
                                       long row, long col);

// Returns the angle (in radians) of the principal orientation of the 
// shape in binary (i.e. of the non-zero pixels). The location of the
// center of mass of the shape is placed into mean_row and mean_col.
double vPrincipalOrientation(grayscale_image * binary, 
                              double * mean_row, double * mean_col);


// Suppose we have an image of size rows x cols, and a line
// that goes through (row, col) whose orientation is specified
// by angle. Where will that line intersect the boundary of the
// image? This function computes the answer and puts the 
// two intersection points into (row1_p, col1_p) and
// (row2_p, col2_p). The function returns 1 if successful. It
// returns 0 when it fails. It fails when (row, col) is not
// inside the image.
long vBoundaryPoints(double row, double col, double angle, 
                      long rows, long cols, 
                      double * row1_p, double * col1_p,
                      double * row2_p, double * col2_p);


// This is an interactive function that takes in row, col, angle,
// rows and cols from the user, and it calls vBoundaryPoints.
// This way we can easily check that vBoundaryPoints is correct.
// The testing stops when the user enters -1 for row.
long vTestBoundaryPoints();

                  
// Finds heuristically the cutoff point where the palm is separated
// from the rest of the arm. The heuristic algorithm is the following:
// - Find the principal axis of the shape
// - For each point P of the axis, count Proj(P), the number of pixels
//   from the shape that project on that axis
// - Find the point W of max width, i.e. the point W on the axis where
//   Proj(W) is maximized
// - Starting at W and going away from the center of the image, find
//   C, the point of least width, i.e. the point where Proj(C) is
//   minimized.
// - Draw a line perpendicilar to the principal axis that goes through
//   C, and find the first and last pixel of the shape that belong
//   to that line.
// - We return the locations of those two pixels. Basically, everything
//   on one side of the line connecting those pixels (the side that
//   is towards the center of the image) is the palm, everything else
//   is rejected.
// 
// We also store in hand_pixel_row and hand_pixel_col the coordinates
// of a point that is on the palm side of the cutoff line, so that
// the caller of vPalmCutpoint knows which way the palm lies.
long vPalmCutpoint(grayscale_image * binary, long * row1, long * col1,
                    long * row2, long * col2, 
                    long * hand_pixel_row, long * hand_pixel_col);

// vKeepPalm uses vPalmCutpoint to figure out where the palm should
// be separated from the rest of the arm, and returns an image
// containing just the palm (zeros are put everywhere else).
grayscale_image * vKeepPalm(grayscale_image * binary);


// An alternative way to find the cutpoint.
long vPalmCutpoint2(grayscale_image * binary, long * row1, long * col1,
                    long * row2, long * col2, 
                    long * hand_pixel_row, long * hand_pixel_col);

grayscale_image * vKeepPalm2(grayscale_image * binary, vLineSegmentLong * cutoff);

grayscale_image * vKeepPalm3(grayscale_image * binary, v3dMatrix<float> * values,
                          vLineSegmentLong * cutoff);


// This function draws a line that goes through row and col and makes
// an angle with the x axis specified by orientation and counts how 
// many non-zero pixels of binary belong to that line. The boundary
// flag is set to 0 if no non-zero boundary pixels belong to that line,
// and it is set to 1 otherwise. The reason we care about the boundary
// flag is that we use vCountProjectedPoints as an auxiliary function
// in finding where we should separate the palm from the rest of the
// hand, and we don't want to consider cutoff lines at corners of the
// image, where the number of projected pixels is small simply because
// part of the arm is not visible in the image.
long vCountProjectedPoints(grayscale_image * binary, double row, double col,
                            double orientation, long * boundary_flag);

// like vCountProjectedPoints, but here we sum up the values, as opposed
// to just counting the points with non-zero values.
double vSumProjectedPoints(v3dMatrix<float> * values, double row, double col,
                            double orientation, long * boundary_flag);

// This function considers the line going through (row, col) that makes
// the specified angle with the x axis, and finds the most distant
// pair of contour points in binary that belong to that line.
long vLineContourIntersection(grayscale_image * binary, long row, long col,
                               double angle, long * row1, long * col1,
                               long * row2, long * col2);


// Pick out_points at random fron in_rows and in_cols and put them
// in out_rows and out_cols. The i-th input point has row (*in_rows)[i]
// and col (*in_cols)[i]. The same input point cannot be picked twice.
long vSamplePoints1(vector<long> * in_rows, vector<long> * in_cols, 
                    vector<long> * out_rows, vector<long> * out_cols,
                    long out_points);


long vSamplePoints2(vector<long> * in_rows, vector<long> * in_cols, 
                    vector<long> * out_rows, vector<long> * out_cols,
                    long sampling_rate);



// rows and cols contain coordinates of points. The i-th point
// has row = (*rows)[i] and col = (*cols)[i]. We want to save
// them in a file that is MATLAB-readable, as a matrix
// whose name is specified by the argument name, and whose
// i+1-th row will contain the row and col of the i-th point.
// This function was originally written to experiment with
// the shape context code I downloaded, which is in MATLAB.
long vSavePoints(vector<long> * rows, vector<long> * cols,
                  const char * name, const char * file);


// For each pixel, consider a line segment of given length and orientation
// centered at the pixel, and find the number of pixels on that segment
// for which distances in the dt_image (assumed to be the distance transform
// of an edge image) are below threshold. A simple (maybe naive) way to detect
// lines of given orientation. Fast version
v3dMatrix<long> * vOrientedLines(v3dMatrix<float> * dt_image, 
                                    double orientation, long length, 
                                    long threshold);

// For each pixel, consider a line segment of given length and orientation
// centered at the pixel, and find the number of pixels on that segment
// for which distances in the dt_image (assumed to be the distance transform
// of an edge image) are below threshold. A simple (maybe naive) way to detect
// lines of given orientation. Slow version
v3dMatrix<long> * vOrientedLines2(v3dMatrix<float> * dt_image, 
                                     double orientation, long length, 
                                     long threshold);

long vOrientedLines7(v3dMatrix<float> * dt_image, vector<long> * rows,
                      vector<long> * cols, v3dMatrix<long> * votes_matrix,
                      double orientation, long length, long threshold);

long vOrientedLines7b(grayscale_image * binary, vector<long> * rows,
                       vector<long> * cols, v3dMatrix<long> * votes_matrix,
                       double orientation, long length);

// For each pixel, consider a line segment of given length and orientation
// centered at the pixel, and find the number of pixels on that segment
// that are within the given distance (expressed by threshold) from an edge
// pixel whose orientation is within a certain tolerance of orientation
v3dMatrix<long> * vOrientedLines_c(grayscale_image * the_image, grayscale_image * edges,
                                      double orientation, long length, 
                                      long threshold, double tolerance);


double vShapeWidth(grayscale_image * the_image);

double vShapeLength(grayscale_image * the_image);

double vShapeOrientation(grayscale_image * the_image);

vLineSegment vProjectionEndpoints(grayscale_image * the_image, 
                                   long row, long col, double orientation);



class vEllipse
{
public:
  double center_row;
  double center_col;
  double width;
  double length;
  // major axis angle
  double angle;
};


// We have a line with given angle, going through x0, y0. We position
// ourselves at (x0, y0) and look towards the direction of the angle.
// Every other point (x,y) in the plane is either right in front of us
// (i.e. it belongs to the line), or is on the left half plane, or on 
// the right half plane. This function returns 0 if (x,y) is on the line,
// -1 if it is on the left, and 1 if it is on the right.
long vFindHalfPlane(double x0, double y0, double angle, double x, double y);

long vDoubleToLongVector(vector<long> * target, vector<double> * source);

long vLongToDoubleVector(vector<double> * target, vector<long> * source);

class vLineSegmentMatch
{
public:
  vLineSegmentLong segment;
  float orientation;
  float all_votes;
  float positive_votes;
  float length;
  float strength;
  float vote_density;
  long center_row, center_col;

  float score;
  long d, o, l, v, s;

  vLineSegmentMatch() {}
  long Print1();
};

// It has some more variables than vLineSegmentMatch
class vLineSegmentMatch2 : public vLineSegmentMatch
{
public:
  long step;
  double angle_difference;
};

class vLineSegmentMatchLess
{
public:
  ushort operator() (const vLineSegmentMatch & match1, const vLineSegmentMatch & match2)
  {
//    vPrint("score1 = %lf, score2 = %lf, (%lf, %lf), (%lf, %lf)\n",
//            match1.score, match2.score, 
//            match1.segment.row1, match1.segment.col1, match1.segment.row2, match1.segment.col2);
    return (match1.score < match2.score);
  }
};


class vLineSegmentSearchParams
{
public:
  long orientations;
  double low_length;
  double high_length;
  long lengths;
  double angle_tolerance;
  long distance_tolerance;

  vLineSegmentSearchParams(long in_orientations, double in_low_length, 
                            double in_high_length,
                            long in_lengths, double in_angle_tolerance, 
                            long in_distance_tolerance)
  {
    orientations = in_orientations;
    low_length = in_low_length;
    high_length = in_high_length;
    lengths = in_lengths;
    angle_tolerance = in_angle_tolerance;
    distance_tolerance = in_distance_tolerance;
  }

  inline long IsValid()
  {
    if (orientations < 2) return 0;
    if (low_length < 1) return 0;
    if (high_length < low_length) return 0;
    if (lengths < 2) return 0;
    if (angle_tolerance < 0) return 0;
    if (distance_tolerance < 0) return 0;

    return 1;
  }
};


class vPointerScore
{
public:
  void * pointer;
  double score;
  vPointerScore () {}
  vPointerScore (void * in_pointer, double in_score)
  {
    pointer = in_pointer;
    score = in_score;
  }
};


class vPointerScoreLess
{
public:
  ushort operator() (const vPointerScore & p1, const vPointerScore & p2)
  {
    return (p1.score < p2.score);
  }
};




long vLineSegmentMatches(grayscale_image * the_image, grayscale_image * edges, 
                          double angle, double length, 
                          double threshold, double tolerance, 
                          vector<vLineSegmentMatch2> * results);


long vOrientedLines_d(v3dMatrix<double> * orientations_image, 
                       grayscale_image * edges, 
                       vector<long> * edge_rows, vector<long> * edge_cols,
                       double orientation, vLineSegmentSearchParams params,
                       vector<vLineSegmentMatch> * results);

long vGetSearchRegion(double orientation, double high_length, 
                       double angle_tolerance, long distance_tolerance,
                       vector<long> * search_rows, vector<long> * search_cols, 
                       vector<long> * search_lengths);

v3dMatrix<double> * vTestOrientedLinesD(grayscale_image * the_image, grayscale_image * edges,
                                           double orientation, double length,
                                           long distance_tolerance, double angle_tolerance);

long vGetLineSegments(grayscale_image * the_image, grayscale_image * edges, 
                       vLineSegmentSearchParams params, 
                       vector<vLineSegmentMatch> * results);

long vSortLineSegments(vector<vLineSegmentMatch> * segments);

long vRankLineSegments(vector<vLineSegmentMatch> * segments, vLineSegment segment);

// For each segment in segments, compute its DOL difference to segment.
long vLineSegmentDifferences(vector<vLineSegmentMatch> * segments, 
                              vLineSegment segment);

// It produces two images that can help visualize the line segments in
// segments. One image (*ranks_p) shows the rank of the segment (i.e.
// the index of the segment in segments). The other includes information
// about a segment in various bands. That information is the rank, orientation,
// length, strength and all_votes.
// Note that we have many (even 1000) segments defined at each pixel. Only
// the highest ranking segment is used here.
// k specifies how many segments we should use (for example, k=10 means
// that we should only use the top 10 segments). A value of k=-1 means
// that we should use all line segments.
long vLineSegmentRankImage(vector<vLineSegmentMatch> * segments, long rows, long cols,
                            v3dMatrix<long> ** ranks_p, v3dMatrix<long> ** info_p,
                            long k);

long vScaleSearchParams(vLineSegmentSearchParams * params, vBoundingBox * box, 
                         long new_size);


long vNormalizeLineSegments(vector<vLineSegmentMatch> * segments, 
                             vBoundingBox * box, long new_size,
                             long new_center_row, long new_center_col);

// This class holds information about the principal axis of a region
// and the number of region pixels that project to each point on that 
// axis. A region is represented (among other ways) as a binary image
// where the region pixels have non-zero values and non-region pixels
// have zero value. Pixels project on a given axis point if they are 
// part of the line that is normal to the axis at that axis point.
class vAxisProjection
{
public:
  long number_of_points;
  double length;
  double orientation;
  double center_row, center_col;

  vArray(long) rows;
  vArray(long) cols;
  vArray(float) counts;

  vAxisProjection() {}

  // Figure out the principal axis of the region. Store in rows
  // and cols the pixels that belong to that axis, and in counts
  // the number of region pixels that project to the corresponding
  // axis point.
  vAxisProjection(grayscale_image * region);
  vAxisProjection(v3dMatrix<float> * region);
  ~vAxisProjection();

  // Return an image where, for each axis point, we draw a line
  // perpendicular to the axis, centered at that point, 
  // with length equal to the corresponding value in counts (the
  // number of region pixels that project onto that pixel).
  grayscale_image * DrawProjections();

  // Return the position at with at the axis point with the smallest 
  // width
  long MinWidth(long * indexp, long * rowp, long * colp, long * widthp);
  // Return the position at with at the axis point with the biggest 
  // width
  long MaxWidth(long * indexp, long * rowp, long * colp, long * widthp);

  // Returns a copy of this object.
  vAxisProjection * Copy();

  // Cuts the axis, keeping the region from startpoint to endpoint. 
  // Information about the rest of the points is discarded.
  long Cut(long start, long end);

  // The largest row and largest column occurring on the axis points.
  long MaxRowCol(long * rowp, long * colp);

  // Reverses the arrays rows, cols and counts.
  long Reverse();
};


// I have made an interface that lets me click twice to specify
// the palm cutoff line segment of a model image. To save time, 
// I have only labeled one out of every 12 frames (so, for each 
// set of 12 frames that are image-plane-rotated versions of each 
// other I have only labeled the first). The locations of the 
// cutoff segments of the remaining 11 frames can be derived by
// figuring out the appropriate translation and scale factor.
// This function does that, and returns the cutoff segments
// of all database views. Furthermore, for each frame of the database,
// it generates the cutoff segments that we get by rotating the 
// frame by 90, 180 and 270 degrees.
v3dMatrix<uchar> * vExtendCutoffs(v3dMatrix<uchar> * basic_cutoffs,
                                     vVideoFile * database);


// Returns the maximum distance between two points in the image. We
// assume that the image is connected and we can just consider
// contour points.
double vDiameter(grayscale_image * binary);


// Apply the transformation defined by t_row, t_col (for translation),
// scale and angle (for rotation) to the point (in_row, in_col). The
// results are put in out_rowp and out_colp. The two origins are 
// specified in (center_row1, center_col1) and (center_row2, center_col2);
// This is what we do:
// - Translate the point so that its center becomes (0,0) from 
// (center_row1, center_col1).
// - Rotate the point by angle
// - Scale the point by scale
// - Translate the point so that the center becomes (center_row2, center_col2)
// from (0,0)
long vWarpPoint(double in_row, double in_col, 
                 double * out_rowp, double * out_colp,
                 double center_row1, double center_col1,
                 double center_row2, double center_col2,
                 double scale, double angle);


vLineSegmentLong vNormalizeSegment(vLineSegmentLong * input, vBoundingBox * box,
                                     long size, long window_size);


// We want to sample points uniformly from the space [0, ..., max_value] so
// that we have "samples" points in the sample, and the max distance
// between any point in [0,...,max_value] and its closest sample point
// is minimuzed.
vArray(long) vUniformSpacing(long max_value, long samples);

// Here, unlike vUniformSpacing, we deal with the case where our
// space is circular, and max_value is the same as 0. This is,
// for example, the case when we are dealing with orientations, where 
// M_PI, or 2*M_PI (depending on the case) are identified with 0.
vArray(long) vCircularUniformSpacing(long max_value, long samples);


// We find the top-most, bottom-most, left-most and right-most coordinate
// values of non-zero pixels in input, and record them into the result's
// member variables. In a multi-band image, a pixel is non-zero if any
// of its bands is non-zero.
vBoundingBox vFindBoundingBox(vArray(long) rows, vArray(long) cols, 
                                long number);

vCircleLong vMinEnclosingCircle(general_image * the_image);

vCircleLong vMinEnclosingCircle(grayscale_image * the_image);

vCircleLong vMinEnclosingCircle3(vArray(long) rows, vArray(long) cols,
                                   long number);

// Returns the angle value, between 0 and 360 degrees, that is equivalent
// to in_degrees.
inline double vNormalizeDegrees(double in_degrees)
{
  long factor = (long) (in_degrees / 360);
  double result;
  if (in_degrees > 0)
  {
    result = in_degrees - factor * 360;
  }
  else
  {
    result = in_degrees - (factor-1) * 360;
  }
  if (result >= 360) result -= 360;
  else if (result < 0) result += 360;
  return result;
}


// Returns a binary image with the specified size, that contains
// "number" non-zero pixels, each of which is no farther than
// "radius" distance from the center of the image.
grayscale_image * vRandomEdgeImage(ulong rows, ulong cols, ulong unumber,
                                ulong radius);



// Returns a gray scale image that is the distance transform
// of color_image. It calls the OpenCV function, and converts
// the result from float to uchar. Values over 255 are
// cropped to 255.
grayscale_image * vDistanceTransformUchar(general_image * the_image);

// I'm writing this function just out of curiosity to see the 
// Voronoi diagram for a set of polygons. Each region in the 
// diagram is the set of points that are closer to a specific
// polygon than to any other polygon. The output is an image
// where the regions are colored randomly, and each polygon
// is drawn in black.
color_image * vLineSetVoronoi(vLineSet * line_set);

long vLineSetVoronoi2(vLineSet * line_set, color_image * the_image);


// This function is useful for getting a frame exported by Poser (using,
// for example, save1_main.py) and retaining a smaller portion of the frame,
// centered on the hand.
color_image * vPoserExtractHand(color_image * the_image, long result_size, long bbox_size);



#endif // VASSILIS_ALGORITHMS_H


