#include "stdafx.h"

#include "angle_sampler.h"
#include "basics/precomputed.h"


using namespace std;

static const double rad_factor = M_PI / 180;
static const double two_pi = M_PI * 2;

// Suppose we inscribe an n-gon into a circle of given radius.
// How long is each side of the n-gon? This is what this
// function computes
double vVertexDistance(long n, double radius)
{
  // The result is computed as follows: Suppose the unit circle is
  // centered at the origin, and the first vertex is at position
  // (1,0). Then, the second vertex is at (cos(f), sin(f)) where 
  // f = 360/n. The distance between the two vertices is the answer
  // for a circumference of 2*PI.
  static const double junk_value1 = vPrecomputedSin::PrecomputeValues();
  static const double junk_value2 = vPrecomputedCos::PrecomputeValues();

  static const double first_x = 1;
  static const double first_y = 0;

  double rad = two_pi / n;
  double circumference = radius * two_pi;

  // compute x and y coordinates of second vertex
  double second_x = vPrecomputedCos::Cos(rad);
  double second_y = vPrecomputedSin::Sin(rad);
  
  // compute vertex distance for unit circle
  double x_diff = second_x - first_x;
  double y_diff = second_y - first_y;
  double unit_distance = sqrt(x_diff * x_diff + y_diff * y_diff);

  // adjust vertex distance for circle of given radius
  double distance = unit_distance * circumference / two_pi;
  return distance;
}



double S_OldLatitudeCircumference(double x_angle)
{
  static const double junk_value1 = vPrecomputedSin::PrecomputeValues();
  static const double junk_value2 = vPrecomputedCos::PrecomputeValues();

  // Convert angle to radians
  x_angle = vAdjustAngle(x_angle);
  double rad_x = x_angle * rad_factor;

  // Figure out the y part of the sphere radius that goes to a point
  // at that latitude, and the non_y part (that is the radius of the
  // circle at that latitude)
  double y_part = vPrecomputedCos::Cos(rad_x);
  double radius = sqrt(1 - y_part * y_part);
  double circumference = radius * two_pi;
  return circumference;
}


// Computes the distance between points (0,0) and (degrees, 0). In other
// words, the distance between two points on a great circle of the unit
// sphere separated by degrees radians.
double S_DistanceLimit(double degrees)
{
  // Get the locations of the two points
  vPoint3 point1 = vOldSphereSurfaceTo3D(0,0);
  vPoint3 point2 = vOldSphereSurfaceTo3D(degrees, 0);

  // Compute the distance between the two points
  double x_diff = point1.x - point2.x;
  double y_diff = point1.y - point2.y;
  double z_diff = point1.z - point2.z;

  double distance = sqrt(x_diff*x_diff + y_diff*y_diff + z_diff*z_diff);
  return distance;
}


// S_YPointsToSample computes the number of points that must be sampled
// at a given latitude (i.e. given distance from the north pole) 
// in order to guarantee that they are uniformly spaced and two
// consecutive points have distance no greater than the distance between
// points (0,0) and (degrees, 0)
long S_OldYPointsToSample(double x_angle, double degrees)
{
  // Figure out the circumference of the latitude circle
  double circumference = S_OldLatitudeCircumference(x_angle);
  double radius = circumference / two_pi;
  // Figure out the max distance we can allow between consecutive
  // points (the distance between (0,0) and (0,degrees)).
  double distance_limit = S_DistanceLimit(degrees);
  
  // adjust distance limit to account for rounding errors
  distance_limit = distance_limit * 1.000001;

  // We know we cannot satisfy distance_limit using
  // first_estimate points, or fewer
  long first_estimate = (long) (circumference / distance_limit - 1);
  if (first_estimate < 1) first_estimate = 1;
  long current_estimate = first_estimate;

  // Now keep increasing current_estimate until we can satisfy
  // distance_limit.
  while(1)
  {
    double distance;
    // If current_estimate is 1, vVertexDistance would return 0,
    // which is clearly not useful right here.
    if (current_estimate == 1) distance = 4*radius;
    else distance = vVertexDistance(current_estimate, radius);
    if (distance <= distance_limit) return current_estimate;
    current_estimate++;
  }

  // We should never get here.
  exit_error("S_YPointsToSample error\n");
}


// Suppose we want to sample points on the surface of the unit sphere,
// so that all sample points are x degrees away from the north pole,
// the points are spaced
// uniformly, and the distance between successive points is the distance
// between point (0, 0) and point (degrees, 0) on the surface of the unit 
// sphere. By point(A,B) here I mean point with latitude A (A degrees from
// the north pole) and Y degrees east. This is what this function computes
// If samples is not empty, the previous contents are retained
// and more contents are added to the back. Points are returned as 3D
// coordinates, assuming the sphere is a unit sphere centered at the origin.
// This function should be rewritten, to use vSampleViewpointYs.
long vOldSampleSphereYs(double x_angle, double degrees, vector<vPoint3> * samples)
{
  // Figure out how many points we should sample
  long number_of_points = S_OldYPointsToSample(x_angle, degrees);
  double y_degrees = 360.0 / number_of_points;
  double y_angle;
  for (y_angle = 0; y_angle < 360.0; y_angle += y_degrees)
  {
    vPoint3 point = vOldSphereSurfaceTo3D(x_angle, y_angle);
    samples->push_back(point);
  }

  return 1;
}



// This function samples points more or less evenly spaced along the surface
// of the unit sphere, such that no point on that surface is further than 
// the given number of degrees from a point in the sample. Point A is 
// d degrees away from point B if the smallest rotation that moves A to B
// is d degrees. If samples is not empty, the previous contents are retained
// and more contents are added to the back. Points are returned as 3D
// coordinates, assuming the sphere is a unit sphere centered at the origin.
// This function should be rewritten, to use vSampleViewpoints.
long vOldSampleSpherePoints(double degrees, std::vector<vPoint3> * samples)
{
  // Figure out how many values we will consider for the x axis (the 
  // north-south axis) and how far apart they will be spaced.
  long number_of_x_values = (long) (180.0 / degrees);
  // We adjust 180 degrees to avoid the effects of rounding errors
  double limit = 180.0 / 1.000001;
  if (number_of_x_values * degrees < limit)
  {
    number_of_x_values++;
  }
  double x_degrees = 180.0 / number_of_x_values;

  double x_angle;
  for (x_angle = 0; x_angle <= 180.0; x_angle += x_degrees)
  {
    vOldSampleSphereYs(x_angle, degrees, samples);
  }

  return 1;
}


// Suppose we want to sample points on the surface of the unit sphere,
// so that all sample points are x degrees away from the north pole,
// the points are spaced
// uniformly, and the distance between successive points is the distance
// between point (0, 0) and point (degrees, 0) on the surface of the unit 
// sphere. By point(A,B) here I mean point with latitude A (A degrees from
// the north pole) and Y degrees east. This is what this function computes
// If samples is not empty, the previous contents are retained
// and more contents are added to the back. The points are returned with 
// two coordinates, that specify their location on the sphere surface.
long vOldSampleViewpointYs(double x_angle, double degrees, vector<vPoint2> * samples)
{
  // Figure out how many points we should sample
  long number_of_points = S_OldYPointsToSample(x_angle, degrees);
  double y_degrees = 360.0 / number_of_points;
  double y_angle;
  for (y_angle = 0; y_angle < 360.0; y_angle += y_degrees)
  {
    samples->push_back(vPoint2(y_angle, x_angle));
  }

  return 1;
}


// This function samples points more or less evenly spaced along the surface
// of the unit sphere, such that no point on that surface is further than 
// the given number of degrees from a point in the sample. Point A is 
// d degrees away from point B if the smallest rotation that moves A to B
// is d degrees. If samples is not empty, the previous contents are retained
// and more contents are added to the back. The points are returned with 
// two coordinates, that specify their location on the sphere surface.
long vOldSampleViewpoints(double degrees, std::vector<vPoint2> * samples)
{
  // Figure out how many values we will consider for the x axis (the 
  // north-south axis) and how far apart they will be spaced.
  long number_of_x_values = (long) (180.0 / degrees);
  // We adjust 180 degrees to avoid the effects of rounding errors
  double limit = 180.0 / 1.000001;
  if (number_of_x_values * degrees < limit)
  {
    number_of_x_values++;
  }
  double x_degrees = 180.0 / number_of_x_values;

  double x_angle;
  for (x_angle = 0; x_angle <= 180.0; x_angle += x_degrees)
  {
    vOldSampleViewpointYs(x_angle, degrees, samples);
  }

  return 1;
}


// This function samples points more or less evenly spaced along the surface
// of the unit sphere, such that no point on that surface is further than 
// the given number of degrees from a point in the sample. Point A is 
// d degrees away from point B if the smallest rotation that moves A to B
// is d degrees. If samples is not empty, the previous contents are retained
// and more contents are added to the back. The points are returned with 
// two coordinates, that specify their location on the sphere surface.
long vPoserSampleViewpoints(double degrees, std::vector<vPoint2> * samples)
{
  vOldSampleViewpoints(degrees, samples);
  long number = samples->size();
  long i;
  for (i = 0; i < number; i++)
  {
    (*samples)[i].col -= 90.0;
  }

  return 1;
}


// This function is a lot more simple. We simply sample orientations spaced
// apart as specified by degrees, starting at 0 and ending before 360.
long vOldSampleOrientations(double degrees, vector<double> * samples)
{
  if (degrees <= 0) return 0;
  long start_size = samples->size();
  double i;
  for (i = 0.0; i < 360.0; i += degrees)
  {
    samples->push_back(i);
  }

  long end_size = samples->size();
  return end_size - start_size;
}
