// this file contains code for generating a database of edge images
// of hands, given images generated from Poser, and code for computing
// the matrices of chamfer distances that are needed in order to apply
// BoostMap to a data set of hand images.


#include "stdafx.h"

#include "chamfer_distances.h"
#include "basics/image_algorithms.h"
#include "basics/image_templates.h"
#include "basics/wrapper.h"
#include "algo_templates.h"
#include "opencv_utils.h"
#include "local_data.h"

#include "basics/definitions.h"



// given, for a particular hand pose, the poser
// cartoon output and the poser color output (where the colors
// are set based on the convention I used to use with the
// VirtualHand software), this function generates
// a synthetic edge image.
grayscale_image * poser_edge_image(color_image * cartoon, color_image * colors)
{
  // convert the cartoon image into an edge image
  vector<uchar> color(3);
  color[0] = 25;
  color[1] = 25;
  color[2] = 25;

  grayscale_image * cartoon_edges = vSelectColor(cartoon, & color);
  long number = count_non_zero(cartoon_edges);
  if (number == 0)
  {
    exit_error("error: empty cartoon edge image\n");
  }

  color_image * modified_colors = function_copy_replace_color(colors, 0, 0, 0, 1, 1, 1);
  function_replace_color(modified_colors, 157, 140, 140, 0, 0, 0);
  grayscale_image * selection = (grayscale_image *) modified_colors->threshold(1);
  grayscale_image * result = (grayscale_image *) vSelectPixels(cartoon_edges, selection);

  function_delete(cartoon_edges);
  function_delete(modified_colors);
  function_delete(selection);

  return result;
}


// here the input image first_edges is the type of image that
// is produced by poser_edge_image. This function processes that image
// by normalizing it so that its enclosing circle has the specified
// radius, and the result image has the specified size. In addition,
// the image is flipped so that corresponds to the right hand.
grayscale_image * poser_normalized_edges(grayscale_image * first_edges, long radius,
                                         long new_vertical, long new_horizontal)
{
  uchar_image * flipped = switch_top_bottom(first_edges);
  long center_vertical = new_vertical / 2;
  long center_horizontal = new_horizontal / 2;

  uchar_image * normalized = normalize_enclosing_circle4(flipped, radius, 
                                                         center_vertical, center_horizontal);
  grayscale_image * result = (grayscale_image *) normalized->copy(0, new_horizontal-1,
                                                                  0, new_vertical-1);

  function_delete(flipped);
  function_delete(normalized);

  return result;
}


/////////////////////////////////////////////////////////////////////
//                                                                 //
//      Beginning of implementation of class_pixel_sequence        //
//                                                                 //
/////////////////////////////////////////////////////////////////////


class_pixel_sequence::class_pixel_sequence(class_sequence * sequence)
{
  rows = sequence->Rows();
  cols = sequence->Cols();
  if ((rows > 256) || (cols > 256))
  {
    exit_error("error: coordinates exceeding 256 in class_pixel_sequence\n");
  }

  number_of_frames = sequence->NumberOfFrames();

  lengths = vnew(long, number_of_frames);
  indices = vnew(long, number_of_frames);

  // First store the points into vectors, since we don't know how
  // many points we will find.
  vector<uchar> temp_rows, temp_cols;

  long i;
  vPrint("\n");
  for (i = 0; i < number_of_frames; i++)
  {
    grayscale_image * gray = sequence->ReadFrameGray(i);
    if (gray == 0)
    {
      exit_error("error in reading frame %li\n", i);
    }

    vector<long> current_rows, current_cols;
    vGetNonZeroPoints(gray, &current_rows, &current_cols);
    long length = current_rows.size();
    lengths[i] = length;
    if (i == 0)
    {
      indices[i] = 0;
    }
    else 
    {
      indices[i] = indices[i-1] + lengths[i-1];
    }
    long j;
    for (j = 0; j < length; j++)
    {
      temp_rows.push_back((uchar) current_rows[j]);
      temp_cols.push_back((uchar) current_cols[j]);
    }
    vdelete(gray);
    vPrint("processed frame %li of %li\r", i, number_of_frames);
  }
  vPrint("\n");

  // Now that we have all the point coordinates, store them into the arrays.
  total_length = temp_rows.size();
  pixel_rows = vnew(uchar, total_length);
  pixel_cols = vnew(uchar, total_length);
  for (i = 0; i < total_length; i++)
  {
    pixel_rows[i] = temp_rows[i];
    pixel_cols[i] = temp_cols[i];
  }
}

  
class_pixel_sequence::class_pixel_sequence(class_pointer(grayscale_image *) images, long number)
{
  if (number <= 0) 
  {
    exit_error("class_pixel_sequence, error 1\n");
  }
  
  rows = images[0]->Rows();
  cols = images[0]->Cols();
  if ((rows > 256) || (cols > 256))
  {
    exit_error("error: coordinates exceeding 256 in class_pixel_sequence\n");
  }

  number_of_frames = number;

  lengths = vnew(long, number_of_frames);
  indices = vnew(long, number_of_frames);

  // First store the points into vectors, since we don't know how
  // many points we will find.
  vector<uchar> temp_rows, temp_cols;

  long i;
//  vPrint("\n");
  for (i = 0; i < number_of_frames; i++)
  {
    grayscale_image * gray = images[i];
    if (gray == 0)
    {
      exit_error("error in reading frame %li\n", i);
    }

    vector<long> current_rows, current_cols;
    vGetNonZeroPoints(gray, &current_rows, &current_cols);
    long length = current_rows.size();
    lengths[i] = length;
    if (i == 0)
    {
      indices[i] = 0;
    }
    else 
    {
      indices[i] = indices[i-1] + lengths[i-1];
    }
    long j;
    for (j = 0; j < length; j++)
    {
      temp_rows.push_back((uchar) current_rows[j]);
      temp_cols.push_back((uchar) current_cols[j]);
    }
    //vPrint("processed frame %li of %li\r", i, number_of_frames);
  }
//  vPrint("\n");

  // Now that we have all the point coordinates, store them into the arrays.
  total_length = temp_rows.size();
  pixel_rows = vnew(uchar, total_length);
  pixel_cols = vnew(uchar, total_length);
  for (i = 0; i < total_length; i++)
  {
    pixel_rows[i] = temp_rows[i];
    pixel_cols[i] = temp_cols[i];
  }
}

  
class_pixel_sequence::~class_pixel_sequence()
{
  vdelete2(lengths);
  vdelete2(indices);
  vdelete2(pixel_rows);
  vdelete2(pixel_cols);
}


class_pixel_sequence * class_pixel_sequence::load(const char * filename)
{
  FILE * fp = fopen(filename, vFOPEN_READ);
  if (fp == 0) return 0;
  class_pixel_sequence * result = new class_pixel_sequence();
  long items = 0;
  items += vReadPCLongs(fp, &(result->rows), 1);
  items += vReadPCLongs(fp, &(result->cols), 1);
  items += vReadPCLongs(fp, &(result->number_of_frames), 1);
  items += vReadPCLongs(fp, &(result->total_length), 1);

  if ((items != 4) || (result->rows < 0) || (result->cols < 0))
  {
    vPrint("Error in loading point set sequence\n");
    vdelete(result);
    fclose(fp);
    return 0;
  }

  result->lengths = vnew(long, result->number_of_frames);
  result->indices = vnew(long, result->number_of_frames);
  result->pixel_rows = vnew(uchar, result->total_length);
  result->pixel_cols = vnew(uchar, result->total_length);

  items = vReadPCLongs(fp, vData(result->lengths), result->number_of_frames);
  items += vReadPCLongs(fp, vData(result->indices), result->number_of_frames);
  items += fread(vData(result->pixel_rows), sizeof(uchar), result->total_length, fp);
  items += fread(vData(result->pixel_cols), sizeof(uchar), result->total_length, fp);
  fclose(fp);
  if (items != 2 * result->number_of_frames + 2 * result->total_length)
  {
    vPrint("Error in reading point set sequence\n");
    vdelete(result);
    return 0;
  }
  return result;
}


long class_pixel_sequence::save(const char * filename)
{
  FILE * fp = fopen(filename, vFOPEN_WRITE);
  if (fp == 0) return 0;
  long items = 0;
  items += vWritePCLongs(fp, &rows, 1);
  items += vWritePCLongs(fp, &cols, 1);
  items += vWritePCLongs(fp, &number_of_frames, 1);
  items += vWritePCLongs(fp, &total_length, 1);

  if (items != 4)
  {
    vPrint("Error in saving point set sequence\n");
    fclose(fp);
    return 0;
  }

  items = vWritePCLongs(fp, vData(lengths), number_of_frames);
  items += vWritePCLongs(fp, vData(indices), number_of_frames);
  items += fwrite(vData(pixel_rows), sizeof(uchar), total_length, fp);
  items += fwrite(vData(pixel_cols), sizeof(uchar), total_length, fp);
  fclose(fp);
  if (items != 2 * number_of_frames + 2 * total_length)
  {
    vPrint("Error in writing point set sequence\n");
    return 0;
  }
  return 1;
}


uchar_matrix class_pixel_sequence::get_pixels(long frame)
{
  long number = length(frame);
  // Get the (row, col) coordinates for given frame and view 0.
  vArray(uchar) frame_rows = pixel_verticals(frame);
  vArray(uchar) frame_cols = pixel_horizontals(frame);
  
  // Copy over the coordinates into the result array.
  uchar_matrix result(number, 2);
  long i;
  for (i = 0; i < number; i++)
  {
    result(i, 0) = frame_rows[i];
    result(i, 1) = frame_cols[i];
  }

  return result;
}


// This function is used mainly for debuging, to visualise the points
// stored for each frame.
grayscale_image * class_pixel_sequence::pixel_image(long frame)
{
  if ((frame < 0) || (frame >= number_of_frames)) return 0;
  grayscale_image * result = new grayscale_image(rows, cols);
  vWriteValue(result, (uchar) 0);
  vArray(uchar) point_rows = pixel_verticals(frame);
  vArray(uchar) point_cols = pixel_horizontals(frame);
  long frame_length = length(frame);
  vWriteValue3(result, point_rows, point_cols, frame_length, (uchar) 255);
  return result;
}

  
/////////////////////////////////////////////////////////////////////
//                                                                 //
//         End of implementation of class_pixel_sequence           //
//                                                                 //
/////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////
//                                                                 //
//       Start of implementation of class_chamfer_features         //
//                                                                 //
/////////////////////////////////////////////////////////////////////


class_chamfer_features::class_chamfer_features()
{
  pixel_sequence = 0;
  distance_transforms = 0;
  erase_features_flag = 0;
}

class_chamfer_features::class_chamfer_features(class_pixel_sequence * in_pixel_sequence,
                                               grayscale_image ** in_distance_transforms)
{
  pixel_sequence = in_pixel_sequence;
  distance_transforms = in_distance_transforms;
  erase_features_flag = 0;
}


// this is the constructor we should call to preprocess the query image
class_chamfer_features::class_chamfer_features(grayscale_image * unnormalized_image, class_hand_database * database)
{
  erase_features_flag = 1;
  grayscale_image * query_image = database->normalize_query(unnormalized_image);
  
  if ((query_image->vertical() != 256) || (query_image->horizontal() != 256))
  {
    exit_error("error: unexpected query image size in class_checker_features constructor\n");
  }

  long number_of_orientations = database->get_orientation_number();
  grayscale_image ** rotated = new grayscale_image*[number_of_orientations];
  distance_transforms = new grayscale_image*[number_of_orientations];

  long i;
  for (i = 0; i < number_of_orientations; i++)
  {
    rotated[i] = database->inverse_rotate(query_image, i);

    v3dMatrix<float> * dt = vCVDistanceTransform(rotated[i]);
    distance_transforms[i] = (grayscale_image *) (new uchar_image(dt));
    vdelete(dt);
  }

  pixel_sequence = new class_pixel_sequence(rotated, number_of_orientations);

  for (i = 0; i < number_of_orientations; i++)
  {
    vdelete(rotated[i]);
  }
  vdelete2(rotated);
  function_delete(query_image);
}


class_chamfer_features::~class_chamfer_features()
{
  if (erase_features_flag)
  {
    erase_data();
  }

  pixel_sequence = 0;
  distance_transforms = 0;
}


long class_chamfer_features::erase_data()
{
  if (pixel_sequence == 0) return 0;
  long number_of_frames = pixel_sequence->get_frame_number();

  long frame;
  for (frame = 0; frame < number_of_frames; frame++)
  {
    vdelete(distance_transforms[frame]);
  }

  vdelete2(distance_transforms);
  vdelete(pixel_sequence);
  return 1;
}


/////////////////////////////////////////////////////////////////////
//                                                                 //
//        End of implementation of class_chamfer_features          //
//                                                                 //
/////////////////////////////////////////////////////////////////////



/////////////////////////////////////////////////////////////////////
//                                                                 //
//       Beginning of implementation of class_hand_database        //
//                                                                 //
/////////////////////////////////////////////////////////////////////



class_hand_database::class_hand_database()
{
  initialize();
}


class_hand_database::~class_hand_database()
{
  vdelete2(name);

  vdelete(edge_sequence);
  clear_distances();
  clear_pixel_sequence();
  destroy_rotation_templates();
}


class_hand_database::class_hand_database(const char * in_name)
{
  initialize();
  name = vCopyString(in_name);

  char * edge_sequence_name = edges_filename();
  edge_sequence = (vRLE_Sequence *) vOpenReadVideoFile(edge_sequence_name);
  if (edge_sequence != 0)
  {
    if (edge_sequence->NumberOfFrames() < number_of_frames)
    {
      exit_error("Bad image sequence read from class_hand_database\n");
    }
  }
  vdelete2(edge_sequence_name);
  load_pixel_sequence();

  is_valid = 1;
}


void class_hand_database::initialize()
{
  is_valid = 0;
  
  number_of_shapes = 20;
  number_of_viewpoints = 84;
  number_of_orientations = 48;
  number_of_frames = number_of_shapes * number_of_viewpoints;
  frames_per_shape = number_of_viewpoints;
  number_of_poses = number_of_frames * number_of_orientations;

  bounding_radius = 120;
  image_size = 256;
  name = 0;

  edge_sequence = 0;
  distance_transforms = 0;
  pixel_sequence = 0;
  rotation_templates = vZero(v3dMatrix<short> *);
}


char * class_hand_database::make_pathname(const char * filename)
{
  char * database_directory = vJoinPaths(sequences_directory, "poser");
  char * result = vJoinPaths(database_directory, filename);
  vdelete2(database_directory);
  return result;
}


// This function takes in an ending and returns the full path
// (i.e., adds to the beginning the directory and the name
// of the database). It is used for filenames refering to
// sequences where the size of the hand has been normalized.
char * class_hand_database::filename_with_ending(const char * ending)
{
  char * name1 = vMergeStrings2(name, ending);
  char * name2 = make_pathname(name1);
  vdelete2(name1);
  return name2;
}

  
// This function takes as input an ending specifying
// a particular edge image sequence. It returns the full
// path name of a file that is supposed to store the distance
// transforms for that edge image sequence.
char * class_hand_database::distances_filename()
{
  char * filename = filename_with_ending("_dt.vimseq1");
  return filename;
}


// This function takes as input an ending specifying
// a particular edge image sequence. It returns the full
// path name of a file that is supposed to store the point
// sequence for that edge image sequence.
char * class_hand_database::pixel_sequence_filename()
{
  char * filename = filename_with_ending("_edges.point_set");
  return filename;
}  


// Returns edge image indexed by shape_id, viewpoint_index, orientation_index
// The caller is assumed to know what actual shapes and 3D orientations
// these indices correspond to.
grayscale_image * class_hand_database::get_edges_three(long shape_id, long viewpoint_index, 
                                                       long orientation_index)
{
  long frame = shape_id * number_of_viewpoints + viewpoint_index;
  return get_edges_two(frame, orientation_index);
}

grayscale_image * class_hand_database::get_edges_two(long frame, long orientation)
{
  if (edge_sequence == 0)
  {
    return 0;
  }
  create_rotation_templates();
  if ((orientation < 0) || (orientation >= number_of_orientations))
  {
    return 0;
  }

  grayscale_image * original = edge_sequence->ReadFrameGray(frame);
  if (orientation == 0)
  {
    return original;
  }

  grayscale_image * rotated = (grayscale_image *) inverse_warp(original, rotation_templates[orientation]);

  function_delete(original);
  return rotated;
}


grayscale_image * class_hand_database::get_edges(long pose_number)
{
  long frame = pose_number / number_of_orientations;
  long orientation = pose_number % number_of_orientations;
  return get_edges_two(frame, orientation);
}

  
// This function simply extracts
// the non-zero points (edge pixels) of each frame, and stores them.
// This gives us a representation that, together with distance transforms,
// can be used for efficient implementation of the chamfer distance,
// in terms of both time and memory.
long class_hand_database::create_pixel_sequence()
{
  char * edge_sequence_name = edges_filename();
  vVideoFile * edge_sequence = open_read_sequence(edge_sequence_name);
  vdelete2(edge_sequence_name);
  if (edge_sequence == 0) 
  {
    return 0;
  }

  class_pixel_sequence sequence(edge_sequence);
  char * filename = pixel_sequence_filename();
  long success = sequence.save(filename);
  vdelete2(filename);
  close_sequence(edge_sequence);
  return success;
}


// Loads the specified distance transform into distance transforms, 
// and, if successful, sets it as the current distance transform.
long class_hand_database::load_distances()
{
  if (predicate_zero(distance_transforms) == 0)
  {
    return 1;
  }

  char * filename = distances_filename();
  vPrint("filename = %s\n", filename);
  vVideoFile * dt_sequence = vOpenReadVideoFile(filename);
  vdelete2(filename);
  if (dt_sequence == 0) return 0;
  if (dt_sequence->NumberOfFrames() < number_of_frames)
  {
    vCloseVideoFile(dt_sequence);
    return 0;
  }

  distance_transforms = vnew(grayscale_image *, number_of_frames);
  long i;
  vPrint("\n");
  for (i = 0; i < number_of_frames; i++)
  {
    distance_transforms[i] = dt_sequence->ReadFrameGray(i);
    if (distance_transforms[i] == 0)
    {
      exit_error("load_distances: error1\n");
    }
    vPrint("Processed frame %li of %li\r", i, number_of_frames);
  }
  vPrint("\n");

  vCloseVideoFile(dt_sequence);
  return 1;
}


// Removes all distance transforms from memory.
long class_hand_database::clear_distances()
{
  if (predicate_zero(distance_transforms))
  {
    return 1;
  }

  long j;
  for (j = 0; j < number_of_frames; j++)
  {
    vdelete(distance_transforms[j]);
  }

  vdelete2(distance_transforms);
  distance_transforms = 0;

  return 1;
}


// Returns the i-th frame of the current DT sequence.
// Used mostly for debugging.
grayscale_image * class_hand_database::distance_transform(long frame, long orientation)
{
  if (distance_transforms == 0) return 0;
  if (frame < 0) return 0;
  if (frame >= number_of_frames) return 0;
  
  grayscale_image * original = distance_transforms[frame];
  if (orientation == 0)
  {
    return original->ToGrayCopy();
  }

  float degrees = orientation_degrees(orientation);
  grayscale_image * result = (grayscale_image *) function_rotate_image(original, center(), center(), degrees);
  return result;
}


// Loads the specified point_sequence into point_sequences, 
// and, if successful, sets it as the current point_sequence.
long class_hand_database::load_pixel_sequence()
{
  if (pixel_sequence != 0) 
  {
    return 1;
  }

  char * filename = pixel_sequence_filename();
  pixel_sequence = class_pixel_sequence::load(filename);
  vdelete2(filename);
  if (pixel_sequence == 0) return 0;
  if (pixel_sequence->get_frame_number() < number_of_frames)
  {
    exit_error("error: in load_pixel_sequence\n");
    return 0;
  }

  return 1;
}


// Removes the specified distance transform from memory
long class_hand_database::clear_pixel_sequence()
{
  vdelete(pixel_sequence);
  pixel_sequence = 0;
  return 1;
}


// Returns the i-th frame of the current point sequence.
// Used mostly for debugging.
grayscale_image * class_hand_database::pixel_image(long frame, long orientation)
{
  if (pixel_sequence == 0) return 0;
  if (frame < 0) return 0;
  if (frame >= pixel_sequence->get_frame_number()) return 0;
  if ((orientation < 0) || (orientation >= number_of_orientations))
  {
    return 0;
  }
  
  uchar_matrix pixels = pixel_sequence->get_pixels(frame);
  uchar_matrix rotated = rotate_pixels(pixels, orientation);

  grayscale_image * result = new grayscale_image(image_size, image_size);
  function_enter_value(result, (uchar) 0);
  function_enter_values(result, long_matrix(&rotated), (uchar) 255);

  return result;
}


// Create or delete the rotation templates that can be used to 
// generate different image plane rotations of a query image.
long class_hand_database::create_rotation_templates()
{
  if (vIsZero(rotation_templates) == 0) return 1;

  rotation_templates = vnew(v3dMatrix<short> *, number_of_orientations);
  double step = 360.0 / (double) number_of_orientations;

  long i;
  for (i = 0; i < number_of_orientations; i ++)
  {
    // angle is positive. Positive angles represent clockwise rotation.
    // Rotating the query by a negative angle and matching
    // with a database frame is the same as matching the unrotated query
    // with the database frame rotated by the equivalent positive angle.
    // However, because the database was built using left hands, the 
    // angles must be inverted.
    double angle = -step * (double) i;
    rotation_templates[i] = 
      vRotationTemplate(image_size, image_size, center(), center(), angle);
  }
  return 1;
}


long class_hand_database::destroy_rotation_templates()
{
  if (vIsZero(rotation_templates)) return 1;
  long i;
  for (i = 0; i < number_of_orientations; i++)
  {
    vdelete(rotation_templates[i]);
  }
  vdelete2(rotation_templates);
  return 1;
}


// For debugging:
v3dMatrix<short> * class_hand_database::rotation_template(long rotation_index)
{
  create_rotation_templates();
  if (rotation_index < 0) return 0;
  if (rotation_index >= number_of_orientations) return 0;
  v3dMatrix<short> * result = 
    new v3dMatrix<short>(rotation_templates[rotation_index]);
  return result;
}


// Prepare the chamfer featuers of a query input image (that
// is assumed to be an edge image). Those features
// can be used to efficiently calculate chamfer distances.
class_chamfer_features * class_hand_database::query_chamfer_features(grayscale_image * query_image)
{
  return new class_chamfer_features(query_image, this);
}


class_chamfer_features * class_hand_database::database_chamfer_features()
{
  if ((pixel_sequence == 0) || (predicate_zero(distance_transforms)))
  {
    return 0;
  }
  return new class_chamfer_features(pixel_sequence, 
                                    distance_transforms);
}


uchar_matrix class_hand_database::rotate_pixels(uchar_matrix pixels, long orientation)
{
  create_rotation_templates();
  if ((orientation <0) || (orientation >= number_of_orientations))
  {
    return uchar_matrix();
  }

  short_image * rotation_template = rotation_templates[orientation];
  long number = pixels.vertical();
  uchar_matrix result(number, 2);

  long counter;
  for (counter = 0; counter < number; counter++) 
  {
    long vertical = pixels(counter, 0);
    long horizontal = pixels(counter, 1);
    long new_vertical = (*rotation_template)(0, vertical, horizontal);
    long new_horizontal = (*rotation_template)(1, vertical, horizontal);

    if ((new_vertical <0) || (new_vertical >= 256) ||
        (new_horizontal <0) || (new_horizontal >= 256) )
    {
      // we should never get here. The original pixels should all be
      // within a circle of 120 pixels away from (128, 128).
      exit_error ("\nerror: unexpected pixel values\n");
    }
    result(counter, 0) = (uchar) new_vertical;
    result(counter, 1) = (uchar) new_horizontal;
  }

  return result;
}


long class_hand_database::print()
{
  function_print("\n\nhand database %s\n", name);
  function_print("%li shapes, %li viewpoints, %li orientations\n",
                 number_of_shapes, number_of_viewpoints, number_of_orientations);
  function_print("%li stored frames, %li database poses\n",
                 number_of_frames, number_of_frames * number_of_orientations);
  function_print("%li poses per shape\n", number_of_viewpoints * number_of_orientations);

  return 1;
}


grayscale_image * class_hand_database::rotate(grayscale_image * image, long orientation)
{
  create_rotation_templates();
  grayscale_image * result = (grayscale_image *) inverse_warp(image, rotation_templates[orientation]);
  return result;
}


grayscale_image * class_hand_database::inverse_rotate(grayscale_image * image, long orientation)
{
  create_rotation_templates();
  long inverse_orientation = (number_of_orientations - orientation) % number_of_orientations;
  grayscale_image * result = (grayscale_image *) inverse_warp(image, rotation_templates[inverse_orientation]);
  return result;
}

// normalizes  the edge image given as argument, so that its center,
// radius of enclosing circle, and overall size have the appropriate
// values for the purposes of computing chamfer distances
// with all database poses.
grayscale_image * class_hand_database::normalize_query(grayscale_image * query)
{
  long original_vertical = query->vertical();
  long original_horizontal = query->horizontal();
  long new_vertical = function_maximum(original_vertical, image_size);
  long new_horizontal = function_maximum(original_horizontal, image_size);
  
  // note that this is wasteful and inefficient when the query
  // is larger than the required image size
  uchar_image * padded = vPadCentered(query, new_vertical, new_horizontal);

  long center = image_size /2;
  uchar_image * normalized = normalize_enclosing_circle4(padded, bounding_radius, center, center);
  uchar_image * cropped = function_copy_rectangle(normalized, 0, image_size -1, 0, image_size -1);

  function_delete(padded);
  function_delete(normalized);

  return (grayscale_image *) cropped;
}


/////////////////////////////////////////////////////////////////////
//                                                                 //
//          end of implementation of class_hand_database           //
//                                                                 //
/////////////////////////////////////////////////////////////////////


float fast_chamfer_distance(class_chamfer_features * query, long query_orientation,
                            class_chamfer_features * database, long database_frame,
                            float * query_model_distance, float * model_query_distance)
{
  vArray(uchar) query_rows = query->pixel_sequence->pixel_verticals(query_orientation);
  vArray(uchar) query_cols = query->pixel_sequence->pixel_horizontals(query_orientation);
  long query_pixels = query->pixel_sequence->length(query_orientation);
  if (query_pixels == 0) 
  {
    return -1.0f;
  }
  v3dMatrix<uchar> * query_dt = query->distance_transforms[query_orientation];
  vArray2(uchar) query_dt2 = query_dt->Matrix2(0);
  
  vArray(uchar) database_rows = database->pixel_sequence->pixel_verticals(database_frame);
  vArray(uchar) database_cols = database->pixel_sequence->pixel_horizontals(database_frame);
  long database_pixels = database->pixel_sequence->length(database_frame);
  if (database_pixels == 0)
  {
    exit_error("error: chamfer_distances, no pixels for database frame\n");
  }
  v3dMatrix<uchar> * database_dt = database->distance_transforms[database_frame];
  vArray2(uchar) database_dt2 = database_dt->Matrix2(0);

  long j;

  float im_score = 0; // image-to-model score;
  for (j = 0; j < query_pixels; j++)
  {
    im_score += database_dt2[query_rows[j]][query_cols[j]];
  }

  float mi_score = 0; // model-to-image score;
  for (j = 0; j < database_pixels; j++)
  {
    mi_score += query_dt2[database_rows[j]][database_cols[j]];
  }

  im_score = im_score / ((float) query_pixels);
  mi_score = mi_score / ((float) database_pixels);
  
  if (query_model_distance != 0)
  {
    *query_model_distance = im_score;
  }
  if (model_query_distance != 0)
  {
    *model_query_distance = mi_score;
  }

  float result = im_score + mi_score;
  return result;
}


vMatrix<float> chamfer_distances(grayscale_image * query, 
                                 class_hand_database * database)
{
  class_chamfer_features * database_features = database->database_chamfer_features();
  if (database_features == 0)
  {
    function_print("\nfailed to extract database features\n");
    return float_matrix();
  }

  class_chamfer_features query_features(query, database);
  float_matrix result = chamfer_distances(& query_features, database_features);
  function_delete(database_features);

  return result;
}


vMatrix<float> chamfer_distances(class_chamfer_features * query, 
                                 class_chamfer_features * database)
{
  long frames = database->pixel_sequence->get_frame_number();
  long orientations = query->pixel_sequence->get_frame_number();
  long number = frames * orientations;

  vMatrix<float> result_matrix(1, number);
  vArray(float) result = result_matrix.Matrix();

//  function_print("\n");
  long frame, orientation;
  for (frame = 0; frame < frames; frame++)
  {
    for (orientation = 0; orientation < orientations; orientation++)
    {
      long index = frame * orientations + orientation;
      result[index] = fast_chamfer_distance(query, orientation, database, frame, 0, 0);
//      function_print("processed %li of %li frames\r", index +1, number);
    }
  }
//  function_print("\n");

  return result_matrix;
}


// produces the file where we should save the results
// of computing chamfer distances between the sequence stored at the given
// sequence_name location and the hand database.
char * chamfer_distances_filename(const char * sequence_name)
{
  char * result = function_replace_ending(sequence_name, ".chamfer");
  return result;
}


// compute distances between every frame of the sequence
// and every database pose, save those distances to file
// and return a matrix of those distances
float_matrix sequence_chamfer_distances(class_sequence * sequence,
                                        class_hand_database * database)
{
  if (sequence->valid() <= 0)
  {
    return float_matrix();
  }

  long query_number = sequence->NumberOfFrames();
  long pose_number = database->get_pose_number();
  float_matrix result(query_number, pose_number);

  function_print("\n");
  long frame;
  for (frame = 0; frame < query_number; frame++)
  {
    grayscale_image * query = sequence->ReadFrameGray(frame);
    if (query == 0)
    {
      exit_error("\nerror: failed to read query %li\n", frame);
    }

    float_matrix distances = chamfer_distances(query, database);
    function_delete(query);
    if (distances.valid() <= 0)
    {
      function_print("\nfailed to compute distances\n");
      return float_matrix();
    }
    function_put_row(& distances, & result, frame);
    function_print("\nprocessed frame %li of %li\r", frame +1, query_number);
  }
  function_print("\n");

  const char * sequence_name = sequence->get_filename();
  char * filename = chamfer_distances_filename(sequence_name);
  long success = result.Write(filename);
  if (success <= 0)
  {
    function_print("failed to save result to %s\n", filename);
  }
  else
  {
    function_print("saved results to %s\n", filename);
  }

  delete_pointer(filename);
  return result;
}


// from the matrix produced using sequence_chamfer_distances, we read
// the results corresponding to a specific frame, and then we adjust 
// those results so that they correspond to the given orientation index.
float_matrix load_pose_chamfer_distances(class_sequence * sequence,
                                         const long query_frame, const long query_orientation)
{
  const long number_of_orientations = 48;
  if ((query_frame < 0) || (query_orientation < 0) || (query_orientation >= number_of_orientations))
  {
    function_print("\ninvalid arguments\n");
    return float_matrix();
  }

  const char * sequence_name = sequence->get_filename();
  char * distances_filename = chamfer_distances_filename(sequence_name);
  file_handle * file_pointer = open_read_file(distances_filename);
  if (file_pointer == 0)
  {
    function_print("\nfailed to open %s\n", distances_filename);
    delete_pointer(distances_filename);
    return float_matrix();
  }
  delete_pointer(distances_filename);

  vector<long> header;
  long success = float_image::ReadHeader(file_pointer, & header);
  if ((success <= 0) || (header[0] != 4) || (header[3] != 1))
  {
    function_print("\ninvalid header\n");
    close_file(file_pointer);
    return float_matrix();
  }

  long vertical = header[1];
  long horizontal = header[2];
  long database_poses = horizontal;
  long sequence_frames = vertical;

  if (query_frame >= sequence_frames)
  {
    function_print("\nquery_frame = %li, sequence_frames = %li\n", query_frame, sequence_frames);
    close_file(file_pointer);
    return float_matrix();
  }

  vint8 converted_horizontal = horizontal;
  vint8 float_size = sizeof(float);
  vint8 converted_frame = query_frame;
  vint8 header_size = header.size() * sizeof(long);
  vint8 offset = converted_horizontal * float_size * converted_frame + header_size;
  success = function_seek(file_pointer, offset, SEEK_SET);
  if (success <= 0)
  {
    function_print("\nfailed in file seek operation\n");
    close_file(file_pointer);
    return float_matrix();
  }

  float_matrix distances(1, database_poses);
  success = distances.read_bandwise(file_pointer);
  if (success <= 0)
  {
    function_print("\nfailed to read distances from file\n");
    close_file(file_pointer);
    return float_matrix();
  }
  close_file(file_pointer);

  float_matrix result(1, database_poses);
  long counter;
  for (counter = 0; counter < database_poses; counter++)
  {
    long frame = counter / number_of_orientations;
    long orientation = counter % number_of_orientations;
    long frame_base = frame * number_of_orientations;
    long adjusted_orientation = (orientation + query_orientation) % number_of_orientations;
    long target_counter = frame_base + adjusted_orientation;
    result(target_counter) = distances(counter);
  }

  return result;
}


// used to generate traintrain chamfer distances
long full_distance_matrix(class_sequence * sequence, const char * filename)
{
  const long number_of_orientations = 48;

  float_matrix first_distances = load_pose_chamfer_distances(sequence, 0, 0);
  long database_poses = first_distances.length();

  long sequence_frames = sequence->NumberOfFrames();
  long sequence_poses = sequence_frames * number_of_orientations;

  file_handle * file_pointer = open_save_file(filename);
  if (file_pointer == 0)
  {
    function_print("\nfailed to open %s\n", filename);
    return 0;
  }

  //const long frames_to_save = 2;
  const long frames_to_save = sequence_frames;

  const long poses_to_save = frames_to_save * number_of_orientations;
  long success = vWriteHeader(file_pointer, FloatType, poses_to_save, database_poses, 1);
  if (success == 0)
  {
    function_print("\nfailed to open %s\n", filename);
    return 0;
  }

  long frame_counter, orientation_counter;
  function_print("\n");
  for (frame_counter = 0; frame_counter < frames_to_save; frame_counter++)
  {
    for (orientation_counter = 0; orientation_counter < number_of_orientations; orientation_counter++)
    {
      float_matrix distances = load_pose_chamfer_distances(sequence, frame_counter, orientation_counter);
      if (distances.valid() <= 0)
      {
        function_print("\nfailed to load distances for frame %li, orientation %li\n", 
                       frame_counter, orientation_counter);
        close_file(file_pointer);
        return 0;
      }

      success = distances.store_bandwise(file_pointer);
      if (success <= 0)
      {
        function_print("\nfailed to save distances for frame %li, orientation %li\n", 
                       frame_counter, orientation_counter);
        close_file(file_pointer);
        return 0;
      }
      
      function_print("processed frame %li, orientation %li\r", 
                     frame_counter, orientation_counter);
    }
  }
  function_print("\n");

  close_file(file_pointer);
  return 1;
}












































