#ifndef VASSILIS_CHAMFER_DISTANCES_H
#define VASSILIS_CHAMFER_DISTANCES_H

#include "basics/algebra.h"
#include "basics/video_common.h"
#include "basics/rle_encoding.h"

// given, for a particular hand pose, the poser
// cartoon output and the poser color output (where the colors
// are set based on the convention I used to use with the
// VirtualHand software), this function generates
// a synthetic edge image.
grayscale_image * poser_edge_image(color_image * cartoon, color_image * colors);


// here the input image first_edges is the type of image that
// is produced by poser_edge_image. This function processes that image
// by normalizing it so that its enclosing circle has the specified
// radius, and the result image has the specified size. In addition,
// the image is flipped so that corresponds to the right hand.
grayscale_image * poser_normalized_edges(grayscale_image * first_edges, long radius,
                                         long new_vertical, long new_horizontal);




// This class is a simplified version of the class vPointSetSequence,
// defined in fingers.h and fingers.cpp, and class vPointSequence 
// from hand_database.h. It is used as an interface 
// to access directly the set of 
// points in the sequence of database edge images. We can load up
// the set of points in all edge images in memory, and access
// the set of points of each image separately.
// We don't allow point coordinates to exceed 255, so that they
// can fit into 8 bits.
// The reason why I am not using vPointSetSequence here and I define
// a new class is that vPointSetSequence was too specialized and 
// it was doing things that are not needed here. Hopefully when I rewrite
// these parts of the code, I will be able to erase old code that I don't 
// use anymore.
class class_pixel_sequence
{
private:
  long rows, cols;
  long number_of_frames;
  long total_length;
  vArray(long) lengths;
  vArray(long) indices;
  vArray(uchar) pixel_rows;
  vArray(uchar) pixel_cols;

  inline void check_frame(long frame)
  {
    if ((frame < 0) || (frame >= number_of_frames))
    {
      exit_error("PointSetSequence error\n");
    }
  }

public:
  class_pixel_sequence() {}
  class_pixel_sequence(vVideoFile * sequence);
  class_pixel_sequence::class_pixel_sequence(class_pointer(grayscale_image *) images, long number);
  ~class_pixel_sequence();

  static class_pixel_sequence * load(const char * filename);
  long save(const char * filename);

  inline long vertical()
  {
    return rows;
  }

  inline long horizontal()
  {
    return cols;
  }

  inline long get_frame_number()
  {
    return number_of_frames;
  }

  inline long get_total_length()
  {
    return total_length;
  }

  inline long length(long frame)
  {
    check_frame(frame);
    return lengths[frame];
  }

  inline vArray(uchar) pixel_verticals(long frame)
  {
    check_frame(frame);
    long index = indices[frame];
    return pixel_rows + index;
  }

  inline vArray(uchar) pixel_horizontals(long frame)
  {
    check_frame(frame);
    long index = indices[frame];
    return pixel_cols + index;
  }

  // returns and matrix where every row corresponds to a pixel and
  // has two values: the vertical and horizontal coordinates of the pixel.
  uchar_matrix get_pixels(long frame);

  // This function is used mainly for debugging, to visualise the points
  // stored for each frame.
  grayscale_image * pixel_image(long frame);
};


class class_hand_database;

// class vChamferFeatures is used to represent the set of features
// (point set and distance transform) for a sequence of images. That
// sequence of images could represent different image plane rotations
// of the query image, or different poses in the database.
class class_chamfer_features
{
protected:
  long erase_data();

public:
  class_pixel_sequence * pixel_sequence;
  grayscale_image ** distance_transforms;
  
  // erase_features_flag is set to zero if this subject was constructed
  // with a constructor that takes in the pixel sequence
  // and the distance transforms as arguments. In that case,
  // the destructor should NOT delete those variables because they were created
  // me for the object itself was created. If the other constructor is used,
  // which creates pixel_sequence and distance_transforms, then
  // erase_features_flag is set to 1.
  long erase_features_flag;
  
  class_chamfer_features();

  class_chamfer_features(class_pixel_sequence * in_pixel_sequence,
                         grayscale_image ** in_distance_transforms);

  // this is the constructor we should call to preprocess the query image.
  // it is assumed, and not checked here, that the query image has already
  // been normalized, so that the minimum enclosing circle is centered
  // at 128,128, and has radius 120, and the image size is 256x256.
  class_chamfer_features(grayscale_image * image, class_hand_database * database);

  ~class_chamfer_features();

};


// This class will be used as an interface to the 
// selected_asl database of edge images of hands.
// I am implementing it in a hurry for my thesis,
// but in principle it should require only few modifications
// to work with other databases generated using poser,
// with different numbers of shapes, viewpoints,
// and image plane orientations
class class_hand_database
{
protected:
  long is_valid;
  char * name; // "selected_asl" for the thesis.

  vRLE_Sequence * edge_sequence; // the original edge images
  grayscale_image ** distance_transforms;
  class_pixel_sequence * pixel_sequence;


  long number_of_shapes;
  // number of viewpoints (camera locations) 
  long number_of_viewpoints;
  long number_of_orientations; // image plane orientations

  // Number of frames stored in the sequence.
  long number_of_frames;
  // Number of frames stored per shape.
  long frames_per_shape;

  // number of poses represented in the database
  long number_of_poses;

  // The radius of the enclosing circle for all images
  long bounding_radius;

  // The rows and cols of the normalized images
  long image_size;

  class_pointer(short_image *) rotation_templates;

protected:
  // Assign default values to member variables.
  void initialize();
  
  // This function takes in an ending and returns the full path
  // (i.e., adds to the beginning the directory and the name
  // of the database). It is used for filenames refering to
  // sequences where the size of the hand has been normalized.
  char * filename_with_ending(const char * ending);

  // This function takes as input an ending specifying
  // a particular edge image sequence. It returns the full
  // path name of a file that is supposed to store the distance
  // transforms for that edge image sequence.
  char * distances_filename();

  // This function takes as input an ending specifying
  // a particular edge image sequence. It returns the full
  // path name of a file that is supposed to store the point
  // sequence for that edge image sequence.
  char * pixel_sequence_filename();

public:
  class_hand_database();
  class_hand_database(const char * argument_name);
  ~class_hand_database();

  // Returns edge image indexed by shape_id, viewpoint_index, orientation_index
  // The caller is assumed to know what actual shapes and 3D orientations
  // these indices correspond to.
  grayscale_image * get_edges_three(long shape_id, long viewpoint_index, 
                                    long orientation_index);

  grayscale_image * get_edges_two(long frame, long orientation);

  grayscale_image * get_edges(long pose_number);

  // Add the standard database directory to a filename that
  // has no path information.
  char * make_pathname(const char * filename);

  inline const char * get_name()
  {
    return name;
  }

  inline long get_shape_number()
  {
    return number_of_shapes;
  }

  inline long get_viewpoint_number()
  {
    return number_of_viewpoints;
  }

  inline long get_orientation_number()
  {
    return number_of_orientations;
  }

  inline long get_pose_number()
  {
    return number_of_poses;
  }

  inline long valid()
  {
    return is_valid;
  }

  inline long get_bounding_radius()
  {
    return bounding_radius;
  }

  inline long get_frame_number()
  {
    return number_of_frames;
  }
  
  // The center is useful for figuring out the center of normalized 
  // images.
  inline long center()
  {
    return image_size / 2;
  }

  inline char * edges_filename()
  {
    return filename_with_ending("_edges.rle");
  }

  // Loads the specified distance transform into distance transforms, 
  // and, if successful, sets it as the current distance transform.
  long load_distances();

  long clear_distances();

  // Returns the i-th frame of the current DT sequence.
  // Used mostly for debugging.
  grayscale_image * distance_transform(long frame, long orientation);

  // This function simply extracts
  // the non-zero points (edge pixels) of each frame, and stores them.
  // This gives us a representation that, together with distance transforms,
  // can be used for efficient implementation of the chamfer distance,
  // in terms of both time and memory.
  long create_pixel_sequence();

  // Loads the specified point_sequence into point_sequences, 
  // and, if successful, sets it as the current point_sequence.
  long load_pixel_sequence();

  // Removes the specified distance transform from memory
  long clear_pixel_sequence();

  // Returns the i-th frame of the current point sequence.
  // Used mostly for debugging.
  grayscale_image * pixel_image(long frame, long orientation = 0);

  // Create or delete the rotation templates that can be used to 
  // generate different image plane rotations of a query image.
  long create_rotation_templates();
  long destroy_rotation_templates();

  // For debugging:
  short_image * rotation_template(long rotation_index);

  // Prepare the chamfer features of a query input image (that
  // is assumed to be an edge image). Those features
  // can be used to efficiently calculate chamfer distances.
  class_chamfer_features * query_chamfer_features(grayscale_image * query_image);
  class_chamfer_features * database_chamfer_features();

  inline float orientation_degrees(long orientation)
  {
    return (360.0f / number_of_orientations) * (float) orientation;
  }

  uchar_matrix rotate_pixels(uchar_matrix pixels, long orientation);
  grayscale_image * rotate(grayscale_image * image, long orientation);
  grayscale_image * inverse_rotate(grayscale_image * image, long orientation);
  
  // normalizes  the edge image given as argument, so that its center,
  // radius of enclosing circle, and overall size have the appropriate
  // values for the purposes of computing chamfer distances
  // with all database poses.
  grayscale_image * normalize_query(grayscale_image * query);

  long print();
};


float fast_chamfer_distance(class_chamfer_features * query, long query_orientation,
                            class_chamfer_features * database, long database_frame,
                            float * query_model__distance, float * model_query__distance);


// compute distances between the query and every database pose
vMatrix<float> chamfer_distances(grayscale_image * query, 
                                 class_hand_database * database);

// compute distances between the query and every database pose
vMatrix<float> chamfer_distances(class_chamfer_features * query, 
                                 class_chamfer_features * database);

// produces the file where we should save the results
// of computing chamfer distances between the sequence stored at the given
// sequence_name location and the hand database.
// note that the result filename does not depend on the database.
// in principle, if we wanted to save results to two different databases
// then we would have to change this function so that it takes the database
// as a second argument.
char * chamfer_distances_filename(const char * sequence_name);

// compute distances between every frame of the sequence
// and every database pose, save those distances to file
// and return a matrix of those distances
float_matrix sequence_chamfer_distances(class_sequence * sequence,
                                        class_hand_database * database);

// from the matrix produced using sequence_chamfer_distances, we read
// the results corresponding to a specific frame, and then we adjust 
// those results so that they correspond to the given orientation index.
float_matrix load_pose_chamfer_distances(class_sequence * sequence,
                                         long frame, long orientation_index);

// used to generate traintrain chamfer distances
long full_distance_matrix(class_sequence * sequence, const char * filename);













#endif // VASSILIS_CHAMFER_DISTANCES_H
