#include "stdafx.h"

#include "chamfer_matching.h"
#include "basics/image_algorithms.h"
#include "basics/image_templates.h"
#include "algo_templates.h"
#include "opencv_utils.h"
#include "hausdorff_matching.h"

#include "basics/definitions.h"



// query is the points and dts of all rotated versions of the query.
// database is the points and dts of all database frames. We return
// a matrix that matches each rotation of the query to each database frame.
// The order of the scores is: 
// query rotation 0 with frame 0
// query rotation 1 with frame 0
// ...
// query rotation rotations-1 with frame 0
// query rotation 0 with frame 1
// 
// and so on.
vMatrix<float> vChamferDistances2(vChamferFeatures query, 
                                    vChamferFeatures database)
{
  long frames = database.point_sequence->NumberOfFrames();
  long orientations = query.point_sequence->NumberOfFrames();
  vPoseSet all_poses = vHandDatabase1::AllPoses(frames, orientations);
  return vChamferDistances3(query, database, all_poses);
}


// Same as vChamferDistances2, but it takes an additional argument, 
// which is the set of database poses that the query should be matched
// with. This is important if we only want to match the query to, say,
// 1000 database poses.
vMatrix<float> vChamferDistances3(vChamferFeatures query, 
                                    vChamferFeatures database,
                                    vPoseSet poses)
{
  long frames = database.point_sequence->NumberOfFrames();
  long orientations = query.point_sequence->NumberOfFrames();
  long number = poses.Size();
  vMatrix<float> result_matrix(1, number);
  vArray(float) result = result_matrix.Matrix();

  long i;
  for (i = 0; i < number; i++)
  {
    vDatabasePose & pose = poses.Pose(i);

    vArray(uchar) query_rows = query.point_sequence->PointRows(pose.rotation);
    vArray(uchar) query_cols = query.point_sequence->PointCols(pose.rotation);
    long query_points = query.point_sequence->Length(pose.rotation);
    if (query_points == 0) 
    {
      result[i] = 0;
      continue;
    }
    v3dMatrix<uchar> * query_dt = query.distance_transforms[pose.rotation];
    vArray2(uchar) query_dt2 = query_dt->Matrix2(0);
    
    vArray(uchar) database_rows = database.point_sequence->PointRows(pose.frame);
    vArray(uchar) database_cols = database.point_sequence->PointCols(pose.frame);
    long database_points = query.point_sequence->Length(pose.rotation);
    if (database_points == 0)
    {
      exit_error("vChamferDistances3: error 1\n");
    }
    v3dMatrix<uchar> * database_dt = database.distance_transforms[pose.frame];
    vArray2(uchar) database_dt2 = database_dt->Matrix2(0);

    long j;

    float im_score = 0; // image-to-model score;
    for (j = 0; j < query_points; j++)
    {
      im_score += database_dt2[query_rows[j]][query_cols[j]];
    }

    float mi_score = 0; // model-to-image score;
    for (j = 0; j < database_points; j++)
    {
      mi_score += query_dt2[database_rows[j]][database_cols[j]];
    }
    result[i] = im_score / ((float) query_points) + mi_score / ((float) database_points);
  }

  return result_matrix;
}


vMatrix<float> * vChamferEmbedding(grayscale_image * object, 
                                     vArray(v3dMatrix<float> *) dts,
                                     vPointSetSequence * points,
                                     v3dMatrix<float> * reference_images,
                                     long start, long dimensions)
{
  long i;
  if (start < 0) start = 0;
  long end = start + dimensions - 1;
  long number_of_references = reference_images->Rows();
  if (end >= number_of_references) end = number_of_references - 1;
  dimensions = end - start + 1;

  grayscale_image * scaled = 
    (grayscale_image *) vScaleBoundingBox6(object, 192, 128, 128, 
                                        256, 256);

  vArray(long) object_rows;
  vArray(long) object_cols;
  long object_number;
  vGetNonZeroPoints4(scaled, &object_rows, &object_cols, &object_number);

  v3dMatrix<float> * object_dt_matrix = vCVDistanceTransform(scaled);
  vArray2(float) object_dt = object_dt_matrix->Matrix2(0);

  vMatrix<float> * result_matrix = new vMatrix<float>(1, dimensions);
  vArray(float) result = result_matrix->Matrix();
  vArray2(float) references = reference_images->Matrix2(0);

  for (i = start; i <= end; i++)
  {
    long frame = vRound(references[i][0]);
    long rotation = vRound(references[i][1]);
    vArray(long) model_rows;
    vArray(long) model_cols;
    long model_number = points->GetPoints(frame, rotation, 
                                          &model_rows, &model_cols);
      float mi_distance = (float) vChamferFromDT(model_rows, model_cols, 
                                          model_number, object_dt_matrix);
      float im_distance = (float) vChamferFromDT(object_rows, object_cols,
                                          object_number, dts[i]);
      result[i - start] = mi_distance + im_distance;
    vdelete2(model_rows);
    vdelete2(model_cols);
  }

  vdelete2(object_rows);
  vdelete2(object_cols);
  vdelete(scaled);
  vdelete(object_dt_matrix);
  return result_matrix;;
}


/////////////////////////////////////////////////////////////////////
//                                                                 //
//      Beginning of implementation of vAllChamferDistances       //
//                                                                 //
/////////////////////////////////////////////////////////////////////


vAllChamferDistances::vAllChamferDistances(vRLE_Sequence * database)
{
  sequence_name = vCopyString(database->GetFilename());
  directory = Directory();
  base_name = BaseName();
  sequence1_name = 0;
  sequence2_name = 0;
  number_of_sequences = 1;
}


// All that we do in this constructor (and the previous one) is
// set the various strings right (sequence_name, directory, base_name).
vAllChamferDistances::vAllChamferDistances(vRLE_Sequence * database1, 
                                             vRLE_Sequence * database2)
{
  number_of_sequences = 2;
  sequence1_name = vCopyString(database1->GetFilename());
  sequence2_name = vCopyString(database2->GetFilename());
  // Make sure that the sequences are ordered alphabetically.
  if (strcmp(sequence1_name, sequence2_name) > 0)
  {
    char * temp = sequence1_name;
    sequence_name = sequence2_name;
    sequence2_name = temp;
  }

  char * parent = vParentDir(sequence1_name, platform_directory_slash);
  char * simple1 = vSimpleName(sequence1_name, platform_directory_slash);
  char * simple2 = vSimpleName(sequence2_name, platform_directory_slash);
  long length1 = strlen(simple1);
  long length2 = strlen(simple2);
  char * buffer = new char[length1 + length2 + 100];

  long index = 0;
  long i;
  for (i = 0; i < length1; i++)
  {
    char c = simple1[i];
    if ((c == '.') || (c == '_'))
    {
      break;
    }
    buffer[index++] = c;
  }
  
  buffer[index++] = '_';

  for (i = 0; i < length2; i++)
  {
    char c = simple2[i];
    if ((c == '.') || (c == '_'))
    {
      break;
    }
    buffer[index++] = c;
  }
  buffer[index++] = 0;
  
  sequence_name = vJoinPaths(parent, buffer);
  directory = Directory();
  base_name = BaseName();
  vdelete2(simple1);
  vdelete2(simple2);
  vdelete2(buffer);
  vdelete2(parent);
  vPrint("base_name = %s\n", base_name);
}

  
vAllChamferDistances::~vAllChamferDistances()
{
  vdelete2(sequence_name);
  vdelete2(sequence1_name);
  vdelete2(sequence2_name);
  vdelete2(directory);
  vdelete2(base_name);
}


char * vAllChamferDistances::Directory()
{
  char * result = vParentDir(sequence_name, platform_directory_slash);
  return result;
}


char * vAllChamferDistances::BaseName()
{
  char * simple = vSimpleName(sequence_name, platform_directory_slash);
  char * simple_base = vMergeStrings2(simple, ".distances");
  char * result = vJoinPaths3(directory, simple_base, "distance");
  
  vdelete2(simple);
  vdelete2(simple_base);
  return result;
}


// Order is 0 when we only have one sequence, 1 for sequence2->sequence1
// distances, 2 for sequence1->sequence2 distances.
char * vAllChamferDistances::DirectedName(long frame, long view, long order)
{
  if ((order < 0) || (order > 2)) return 0;
  if ((order == 0) && (number_of_sequences == 2)) return 0;
  if ((order > 0) && (number_of_sequences == 1)) return 0;

  char * order_string = vLongToString(order);
  char * frame_string = vLongToString(frame);
  char * result;

  if (order == 0)
  {
    result = vMergeStrings3(base_name, "_directed.", frame_string);
  }
  else
  {
    result = vMergeStrings5(base_name, "_directed", 
                             order_string, ".", frame_string);
  }

  vdelete2(frame_string);
  vdelete2(order_string);
  return result;
}


char * vAllChamferDistances::UndirectedName(long frame, long view)
{
  char * frame_string = vLongToString(frame);
  char * result = vMergeStrings3(base_name, ".", frame_string);
  vdelete2(frame_string);
  return result;
}


long vAllChamferDistances::SaveAllDirectedDistances(vRLE_Sequence * database,
                                                     vPointSetSequence * points)
{
  long start = 0;
  long end = database->NumberOfFrames() - 1;
  return SaveDirectedDistances(database, points, start, end);
}


long vAllChamferDistances::SaveDirectedDistances(vRLE_Sequence * database,
                                                  vPointSetSequence * points,
                                                  long start, long end)
{
  long order = -1;
  if (number_of_sequences == 1) 
  {
    order = 0;
  }
  else if (strcmp(database->GetFilename(), sequence1_name) == 0)
  {
    order = 1;
  }
  else if (strcmp(database->GetFilename(), sequence2_name) ==0 )
  {
    order = 2;
  }
  else
  {
    return 0;
  }

  vPrint("order = %li\n", order);
  long number_of_frames = database->NumberOfFrames();
  const long number_of_rotations = 4;
  long number_of_poses = number_of_rotations * number_of_frames;
  if (end < 0) end = number_of_frames - 1;

  // Prepare arguments on which we will call vOneWayChamfer2:
  vMatrix<long> frames(1, number_of_poses);
  vMatrix<long> rotations(1, number_of_poses);
  long i;
  for (i = 0; i < number_of_poses; i++)
  {
    frames(i) = i / number_of_rotations;
    rotations(i) = i % number_of_rotations;
  }

  vPrint("\n");
  for (i = start; i <= end; i++)
//  for (i = start; i <= -1; i++)
  {
    // Read image from the database of edge images, 
    // compute its distance tranform.
    grayscale_image * edge1 = database->ReadFrameGray(i);
    v3dMatrix<float> * dt1 = vCVDistanceTransform(edge1);
    vdelete(edge1);

    // Now, for each database image, compute its directed chamfer distance
    // to dt1.
    vMatrix<double> frame_distances = vOneWayChamfer2(dt1, points, 
                                                        &frames, &rotations);
    vdelete(dt1);

    v3dMatrix<float> float_distances(&frame_distances);
    char * output_file = DirectedName(i, 0, order);
    long success = float_distances.Write(output_file);
    if (success <= 0)
    {
      vPrint("Failed to save to %s\n", output_file);
      vdelete2(output_file);
      return success;
    }

    vdelete2(output_file);
    vPrint("Done with frame %li of %li\r", i, end);
  }

  vPrint("\n");
  return 1;
}


long vAllChamferDistances::SaveDistances(long first_frame, long last_frame,
                                          long chunk)
{
  long order1, order2;
  if (number_of_sequences == 1)
  {
    order1 = 0;
    order2 = 0;
  }
  else
  {
    order1 = 1;
    order2 = 2;
  }

  v3dMatrix<float> * junk = DirectedDistances(0, 0, order1);
  long number_of_poses = junk->Size();
  vdelete(junk);
  long number_of_frames = number_of_poses / 4;
  if (last_frame < 0) last_frame = number_of_frames - 1;

  // Will be set to zero or less if something goes wrong.
  long result = 1;

  long start;
  vMatrix<float> original(1, number_of_poses);
  vArray(v3dMatrix<float> *) matrix_cache = 
    vnew(v3dMatrix<float> *, chunk);
  vArray2(float) cache = vnew(vArray(float), chunk);
  for (start = first_frame; start <= last_frame; start = start + chunk)
  {
    long end = start + chunk - 1;
    if (end > last_frame) end = last_frame;

    // Load into the cache
    long load_flag = 0;  // Will be set to 1 if at least one
                         // matrix gets loaded.
    long number_loaded = end - start + 1;
    long i;
    for (i = 0; i < number_loaded; i++)
    {
      matrix_cache[i] = DirectedDistances(start + i, 0, order1);
      if (matrix_cache[i] == 0)
      {
        vPrint("Failed to load %li into cache\n", start + i);
        cache[i] = vZero(float);
        result = 0;
      }
      else
      {
        cache[i] = matrix_cache[i]->Matrix();
        load_flag = 1;
      }
    }
    vPrint("Loaded into cache %li to %li\n", start, end);
    if (load_flag == 0) continue;

    long frame, view;
    // Now load all the other matrices, and add the information.
    vPrint("\n");
    for (frame = 0; frame < number_of_frames; frame++)
    {
      v3dMatrix<float> * distances_matrix = DirectedDistances(frame, 0, order2);
      if (distances_matrix == 0) 
      {
        result = -1;
        continue;
      }
      vArray(float) distances = distances_matrix->Matrix();
      vCopyImage(&original, distances_matrix);
      for (view = 0; view < 4; view++)
      {
        vCopyImage(distances_matrix, &original);
        ConvertToView(distances_matrix, view);
        long index = frame * 4 + view;

        for (i = 0; i < number_loaded; i++)
        {
          if (matrix_cache[i] == 0) continue;
          long index2 = 4 * (i + start);
          cache[i][index] = cache[i][index] + distances[index2];
        }
      }
      vdelete(distances_matrix);
      vPrint("processed frame %li of %li\r", frame+1, number_of_frames);
    }
    vPrint("\n");

    // Save the cache into files and delete it.
    for (i = 0; i < number_loaded; i++)
    {
      if (matrix_cache[i] == 0) continue;
      long frame = start + i;
      char * filename = UndirectedName(frame, 0);
      long success = matrix_cache[i]->Write(filename);
      if (success <= 0)
      {
        vPrint("Failed to save to %s\n", filename);
        result = -2;
      }
      vdelete2(filename);
      vdelete(matrix_cache[i]);
    }
  }
  vdelete2(matrix_cache);
  vdelete2(cache);

  return result;
}


v3dMatrix<float> * vAllChamferDistances::DirectedDistances(long frame, 
                                                              long view, long order)
{
  if (number_of_sequences == 1) order = 0;
  char * input_file = DirectedName(frame, view, order);
  if (input_file == 0) return 0;
  v3dMatrix<float> * result = v3dMatrix<float>::Read(input_file);
  vdelete2(input_file);
  ConvertToView(result, view);
  return result;}


v3dMatrix<float> * vAllChamferDistances::Distances(long frame, 
                                                      long view)
{
  char * input_file = UndirectedName(frame, view);
  v3dMatrix<float> * result = v3dMatrix<float>::Read(input_file);
  vdelete2(input_file);
  ConvertToView(result, view);
  return result;
}


long vAllChamferDistances::ConvertToView(v3dMatrix<float> * distances, 
                                          long view)
{
  if (distances == 0) return 0;
  if ((view < 0) || (view > 3))
  {
    return 0;
  }
  if (view == 0) return 1;

  long size = distances->Size();
  vArray(float) values = distances->Matrix();

  float temp[4];
  long i, j;
  for (i = 0; i < size; i = i + 4)
  {
    for (j = 0; j < 4; j++)
    {
      temp[j] = values[i+j];
    }
    for (j = 0; j < 4; j++)
    {
      values[i+j] = temp[(j + 4 - view) % 4];
    }
  }
  return 1;
}


// For the (frame, view) database image, find the k-th closest database
// image (besides itself) based on the chamfer distance. It stores
// the frame and view of the nearest neighbor in the output arguments,
// and returns the distance.
float vAllChamferDistances::NearestNeighbor(long frame, long view, long k,
                                             long * nn_frame, long * nn_view)
{
  v3dMatrix<float> * distances_matrix = Distances(frame, view);
  float min_distance = NearestNeighbor(frame, view, k, nn_frame, nn_view,
                                       distances_matrix);
  vdelete(distances_matrix);
  return 0;
}


// Here we also pass the distances from (frame, view) to
// all database images, so that when we already have these
// distances available we call this function, and we don't
// need to read the distances from disk.
float vAllChamferDistances::NearestNeighbor(long frame, long view, long k,
                                             long * nn_frame, long * nn_view,
                                             v3dMatrix<float> * distances_matrix)
{
  if (distances_matrix == 0) return -1;
  const long number_of_rotations = 4;

  long k_index = 0;
  // We need to pass in k+1 to omit the (frame, view) itself.
  float min_distance = kth_smallest_cb(k+1, distances_matrix, &k_index);

  *nn_frame = k_index / number_of_rotations;
  *nn_view = k_index % number_of_rotations;
  return min_distance;
}


// For (frame, view), figure out how many database images have embeddings
// whose weighted distance to the embedding of (frame, view) is 
// smaller than the distance of the embedding of (frame2, view2) to
// the embedding of (frame, view).
long vAllChamferDistances::ApproximateRank(long frame, long view,
                                            long frame2, long view2,
                                            v3dMatrix<float> * reference_images,
                                            v3dMatrix<float> * database_embeddings,
                                            long start, long dimensions)
{
  const long number_of_rotations = 4;
  if (start < 0) return -1;
  if (dimensions <= 0) return 1;
  long index1 = frame * number_of_rotations + view;
  long index2 = frame2 * number_of_rotations + view2;
  long number = database_embeddings->Rows();
  long end = start + dimensions - 1;
  if (end >= database_embeddings->Cols()) return -1;

  vArray2(float) embeddings = database_embeddings->Matrix2(0);
  vArray2(float) references = reference_images->Matrix2(0);
  vArray(float) embedding1 = embeddings[index1];
  vArray(float) embedding2 = embeddings[index2];

  // Calculate the distance betweent the two embeddings.
  long i;
  float distance12 = 0;
  for (i = start; i <= end; i++)
  {
    float diff = vAbs(embedding1[i] - embedding2[i]);
    float weight = references[i][2];
    distance12 = distance12 + weight * diff;
  }

  // Count number of database images whose embeddings
  // are closer than distance12 to the embedding of (frame1, view1).
  long counter = 0;
  for (i = 0; i < number; i++)
  {
    if (i == index1) continue;
    if (i == index2) continue;
    float distance = 0;
    long j;
    for (j = start; j <= end; j++)
    {
      float diff = vAbs(embedding1[j] - embeddings[i][j]);
      float weight = references[j][2];
      distance = distance + weight * diff;
    }
    if (distance < distance12)
    {
      counter = counter+1;
    }
  }

  return counter;
}


// For (frame, view), first call NearestNeighbor to get its
// k-nearest neighbor, and then call ApproximateRank to get
// the rank of that neighbor using the embedding. That represents
// the "loss" incurred by using the embedding instead of the
// original distance.
long vAllChamferDistances::ApproximateLoss(long frame, long view, long k,
                                            v3dMatrix<float> * reference_images,
                                            v3dMatrix<float> * database_embeddings,
                                            long start, long dimensions)
{
  long nn_frame = 0, nn_view = 0;
  NearestNeighbor(frame, view, k, &nn_frame, &nn_view);
  long result = ApproximateRank(frame, view, nn_frame, nn_view, 
                                reference_images, database_embeddings,
                                start, dimensions);
  return result;
}


// Pick "times" random database images, and for each of them call
// ApproximateLoss and save the result into ranks. This is a way
// to get quick statistics on how well the embedding preserves
// nearest neighbors of database images.
long vAllChamferDistances::ApproximateLossStats(long k, v3dMatrix<float> * reference_images,
                                                 v3dMatrix<float> * database_embeddings,
                                                 long start, long dimensions,
                                                 long times, vector<long> * ranks)
{
  const long number_of_rotations = 4;
  long number = database_embeddings->Rows();

  long i;
  vPrint("\n");
  for (i = 0; i < times; i++)
  {
    long database_pick = vRandomLong2(0, number-1);
    long frame = database_pick / number_of_rotations;
    long view = database_pick % number_of_rotations;
    long result = ApproximateLoss(frame, view, k, reference_images,
                                  database_embeddings, start, dimensions);
    ranks->push_back(result);
    vPrint("Done with step %li of %li\r", i+1, times);
  }
  vPrint("\n");

  std::sort(ranks->begin(), ranks->end(), less<long>());
  return 1;
}


// returns the size of the hand database, 
// i.e. number of frames * number of views.
long vAllChamferDistances::NumberOfObjects()
{
  v3dMatrix<float> * distances = Distances(0, 0);
  if (distances == 0)
  {
    return 0;
  }
  else
  {
    long result = distances->Size();
    vdelete(distances);
    return result;
  }
}


/////////////////////////////////////////////////////////////////////
//                                                                 //
//         End of implementation of vAllChamferDistances          //
//                                                                 //
/////////////////////////////////////////////////////////////////////



/////////////////////////////////////////////////////////////////////
//                                                                 //
//      Beginning of implementation of vChamferBoostInfo          //
//                                                                 //
/////////////////////////////////////////////////////////////////////


vChamferBoostInfo::vChamferBoostInfo()
{
  object_number = 0;
  training_number = 0;
  sampling_rate = 0;
  sequence_name = 0;
  boost_map = 0;
  sparse_map = 0;
  fast_map = 0;
  max_distance = 0;
  original_max_distance = 0;
}


vChamferBoostInfo::vChamferBoostInfo(const char * in_sequence_name, long in_sampling)
{
  sequence_name = vCopyString(in_sequence_name);
  sampling_rate = in_sampling;
  boost_map = 0;
  sparse_map = 0;
  fast_map = 0;
  max_distance = 0;
  original_max_distance = 0;
  long success = Load();
  is_valid = success;
}

  
vChamferBoostInfo::vChamferBoostInfo(vRLE_Sequence * database, 
                                       vPointSetSequence * points,
                                       long in_sampling,
                                       long in_training)
{
  boost_map = 0;
  sparse_map = 0;
  fast_map = 0;
  sampling_rate = in_sampling;
  training_number = in_training;
  max_distance = 0;
  long frames = database->NumberOfFrames();
  long poses = frames * 4;
  sequence_name = vCopyString(database->GetFilename());
  object_number = poses / sampling_rate;

  PickIndices(frames, 4, object_number);
  ComputeDistances(database, points);
  PickTraining(training_number);
  original_max_distance = CurrentMaxDistance();
}


vChamferBoostInfo::~vChamferBoostInfo()
{
  vdelete2(sequence_name);
  vdelete(boost_map);
  vdelete(sparse_map);
  vdelete(fast_map);
}


float vChamferBoostInfo::CurrentMaxDistance()
{
  long i, j;
  if (object_number <= 0) return -1;
  float current_max = distances[0][0];
  for (i = 0; i < object_number; i++)
  { 
    for (j = 0; j <= i; j++)
    {
      float current = distances[i][j];
      if (current > current_max)
      {
        current_max = current;
      }
    }
  }

  return current_max;
}


char * vChamferBoostInfo::Filename()
{
  char * name1 = vReplaceEnding2(sequence_name, ".boost_info");
  char * sampling_string = vLongToString(sampling_rate);
  char * name2 = vMergeStrings2(name1, sampling_string);
  
  char * result = name2;
  vdelete2(name1);
  vdelete2(sampling_string);
  return result;
}


long vChamferBoostInfo::PickIndices(long frames, long rotations, long number)
{
  if (number >= frames * rotations)
  {
    exit_error("Error: PickIndices(%li, %li, %li)\n", 
                    frames, rotations, number);
  }

  // picked[frame][rotation] will store whether that pose has been
  // picked yet.
  vMatrix<uchar> picked_matrix(frames, rotations);
  vWriteValue(&picked_matrix, (uchar) 0);
  vArray2(uchar) picked = picked_matrix.Matrix2();

  indices_matrix = vMatrix<long>(number, 2);
  indices = indices_matrix.Matrix2();

  long i;
  for (i = 0; i < number; i++)
  {
    long frame, rotation;
    // Pick a frame and a rotation that have not been picked yet.
    do
    {
      frame = vRandomLong2(0, frames-1);
      rotation = vRandomLong2(0, rotations-1);
    }
    while (picked[frame][rotation] != 0);

    picked[frame][rotation] = 1;
    indices[i][0] = frame;
    indices[i][1] = rotation;
  }

  return 1;
}


long vChamferBoostInfo::ComputeDistances(vRLE_Sequence * database,
                                          vPointSetSequence * points)
{
  if (object_number == 0) return 0;
  distances_matrix = vMatrix<float>(object_number, object_number);
  distances = distances_matrix.Matrix2();

  // Prepare arguments on which we will call vOneWayChamfer2:
  vMatrix<long> frames(1, object_number);
  vMatrix<long> rotations(1, object_number);
  long i, j;
  for (i = 0; i < object_number; i++)
  {
    frames(i) = indices[i][0];
    rotations(i) = indices[i][1];
  }

  vPrint("\n");
  for (i = 0; i < object_number; i++)
  {
    // Read image from the database of edge images, 
    // compute its distance tranform.
    long frame1 = indices[i][0];
    long rotation1 = indices[i][1];
    grayscale_image * edge1 = database->ReadFrameGray(frame1);
    v3dMatrix<float> * dt = vCVDistanceTransform(edge1);
    v3dMatrix<float> * dt1 = vRotateRight(dt, rotation1 * 90);
    vdelete(dt);
    vdelete(edge1);

    // Now, for each database image, compute its directed chamfer distance
    // to dt1.
    vMatrix<double> frame_distances = vOneWayChamfer2(dt1, points, 
                                                       &frames, &rotations);
    vdelete(dt1);

    vArray(double) distances_array = frame_distances.Matrix();
    for (j = 0; j < object_number; j++)
    {
      distances[i][j] = (float) distances_array[j];
    }

    vPrint("Done with frame %li of %li\r", i+1, object_number);
  }

  for (i = 0; i < object_number; i++)
  {
    for (j = 0; j <= i; j++)
    {
      float ij = distances[i][j];
      float ji = distances[j][i];
      float sum = ij + ji;
      distances[i][j] = sum;
      distances[j][i] = sum;
    }
  }

  vPrint("\n");
  return 1;
}


long vChamferBoostInfo::PickTraining(long in_training)
{
  if (object_number == 0)
  {
    exit_error("Error: PickTraining called with zero objects\n");
  }
  if (distances_matrix.Rows() != object_number)
  {
    exit_error("Error: uninitialized distances in PickTraining\n");
  }

  training_number = in_training;
  triples_matrix = vMatrix<long>(training_number, 3);
  triples = triples_matrix.Matrix2();
  
  long i;
  for (i = 0; i < training_number; i++)
  {
    long q_index = vRandomLong2(0, object_number - 1);
    long a_index = vRandomLong2(0, object_number - 1);
    long b_index = vRandomLong2(0, object_number - 1);

    // Make sure the distances are different.
    float qa = distances[q_index][a_index];
    float qb = distances[q_index][b_index];
    float ab = distances[a_index][b_index];

    if ((vAbs(qa - qb) < 0.000001) ||
        (vAbs(qa - ab) < 0.000001) ||
        (vAbs(qb - ab) < 0.000001))
    {
      i--;
    }
    else
    {
      triples[i][0] = q_index;
      triples[i][1] = a_index;
      triples[i][2] = b_index;
    }
  }

  return 1;
}


vMatrix<long> vChamferBoostInfo::Indices()
{
  return indices_matrix;
}


vMatrix<float> vChamferBoostInfo::Distances()
{
  return distances_matrix;
}


vMatrix<long> vChamferBoostInfo::Triples()
{
  return triples_matrix;
}


long vChamferBoostInfo::Save()
{
  char * filename = Filename();
  FILE * fp = fopen(filename, vFOPEN_WRITE);
  if (fp == 0)
  {
    vPrint("Failed to open %s\n", filename);
    vdelete2(filename);
    return 0;
  }

  long success = 1;
  long temp_success;
  
  temp_success = vWritePCString(fp, sequence_name);
  if (temp_success <= 0) success = 0;

  temp_success = vWritePCLongs(fp, &object_number, 1);
  if (temp_success <= 0) success = 0;

  temp_success = vWritePCLongs(fp, &training_number, 1);
  if (temp_success <= 0) success = 0;

  temp_success = vWritePCLongs(fp, &sampling_rate, 1);
  if (temp_success <= 0) success = 0;

  temp_success = indices_matrix.WriteBandwise(fp);
  if (temp_success <= 0) success = 0;

  temp_success = distances_matrix.WriteBandwise(fp);
  if (temp_success <= 0) success = 0;

  temp_success = triples_matrix.WriteBandwise(fp);
  if (temp_success <= 0) success = 0;
  fclose(fp);

  if (success == 0)
  {
    vPrint("Failed to save to %s\n", filename);
  }
  vdelete2(filename);
  return success;
}


long vChamferBoostInfo::Load()
{
  char * filename = Filename();
  FILE * fp = fopen(filename, vFOPEN_READ);
  if (fp == 0)
  {
    vPrint("Failed to load from %s\n", filename);
    vdelete2(filename);
    return 0;
  }

  long success = 1;
  long temp_success;

  char * temp_sequence_name = vReadPCString(fp);
  if (temp_sequence_name == 0) success = 0;

  // The sequence names should be the same, except perhaps for
  // the initial drive.
  else if (strcmp(temp_sequence_name+1, sequence_name+1) != 0) success = 0;
  vdelete2(temp_sequence_name);

  temp_success = vReadPCLongs(fp, &object_number, 1);
  if (temp_success <= 0) success = 0;

  temp_success = vReadPCLongs(fp, &training_number, 1);
  if (temp_success <= 0) success = 0;

  temp_success = vReadPCLongs(fp, &sampling_rate, 1);
  if (temp_success <= 0) success = 0;

  indices_matrix = vMatrix<long>(object_number, 2);
  temp_success = indices_matrix.ReadBandwise(fp);
  if (temp_success <= 0) success = 0;

  distances_matrix = vMatrix<float>(object_number, object_number);
  temp_success = distances_matrix.ReadBandwise(fp);
  if (temp_success <= 0) success = 0;

  triples_matrix = vMatrix<long>(training_number, 3);
  temp_success = triples_matrix.ReadBandwise(fp);
  if (temp_success <= 0) success = 0;
  fclose(fp);

  if (success == 0)
  {
    vPrint("Failed to load from %s\n", filename);
  }
  vdelete2(filename);

  indices = indices_matrix.Matrix2();
  distances = distances_matrix.Matrix2();
  triples = triples_matrix.Matrix2();
  original_max_distance = CurrentMaxDistance();

  return success;
}


long vChamferBoostInfo::PrintIndices()
{
  indices_matrix.Print("indices");
  return 1;
}


long vChamferBoostInfo::PrintTriples()
{
  triples_matrix.Print("triples");
  return 1;
}


long vChamferBoostInfo::PrintDistances()
{
  distances_matrix.Print("distances");
  return 1;
}


long vChamferBoostInfo::MakeBoostMap()
{
  if (object_number == 0) return 0;
  if (training_number == 0) return 0;

  NormalizeDistances();
//  vMatrix<long> candidates = boost_map.AllCandidates(object_number);
//  boost_map = V_BoostMap(distances_matrix, triples_matrix, candidates);
  
  long candidate_range = object_number / 3;
  long training_range = object_number / 3;
  long number_of_picked_candidates = Min(candidate_range, (long) 1000);
  long validation_number = training_number;
  vdelete(boost_map);
  boost_map = new V_BoostMap3(distances_matrix, candidate_range,
                              training_range, number_of_picked_candidates,
                              training_number, validation_number);

  return 1;
}


V_BoostMap3 * vChamferBoostInfo::BoostMap()
{
  return boost_map;
}


char * vChamferBoostInfo::ClassifierDirectory()
{
  const char * directory = "r:\\text\\boost_map";
//  const char * directory = "c:\\users\\athitsos\\sequences\\glove_data\\training\\boost_map";
//  const char * directory = "d:\\users\\athitsos\\text\\boost_map";
  return vCopyString(directory);
}

  
char * vChamferBoostInfo::ClassifierFilename(const char * base, long number)
{
  char * directory = ClassifierDirectory();
  char * step_string = vLongToString(number);
  char * simple_name = vMergeStrings3(base, ".", step_string);
  char * filename = vJoinPaths(directory, simple_name);
  vdelete2(directory);
  vdelete2(step_string);
  vdelete(simple_name);
  return filename;
}


long vChamferBoostInfo::SaveClassifier(const char * base)
{
  if (boost_map == 0) return 0;
  long steps = boost_map->StepsDone();
  if (steps == 0) return 0;
  char * filename = ClassifierFilename(base, steps);
  vPrint("filename = %s\n", filename);

  vector<long> boost_indices;
  vector<float> weights;
  V_BoostMap * bm = (V_BoostMap *) boost_map;
  bm->CleanClassifier(&boost_indices, &weights);
  
  long number_of_references = boost_indices.size();
  vMatrix<float> references(number_of_references, 8);
  long i;
  for (i = 0; i < number_of_references; i++)
  {
    long index = boost_indices[i];
    long frame = indices[index][0];
    long rotation = indices[index][1];
    references(i, 0) = (float) frame;
    references(i, 1) = (float) rotation;
    references(i, 2) = weights[i];
    references(i, 3) = (float) index;

    // With respect to training error, z, and validation error,
    // note that these are saved for each training round. However,
    // the number of reference images that we save is smaller than
    // the number of training rounds, and since there are multiple
    // training rounds corresponding to each reference image, we 
    // cannot store information for each round. Instead, we will
    // store information for each of the last "number_of_references"
    // rounds, and the information will not necessarily be associated
    // with the actual reference image with which it is stored. At
    // any rate, this information is used just for my curiosity, it
    // is not used by the system.
    long j = steps - number_of_references + i;
    references(i, 4) = boost_map->StepTrainingError(j);
    references(i, 5) = boost_map->StepZ(j);
    references(i, 6) = boost_map->StepValidationError(j);
    references(i, 7) = (float) i;
  }

  long success = references.WriteText(filename);
  vdelete2(filename);
  return success;
}


long vChamferBoostInfo::LoadClassifier(const char * base, long number)
{
  if (boost_map == 0) return 0;
  char * filename = ClassifierFilename(base, number);
  v3dMatrix<float> * reference_images = 
    v3dMatrix<float>::ReadText(filename);
  if (reference_images == 0) 
  {
    vPrint("Failed to load from %s\n", filename);
    vdelete2(filename);
    return 0;
  }
  vdelete2(filename);

  long steps = reference_images->Rows();
  long cols = reference_images->Cols();
  vArray2(float) references = reference_images->Matrix2(0);

  long step;
  V_BoostMap * bm = (V_BoostMap *) boost_map;
  for (step = 0; step < steps; step++)
  {
    long index = (long) references[step][3];
    float alpha = references[step][2];
    bm->AddClassifier(index, alpha);
    bm->PrintSummary();
  }

  vdelete(reference_images);
  return 1;
}


long vChamferBoostInfo::FindIndex(long frame, long rotation)
{
  long index = -1;
  long i;
  for (i = 0; i < object_number; i++)
  {
    if ((frame == indices[i][0]) && (rotation = indices[i][1]))
    {
      index = i;
      break;
    }
  }
  return index;
}


// Return the frame/rotation info for this index.
long vChamferBoostInfo::GetPose(long index, long * frame, long * rotation)
{
  if ((index < 0) || (index >= object_number)) 
  {
    return -1;
  }

  *frame = indices[index][0];
  *rotation = indices[index][1];
  return 1;
}


float vChamferBoostInfo::MaxDistance()
{
  return max_distance;
}


long vChamferBoostInfo::MakeSparseMap(long number)
{
  if (object_number == 0) return 0;

  NormalizeDistances();
  long candidate_range = object_number / 3;
  if (number < 2) number = 2;
  if (number > candidate_range) number = candidate_range;
  long validation_range = object_number - 2 * candidate_range;
  long validation_number = training_number;
  vdelete(sparse_map);
  sparse_map = new vChamferSparseMap(distances_matrix, number,
                                      validation_range, validation_number, 
                                      &indices_matrix);

  return 1;
}


vChamferSparseMap * vChamferBoostInfo::SparseMap()
{
  return sparse_map;
}


long vChamferBoostInfo::MakeFastMap()
{
  if (object_number == 0) return 0;

  NormalizeDistances();
  long candidate_range = object_number / 3;
  long validation_range = object_number - 2 * candidate_range;
  long validation_number = training_number;
  vdelete(sparse_map);
  fast_map = new vChamferFastMap(distances_matrix, candidate_range,
                                  validation_range, validation_number, 
                                  &indices_matrix);

  return 1;
}


vChamferFastMap * vChamferBoostInfo::FastMap()
{
  return fast_map;
}


long vChamferBoostInfo::SetTrainingNumber(long in_number)
{
  if (in_number<= 0) return training_number;
  training_number = in_number;
  return training_number;
}


long vChamferBoostInfo::SetValidationNumber(long in_number)
{
  if (in_number<= 0) return validation_number;
  validation_number = in_number;
  return validation_number;
}


// make sure max distance is at most 1.
long vChamferBoostInfo::NormalizeDistances()
{
  // Find max_distance.
  max_distance = 0;
  long i, j;
  for (i = 0; i < object_number; i++)
  {
    for (j = 0; j <= i; j++)
    {
      float distance = distances[i][j];
      if (distance > max_distance)
      {
        max_distance = distance;
      }
    }
  }

  if (max_distance == 0)
  {
    exit_error("max_distance = 0\n");
  }

  // Normalize distances
  for (i = 0; i < object_number; i++)
  {
    for (j = 0; j < object_number; j++)
    {
      distances[i][j] = distances[i][j] / max_distance;
    }
  }
  return 1;
}

  
// This function takes in a matrix like the one returned 
// from V_BoostMap3::ClassifierMatrix, and returns the
// corresponding distances between pairs of points.
vMatrix<float> vChamferBoostInfo::PairDistances(vMatrix<float> classifiers)
{
  long number = classifiers.Rows();
  vMatrix<float> result(1, number);
  vWriteValue(&result, (float) -1);

  long i;
  for (i = 0; i < number; i++)
  {
    long type = vRound(classifiers(i, 0));
    if (type != 1) 
    {
      continue;
    }
    long index1 = vRound(classifiers(i, 1));
    long index2 = vRound(classifiers(i, 2));
    if (distances_matrix.check_bounds(index1, index2) == 0)
    {
      exit_error("Error: Bad indices in PairDistances: %li %li\n",
                      index1, index2);
    }
    
    float distance = distances[index1][index2] * original_max_distance;
    result(i) = distance;
  }

  return result;
}

  
  
/////////////////////////////////////////////////////////////////////
//                                                                 //
//         End of implementation of vChamferBoostInfo             //
//                                                                 //
/////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////
//                                                                 //
//           Start of implementation of vSignBoostInfo            //
//                                                                 //
/////////////////////////////////////////////////////////////////////


vSignBoostInfo::vSignBoostInfo()
{
  Zero();
}


vSignBoostInfo::vSignBoostInfo(long in_training_number)
{
  Zero();
  distances_matrix = ReadDistances();
  training_number = in_training_number;
  object_number = distances_matrix.Rows();
  distances = distances_matrix.Matrix2();

  ReadIndices();
  SetCandidateIndices();
  SetTraining();
  SetValidation();
}


vSignBoostInfo::~vSignBoostInfo()
{
  vdelete(boost_map);
  vdelete(sparse_map);
  vdelete(fast_map);
}


long vSignBoostInfo::Zero()
{
  object_number = 0;
  training_number = 0;
  max_distance = 0;

  boost_map = 0;
  sparse_map = 0;
  fast_map = 0;
  return 1;
}


char * vSignBoostInfo::Directory()
{
//  char * directory = "d:\\users\\athitsos\\sequences\\joni_dataset";
  char * directory = "d:\\users\\athitsos\\sequences\\joni_dataset";
  return vCopyString(directory);
}


vMatrix<float> vSignBoostInfo::Distances()
{
  return vMatrix<float>(&distances_matrix);
}


long vSignBoostInfo::NormalizeDistances()
{
  // Find max_distance.
  max_distance = 0;
  long i, j;
  for (i = 0; i < object_number; i++)
  {
    for (j = 0; j <= i; j++)
    {
      float distance = distances[i][j];
      if (distance > max_distance)
      {
        max_distance = distance;
      }
    }
  }

  if (max_distance == 0)
  {
    exit_error("max_distance = 0\n");
  }

  // Normalize distances
  for (i = 0; i < object_number; i++)
  {
    for (j = 0; j < object_number; j++)
    {
      distances[i][j] = distances[i][j] / max_distance;
    }
  }
  return 1;
}


long vSignBoostInfo::Save(const char * filename)
{
  return -1;
}


long vSignBoostInfo::Load()
{
  return -1;
}


long vSignBoostInfo::MakeBoostMap()
{
  if (object_number < 5) return 0;

  NormalizeDistances();
//  vMatrix<long> candidates = boost_map.AllCandidates(object_number);
//  boost_map = V_BoostMap(distances_matrix, triples_matrix, candidates);
  
  long number_of_picked_candidates = candidate_indices.Size();
  long validation_number = training_number;
  vdelete(boost_map);
  boost_map = new V_BoostMap3(distances_matrix, candidate_indices,
                              training_triples, validation_triples,
                              number_of_picked_candidates);
  boost_map->SetProjectionCandidateNumber(1000);
  return 1;
}


V_BoostMap3 * vSignBoostInfo::BoostMap()
{
  return boost_map;
}


long vSignBoostInfo::MakeSparseMap(long number)
{
  if (object_number < 5) return 0;

  NormalizeDistances();
  vdelete(sparse_map);

  // number specifies the number of objects we should use
  // to build the sets for the Bourgain embedding.

  if (number <= 1) number = 2;
  if (number >= candidate_indices.Size())
  {
    number = candidate_indices.Size();
  }

  vMatrix<long> sm_indices(1, number);
  vPrint("number = %li\n", number);
  long i;
  for(i = 0; i < number; i++)
  {
    sm_indices(i) = candidate_indices(i);
  }

  sparse_map = new V_SparseMap(distances_matrix, sm_indices,
                               validation_triples);

  return 1;
}


V_SparseMap * vSignBoostInfo::SparseMap()
{
  return sparse_map;
}


long vSignBoostInfo::MakeFastMap()
{
  if (object_number < 5) return 0;

  NormalizeDistances();
  vdelete(fast_map);
  fast_map = new V_FastMap(distances_matrix, candidate_indices,
                           validation_triples);

  return 1;
}

V_FastMap * vSignBoostInfo::FastMap()
{
  return fast_map;
}


// This group of functions should be identical to vChamferBoostInfo.
long vSignBoostInfo::PrintDistances()
{
  distances_matrix.Print("distances");
  return 1;
}


char * vSignBoostInfo::ClassifierDirectory()
{
  return vChamferBoostInfo::ClassifierDirectory();
}


char * vSignBoostInfo::ClassifierFilename(const char * base, long number)
{
  return vChamferBoostInfo::ClassifierFilename(base, number);
}


long vSignBoostInfo::SaveClassifier(const char * base)
{
  return 0;
}


long vSignBoostInfo::LoadClassifier(const char * base, long number)
{
  return 0;
}


float vSignBoostInfo::MaxDistance()
{
  return max_distance;
}


long vSignBoostInfo::SetTrainingNumber(long in_number)
{
  if (in_number <= 0) return 0;
  training_number = in_number;
  return training_number;
}


vMatrix<float> vSignBoostInfo::ReadDistances()
{
  char * directory = Directory();
  char * name = "signs_dist_mat.bin";
  char * filename = vJoinPaths(directory, name);
  vdelete2(directory);

  const long item_size = 8;
  long file_size = vFilesize(filename);
  long entries = file_size / 8;
  long number = vRound(sqrt((float) entries));

  if (number * number * item_size != file_size)
  {
    exit_error("Error: bad matrix in ReadDistances\n");
  }

  vMatrix<double> temp(number, number);
  long success = temp.ReadBandwise(filename, 0);
  if (success <= 0)
  {
    exit_error("Failed to read distances from %s\n",
                    filename);
  }

  vMatrix<float> result(number, number);
  vArray(float) r = result.Matrix();
  vArray(double) t = temp.Matrix();
  long i;
  for (i = 0; i < entries; i++)
  {
    r[i] = (float) t[i];
  }

  vdelete2(filename);
  return result;
}


vMatrix<long> vSignBoostInfo::ChooseTriples(vMatrix<long> query_indices,
                                              vMatrix<long> model_indices,
                                              long number)
{
  vMatrix<long> result_matrix(number, 3);
  vArray2(long) result = result_matrix.Matrix2();

  long query_number = query_indices.Size();
  long model_number = model_indices.Size();
  long i;
  for (i = 0; i < number; i++)
  {
    long q_index = vRandomLong2(0, query_number-1);
    long a_index = vRandomLong2(0, model_number-1);
    long b_index = vRandomLong2(0, model_number-1);
    if (a_index == b_index)
    {
      i--;
      continue;
    }

    long q = query_indices(q_index);
    long a = model_indices(a_index);
    long b = model_indices(b_index);
    result[i][0] = q;
    result[i][1] = a;
    result[i][2] = b;
  }

  return result_matrix;
}


long vSignBoostInfo::ReadIndices()
{
  char * directory = Directory();
  long success = 1;
  long temp_success;
  char * name;
  char * filename;

  ben_indices = vMatrix<long>(1, object_number);
  norma_indices = vMatrix<long>(1, object_number);
  ben_test_indices = vMatrix<long>(1, object_number);
  norma_test_indices = vMatrix<long>(1, object_number);
  class_labels = vMatrix<long>(1, object_number);
  test_labels = vMatrix<long>(1, object_number);

  name = "ben_ind2.bin";
  filename = vJoinPaths(directory, name);
  temp_success = ben_indices.ReadBandwise(filename, 0);
  if (temp_success <= 0)
  {
    success = 0;
    vPrint("Failed to read ben_indices\n");
  }
  vdelete2(filename);

  name = "norma_ind2.bin";
  filename = vJoinPaths(directory, name);
  temp_success = norma_indices.ReadBandwise(filename, 0);
  if (temp_success <= 0)
  {
    success = 0;
    vPrint("Failed to read norma_indices\n");
  }
  vdelete2(filename);

  name = "ben_ind.bin";
  filename = vJoinPaths(directory, name);
  temp_success = ben_test_indices.ReadBandwise(filename, 0);
  if (temp_success <= 0)
  {
    success = 0;
    vPrint("Failed to read ben_test_indices\n");
  }
  vdelete2(filename);

  name = "norma_ind.bin";
  filename = vJoinPaths(directory, name);
  temp_success = norma_test_indices.ReadBandwise(filename, 0);
  if (temp_success <= 0)
  {
    success = 0;
    vPrint("Failed to read norma_test_indices\n");
  }
  vdelete2(filename);

  name = "class.bin";
  filename = vJoinPaths(directory, name);
  temp_success = class_labels.ReadBandwise(filename, 0);
  if (temp_success <= 0)
  {
    success = 0;
    vPrint("Failed to read class_labels\n");
  }

  vdelete2(filename);
  vdelete2(directory);

  // Set values for test_labels.
  vWriteValue(&test_labels, (long) 0);
  long i;
  for (i = 0; i < object_number; i++)
  {
    long ben_i = ben_test_indices(i);
    long norma_i = norma_test_indices(i);
    if ((ben_i == 0) && (norma_i == 0)) continue;
    long label = class_labels(i);
    test_labels(label) = 1;
  }

  return success;
}


long vSignBoostInfo::SetCandidateIndices()
{
  long number = object_number / 3;
  // long number = 88;
  vector<long> candidates;

  candidate_chi = vMatrix<long>(1, object_number);
  vWriteValue(&candidate_chi, (long) 0);
  long counter = 0;
  long tries = 0;
  long i;
  while(1)
  {
    i = vRandomLong2(0, object_number-1);
//    long label = class_labels(i);
//    long is_test = test_labels(label);
    long label = class_labels(i);
    long is_test = test_labels(label);
    if (is_test == 0)
    {
      candidates.push_back(i);
      candidate_chi(i) = 1;
      counter++;
      if (counter == number) break;
    }
    tries++;
    if (tries == object_number) break;
  }

  candidate_indices = vMatrix<long>(1, counter);
  for (i = 0; i < counter; i++)
  {
    candidate_indices(i) = candidates[i];
  }

  return counter;
}


long vSignBoostInfo::SetTraining()
{
  // We will pick as query indices signs by Norma that
  // are not used for validation or as candidates, and that
  // do not have the same label as the norma_test_indices.
  vector<long> query_indices;
  long i;
  for (i = 0; i < object_number; i++)
  {
    if (norma_indices(i) == 0) continue;
    if (candidate_chi(i) != 0) continue;
    long label = class_labels(i);
    if (test_labels(label) != 0) continue;
    query_indices.push_back(i);
  }

  long query_number = query_indices.size();
  training_query_indices = vMatrix<long>(1, query_number);
  for (i = 0; i < query_number; i++)
  {
    training_query_indices(i) = query_indices[i];
  }

  // For the model indices we will pick signs by Ben that
  // are not used as candidates.
  vector<long> model_indices;
  for (i = 0; i < object_number; i++)
  {
    if (ben_indices(i) == 0) continue;
    if (candidate_chi(i) != 0) continue;
    model_indices.push_back(i);
  }

  long model_number = model_indices.size();
  training_model_indices = vMatrix<long>(1, model_number);
  for (i = 0; i < model_number; i++)
  {
    training_model_indices(i) = model_indices[i];
  }

  training_triples = ChooseTriples(training_query_indices, 
                                   training_model_indices,
                                   training_number);
  return 1;
}


long vSignBoostInfo::SetValidation()
{
  // We will pick as query indices signs by Norma that
  // belong to norma_test_indices.
  vector<long> query_indices;
  long i;
  for (i = 0; i < object_number; i++)
  {
    if (norma_test_indices(i) == 0) continue;
    query_indices.push_back(i);
  }

  long query_number = query_indices.size();
  validation_query_indices = vMatrix<long>(1, query_number);
  for (i = 0; i < query_number; i++)
  {
    validation_query_indices(i) = query_indices[i];
  }

  // For the model indices we will the same signs as
  // the ones we used for training
  validation_model_indices = training_model_indices;
  validation_triples = ChooseTriples(validation_query_indices, 
                                     validation_model_indices,
                                     training_number);

  return 1;
}


long vSignBoostInfo::Print()
{
  candidate_indices.Print("candidate_indices");
  training_query_indices.Print("training_query_indices");
  training_model_indices.Print("training_model_indices");
  validation_query_indices.Print("validation_query_indices");
  validation_model_indices.Print("validation_model_indices");
  
  long i;
  vPrint("training triples:\n");
  for (i = 0; i < 100; i++)
  {
    long index = vRandomLong2(0, training_number-1);
    long q = training_triples(index, 0);
    long a = training_triples(index, 1);
    long b = training_triples(index, 2);
    vPrint("index = %li: (%li, %li, %li)\n", index, q, a, b);
  }

  vPrint("\n\nvalidation triples:\n");
  for (i = 0; i < 100; i++)
  {
    long index = vRandomLong2(0, training_number-1);
    long q = validation_triples(index, 0);
    long a = validation_triples(index, 1);
    long b = validation_triples(index, 2);
    vPrint("index = %li: (%li, %li, %li)\n", index, q, a, b);
  }

  return 1;
}
  

// Compute embeddings, based on the current state of boost_map.
long vSignBoostInfo::ComputeBMEmbeddings()
{
  if (boost_map == 0) return 0;
  long dimensions = boost_map->Dimensions();
  if (dimensions <= 0) return 0;

  vMatrix<float> weights_matrix = boost_map->WeightsMatrix();
  vArray(float) weights = weights_matrix.Matrix();
  bm_embeddings_matrix = vMatrix<float>(object_number, dimensions);
  vArray2(float) embeddings = bm_embeddings_matrix.Matrix2();

  long i, j;
  for (i = 0; i < object_number; i++)
  {
    vMatrix<float> current_embedding = boost_map->Embedding1(i);
    vArray(float) current = current_embedding.Matrix();
    long current_dimensions = current_embedding.Size();
    if (current_dimensions != dimensions)
    {
      exit_error("Error: current_dimensions = %li, dimensions = %li\n",
                      current_dimensions, dimensions);
    }

    for (j = 0; j < dimensions; j++)
    {
      embeddings[i][j] = current[j];
    }
  }
  
  return 1;
}


long vSignBoostInfo::ComputeSMEmbeddings()
{
  if (sparse_map == 0) return 0;
  long dimensions = sparse_map->Dimensions();
  if (dimensions <= 0) return 0;

  sm_embeddings_matrix = sparse_map->ComputeEmbeddings0();
  long rows = sm_embeddings_matrix.Rows();
  long cols = sm_embeddings_matrix.Cols();
  if ((rows != object_number) ||
      (cols != dimensions))
  {
    exit_error("Error: sm_embeddings size = %li x li instead of %li x %li\n",
                    rows, cols, object_number, dimensions);
  }
  
  return 1;
}


long vSignBoostInfo::ComputeFMEmbeddings()
{
  if (fast_map == 0) return 0;
  long dimensions = fast_map->Dimensions();
  if (dimensions <= 0) return 0;

  vMatrix<float> weights_matrix = vMatrix<float>(1, dimensions);
  vWriteValue(&weights_matrix, (float) 1);
  vArray(float) weights = weights_matrix.Matrix();

  fm_embeddings_matrix = vMatrix<float>(object_number, dimensions);
  vArray2(float) embeddings = fm_embeddings_matrix.Matrix2();

  vMatrix<long> pivots = fast_map->PivotsMatrix();
  vMatrix<float> pivot_distances = fast_map->OriginalPivotDistances(pivots);
  vMatrix<float> intrapivot_distances = 
    fast_map->OptimizedPivotDistances(pivot_distances);

  long i, j;
  for (i = 0; i < object_number; i++)
  {
    vMatrix<float> current_embedding = 
      fast_map->Embedding1(i, intrapivot_distances);
    vArray(float) current = current_embedding.Matrix();
    long current_dimensions = current_embedding.Size();
    if (current_dimensions != dimensions)
    {
      exit_error("Error: current_dimensions = %li, dimensions = %li\n",
                      current_dimensions, dimensions);
    }

    for (j = 0; j < dimensions; j++)
    {
      embeddings[i][j] = current[j];
    }
  }
  
  vPrint("\nComputed embeddings, now computing approximate distances:\n");

  return 1;
}


// This function is written so that we can update the embedding distances by
// adding one dimension at a time. That's why the result matrix 
// (embedding_distances_matrix) is passed in
// as an argument. If the result matrix already contains some entries and
// is not zero, we will just add to each of those entries the contribution
// of the specified dimensions.
vMatrix<float> vSignBoostInfo::ComputeEmbeddingDistances(vMatrix<float> embeddings_matrix,
                                                           vMatrix<float> weights_matrix,
                                                           vMatrix<float> embedding_distances_matrix,
                                                           long start, long dimensions)
{
  if ((embedding_distances_matrix.Rows() != object_number) ||
      (embedding_distances_matrix.Rows() != object_number))
  {
    exit_error("Error: non-square matrix in ComputeEmbeddingDistances\n");
  }

  vArray2(float) embeddings = embeddings_matrix.Matrix2();
  vArray2(float) embedding_distances = embedding_distances_matrix.Matrix2();
  vArray(float) weights = weights_matrix.Matrix();
  if (start < 0) start = 0;

  if (start+dimensions > embeddings_matrix.Cols())
  {
    exit_error("Error: too many dimensions in ComputeEmbeddingDistances\n");
  }

  long i, j;
  vPrint("\n");
  for (i = 0; i < object_number; i++)
  {
    embedding_distances[i][i] = 0;
    for (j = 0; j < i; j++)
    {
      long k;
      float sum = embedding_distances[i][j];
      for (k = start; k < start+dimensions; k++)
      {
        float coord1 = embeddings[i][k];
        float coord2 = embeddings[j][k];
        float diff = vAbs(coord1 - coord2);
        float weight = weights[k];
        sum = sum + weight * diff;
      }

      embedding_distances[i][j] = sum;
      embedding_distances[j][i] = sum;
    }

    vPrint("Done with object %li of %li\r", i+1, object_number);
  }
  vPrint("\n");

  return embedding_distances_matrix;
}

  
vMatrix<float> vSignBoostInfo::BM_Embeddings()
{
  return bm_embeddings_matrix;
}


vMatrix<float> vSignBoostInfo::SM_Embeddings()
{
  return sm_embeddings_matrix;
}


vMatrix<float> vSignBoostInfo::FM_Embeddings()
{
  return fm_embeddings_matrix;
}


// CandidateMatches(i) is non-zero iff the i-th object should be used
// as a candidate match for queries in ApproximateRankStats. We use
// 180 signs by Norma as queries. As candidate matches we consider 
// all sequences by Ben and Mike, except for sequences by Mike that
// have the same label as one of the queries, and sequences by Ben
// which have the same label as one of the queries AND are set to 0
// in ben_test_indices.
vMatrix<long> vSignBoostInfo::CandidateMatches()
{
  long i;
  vector<long> candidates;
  for (i = 0; i < object_number; i++)
  {
    if (norma_indices(i) != 0) continue;
    long label = class_labels(i);
    long is_test = test_labels(label);
    if (is_test != 0)
    {
      if (ben_test_indices(i) == 0)
      { 
        continue;
      }
    }
    candidates.push_back(i);
  }

  long number = candidates.size();
  vMatrix<long> result_matrix(1, number);
  vArray(long) result = result_matrix.Matrix();
  for (i = 0; i < number; i++)
  {
    result[i] = candidates[i];
  }
  
  result_matrix.Print("candidate_matches");
  vPrint("number = %li\n", number);
  return result_matrix;
}


v3dMatrix<float> vSignBoostInfo::ApproximateRankStatsBM(long max_dimensions, long max_k)
{
  if (boost_map == 0) return v3dMatrix<float>();
  if (boost_map->StepsDone() == 0) return v3dMatrix<float>();
  ComputeBMEmbeddings();
  vMatrix<float> weights_matrix = boost_map->WeightsMatrix();
  v3dMatrix<float> result = ApproximateRankStats(max_dimensions, max_k, 
                                                   distances_matrix, 
                                                   bm_embeddings_matrix,
                                                   weights_matrix);
  return result;
}


v3dMatrix<float> vSignBoostInfo::ApproximateRankStatsFM(long max_dimensions, long max_k)
{
  if (fast_map == 0) return v3dMatrix<float>();
  if (fast_map->StepsDone() == 0) return v3dMatrix<float>();
  ComputeFMEmbeddings();
  vMatrix<float> weights_matrix(1, fast_map->Dimensions());
  vWriteValue(&weights_matrix, (float) 1);
  v3dMatrix<float> result = ApproximateRankStats(max_dimensions, max_k, 
                                                   distances_matrix, 
                                                   fm_embeddings_matrix,
                                                   weights_matrix);
  return result;
}


v3dMatrix<float> 
vSignBoostInfo::ApproximateRankStatsBM4(long min_dimensions, long max_dimensions,
                                         long min_k, long max_k)
{
  if (boost_map == 0) return v3dMatrix<float>();
  ComputeBMEmbeddings();
  vMatrix<float> weights_matrix = boost_map->WeightsMatrix();
  v3dMatrix<float> result = ApproximateRankStats7(min_dimensions, max_dimensions, 
                                                    min_k, max_k, 
                                                    distances_matrix, 
                                                    bm_embeddings_matrix,
                                                    weights_matrix);
  return result;
}


v3dMatrix<float> 
vSignBoostInfo::ApproximateRankStatsSM4(long min_dimensions, long max_dimensions, 
                                         long min_k, long max_k)
{
  if (sparse_map == 0) return v3dMatrix<float>();
  ComputeSMEmbeddings();
  vMatrix<float> weights_matrix(1, sparse_map->Dimensions());
  vWriteValue(&weights_matrix, (float) 1);
  v3dMatrix<float> result = ApproximateRankStats7(min_dimensions, max_dimensions, 
                                                    min_k, max_k, 
                                                    distances_matrix, 
                                                    sm_embeddings_matrix,
                                                    weights_matrix);
  return result;
}


v3dMatrix<float> 
vSignBoostInfo::ApproximateRankStatsFM4(long min_dimensions, long max_dimensions, 
                                         long min_k, long max_k)
{
  if (fast_map == 0) return v3dMatrix<float>();
  ComputeFMEmbeddings();
  vMatrix<float> weights_matrix(1, fast_map->Dimensions());
  vWriteValue(&weights_matrix, (float) 1);
  v3dMatrix<float> result = ApproximateRankStats7(min_dimensions, max_dimensions, 
                                                    min_k, max_k, 
                                                    distances_matrix, 
                                                    fm_embeddings_matrix,
                                                    weights_matrix);
  return result;
}
  

  
v3dMatrix<float> vSignBoostInfo::ApproximateRankStats(long max_dimensions, long max_k,
                                                         vMatrix<float> actual_distances,
                                                         vMatrix<float> embeddings_matrix,
                                                         vMatrix<float> weights_matrix)
{
  // Figure out what the set of queries is and what the set of matches is.
  long i, j, k;
  vector<long> queries;
  for (i = 0; i < object_number; i++)
  {
    if (norma_test_indices(i) != 0)
    {
      queries.push_back(i);
    }
  }

  vMatrix<long> candidate_matches = CandidateMatches();
  vArray(long) candidates = candidate_matches.Matrix();
  long number_of_queries = queries.size();
  long number_of_candidates = candidate_matches.Size();
  if (max_k < 1) max_k = 1;
  if (max_k > number_of_candidates) max_k = number_of_candidates;
  if (max_dimensions < 1) max_dimensions = 1;
  if (max_dimensions > embeddings_matrix.Cols()) 
  {
    max_dimensions = embeddings_matrix.Cols();
  }

  v3dMatrix<float> result_matrix(number_of_queries, max_k,
                                   max_dimensions);
  vArray3(float) result = result_matrix.Matrix3();

  vMatrix<float> embedding_distances(object_number, object_number);
  vWriteValue(&embedding_distances, (float) 0);
  vArray2(float) actual = actual_distances.Matrix2();
  vArray2(float) embedding = embedding_distances.Matrix2();

  for (i = 0; i < max_dimensions; i++)
  {
    ComputeEmbeddingDistances(embeddings_matrix, weights_matrix, 
                              embedding_distances, i, 1);
    
    for (j = 0; j < number_of_queries; j++)
    {
      long query_index = queries[j];
      vector<float> actual1(number_of_candidates);
      vector<float> approximate1(number_of_candidates);
      long c;
      for (c = 0; c < number_of_candidates; c++)
      {
        long match_index = candidates[c];
        actual1[c] = actual[query_index][match_index];
        approximate1[c] = embedding[query_index][match_index];
      }

      for (k = 0; k < max_k; k++)
      {
        long k_index = 0;
        // Get the index of the k-th nearest neighbor
        kth_smallest_ca(k+1, &actual1, &k_index);
        float k_approx_distance = approximate1[k_index];
        float counter = (float) 0;
        for (c = 0; c < number_of_candidates; c++)
        {
          if (c == k_index) continue;
          float approx_distance = approximate1[c];
          if (approx_distance < k_approx_distance)
          {
            counter = counter + (float) 1;
          }
          else if (approx_distance == k_approx_distance)
          {
            counter = counter + (float) 0.5;
          }
        }

        result[i][j][k] = counter;
      }
    }
    
    vPrint("done with dimension %li of %li\n", i+1, max_dimensions);
  }

  return result_matrix;
}


v3dMatrix<float> 
vSignBoostInfo::ApproximateRankStats7(long min_dimensions, long max_dimensions, 
                                       long min_k, long max_k,
                                       vMatrix<float> actual_distances,
                                       vMatrix<float> embeddings_matrix,
                                       vMatrix<float> weights_matrix)
{
  // Figure out what the set of queries is and what the set of matches is.
  long i, j, k;
  vector<long> queries;
  for (i = 0; i < object_number; i++)
  {
    if (norma_test_indices(i) != 0)
    {
      queries.push_back(i);
    }
  }

  vMatrix<long> candidate_matches = CandidateMatches();
  vArray(long) candidates = candidate_matches.Matrix();
  long number_of_queries = queries.size();
  long number_of_candidates = candidate_matches.Size();
  if (min_k < 1) min_k = 1;
  if (max_k < min_k) max_k = min_k;
  if (max_k > number_of_candidates) max_k = number_of_candidates;
  if (min_dimensions < 1) min_dimensions = 1;
  if (max_dimensions < min_dimensions) max_dimensions = min_dimensions;
  if (max_dimensions > embeddings_matrix.Cols()) 
  {
    max_dimensions = embeddings_matrix.Cols();
  }

  long number_of_ks = max_k - min_k + 1;
  long number_of_dimensions = max_dimensions - min_dimensions + 1;

  v3dMatrix<float> result_matrix(number_of_queries, number_of_ks,
                                   number_of_dimensions);
  vArray3(float) result = result_matrix.Matrix3();

  vMatrix<float> embedding_distances(object_number, object_number);
  vWriteValue(&embedding_distances, (float) 0);
  vArray2(float) actual = actual_distances.Matrix2();
  vArray2(float) embedding = embedding_distances.Matrix2();

  ComputeEmbeddingDistances(embeddings_matrix, weights_matrix, 
                            embedding_distances, 0, min_dimensions-1);

  for (i = min_dimensions; i <= max_dimensions; i++)
  {
    ComputeEmbeddingDistances(embeddings_matrix, weights_matrix, 
                              embedding_distances, i-1, 1);
    
    for (j = 0; j < number_of_queries; j++)
    {
      long query_index = queries[j];
      vector<float> actual1(number_of_candidates);
      vector<float> approximate1(number_of_candidates);
      long c;
      for (c = 0; c < number_of_candidates; c++)
      {
        long match_index = candidates[c];
        actual1[c] = actual[query_index][match_index];
        approximate1[c] = embedding[query_index][match_index];
      }

      for (k = min_k; k <= max_k; k++)
      {
        long k_index = 0;
        // Get the index of the k-th nearest neighbor
        kth_smallest_ca(k, &actual1, &k_index);
        float k_approx_distance = approximate1[k_index];
        float counter = (float) 0;
        for (c = 0; c < number_of_candidates; c++)
        {
          if (c == k_index) continue;
          float approx_distance = approximate1[c];
          if (approx_distance < k_approx_distance)
          {
            counter = counter + (float) 1;
          }
          else if (approx_distance == k_approx_distance)
          {
            counter = counter + (float) 0.5;
          }
        }

        result[i-min_dimensions][j][k-min_k] = counter;
      }
    }
    
    vPrint("done with dimension %li of %li\n", i, max_dimensions);
  }

  return result_matrix;
}



v3dMatrix<float> 
vSignBoostInfo::HRCM_ApproxStatsBM4(long min_dimensions, long max_dimensions,
                                     long min_k, long max_k)
{
  if (boost_map == 0) return v3dMatrix<float>();
  ComputeBMEmbeddings();
  vMatrix<float> weights_matrix = boost_map->WeightsMatrix();
  v3dMatrix<float> result = HRCM_ApproxStats7(min_dimensions, max_dimensions, 
                                                min_k, max_k, 
                                                distances_matrix, 
                                                bm_embeddings_matrix,
                                                weights_matrix);
  return result;
}


v3dMatrix<float> 
vSignBoostInfo::HRCM_ApproxStatsFM4(long min_dimensions, long max_dimensions, 
                                     long min_k, long max_k)
{
  if (fast_map == 0) return v3dMatrix<float>();
  ComputeFMEmbeddings();
  vMatrix<float> weights_matrix(1, fast_map->Dimensions());
  vWriteValue(&weights_matrix, (float) 1);
  v3dMatrix<float> result = HRCM_ApproxStats7(min_dimensions, max_dimensions, 
                                                min_k, max_k, 
                                                distances_matrix, 
                                                fm_embeddings_matrix,
                                                weights_matrix);
  return result;
}
  

v3dMatrix<float> 
vSignBoostInfo::HRCM_ApproxStatsSM4(long min_dimensions, long max_dimensions, 
                                     long min_k, long max_k)
{
  if (sparse_map == 0) return v3dMatrix<float>();
  ComputeSMEmbeddings();
  vMatrix<float> weights_matrix(1, sparse_map->Dimensions());
  vWriteValue(&weights_matrix, (float) 1);
  v3dMatrix<float> result = HRCM_ApproxStats7(min_dimensions, max_dimensions, 
                                                min_k, max_k, 
                                                distances_matrix, 
                                                sm_embeddings_matrix,
                                                weights_matrix);
  return result;
}
  

  
v3dMatrix<float> 
vSignBoostInfo::HRCM_ApproxStats7(long min_dimensions, long max_dimensions, 
                                   long min_k, long max_k,
                                   vMatrix<float> actual_distances,
                                   vMatrix<float> embeddings_matrix,
                                   vMatrix<float> weights_matrix)
{
  // Figure out what the set of queries is and what the set of matches is.
  long i, j, k;
  vector<long> queries;
  for (i = 0; i < object_number; i++)
  {
    if (norma_test_indices(i) != 0)
    {
      queries.push_back(i);
    }
  }

  vMatrix<long> candidate_matches = CandidateMatches();
  vArray(long) candidates = candidate_matches.Matrix();
  long number_of_queries = queries.size();
  long number_of_candidates = candidate_matches.Size();
  if (min_k < 1) min_k = 1;
  if (max_k < min_k) max_k = min_k;
  if (max_k > number_of_candidates) max_k = number_of_candidates;
  if (min_dimensions < 1) min_dimensions = 1;
  if (min_dimensions >= embeddings_matrix.Cols())
  {
    min_dimensions = embeddings_matrix.Cols();
  }
  if (max_dimensions < min_dimensions) max_dimensions = min_dimensions;
  if (max_dimensions > embeddings_matrix.Cols()) 
  {
    max_dimensions = embeddings_matrix.Cols();
  }

  long number_of_ks = max_k - min_k + 1;
  long number_of_dimensions = max_dimensions - min_dimensions + 1;

  v3dMatrix<float> result_matrix(number_of_queries, number_of_ks,
                                   number_of_dimensions);
  vArray3(float) result = result_matrix.Matrix3();

  vMatrix<float> embedding_distances(object_number, object_number);
  vWriteValue(&embedding_distances, (float) 0);

  vArray2(float) actual = actual_distances.Matrix2();
  vArray2(float) embedding = embedding_distances.Matrix2();

  ComputeEmbeddingDistances(embeddings_matrix, weights_matrix, 
                            embedding_distances, 0, min_dimensions-1);

  for (i = min_dimensions; i <= max_dimensions; i++)
  {
    ComputeEmbeddingDistances(embeddings_matrix, weights_matrix, 
                              embedding_distances, i-1, 1);
    
    for (j = 0; j < number_of_queries; j++)
    {
      long query_index = queries[j];
      long class_label = class_labels(query_index);
      vector<float> actual1(number_of_candidates);
      vector<float> approximate1(number_of_candidates);
      
      // max_correct_distance will hold the maximum distance (exact distance,
      // not approximate distance), between the query and a correct match.
      float max_correct_distance = 0;

      // correct_counter will count the number of correct matches found.
      // In this implementation and with this dataset, that should always
      // be 20.
      long correct_counter = 0;

      long c;
      for (c = 0; c < number_of_candidates; c++)
      {
        long match_index = candidates[c];

        // The next line leads to finding the embedding-based rank of the k-th
        // correct match under the exact distance.
        //actual1[c] = actual[query_index][match_index];

        // The next line leads to finding the embedding-based k-th
        // correct match
        actual1[c] = embedding[query_index][match_index];

        approximate1[c] = embedding[query_index][match_index];
        long match_label = class_labels(match_index);
        if (match_label == class_label)
        {
          correct_counter++;
          if (actual1[c] > max_correct_distance)
          {
            max_correct_distance = actual1[c];
          }
        }
      }

      if (max_correct_distance < 0)
      {
        exit_error("Error: max_correct_distance = %f\n", max_correct_distance);
      }

      if (correct_counter != 20)
      {
        exit_error("Error: correct_counter = %li\n", correct_counter);
      }

      // The lext for-loop is a hack, to ensure that when we call kth_smallest_ca
      // we get the index of the k-th correct match, not the index of the k-th
      // nearest neighbor. We simply modify the actual1 distances of all incorrect
      // matches to be higher than the highest distance of a correct match.
      for (c = 0; c < number_of_candidates; c++)
      {
        long match_index = candidates[c];
        long match_label = class_labels(match_index);
        if (match_label != class_label)
        {
          actual1[c] = vAbs(actual1[c]) + vAbs(max_correct_distance) + (float) 1.0;
        }
      }

      for (k = min_k; k <= max_k; k++)
      {
        long k_index = 0;
        // Get the index of the k-th nearest neighbor
        kth_smallest_ca(k, &actual1, &k_index);
        float k_approx_distance = approximate1[k_index];
        float counter = (float) 0;
        for (c = 0; c < number_of_candidates; c++)
        {
          if (c == k_index) continue;
          float approx_distance = approximate1[c];
          if (approx_distance < k_approx_distance)
          {
            counter = counter + (float) 1;
          }
          else if (approx_distance == k_approx_distance)
          {
            counter = counter + (float) 0.5;
          }
        }

        result[i-min_dimensions][j][k-min_k] = counter;
      }
    }
    
    vPrint("done with dimension %li of %li\n", i, max_dimensions);
  }

  return result_matrix;
}


v3dMatrix<float> vSignBoostInfo::HRCM_Stats2(long min_k, long max_k)
{
  v3dMatrix<float> result = HRCM_Stats3(min_k, max_k, distances_matrix);
  return result;
}


v3dMatrix<float> vSignBoostInfo::HRCM_Stats3(long min_k, long max_k,
                                                vMatrix<float> actual_distances)
{
  // Figure out what the set of queries is and what the set of matches is.
  long i, j, k;
  vector<long> queries;
  for (i = 0; i < object_number; i++)
  {
    if (norma_test_indices(i) != 0)
    {
      queries.push_back(i);
    }
  }

  vMatrix<long> candidate_matches = CandidateMatches();
  vArray(long) candidates = candidate_matches.Matrix();
  long number_of_queries = queries.size();
  long number_of_candidates = candidate_matches.Size();
  if (min_k < 1) min_k = 1;
  if (max_k < min_k) max_k = min_k;
  if (max_k > number_of_candidates) max_k = number_of_candidates;

  long number_of_ks = max_k - min_k + 1;

  v3dMatrix<float> result_matrix(number_of_queries, number_of_ks,
                                   1);
  vArray2(float) result = result_matrix.Matrix2(0);

  vArray2(float) actual = actual_distances.Matrix2();

  for (j = 0; j < number_of_queries; j++)
  {
    long query_index = queries[j];
    long class_label = class_labels(query_index);
    vector<float> actual1(number_of_candidates);
    vector<float> modified_actual1(number_of_candidates);
    
    // max_correct_distance will hold the maximum distance (exact distance,
    // not approximate distance), between the query and a correct match.
    float max_correct_distance = 0;

    // correct_counter will count the number of correct matches found.
    // In this implementation and with this dataset, that should always
    // be 20.
    long correct_counter = 0;

    long c;
    for (c = 0; c < number_of_candidates; c++)
    {
      long match_index = candidates[c];
      actual1[c] = actual[query_index][match_index];
      modified_actual1[c] = actual1[c];
      long match_label = class_labels(match_index);
      if (match_label == class_label)
      {
        correct_counter++;
        if (actual1[c] > max_correct_distance)
        {
          max_correct_distance = actual1[c];
        }
      }
    }

    if (max_correct_distance < 0)
    {
      exit_error("Error: max_correct_distance = %f\n", max_correct_distance);
    }

    if (correct_counter != 20)
    {
      exit_error("Error: correct_counter = %li\n", correct_counter);
    }

    // The lext for-loop is a hack, to ensure that when we call kth_smallest_ca
    // we get the index of the k-th correct match, not the index of the k-th
    // nearest neighbor. We simply modify the actual1 distances of all incorrect
    // matches to be higher than the highest distance of a correct match.
    for (c = 0; c < number_of_candidates; c++)
    {
      long match_index = candidates[c];
      long match_label = class_labels(match_index);
      if (match_label != class_label)
      {
        modified_actual1[c] = vAbs(modified_actual1[c]) + vAbs(max_correct_distance) + (float) 1.0;
      }
    }

    for (k = min_k; k <= max_k; k++)
    {
      long k_index = 0;
      // Get the index of the k-th nearest neighbor
      kth_smallest_ca(k, &modified_actual1, &k_index);
      float k_correct_distance = modified_actual1[k_index];
      float counter = (float) 0;
      for (c = 0; c < number_of_candidates; c++)
      {
        if (c == k_index) continue;
        float distance  = actual1[c];
        if (distance< k_correct_distance)
        {
          counter = counter + (float) 1;
        }
        else if (distance == k_correct_distance)
        {
          counter = counter + (float) 0.5;
        }
      }

      result[j][k-min_k] = counter;
    }
  }

  return result_matrix;
}


// beginning of functions written for submission to KDD 2004
// indices to be used for the training set.
vMatrix<long> vSignBoostInfo::TrainingIndices()
{
  vMatrix<long> result = CandidateMatches();
  long size = result.Size();
  vPrint("There are %li training indices\n", size);
  return result;
}


// Labels of training set.
vMatrix<long> vSignBoostInfo::TrainingLabels()
{
  vMatrix<long> indices = TrainingIndices();
  long size = indices.Size();
  vMatrix<long> result(1, size);

  long i;
  for (i = 0; i < size; i++)
  {
    long index = indices(i);
    if ((index < 0) || (index >= class_labels.Size()))
    {
      exit_error("error: traininglabels, index = %li\n",
                      index);
    }
    result(i) = class_labels(index);
  }

  return result;
}


// distance matrix for training set (training to training)
vMatrix<float> vSignBoostInfo::TrainingDistances()
{
  vMatrix<long> indices = TrainingIndices();
  long size = indices.Size();
  vMatrix<float> result(size, size);

  long i;
  for (i = 0; i < size; i++)
  {
    long index1 = indices(i);
    if ((index1 < 0) || (index1 >= distances_matrix.Rows()))
    {
      exit_error("error: trainingdistances, index1 = %li\n",
                      index1);
    }
    long j;
    for (j = 0; j < size; j++)
    {
      long index2 = indices(j);
      if ((index2 < 0) || (index2 >= distances_matrix.Rows()))
      {
        exit_error("error: trainingdistances, index2 = %li\n",
                        index2);
      }
      float distance = distances_matrix(index1, index2);
      result(i, j) = distance;
    }
  }

  vPrint("distance size is %li x %li\n", result.Rows(), result.Cols());
  return result;
}


vMatrix<float> vSignBoostInfo::TrainingPdistances()
{
  vMatrix<long> indices = TrainingIndices();
  long size = indices.Size();
  vMatrix<float> result(size, size);

  long i;
  for (i = 0; i < size; i++)
  {
    long index1 = indices(i);
    if ((index1 < 0) || (index1 >= class_labels.Size()))
    {
      exit_error("error: trainingpdistances, index1 = %li\n",
                      index1);
    }
    long class1 = class_labels(index1);
    long j;
    for (j = 0; j < size; j++)
    {
      long index2 = indices(j);
      if ((index1 < 0) || (index1 >= class_labels.Size()))
      {
        exit_error("error: trainingpdistances, index2 = %li\n",
                        index2);
      }
      long class2 = class_labels(index2);
      float distance;
      if ((class1 == class2) && (test_labels(class1) == 1))
      {
        distance = (float) 0;
      }
      else
      { 
        distance = (float) 1;
      }
      result(i, j) = distance;
    }
  }

  vPrint("pdistance size is %li x %li\n", result.Rows(), result.Cols());
  return result;
}


// indices to be used for the test set
vMatrix<long> vSignBoostInfo::TestIndices()
{
  vector<long> queries;
  long i;
  for (i = 0; i < object_number; i++)
  {
    if (norma_test_indices(i) != 0)
    {
      queries.push_back(i);
    }
  }
  vMatrix<long> result = matrix_from_vector(&queries);
  long size = result.Size();
  vPrint("There are %li test indices\n", size);
  return result;
}


// Labels of Test set.
vMatrix<long> vSignBoostInfo::TestLabels()
{
  vMatrix<long> indices = TestIndices();
  long size = indices.Size();
  vMatrix<long> result(1, size);

  long i;
  for (i = 0; i < size; i++)
  {
    long index = indices(i);
    if ((index < 0) || (index >= class_labels.Size()))
    {
      exit_error("error: testlabels, index = %li\n",
                      index);
    }
    result(i) = class_labels(index);
  }

  return result;
}


// distance matrix for Test set (Test to Test)
vMatrix<float> vSignBoostInfo::TestDistances()
{
  vMatrix<long> test_indices = TestIndices();
  vMatrix<long> training_indices = TrainingIndices();
  long test_size = test_indices.Size();
  long training_size = training_indices.Size();
  vMatrix<float> result(test_size, training_size);

  long i;
  for (i = 0; i < test_size; i++)
  {
    long index1 = test_indices(i);
    if ((index1 < 0) || (index1 >= distances_matrix.Rows()))
    {
      exit_error("error: testdistances, index1 = %li\n",
                      index1);
    }
    long j;
    for (j = 0; j < training_size; j++)
    {
      long index2 = training_indices(j);
      if ((index2 < 0) || (index2 >= distances_matrix.Rows()))
      {
        exit_error("error: testindices, index2 = %li\n",
                        index2);
      }
      float distance = distances_matrix(index1, index2);
      result(i, j) = distance;
    }
  }

  vPrint("distance size is %li x %li\n", result.Rows(), result.Cols());
  return result;
}


vMatrix<float> vSignBoostInfo::TestPdistances()
{
  vMatrix<long> test_indices = TestIndices();
  vMatrix<long> training_indices = TrainingIndices();
  long test_size = test_indices.Size();
  long training_size = training_indices.Size();
  vMatrix<float> result(test_size, training_size);

  long i;
  for (i = 0; i < test_size; i++)
  {
    long index1 = test_indices(i);
    if ((index1 < 0) || (index1 >= class_labels.Size()))
    {
      exit_error("error: testpdistances, index1 = %li\n",
                      index1);
    }
    long class1 = class_labels(index1);
    long j;
    for (j = 0; j < training_size; j++)
    {
      long index2 = training_indices(j);
      if ((index1 < 0) || (index1 >= class_labels.Size()))
      {
        exit_error("error: testpdistances, index2 = %li\n",
                        index2);
      }
      long class2 = class_labels(index2);
      float distance;
      if (class1 == class2)
      {
        distance = (float) 0;
      }
      else
      { 
        distance = (float) 1;
      }
      result(i, j) = distance;
    }
  }

  vPrint("pdistance size is %li x %li\n", result.Rows(), result.Cols());
  return result;
}


  

/////////////////////////////////////////////////////////////////////
//                                                                 //
//            End of implementation of vSignBoostInfo             //
//                                                                 //
/////////////////////////////////////////////////////////////////////


vChamferEmbeddingBM::vChamferEmbeddingBM(vMatrix<float> in_classifiers,
                                           vMatrix<float> in_pivot_distances,
                                           vMatrix<long> indices_to_poses)
{
  is_valid = 0;
  Initialize3(in_classifiers, in_pivot_distances, indices_to_poses);
}


vChamferEmbeddingBM::vChamferEmbeddingBM(const char * filename)
{
  is_valid = 0;
  char * pathname = Pathname1(filename);
  FILE * fp = fopen(pathname, vFOPEN_READ);
  if (fp == 0) 
  {
    vPrint("Failed to open %s\n", pathname);
    vdelete2(pathname);
    return;
  }

  vdelete2(pathname);
  v3dMatrix<float> * temp_pivot_distances = v3dMatrix<float>::ReadText(fp);
  if (temp_pivot_distances == 0)
  {
    fclose(fp);
    return;
  }

  pivot_distances = vMatrix<float>(temp_pivot_distances);
  vdelete(temp_pivot_distances);

  v3dMatrix<float> * temp_unique_classifiers = v3dMatrix<float>::ReadText(fp);
  if (temp_unique_classifiers == 0)
  {
    fclose(fp);
    return;
  }

  unique_classifiers = vMatrix<float>(temp_unique_classifiers);
  vdelete(temp_unique_classifiers);

  v3dMatrix<long> * temp_indices_to_poses = v3dMatrix<long>::ReadText(fp);
  if (temp_indices_to_poses == 0)
  {
    fclose(fp);
    return;
  }

  indices_to_poses = vMatrix<long>(temp_indices_to_poses);
  vdelete(temp_indices_to_poses);
  fclose(fp);

  Initialize3(unique_classifiers, pivot_distances, indices_to_poses);
}


vChamferEmbeddingBM::~vChamferEmbeddingBM()
{
}


long vChamferEmbeddingBM::Initialize3(vMatrix<float> in_unique_classifiers,
                                       vMatrix<float> in_pivot_distances,
                                       vMatrix<long> in_indices_to_poses)
{
  pivot_distances = in_pivot_distances;
  unique_classifiers = in_unique_classifiers;
  indices_to_poses = in_indices_to_poses;

  if (pivot_distances.Size() != unique_classifiers.Rows())
  {
    return 0;
  }

  types_matrix = V_BoostMap3::ExtractTypes(unique_classifiers);
  dimensions = pivot_distances.Size();
  reference_images = ComputeReferenceImages(unique_classifiers, 
                                            indices_to_poses);
  v3dMatrix<float> * temp_unique = V_BoostMap::CleanUp(107328, 
                                                         &reference_images);
  unique_reference_images = vMatrix<float>(temp_unique);
  vdelete(temp_unique);
  unique_indices = ComputeUniqueIndices(reference_images, 
                                        unique_reference_images);

  is_valid = 1;
  return 1;
}

long vChamferEmbeddingBM::Print()
{
//  indices_to_poses.Print("indices_to_poses");
  vPrint("ChamferEmbeddingBM: dimensions = %li\n", dimensions);
  pivot_distances.Print("pivot_distances");
  types_matrix.Print("types_matrix");
  unique_classifiers.Print("unique_classifiers");
  reference_images.Print("reference_images");
  unique_reference_images.Print("unique_reference_images");
  unique_indices.Print("unique_indices");
  return 1;
}

long vChamferEmbeddingBM::Save(const char * filename)
{
  char * pathname1 = Pathname1(filename);
  FILE * fp = fopen(pathname1, vFOPEN_WRITE);
  if (fp == 0) 
  {
    vdelete2(pathname1);
    return 0;
  }

  long success;

  success = pivot_distances.WriteText(fp);
  if (success <= 0)
  {
    fclose(fp);
    vdelete2(pathname1);
    return 0;
  }

  success = unique_classifiers.WriteText(fp);
  if (success <= 0)
  {
    fclose(fp);
    vdelete2(pathname1);
    return 0;
  }

  success = indices_to_poses.WriteText(fp);
  if (success <= 0)
  {
    fclose(fp);
    vdelete2(pathname1);
    return 0;
  }

  fclose(fp);
  char * pathname2 = Pathname2(filename);
  success = vCopyFile(pathname1, pathname2);
  if (success <= 0)
  {
    vPrint("Failed to copy %s to %s\n", pathname1, pathname2);
  }
  else
  {
    vPrint("Saved ceb successfully to %s and %s\n", 
            pathname1, pathname2);
  }
  vdelete2(pathname1);
  vdelete2(pathname2);
  return 1;
}


vMatrix<float> vChamferEmbeddingBM::ReferenceImages()
{
  return vMatrix<float>(&reference_images);
}


vMatrix<float> vChamferEmbeddingBM::UniqueReferenceImages()
{
  return vMatrix<float>(&unique_reference_images);
}


vMatrix<long> vChamferEmbeddingBM::UniqueIndices()
{
  return vMatrix<long>(&unique_indices);
}


vMatrix<float> vChamferEmbeddingBM::UniqueClassifiers()
{
  return vMatrix<float>(&unique_classifiers);
}


vMatrix<float> vChamferEmbeddingBM::Embedding(vMatrix<float> query_distances)
{
  vMatrix<float> result = V_BoostMap3::Embedding3(query_distances, pivot_distances,
                                                   types_matrix);
  return result;
}


vMatrix<float> 
vChamferEmbeddingBM::ComputeReferenceImages(vMatrix<float> classifiers,
                                             vMatrix<long> indices_to_poses)
{
  long number = classifiers.Rows();
  vMatrix<float> result(2*number, 4);
  vWriteValue(&result, (float) -1);

  long i;
  for (i = 0; i < number; i++)
  {
    long type = vRound(classifiers(i, 0));
    long index1 = vRound(classifiers(i, 1));
    long index2 = vRound(classifiers(i, 2));
    float weight = classifiers(i, 3);
    long frame = indices_to_poses(index1, 0);
    long rotation = indices_to_poses(index1, 1);
    result(2*i, 0) = (float) frame;
    result(2*i, 1) = (float) rotation;
    // Column 2 is the weight. By making all weights 1
    // we ensure that no images are removed by V_BoostMap::CleanUp.
    // There is no meaningful weight to put here, anyway.
    result(2*i, 2) = 1;
    result(2*i, 3) = (float) index1;

    if (type != 1)
    {
      if (type != 0)
      {
        exit_error("Error: bad type\n");
      }
      continue;
    }
    frame = indices_to_poses(index2, 0);
    rotation = indices_to_poses(index2, 1);
    result(2*i+1, 0) = (float) frame;
    result(2*i+1, 1) = (float) rotation;
    result(2*i+1, 2) = 1;
    result(2*i+1, 3) = (float) index2;
  }

  return result;
}


// unique_distances are the distances of the query to each image in
// unique_reference_images. We return a result such that result(i)
// is the distance to the reference image stored in the i-th row of
// reference_images. The result can be passed to the Embedding function
// to get a valid embedding.
vMatrix<float> 
vChamferEmbeddingBM::ConvertUniqueDistances(vMatrix<float> unique_distances)
{
  vMatrix<float> result(1, 2*dimensions);
  vWriteValue(&result, (float) 0);

  if (unique_indices.Size() != 2*dimensions)
  {
    exit_error("Error: unique_indices.size = %li, != %li\n",
                    unique_indices.Size(), 2*dimensions);
  }
  long unique_number = unique_distances.Size();
  long i;
  for (i = 0; i < 2*dimensions; i++)
  {
    long index = unique_indices(i);
    if (index < 0) continue;
    if (index >= unique_number)
    {
      exit_error("Error: index >= unique_number: %li %li\n",
                      index, unique_number);
    }
    result(i) = unique_distances(index);
  }
  return result;
}


// First we call ConvertUniqueDistances to get query_distances, and then
// we call Embedding(query_distances).
vMatrix<float> 
vChamferEmbeddingBM::EmbeddingFromUniqueDistances(vMatrix<float> unique_distances)
{
  vMatrix<float> query_distances = ConvertUniqueDistances(unique_distances);
  vMatrix<float> result = Embedding(query_distances);
  return result;
}


// Here we build a mapping so that result(j) = i iff the 
// i-th row of unique_reference_images corresponds to j.
vMatrix<long> 
vChamferEmbeddingBM::ComputeUniqueIndices(vMatrix<float> reference_images, 
                                           vMatrix<float> unique_reference_images)
{
  long indices_length = 2*dimensions;
  vMatrix<long> result_matrix(1, indices_length);
  vWriteValue(&result_matrix, (long) -1);
  vArray(long) result = result_matrix.Matrix();

  long i, j;
  long unique_number = unique_reference_images.Rows();
  vArray2(float) references = reference_images.Matrix2();
  vArray2(float) unique = unique_reference_images.Matrix2();

  for (i = 0; i < indices_length; i++)
  {
    long frame1 = vRound(references[i][0]);
    long view1 = vRound(references[i][1]);
    long number_found = 0;
    if (frame1 < 0) 
    {
      continue;
    }
    for (j = 0; j < unique_number; j++)
    {
      long frame2 = vRound(unique[j][0]);
      long view2 = vRound(unique[j][1]);
      if ((frame1 == frame2) && (view1 == view2))
      {
        number_found++;
        result[i] = j;
      }
    }
    if (number_found != 1)
    {
      exit_error("Error: %li matches found for (%li, %li)\n",
                      number_found, frame1, view1);
    }
  }

  return result_matrix;
}


char * vChamferEmbeddingBM::Directory1()
{
  char * directory = vJoinPaths3(sequences_directory, "glove_data",
                                  "ceb_classifiers");
  return directory;
}


char * vChamferEmbeddingBM::Directory2()
{
  char * directory = "r:\\nessie\\source\\ceb_classifiers";
  char * result = vCopyString(directory);
  return result;
}


char * vChamferEmbeddingBM::Pathname1(const char * filename)
{
  char * directory = Directory1();
  char * result = vJoinPaths(directory, filename);
  vdelete2(directory);
  return result;
}


char * vChamferEmbeddingBM::Pathname2(const char * filename)
{
  char * directory = Directory2();
  char * result = vJoinPaths(directory, filename);
  vdelete2(directory);
  return result;
}


// I wrote this function on 12/21/2003, before going to Greece. The goal is to 
// have a big file containing distances from some of the database objects
// to the entire database.
long vWriteBigDistancesFile(long number)
{
  char * sequence = "g:\\users\\athitsos\\sequences\\glove_data\\training\\training7_edges.rle";
  vRLE_Sequence rle(sequence);
  if (rle.IsValid() == 0)
  {
    vPrint("Failed to load sequence from %s\n", sequence);
    return 0;
  }

  vAllChamferDistances acd(&rle);
  
  vMatrix<long> samples = vSampleWithoutReplacement(0, 26831, number);
  samples.Print("samples");

  char * output = "g:\\users\\athitsos\\sequences\\glove_data\\training\\training7_edges.big3";
  FILE * fp = fopen(output, vFOPEN_WRITE);
  if (fp == 0) 
  {
    vPrint("Failed to open %s\n", output);
    return 0;
  }
   
  long result = 1;
  long success = samples.WriteBandwise(fp);
  if (success <= 0)
  {
    vPrint("Failed to save samples\n");
    result = 0;
  }

  vPrint("\n");
  long i;
  for (i = 0; i < number; i++)
  {
    v3dMatrix<float> * distances = acd.Distances(samples(i), 0);
    success = distances->WriteBandwise(fp);
    vdelete(distances);
    if (success <= 0)
    {
      vPrint("Failed to write distances for %li frame, frame = %li\n",
              i, samples(i));
      result = 0;
      break;
    }

    vPrint("Done with frame %li of %li\r", i+1, number);
  }
  vPrint("\n");

  fclose(fp);
  return result;
}


vMatrix<float> vReadBigDistancesFile(vMatrix<long> samples, long frame)
{
  long number = samples.Size();

  char * input = "g:\\users\\athitsos\\sequences\\glove_data\\training\\training7_edges.bigg";
  FILE * fp = fopen(input, vFOPEN_READ);
  if (fp == 0) 
  {
    vPrint("Failed to open %s\n", input);
    return vMatrix<float>();
  }
   
  long result = 1;
  
  long success = samples.ReadBandwise(fp);
  samples.Print("samples");
  if (success <= 0)
  {
    vPrint("Failed to read samples\n");
    result = 0;
  }

  vMatrix<float> distances(1, 107328);
  long position = number * 4 + frame * 107328 * 4;
  fseek(fp, position, SEEK_SET);

  success = distances.ReadBandwise(fp);
  if (success <= 0)
  {
    vPrint("Failed to read distances.\n");
    result = 0;
  }

  fclose(fp);
  if (result <= 0)
  {
    return vMatrix<float>();
  }
  else
  {
    return distances;
  }
}

