#ifndef VASSILIS_CHAMFER_MATCHING_H
#define VASSILIS_CHAMFER_MATCHING_H


#include "hand_database.h"
#include "fingers.h"
#include "learning/boost_map.h"

// query is the points and dts of all rotated versions of the query.
// database is the points and dts of all database frames. We return
// a matrix that matches each rotation of the query to each database frame.
// The order of the scores is: 
// query rotation 0 with frame 0
// query rotation 1 with frame 0
// ...
// query rotation rotations-1 with frame 0
// query rotation 0 with frame 1
// 
// and so on.
vMatrix<float> vChamferDistances2(vChamferFeatures query, 
                                    vChamferFeatures database);

// Same as vChamferDistances2, but it takes an additional argument, 
// which is the set of database poses that the query should be matched
// with. This is important if we only want to match the query to, say,
// 1000 database poses.
vMatrix<float> vChamferDistances3(vChamferFeatures query, 
                                    vChamferFeatures database,
                                    vPoseSet poses);

vMatrix<float> * vChamferEmbedding(grayscale_image * object, 
                                     vArray(v3dMatrix<float> *) dts,
                                     vPointSetSequence * points,
                                     v3dMatrix<float> * reference_images,
                                     long start, long dimensions);


// vAllChamferDistances is a class that is used to compute, save
// and load all distances between pairs of database images (for
// the old database). The chamfer distances from each database frame
// to all (frame, view) combinations of the database are saved
// into a separate file. In all, about 27000 files are created.
// On 10/25/2003 I extended it to handle cases where we want to
// compute distances between all pairs of images such that one comes
// from one database (say training7) and the other comes from another 
// database (say poser1).
class vAllChamferDistances
{
protected:
  char * sequence_name;
  char * sequence1_name;
  char * sequence2_name;
  char * directory;
  char * base_name;

  long number_of_sequences;

  char * Directory();
  char * BaseName();

  // Order is 0 when we only have one sequence, 1 for sequence2->sequence1
  // distances, 2 for sequence1->sequence2 distances.
  char * DirectedName(long frame, long view, long order);
  char * UndirectedName(long frame, long view);

public:
  vAllChamferDistances(vRLE_Sequence * database);

  // All that we do in this constructor (and the previous one) is
  // set the various strings right (sequence_name, directory, base_name).
  vAllChamferDistances(vRLE_Sequence * database1, 
                        vRLE_Sequence * database2);

  ~vAllChamferDistances();

  inline long NumberOfSequences()
  {
    return number_of_sequences;
  }

  // Save all directed distances from frame1/view1 to frame2/view2.
  // Directed distances can be saved very quickly using 
  // distance transforms. Actually, we only need to consider
  // all frame1/view1 frame2 combinations, because the distances
  // for different views of frame2 can be figured out from the 
  // distances between (frame2, 0) and al other database images.
  long SaveAllDirectedDistances(vRLE_Sequence * database,
                                vPointSetSequence * points);

  long SaveDirectedDistances(vRLE_Sequence * database,
                             vPointSetSequence * points,
                             long start, long end);

  // Assuming that all directed distances have already been saved,
  // we can save the undirected distances. In order to do that,
  // we need to load directed distances from a set of frames A
  // to all database images, and then (keeping the distances of A 
  // in memory) load in sequence, for every database image B, its directed
  // distances in memory, and use them to compute undirected distances
  // between all frames in A and frame B.
  long SaveDistances(long first_frame, long last_frame, long chunk);
  v3dMatrix<float> * DirectedDistances(long frame, long view, long order);
  v3dMatrix<float> * Distances(long frame, long view);

  // Given distances from frame/0 to all database images, convert
  // them to distances from frame/view to all database images.
  long ConvertToView(v3dMatrix<float> * distances, long view);

  // For the (frame, view) database image, find the k-th closest database
  // image (besides itself) based on the chamfer distance. It stores
  // the frame and view of the nearest neighbor in the output arguments,
  // and returns the distance.
  float NearestNeighbor(long frame, long view, long k,
                        long * nn_frame, long * nn_view);

  // Here we also pass the distances from (frame, view) to
  // all database images, so that when we already have these
  // distances available we call this function, and we don't
  // need to read the distances from disk.
  float NearestNeighbor(long frame, long view, long k,
                        long * nn_frame, long * nn_view,
                        v3dMatrix<float> * distances);

  // For (frame, view), figure out how many database images have embeddings
  // whose weighted distance to the embedding of (frame, view) is 
  // smaller than the distance of the embedding of (frame2, view2) to
  // the embedding of (frame, view). 
  long ApproximateRank(long frame, long view,
                       long frame2, long view2,
                       v3dMatrix<float> * reference_images,
                       v3dMatrix<float> * database_embeddings,
                       long start, long dimensions);

  // For (frame, view), first call NearestNeighbor to get its
  // k-nearest neighbor, and then call ApproximateRank to get
  // the rank of that neighbor using the embedding. That represents
  // the "loss" incurred by using the embedding instead of the
  // original distance.
  long ApproximateLoss(long frame, long view, long k,
                       v3dMatrix<float> * reference_images,
                       v3dMatrix<float> * database_embeddings,
                       long start, long dimensions);

  // Pick "times" random database images, and for each of them call
  // ApproximateLoss and save the result into ranks. This is a way
  // to get quick statistics on how well the embedding preserves
  // nearest neighbors of database images.
  long ApproximateLossStats(long k, v3dMatrix<float> * reference_images,
                            v3dMatrix<float> * database_embeddings,
                            long start, long dimensions,
                            long times, vector<long> * ranks);

  // returns the size of the hand database, 
  // i.e. number of frames * number of views.
  long NumberOfObjects();
};


// class vChamferBoostInfo is used to create and store information
// that can be used to apply BoostMap to the chamfer distance between
// hand edge images.
class vChamferBoostInfo
{
protected:
  long object_number;
  long training_number;
  long validation_number;
  long sampling_rate;
  float max_distance;  

  // The original max distance from the distances that have been loaded.
  float original_max_distance;

  // contains the (frame, view) parameters of the indices that were chosen.
  vMatrix<long> indices_matrix;
  vMatrix<float> distances_matrix;
  vMatrix<long> triples_matrix;

  vArray2(long) indices;
  vArray2(float) distances;
  vArray2(long) triples;

  char * sequence_name;
  V_BoostMap3 * boost_map;
  vChamferSparseMap * sparse_map;
  vChamferFastMap * fast_map;

  long is_valid;

  float CurrentMaxDistance();

public:
  vChamferBoostInfo();
  vChamferBoostInfo(const char * in_sequence_name, long in_sampling);
  vChamferBoostInfo(vRLE_Sequence * database, vPointSetSequence * points,
                     long in_sampling, long in_training);
  ~vChamferBoostInfo();

  char * Filename();

  long PickIndices(long frames, long rotations, long number);
  long ComputeDistances(vRLE_Sequence * database, vPointSetSequence * points);
  long PickTraining(long in_training);

  vMatrix<long> Indices();
  vMatrix<float> Distances();
  vMatrix<long> Triples();

  long Save();
  long Load();

  long PrintIndices();
  long PrintTriples();
  long PrintDistances();

  long MakeBoostMap();
  V_BoostMap3 * BoostMap();
  static char * ClassifierDirectory();
  static char * ClassifierFilename(const char * base, long number);
  long SaveClassifier(const char * base);
  long LoadClassifier(const char * base, long number);
  long FindIndex(long frame, long rotation);

  // For debugging mainly
  // Return the frame/rotation info for this index.
  long GetPose(long index, long * frame, long * rotation);
  float MaxDistance();
  // End of functions that are used mainly for debugging.
  
  long MakeSparseMap(long number);
  vChamferSparseMap * SparseMap();
  long MakeFastMap();
  vChamferFastMap * FastMap();
  long SetTrainingNumber(long in_number);
  long SetValidationNumber(long in_number);
  // make sure max distance is at most 1.
  long NormalizeDistances();

  // This function takes in a matrix like the one returned 
  // from V_BoostMap3::ClassifierMatrix, and returns the
  // corresponding distances between pairs of points.
  vMatrix<float> PairDistances(vMatrix<float> classifiers);
  inline long IsValid()
  {
	return is_valid;
  }
};


// vSignBoostInfo provides an interface for applying BoostMap to
// Joni's ASL database. It has many things in common with 
// vChamferBoostInfo.
class vSignBoostInfo
{
protected:
  long object_number;
  long training_number;
  float max_distance;  

  vMatrix<float> distances_matrix;
  // ben_indices(i) is non-zero iff the i-th sign was signed by Ben.
  vMatrix<long> ben_indices;
  vMatrix<long> norma_indices;
  // norma_test_indices is a subset of norma_indices, and
  // ben_test_indices is a subset of ben_indices.
  vMatrix<long> norma_test_indices;
  vMatrix<long> ben_test_indices;
  vMatrix<long> class_labels;
  // test_labels(i) is non-zero iff label i is the label of at
  // least one of the signs in ben_test_indices and norma_test_indices.
  vMatrix<long> test_labels;

  vMatrix<long> candidate_indices;
  // candidate_chi(i) is non-zero iff i is a candidate object.
  vMatrix<long> candidate_chi;
  vMatrix<long> training_query_indices;
  vMatrix<long> training_model_indices;
  vMatrix<long> validation_query_indices;
  vMatrix<long> validation_model_indices;

  vMatrix<long> training_triples;
  vMatrix<long> validation_triples;
  vMatrix<float> bm_embeddings_matrix;
  vMatrix<float> sm_embeddings_matrix;
  vMatrix<float> fm_embeddings_matrix;
  
  // embedding_distances(i, j) is the approximate distance between
  // object i and object j, calculated based on their embeddings.
  vMatrix<float> bm_embedding_distances;
  vMatrix<float> sm_embedding_distances;
  vMatrix<float> fm_embedding_distances;

  vArray2(float) distances;

  V_BoostMap3 * boost_map;
  V_SparseMap * sparse_map;
  V_FastMap * fast_map;

  static char * Directory();

  long ReadIndices();
  long SetCandidateIndices();
  long SetTraining();
  long SetValidation();

public:
  vSignBoostInfo();
  vSignBoostInfo(long in_training_number);
  ~vSignBoostInfo();

  long Zero();

  vMatrix<float> Distances();
  long NormalizeDistances();

  long Save(const char * filename);
  long Load();
  long MakeBoostMap();
  V_BoostMap3 * BoostMap();
  long MakeSparseMap(long number);
  V_SparseMap * SparseMap();
  long MakeFastMap();
  V_FastMap * FastMap();

  // This group of functions should be identical to vChamferBoostInfo.
  long PrintDistances();
  static char * ClassifierDirectory();
  static char * ClassifierFilename(const char * base, long number);
  long SaveClassifier(const char * base);
  long LoadClassifier(const char * base, long number);
  float MaxDistance();
  long SetTrainingNumber(long in_number);

  static vMatrix<float> ReadDistances();
  static vMatrix<long> ChooseTriples(vMatrix<long> query_indices,
                                      vMatrix<long> model_indices,
                                      long number);

  long Print();

  // Compute embeddings, based on the current state of boost_map.
  long ComputeBMEmbeddings();
  long ComputeSMEmbeddings();
  long ComputeFMEmbeddings();
  vMatrix<float> ComputeEmbeddingDistances(vMatrix<float> embeddings_matrix,
                                            vMatrix<float> weights_matrix,
                                            vMatrix<float> embedding_distances_matrix,
                                            long start, long dimensions);
  vMatrix<float> BM_Embeddings();
  vMatrix<float> SM_Embeddings();
  vMatrix<float> FM_Embeddings();

  // CandidateMatches(i) is non-zero iff the i-th object should be used
  // as a candidate match for queries in ApproximateRankStats. We use
  // 180 signs by Norma as queries. As candidate matches we consider 
  // all sequences by Ben and Mike, except for sequences by Mike that
  // have the same label as one of the queries, and sequences by Ben
  // which have the same label as one of the queries AND are set to 0
  // in ben_test_indices.
  vMatrix<long> CandidateMatches();

  v3dMatrix<float> ApproximateRankStatsBM(long max_dimensions, long max_k);
  v3dMatrix<float> ApproximateRankStatsFM(long max_dimensions, long max_k);

  v3dMatrix<float> ApproximateRankStatsBM4(long min_dimensions, long max_dimensions,
                                             long min_k, long max_k);
  v3dMatrix<float> ApproximateRankStatsSM4(long min_dimensions, long max_dimensions, 
                                             long min_k, long max_k);
  v3dMatrix<float> ApproximateRankStatsFM4(long min_dimensions, long max_dimensions, 
                                             long min_k, long max_k);
  
  v3dMatrix<float> ApproximateRankStats(long max_dimensions, long max_k,
                                          vMatrix<float> actual_distances,
                                          vMatrix<float> embeddings,
                                          vMatrix<float> weights_matrix);

  v3dMatrix<float> ApproximateRankStats7(long min_dimensions, long max_dimensions, 
                                           long min_k, long max_k,
                                           vMatrix<float> actual_distances,
                                           vMatrix<float> embeddings,
                                           vMatrix<float> weights_matrix);


  v3dMatrix<float> HRCM_Stats2(long min_k, long max_k);
  v3dMatrix<float> HRCM_Stats3(long min_k, long max_k,
                                 vMatrix<float> actual_distances);


  v3dMatrix<float> HRCM_ApproxStatsBM4(long min_dimensions, long max_dimensions,
                                         long min_k, long max_k);
  v3dMatrix<float> HRCM_ApproxStatsFM4(long min_dimensions, long max_dimensions, 
                                         long min_k, long max_k);
  v3dMatrix<float> HRCM_ApproxStatsSM4(long min_dimensions, long max_dimensions, 
                                         long min_k, long max_k);
  
  v3dMatrix<float> HRCM_ApproxStats7(long min_dimensions, long max_dimensions, 
                                       long min_k, long max_k,
                                       vMatrix<float> actual_distances,
                                       vMatrix<float> embeddings,
                                       vMatrix<float> weights_matrix);

  // beginning of functions written for submission to KDD 2004
  // indices to be used for the training set.
  vMatrix<long> TrainingIndices();

  // Labels of training set.
  vMatrix<long> TrainingLabels();

  // distance matrix for training set (training to training)
  vMatrix<float> TrainingDistances();

  vMatrix<float> TrainingPdistances();

  // indices to be used for the test set
  vMatrix<long> TestIndices();

  // Labels of Test set.
  vMatrix<long> TestLabels();

  // distance matrix for Test set (Test to Test)
  vMatrix<float> TestDistances();

  vMatrix<float> TestPdistances();
};


// vChamferEmbeddingBM will hold enough information to obtain the embedding
// of a query hand image, and to compare it to database images.
class vChamferEmbeddingBM
{
protected:
  vMatrix<float> pivot_distances;
  vMatrix<long> types_matrix;
  vMatrix<float> unique_classifiers;
  vMatrix<float> reference_images;
  vMatrix<long> indices_to_poses;
  vMatrix<float> unique_reference_images;
  vMatrix<float> weights_matrix;

  // If reference_images(i) corresponds to unique_reference_images(j),
  // then unique_indices(i) = j.
  vMatrix<long> unique_indices;

  long dimensions;
  long is_valid;

  long Initialize3(vMatrix<float> in_unique_classifiers,
                   vMatrix<float> in_pivot_distances,
                   vMatrix<long> in_indices_to_poses);

public:
  vChamferEmbeddingBM(vMatrix<float> in_unique_classifiers,
                       vMatrix<float> in_pivot_distances,
                       vMatrix<long> in_indices_to_poses);

  vChamferEmbeddingBM(const char * filename);

  ~vChamferEmbeddingBM();

  long Print();
  long Save(const char * filename);
  vMatrix<float> ReferenceImages();
  vMatrix<float> UniqueReferenceImages();
  vMatrix<float> UniqueClassifiers();
  vMatrix<long> UniqueIndices();
  vMatrix<float> Embedding(vMatrix<float> query_distances);
  vMatrix<float> ComputeReferenceImages0();
  static vMatrix<float> ComputeReferenceImages(vMatrix<float> classifiers,
                                                vMatrix<long> indices_to_poses);
  long Dimensions()
  {
    return dimensions;
  }

  long IsValid()
  {
    return is_valid;
  }

  // unique_distances are the distances of the query to each image in
  // unique_reference_images. We return a result such that result(i)
  // is the distance to the reference image stored in the i-th row of
  // reference_images. The result can be passed to the Embedding function
  // to get a valid embedding.
  vMatrix<float> ConvertUniqueDistances(vMatrix<float> unique_distances);

  // First we call ConvertUniqueDistances to get query_distances, and then
  // we call Embedding(query_distances).
  vMatrix<float> EmbeddingFromUniqueDistances(vMatrix<float> unique_distances);

  // We assume that unique_reference_images(i, 4) = j iff the j-th row
  // of reference_images corresponds to the i-th row of 
  // unique_reference_images. Here we build the inverse mapping, so that
  // result(j) = i iff unique_reference_images(i, 4) = j.
  vMatrix<long> ComputeUniqueIndices(vMatrix<float> reference_images,
                                      vMatrix<float> unique_reference_images);

  // We save to two directories, to make sure that we have backups of the
  // classifiers.
  char * Directory1();
  char * Directory2();
  char * Pathname1(const char * filename);
  char * Pathname2(const char * filename);
};                                       


// I wrote this function on 12/21/2003, before going to Greece. The goal is to 
// have a big file containing distances from some of the database objects
// to the entire database.
long vWriteBigDistancesFile(long number);

vMatrix<float> vReadBigDistancesFile(vMatrix<long> samples, long frame);


#endif // VASSILIS_CHAMFER_MATCHING_H
