#ifndef VASSILIS_CONTOUR_H
#define VASSILIS_CONTOUR_H

#include "basics/auxiliaries.h"
#include "basics/image.h"
#include "basics/algebra.h"
#include "algorithms.h"

// class vElongation is used to represent results of the call to 
// vContour::MeasureElongation. It is essentially a vMatrix<double>
// in which we store information about the elongation at a point. 
// All elongations are measured with respect to a specific point P and 
// a specific contour (to which that point is a part). With respect
// to that point, the contour is split into left points and right points.
// We want, starting at that point, to match a left segment to a right
// segment, so that the mapping is monotonic and surjective, the 
// maximum distance between two match points is minimized and the sum of
// lengths of the left and right segments is maximized. Then, the 
// "elongation" of that point for a specified min distance is the 
// ratio length/distance, where length is the sum of the lengths of the 
// segments. We get a different elongation for different minimum distances
// (I use the term "width" for minimum distance in this context.
// vElongation stores, for every width (up to a maximum), the segment (composed
// of points to the left of P, P and points to the right of P) that achieves
// the highest elongation.
class vElongation : public vMatrix<double>
{
private:
  static const long extra_number;
  static const long number_of_rows;

  long Initialize();

public:
  // Each column of the matrix corresponds to a width, and each row 
  // corresponds to a different piece of information about the width.
  // Here I basically create variables that will give a name to each
  // row of the matrix
  vArray(double) lengths;
  vArray(double) start_rows;
  vArray(double) start_cols;
  vArray(double) start_indices;
  vArray(double) end_rows;
  vArray(double) end_cols;
  vArray(double) end_indices;
  vArray(double) extra;

  // The "extra" row has information about the maximum elongation
  // found. The entries have no relation to their index (not indexed
  // by width, like the rest of the rows). Here I give names for each
  // entry in the "extra" row. Note that only a fixed number of entries
  // in the "extra" row carry information, the rest of the entries
  // are not used.
  double * max_elongation;
  double * max_length;
  double * max_width;
  double * max_index;
  double * max_start_row;

  double * max_start_col;
  double * max_start_index;
  double * max_end_row;
  double * max_end_col;
  double * max_end_index;
  
  double * starting_point_row;
  double * starting_point_col;
  double * starting_point_index;

  
  vElongation() {};
  vElongation(long columns);
  ~vElongation();

  // Max length attained for a given width, returns a negative 
  // number if width is invalid.
  inline double GetLength(double in_width)
  {
    long width = (long) in_width;
    if (width < 0) return 0;
    if (width >= cols) return 0;
    return (lengths[width]);
  }

  // Max elongation (length / width) attained for a given width, returns
  // a negative number if width is invalid.
  inline double GetElongation(double in_width)
  {
    long width = (long) in_width;
    if (width < 0) return 0;
    if (width >= cols) return 0;
    return (lengths[width]) / in_width;
  }

  // Endpoints of longest segment for a width.
  inline long LongestSegment(double in_width, long * start_row, long * start_col,
                            long * end_row, long * end_col)
  {
    long width = (long) in_width;
    if (width < 0) return 0;
    if (width >= cols) return 0;
    *start_row = (long) (start_rows[width]);
    *start_col = (long) (start_cols[width]);
    *end_row = (long) (end_rows[width]);
    *end_col = (long) (end_cols[width]);
  }

  // Returns the coordinates and index of point P.
  inline long StartingPoint(long * row, long * col, long * index)
  {
    *row = (long) (*starting_point_row);
    *col = (long) (*starting_point_col);
    *index = (long) (*starting_point_index);
    return 1;
  }

  inline double MaxElongation()
  {
    return *max_elongation;
  }

  inline double MaxElongationLength()
  {
    return *max_length;
  }

  inline double MaxElongationWidth()
  {
    return *max_width;
  }

  inline long MaxSegment(long * start_row, long * start_col, 
                         long * end_row, long * end_col)
  {
    *start_row = (long) *max_start_row;
    *start_col = (long) *max_start_col;
    *end_row = (long) *max_end_row;
    *end_col = (long) *max_end_col;
    return 1;
  }

  inline long MaxSegmentIndices(long * start_index, long * end_index)
  {
    *start_index = (long) *max_start_index;
    *end_index = (long) *max_end_index;
    return 1;
  }

  long Print();
};

// As of 10/8/2001 I have never used this class
// This class was created to be used in the following task: Suppose
// we have lots of views of hands. For each hand, we run Elongation
// Statistics to get the elongation characteristics of some protrusions.
// We want to train on those views, so that, later, when we get a test
// image of a hand and we get a protrusion (a width/length combination),
// we can answer the question: What is the probability that the tip 
// of this protrusion belongs to a finger?
//
// The way to do that is by looking at each protrusion in the training set,
// classifying its tip as finger or non-finger, and incrementing, for
// each width/length combination of the protrusion an appropriate bin
// either in finger_bins or in non_finger_bins (very similar to how I do
// skin detection).
// As of 10/8/2001 I have never used this class
class vElongationBins
{
private:
  v3dMatrix<long> * finger_bins;
  v3dMatrix<long> * non_finger_bins;
  long width_bins, length_bins;
  long width_factor, length_factor;

  long is_valid;

  inline long WidthIndex(double width)
  {
    double double_index = width / width_factor;
    long temp_index = (long) double_index;
    return Min<long>(temp_index, width_bins - 1);
  }

  inline long LengthIndex(double length)
  {
    double double_index = length * (double) length_factor;
    long temp_index = (long) double_index;
    return Min<long>(temp_index, length_bins - 1);
  }

  long ProcessElongation(vElongation * elongation, v3dMatrix<long> * bins);

public:
  vElongationBins(const char * filename);
  vElongationBins(long in_width_bins, long in_length_bins,
                   long in_width_factor, long in_length_factor);
  ~vElongationBins();

  long CleanUp();
  long ProcessFingerElongation(vElongation * elongation);
  long ProcessNonFingerElongation(vElongation * elongation);
  long Save(const char * filename);
  long IsValid();
  double FingerProbability(double width, double length);
};


static inline double S_Distance(double x1, double y1, double x2, double y2)
{
  double diff1 = x1 - x2;
  double diff2 = y1 - y2;
  return (diff1 * diff1 + diff2 * diff2);
}


// This class is used to represent contours of shapes in images. Its
// functions know how to extract contours under certain conditions
class vContour
{
private:
  // xs[i] and ys[i] will be the x and y coordinates of the i-th contour
  // point.
  vector<double> xs;
  vector<double> ys;

  // initialized is 0 if no contour has been computed yet (which happens
  // if we create a contour object using the default constructor.
  ulong initialized;

  // Ranges of values that are considered foreground.
  uchar foreground_low, foreground_high;

  // Dimensions of the image from which the contour was computed.
  long image_rows, image_cols;

  // This variable is not automatically updated, a function should call
  // number_of_points = NumberOfPoints() to make sure that
  // number_of_points has the right value, otherwise number_of_points
  // may have an outdated value.
  long number_of_points;

  // Private member functions:
  // Sets member variables to default values.
  long Initialize();

  long StartingPoint(grayscale_image * the_image, long * row, long * col);

  // Pack goes through the contour points and deletes points such that
  // their preceding and their next point are neighbors.
  long Pack();

  // Pack2 goes through the contour points and deletes repeated points.
  long Pack2();
  
  inline ulong IsForeground(uchar value)
  {
    return ((value >= foreground_low) && (value <= foreground_high));
  }

  inline long DoubleDistance(long row, long col, long index1, long index2)
  {
    long row1 = (long) ys[index1];
    long col1 = (long) xs[index1];
    long row2 = (long) ys[index2];
    long col2 = (long) xs[index2];
    long distance1 = (long) S_Distance(row, col, row1, col1);
    long distance2 = (long) S_Distance(row, col, row2, col2);
    return Max(distance1, distance2);
  }

public:
  // If we want to immediately create a contour of an image, we can call
  // the constructor(s) that take images as arguments. However, we might
  // want to specify what values should be considered foreground and 
  // what values should be considered background. The other constructors
  // treat (by default) 0 as background and non-zero as foreground. By
  // using the default constructor we can then set the foreground and 
  // background values as desired, and then call the Process member
  // function to actually get the contour.
  vContour();
  
  vContour(grayscale_image * the_image);

  ~vContour();

  // Deletes all data stored in this object.
  long CleanUp();

  inline long NumberOfPoints()
  {
    return xs.size();
  }

  vContour * Copy();

  // Sets the range of values that will be considered foreground.
  long SetForegroundRange(uchar low, uchar high);

  // Gets the values of foreground_low and foreground_high.
  long GetForegroundRange(uchar * low, uchar * high);

  // Get the vectors with the actual contour point coordinates.
  inline vector<double> * GetXs()
  {
    return &xs;
  }

  inline vector<double> * GetYs()
  {
    return &ys;
  }

  vector<long> * GetXsLong();

  vector<long> * GetYsLong();

  inline long Increment(long index)
  {
    return ((index + 1) % number_of_points);
  }

  inline long Decrement(long index)
  {
    if (index > 0) return (index - 1);
    else return (number_of_points - 1);
  }

  // Computes the outer contour of the foreground in an image. A problem
  // with this function is that, in going around the contour, represent
  // its current position with the pixel location in that position. This
  // way there are certain junctions in which there are more than one
  // possibilities for the next point, and this function just picks
  // one of those arbitrary possibilities. To address this problem
  // we need to keep track, for the current position, of the actual
  // pixel side that corresponds to the contour (i.e that is between
  // the pixel and the background) and the direction in which we 
  // moved along that side. That approach is taken in OuterContour2.
  long OuterContour2(grayscale_image * the_image);

  // This function addresses the weakness of OuterContour2 described
  // in the comments preceding OuterContour.
  long OuterContour(grayscale_image * the_image);

  
  
  // Computes the outer contour of the foreground in an image
  // starting from the given starting point. 
  //long OuterContour3(grayscale_image * the_image, long row, long col);

  // This function smooths the contour. It basically creates a new
  // contour, on which point i is the average of a number of 
  // nearby pixels in the old contour. The number of pixels used
  // is specified by neighborhood (it is 2 * neighborhood + 1)
  long Smooth1(long neighborhood);

  // This function is like Smooth1, but instead of modifying the 
  // object from which it is called, it leaves the object unchanged
  // and returns the results in another object.
  vContour * Smooth2(long neighborhood);

  // Returns an image that is 0 everywhere except for the contour
  // points, which have the value 255.
  grayscale_image * ContourImage(uchar value = 255);

  // Draws the contour onto the image (sets contour points to value).
  long ContourImage2(grayscale_image * the_image, uchar value = 255);

  // Returns an image that is 0 everywhere except for the contour points 
  // between the start index and the end index (in increasing order), where
  // it has the specified value.
  grayscale_image * SegmentImage(long start_index, long end_index, 
                             uchar value = 255);

  // Sets the pixel value of points in the segment specified by start_index
  // and end_index to value
  long SegmentImage2(grayscale_image * the_image, long start_index, long end_index, 
                     uchar value = 255);

  long SegmentImage3(color_image * the_image, long start_index, 
                    long end_index, uchar value1, uchar value2, uchar value3);
							  
  // Region boundary points are points that segment a contour 
  // into segments, such
  // that each segment is a (more or less) convex region. The "threshold"
  // argument specifies how much we tolerate non-convexities. Basically,
  // if we have a cut point A, to find the next, we look at each point in
  // order starting at the point after the cut point. First we find a point
  // B whose orientation is different than the cut point's by more than the
  // threshold. From that point on we know (based on B) the way the next
  // segment will turn (left or right). Now we are looking for the endpoint.
  // Whenever we find a point whose difference in orientation from A is 
  // the most we have seen so far (for point A) we keep track of that. When
  // we find a point C whose difference in orientation with respect to A is
  // less the max difference we have seen so far minus the threshold, then
  // we pick B as the next cut point. 
  // 
  // The neighborhood argument specifies the neighborhood that we use
  // to calculate first and second derivatives, it is passed as is to 
  // the vCircularDerivative function.
  // We put the indices of the boundary points in "indices". The direction
  // of the segment starting at (*indices)[i] is saved in (*directions)[i],
  // it is either 1 or 0. The length of each curve segment is saved in 
  // lengths.
  long RegionBoundaries5(vector<long> * indices, vector<long> * directions,
                  vector<long> * lengths,
                  long neighborhood, double threshold);

  // Like RegionBoundaries5, but it also returns the magnitudes and orientations
  // of the contour points in the output arguments "magnitudes" and
  // "orientations"
  long RegionBoundaries7(vector<long> * indices, vector<long> * directions,
                  vector<long> * lengths,
                  long neighborhood, double threshold,
                  vector<double> * magnitudes, vector<double> * orientations);

  // Cut points are like region boundaries, but there is more of 
  // them. In particular, whereas a contour segment between two
  // boundary points can change rotation by more than 180 degrees,
  // CutPoints makes sure that this doesn't happen. This is useful
  // in things like locating fingers, where, if we allow a segment
  // to go over 180 degrees in rotation we can have two fingertips
  // in the same segment.
  long CutPoints5(vector<long> * indices, vector<long> * directions,
                  vector<long> * lengths,
                  long neighborhood, double threshold);

  long CutPoints7(vector<long> * indices, vector<long> * directions,
                  vector<long> * lengths,
                  long neighborhood, double threshold,
                  vector<double> * magnitudes, vector<double> * orientations);

  // Here we find points of maximal curvature in the contour. The
  // arguments "indices", "values" and "directions" are output arguments,
  // that's where we put the indices, values of the magnitude of the 
  // second derivative, and directions of curve rotation for each maximum.
  long Maxima5(vector<long> * indices, vector<double> * values,
              vector<long> * directions,
              long neighborhood, double threshold);

  // This function is like the previous one, but instead of computing
  // the magnitudes of the second derivative and the orientation angles
  // at each contour point, it gets them as arguments.
  long Maxima8(vector<long> * indices, vector<double> * values,
               vector<long> * directions, long neighborhood,
               double threshold, vector<long> * cuts,
               vector<double> * magnitudes, vector<double> * orientations);

  // This function takes in as argument the index to a point in the contour.
  // It returns a matrix whose i-th entry is the maximum length of a contour
  // segment that contains the indexed point and has width i. Widths and 
  // lengths are measured in pixels. The length of a contour segment is 
  // defined as the number of contour points between the endpoints of the segment.
  // The width definition is much trickier, and depends on the "index" argument.
  // It is obtained as follows: Match point-by-point
  // the contour segment on the left of the 
  // indexed point to the segment on the right of the indexed point using
  // dynamic programming, so that the maximum distance between any two
  // matched points is minimized. That maximum distance is defined as the width.
  //
  // I should note the following: I wrote this function so that I could detect
  // the extent of a finger in a hand contour, assuming that the index is 
  // the location of a finger tip (found with contour_maxima and having the
  // right curvature). It has the following problem: It often overshoots the 
  // finger ends and picks up neighboring points, because all it cares about
  // is finding contour segments of maximum length for a given width and 
  // starting at the index. The FingerFilter function can be used to
  // trim the maximum elongation found by this function and decide if
  // it can possibly be a finger or not (segments with overall negative 
  // curvatures, for example, should be rejected).
  vElongation
  MeasureElongation(long index, long dp_length, double threshold);

  double MeasureElongation2(long index, long dp_length, double threshold);

  vBoundingBox BoundingBox();

  // This function basically calls ElongationStatistics3, and supplies
  // to it reasonable default arguments for dp_length and threshold.
  vElongation ElongationStatistics1(long index);


  // This function behaves like MeasureElongation, takes the same arguments
  // and returns the same type of result. The only difference is that, in the
  // result, the unit size
  // in which widths and lengths are measured is 1% of dp_length 
  vElongation ElongationStatistics3(long index, long dp_length, 
                                     double threshold);

  // This function was used to try out an idea of mine about how
  // to get a skeleton-like feature efficiently. It ends up
  // getting stuck if some local orientation is not cooperating. The idea 
  // was, start by picking left point, right point, and skeleton point,
  // all being the index point. Then, at each step, advance one of
  // either left point or right point, and look at the neighborhood
  // of current skeletal point for the next skeletal point. The next 
  // skeletal point is chosen as follow. Get the max distance from
  // distances from that point to the
  // next left point and current right point. Then, do the same for
  // that point and the current left and next right point. Pick 
  // the minimum of those maxes. Choose as next skeletal point the
  // one whose minimum is the lowest.
  // 
  // At any rate, I'm keeping the function here, maybe I'll get 
  // a better idea on how to fix it later.
  vMatrix<double> GrowSkeleton(long index, long dp_length, 
                                grayscale_image * silhouette);

  // FingerFilter takes as input an elongation that can possibly be
  // a finger. It trims the endpoints of the max elongation, using 
  // the following rule: Consider the line connecting the two finger
  // endpoints. If we start at one of the endpoints and we start following
  // that line, the next pixel after the endpoint should be an interior 
  // pixel of the hand. In most cases that will also mean that the line
  // is entirely in the interior of the hand (so it can possibly
  // be the base line of the finger).
  long FingerFilter(vElongation * elongation);

  // A quick and dirty finger detector. dp_length and width_threshold
  // are the parameters passed to ElongationStatistics3. neighborhood
  // and curvature_threshold are the arguments passed to 
  // ContourMaxima. Fingers whose contour length is less than 
  // length_threshold are rejected, and if two fingers share a 
  // contour segment whose length is more than overlap_threshold,
  // then only the longest of them is retained.
  long FindFingers9(vector<vElongation> * results, long dp_length,
                    double width_threshold, long neighborhood,
                    double curvature_threshold, double elongation_threshold,
                    long length_threshold, long min_width, long overlap_threshold);

  // Like FindFingers9. It calls FindFingers9, but it first picks some 
  // default values for the arguments.
  long FindFingers1(vector<vElongation> * results);

  // Like FindFingers1, but it lets the caller also specify the 
  // elongation threshold.
  long FindFingers2(vector<vElongation> * results, 
                    double elongation_threshold);

  long BorderContour(grayscale_image * the_image, long start_row, long start_col,
                     grayscale_image * to_ignore);
};


class vFingerInfo
{
public:
  long left_row;
  long left_col;
  long right_row;
  long right_col;
  long tip_row;
  long tip_col;
  double width;
  double elongation;
  double orientation;
  double base_row;
  double base_col;

  vFingerInfo() {}
  vFingerInfo(vElongation elongation);

  long Print();
};


class vSequenceFingerInfo
{
public:
  long is_valid;
  long number_of_frames;

  // indices[i] is the index of the first finger info for frame i,
  // in the frame_fingers array.
  long * indices;

  // numbers[i] is the number of finger infos saved for frame i.
  long * numbers;

  // The finger_info objects for all frames.
  vector<vFingerInfo> frame_fingers;

  vSequenceFingerInfo(const char * filename);
  ~vSequenceFingerInfo();

  inline vFingerInfo & Get(long frame, long in_frame_index)
  {
    long index = indices[frame] + in_frame_index;
    return frame_fingers[index];
  }

  long Print(long start_frame, long end_frame);

  vMatrix<double> Distances(vector<vElongation> * elongations);
  double Distance(long frame, vector<vElongation> * elongations);
};


// For each frame in sequence_file, it finds fingers using 
// vContour::FindFingers2, and saves the results in a .fingers
// file. The format of the file is:
// frame1 ... frame_n
// where each frame is:
// left_start_row
// left_start_col
// right_start_row
// right_start_col
// tip_row
// tip_col
// width
// elongation
// orientation
// If end_frame is negative, it is set to the last frame in the sequence.
// It ignores fingers whose finger_tip has the value ignore_value
long vSaveFingerInfo(const char * sequence_file, double low_threshold,
                      long start_frame, long end_frame, long ignore_value);




#endif // VASSILIS_CONTOUR_H
