#include "stdafx.h"

#include "features.h"
#include "algo_templates.h"


#include "basics/definitions.h"


// vDirectionChanges computes the average change in pixel values,
// as we go from one line pixel to a neighboring line pixel. The line
// is specified by center (row and column), length, and direction (in radians).
// This is useful for the vDominantOrientation function.
float vDirectionChanges(v3dMatrix<float> * the_image, long row, 
                        long col, long lengthl, float direction)
{
  float length = (float) lengthl;
  long rows = the_image->Rows();
  long cols = the_image->Cols();

  float col_step = cos(direction);
  float row_step = sin(direction);
  float half_length = length / (float) 2.0;

  if ((row + half_length >= rows) || (col + half_length >= cols) ||
      (half_length > row) || (half_length > col))
  {
    return -1;
  }
  
  float start_row = ((float) row) - row_step * half_length;
  float start_col = ((float) col) - col_step * half_length;

  float end_row = ((float) row) + row_step * half_length;
  float end_col = ((float) col) + col_step * half_length;

  if ((start_row > rows-2) || (start_row < 1) ||
      (end_row > rows-2) || (end_row < 1) ||
      (start_col > cols-2) || (start_col < 1) ||
      (end_col > cols-2) || (end_col < 1))
  {
    return -1;
  }

  float sum = 0;
  long bands = the_image->Bands();
  long i;
  float previous_row = start_row;
  float previous_col = start_col;

  for (i = 1; i <= lengthl; i++)
  {
    long band;
    float current_row = start_row + ((float) i) * row_step;
    float current_col = start_col + ((float) i) * col_step;
    for (band = 0; band < bands; band++)
    {
      float previous_value = vBilinearInterpolation(the_image->Matrix2(band), 
                                                     previous_row, previous_col);
      float current_value = vBilinearInterpolation(the_image->Matrix2(band), 
                                                    current_row, current_col);
      sum = sum + vAbs(current_value - previous_value);
    }
    previous_row = current_row;
    previous_col = current_col;
  }

  float result = sum / length;
  return result;
}


// Returns a matrix containing results of vDirectionChanges for 
// many different orientations, sampled uniformly. The number of
// samples is specified by the "number" argument.
vMatrix<float> vAllDirectionChanges(v3dMatrix<float> * the_image, long row, 
                                    long col, long lengthl, long numberl)
{
  float numberf = (float) numberl;
  vMatrix<float> result(1, numberl);

  long i;
  for (i = 0; i < numberl; i++)
  {
    float direction = (float) M_PI / numberf * (float) i;
    result(i) = vDirectionChanges(the_image, row, col, lengthl, direction);
  }

  return result;
}


// here we find, for the point specified at row,col, and the scale
// specified by length, the orientation where the pixel values change
// the least.
vMatrix<float> vDominantOrientation(v3dMatrix<float> * the_image, long row,
                                    long col, long length)
{
  vMatrix<float> changes = vAllDirectionChanges(the_image, row, col, length, 60);

  long number = changes.Size();
  long min_index = -1;
  float min_value = 1000000000;
  float max_index = -1;
  float max_value = -1;
  long i;
  for (i = 0; i < number; i++)
  {
    float current = changes(i);
    if (current < 0)
    {
      continue;
    }
    if (current < min_value)
    {
      min_value = current;
      min_index = i;
    }
    if (current > max_value)
    {
      max_value = current;
      max_index = (float) i;
    }
  }

  vMatrix<float> result(1, 4);
  result(0) = ((float) min_index) * (float) M_PI / (float) number;
  result(1) = min_value;
  result(2) = ((float) max_index) * (float) M_PI / (float) number;
  result(3) = max_value;

  return result;
}


// Uniformly samples "number" dimensions from 0 to PI.
vMatrix<float> vSampleDirections(long number)
{
  vMatrix<float> result(1, number);
  float numberf = (float) number;

  long i;
  for (i = 0; i < number; i++)
  {
    float direction = (float) M_PI / numberf * (float) i;
    result(i) = direction;
  }

  return result;
}


// Uniformly samples "number" numbers from min_value to max_value.
vMatrix<long> vSampleLongs(long min_value, long max_value, long number)
{
  vMatrix<long> result(1, number);
  float numberf = (float) number;
  float diff = (float) (max_value - min_value)
    ;
  float minf = (float) min_value;
  float maxf = (float) max_value;
  float step = diff / (numberf - (float) 1.0);

  long i;
  for (i = 0; i < number; i++)
  {
    float sample = min_value + ((float) i) * step;
    result(i) = (long) sample;
  }

  return result;
}



// for the i-th direction in directions, we store in the i-th row
// of the result two values: the row_diff and col_diff that correspond
// to a unit vector going towards that direction.
vMatrix<float> vDirectionDiffs(v3dMatrix<float> * directions)
{
  long number = directions->Size();
  vMatrix<float> result(number, 2);
  vArray(float) values = directions->Matrix();
    
  long i;
  for (i = 0; i < number; i++)
  {
    float direction = values[i];
    float col_step = cos(direction);
    float row_step = sin(direction);
    result(i, 0) = row_step;
    result(i, 1) = col_step;
  }

  return result;
}




// We return a matrix containing the average change in pixel values as
// we go from one pixel to its neighboring pixels along a given direction.
// The number of directions evaluated is specified by argument "number".
// Directions are sampled uniformly. Pixels whose intensity is 0 in
// all bands are not used.
vMatrix<float> vImageDirectionChanges(v3dMatrix<float> * the_image, long number)
{
  vMatrix<float> directions = vSampleDirections(number);
  vMatrix<float> diffs = vDirectionDiffs(&directions);

  float count = 0;
  vMatrix<float> sums(1, number);
  vWriteValue(&sums, (float) 0.0);

  vArray3(float) values = the_image->Matrix3();
  long rows = the_image->Rows();
  long cols = the_image->Cols();
  long bands = the_image->Bands();

  long row, col, band;
  for (row = 2; row < rows - 2; row++)
  {
    for (col = 2; col < cols - 2; col++)
    {
      float sum = 0;
      for (band = 0; band < bands; band++)
      {
        sum = sum + values[band][row][col];
      }
      if (sum == (float) 0)
      {
        continue;
      }
      count = count + 1;
      
      long i;
      for (i = 0; i < number; i++)
      {
        float row_diff = diffs(i, 0);
        float col_diff = diffs(i, 1);
        float previous_row = ((float) row) - row_diff;
        float previous_col = ((float) col) - col_diff;
        float next_row = ((float) row) + row_diff;
        float next_col = ((float) col) + col_diff;
        float sum = 0;
        for (band = 0; band < bands; band++)
        {
          float previous_value = vBilinearInterpolation(the_image->Matrix2(band), 
                                                         previous_row, previous_col);
          float current_value = values[band][row][col];
          float next_value = vBilinearInterpolation(the_image->Matrix2(band), 
                                                     next_row, next_col);
          sum = sum + vAbs(current_value - previous_value);
          sum = sum + vAbs(current_value - next_value);
        }
        sums(i) = sums(i) + sum;
      }
    }
  }

  long i;
  for (i = 0; i < number; i++)
  {
    sums(i) = sums(i) / count;
  }

  return sums;
}


vImageOrientations::vImageOrientations(v3dMatrix<uchar> * the_image, 
                                       long in_orientations,
                                       long in_min_scale, long in_max_scale, 
                                       long in_scale_steps)
{
  Initialize();

  if ((the_image == 0) || (in_orientations < 2) || (in_min_scale < 2) ||
      (in_max_scale < in_min_scale) || (in_scale_steps < 2))
  {
    return;
  }

  orientations = vSampleDirections(in_orientations);
  scales = vSampleLongs(in_min_scale, in_max_scale, in_scale_steps);

  rows = the_image->Rows();
  cols = the_image->Cols();
  long bands = the_image->Bands();
  vHyperArray2(&values, rows, cols);

  vArray3(uchar) matrix = the_image->Matrix3();
  long row, col, band;
  v3dMatrix<float> * displacements = Displacements(orientations, scales);
  vPrint("\n");

  for (row = 0; row < rows; row++)
  {
    for (col = 0; col < cols; col++)
    {
      long zero_flag = 1;
      for (band = 0; band < bands; band++)
      {
        if (matrix[band][row][col] != 0)
        {
          zero_flag = 0;
        }
      }

      if (zero_flag == 1)
      {
        values[row][col] = vZero(vArray(float));
        continue;
      }

      // compute the direction changes for that pixel, for 
      // all orientations and scales.
      vMatrix<float> changesm = DirectionChanges(the_image, row, col, displacements);
      vHyperArray2(&(values[row][col]), changesm.Rows(), changesm.Cols());
      vArray2(float) current = values[row][col];
      long i, j;
      for (i = 0; i < changesm.Rows(); i++)
      {
        for (j = 0; j < changesm.Cols(); j++)
        {
          float value = changesm(i,j);
          if (value >= 0)
          {
            long max_dist = scales(j) / 2;
            value = value / (float) (2 * max_dist);
          }
          current[i][j] = value;
        }
      }
      vPrint("done with (%5li, %5li)\r", row, col);
    }
  }
  vPrint("\n");

  vdelete(displacements);
}


long vImageOrientations::Initialize()
{
  values = vZero(vArray3(float));
  return 1;
}


vImageOrientations::~vImageOrientations()
{
  long row, col;
  for (row = 0; row < rows; row++)
  {
    for (col = 0; col < cols; col++)
    {
      if (!vIsZero(values[row][col]))
      {
        vDelete2(values[row][col], orientations.Size());
      }
    }
  }

  vDelete2(values, rows);  
}


v3dMatrix<float> * vImageOrientations::Displacements(vMatrix<float> orientations, 
                                                     vMatrix<long> scales)
{
  if ((orientations.Size() < 1) || (scales.Size() < 1) || 
      (orientations.IsValid() <= 0) || (scales.IsValid() <= 0))
  {
    vPrint("invalid arguments to vImageOrientations::Displacements\n");
    return 0;
  }

  // The scales should appear in strictly increasing order.
  long scales_error_flag = 0;
  long current_scale = scales(0);
  long i, j;
  for (i = 1; i < scales.Size(); i++)
  {
    long scale = scales(i);
    if (scale < current_scale)
    {
      scales_error_flag = 1;
      break;
    }
    current_scale = scale;
  }

  if (scales_error_flag != 0)
  {
    exit_error("error: in Displacements(), decreasing scales\n");
  }

  long max_dist = scales(scales.Size() - 1) / 2;

  // scale_indices maps lengths to indices of scales. Each length maps
  // to the index of the smallest scale that is at least as large as
  // that length.
  vMatrix<long> scale_indices(1, max_dist+1);
  long index = 0;
  for (i = 0; i <= max_dist; i++)
  {
    while(i > scales(index) / 2)
    {
      index++;
      if (index > max_dist)
      {
        exit_error("error: something is wrong in Displacements\n");
      }
    }
    scale_indices(i) = index;
  }

  v3dMatrix<float> * resultm = new v3dMatrix<float>(3, 2 * max_dist + 1, 
                                                    orientations.Size());
  vArray3(float) result = resultm->Matrix3();

  for (i = 0; i < orientations.Size(); i++)
  {
    float direction = orientations(i);
    float col_step = cos(direction);
    float row_step = sin(direction);
    long counter = 0;
    
    for (j = -max_dist; j <= max_dist; j++)
    {
      float row = ((float) j) * row_step;
      float col = ((float) j) * col_step;
      long index = scale_indices(vAbs(j));

      result[i][0][counter] = row;
      result[i][1][counter] = col;
      result[i][2][counter] = (float) index;
      counter++;
    }
  }

  return resultm;
}


vMatrix<float> vImageOrientations::DirectionChanges(v3dMatrix<uchar> * the_image, long row,
                                                    long col, v3dMatrix<float> * displacements)
{
  long orientations = displacements->Bands();
  long lengths = displacements->Cols();
  vArray3(float) disp = displacements->Matrix3();

  long max_scale_index = 0;
  long i, j;
  for (i = 0; i < lengths; i++)
  {
    long current_scale_index = vRound(disp[0][2][i]);
    if (current_scale_index > max_scale_index)
    {
      max_scale_index = current_scale_index;
    }
  }

  long scales = max_scale_index + 1;
  vMatrix<float> resultm(orientations, scales);
  vArray2(float) result = resultm.Matrix2();
  long rows = the_image->Rows();
  long cols = the_image->Cols();
  long bands = the_image->Bands();
  vMatrix<float> previousm(1, bands);
  vArray(float) previous = previousm.Matrix();
  vMatrix<float> changesm(1, scales);
  vArray(float) changes = changesm.Matrix();
  vMatrix<short> out_of_bounds_flags(1, scales);
  vArray(short) flags = out_of_bounds_flags.Matrix();

  for (i = 0; i < orientations; i++)
  {
    float row_diff = disp[i][0][0];
    float col_diff = disp[i][1][0];
    long index = vRound(disp[i][2][0]);
    float previous_row = row + row_diff;
    float previous_col = col + col_diff;

    long band, previous_flag, current_flag;
    if ((previous_row > rows-2) || (previous_row < 1) ||
        (previous_col > cols-2) || (previous_col < 1))
    {
      flags[index] = 1;
      previous_flag = 1;
    }
    else
    {
      for (band = 0; band < bands; band++)
      {
        previous[band] = vBilinearInterpolation(the_image->Matrix2(band),
                                                 previous_row, previous_col);
      }
      previous_flag = 0;
    }

    vWriteValue(&changesm, (float) 0);
    vWriteValue(&out_of_bounds_flags, (short) 0);
    for (j = 1; j < lengths; j++)
    {
      current_flag = 0;
      row_diff = disp[i][0][j];
      col_diff = disp[i][1][j];
      index = vRound(disp[i][2][j]);
      float current_row = row + row_diff;
      float current_col = col + col_diff;
      if ((current_row > rows-2) || (current_row < 1) ||
          (current_col > cols-2) || (current_col < 1))
      {
        current_flag = 1;
        flags[index] = 1;
      }
      
      float sum = 0;
      if (previous_flag == 1)
      {
        flags[index] = 1;
      }

      if (current_flag == 0)
      {
        for (band = 0; band < bands; band++)
        {
//          float current_value = vBilinearInterpolation(the_image->Matrix2(band),
//                                                        current_row, current_col);
          float current_value = 0;
          sum += vAbs(previous[band] - current_value);
          previous[band] = current_value;
        }
        changes[index] += sum;
      }
      previous_flag = current_flag;
      
    }

    if (flags[0] == 1)
    {
      result[i][0] = -1;
    }
    else
    {
      result[i][0] = changes[0];
    }
    for (j = 1; j < scales; j++)
    {
      if (flags[j] == 1)
      {
        result[i][j] = -1;
      }
      else
      {
        // just in case there is a bug:
        if (flags[j-1] == 1)
        {
          exit_error("error: something is wrong in DirectionChanges\n");
        }
        result[i][j] = result[i][j-1] + changes[j];
      }
    }
  }

  return resultm;
}


vMatrix<float> vImageOrientations::Orientations()
{
  return vMatrix<float>(&orientations);
}


vMatrix<long> vImageOrientations::Scales()
{
  return vMatrix<long>(&scales);
}


vMatrix<float> vImageOrientations::PixelValues(long row, long col)
{
  if (vIsZero(values[row][col]))
  {
    return vMatrix<float>();
  }

  vMatrix<float> resultm(orientations.Size(), scales.Size());
  vArray2(float) pixel_values = values[row][col];
  long i, j;
  for (i = 0; i < orientations.Size(); i++)
  {
    for (j = 0; j < scales.Size(); j++)
    {
      resultm(i, j) = pixel_values[i][j];
    }
  }

  return resultm;
}
