#ifndef VASSILIS_FEATURES_H
#define VASSILIS_FEATURES_H

#include "basics/algebra.h"

// This file should contain code that extracts features from 
// hand images.



// vDirectionChanges computes the average change in pixel values,
// as we go from one line pixel to a neighboring line pixel. The line
// is specified by center (row and column), length, and direction (in radians).
// This is useful for the vDominantOrientation function.
float vDirectionChanges(v3dMatrix<float> * the_image, long row, 
                        long col, long length, float direction);


// Returns a matrix containing results of vDirectionChanges for 
// many different orientations, sampled uniformly. The number of
// samples is specified by the "number" argument.
vMatrix<float> vAllDirectionChanges(v3dMatrix<float> * the_image, long row, 
                                    long col, long length, long number);


// here we find, for the point specified at row,col, and the scale
// specified by length, the orientation where the pixel values change
// the least.
vMatrix<float> vDominantOrientation(v3dMatrix<float> * the_image, long row,
                                    long col, long length);

// Uniformly samples "number" dimensions from 0 to PI.
vMatrix<float> vSampleDirections(long number);

// Uniformly samples "number" numbers from min_value to max_value.
vMatrix<long> vSampleLongs(long min_value, long max_value, long number);

// for the i-th direction in directions, we store in the i-th row
// of the result two values: the row_diff and col_diff that correspond
// to a unit vector going towards that direction.
vMatrix<float> vDirectionDiffs(v3dMatrix<float> * directions);

// We return a matrix containing the average change in pixel values as
// we go from one pixel to its neighboring pixels along a given direction.
// The number of directions evaluated is specified by argument "number".
// Directions are sampled uniformly. Pixels whose intensity is 0 in
// all bands are not used.
vMatrix<float> vImageDirectionChanges(v3dMatrix<float> * the_image, long number);



// given an image of a hand, we want to know for each non-zero
// pixel location, each scale, and each orientation the sum of 
// intensity changes along the line centered at that location and
// with the given orientation and scale. 
class vImageOrientations
{
protected:
  vArray4(float) values;

  vMatrix<float> orientations;
  vMatrix<long> scales;
  long rows, cols;

  long Initialize();

public:
  vImageOrientations(v3dMatrix<uchar> * the_image, long in_orientations,
                     long in_min_scale, long in_max_scale, long in_scale_steps);
  ~vImageOrientations();

  static v3dMatrix<float> * Displacements(vMatrix<float> orientations, 
                                          vMatrix<long> scales);
                       
  static vMatrix<float> DirectionChanges(v3dMatrix<uchar> * the_image, long row,
                                         long col, v3dMatrix<float> * displacements);

  vMatrix<float> Orientations();
  vMatrix<long> Scales();
  vMatrix<float> PixelValues(long row, long col);
};






#endif // VASSILIS_FEATURES_H
