#ifndef VASSILIS_FINGERS_H
#define VASSILIS_FINGERS_H


#include "algorithms.h"
#include "contour.h"
#include "ranking.h"

// This class allows us to conveniently package together a contour segment
// and the id of the link that it corresponds to
class vFingerContour
{
public:
  long link_id;
  vContour * contour;

  vFingerContour(long in_id, vContour * in_contour)
  {
    link_id = in_id;
    contour = in_contour;
  }
};


// This class is used as an interface to access directly the set of 
// points in the sequence of database edge images. We can load up
// the set of points in all edge images in memory, and access
// the set of points of each image separately.
// We don't allow point coordinates to exceed 255, so that they
// can fit into 8 bits.
// We make a distinction between points in general and leftover points.
// Leftover points are points that do not correspond to lines (of course,
// what points are left over depends on what lines we extract). At any
// rate, we use the convention here that of all points of a particular image,
// the leftover points are stored first, and leftover_lengths indicates
// how many leftover points there are.
class vPointSetSequence
{
private:
  long rows, cols;
  long number_of_frames;
  long total_length;
  vArray(long) lengths;
  vArray(long) leftover_lengths;
  vArray(long) indices;
  vArray(uchar) point_rows;
  vArray(uchar) point_cols;

  inline void CheckFrame(long frame)
  {
    if ((frame < 0) || (frame >= number_of_frames))
    {
      exit_error("PointSetSequence error\n");
    }
  }

public:
  vPointSetSequence(vVideoFile * sequence);
  vPointSetSequence() {}
  ~vPointSetSequence();

  static vPointSetSequence * Load(const char * filename);
  long Save(const char * filename);

  inline long Rows()
  {
    return rows;
  }

  inline long Cols()
  {
    return cols;
  }

  inline long NumberOfFrames()
  {
    return number_of_frames;
  }

  inline long TotalLength()
  {
    return total_length;
  }

  inline long Length(long frame)
  {
    CheckFrame(frame);
    return lengths[frame];
  }

  inline long LeftoverLength(long frame)
  {
    CheckFrame(frame);
    return leftover_lengths[frame];
  }

  inline vArray(uchar) PointRows(long frame)
  {
    CheckFrame(frame);
    long index = indices[frame];
    return point_rows + index;
  }

  inline vArray(uchar) PointCols(long frame)
  {
    CheckFrame(frame);
    long index = indices[frame];
    return point_cols + index;
  }

  // returns the number of points.
  // In rowsp and colsp we put arrays of the (row, col)
  // coordinates of the set of points corresponding
  // to the given frame and given view
  long GetPoints(long frame, long view, 
                 vArray(long) * rowsp, vArray(long) * colsp);

  long MarkLeftover(long frame, vector<long> * leftover_rows, 
                    vector<long> * leftover_cols);

  long MarkLeftovers(vVideoFile * model_images,
                     vVideoFile * model_edges,
                     v3dMatrix<uchar> * all_model_lines);
};


// This is an auxiliary class, used to identify link groups (consisting of
// one, two or three links) that belong to the same finger and that appear
// more or less in a straight line. Link values of 0 mean that there is no 
// link (for cases where the group consists of only one or two links).
// quality indicates the "goodness" of the link, i.e. how reliable it is for
// giving good and stable features that will likely be observed in the 
// input image. The two lines are the more or less parallel lines that 
// delineate the link group
class vStraightLinks
{
public:
  long link1, link2, link3;
  double quality;
  vEllipse ellipse;
  vLineSegment line1;
  vLineSegment line2;
};


class vDOLVS_Histogram
{
private:
  long D_length, O_length, L_length, vlength, S_length;
  long D_values, O_values, L_values, vvalues, S_values;

  long histogram[25][10][10][10][10];
  long D_indices[400];
  long O_indices[360];
  long L_indices[500];
  long vindices[1000];
  long S_indices[2000];

  long D_inverse[25];
  long O_inverse[10];
  long L_inverse[10];
  long vinverse[10];
  long S_inverse[10];

  // 5/12/2003: what is DOL used for?
  long DOL[25][10][10];
  vArray(long) input_summary;
  long input_length;
  long items_per_segment;

public:
  vDOLVS_Histogram();
  ~vDOLVS_Histogram();

  long Add(vDOLVS_Histogram * histogram);
  long Zero();
  vDOLVS_Histogram * Copy();

  // In preparation for computing the existential DOLVS and the PDF,
  // here we create an array of longs (input_summary) that has,
  // precomputed, all the information that we need from input_lines.
  // input_summary has 9 (items_per_segment) entries per input line,
  // they occur in the following order:
  // center_row, center_col, orientation (degrees), length, votes, strength
  // The last 3 of the nine entries are used to save the differences between
  // the input segment and some model line in DOL, that is used
  // in MarkEntries.
  long MakeInputSummary(vector<vLineSegmentMatch> * input_lines);

  // If histogram2 is equal to this, it returns 1. Else, 
  // it returns the negative of the number of bins where they differ.
  long Equal(vDOLVS_Histogram * histogram2);
  long MakeCumulative();
  long MakeDOLCumulative();

  // Here, instead of summing up everything that occured before, we simply
  // put a 1 if something has occured before, 0 otherwise.
  long MakeExistentialCumulative2();

  // Here we assume that individual matches are marked as bits, so that
  // we can process entries from 32 model lines in parallel. See comments
  // on ExistentialFrequencies.
  long MakeExistentialCumulative();

  // Propagate is like making cumulative, except that instead of summing up it 
  // writes the highest value it has seen so far into the next bins.
  // For example, Cumulative({0,2,0,1,0}) = {0,2,2,3,3}, whereas
  // Propagate({0,2,0,1,0}) = {0,2,2,2,2}
  long Propagate();
  long Save(const char * filename);
  long Load(const char * filename);

  // Here we calculate, for each match between the model link_segment
  // and one of the input segments, the number of matches (between
  // the same link_segment and other input segments) that are better
  // than or equal to this match, in all POLVS dimensions. This is an older
  // and slower version, that I'm just keeping around for the time being. 
  // This version doesn't use input_summary, and it's slow because
  // it is precomputing a lot of things over and over again.
  long ProcessLinkSegment2(vLineSegment link_segment, 
                           vector<vLineSegmentMatch> * input_segments);

  // This function is called when we have already built the existential
  // DOLVS, and each bin contains either frequencies, or probabilities.
  // It finds, for each match of link_segment, the corresponding entry,
  // and it returns the smallest among those entries. It uses input_summary,
  // so it is a faster version of SmallestMatchFrequency2, that computes
  // the same thing.
  long SmallestMatchFrequency(vLineSegment link_segment);

  // Here, in addition to returning the smallest frequency, we store
  // in best_match_index the index of the line segment that attained
  // that smallest frequency.
  long SmallestMatchFrequency(vLineSegment link_segment,
                              long * best_match_index);

  // Older and slower version of SmallestMatchFrequency.
  long SmallestMatchFrequency2(vLineSegment link_segment, 
                              vector<vLineSegmentMatch> * input_segments);

  // This is useful when we have a list of candidate line segments that
  // we should consider (as opposed to considering all input segments).
  // Those candidates can be obtained, for example, using a line grid.
  // Also, this function doesn't require ProcessLinkSegment to be called
  // first, it does the job of that function as well, but only for the 
  // given candidates, so that it saves time.
  long SmallestMatchFrequency3(vLineSegment link_segment, 
                               vArray(long) candidates,
                               long candidate_number); 
  

  // Here we use the dolvs entries in each of the input segments
  // (as computed , for example, by ProcessLinkSegment) and we mark
  // them on to the histogram (each bin receives as many votes
  // as the number of segments whose DOLVS entries fall into that bin).
  // Older and slower version of MarkEntries, this version doesn't use
  // input_summary.
  long MarkEntries2(vector<vLineSegmentMatch> * input_segments);

  // In this function, we assume that the DOLVS of each segment with respect
  // to a given model line has already
  // been computed (and stored in that segment's dolvs variables). This function
  // gives as score to each segment the corresponding entry in the histogram,
  // and sorts the segments in ascending order.
  long RankInputSegments(vector<vLineSegmentMatch> * segments);

  // Return all input segments, ranked in order of similarity to 
  // the model line specified by the argument. It calls FrequencyRanking.
  vArray(long) FrequencyRanking1(vLineSegment model_line);

  // Return all input segments, ranked in order of similarity to 
  // the model line specified by the argument. It calls FrequencyRanking5.
  vArray(long) FrequencyRanking2(vLineSegment model_line, 
                                  vector<long> * frequencies);

  // Return all input segments, ranked in order of similarity to 
  // the model line specified by the arguments.
  vArray(long) FrequencyRanking4(long center_row, long center_col, 
                                  long orientation, long length);

  // Return all input segments, ranked in order of similarity to 
  // the model line specified by the arguments. It puts the 
  // frequency corresponding to each input segment into the
  // frequencies array.
  vArray(long) FrequencyRanking5(long center_row, long center_col, 
                                  long orientation, long length,
                                  vector<long> * frequencies);

  inline void GetIndices(long d, long o, long l, long v, long s,
                         long & id, long & io, long & il, long & iv, long & is)
  {
    id = D_indices[d];
    io = O_indices[o];
    il = L_indices[l];
    iv = vindices[v];
    is = S_indices[s];
  }


  // What we compute here is the following: For each DOLVS bin, how common
  // is it for a random model line to have among the input lines a match
  // that is as good or better than that DOLVS bin. Model lines
  // are sampled from model_lines_matrix uniformly, as specified by
  // the arguments "start" and "period".
  // Older and slower version of ExistentialFrequencies, creates a 
  // cumulative DOLVS histogram for each model line.
  long ExistentialFrequencies2(v3dMatrix<uchar> * model_lines_matrix,
                               vector<vLineSegmentMatch> * input_lines,
                               long start, long period);

  // Here we assume that this object is an existential DOLVS. We return
  // a vector, which, at entry i, indicates the number of random model
  // lines for which the best DOLVS match had existential frequency less than or 
  // equal to i. 
  // Faster version of PDF2, that uses input_summary
  vMatrix<long> PDF(v3dMatrix<uchar> * model_lines_matrix,
                     long start, long period);

  // Older and slower version of PDF.
  vMatrix<long> PDF2(v3dMatrix<uchar> * model_lines_matrix,
                      vector<vLineSegmentMatch> * input_lines,
                      long start, long period);

  // Returns histogram[d][o][l][v][s].
  inline long BinValue(long d, long o, long l, long v, long s)
  {
    if ((d < 0) || (o < 0) || (l < 0) || (v < 0) || (s < 0))
    {
      return -1;
    }
    if ((d >= D_length) || (o >= O_length) || (l >= L_length) || 
        (v >= vlength) || (s >= S_length))
    {
      return -1;
    }
    return histogram[d][o][l][v][s];
  }

  // Returns the entry corresponding to given DOLVS (after we map
  // DOLVS values to indices).
  inline long DOLVS(long d, long o, long l, long v, long s)
  {
    if ((d < 0) || (o < 0) || (l < 0) || (v < 0) || (s < 0))
    {
      exit_error("Error:\n");
      return -1;
    }
    if ((d >= D_values) || (o >= O_values) || (l >= L_values) || 
        (v >= vvalues) || (s >= S_values))
    {
      exit_error("Error:\n");
      return -1;
    }

    long d1, o1, l1, v1, s1;
    d1 = o1 = l1 = v1 = s1 = 0;
    GetIndices(d, o, l, v, s, d1, o1, l1, v1, s1);
    return histogram[d1][o1][l1][v1][s1];
  }

  long CountValue(long value);
  long CountGequal(long value);
  long CountLequal(long value);
  long MaxValue();

  // Converts each entry in the DOLVS histogram by looking
  // up the corresponding probability of that entry at 
  // dolvs_pdf_matrix.
  long ToPDF(vMatrix<long> * dolvs_pdf_matrix);

  // Here we calculate, for each match between the model link_segment
  // and one of the input segments, the number of matches (between
  // the same link_segment and other input segments) that are better
  // than or equal to this match, in all POLVS dimensions. This is an older
  // and slower version, that I'm just keeping around for the time being. 
  // Newer and faster version of ProcessLinkSegment2, that uses input_summary.
  long ProcessLinkSegment(vLineSegment link_segment);

  long ProcessMarkSegment2(vLineSegment link_segment, 
                           vArray(long) input_segments, long number);

  // Here we mark bits, so that we only need to accumulate once every
  // 32 model lines.
  long ProcessMarkSegment3(vLineSegment link_segment, 
                            long counter);

  // Here we use the dolvs entries in each of the input segments
  // (as computed , for example, by ProcessLinkSegment) and we mark
  // them on to the histogram (each bin receives as many votes
  // as the number of segments whose DOLVS entries fall into that bin).
  // Newer and faster version of MarkEntries2, uses input_summary.
  // It assumes that for input line i, entries 9*i + 6,7,8 in the 
  // input_summary are dd, do, dl (differences in DOL between 
  // line i and some fixed model line).
  long MarkEntries();


  // What we compute here is the following: For each DOLVS bin, how common
  // is it for a random model line to have among the input lines a match
  // that is as good or better than that DOLVS bin. Model lines
  // are sampled from model_lines_matrix uniformly, as specified by
  // the arguments "start" and "period".
  // This is a newer and faster version of ExistentialFrequencies2. Instead
  // of creating, filling in and converting to cumulative a DOLVS histogram
  // for each model line, we instead create a single histogram to which
  // we mark entries for 32 model lines before we accumulate. This cuts down
  // the time spent for accumulation (that has to go through all DOLVS bins)
  // by a factor of 32. The trick is that when we mark entries for each
  // of 32 model lines, we simply set a bit in the corresponding bin, that
  // depends on the number of the line (0-31). Then, we can accumulate in
  // two steps. In the first step we go from lower to higher bins taking
  // logical ORs between the previous and next bin, so that each bit
  // gets propagated separately. Then, for each bin, we replace its entry
  // with the number of ones in the binary number representing that entry.
  long ExistentialFrequencies(v3dMatrix<uchar> * model_lines_matrix,
                              vector<vLineSegmentMatch> * input_lines,
                              long start, long period);

  inline long InputLength()
  {
    return input_length;
  }

  // segment1 and segment2 are two model line segments. We want to
  // find the pair of input line segments that best match the pair
  // of model lines. indices1[i] and indices2[i] are indices of the i-th 
  // best pair of input lines. 
  long LinePairMatches(vLineSegment segment1, vLineSegment segment2,
                       vector<long> * indices1, vector<long> * indices2,
                       vector<float> * scores);                      
};


class vLineGrid;


// This class is used to store the score of the best match of
// each line stored in a vector<vStraightLinks> object among
// some input segments. We use it because, at this time, it
// takes several minutes to compute the necessary DOLVS histogram
// and to find the best match after that.
class vLineMatchFrameStats : private vUnique
{
public:
  long frame;
  long view;
  long number_of_lines;
  vArray(long) scores;
  double average;

  vLineMatchFrameStats() 
  {
    frame = -1; 
    view = -1; 
    number_of_lines = 0;
    scores = vZero(long);
  }

  vLineMatchFrameStats(vDOLVS_Histogram * histogram,
                        vector<vStraightLinks> * model_links);
  
  vLineMatchFrameStats(vDOLVS_Histogram * histogram,
                        vector<vStraightLinks> * model_links,
                        long frame, long view);

  vLineMatchFrameStats(vDOLVS_Histogram * histogram,
                        vMatrix<long> * pdf_matrix,
                        vector<vStraightLinks> * model_links,
                        vector<vLineSegmentMatch> * input_segments);

  vLineMatchFrameStats(vDOLVS_Histogram * histogram,
                        vMatrix<long> * pdf_matrix,
                        vector<vStraightLinks> * model_links,
                        vector<vLineSegmentMatch> * input_segments,
                        long frame, long view);

  vLineMatchFrameStats(vDOLVS_Histogram * histogram,
                        vLineGrid * grid,
                        vector<vStraightLinks> * model_links,
                        double fraction);
  
  vLineMatchFrameStats(vDOLVS_Histogram * histogram,
                        vLineGrid * grid,
                        vector<vStraightLinks> * model_links,
                        double fraction, long frame, long view);
  
  long Initialize(vDOLVS_Histogram * histogram,
                  vector<vStraightLinks> * model_links,
                  long in_frame, long in_view);

  long Initialize2(vDOLVS_Histogram * histogram,
                   vMatrix<long> * pdf_matrix,
                   vector<vStraightLinks> * model_links,
                   vector<vLineSegmentMatch> * input_segments,
                   long in_frame, long in_view);

  long Initialize3(vDOLVS_Histogram * histogram,
                   vLineGrid * grid,
                   vector<vStraightLinks> * model_links,
                   double fraction, long frame, long view);

  ~vLineMatchFrameStats()
  {
    DecrementReferences();
  }

  inline double Average()
  {
    return average;
  }

  long Print();

  virtual void DeleteUnique()
  {
    vdelete2(scores);
  }
};


// This class is used to store the vLineMatchFrameStats of many
// database images, so that we can conveniently write them to disk
// and read them from disk.
class vLineMatchStats
{
public:
  vLineMatchStats() {}
  ~vLineMatchStats() {}
  vector<vLineMatchFrameStats> frame_stats;

  char * Filename(const char * path_name, long frame,
                  long start, long period);

  long Load(const char * path_name, long frame,
            long start, long period);
  long Load(const char * filename);

  long Save(const char * path_name, long frame,
            long start, long period);
  long Save(const char * filename);

  // Tries to see if there is a vLineMatchFrameStats object
  // in frame_stats that is equal to frame_stats2 in frame, view,
  // number_of_lines and scores. If there is no object equal
  // in frame and view it returns -1. If there is an object that
  // is equal in frame and view but differs on the rest it returns 
  // 0. Otherwise it returns 1.
  long Compare(vLineMatchFrameStats * frame_stats2);

  vMatchScores * MatchScores();
};


long vFingerLines(grayscale_image * the_image, vector<long> * point_rows, 
                   vector<long> * point_cols, vector<vLineSegment> * result);


long vLinkBoundaries(color_image * the_image, grayscale_image * edge_image,
                      vArray(vector<long>) * point_rows_p, 
                      vArray(vector<long>) * point_cols_p);

color_image * vLinkBoundariesImage(color_image * the_image, vArray(vector<long>) point_rows, 
                                vArray(vector<long>) point_cols);


v3dMatrix<double> * vContourLineVotes(grayscale_image * the_image, 
                                         vector<long> * point_rows,
                                         vector<long> * point_cols);


// A database hand image typically has every link colored in a different way,
// and the link id for a pixel is the remainder of dividing the red value
// of that pixel by 15. Here we put that id onto the blue band.
long vLinkIDs(color_image * the_image);

grayscale_image * vLinkIDsB(color_image * the_image);


vector<vFingerContour> vFingerContours(color_image * the_image,
                                         vArray(vector<long>) point_rows, 
                                         vArray(vector<long>) point_cols);

vLineSegment vBestContourLine(grayscale_image * the_image);

// Input: A gray scale image, where every pixel is labeled with the id
// of the hand link that it belongs to, according to the conventions 
// followed in the hand database. Also, a vStraightLinks pointer,
// that describes the link or group of links that we are interested in.
// Output: two line segments (line1 and line2 in the vStraightLinks object) 
// that are close to parallel to the major axis of the straight link, 
// and that more or less go along the contour of the link.
long vEstimateFingerLines(grayscale_image * link_ids, vStraightLinks * finger);

// Input: color_image, a color image belonging to the hand database. The id
// of the link that a pixel belongs to is the remainder of dividing its
// red band by 70.
// Goal: To find the links (or finger/subfinger groups of links) that 
// look straight and are likely to give nice, easily detectable straight lines
// in input images.
// Output: A vector of such links (or link groupings), described by
// vStraightLinks objects.
long vGoodLinksForLines(color_image * the_image, vector<vStraightLinks> * links);

long vLinkAreas(grayscale_image * link_ids, long * areas);

long vLinkEllipseParams(grayscale_image * link_ids, vEllipse * ellipse_params);

long vFindOuterLines(vContour * contour, double row, double col, double angle,
                        vector<double> * rows1, vector<double> * cols1,
                        vector<double> * rows2, vector<double> * cols2);


vLineSegment vFitLineSegment(vector<double> * rows1, vector<double> * cols1);

// Here we also pass in as input an edge image. A pixel in color_image should
// not be considered an edge pixel unless it is within distance 1 of
// an edge pixel in edges. Also, vGoodLinksForLines doesn't try
// to group links into larger groups. Here we try to do that.
long vGoodLinksForLines2(color_image * the_image, grayscale_image * edges, 
                          vector<vStraightLinks> * links);

// Like vEstimateFingerLines, but here we also take an edge image
// as input (see comments for vGoodLinksForLines2).
long vEstimateFingerLines2(grayscale_image * link_ids, grayscale_image * edges,
                            vStraightLinks * finger);

long vMergeLinks(grayscale_image * link_ids, grayscale_image * edges, 
                  vector<vStraightLinks> * links, 
                  long * areas, vEllipse * ellipses);

long vGroupUnusedLinks(grayscale_image * link_ids, grayscale_image * edges, 
                        vector<vStraightLinks> * links);

// Determine whether two consecutive links that have already 
// been identified as straight links should be merged into one group
long vMergeLinkTest2(grayscale_image * link_ids, grayscale_image * edges, 
                      vector<vStraightLinks> * links, long index1, long index2,
                      long * areas, vEllipse * ellipses);


// Determine if the link is good enough to be a part of a
// straight links group (doesn't need special elongation, 
// but needs min width requirements).
long vLinkTest1(long index1, long * areas, vEllipse * ellipses, 
                 double max_area, double max_width);

// Create the vStraightLinks object that best describes the group
// of links with ids from low to high.
vStraightLinks vGetStraightLinks(grayscale_image * link_ids, grayscale_image * edges,
                                   long low, long high);

// Determine if two straight link groups, one consisting of a single link,
// and the second consisting of that same link plus one or two more, are
// such that the second is a natural extension of the first. This is what
// we use to decide if a single link group should be replaced by an extended
// group of two or three links that includes the original link.
long vParallelTest1(vStraightLinks link1, vStraightLinks link2);

// Determine if two disjoint straight link groups should be merged
// into one.    
long vParallelTest2(vStraightLinks link1, vStraightLinks link2);
    
// Return true if the two lines are close to parallel, one endpoint
// of line1 is close to one endpoint of line2 and the other endpoint
// of line1 is relatively far from the other endpoint of line2.
// Basically, it should look like they share start points, but line2 
// is longer.
long vLineSegmentTest1(vLineSegment line1, vLineSegment line2);

// Return true if the two lines are close to parallel, their start points
// are close to each other and the lines go to opposite directions
// from the start points.
long vLineSegmentTest2(vLineSegment line1, vLineSegment line2);


long vEndpointTest1(double row1a, double col1a, double row2a, double col2a,
                     double row1b, double col1b, double row2b, double col2b);

long vEndpointTest2(double row1a, double col1a, double row2a, double col2a,
                     double row1b, double col1b, double row2b, double col2b);


// Reorder the line segments in segments so that the ones that
// are better (or worse) than d and o and l and v and s are placed first.
// direction specifies if they should be better (positive values)
// or worse (negative values or 0). We return the number of matches that
// are better (or worse). If one of dolvs is negative, we don't use it.
// A value of 0 or negative chooses bigger DOL and smaller VS.
// A value of 1 chooses smaller DOL and larger VS. A value of 2 chooses
// smaller DO and larger LVS.
long vCountLineSegments(vector<vLineSegmentMatch> * segments, long score,
                         long d, long o, long l, long v, long s, long direction);


// Erase all line segments except for the ones that
// are better (or worse) than the givend and o and l and v and s.
// direction specifies if they should be better (positive values)
// or worse (negative values or 0). We return the number of matches that
// are better (or worse). If one of dolvs is negative, we don't use it.
// A value of 0 or negative chooses bigger DOL and smaller VS.
// A value of 1 chooses smaller DOL and larger VS. A value of 2 chooses
// smaller DO and larger LVS.
long vSelectLineSegments(vector<vLineSegmentMatch> * segments, long score,
                          long d, long o, long l, long v, long s, long direction);




// If the database has 107328 images, the result is of size 107328x30x4.
// band corresponds to database image. row corresponds to a line for a 
// link corresponding to that row.
// cols are the coordinates of the two endpoints of that line 
// (in the order row1, col1, row2, col2). We use some conventions.
// A line that covers more than one link is saved at the position
// corresponding to the link with the lowest id. The entry at another link
// will be (1, i, 0, 0) where 1 signifies that the link has a line, that is shared.
// i is the row (0 to 29) where the line is stored.
// For some links no lines are found. For those links we have entries
// (0,0,0,0) in the corresponding rows.
v3dMatrix<uchar> * vAllModelLines(vVideoFile * database, 
                                     vVideoFile * edges);


// model_lines_matrix is a matrix of the kind that we create and return
// at vAllModelLines. This function writes information for a particular
// database image, to band image_id, using the information about lines
// that is held in link_lines.
long vWriteModelLines(v3dMatrix<uchar> * model_lines_matrix, 
                       vector<vStraightLinks> * link_lines, 
                       long image_id);


// vReadModelLines is the inverse of vWriteModelLines: it creates
// vStraightLinks objects based on the entries in the 
// model_line_matrix.
vector<vStraightLinks> * vReadModelLines(v3dMatrix<uchar> * model_lines_matrix,
                                           long image_id);

// vReadModelLines is the inverse of vWriteModelLines: it creates
// vStraightLinks objects based on the entries in the 
// model_line_matrix.
vector<vStraightLinks> * vReadModelLines3(v3dMatrix<uchar> * model_lines_matrix,
                                            long model_frame, long model_view);


// Return the line that is saved at the corresponding index in
// the model_lines_matrix. index -> (band, row), where
// band = index/30 and row = index % 30. If there is 
// no line saved at that position, we return a segment
// with -1 for row1.
vLineSegment vReadLineSegment(v3dMatrix<uchar> * model_lines_matrix,
                                long index);


// For every index from {start, start + period, start+2*period, ..., bands * rows-1}
// we read the line segment that is described there (if any) and we save it 
// at segments. 
long vSampleLineSegments(v3dMatrix<uchar> * model_lines_matrix,
                          long start, long period,
                          vector<vLineSegment> * segments);

      
// We compute all model lines calling vAllModelLines, and we save
// them in a file that shares name with database, but has the 
// extension .model_lines.
v3dMatrix<uchar> * vSaveAllModelLines(vVideoFile * database,
                                         vVideoFile * edges);

v3dMatrix<uchar> * vLoadAllModelLines(vVideoFile * database);


// Compute line match stats (i.e. the information about how well
// each model line matched the input) for the 
// hand pose images specified by database_frames and database_views.
// This is like vComputeLineMatchStats3, but is slower and more exact, 
// because it does not use a grid.
vLineMatchStats * vComputeLineMatchStats(vDOLVS_Histogram * histogram,
                                           v3dMatrix<uchar> * all_model_lines,
                                           vector<long> * database_frames,
                                           vector<long> * database_views);

// Computes line match stats for all hand pose images in the database.
// It basically sets up the right arguments and calls 
// vComputeLineMatchStats3.
vLineMatchStats * vComputeLineMatchStats2(vDOLVS_Histogram * histogram,
                                            vLineGrid * grid,
                                            v3dMatrix<uchar> * all_model_lines,
                                            double fraction);


// This is like vComputeLineMatchStats, but it is faster and a bit 
// inaccurate, because it uses a grid.
vLineMatchStats * vComputeLineMatchStats3(vDOLVS_Histogram * histogram,
                                            vLineGrid * grid,
                                            v3dMatrix<uchar> * all_model_lines,
                                            vector<long> * database_frames,
                                            vector<long> * database_views,
                                            double fraction);


long vSaveLineMatchFrameStats(vDOLVS_Histogram * histogram,
                               v3dMatrix<uchar> * all_model_lines,
                               long start, long period, long frame,
                               vector<long> * database_frames,
                               vector<long> * database_views,
                               vector<vLineSegmentMatch> * input_segments,
                               const char * input_filename);

// Returns a filename to which we should save the existential_DOLVS
// histogram of the given frame from video file "path_name", if
// the existential_DOLVS histogram was computed with a sampling
// of model lines described by start and period.
char * vDOLVSFrameFile(const char * path_name, long frame, 
                        long start, long period);

char * vDOLVS_PDF_File(const char * path_name, long frame, 
                        long start, long period);

long vFrequencyToPDF(vMatrix<long> & dolvs_pdf, 
                      vector<vLineSegmentMatch> * line_segments);

// Creates clusters of the line segments in line_segments, and puts
// the best (meaning highest ranked) segment of each cluster into
// samples. A cluster is defined kind of loosely. We scan line segments,
// in order of their appearance in line_segments, and for every segment S
// that we find that doesn't have a cluster yet, we create a cluster
// centered around it, we put the segment S into samples, and we mark
// the area around the segment S as already belonging to a cluster. How
// large an area we mark is specified using location_diff, orientation_diff
// length_diff. Any segment whose DOL difference to S is less than or equal
// to these three parameters will be represented by the segment S.
long vLineSegmentSamples(vector<vLineSegmentMatch> * line_segments,
                          vector<vLineSegmentMatch> * samples, 
                          long location_diff, long orientation_diff,
                          long length_diff);


// Pick all segments S in line segments that are maximal elements (if 
// we have identical segments pick one of them only), and put them into 
// best segments.
long vSelectBestSegments(vector<vLineSegmentMatch> * line_segments,
                          vector<vLineSegmentMatch> * best_segments);


class vLinePairing
{
public:
  vLinePairing() {}
  ~vLinePairing() {}

  vLineSegment model_line;
  vLineSegmentMatch input_line;
};


class vLineChamferInfo
{
public:
  v3dMatrix<float>  * dt;
  vector<vLinePairing> * pairings;
  vPointSetSequence * point_sets;

  // Set of points in each model edge image that are not covered
  // by the extracted lines.
  vDOLVS_Histogram * dolvs;
  v3dMatrix<uchar> * all_model_lines;
  std::vector<vLineSegmentMatch> * input_lines;

  vArray(long) input_rows;
  vArray(long) input_cols;
  long input_points;

  vLineChamferInfo() 
  {
    dt = 0;
    point_sets = 0;
    dolvs = 0;
    all_model_lines = 0;
    input_lines = 0;
    pairings = 0;

    input_points = 0;
    input_rows = vZero(long);
    input_cols = vZero(long);
  }

  ~vLineChamferInfo() 
  {
    // The variables that are not deleted here should
    // be deleted by whoever created them in the first place.
    vdelete2(dt);
    vdelete(pairings);
    vdelete2(input_rows);
    vdelete2(input_cols);
  }

};


vMatchScores * vLineChamferScores(grayscale_image * the_image, vector<long> * frames,
                                    vector<long> * views, 
                                    vLineChamferInfo * info);

vector<vLinePairing> * 
vFindLinePairings(long model_frame, long model_view,
                   v3dMatrix<uchar> * all_model_lines,
                   vector<vLineSegmentMatch> * input_lines,
                   vDOLVS_Histogram * dolvs);

double vLineChamferDistance(vLineChamferInfo * info, long frame, long view);


// This is just a wrapper around vLeftoverPoints, where instead of passing
// in a vector of line segments, we pass in a vector of line segment pairs.
// Each of those pairs contains a line from the model and its best match
// in the image. Based on the value model_or_image (1 for model, 2 for image)
// we select the appropriate line segments from pairings, and then call
// vLeftoverPoints.
long vLeftoverPoints2(vArray(long) point_rows, vArray(long) point_cols, 
                       long number_of_points, vector<vLinePairing> * pairings, 
                       long model_or_image,
                       vArray(long) * leftover_rowsp, 
                       vArray(long) * leftover_colsp, 
                       long * leftover_numberp);


// We have a set "number_of_points" points, whose coordinates are stored
// in point_rows and point_cols. We want to figure out which of those
// points do NOT belong to the line segments stored in lines, and store
// the coordinates of these leftover points into leftover_rowsp and
// leftover_colsp (which will contain arrays that we create in this
// function). We put the number of those leftover points into 
// leftover_numberp.
long vLeftoverPoints(vArray(long) point_rows, vArray(long) point_cols, 
                      long number_of_points, vector<vLineSegmentLong> * lines, 
                      vArray(long) * leftover_rowsp, 
                      vArray(long) * leftover_colsp, 
                      long * leftover_numberp);


// In vEstimateFingerLines2 we fit lines into some finger edges. Here
// we want to figure out which points are used to fit those lines, because
// we want to build the set of model image points that are NOT used to 
// fit those lines (also called model leftover points elsewhere in the code).
// Those leftover points are important, because they are not described
// by the extracted straight lines and they need to also be matched in order
// to have a more accurate similarity measure.
// The line finding code from vEstimateFingerLines2 is repeated here.
// The coordinates of the points that are used to fit the lines are stored
// in out_rows and out_cols.
long vPointsCoveredByLines(grayscale_image * link_ids, grayscale_image * edges,
                            vStraightLinks * finger,
                            vector<long> * out_rows, vector<long> * out_cols);


// We compute here the coordinates of the model edge points that were NOT 
// used to create the lines contained in the vStraightLink objects.
long vLeftoverModelPoints(grayscale_image * link_ids, grayscale_image * edges,
                           grayscale_image * dilated,
                           vector<vStraightLinks> * finger_lines,
                           vector<long> * out_rows, vector<long> * out_cols);


double vModelToImagePairingScore(vector<vLinePairing> * pairings);

double vImageToModelPairingScore(vector<vLinePairing> * pairings);

// This function is used as follows: suppose I want to know how model
// images that have 8 lines tend to score using line scores. Then,
// I call this function, giving it a dolvs for the current input
// image, specifying line_number = 8, and asking it to find a number
// of model images with 8 lines and collect statistics on them. 
long vLineNumberStatistics(v3dMatrix<uchar> * all_model_lines,
                            vDOLVS_Histogram * dolvs,
                            long number);

vMatchScores * vPalmCutoffScores(v3dMatrix<uchar> * cutoffs,
                                   vLineSegmentLong * cutoff);


class VS_Histogram;

// vLineGrid is used to create a uniform sampling of lines. Each of those
// lines will be matched up with each line segment of the input, and it
// will remember the relative order of those matches. Then, for any model line,
// we can narrow the search for the best match to only the top matches for the
// nearest grid line.
class vLineGrid
{
private:
  long d_samples;
  long o_samples;
  long l_samples;

  // The "values" arrays give the DOL values for the i-th sample.
  vArray(long) d_values;
  vArray(long) o_values;
  vArray(long) l_values;

  // The "indices" arrays give the DOL indices for nearest neighbor 
  // of given DOL values.
  vArray(long) d_indices;
  vArray(long) o_indices;
  vArray(long) l_indices;

public:
  vArray5(long) rankings;
  long number_of_segments;

  vLineGrid(long in_d_samples, long in_o_samples, long in_l_samples);
  ~vLineGrid();

  long PrintValues();
  long ComputeRankings(vDOLVS_Histogram * dolvs);
  long ComputeRankings2(VS_Histogram * vs_histogram,
                        vector<vLineSegmentMatch> * input_lines);
  long SmallestFrequency(vDOLVS_Histogram * dolvs, vLineSegment segment,
                         double fraction);
};


// This class is supposed to be a votes/saliency histogram,
// built based on the lines extracted from an input image.
// Overall, in this code, v stands for votes and s stands for
// saliency.
// One use of this histogram is for evaluating quality of line matches.
// If we assume that saliency, D,O and L are independent, we can just
// multiply the frequencies corresponding to each of them. The results
// seem comparable to using a full DOLVS histogram.
class VS_Histogram : public vMatrix<double>
{
protected:
  // max_v is the maximum value for votes, plus 1. Same for max_s.
  long max_v, max_s;
  long number_of_lines;

  long ProcessInputLines(vector<vLineSegmentMatch> * input_lines);
  long MakeCumulative();
  

public:
  VS_Histogram(vector<vLineSegmentMatch> * input_lines);
  ~VS_Histogram();

  vArray(long) FrequencyRanking(vector<vLineSegmentMatch> * input_lines, 
                                 long row, long col, long orientation, 
                                 long len);
};


// vPointSetOrientation is a class that lets us store the orientation
// of all edge pixels in all images of a database. It should be
// used together with the vPointSetSequence class, which lets us store
// the pixel locations of all edge pixels in the database. The point 
// orientations are stored here in the same order in which their locations
// are stored in the vPointSetSequence class.
class vPointSetOrientation
{
private:
  long rows, cols;
  long number_of_frames;
  long total_length;
  vArray(long) lengths;
  vArray(long) indices;
  vArray(uchar) point_orientations;

  inline void CheckFrame(long frame)
  {
    if ((frame < 0) || (frame >= number_of_frames))
    {
      exit_error("PointSetOrientation error\n");
    }
  }

public:
  vPointSetOrientation(vVideoFile * sequence, vPointSetSequence * locations);
  vPointSetOrientation() {}
  ~vPointSetOrientation();

  // Assuming that filename specifies a database, we return a string 
  // that is the same as the filename, except that the ending has been
  // adjusted to the default ending for files where vPointSetOrientation
  // objects are saved.
  static char * AdjustEnding(const char * filename);
  
  static vPointSetOrientation * Load(const char * filename);
  long Save(const char * filename);

  inline long Rows()
  {
    return rows;
  }

  inline long Cols()
  {
    return cols;
  }

  inline long NumberOfFrames()
  {
    return number_of_frames;
  }

  inline long Length(long frame)
  {
    CheckFrame(frame);
    return lengths[frame];
  }

  inline long TotalLength()
  {
    return total_length;
  }

  inline vArray(uchar) PointOrientations(long frame)
  {
    CheckFrame(frame);
    long index = indices[frame];
    return point_orientations + index;
  }

  // returns the number of points.
  // In orientationsp we put store the orientations of
  // each point in the set of points corresponding
  // to the given frame and given view
  long GetOrientations(long frame, long view, vArray(float) * orientationsp);

  // Orientations range from 0 to pi, but they are saved as unsigned chars,
  // that range from 0 to 255. Therefore, we need to be able to convert 
  // values from one type to the other. 
  inline float UcharToOrientation(float uchar_value)
  {
    return uchar_value / (float) 255.0 * (float) M_PI;
  }

  inline uchar OrientationToUchar(float orientation)
  {
    long int_value = vRound(orientation * 255.0 / M_PI);
    if ((int_value < 0) || (int_value > 255))
    {
      exit_error("Bad OrientationToUchar result: %f -> %li\n", 
                      orientation, int_value);
    }
    return (uchar) int_value;
  }

  // Assuming orientation1 and orientation2 are between 0 and M_PI,
  // we add them and adjust the result to make sure its range is also
  // between 0 and M_PI.
  inline float AddOrientations(float orientation1, float orientation2)
  {
    float result = orientation1 + orientation2;
    if (result >= (float) M_PI) result = result - (float) M_PI;
    return result;
  }
};


// Returns the equivalent of the distance transform when we 
// want to take into account edge orientation in addition to
// edge location. Location (i,j,k) contains the distance to
// the closest location/orientation combination.
v3dMatrix<float> * vDO_DistanceTransform(grayscale_image * edges,
                                            v3dMatrix<double> * orientations);



// Inputs: do_dt is a result of vDo_DistanceTransform. locations and 
// orientations specify the location and orientation of every edge pixel
// in every image of the database. The result is the scores of all 
// database images in the distance/orientation model-to-image version 
// of the chamfer distance.
vMatchScores *
vChamfer_DO_MI3(v3dMatrix<float> * do_dt, vPointSetSequence * locations,
                 vPointSetOrientation * orientations);


// Same as vChamfer_DO_MI3, but the scores are only for a subset of
// database images.
vMatchScores *
vChamfer_DO_MI5(v3dMatrix<float> * do_dt, vPointSetSequence * locations,
                 vPointSetOrientation * orientations, 
                 vector<long> * frames, vector<long> * views);


// Computes scores for the given database images (specified by frames
// and views) using the distance/orientation version of the chamfer distance.
vMatchScores * vChamfer_DO_Scores4(grayscale_image * edges,
                                     v3dMatrix<double> * orientations,
                                     vPointSetSequence * database_points,
                                     vPointSetOrientation * database_orientations);

                                          
// Computes scores for the given database images (specified by frames
// and views) using the distance/orientation version of the chamfer distance.
vMatchScores * vChamfer_DO_Scores6b(grayscale_image * edges,
                                      v3dMatrix<double> * orientations,
                                      vPointSetSequence * database_points,
                                      vPointSetOrientation * database_orientations,
                                      vMatchScores * scores,
                                      long scores_to_keep);

                                          
// Computes scores for the given database images (specified by frames
// and views) using the distance/orientation version of the chamfer distance.
vMatchScores * vChamfer_DO_Scores6(grayscale_image * edges,
                                     v3dMatrix<double> * orientations,
                                     vPointSetSequence * database_points,
                                     vPointSetOrientation * database_orientations,
                                     vector<long> * frames,
                                     vector<long> * views);


// This class will be used to define a template of what
// a finger curve looks like, and to use that template
// in order to locate fingers in an image.
class vFingerCurve
{
private:
  long number_of_scales;
  long number_of_orientations;
  // Total number of scales/orientation combinations
  long number_of_templates;
  long height;
  long width;

  // The prototype, in canonical scale and orientation (upright, 50x50).
  grayscale_image * prototype;

  // Versions of the prototype at different scales and orientations.
  vArray(grayscale_image *) all_templates;

  // scales[i] will be the scale of all_templates[i]
  vArray(double) scales;

  // orientations[i] will be the orientation of all_templates[i], in radians.
  vArray(double) orientations;

  // all_point_rows[i] will contain row coordinates of all non-zero points 
  // of all_templates[i]
  vArray2(long) all_point_rows;
  vArray2(long) all_point_cols;

  // all_lengths[i] is the number of non-zero pixels of all_templates[i]
  vArray(long) all_lengths;

public:
  vFingerCurve(long in_number_of_scales, long in_number_of_orientations);
  ~vFingerCurve();

  grayscale_image * MakePrototype();
  v3dMatrix<double> * Match(grayscale_image * the_image);
  v3dMatrix<double> * Match(v3dMatrix<float> * dt);
};


// result[row][col] is the (rounded) distance between location (row, col) and
// the nearest location in the input image that has a value higher than
// or equal to that of location (row, col). For the pixel in the image
// with max magnitude, the value is equal to the length of the diagonal
// of the image plus 1.
v3dMatrix<long> * vDominanceTransform(general_image * input_image);


// An auxiliary function for vDominanceTransform. For every point in
// points, it checks whether there exists a point in input, within
// the specified distance, that has a greater value. If the answer
// is yes, then we store the value distance in the corresponding location
// in output. This can be done in more efficient ways, but initially I 
// will go for ease of implementation. This function is more efficient
// than S_ListBasedDominance when we have lots of points and a small
// distance.
long S_WindowBasedDominance(vArray2(float) input, vArray2(long) result, 
                            std::list<vPoint> * points, long distance, 
                            vArray2(long) lengths, long rows, long cols);


// An auxiliary function for vDominanceTransform. For every point p1 in
// points, it finds the point p2 in points that is the farthest from p1,
// among all points in points that have a greater value (in input) than
// p1. This function is more efficient than S_WindowBasedDominance when
// the number of points is sufficiently small and the min distance
// between those points is sufficiently large.
long S_ListBasedDominance(vArray2(float) input, vArray2(long) result, 
                            std::list<vPoint> * points, 
                            vArray2(long) lengths, long rows, long cols);


// This function is built for domt arguments that are a dominance transforms.
// For the top number pixels, we draw the circle of their dominance on output.
long vDrawTopCircles(v3dMatrix<long> * domt, color_image * output, long number);


// corners gives the three corners of a triangle.
// We store into points a sampling 
// of the 3D surface of the triangle that guarantees that any point
// on the surface of the triangle is within max_distance distance
// of one of the points stored in points.
long vTrianglePixels3D(vPoint3f * corners, 
                        vector<vPoint3f> * points, float max_distance);


// result(i, j) will be the mean of all entries in band i, column j.
vMatrix<float> vColumnMeans(v3dMatrix<float> * values_matrix);

// result(i, j) will be the median of all entries in band i, column j.
vMatrix<float> vColumnMedians(v3dMatrix<float> * values_matrix);

// result(i, j) will be the k-th largest value (starting at 1) of all entries in 
// band i, column j.
vMatrix<float> vColumnKth(v3dMatrix<float> * values_matrix, long k);



#endif // VASSILIS_FINGERS_H
