#include "stdafx.h"

#include "gesture_data.h"
#include "hand_database.h"
#include "basics/image_algorithms.h"
#include "basics/image_templates.h"
#include "algorithms.h"
#include "algo_templates.h"
#include "quaternion.h"
#include "angle_sampler.h"
#include "fingers.h"
#include "opencv_utils.h"
#include "avi_file.h"

#include "basics/definitions.h"


vMatrix<double> vLoadViewpoints(const char * filename)
{
  vector<double> numbers;
  long success = vReadDoubleVectorB(filename, &numbers);
  if (success <= 0) return vMatrix<double>();
  vMatrix<double> result(&numbers, 2);
  return result;
}


long vSaveCameraParams(const char * filename, 
                        vMatrix<double> camera_params)
{
  long success = camera_params.WriteText(filename);
  if (success <= 0) 
  {
    vPrint("\nError in saving camera params to %s\n", filename);
  }
  return success;
}


vOldDatabase::vOldDatabase()
{
  max_x = 200.0;
  max_y = 400.0;
  map_factor = 1;
}


vOldDatabase::~vOldDatabase()
{
}


char * vOldDatabase::AddDirectory(const char * filename)
{
  char * database_directory = vJoinPaths3(sequences_directory, "glove_data",
                                           "training");
  char * result = vJoinPaths(database_directory, filename);
  vdelete2(database_directory);
  return result;
}


char * vOldDatabase::DatabaseFilename()
{
  return DatabaseFilename1(name);
}


char * vOldDatabase::DatabaseFilename1(const char * in_name)
{
  char * name1 = vReplaceEnding2(in_name, ".database");
  char * name2 = AddDirectory(name1);
  vdelete2(name1);
  return name2;
}


char * vOldDatabase::ImageSequenceName()
{
  return ImageSequenceName1(name);
}


char * vOldDatabase::ImageSequenceName1(const char * in_name)
{
  char * name1 = vReplaceEnding2(in_name, ".rle");
  char * name2 = AddDirectory(name1);
  vdelete2(name1);
  return name2;
}

char * vOldDatabase::EdgeSequenceName()
{
  return EdgeSequenceName1(name);
}


char * vOldDatabase::EdgeSequenceName1(const char * in_name)
{
  char * name1 = vReplaceEnding2(in_name, "_edges.rle");
  char * name2 = AddDirectory(name1);
  vdelete2(name1);
  return name2;
}


// Filename where the pose_to_frame_map is stored.
char * vOldDatabase::PoseFrameName()
{
  return PoseFrameName1(name);
}


char * vOldDatabase::PoseFrameName1(const char * in_name)
{
  char * name1 = vReplaceEnding2(in_name, ".pose_to_frame");
  char * name2 = AddDirectory(name1);
  vdelete2(name1);
  return name2;
}


long vOldDatabase::ComputePoseToFrameMap()
{
  double angle_step = ((double) 1) / map_factor;
  double adjustment = angle_step / 2;
  long rows = vRound(max_x * map_factor);
  long cols = vRound(max_y * map_factor);
  pose_to_frame_map = vMatrix<long>(rows, cols);
  pose_to_frame = pose_to_frame_map.Matrix2();
  orientation_adjustment_map = vMatrix<float>(rows, cols);
  orientation_adjustment = orientation_adjustment_map.Matrix2();
  long i, j;
  vHandPose pose;
  pose.shape_id = 0;
  pose.orientation = 0;
  long frame = 0;
  long view = 0;
  vPrint("\n");
  for (i = 0; i < rows; i++)
  {
    pose.camera_x = ((double) i) / map_factor + adjustment;
    for (j = 0; j < cols; j++)
    {
      pose.camera_y = ((double) j) / map_factor + adjustment;
      GetPoseFrame(pose, &frame, &view);
      pose_to_frame[i][j] = frame;
      orientation_adjustment[i][j] = (float) OrientationAdjustment(pose, frame);
    }
    vPrint("Done with row %li of %li\r", i, rows);
  }
  vPrint("\n");
  long result = SavePoseToFrameMap();
  if (result <= 0)
  {
    vPrint("Failed to save pose_to_frame_map\n");
  }
  return result;
}


long vOldDatabase::LoadPoseToFrameMap()
{
  long rows = vRound(max_x * map_factor);
  long cols = vRound(max_y * map_factor);
  pose_to_frame_map = vMatrix<long>(rows, cols);
  orientation_adjustment_map = vMatrix<float>(rows, cols);
  char * filename = PoseFrameName();
  long result = 0;
  FILE * fp = fopen(filename, vFOPEN_READ);
  if (fp != 0)
  {
    result = pose_to_frame_map.ReadBandwise(filename, 0);
    if (result != 0) 
    {
      result = orientation_adjustment_map.ReadBandwise(fp);
    }
    fclose(fp);
  }
  vdelete2(filename);
  if (result <= 0)
  {
    result = ComputePoseToFrameMap();
  }

  pose_to_frame = pose_to_frame_map.Matrix2();
  orientation_adjustment = orientation_adjustment_map.Matrix2();
  return result;
}


long vOldDatabase::SavePoseToFrameMap()
{
  char * filename = PoseFrameName();
  long result = 0;
  FILE * fp = fopen(filename, vFOPEN_WRITE);
  if (fp != 0)
  {
    result = pose_to_frame_map.WriteBandwise(fp);
    if (result != 0) 
    {
      result = orientation_adjustment_map.WriteBandwise(fp);
    }
    fclose(fp);
  }
  vdelete2(filename);
  return result;
}


double vOldDatabase::OrientationAdjustment(vHandPose & pose1, long frame)
{
  vHandPose pose2 = GetPose(frame);
  double angle_step = ((double) 1) / (map_factor * 4);
  double z;
  double limit = 360 - angle_step/2;
  double min_distance = 100000;
  double min_z = -1;
  for (z = 0; z < limit; z++)
  {
    pose2.orientation = z;
    double distance = vOldPoseDistance(pose1, pose2);
    if (distance < min_distance)
    {
      min_distance = distance;
      min_z = z;
    }
  }
  return min_z;
}



vOldDatabase1::vOldDatabase1() 
{
  Zero();

  number_of_shapes = 26;
  number_of_viewpoints = 86;
  number_of_orientations = 48;
  number_of_images = number_of_shapes * number_of_viewpoints * number_of_orientations;
  number_of_views = 4;
  number_of_frames = number_of_images / number_of_views;
  frames_per_shape = number_of_frames / number_of_shapes;
  name = vCopyString("training7");
  LoadPoseToFrameMap();
}


vOldDatabase1::~vOldDatabase1() 
{
  vdelete2(name);
  vdelete(image_sequence);
  vdelete(edge_sequence);
}


void vOldDatabase1::Zero()
{
  name = 0;
  number_of_images = 0;
  number_of_shapes = 0;
  number_of_viewpoints = 0;
  number_of_views = 0; 
  number_of_orientations = 0;
  number_of_frames = 0;
  frames_per_shape = 0;
  image_sequence = 0;
  edge_sequence = 0;
  is_valid = 0;
}


// Returns image indexed by given shape id and 3D orientation
color_image * vOldDatabase1::PoseImage(vHandPose & pose)
{
  exit_error("Error: call to unimplemented vOldDatabase1 function\n");
  return 0;
}


grayscale_image * vOldDatabase1::RotateImage(grayscale_image * image, long view)
{
  exit_error("Error: call to unimplemented vOldDatabase1 function\n");
  return 0;
}



color_image * vOldDatabase1::RotateImage(color_image * image, long view)
{
  exit_error("Error: call to unimplemented vOldDatabase1 function\n");
  return 0;
}



// Returns image indexed by shape_id, viewpoint_index, orientation_index
// The caller is assumed to know what actual shapes and 3D orientations
// these indices correspond to.
color_image * vOldDatabase1::GetImage3(long shape_id, long viewpoint_index, 
                                     long orientation_index)
{                                     
  exit_error("Error: call to unimplemented vOldDatabase1 function\n");
  return 0;
}



color_image * vOldDatabase1::GetImage2(long frame, long view)
{
  exit_error("Error: call to unimplemented vOldDatabase1 function\n");
  return 0;
}



// Returns edge image indexed by given shape id and 3D orientation
grayscale_image * vOldDatabase1::PoseEdges(vHandPose & pose)
{
  exit_error("Error: call to unimplemented vOldDatabase1 function\n");
  return 0;
}



// Returns edge image indexed by shape_id, viewpoint_index, orientation_index
// The caller is assumed to know what actual shapes and 3D orientations
// these indices correspond to.
grayscale_image * vOldDatabase1::GetEdges3(long shape_id, long viewpoint_index, 
                                         long orientation_index)
{
  exit_error("Error: call to unimplemented vOldDatabase1 function\n");
  return 0;
}



grayscale_image * vOldDatabase1::GetEdges2(long frame, long view)
{
  exit_error("Error: call to unimplemented vOldDatabase1 function\n");
  return 0;
}



vHandPose vOldDatabase1::GetPose2(long frame, long view)
{
  vHandPose pose;
  finder.GetViewParams(frame, view, &(pose.shape_id), &(pose.camera_x),
                       &(pose.camera_y), &(pose.orientation));
  return pose;
}


vHandPose vOldDatabase1::GetPose(long frame)
{
  vHandPose pose;
  finder.GetViewParams(frame, 0, &(pose.shape_id), &(pose.camera_x),
                       &(pose.camera_y), &(pose.orientation));
  return pose;
}


long vOldDatabase1::GetPoseFrame(vHandPose pose, 
                                   long * framep, long * viewp)
{
  pose.Normalize();
  long shape_base = finder.FirstPoseView(pose.shape_id);
  long frames_per_viewpoint = number_of_orientations / number_of_views;
  long frames_per_shape = number_of_viewpoints * frames_per_viewpoint;
  long shape_end = shape_base + frames_per_shape - 1;

  // First, find the closest frame/view combination.
  long shape_id;
  double current_x, current_y, current_z;
  current_x = current_y = current_z = shape_id = 0;
  double min_distance = 100000;
  long min_frame = -1;
  long min_view = -1;
  long frame;
  for (frame = shape_base; frame <= shape_end; frame++)
  {
    long view;
    for (view = 0; view < number_of_views; view++)
    {
      finder.GetViewParams(frame, view, &shape_id, &current_x, &current_y, &current_z);
      double distance = vOldViewDistance2(pose.camera_x, pose.camera_y, pose.orientation,
                                           current_x, current_y, current_z);
      if (distance < min_distance)
      {
        min_distance = distance;
        min_frame = frame;
        min_view = view;
      }
    }
  }
  *framep = min_frame;
  *viewp = min_view;
  return 1;
}


long vOldDatabase1::FastPoseFrame(vHandPose & pose, 
                                    long * framep, long * viewp)
{

  const double orientation_sampling = 360.0 / ((double) number_of_orientations);
  const double half_sampling = orientation_sampling / 2;
  long x_index = vRound(pose.camera_x * map_factor);
  long y_index = vRound(pose.camera_y * map_factor);
  long frame = pose_to_frame[x_index][y_index];
  frame = frame - frame % 12;
  double orientation = pose.orientation + orientation_adjustment[x_index][y_index];
  orientation = vNormalizeDegrees(orientation - half_sampling);
  long rotation_index = vRound(orientation / orientation_sampling);
  if (rotation_index == 48) rotation_index = 0;

  long frame2 = finder.FrameIndex(rotation_index);
  long view = finder.RotationIndex(rotation_index);
  long shape_base = pose.shape_id * frames_per_shape;
  long final_frame = shape_base + frame + frame2;
  *framep = final_frame;
  *viewp = view;
  return 1;
}









vOldDatabase2::vOldDatabase2()
{
  Zero();
}


void vOldDatabase2::Zero()
{
  name = 0;
  number_of_images = 0;
  number_of_shapes = 0;
  number_of_viewpoints = 0;
  number_of_views = 0; 
  number_of_orientations = 0;
  number_of_frames = 0;
  frames_per_shape = 0;
  image_sequence = 0;
  edge_sequence = 0;
  rotation_templates = vZero(v3dMatrix<short> *);
  x_angles = vZero(double);
  y_angles = vZero(double);
  orientations = vZero(double);
  view_density = 0;
  is_valid = 0;
}


vOldDatabase2::vOldDatabase2(const char * in_name)
{
  Zero();
  name = vCopyString(in_name);
  char * database_filename = DatabaseFilename();
  FILE * fp = fopen(database_filename, vFOPEN_READ);
  vdelete2(database_filename);
  if (fp == 0) return;

  fscanf(fp, "%lf", &view_density);
  if (view_density <= 0) return;
  fclose(fp);

  // Load sequences
  char * sequence_name = ImageSequenceName();
  image_sequence = new vRLE_Sequence(sequence_name);
  vdelete2(sequence_name);
  if (image_sequence == 0) return;

  sequence_name = EdgeSequenceName();
  edge_sequence = new vRLE_Sequence(sequence_name);
  vdelete2(sequence_name);
  if (edge_sequence == 0) return;

  number_of_frames = image_sequence->NumberOfFrames();
  long rows = image_sequence->Rows();
  long cols = image_sequence->Cols();
  if (number_of_frames != edge_sequence->NumberOfFrames()) return;
  if (rows != edge_sequence->Rows()) return;
  if (cols != edge_sequence->Cols()) return;

  // Get viewpoint information
  vector<vPoint2> sample_viewpoints;
  vOldSampleViewpoints(view_density, &sample_viewpoints);
  number_of_viewpoints = sample_viewpoints.size();
  if (number_of_viewpoints <= 0) return;
  number_of_shapes = number_of_frames / number_of_viewpoints;
  if (number_of_shapes * number_of_viewpoints != number_of_frames) return;
  frames_per_shape = number_of_frames / number_of_shapes;

  x_angles = vnew(double, number_of_viewpoints);
  y_angles = vnew(double, number_of_viewpoints);
  long i;
  for (i = 0; i < number_of_viewpoints; i++)
  {
    x_angles[i] = sample_viewpoints[i].col;
    y_angles[i] = sample_viewpoints[i].row;
  }
  
  vector<double> orientations_vector;
  number_of_views = vOldSampleOrientations(view_density, &orientations_vector);
  number_of_orientations = number_of_views;
  number_of_images = number_of_shapes * frames_per_shape * number_of_orientations;
  if (number_of_views <= 0) return;
  orientations = vnew(double, number_of_views);
  for (i = 0; i < number_of_views; i++)
  {
    orientations[i] = orientations_vector[i];
  }

  // Initialize rotation templates
  rotation_templates = vnew(v3dMatrix<short> *, number_of_views);
  long center_row = rows/2;
  long center_col = cols/2;
  for (i = 0; i < number_of_views; i++)
  {
    rotation_templates[i] = vRotationTemplate(rows, cols, center_row, center_col,
                                               orientations[i]);
  }

  LoadPoseToFrameMap();
  is_valid = 1;
}


vOldDatabase2::~vOldDatabase2()
{
  vdelete2(name);
  vdelete(image_sequence);
  vdelete(edge_sequence);
  long i;
  for (i = 0; i < number_of_views; i++)
  {
    vdelete(rotation_templates[i]);
  }
  vdelete2(rotation_templates);
  vdelete2(x_angles);
  vdelete2(y_angles);
  vdelete2(orientations);
}


grayscale_image * vOldDatabase2::RotateImage(grayscale_image * image, long view)
{
  v3dMatrix<short> * sources = rotation_templates[view];
  return (grayscale_image *) wrma(image).Warp_a(sources);
}


color_image * vOldDatabase2::RotateImage(color_image * image, long view)
{
  v3dMatrix<short> * sources = rotation_templates[view];
  return (color_image *) wrma(image).Warp_a(sources);
}


// Returns image indexed by given shape id and 3D orientation
color_image * vOldDatabase2::PoseImage(vHandPose & pose)
{
  long frame = 0;
  long view = 0;
  GetPoseFrame(pose, &frame, &view);
  return GetImage2(frame, view);
}



// Returns image indexed by shape_id, viewpoint_index, orientation_index
// The caller is assumed to know what actual shapes and 3D orientations
// these indices correspond to.
color_image * vOldDatabase2::GetImage3(long shape_id, long viewpoint_index, 
                                         long orientation_index)
{
  long frame = number_of_shapes * shape_id + viewpoint_index;
  return GetImage2(frame, orientation_index);
}


color_image * vOldDatabase2::GetImage2(long frame, long view)
{
  if (image_sequence == 0) return 0;
  if (frame < 0) return 0;
  if (frame >= image_sequence->NumberOfFrames()) return 0;
  color_image * frame_image = image_sequence->ReadFrame(frame);
  if (view == 0) 
  {
    return frame_image;
  }
  else
  {
    color_image * rotated = RotateImage(frame_image, view);
    vdelete(frame_image);
    return rotated;
  }
}


// Returns image indexed by given shape id and 3D orientation
grayscale_image * vOldDatabase2::PoseEdges(vHandPose & pose)
{
  long frame = 0;
  long view = 0;
  GetPoseFrame(pose, &frame, &view);
  return GetEdges2(frame, view);
}



// Returns image indexed by shape_id, viewpoint_index, orientation_index
// The caller is assumed to know what actual shapes and 3D orientations
// these indices correspond to.
grayscale_image * vOldDatabase2::GetEdges3(long shape_id, long viewpoint_index, 
                                         long orientation_index)
{
  long frame = number_of_shapes * shape_id + viewpoint_index;
  return GetEdges2(frame, orientation_index);
}


grayscale_image * vOldDatabase2::GetEdges2(long frame, long view)
{
  if (edge_sequence == 0) return 0;
  if (frame < 0) return 0;
  if (frame >= edge_sequence->NumberOfFrames()) return 0;
  grayscale_image * frame_image = edge_sequence->ReadFrameGray(frame);
  if (view == 0) 
  {
    return frame_image;
  }
  else
  {
    grayscale_image * rotated = RotateImage(frame_image, view);
    vdelete(frame_image);
    return rotated;
  }
}


vHandPose vOldDatabase2::GetPose2(long frame, long view)
{
  vHandPose pose;
  pose.shape_id = frame / number_of_viewpoints;
  long viewpoint_index = frame % number_of_viewpoints;
  pose.camera_x = x_angles[viewpoint_index];
  pose.camera_y = y_angles[viewpoint_index];
  pose.orientation = orientations[view];
  return pose;
}


vHandPose vOldDatabase2::GetPose(long frame)
{
  vHandPose pose;
  pose.shape_id = frame / number_of_viewpoints;
  long viewpoint_index = frame % number_of_viewpoints;
  pose.camera_x = x_angles[viewpoint_index];
  pose.camera_y = y_angles[viewpoint_index];
  pose.orientation = 0;
  return pose;
}


long vOldDatabase2::GetPoseFrame(vHandPose pose, long * framep, long * viewp)
{
  if (number_of_viewpoints <= 0) return -1;
  if (number_of_views <= 0) return -1;
  pose.Normalize();
  double pose_x = pose.camera_x;
  double pose_y = pose.camera_y;
  double pose_z = pose.orientation;

  long min_index = -1;
  double min_distance = 1000000;
  long i;
  long start = pose.shape_id * number_of_viewpoints;
  long end = start + number_of_viewpoints - 1;
  if (end >= edge_sequence->NumberOfFrames()) return -1;

  // First find the best viewpoint;
  for (i = start; i <= end; i++)
  {
    long viewpoint_index = i % number_of_viewpoints;
    double current_x = x_angles[viewpoint_index];
    double current_y = y_angles[viewpoint_index];
    double distance = vOldViewDistance2(pose_x, pose_y, 0,
                                         current_x, current_y, 0);
    if (distance < min_distance)
    {
      min_index = i;
      min_distance = distance;
    }
  }

  // Now find the best view for the best viewpoint;
  double best_x = x_angles[min_index % number_of_viewpoints];
  double best_y = y_angles[min_index % number_of_viewpoints];
  *framep = min_index;

  min_index = -1;
  min_distance = 1000000;
  for (i = 0; i < number_of_views; i++)
  {
    double current_z = orientations[i];
    double distance = vOldViewDistance2(pose_x, pose_y, pose_z,
                                      best_x, best_y, current_z);
    if (distance < min_distance)
    {
      min_index = i;
      min_distance = distance;
    }
  }

  *viewp = min_index;
  return 1;
}




vOldDatabase2 * vOldDatabase2::Load(const char * in_name)
{
  vOldDatabase2 * database = new vOldDatabase2(in_name);
  if (database->IsValid() <= 0)
  {
    vdelete(database);
    return 0;
  }
  else
  {
    return database;
  }
}


vOldDatabase2 * vOldDatabase2::Load(vRLE_Sequence * in_sequence)
{
  const char * name = in_sequence->GetFilename();
  char * simple_name = vSimpleName(name, platform_directory_slash);
  vOldDatabase2 * result = Load(simple_name);
  vdelete2(simple_name);
  return result;
}


long vOldDatabase2::Save()
{
  char * database_filename = DatabaseFilename();
  // If the file already exists, we don't overwrite it, as a safety
  // measure. The user should doubleheck and delete the existing file,
  // befure calling this function again.
  FILE * fp = fopen(database_filename, vFOPEN_READ);
  if (fp != 0)
  {
    vPrint("%s exists, can't overwrite\n", database_filename);
    vdelete2(database_filename);
    return 0;
  }

  fp = fopen(database_filename, vFOPEN_WRITE);
  vdelete2(database_filename);
  if (fp == 0) return 0;
  fprintf(fp, "%lf\r\n", view_density);
  fclose(fp);
  return 1;
}


long vOldDatabase2::FastPoseFrame(vHandPose & pose, 
                                    long * framep, long * viewp)
{
  const double orientation_sampling = 360.0 / ((double) number_of_orientations);
  long x_index = vRound(pose.camera_x * map_factor);
  long y_index = vRound(pose.camera_y * map_factor);
  long frame = pose_to_frame[x_index][y_index];
  double orientation = pose.orientation + orientation_adjustment[x_index][y_index];
  orientation = vNormalizeDegrees(orientation);
  long rotation_index = vRound(orientation / orientation_sampling);

  long shape_base = pose.shape_id * frames_per_shape;
  long final_frame = shape_base + frame;
  *framep = final_frame;
  *viewp = rotation_index;
  return 1;
}


/////////////////////////////////////////////////////////////////////
//                                                                 //
//         Beginning of implementation of vPointSequence          //
//                                                                 //
/////////////////////////////////////////////////////////////////////


vPointSequence::vPointSequence(vVideoFile * sequence)
{
  rows = sequence->Rows();
  cols = sequence->Cols();
  if ((rows > 256) || (cols > 256))
  {
    vPrint("Warning: coordinates exceeding 256 in vPointSequence\n");
  }

  number_of_frames = sequence->NumberOfFrames();

  lengths = vnew(long, number_of_frames);
  indices = vnew(long, number_of_frames);

  // First store the points into vectors, since we don't know how
  // many points we will find.
  vector<uchar> temp_rows, temp_cols;

  long i;
  vPrint("\n");
  for (i = 0; i < number_of_frames; i++)
  {
    grayscale_image * gray = sequence->ReadFrameGray(i);
    if (gray == 0)
    {
      exit_error("error in reading frame %li\n", i);
    }

    vector<long> current_rows, current_cols;
    vGetNonZeroPoints(gray, &current_rows, &current_cols);
    long length = current_rows.size();
    lengths[i] = length;
    if (i == 0)
    {
      indices[i] = 0;
    }
    else 
    {
      indices[i] = indices[i-1] + lengths[i-1];
    }
    long j;
    for (j = 0; j < length; j++)
    {
      temp_rows.push_back((uchar) current_rows[j]);
      temp_cols.push_back((uchar) current_cols[j]);
    }
    vdelete(gray);
    vPrint("processed frame %li of %li\r", i, number_of_frames);
  }
  vPrint("\n");

  // Now that we have all the point coordinates, store them into the arrays.
  total_length = temp_rows.size();
  point_rows = vnew(uchar, total_length);
  point_cols = vnew(uchar, total_length);
  for (i = 0; i < total_length; i++)
  {
    point_rows[i] = temp_rows[i];
    point_cols[i] = temp_cols[i];
  }
}


#ifdef VASSILIS_USE_vArray
vPointSequence::vPointSequence(grayscale_image ** images, long number)
{
  if (number <= 0) 
  {
    exit_error("vPointSequence, error 1\n");
  }
  
  rows = images[0]->Rows();
  cols = images[0]->Cols();
  if ((rows > 256) || (cols > 256))
  {
    vPrint("Warning: coordinates exceeding 256 in vPointSequence\n");
  }

  number_of_frames = number;

  lengths = vnew(long, number_of_frames);
  indices = vnew(long, number_of_frames);

  // First store the points into vectors, since we don't know how
  // many points we will find.
  vector<uchar> temp_rows, temp_cols;

  long i;
//  vPrint("\n");
  for (i = 0; i < number_of_frames; i++)
  {
    grayscale_image * gray = images[i];
    if (gray == 0)
    {
      exit_error("error in reading frame %li\n", i);
    }

    vector<long> current_rows, current_cols;
    vGetNonZeroPoints(gray, &current_rows, &current_cols);
    long length = current_rows.size();
    lengths[i] = length;
    if (i == 0)
    {
      indices[i] = 0;
    }
    else 
    {
      indices[i] = indices[i-1] + lengths[i-1];
    }
    long j;
    for (j = 0; j < length; j++)
    {
      temp_rows.push_back((uchar) current_rows[j]);
      temp_cols.push_back((uchar) current_cols[j]);
    }
    //vPrint("processed frame %li of %li\r", i, number_of_frames);
  }
//  vPrint("\n");

  // Now that we have all the point coordinates, store them into the arrays.
  total_length = temp_rows.size();
  point_rows = vnew(uchar, total_length);
  point_cols = vnew(uchar, total_length);
  for (i = 0; i < total_length; i++)
  {
    point_rows[i] = temp_rows[i];
    point_cols[i] = temp_cols[i];
  }
}
#endif VASSILIS_USE_vArray

  
vPointSequence::vPointSequence(class_pointer(grayscale_image *) images, long number)
{
  if (number <= 0) 
  {
    exit_error("vPointSequence, error 1\n");
  }
  
  rows = images[0]->Rows();
  cols = images[0]->Cols();
  if ((rows > 256) || (cols > 256))
  {
    vPrint("Warning: coordinates exceeding 256 in vPointSequence\n");
  }

  number_of_frames = number;

  lengths = vnew(long, number_of_frames);
  indices = vnew(long, number_of_frames);

  // First store the points into vectors, since we don't know how
  // many points we will find.
  vector<uchar> temp_rows, temp_cols;

  long i;
//  vPrint("\n");
  for (i = 0; i < number_of_frames; i++)
  {
    grayscale_image * gray = images[i];
    if (gray == 0)
    {
      exit_error("error in reading frame %li\n", i);
    }

    vector<long> current_rows, current_cols;
    vGetNonZeroPoints(gray, &current_rows, &current_cols);
    long length = current_rows.size();
    lengths[i] = length;
    if (i == 0)
    {
      indices[i] = 0;
    }
    else 
    {
      indices[i] = indices[i-1] + lengths[i-1];
    }
    long j;
    for (j = 0; j < length; j++)
    {
      temp_rows.push_back((uchar) current_rows[j]);
      temp_cols.push_back((uchar) current_cols[j]);
    }
    //vPrint("processed frame %li of %li\r", i, number_of_frames);
  }
//  vPrint("\n");

  // Now that we have all the point coordinates, store them into the arrays.
  total_length = temp_rows.size();
  point_rows = vnew(uchar, total_length);
  point_cols = vnew(uchar, total_length);
  for (i = 0; i < total_length; i++)
  {
    point_rows[i] = temp_rows[i];
    point_cols[i] = temp_cols[i];
  }
}

  
vPointSequence::~vPointSequence()
{
  vdelete2(lengths);
  vdelete2(indices);
  vdelete2(point_rows);
  vdelete2(point_cols);
}


vPointSequence * vPointSequence::Load(const char * filename)
{
  FILE * fp = fopen(filename, vFOPEN_READ);
  if (fp == 0) return 0;
  vPointSequence * result = new vPointSequence();
  long items = 0;
  items += vReadPCLongs(fp, &(result->rows), 1);
  items += vReadPCLongs(fp, &(result->cols), 1);
  items += vReadPCLongs(fp, &(result->number_of_frames), 1);
  items += vReadPCLongs(fp, &(result->total_length), 1);

  if ((items != 4) || (result->rows < 0) || (result->cols < 0))
  {
    vPrint("Error in loading point set sequence\n");
    vdelete(result);
    fclose(fp);
    return 0;
  }

  result->lengths = vnew(long, result->number_of_frames);
  result->indices = vnew(long, result->number_of_frames);
  result->point_rows = vnew(uchar, result->total_length);
  result->point_cols = vnew(uchar, result->total_length);

  items = vReadPCLongs(fp, vData(result->lengths), result->number_of_frames);
  items += vReadPCLongs(fp, vData(result->indices), result->number_of_frames);
  items += fread(vData(result->point_rows), sizeof(uchar), result->total_length, fp);
  items += fread(vData(result->point_cols), sizeof(uchar), result->total_length, fp);
  fclose(fp);
  if (items != 2 * result->number_of_frames + 2 * result->total_length)
  {
    vPrint("Error in reading point set sequence\n");
    vdelete(result);
    return 0;
  }
  return result;
}


long vPointSequence::Save(const char * filename)
{
  FILE * fp = fopen(filename, vFOPEN_WRITE);
  if (fp == 0) return 0;
  long items = 0;
  items += vWritePCLongs(fp, &rows, 1);
  items += vWritePCLongs(fp, &cols, 1);
  items += vWritePCLongs(fp, &number_of_frames, 1);
  items += vWritePCLongs(fp, &total_length, 1);

  if (items != 4)
  {
    vPrint("Error in saving point set sequence\n");
    fclose(fp);
    return 0;
  }

  items = vWritePCLongs(fp, vData(lengths), number_of_frames);
  items += vWritePCLongs(fp, vData(indices), number_of_frames);
  items += fwrite(vData(point_rows), sizeof(uchar), total_length, fp);
  items += fwrite(vData(point_cols), sizeof(uchar), total_length, fp);
  fclose(fp);
  if (items != 2 * number_of_frames + 2 * total_length)
  {
    vPrint("Error in writing point set sequence\n");
    return 0;
  }
  return 1;
}


// returns the number of points.
// In rowsp and colsp we put arrays of the (row, col)
// coordinates of the set of points corresponding
// to the given frame and given view
long vPointSequence::GetPoints(long frame, long view, 
                                   vArray(long) * rowsp, vArray(long) * colsp)
{
  long number = Length(frame);
  // Get the (row, col) coordinates for given frame and view 0.
  vArray(uchar) frame_rows = PointRows(frame);
  vArray(uchar) frame_cols = PointCols(frame);
  
  // Copy over the coordinates into the result array.
  vArray(long) result_rows = vnew(long, number);
  vArray(long) result_cols = vnew(long, number);
  long i;
  for (i = 0; i < number; i++)
  {
    result_rows[i] = frame_rows[i];
    result_cols[i] = frame_cols[i];
  }

  // Rotate points by rotation corresponding to given view.
  long degrees = view * 90;
  vRotatePointsRight(result_rows, result_cols, result_rows, result_cols,
                      number, rows, cols, degrees);
  *rowsp = result_rows;
  *colsp = result_cols;
  return number;
}


// This function is used mainly for debuging, to visualise the points
// stored for each frame.
grayscale_image * vPointSequence::PointImage(long frame)
{
  if ((frame < 0) || (frame >= number_of_frames)) return 0;
  grayscale_image * point_image = new grayscale_image(rows, cols);
  vWriteValue(point_image, (uchar) 0);
  vArray(uchar) point_rows = PointRows(frame);
  vArray(uchar) point_cols = PointCols(frame);
  long length = Length(frame);
  vWriteValue3(point_image, point_rows, point_cols, length, (uchar) 255);
  return point_image;
}

  
/////////////////////////////////////////////////////////////////////
//                                                                 //
//            End of implementation of vPointSequence             //
//                                                                 //
/////////////////////////////////////////////////////////////////////


long vChamferFeatures::EraseData()
{
  if (point_sequence == 0) return 0;
  long number_of_frames = point_sequence->NumberOfFrames();

  long frame;
  for (frame = 0; frame < number_of_frames; frame++)
  {
    vdelete(distance_transforms[frame]);
  }

  vdelete2(distance_transforms);
  vdelete(point_sequence);
  return 1;
}


/////////////////////////////////////////////////////////////////////
//                                                                 //
//         Beginning of implementation of vHandDatabase1          //
//                                                                 //
/////////////////////////////////////////////////////////////////////

vHandDatabase1::vHandDatabase1()
{
  Initialize();
}


vHandDatabase1::~vHandDatabase1()
{
  vdelete2(name);
  vdelete(color_map);
  vdelete(bounding_boxes);

  vdelete(image_sequence);
  vdelete(big_image_sequence);
  vdelete(definite_edges);
  vdelete(big_definite_edges);
  vdelete(possible_edges);
  vdelete(big_possible_edges);
  vdelete(cartoon_sequence);
  vdelete(big_cartoon_sequence);
  vdelete(old_edges);
  vdelete(big_old_edges);

  ClearAllDistanceTransforms();
  ClearAllPointSequences();
  DeleteRotationTemplates();
}


vHandDatabase1::vHandDatabase1(const char * in_name)
{
  Initialize();
  name = vCopyString(in_name);

  char * bboxes_name = BoundingBoxesName();
  bounding_boxes = v3dMatrix<long>::Read(bboxes_name);
  if (bounding_boxes != 0)
  {
    if ((bounding_boxes->Rows() < 4) ||
        (bounding_boxes->Cols() < number_of_frames))
    {
      vPrint("rows = %li, cols = %li\n", bounding_boxes->Rows(), bounding_boxes->Cols());
      vPrint("number of frames = %li\n", number_of_frames);
      exit_error("Error reading bounding boxes\n");
    }
  }
  vdelete2(bboxes_name);

  char * image_sequence_name = ImageSequenceName();
  image_sequence = (vRLE_Sequence *) vOpenReadVideoFile(image_sequence_name);
  if (image_sequence != 0)
  {
    if (image_sequence->NumberOfFrames() < number_of_frames)
    {
      exit_error("Bad image sequence read from vHandDatabase1\n");
    }
  }
  vdelete2(image_sequence_name);

  char * big_image_sequence_name = BigImageSequenceName();
  big_image_sequence = (vRLE_Sequence *) vOpenReadVideoFile(big_image_sequence_name);
  if (big_image_sequence != 0)
  {
    if (big_image_sequence->NumberOfFrames() < number_of_frames)
    {
      exit_error("Bad big image sequence read from vHandDatabase1\n");
    }
  }
  vdelete2(big_image_sequence_name);

  char * definite_edges_name = DefiniteEdgesName();
  definite_edges = (vRLE_Sequence *) vOpenReadVideoFile(definite_edges_name);
  if (definite_edges != 0)
  {
    if (definite_edges->NumberOfFrames() < number_of_frames)
    {
      exit_error("Bad definite edges read from vHandDatabase1\n");
    }
  }
  vdelete2(definite_edges_name);

  char * big_definite_edges_name = BigDefiniteEdgesName();
  big_definite_edges = (vRLE_Sequence *) vOpenReadVideoFile(big_definite_edges_name);
  if (big_definite_edges != 0)
  {
    if (big_definite_edges->NumberOfFrames() < number_of_frames)
    {
      exit_error("Bad big_definite_edges read from vHandDatabase1\n");
    }
  }
  vdelete2(big_definite_edges_name);

  char * possible_edges_name = PossibleEdgesName();
  possible_edges = (vRLE_Sequence *) vOpenReadVideoFile(possible_edges_name);
  if (possible_edges != 0)
  {
    if (possible_edges->NumberOfFrames() < number_of_frames)
    {
      exit_error("Bad possible edges read from vHandDatabase1\n");
    }
  }
  vdelete2(possible_edges_name);

  char * big_possible_edges_name = BigPossibleEdgesName();
  big_possible_edges = (vRLE_Sequence *) vOpenReadVideoFile(big_possible_edges_name);
  if (big_possible_edges != 0)
  {
    if (big_possible_edges->NumberOfFrames() < number_of_frames)
    {
      exit_error("Bad big possible edges read from vHandDatabase1\n");
    }
  }
  vdelete2(big_possible_edges_name);

  char * cartoon_sequence_name = CartoonSequenceName();
  cartoon_sequence = (vRLE_Sequence *) vOpenReadVideoFile(cartoon_sequence_name);
  if (cartoon_sequence != 0)
  {
    if (cartoon_sequence->NumberOfFrames() < number_of_frames)
    {
      exit_error("Bad cartoon sequence read from vHandDatabase1\n");
    }
  }
  vdelete2(cartoon_sequence_name);

  char * big_cartoon_sequence_name = BigCartoonSequenceName();
  big_cartoon_sequence = (vRLE_Sequence *) vOpenReadVideoFile(big_cartoon_sequence_name);
  if (big_cartoon_sequence != 0)
  {
    if (big_cartoon_sequence->NumberOfFrames() < number_of_frames)
    {
      exit_error("Bad big cartoon sequence read from vHandDatabase1\n");
    }
  }
  vdelete2(big_cartoon_sequence_name);

  char * old_edges_name = OldEdgesName();
  old_edges = (vRLE_Sequence *) vOpenReadVideoFile(old_edges_name);
  if (old_edges != 0)
  {
    if (old_edges->NumberOfFrames() < number_of_frames)
    {
      exit_error("Bad cartoon sequence read from vHandDatabase1\n");
    }
  }
  vdelete2(old_edges_name);

  char * big_old_edges_name = BigOldEdgesName();
  big_old_edges = (vRLE_Sequence *) vOpenReadVideoFile(big_old_edges_name);
  if (big_old_edges != 0)
  {
    if (big_old_edges->NumberOfFrames() < number_of_frames)
    {
      exit_error("Bad big cartoon sequence read from vHandDatabase1\n");
    }
  }
  vdelete2(big_old_edges_name);

  is_valid = 1;
}


void vHandDatabase1::Initialize()
{
  is_valid = 0;
  
  background_id = 0;
  palm_id = 1;
  thumb1_id = 2;
  thumb2_id = 3;
  thumb3_id = 4;
  index1_id = 5;
  index2_id = 6;
  index3_id = 7;
  middle1_id = 8;
  middle2_id = 9;
  middle3_id = 10;
  ring1_id = 11;
  ring2_id = 12;
  ring3_id = 13;
  pinky1_id = 14;
  pinky2_id = 15;
  pinky3_id = 16;
  forearm_id = 17;
  number_of_ids = 18;

  number_of_shapes = 26;
  number_of_viewpoints = 86;
  number_of_orientations = 48;
  number_of_frames = number_of_shapes * number_of_viewpoints;
  frames_per_shape = number_of_viewpoints;

  bounding_length = 192;
  image_size = 288;
  name = 0;
  color_map = DefaultColorMap();
  bounding_boxes = 0;
  adjacencies_matrix = FingerAdjacencies();
  adjacencies = adjacencies_matrix.Matrix2();

  image_sequence = 0;
  big_image_sequence = 0;
  definite_edges = 0;
  big_definite_edges = 0;
  possible_edges = 0;
  big_possible_edges = 0;
  cartoon_sequence = 0;
  big_cartoon_sequence = 0;
  old_edges = 0;
  big_old_edges = 0;

  current_dt_sequence = 0;
  current_point_sequence = 0;
  rotation_templates = vZero(v3dMatrix<short> *);
}


char * vHandDatabase1::AddDirectory(const char * filename)
{
  char * database_directory = vJoinPaths3(sequences_directory, "databases",
                                           name);
  char * result = vJoinPaths(database_directory, filename);
  vdelete2(database_directory);
  return result;
}


// This function takes in an ending and returns the full path
// (i.e., adds to the beginning the directory and the name
// of the database). It is used for filenames refering to
// sequences where the size of the hand has been normalized.
char * vHandDatabase1::FilenameWithEnding(const char * ending)
{
  char * name1 = vMergeStrings2(name, ending);
  char * name2 = AddDirectory(name1);
  vdelete2(name1);
  return name2;
}

  
// Same as FilenameWithEnding, but for sequences that are not
// normalized for the size of the hand.
char * vHandDatabase1::BigFilenameWithEnding(const char * ending)
{
  char * name1 = vMergeStrings3(name, "_big", ending);
  char * name2 = AddDirectory(name1);
  vdelete2(name1);
  return name2;
}


// This function takes as input an ending specifying
// a particular edge image sequence. It returns the full
// path name of a file that is supposed to store the distance
// transforms for that edge image sequence.
char * vHandDatabase1::DistanceTransformFilename(const char * ending)
{
  char * new_ending = vReplaceEnding2(ending, "_dt.avi");
  char * filename = FilenameWithEnding(new_ending);
  vdelete2(new_ending);
  return filename;
}


// This function takes as input an ending specifying
// a particular edge image sequence. It returns the full
// path name of a file that is supposed to store the point
// sequence for that edge image sequence.
char * vHandDatabase1::PointSequenceFilename(const char * ending)
{
  char * new_ending = vReplaceEnding2(ending, ".point_set");
  char * filename = FilenameWithEnding(new_ending);
  vdelete2(new_ending);
  return filename;
}
  

// This is the colormap that is used to efficiently save images
// in RLE sequences, so that each hand link has a different color. An almost
// identical colormap is defined in hand_renderer.cpp, but that file is not
// included in the interface project, so I replicate the function here. 
// This function, unlike the one in hand_renderer.cpp, also handles
// the forearm.
// One thing to note is that the link ID of a pixel is always
// the red value of that pixel modulo 70.
vColorMap * vHandDatabase1::DefaultColorMap()
{
  const long number_of_links = 18;
  vColorMap * result = new vColorMap();
  result->number_of_colors = number_of_links;
  vector<uchar> * R = &(result->R);
  vector<uchar> * G = &(result->G);
  vector<uchar> * B = &(result->B);

  // The color of link 0 (which stands for the background, or,
  // in general, not a part of the hand) is black.
  R->push_back(0);
  G->push_back(0);
  B->push_back(0);

  // The color of the palm
  R->push_back(71);
  G->push_back(70);
  B->push_back(70);

  // Note: link1 is the link towards the palm, link3 is the outermost one.
  // The color of thumb1
  R->push_back(72);
  G->push_back(0);
  B->push_back(0);

  // The color of thumb2
  R->push_back(143);
  G->push_back(0);
  B->push_back(0);

  // The color of thumb3
  R->push_back(214);
  G->push_back(0);
  B->push_back(0);

  // The color of index1
  R->push_back(5);
  G->push_back(70);
  B->push_back(0);

  // The color of index2
  R->push_back(6);
  G->push_back(140);
  B->push_back(0);

  // The color of index3
  R->push_back(7);
  G->push_back(210);
  B->push_back(0);

  // The color of middle1
  R->push_back(8);
  G->push_back(0);
  B->push_back(70);

  // The color of middle2
  R->push_back(9);
  G->push_back(0);
  B->push_back(140);

  // The color of middle3
  R->push_back(10);
  G->push_back(0);
  B->push_back(210);

  // The color of ring1
  R->push_back(81);
  G->push_back(70);
  B->push_back(0);

  // The color of ring2
  R->push_back(152);
  G->push_back(140);
  B->push_back(0);

  // The color of ring3
  R->push_back(223);
  G->push_back(210);
  B->push_back(0);

  // The color of pinky1
  R->push_back(84);
  G->push_back(0);
  B->push_back(70);

  // The color of pinky2
  R->push_back(155);
  G->push_back(0);
  B->push_back(140);

  // The color of pinky3
  R->push_back(226);
  G->push_back(0);
  B->push_back(210);

  // The color of the forearm (for Poser-generated images)
  R->push_back(157);
  G->push_back(140);
  B->push_back(140);

  return result;
}


// Again, a similar function (that, however, does not take care
// of the forearm) is defined in hand_renderer.cpp.
vMatrix<ushort> vHandDatabase1::FingerAdjacencies()
{
  const long link_id_limit = 18;
  vMatrix<ushort> adjacencies_matrix(link_id_limit, link_id_limit);
  vArray2(ushort) adjacencies = adjacencies_matrix.Matrix2();

  vWriteValue(&adjacencies_matrix, (ushort) 0);

  // Add the adjacencies for the palm
  adjacencies[1][2] = 1;
  adjacencies[1][5] = 1;
  adjacencies[1][8] = 1;
  adjacencies[1][11] = 1;
  adjacencies[1][14] = 1;
  adjacencies[1][17] = 1;

  // Adjacencies for inner thumb
  adjacencies[2][1] = 1;
  adjacencies[2][3] = 1;

  // Adjacencies for middle thumb
  adjacencies[3][2] = 1;
  adjacencies[3][4] = 1;

  // Adjacencies for outer thumb
  adjacencies[4][3] = 1;

  // Adjacencies for inner index
  adjacencies[5][1] = 1;
  adjacencies[5][6] = 1;

  // Adjacencies for middle index
  adjacencies[6][5] = 1;
  adjacencies[6][7] = 1;

  // Adjacencies for outer index
  adjacencies[7][6] = 1;

  // Adjacencies for inner middle
  adjacencies[8][1] = 1;
  adjacencies[8][9] = 1;

  // Adjacencies for middle middle
  adjacencies[9][8] = 1;
  adjacencies[9][10] = 1;

  // Adjacencies for outer middle
  adjacencies[10][9] = 1;

  // Adjacencies for inner ring
  adjacencies[11][1] = 1;
  adjacencies[11][12] = 1;

  // Adjacencies for middle ring
  adjacencies[12][11] = 1;
  adjacencies[12][13] = 1;

  // Adjacencies for outer ring
  adjacencies[13][12] = 1;

  // Adjacencies for inner pinky
  adjacencies[14][1] = 1;
  adjacencies[14][15] = 1;

  // Adjacencies for middle pinky
  adjacencies[15][14] = 1;
  adjacencies[15][16] = 1;

  // Adjacencies for outer pinky
  adjacencies[16][15] = 1;

  // Adjacencies for forearm
  adjacencies[17][1] = 1;

  // For sanity check, verify that the matrix is symmetric
  long row, col;
  for (row = 0; row < link_id_limit; row++)
  {
    for (col = 0; col < link_id_limit; col++)
    {
      if (adjacencies[row][col] != adjacencies[col][row])
      {
        exit_error("Error in vFingerAdjacecies, (%li, %li)\n", row, col);
      }
    }
  }

  return adjacencies_matrix;
}


grayscale_image * vHandDatabase1::LinkIDs(color_image * colored_link_image)
{
  return vLinkIDsB(colored_link_image);
}


// This function processes the original poser sequence (which
// we assume already exists) and generates the image sequence
// based on that. The original poser sequence is unnormalized,
// includes the forearm, and has a "LHand Camera" caption on the
// top left.
long vHandDatabase1::CreateImageSequence()
{
  if (big_image_sequence == 0) return 0;

  long frames = big_image_sequence->NumberOfFrames();
  if (frames != number_of_frames)
  {
    exit_error("Bad number of frames (%li) in CreateImageSequence\n",
                    frames);
  }

  char * image_sequence_name = ImageSequenceName();
  vRLE_Sequence * out_sequence = 
    new vRLE_Sequence(image_sequence_name, image_size, image_size);
  if (out_sequence == 0)
  {
    exit_error("Failed to open image_sequence %s\n",
                    image_sequence_name);
  }
  vdelete2(image_sequence_name);
  
  long frame;
  vPrint("\n");
  for (frame = 0; frame < frames; frame++)
  {
    color_image * original = big_image_sequence->ReadFrame(frame);
    if (original == 0)
    {
      exit_error("Failed to read frame %li\n", frame);
    }

    color_image * the_image = RemoveForearm(original);
    vdelete(original);

    long center = Center();
    color_image * normalized = (color_image *)
      vScaleBoundingBox6(the_image, bounding_length, center, center,
                          image_size, image_size);
                    
    grayscale_image * link_ids = LinkIDs(normalized);
    long success = out_sequence->write_frameGray2(link_ids, 2, color_map);
    vdelete(the_image);
    vdelete(normalized);
    vdelete(link_ids);
    if (success <= 0)
    {
      exit_error("Failed to write frame %li.\n", frame);
    }
    vPrint("processed frame %li\r", frame);
  }
  vPrint("\n");

  long success = out_sequence->FinishWriting();
  vdelete(out_sequence);
  return success;
}


// We change the color of the forearm, so that it gets erased
// it becomes like background). We keep, however, forearm pixels
// that are neighbors to palm pixels
color_image * vHandDatabase1::RemoveForearm(color_image * colored_link_image)
{
  // We assume that the borders of the image are all black pixels.
  long rows = colored_link_image->Rows();
  long cols = colored_link_image->Cols();

  // Get the colors for the background, palm and forearm.
  long palm_r, palm_g, palm_b, forearm_r, forearm_g, forearm_b;
  long background_r, background_g, background_b;
  palm_r = palm_g = palm_b = forearm_r = forearm_g = forearm_b = 0;
  background_r = background_g = background_b = 0;
  ObjectColor(background_id, &background_r, &background_g, &background_b);
  ObjectColor(palm_id, &palm_r, &palm_g, &palm_b);
  ObjectColor(forearm_id, &forearm_r, &forearm_g, &forearm_b);

  color_image * result = (color_image *) vCopyImage(colored_link_image);
  vArray2(uchar) result_r = result->GetR2();
  vArray2(uchar) result_g = result->GetG2();
  vArray2(uchar) result_b = result->GetB2();
  long neighbor_drow[] = {-1, -1, -1, 0, 0, 1, 1, 1};
  long neighbor_dcol[] = {-1, 0, 1, -1, 1, -1, 0, 1};
  long neighbors = 8;
  long row, col, neighbor;

  // Go through the image, pixel by pixel
  for (row = 1; row < rows-1; row++)
  {
    for (col = 1; col < cols-1; col++)
    {
      // if the current pixel is not forearm, we don't change it.
      if (result_r[row][col] != forearm_r) continue;
      // If it is forearm, check if it is neighboring the palm
      long palm_neighbor = 0;
      for (neighbor = 0; neighbor < neighbors; neighbor++)
      {
        long neighbor_row = row + neighbor_drow[neighbor];
        long neighbor_col = col + neighbor_dcol[neighbor];
        if (result_r[neighbor_row][neighbor_col] == palm_r)
        {
          palm_neighbor = 1;
          break;
        }
      }
      // If it is not neighboring the palm, color it as background.
      if (palm_neighbor <= 0) 
      {
        result_r[row][col] = (uchar) background_r;
        result_g[row][col] = (uchar) background_g;
        result_b[row][col] = (uchar) background_b;
      }
    }
  }

  return result;
}


// We save the transformation that we applied to normalize the bounding
// box of each frame in the original Poser sequence, so that we 
// can apply the same transformation to any features extracted from 
// that sequence (or its associated cartoon sequence).
// The transformation is saved in matrix form, it is a matrix
// with long entries, one band, 4 rows, and frames columns.
long vHandDatabase1::SaveBoundingBoxes()
{
  if (big_image_sequence == 0) return 0;

  long frames = big_image_sequence->NumberOfFrames();
  if (frames != number_of_frames)
  {
    exit_error("Bad number of frames (%li) in CreateImageSequence\n",
                    frames);
  }

  v3dMatrix<long> bounding_boxes_matrix(4, frames, 1);
  vArray2(long) bounding_boxes = bounding_boxes_matrix.Matrix2(0);

  long frame;
  vPrint("\n");
  for (frame = 0; frame < frames; frame++)
  {
    color_image * original = big_image_sequence->ReadFrame(frame);
    if (original == 0)
    {
      exit_error("Failed to read frame %li\n", frame);
    }

    color_image * the_image = RemoveForearm(original);
    vdelete(original);
    vBoundingBox box = vBinaryBoundingBox(the_image);
    vdelete(the_image);
    bounding_boxes[0][frame] = box.top;
    bounding_boxes[1][frame] = box.bottom;
    bounding_boxes[2][frame] = box.left;
    bounding_boxes[3][frame] = box.right;
    vPrint("processed frame %li\r", frame);
  }
  vPrint("\n");

  char * bounding_boxes_file = BoundingBoxesName();
  long success = bounding_boxes_matrix.Write(bounding_boxes_file);
  vdelete2(bounding_boxes_file);
  vdelete(big_image_sequence);
  return success;
}
  

// Extracts noise-free edges from the hand image, which, we
// assume, follows the color code that this database uses.
// The problem with this function is that it considers borders
// between neighboring links not to be edges. Sometimes those
// borders represent discontinuities in 3D, and in those
// cases they should be considered as edges.
grayscale_image * vHandDatabase1::ArticulatedEdges(color_image * object)
{
  vMatrix<ushort> adjacencies = FingerAdjacencies();
  long rows = object->Rows();
  long cols = object->Cols();
  grayscale_image * result = new grayscale_image(rows, cols);
  vArray2(ushort) neighbors = adjacencies.Matrix2();

  // red will be the red band for a 3-band image, but it will simply
  // be the intensity band for a one-band (grayscale) image.
  vArray2(uchar) red = object->Matrix2(0);
  vArray2(uchar) edges = result->GetI2();

  long row, col;
  vWriteBorderValue(result, (uchar) 0);
  
  for (row = 1; row < rows - 1; row++)
  {
    for (col = 1; col < cols - 1; col++)
    {
      // Read the values of the center and the 4 neighbors.
      uchar center = red[row][col];
      if (center == 0) 
      {
        edges[row][col] = 0;
        continue;
      }
      // The id of the link of the center is obtained by
      // taking the remainder of dividing center by 70.
      long center_id = LinkID(center);
      long around[4];
      around[0] = red[row][col-1];
      around[1] = red[row][col+1];
      around[2] = red[row-1][col];
      around[3] = red[row+1][col];
 
      // Now check if center is an edge pixel. Compare
      // center to each neighbor.
      long i;
      long edge_flag = 0;
      for (i = 0; i < 4; i++)
      {
        // If neighbor is same as center, then no edge has been
        // found so far.
        if (center == around[i]) continue;

        // If neighbor is not same as center, check if neighbor
        // belongs to an anatomically neighboring link. If not,
        // then center is a "finger edge".
        uchar neighbor_id = (uchar) LinkID((uchar) around[i]);
        if (neighbors[center_id][neighbor_id] == 1) continue;
        // If we get here, we do have an edge pixel. However,
        // its neighbor is also an edge pixel. To avoid having edge lines
        // that are two pixels thick, we will label as an edge pixel
        // only the pixel with the highest id. This has the additional
        // advantage that the forearm/palm intersection will be labeled
        // with the forearm id.
        else if (center_id > neighbor_id)
        {
          edge_flag = 1;
          break;
        }
      }

      if (edge_flag == 1)
      {
        // We have a finger edge. Mark it with the id of the center, and
        // add 210 to make it more visible. The convention that we follow
        // anyway is that id = intensity % 70.
        edges[row][col] = (uchar) (210 + center_id);
      }
      else 
      {
        // center is not an edge, we mark the pixel with 0.
        edges[row][col] = 0;
      }
    }
  }
  return result;
}  

// This function takes in a frame number, and returns the 
// edge image for the corresponding database pose, using
// the original frame and the cartoon drawing of that frame.
// By using both images, edges between neighboring links that
// represent discontinuities in 3D can be detected with 
// reasonable accuracy, since, in such edges, there is 
// an edge pixel in the cartoon image.
grayscale_image * vHandDatabase1::EdgeImage(long frame)
{
  if (big_image_sequence == 0) return 0;
  if (big_cartoon_sequence == 0) return 0;
  if ((frame < 0) || (frame >= number_of_frames)) return 0;

  grayscale_image * cartoon_image = big_cartoon_sequence->ReadFrameGray(frame);
  if (cartoon_image == 0) exit_error("EdgeImage1\n");
  color_image * link_ids_image1 = big_image_sequence->ReadFrame(frame);
  if (link_ids_image1 == 0) exit_error("EdgeImage2\n");
  color_image * link_ids_image = RemoveForearm(link_ids_image1);
  vdelete(link_ids_image1);

  long rows = link_ids_image->Rows();
  long cols = link_ids_image->Cols();
  grayscale_image * result = new grayscale_image(rows, cols);

  // red will be the red band for a 3-band image, but it will simply
  // be the intensity band for a one-band (grayscale) image.
  vArray2(uchar) link_ids = link_ids_image->Matrix2(0);
  vArray2(uchar) edges = result->GetI2();
  vArray2(uchar) cartoon = cartoon_image->Matrix2(0);

  long row, col;
  vWriteBorderValue(result, (uchar) 0);
  
  for (row = 1; row < rows - 1; row++)
  {
    for (col = 1; col < cols - 1; col++)
    {
      // Read the values of the center and the 4 neighbors.
      long center = LinkID(link_ids[row][col]);
      if (center == 0) 
      {
        edges[row][col] = 0;
        continue;
      }
      // The id of the link of the center is obtained by
      // taking the remainder of dividing center by 70.
      long neighbor_rows[4], neighbor_cols[4];
      vFourNeighbors(row, col, neighbor_rows, neighbor_cols);
      
      // Now check if center is an edge pixel. Compare
      // center to each neighbor.
      long i;
      long edge_flag = 0;
      for (i = 0; i < 4; i++)
      {
        long neighbor_row = neighbor_rows[i];
        long neighbor_col = neighbor_cols[i];
        long neighbor = LinkID(link_ids[neighbor_row][neighbor_col]);

        // If neighbor is same as center, then no edge has been
        // found so far.
        // Also, to avoid getting edge lines that are two pixels thick, when two
        // pixels represent a boundary, we assign an edge pixel only
        // to the pixel among the two that has the highest link ID. This
        // has the additional advantage that edges between the forearm and
        // other links will get the forearm ID, so it will be easy to 
        // identify the border between palm and forearm in the result.
        if (center <= neighbor) continue;

        // If neighbor is not same as center, check if neighbor
        // belongs to an anatomically neighboring link. If not,
        // then center is a "finger edge".

        // This if statement handles the case where the pixels are
        // neighbors. In that case, usually there is no edge pixel,
        // but sometimes there should be an edge pixel, if there
        // is a significant depth discontinuity at that point. To determine
        // the presence of discontinuity, we consult the cartoon image. If there
        // is an edge pixel in the cartoon image at this point or at one
        // of the four neighbors, then this is an edge.
        if (adjacencies[center][neighbor] == 1)
        {
          if ((cartoon[row][col] != 0) ||
              (cartoon[neighbor_row][neighbor_col] != 0))
          {
            edge_flag = 1;
            break;
          }
        }

        // If we get here, we do have an edge pixel. However,
        // its neighbor is also an edge pixel. 
        else if (center > neighbor)
        {
          edge_flag = 1;
          break;
        }
      }

      if (edge_flag == 1)
      {
        // We have a finger edge. Mark it with the id of the center, and
        // add 210 to make it more visible. The convention that we follow
        // anyway is that id = intensity % 70.
        edges[row][col] = (uchar) (210 + center);
      }
      else 
      {
        // center is not an edge, we mark the pixel with 0.
        edges[row][col] = 0;
      }
    }
  }

  vdelete(cartoon_image);
  vdelete(link_ids_image);
  return result;
}

  
// Return the bounding box of the given frame, in the original 
// poser sequence.
vBoundingBox vHandDatabase1::BoundingBox(long frame)
{
  if ((frame < 0) || (frame >= number_of_frames))
  {
    return vBoundingBox();
  }

  if (bounding_boxes == 0)
  {
    return vBoundingBox();
  }
  vBoundingBox result;
  vArray2(long) boxes = bounding_boxes->Matrix2(0);
  result.top = boxes[0][frame];
  result.bottom = boxes[1][frame];
  result.left = boxes[2][frame];
  result.right = boxes[3][frame];
  return result;
}


// The goal of this function is to create a sequence of edge
// images that is as similar as possible to the edge images
// generated using VirtualHand.
long vHandDatabase1::RecreateOldEdgeSequence()
{
  if (big_image_sequence == 0) return 0;
  if (big_cartoon_sequence == 0) return 0;
  if (bounding_boxes == 0) return 0;
  long rows = big_image_sequence->Rows();
  long cols = big_image_sequence->Cols();

  char * filename1 = BigOldEdgesName();
  char * filename2 = OldEdgesName();
  vRLE_Sequence * out_sequence1 = new vRLE_Sequence(filename1, rows, 
                                                     cols);
  vRLE_Sequence * out_sequence2 = new vRLE_Sequence(filename2, image_size, 
                                                     image_size);
  if ((out_sequence1->IsValid() <= 0) || (out_sequence2->IsValid() <= 0))
  {
    exit_error("RecreateOldEdgeSequence, error 1\n");
  }
  vdelete2(filename1);
  vdelete2(filename2);

  long center = Center();

  vPrint("\n");
  long frame;
//  for (frame = 0; frame < 2; frame++)
  for (frame = 0; frame < number_of_frames; frame++)
  {
    grayscale_image * edges = EdgeImage(frame);
    if (edges == 0)
    {
      exit_error("RecreateOldEdgeSequence, error 2\n");
    }
    long success = out_sequence1->write_frameGray(edges);
    if (success <= 0)
    {
      exit_error("RecreateOldEdgeSequence, error 3\n");
    }

    vBoundingBox bbox = BoundingBox(frame);
    grayscale_image * normalized_edges = (grayscale_image *)
      vSparseScaleBoundingBox7(edges, bbox, bounding_length,
                                center, center, image_size, image_size);
    success = out_sequence2->write_frameGray(normalized_edges);
    if (success <= 0)
    {
      exit_error("RecreateOldEdgeSequence, error 3.5\n");
    }

    vdelete(edges);
    vdelete(normalized_edges);
    vPrint("Processed frame %li of %li\r", frame, number_of_frames);
  }
  vPrint("\n");
  long success = out_sequence1->FinishWriting();
  if (success <= 0) 
  {
    exit_error("RecreateOldEdgeSequence, error 4\n");
  }

  success = out_sequence2->FinishWriting();
  if (success <= 0) 
  {
    exit_error("RecreateOldEdgeSequence, error 5\n");
  }
  vdelete(out_sequence1);
  vdelete(out_sequence2);
  return 1;
}


// This function processes the unnormalized cartoon edge
// images and creates normalized versions of them, in which
// the forearm is removed.
long vHandDatabase1::CreateCartoonSequence()
{
  if (big_image_sequence == 0) return 0;
  if (big_cartoon_sequence == 0) return 0;

  long frames = big_cartoon_sequence->NumberOfFrames();
  if (frames != number_of_frames)
  {
    exit_error("Bad number of frames (%li) in CreateCartoonSequence\n",
                    frames);
  }

  char * cartoon_sequence_name = CartoonSequenceName();
  vRLE_Sequence * out_sequence = 
    new vRLE_Sequence(cartoon_sequence_name, image_size, image_size);
  if (out_sequence == 0)
  {
    exit_error("Failed to open cartoon_sequence %s\n",
                    cartoon_sequence_name);
  }
  vdelete2(cartoon_sequence_name);
  
  long center = Center();
  long frame;
  long rows = big_image_sequence->Rows();
  long cols = big_image_sequence->Cols();
  long size = rows * cols;
  vPrint("\n");
//  for (frame = 0; frame < 3; frame++)
  for (frame = 0; frame < frames; frame++)
  {
    grayscale_image * original = big_cartoon_sequence->ReadFrameGray(frame);
    // We also need to read the color-coded image, to remove 
    // the forearm.
    color_image * link_ids1 = big_image_sequence->ReadFrame(frame);
    if ((original == 0) || (link_ids1 == 0))
    {
      exit_error("Failed to read frame %li\n", frame);
    }

    // Now, erase the forearm. We will remove the forearm from the colored
    // link image, we will apply the distance transform to the result,
    // and we will remove all cartoon pixels that have a value of more
    // than 2 or 3 in the distance transform.
    color_image * link_ids = RemoveForearm(link_ids1);
    v3dMatrix<float> * dt_image = vCVDistanceTransform(link_ids);
    vArray(float) dt = dt_image->Matrix(0);
    vArray(uchar) cartoon = original->Matrix(0);
    long i;
    vArray(uchar) ids = link_ids->Matrix(0);
    for (i = 0; i < size; i++)
    {
      if (dt[i] > 2) cartoon[i] = 0;
    }
    
    // Now we can normalize the cartoon edge image.
    vBoundingBox bbox = BoundingBox(frame);
    long center = Center();
    grayscale_image * normalized = (grayscale_image *)
      vSparseScaleBoundingBox7(original, bbox, bounding_length, 
                                center, center, image_size, image_size);
                    
    long success = out_sequence->write_frameGray(normalized);
    vdelete(link_ids1);
    vdelete(link_ids);
    vdelete(dt_image);
    vdelete(normalized);
    vdelete(original);
    if (success <= 0)
    {
      exit_error("Failed to write frame %li.\n", frame);
    }
    vPrint("processed frame %li\r", frame);
  }
  vPrint("\n");

  long success = out_sequence->FinishWriting();
  vdelete(out_sequence);
  return success;
}


// CreateDistanceTransform takes in as argument an ending, like the
// endings we pass to FilenameWithEnding, that specify sequences that
// store particular pose features. We assume that the normalized edge
// sequence specified by that ending exists. This function creates
// another sequence, that contains the distance transform of each
// frame of the normalized edge sequence.
long vHandDatabase1::CreateDistanceTransform(const char * ending)
{
  char * edge_sequence_name = FilenameWithEnding(ending);
  vVideoFile * edge_sequence = vOpenReadVideoFile(edge_sequence_name);
  vdelete2(edge_sequence_name);
  if (edge_sequence == 0) return 0;

  long rows = edge_sequence->Rows();
  long cols = edge_sequence->Cols();
  long size = rows * cols;
  char * dt_filename = DistanceTransformFilename(ending);
  vAVI_Stream * dt_sequence = new vAVI_Stream(dt_filename, rows, cols, 1);
  vdelete2(dt_filename);
  if (dt_sequence->IsValid() <= 0)
  {
    exit_error("CreateDistanceTransform: error 1\n");
  }

  long frames = edge_sequence->NumberOfFrames();
  long frame;
  vPrint("\n");
//  for (frame = 0; frame < 2; frame++)
  for (frame = 0; frame < frames; frame++)
  {
    grayscale_image * edge_image = edge_sequence->ReadFrameGray(frame);
    if (edge_image == 0)
    {
      exit_error("CreateDistanceTransform: error 2\n");
    }
    v3dMatrix<float> * dt_image = vCVDistanceTransform(edge_image);
    grayscale_image * dt_frame = new grayscale_image(rows, cols);
    vArray(float) dt_float = dt_image->Matrix(0);
    vArray(uchar) dt_uchar = dt_frame->Matrix(0);
    long i;
    for (i = 0; i < size; i++)
    {
      long value = vRound(dt_float[i]);
      if (value > 255) dt_uchar[i] = (uchar) 255;
      else dt_uchar[i] = (uchar) value;
    }

    long success = dt_sequence->write_frameGray(dt_frame);
    if (success == 0)
    {
      exit_error("CreateDistanceTransform: error 3\n");
    }
    vdelete(edge_image);
    vdelete(dt_image);
    vdelete(dt_frame);
    vPrint("processed frame %li of %li\r", frame, frames);
  }
  vPrint("\n");
  long success = dt_sequence->FinishWriting();
  if (success <= 0)
  {
    exit_error("CreateDistanceTransform: error 5\n");
  }
  vCloseVideoFile(edge_sequence);
  vdelete(dt_sequence);
  return 1;
}


// CreatePointSequence, like CreateDistanceTransform,
// takes in as argument an ending, like the
// endings we pass to FilenameWithEnding, that specify sequences that
// store particular pose features. We assume that the normalized edge
// sequence specified by that ending exists. This function simply extracts
// the non-zero points (edge pixels) of each frame, and stores them.
// This gives us a representation that, together with distance transforms,
// can be used for efficient implementation of the chamfer distance,
// in terms of both time and memory.
long vHandDatabase1::CreatePointSequence(const char * ending)
{
  char * edge_sequence_name = FilenameWithEnding(ending);
  vVideoFile * edge_sequence = vOpenReadVideoFile(edge_sequence_name);
  vdelete2(edge_sequence_name);
  if (edge_sequence == 0) return 0;

  vPointSequence sequence(edge_sequence);
  char * filename = PointSequenceFilename(ending);
  long success = sequence.Save(filename);
  vdelete2(filename);
  vCloseVideoFile(edge_sequence);
  return success;
}


// Loads the specified distance transform into distance transforms, 
// and, if successful, sets it as the current distance transform.
long vHandDatabase1::LoadDistanceTransform(const char * ending)
{
  if (distance_transforms.Get(ending) != 0) 
  {
    SetCurrentDistanceTransform(ending);
    return 1;
  }

  char * filename = DistanceTransformFilename(ending);
  vPrint("filename = %s\n", filename);
  vVideoFile * dt_sequence = vOpenReadVideoFile(filename);
  vdelete2(filename);
  if (dt_sequence == 0) return 0;
  if (dt_sequence->NumberOfFrames() < number_of_frames)
  {
    vCloseVideoFile(dt_sequence);
    return 0;
  }

  vArray(grayscale_image *) new_dt_sequence = vnew(grayscale_image *, 
                                                number_of_frames);
  long i;
  vPrint("\n");
  for (i = 0; i < number_of_frames; i++)
  {
    new_dt_sequence[i] = dt_sequence->ReadFrameGray(i);
    if (new_dt_sequence[i] == 0)
    {
      exit_error("LoadDistanceTransform: error1\n");
    }
    vPrint("Processed frame %li of %li\r", i, number_of_frames);
  }
  vPrint("\n");

  distance_transforms.Put(ending, (void *) actual_pointer (new_dt_sequence));
  SetCurrentDistanceTransform(ending);
  vCloseVideoFile(dt_sequence);
  return 1;
}


// Sets the specified distance transform sequence as the current one.
long vHandDatabase1::SetCurrentDistanceTransform(const char * ending)
{
  grayscale_image ** dt_sequence = 
    (grayscale_image **) distance_transforms.Get(ending);
  if (dt_sequence != 0) 
  {
    current_dt_sequence = dt_sequence;
    return 1;
  }
  else
  {
    return 0;
  }
}


// Removes the specified distance transform from memory.
long vHandDatabase1::ClearDistanceTransform(const char * ending)
{
  grayscale_image ** dt_sequence = 
    (grayscale_image **) distance_transforms.Remove(ending);
  if (current_dt_sequence == dt_sequence)
  {
    current_dt_sequence = 0;
  }
  if (dt_sequence != 0)
  {
    long i;
    for (i = 0; i < number_of_frames; i++)
    {
      vdelete(dt_sequence[i]);
    }
    vdelete2(dt_sequence);
    return 1;
  }
  else
  {
    return 0;
  }
}


// Removes all distance transforms from memory.
long vHandDatabase1::ClearAllDistanceTransforms()
{
  long size = distance_transforms.Size();
  long i;
  for (i = size-1; i >= 0; i--)
  {
    grayscale_image ** dt_sequence = 
      (grayscale_image **) distance_transforms.RemoveIndex(i);
    long j;
    for (j = 0; j < number_of_frames; j++)
    {
      vdelete(dt_sequence[j]);
    }
    vdelete2(dt_sequence);
  }
  current_dt_sequence = 0;
  return 1;
}


// Returns the i-th frame of the current DT sequence.
// Used mostly for debugging.
grayscale_image * vHandDatabase1::GetDistanceTransform(long frame)
{
  if (current_dt_sequence == 0) return 0;
  if (frame < 0) return 0;
  if (frame >= number_of_frames) return 0;
  return current_dt_sequence[frame]->ToGrayCopy();
}


// Loads the specified point_sequence into point_sequences, 
// and, if successful, sets it as the current point_sequence.
long vHandDatabase1::LoadPointSequence(const char * ending)
{
  if (point_sequences.Get(ending) != 0) 
  {
    SetCurrentPointSequence(ending);
    return 1;
  }

  char * filename = PointSequenceFilename(ending);
  vPointSequence * point_sequence = vPointSequence::Load(filename);
  vdelete2(filename);
  if (point_sequence == 0) return 0;
  if (point_sequence->NumberOfFrames() < number_of_frames)
  {
    vdelete(point_sequence);
    return 0;
  }

  point_sequences.Put(ending, (void *) point_sequence);
  SetCurrentPointSequence(ending);
  return 1;
}

// Sets the specified point sequenceas current, if it is 
// already loaded.
long vHandDatabase1::SetCurrentPointSequence(const char * ending)
{
  vPointSequence * point_sequence = 
    (vPointSequence *) point_sequences.Get(ending);

  if (point_sequence != 0) 
  {
    current_point_sequence = point_sequence;
    return 1;
  }
  else
  {
    return 0;
  }
}

// Removes the specified distance transform from memory
long vHandDatabase1::ClearPointSequence(const char * ending)
{
  vPointSequence * point_sequence = 
    (vPointSequence *) point_sequences.Remove(ending);
  if (current_point_sequence == point_sequence)
  {
    current_point_sequence = 0;
  }
  if (point_sequence != 0)
  {
    vdelete(point_sequence);
    return 1;
  }
  else
  {
    return 0;
  }
}


// Removes all distance transforms from memory.
long vHandDatabase1::ClearAllPointSequences()
{
  long size = point_sequences.Size();
  long i;
  for (i = size-1; i >= 0; i--)
  {
    vPointSequence * point_sequence = 
      (vPointSequence *) point_sequences.RemoveIndex(i);
    vdelete(point_sequence);
  }
  current_point_sequence = 0;
  return 1;
}


// Returns the i-th frame of the current point sequence.
// Used mostly for debugging.
grayscale_image * vHandDatabase1::GetPointImage(long frame)
{
  if (current_point_sequence == 0) return 0;
  if (frame < 0) return 0;
  if (frame >= current_point_sequence->NumberOfFrames()) return 0;
  return current_point_sequence->PointImage(frame);
}


// Create or delete the rotation templates that can be used to 
// generate different image plane rotations of a query image.
long vHandDatabase1::CreateRotationTemplates()
{
  if (vIsZero(rotation_templates) == 0) return 1;

  rotation_templates = vnew(v3dMatrix<short> *, number_of_orientations);
  double step = 360.0 / (double) number_of_orientations;
  long i;
  long center = Center();
  for (i = 0; i < number_of_orientations; i ++)
  {
    // angle is positive. Positive angles represent clockwise rotation.
    // Rotating the query by a negative angle and matching
    // with a database frame is the same as matching the unrotated query
    // with the database frame rotated by the equivalent positive angle.
    // However, because the database was built using left hands, the 
    // angles must be inverted.
    double angle = step * (double) i;
    rotation_templates[i] = 
      vRotationTemplate(image_size, image_size, center, center, angle);
  }
  return 1;
}


long vHandDatabase1::DeleteRotationTemplates()
{
  if (vIsZero(rotation_templates)) return 1;
  long i;
  for (i = 0; i < number_of_orientations; i++)
  {
    vdelete(rotation_templates[i]);
  }
  vdelete2(rotation_templates);
  return 1;
}


// For debugging:
v3dMatrix<short> * vHandDatabase1::RotationTemplate(long rotation_index)
{
  CreateRotationTemplates();
  if (rotation_index < 0) return 0;
  if (rotation_index >= number_of_orientations) return 0;
  v3dMatrix<short> * result = 
    new v3dMatrix<short>(rotation_templates[rotation_index]);
  return result;
}


// Prepare the chamfer featuers of a query input image (that
// is assumed to be an edge image). Those features
// can be used to efficiently calculate chamfer distances.
vChamferFeatures vHandDatabase1::ChamferFeatures(grayscale_image * query_image)
{
  CreateRotationTemplates();
  vBoundingBox bbox = vBinaryBoundingBox(query_image);
  long limit = vRound((float) image_size / sqrt((float) 2) - (float) 3);
  long center = Center();

  // Before we rotate, we must make sure that there are no pixels that would
  // move out of bounds by rotating. Scaling introduces errors of up to one
  // pixel, so we avoid it if possible. However, we do scale down if 
  // the bounding box indicates that some pixels may fall out of bounds.
  grayscale_image * small_query;
  if (bbox.MaxSide() > limit)
  {
    small_query = (grayscale_image *) 
      vSparseScaleBoundingBox6(query_image, limit, center, center,
                                image_size, image_size);
  }
  else
  {
    small_query = (grayscale_image *) 
      vSparseScaleBoundingBox6(query_image, bbox.MaxSide(), center, center,
                                image_size, image_size);
  }

  grayscale_image ** rotated = new grayscale_image*[number_of_orientations];
  grayscale_image ** dts = new grayscale_image*[number_of_orientations];

  long i;
  for (i = 0; i < number_of_orientations; i++)
  {
    grayscale_image * rotated_query = (grayscale_image *) 
      vWarpImage(small_query, rotation_templates[i]);
    rotated[i] = (grayscale_image *) 
      vSparseScaleBoundingBox6(rotated_query, bounding_length, 
                                center, center, image_size, image_size);
    v3dMatrix<float> * dt = vCVDistanceTransform(rotated[i]);
    dts[i] = (grayscale_image *) vCroppedUchar(dt);
    vdelete(rotated_query);
    vdelete(dt);
  }

  vPointSequence * point_sequence = 
    new vPointSequence(rotated, number_of_orientations);
  vdelete(small_query);
  for (i = 0; i < number_of_orientations; i++)
  {
    vdelete(rotated[i]);
  }
  vdelete2(rotated);
  return vChamferFeatures(point_sequence, dts);
}


// Generate the set of all valid database poses, in order 
// of increasing frames, and increasing orientations between
// poses of the same frame.
vPoseSet vHandDatabase1::AllPoses(long frames, long orientations)
{
  vPoseSet poses;
  long frame, orientation;
  for (frame = 0; frame < frames; frame++)
  {
    for(orientation = 0; orientation < orientations; orientation++)
    {
      poses.AddPose(vDatabasePose(frame, orientation));
    }
  }
  return poses;
}


// returns a sequence saved in a file ending with "_real1.avi".
// that sequence should contain realistic images of hands.
vVideoFile * vHandDatabase1::RealImages()
{
  char * filename = FilenameWithEnding("_real1.avi");
  vVideoFile * sequence = vOpenReadVideoFile(filename);
  if (sequence == 0)
  {
    vPrint("failed to read from %s\n", filename);
  }

  vdelete2(filename);
  return sequence;
}
