#ifndef VASSILIS_HAND_DATABASE_H
#define VASSILIS_HAND_DATABASE_H


#include "algorithms.h"
#include "virtual_hand/FrameViewFinder.h"
#include "quaternion.h"

class vHandPose
{
public:
  long shape_id;
  double camera_x, camera_y, orientation;

  inline long Print()
  {
    return vPrint("shape: %li, 3D orientation: %.2lf %.2lf %.2lf\n",
                   shape_id, camera_x, camera_y, orientation);
  }

  inline long Scan()
  {
    return vScan("%li %lf %lf %lf", &shape_id, 
                  &camera_x, &camera_y, &orientation);
  }

  // Make sure camera_x is between 0 and 180 degrees, and camera_y and
  // orientation are between 0 and 360 degrees
  inline void Normalize()
  {
    camera_x = vNormalizeDegrees(camera_x);
    if (camera_x > 180)
    {
      camera_x = 360 - camera_x;
      camera_y += 180;
      orientation += 180;
    }
    camera_y = vNormalizeDegrees(camera_y);
    orientation = vNormalizeDegrees(orientation);
  }
};


inline double vOldPoseDistance(vHandPose & pose1, vHandPose & pose2)
{
  return vOldViewDistance2(pose1.camera_x, pose1.camera_y, pose1.orientation,
                            pose2.camera_x, pose2.camera_y, pose2.orientation);
}

// Note: As of 7/24/2003, this is old code. I just preserve it for future 
// reference. To manipulate the databases generated using Poser I will
// create a new vHandDatabase class. There is no complete implementation
// of anything using vOldDatabase.
// This is an abstract class. Actual database implementations may vary, but
// they have to support the common interface.
// A general note: hand images are indexed by frame and view. "frame" refers
// to the frame number in the sequence that corresponds to the image, and "view"
// describes (in a way that depends on the actual subclass of the vOldDatabase
// class) how that frame should be rotated. Since image-plane rotations are easy 
// to do on the fly, they don't have to be saved on disk.
class vOldDatabase
{
protected:
  char * name;
  long is_valid;

  vRLE_Sequence * image_sequence;
  vRLE_Sequence * edge_sequence;

  long number_of_images;
  long number_of_shapes;
  // number of viewpoints (camera locations) 
  long number_of_viewpoints;
  long number_of_orientations;
  // Number of image-plane rotation variations that are derived
  // from a single frame.
  long number_of_views; 
  // Number of frames stored in the sequence.
  long number_of_frames;
  // Number of frames stored per shape.
  long frames_per_shape;
  // pose_to_frame_map contains precomputed information about what
  // database frame a given pose corresponds to. The space
  // (x,y) is subsampled so that x angles are picked at 0.2 intervals and
  // y angles are picked also at 0.2 intervals. This way map_factor is
  // equal to 5.
  vMatrix<long> pose_to_frame_map;
  vArray2(long) pose_to_frame;
  // For some pairs of (x,y), that pose_to_frame maps to (x', y'),
  // it could be that the pose (x, y, 0) is significantly far from
  // the pose (x', y', 0). orientation_adjustment will store the value z'
  // such that (x, y, z) is as close as possible to (x', y', z+z').
  vMatrix<float> orientation_adjustment_map;
  vArray2(float) orientation_adjustment;
  double map_factor;
  // maximum x angle that will be represented in the map.
  double max_x;
  double max_y;
  virtual void Zero() = 0;
  virtual long ComputePoseToFrameMap();
  virtual long LoadPoseToFrameMap();
  virtual long SavePoseToFrameMap();

public:
  vOldDatabase();
  virtual ~vOldDatabase();

  // Returns image indexed by given shape id and 3D orientation
  virtual color_image * PoseImage(vHandPose & pose) = 0;

  // Rotate an image the way a database frame should be rotated to
  // correspond to the particular view.
  virtual grayscale_image * RotateImage(grayscale_image * image, long view) = 0;

  virtual color_image * RotateImage(color_image * image, long view) = 0;

  // Returns image indexed by shape_id, viewpoint_index, orientation_index
  // The caller is assumed to know what actual shapes and 3D orientations
  // these indices correspond to.
  virtual color_image * GetImage3(long shape_id, long viewpoint_index, 
                                  long orientation_index) = 0;

  virtual color_image * GetImage2(long frame, long view) = 0;

  // Returns edge image indexed by given shape id and 3D orientation
  virtual grayscale_image * PoseEdges(vHandPose & pose) = 0;

  // Returns edge image indexed by shape_id, viewpoint_index, orientation_index
  // The caller is assumed to know what actual shapes and 3D orientations
  // these indices correspond to.
  virtual grayscale_image * GetEdges3(long shape_id, long viewpoint_index, 
                                  long orientation_index) = 0;

  virtual grayscale_image * GetEdges2(long frame, long view) = 0;

  // Get the hand pose that corresponds to the given frame and view.
  virtual vHandPose GetPose2(long frame, long view) = 0;

  virtual vHandPose GetPose(long frame) = 0;

  virtual long GetPoseFrame(vHandPose pose, long * framep, long * viewp) = 0;

  virtual long FastPoseFrame(vHandPose & pose, long * framep, long * viewp) = 0;

  static char * AddDirectory(const char * filename);

  inline const char * Name()
  {
    return name;
  }

  inline long NumberOfShapes()
  {
    return number_of_shapes;
  }

  inline long NumberOfViewpoints()
  {
    return number_of_viewpoints;
  }

  inline long NumberOfViews()
  {
    return number_of_views;
  }

  inline long IsValid()
  {
    return is_valid;
  }

  inline long NumberOfImages()
  {
    return number_of_images;
  }

  virtual double OrientationAdjustment(vHandPose & pose1, long frame);

  char * DatabaseFilename();
  static char * DatabaseFilename1(const char * in_name);
  char * ImageSequenceName();
  static char * ImageSequenceName1(const char * in_name);
  char * EdgeSequenceName();
  static char * EdgeSequenceName1(const char * in_name);
  // Filename where the pose_to_frame_map is stored.
  char * PoseFrameName();
  static char * PoseFrameName1(const char * in_name);
};


// This class was an attempt to implement the  first database, where we have
// 26 basic hand shapes, and for each shape we have 86 viewpoints and 48
// orientations. Images are normalized so that the bounding box size is 192.
// As of 7/24/2003, I still haven't finished this implementation, and I don't
// intend to finish it.
class vOldDatabase1 : public vOldDatabase
{
private:
  long images_per_viewpoint;
  long number_of_viewpoints;
  long number_of_shapes;
  long number_of_frames;
  vFrameViewFinder finder;

protected:
  virtual void Zero();

public:
  vOldDatabase1();
  virtual ~vOldDatabase1();

  // Returns image indexed by given shape id and 3D orientation
  color_image * PoseImage(vHandPose & pose);

  virtual grayscale_image * RotateImage(grayscale_image * image, long view);

  virtual color_image * RotateImage(color_image * image, long view);

  // Returns image indexed by shape_id, viewpoint_index, orientation_index
  // The caller is assumed to know what actual shapes and 3D orientations
  // these indices correspond to.
  virtual color_image * GetImage3(long shape_id, long viewpoint_index, 
                                  long orientation_index);

  virtual color_image * GetImage2(long frame, long view);

  // Returns edge image indexed by given shape id and 3D orientation
  virtual grayscale_image * PoseEdges(vHandPose & pose);

  // Returns edge image indexed by shape_id, viewpoint_index, orientation_index
  // The caller is assumed to know what actual shapes and 3D orientations
  // these indices correspond to.
  virtual grayscale_image * GetEdges3(long shape_id, long viewpoint_index, 
                                  long orientation_index);

  virtual grayscale_image * GetEdges2(long frame, long view);

  virtual vHandPose GetPose2(long frame, long view);

  virtual vHandPose GetPose(long frame);

  virtual long GetPoseFrame(vHandPose pose, long * framep, long * viewp);

  virtual long FastPoseFrame(vHandPose & pose, long * framep, long * viewp);
};


// This class implements the second version of the database. Here we 
// normalize images based on the center and radius of their minimum
// enclosing circle. This way we only need to save on disk one image
// per viewpoint.
class vOldDatabase2 : public vOldDatabase
{
private:
  // These should probably be members of the superclass.
  double view_density;

  // The camera x_angle and y_angle corresponding to a given viewpoint index.
  vArray(double) x_angles;
  vArray(double) y_angles;
  // The image plane rotation corresponding to a given view index.
  vArray(double) orientations;

  vArray(v3dMatrix<short> *) rotation_templates;


protected:
  virtual void Zero();

public:
  vOldDatabase2();
  virtual ~vOldDatabase2();
  vOldDatabase2(const char * in_name);

  // Returns image indexed by given shape id and 3D orientation
  virtual color_image * PoseImage(vHandPose & pose);

  // Returns image indexed by shape_id, viewpoint_index, orientation_index
  // The caller is assumed to know what actual shapes and 3D orientations
  // these indices correspond to.
  virtual color_image * GetImage3(long shape_id, long viewpoint_index, 
                                  long orientation_index);

  virtual color_image * GetImage2(long frame, long view);

  // Returns edge image indexed by given shape id and 3D orientation
  virtual grayscale_image * PoseEdges(vHandPose & pose);

  // Returns edge image indexed by shape_id, viewpoint_index, orientation_index
  // The caller is assumed to know what actual shapes and 3D orientations
  // these indices correspond to.
  virtual grayscale_image * GetEdges3(long shape_id, long viewpoint_index, 
                                  long orientation_index);

  virtual grayscale_image * GetEdges2(long frame, long view);

  virtual grayscale_image * RotateImage(grayscale_image * image, long view);

  virtual color_image * RotateImage(color_image * image, long view);

  virtual vHandPose GetPose2(long frame, long view);

  virtual vHandPose GetPose(long frame);

  virtual long GetPoseFrame(vHandPose pose, long * framep, long * viewp);

  virtual long FastPoseFrame(vHandPose & pose, long * framep, long * viewp);

  inline const vArray(double) X_Angles()
  {
    return x_angles;
  }
  
  inline const vArray(double) Y_Angles()
  {
    return y_angles;
  }
  
  inline const vArray(double) Orientations()
  {
    return orientations;
  }
  
  static vOldDatabase2 * Load(const char * in_name);
  static vOldDatabase2 * Load(vRLE_Sequence * in_sequence);
  long Save();
};


// This class is a simplified version of the class vPointSetSequence,
// defined in fingers.h and fingers.cpp. It is used as an interface 
// to access directly the set of 
// points in the sequence of database edge images. We can load up
// the set of points in all edge images in memory, and access
// the set of points of each image separately.
// We don't allow point coordinates to exceed 255, so that they
// can fit into 8 bits.
// The reason why I am not using vPointSetSequence here and I define
// a new class is that vPointSetSequence was too specialized and 
// it was doing things that are not needed here. Hopefully when I rewrite
// these parts of the code, I will be able to erase old code that I don't 
// use anymore.
class vPointSequence
{
private:
  long rows, cols;
  long number_of_frames;
  long total_length;
  vArray(long) lengths;
  vArray(long) indices;
  vArray(uchar) point_rows;
  vArray(uchar) point_cols;

  inline void CheckFrame(long frame)
  {
    if ((frame < 0) || (frame >= number_of_frames))
    {
      exit_error("PointSetSequence error\n");
    }
  }

public:
  vPointSequence() {}
  vPointSequence(vVideoFile * sequence);
#ifdef VASSILIS_USE_vArray
  vPointSequence(grayscale_image ** images, long number);
#endif
  vPointSequence(class_pointer(grayscale_image *) images, long number);
  ~vPointSequence();

  static vPointSequence * Load(const char * filename);
  long Save(const char * filename);

  inline long Rows()
  {
    return rows;
  }

  inline long Cols()
  {
    return cols;
  }

  inline long NumberOfFrames()
  {
    return number_of_frames;
  }

  inline long TotalLength()
  {
    return total_length;
  }

  inline long Length(long frame)
  {
    CheckFrame(frame);
    return lengths[frame];
  }

  inline vArray(uchar) PointRows(long frame)
  {
    CheckFrame(frame);
    long index = indices[frame];
    return point_rows + index;
  }

  inline vArray(uchar) PointCols(long frame)
  {
    CheckFrame(frame);
    long index = indices[frame];
    return point_cols + index;
  }

  // returns the number of points.
  // In rowsp and colsp we put arrays of the (row, col)
  // coordinates of the set of points corresponding
  // to the given frame and given view
  long GetPoints(long frame, long view, 
                 vArray(long) * rowsp, vArray(long) * colsp);

  // This function is used mainly for debugging, to visualise the points
  // stored for each frame.
  grayscale_image * PointImage(long frame);
};

typedef vPointSequence class_pixel_sequence; 


// class vChamferFeatures is used to represent the set of features
// (point set and distance transform) for a sequence of images. That
// sequence of images could represent different image plane rotations
// of the query image, or different poses in the database.
class vChamferFeatures
{
public:
  vPointSequence * point_sequence;
  grayscale_image ** distance_transforms;

  vChamferFeatures()
  {
    point_sequence = 0;
    distance_transforms = 0;
  }

  vChamferFeatures(vPointSequence * in_point_sequence,
                   grayscale_image ** in_distance_transforms)
  {
    point_sequence = in_point_sequence;
    distance_transforms = in_distance_transforms;
  }

  // The destructor does nothing. Call EraseData to actually erase
  // the point sequence and the distance transforms. The reason
  // the destructor does nothing is that we may possibly want
  // to reuse those features.
  ~vChamferFeatures()
  {
  }

  long EraseData();
};

typedef vChamferFeatures class_chamfer_features;


// vDatabasePose is a class that is useful for representing
// a single database pose, as a (frame, rotation_index) pair.
class vDatabasePose
{
public:
  long frame;
  long rotation;

  vDatabasePose();
  vDatabasePose(long in_frame, long in_rotation)
  {
    frame = in_frame;
    rotation = in_rotation;
  }

  ~vDatabasePose()
  {
  }
};


// vPoseSet is used to specify a set of database poses. A 
// database pose is a pair, (frame, rotation_index), represented
// using the class vDatabasePose
class vPoseSet : public vUnique
{
protected:
  virtual void DeleteUnique()
  {
    vdelete(poses);
  }

public:
  vector<vDatabasePose> * poses;

  vPoseSet()
  {
    poses = new vector<vDatabasePose>;
  }

  ~vPoseSet()
  {
    DecrementReferences();
  }

  inline vDatabasePose & Pose(long index)
  {
    return (*poses)[index];
  }

  inline void AddPose(vDatabasePose & pose)
  {
    poses->push_back(pose);
  }

  inline long Size()
  {
    return poses->size();
  }
};


// This class will be used as an interface to the first database
// created by poser, which is a regeneration of the old database
// that I had produced using VirtualHand, with 26 shapes, 86
// viewpoints and 48 image plane rotations (the image plane rotations
// are not saved to disk here, they were saved to disk for the 
// VirtualHand database). The class should be written in 
// a general way, so that it allows different numbers of shapes,
// viewpoints and orientations. I'm not sure, as of 7/25/2003, if
// it should also handle transitions between shapes or not.
class vHandDatabase1
{
protected:
  // index1 is index link next to palm, index3 is outer index link 
  // Maybe these variables should be static, but I don't like
  // initializing static variables outside of all functions.
  long background_id, forearm_id, palm_id;
  long thumb1_id, thumb2_id, thumb3_id;
  long index1_id, index2_id, index3_id;
  long middle1_id, middle2_id, middle3_id;
  long ring1_id, ring2_id, ring3_id;
  long pinky1_id, pinky2_id, pinky3_id;
  long number_of_ids;
  
  char * name;
  long is_valid;
  // adjacencies_matrix(i,j) will be 1 if links i and j
  // are neighboring links in the hand, 0 otherwise.
  vMatrix<ushort> adjacencies_matrix;
  vArray2(ushort) adjacencies;

  // The color map specifies the colors assigned to each link.
  // Again, maybe this should be static.
  vColorMap * color_map;

  // For most of the image sequences that represent different features
  // of the database poses, there is an unnormalized (big) version and
  // a normalized version. The same scaling and translation has to
  // be applied to all those sequences, to ensure that the normalized
  // versions are aligned with each other. The matrix bounding_boxes
  // contains the bounding box info for the original hand image (with
  // colored links) of each database pose. Based on that bounding box,
  // we can deduce what transformation needs to be applied to each image.
  v3dMatrix<long> * bounding_boxes;

  // Various sequences that are useful for extracting features from
  // database poses. Every sequence comes in two versions: the big
  // version, in the original resolution in which it was saved by poser,
  // and a normalized version, where the hand has been scaled to a canonical
  // size and the forearm is frequently removed.
  
  // image_sequence: color-coded links
  vRLE_Sequence * image_sequence;
  vRLE_Sequence * big_image_sequence;

  // definite_edges: edge pixels are non-zero, id = intensity % 70
  vRLE_Sequence * definite_edges;
  vRLE_Sequence * big_definite_edges;

  // possible_edges: edge pixels are non-zero, id = intensity % 70
  vRLE_Sequence * possible_edges;
  vRLE_Sequence * big_possible_edges;

  // edges from cartoon images (generated by Poser).
  vRLE_Sequence * cartoon_sequence;
  vRLE_Sequence * big_cartoon_sequence;

  // edges that resemble the edges of the old database, that was created
  // using VirtualHand.
  vRLE_Sequence * old_edges;
  vRLE_Sequence * big_old_edges;

  long number_of_shapes;
  // number of viewpoints (camera locations) 
  long number_of_viewpoints;
  long number_of_orientations;

  // Number of frames stored in the sequence.
  long number_of_frames;
  // Number of frames stored per shape.
  long frames_per_shape;
  // The length of the side of the bounding square for
  // the normalized images.
  long bounding_length;
  // The rows and cols of the normalized images
  long image_size;

  // This table will hold distance transforms for 
  vStringTable distance_transforms;
  vStringTable point_sequences;
  
  grayscale_image ** current_dt_sequence;
  vPointSequence * current_point_sequence;

  vArray(v3dMatrix<short> *) rotation_templates;

protected:
  // Assign default values to member variables.
  void Initialize();
  
  // This function takes in an ending and returns the full path
  // (i.e., adds to the beginning the directory and the name
  // of the database). It is used for filenames refering to
  // sequences where the size of the hand has been normalized.
  char * FilenameWithEnding(const char * ending);
  // Same as FilenameWithEnding, but for sequences that are not
  // normalized for the size of the hand.
  char * BigFilenameWithEnding(const char * ending);

  // This function takes as input an ending specifying
  // a particular edge image sequence. It returns the full
  // path name of a file that is supposed to store the distance
  // transforms for that edge image sequence.
  char * DistanceTransformFilename(const char * ending);

  // This function takes as input an ending specifying
  // a particular edge image sequence. It returns the full
  // path name of a file that is supposed to store the point
  // sequence for that edge image sequence.
  char * PointSequenceFilename(const char * ending);

public:
  vHandDatabase1();
  vHandDatabase1(const char * name);
  ~vHandDatabase1();

  // Returns image indexed by shape_id, viewpoint_index, orientation_index
  // The caller is assumed to know what actual shapes and 3D orientations
  // these indices correspond to.
  color_image * GetImage3(long shape_id, long viewpoint_index, 
                              long orientation_index);

  color_image * GetImage2(long frame, long orientation);

  // Returns edge image indexed by shape_id, viewpoint_index, orientation_index
  // The caller is assumed to know what actual shapes and 3D orientations
  // these indices correspond to.
  grayscale_image * GetEdges3(long shape_id, long viewpoint_index, 
                          long orientation_index);

  grayscale_image * GetEdges2(long frame, long orientation);

  // Get the hand pose that corresponds to the given frame and view.
  //vHandPose GetPose2(long frame, long orientation);

  //vHandPose GetPose(long frame);

  // Add the standard database directory to a filename that
  // has no path information.
  char * AddDirectory(const char * filename);

  inline const char * Name()
  {
    return name;
  }

  inline long NumberOfShapes()
  {
    return number_of_shapes;
  }

  inline long NumberOfViewpoints()
  {
    return number_of_viewpoints;
  }

  inline long NumberOfOrientations()
  {
    return number_of_orientations;
  }

  inline long IsValid()
  {
    return is_valid;
  }

  inline long BoundingLength()
  {
    return bounding_length;
  }

  inline long NumberOfFrames()
  {
    return number_of_frames;
  }
  
  // The center is useful for figuring out the center of normalized 
  // images.
  inline long Center()
  {
    return image_size / 2;
  }

  inline long ObjectColor(long index, long * r, long * g, long * b)
  {
    if (color_map == 0) 
    {
      exit_error("zero color map in HandDatabase1\n");
    }
    return color_map->Color(index, r, g, b);
  }

  // The link id of a pixel can be always obtained using this function
  inline long LinkID(uchar r)
  {
    return r % 70;
  }

  grayscale_image * LinkIDs(color_image * colored_link_image);

  inline char * ImageSequenceName()
  {
    return FilenameWithEnding(".rle");
  }
  inline char * BigImageSequenceName()
  {
    return BigFilenameWithEnding(".rle");
  }

  inline char * DefiniteEdgesName()
  {
    return FilenameWithEnding("_edges.rle");
  }
  inline char * BigDefiniteEdgesName()
  {
    return BigFilenameWithEnding("_edges.rle");
  }

  inline char * PossibleEdgesName()
  {
    return FilenameWithEnding("_edges.rle");
  }
  inline char * BigPossibleEdgesName()
  {
    return BigFilenameWithEnding("_edges.rle");
  }

  inline char * CartoonSequenceName()
  {
    return FilenameWithEnding("_cartoon.rle");
  }
  inline char * BigCartoonSequenceName()
  {
    return BigFilenameWithEnding("_cartoon.rle");
  }

  inline char * OldEdgesName()
  {
    return FilenameWithEnding("_old_edges.rle");
  }
  inline char * BigOldEdgesName()
  {
    return BigFilenameWithEnding("_old_edges.rle");
  }

  char * BoundingBoxesName()
  {
    return FilenameWithEnding("_bbox.bin");
  }

  // This is the colormap that is used to efficiently save images
  // in RLE sequences, so that each hand link has a different color. An almost
  // identical colormap is defined in hand_renderer.cpp, but that file is not
  // included in the interface project, so I replicate the function here. 
  // This function, unlike the one in hand_renderer.cpp, also handles
  // the forearm.
  static vColorMap * DefaultColorMap();

  // Again, a similar function (that, however, does not take care
  // of the forearm) is defined in hand_renderer.cpp.
  static vMatrix<ushort> FingerAdjacencies();

  // This function processes the original poser sequence (which
  // we assume already exists) and generates the image sequence
  // based on that. The original poser sequence is unnormalized,
  // and includes the forearm.
  long CreateImageSequence();

  // Removes the forearm from the image, assuming that every
  // pixel of the forearm has the preassigned forearm color.
  // It keeps a line of forearm pixels adjacent to the palm,
  // so that we can easily identify the boundary between palm
  // and wrist in the result image.
  color_image * RemoveForearm(color_image * colored_link_image);

  // We save the transformation that we applied to normalize the bounding
  // box of each frame in the original Poser sequence, so that we 
  // can apply the same transformation to any features extracted from 
  // that sequence (or its associated cartoon sequence).
  // The transformation is saved in matrix form, it is a matrix
  // with long entries, one band, 4 rows, and frames columns.
  long SaveBoundingBoxes();

  // Extracts noise-free edges from the hand image, which, we
  // assume, follows the color code that this database uses.
  // The problem with this function is that it considers borders
  // between neighboring links not to be edges. Sometimes those
  // borders represent discontinuities in 3D, and in those
  // cases they should be considered as edges.
  grayscale_image * ArticulatedEdges(color_image * hand_image);

  // This function takes in a frame number, and returns the 
  // edge image for the corresponding database pose, using
  // the original frame and the cartoon drawing of that frame.
  // By using both images, edges between neighboring links that
  // represent discontinuities in 3D can be detected with 
  // reasonable accuracy, since, in such edges, there is 
  // an edge pixel in the cartoon image.
  grayscale_image * EdgeImage(long frame);

  // Return the bounding box of the given frame, in the original 
  // poser sequence.
  vBoundingBox BoundingBox(long frame);

  // The goal of this function is to create a sequence of edge
  // images that is as similar as possible to the edge images
  // generated using VirtualHand.
  long RecreateOldEdgeSequence();  

  // This function processes the unnormalized cartoon edge
  // images and creates normalized versions of them, in which
  // the forearm is removed.
  long CreateCartoonSequence();

  // CreateDistanceTransform takes in as argument an ending, like the
  // endings we pass to FilenameWithEnding, that specify sequences that
  // store particular pose features. We assume that the normalized edge
  // sequence specified by that ending exists. This function creates
  // another sequence, that contains the distance transform of each
  // frame of the normalized edge sequence.
  long CreateDistanceTransform(const char * ending);

  // CreatePointSequence, like CreateDistanceTransform,
  // takes in as argument an ending, like the
  // endings we pass to FilenameWithEnding, that specify sequences that
  // store particular pose features. We assume that the normalized edge
  // sequence specified by that ending exists. This function simply extracts
  // the non-zero points (edge pixels) of each frame, and stores them.
  // This gives us a representation that, together with distance transforms,
  // can be used for efficient implementation of the chamfer distance,
  // in terms of both time and memory.
  long CreatePointSequence(const char * ending);

  // Loads the specified distance transform into distance transforms, 
  // and, if successful, sets it as the current distance transform.
  long LoadDistanceTransform(const char * ending);

  // Sets the specified distance transform sequence as the current one.
  long SetCurrentDistanceTransform(const char * ending);

  // Removes the specified distance transform from memory.
  long ClearDistanceTransform(const char * ending);

  // Removes all distance transforms from memory.
  long ClearAllDistanceTransforms();

  // Returns the i-th frame of the current DT sequence.
  // Used mostly for debugging.
  grayscale_image * GetDistanceTransform(long frame);

  // Loads the specified point_sequence into point_sequences, 
  // and, if successful, sets it as the current point_sequence.
  long LoadPointSequence(const char * ending);

  // Sets the specified point sequenceas current, if it is 
  // already loaded.
  long SetCurrentPointSequence(const char * ending);

  // Removes the specified distance transform from memory
  long ClearPointSequence(const char * ending);

  // Removes all point sequences from memory
  long ClearAllPointSequences();

  // Returns the i-th frame of the current point sequence.
  // Used mostly for debugging.
  grayscale_image * GetPointImage(long frame);

  // Create or delete the rotation templates that can be used to 
  // generate different image plane rotations of a query image.
  long CreateRotationTemplates();
  long DeleteRotationTemplates();

  // For debugging:
  v3dMatrix<short> * RotationTemplate(long rotation_index);

  // Prepare the chamfer featuers of a query input image (that
  // is assumed to be an edge image). Those features
  // can be used to efficiently calculate chamfer distances.
  vChamferFeatures ChamferFeatures(grayscale_image * query_image);
  inline vChamferFeatures DatabaseChamferFeatures()
  {
    return vChamferFeatures(current_point_sequence, 
                             current_dt_sequence);
  }

  // Generate the set of all valid database poses, in order 
  // of increasing frames, and increasing orientations between
  // poses of the same frame.
  static vPoseSet AllPoses(long frames, long orientations);

  // returns a sequence saved in a file ending with "_real1.avi".
  // that sequence should contain realistic images of hands.
  vVideoFile * RealImages();
};


vMatrix<double> vLoadViewpoints(const char * filename);

long vSaveCameraParams(const char * filename, 
                        vMatrix<double> camera_params);


#endif // VASSILIS_HAND_DATABASE_H
