#include "stdafx.h"

#include <math.h>

#include "hausdorff_matching.h"
#include "basics/drawing_temp.h"
#include "opencv_utils.h"
#include "algorithms.h"
#include "basics/image_algorithms.h"
#include "basics/image_templates.h"
#include "algo_templates.h"
#include "fingers.h"
#include "basics/drawing.h"

#include "basics/definitions.h"





// This function computes the similarity between edge_image and
// some database views. The database edge images are stored in
// training_filename, and the database views that should be 
// compared to the test image are the highest ranking frames
// in database_frames. number_of_frames specifies how many
// of those frames should be used.
vMatchScores * vChamferRanks(grayscale_image * test_gray, 
                               const char * training_filename, 
                               vMatchScores * database_frames,
                               long number_of_frames)
{
  vVideoFile * training_sequence = vOpenReadVideoFile(training_filename);
  if (training_sequence == 0) return 0;
  if (strcmp(training_sequence->Type(), "vRLE_Sequence") != 0)
  {
    vCloseVideoFile(training_sequence);
    return 0;
  }
  v3dMatrix<long> training_frames_matrix(1, number_of_frames, 1);
  v3dMatrix<long> training_views_matrix(1, number_of_frames, 1);
  vArray(long) training_frames = training_frames_matrix.Matrix();
  vArray(long) training_views= training_views_matrix.Matrix();
  vArray(vMatchScore) frame_sorted = database_frames->GetFrameSorted();
  long i;
  for (i = 0; i < number_of_frames; i++)
  {
    training_frames[i] = frame_sorted[i].frame;
    training_views[i] = frame_sorted[i].view;
  }

  vMatrix<double> distances = 
    vDatabasePicksChamfer(test_gray, (vRLE_Sequence *) training_sequence, 
                           &training_frames_matrix, &training_views_matrix, 0, 
                           number_of_frames, -1, -1);

  vCloseVideoFile(training_sequence);
  vArray(vMatchScore) scores = vnew(vMatchScore, number_of_frames);
  for (i = 0; i < number_of_frames; i++)
  {
    scores[i].frame = training_frames[i];
    scores[i].view = training_views[i];
    scores[i].score = (float) distances(i);
    scores[i].is_valid = 1;
  }

  vMatchScores * result = new vMatchScores(scores, number_of_frames);
  return result;
}


vMatchScores * vOneWayChamferRanks(grayscale_image * test_gray, 
                                     const char * training_filename)
{
  const long number_of_views = 4;
  vVideoFile * training_sequence = vOpenReadVideoFile(training_filename);
  if (training_sequence == 0) return 0;
  if (strcmp(training_sequence->Type(), "vRLE_Sequence") != 0)
  {
    vCloseVideoFile(training_sequence);
    return 0;
  }

  long number_of_frames = training_sequence->NumberOfFrames();
  long number_of_scores = number_of_frames * number_of_views;
  v3dMatrix<long> training_frames_matrix(1, number_of_scores, 1);
  v3dMatrix<long> training_views_matrix(1, number_of_scores, 1);
  vArray(long) training_frames = training_frames_matrix.Matrix();
  vArray(long) training_views= training_views_matrix.Matrix();
  long i;
  for (i = 0; i < number_of_frames; i++)
  {
    long index = number_of_views * i;
    long j;
    for (j = 0; j < number_of_views; j++)
    {
      training_frames[index] = i;
      training_views[index] = j;
      index++;
    }
  }

  v3dMatrix<float> * dt = vCVDistanceTransform(test_gray);

  vMatrix<double> distances = 
    vOneWayChamfer(dt, (vRLE_Sequence *) training_sequence, 
                   &training_frames_matrix, &training_views_matrix);

  vCloseVideoFile(training_sequence);
  vdelete(dt);
  vArray(vMatchScore) scores = vnew(vMatchScore, number_of_scores);
  for (i = 0; i < number_of_scores; i++)
  {
    scores[i].frame = training_frames[i];
    scores[i].view = training_views[i];
    scores[i].score = (float) distances(i);
    scores[i].is_valid = 1;
  }

  vMatchScores * result = new vMatchScores(scores, number_of_scores);
  return result;
}


// Computes the directed chamfer distance from each database image
// to test_gray.
vMatchScores * vOneWayChamferRanks2(grayscale_image * test_gray, 
                                      vPointSetSequence * database)
{
  const long number_of_views = 4;

  long number_of_frames = database->NumberOfFrames();
  long number_of_scores = number_of_frames * number_of_views;
  v3dMatrix<long> training_frames_matrix(1, number_of_scores, 1);
  v3dMatrix<long> training_views_matrix(1, number_of_scores, 1);
  vArray(long) training_frames = training_frames_matrix.Matrix();
  vArray(long) training_views= training_views_matrix.Matrix();
  long i;
  for (i = 0; i < number_of_frames; i++)
  {
    long index = number_of_views * i;
    long j;
    for (j = 0; j < number_of_views; j++)
    {
      training_frames[index] = i;
      training_views[index] = j;
      index++;
    }
  }

  v3dMatrix<float> * dt = vCVDistanceTransform(test_gray);

  vMatrix<double> distances = 
    vOneWayChamfer2(dt, database, 
                    &training_frames_matrix, &training_views_matrix);

  vdelete(dt);
  vArray(vMatchScore) scores = vnew(vMatchScore, number_of_scores);
  for (i = 0; i < number_of_scores; i++)
  {
    scores[i].frame = training_frames[i];
    scores[i].view = training_views[i];
    scores[i].score = (float) distances(i);
    scores[i].is_valid = 1;
  }

  vMatchScores * result = new vMatchScores(scores, number_of_scores);
  return result;
}



// This function computes the similarity between edge_image and
// some database views. The database edge images are stored in
// training_filename, and the database views that should be 
// compared to the test image are the highest ranking frames
// in database_frames. number_of_frames specifies how many
// of those frames should be used.
vMatchScores * vChamferRanks2(grayscale_image * test_gray, 
                                const char * training_filename, 
                                vMatchScores * database_frames,
                                long number_of_frames)
{
  // Declare and initialize certain useful variables.
  ushort success = 0;
  long score = 0;
  vVideoFile * training_sequence = 0;
  grayscale_image * binary0, * binary1, * binary2, * binary3;
  binary1 = binary2 = binary3 = binary0 = 0;
  grayscale_image * training_gray = 0;

  long training_frame = 0, score_frame = 0;
  long number_of_training_frames = 0;
  long scores_frames = 0;
  long orientation = 0;

  // Prepare test image before matching. We convert test_gray
  // to a normalized binary. We then rotate it
  // by 90, 180 and 270 degrees. We compare each of those 
  // views to each of the test frames.
  binary0 = (grayscale_image *) vScaleBoundingBox6(test_gray, 192, 128, 128, 
                                                256, 256);
  binary1 = (grayscale_image *) vRotateLeft(binary0);
  binary2 = (grayscale_image *) vRotateLeft(binary1);
  binary3 = (grayscale_image *) vRotateLeft(binary2);
  grayscale_image * test_binaries [] = {binary0, binary1, binary2, binary3};
  vArray(vMatchScore) chamfer_scores = vnew(vMatchScore, number_of_frames + 1);

  // Open the database edge file.
  training_sequence = vOpenReadVideoFile(training_filename);
  if (training_sequence == 0)
  {
    vPrint("Can't open training sequence %s\n", training_filename);
    goto exit;
  }

  number_of_training_frames = training_sequence->NumberOfFrames();
  vPrint("%15s\n", "training_frame");
  scores_frames = database_frames->NumberOfScores();
  if (number_of_frames > scores_frames)
  {
    number_of_frames = scores_frames;
  }

  // Do the matching
  for (score_frame = 0; score_frame < number_of_frames; 
       score_frame++)
  {
    // Get the frame info for the score
    vMatchScore score = database_frames->GetScore(score_frame);

    training_frame = score.frame;
    long training_view = score.view;

    // Read in the training frame.
    training_gray = training_sequence->ReadFrameGray(training_frame);
    if (training_gray == 0) goto exit;
    grayscale_image * training_binary = (grayscale_image *) training_gray->threshold(1);
    vdelete(training_gray);
    training_gray = 0;
    
    // Compare training frame to each of the four test views
    score.score = (float) vCompareContoursFast(test_binaries[training_view], 
                                                training_binary, 10);
    chamfer_scores[score_frame] = score;
    vdelete(training_binary);
    training_binary = 0;
    vPrint("%15li               \r", score_frame);
  }

  success = 1;
  vPrint("\n");

exit:
  vCloseVideoFile(training_sequence);
  vdelete(training_gray);
  vdelete(binary0);
  vdelete(binary1);
  vdelete(binary2);
  vdelete(binary3);
  if (success == 0)
  {
    vdelete2(chamfer_scores);
    return 0;
  }
  else
  {
    vMatchScores * result = new vMatchScores(chamfer_scores, 
                                               number_of_frames);
    return result;
  }
}



// Compute the flexible (allowing translation and scaling) chamfer distances
// between edge_image and some database views. The database views that
// we should check are specified in the frames_to_check objects,
// they are the first number_of_frames images in the score_sorted array
vMatchScores * vFlexibleChamferRanks(grayscale_image * edge_image, 
                                       vVideoFile * database,
                                       vMatchScores * frames_to_check,
                                       long number_of_frames)
{
  // Declare and initialize certain useful variables.
  long score = 0;
  grayscale_image * binary0, * binary1, * binary2, * binary3;
  binary1 = binary2 = binary3 = binary0 = 0;
  grayscale_image * training_gray;
  training_gray = 0;

  long training_frame;
  training_frame = 0;
  long orientation = 0;
  vMatchScores * scores = 0;

  //number_of_training_frames = 3;
  vPrint("%15s\n", "training_frame");

  // Prepare test image before matching. We read a gray-scale
  // frame, and we convert it to a normalized binary. We then rotate it
  // by 90, 180 and 270 degrees. We compare each of those 
  // views to each of the test frames.
  //    binary0 = (grayscale_image *) vNormalizedBinary(test_gray, 96.0, 256, 256);
  binary0 = (grayscale_image *) vScaleBoundingBox6(edge_image, 192, 128, 128, 
                                                 256, 256);
  binary1 = (grayscale_image *) vRotateLeft(binary0);
  binary2 = (grayscale_image *) vRotateLeft(binary1);
  binary3 = (grayscale_image *) vRotateLeft(binary2);
  grayscale_image * test_binaries [] = {binary0, binary1, binary2, binary3};
   // binary0->Save("D:\\vassilis\\mine\\sequences\\binary0.pgm");

  if (strcmp(database->Type(), "vRLE_Sequence") != 0)
  {
    vCloseVideoFile(database);
    return 0;
  }
  v3dMatrix<long> training_frames_matrix(1, number_of_frames, 1);
  v3dMatrix<long> training_views_matrix(1, number_of_frames, 1);
  vArray(long) training_frames = training_frames_matrix.Matrix();
  vArray(long) training_views= training_views_matrix.Matrix();
  vArray(vMatchScore) frame_sorted = frames_to_check->GetFrameSorted();

  vArray(vMatchScore) frame_scores = vnew(vMatchScore, number_of_frames + 1);
 
  long score_frame = 0;
  // Do the matching
  for (score_frame = 0; score_frame < number_of_frames; score_frame++)
  {
    vMatchScore score = frame_sorted[score_frame];
    long training_frame = score.frame;
    long training_view = score.view;
    // Read in the training frame.
    training_gray = database->ReadFrameGray(training_frame);
    if (training_gray == 0) goto exit;
    grayscale_image * training_binary = (grayscale_image *) training_gray->threshold(1);
    vdelete(training_gray);
    training_gray = 0;
    
    // Compare training frame to each of the four test views
    double scaling;
    long translation_rows, translation_cols;
    scaling = translation_rows = translation_cols = 0;
    double distance = 0;
    vChamferDistances(test_binaries[training_view], training_binary, &distance, 
                         &translation_rows, &translation_cols, &scaling);
    score.score = (float) distance;
    frame_scores[score_frame] = score;
    vdelete(training_binary);
    training_binary = 0;
    vPrint("%15li              \r", training_frame);
  }

  vPrint("\n");
  scores = new vMatchScores(frame_scores, number_of_frames);

exit:
  vdelete(training_gray);
  vdelete(binary0);
  vdelete(binary1);
  vdelete(binary2);
  vdelete(binary3);
  return scores;
}


// test_gray is supposed to be a binary image of feature points (edges 
// in my 
// current application). database is a sequence of model images (also
// represented as images of feature points). If the database has K
// images, the result will have 4*K values, each of them being the chamfer
// distance between test_gray and a model image. For each image in the
// database we actually use 4 images, the results of rotating that image
// by 0, 90, 180 and 270 degrees.
vMatrix<double> 
vDatabaseChamfer(grayscale_image * test_gray, vRLE_Sequence * database,
                  long grid_size, long subsampling, long training_frames)
{
  // Declare and initialize certain useful variables.
  long score = 0;
  long i = 0;
  grayscale_image * binary0, * binary1, * binary2, * binary3;
  v3dMatrix<float> * dt_0, * dt_1, * dt_2, * dt_3;
  binary1 = binary2 = binary3 = binary0 = 0;
  dt_0 = dt_1 = dt_2 = dt_3 = 0;
  grayscale_image * training_gray = 0;
  double scores[4];

  long training_frame = 0;
  long number_of_training_frames = 0;
  long orientation = 0;
  vArray(long) rows = vZero(long);
  vArray(long) cols = vZero(long);
  long training_rows, training_cols, training_pixels;
  training_rows = training_cols = training_pixels = 0;

  number_of_training_frames = database->NumberOfFrames();
  training_rows = database->Rows();
  training_cols = database->Cols();
  training_pixels = training_rows * training_cols;
  rows = vnew(long, training_pixels);
  cols = vnew(long, training_pixels);
  vMatrix<double> result;
  vArray(double) result_matrix = vZero(double);
  long result_index = 0;

  // Set grid_size, subsampling and training_frames to default values 
  // if they are -1, reject them if they have bad values.
  if (grid_size == -1) grid_size = 10;
  if ((grid_size < 4) || (grid_size > 256)) goto exit;
  if (subsampling < 0) subsampling = 1;
  if (training_frames == -1) training_frames = number_of_training_frames;
  if ((training_frames < 0) || 
      (training_frames > number_of_training_frames)) goto exit;

  result = vMatrix<double>(1, training_frames*4);
  result_matrix = result.Matrix();

  // The following part of the code is enclosed in braces so that I don't get 
  // the stupid initialization warning about "initialization skipped by
  // jumping to exit".
  {
    vArray(long) binary0_rows;
    vArray(long) binary0_cols;
    vArray(long) binary1_rows;
    vArray(long) binary1_cols;
    vArray(long) binary2_rows;
    vArray(long) binary2_cols;
    vArray(long) binary3_rows;
    vArray(long) binary3_cols;

    binary0 = (grayscale_image *) vScaleBoundingBox6(test_gray, 192, 128, 128, 
                                                  256, 256);
    binary1 = (grayscale_image *) vRotateLeft(binary0);
    binary2 = (grayscale_image *) vRotateLeft(binary1);
    binary3 = (grayscale_image *) vRotateLeft(binary2);

    dt_0 = vCVDistanceTransform(binary0);
    dt_1 = vCVDistanceTransform(binary1);
    dt_2 = vCVDistanceTransform(binary2);
    dt_3 = vCVDistanceTransform(binary3);

    long binary0_number, binary1_number, binary2_number, binary3_number;
    binary0_number = binary1_number = binary2_number = binary3_number = 0;
    vGetNonZeroPoints4(binary0, &binary0_rows, &binary0_cols, &binary0_number);
    vGetNonZeroPoints4(binary1, &binary1_rows, &binary1_cols, &binary1_number);
    vGetNonZeroPoints4(binary2, &binary2_rows, &binary2_cols, &binary2_number);
    vGetNonZeroPoints4(binary3, &binary3_rows, &binary3_cols, &binary3_number);

    grayscale_image * binaries[4] = {binary0, binary1, binary2, binary3};
    v3dMatrix<float> * dts[4] = {dt_0, dt_1, dt_2, dt_3};
    vArray(long) binary_rows[4] = {binary0_rows, binary1_rows, 
                             binary2_rows, binary3_rows};
    vArray(long) binary_cols[4] = {binary0_cols, binary1_cols, 
                             binary2_cols, binary3_cols};
    long binary_numbers[4] = {binary0_number, binary1_number, 
                              binary2_number, binary3_number};

    // Do the matching
    for (training_frame = 0; training_frame < training_frames; 
         training_frame++)
    {
      // Read in the training frame.
      vRLE_Gray * training_rle = (vRLE_Gray *) database->ReadRLEFrame(training_frame);
      if (training_rle == 0) goto exit;
      long number = 0;
      vRLENonZeroPoints(training_rle, rows, cols, &number);

      for (i = 0; i < 4; i++)
      {
        double distance1 = vOptimizedChamfer2(binary_rows[i], binary_cols[i], 
                                                 binary_numbers[i], rows, cols,
                                                 number, grid_size, subsampling);
        double distance2 = vChamferFromDT(rows, cols, number, dts[i]);
        double distance = distance1 + distance2;
        scores[i] = distance;
        result_matrix[result_index++] = distance;

      }
      vdelete(training_rle);
      training_rle = 0;
      vPrint("%15li               \r", training_frame);
    }

    vdelete2(binary0_rows);
    vdelete2(binary0_cols);
    vdelete2(binary1_rows);
    vdelete2(binary1_cols);
    vdelete2(binary2_rows);
    vdelete2(binary2_cols);
    vdelete2(binary3_rows);
    vdelete2(binary3_cols);
  }

exit:
  vPrint("\n");

  vdelete(training_gray);
  vdelete(binary0);
  vdelete(binary1);
  vdelete(binary2);
  vdelete(binary3);
  vdelete(dt_0);
  vdelete(dt_1);
  vdelete(dt_2);
  vdelete(dt_3);
  vdelete2(rows);
  vdelete2(cols);
  return result;
}


// test_gray is supposed to be a binary image of feature points (edges 
// in my 
// current application). database is a sequence of model images (also
// represented as images of feature points). frames and views specify
// the database frames that we want to compare test_gray with, and 
// the orientation from which we want to compare them (there are four
// possible orientations, corresponding to rotating the training image
// to the right by 0, 90, 180 or 270 degrees).
vMatrix<double> 
vDatabasePicksChamfer(grayscale_image * test_gray, vPointSetSequence * database,
                       v3dMatrix<long> * frames, v3dMatrix<long> * views,
                       long start, long dimensions,
                       long grid_size, long subsampling)
{
  // Declare and initialize certain useful variables.
  long score = 0;
  long i = 0;
  grayscale_image * binary = 0;
  v3dMatrix<float> * dt = 0;
  grayscale_image * training_gray = 0;

  vArray(long) rows = vZero(long);
  vArray(long) cols = vZero(long);
  long training_rows, training_cols, training_pixels;
  training_rows = training_cols = training_pixels = 0;

  long number_of_training_frames = database->NumberOfFrames();
  long number_of_picks = frames->Size();
  if (number_of_picks != views->Size()) return vMatrix<double>();
  if ((start < 0) || (dimensions < 0)) return vMatrix<double>();
  if (start+dimensions > number_of_picks) return vMatrix<double>();

  training_rows = database->Rows();
  training_cols = database->Cols();
  training_pixels = training_rows * training_cols;
  vMatrix<double> result;
  vArray(double) result_matrix = vZero(double);
  long result_index = 0;

  // Set grid_size, subsampling and training_frames to default values 
  // if they are -1, reject them if they have bad values.
  if (grid_size == -1) grid_size = 10;
  if ((grid_size < 4) || (grid_size > 256)) return vMatrix<double>();
  if (subsampling < 0) subsampling = 1;


  result = vMatrix<double>(1, dimensions);
  result_matrix = result.Matrix();

  vArray(long) binary_rows;
  vArray(long) binary_cols;
  binary = (grayscale_image *) vScaleBoundingBox6(test_gray, 192, 128, 128, 
                                               256, 256);

  dt = vCVDistanceTransform(binary);

  long binary_number = 0;
  vGetNonZeroPoints4(binary, &binary_rows, &binary_cols, &binary_number);

  // Do the matching
  long error_flag = 0;
  vArray(long) picked_frames = frames->Matrix();
  vArray(long) frame_views = views->Matrix();
  long pick = 0;
  long end_pick = start + dimensions - 1;
  vPrint("\n");
  for (pick = start; pick <= end_pick; pick++)
  {
    long training_frame = picked_frames[pick];
    long view = frame_views[pick];
    if ((training_frame < 0) || (training_frame > number_of_training_frames) ||
        (view < 0) || (view >= 4))
    {
      vPrint("Bad frame or view pick: %li %li\n", training_frame, view);
      error_flag = 1;
      goto exit;
    }
  
    database->GetPoints(training_frame, view, &rows, &cols);
    long number = database->Length(training_frame);
    double distance1 = vOptimizedChamfer2(binary_rows, binary_cols, 
                                             binary_number, rows, cols,
                                             number, grid_size, subsampling);
    double distance2 = vChamferFromDT(rows, cols, number, dt);
    double distance = distance1 + distance2;
    result_matrix[result_index++] = distance;

    vdelete2(rows);
    vdelete2(cols);
    vPrint("frame %li of %li, index = %li\r", 
            pick - start + 1, dimensions, training_frame);
  }

  vdelete2(binary_rows);
  vdelete2(binary_cols);

exit:
  vPrint("\n");

  vdelete(training_gray);
  vdelete(binary);
  vdelete(dt);
  return result;
}


vMatrix<double> 
vDatabasePicksChamfer(grayscale_image * test_gray, vRLE_Sequence * database,
                       v3dMatrix<long> * frames, v3dMatrix<long> * views,
                       long start, long dimensions,
                       long grid_size, long subsampling)
{
  // Declare and initialize certain useful variables.
  long score = 0;
  long i = 0;
  grayscale_image * binary = 0;
  v3dMatrix<float> * dt = 0;
  grayscale_image * training_gray = 0;

  vArray(long) rows = vZero(long);
  vArray(long) cols = vZero(long);
  long training_rows, training_cols, training_pixels;
  training_rows = training_cols = training_pixels = 0;

  long number_of_training_frames = database->NumberOfFrames();
  long number_of_picks = frames->Size();
  if (number_of_picks != views->Size()) return vMatrix<double>();
  if ((start < 0) || (dimensions < 0)) return vMatrix<double>();
  if (start+dimensions > number_of_picks) return vMatrix<double>();

  training_rows = database->Rows();
  training_cols = database->Cols();
  training_pixels = training_rows * training_cols;
  rows = vnew(long, training_pixels);
  cols = vnew(long, training_pixels);
  vMatrix<double> result;
  vArray(double) result_matrix = vZero(double);
  long result_index = 0;

  // Set grid_size, subsampling and training_frames to default values 
  // if they are -1, reject them if they have bad values.
  if (grid_size == -1) grid_size = 10;
  if ((grid_size < 4) || (grid_size > 256)) return vMatrix<double>();
  if (subsampling < 0) subsampling = 1;


  result = vMatrix<double>(1, dimensions);
  result_matrix = result.Matrix();

  vArray(long) binary_rows;
  vArray(long) binary_cols;
  binary = (grayscale_image *) vScaleBoundingBox6(test_gray, 192, 128, 128, 
                                               256, 256);

  dt = vCVDistanceTransform(binary);

  long binary_number = 0;
  vGetNonZeroPoints4(binary, &binary_rows, &binary_cols, &binary_number);

  // Do the matching
  long error_flag = 0;
  vArray(long) picked_frames = frames->Matrix();
  vArray(long) frame_views = views->Matrix();
  long pick = 0;
  long end_pick = start + dimensions - 1;
  vPrint("\n");
  for (pick = start; pick <= end_pick; pick++)
  {
    long training_frame = picked_frames[pick];
    long view = frame_views[pick];
    if ((training_frame < 0) || (training_frame > number_of_training_frames) ||
        (view < 0) || (view >= 4))
    {
      vPrint("Bad frame or view pick: %li %li\n", training_frame, view);
      error_flag = 1;
      goto exit;
    }
  
    // Read in the training frame.
    vRLE_Gray * training_rle = (vRLE_Gray *) database->ReadRLEFrame(training_frame);
    if (training_rle == 0) goto exit;
    long number = 0;
    vRLENonZeroPoints(training_rle, rows, cols, &number);
    vRotatePointsRight(rows, cols, rows, cols, number, training_rows, 
                        training_cols, 90 * view);

    double distance1 = vOptimizedChamfer2(binary_rows, binary_cols, 
                                             binary_number, rows, cols,
                                             number, grid_size, subsampling);
    double distance2 = vChamferFromDT(rows, cols, number, dt);
    double distance = distance1 + distance2;
    result_matrix[result_index++] = distance;

    vdelete(training_rle);
    training_rle = 0;
    vPrint("frame %li of %li, index = %li\r", 
            pick - start + 1, dimensions, training_frame);
  }

  vdelete2(binary_rows);
  vdelete2(binary_cols);

exit:
  vPrint("\n");

  vdelete(training_gray);
  vdelete(binary);
  vdelete(dt);
  vdelete2(rows);
  vdelete2(cols);
  return result;
}



vMatrix<double> 
vDatabasePicksChamfer(grayscale_image * test_gray, vRLE_Sequence * database,
                       v3dMatrix<float> * reference_images,
                       long start, long dimensions,
                       long grid_size, long subsampling)
{
  // Declare and initialize certain useful variables.
  long score = 0;
  long i = 0;
  grayscale_image * binary = 0;
  v3dMatrix<float> * dt = 0;
  grayscale_image * training_gray = 0;

  vArray(long) rows = vZero(long);
  vArray(long) cols = vZero(long);
  long training_rows, training_cols, training_pixels;
  training_rows = training_cols = training_pixels = 0;

  long number_of_training_frames = database->NumberOfFrames();
  long number_of_picks = reference_images->Rows();
  if ((start < 0) || (dimensions < 0)) return vMatrix<double>();
  if (start+dimensions > number_of_picks) return vMatrix<double>();

  training_rows = database->Rows();
  training_cols = database->Cols();
  training_pixels = training_rows * training_cols;
  rows = vnew(long, training_pixels);
  cols = vnew(long, training_pixels);
  vMatrix<double> result;
  vArray(double) result_matrix = vZero(double);
  long result_index = 0;

  // Set grid_size, subsampling and training_frames to default values 
  // if they are -1, reject them if they have bad values.
  if (grid_size == -1) grid_size = 10;
  if ((grid_size < 4) || (grid_size > 256)) return vMatrix<double>();
  if (subsampling < 0) subsampling = 1;


  result = vMatrix<double>(1, dimensions);
  result_matrix = result.Matrix();

  vArray(long) binary_rows;
  vArray(long) binary_cols;

  binary = (grayscale_image *) vScaleBoundingBox6(test_gray, 192, 128, 128, 
                                               256, 256);

  dt = vCVDistanceTransform(binary);

  long binary_number = 0;
  vGetNonZeroPoints4(binary, &binary_rows, &binary_cols, &binary_number);

  // Do the matching
  long error_flag = 0;
  vArray2(float) references = reference_images->Matrix2(0);
  long pick = 0;
  long end_pick = start + dimensions - 1;
  vPrint("\n");
  for (pick = start; pick <= end_pick; pick++)
  {
    long training_frame = vRound(references[pick][0]);
    long view = vRound(references[pick][1]);
    if ((training_frame < 0) || (training_frame > number_of_training_frames) ||
        (view < 0) || (view >= 4))
    {
      vPrint("Bad frame or view pick: %li %li\n", training_frame, view);
      error_flag = 1;
      goto exit;
    }
  
    // Read in the training frame.
    vRLE_Gray * training_rle = (vRLE_Gray *) database->ReadRLEFrame(training_frame);
    if (training_rle == 0) goto exit;
    long number = 0;
    vRLENonZeroPoints(training_rle, rows, cols, &number);
    vRotatePointsRight(rows, cols, rows, cols, number, training_rows, 
                        training_cols, 90 * view);

    double distance1 = vOptimizedChamfer2(binary_rows, binary_cols, 
                                             binary_number, rows, cols,
                                             number, grid_size, subsampling);
    double distance2 = vChamferFromDT(rows, cols, number, dt);
    double distance = distance1 + distance2;
    result_matrix[result_index++] = distance;

    vdelete(training_rle);
    training_rle = 0;
    vPrint("frame %li of %li, index = %li\r", 
            pick - start + 1, dimensions, training_frame);
  }

  vdelete2(binary_rows);
  vdelete2(binary_cols);

exit:
  vPrint("\n");

  vdelete(training_gray);
  vdelete(binary);
  vdelete(dt);
  vdelete2(rows);
  vdelete2(cols);
  return result;
}


// Here we only care about distances from image to model (not from model
// to image).
vMatrix<double> 
vDatabasePicksChamferIM(grayscale_image * test_gray, vPointSetSequence * database,
                        v3dMatrix<long> * frames, v3dMatrix<long> * views,
                        long start, long dimensions,
                        long grid_size, long subsampling)
{
  // Declare and initialize certain useful variables.
  long score = 0;
  long i = 0;
  grayscale_image * binary = 0;

  vArray(long) rows = vZero(long);
  vArray(long) cols = vZero(long);

  long number_of_training_frames = database->NumberOfFrames();
  long number_of_picks = frames->Size();
  if (number_of_picks != views->Size()) return vMatrix<double>();
  if ((start < 0) || (dimensions < 0)) return vMatrix<double>();
  if (start+dimensions > number_of_picks) return vMatrix<double>();
  vMatrix<double> result;
  vArray(double) result_matrix = vZero(double);
  long result_index = 0;

  // Set grid_size, subsampling and training_frames to default values 
  // if they are -1, reject them if they have bad values.
  if (grid_size == -1) grid_size = 10;
  if ((grid_size < 4) || (grid_size > 256)) return vMatrix<double>();
  if (subsampling < 0) subsampling = 1;


  result = vMatrix<double>(1, dimensions);
  result_matrix = result.Matrix();

  vArray(long) binary_rows;
  vArray(long) binary_cols;
  binary = (grayscale_image *) vScaleBoundingBox6(test_gray, 192, 128, 128, 
                                               256, 256);

  long binary_number = 0;
  vGetNonZeroPoints4(binary, &binary_rows, &binary_cols, &binary_number);

  // Do the matching
  long error_flag = 0;
  vArray(long) picked_frames = frames->Matrix();
  vArray(long) frame_views = views->Matrix();
  long pick = 0;
  long end_pick = start + dimensions - 1;
  for (pick = start; pick <= end_pick; pick++)
  {
    long training_frame = picked_frames[pick];
    long view = frame_views[pick];
    if ((training_frame < 0) || (training_frame > number_of_training_frames) ||
        (view < 0) || (view >= 4))
    {
      vPrint("Bad frame or view pick: %li %li\n", training_frame, view);
      error_flag = 1;
      goto exit;
    }
  
    database->GetPoints(training_frame, view, &rows, &cols);
    long number = database->Length(training_frame);
    double distance = vOptimizedChamfer2(binary_rows, binary_cols, 
                                            binary_number, rows, cols,
                                            number, grid_size, subsampling);
    result_matrix[result_index++] = distance;

    vdelete2(rows);
    vdelete2(cols);
    vPrint("%15li               \r", training_frame);
  }

  vdelete2(binary_rows);
  vdelete2(binary_cols);

exit:
  vPrint("\n");

  vdelete(binary);
  return result;
}


// Here we only care about distances from image to model (not from model
// to image).
vMatrix<double> 
vDatabasePicksChamferIM(grayscale_image * test_gray, vPointSetSequence * database,
                        v3dMatrix<float> * reference_images,
                        long start, long dimensions,
                        long grid_size, long subsampling)
{
  // Declare and initialize certain useful variables.
  long score = 0;
  long i = 0;
  grayscale_image * binary = 0;

  vArray(long) rows = vZero(long);
  vArray(long) cols = vZero(long);

  long number_of_training_frames = database->NumberOfFrames();
  long number_of_picks = reference_images->Rows();
  if ((start < 0) || (dimensions < 0)) return vMatrix<double>();
  if (start+dimensions > number_of_picks) return vMatrix<double>();
  vMatrix<double> result;
  vArray(double) result_matrix = vZero(double);
  long result_index = 0;

  // Set grid_size, subsampling and training_frames to default values 
  // if they are -1, reject them if they have bad values.
  if (grid_size == -1) grid_size = 10;
  if ((grid_size < 4) || (grid_size > 256)) return vMatrix<double>();
  if (subsampling < 0) subsampling = 1;

  result = vMatrix<double>(1, dimensions);
  result_matrix = result.Matrix();

  vArray(long) binary_rows;
  vArray(long) binary_cols;
  binary = (grayscale_image *) vScaleBoundingBox6(test_gray, 192, 128, 128, 
                                               256, 256);

  long binary_number = 0;
  vGetNonZeroPoints4(binary, &binary_rows, &binary_cols, &binary_number);

  // Do the matching
  long error_flag = 0;
  vArray2(float) references = reference_images->Matrix2(0);
  long pick = 0;
  long end_pick = start + dimensions - 1;
  for (pick = start; pick <= end_pick; pick++)
  {
    long training_frame = vRound(references[pick][0]);
    long view = vRound(references[pick][1]);
    float weight = references[pick][2];
    if ((training_frame < 0) || (training_frame > number_of_training_frames) ||
        (view < 0) || (view >= 4))
    {
      vPrint("Bad frame or view pick: %li %li\n", training_frame, view);
      error_flag = 1;
      goto exit;
    }
  
    database->GetPoints(training_frame, view, &rows, &cols);
    long number = database->Length(training_frame);
    double distance = vOptimizedChamfer2(binary_rows, binary_cols, 
                                            binary_number, rows, cols,
                                            number, grid_size, subsampling);
    result_matrix[result_index++] = weight * distance;

    vdelete2(rows);
    vdelete2(cols);
    vPrint("%15li               \r", training_frame);
  }

  vdelete2(binary_rows);
  vdelete2(binary_cols);

exit:
  vPrint("\n");

  vdelete(binary);
  return result;
}


// dt is the distance transform of the input image.
// database is a sequence of model images (also
// represented as images of feature points). frames and views specify
// the database frames that we want to compare test_gray with, and 
// the orientation from which we want to compare them (there are four
// possible orientations, corresponding to rotating the training image
// to the right by 0, 90, 180 or 270 degrees).
// We calculate the one-way chamfer distance from model to image.
vMatrix<double> 
vOneWayChamfer(v3dMatrix<float> * dt, vRLE_Sequence * database)
{
  long number_of_frames = database->NumberOfFrames();
  const long number_of_views = 4;
  long i, j;
  vMatrix<long> frames(1, number_of_frames * number_of_views);
  vMatrix<long> views(1, number_of_frames * number_of_views);
  for (i = 0; i < number_of_frames; i++)
  {
    for (j = 0; j < number_of_frames; j++)
    {
      long index = i*4+j;
      frames(index) = i;
      views(index) = j;
    }
  }

  return vOneWayChamfer(dt, database, &frames, &views);
}

                
vMatrix<double> 
vOneWayChamfer(v3dMatrix<float> * dt, vRLE_Sequence * database,
                v3dMatrix<long> * frames, v3dMatrix<long> * views)
{
  // Declare and initialize certain useful variables.
  long number_of_training_frames = database->NumberOfFrames();
  long number_of_picks = frames->Size();
  if (number_of_picks != views->Size()) return vMatrix<double>();

  long training_rows = database->Rows();
  long training_cols = database->Cols();
  long training_pixels = training_rows * training_cols;
  vArray(long) rows = vnew(long, training_pixels);
  vArray(long) cols = vnew(long, training_pixels);
  vMatrix<double> result;
  vArray(double) result_matrix = vZero(double);
  long result_index = 0;

  result = vMatrix<double>(1, number_of_picks);
  result_matrix = result.Matrix();

  // Do the matching
  long error_flag = 0;
  vArray(long) picked_frames = frames->Matrix();
  vArray(long) frame_views = views->Matrix();
  long pick;
  vPrint("\n");
  for (pick = 0; pick < number_of_picks; pick++)
  {
    long training_frame = picked_frames[pick];
    long view = frame_views[pick];
    if ((training_frame < 0) || (training_frame >= number_of_training_frames) ||
        (view < 0) || (view >= 4))
    {
      vPrint("Bad frame or view pick: %li %li\n", training_frame, view);
      error_flag = 1;
      break;
    }
  
    // Read in the training frame.
    vRLE_Gray * training_rle = (vRLE_Gray *) database->ReadRLEFrame(training_frame);
    if (training_rle == 0) break;
    long number = 0;
    vRLENonZeroPoints(training_rle, rows, cols, &number);
    vRotatePointsRight(rows, cols, rows, cols, number, training_rows, 
                        training_cols, 90 * view);

    double distance = vChamferFromDT(rows, cols, number, dt);
    result_matrix[result_index++] = distance;

    vdelete(training_rle);
    training_rle = 0;
    if (pick % 100 == 0)
    {
      vPrint("%15li\r", training_frame);
    }
  }
  vPrint("\n");

  vdelete2(rows);
  vdelete2(cols);
  return result;
}


vMatrix<double> 
vOneWayChamfer2(v3dMatrix<float> * dt, vPointSetSequence * database)
{
  long number_of_frames = database->NumberOfFrames();
  const long number_of_views = 4;
  long i, j;
  vMatrix<long> frames(1, number_of_frames * number_of_views);
  vMatrix<long> views(1, number_of_frames * number_of_views);
  for (i = 0; i < number_of_frames; i++)
  {
    for (j = 0; j < number_of_views; j++)
    {
      long index = i*4+j;
      frames(index) = i;
      views(index) = j;
    }
  }

  return vOneWayChamfer2(dt, database, &frames, &views);
}

                
vMatrix<double> 
vOneWayChamfer2(v3dMatrix<float> * dt, vPointSetSequence * database,
                 v3dMatrix<long> * frames, v3dMatrix<long> * views)
{
  // Declare and initialize certain useful variables.
  long number_of_training_frames = database->NumberOfFrames();
  long number_of_picks = frames->Size();
  if (number_of_picks != views->Size()) return vMatrix<double>();

  long training_rows = database->Rows();
  long training_cols = database->Cols();
  vMatrix<double> result;
  vArray(double) result_matrix = vZero(double);
  long result_index = 0;

  result = vMatrix<double>(1, number_of_picks);
  result_matrix = result.Matrix();

  v3dMatrix<float> * dt1 = vRotateLeft(dt);
  v3dMatrix<float> * dt2 = vRotateLeft(dt1);
  v3dMatrix<float> * dt3 = vRotateLeft(dt2);

  v3dMatrix<float> * dts[] = {dt, dt1, dt2, dt3};

  // Do the matching
  long error_flag = 0;
  vArray(long) picked_frames = frames->Matrix();
  vArray(long) frame_views = views->Matrix();
  long pick;
  for (pick = 0; pick < number_of_picks; pick++)
  {
    long training_frame = picked_frames[pick];
    long view = frame_views[pick];
    if ((training_frame < 0) || (training_frame >= number_of_training_frames) ||
        (view < 0) || (view >= 4))
    {
      vPrint("Bad frame or view pick: %li %li\n", training_frame, view);
      error_flag = 1;
      break;
    }
  
    // Read in the points of the training frame.
    long number = database->Length(training_frame);
    vArray(uchar) rows = database->PointRows(training_frame);
    vArray(uchar) cols = database->PointCols(training_frame);
    double distance = vChamferFromDT(rows, cols, number, dts[view]);
    result_matrix[result_index++] = distance;
  }
  vdelete(dt1);
  vdelete(dt2);
  vdelete(dt3);
  return result;
}


vMatchScores *
vOneWayChamfer2(grayscale_image * gray, vPointSetSequence * database,
                 vector<long> * frames, vector<long> * views)
{
  v3dMatrix<float> * dt = vCVDistanceTransform(gray);
  long number_of_frames = frames->size();
  v3dMatrix<long> frames_matrix(1, number_of_frames, 1);
  v3dMatrix<long> views_matrix(1, number_of_frames, 1);
  vArray(long) frames1 = frames_matrix.Matrix();
  vArray(long) views1 = views_matrix.Matrix();
  long i;
  for (i = 0; i < number_of_frames; i++)
  {
    frames1[i] = (*frames)[i];
    views1[i] = (*views)[i];
  }

  vMatrix<double> distances = vOneWayChamfer2(dt, database, 
                                                &frames_matrix, &views_matrix);
  vdelete(dt);
  vArray(vMatchScore) scores = vnew(vMatchScore, number_of_frames);
  for (i = 0; i < number_of_frames; i++)
  {
    scores[i].frame = frames1[i];
    scores[i].view = views1[i];
    scores[i].score = (float) distances(i);
    scores[i].is_valid = 1;
  }

  vMatchScores * result = new vMatchScores(scores, number_of_frames);
  return result;
}


// vRandomChamferDistances picks some random views from the database, and
// for each picked view it calculates the chamfer distances between
// that view and the entire database. Then it pieces all those distances
// together in one big file.
long vRandomChamferDistances(vRLE_Sequence * database, long number,
                              const char * target, long training_frames)
{
  vMatrix<long> random_views(1, number);
  vMatrix<long> random_orientations(1, number);
  long database_views = database->NumberOfFrames();
  long i;
  // Make the random picks for views.
  for (i = 0; i < number; i++)
  {
    random_views(i) = vRandomLong2(0, database_views-1);
    random_orientations(i) = i % 4;
  }

  // Save the random pick values, so that if something goes wrong we
  // can still get that information.
  char * random_pick_file = vMergeStrings2(target, ".random");
  FILE * fp = fopen(random_pick_file, vFOPEN_WRITE);
  if (fp == 0)
  {
    vPrint("Error in opening %s\n", random_pick_file);
	vdelete2(random_pick_file);
    return 0;
  }

  long success1 = random_views.Write(fp);
  long success2 = random_orientations.Write(fp);
  fclose(fp);
  vdelete2(random_pick_file);
  if ((success1 <= 0) || (success2 <= 0))
  {
    vPrint("Error in saving random pick info.\n");
    return 0;
  }

  // Now, for each randomly picked view, save its chamfer distances with the 
  // entire database
  for (i = 0; i < number; i++)
  {
    long view = random_views(i);
    long orientation = random_orientations(i);
    grayscale_image * original_view = database->ReadFrameGray(view);
    grayscale_image * binary_view = (grayscale_image *) original_view->threshold(1);
    grayscale_image * view_image = (grayscale_image *) vRotateRight(binary_view, orientation * 90);
    vdelete(original_view);
    vdelete(binary_view);
    vMatrix<double> distances = vDatabaseChamfer(view_image, database, -1, -1, training_frames);
    vdelete(view_image);
    char * i_string = vLongToString(i);
    char * temp_file = vMergeStrings3(target, ".", i_string);
    vdelete2(i_string);
    success1 = distances.Write(temp_file);
    if (success1 <= 0)
    {
      vPrint("Error in saving to %s\n", temp_file);
      vdelete2(temp_file);
      return 0;
    }
    vPrint("Processed frame %li\n", i);
    vdelete2(temp_file);
  }

  return 1;
}



// vMergeRandomChamferDistances takes a sequence of files created by
// vRandomChamferDistances and puts them together in one big file.
// The individual files have random matrices of type v3dMatrix<double>.
// The result is saved as v3dMatrix<ushort> to save disk space.
long vMergeRandomChamferDistances(const char * base, long start, long end)
{
  if (end < start) return 0;

  // Read the first matrix to figure out the number of entries in each 
  // matrix.
  char * start_string = vLongToString(start);
  char * first_file = vMergeStrings3(base, ".", start_string);
  vdelete2(start_string);
  v3dMatrix<double> * first_distances = v3dMatrix<double>::Read(first_file);
  vdelete2(first_file);
  if (first_distances == 0)
  {
    vPrint("Error in reading first file, base = %s, start = %li\n", 
            base, start);
    return 0;
  }

  // The individual distance matrices are 1 x database_frames (1 row).
  // The merged matrix is database_frames x (end-start+1).
  long merge_rows = first_distances->Cols();
  long merge_cols = end - start + 1;
  v3dMatrix<float> * merged_matrix = 
    new v3dMatrix<float>(merge_rows, merge_cols, 1);
  vArray2(float) merged2 = merged_matrix->Matrix2(0);
  vdelete(first_distances);

  // Read in all matrices and store their entries in merged_matrix.
  long i;
  for (i = start; i <= end; i++)
  {
    char * i_string = vLongToString(i);
    char * i_file = vMergeStrings3(base, ".", i_string);
    vdelete2(i_string);
    v3dMatrix<double> * i_distances_matrix = v3dMatrix<double>::Read(i_file);
    vdelete2(i_file);
    if ((i_distances_matrix == 0) || (i_distances_matrix->Size() != merge_rows))
    {
      vPrint("Error in reading distance matrix %li\n", i);
      vdelete(i_distances_matrix);
      vdelete(merged_matrix);
      return 0;
    }
    vArray(double) i_distances = i_distances_matrix->Matrix();
    long entry;
    for (entry = 0; entry < merge_rows; entry++)
    {
      merged2[entry][i] = (float) i_distances[entry];
    }
    vdelete(i_distances_matrix);
    vPrint("Processed distance matrix %li\r", i);
  }
  vPrint("\n");
  long success = vSaveMatrixAsUshorts(merged_matrix, base);
  vdelete(merged_matrix);
  return success;
}


// database_edges is a video file with edge images of the model.
// point_sequence is a structure that contains, for every database frame,
// the list of all non-zero pixels in the corresponding edge image in 
// database_edges. (frames[i], views[i]) specifies the i-th base image.
// For each of those base images, we compute the one way chamfer distances
// from each database image to them (the order matters). The result is
// a matrix whose rows are equal to the number of database images, and
// whose cols are equal to the number of base images, so that entry i,j
// is the one-way distance from the i-th database image to the j-th base image.
// This matrix is saved into a file that is obtained by replacing the ending
// of the filename of database_edges with ".chamfer1_index".
long vSaveChamferIndexIM(vRLE_Sequence * database_edges,
                          vPointSetSequence * point_sequence, 
                          v3dMatrix<long> * frames,
                          v3dMatrix<long> * views)
{
  long number_of_frames = database_edges->NumberOfFrames();
  long number_of_bases = frames->Size();
  if (number_of_bases != views->Size()) return 0;
  if (frames->Bands() != 1) return 0;
  if (views->Bands() != 1) return 0;
  if (frames->Rows() != 1) return 0;
  if (views->Rows() != 1) return 0;
  vPrint("600\n");
  vPrint("number_of_bases = %li\n", number_of_bases);

  number_of_bases = 200;
  vMatrix<double> distances(number_of_frames*4, number_of_bases);
  long success = 1;
  long i;
  vPrint("%\n");
  for (i = 0; i < number_of_bases; i++)
  {
    long frame = frames->Matrix()[i];
    long view = views->Matrix()[i];

    grayscale_image * edge_image0 = database_edges->ReadFrameGray(frame);
    if (edge_image0 == 0) 
    {
      success = 0;
      break;
    }
    grayscale_image * edge_image1 = (grayscale_image *) vRotateRight(edge_image0, view * 90);
    vdelete(edge_image0);

    v3dMatrix<float> * edge_dt = vCVDistanceTransform(edge_image1);
    vdelete(edge_image1);
    vMatrix<double> base_distances = vOneWayChamfer2(edge_dt, point_sequence);
    long j;
    for (j = 0; j < number_of_frames * 4; j++)
    {
      distances(j,i) = base_distances(j);
    }
    vdelete(edge_dt);
    vPrint("Processed frame %li of %li\r", i, number_of_bases);
  }
  vPrint("\n");
  if (success <= 0)
  {
    vPrint("Errors occurred\n");
    return success;
  }

  char * output_file = vReplaceEnding2(database_edges->GetFilename(), ".chamfer1_index");
  success = vSaveMatrixAsUshorts(&distances, output_file);
  if (success <= 0)
  {
    vPrint("Failed to save result into %s\n", output_file);
  }
  vdelete2(output_file);
  return success;
}

// Beginning of functions used for the GW2003 paper.
// Saves all directed chamfer distances between every 
// shape/viewpoint combination in the database and every 
// database image. The direction is from database image
// to shape viewpoint combination.
// database_edges is the video sequence of database edge images.
// points is an object that stores all the non-zero pixels (i.e.
// edge pixels) in those images.
v3dMatrix<uchar> * vAll_MI_Distances(vVideoFile * database_edges, 
                                        vPointSetSequence * points)
{
  if (database_edges == 0) return 0;
  if (points == 0) return 0;

  // Compute the necessary size to store all distances.
  long database_frames = database_edges->NumberOfFrames();
  const long views_per_frame = 4;
  const long rotations_per_viewpoint = 12;
  long number_of_scores = database_frames * views_per_frame;
  long rows = database_frames / rotations_per_viewpoint;
  long cols = database_frames * views_per_frame;
  v3dMatrix<uchar> * result_matrix = new v3dMatrix<uchar>(rows, cols, 1);
  vArray2(uchar) result = result_matrix->Matrix2(0);
  
  // Store in column vectors the frame and view of each database image.
  // We do that so that we can conveniently call the vOneWayChamfer2
  // function a bit later.
  v3dMatrix<long> training_frames_matrix(1, number_of_scores, 1);
  v3dMatrix<long> training_views_matrix(1, number_of_scores, 1);
  vArray(long) training_frames = training_frames_matrix.Matrix();
  vArray(long) training_views= training_views_matrix.Matrix();
  long i;
  for (i = 0; i < database_frames; i++)
  {
    long index = views_per_frame * i;
    long j;
    for (j = 0; j < views_per_frame; j++)
    {
      training_frames[index] = i;
      training_views[index] = j;
      index++;
    }
  }

  long frame;
  vPrint("\n");
  // For each shape/viewpoint combination (every 12th frame in the database) 
  // compute the directed chamfer distance from each database image to the 
  // shape/viewpoint combination.
//  for (frame = 0; frame < 24; frame+= rotations_per_viewpoint) // This line used for debugging.
  for (frame = 0; frame < database_frames; frame += rotations_per_viewpoint)
  {
    // Read image (shape/viewpoint combination) from the database edge images, 
    // compute its distance tranform.
    grayscale_image * edge1 = database_edges->ReadFrameGray(frame);
    v3dMatrix<float> * dt1 = vCVDistanceTransform(edge1);
    vdelete(edge1);

    // Now, for each database image, compute its directed chamfer distance
    // to dt1.
    vMatrix<double> distances = 
      vOneWayChamfer2(dt1, points, 
                      &training_frames_matrix, &training_views_matrix);
    vdelete(dt1);

    // Convert distances to uchar, so that they can be stored in little memory.
    vArray(double) distances_array = distances.Matrix();
    long row = frame / rotations_per_viewpoint;
    for (i = 0; i < number_of_scores; i++)
    {
      double distance = distances_array[i];
      result[row][i] = double_to_uchar_mi_convert(distance);
    }
    vPrint("Done with frame %li\r", frame);
  }
  vPrint("\n");

  return result_matrix;
}


// Saves all directed chamfer distances between every 
// shape/viewpoint combination in the database and every 
// database image. The direction is from shape/viewpoint combination
// to database image. See comments for vAll_MI_Distances for
// more information on how it works.
v3dMatrix<uchar> * vAll_IM_Distances(vVideoFile * database_edges, 
                                        vPointSetSequence * points)
{
  if (database_edges == 0) return 0;
  if (points == 0) return 0;

  long database_frames = database_edges->NumberOfFrames();
  const long views_per_frame = 4;
  const long rotations_per_viewpoint = 12;
  long number_of_scores = database_frames * views_per_frame;
  long rows = database_frames / rotations_per_viewpoint;
  long cols = database_frames * views_per_frame;
  v3dMatrix<uchar> * result_matrix = new v3dMatrix<uchar>(rows, cols, 1);
  vArray2(uchar) result = result_matrix->Matrix2(0);
  
  v3dMatrix<long> training_frames_matrix(1, rows, 1);
  v3dMatrix<long> training_views_matrix(1, rows, 1);
  vArray(long) training_frames = training_frames_matrix.Matrix();
  vArray(long) training_views= training_views_matrix.Matrix();
  long i;
  for (i = 0; i < rows; i++)
  {
    training_frames[i] = i * rotations_per_viewpoint;
    training_views[i] = 0;
  }

  long frame2;
  vPrint("\n");
  for (frame2 = 0; frame2 < database_frames; frame2++)
//  for (frame2 = 0; frame2 < 10; frame2++) // For debugging.
  {
    // Read next image from the database edge images, compute its distance
    // tranform.
    grayscale_image * edge2 = database_edges->ReadFrameGray(frame2);
    v3dMatrix<float> * dt2 = vCVDistanceTransform(edge2);
    vdelete(edge2);

    long view;
    for (view = 0; view < views_per_frame; view++)
    {
      v3dMatrix<float> * dt = vRotateRight(dt2, view * 90);
      // Now, for each shape/viewpoint combination, compute its directed chamfer distance
      // to dt.

      vMatrix<double> distances = 
        vOneWayChamfer2(dt, points, 
                         &training_frames_matrix, &training_views_matrix);
      vdelete(dt);

      vArray(double) distances_array = distances.Matrix();
      long col = frame2 * views_per_frame + view;
      for (i = 0; i < rows; i++)
      {
        double distance = distances_array[i];
        result[i][col] = double_to_uchar_mi_convert(distance);
      }
    }

    vdelete(dt2);
    vPrint("Done with frame %li\r", frame2);
  }
  vPrint("\n");

  return result_matrix;
}


// Given a database image described as (frame, view), and
// an image rotation index A (from 0 to 48), we want to
// find the index B, such that the angle between (frame, view)
// and A, in terms of image rotation, is the same as the angle
// between (0, 0) and B, in terms of image rotation. That gets
// stored in adjusted[A].
long S_AdjustImageIndex(long frame, long view, long * adjusted)
{
  static vFrameViewFinder finder;

  long rotation1 = finder.ActualRotation2(frame, view);
  long i;
  for (i = 0; i < 48; i++)
  {
    long rotation2 = finder.ActualRotation1(i);
    long difference;
    if (rotation1 > rotation2)
    {
      difference = rotation2 + 48 - rotation1;
    }
    else 
    {
      difference = rotation2 - rotation1;
    }
    long frame2 = finder.FrameIndex(difference);
    long view2 = finder.RotationIndex(difference);
    adjusted[i] = frame2 * 4 + view2;
  }
  return 1;
}


// Given a matrix that contains the chamfer distances between
// every shape/viewpoint combination and all database images,
// we want to extract the distances between a given database 
// image A and all other database images. Essentially, for every
// database image B, the distance between A and B is the same
// as the distance between the shape/viewpoint combination that
// corresponds to A and some image of the same shape/viewpoint
// combination as B. We have to figure out, for every A and B,
// what the right C is.
vMatrix<double> vGet_CD_Distances(long frame, long view, 
                                    v3dMatrix<uchar> * cd_distances)
{
  long adjusted[48];
  S_AdjustImageIndex(frame, view, adjusted);

  long cols = cd_distances->Cols();
  long row = frame / 12;
  vMatrix<double> result_matrix(1, cols);
  vArray(double) result = result_matrix.Matrix();
  vArray2(uchar) cd = cd_distances->Matrix2(0);

  long i;
  for (i = 0; i < cols; i++)
  {
    long rotation = i % 48;
    long base = i - rotation;
    long adjusted_i = base + adjusted[rotation];
    long distance = cd[row][adjusted_i];
    result[i] = uchar_to_double_cd_convert(distance);
  }

  return result_matrix;
}


// This function is used to test and debug the way indices are adjusted
// in vGet_CD_Distances, which is based on S_AdjustImageIndex.
long vTestIndexAdjustment(long frame, long view, long index)
{
  long adjusted[48];
  S_AdjustImageIndex(frame, view, adjusted);

  long i = index;
  long rotation = i % 48;
  long base = i - rotation;
  long adjusted_i = base + adjusted[rotation];
  return adjusted_i;
}


// We produce "number" pairs of frame/view combinations that
// correspond to database images. The first frame/view combination
// corresponds to a shape/viewpoint combination. Since the 
// database contains 12 frames per shape/viewpoint combination, only
// one of those 12 frames (the first one) is represented here.
v3dMatrix<long> * vRandom_CD_Pairs(long number)
{
  long rows = 4;
  v3dMatrix<long> * result_matrix = new v3dMatrix<long>(rows, number, 1);
  vArray2(long) result = result_matrix->Matrix2(0);

  // For frame1 we have only as many choices as the shape/viewpoint 
  // combinations that exist in the database.
  long frame_limit1 = 2235;
  // For frame2 we have as many choices as the database frames.
  long frame_limit2 = (frame_limit1+1) * 12 - 1;
  long view_limit = 3;

  long i;
  for (i = 0; i < number; i++)
  {
    long frame1 = vRandomLong2(0, frame_limit1) * 12;
    long view1 = vRandomLong2(0, view_limit);
    long frame2 = vRandomLong2(0, frame_limit2);
    long view2 = vRandomLong2(0, view_limit);

    result[0][i] = frame1;
    result[1][i] = view1;
    result[2][i] = frame2;
    result[3][i] = view2;
  }

  return result_matrix;
}


// cd_pairs specifies some pairs of hand poses represented in 
// the database (frame/view combinations). cd_distances contains 
// precomputed chamfer distances between every shape/viewpoint 
// combination (every 12th frame stored in the database) and 
// every hand pose represented in the database (4 poses are 
// represented by each frame in the database). This function
// looks up the distances between the pairs of poses stored in cd_pairs,
// using cd_distances, and stores those distances in the result matrix,
// which is a one-dimensional column vector.
v3dMatrix<double> * vCD_Pair_Distances(v3dMatrix<long> * cd_pairs, 
                                          v3dMatrix<uchar> * cd_distances)
{
  long cols = cd_pairs->Cols();

  // Create the result.
  v3dMatrix<double> * result_matrix = new v3dMatrix<double>(1, cols, 1);
  vArray(double) result = result_matrix->Matrix();
  vArray2(long) pairs = cd_pairs->Matrix2(0);
  vArray2(uchar) distances = cd_distances->Matrix2(0);
  // finder is used to do some arithmetic that will tell us what
  // entry in cd_distances contains the distance we are looking for.
  static vFrameViewFinder finder;

  long i;
  vPrint("\n");
  for (i = 0; i < cols; i++)
  {
    // Get the hand pose images the distance between which we must
    // look up.
    long frame1 = pairs[0][i];
    long view1 = pairs[1][i];
    long frame2 = pairs[2][i];
    long view2 = pairs[3][i];

    // Figure out where to look up this distance. 
    // First, figure out what row in cd_distances we should look at,
    // i.e. what shape/viewpoint combination the first hand pose
    // corresponds to 
    long row1 = frame1 / 12;
    // Now, figure out what column we should
    // look at, i.e. what hand pose image is as far from the
    // shape/viewpoint combination as the two hand poses
    // (frame1/view1 and frame2/view2) are from each other.
    long rotation1 = finder.ActualRotation2(frame1, view1);
    long index2 = frame2 * 4 + view2;
    long rotation2 = finder.ActualRotation2(frame2, view2);
    long base2 = index2 - (index2 % 48);

    long difference;
    if (rotation1 > rotation2)
    {
      difference = rotation2 + 48 - rotation1;
    }
    else 
    {
      difference = rotation2 - rotation1;
    }
    long frame3 = finder.FrameIndex(difference);
    long view3 = finder.RotationIndex(difference);
    long adjustment = frame3 * 4 + view3;
    long adjusted2 = base2 + adjustment;

    // Look up the distance and store it.
    long distance = distances[row1][adjusted2];
    result[i] = uchar_to_double_cd_convert(distance);
    if (i % 1000 == 0)
    {
      vPrint("Done with pair %i\r", i);
    }
  }
  vPrint("\n");  

  return result_matrix;
}


// This function receives the following arguments:
// all_cd_distances: precomputed chamfer distances between each
// shape/viewpoint combination in the database and each hand pose
// image represented by the database. 
// cd_pairs: randomly selected pairs of hand pose images represented
// in the database.
// cd_pair_distances: precomputed chamfer distances between the 
// pairs of hand pose images stored in cd_pairs.
// estimated_distances.
// estimated_distances: contains the distances for the pairs in
// cd_pairs, as computed using Lipschitz embeddings for some
// choice of reference images.
// factor: specifies one out of how many pairs from cd_pairs we 
// should use. This is important, since it can take too much time
// if we use all pairs in cd_pairs.
// The goal of this function is to compute the average distortion
// obtained by trying all possible sets of reference images obtained
// by adding another database image to the previous set of reference
// images. To compute that we do not need the previous set of 
// reference images, just the L1 distances for the embeddings of the 
// cd_pairs using that previous reference set, as stored i 
// estimated_distances.
// The set of candidates (for efficiency) does not contain all hand poses
// in the database, but only four poses per shape/viewpoint combination,
// corresponding to 0, 90, 180 and 270 degrees image-plane rotation of the 
// image of that combination.
vMatrix<double> vCalculateSTDs(v3dMatrix<uchar> * all_cd_distances,
                                 v3dMatrix<long> * cd_pairs,
                                 v3dMatrix<double> * cd_pair_distances,
                                 v3dMatrix<double> * estimated_distances,
                                 long factor)
{
  long candidates = all_cd_distances->Rows() * 4;
  vMatrix<double> distortions_m = vMatrix<double>(2, candidates);
  vArray2(double) distortions = distortions_m.Matrix2();
  double mean, std;
  mean = std = 0;
  long i;
  vPrint("\n");
  // Calculate the mean scaling and average distortion achieved by
  // each candidate. The mean scaling is computed just out of curiosity,
  // it is not really used anywhere.
  for (i = 0; i < candidates; i++)
  {
    vCalculateSTD(all_cd_distances, cd_pairs, cd_pair_distances,
                   estimated_distances, i, &mean, &std, factor);
    distortions[0][i] = mean;
    distortions[1][i] = std;
    vPrint("Processed candidate %li of %li\r", i, candidates);
  }
  vPrint("\n");

  return distortions_m;
}


// This function computes the mean scaling and average distortion
// achieved by each candidate to be added to a previously chosen 
// set of reference images used to define a Lipschitz embedding.
// For more information see comments on vCalculateSTDs. meanp
// and stdp are output parameters.
long vCalculateSTD(v3dMatrix<uchar> * all_cd_distances,
                    v3dMatrix<long> * cd_pairs,
                    v3dMatrix<double> * cd_pair_distances,
                    v3dMatrix<double> * estimated_distances,
                    long candidate, double * meanp, double * stdp,
                    long factor)
{
  if ((candidate < 0) || (candidate >= 86 * 26 * 4))
  {
    return 0;
  }

  vArray2(uchar) cd_distances = all_cd_distances->Matrix2(0);
  vArray2(long) pairs = cd_pairs->Matrix2(0);
  vArray(double) pair_distances = cd_pair_distances->Matrix();
  vArray(double) estimated = estimated_distances->Matrix();
  long number = cd_pairs->Cols() / factor;
  // Find the hand pose image corresponding to this candidate.
  long frame = (candidate / 4) * 12;
  long view = candidate % 4;

  // Get the chamfer distances between the candidate image and 
  // all hand pose images represented in the database, using 
  // the precomputed distances stored in all_cd_distances.
  vMatrix<double> candidate_distances_m = vGet_CD_Distances(frame, view, all_cd_distances);
  // Create some arrays where, for each pair of hand pose images specified
  // in cd_pairs,
  // we will store information about how well the set of already chosen
  // reference images, together with this candidate, can be used
  // to approximate the true chamfer distance between the images of that pair.
  vMatrix<double> new_distances_m(1, number);
  vMatrix<double> total_distances_m(1, number);
  vMatrix<double> fraction_distances_m(1, number);
  vMatrix<double> scaled_distances_m(1, number);

  vArray(double) candidate_distances = candidate_distances_m.Matrix();
  vArray(double) new_distances = new_distances_m.Matrix();
  vArray(double) total_distances = total_distances_m.Matrix();
  vArray(double) fraction_distances = fraction_distances_m.Matrix();
  vArray(double) scaled_distances = scaled_distances_m.Matrix();

  long i;
  // Get new estimated distances
  long counter = 0; // Will count pairs for which actual_distance != 0.
  double total = 0.0;
  for (i = 0; i < number; i++)
  {
    // Look up the frame/view information for the current pair of hand
    // pose images.
    long frame1 = pairs[0][i];
    long view1 = pairs[1][i];
    long index1 = frame1 * 4 + view1;

    long frame2 = pairs[2][i];
    long view2 = pairs[3][i];
    long index2 = frame2 * 4 + view2;

    // Look up the precomputed chamfer distances between each
    // image of the pair and the candidate.
    double distance1 = candidate_distances[index1];
    double distance2 = candidate_distances[index2];
    // Look up the contribution to the L1 distance between the
    // Lipschitz embeddings of the two images of the pair that is
    // made by looking at their distances to the candidate.
    new_distances[i] = vAbs(distance1 - distance2);
    // Add this contribution to the contributions made by all
    // the previously chosen reference images.
    total_distances[i] = estimated[i] + new_distances[i];
    // Look up the actual distance for the current pair of images.
    double actual_distance = pair_distances[i];
    // Estimate the ratio between the distance of the embeddings
    // and the original distance.
    if (actual_distance == 0)
    {
      fraction_distances[i] = -1;
    }
    else
    {
      fraction_distances[i] = total_distances[i] / actual_distance;
      counter++;
      total += fraction_distances[i];
    }
  }

  if ((counter == 0) || (total == 0))
  {
    vPrint("Error: counter or total = 0\n");
    return 0;
  }

  // Calculate the mean ratio between distance calculated using
  // the embedding and the original distance.
  double mean = total / (double) counter;
  if (mean == 0)
  {
    vPrint("Error: mean = 0\n");
    return 0;
  }

  // Scale the embedded distances so that the mean is 1. This is a 
  // heuristic. Overall, we should find the scale factor that minimizes
  // the average distortion. Here, instead of searching for that 
  // optimal scale factor, we assume that the scale factor that
  // makes the mean of fraction_distances 1 is a pretty good guess,
  // and that is the scale factor that we use.
  for (i = 0; i < number; i++)
  {
    scaled_distances[i] = fraction_distances[i] / mean;
  }

  // Calculate the average distortion (called std in the code, which
  // is a somewhat incorrect term, since std is the square root of the
  // mean-squared-error, whereas here we just take the mean absolute
  // error). 
  // Clip scaled distances, to achieve robustness. The maximum contribution
  // to the sum of distortions is 10.
  double std_sum = 0;
  // First, sum up all average distortions.
  for (i = 0; i < number; i++)
  {
    // Skip cases where the actual distance is 0.
    if (fraction_distances[i] < -0.5) continue;
    // Get the current distortion.
    double current_std = vAbs(scaled_distances[i] - 1);
    // Clip the distortion if it is over 10, for robustness.
    if (current_std > 10) current_std = 10;
    std_sum += current_std;
  }

  double std = std_sum / counter;
  *meanp = mean;
  *stdp = std;

  return 1;
}


// See comments on vCalculateSTDs for info on all_cd_distances,
// cd_pairs, estimated_distances, and factor. frame and view
// specify the candidate that was chosen to be added to the
// set of reference images. At the time we call the function,
// estimated_distances contains the L1 distances in the embedded
// space for the pairs of images specified in cd_pairs, for the 
// embedding obtained using the previous set of reference images.
// At the end of the function, estimated_distances contains the
// L1 distances for the pairs of images specified in cd_pairs, for
// the new set of reference images, which consists of the previous
// reference images and the new reference image, specified by
// frame and view.
long vUpdateEstimatedDistances(v3dMatrix<uchar> * all_cd_distances,
                                v3dMatrix<long> * cd_pairs,
                                v3dMatrix<double> * estimated_distances,
                                long frame, long view, long factor)
{
  vArray2(long) pairs = cd_pairs->Matrix2(0);
  vArray(double) estimated = estimated_distances->Matrix();
  long number = cd_pairs->Cols() / factor;

  // Get the distances between the new reference image (the chosen
  // candidate) and all hand pose images represented in the database.
  vMatrix<double> candidate_distances_m = vGet_CD_Distances(frame, view, all_cd_distances);
  vArray(double) candidate_distances = candidate_distances_m.Matrix();

  // Update estimated_distances for each pair in cd_pairs.
  long i;
  for (i = 0; i < number; i++)
  {
    // Look up the frame/view for each image in the current pair.
    long frame1 = pairs[0][i];
    long view1 = pairs[1][i];
    long index1 = frame1 * 4 + view1;

    long frame2 = pairs[2][i];
    long view2 = pairs[3][i];
    long index2 = frame2 * 4 + view2;

    // Compute the contribution to the L1 distance in the Euclidean space,
    // between the targets of the two images of the current pair, that is 
    // made by looking at the distance between each of those images and the
    // new reference image.
    double distance1 = candidate_distances[index1];
    double distance2 = candidate_distances[index2];
    double distance = vAbs(distance1 - distance2);
    // Add this contribution to estimated_distances.
    estimated[i] += distance;
  }

  return 1;
}


// Choose basis elements in a greedy way, to minimize expected distortion
// (measured as a fraction of the actual distance). 
// For information on the arguments, see comments for vCalculateSTDs.
// number is the total number of reference images we want to choose.
vMatrix<double> vChooseBasis1(v3dMatrix<uchar> * all_cd_distances,
                                v3dMatrix<long> * cd_pairs,
                                v3dMatrix<double> * cd_pair_distances,
                                long factor, long number)
{
  if ((number <= 0) || (factor <= 0)) return vMatrix<double>();

  // Create estimated_distances, that will store, at each point,
  // the L1 distances (i.e. approximate chamfer distances) between
  // images in cd_pairs that are obtained using the current set of reference
  // images.
  long rows = cd_pair_distances->Rows();
  long cols = cd_pair_distances->Cols() / factor;
  vMatrix<double> estimated_distances(1, cols);
  vWriteValue(&estimated_distances, (double) 0);
  
  // Create the result matrix, that will contain one row for each
  // reference image, specifying the row number, frame/view of the
  // reference image, and the mean scaling and average distortion
  // obtained by the reference images stored up to the current row.
  // The mean scaling is not really important, I was just curious
  // to see what it is, that's why it is stored here.
  vMatrix<double> result_matrix(number, 5);
  vWriteValue(&result_matrix, (double) 0);
  vArray2(double) result = result_matrix.Matrix2();

  long i;
  for (i = 0; i < number; i++)
  {
    // Calculate the average distortions for each candidate new
    // reference image.
    vMatrix<double> distortions =
      vCalculateSTDs(all_cd_distances, cd_pairs, cd_pair_distances,
                      &estimated_distances, factor);

    // Find the best candidate, i.e. the one that, combined with 
    // the already chosen reference images, achieves the lowest distortion.
    // That candidate will be the next chosen reference image.
    long min_std_index = 0;
    double min_std = color_imageRowMin3(&distortions, 1, &min_std_index);
    double mean = distortions(0, min_std_index);
    // Compute the frame/view info for the new reference image.
    long frame = (min_std_index / 4) * 12;
    long view = (min_std_index % 4);
    vPrint("%li: image (%li, %li), mean[%li] = %lf, std = %lf\n", 
            i, frame, view, min_std_index, mean, min_std);
    // Store information in the result about the new reference image.
    result[i][0] = i;
    result[i][1] = frame;
    result[i][2] = view;
    result[i][3] = mean;
    result[i][4] = min_std;
    char simple_name[1000];
    sprintf(simple_name, "basis1c.%li", i);
    char * filename1 = vJoinPaths("g:\\users\\athitsos\\sequences\\glove_data\\training\\gw2003",
                                   simple_name);
    char * filename2 = vJoinPaths("r:\\text\\experiments\\gw2003",
                                   simple_name);
    // Output some useful information into two files, one in UNIX and the other
    // on the local PC.
    long success1 = result_matrix.WriteText(filename1);
    long success2 = result_matrix.WriteText(filename2);
    vPrint("Success1 = %li, success2 = %li\n", success1, success2);
    vdelete2(filename1);
    vdelete2(filename2);

    // Update estimated distances, to reflect the L1 distances (approximate
    // chamfer distances) between the images in each pair in cd_pairs, 
    // that are obtained using the reference images chosen so far (including
    // the newly chosen reference image.
    vUpdateEstimatedDistances(all_cd_distances, cd_pairs, 
                               &estimated_distances, frame, view, factor);
  }

  return result_matrix;
}


// Choose basis elements randomly. See comments on vChooseBasis1
// for more information. This function is structured in a pretty 
// similar way to vChooseBasis1. The only difference is that here,
// when we choose the next reference image to add to the set of 
// reference images, we make the choice randomly. However, we
// still compute the same information as in vChooseBasis1, so that
// we can easily compare the average distortions obtained with each
// method of choosing bases. 
// This function is grossly inefficient, since it computes 
// average distortions attained by each candidate, which are useless,
// since those distortions are not used for anything, except for 
// the distortion. It was written this way because it was very fast
// to write, by just modifying vChooseBasis1.
vMatrix<double> vChooseBasis2(v3dMatrix<uchar> * all_cd_distances,
                                v3dMatrix<long> * cd_pairs,
                                v3dMatrix<double> * cd_pair_distances,
                                long factor, long number)
{
  if ((number <= 0) || (factor <= 0)) return vMatrix<double>();

  long rows = cd_pair_distances->Rows();
  long cols = cd_pair_distances->Cols() / factor;
  vMatrix<double> estimated_distances(1, cols);
  vWriteValue(&estimated_distances, (double) 0);
  
  vMatrix<double> result_matrix(number, 5);
  vWriteValue(&result_matrix, (double) 0);
  vArray2(double) result = result_matrix.Matrix2();

  long i;
  for (i = 0; i < number; i++)
  {
    vMatrix<double> distortions =
      vCalculateSTDs(all_cd_distances, cd_pairs, cd_pair_distances,
                      &estimated_distances, factor);

    long index = vRandomLong2(0, 86 * 26 * 4);
    double mean = distortions(0, index);
    double std = distortions(1, index);
    long frame = (index / 4) * 12;
    long view = (index % 4);
    vPrint("%li: image (%li, %li), mean[%li] = %lf, std = %lf\n", 
            i, frame, view, index, mean, std);
    result[i][0] = i;
    result[i][1] = frame;
    result[i][2] = view;
    result[i][3] = mean;
    result[i][4] = std;
    char simple_name[1000];
    sprintf(simple_name, "basis2.%li", i);
    char * filename1 = vJoinPaths("g:\\users\\athitsos\\sequences\\glove_data\\training\\gw2003",
                                   simple_name);
    char * filename2 = vJoinPaths("r:\\text\\experiments\\gw2003",
                                   simple_name);
    long success1 = result_matrix.WriteText(filename1);
    long success2 = result_matrix.WriteText(filename2);
    vPrint("Success1 = %li, success2 = %li\n", success1, success2);
    vdelete2(filename1);
    vdelete2(filename2);

    // Update estimated distances;
    vUpdateEstimatedDistances(all_cd_distances, cd_pairs, 
                               &estimated_distances, frame, view, factor);
  }

  return result_matrix;
}


// This function assumes that a set of reference images has been
// chosen by one of the vChooseBasis? functions (where the "basis"
// argument specifies the value for ?, i.e. which vChooseBasis 
// function was used). "index" specifies which particular file of
// the many files generated by the vChooseBasis function should
// be used to read the reference images. Essentially, the number
// of reference images is index+1. The information about 
// all reference images is read, and the distances between each
// hand pose image and the reference images is looked up using
// all_cd_distances. The output parameters random_training_distances_p,
// random_picks_p and random_views_p store the embeddings of 
// all database images and info about the reference images, so that
// the approximate chamfer distance between any image and the database
// images can be easily computed.
long vMakeLipschitzBasis(v3dMatrix<uchar> * all_cd_distances, 
                          long basis, long index,
                          v3dMatrix<float> ** random_training_distances_p, 
                          v3dMatrix<long> ** random_picks_p, 
                          v3dMatrix<long> ** random_views_p)
{
  // The number of reference images is one plus the index.
  long dimensions = index+1;
  // Figure out the filename where the information has been saved.
  char simple_name[100];
  sprintf(simple_name, "basis%lib.%li", basis, index);
  char * filename = vJoinPaths("g:\\users\\athitsos\\sequences\\glove_data\\training\\gw2003",
                                simple_name);
  // Read the information about the reference images.
  v3dMatrix<double> * basis_info = 
    v3dMatrix<double>::ReadText(filename);
  // If we failed to read the information, return.
  if (basis_info == 0)
  {
    vPrint("Error in reading %s\n", filename);
    vdelete2(filename);
    return 0;
  }
  vdelete2(filename);

  long rows = basis_info->Rows();
  long cols = basis_info->Cols();
  if ((rows < dimensions) || (cols != 5))
  {
    vPrint("Bad basis info\n");
    vdelete(basis_info);
    return 0;
  }

  vArray2(double) info = basis_info->Matrix2(0);

  // Store the information that needs to be stored to compute 
  // approximate chamfer distances, i.e. frame/view info
  // of reference images, and Lipschitz embeddings of all
  // database images using these reference images.
  v3dMatrix<long> * random_picks_m = new v3dMatrix<long>(1, dimensions, 1);
  v3dMatrix<long> * random_views_m = new v3dMatrix<long>(1, dimensions, 1);
  v3dMatrix<float> * random_distances_m = new v3dMatrix<float>(107328, dimensions, 1);
  vArray2(float) random_distances = random_distances_m->Matrix2(0);
  vArray(long) random_picks = random_picks_m->Matrix();
  vArray(long) random_views = random_views_m->Matrix();

  long i;
  for (i = 0; i < dimensions; i++)
  {
    // Get info on the i-th reference image.
    long frame = vRound(info[i][1]);
    long view = vRound(info[i][2]);
    random_picks[i] = frame;
    random_views[i] = view;

    // Get distances between all database images and the i-th reference image.
    vMatrix<double> distances_matrix = vGet_CD_Distances(frame, view, all_cd_distances);
    long j;
    long number = distances_matrix.Cols();
    vArray(double) distances = distances_matrix.Matrix();
    for (j = 0; j < number; j++)
    {
      random_distances[j][i] = (float) distances[j];
    }
  }

  *random_training_distances_p = random_distances_m;
  *random_picks_p = random_picks_m;
  *random_views_p = random_views_m;
  vdelete(basis_info);

  return 1;
}


long vMakeLipschitzBasis(v3dMatrix<uchar> * all_cd_distances, 
                          long basis, long index,
                          v3dMatrix<float> ** database_embeddings, 
                          v3dMatrix<float> ** reference_images_p)
{
  // The number of reference images is one plus the index.
  long dimensions = index+1;
  // Figure out the filename where the information has been saved.
  char simple_name[100];
  sprintf(simple_name, "basis%lib.%li", basis, index);
  char * filename = vJoinPaths("g:\\users\\athitsos\\sequences\\glove_data\\training\\gw2003",
                                simple_name);
  // Read the information about the reference images.
  v3dMatrix<double> * basis_info = 
    v3dMatrix<double>::ReadText(filename);
  // If we failed to read the information, return.
  if (basis_info == 0)
  {
    vPrint("Error in reading %s\n", filename);
    vdelete2(filename);
    return 0;
  }
  vdelete2(filename);

  long rows = basis_info->Rows();
  long cols = basis_info->Cols();
  if ((rows < dimensions) || ((cols != 5) && (cols != 3)))
  {
    vPrint("Bad basis info\n");
    vdelete(basis_info);
    return 0;
  }

  vArray2(double) info = basis_info->Matrix2(0);

  // Store the information that needs to be stored to compute 
  // approximate chamfer distances, i.e. frame/view info
  // of reference images, and Lipschitz embeddings of all
  // database images using these reference images.
  v3dMatrix<long> * random_picks_m = new v3dMatrix<long>(1, dimensions, 1);
  v3dMatrix<long> * random_views_m = new v3dMatrix<long>(1, dimensions, 1);
  v3dMatrix<float> * random_distances_m = new v3dMatrix<float>(107328, dimensions, 1);
  vArray2(float) random_distances = random_distances_m->Matrix2(0);
  v3dMatrix<float> * reference_images = new v3dMatrix<float>(dimensions, 3, 1);
  vArray2(float) references = reference_images->Matrix2(0);

  long i;
  for (i = 0; i < dimensions; i++)
  {
    long frame, view;
    // Get info on the i-th reference image.
    if (cols == 5)
    {
      frame = vRound(info[i][1]);
      view = vRound(info[i][2]);
      references[i][0] = (float) frame;
      references[i][1] = (float) view;
      references[i][2] = (float) 1.;
    }
    else
    {
      frame = vRound(info[i][0]);
      view = vRound(info[i][1]);
      references[i][0] = (float) frame;
      references[i][1] = (float) view;
      references[i][2] = (float) info[i][2];
    }

    float weight = references[i][2];
    // Get distances between all database images and the i-th reference image.
    vMatrix<double> distances_matrix = vGet_CD_Distances(frame, view, all_cd_distances);
    long j;
    long number = distances_matrix.Cols();
    vArray(double) distances = distances_matrix.Matrix();
    for (j = 0; j < number; j++)
    {
      random_distances[j][i] = weight * (float) distances[j];
    }
  }

  *database_embeddings = random_distances_m;
  *reference_images_p = reference_images;
  vdelete(basis_info);
  vdelete(random_picks_m);
  vdelete(random_views_m);

  return 1;
}


long vMakeLipschitzBasis3(long basis, long index,
                           v3dMatrix<float> ** reference_images_p)
{
  // The number of reference images is one plus the index.
  long dimensions = index+1;
  // Figure out the filename where the information has been saved.
  char simple_name[100];
  sprintf(simple_name, "basis%lib.%li", basis, index);
  char * filename = vJoinPaths("g:\\users\\athitsos\\sequences\\glove_data\\training\\gw2003",
                                simple_name);
  // Read the information about the reference images.
  v3dMatrix<double> * basis_info = 
    v3dMatrix<double>::ReadText(filename);
  // If we failed to read the information, return.
  if (basis_info == 0)
  {
    vPrint("Error in reading %s\n", filename);
    vdelete2(filename);
    return 0;
  }
  vdelete2(filename);

  long rows = basis_info->Rows();
  long cols = basis_info->Cols();
  if ((rows < dimensions) || ((cols != 5) && (cols != 3)))
  {
    vPrint("Bad basis info\n");
    vdelete(basis_info);
    return 0;
  }

  vArray2(double) info = basis_info->Matrix2(0);

  v3dMatrix<float> * reference_images = new v3dMatrix<float>(dimensions, 4, 1);
  vArray2(float) references = reference_images->Matrix2(0);

  long i;
  for (i = 0; i < dimensions; i++)
  {
    long frame, view;
    // Get info on the i-th reference image.
    if (cols == 5)
    {
      frame = vRound(info[i][1]);
      view = vRound(info[i][2]);
      references[i][0] = (float) frame;
      references[i][1] = (float) view;
      references[i][2] = (float) 1.;
      references[i][3] = (float) (frame * 4 + view);
    }
    else
    {
      frame = vRound(info[i][0]);
      view = vRound(info[i][1]);
      references[i][0] = (float) frame;
      references[i][1] = (float) view;
      references[i][2] = (float) info[i][2];
    }
  }

  *reference_images_p = reference_images;
  vdelete(basis_info);

  return 1;
}


// This function is similar to vCalculateSTDs (see comments there).
// The difference is that here reference images are chosen based
// on a different criterion: the worst case distortion. We pick
// as the next reference image the one that, together with the 
// already chosen reference images, achieves the smallest worst
// case distortion, defined as the ratio between the largest and
// the smallest scaling observed by looking at ratios of
// approximate chamfer distances over exact chamfer distances, measured
// for images in cd_pairs.
vMatrix<double> vCalculateDistortions(v3dMatrix<uchar> * all_cd_distances,
                                        v3dMatrix<long> * cd_pairs,
                                        v3dMatrix<double> * cd_pair_distances,
                                        v3dMatrix<double> * estimated_distances,
                                        long factor)
{
  long candidates = all_cd_distances->Rows() * 4;
  vMatrix<double> distortions_m = vMatrix<double>(2, candidates);
  vArray2(double) distortions = distortions_m.Matrix2();
  double mean, distortion;
  mean = distortion = 0;
  long i;
  vPrint("\n");
  for (i = 0; i < candidates; i++)
  {
    vCalculateDistortion(all_cd_distances, cd_pairs, cd_pair_distances,
                          estimated_distances, i, &mean, &distortion, factor);
    distortions[0][i] = mean;
    distortions[1][i] = distortion;
    vPrint("Processed candidate %li of %li\r", i, candidates);
  }
  vPrint("\n");

  return distortions_m;
}


// Similar to vCalculateSTD (see comments), except that here we 
// compute worst-case distortion as opposed to average distortion.
// See comments on vCalculateDistortions.
long vCalculateDistortion(v3dMatrix<uchar> * all_cd_distances,
                           v3dMatrix<long> * cd_pairs,
                           v3dMatrix<double> * cd_pair_distances,
                           v3dMatrix<double> * estimated_distances,
                           long candidate, double * meanp, double * distortionp,
                           long factor)
{
  if ((candidate < 0) || (candidate >= 86 * 26 * 4))
  {
    return 0;
  }

  vArray2(uchar) cd_distances = all_cd_distances->Matrix2(0);
  vArray2(long) pairs = cd_pairs->Matrix2(0);
  vArray(double) pair_distances = cd_pair_distances->Matrix();
  vArray(double) estimated = estimated_distances->Matrix();
  long number = cd_pairs->Cols() / factor;
  long frame = (candidate / 4) * 12;
  long view = candidate % 4;

  vMatrix<double> candidate_distances_m = vGet_CD_Distances(frame, view, all_cd_distances);
  vMatrix<double> new_distances_m(1, number);
  vMatrix<double> total_distances_m(1, number);
  vMatrix<double> fraction_distances_m(1, number);
  vMatrix<double> scaled_distances_m(1, number);

  vArray(double) candidate_distances = candidate_distances_m.Matrix();
  vArray(double) new_distances = new_distances_m.Matrix();
  vArray(double) total_distances = total_distances_m.Matrix();
  vArray(double) fraction_distances = fraction_distances_m.Matrix();

  long i;
  // Get new estimated distances
  long counter = 0; // Will count pairs for which actual_distance != 0.
  double total = 0.0;
  for (i = 0; i < number; i++)
  {
    long frame1 = pairs[0][i];
    long view1 = pairs[1][i];
    long index1 = frame1 * 4 + view1;

    long frame2 = pairs[2][i];
    long view2 = pairs[3][i];
    long index2 = frame2 * 4 + view2;

    double distance1 = candidate_distances[index1];
    double distance2 = candidate_distances[index2];
    new_distances[i] = vAbs(distance1 - distance2);
    total_distances[i] = estimated[i] + new_distances[i];
    double actual_distance = pair_distances[i];
    if (actual_distance == 0)
    {
      fraction_distances[i] = -1;
    }
    else
    {
      fraction_distances[i] = total_distances[i] / actual_distance;
      counter++;
      total += fraction_distances[i];
    }
  }

  if ((counter == 0) || (total == 0))
  {
    vPrint("Error: counter or total = 0\n");
    return 0;
  }

  double mean = total / (double) counter;
  if (mean == 0)
  {
    vPrint("Error: mean = 0\n");
    return 0;
  }


  // Calculate the worst-case distortion, by finding the maximum
  // and minimum scaling attained using Lipschitz embeddings with
  // the current set of reference images (already chosen reference
  // images plus the current candidate).
  double min_scaling = 1000000000;
  double max_scaling = 0;
  double limit = 60 / 255.0;
  for (i = 0; i < number; i++)
  {
    double current_scaling = fraction_distances[i];
    if (current_scaling < -0.5) continue;
    if (current_scaling <= 0) current_scaling = limit;
    if (current_scaling < min_scaling) min_scaling = current_scaling;
    if (current_scaling > max_scaling) max_scaling = current_scaling;
  }

  if (min_scaling == 0) return 0;
  double distortion = max_scaling / min_scaling;
  *meanp = mean;
  *distortionp = distortion;

  return 1;
}


// Choose basis elements in a greedy way, to minimize worst-case distortion.
// This function is almost identical to vChooseBasis1, but it calls
// vCalculateDistortions as opposed to vCalculateSTDs.
vMatrix<double> vChooseBasis3(v3dMatrix<uchar> * all_cd_distances,
                                v3dMatrix<long> * cd_pairs,
                                v3dMatrix<double> * cd_pair_distances,
                                long factor, long number)
{
  if ((number <= 0) || (factor <= 0)) return vMatrix<double>();

  long rows = cd_pair_distances->Rows();
  long cols = cd_pair_distances->Cols() / factor;
  vMatrix<double> estimated_distances(1, cols);
  vWriteValue(&estimated_distances, (double) 0);
  
  vMatrix<double> result_matrix(number, 5);
  vWriteValue(&result_matrix, (double) 0);
  vArray2(double) result = result_matrix.Matrix2();

  long i;
  for (i = 0; i < number; i++)
  {
    vMatrix<double> distortions =
      vCalculateDistortions(all_cd_distances, cd_pairs, cd_pair_distances,
                             &estimated_distances, factor);

    // Write over entries corresponding to images already chosen for the basis,
    // so that they will not get chosen twice.
    long j;
    for (j = 0; j < i; j++)
    {
      long frame = (long) result[j][1];
      long view = (long) result[j][2];
      long index = (frame / 12) * 4 + view;
      distortions(1, index) = 1000000000;
    }

    long min_distortion_index = 0;
    double min_distortion = color_imageRowMin3(&distortions, 1, &min_distortion_index);
    double mean = distortions(0, min_distortion_index);
    long frame = (min_distortion_index / 4) * 12;
    long view = (min_distortion_index % 4);
    vPrint("%li: image (%li, %li), mean[%li] = %lf, distortion = %lf\n", 
            i, frame, view, min_distortion_index, mean, min_distortion);
    result[i][0] = i;
    result[i][1] = frame;
    result[i][2] = view;
    result[i][3] = mean;
    result[i][4] = min_distortion;
    char simple_name[1000];
    sprintf(simple_name, "basis3b.%li", i);
    char * filename1 = vJoinPaths("g:\\users\\athitsos\\sequences\\glove_data\\training\\gw2003",
                                   simple_name);
    char * filename2 = vJoinPaths("r:\\text\\experiments\\gw2003",
                                   simple_name);
    long success1 = result_matrix.WriteText(filename1);
    long success2 = result_matrix.WriteText(filename2);
    vPrint("Success1 = %li, success2 = %li\n", success1, success2);
    vdelete2(filename1);
    vdelete2(filename2);

    // Update estimated distances;
    vUpdateEstimatedDistances(all_cd_distances, cd_pairs, 
                               &estimated_distances, frame, view, factor);
  }

  return result_matrix;
}


// Choose basis elements randomly. This function is almost identical to 
// vChooseBasis2. The only difference is that here, for each random 
// basis, we print out the worst-case distortion achieved by that basis,
// whereas in vChooseBasis2 we print out the average distortion achieved
// by that basis. The result is the same, but different information is
// displayed on the screen and saved on files, that is useful for 
// comparing the worst-case distortion attained using random reference
// images, versus using reference images selected to minimize the 
// worst-case distortion.
vMatrix<double> vChooseBasis4(v3dMatrix<uchar> * all_cd_distances,
                                v3dMatrix<long> * cd_pairs,
                                v3dMatrix<double> * cd_pair_distances,
                                long factor, long number)
{
  if ((number <= 0) || (factor <= 0)) return vMatrix<double>();

  long rows = cd_pair_distances->Rows();
  long cols = cd_pair_distances->Cols() / factor;
  vMatrix<double> estimated_distances(1, cols);
  vWriteValue(&estimated_distances, (double) 0);
  
  vMatrix<double> result_matrix(number, 5);
  vWriteValue(&result_matrix, (double) 0);
  vArray2(double) result = result_matrix.Matrix2();

  long i;
  for (i = 0; i < number; i++)
  {
    vMatrix<double> distortions =
      vCalculateDistortions(all_cd_distances, cd_pairs, cd_pair_distances,
                             &estimated_distances, factor);

    long index = vRandomLong2(0, 86 * 26 * 4);
    double mean = distortions(0, index);
    double distortion = distortions(1, index);
    long frame = (index / 4) * 12;
    long view = (index % 4);
    vPrint("%li: image (%li, %li), mean[%li] = %lf, distortion = %lf\n", 
            i, frame, view, index, mean, distortion);
    result[i][0] = i;
    result[i][1] = frame;
    result[i][2] = view;
    result[i][3] = mean;
    result[i][4] = distortion;
    char simple_name[1000];
    sprintf(simple_name, "basis4.%li", i);
    char * filename1 = vJoinPaths("g:\\users\\athitsos\\sequences\\glove_data\\training\\gw2003",
                                   simple_name);
    char * filename2 = vJoinPaths("r:\\text\\experiments\\gw2003",
                                   simple_name);
    long success1 = result_matrix.WriteText(filename1);
    long success2 = result_matrix.WriteText(filename2);
    vPrint("Success1 = %li, success2 = %li\n", success1, success2);
    vdelete2(filename1);
    vdelete2(filename2);

    // Update estimated distances;
    vUpdateEstimatedDistances(all_cd_distances, cd_pairs, 
                               &estimated_distances, frame, view, factor);
  }

  return result_matrix;
}


// This function is useful for the method of picking reference images
// that works by picking the reference images that maximize classification
// accuracy on a set of real hand images. test_images is the video sequence
// containing the real hand images, and database_images contains the
// edge pixels of all database images. All chamfer distances between
// each real hand image and each candidate reference image are computed.
// Note that, despite the name of the function, the chamfer distances are
// not saved on disk here, just stored in memory, in the result matrix.
vMatrix<float> vSaveReferenceDistances(vVideoFile * test_images,
                                         vPointSetSequence * database_images)
{
  const long pick_number = 86*26*4;
  long test_frames = test_images->NumberOfFrames();
  vMatrix<long> frames(1, pick_number);
  vMatrix<long> views(1, pick_number);
  vMatrix<float> result(test_frames, pick_number);

  // Fill in the frames that are possible candidates for reference images.
  // For memory and time efficiency, not every hand pose represented in
  // the database is a candidate. Only four images per shape/viewpoint
  // combination are candidates (which are rotated versions of one image,
  // by 0, 90, 180 and 270 degrees).
  long i;
  for (i = 0; i < pick_number; i++)
  {
    long frame = (i / 4) * 12;
    long view = (i % 4);
    frames(0, i) = frame;
    views(0, i) = view;
  }

  // Now, for each test image, calculate and fill in the chamfer distances
  for (i = 0; i < test_frames; i++)
  {
    grayscale_image * the_image = test_images->ReadFrameGray(i);
    vMatrix<double> distances = vDatabasePicksChamfer(the_image, database_images, 
                                                        &frames, &views, 0, pick_number, 
                                                        -1, -1);
    long j;
    for (j = 0; j < pick_number; j++)
    {
      result(i, j) = (float) distances(j);
    }
    // Print out some information so that the user can see
    // how far along we are in the computation.
    vPrint("result(%li, %li) = %f\n", i, 0, result(i, 0));
    vPrint("result(%li, 5000) = %f\n", i, result(i, 5000));
    vPrint("result(%li, %li) = %f\n", i, pick_number-1, result(i, pick_number-1));
    vdelete(the_image);
  }
  return result;
}


// This function calculates the median of the highest ranking correct match
// attained by combining each candidate new reference image with the set
// of already chosen reference images. For explanations on some of the
// arguments see comments on vCalculateSTDs. The rest of the arguments are:
// reference_distances: precomputed chamfer distances between the set
// of real hand images and all candidate images.
// ground_distances: ground truth distances between each test image (real hand
// image) and each database image. This is a binarized matrix: the ground
// truth distance is 0 if the corresponding match is correct, and 1 if
// the corresponding match is incorrect.
// current_candidates: a subset of all candidate reference images. We don't
// consider all reference images every time, because it would be too slow.
// frames_to_use: number of test images to use in picking each new reference
// image. Again, it may be too slow to consider all test images all the time.
vMatrix<double> vCalculateAccuracies(v3dMatrix<uchar> * all_cd_distances,
                                       v3dMatrix<float> * reference_distances,
                                       v3dMatrix<ushort> * ground_distances,
                                       v3dMatrix<double> * estimated_distances,
                                       vector<long> * current_candidates,
                                       long frames_to_use)
{
  long test_number = reference_distances->Rows();
  if (frames_to_use > test_number) return 0;
  long candidates = current_candidates->size();
  vMatrix<double> accuracies_m = vMatrix<double>(2, candidates);
  vArray2(double) accuracies = accuracies_m.Matrix2();
  double accuracy = 0;
  long i;
  vPrint("\n");

  // Choose a random set of test frames
  vector<long> test_frames;
  vPickWithoutRepetitions2(0, test_number - 1, frames_to_use, &test_frames);

  for (i = 0; i < candidates; i++)
  {
    long candidate = (*current_candidates)[i];
    vCalculateAccuracy(all_cd_distances, reference_distances, ground_distances,
                        estimated_distances, candidate, &accuracy, &test_frames);
    accuracies[0][i] = 0;
    accuracies[1][i] = accuracy;
    vPrint("Processed candidate %li of %li\r", i, candidates);
  }
  vPrint("\n");

  return accuracies_m;
}


// Function similar to vCalculateSTD, but here we compute for each
// candidate reference image the classification accuracy attained
// by adding that candidate to the set of previously chosen reference
// images. See comments on vCalculateAccuracies. frames_to_use are
// a subset of all test frames.
long vCalculateAccuracy(v3dMatrix<uchar> * all_cd_distances,
                         v3dMatrix<float> * reference_distances,
                         v3dMatrix<ushort> * ground_distances,
                         v3dMatrix<double> * estimated_distances,
                         long candidate, double * accuracyp,
                         vector<long> * frames_to_use)
{
  if ((candidate < 0) || (candidate >= 86 * 26 * 4))
  {
    return 0;
  }

  long database_images = all_cd_distances->Cols();

  vArray2(uchar) cd_distances = all_cd_distances->Matrix2(0);
  vArray2(float) references = reference_distances->Matrix2(0);
  vArray2(ushort) ground = ground_distances->Matrix2(0);
  vArray2(double) estimated = estimated_distances->Matrix2(0);
  long frame = (candidate / 4) * 12;
  long view = candidate % 4;

  vMatrix<double> candidate_distances_m = vGet_CD_Distances(frame, view, all_cd_distances);
  vMatrix<double> total_distances_m(frames_to_use->size(), database_images);
  vMatrix<double> min_correct_distances(1, frames_to_use->size());
  vWriteValue(&min_correct_distances, (double) 1000000000);

  vArray(double) candidate_distances = candidate_distances_m.Matrix();
  vArray2(double) total_distances = total_distances_m.Matrix2();
  vArray(double) min_correct = min_correct_distances.Matrix();

  long i;
  // Get the distances for each test frame and each database image.
  long counter = 0; // Will count pairs for which actual_distance != 0.
  double total = 0.0;
  for (i = 0; i < (long) frames_to_use->size(); i++)
  { 
    long test_frame = (*frames_to_use)[i];
    double distance1 = references[test_frame][candidate];
    long j;
    for (j = 0; j < database_images; j++)
    {
      double distance2 = candidate_distances[j];
      double new_distance = vAbs(distance1 - distance2);
      // total_distance is the approximate chamfer distance between
      // the test image and the database image, using the set
      // of reference images consisting of the previously chosen
      // images and the current candidate.
      double total_distance = estimated[test_frame][j] + new_distance;
      total_distances[i][j] = total_distance;
      // Check if this database image is the closest correct match
      // seen so far.
      if (ground[test_frame][j] == 0)
      {
        if (total_distance < min_correct[i])
        {
          min_correct[i] = total_distance;
        }
      }
    }
  }

  // calculate accuracy. For each test frame under consideration,
  // count how many database images have scores less than or 
  // equal to the best score by a correct match. This is equivalent
  // to computing the rank of the highest ranking correct match.
  vector<long> counters(frames_to_use->size());
  for (i = 0; i < (long) frames_to_use->size(); i++)
  {
    counters[i] = 0;
    double min_distance = min_correct[i];
    long j;
    for (j = 0; j < database_images; j++)
    {
      if (total_distances[i][j] <= min_distance)
      {
        counters[i]++;
      }
    }
  }

  // Sort the ranks, and find the median rank of the highest
  // ranking correct matches.
  long median_index = frames_to_use->size() / 2;
  std::sort(counters.begin(), counters.end(), less<long>());
  double accuracy = counters[median_index];
  *accuracyp = accuracy;
  return 1;
}


// Choose basis elements in a greedy way, to minimize rank
// of highest ranking correct match. See comments on
// vChooseBasis1 and vCalculateAccuracies.
vMatrix<double> vChooseBasis5(v3dMatrix<uchar> * all_cd_distances,
                                v3dMatrix<float> * reference_distances,
                                const char * filename,
                                long frames_to_use, long number)
{
  long candidate_limit = 20;
  if (candidate_limit > all_cd_distances->Rows() * 4)
  {
    candidate_limit = all_cd_distances->Rows() * 4;
  }
  v3dMatrix<ushort> * ground_distances = vPrepareGroundDistances(filename);
  if (ground_distances == 0) return 0;

  if ((frames_to_use > ground_distances->Rows()) || 
      (number <= 0)) 
  {
    vdelete(ground_distances);
    return vMatrix<double>();
  }

  long rows = ground_distances->Rows();
  long cols = ground_distances->Cols();
  vMatrix<double> estimated_distances(rows, cols);
  vWriteValue(&estimated_distances, (double) 0);
  
  vMatrix<double> result_matrix(number, 5);
  vWriteValue(&result_matrix, (double) 0);
  vArray2(double) result = result_matrix.Matrix2();

  long i;
  for (i = 0; i < number; i++)
  {
    // Choose a set of current candidates
    vector<long> current_candidates;
    vPickWithoutRepetitions2(0, all_cd_distances->Rows() * 4, 
                              candidate_limit, &current_candidates);

    
    vMatrix<double> accuracies =
      vCalculateAccuracies(all_cd_distances, reference_distances, ground_distances,
                            &estimated_distances, &current_candidates, frames_to_use);

    long min_median_index = 0;
    double min_median = color_imageRowMin3(&accuracies, 1, &min_median_index);
    long candidate_index = current_candidates[min_median_index];
    long frame = (candidate_index / 4) * 12;
    long view = (candidate_index % 4);
    vPrint("%li: image (%li, %li), median = %lf\n", 
            i, frame, view, min_median);
    result[i][0] = i;
    result[i][1] = frame;
    result[i][2] = view;
    result[i][3] = 0;
    result[i][4] = min_median;
    char simple_name[1000];
    sprintf(simple_name, "basis5b.%li", i);
    char * filename1 = vJoinPaths("g:\\users\\athitsos\\sequences\\glove_data\\training\\gw2003",
                                   simple_name);
    char * filename2 = vJoinPaths("r:\\text\\experiments\\gw2003",
                                   simple_name);
    long success1 = result_matrix.WriteText(filename1);
    long success2 = result_matrix.WriteText(filename2);
    vPrint("Success1 = %li, success2 = %li\n", success1, success2);
    vdelete2(filename1);
    vdelete2(filename2);

    // Update estimated distances;
    vUpdateEstimatedDistances2(all_cd_distances, reference_distances, 
                                &estimated_distances, frame, view);
  }

  vdelete(ground_distances);
  return result_matrix;
}


// This function is the equivalent of vUpdateEstimatedDistances for 
// the method of choosing reference images that relies on the classification
// accuracy on a test set of real images. As opposed to keeping track of the
// approximate chamfer distances between database images, here we must keep
// track of the distances between test images and database images.
long vUpdateEstimatedDistances2(v3dMatrix<uchar> * all_cd_distances,
                                 v3dMatrix<float> * reference_distances,
                                 v3dMatrix<double> * estimated_distances,
                                 long frame, long view)
{
  vArray2(double) estimated = estimated_distances->Matrix2(0);
  long test_number = reference_distances->Rows();
  long database_number = all_cd_distances->Cols();

  vMatrix<double> candidate_distances_m = vGet_CD_Distances(frame, view, all_cd_distances);
  vArray(double) candidate_distances = candidate_distances_m.Matrix();
  vArray2(float) references = reference_distances->Matrix2(0);
  long index = (frame / 12) * 4 + view;

  long i;
  for (i = 0; i < test_number; i++)
  {
    double distance1 = references[i][index];
    long j;
    for (j = 0; j < database_number; j++)
    {
      double distance2 = candidate_distances[j];
      double distance = vAbs(distance1 - distance2);
      estimated[i][j] += distance;
    }
  }

  return 1;
}


// Prepares a ground truth matrix that is appropriate to use in
// vCalculateAccuracies, which is used by vChooseBasis5. The 
// ground truth matrix is binarized, to be 0 if a match is correct
// and 1 if a match is incorrect.
v3dMatrix<ushort> * vPrepareGroundDistances(const char * filename)
{
  char * ground_truth_file = vReplaceEnding2(filename, ".gt");
  vector<vViewGroundTruth> * ground_truth =
    vViewGroundTruth::Load(ground_truth_file);
  vdelete2(ground_truth_file);
  if (ground_truth == 0) 
  {
    return 0;
  }

  char * ground_distances_file = vReplaceEnding2(filename, ".gt_distances");
  v3dMatrix<ushort> * ground_distances_m = v3dMatrix<ushort>::Read(ground_distances_file);
  vArray2(ushort) ground_distances = ground_distances_m->Matrix2(0);
  vdelete2(ground_distances_file);
  if (vIsZero(ground_distances))
  {
    exit_error("ground distances file not found\n");
  }

  long rows = ground_truth->size();
  long cols = ground_distances_m->Cols() * 26;
  v3dMatrix<ushort> * result = new v3dMatrix<ushort>(rows, cols, 1);
  vArray2(ushort) binary = result->Matrix2(0);
  vWriteValue(result, (ushort) 1);
  long i, j;
  for (i = 0; i < rows; i++)
  {
    long pose = (*ground_truth)[i].training_pose_index;
    long start = pose * 86 * 48;
    for (j = 0; j < ground_distances_m->Cols(); j++)
    {
      if (ground_distances[i][j] <= 300)
      { 
        binary[i][start+j] = 0;
      }
    }
  }

  vdelete(ground_truth);
  vdelete(ground_distances_m);
  return result;
}

// End of functions used for the GW2003 paper.



/////////////////////////////////////////////////////////////////////
//                                                                 //
// Start of implementation of vDatabaseEdgeSummary and subclasses //
//                                                                 //
/////////////////////////////////////////////////////////////////////


char * vDatabaseEdgeSummary::DT_Filename(const char * filename)
{
  char * result = vReplaceEnding3(filename, "_edges.rle", "_dt.avi");
  return result;
}


char * vDatabaseEdgeSummary::PointsFilename(const char * filename)
{
  char * result = vReplaceEnding3(filename, "_edges.rle", ".points");
  return result;
}


long vDatabaseEdgeSummary::Load(const char * filename)
{
  // Load DTs.
  
  char * dt_name = DT_Filename(filename);
  vVideoFile * input = vOpenReadVideoFile(dt_name);
  if (input == 0) 
  {
    vdelete2(dt_name);
    vPrint("Error in loading dts\n");
    return 0;
  }
  input->SetBounds(18, 18, 220, 220);

  number_of_frames = input->NumberOfFrames();
  dts = vnew(grayscale_image *, number_of_frames);
  long frame;
  long success = 1;
  for (frame = 0; frame < number_of_frames; frame++)
  {
    dts[frame] = input->ReadFrameGray(frame);
    if (dts[frame] == 0)
    {
      vPrint("Error in loading dt %li\n", frame);
      success = 0;
      break;
    }
  }
  vCloseVideoFile(input);
  vdelete2(dt_name);

  // Load points.
  char * points_name = PointsFilename(filename);
  FILE * fp = fopen(points_name, vFOPEN_READ);
  if (fp == 0)
  {
    vPrint("Error in opening %s\n", points_name);
    vdelete2(points_name);
    return -2;
  }

  long success2 = 1;
  long items;
  long number_of_frames2 = 0;
  items = vReadPCLongs(fp, &number_of_frames2, 1);
  if (items != 1) success2 = -2;
  if (number_of_frames2 != number_of_frames) success2 = -2;

  lengths = vnew(long, number_of_frames2);
  rows = vnew(vArray(uchar), number_of_frames2);
  cols = vnew(vArray(uchar), number_of_frames2);

  items = vReadPCLongs(fp, vData(lengths), number_of_frames2);
  if (items != number_of_frames2) success2 = -2;
  for (frame = 0; frame < number_of_frames2; frame++)
  {
    long length = lengths[frame];
    rows[frame] = vnew(uchar, length);
    cols[frame] = vnew(uchar, length);
    items = fread(actual_pointer (rows[frame]), sizeof(uchar), length, fp);
    if (items != length) success2 = -2;
    items = fread(actual_pointer (cols[frame]), sizeof(uchar), length, fp);
    if (items != length) success2 = -2;
  }

  fclose(fp);
  vdelete2(points_name);
  if (success2 <= 0)
  {
    vPrint("Failed to load points successfully\n");
  }

  if ((success > 0) && (success2 > 0)) 
  {
    is_valid = 1;
    return 1;
  }
  else return success + success2;
}


long vDatabaseEdgeSummary::Initialize()
{
  dts = function_zero (grayscale_image*);
  rows = vZero(vArray(uchar));
  cols = vZero(vArray(uchar));
  number_of_frames = 0;
  lengths = vZero(long);
  is_valid = 0;
  return 1;
}


vDatabaseEdgeSummary::vDatabaseEdgeSummary()
{
  Initialize();
}


vDatabaseEdgeSummary::vDatabaseEdgeSummary(const char * filename)
{
  Initialize();
  Load(filename);
}


vDatabaseEdgeSummary::vDatabaseEdgeSummary(vVideoFile * sequence) 
{
  Initialize();
  const char * filename = sequence->GetFilename();
  Load(filename);
}


vDatabaseEdgeSummary::~vDatabaseEdgeSummary()
{
  vdelete2(lengths);
  long i;
  for (i = 0; i < number_of_frames; i++)
  {
    vdelete(dts[i]);
    vdelete2(rows[i]);
    vdelete2(cols[i]);
  }
  vdelete2(dts);
  vdelete2(rows);
  vdelete2(cols);
}


long vDatabaseEdgeSummary1::Initialize()
{
  basic_shapes = 26;
  viewpoints = 86;
  saved_orientations = 12;
  orientations = 48;
  views_per_frame = 4;

  long index = 0;
  long frame, view;
  for (frame = 0; frame < saved_orientations; frame++)
  {
    for (view = 0; view < views_per_frame; view++)
    {
      double steps = frame - view * 12;
      if (steps < 0) steps += 48;
      double rotation = steps / 48.0 * 2 * M_PI;
      rotations[index] = steps;
      rotation_templates[index] = vRotationTemplateR(220, 220, 109, 109, rotation);
      index++;
      vPrint("frame = %li, view = %li, rotation = %.2lf\n",
               frame, view, rotation * 180 / M_PI);
    }
  }

  return 1;
}


vDatabaseEdgeSummary1::vDatabaseEdgeSummary1()
{
  Initialize();
}


vDatabaseEdgeSummary1::vDatabaseEdgeSummary1(const char * filename) :
vDatabaseEdgeSummary(filename)
{
  Initialize();
}

vDatabaseEdgeSummary1::vDatabaseEdgeSummary1(vVideoFile * sequence) :
vDatabaseEdgeSummary(sequence)
{
  Initialize();
}


vDatabaseEdgeSummary1::~vDatabaseEdgeSummary1()
{
  long i;
  for (i = 0; i < 48; i++)
  {
    vdelete(rotation_templates[i]);
  }
}


// Saves DTs.
long vDatabaseEdgeSummary1::SaveDTs(vVideoFile * sequence)
{
  exit_error("\nerror: uses NormalizeMEC3, must be reimplemented\n");

  const char * edges_name = sequence->GetFilename();
  char * dt_name = vReplaceEnding3(edges_name, "_edges.rle", "_dt.avi");
  vVideoFile * output = vOpenWriteVideoFile(dt_name, sequence->Rows(), 
                                              sequence->Cols(), 30, 1, 1);

  long frames = sequence->NumberOfFrames();
  long frame;
  vPrint("\n");
  long success = 1;
  for (frame = 0; frame < frames; frame+=12)
  {
    grayscale_image * pre_edges = sequence->ReadFrameGray(frame);
    if (pre_edges == 0) continue;
//    grayscale_image * edges = (grayscale_image *) wrma(pre_edges).NormalizeMEC3(127, 127, 96);
    vdelete(pre_edges);
//    grayscale_image * dt = vDistanceTransformUchar(edges);
//    success = output->write_frameGray(dt);
//    vdelete(edges);
//    vdelete(dt);
    if (success <= 0)
    {
      vPrint("Failed to save frame %li\n", frame);
    }
    else
    {
      vPrint("Saved frame %li\r", frame);
    }
  }
  vPrint("\n");
  output->FinishWriting();
  vCloseVideoFile(output);
  vdelete2(dt_name);
  if (success <= 0) return -1;
  else return 1;
}


// Saves Points.
long vDatabaseEdgeSummary1::SavePoints(vVideoFile * sequence)
{
  exit_error("\nerror: uses NormalizeMEC3, must be reimplemented\n");

  const char * edges_name = sequence->GetFilename();
  char * points_name = PointsFilename(edges_name);
  FILE * fp = fopen(points_name, vFOPEN_WRITE);
  if (fp == 0)
  {
    vPrint("Error in opening %s\n", points_name);
    vdelete2(points_name);
    return -2;
  }

  long frames = sequence->NumberOfFrames();
  long sampled_frames = frames/12;
  vArray(long) seq_lengths = vnew(long, sampled_frames);
  vArray2(uchar) seq_rows = vnew(vArray(uchar), sampled_frames);
  vArray2(uchar) seq_cols = vnew(vArray(uchar), sampled_frames);
  long frame;
  vPrint("\n");
  long success = 1;
  for (frame = 0; frame < sampled_frames; frame++)
  {
    grayscale_image * pre_edges = sequence->ReadFrameGray(frame*12);
    if (pre_edges == 0) continue;
//    grayscale_image * edges = (grayscale_image *) wrma(pre_edges).NormalizeMEC3(127, 127, 96);
    vdelete(pre_edges);

    vector<long> rows, cols;
//    vGetNonZeroPoints(edges, &rows, &cols);
    long length = rows.size();
    seq_lengths[frame] = length;
    seq_rows[frame] = vnew(uchar, length);
    seq_cols[frame] = vnew(uchar, length);
    long i;
    for (i = 0; i < length; i++)
    {
      // We subtract 18, because when we read DTs and points
      // we essentially crop the 18-thick outer region of the 
      // image.
      seq_rows[frame][i] = (uchar) rows[i] - 18;
      seq_cols[frame][i] = (uchar) cols[i] - 18;
    }
    vPrint("Processed frame %li\r", frame);
  }
  long items;
  items = vWritePCLongs(fp, &sampled_frames, 1);
  if (items != 1) success = -2;
  items = vWritePCLongs(fp, vData(seq_lengths), sampled_frames);
  if (items != sampled_frames) success = -2;
  vPrint("\n");
  for (frame = 0; frame < sampled_frames; frame++)
  {
    long length = seq_lengths[frame];
    items = fwrite(actual_pointer (seq_rows[frame]), sizeof(uchar), length, fp);
    if (items != length) success = -2;
    items = fwrite(actual_pointer (seq_cols[frame]), sizeof(uchar), length, fp);
    if (items != length) success = -2;
    vdelete2(seq_rows[frame]);
    vdelete2(seq_cols[frame]);
    vPrint("Saved frame %li\r", frame);
  }
  vPrint("\n");

  vdelete2(seq_rows);
  vdelete2(seq_cols);
  vdelete2(seq_lengths);
  fclose(fp);
  vdelete2(points_name);
  if (success <= 0)
  {
    vPrint("Failed to save points successfully\n");
  }
  return success;
}


vMatrix<double> vDatabaseEdgeSummary1::ChamferDistances(grayscale_image * edge_image)
{
  long number_of_scores = number_of_frames * saved_orientations * views_per_frame;
  vMatrix<double> distances(1, number_of_scores);
  vArray(double) distances1 = distances.Matrix();
  grayscale_image * dt_array[48];
  vector<long> rows_array[48];
  vector<long> cols_array[48];

  PrepareForChamfer(edge_image, dt_array, rows_array, cols_array);
  long i;

  for (i = 0; i < number_of_scores; i+= 48)
  {
    long database_index = i / 48;
    grayscale_image * database_dt = dts[database_index];
    vArray(uchar) database_rows = rows[database_index];
    vArray(uchar) database_cols = cols[database_index];
    long length = lengths[database_index];

    long j;
    for (j = 0; j < 48; j++)
    {
      grayscale_image * input_dt = dt_array[j];
      vector<long> & input_rows = rows_array[j];
      vector<long> & input_cols = cols_array[j];
      double distance1 = vChamferFromDT(&input_rows, &input_cols, database_dt);
      double distance2 = vChamferFromDT(database_rows, database_cols, length,
                                           input_dt);
      distances1[i+j] = distance1 + distance2;
    }
  }


  for (i = 0; i < 48; i++)
  {
    vdelete(dt_array[i]);
  }
  return distances;
}

vMatrix<double> vDatabaseEdgeSummary1::ChamferDistances(grayscale_image * edge_image,
                                                          vector<long> * frames,
                                                          vector<long> * views)
{
  return vMatrix<double>();
}
  

// Extracts the distance transform of the_image, rotates it
// 48 times, and also extracts the non-zero pixel locations
// and rotates them 48 times. The image plane rotations
// appear in the result vectors in the order in which
// the match scores are supposed to appear (4 per database frame).
long vDatabaseEdgeSummary1::PrepareForChamfer(grayscale_image * the_image, 
                                               grayscale_image ** dt_array,
                                               vector<long> * rows_array, 
                                               vector<long> * cols_array)
{
  exit_error("\nerror: uses NormalizeMEC3, must be reimplemented\n");

//  long i;
//  grayscale_image * normalized = 
//    (grayscale_image *) wrma(the_image).NormalizeMEC3_a(109, 109, 96);

 ///* grayscale_image * cropped = 
 //   (grayscale_image *) vCopyRectangle(normalized, 0, 219, 0, 219);

 // vdelete(normalized);

 // grayscale_image * dt1 = vDistanceTransformUchar(cropped);
 // for (i = 0; i < 48; i++)
 // {
 //   v3dMatrix<short> * sources = rotation_templates[i];
 //   dt_array[i] = (grayscale_image *) wrma(dt1).Warp_a(sources);
 //   grayscale_image * rotated = (grayscale_image *) wrma(cropped).Warp_a(sources);
 //   vGetNonZeroPoints(rotated, &(rows_array[i]), &(cols_array[i]));
 //   vdelete(rotated);
 // }

 // vdelete(dt1);
 // vdelete(cropped);*/
  return 1;
}


// Returns the chamfer distance between two database poses,
// specified by frame and view.
double vDatabaseEdgeSummary1::PoseChamferDistance(long frame1, long view1, 
                                                   long frame2, long view2)
{
  exit_error("PoseChamferDistance: implementation not done yet\n");
  double steps1 = frame1 - view1 * 12;
  if (steps1 < 0) steps1 += 48;
  double rotation1 = steps1 / 48.0 * 2 * M_PI;
  long index1 = frame1 / 12;
  grayscale_image * dt1 = dts[index1];
  vArray(uchar) & pre_rows1 = rows[index1];
  vArray(uchar) & pre_cols1 = cols[index1];
  long length1 = lengths[index1];
  vector<uchar> rows1(length1);
  vector<uchar> cols1(length1);

  double steps2 = frame1 - view1 * 12;
  if (steps1 < 0) steps1 += 48;
  double rotation2 = steps1 / 48.0 * 2 * M_PI;
  long index2 = frame2 / 12;

  
  //vWarpPoints(pre_rows1, pre_cols1, &rows1, &cols1, rotation_templates[steps1];

  return 0;
}


/////////////////////////////////////////////////////////////////////
//                                                                 //
// End of implementation of vDatabaseEdgeSummary and subclasses   //
//                                                                 //
/////////////////////////////////////////////////////////////////////


// This function is a modified version of vChooseBasis, that is used
// to run multiple experiments comparing the method of minimizing
// average distortion versus the method of choosing images at random.
// The extra argument "shape" specifies which hand shape to use. Only
// database images of that hand shape are considered.
vMatrix<double> vChooseBasis6(v3dMatrix<uchar> * all_cd_distances,
                                v3dMatrix<long> * cd_pairs,
                                v3dMatrix<double> * cd_pair_distances,
                                long factor, long number, long shape)
{
  if ((number <= 0) || (factor <= 0)) return vMatrix<double>();
  if ((shape < 0) || (shape > 25)) return vMatrix<double>();

  // Create estimated_distances, that will store, at each point,
  // the L1 distances (i.e. approximate chamfer distances) between
  // images in cd_pairs that are obtained using the current set of reference
  // images.
  long rows = cd_pair_distances->Rows();
  long cols = cd_pair_distances->Cols() / factor;
  vMatrix<double> estimated_distances(1, cols);
  vWriteValue(&estimated_distances, (double) 0);

  // Create the result matrix, that will contain one row for each
  // reference image, specifying the row number, frame/view of the
  // reference image, and the mean scaling and average distortion
  // obtained by the reference images stored up to the current row.
  // The mean scaling is not really important, I was just curious
  // to see what it is, that's why it is stored here.
  vMatrix<double> result_matrix(number, 5);
  vWriteValue(&result_matrix, (double) 0);
  vArray2(double) result = result_matrix.Matrix2();

  long number_of_candidates = 86 * 4;
  long start = shape * number_of_candidates;

  long i;
  for (i = 0; i < number; i++)
  {
    // Calculate the average distortions for each candidate new
    // reference image.
    vMatrix<double> distortions =
      vCalculateSTDs6(all_cd_distances, cd_pairs, cd_pair_distances,
                       &estimated_distances, factor, start,
                       number_of_candidates);

    // Find the best candidate, i.e. the one that, combined with 
    // the already chosen reference images, achieves the lowest distortion.
    // That candidate will be the next chosen reference image.
    long min_std_index = 0;
    double min_std = color_imageRowMin3(&distortions, 1, &min_std_index);
    double mean = distortions(0, min_std_index);
    // Compute the frame/view info for the new reference image.
    long frame = (min_std_index / 4) * 12;
    long view = (min_std_index % 4);
    vPrint("%li: image (%li, %li), mean[%li] = %lf, std = %lf\n", 
            i, frame, view, min_std_index, mean, min_std);
    // Store information in the result about the new reference image.
    result[i][0] = i;
    result[i][1] = frame;
    result[i][2] = view;
    result[i][3] = mean;
    result[i][4] = min_std;
    char simple_name[1000];
    sprintf(simple_name, "basis6%lib.%li", shape, i);
    char * filename1 = vJoinPaths("g:\\users\\athitsos\\sequences\\glove_data\\training\\gw2003",
                                   simple_name);
    char * filename2 = vJoinPaths("r:\\text\\experiments\\gw2003",
                                   simple_name);
    // Output some useful information into two files, one in UNIX and the other
    // on the local PC.
    long success1 = result_matrix.WriteText(filename1);
    long success2 = result_matrix.WriteText(filename2);
    vPrint("Success1 = %li, success2 = %li\n", success1, success2);
    vdelete2(filename1);
    vdelete2(filename2);

    // Update estimated distances, to reflect the L1 distances (approximate
    // chamfer distances) between the images in each pair in cd_pairs, 
    // that are obtained using the reference images chosen so far (including
    // the newly chosen reference image.
    vUpdateEstimatedDistances(all_cd_distances, cd_pairs, 
                               &estimated_distances, frame, view, factor);
  }

  return result_matrix;
}


// This is like vCalculateSTDs, but modified to work with vChooseBasis6.
// Basically, here we can specify a range of candidates to be examined.
// Only a limited number of candidates are considered, and therefore
// we can run multiple experiments to see how minimizing average distortion
// compares to picking random images.
vMatrix<double> vCalculateSTDs6(v3dMatrix<uchar> * all_cd_distances,
                                  v3dMatrix<long> * cd_pairs,
                                  v3dMatrix<double> * cd_pair_distances,
                                  v3dMatrix<double> * estimated_distances,
                                  long factor,
                                  long start, long number)
{
  long candidates = all_cd_distances->Rows() * 4;
  vMatrix<double> distortions_m = vMatrix<double>(2, candidates);
  // By writing extremenly high values at first over all distortions,
  // we ensure that the minimum distortion will come from
  // the entries that are replaced by actual 
  // values, i.e. the entries that correspond to the candidates
  // specified by arguments "start" and "number".
  vWriteValue(&distortions_m, (double) 1000000000);
  vArray2(double) distortions = distortions_m.Matrix2();
  double mean, std;
  mean = std = 0;
  long i;
  vPrint("\n");
  // Calculate the mean scaling and average distortion achieved by
  // each candidate. The mean scaling is computed just out of curiosity,
  // it is not really used anywhere.
  for (i = start; i < start + number; i++)
  {
    vCalculateSTD(all_cd_distances, cd_pairs, cd_pair_distances,
                   estimated_distances, i, &mean, &std, factor);
    distortions[0][i] = mean;
    distortions[1][i] = std;
    vPrint("Processed candidate %li of %li\r", i, candidates);
  }
  vPrint("\n");

  return distortions_m;
}


// Choose reference images randomly, considering as candidate reference
// images only database images that correspond to the hand shape
// specified by the argument "shape" (shape is between 0 and 25).
vMatrix<double> vChooseBasis7(v3dMatrix<uchar> * all_cd_distances,
                                v3dMatrix<long> * cd_pairs,
                                v3dMatrix<double> * cd_pair_distances,
                                long number, long shape)
{
  if (number <= 0) return vMatrix<double>();
  if ((shape < 0) || (shape > 25)) return vMatrix<double>();

  vMatrix<double> result_matrix(number, 5);
  vWriteValue(&result_matrix, (double) 0);
  vArray2(double) result = result_matrix.Matrix2();
  long number_of_candidates = 86 * 4;
  long start = number_of_candidates * shape;
  long end = start + number_of_candidates - 1;

  long i;
  for (i = 0; i < number; i++)
  {
    long index = vRandomLong2(start, end);
    double mean = -1;
    double std = -1;
    long frame = (index / 4) * 12;
    long view = (index % 4);
    vPrint("%li: image (%li, %li), mean[%li] = %lf, std = %lf\n", 
            i, frame, view, index, mean, std);
    result[i][0] = i;
    result[i][1] = frame;
    result[i][2] = view;
    result[i][3] = mean;
    result[i][4] = std;
    char simple_name[1000];
    sprintf(simple_name, "basis7%lib.%li", shape, i);
    char * filename1 = vJoinPaths("g:\\users\\athitsos\\sequences\\glove_data\\training\\gw2003",
                                   simple_name);
    char * filename2 = vJoinPaths("r:\\text\\experiments\\gw2003",
                                   simple_name);
    long success1 = result_matrix.WriteText(filename1);
    long success2 = result_matrix.WriteText(filename2);
    vPrint("Success1 = %li, success2 = %li\n", success1, success2);
    vdelete2(filename1);
    vdelete2(filename2);
  }

  return result_matrix;
}


// Returns the equivalent of the distance transform when we 
// want to take into account edge orientation in addition to
// edge location. Location (i,j,k) contains the distance to
// the closest location/orientation combination.
v3dMatrix<float> * vDO_DistanceTransform(grayscale_image * edges,
                                            v3dMatrix<double> * orientations)
{
  const long orientation_bins = 15;
  long rows = edges->Rows();
  long cols = edges->Cols();

  if (orientations->Rows() != rows) return 0;
  if (orientations->Cols() != cols) return 0;

  // For the time being, in the for-loops, we making row and col range
  // from 25 to 230, assuming that the bounding box is 192x192, 
  // centered, in a 256x256 image. So, we should make sure 
  // rows and cols are 256. To allow arbitrary rows and cols, I should
  // just make row and col in the for-loop range from 0 to rows/cols.
  if ((rows != 256) || (cols != 256))
  {
    exit_error("vDO_DistanceTransform: can't handle (rows,cols) = %li,%li\n",
                    rows, cols);
  }

  vector<long> point_rows, point_cols;
  vGetNonZeroPoints(edges, &point_rows, &point_cols);

  long number = point_rows.size();
  vPrint("number = %li\n", number);
  vArray2(double) o2 = orientations->Matrix2(0);
  vector<float> point_os(number); // Will store point orientations
  
  long i;
  for (i = 0; i < number; i++)
  {
    long row = point_rows[i];
    long col = point_cols[i];
    point_os[i] = (float) o2[row][col];
  }

  v3dMatrix<float> * result = new v3dMatrix<float>(rows, cols, orientation_bins);
  vArray3(float) distances3 = result->Matrix3();

  long bin, row, col;
 
  // For each possible row, col and orientation, we will precompute the distances
  // to all edge pixels.
  vMatrix<long> row_diffs_matrix(rows, number);
  vMatrix<long> col_diffs_matrix(rows, number);
  vMatrix<float> orientation_diffs_matrix(orientation_bins, number);
  vArray2(long) row_diffs = row_diffs_matrix.Matrix2();
  vArray2(long) col_diffs = col_diffs_matrix.Matrix2();
  vArray2(float) orientation_diffs = orientation_diffs_matrix.Matrix2();

  // Fill in the row difference information (squared row difference)
//  for (row = 0; row < rows; row++)
  for (row = 25; row < 231; row++)
  {
    for (i = 0; i < number; i++)
    {
      long point_row = point_rows[i];
      long diff = point_row - row;
      row_diffs[row][i] = diff * diff;
    }
  }

  // Fill in the col difference information (squared col difference)
//  for (col = 0; col < cols; col++)
  for (col = 25; col < 231; col++)
  {
    for (i = 0; i < number; i++)
    {
      long point_col = point_cols[i];
      long diff = point_col - col;
      col_diffs[col][i] = diff * diff;
    }
  }

  // Fill in the orientation difference information
  for (bin = 0; bin < orientation_bins; bin++)
  {
    float orientation1 = bin * (float) M_PI / (float) orientation_bins;
    for (i = 0; i < number; i++)
    {
      float orientation2 = point_os[i];
      orientation_diffs[bin][i] = vAbs(vCircularDistance(orientation1, orientation2, (float) M_PI));
      // Make sure the orientation diff is never 0. 
      orientation_diffs[bin][i] += (float) M_PI / (float) orientation_bins;
    }
  }

  float min_dissimilarities[orientation_bins];
  const float initial_min_dissimilarity = (rows * rows + cols * cols) * (float) M_PI;

  vPrint("\n");
//  for (row = 0; row < rows; row++)
  for (row = 25; row < 231; row+=2)
  {
//    for (col = 0; col < cols; col++)
    for (col = 25; col < 231; col+=2)
    {
      // Initialize all min dissimilarities.
      for (bin = 0; bin < orientation_bins; bin++)
      {
        min_dissimilarities[bin] = initial_min_dissimilarity;
      }

      // Find the most similar edge pixel for (orientation, row, col).
      for (i = 0; i < number; i++)
      {
        float row_diff = (float) row_diffs[row][i];
        float col_diff = (float) col_diffs[col][i];
        // We add 1 to location_distance to make sure that the 
        // overall dissimilarity is never 0 (otherwise, a pixel
        // of different orientation that was at the same position
        // would always have dissimilarity 0, since we are multiplying).
        float location_distance = sqrt(row_diff + col_diff + 1);
        
        for (bin = 0; bin < orientation_bins; bin++)
        {
          float angle = orientation_diffs[bin][i];
          float dissimilarity = location_distance * angle;
          if (dissimilarity < min_dissimilarities[bin])
          {
            min_dissimilarities[bin] = dissimilarity;
          }
        }
      }

      // Write all dissimilarity values.
      for (bin = 0; bin < orientation_bins; bin++)
      {
        distances3[bin][row][col] = min_dissimilarities[bin];
        distances3[bin][row+1][col] = min_dissimilarities[bin];
        distances3[bin][row][col+1] = min_dissimilarities[bin];
        distances3[bin][row+1][col+1] = min_dissimilarities[bin];
      }
    }
    vPrint("Processed row %li\r", row);
  }

  vPrint("\n");
  return result;
}


vMatchScores *
vChamfer_DO_MI3(v3dMatrix<float> * do_dt, vPointSetSequence * locations,
                 vPointSetOrientation * orientations)
{
  long number_of_frames = locations->NumberOfFrames();
  const long number_of_views = 4;
  long number_of_scores = number_of_frames * number_of_views;
  vector<long> frames(number_of_scores);
  vector<long> views(number_of_scores);

  long index = 0;
  long i;
  for (i = 0; i < number_of_frames; i++)
  {
    long j;
    for (j = 0; j < number_of_views; j++)
    {
      frames[index] = i;
      views[index] = j;
      index++;
    }
  }

  return vChamfer_DO_MI5(do_dt, locations, orientations, &frames, &views);
}


vMatchScores *
vChamfer_DO_MI5(v3dMatrix<float> * do_dt, vPointSetSequence * locations,
                 vPointSetOrientation * orientations, 
                 vector<long> * frames, vector<long> * views)
{
  long number_of_frames = frames->size();
  long i;
  long orientation_bins = do_dt->Bands();
  vArray3(float) values = do_dt->Matrix3();

  vArray(vMatchScore) scores = vnew(vMatchScore, number_of_frames);
  for (i = 0; i < number_of_frames; i++)
  {
    long frame = (*frames)[i];
    long view = (*views)[i];
    scores[i].frame = frame;
    scores[i].view = view;
    scores[i].is_valid = 1;

    vArray(long) rows;
    vArray(long) cols;
    vArray(float) angles;
    long number = locations->GetPoints(frame, view, &rows, &cols);
    orientations->GetOrientations(frame, view, &angles);

    long j;
    float sum = 0;
    for (j = 0; j < number; j++)
    {
      long row = rows[j];
      if (row % 2 == 0) row++;
      long col = cols[j];
      if (col % 2 == 0) col++;
      long bin = vRound(angles[j] / M_PI * orientation_bins);
      if (bin >= orientation_bins) bin = orientation_bins - 1;
      sum += values[bin][row][col];
    }

    scores[i].score = sum / (float) number;
    vdelete2(rows);
    vdelete2(cols);
    vdelete2(angles);
  }

  vMatchScores * result = new vMatchScores(scores, number_of_frames);
  return result;
}

                
// Computes scores for all database images
//  using the distance/orientation version of the chamfer distance.
vMatchScores * vChamfer_DO_Scores4(grayscale_image * edges,
                                     v3dMatrix<double> * orientations,
                                     vPointSetSequence * database_points,
                                     vPointSetOrientation * database_orientations)
{
  static const long number_of_frames = 86*12*26;
  static const long number_of_views = 4;
  static const long number_of_database_images = number_of_views * number_of_frames;
  vector<long> frames(number_of_database_images), views(number_of_database_images);
  long frame, view;
  long index = 0;

  for (frame = 0; frame < number_of_frames; frame++)
  {
    for (view = 0; view < number_of_views; view++)
    {
      frames[index] = frame;
      views[index] = view;
      index++;
    }
  }

  return vChamfer_DO_Scores6(edges, orientations, database_points,
                              database_orientations, &frames, &views);
}

                                          
// Computes scores for the given database images (specified by frames
// and views) using the distance/orientation version of the chamfer distance.
vMatchScores * vChamfer_DO_Scores6b(grayscale_image * edges,
                                      v3dMatrix<double> * orientations,
                                      vPointSetSequence * database_points,
                                      vPointSetOrientation * database_orientations,
                                      vMatchScores * scores,
                                      long scores_to_keep)
{
  vector<long> frames(scores_to_keep), views(scores_to_keep);
  vArray(vMatchScore) frame_scores = scores->GetScores();
  long i;
  for (i = 0; i < scores_to_keep; i++)
  {
    frames[i] = frame_scores[i].frame;
    views[i] = frame_scores[i].view;
  }

  return vChamfer_DO_Scores6(edges, orientations, database_points,
                              database_orientations, &frames, &views);
}


// Computes scores for the given database images (specified by frames
// and views) using the distance/orientation version of the chamfer distance.
vMatchScores *  vChamfer_DO_Scores6(grayscale_image * edges,
                                      v3dMatrix<double> * orientations,
                                      vPointSetSequence * database_points,
                                      vPointSetOrientation * database_orientations,
                                      vector<long> * frames,
                                      vector<long> * views)
{
  const long orientation_bins = 15;
  long rows = edges->Rows();
  long cols = edges->Cols();

  if (orientations->Rows() != rows) return 0;
  if (orientations->Cols() != cols) return 0;
  if (frames->size() == 0) return 0;
  if (frames->size() != views->size()) return 0;

  if ((rows > 256) || (cols > 256))
  {
    exit_error("vChamfer_Do_Scores6: image dimensions (%li, %li) over 256\n",
                    rows, cols);
  }

  vector<long> point_rows, point_cols;
  vGetNonZeroPoints(edges, &point_rows, &point_cols);

  long number = point_rows.size();
  vPrint("number = %li\n", number);
  if (number == 0) return 0;
  vArray2(double) o2 = orientations->Matrix2(0);
  vector<float> point_os(number); // Will store point orientations
  
  long i, j;
  for (i = 0; i < number; i++)
  {
    long row = point_rows[i];
    long col = point_cols[i];
    point_os[i] = (float) o2[row][col];
  }

  long number_of_database_images = frames->size();
  vArray(vMatchScore) scores = vnew(vMatchScore, number_of_database_images);
  vPrint("\n");

  // Collect scores for all database images.
  long counter;
  for (counter = 0; counter < number_of_database_images; counter++)
  {
    long frame = (*frames)[counter];
    long view = (*views)[counter];
    
    vArray(long) point_rows2;
    vArray(long) point_cols2;
    vArray(float) point_os2;
    long number2 = database_points->GetPoints(frame, view, &point_rows2, 
                                    &point_cols2);
    database_orientations->GetOrientations(frame, view, &point_os2);
 
    double sum1, sum2;
    sum1 = sum2 = 0;
    // calculate the image-to-model distance.
    for (i = 0; i < number; i++)
    {
      long row = point_rows[i];
      long col = point_cols[i];
      float orientation = point_os[i];
    

      float min_dissimilarity = 1000000000;
      // Find the most similar edge pixel for (orientation, row, col).
      for (j = 0; j < number2; j++)
      {
        long row2 = point_rows2[j];
        long col2 = point_cols2[j];
        float orientation2 = point_os2[j];
        float row_diff = (float) (row2 - row);
        float col_diff = (float) (col2 - col);
        float orientation_diff = vAbs(vCircularDistance(orientation, orientation2, (float) M_PI));
        float dissimilarity = sqrt(row_diff * row_diff + col_diff * col_diff) * orientation_diff;

        if (dissimilarity < min_dissimilarity)
        {
          min_dissimilarity = dissimilarity;
        }
      }
      sum1 += min_dissimilarity;
    }

    // calculate the model-to-image distance.
    for (j = 0; j < number2; j++)
    {
      long row2 = point_rows2[j];
      long col2 = point_cols2[j];
      float orientation2 = point_os2[j];
      float min_dissimilarity = 1000000000;

      // Find the most similar edge pixel for (orientation, row, col).
      for (i = 0; i < number; i++)
      {
        long row = point_rows[i];
        long col = point_cols[i];
        float orientation = point_os[i];

        float row_diff = (float) (row2 - row);
        float col_diff = (float) (col2 - col);
        float orientation_diff = vAbs(vCircularDistance(orientation, orientation2, (float) M_PI));
        float dissimilarity = sqrt(row_diff * row_diff + col_diff * col_diff) * orientation_diff;
  
        if (dissimilarity < min_dissimilarity)
        {
          min_dissimilarity = dissimilarity;
        }
      }
      sum2 += min_dissimilarity;
    }

    vdelete2(point_rows2);
    vdelete2(point_cols2);
    vdelete2(point_os2);
    scores[counter].frame = frame;
    scores[counter].view = view;
    scores[counter].score = (float) sum1 / ((float) number) + (float) sum2 / ((float) number2);
    scores[counter].is_valid = 1;
    vPrint("Done with frame %li of %li\r", counter, number_of_database_images);
  }

  vPrint("\n");
  vMatchScores * result = new vMatchScores(scores, number_of_database_images);
  return result;
}


// cd_triples specifies some triples of hand poses represented in 
// the database (frame/view combinations). cd_distances contains 
// precomputed chamfer distances between every shape/viewpoint 
// combination (every 12th frame stored in the database) and 
// every hand pose represented in the database (4 poses are 
// represented by each frame in the database). This function
// looks up, for each triple (A,B,C), the distances AB, AC, BC,
// using cd_distances. Those distances are stored in the result matrix,
// which is, if we have K triples, a Kx3 matrix.
v3dMatrix<float> * vCD_Triple_Distances(v3dMatrix<long> * cd_triples,
                                           v3dMatrix<uchar> * cd_distances)
{
  long number_of_triples = cd_triples->Rows();

  // Create the result.
  v3dMatrix<float> * result_matrix = new v3dMatrix<float>(number_of_triples, 3, 1);
  vArray2(float) result = result_matrix->Matrix2(0);
  vArray2(long) triples = cd_triples->Matrix2(0);
  vArray2(uchar) distances = cd_distances->Matrix2(0);
  // finder is used to do some arithmetic that will tell us what
  // entry in cd_distances contains the distance we are looking for.
  static vFrameViewFinder finder;

  long i;
  vPrint("\n");
  for (i = 0; i < number_of_triples; i++)
  {
    // Get the hand pose images the distance between which we must
    // look up.
    long frame1 = triples[i][0];
    long view1 = triples[i][1];
    long frame2 = triples[i][2];
    long view2 = triples[i][3];
    long frame3 = triples[i][4];
    long view3 = triples[i][5];

    // Figure out where to look up this distance. 
    // First, figure out what row in cd_distances we should look at,
    // i.e. what shape/viewpoint combination the first hand pose
    // corresponds to 
    long row1 = frame1 / 12;
    // Now, figure out what column we should
    // look at, i.e. what hand pose image is as far from the
    // shape/viewpoint combination as the two hand poses
    // (frame1/view1 and frame2/view2) are from each other.
    long rotation1 = finder.ActualRotation2(frame1, view1);

    long index2 = frame2 * 4 + view2;
    long rotation2 = finder.ActualRotation2(frame2, view2);
    long base2 = index2 - (index2 % 48);

    long difference12;
    if (rotation1 > rotation2)
    {
      difference12 = rotation2 + 48 - rotation1;
    }
    else 
    {
      difference12 = rotation2 - rotation1;
    }
    long viewpoint_adjustment12 = finder.FrameIndex(difference12);
    long rotation_adjustment12 = finder.RotationIndex(difference12);
    long adjustment12 = viewpoint_adjustment12 * 4 + rotation_adjustment12;
    long adjusted_index12 = base2 + adjustment12;

    // Look up the distance and store it.
    long distance12 = distances[row1][adjusted_index12];
    result[i][0] = (float) uchar_to_double_cd_convert(distance12);

    long index3 = frame3 * 4 + view3;
    long rotation3 = finder.ActualRotation2(frame3, view3);
    long base3 = index3 - (index3 % 48);

    long difference13;
    if (rotation1 > rotation3)
    {
      difference13 = rotation3 + 48 - rotation1;
    }
    else 
    {
      difference13 = rotation3 - rotation1;
    }
    long viewpoint_adjustment13 = finder.FrameIndex(difference13);
    long rotation_adjustment13 = finder.RotationIndex(difference13);
    long adjustment13 = viewpoint_adjustment13 * 4 + rotation_adjustment13;
    long adjusted_index13 = base3 + adjustment13;

    // Look up the distance and store it.
    long distance13 = distances[row1][adjusted_index13];
    result[i][1] = (float) uchar_to_double_cd_convert(distance13);

    long row2 = frame2 / 12;

    long difference23;
    if (rotation2 > rotation3)
    {
      difference23 = rotation3 + 48 - rotation2;
    }
    else 
    {
      difference23 = rotation3 - rotation2;
    }
    long viewpoint_adjustment23 = finder.FrameIndex(difference23);
    long rotation_adjustment23 = finder.RotationIndex(difference23);
    long adjustment23 = viewpoint_adjustment23 * 4 + rotation_adjustment23;
    long adjusted_index23 = base3 + adjustment23;

    // Look up the distance and store it.
    long distance23 = distances[row2][adjusted_index23];
    result[i][2] = (float) uchar_to_double_cd_convert(distance23);

    if (i % 1000 == 0)
    {
      vPrint("Done with triple %i\r", i);
    }
  }
  vPrint("\n");  

  return result_matrix;
}


double function_Hausdorff_distance(vArray(uchar) rows1, vArray(uchar) cols1, 
                                   long number1, v3dMatrix<uchar> * dt_image)
{
  if (number1 == 0) 
  {
    return dt_image->Rows() + dt_image->Cols();
  }
  vArray2(uchar) distances = dt_image->Matrix2(0);
  long i;
  float largest_distance = 0;
  for (i = 0; i < number1; i++)
  {
    long row = rows1[i];
    long col = cols1[i];
    float distance = distances[row][col];
    if (distance > largest_distance)
    {
      largest_distance = distance;
    }
  }

  return largest_distance;
}


/////////////////////////////////////////////////////////////////////
//                                                                 //
//     Beginning of implementation of class vPointSetSequence     //
//                                                                 //
/////////////////////////////////////////////////////////////////////


vPointSetSequence::vPointSetSequence(vVideoFile * sequence)
{
  rows = sequence->Rows();
  cols = sequence->Cols();
  if ((rows > 256) || (cols > 256))
  {
    exit_error("Coordinates exceeding 256 in vPointSetSequence\n");
  }

  number_of_frames = sequence->NumberOfFrames();

  lengths = vnew(long, number_of_frames);
  leftover_lengths = vnew(long, number_of_frames);
  indices = vnew(long, number_of_frames);

  // First store the points into vectors, since we don't know how
  // many points we will find.
  vector<uchar> temp_rows, temp_cols;

  long i;
  vPrint("\n");
  for (i = 0; i < number_of_frames; i++)
  {
    grayscale_image * gray = sequence->ReadFrameGray(i);
    if (gray == 0)
    {
      exit_error("error in reading frame %li\n");
    }

    vector<long> current_rows, current_cols;
    vGetNonZeroPoints(gray, &current_rows, &current_cols);
    long length = current_rows.size();
    lengths[i] = length;
    leftover_lengths[i] = 0;
    if (i == 0)
    {
      indices[i] = 0;
    }
    else 
    {
      indices[i] = indices[i-1] + lengths[i-1];
    }
    long j;
    for (j = 0; j < length; j++)
    {
      temp_rows.push_back((uchar) current_rows[j]);
      temp_cols.push_back((uchar) current_cols[j]);
    }
    vdelete(gray);
    if (i % 100 == 0)
    {
      vPrint("processed frame %li of %li\r", i, number_of_frames);
    }
  }
  vPrint("\n");

  // Now that we have all the point coordinates, store them into the arrays.
  total_length = temp_rows.size();
  point_rows = vnew(uchar, total_length);
  point_cols = vnew(uchar, total_length);
  for (i = 0; i < total_length; i++)
  {
    point_rows[i] = temp_rows[i];
    point_cols[i] = temp_cols[i];
  }
}


vPointSetSequence::~vPointSetSequence()
{
  vdelete2(lengths);
  vdelete2(leftover_lengths);
  vdelete2(indices);
  vdelete2(point_rows);
  vdelete2(point_cols);
}


vPointSetSequence * vPointSetSequence::Load(const char * filename)
{
  FILE * fp = fopen(filename, vFOPEN_READ);
  if (fp == 0) return 0;
  vPointSetSequence * result = new vPointSetSequence();
  long items = 0;
  items += vReadPCLongs(fp, &(result->rows), 1);
  items += vReadPCLongs(fp, &(result->cols), 1);
  items += vReadPCLongs(fp, &(result->number_of_frames), 1);
  items += vReadPCLongs(fp, &(result->total_length), 1);

  if ((items != 4) || (result->rows < 0) || (result->cols < 0) ||
      (result->rows > 256) || (result->cols > 256))
  {
    vPrint("Error in loading point set sequence\n");
    vdelete(result);
    fclose(fp);
    return 0;
  }

  result->lengths = vnew(long, result->number_of_frames);
  result->leftover_lengths = vnew(long, result->number_of_frames);
  result->indices = vnew(long, result->number_of_frames);
  result->point_rows = vnew(uchar, result->total_length);
  result->point_cols = vnew(uchar, result->total_length);

  items = vReadPCLongs(fp, vData(result->lengths), result->number_of_frames);
  items += vReadPCLongs(fp, vData(result->leftover_lengths), result->number_of_frames);
  items += vReadPCLongs(fp, vData(result->indices), result->number_of_frames);
  items += fread(vData(result->point_rows), sizeof(uchar), result->total_length, fp);
  items += fread(vData(result->point_cols), sizeof(uchar), result->total_length, fp);
  fclose(fp);
  if (items != 3 * result->number_of_frames + 2 * result->total_length)
  {
    vPrint("Error in reading point set sequence\n");
    vdelete(result);
    return 0;
  }
  return result;
}


long vPointSetSequence::Save(const char * filename)
{
  FILE * fp = fopen(filename, vFOPEN_WRITE);
  if (fp == 0) return 0;
  long items = 0;
  items += vWritePCLongs(fp, &rows, 1);
  items += vWritePCLongs(fp, &cols, 1);
  items += vWritePCLongs(fp, &number_of_frames, 1);
  items += vWritePCLongs(fp, &total_length, 1);

  if (items != 4)
  {
    vPrint("Error in saving point set sequence\n");
    fclose(fp);
    return 0;
  }

  items = vWritePCLongs(fp, vData(lengths), number_of_frames);
  items += vWritePCLongs(fp, vData(leftover_lengths), number_of_frames);
  items += vWritePCLongs(fp, vData(indices), number_of_frames);
  items += fwrite(vData(point_rows), sizeof(uchar), total_length, fp);
  items += fwrite(vData(point_cols), sizeof(uchar), total_length, fp);
  fclose(fp);
  if (items != 3 * number_of_frames + 2 * total_length)
  {
    vPrint("Error in writing point set sequence\n");
    return 0;
  }
  return 1;
}


// returns the number of points.
// In rowsp and colsp we put arrays of the (row, col)
// coordinates of the set of points corresponding
// to the given frame and given view
long vPointSetSequence::GetPoints(long frame, long view, 
                                   vArray(long) * rowsp, vArray(long) * colsp)
{
  long number = Length(frame);
  // Get the (row, col) coordinates for given frame and view 0.
  vArray(uchar) frame_rows = PointRows(frame);
  vArray(uchar) frame_cols = PointCols(frame);
  
  // Copy over the coordinates into the result array.
  vArray(long) result_rows = vnew(long, number);
  vArray(long) result_cols = vnew(long, number);
  long i;
  for (i = 0; i < number; i++)
  {
    result_rows[i] = frame_rows[i];
    result_cols[i] = frame_cols[i];
  }

  // Rotate points by rotation corresponding to given view.
  long degrees = view * 90;
  vRotatePointsRight(result_rows, result_cols, result_rows, result_cols,
                      number, rows, cols, degrees);
  *rowsp = result_rows;
  *colsp = result_cols;
  return number;
}


// Here we assume we already have all the information stored in the
// vPointSetSequence object except for the leftover points. Given
// the leftover points for a particular frame, we rearrange the 
// order of points for that frame, so that the leftover points
// show up first.
long vPointSetSequence::MarkLeftover(long frame, vector<long> * leftover_rows,
                                      vector<long> * leftover_cols)
{
  CheckFrame(frame);

  vArray(uchar) frame_rows = PointRows(frame);
  vArray(uchar) frame_cols = PointCols(frame);
  long length = Length(frame);
  long leftover_length = leftover_rows->size();
  leftover_lengths[frame] = leftover_length;

  v3dMatrix<uchar> scratch_matrix(rows, cols, 1);
  vArray2(uchar) scratch = scratch_matrix.Matrix2(0);

  // Mark all points as zero on the scratch array
  long i;
  for (i = 0; i < length; i++)
  {
    long row = frame_rows[i];
    long col = frame_cols[i];
    scratch[row][col] = 0;
  }

  // Mark points on leftover_rows with 1.
  for (i = 0; i < leftover_length; i++)
  {
    long row = (*leftover_rows)[i];
    long col = (*leftover_cols)[i];
    scratch[row][col] = 1;
  }

  // Now put all points that are still marked with 0 into
  // a new set of points (used_rows/used_cols).
  vector<long> used_rows, used_cols;
  for (i = 0; i < length; i++)
  {
    long row = frame_rows[i];
    long col = frame_cols[i];
    if (scratch[row][col] != 0) continue;
    used_rows.push_back(row);
    used_cols.push_back(col);
  }

  long used_points = used_rows.size();
  if (used_points + leftover_length != length)
  {
    vPrint("There must be a bug, points don't add up\n");
    return -1;
  }

  // Now reorder the points in point_rows/point_cols.
  // First put in the leftovers
  long index = indices[frame];
  for (i = 0; i < leftover_length; i++)
  {
    point_rows[index] = (uchar) (*leftover_rows)[i];
    point_cols[index] = (uchar) (*leftover_cols)[i];
    index++;
  }

  // Now put in the used points (non-leftovers)
  for (i = 0; i < used_points; i++)
  {
    point_rows[index] = (uchar) used_rows[i];
    point_cols[index] = (uchar) used_cols[i];
    index++;
  }

  return 1;
}


long vPointSetSequence::MarkLeftovers(vVideoFile * model_images,
                                       vVideoFile * model_edges,
                                       v3dMatrix<uchar> * all_model_lines)
{
  if (model_images->NumberOfFrames() != number_of_frames) return 0;
  if (model_edges->NumberOfFrames() != number_of_frames) return 0;

  long i;
  long success = 1;
  vPrint("\n");
  for (i = 0; i < number_of_frames; i++)
  {
    grayscale_image * edges = model_edges->ReadFrameGray(i);
    color_image * links = model_images->ReadFrame(i);
    if ((edges == 0) || (links == 0)) 
    {
      vPrint("Failed to read images\n");
      success = 0;
      vdelete(edges);
      vdelete(links);
      break;
    }

    grayscale_image * dilated = vDilateImage(edges, 1, 255);
    grayscale_image * link_ids = vLinkIDsB(links);
    vector<vStraightLinks> * model_straight_links = 
      vReadModelLines3(all_model_lines, i, 0);
    vector<long> leftover_rows, leftover_cols;
    vLeftoverModelPoints(link_ids, edges, dilated, model_straight_links,
                          &leftover_rows, &leftover_cols);
    MarkLeftover(i, &leftover_rows, &leftover_cols);
    vdelete(edges);
    vdelete(links);
    vdelete(dilated);
    vdelete(link_ids);
    vdelete(model_straight_links);
    vPrint("Processed frame %li of %li\r", i+1, number_of_frames);
  }

  vPrint("\n");
  return success;
}



/////////////////////////////////////////////////////////////////////
//                                                                 //
//     End of implementation of class vPointSetSequence           //
//                                                                 //
/////////////////////////////////////////////////////////////////////

