#ifndef VASSILIS_HAUSDORFF_MATCHING_H
#define VASSILIS_HAUSDORFF_MATCHING_H


#include "basics/image.h"
#include "basics/algebra.h"
#include "basics/rle_encoding.h"
#include "simple_chamfer.h"
#include "ranking.h"


// DatabaseEdgeSummary is a class that will contain distance
// transforms and extracted edge pixels from database images, 
// for efficient computation of the chamfer distance. It
// is an interface that subclasses should follow.
class vDatabaseEdgeSummary
{
private:
  long Initialize();

protected:
  // dts[i] will contain the distance transform of some
  // database image.
  class_pointer (grayscale_image *) dts;

  // rows[i] and cols[i] will contain coordinates of points
  // in some database edge image.
  vArray2(uchar) rows;
  vArray2(uchar) cols;

  long number_of_frames;
  vArray(long) lengths;
  long is_valid;

  // Filename in these three functions is the name of
  // the database edge sequence. From that, the filenames
  // of the DTs and points are derived.
  virtual char * DT_Filename(const char * filename);
  virtual char * PointsFilename(const char * filename);
  virtual long Load(const char * filename);

public:
  vDatabaseEdgeSummary();
  vDatabaseEdgeSummary(const char * filename);
  vDatabaseEdgeSummary(vVideoFile * sequence);
  ~vDatabaseEdgeSummary();

  // Saves DTs and points.
  virtual long SaveInfo(vVideoFile * sequence)
  {
    long success1 = SaveDTs(sequence);
    long success2 = SavePoints(sequence);
    return success1 + success2;
    return 1;
  }

  // Saves DTs.
  virtual long SaveDTs(vVideoFile * sequence) = 0;

  // Saves Points.
  virtual long SavePoints(vVideoFile * sequence) = 0;


  virtual vMatrix<double> ChamferDistances(grayscale_image * edge_image) = 0;
  
  virtual vMatrix<double> ChamferDistances(grayscale_image * edge_image,
                                            vector<long> * frames,
                                            vector<long> * views) = 0;
  
  vMatchScores * ChamferDistanceScores(grayscale_image * edge_image)
  {
    vMatrix<double> distances = ChamferDistances(edge_image);
    vMatchScores * result = new vMatchScores(&distances);
    return result;
  }

  vMatchScores * ChamferDistanceScores(grayscale_image * edge_image,
                                        vector<long> * frames,
                                        vector<long> * views)
  {
    vMatrix<double> distances = ChamferDistances(edge_image, frames, views);
    long number_of_scores = distances.Size();
    vArray(vMatchScore) scores = vnew(vMatchScore, number_of_scores);
    vArray(double) distances1 = distances.Matrix();
    long i;
    for (i = 0; i < number_of_scores; i++)
    {
      scores[i].score = (float) distances1[i];
      scores[i].frame = (*frames)[i];
      scores[i].view = (*views)[i];
    }
    vMatchScores * result = new vMatchScores(scores, number_of_scores);
    return result;
  }

  // Returns the chamfer distance between two database poses,
  // specified by frame and view.
  double PoseChamferDistance(long frame1, long view1, 
                             long frame2, long view2);

  inline long IsValid()
  {
    return is_valid;
  }
};


// This subclass of vDatabaseEdgeSummary is geared towards the
// first version of the database, where we save 12 image plane
// rotations per viewpoint.
class vDatabaseEdgeSummary1 : public vDatabaseEdgeSummary
{
private:
  long Initialize();

protected:
  long basic_shapes;
  long viewpoints;
  long saved_orientations;
  long orientations;
  long views_per_frame;
  
  // Rotation templates, that will be used to rotate an
  // input image before we compute the chamfer distances.
  v3dMatrix<short> * rotation_templates[48];

  // rotations[i] will be the amount of rotation corresponding
  // to rotation_templates[i];
  double rotations[48];


public:
  vDatabaseEdgeSummary1();
  vDatabaseEdgeSummary1(const char * filename);
  vDatabaseEdgeSummary1(vVideoFile * sequence);
  ~vDatabaseEdgeSummary1();

  // Extracts the distance transform of color_image, rotates it
  // 48 times, and also extracts the non-zero pixel locations
  // and rotates them 48 times. The image plane rotations
  // appear in the result vectors in the order in which
  // the match scores are supposed to appear (4 per database frame).
  long PrepareForChamfer(grayscale_image * the_image, grayscale_image ** dt_array,
                         vector<long> * rows_array, vector<long> * cols_array);

  // Saves DTs.
  virtual long SaveDTs(vVideoFile * sequence);

  // Saves Points.
  virtual long SavePoints(vVideoFile * sequence);

  virtual vMatrix<double> ChamferDistances(grayscale_image * edge_image);
  
  virtual vMatrix<double> ChamferDistances(grayscale_image * edge_image,
                                            vector<long> * frames,
                                            vector<long> * views);
  double PoseChamferDistance(long frame1, long view1, 
                             long frame2, long view2);
};  

  
// This function is essentially a wrapper around vOneWayChamfer, it
// produces match scores for the whole database using vOneWayChamfer.
vMatchScores * vOneWayChamferRanks(grayscale_image * test_gray, 
                                     const char * training_filename);


vMatchScores *
vOneWayChamfer2(grayscale_image * gray, vPointSetSequence * database,
                 vector<long> * frames, vector<long> * views);



// This function is essentially a wrapper around vOneWayChamfer, it
// produces match scores for the whole database using vOneWayChamfer.
vMatchScores * vOneWayChamferRanks2(grayscale_image * test_gray, 
                                      vPointSetSequence * database);



// This function computes the similarity between edge_image and
// some database views. The database edge images are stored in
// training_filename, and the database views that should be 
// compared to the test image are the highest ranking frames
// in database_frames. number_of_frames specifies how many
// of those frames should be used.
vMatchScores * vChamferRanks(grayscale_image * edge_image, 
                                const char * training_filename, 
                                vMatchScores * database_frames,
                                long number_of_frames);


// Like vChamferRanks, but allows the test image to translate and scale
// a bit to better match each database image
vMatchScores * vFlexibleChamferRanks(grayscale_image * edge_image, 
                                       vVideoFile * database,
                                       vMatchScores * frames_to_check,
                                       long number_of_frames);

                                       
// These conversions are used so that we can save distances (in
// vAll_MI_Distances etc) using uchars.
inline double uchar_to_double_mi_convert(long converted)
{
  const double factor = 255.0 / 60;
  return ((double) converted) / factor;
}

inline uchar double_to_uchar_mi_convert(double original)
{
  const double factor = 255.0 / 60;
  long result = (long) (original * factor);
  if (result < 0) result = 0;
  if (result > 255) result = 255;
  return (uchar) result;
}

inline double uchar_to_double_cd_convert(long converted)
{
  const double factor = 255.0 / 120;
  return ((double) converted) / factor;
}

inline uchar double_to_uchar_cd_convert(double original)
{
  const double factor = 255.0 / 120;
  long result = (long) (original * factor);
  if (result < 0) result = 0;
  if (result > 255) result = 255;
  return (uchar) result;
}


// Given a matrix that contains the chamfer distances between
// every shape/viewpoint combination and all database images,
// we want to extract the distances between a given database 
// image A and all other database images. Essentially, for every
// database image B, the distance between A and B is the same
// as the distance between the shape/viewpoint combination that
// corresponds to A and some image of the same shape/viewpoint
// combination as B. We have to figure out, for every A and B,
// what the right C is.
vMatrix<double> vGet_CD_Distances(long frame, long view, 
                                    v3dMatrix<uchar> * cd_distances);

// This function is used to test and debug the way indices are adjusted
// in vGet_CD_Distances, which is based on S_AdjustImageIndex.
long vTestIndexAdjustment(long frame, long view, long index);

v3dMatrix<long> * vRandom_CD_Pairs(long number);

v3dMatrix<double> * vCD_Pair_Distances(v3dMatrix<long> * cd_pairs, 
                                          v3dMatrix<uchar> * cd_distances);

vMatrix<double> vCalculateSTDs(v3dMatrix<uchar> * all_cd_distances,
                                 v3dMatrix<long> * cd_pairs,
                                 v3dMatrix<double> * cd_pair_distances,
                                 v3dMatrix<double> * estimated_distances,
                                 long factor);

long vCalculateSTD(v3dMatrix<uchar> * all_cd_distances,
                    v3dMatrix<long> * cd_pairs,
                    v3dMatrix<double> * cd_pair_distances,
                    v3dMatrix<double> * estimated_distances,
                    long candidate, double * mean, double * std,
                    long factor);


long vUpdateEstimatedDistances(v3dMatrix<uchar> * all_cd_distances,
                                v3dMatrix<long> * cd_pairs,
                                v3dMatrix<double> * estimated_distances,
                                long frame, long view, long factor);


vMatrix<double> vChooseBasis1(v3dMatrix<uchar> * all_cd_distances,
                                v3dMatrix<long> * cd_pairs,
                                v3dMatrix<double> * cd_pair_distances,
                                long factor, long number);

vMatrix<double> vChooseBasis2(v3dMatrix<uchar> * all_cd_distances,
                                v3dMatrix<long> * cd_pairs,
                                v3dMatrix<double> * cd_pair_distances,
                                long factor, long number);


vMatrix<double> vChooseBasis3(v3dMatrix<uchar> * all_cd_distances,
                                v3dMatrix<long> * cd_pairs,
                                v3dMatrix<double> * cd_pair_distances,
                                long factor, long number);

vMatrix<double> vChooseBasis4(v3dMatrix<uchar> * all_cd_distances,
                                v3dMatrix<long> * cd_pairs,
                                v3dMatrix<double> * cd_pair_distances,
                                long factor, long number);


// Choose basis elements in a greedy way, to minimize rank
// of highest ranking correct match
vMatrix<double> vChooseBasis5(v3dMatrix<uchar> * all_cd_distances,
                                v3dMatrix<float> * reference_distances,
                                const char * filename,
                                long frames_to_use, long number);

long vMakeLipschitzBasis(v3dMatrix<uchar> * all_cd_distances, 
                          long basis, long index,
                          v3dMatrix<float> ** random_training_distances, 
                          v3dMatrix<long> ** random_picks, 
                          v3dMatrix<long> ** random_views);


long vMakeLipschitzBasis(v3dMatrix<uchar> * all_cd_distances, 
                          long basis, long index,
                          v3dMatrix<float> ** database_embeddings, 
                          v3dMatrix<float> ** reference_images); 


long vMakeLipschitzBasis3(long basis, long index,
                           v3dMatrix<float> ** reference_images); 


vMatrix<double> vCalculateDistortions(v3dMatrix<uchar> * all_cd_distances,
                                        v3dMatrix<long> * cd_pairs,
                                        v3dMatrix<double> * cd_pair_distances,
                                        v3dMatrix<double> * estimated_distances,
                                        long factor);


long vCalculateDistortion(v3dMatrix<uchar> * all_cd_distances,
                           v3dMatrix<long> * cd_pairs,
                           v3dMatrix<double> * cd_pair_distances,
                           v3dMatrix<double> * estimated_distances,
                           long candidate, double * mean, double * std,
                           long factor);


vMatrix<float> vSaveReferenceDistances(vVideoFile * test_images,
                                         vPointSetSequence * database_images);


vMatrix<double> vCalculateAccuracies(v3dMatrix<uchar> * all_cd_distances,
                                       v3dMatrix<float> * reference_distances,
                                       v3dMatrix<ushort> * ground_distances,
                                       v3dMatrix<double> * estimated_distances,
                                       vector<long> * current_candidates,
                                       long frames_to_use);


long vCalculateAccuracy(v3dMatrix<uchar> * all_cd_distances,
                         v3dMatrix<float> * reference_distances,
                         v3dMatrix<ushort> * ground_distances,
                         v3dMatrix<double> * estimated_distances,
                         long candidate, double * accuracyp,
                         vector<long> * frames_to_use);


long vUpdateEstimatedDistances2(v3dMatrix<uchar> * all_cd_distances,
                                 v3dMatrix<float> * reference_distances,
                                 v3dMatrix<double> * estimated_distances,
                                 long frame, long view);


v3dMatrix<ushort> * vPrepareGroundDistances(const char * filename);


class vEmbedding
{
protected:
  vDatabaseEdgeSummary * edge_summary;
  virtual char * DatabaseDistancesFile() = 0;
  virtual char * ReferenceSetFile() = 0;

public:
  vEmbedding(vDatabaseEdgeSummary * in_edge_summary) 
  {
    edge_summary = in_edge_summary;
  }
  ~vEmbedding() {}

  virtual vMatrix<float> Embed(grayscale_image * the_image) = 0;
  virtual vMatchScores * ComputeScores(grayscale_image * the_image) = 0;
  virtual long SaveDatabaseDistances() = 0;
  virtual long Load() = 0;
};


// Reference images consist of random points.
class vEmbedding1 : public vEmbedding
{
protected:
  virtual char * DatabaseDistancesFile();
  virtual char * ReferenceSetFile();

public:
  vEmbedding1(vDatabaseEdgeSummary * in_edge_summary);
  ~vEmbedding1();

  virtual vMatrix<float> Embed(grayscale_image * the_image);
  virtual vMatchScores * ComputeScores(grayscale_image * the_image);
  virtual long SaveDatabaseDistances();
  virtual long Load();
};


// This function is a modified version of vChooseBasis, that is used
// to run multiple experiments comparing the method of minimizing
// average distortion versus the method of choosing images at random.
// The extra argument "shape" specifies which hand shape to use. Only
// database images of that hand shape are considered.
vMatrix<double> vChooseBasis6(v3dMatrix<uchar> * all_cd_distances,
                                v3dMatrix<long> * cd_pairs,
                                v3dMatrix<double> * cd_pair_distances,
                                long factor, long number, long shape);


// This is like vCalculateSTDs, but modified to work with vChooseBasis6.
// Basically, here we can specify a range of candidates to be examined.
// Only a limited number of candidates are considered, and therefore
// we can run multiple experiments to see how minimizing average distortion
// compares to picking random images.
vMatrix<double> vCalculateSTDs6(v3dMatrix<uchar> * all_cd_distances,
                                  v3dMatrix<long> * cd_pairs,
                                  v3dMatrix<double> * cd_pair_distances,
                                  v3dMatrix<double> * estimated_distances,
                                  long factor,
                                  long start, long number);


// Choose reference images randomly, considering as candidate reference
// images only database images that correspond to the hand shape
// specified by the argument "shape" (shape is between 0 and 25).
vMatrix<double> vChooseBasis7(v3dMatrix<uchar> * all_cd_distances,
                                v3dMatrix<long> * cd_pairs,
                                v3dMatrix<double> * cd_pair_distances,
                                long number, long shape);

double function_Hausdorff_distance(vArray(uchar) rows1, vArray(uchar) cols1, 
                                   long number1, v3dMatrix<uchar> * dt_image);






#endif // VASSILIS_HAUSDORFF_MATCHING_H
