#ifndef VASSILIS_INTERPRETER1_H
#define VASSILIS_INTERPRETER1_H

#include "basics/v_types.h"
#include "basics/video_file.h"
#include "basics/auxiliaries.h"
#include "basics/image_algorithms.h"
#include "gestures/algorithms.h"
#include "gestures/algo_templates.h"
#include "gestures/volumetric.h"
#include "leon/SkinTracker.h"
#include "window.h"
#include "matching.h"
#include "ranking.h"
#include "basics/base_interpreter.h"

#include <vector>


class vGestureInterfaceGlobals
{
public:
  vWindow * window; 
  vWindow * pixel_info_window;
  vWindow * frame_info_window;
  long min_width; // min_width of the main window
  long extra_height; // height in addition to the image, that the window requires
};


class class_sequence_information
{
private:

public:
  // The actual input video file
  vVideoFile * sequence;
  // Current frame in the video file
  long current_frame;
  // The path name from which we opened sequence
  char * path_name;
  // The simple name (without the path part) for the sequence
  char * simple_name;
  // Number of frames the sequence has
  long number_of_frames; 
  // sequence_step determines how many frames we should move forward at each 
  // step. This is used when we just step forward through the sequence using
  // the graphical interfce, but also with things like motion differencing,
  // where it determines which frame should be used as the "previous" frame
  // and which frame should be used as the "next" frame
  long sequence_step;
  // The window row where the top of the frame should be displayed
  long display_top;
  // The window column where the left of the frame should be displayed
  long display_left;

  // Resolution = 0 means we display the frame as we read it. resolution = k-1
  // means we should display the frame at half the resolution that we would use
  // for resolution = k. So, if our frame is 200x100 and resolution is -2 we
  // display a subsampled version of the frame of size 50x25. We can also have
  // positive resolution, in which case we only display a magnified version of
  // the image. In that case, we can't display the whole image anymore, so we
  // have to choose a center row and center col for the region that we will
  // magnify.
  long resolution;

  long center_row, center_col;
  // If we have a resolution > 0, we may want to display the image in a region
  // larger than the original size of the image. The user can set the size
  // of the region, but, by default, it is magnified_rows x magnified_cols.
  //long magnified_rows, magnified_cols;

  // automatic_centering is 1 if center_row, center_col should be set to
  // the center of each image that we show, 0 if they should stay fixed
  // (which is what we want when we set the center manually).
  long automatic_centering; 

  // When we display an image, we also store it in a most-recently-used buffer
  // (also called ring)
  vArray(general_image *) image_buffer;
  long buffer_index;    // Pointing to the current image in the ring
  general_image * current_image; // image_ring[ring_index];
  long buffer_size;

  // If we find a face in some frame in the video sequence, we may want to 
  // remember its color. That can help with finding the face and hands in
  // other frames of the sequence. The face color is represented as a color grid,
  // where for each rg color we have stored its block metric distance from
  // the average rg color of the face.
  
  vSearchGrid * face_color;

  // Camera input may also be represented as a class_sequence_information object, but in
  // that case we should treat it differently in a few places. For example,
  // it makes it convenient if we don't allow a call to 
  // vInterpreter1::RemoveInput to remove a camera input.
  long is_camera_input;

  class_sequence_information(long in_buffer_size);
  ~class_sequence_information();
  ushort ClearImageBuffer();
  ushort SetImageBufferSize(long in_buffer_size);
  long CurrentFrame();
  ushort SetCurrentFrame(long index);
};


typedef class_sequence_information class_sequence_information;

class vInterpreter1 : public class_module
{
public:
  // A variable of this type is used so that the ProcessMouseClick function knows
  // what it should do when it gets the next mouse click.
  enum MouseClickResponse {pixel_info_response, MoveSequenceResponse, 
                           LineResponse, CenterResponse, CutoffLabelResponse,
                           AddFeaturesResponse, choose_rectangle_response,
                           spline_response};
  MouseClickResponse previous_mouse_response;
  MouseClickResponse mouse_click_response;

  // The main window, can be 0 (then nothing gets displayed)
  vWindow * window; 
  vWindow * pixel_info_window;
  vWindow * frame_info_window;
  long min_width; // min_width of the main window
  long extra_height; // height in addition to the image, that the window requires

  // The default source of input for text commands (as defined in 
  // commands.cpp) is standard input. However, sometimes we may want
  // to execute batch commands, saved in a file. In those cases,
  // input_fp will represent that file.

  FILE * input_fp;

  // Data related to the skin detector
  vArray(vSkinTableBase *) skin_tables;
  vSkinTableBase * skin_table;
  ushort skin_data_initialized;
  ulong current_table_index;
  const char * skin_directory;
  const char * non_skin_directory;

  vector<class_sequence_information *> inputs;
  long default_buffer_size;

  // Data related to current input video files
  long current_input_index;
  long number_of_inputs;
  class_sequence_information * input;

  // Data related to current output video files
  long current_output;
  long number_of_outputs;
  std::vector<void *> output_files;
  std::vector<void *> output_names;

  // Various objects that are occasionally needed (for specific applications)
  vBackgroundModel * background_model;
  vCalibrationInfo * calibration;
  #ifndef VASSILIS_LINUX_PLATFORM
  skinTracker * skin_tracker;
  #endif 
  vMatchScores * match_scores;
  char * score_base;
  long score_frame;
  // score_order is 0 if we want to sort the scores in ascending order, and
  // 1 if we want to sort the scores in descending order.
  ushort score_order;
  char * test_score_file;
  double ground_truth_distance;
  v3dMatrix<ushort> * ground_distances;

  // Various parameters that are needed to supply default arguments to
  // some functions
  long filter_blur_size; // Used in convolutions.
  // Used in detecting hands.
  long hand_threshold;

  // Miscellaneous info about the state of the program
  char * current_directory;
  std::vector<void *> images; // Images of interest that we want to save.
  long number_of_images;
  vLineSet * line_set;
  vBoundingBox face_box;  

  // If left_hand = 1, then we assume that a left hand is classified.
  // Then, we should show database matches inverted upside down, to
  // look like left hands.
  long left_hand;

  long histogram_equalization_flag;
  long equalization_window_size;
  long preserve_current_frame;
  long frame_rate;

public:
  vInterpreter1(class_interpreter* in_interpreter);
  ~vInterpreter1();
  virtual void Initialize(vGestureInterfaceGlobals * in_globals);
  virtual void CleanUp();
  long DeleteSkinData();
    
  long vGetNonEmptyLine(char * buffer, long max_length);

  // Sets up the image as current image of the current input, and
  // calls DrawCurrentImage to display it.
  long ShowImage(general_image * image);
  long store_image(general_image * image);
  inline long show_image(general_image * image)
  {
    return ShowImage(image);
  }
  long ChangeDirectory(const char * filename);
  long ChangeDirectory();
  
  long AddInput();
  long AddInput(const char * filename);
  long AddInputSilent(const char * filename);
  long open_input(const char * filename)
  {
    return AddInput(filename);
  }
  long AddInput(vVideoFile * new_video_file);
  long open_input(class_sequence * video_file)
  {
    AddInput(video_file);
  }

  long default_input0();
  long default_input(long rows, long cols);
  long RemoveInput(long index);
  long ClearInputs();
  vVideoFile * GetInput(long index);
  long PrintInputs();
  // CurrentImage(i) returns the current image of the i-th
  // input file, or 0 if it doesn't exist.
  general_image * CurrentImage(long index);

  long AddOutput();
  long AddOutput(const char * filename);
  long add_output_stream();
  long AddOutput(vVideoFile * output);
  long RemoveOutput(long index);
  long RemoveCurrentOutput();
  long ClearOutputs();
  vVideoFile * GetOutput(long index);

  long write_frame();
  // Write consecutive frames, by calling alternatingly
  // ShowNextImage and write_frame.
  long write_frames0();
  long write_frames(long number_of_frames);

  long AddImage(long sequence, long frame);
  long RemoveImage(long index);
  long ClearImages();

  general_image * GetImage(long index);
  color_image * GetColorImage(long index);
  grayscale_image * GetGrayImage(long index);
  
  vWindow * GetWindow();

  long CompareContours(long frame1, long frame2);

  long CompareNormalizedContours(long frame1, long frame2);

  long CompareRescaledContours(long frame1, long frame2);

  long CompareRescaledContoursFast(long frame1, long frame2, long parameter);

  long CompareFrames(long frame1, long frame2);

  long CompareFrames(const char * filename1, long frame1,
                       const char * filename2, long frame2);

  long MatchingScores(const char * test_filename, 
                        long start, long end,
                        const char * training_filename, 
                        const char * results_base_name);

  long MakeSequence(const char * sequence_name, 
                      const char * pattern);

  long Recenter();
  long Center0();
  long Center(long center_row, long center_col);
  long CenterWithClick();
  long SetMagnifiedSize0();
  long ResetMagnifiedSize();
  long SetMagnifiedSize(long rows, long cols);

  long ConnectedComponents();
  long BiggestComponent();
  long SelectComponent();
  long SelectComponents();
  long DetectFaces0();
  long DetectFaces(FaceLocation ** locs);
  long PrintImage(color_image * the_image);
  long RedrawWindow();
  long RedrawImages();
  // Draws the current image of the current input, so that it
  // gets displayed on the window.
  long DrawCurrentImage();
  long SetBounds();
  long ResetBounds();
  long SetBounds(long top, long left, 
        	       long rows, long cols);
  long sequence_rectangle(long top, long left, 
                  	      long rows, long cols)
  {
    SetBounds (top, left, rows, cols);
  }
  // If resolution <= 0, we don't need to normalize coordinates.
  // Otherwise, we need to normalize them based on the magnification
  // factor (2 to the power of resolution), and the center_row and
  // center_col of the input.
  long NormalizedCoordinates(long in_row, long in_col,
                             long * out_row, long * out_col);
  long PrintNormalizedCoordinates0();
  long PrintNormalizedCoordinates(long in_row, long in_col);

  long ShowPixelInfo(long row, long col);
  long PrintPixelInfo();
  long PrintPixelInfo(long row, long col);
  long entry_value();
  long ShowFrameInfo();
  long FrameDifferences();

  // Makes the current image all black
  long ClearImage();
  long SetPixel();
  long SetPixel(long row, long col, uchar value);

  long ComputeMotion0()
  {
    ComputeMotion ();

    return 1;
  }
  v3dMatrix<double> * ComputeMotion();
  v3dMatrix<double> * ComputeMotion(long win_rows,
                        					    long win_cols);

  long ThresholdCurrentImage();
  long ThresholdCurrentImage(double threshold);
  long InverseThresholdImage();
  // Subtracts one (mod ring size) from the image buffer index.
  long ShowPreviousImage();
  // Adds one (mod ring size) to the image buffer index
  long ShowNextImage();

  //Subtract from or add a given number to the image buffer index.
  long RewindImageBuffer0();
  long ForwardImageBuffer0();

  long RewindImageBuffer(long frames_to_skip);
  long ForwardImageBuffer(long frames_to_skip);

  long UndirectedEdges();
  long AmplifyImage();
  long MultiplyImage();
  long FindSkin2();
  long ProcessPolygon();
  long InitializeSkinData();
  long PickSkinTables();
  long DetectSkin();
  long SaveCurrentImage();
  long ApplyProcedure(vPixelPred proc);
  long HeightAdjustment();
  long Trash();
  long Trash2();
  long Trash3();

  long DrawInput(long index);
  long SetCurrentInput();
  long SetCurrentInput(long index);
  long PreviousInput();
  long NextInput();

  // these functions set the input without redrawing. Useful when switching
  // inputs and immediately displaying something new, it prevents
  // the "flashing" effect caused by first drawing whatever the current image
  // was for that input and then drawing the new image.
  long SetCurrentInputSilent(long index);
  long PreviousInputSilent();
  long NextInputSilent();

  long RemoveCurrentInput();
  long remove_inputs();
  long MoveSequence();
  long MoveSequenceWithClick();
  long MoveSequence(long top, long left); 
  long move_sequence(long top, long left)
  {
    return MoveSequence(top, left);
  }

  long SetCurrentCounter();
  long SetCurrentCounterPercent(double percentage);
  long SaveImage();
  long SavePGM();
  long NextInSequence();
  long AdvanceSequence(long num_frames);
  // Moves all input sequences to the same frame as that of the current input sequence
  long SynchronizeSequences();

  long ShowCurrentFrame();
  long StoreCurrentFrame();
  long JumpToFrame0();
  long JumpToFrame(long frame);
  long make_current_frame(long frame)
  {
    return JumpToFrame(frame);
  }

  general_image * ReadCurrentImage();
  long SetCurrentFrame(long index);
  long PreviousInSequence();
  long FindMaximalMotion();
  long BackgroundStatistics();
  long Blur();
  long FindHands();
  long BrowseImage0();
  long browse_image(long number);
  long SetSequenceStep();
  long SetSequenceStep(long step);
  long AverageRegions();

  long InputCalibration();
  long UndistortFrame();
  long EstimateForeground();

  // Functions interfacing with vMatchScores
  long UpdateMatchScores(long frame_number);
  long LoadMatchScores();
  long LoadMatchScores(char * filename);
  long LoadMatchScoresN(long frame_number);
  long LoadMatchScoresN0();
  long SetScoreBase0();
  long SetScoreBase(const char * filename, long score_frame);
  long LoadCurrentScores();
  long LoadNextScores();
  long JumpToMatch();
  long JumpToMatch(long index);
  long ViewRank0();
  long ViewRank(long frame, long view);
  long SetScoreOrder(ushort order);
  long ScoresToBinary0();

  // Functions interfacing with vMatchScores and related
  // to viewing ground truth information
  long SetTestFilename0();
  long GroundMatch0();
  long CorrectMatch0();
  long SetTestFrame0();
  long GroundTruthStatistics0();

  long SetTestFilename(long input_index);
  long GroundMatch(long rank);
  long CorrectMatch(long rank);
  long SetTestFrame(long in_test_frame);
  long GroundTruthStatistics(long limit);
 // End of functions interfacing with vMatchScores

  long LoadImage();
  long WritePPM();
  long ProcessRegions();

  // Here we assume that the current image is a mask showing what we want to
  // track, and the current frame in the avi file corresponds to that mask.
  long InitializeSkinTracker();

  long SkinTrackNextFrame();
  long HandTrackSequences();
  long HandTrackSequence();
  long HandTrackSequence(const char * filename);

  // Detect and segment the hand from a video sequence, using
  // skin detection.
  long SkinSegmentSequence0a();
  long SkinSegmentSequence0b();
  long SkinSegmentSequence0c();
  long SkinSegmentSequence0d();

  long SkinSegmentSequence2(const char * input_file, const char * output_file);
  
  // Segments hands in the input sequence, and writes the segmented hands to 
  // the output sequence. It processes frame in the input sequence in the
  // range [start, ...,  end];
  long SkinSegmentSequence4(const char * input_file, const char * output_file, 
                              long start, long end);
  long SkinSegmentSequence5(const char * input_file, const char * output_file, 
                              long start, long end, long threshold);

  // Like SkinSegmentSequence4, but this one accepts a few more parameters
  // to make the segmentation work when the default parameters don't work.
  // Threshold is the threshold we use for classifying pixels as skin and
  // non-skin. (row, col) is the location of a pixel that must be in the
  // connected component that is chosen as the hand. -1 for any of those values
  // (start, end, threshold, row, col) means "don't care, use a default value".
  // For the row and col, default is not using at all a row and col, just
  // selecting the largest component.
  long SkinSegmentSequence7(const char * input_file, const char * output_file,
                              long start, long end, long threshold,
                              long row, long col);

  long SkinSegmentHand0a();
  long SkinSegmentHand0b();
  long SkinSegmentHand0c();
  color_image * SkinSegmentHand1(color_image * source_image);
  color_image * SkinSegmentHand2(color_image * source_image, long threshold);
  color_image * SkinSegmentHand4(color_image * source_image, long threshold, 
                             long row, long col);
  long NormalizeImage(double distance, double center_row, double center_col);
  long NormalizeLabeledImage(double distance, double center_row, 
                               double center_col);
  long CenterOfMass();
  long MatchWithRLESequence();
  long RotateLeft0();
  long RotateLeft(long left_turns);
  long RotateRight0();
  // Rotates the current image by 90 * right_turns degrees
  long RotateRight(long right_turns);
  
  long NewImage();
  long NewImage(long rows, long cols);
  long DrawLine();
  long DrawLineB();
  long DrawLine(long row1, long col1, long row2, long col2, long value);
  long DrawLine2(long row, long col, double angle, long length, long value);
  long DrawRectangle();
  long DrawRectangle(long row1, long col1, long row2, long col2, long value);
  long DrawCircle();
  long DrawCircle(long row, long col, long radius, long value);
  long DrawFilledRectangle();
  long DrawFilledRectangle(long row1, long col1, long row2, long col2, long value);
  long DrawFilledCircle();
  long DrawFilledCircle(long row, long col, long radius, long value);

  long SelectValue0();
  long SelectValue(long value);

  long CreateLineSet();
  long DestroyLineSet();
  long StartPolygon();
  long AddVertex(long row, long col);
  long DeleteVertex();
  // Draws the set of lines on a black background
  long DrawLineSet();
  // Draws the set of lines on top of the current image.
  long OverlayLineSet();
  long DecreaseResolution();
  long IncreaseResolution();
  long SetResolution0();
  long SetResolution(long resolution);

  // Start of section using the IPL library
  long CannyDetector0a();
  long CannyDetector0b();
  long CannyDetector(double low);
  long CannyDetector3(double low, double high, int aperture);
  // End of section using the IPL library

  // In every video sequence, we may want to store 
  // the color of the face 
  long SetFaceColor0();

  long SetHandThreshold0();
  long SetHandThreshold(long new_threshold);
  long DepthNeighbors();
  long MaxNeighbors();

  long Histogram0();
  long Histogram(long number_of_bins);
  long Histogram30();
  long Histogram3(long number_of_bins, double lower, double upper);

  long FingerEdges0();

  // Calls vArticulatedEdges on each frame of input video file with index
  // link_input, using for depth the corresponding frame of the video file
  // with index depth_input. It saves the results into the output
  // video sequence named with output.
  long FingerEdges(long link_input, long depth_input, const char * output);

  long BoundingBox0();
  long Translate0();
  long Translate(long row_diff, long col_diff);
  long ScaleImage0();
  long ScaleImage(double scaling);
  long ResizeImage0();
  long ResizeImage(long new_rows, long new_cols);
  long ScaleBoundingBox0();
  long ScaleBoundingBox0b();
  long ScaleBoundingBox(long new_box_size);
  long ScaleBoundingBox2(long new_box_size, long new_rows, long new_cols);

  long ContourScores0();
  long ContourScoresFast0();
  long ContourScores(long test_input, long training_input, 
                       long low_frame, long high_frame, 
                       const char * scores_file);

  long ContourScoresFast(long test_input, long training_input, 
                           long low_frame, long high_frame, long training_frames, 
                           long grid_size, long subsampling,
                           const char * scores_file);

  long ChamferScores0();
  long ChamferScores(long test_input, long training_input, 
                         long low_frame, long high_frame, 
                         const char * scores_file);

  long DistanceTransformSlow();
  long DistanceTransform();
  long SetDistance0();
  long SetDistance(long input1, long input2);

  long ChamferDistances0();
  long ChamferDistances(long input1, long input2);
  long ChamferDistancesSlow0();
  long ChamferDistancesSlow(long input1, long input2);

  long ImageMoments();
  long SequenceImageMoments_a();
  long SequenceImageMoments_b();
  long SequenceImageMoments1(const char * output_file);

  long LeftHand();
  long RightHand();
  long TopBottom();

  // Auxiliary functions for DatabaseFrame
  // This one is for the old database
  color_image * Training7Frame(long frame, long view);

  // This one is for the poser1 database
  color_image * Poser1Frame(long frame, long view);

  color_image * ReadDatabaseFrame(long frame, long view);
  long DatabaseFrame0();
  long DatabaseFrame(long frame, long view);
  long Pad();

  long make_short();
  long make_long();
  long make_ushort();
  long make_ulong();
  long make_float();
  long make_double();

  long histogram_equalization_on();
  long histogram_equalization_off();
  long histogram_equalization_size ();
  long histogram_equalization ();
  long histogram_equalization_offset ();
  long histogram_equalization_weights_test ();

  long toggle_preserve_current_frame();
  long current_percentage();
  long match_window_size();
  long set_frame_rate();
  long get_frame_rate();
  general_image * current_image ();
  class_sequence_information * current_sequence ();

  long set_buffer_size();
  long set_default_buffer_size();
  long check_input();
};







#endif // VASSILIS_INTERPRETER1_H
