#ifndef VASSILIS_INTERPRETER3_H
#define VASSILIS_INTERPRETER3_H


#include "interpreter2.h"
#include "hausdorff_matching.h"
#include "hand_database.h"
#include "learning/boost_map.h"
#include "chamfer_matching.h"
#include "gesture_data.h"
#include "learning/uci_learning.h"
#include "learning/boost_nn.h"
#include "learning/mnist.h"
#include "algorithms/embeddings.h"

// This class holds enough information to apply SparseMap to
// hand image retrieval (using the old database).
class class_SparseMapInfo
{
public:
  // Sets of reference images, in (frame, rotation)
  vector<v3dMatrix<long> > sets;
  
  // if indices[i][j][k] = x, then flattened_images(x) is a row 
  // that contains the frame and rotation of sets[i][j][k] (i.e.
  // of the entries in the 0 and 1 band of sets[i], for 
  // row = j and col = k).
  vector<vMatrix<long> > indices;
  vMatrix<long> flattened_images;

  // image_to_sets[i] will be the list of reference sets
  // in which flattened_images[i] occurs.
  vector<vector<long> > image_to_sets;

  char * Directory()
  {
    char * result = vJoinPaths3(sequences_directory, "glove_data",
                                 "sparse_map");
    return result;
  }

  class_SparseMapInfo() {}
};


// This class holds enough information to apply SparseMap to
// hand image retrieval (using the old database).
class class_FastMapInfo
{
public:
  // Sets of reference images, in (frame, rotation)
  vMatrix<long> pivots_matrix;
  // indices(i) will be the row containing the indices
  // of the i-th pair pivot images into reference images.
  vMatrix<long> indices_matrix;
  vMatrix<float> pivot_distances_matrix;
  vMatrix<float> optimized_pivot_distances;

  char * Directory()
  {
    char * result = vJoinPaths3(sequences_directory, "glove_data",
                                 "sparse_map");
    return result;
  }

  class_FastMapInfo() {}
};


class vInterpreter3 : public vInterpreter2
{
protected:
  // Distance transforms of database images, one distance
  // transform per viewpoint.
  vDatabaseEdgeSummary1 * edge_summary;
  VS_Histogram * vs_histogram;
  vPointSetOrientation * edge_orientations;
  vHandDatabase1 * hand_database;
  V_BoostMap * boost_map;
  vMatrix<float> bm_points;
  vChamferBoostInfo * cbi;
  vSignBoostInfo * sbi;
  vAllChamferDistances * acd;
  class_SparseMapInfo * sm_info;
  v3dMatrix<float> * sm_embeddings;
  class_FastMapInfo * fm_info;
  v3dMatrix<float> * fm_embeddings;
  vChamferEmbeddingBM * ceb;
  v3dMatrix<float> * ceb_embeddings;
  v3dMatrix<float> approximate_stats;
  class_BoostMap * & bm;
  BoostMap_dataset * & bm_data;

  vMnistSet * training_mnist;
  vMnistSet * test_mnist;

  class_FastMap * fm;

  double foreground_threshold;

  // Scales that hand shapes should be scaled to
  float x_scale, y_scale, z_scale;

  // Canonical sizes that rendered hand shape images should have
  // (for image size and bounding box size)
  long canonical_image_size;
  long canonical_bbox_size;
  long acd_input1;
  long acd_input2;

public:
  vInterpreter3(class_interpreter* in_interpreter);
  ~vInterpreter3();

  virtual void Initialize(vGestureInterfaceGlobals * in_globals);

  // An interesting thing with the CleanUp functions of each interpreter
  // (i.e. interpreter1, 2, 3), is that each of them gets called by
  // the destructor of that class. I don't understand why that happens.
  // I would expect that, since the CleanUp functions are all virtual,
  // only the CleanUp function of the lowest subclass (interpreter3
  // currently) should be called. I don't know if the way things work 
  // now is according to the C++ standard or if it is a Virtual Studio bug.
  virtual void CleanUp();

  vRLE_Sequence * DatabaseEdges();

  long WaitSeconds0();
  long WaitSeconds(long seconds);

  long DoubleSize();

  long SequenceDoubleSize();

  long CorrectToPrint();

  long RandomEdges();

  long SaveDTs();

  long LoadEdgeSummary();

  long SaveDatabasePoints();

  long TestPrepareChamfer();

  long QuickChamferScores();

  long ReplaceColor0();

  long ReplaceColor(uchar r1, uchar g1, uchar b1,
                    uchar r2, uchar g2, uchar b2);

  // This function is used to generate some data for George Kollios,
  // for the class project of a student. We generate a distance
  // matrix with pairwise chamfer distances between all database images
  // in a subset of the database.
  long DistanceMatrix();

  // For debugging
  long CheckDistanceMatrix();

  // This function is used in conjunction with DistanceMatrix. DistanceMatrix
  // generates chamfer distances between pairs of database images. 
  // GroundDistanceMatrix generates ground truth distances between pairs of 
  // database images.
  long OldGroundDistanceMatrix();

  // For debugging
  long CheckOldGroundDistanceMatrix();

  // George Kollios gave me a 10-dimensional embedding of the images 
  // used in DistanceMatrix() to generate the distance matrix. That
  // embedding was produced using Isomap. LipschitzVectors produces
  // an n-dimensional embedding of the same images. The argument 
  // "dimensions" specifies n. We assume that an appropriate 
  // reference set has already been loaded. The goal is to compare
  // the accuracy of the Lipschitz embedding with the accuracy of the
  // Isomap embedding. 
  long LipschitzVectors0();
  long LipschitzVectors(long dimensions);

  // For debugging
  long CheckLipschitzVectors();

  //-----------------------------------------------------------------

  // The next functions are used for matching lines and evaluating 
  // image similarity based on line matching scores.
  
  // Construct a votes/saliency histogram for input lines.
  long VS_Stats();

  // Match an individual model line segment to input line segments
  // using the VS histogram.
  // I should call this after calling VS_Stats() and one of the 
  // MatchModelLines variants.
  long MatchLineSegmentVS0();

  long MatchLineSegmentVS(long link_id, long line);
  
  // Get quick line scores, matching the input image to all database images.
  long QuickLineScores();

  // Find the best match for a pair of model lines
  long LinePairMatch0();
  long LinePairMatch(long link1, long side1, long link2, long side2);

  //-----------------------------------------------------------------

  // Saves the edge orientations of all database images.
  long SaveEdgeOrientations();

  long LoadEdgeOrientations();

  // Displayes the orientations of database edge pixels for a particular
  // frame and view, as stored in point_sequence an edge_orientations.
  // Useful for debugging.
  long ShowOrientations0();
  long ShowOrientations(long frame, long view);

  // This function is used to verify that the orientations saved 
  // using SaveEdgeOrientations are similar to the orientations
  // computed directly from the database images.
  long VerifyOrientations();

  // Used to see the result of scaling edges and orientations the way
  // we do in preparation for the distance/orientation transform.
  long TestEdgesOrientations();

  // Used to see the results of vDO_DistanceTransform();
  long Test_DO_Transform();

  // scores based on an extension of the chamfer distance, that takes into
  // account edge orientation in addition to edge location.
  long Chamfer_DO_MI_Scores();

  long Chamfer_DO_Scores();

  // This function initializes the same structures as SkinHand(),
  // but it simply assumes that the current input image is the
  // hand image. This is the right thing for presegmented sequences,
  // like hands1.avi
  long TrivialSegmentation();

  // Picks a random bounding box, of size canonical_bbox_size.
  long RandomSegmentation();

  long NormalizeIntensity();

  long WriteRange0();

  long WriteRange(long low, long high);

  long InputFaceColor0();

  long InputFaceColor(long r, long g);

  long OldSphereSurfaceTo3D0();
  long OldSphereSurfaceTo3D(double x, double y);

  long OldThreeDToSphereSurface0();
  long OldThreeDToSphereSurface(double x, double y, double z);

  // These functions are useful to convert BMP files generated by Poser
  // to RLE sequences containing many images. Poser (using my
  // save_viewpoints.py script) generates images named with a 
  // base_number1_number2.bmp template, where number1 is handshape
  // and number2 is viewpoint number.
  long BMP_To_RLE0();
  long BMP_To_RLE(const char * base);

  long BMP_To_RLE0b();
  long BMP_To_RLE2(const char * base);

  // These functions are used to look directly at values stored at the 
  // skin table.
  long DebugSkinTable0();
  long DebugSkinTable(long r, long g, long b);

  // Load the specified hand database.
  long SetHandDatabase0();
  long SetHandDatabase(const char * name);

  // Reload the hand database (this is necessary if we have 
  // created new sequences that the database now needs to load).
  long ReloadHandDatabase();

  // Prints the name of the currently loaded hand database.
  long DatabaseName();

  // Color the forearm as background.
  long RemoveForearm();

  // Create the database image sequence, from the original
  // poser sequence.
  long CreateImageSequence();

  long SaveDatabaseBoundingBoxes();

  // Applies the articulated edge extraction as specified in 
  // the database with that name. This is just for debugging.
  long ArticulatedEdges();

  // This function is used to verify that the SaveBoundingBoxes
  // function in vHandDatabase1 does the right thing.
  long ReadBoundingBox0();
  long ReadBoundingBox(long frame);

  // This function is used to display the results of 
  // vHandDatabase1::EdgeImage
  long DatabaseEdgeImage0();
  // This version works on the currently displayed frame, assuming
  // it comes from the big image sequence.
  long DatabaseEdgeImage();
  long DatabaseEdgeImage1(long frame);

  // Recreate, based on Poser, a sequence of edge images 
  // similar to the edge images I
  // had generated using VirtualHand.
  long RecreateOldEdgeSequence();

  // Create normalized cartoon sequence based on the unnormalized
  // sequence generated by Poser.
  long CreateCartoonSequence();

  // Take in a database that is saved as one image per shape/viewpoint
  // combination and convert it into a database that saves 12
  // images per shape/viewpoint combination, as in the original
  // training7 database.
  long SaveOldFormat0();
  long SaveOldFormat(const char * name);

  // The next functions are used to test and debug vScaleImageDown8
  long ScaleImageA0();
  long ScaleImageA(long bbox_size, long image_size);
  long ScaleImageB0();
  long ScaleImageB(double scaling, long center_row, long center_col,
                   long new_center_row, long new_center_col,
                   long new_rows, long new_cols);

  // Create sequence of DT transforms for a database edge sequence.
  // inputs: database name, and ending specifying the particular
  // edge sequence we want to work on (cartoon edges, articulated edges,
  // or whatever else).
  long CreateDTSequence0();
  long CreateDTSequence(const char * ending);

  // Create point sequence for a database edge sequence.
  // inputs: database name, and ending specifying the particular
  // edge sequence we want to work on (cartoon edges, articulated edges,
  // or whatever else).
  long CreatePointSequence0();
  long CreatePointSequence(const char * ending);

  long DatabaseLoadDT0();
  long DatabaseLoadDT(const char * ending);

  long DatabaseSetCurrentDT0();
  long DatabaseSetCurrentDT(const char * ending);

  long DatabaseClearDT0();
  long DatabaseClearDT(const char * ending);

  long DatabaseClearAllDTs();
  
  long DatabaseGetDT0();
  long DatabaseGetDT(long frame);

  long DatabaseLoadPointSequence0();
  long DatabaseLoadPointSequence(const char * ending);

  long DatabaseSetCurrentPointSequence0();
  long DatabaseSetCurrentPointSequence(const char * ending);

  long DatabaseClearPointSequence0();
  long DatabaseClearPointSequence(const char * ending);

  long DatabaseClearAllPointSequences();

  long DatabasePointImage0();
  long DatabasePointImage(long frame);

  long DatabaseRotationTemplate0();
  long DatabaseRotationTemplate(long index);

  // Test the vWarpImage and vWarpImageSparse functions, by getting
  // rotation templates and applying them to the current image.
  long TestWarpImage0();
  long TestWarpImage(long degrees);

  long TestWarpImageSparse0();
  long TestWarpImageSparse(long degrees);

  long CroppedUchar();

  long TestChamferFeatures0();
  long TestChamferFeatures(long index);

  // Tests the function vFrameViewFinder::ConvertPoser1Pose.
  long ConvertPose0();
  long ConvertPose(long frame, long view);

  // Tests the function vFrameViewFinder::ConvertToPoser1Pose.
  long Poser1Pose0();
  long Poser1Pose(long frame, long view);

  // Chamfer scores using the database "poser1". Operates
  // on the current hand_edge_image.
  long Poser1ChamferScores();

  // Set the current image as the current hand edge image.
  long SetHandEdgeImage();

  long OldLineOrientation0();
  long OldLineOrientation(double camera_x1, double camera_y1, 
                       double camera_z1, double angle1,
                       double camera_x2, double camera_y2, 
                       double camera_z2, double angle2);

  // These functions are for debugging LineOrientation.
  long VerifyMultiplication();
  long VerifyInverse();
  long VerifyRotation();
  long VerifyRotationMatrix();
  long OldVerifyPlaneIntersection0();
  long OldVerifyPlaneIntersection(double camera_x1, double camera_y1,
                               double camera_z1, double angle1,
                               double camera_x2, double camera_y2,
                               double camera_z2, double angle2);

  long OldVerifyPlaneIntersection20();
  long OldVerifyPlaneIntersection2(double camera_x1, double camera_y1,
                               double camera_z1, double angle1,
                               double camera_x2, double camera_y2,
                               double camera_z2, double angle2);

  long LineVoronoi();

  long DominanceTransform();

  // Draws circles for the top number points (with highest values)
  // on current image. The circles are drawn on a copy of the
  // image with the specified buffer index.
  long TopCircles0();
  long TopCircles(long number, long buffer_index);

  // Draws a random quadrilateral, so that we can test the 
  // vQuadPixels function.
  long RandomQuadrilateral();
  long DrawQuadrilateral0();
  long DrawQuadrilateral(class_pointer (long) rows, class_pointer (long) cols);

  long DrawQuadrilateral0b();
  long DrawQuadrilateral2(class_pointer (long) rows, class_pointer (long) cols);
  
  long DrawTriangle0();
  long DrawTriangle(class_pointer (long) rows, class_pointer (long) cols);

  // Selects a color from the current image (assuming it is a color image).
  long SelectColor0();
  long SelectColor(uchar r, uchar g, uchar b);

  // Counts the pixels having the specified color in the current image 
  // (assuming it is a color image).
  long CountColor0();
  long CountColor(uchar r, uchar g, uchar b);

  // Compares the RGB versions of the current images of the two
  // inputs.
  long CompareInputs0();
  long CompareInputs(long input1, long input2);
  long compare_images();
  long PreprocessPolygons();

  long HandPointsSlow0();
  long HandPointsSlow(const char * shape, double roll, 
                  double pitch, double yaw);
  long TimeHandPointsSlow0();
  long TimeHandPointsSlow(const char * shape, double roll, 
                      double pitch, double yaw, long times);
  long HandPolygonsSlow0();
  long HandPolygonsSlow(const char * shape, double roll, 
                  double pitch, double yaw);
  long TimeHandPolygonsSlow0();
  long TimeHandPolygonsSlow(const char * shape, double roll, 
                      double pitch, double yaw, long times);

  long HandLinksSlow0();
  long HandLinksSlow(const char * shape, double roll, 
                 double pitch, double yaw);
  long TimeHandLinksSlow0();
  long TimeHandLinksSlow(const char * shape, double roll, 
                     double pitch, double yaw, long times);

  long HandPoints0();
  long HandPoints(const char * shape, double roll, 
                  double pitch, double yaw);
  long TimeHandPoints0();
  long TimeHandPoints(const char * shape, double roll, 
                      double pitch, double yaw, long times);
  long HandPolygons0();
  long HandPolygons(const char * shape, double roll, 
                  double pitch, double yaw);
  long TimeHandPolygons0();
  long TimeHandPolygons(const char * shape, double roll, 
                      double pitch, double yaw, long times);

  long HandLinks0();
  long HandLinks(const char * shape, double roll, 
                 double pitch, double yaw);
  long TimeHandLinks0();
  long TimeHandLinks(const char * shape, double roll, 
                     double pitch, double yaw, long times);

  long TestScaleX0();
  long TestScaleX(const char * shape, double roll, 
                  double pitch, double yaw, double x_scale);

  long AltHandPoints0();
  long AltHandPoints(const char * order, const char * shape, double roll, 
                     double pitch, double yaw);

  long AltHandPolygons0();
  long AltHandPolygons(const char * order, const char * shape, double roll, 
                       double pitch, double yaw);

  long AltHandLinks0();
  long AltHandLinks(const char * order, const char * shape, double roll, 
                    double pitch, double yaw);

  long SetScale0();
  long SetScale(double x, double y, double z);

  long SetCanonicalSizes0();
  long SetCanonicalSizes(long size, long bbox_size);

  long ShapeMatrices0();
  long ShapeMatrices(const char * order, const char * shape);

  long LinkMatrices0();
  long LinkMatrices(const char * order, const char * shape, long link_id);

  long SideLengthStats();

  long CutTriangle0();
  long CutTriangle(class_pointer (long) rows, class_pointer (long) cols, long max_side);

  long TrianglePoints3D0();
  long TrianglePoints3D(vPoint3f * vertices, float max_distance);

  long TrianglePixels0();
  long TrianglePixels(class_pointer (long) rows, class_pointer (long) cols);
  long TimeTrianglePixels0();
  long TimeTrianglePixels(class_pointer (long) rows, class_pointer (long) cols, long times);

  long TrianglePixels0b();
  long TrianglePixels2(class_pointer (long) rows, class_pointer (long) cols);
  long TimeTrianglePixels0b();
  long TimeTrianglePixels2(class_pointer (long) rows, class_pointer (long) cols, long times);

  long TestLinePoints50();
  long TestLinePoints5(long row1, long col1, long row2, long col2);

  long TestLineCols0();
  long TestLineCols(long row, long col);
  long test_line_pixels ();

  long OnlineSegmentation();
  long OnlineRanking();
  long CameraBackground();
  long ShowBackground();
  long ForegroundThreshold0();
  long ForegroundThreshold(double threshold);
  long ForegroundHand();

  long PrecomputedTriangle0();
  long PrecomputedTriangle(long * rows, long * cols);

  long TimeVertexOperations0();
  long TimeVertexOperations(const char * shape, double roll, 
                            double pitch, double yaw, long times);

  long TimeImage0();
  long TimeImage(long times);

  // Beginning of functions written for the submission to CVPR 2004.
  // Create a boost map for 2D points.
  long PointBM0();
  long PointBM(long points, long triples);
  long DeleteBM();
  long BM_PointImage();
  long BM_BasisImage();
  long BM_PrintSummary();
  long BM_PrintClassifier();
  long BM_PrintAll();
  long BM_Step0();
  long BM_Step(const char * filename);
  long BM_Steps0();
  long BM_Steps(char * filename, long steps);
  long BM_Random0();
  long BM_Random(long number, long times);
  long BM_Training0();
  long BM_Training(long number);
  long BM_Validation0();
  long BM_Validation(long number);
  long BM_AllowNegative0();
  long BM_AllowNegative(long value);
  long BM_AllowRemovals0();
  long BM_AllowRemovals(long value);
  long BM_AllowLipschitz0();
  long BM_AllowLipschitz(long value);
  long BM_AllowProjections0();
  long BM_AllowProjections(long value);
  long BM_Load0();
  long BM_Load(const char * filename, long dimensions);

  // ChamferBoostInfo classes.
  long MakeCBI0();
  long MakeCBI(long sampling, long training);
  long DeleteCBI();
  long SaveCBI();
  long LoadCBI0();
  long LoadCBI(long sampling);
  long CBI_Indices();
  long CBI_Distances();
  long CBI_Distance0();
  long CBI_Distance(long i, long j);
  long CBI_Triples();

  long CBI_BM();
  long CBI_Steps0();
  long CBI_Steps(char * base, long steps);
  long CBI_Random0();
  long CBI_Random(long number);
  long CBI_LoadBM0();
  long CBI_LoadBM(char * base, long steps);

  long CBI_Training0();
  long CBI_Training(long number);
  long CBI_Validation0();
  long CBI_Validation(long number);
  long CBI_SM0();
  long CBI_SM(long number);
  long CBI_FM();

  long TimeExp0();
  long TimeExp(long times, float argument);
  long TimePrecomputedExp0();
  long TimePrecomputedExp(long times, float argument);

  long ACD_CheckInputs();
  long ACD_SetInputs0();
  long ACD_SetInputs(long input1, long input2);
  long ACD_SaveDirected0();
  long ACD_SaveDirected(long start, long end, long order);
  long ACD_TestDirected0();
  long ACD_TestDirected(long frame1, long view1, 
                        long frame2, long view2, long order);
  long ACD_Save0();
  long ACD_Save(long start, long end, long chunk);
  long ACD_Test0();
  long ACD_Test(long frame1, long view1, long frame2, long view2);

  long ACD_NN0();
  long ACD_NN(long frame, long view, long k);
  long ACD_ApproximateRank0();
  long ACD_ApproximateRank(long frame1, long view1,
                           long frame2, long view2);
  long ACD_ApproximateLoss0();
  long ACD_ApproximateLoss(long frame1, long view1, long k);
  long ACD_LossStats0();
  long ACD_LossStats(long k, long times, long print_all);


  long LoadBoostMap0();
  long LoadBoostMap(const char * base, long steps);
  long BM_Scores();
  long PrintReferenceImages();

  long ShowEmbedding();
  long CompareEmbeddings0();
  long CompareEmbeddings(long rank1, long rank2);
  long PositiveWeights();

  long TestKthSmallest0();
  long TestKthSmallest(long k, long size, long range);
  long TestSample0();
  long TestSample(long low, long high, long number);

  long PointSM0();
  long PointSM(long points, long triples);
  long SM_Features0();
  long SM_Features(long in_features);
  long SparseMapInfo();
  long SparseMapInfoB();
  long SM_Scores();
  // Compute the training embedding using SparseMap
  long SM_TrainingEmbedding();

  long SM_SaveEmbedding0();
  long SM_SaveEmbedding(const char * name);
  long SM_LoadEmbedding0();
  long SM_LoadEmbedding(const char * name);
  long SM_ReferenceSet0();
  long SM_ReferenceSet(long dimension);
  // Look at a specific coordinate of the SparseMap embedding
  // for a particular database image, and all the distances
  // to images in the corresponding reference set.
  long SM_TrainingCoordinate0();
  long SM_TrainingCoordinate(long frame, long rotation);
  long SM_TestCoordinate0();
  long SM_TestCoordinate(long dimension);

  long ShowOriginalFrame();
  long TimeDT0();
  long TimeDT(long times);
  long EmbeddingDTs();
  long PrintTestEmbedding();

  long PointFM0();
  long PointFM(long points, long triples);
  long FM_PointImage();
  long FM_BasisImage();
  long FM_Pivots();

  long FM_LoadEmbedding0();
  long FM_LoadEmbedding(const char * name);
  // Compute the training embedding using FastMap
  long FM_TrainingEmbedding();
  long FM_TestEmbedding();
  long FM_Scores();
  long FM_PrintInfo();
  long FM_TrainingCoords0();
  long FM_TrainingCoords(long frame, long view);
  long FM_TestCoords();
  long FM_LineProjection0();
  long FM_LineProjection(float ai, float bi, float ab);
  long FM_Distances();

  long MakeSBI0();
  long MakeSBI(long training);
  long DeleteSBI();
  long SaveSBI();
  long LoadSBI0();
  long LoadSBI(long sampling);
  long SBI_Print();

  long SBI_BM();
  long SBI_Training0();
  long SBI_Training(long number);
  long SBI_SM0();
  long SBI_SM(long candidates);
  long SBI_FM();

  long SBI_BM_Stats0();
  long SBI_BM_Stats(long max_dimensions, long max_k);
  long SBI_FM_Stats0();
  long SBI_FM_Stats(long max_dimensions, long max_k);

  long SBI_BM_Stats0d();
  long SBI_BM_Stats4(long min_dimensions, long max_dimensions, 
                     long min_k, long max_k);
  long SBI_FM_Stats0d();
  long SBI_FM_Stats4(long min_dimensions, long max_dimensions, 
                     long min_k, long max_k);
  long SBI_SM_Stats0d();
  long SBI_SM_Stats4(long min_dimensions, long max_dimensions, 
                     long min_k, long max_k);

  // HRCM stands for highest-ranking correct match, and AHRCM
  // stands for approximate highest-ranking correct match.
  long SBI_BM_AHRCM0d();
  long SBI_BM_AHRCM4(long min_dimensions, long max_dimensions, 
                     long min_k, long max_k);
  long SBI_FM_AHRCM0d();
  long SBI_FM_AHRCM4(long min_dimensions, long max_dimensions, 
                     long min_k, long max_k);
  long SBI_SM_AHRCM0d();
  long SBI_SM_AHRCM4(long min_dimensions, long max_dimensions, 
                     long min_k, long max_k);
  long SBI_HRCM0();
  long SBI_HRCM(long min_k, long max_k);

  long SbiDataset();

  long KthRankStats0();
  long KthRankStats(long k);

  long CEB_Create();
  long CEB_Print();
  long CEB_Save0();
  long CEB_Save(const char * filename);
  long CEB_Load0();
  long CEB_Load(const char * filename);
  long CEB_Embeddings();
  long CEB_Scores();

  // Based on the current sequence_match_stats object, which is
  // supposed to know for each frame the exact nearest neighbor,
  // figure out the rank of that neighbor at each frame as computed
  // bo score_script.
  long NN_RankStats0();
  long NN_RankStats(const char * score_script, long start_frame, 
                    long end_frame);

  long CombineSequenceStats0();
  long CombineSequenceStats(const char * in_file1, const char * in_file2, 
                            const char * out_file);

  long WriteBigDistances0();
  long WriteBigDistances(long number);
  long ReadBigDistances0();
  long ReadBigDistances(long number, long frame);

  // End of functions written for the submission to CVPR 2004.

  // Test the function vComplement.
  long Complement0();
  long Complement(long number, long start, long end);

  // Test the function vPermutation
  long Permutation0();
  long Permutation(long start, long end);


  // Beginning of functions written for potential submission to KDD 2004.
  long MnistLoad0();
  long MnistLoad(long dataset);
  long MnistLoad30();
  long MnistLoad3(long dataset, long start, long end);

  long MnistShow0();
  long MnistShow(long number);

  long MnistDataset20();
  long MnistDataset2(const char * name, long number);
  long MnistDataset40();
  long MnistDataset4(const char * name, long candidates, 
                     long training, long validation);
  long MnistCandidates();
  long MnistTraining();
  long MnistValidation();

  long MnistWholeDataset();
  long MnistLabels();
  long MnistTestTrain();
  long MnistTrainTrain();

  long MnistChamfer0();
  long MnistChamfer(long index1, long index2);

  // these functions read my results
  char * EmbeddingDirectory2();
  long FastMapFilterRefine();
  long FastMapKnn();
  long FastMapIndexErrors();

  // these functions read Panagiotis' results
  char * EmbeddingDirectory();

  long FmMake20();
  long FmMake2(const char * dataset, const char * sample_name);
  long FmMake30();
  long FmMake3(const char * dataset, const char * sample_name,
               const char * saved_file);
  long FmDelete();

  long FmNextStep();
  long FmNextSteps0();
  long FmNextSteps(long steps);
  long FmEmbedTest();
  long FmEmbedTraining();
  long FmSave0();
  long FmSave(const char * filename);
  long FmPrintSample0();
  long FmPrintSample(long i);
  long FmPrintTraining0();
  long FmPrintTraining(long i);
  long FmPrintPivots();
  long FmTest();

  // End of functions written for the attempted (but not 
  // completed) KDD 2004 submission.

  // SaveEdgeImages is used to save the database of edge
  // images, so that I can give it to other people or just
  // put it on the web. For the time being, I will
  // just save it all as an uncompressed AVI file.
  long SaveEdgeImages0();

  long SaveEdgeImages(const char * filename);

  // ConvertSequence takes a sequence of frames, from the current
  // input, and converts each frame by applying to it a given script,
  // and then saves the output of that frame to the output sequence.
  long ConvertSequence0();

  long ConvertSequence(const char * script_file, long start, long end, long step);
  long PrintClassificationError(vMatrix<float> results);
};











#endif // VASSILIS_INTERPRETER3_H

