#ifndef VASSILIS_PARTICLE_TRACKER_H
#define VASSILIS_PARTICLE_TRACKER_H


#include "basics/auxiliaries.h"
#include "ranking.h"

class vTrackingEstimate
{
public:
  double probability;

public:
  vTrackingEstimate();
  ~vTrackingEstimate();
  inline double GetProbability() { return probability; }
  inline void SetProbability(double in_probability) { probability = in_probability; }
};


class vTrackingObservation
{
protected:
  // This will be used to verify that the vTrackingObservation object 
  // is of the correct subtype.
  long type;

public:
  vTrackingObservation();
  ~vTrackingObservation();
  long Type()
  {
    return type;
  }
  virtual long ClassType() = 0;
};


class vParticleTracker
{
protected:
  long number_of_samples;
  vArray(vTrackingEstimate *) particles;

  virtual long Propagate() = 0;
  virtual long Resample() = 0;

public:
  vParticleTracker();
  ~vParticleTracker();
  
  virtual long SetNumberOfSamples(long in_number_of_samples) = 0;
  
  inline long GetNumberOfSamples()
  {
    return number_of_samples;
  }

  inline vArray(vTrackingEstimate *) GetEstimates()
  {
    return particles;
  }

  virtual long InitializeEstimates(vTrackingObservation * observation) = 0;

  virtual long ProcessObservation(vTrackingObservation * observation) = 0;
};



class vHandShapeEstimate : public vTrackingEstimate
{
protected:
  long shape_id;
  double x_angle, y_angle, z_angle;
  long database_frame;
  long database_view;

public:
  vHandShapeEstimate();
  vHandShapeEstimate(long database_frame, long database_view);
  ~vHandShapeEstimate();

  inline long ShapeID() { return shape_id; }
  inline double X_Angle() { return x_angle; }
  inline double Y_Angle() { return y_angle; }
  inline double Z_Angle() { return z_angle; }
  inline long DatabaseFrame() { return database_frame; }
  inline long DatabaseView() { return database_view; }
  inline void SetEstimate(vFrameViewFinder * finder, long in_frame, long in_view)
  {
    database_frame = in_frame;
    database_view = in_view;
    finder->GetViewParams(database_frame, database_view, &shape_id,
                          &x_angle, &y_angle, &z_angle);
  }
};


class vHandShapeEstimate2 : public vHandShapeEstimate
{
public:
  vArray(double) speed_probabilities;

public:
  vHandShapeEstimate2();
  vHandShapeEstimate2(long database_frame, long database_view, 
                       long number_of_speeds);
  ~vHandShapeEstimate2();
};



class vHandShapeObservation : public vTrackingObservation
{
private:
  vMatchScores * scores;

public:
  vHandShapeObservation();
  vHandShapeObservation(vMatchScores * in_scores);
  ~vHandShapeObservation();
  virtual long ClassType();

  inline vMatchScores * GetScores() { return scores; }
};


// The class vHandShapeTracker is a hand tracker that tracks hand shape
// and position. No estimates of velocity/acceleration are maintained.
// Essentially every particle corresponds to an image in the database
// of hand images.

class vHandShapeTracker : public vParticleTracker
{
protected:
  vFrameViewFinder view_finder;
  vMatrix<uchar> angle_distances_matrix;
  long number_of_poses;
  long number_of_shapes;
  long number_of_orientations;
  long number_of_viewpoints;
  long number_of_hand_views;

  // angle_distances[i][j] will store the angle distance between
  // camera params (x_angle_i, y_angle_i, 0) and the camera params
  // of frame j. x_angle_i is the contents of angles_x[i] in 
  // FrameViewFinder.
  vArray2(uchar) angle_distances;

  // angle_probabilities[i] will be the probability of the next
  // frame having camera params at distance i from the camera params
  // at the current frame. This is defined ad hoc as a prior.
  double angle_probabilities[360];

  // The way the database views and scores are ordered,
  // the frames adjacent to frame 11, view 0, in camera parameter space
  // are frame 10, view 0, and frame 0, view 3. It would have been more
  // reasonable if it were frame 0, view 1 instead of frame 0, view 3.
  // At any rate, here we want to map frame/view pairs to a number
  // between 0 and 48 such that the z angle difference between
  // two frame/view pairs that map onto indices I and J depends only
  // on J-I.
  long z_angle_index[12][4];

  double RankProbability(long rank);

  double CameraParamsProbability(double current_x, double current_y, double current_z, 
                                 double next_x, double next_y, double next_z);

  double AngleDistanceProbability(double distance);
  
  // This function ignores the shape information that corresponds to frame1 and 
  // frame2. It simply returns the probability that, given that the camera params
  // at frame i correspond to frame1/view1, the camera params at frame i+1
  // will correspond to frame2/view2.
  inline double CameraParamsProbability(long frame1, long view1, 
                                        long frame2, long view2)
  {
    long distance = CameraParamsDistance(frame1, view1, frame2, view2);
    return angle_probabilities[distance];
  }


  inline long CameraParamsDistance(long frame1, long view1, 
                                   long frame2, long view2)
  {
    const long frames_per_pose = view_finder.FramesPerPose();
    const long number_of_orientations = view_finder.NumberOfOrientations();
    long index1 = frame1 / frames_per_pose;
    long angle_index1 = index1 / number_of_orientations;
    long temp_z_index1 = index1 % number_of_orientations;
    long z_index1 = z_angle_index[temp_z_index1][view1];

    long index2 = frame2 / frames_per_pose;
    long angle_index2 = index2 / number_of_orientations;
    long temp_z_index2 = index2 % number_of_orientations;
    long z_index2 = z_angle_index[temp_z_index2][view2];

    long index2_base = angle_index2 * 48;
    long displacement = z_index2 - z_index1;
    if (displacement < 0) displacement += 48;
    return angle_distances[angle_index1][index2_base + displacement];
  }

  double ShapeProbability(long current_shape_id, long next_shape_id);
  long PrecomputeAngleDistances();

protected:
  virtual long Resample();
  long ClearAll();
  long ZeroAll();
  long SetNumberOfSamples(long in_number);

public:
  vHandShapeTracker();
  ~vHandShapeTracker();

  virtual long InitializeEstimates (vTrackingObservation * observation);
  virtual long Propagate();
  virtual long ProcessObservation(vTrackingObservation * observation);
  virtual vMatchScores * ParticleScores();
};


// class vHandShapeTracker2 is different than vHandShapeTracker
// in that, in addition to tracking hand shape and camera parameters,
// it also tracks velocity.
class vHandShapeTracker2 : public vHandShapeTracker
{
protected:
  // This array will contain in neighbors[i][j] the index number
  // for the j-th neighbor to camera parameters indexed by i.
  // When I say "camera parameters indexed by i" I mean the 4128 different
  // possible camera parameters under which each frame is rendered.
  // Neighbors don't have to be ordered in order of distance from pose i,
  // I'm not sure what the most convenient way to do it will be.
  v3dMatrix<long> neighbors_matrix;
  vArray2(long) neighbors;

  // neighbor_angles[i][j] will be the difference in angles between camera parameters
  // i and j.
  v3dMatrix<ushort> neighbor_angles_matrix;
  vArray2(ushort) neighbor_angles;

  // max_neighbors is the maximum number of neighbors that a set of camera
  // parameters can influence. That determines the second dimension in
  // the neighbors and angles arrays. Determined by max_distance;
  long max_neighbors;

  // max_distance is the maximum distance allowed between two sets of camera
  // parameters if we want one of them to influence the other one.
  long max_distance;
  
  long PrecomputeAngleDistances();

protected:
  virtual long Resample();
  long ClearAll();
  long ZeroAll();
  // If SetNumberOfSamples becomes public then we should make it virtual,
  // and we should make sure it is not called by the default constructors.
  long SetNumberOfSamples(long in_number);

public:
  vHandShapeTracker2();
  ~vHandShapeTracker2();

  virtual long InitializeEstimates (vTrackingObservation * observation);
  virtual long Propagate();
  virtual long ProcessObservation(vTrackingObservation * observation);
  virtual vMatchScores * ParticleScores();
};








#endif // VASSILIS_PARTICLE_TRACKER_H
