#include "stdafx.h"

#include "gesture_data.h"
#include "projections.h"
#include "quaternion.h"
#include "hand_database.h"
#include "algorithms.h"
#include "algo_templates.h"
#include "fingers.h"

#include "basics/definitions.h"


static const double pi = M_PI;


vMatrix<float> S_DoubleToFloat(vMatrix<double> matrix)
{
  long rows = matrix.Rows();
  long cols = matrix.Cols();
  vMatrix<float> result(rows, cols);
  vArray(double) data1 = matrix.Matrix();
  vArray(float) data2 = result.Matrix();
  long i;
  long size = rows * cols;
  for (i = 0; i < size; i++)
  {
    data2[i] = (float) data1[i];
  }
  
  return result;
}


vMatrix<double> S_FloatToDouble(vMatrix<float> matrix)
{
  long rows = matrix.Rows();
  long cols = matrix.Cols();
  vMatrix<double> result(rows, cols);
  vArray(float) data1 = matrix.Matrix();
  vArray(double) data2 = result.Matrix();
  long i;
  long size = rows * cols;
  for (i = 0; i < size; i++)
  {
    data2[i] = (double) data1[i];
  }
  
  return result;
}


// result = vLinkCameraParams(camera_params, world_matrix, orientation, viewpoints)
// arguments:
// camera_params: current camera parameters in poser.
// world_matrix: current world matrix of the link that the camera is pointing to.
// orientation: current orientation of the link that the camera is pointing to.
// viewpoints: a list of 3D orientations that we want to see the link under.
//
// result: a list of camera parameters (roll, pitch, yaw) that will make
// the link seen under the desired 3D orientations
vMatrix<double> vLinkCameraParams(vMatrix<double> world_matrix, 
                                    vMatrix<double> orientation,
                                    vMatrix<double> viewpoints)
{
  vMatrix<double> orientation_matrix = vOrientationMatrix(orientation);
  vMatrix<double> world_projection = world_matrix * orientation_matrix;
  long number_of_views = viewpoints.Rows();

  vMatrix<double> result(number_of_views, 3);
  vMatrix<double> z_axis(1, 4);
  z_axis.WriteN(4, 0.,0.,1.,1.);
  vMatrix<double> eye4 = vMatrix<double>::I(4);

  int i;
  for (i = 0; i < number_of_views; i++)
  {
    vMatrix<double> view_matrix = vCameraMatrix(0, viewpoints(i, 0), viewpoints(i, 1), 
                                                  eye4, z_axis);
    vMatrix<double> inv_cm = view_matrix * world_projection.Inv();
    vMatrix<double> cm = inv_cm.Inv();
    vMatrix<double> link_orientation = vLinkOrientation(cm);
    result.WriteRowN(i, 0, 3, link_orientation(0), 
                     link_orientation(1), link_orientation(2));
  }

  return result;
}


// vOrientationMatrix(orientation) returns a matrix 
// that describes the rotation that must be performed
// to link coordinates before we apply the link world
// matrix to them. The argument "orientation"
// is a tuple describing rotations around the x axis,
// y axis, and z axis. The result is the matrix corresponding to 
// the product of these rotations. The order of the multiplications
// is not specified in Poser, I've been trying to figure that out.
vMatrix<double> vOrientationMatrix(vMatrix<double> orientation)
{
  double factor = pi / 180.0;
  double x = orientation(0) * factor;
  double y = orientation(1) * factor;
  double z = orientation(2) * factor;
  vMatrix<double> x_matrix4 = vRotationMatrix4(x, 1, 0, 0);
  vMatrix<double> y_matrix4 = vRotationMatrix4(y, 0, 1, 0);
  vMatrix<double> z_matrix4 = vRotationMatrix4(z, 0, 0, 1);

  // I wasn't sure what the right order is for multiplying the matrices, 
  // I have verified it experimentally.
  vMatrix<double> result = z_matrix4 * y_matrix4 * x_matrix4;

  return result;
}


vMatrix<float> vOrientationMatrixFloat(vMatrix<float> orientation)
{
  vMatrix<double> double_orientation = S_FloatToDouble(orientation);
  vMatrix<double> double_result = vOrientationMatrix(double_orientation);
  vMatrix<float> result = S_DoubleToFloat(double_result);
  return result;
}



// returns the 4x4 projective matrix representing rotating by angle 
// theta around the 3D axis (x,y,z)
vMatrix<double> vRotationMatrix4(double theta, double x, double y, double z)
{
  vMatrix<double> matrix3 = vRotationMatrix(theta, x, y, z);
  vMatrix<double> result = vMatrix<double>::I(4);
  matrix3.Copy(&result, 0, 0);
  return result;
}


// returns the 3x3 matrix representing rotation by theta radians
// around axis going from origin to (x, y, z). Theta is in radians.
vMatrix<double> vRotationMatrix(double theta, double x, double y, double z)
{
  vQuaternion quaternion = vAngleQuaternion(theta, x, y, z);
  vMatrix<double> result = vQuaternionMatrix(quaternion);
  return result;
}


// result = vCameraMatrix(roll, pitch, yaw, canonical_matrix, center)
// arguments:
// roll, pitch, yaw, are the camera parameters (in degrees) as given by Poser.
// canonical_matrix is the camera matrix (as given by Poser) corresponding
// to roll = 0, pitch = 0, yaw = 0. All angles are passed in degrees.
// center is the center of the viewing sphere. We assume that the camera
// is always located on the viewing sphere, on the only position that
// allows it to point at the center while attaining the three orientation
// parameters (roll, pitch and yaw).
// roll rotates along the image plane, without affecting 3D camera position.
// roll does not affect the camera x axis.
// yaw rotates along the world y axis, and it affects the camera x axis.
// pitch rotates along the camera x axis
vMatrix<double> vCameraMatrix(double roll, double pitch, double yaw, 
                                vMatrix<double> canonical_matrix, 
                                vMatrix<double> center)
{
  vMatrix<double> pitch_matrix = vPitchMatrix(canonical_matrix, pitch, center);
  vMatrix<double> yaw_matrix = vAdjustYaw(pitch_matrix, yaw, center);
  vMatrix<double> z_rotation = vRotationMatrix4(roll * pi / 180, 0, 0, 1);
  vMatrix<double> result = yaw_matrix * z_rotation;
//  pitch_matrix.Print("pitch_matrix");
//  yaw_matrix.Print("yaw_matrix");
//  z_rotation.Print("z_rotation");
  return result;
};


vMatrix<float> vCameraMatrixFloat(float roll, float pitch, float yaw, 
                                    vMatrix<double> canonical_matrix, 
                                    vMatrix<double> center)
{
  vMatrix<double> double_result = vCameraMatrix(roll, pitch, yaw, 
                                                  canonical_matrix, center);
  vMatrix<float> result = S_DoubleToFloat(double_result);  
  return result;
};


// result = vPitchMatrix(matrix, pitch_diff, center) assumes that matrix
// is the camera matrix corresponding to roll = 0, pitch = 0, yaw = 0.
// center is the center of the viewing sphere for the camera. It returns the
// matrix that corresponds to roll = 0, pitch = pitch, yaw = 0.
// center is four-dimensional (homogeneous coordinates).
vMatrix<double> vPitchMatrix(vMatrix<double> canonical, double pitch, 
                               vMatrix<double> center)
{
  double pitch_rad = pitch * pi / 180.;

//  vPrint("pitch = %lf\n", pitch);
//  canonical.Print("canonical");
//  center.Print("center");
  vMatrix<double> translation1 = vTranslationMatrix(center).Inv();
//  translation1.Print("translation1");

  vMatrix<double> x_point(4,1);
  x_point.WriteN(4, 1., 0., 0., 1.);
  vMatrix<double> camera_x_point = canonical * x_point;
  vMatrix<double> origin(4, 1);
  origin.WriteN(4, 0., 0., 0., 1.);
  vMatrix<double> camera_origin = canonical * origin;
//  camera_origin.Print("camera_origin");
  vMatrix<double> x_axis = camera_x_point - camera_origin;
  vMatrix<double> x_rotation = vRotationMatrix4(-pitch_rad, x_axis(0), 
                                                 x_axis(1), x_axis(2));

  vMatrix<double> translation2 = vTranslationMatrix(center);
//  translation2.Print("translation2");

  vMatrix<double> world_to_camera = canonical.Inv() * translation2 * x_rotation * translation1;
//  world_to_camera.Print("world_to_camera");
  vMatrix<double> result = world_to_camera.Inv();
  return result;
}


// result = vAdjustYaw(matrix, yaw, center) assumes that matrix
// is the camera matrix corresponding to roll = 0, pitch = x, yaw = 0.
// center is the center of the viewing sphere for the camera. It returns the
// matrix that corresponds to roll = 0, pitch = pitch, yaw = yaw.
// center is four-dimensional (homogeneous coordinates).
vMatrix<double> vAdjustYaw(vMatrix<double> canonical, double yaw, 
                             vMatrix<double> center)
{
  double yaw_rad = yaw * pi / 180.;

  vMatrix<double> translation1 = vTranslationMatrix(center).Inv();
  vMatrix<double> y_rotation = vRotationMatrix4(-yaw_rad, 0., 1., 0);
  vMatrix<double> translation2 = vTranslationMatrix(center);

  vMatrix<double> world_to_camera = canonical.Inv() * translation2 * y_rotation * translation1;
  vMatrix<double> result = world_to_camera.Inv();
  return result;
}


// returns the roll, pitch, yaw parameters, in degrees, that correspond to 
// the 3x3 rotation matrix that is the top left of link_matrix (after
// dividing by the scale of the matrix.
vMatrix<double> vLinkOrientation(vMatrix<double> link_matrix)
{
  vMatrix<double> matrix3(3, 3);
  link_matrix.Copy(&matrix3, 0, 2, 0, 2, 0, 0);
  matrix3 = matrix3 / vProjectionScale(link_matrix);
  vMatrix<double> orientation(1, 3);

  vMatrix<double> z_point(3, 1);
  z_point.WriteN(3, 0., 0., 1.);
  vMatrix<double> viewpoint = matrix3 * z_point;
  double y = viewpoint(1);
  if (y > 1) y = 1;
  else if (y < -1) y = -1;

  double x = viewpoint(0);
  double z = viewpoint(2);

  double pitch_rad = asin(-y);
  double pitch = pitch_rad * 180 / pi;
  double yaw_rad = atan2(x, z);
  if (yaw_rad < 0) yaw_rad = yaw_rad + 2*pi;

  double yaw = yaw_rad * 180 / pi;

  vMatrix<double> eye4 = vMatrix<double>::I(4);
  vMatrix<double> origin(1, 4);
  origin.WriteN(4, 0., 0., 0., 1.);
  vMatrix<double> pitch_yaw = vCameraMatrix(0, pitch, yaw, eye4, origin);
  vMatrix<double> pitch_yaw3(3, 3);
  pitch_yaw.Copy(&pitch_yaw3, 0, 2, 0, 2, 0, 0);
  vMatrix<double> z_rotation = pitch_yaw3.Inv() * matrix3;
  double cos_z = z_rotation(0,0);
  double sin_z = z_rotation(1,0);
  if (sin_z > 1) sin_z = 1;
  else if (sin_z < -1) sin_z = -1;   

  double roll_rad = asin(sin_z);
  if (cos_z < 0) roll_rad = pi - roll_rad;
  if (roll_rad < 0) roll_rad = roll_rad + 2*pi;

  double roll = roll_rad * 180 / pi;
   
  orientation(0) = roll;
  orientation(1) = pitch;
  orientation(2) = yaw;
  return orientation;
}


vMatrix<float> vLinkOrientationFloat(vMatrix<float> link_matrix)
{
  vMatrix<double> double_matrix = S_FloatToDouble(link_matrix);
  vMatrix<double> double_result = vLinkOrientation(double_matrix);
  vMatrix<float> result = S_DoubleToFloat(double_result);
  return result;
}


// Determinant of a 4x4 matrix
double vDeterminant4(vMatrix<double> a)
{
  if ((a.Rows() != 4) || (a.Cols() != 4))
  {
    exit_error("\nvDeterminant4: bad size: %li x %li\n", 
                    a.Rows(), a.Cols());
  }

  double sum1 = a(0,3) * a(1,2) * a(2,1) * a(3,0) - a(0,2) * a(1,3) * a(2,1) * a(3,0);
  double sum2 = -a(0,3) * a(1,1) * a(2,2) * a(3,0);
  double sum3 = a(0,1) * a(1,3) * a(2,2) * a(3,0) + a(0,2) * a(1,1) * a(2,3) * a(3,0);
  double sum4 = -a(0,1) * a(1,2) * a(2,3) * a(3,0) - a(0,3) * a(1,2) * a(2,0) * a(3,1); 
  double sum5 = a(0,2) * a(1,3) * a(2,0) * a(3,1) + a(0,3) * a(1,0) * a(2,2) * a(3,1);
  double sum6 = -a(0,0) * a(1,3) * a(2,2) * a(3,1) - a(0,2) * a(1,0) * a(2,3) * a(3,1);  
  double sum7 = a(0,0) * a(1,2) * a(2,3) * a(3,1) + a(0,3) * a(1,1) * a(2,0) * a(3,2);
  double sum8 = -a(0,1) * a(1,3) * a(2,0) * a(3,2) - a(0,3) * a(1,0) * a(2,1) * a(3,2);  
  double sum9 = a(0,0) * a(1,3) * a(2,1) * a(3,2) + a(0,1) * a(1,0) * a(2,3) * a(3,2);
  double sum10 = -a(0,0) * a(1,1) * a(2,3) * a(3,2) - a(0,2) * a(1,1) * a(2,0) * a(3,3);  
  double sum11 = a(0,1) * a(1,2) * a(2,0) * a(3,3) + a(0,2) * a(1,0) * a(2,1) * a(3,3);
  double sum12 = -a(0,0) * a(1,2) * a(2,1) * a(3,3) - a(0,1) * a(1,0) * a(2,2) * a(3,3); 
  double sum13 = a(0,0) * a(1,1) * a(2,2) * a(3,3);
  double result = sum1 + sum2 + sum3 + sum4 + sum5 + sum6 + sum7 + sum8 + sum9;
  result += sum10 + sum11 + sum12 + sum13;
  return result;
}


// projection is a 4x4 matrix projecting from link coordinates to world
// coordinates.
// returns the scale part of the projection. We assume that the projection 
// consists of rotating, scaling, and translating (in that order).
double vProjectionScale(vMatrix<double> projection)
{
  double result = pow(vDeterminant4(projection), (1.0/3.0));
  return result;
}


// translation is given in four dimensions (homogeneous coordinates),
// and represents a 3D displacement. We return the 4x4 matrix
// that applies that translation.
vMatrix<double> vTranslationMatrix(vMatrix<double> translation)
{
  vMatrix<double> result = vMatrix<double>::I(4);
  double factor = translation(3);
  result.WriteCol(3, translation(0) / factor, translation(1) / factor,
                  translation(2) / factor, translation(3) / factor);
  return result;
}


class vMatrixDouble : public vMatrix<double>
{
public:
  vMatrixDouble(long in_rows, long in_cols, ...) :
  vMatrix<double>(in_rows, in_cols)
  {
    va_list arguments;
    va_start(arguments, in_cols);
    long i;
    for (i = 0; i < size; i++)
    {
      matrix[i] = va_arg(arguments, double);
    }
    va_end(arguments);
    
  }
  ~vMatrixDouble() {}
};


long vTestProjectionFunctions()
{
  vMatrixDouble li3w1(4,4,
    5.49999904633, 0.00281232153066, 0.0, -0.001630088198, 
    -0.00281232153066, 5.49999904633, 0.0, 0.000969063083176, 
    0.0, 0.0, 5.5, 0.0, 
    0.0, 0.0, 0.0, 1.0);

  vMatrixDouble li3o1(1,3, 0.0, 0.0, -40.0780982971);

  vMatrixDouble lt1w1(4,4,
    5.49999952316, 0.00263843941502, -0.000914786243811, -0.0015238764463, 
    -0.00263809645548, 5.49999952316, 0.00236902828328, 0.000834598031361,
    0.000915809941944, -0.00236857193522, 5.49999952316, 0.00107802182902,
    0.0, 0.0, 0.0, 1.0);

  vMatrixDouble lt1o1(1,3, 5.68780994415, -24.0263004303, -26.4734992981);

  vMatrix<double> lt1pm = lt1w1 * vOrientationMatrix(lt1o1);

  vMatrix<double> li3pm = li3w1 * vOrientationMatrix(li3o1);

  vMatrixDouble views(3,2,
    0.000000, 0.000000,
    90.000000, 112.500000,
    112.500000, 264.000000);
  

  vMatrix<double> params = vLinkCameraParams(lt1w1, lt1o1, views);
  params.Print("params");

  return 1;
}


//long vHandModel::Remesh(double max_distance)
//{
//  return 1;
//}


/////////////////////////////////////////////////////////////////////
//                                                                 //
//          End of implementation of vHandModel                   //
//                                                                 //
/////////////////////////////////////////////////////////////////////


vHandModel::vHandModel()
{
  is_valid = 0;
  weld_vertices = 0;
  polygons = 0;
  number_of_links = NumberOfLinks();
  link_names = vnew(char *, number_of_links);
  link_names[0] = vCopyString("Background-Body");
  link_names[1] = vCopyString("Left Hand");
  link_names[2] = vCopyString("Left Thumb 1");
  link_names[3] = vCopyString("Left Thumb 2");
  link_names[4] = vCopyString("Left Thumb 3");
  link_names[5] = vCopyString("Left Index 1");
  link_names[6] = vCopyString("Left Index 2");
  link_names[7] = vCopyString("Left Index 3");
  link_names[8] = vCopyString("Left Mid 1");
  link_names[9] = vCopyString("Left Mid 2");
  link_names[10] = vCopyString("Left Mid 3");
  link_names[11] = vCopyString("Left Ring 1");
  link_names[12] = vCopyString("Left Ring 2");
  link_names[13] = vCopyString("Left Ring 3");
  link_names[14] = vCopyString("Left Pinky 1");
  link_names[15] = vCopyString("Left Pinky 2");
  link_names[16] = vCopyString("Left Pinky 3");
  link_names[17] = vCopyString("Left Forearm");

  // Load local vertex information (each vertex is described
  // with coordinates in the local link frame).
  char * directory = ModelDirectory();

  char * local_vertices_bin = vJoinPaths(directory, "local_vertices.bin");
  vertices = v3dMatrix<double>::Read(local_vertices_bin);
  if (vertices == 0)
  {
    char * local_vertices_text = vJoinPaths(directory, "local_vertices.txt");
    vertices = v3dMatrix<double>::ReadText(local_vertices_text);
    if (vertices == 0) exit_error("\n%s not found\n", local_vertices_text);
    vertices->Write(local_vertices_bin);
    vdelete2(local_vertices_text);
  }
  vdelete2(local_vertices_bin);

  // Load polygon information.
  char * polygon_file = PolygonFile();
  if (!vFileExists(polygon_file))
  {
    vdelete2(polygon_file);
    vdelete2(directory);
    return;
  }

  polygons = v3dMatrix<long>::Read(polygon_file);
  vdelete2(polygon_file);
  if (polygons == 0)
  {
    exit_error("Invalid polygon file\n");
  }

  vdelete2(directory);
  weld_vertices = 0;
  is_valid = 1;
}


vHandModel::~vHandModel()
{
  long i;
  for (i = 0; i < number_of_links; i++)
  {
    vdelete2(link_names[i]);
  }
  vdelete2(link_names);
  vdelete(vertices);
  vdelete(weld_vertices);
  vdelete(polygons);
  is_valid = 0;
}


// PreprocessPolygons takes in a matrix describing all the polygons
// in the hand model. The matrix has, at each row, information
// describing a vertex of a polygon, in the following format:
// (link_id, polygon_id, number_of_vertices, x, y, z)
// where (x,y,z) are the coordinates of the vertex in the local
// coordinate system of the link. Then, a new matrix is created,
// where each of the vertices is mapped to an index into the vertices
// matrix, so that we get rows of the form 
// (link_id, polygon_id, number_of_vertices, index). This way, we
// can easily look up the corresponding vertices in a particular
// hand shape (vHandShapeSlow object), where those vertices are represented
// in world coordinates.
long vHandModel::PreprocessPolygons()
{
  if (vertices == 0) return 0;
  char * raw_file = RawPolygonFile();
  v3dMatrix<double> * raw_polygons = 
    v3dMatrix<double>::ReadText(raw_file);
  if ((raw_polygons == 0) || (raw_polygons->Cols() != 6))
  {
    vPrint("\nFailed to read raw polygons file %s\n", raw_file);
    vdelete(raw_file);
    return 0;
  }
  vdelete2(raw_file);
  
  long number_of_polygons = raw_polygons->Rows();
  long number_of_vertices = vertices->Rows();
  vdelete(polygons);
  polygons = new v3dMatrix<long>(number_of_polygons, 4, 1);

  vArray2(double) v = vertices->Matrix2(0);
  vArray2(double) rawp = raw_polygons->Matrix2(0);
  vArray2(long) p = polygons->Matrix2(0);
  long success = 1;

  long i;
  vPrint("\n");
  for (i = 0; i < number_of_polygons; i++)
  {
    p[i][0] = vRound(rawp[i][0]);
    p[i][1] = vRound(rawp[i][1]);
    p[i][2] = vRound(rawp[i][2]);

    // Find the vertex index.
    double x = rawp[i][3];
    double y = rawp[i][4];
    double z = rawp[i][5];

    long j;
    for (j = 0; j < number_of_vertices; j++)
    {
      if ((x != v[j][1]) || (y != v[j][2]) || (z != v[j][3]))
      {
        continue;
      }
      if (p[i][0] != vRound(v[j][0]))
      {
        continue;
      }

      p[i][3] = j;
      break;
    }

    if (j == number_of_vertices)
    {
      vPrint("\nFailed to find a match for vertex %li of polygons\n", i);
      success = 0;
    }
    vPrint("processed vertex %li\r", i);
  }
  vPrint("\n");
 
  char * polygon_file = PolygonFile();
  long success2 = polygons->Write(polygon_file);
  if (success2 <= 0)
  {
    vPrint("Failed to save polygons to %s\n", polygon_file);
    success = 0;
  }
  vdelete2(polygon_file);
  vdelete(raw_polygons);
  vPrint("preprocessed polygons successfully\n");
  return 1;
}


// File where the polygon information is stored.
char * vHandModel::PolygonFile()
{
  char * directory = ModelDirectory();
  char * result = vJoinPaths(directory, "polygons.bin");
  vdelete2(directory);
  return result;
}


// File where the raw polygon information is stored (from which
// the regular polygon matrix is built, using PreprocessPolygons);
char * vHandModel::RawPolygonFile()
{
  char * directory = ModelDirectory();
  char * result = vJoinPaths(directory, "raw_polygons.txt");
  vdelete2(directory);
  return result;
}
  

// Returns a default directory where information about the hand
// model is stored.
char * vHandModel::ModelDirectory()
{
  // Load local vertex information (each vertex is described
  // with coordinates in the local link frame).
  char * directory = vJoinPaths3(sequences_directory, 
                                  "databases", "hand_model");
  return directory;
}


// This function is used to obtain some statistics that would be useful
// for remeshing.
long vHandModel::SideLengthStats()
{
  if (is_valid <= 0) return 0;

  vArray2(long) polygons2 = polygons->Matrix2(0);
  vArray2(double) vertices2 = vertices->Matrix2(0);
  vPrint("rows = %li, cols = %li\n", vertices->Rows(), vertices->Cols());
  long number_of_corners = polygons->Rows();
  long index = 0;
  float corner_x[4];
  float corner_y[4];
  float corner_z[4];

  vector<double> stats;
  
  while(index < number_of_corners)
  {
    long corners = polygons2[index][2];
    long i;
    for (i = 0; i < corners; i++)
    {
      long v = polygons2[index+i][3];
      corner_x[i] = (float) vertices2[v][1];
      corner_y[i] = (float) vertices2[v][2];
      corner_z[i] = (float) vertices2[v][3];
    }

    for (i = 0; i < corners; i++)
    {
      long j = (i + 1) % corners;
      double dx = corner_x[i] - corner_x[j];
      double dy = corner_y[i] - corner_y[j];
      double dz = corner_z[i] - corner_z[j];
      double length = sqrt(dx * dx + dy * dy + dz * dz);

      stats.push_back(length);
    }

    if (corners == 4)
    {
      for (i = 0; i < 2; i++)
      {
        long j = (i + 2) % corners;
        double dx = corner_x[i] - corner_x[j];
        double dy = corner_y[i] - corner_y[j];
        double dz = corner_z[i] - corner_z[j];
        double length = sqrt(dx * dx + dy * dy + dz * dz);

        stats.push_back(length);
      }
    }

    index = index + corners;
  }

  std::sort(stats.begin(), stats.end(), less<double>());
  long i;
  for (i = 0; i < (long) stats.size(); i += 100)
  {
    vPrint("%5li: %lf\n", i, stats[i]);
  }
  for (i = i-100; i < (long) stats.size(); i++)
  {
    vPrint("%5li: %lf\n", i, stats[i]);
  }

  return 1;
}


/////////////////////////////////////////////////////////////////////
//                                                                 //
//          End of implementation of vHandModel                   //
//                                                                 //
/////////////////////////////////////////////////////////////////////

  
vHandShapeSlow::vHandShapeSlow()
{
  Zero();
}


vHandShapeSlow::vHandShapeSlow(const char * in_name)
{
  Zero();
  name = vCopyString(in_name);
  color_map = vHandDatabase1::DefaultColorMap();

  if (LoadVertices() <= 0) return;
  scaled_vertices = new v3dMatrix<double>(world_vertices);
  if (LoadMatrices() <= 0) return;
  model = new vHandModel();
  if (model->IsValid() <= 0) return;
  SetParents();
  SetNumberOfVertices();
  FindSharedVertices();
  SetDisplacements();
  is_valid = 1;
}


vHandShapeSlow::~vHandShapeSlow()
{
  vdelete(model);
  vdelete(world_vertices);
  vdelete2(world_matrices);
  vdelete(scaled_vertices); 
  vdelete(color_map);
  vdelete(link_axes);
  vdelete2(name);
  vdelete2(orientation_matrices);
  vdelete2(orientations);
  vdelete2(scale_matrices);
  Zero();
}


long vHandShapeSlow::Zero()
{
  is_valid = 0;
  name = 0;
  model = 0;
  world_vertices = 0;
  world_matrices = vZero(vMatrix<double>);
  orientation_matrices = vZero(vMatrix<double>);
  orientations = vZero(vMatrix<double>);
  scale_matrices = vZero(vMatrix<double>);
  scaled_vertices = 0;
  link_axes = 0;
  color_map = 0;
  return 1;
}


char * vHandShapeSlow::ShapeDirectory()
{
  char * directory = vJoinPaths3(sequences_directory, 
                                  "databases", "hand_model");
  return directory;
}


long vHandShapeSlow::LoadVertices()
{
  char * world_vertices_bin = VertexFile();
  world_vertices = v3dMatrix<double>::Read(world_vertices_bin);
  if (world_vertices == 0)
  {
    char * world_vertices_text = TextVertexFile();
    world_vertices = v3dMatrix<double>::ReadText(world_vertices_text);
    if (world_vertices == 0) 
    {
      vdelete2(world_vertices_bin);
      vdelete2(world_vertices_text);
      return 0;
    }
    world_vertices->Write(world_vertices_bin);
    vdelete2(world_vertices_text);
  }
  vdelete2(world_vertices_bin);
  return 1;
}


// An auxiliary function for LoadMatrices, it reads one text
// matrix.
long vHandShapeSlow::ReadTextMatrix(FILE * fp, vMatrix<double> * matrix,
                                 char letter1, char letter2)
{
  long success = 1;
  char buffer[100];
  char char1, char2;
  char1 = char2 = 0;
  long items;
  if (letter2 == '0')
  {
    items = fscanf(fp, "%cwm_%s = [", &char1, buffer);
    if (items != 2) 
    {
      success = 0;
    }
  }

  else 
  {
    items = fscanf(fp, "%c%cwm_%s = [", &char1, &char2, buffer);
    if (items != 3) 
    {
      success = 0;
    }
  }

  if (char1 != letter1) 
  {
    success = 0;
  }

  if ((letter2 != '0') && (char2 != letter2))
  {
    success = 0;
  }

  if (strcmp(buffer, name) != 0) 
  {
    success = 0;
  }

  long read_success = matrix->ReadTextBandwise(fp);
  if (read_success <= 0) 
  {
    success = 0;
  }

  items = fscanf(fp, "%c%c%c%c%c%c%c", buffer, buffer+1, buffer+2, buffer+3,
                 buffer+4, buffer+5, buffer+6);
  if (items != 7) 
  {
    success = 0;
  }
  if (memcmp(buffer, "]';\r\n\r\n", 7) != 0)
  {
    success = 0;
  }


  return success;
}


// Read the world matrices for each link. What I consider world matrix
// is the product of two matrices: what poser considers as world matrix, 
// and the matrix corresponding to the orientation of the link.
// Matrices are exported from poser using a Python script, in text format.
// The first time they are read, they are converted to binary format.
long vHandShapeSlow::LoadMatrices()
{
  const long number_of_links = NumberOfLinks();
  char * bin_file = MatrixFile();
  long success = 1;
  long i;

  if (!vFileExists(bin_file))
  {
    // Initialize the world matrices (in poser's sense of the term).
    vector<vMatrix<double> > world1_matrices(number_of_links);
    for (i = 0; i < number_of_links; i++)
    {
      world1_matrices[i] = vMatrix<double>(4, 4);
    }

    // Read the text format.
    char * text_file = TextMatrixFile();
    FILE * fp = fopen(text_file, vFOPEN_READ);
    vdelete2(text_file);
    if (fp == 0)
    {
      vdelete2(bin_file);
      return 0;
    }

    // Read the matrices one by one.
    long read_success;
    long index;

    index = 0;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'b', '0');
    if (read_success <= 0) success = 0;
    index = 1;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'f', '0');
    if (read_success <= 0) success = 0;
    index = 17;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'h', '0');
    if (read_success <= 0) success = 0;

    index = 2;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 't', '1');
    if (read_success <= 0) success = 0;
    index = 5;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'i', '1');
    if (read_success <= 0) success = 0;
    index = 8;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'm', '1');
    if (read_success <= 0) success = 0;
    index = 11;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'r', '1');
    if (read_success <= 0) success = 0;
    index = 14;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'p', '1');
    if (read_success <= 0) success = 0;

    index = 3;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 't', '2');
    if (read_success <= 0) success = 0;
    index = 4;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 't', '3');
    if (read_success <= 0) success = 0;

    index = 6;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'i', '2');
    if (read_success <= 0) success = 0;
    index = 7;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'i', '3');
    if (read_success <= 0) success = 0;

    index = 9;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'm', '2');
    if (read_success <= 0) success = 0;
    index = 10;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'm', '3');
    if (read_success <= 0) success = 0;

    index = 12;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'r', '2');
    if (read_success <= 0) success = 0;
    index = 13;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'r', '3');
    if (read_success <= 0) success = 0;

    index = 15;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'p', '2');
    if (read_success <= 0) success = 0;
    index = 16;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'p', '3');
    if (read_success <= 0) success = 0;


    // Read orientations.
    vMatrix<double> orientations_matrix(number_of_links, 3);
    vArray2(double) orientation_array = orientations_matrix.Matrix2();
    long items;
    char buffer[100];

    index = 0;
    items = fscanf(fp, "bo_%s = [%lf %lf %lf]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 1;
    items = fscanf(fp, "fo_%s = [%lf %lf %lf]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 17;
    items = fscanf(fp, "ho_%s = [%lf %lf %lf]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 2;
    items = fscanf(fp, "t1o_%s = [%lf %lf %lf]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 5;
    items = fscanf(fp, "i1o_%s = [%lf %lf %lf]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 8;
    items = fscanf(fp, "m1o_%s = [%lf %lf %lf]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 11;
    items = fscanf(fp, "r1o_%s = [%lf %lf %lf]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 14;
    items = fscanf(fp, "p1o_%s = [%lf %lf %lf]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 3;
    items = fscanf(fp, "t2o_%s = [%lf %lf %lf]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 4;
    items = fscanf(fp, "t3o_%s = [%lf %lf %lf]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 6;
    items = fscanf(fp, "i2o_%s = [%lf %lf %lf]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 7;
    items = fscanf(fp, "i3o_%s = [%lf %lf %lf]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 9;
    items = fscanf(fp, "m2o_%s = [%lf %lf %lf]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 10;
    items = fscanf(fp, "m3o_%s = [%lf %lf %lf]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 12;
    items = fscanf(fp, "r2o_%s = [%lf %lf %lf]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 13;
    items = fscanf(fp, "r3o_%s = [%lf %lf %lf]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 15;
    items = fscanf(fp, "p2o_%s = [%lf %lf %lf]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 16;
    items = fscanf(fp, "p3o_%s = [%lf %lf %lf]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    if (success <= 0)
    {
      vPrint("Failed to read text file\n");
    }
    fclose(fp);
    
    world_matrices = vnew(vMatrix<double>, number_of_links);
    orientation_matrices = vnew(vMatrix<double>, number_of_links);
    orientations = vnew(vMatrix<double>, number_of_links);
    long i;

    // compute the actual world matrices, and save them.
    if (success > 0)
    {
      fp = fopen(bin_file, vFOPEN_WRITE);
      if (fp == 0) 
      {
        success = 0;
      }

      for (i = 0; i < number_of_links; i++)
      {
  //      vPrint("\ni = %li\n", i);
        orientations[i] = vMatrix<double>(1, 3);
        orientations[i].WriteN(3, orientation_array[i][0], orientation_array[i][1],
                                 orientation_array[i][2]);
        orientation_matrices[i] = vOrientationMatrix(orientations[i]);
        world_matrices[i] = world1_matrices[i].Trans() * orientation_matrices[i];
  //      world1_matrices[i].Print("world1 matrix");
  //      temp_orientations.Print("orientation");
  //      world_matrices[i].Print("world matrix");
  //      vMatrix<double> link_orientation = vLinkOrientation(world_matrices[i]);
  //      link_orientation.Print("link orientation");
        long temp_success;
        temp_success = world_matrices[i].WriteBandwise(fp);
        if (temp_success <= 0) success = 0;
        temp_success = orientation_matrices[i].WriteBandwise(fp);
        if (temp_success <= 0) success = 0;
        temp_success = orientations[i].WriteBandwise(fp);
        if (temp_success <= 0) success = 0;
      }
      fclose(fp);
    }
  }

  else
  {
    FILE * fp = fopen(bin_file, vFOPEN_READ);
    if (fp == 0) 
    {
      success = 0;
    }

    world_matrices = vnew(vMatrix<double>, number_of_links);
    orientation_matrices = vnew(vMatrix<double>, number_of_links);
    orientations = vnew(vMatrix<double>, number_of_links);
    long i;
    for (i = 0; i < number_of_links; i++)
    {
      world_matrices[i] = vMatrix<double>(4, 4);
      orientation_matrices[i] = vMatrix<double>(4, 4);
      orientations[i] = vMatrix<double>(1, 3);

      long temp_success;
      temp_success = world_matrices[i].ReadBandwise(fp);
      if (temp_success <= 0) success = 0;
      temp_success = orientation_matrices[i].ReadBandwise(fp);
      if (temp_success <= 0) success = 0;
      temp_success = orientations[i].ReadBandwise(fp);
      if (temp_success <= 0) success = 0;

      vMatrix<double> link_orientation = vLinkOrientation(world_matrices[i]);
//      link_orientation.Print("link orientation b");
    }
  }

  if (success > 0)
  {
    link_axes = new v3dMatrix<double>(4, 3, number_of_links);
    vArray3(double) axes = link_axes->Matrix3();
    vMatrix<double> x_vector(4, 1), y_vector(4, 1), z_vector(4, 1), origin(4, 1);
    x_vector.WriteN(4, (double) 1, (double) 0, (double) 0, (double) 1);
    y_vector.WriteN(4, (double) 0, (double) 1, (double) 0, (double) 1);
    z_vector.WriteN(4, (double) 0, (double) 0, (double) 1, (double) 1);
    origin.WriteN(4, (double) 0, (double) 0, (double) 0, (double) 1);

    for(i = 0; i < number_of_links; i++)
    {
      vMatrix<double> & world_matrix = world_matrices[i];

      vMatrix<double> link_origin = world_matrix * origin;
      vMatrix<double> link_x_vector = world_matrix * x_vector - link_origin;
      vMatrix<double> link_y_vector = world_matrix * y_vector - link_origin;
      vMatrix<double> link_z_vector = world_matrix * z_vector - link_origin;
      long j;
      double x_norm, y_norm, z_norm;
      x_norm = y_norm = z_norm = 0;
      for (j = 0; j < 3; j++)
      {
        axes[i][3][j] = link_origin(j);
        double value;

        value = link_x_vector(j);
        axes[i][0][j] = value;
        x_norm += value * value;;

        value = link_y_vector(j);
        axes[i][1][j] = value;
        y_norm += value * value;;

        value = link_z_vector(j);
        axes[i][2][j] = value;
        z_norm += value * value;;
      }
      x_norm = sqrt(x_norm);
      y_norm = sqrt(y_norm);
      z_norm = sqrt(z_norm);
      for (j = 0; j < 3; j++)
      {
        if (x_norm != 0) axes[i][0][j] /= x_norm;
        if (y_norm != 0) axes[i][1][j] /= y_norm;
        if (z_norm != 0) axes[i][2][j] /= z_norm;
      }
    }
  //  link_axes->Print2("link_axes");
  }

  vdelete2(bin_file);

  scale_matrices = vnew(vMatrix<double>, number_of_links);
  for (i = 0; i < number_of_links; i++)
  {
    scale_matrices[i] = vMatrix<double>::I(4);
  }

  return success;
}


char * vHandShapeSlow::BinFile(const char * prefix)
{
  return InfoFile(prefix, ".bin");
}


char * vHandShapeSlow::TextFile(const char * prefix)
{
  return InfoFile(prefix, ".txt");
}


char * vHandShapeSlow::InfoFile(const char * prefix, const char * suffix)
{
  char * simple_name = vMergeStrings4(prefix, "_", name, suffix);
  char * directory = ShapeDirectory();
  char * result = vJoinPaths(directory, simple_name);
  vdelete2(directory);
  vdelete2(simple_name);
  return result;
}

  
char * vHandShapeSlow::VertexFile()
{
  return BinFile("v");
}
  

char * vHandShapeSlow::TextVertexFile()
{
  return TextFile("v");
}


char * vHandShapeSlow::MatrixFile()
{
  return BinFile("m");
}


char * vHandShapeSlow::TextMatrixFile()
{
  return TextFile("m");
}


long vHandShapeSlow::SetDisplacements()
{
  long number_of_vertices = world_vertices->Rows();
  char * name1 = vMergeStrings2(name, "_111010");
  char * name2 = vMergeStrings2(name, "_101110");
  char * name3 = vMergeStrings2(name, "_101011");

  vHandShapeSlow shape_x(name1);
  vHandShapeSlow shape_y(name2);
  vHandShapeSlow shape_z(name3);

  vdelete2(name1);
  vdelete2(name2);
  vdelete2(name3);

  if ((shape_x.IsValid() <= 0) ||
      (shape_y.IsValid() <= 0) ||
      (shape_z.IsValid() <= 0))
  {
    return 0;
  }

  displacements = vMatrix<double>(number_of_vertices, 9);

  vArray2(double) normal = world_vertices->Matrix2(0);
  vArray2(double) x_vertices = shape_x.world_vertices->Matrix2(0);
  vArray2(double) y_vertices = shape_y.world_vertices->Matrix2(0);
  vArray2(double) z_vertices = shape_z.world_vertices->Matrix2(0);
  
  long i;
  for (i = 0; i < number_of_vertices; i++)
  {
    double x = normal[i][1];
    double y = normal[i][2];
    double z = normal[i][3];

    double xx = x_vertices[i][1];
    double xy = x_vertices[i][2];
    double xz = x_vertices[i][3];

    double yx = y_vertices[i][1];
    double yy = y_vertices[i][2];
    double yz = y_vertices[i][3];

    double zx = z_vertices[i][1];
    double zy = z_vertices[i][2];
    double zz = z_vertices[i][3];

    displacements(i, 0) = (xx - x) * 10.0;
    displacements(i, 1) = (xy - y) * 10.0;
    displacements(i, 2) = (xz - z) * 10.0;

    displacements(i, 3) = (yx - x) * 10.0;
    displacements(i, 4) = (yy - y) * 10.0;
    displacements(i, 5) = (yz - z) * 10.0;
    
    displacements(i, 6) = (zx - x) * 10.0;
    displacements(i, 7) = (zy - y) * 10.0;
    displacements(i, 8) = (zz - z) * 10.0;
  }
  
  return 1;
}


// Stores information into parents.
long vHandShapeSlow::SetParents()
{
  long number = NumberOfLinks();
  parents = vMatrix<long>(1, number);
  parents(0) = -1; // body
  parents(17) = 0; // forearm
  parents(1) = 17; // hand

  // inner links
  parents(2) = parents(5) = parents(8) = parents(11) = parents(14) = 1;

  // thumb
  parents(3) = 2;
  parents(4) = 3;

  // index
  parents(6) = 5;
  parents(7) = 6;

  // middle
  parents(9) = 8;
  parents(10) = 9;

  // ring
  parents(12) = 11;
  parents(13) = 12;

  // pinky
  parents(15) = 14;
  parents(16) = 15;

  return 1;
}


// Stores information into num_vertices and starting_index.
long vHandShapeSlow::SetNumberOfVertices()
{
  long number_of_links = NumberOfLinks();
  long number_of_vertices = world_vertices->Rows();
  num_vertices = vMatrix<long>(1, number_of_links);
  starting_index = vMatrix<long>(1, number_of_links);
  vArray2(double) vertices = world_vertices->Matrix2(0);

  long index = 0;
  while (index < number_of_vertices)
  {
    long id = vRound(vertices[index][0]);
    starting_index(id) = index;
    while (index < number_of_vertices)
    {
      if (vRound(vertices[index][0]) != id)
      {
        break;
      }
      index++;
    }
    num_vertices(id) = index - starting_index(id);
  }

  num_vertices(0) = 0;
  starting_index(0) = 0;
  return 1;
}


// Stores information into shared_vertices
long vHandShapeSlow::FindSharedVertices()
{
  long number_of_vertices = world_vertices->Rows();
  vArray2(double) vertices = world_vertices->Matrix2(0);
  vector<long> link_ids;
  vector<long> indices;
  vector<long> parent_indices;

  long i;
  for (i = 0; i < number_of_vertices; i++)
  {
    long link_id = vRound(vertices[i][0]);
    double x1 = vertices[i][1];
    double y1 = vertices[i][2];
    double z1 = vertices[i][3];

    double min_distance = 10000000;
    long min_index = -1;
    long parent = parents(link_id);
    long start = starting_index(parent);
    long number = num_vertices(parent);
    long limit = start + number;
    long j;
    for (j = start; j < limit; j++)
    {
      double x = vertices[j][1] - x1;
      double y = vertices[j][2] - y1;
      double z = vertices[j][3] - z1;

      double distance = x*x + y*y + z*z;
      if (distance < min_distance) 
      {
        min_distance = distance;
        min_index = j;
      }
    }
    if (sqrt(min_distance) < 0.000001)
    {
      link_ids.push_back(link_id);
      indices.push_back(i);
      parent_indices.push_back(min_index);
    }
  }

  long number = link_ids.size();
  shared_vertices = vMatrix<long>(number, 3);
  for (i = 0; i < number; i++)
  {
    shared_vertices(i, 0) = link_ids[i];
    shared_vertices(i, 1) = indices[i];
    shared_vertices(i, 2) = parent_indices[i];
  }

  return 1;
}


  
// Appropriate camera matrix. In the current implementation, the camera
// is placed at the origin. Is that OK or does it lead to problems?
// 
// Question. If the camera points at a given direction, are all camera
// locations equivalent? Do they give the same image up to translation?
// 
// Suppose that the camera direction is canonical. Then, for each x,y
// component, we use the nearest z. Camera location doesn't matter.
// 
// Now, how about doing a rotation. A rotation determines both camera
// orientation and camera location. The camera orientation, again, would
// be the determining factor. So, having the camera at the origin should
// be OK, and is probably the best in terms of optimizing performance.
// 
// The result is a rotation of the scaled vertices, according to the camera
// parameters. Using the result, the image coordinates of each vertex
// can be determined from the x,y values.
vMatrix<double> vHandShapeSlow::CameraPoints(double roll, double pitch, double yaw)
{
  if (is_valid <= 0) return 0;

  // Prepare the points matrix.
  long number_of_points = scaled_vertices->Rows();
  vArray2(double) vertices = scaled_vertices->Matrix2(0);
  vMatrix<double> world_points(4, number_of_points);
  long i;
  for (i = 0; i < number_of_points; i++)
  {
    world_points(0, i) = vertices[i][1];
    world_points(1, i) = vertices[i][2];
    world_points(2, i) = vertices[i][3];
    world_points(3, i) = 1.0;
  }

  // Weld: if a link shares a vertex with its parent, we move its vertex
  // to the same position as the corresponding vertex of the parent.
  long number_of_shared = shared_vertices.Rows();
  for (i = 0; i < number_of_shared; i++)
  {
    long index1 = shared_vertices(i, 1);
    long index2 = shared_vertices(i, 2);
    world_points(0, index1) = world_points(0, index2);
    world_points(1, index1) = world_points(1, index2);
    world_points(2, index1) = world_points(2, index2);
    world_points(3, index1) = world_points(3, index2);
  }

  // Get the camera matrix.
  vMatrix<double> center(1, 4);
  center.WriteN(4, 0., 0., 0., 1.);
  vMatrix<double> cm = vCameraMatrix(roll, pitch, yaw, 
                                       vMatrix<double>::I(4), center);
  vMatrix<double> inv_cm = cm.Inv();


  // Project the points
  vMatrix<double> camera_points = inv_cm * world_points;
  return camera_points;
}


vMatrix<long> vHandShapeSlow::WorldToPixel(double roll, double pitch, double yaw, 
                                         long size, long bbox_size)
{
  vMatrix<double> camera_points = CameraPoints(roll, pitch,  yaw);
  vMatrix<long> pixels = CameraToPixel(camera_points, size, bbox_size);
  return pixels;
}


static vMatrix<long> S_CameraToPixel(vMatrix<double> camera_points, long size,
                                      long bbox_size, double min_x, double max_x,
                                      double min_y, double max_y)
{
  double world_width = max_x - min_x;
  double world_height = max_y - min_y;
  double world_length = Max(world_width, world_height);
  double scale = ((double) bbox_size) / world_length;
  double center_x = (max_x + min_x) / 2.0;
  double center_y = (max_y + min_y) / 2.0;
  double translation = ((double) size) / 2.0;

  long number_of_points = camera_points.Cols();
  vMatrix<long> pixels_matrix(number_of_points, 2);
  vArray2(long) pixels = pixels_matrix.Matrix2();

  // Debug line:
  long i;
  for (i = 0; i < number_of_points; i++)
  {
    pixels[i][0] = size - vRound((camera_points(1, i) - center_y) * scale + translation);
    pixels[i][1] = vRound((camera_points(0, i) - center_x) * scale + translation);
  }  
  return pixels_matrix;
}


vMatrix<long> vHandShapeSlow::CameraToPixel(vMatrix<double> camera_points, 
                                          long size, long bbox_size)
{
  long number_of_points = camera_points.Cols();

  // Find bounding box in 3D
  double min_x, min_y, max_x, max_y;
  min_x = max_x = camera_points(0, 0);
  min_y = max_y = camera_points(1, 0);

  long i;
  for (i = 0; i < number_of_points; i++)
  {
    if (camera_points(0, i) < min_x) min_x = camera_points(0, i);
    if (camera_points(0, i) > max_x) max_x = camera_points(0, i);
    if (camera_points(1, i) < min_y) min_y = camera_points(1, i);
    if (camera_points(1, i) > max_y) max_y = camera_points(1, i);
  }

  vMatrix<long> pixels_matrix = S_CameraToPixel(camera_points, size, bbox_size,
                                                 min_x, max_x, min_y, max_y);
  return pixels_matrix;
}


color_image * vHandShapeSlow::ColorLinks(grayscale_image * link_ids)
{
  long rows = link_ids->Rows();
  long cols = link_ids->Cols();
  long size = rows * cols;
  vArray(uchar) ids = link_ids->Matrix();

  color_image * result = new color_image(rows, cols);
  vArray(uchar) R = result->GetR();
  vArray(uchar) G = result->GetG();
  vArray(uchar) B = result->GetB();

  long r, g, b;
  r = g = b = 0;

  long i;
  for (i = 0; i < size; i++)
  {
    long link_id = ids[i];
    color_map->Color(link_id, &r, &g, &b);
    R[i] = (uchar) r;
    G[i] = (uchar) g;
    B[i] = (uchar) b;
  }

  return result;
}

  
color_image * vHandShapeSlow::DrawPoints2(long size, long bbox_size)
{
  return DrawPoints5(size, bbox_size, 0, 0, 0);
}


color_image * vHandShapeSlow::DrawPoints5(long size, long bbox_size,
                                   double roll, double pitch, double yaw)
{
  if (is_valid <= 0) return 0;

  vMatrix<long> pixel_matrix = WorldToPixel(roll, pitch, yaw, size, bbox_size);
  vArray2(long) pixels = pixel_matrix.Matrix2();

  grayscale_image link_ids(size, size);
  vWriteValue(&link_ids, (uchar) 0);
  vArray2(uchar) ids = link_ids.Matrix2(0);
  vArray2(double) vertices = scaled_vertices->Matrix2(0);
  long number_of_points = pixel_matrix.Rows();

  long i;
  for (i = 0; i < number_of_points; i++)
  {
    long id = vRound(vertices[i][0]);
    long row = pixels[i][0];
    long col = pixels[i][1];
    ids[row][col] = (uchar) id;
  }
  color_image * result = ColorLinks(&link_ids);

  /*
  // This is a temporary hack, for debugging. Plot some more points on the hand.

  // Find bounding box in 3D
  vMatrix<double> camera_points = CameraPoints(roll, pitch, yaw);
  vPrint("\nEnter link id:\n");
  long link_id = 0;
  vScan("%li", &link_id);

  long junk = 0;
  double min_x = color_imageRowMin3(&camera_points, 0, &junk);
  double max_x = color_imageRowMax3(&camera_points, 0, &junk);
  double min_y = color_imageRowMin3(&camera_points, 1, &junk);
  double max_y = color_imageRowMax3(&camera_points, 1, &junk);
  
  vPrint("min_x = %lf, max_x = %lf\n", min_x, max_x);
  vPrint("min_y = %lf, max_y = %lf\n", min_y, max_y);

  // Get the camera matrix.
  vMatrix<double> center(1, 4);
  center.WriteN(4, 0., 0., 0., 1.);
  vMatrix<double> cm = vCameraMatrix(roll, pitch, yaw, 
                                       vMatrix<double>::I(4), center);
  vMatrix<double> inv_cm = cm.Inv();
  cm.Print("cm");
  inv_cm.Print("inv_cm");

  // Prepare the points (origin and x unit vector).
  
  vMatrix<double> extra_points(4, 4);
  extra_points.WriteN(16, (double) 0, (double) 0, (double) 0, (double) 1,
                      (double) 1, (double) 0, (double) 0, (double) 1,
                      (double) 0, (double) 1, (double) 0, (double) 1,
                      (double) 0, (double) 0, (double) 1, (double) 1);
  vMatrix<double> w = world_matrices[link_id % 18] * extra_points.Trans();
  w.Trans().Print("world points transpose");
  vMatrix<double> c = inv_cm * w;
  c.Trans().Print("camera points transpose");

  vMatrix<long> pixels_matrix = S_CameraToPixel(c, size, bbox_size,
                                  min_x, max_x, min_y, max_y);
  pixels_matrix.Print("pixels");
  world_matrices[link_id % 18].Print("link_matrix");

  for (i = 1; i < pixels_matrix.Rows(); i++)
  {
    double row_diff = pixels_matrix(i, 0) - pixels_matrix(0, 0);
    double col_diff = pixels_matrix(i, 1) - pixels_matrix(0, 1);
    double angle = vAngle2(col_diff, row_diff);
    vPrint("row_diff = %lf, col_diff = %lf\n", row_diff, col_diff);

    vector<uchar> colors(3);
    colors[0] = colors[1] = colors[2] = 0;
    colors[i-1] = 255;
    vDrawLineA(result, size/2, size/2, angle, size/2, &colors);
  }
*/

  return result;
}


color_image * vHandShapeSlow::DrawPolygons5(long size, long bbox_size,
                                     double roll, double pitch, double yaw)
{
  if (is_valid <= 0) return 0;

  vMatrix<long> pixel_matrix = WorldToPixel(roll, pitch, yaw, size, bbox_size);
  vArray2(long) pixels = pixel_matrix.Matrix2();

  // Now, draw polygon lines.
  grayscale_image link_ids(size, size);
  vWriteValue(&link_ids, (uchar) 0);

  v3dMatrix<long> polygons_matrix = model->polygons;
  vArray2(long) polygons = polygons_matrix.Matrix2(0);
  long number_of_corners = polygons_matrix.Rows();
  long index = 0;
  while(index < number_of_corners)
  {
    long link_id = polygons[index][0];
    long polygon_id = polygons[index][1];
    long corners = polygons[index][2];
    long first_vertex_index = polygons[index][3];
    long end_index = index + corners - 1;
    long j;
    for (j = index; j < end_index; j++)
    {
      long vertex_index1 = polygons[j][3];
      long vertex_index2 = polygons[j+1][3];
      long row1 = pixels[vertex_index1][0];
      long col1 = pixels[vertex_index1][1];
      long row2 = pixels[vertex_index2][0];
      long col2 = pixels[vertex_index2][1];
      vDrawLine1(&link_ids, row1, col1, row2, col2, (uchar) link_id);      
    }

    long vertex_index1 = polygons[j][3];
    long vertex_index2 = first_vertex_index;
    long row1 = pixels[vertex_index1][0];
    long col1 = pixels[vertex_index1][1];
    long row2 = pixels[vertex_index2][0];
    long col2 = pixels[vertex_index2][1];
    vDrawLine1(&link_ids, row1, col1, row2, col2, (uchar) link_id);      
    index = index + corners;
  }

  color_image * result = ColorLinks(&link_ids);
  return result;
}


color_image * vHandShapeSlow::DrawLinks5(long size, long bbox_size,
                                  double roll, double pitch, double yaw)
{
  if (is_valid <= 0) return 0;
  vMatrix<double> camera_points = CameraPoints(roll, pitch, yaw);
  vMatrix<long> pixels_matrix = CameraToPixel(camera_points, size, bbox_size);
  vArray2(long) pixels = pixels_matrix.Matrix2();
  
  // Now, draw the links
  grayscale_image link_id_matrix(size, size);
  vWriteValue(&link_id_matrix, (uchar) 0);
  v3dMatrix<float> depth_matrix(size, size, 1);
  vWriteValue(&depth_matrix, (float) -1000000);
  vArray2(uchar) link_ids = link_id_matrix.Matrix2(0);
  vArray2(float) depths = depth_matrix.Matrix2(0);

  v3dMatrix<long> polygons_matrix = model->polygons;
  vArray2(long) polygons = polygons_matrix.Matrix2(0);
  long number_of_corners = polygons_matrix.Rows();
  long index = 0;
  class_pointer(long) corner_rows = vnew(long, 4);
  class_pointer(long) corner_cols = vnew(long, 4);
//  vector<long> stats;
  while(index < number_of_corners)
  {
    long link_id = polygons[index][0];
    long polygon_id = polygons[index][1];
    long corners = polygons[index][2];
    double z_sum = 0;
    long i;
    for (i = 0; i < corners; i++)
    {
      long v = polygons[index+i][3];
      corner_rows[i] = pixels[v][0];
      corner_cols[i] = pixels[v][1];
      z_sum += camera_points(2, v);
    }

    double z = z_sum / (double) corners;
    if (corners <= 3)
    {
      for (i = corners; i < 4; i++)
      {
        corner_rows[i] = corner_rows[i-1];
        corner_cols[i] = corner_cols[i-1];
      }
    }
    vector<vPoint> quad_pixels;
    vQuadPixels(corner_rows, corner_cols, &quad_pixels);
    long number = quad_pixels.size();
//    stats.push_back(number);
    for (i = 0; i < number; i++)
    {
      vPoint & point = quad_pixels[i];
      long row = point.row;
      long col = point.col;
      if (z > depths[row][col])
      {
        depths[row][col] = (float) z;
        link_ids[row][col] = (uchar) link_id;
      }
    }

    index = index + corners;
  }

  color_image * result = ColorLinks(&link_id_matrix);

//  std::sort(stats.begin(), stats.end(), less<double>());
//  long i;
//  for (i = 0; i < stats.size(); i += 100)
//  {
//    vPrint("%5li: %li\n", i, stats[i]);
//  }

//  for (i = i-100; i < stats.size(); i++)
//  {
//    vPrint("%5li: %li\n", i, stats[i]);
//  }

  delete_pointer(corner_rows);
  delete_pointer(corner_cols);
  return result;
}


long vHandShapeSlow::ScaleX(double scale)
{
  vArray2(double) vertices = world_vertices->Matrix2(0);
  vArray2(double) scaled = scaled_vertices->Matrix2(0);
  vArray3(double) axes = link_axes->Matrix3();
  long number_of_vertices = world_vertices->Rows();

  long i;
  for (i = 0; i < number_of_vertices; i++)
  {
    // Get the link id and world coordinates.
    long link_id = vRound(vertices[i][0]);
    double x = vertices[i][1];
    double y = vertices[i][2];
    double z = vertices[i][3];
  
    // Get the link origin and x_axis
    vArray(double) o = axes[link_id][3];
    vArray(double) a = axes[link_id][2];

    // subtract the origin
    x -= o[0];
    y -= o[1];
    z -= o[2];

    // Get projection to x axis.
    double dot_product = x * a[0] + y * a[1] * z * a[2];

    // subtract projection to x axis
    x -= dot_product * a[0];
    y -= dot_product * a[1];
    z -= dot_product * a[2];

    double factor = dot_product * scale;
    // add scaled projection to x axis.
    x += factor * a[0];
    y += factor * a[1];
    z += factor * a[2];

    // add the origin
    x += o[0];
    y += o[1];
    z += o[2];

    // store into scaled_vertices;
    scaled[i][1] = x;
    scaled[i][2] = y;
    scaled[i][3] = z;
  }

  return 1;
}


long vHandShapeSlow::Trash()
{
  long number_of_vertices = world_vertices->Rows();
  char * name1 = vMergeStrings2(name, "_111010");
  char * name2 = vMergeStrings2(name, "_101110");
  char * name3 = vMergeStrings2(name, "_101011");

  vHandShapeSlow shape_x(name1);
  vHandShapeSlow shape_y(name2);
  vHandShapeSlow shape_z(name3);

  vdelete2(name1);
  vdelete2(name2);
  vdelete2(name3);

  if ((shape_x.IsValid() <= 0) ||
      (shape_y.IsValid() <= 0) ||
      (shape_z.IsValid() <= 0))
  {
    vPrint("Failed to read shapes\n");
    return 1;
  }

  double x, y, z;
  x = y = z = 0;
  vPrint("\nEnter x, y, z\n");
  vScan("%lf %lf %lf", &x, &y, &z);

  double x_factor = x - 1.0;
  double y_factor = y - 1.0;
  double z_factor = z - 1.0;
  vMatrix<double> displacements(number_of_vertices, 9);

  vArray2(double) normal = world_vertices->Matrix2(0);
  vArray2(double) scaled = scaled_vertices->Matrix2(0);
  vArray2(double) x_vertices = shape_x.world_vertices->Matrix2(0);
  vArray2(double) y_vertices = shape_y.world_vertices->Matrix2(0);
  vArray2(double) z_vertices = shape_z.world_vertices->Matrix2(0);
  
  long i;
  for (i = 0; i < number_of_vertices; i++)
  {
    double x = normal[i][1];
    double y = normal[i][2];
    double z = normal[i][3];

    double xx = x_vertices[i][1];
    double xy = x_vertices[i][2];
    double xz = x_vertices[i][3];

    double yx = y_vertices[i][1];
    double yy = y_vertices[i][2];
    double yz = y_vertices[i][3];

    double zx = z_vertices[i][1];
    double zy = z_vertices[i][2];
    double zz = z_vertices[i][3];

    displacements(i, 0) = (xx - x) * 10.0;
    displacements(i, 1) = (xy - y) * 10.0;
    displacements(i, 2) = (xz - z) * 10.0;

    displacements(i, 3) = (yx - x) * 10.0;
    displacements(i, 4) = (yy - y) * 10.0;
    displacements(i, 5) = (yz - z) * 10.0;
    
    displacements(i, 6) = (zx - x) * 10.0;
    displacements(i, 7) = (zy - y) * 10.0;
    displacements(i, 8) = (zz - z) * 10.0;

    scaled[i][1] = x + displacements(i, 0) * x_factor + displacements(i, 3) * y_factor + displacements(i, 6) * z_factor;
    scaled[i][2] = y + displacements(i, 1) * x_factor + displacements(i, 4) * y_factor + displacements(i, 7) * z_factor;
    scaled[i][3] = z + displacements(i, 2) * x_factor + displacements(i, 5) * y_factor + displacements(i, 8) * z_factor;
  }

  return 1;
}


// This is an auxiliary function that is used to check whether
// the arguments to AltOrientationMatrix and related functions
// are valid. It is useful so that each of those functions does
// not have to repeat the cumbersom checks. If the arguments
// are valid, we return 1, else we return a non-positive
// error code.
static long S_CheckOrientationArgs(long link_id, char first, char second,
                                   char third, long number_of_links)
{
  // Error checking
  if ((link_id < 0) || (link_id >= number_of_links))
  {
    return 0;
  }
  if ((first != '0') || (second != '0') || (third != '0'))
  {
    if ((first == second) || (first == third) || (second == third))
    {
      return -1;
    }

    if ((first != 'x') && (second != 'x') && (third != 'x'))
    {
      return -2;
    }
    
    if ((first != 'y') && (second != 'y') && (third != 'y'))
    {
      return -3;
    }
   
    if ((first != 'z') && (second != 'z') && (third != 'z'))
    {
      return -4;
    }
  }

  return 1;
}


// This function is used to experiment with different orderings
// in producing the orientation matrix. Each argument should be
// 'x', 'y', or 'z', or all arguments should be '0' (in which case
// the identity is returned).
vMatrix<double> 
vHandShapeSlow::AltOrientationMatrix(long link_id, char first, 
                                  char second, char third)
{
  long number = NumberOfLinks();
  long check = S_CheckOrientationArgs(link_id, first, second, third, number);
  if (check <= 0)
  {
    exit_error("\nError %li in AltOrientationMatrix(%li, %c, %c, %c)\n",
                    check, link_id, first, second, third);
  }

  if ((first == '0') && (second == '0') && (third == '0'))
  {
    return vMatrix<double>::I(4);
  }

  double factor = pi / 180.0;
  vMatrix<double> orientation = orientations[link_id];
  double x = orientation(0) * factor;
  double y = orientation(1) * factor;
  double z = orientation(2) * factor;
  vMatrix<double> x_matrix4 = vRotationMatrix4(x, 1, 0, 0);
  vMatrix<double> y_matrix4 = vRotationMatrix4(y, 0, 1, 0);
  vMatrix<double> z_matrix4 = vRotationMatrix4(z, 0, 0, 1);

  vMatrix<double> first_matrix, second_matrix, third_matrix;

  if (first == 'x') first_matrix = x_matrix4;
  else if (first == 'y') first_matrix = y_matrix4;
  else if (first == 'z') first_matrix = z_matrix4;

  if (second == 'x') second_matrix = x_matrix4;
  else if (second == 'y') second_matrix = y_matrix4;
  else if (second == 'z') second_matrix = z_matrix4;

  if (third == 'x') third_matrix = x_matrix4;
  else if (third == 'y') third_matrix = y_matrix4;
  else if (third == 'z') third_matrix = z_matrix4;

  vMatrix<double> result = first_matrix * second_matrix * third_matrix;
  return result;
}


// This function is used to experiment with different orderings
// in producing the orientation matrix. The arguments are passed
// to AltOrientationMatrix to get orientation matrices for each link.
// Then, the world matrices are adjusted accordingly.
long vHandShapeSlow::AltWorldMatrices(char first, char second, char third)
{
  if (is_valid == 0) return 0;
  long number = NumberOfLinks();
  long check = S_CheckOrientationArgs(0, first, second, third, number);
  if (check <= 0) return 0;

  long i;
  for (i = 0; i < number; i++)
  {
    vMatrix<double> old_world = world_matrices[i];
    vMatrix<double> old_orientation = orientation_matrices[i];
    vMatrix<double> orientation = orientations[i];
    vMatrix<double> poser_world = old_world * old_orientation.Inv();
    orientation_matrices[i] = AltOrientationMatrix(i, first, second, third);
    world_matrices[i] = poser_world * orientation_matrices[i];
  }

  return 1;
}


// These functions will test an alternative way of rendering that
// applies world matrices to local vertices (as opposed to reading
// world vertices from the disk).
vMatrix<double> 
vHandShapeSlow::AltCameraPoints(double roll, double pitch, double yaw)
{
  if (is_valid <= 0) return 0;

  // Prepare the points matrix.
  long number_of_points = scaled_vertices->Rows();
  vArray2(double) vertices = model->vertices->Matrix2(0);
  vMatrix<double> local_point(4, 1);
  vMatrix<double> world_points(4, number_of_points);
  long i;
  for (i = 0; i < number_of_points; i++)
  {
    long link_id = vRound(vertices[i][0]);
    local_point(0) = vertices[i][1];
    local_point(1) = vertices[i][2];
    local_point(2) = vertices[i][3];
    local_point(3) = 1.0;
    vMatrix<double> world_point = 
      world_matrices[link_id] * orientation_matrices[link_id] * scale_matrices[link_id] * orientation_matrices[link_id].Inv() * local_point;
    world_points(0, i) = world_point(0);
    world_points(1, i) = world_point(1);
    world_points(2, i) = world_point(2);
    world_points(3, i) = world_point(3);
  }

  // Weld: if a link shares a vertex with its parent, we move its vertex
  // to the same position as the corresponding vertex of the parent.
  long number_of_shared = shared_vertices.Rows();
  for (i = 0; i < number_of_shared; i++)
  {
    long index1 = shared_vertices(i, 1);
    long index2 = shared_vertices(i, 2);
    world_points(0, index1) = world_points(0, index2);
    world_points(1, index1) = world_points(1, index2);
    world_points(2, index1) = world_points(2, index2);
    world_points(3, index1) = world_points(3, index2);
  }

  // Get the camera matrix.
  vMatrix<double> center(1, 4);
  center.WriteN(4, 0., 0., 0., 1.);
  vMatrix<double> cm = vCameraMatrix(roll, pitch, yaw, 
                                       vMatrix<double>::I(4), center);
  vMatrix<double> inv_cm = cm.Inv();


  // Project the points
  vMatrix<double> camera_points = inv_cm * world_points;
  return camera_points;
}


vMatrix<long> 
vHandShapeSlow::AltWorldToPixel(double roll, double pitch, double yaw, 
                             long size, long bbox_size)
{
  vMatrix<double> camera_points = AltCameraPoints(roll, pitch,  yaw);
  vMatrix<long> pixels = CameraToPixel(camera_points, size, bbox_size);
  return pixels;
}



color_image * 
vHandShapeSlow::AltDrawPoints5(long size, long bbox_size,
                            double roll, double pitch, double yaw)
{
  if (is_valid <= 0) return 0;

  vMatrix<long> pixel_matrix = AltWorldToPixel(roll, pitch, yaw, size, bbox_size);
  vArray2(long) pixels = pixel_matrix.Matrix2();

  grayscale_image link_ids(size, size);
  vWriteValue(&link_ids, (uchar) 0);
  vArray2(uchar) ids = link_ids.Matrix2(0);
  vArray2(double) vertices = scaled_vertices->Matrix2(0);
  long number_of_points = pixel_matrix.Rows();

  long i;
  for (i = 0; i < number_of_points; i++)
  {
    long id = vRound(vertices[i][0]);
    long row = pixels[i][0];
    long col = pixels[i][1];
    ids[row][col] = (uchar) id;
  }

  color_image * result = ColorLinks(&link_ids);
  return result;
}


color_image * 
vHandShapeSlow::AltDrawPolygons5(long size, long bbox_size,
                              double roll, double pitch, double yaw)
{
  if (is_valid <= 0) return 0;

  vMatrix<long> pixel_matrix = AltWorldToPixel(roll, pitch, yaw, size, bbox_size);
  vArray2(long) pixels = pixel_matrix.Matrix2();

  // Now, draw polygon lines.
  grayscale_image link_ids(size, size);
  vWriteValue(&link_ids, (uchar) 0);

  v3dMatrix<long> polygons_matrix = model->polygons;
  vArray2(long) polygons = polygons_matrix.Matrix2(0);
  long number_of_corners = polygons_matrix.Rows();
  long index = 0;
  while(index < number_of_corners)
  {
    long link_id = polygons[index][0];
    long polygon_id = polygons[index][1];
    long corners = polygons[index][2];
    long first_vertex_index = polygons[index][3];
    long end_index = index + corners - 1;
    long j;
    for (j = index; j < end_index; j++)
    {
      long vertex_index1 = polygons[j][3];
      long vertex_index2 = polygons[j+1][3];
      long row1 = pixels[vertex_index1][0];
      long col1 = pixels[vertex_index1][1];
      long row2 = pixels[vertex_index2][0];
      long col2 = pixels[vertex_index2][1];
      vDrawLine1(&link_ids, row1, col1, row2, col2, (uchar) link_id);      
    }

    long vertex_index1 = polygons[j][3];
    long vertex_index2 = first_vertex_index;
    long row1 = pixels[vertex_index1][0];
    long col1 = pixels[vertex_index1][1];
    long row2 = pixels[vertex_index2][0];
    long col2 = pixels[vertex_index2][1];
    vDrawLine1(&link_ids, row1, col1, row2, col2, (uchar) link_id);      
    index = index + corners;
  }

  color_image * result = ColorLinks(&link_ids);
  return result;
}


color_image * 
vHandShapeSlow::AltDrawLinks5(long size, long bbox_size,
                           double roll, double pitch, double yaw)
{
  if (is_valid <= 0) return 0;
  vMatrix<double> camera_points = AltCameraPoints(roll, pitch, yaw);
  vMatrix<long> pixels_matrix = CameraToPixel(camera_points, size, bbox_size);
  vArray2(long) pixels = pixels_matrix.Matrix2();
  
  // Now, draw the links
  grayscale_image link_id_matrix(size, size);
  vWriteValue(&link_id_matrix, (uchar) 0);
  v3dMatrix<float> depth_matrix(size, size, 1);
  vWriteValue(&depth_matrix, (float) -1000000);
  vArray2(uchar) link_ids = link_id_matrix.Matrix2(0);
  vArray2(float) depths = depth_matrix.Matrix2(0);

  v3dMatrix<long> polygons_matrix = model->polygons;
  vArray2(long) polygons = polygons_matrix.Matrix2(0);
  long number_of_corners = polygons_matrix.Rows();
  long index = 0;
  class_pointer(long) corner_rows = vnew (long, 4);
  class_pointer(long) corner_cols = vnew (long, 4);
  while(index < number_of_corners)
  {
    long link_id = polygons[index][0];
    long polygon_id = polygons[index][1];
    long corners = polygons[index][2];
    double z_sum = 0;

    long first_v = polygons[index][3];
    double x1 = pixels[first_v][1];
    double y1 = pixels[first_v][0];
    double z1 = camera_points(2, first_v);
    double dz_dx, dz_dy;
    long dz_dx_found = 0;
    long dz_dy_found = 0;

    long i;
    for (i = 0; i < corners; i++)
    {
      long v = polygons[index+i][3];
      corner_rows[i] = pixels[v][0];
      corner_cols[i] = pixels[v][1];
      z_sum += camera_points(2, v);

      if (dz_dx_found == 0) 
      {
        double x = corner_cols[i];
        if (x != x1)
        {
          dz_dx = (camera_points(2, v) - z1) / (x - x1);
          dz_dx_found = 1;
        }
      }

      if (dz_dy_found == 0) 
      {
        double y = corner_rows[i];
        if (y != y1)
        {
          dz_dy = (camera_points(2, v) - z1) / (y - y1);
          dz_dy_found = 1;
        }
      }
    }

    long found = dz_dy_found * dz_dx_found;
    double mean_z = z_sum / (double) corners;
    if (corners <= 3)
    {
      for (i = corners; i < 4; i++)
      {
        corner_rows[i] = corner_rows[i-1];
        corner_cols[i] = corner_cols[i-1];
      }
    }
    vector<vPoint> quad_pixels;
    vQuadPixels(corner_rows, corner_cols, &quad_pixels);
    long number = quad_pixels.size();
    for (i = 0; i < number; i++)
    {
      vPoint & point = quad_pixels[i];
      long row = point.row;
      long col = point.col;
      double z;
//      if (found > 0)
//      {
//        double dx = ((double) col) - x1;
//        double dy = ((double) row) - y1;
//        z = z1 + dx * dz_dx + dy * dz_dy;
//      }
//      else
      {
        z = mean_z;
      }

      if (z > depths[row][col])
      {
        depths[row][col] = (float) z;
        link_ids[row][col] = (uchar) link_id;
      }
    }

    index = index + corners;
  }

  color_image * result = ColorLinks(&link_id_matrix);

  delete_pointer(corner_rows);
  delete_pointer (corner_cols);
  return result;
}


long vHandShapeSlow::SetScale(double x, double y, double z)
{
  if (is_valid == 0) return 0;
  long number_of_vertices = world_vertices->Rows();
  if (displacements.Rows() != number_of_vertices) return 0;
  
  vArray2(double) d = displacements.Matrix2();
  vArray2(double) normal = world_vertices->Matrix2(0);
  vArray2(double) scaled = scaled_vertices->Matrix2(0);
  double x_factor = x - 1.0;
  double y_factor = y - 1.0;
  double z_factor = z - 1.0;
  
  long i;
  for (i = 0; i < number_of_vertices; i++)
  {
    double x = normal[i][1];
    double y = normal[i][2];
    double z = normal[i][3];

    scaled[i][1] = x + d[i][0] * x_factor + d[i][3] * y_factor + d[i][6] * z_factor;
    scaled[i][2] = y + d[i][1] * x_factor + d[i][4] * y_factor + d[i][7] * z_factor;
    scaled[i][3] = z + d[i][2] * x_factor + d[i][5] * y_factor + d[i][8] * z_factor;
  }

  return 1;
}


// This function is used for debugging. For every link it prints out
// its world matrix, orientation matrix, and 
// world matrix * inv(orientation matrix), which is what Poser 
// considers as world matrix.
long vHandShapeSlow::PrintMatrices()
{
  vPrint("\n");
  if (is_valid <= 0) 
  {
    vPrint("Invalid shape\n");
    return 0;
  }

  long number = NumberOfLinks();
  long i;
  for (i = 0; i < number; i++)
  {
    PrintLinkMatrices(i);
  }

  return 1;
}


long vHandShapeSlow::PrintLinkMatrices(long link_id)
{
  if ((is_valid <= 0) || (link_id < 0) || (link_id >= NumberOfLinks()))
  {
    return 0;
  }

  vPrint("%2li, %s:\n", link_id, model->link_names[link_id]);
  vMatrix<double> world_matrix = world_matrices[link_id];
  vMatrix<double> orientation_matrix = orientation_matrices[link_id];
  vMatrix<double> orientation = orientations[link_id];
  vMatrix<double> poser_world = world_matrix * orientation_matrix.Inv();

  orientation.Print("orientation");
  orientation_matrix.Print("orientation_matrix");
  poser_world.Print("poser_world");
  world_matrix.Print("world_matrix");
      
  return 1;
}




/////////////////////////////////////////////////////////////////////
//                                                                 //
//      Beginning of implementation of vHandShape                 //
//                                                                 //
/////////////////////////////////////////////////////////////////////


vHandShape::vHandShape()
{
  Zero();
}


vHandShape::vHandShape(const char * in_name)
{
  Zero();
  name = vCopyString(in_name);
  color_map = vHandDatabase1::DefaultColorMap();

  if (LoadVertices() <= 0) return;
  scaled_vertices = vMatrix<float>(&world_vertices);
  if (LoadMatrices() <= 0) return;
  model = new vHandModel();
  if (model->IsValid() <= 0) return;
  SetParents();
  SetNumberOfVertices();
  FindSharedVertices();
  SetDisplacements();
  is_valid = 1;
}


vHandShape::~vHandShape()
{
  vdelete(model);
  vdelete2(world_matrices);
  vdelete(color_map);
  vdelete(link_axes);
  vdelete2(name);
  vdelete2(orientation_matrices);
  vdelete2(orientations);
  Zero();
}


long vHandShape::Zero()
{
  is_valid = 0;
  name = 0;
  model = 0;
  diameter = -1;
  max_scale = 1.5;
  min_scale = 0.6;
  world_matrices = vZero(vMatrix<float>);
  orientation_matrices = vZero(vMatrix<float>);
  orientations = vZero(vMatrix<float>);
  link_axes = 0;
  color_map = 0;
  return 1;
}


char * vHandShape::ShapeDirectory()
{
  char * directory = vJoinPaths3(sequences_directory, 
                                  "databases", "hand_model");
  return directory;
}


long vHandShape::LoadVertices()
{
  char * world_vertices_bin = VertexFile();

  // temp_world_vertices has a vertex per row, in
  // (link_id, x, y, z) format.
  v3dMatrix<float> * temp_world_vertices = 
    v3dMatrix<float>::Read(world_vertices_bin);
  if (temp_world_vertices == 0)
  {
    char * world_vertices_text = TextVertexFile();
    temp_world_vertices = v3dMatrix<float>::ReadText(world_vertices_text);
    if (temp_world_vertices == 0) 
    {
      vdelete2(world_vertices_bin);
      vdelete2(world_vertices_text);
      return 0;
    }
    temp_world_vertices->Write(world_vertices_bin);
    vdelete2(world_vertices_text);
  }
  vdelete2(world_vertices_bin);

  // Initialize vertex_ids, world_vertices, and scaled_vertices.
  long number = temp_world_vertices->Rows();
  world_vertices = vMatrix<float>(3, number);
  vertex_ids = vMatrix<uchar>(1, number);

  vArray2(float) temp = temp_world_vertices->Matrix2(0);
  vArray2(float) world = world_vertices.Matrix2();
  vArray(uchar) ids = vertex_ids.Matrix();

  long i;
  for (i = 0; i < number; i++)
  {
    long id = vRound(temp[i][0]);
    ids[i] = (uchar) id;
    world[0][i] = temp[i][1];
    world[1][i] = temp[i][2];
    world[2][i] = temp[i][3];
  }
  vdelete(temp_world_vertices);

  return 1;
}


// An auxiliary function for LoadMatrices, it reads one text
// matrix.
long vHandShape::ReadTextMatrix(FILE * fp, vMatrix<float> * matrix,
                                 char letter1, char letter2)
{
  long success = 1;
  char buffer[100];
  char char1, char2;
  char1 = char2 = 0;
  long items;
  if (letter2 == '0')
  {
    items = fscanf(fp, "%cwm_%s = [", &char1, buffer);
    if (items != 2) 
    {
      success = 0;
    }
  }

  else 
  {
    items = fscanf(fp, "%c%cwm_%s = [", &char1, &char2, buffer);
    if (items != 3) 
    {
      success = 0;
    }
  }

  if (char1 != letter1) 
  {
    success = 0;
  }

  if ((letter2 != '0') && (char2 != letter2))
  {
    success = 0;
  }

  if (strcmp(buffer, name) != 0) 
  {
    success = 0;
  }

  long read_success = matrix->ReadTextBandwise(fp);
  if (read_success <= 0) 
  {
    success = 0;
  }

  items = fscanf(fp, "%c%c%c%c%c%c%c", buffer, buffer+1, buffer+2, buffer+3,
                 buffer+4, buffer+5, buffer+6);
  if (items != 7) 
  {
    success = 0;
  }
  if (memcmp(buffer, "]';\r\n\r\n", 7) != 0)
  {
    success = 0;
  }


  return success;
}


// Read the world matrices for each link. What I consider world matrix
// is the product of two matrices: what poser considers as world matrix, 
// and the matrix corresponding to the orientation of the link.
// Matrices are exported from poser using a Python script, in text format.
// The first time they are read, they are converted to binary format.
long vHandShape::LoadMatrices()
{
  const long number_of_links = NumberOfLinks();
  char * bin_file = MatrixFile();
  long success = 1;
  long i;

  if (!vFileExists(bin_file))
  {
    // Initialize the world matrices (in poser's sense of the term).
    vector<vMatrix<float> > world1_matrices(number_of_links);
    for (i = 0; i < number_of_links; i++)
    {
      world1_matrices[i] = vMatrix<float>(4, 4);
    }

    // Read the text format.
    char * text_file = TextMatrixFile();
    FILE * fp = fopen(text_file, vFOPEN_READ);
    vdelete2(text_file);
    if (fp == 0)
    {
      vdelete2(bin_file);
      return 0;
    }

    // Read the matrices one by one.
    long read_success;
    long index;

    index = 0;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'b', '0');
    if (read_success <= 0) success = 0;
    index = 1;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'f', '0');
    if (read_success <= 0) success = 0;
    index = 17;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'h', '0');
    if (read_success <= 0) success = 0;

    index = 2;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 't', '1');
    if (read_success <= 0) success = 0;
    index = 5;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'i', '1');
    if (read_success <= 0) success = 0;
    index = 8;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'm', '1');
    if (read_success <= 0) success = 0;
    index = 11;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'r', '1');
    if (read_success <= 0) success = 0;
    index = 14;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'p', '1');
    if (read_success <= 0) success = 0;

    index = 3;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 't', '2');
    if (read_success <= 0) success = 0;
    index = 4;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 't', '3');
    if (read_success <= 0) success = 0;

    index = 6;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'i', '2');
    if (read_success <= 0) success = 0;
    index = 7;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'i', '3');
    if (read_success <= 0) success = 0;

    index = 9;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'm', '2');
    if (read_success <= 0) success = 0;
    index = 10;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'm', '3');
    if (read_success <= 0) success = 0;

    index = 12;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'r', '2');
    if (read_success <= 0) success = 0;
    index = 13;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'r', '3');
    if (read_success <= 0) success = 0;

    index = 15;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'p', '2');
    if (read_success <= 0) success = 0;
    index = 16;
    read_success = ReadTextMatrix(fp, &(world1_matrices[index]), 'p', '3');
    if (read_success <= 0) success = 0;


    // Read orientations.
    vMatrix<float> orientations_matrix(number_of_links, 3);
    vArray2(float) orientation_array = orientations_matrix.Matrix2();
    long items;
    char buffer[100];

    index = 0;
    items = fscanf(fp, "bo_%s = [%f %f %f]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 1;
    items = fscanf(fp, "fo_%s = [%f %f %f]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 17;
    items = fscanf(fp, "ho_%s = [%f %f %f]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 2;
    items = fscanf(fp, "t1o_%s = [%f %f %f]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 5;
    items = fscanf(fp, "i1o_%s = [%f %f %f]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 8;
    items = fscanf(fp, "m1o_%s = [%f %f %f]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 11;
    items = fscanf(fp, "r1o_%s = [%f %f %f]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 14;
    items = fscanf(fp, "p1o_%s = [%f %f %f]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 3;
    items = fscanf(fp, "t2o_%s = [%f %f %f]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 4;
    items = fscanf(fp, "t3o_%s = [%f %f %f]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 6;
    items = fscanf(fp, "i2o_%s = [%f %f %f]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 7;
    items = fscanf(fp, "i3o_%s = [%f %f %f]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 9;
    items = fscanf(fp, "m2o_%s = [%f %f %f]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 10;
    items = fscanf(fp, "m3o_%s = [%f %f %f]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 12;
    items = fscanf(fp, "r2o_%s = [%f %f %f]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 13;
    items = fscanf(fp, "r3o_%s = [%f %f %f]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 15;
    items = fscanf(fp, "p2o_%s = [%f %f %f]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    index = 16;
    items = fscanf(fp, "p3o_%s = [%f %f %f]';\n\n", buffer, 
                   &(orientation_array[index][0]), &(orientation_array[index][1]),
                   &(orientation_array[index][2]));
    if (items != 4) success = 0;
    if (strcmp(buffer, name) != 0) success = 0;

    if (success <= 0)
    {
      vPrint("Failed to read text file\n");
    }
    fclose(fp);
    
    world_matrices = vnew(vMatrix<float>, number_of_links);
    orientation_matrices = vnew(vMatrix<float>, number_of_links);
    orientations = vnew(vMatrix<float>, number_of_links);
    long i;

    // compute the actual world matrices, and save them.
    if (success > 0)
    {
      fp = fopen(bin_file, vFOPEN_WRITE);
      if (fp == 0) 
      {
        success = 0;
      }

      for (i = 0; i < number_of_links; i++)
      {
        orientations[i] = vMatrix<float>(1, 3);
        orientations[i].WriteN(3, orientation_array[i][0], orientation_array[i][1],
                                 orientation_array[i][2]);
        orientation_matrices[i] = vOrientationMatrixFloat(orientations[i]);
        world_matrices[i] = world1_matrices[i].Trans() * orientation_matrices[i];
        long temp_success;
        temp_success = world_matrices[i].WriteBandwise(fp);
        if (temp_success <= 0) success = 0;
        temp_success = orientation_matrices[i].WriteBandwise(fp);
        if (temp_success <= 0) success = 0;
        temp_success = orientations[i].WriteBandwise(fp);
        if (temp_success <= 0) success = 0;
      }
      fclose(fp);
    }
  }

  else
  {
    FILE * fp = fopen(bin_file, vFOPEN_READ);
    if (fp == 0) 
    {
      success = 0;
    }

    world_matrices = vnew(vMatrix<float>, number_of_links);
    orientation_matrices = vnew(vMatrix<float>, number_of_links);
    orientations = vnew(vMatrix<float>, number_of_links);
    long i;
    for (i = 0; i < number_of_links; i++)
    {
      world_matrices[i] = vMatrix<float>(4, 4);
      orientation_matrices[i] = vMatrix<float>(4, 4);
      orientations[i] = vMatrix<float>(1, 3);

      long temp_success;
      temp_success = world_matrices[i].ReadBandwise(fp);
      if (temp_success <= 0) success = 0;
      temp_success = orientation_matrices[i].ReadBandwise(fp);
      if (temp_success <= 0) success = 0;
      temp_success = orientations[i].ReadBandwise(fp);
      if (temp_success <= 0) success = 0;

      vMatrix<float> link_orientation = vLinkOrientationFloat(world_matrices[i]);
//      link_orientation.Print("link orientation b");
    }
  }

  if (success > 0)
  {
    link_axes = new v3dMatrix<float>(4, 3, number_of_links);
    vArray3(float) axes = link_axes->Matrix3();
    vMatrix<float> x_vector(4, 1), y_vector(4, 1), z_vector(4, 1), origin(4, 1);
    x_vector.WriteN(4, (float) 1, (float) 0, (float) 0, (float) 1);
    y_vector.WriteN(4, (float) 0, (float) 1, (float) 0, (float) 1);
    z_vector.WriteN(4, (float) 0, (float) 0, (float) 1, (float) 1);
    origin.WriteN(4, (float) 0, (float) 0, (float) 0, (float) 1);

    for(i = 0; i < number_of_links; i++)
    {
      vMatrix<float> & world_matrix = world_matrices[i];

      vMatrix<float> link_origin = world_matrix * origin;
      vMatrix<float> link_x_vector = world_matrix * x_vector - link_origin;
      vMatrix<float> link_y_vector = world_matrix * y_vector - link_origin;
      vMatrix<float> link_z_vector = world_matrix * z_vector - link_origin;
      long j;
      float x_norm, y_norm, z_norm;
      x_norm = y_norm = z_norm = 0;
      for (j = 0; j < 3; j++)
      {
        axes[i][3][j] = link_origin(j);
        float value;

        value = link_x_vector(j);
        axes[i][0][j] = value;
        x_norm += value * value;;

        value = link_y_vector(j);
        axes[i][1][j] = value;
        y_norm += value * value;;

        value = link_z_vector(j);
        axes[i][2][j] = value;
        z_norm += value * value;;
      }
      x_norm = sqrt(x_norm);
      y_norm = sqrt(y_norm);
      z_norm = sqrt(z_norm);
      for (j = 0; j < 3; j++)
      {
        if (x_norm != 0) axes[i][0][j] /= x_norm;
        if (y_norm != 0) axes[i][1][j] /= y_norm;
        if (z_norm != 0) axes[i][2][j] /= z_norm;
      }
    }
  //  link_axes->Print2("link_axes");
  }

  vdelete2(bin_file);

  return success;
}


char * vHandShape::BinFile(const char * prefix)
{
  return InfoFile(prefix, "_f.bin");
}


char * vHandShape::TextFile(const char * prefix)
{
  return InfoFile(prefix, ".txt");
}


char * vHandShape::InfoFile(const char * prefix, const char * suffix)
{
  char * simple_name = vMergeStrings4(prefix, "_", name, suffix);
  char * directory = ShapeDirectory();
  char * result = vJoinPaths(directory, simple_name);
  vdelete2(directory);
  vdelete2(simple_name);
  return result;
}

  
char * vHandShape::VertexFile()
{
  return BinFile("v");
}
  

char * vHandShape::TextVertexFile()
{
  return TextFile("v");
}


char * vHandShape::MatrixFile()
{
  return BinFile("m");
}


char * vHandShape::TextMatrixFile()
{
  return TextFile("m");
}


long vHandShape::SetDisplacements()
{
  long number_of_vertices = world_vertices.Cols();
  char * name1 = vMergeStrings2(name, "_111010");
  char * name2 = vMergeStrings2(name, "_101110");
  char * name3 = vMergeStrings2(name, "_101011");

  vHandShape shape_x(name1);
  vHandShape shape_y(name2);
  vHandShape shape_z(name3);

  vdelete2(name1);
  vdelete2(name2);
  vdelete2(name3);

  if ((shape_x.IsValid() <= 0) ||
      (shape_y.IsValid() <= 0) ||
      (shape_z.IsValid() <= 0))
  {
    return 0;
  }

  displacements = vMatrix<float>(number_of_vertices, 9);

  vArray2(float) normal = world_vertices.Matrix2();
  vArray2(float) x_vertices = shape_x.world_vertices.Matrix2();
  vArray2(float) y_vertices = shape_y.world_vertices.Matrix2();
  vArray2(float) z_vertices = shape_z.world_vertices.Matrix2();
  
  long i;
  for (i = 0; i < number_of_vertices; i++)
  {
    float x = normal[0][i];
    float y = normal[1][i];
    float z = normal[2][i];

    float xx = x_vertices[0][i];
    float xy = x_vertices[1][i];
    float xz = x_vertices[2][i];

    float yx = y_vertices[0][i];
    float yy = y_vertices[1][i];
    float yz = y_vertices[2][i];

    float zx = z_vertices[0][i];
    float zy = z_vertices[1][i];
    float zz = z_vertices[2][i];

    displacements(i, 0) = (xx - x) * (float) 10.0;
    displacements(i, 1) = (xy - y) * (float) 10.0;
    displacements(i, 2) = (xz - z) * (float) 10.0;

    displacements(i, 3) = (yx - x) * (float) 10.0;
    displacements(i, 4) = (yy - y) * (float) 10.0;
    displacements(i, 5) = (yz - z) * (float) 10.0;
    
    displacements(i, 6) = (zx - x) * (float) 10.0;
    displacements(i, 7) = (zy - y) * (float) 10.0;
    displacements(i, 8) = (zz - z) * (float) 10.0;
  }
  
  return 1;
}


// Stores information into parents.
long vHandShape::SetParents()
{
  long number = NumberOfLinks();
  parents = vMatrix<long>(1, number);
  parents(0) = -1; // body
  parents(17) = 0; // forearm
  parents(1) = 17; // hand

  // inner links
  parents(2) = parents(5) = parents(8) = parents(11) = parents(14) = 1;

  // thumb
  parents(3) = 2;
  parents(4) = 3;

  // index
  parents(6) = 5;
  parents(7) = 6;

  // middle
  parents(9) = 8;
  parents(10) = 9;

  // ring
  parents(12) = 11;
  parents(13) = 12;

  // pinky
  parents(15) = 14;
  parents(16) = 15;

  return 1;
}


// Stores information into num_vertices and starting_index.
long vHandShape::SetNumberOfVertices()
{
  long number_of_links = NumberOfLinks();
  long number_of_vertices = world_vertices.Cols();
  num_vertices = vMatrix<long>(1, number_of_links);
  starting_index = vMatrix<long>(1, number_of_links);
  vArray(uchar) vids= vertex_ids.Matrix();

  long index = 0;
  while (index < number_of_vertices)
  {
    long id = vids[index];
    starting_index(id) = index;
    while (index < number_of_vertices)
    {
      if (vids[index] != id)
      {
        break;
      }
      index++;
    }
    num_vertices(id) = index - starting_index(id);
  }

  num_vertices(0) = 0;
  starting_index(0) = 0;
  return 1;
}


// Stores information into shared_vertices
long vHandShape::FindSharedVertices()
{
  long number_of_vertices = world_vertices.Cols();
  vArray2(float) vertices = world_vertices.Matrix2();
  vArray(uchar) vids = vertex_ids.Matrix();
  vector<long> link_ids;
  vector<long> indices;
  vector<long> parent_indices;

  long i;
  for (i = 0; i < number_of_vertices; i++)
  {
    long link_id = vids[i];
    float x1 = vertices[0][i];
    float y1 = vertices[1][i];
    float z1 = vertices[2][i];

    float min_distance = 10000000;
    long min_index = -1;
    long parent = parents(link_id);
    if (parent < 0) continue;
    long start = starting_index(parent);
    long number = num_vertices(parent);
    long limit = start + number;
    long j;
    for (j = start; j < limit; j++)
    {
      float x = vertices[0][j] - x1;
      float y = vertices[1][j] - y1;
      float z = vertices[2][j] - z1;

      float distance = x*x + y*y + z*z;
      if (distance < min_distance) 
      {
        min_distance = distance;
        min_index = j;
      }
    }
    if (sqrt(min_distance) < 0.000001)
    {
      link_ids.push_back(link_id);
      indices.push_back(i);
      parent_indices.push_back(min_index);
    }
  }

  long number = link_ids.size();
  shared_vertices = vMatrix<long>(number, 3);
  for (i = 0; i < number; i++)
  {
    shared_vertices(i, 0) = link_ids[i];
    shared_vertices(i, 1) = indices[i];
    shared_vertices(i, 2) = parent_indices[i];
  }

  return 1;
}

  
// Stores information into dense_points, dense_ids, and 
// scaled_dense_points.
long vHandShape::CreateDensePoints(long l_pixel_diameter)
{
  float image_diameter = (float) l_pixel_diameter;
  vector<vPoint3f> points;
  vector<uchar> ids;

  v3dMatrix<long> polygons_matrix = model->polygons;
  vArray2(long) polygons = polygons_matrix.Matrix2(0);
  vArray2(float) vertices = scaled_vertices.Matrix2();
  long number_of_corners = polygons_matrix.Rows();
  long index = 0;
  vPoint3f corners[3];
  long i, j;

  float max_distance = MaxPixelDistance(image_diameter);
  vPrint("\n");
  while(index < number_of_corners)
  {
    long link_id = polygons[index][0];
    long polygon_id = polygons[index][1];
    long polygon_corners = polygons[index][2];

    for (i = 0; i < 3; i++)
    {
      long v = polygons[index+i][3];
      corners[i] = vPoint3f(vertices[0][v], vertices[1][v], vertices[2][v]);
    }

    long number = vTrianglePixels3D(corners, &points, max_distance/(float) 2.0);
    for (j = 0; j < number; j++)
    {
      ids.push_back((uchar) link_id);
    }

    if (polygon_corners == 4)
    {
      for (i = 2; i < 3; i++)
      {
        long v = polygons[index+i][3];
        corners[i-1] = vPoint3f(vertices[0][v], vertices[1][v], vertices[2][v]);
      }
    }

    number = vTrianglePixels3D(corners, &points, max_distance/(float) 2.0);
    for (j = 0; j < number; j++)
    {
      ids.push_back((uchar) link_id);
    }

    index = index + polygon_corners;
    vPrint("processed %li of %li\r", index, number_of_corners);
  }

  vPrint("\n");
  long number = points.size();
  vPrint("number = %li\n", number);
  dense_points = vMatrix<float>(3, number);
  dense_ids = vMatrix<uchar>(1, number);

  vArray2(float) points_matrix = dense_points.Matrix2();
  vArray(uchar) ids_matrix = dense_ids.Matrix();

  for (i = 0; i < number; i++)
  {
    points_matrix[0][i] = points[i].x;
    points_matrix[1][i] = points[i].y;
    points_matrix[2][i] = points[i].z;
    ids_matrix[i] = ids[i];
  }

  return 1;
}


long vHandShape::CalculateDiameter()
{
  long number = world_vertices.Cols();
  if (number == 0) return 0;
  vArray2(float) vertices = world_vertices.Matrix2();

  long i, j;
  double max_diameter = -1;
  for (i = 0; i < number; i++)
  {
    double x1 = vertices[0][i];
    double y1 = vertices[1][i];
    double z1 = vertices[2][i];

    for (j = i+1; j < number; j++)
    {
      double x2 = vertices[0][j];
      double y2 = vertices[1][j];
      double z2 = vertices[2][j];

      double x_diff = x1 - x2;
      double y_diff = y1 - y2;
      double z_diff = z1 - z2;

      double square_distance = x_diff * x_diff + y_diff * y_diff + z_diff * z_diff;
      if (max_diameter < square_distance) max_diameter = square_distance;
    }
  }
  if (max_diameter < 0) return -1;
  diameter = sqrt(max_diameter);
  return 1;
}

  
// Given that the rendered image of the hand shape will
// have at most image_diameter diameter, calculate and
// return the max distance between a 3D point and its nearest
// neighbor in dense_points that would guarantee that
// rendering dense_points doesn't leave any holes in the image.
float vHandShape::MaxPixelDistance(float image_diameter)
{
  float diameter = Diameter();
  float result = diameter / image_diameter / ((float) max_scale / (float) min_scale);
  return result;
}

  
// Appropriate camera matrix. In the current implementation, the camera
// is placed at the origin. Is that OK or does it lead to problems?
// 
// Question. If the camera points at a given direction, are all camera
// locations equivalent? Do they give the same image up to translation?
// 
// Suppose that the camera direction is canonical. Then, for each x,y
// component, we use the nearest z. Camera location doesn't matter.
// 
// Now, how about doing a rotation. A rotation determines both camera
// orientation and camera location. The camera orientation, again, would
// be the determining factor. So, having the camera at the origin should
// be OK, and is probably the best in terms of optimizing performance.
// 
// The result is a rotation of the scaled vertices, according to the camera
// parameters. Using the result, the image coordinates of each vertex
// can be determined from the x,y values.
vMatrix<float> vHandShape::CameraPoints(float roll, float pitch, float yaw)
{
  if (is_valid <= 0) return 0;

  // Prepare the points matrix.
  long number_of_points = scaled_vertices.Cols();

  // Get the camera matrix.
  vMatrix<double> center(1, 4);
  center.WriteN(4, 0., 0., 0., 1.);
  vMatrix<float> temp_cm = vCameraMatrixFloat(roll, pitch, yaw, 
                                                vMatrix<double>::I(4), center);
  vMatrix<float> cm(3, 3);
  temp_cm.Copy(&cm, 0, 2, 0, 2, 0, 0);
  vMatrix<float> inv_cm = cm.Inv();

  // Project the points
  vMatrix<float> camera_points = inv_cm * scaled_vertices;
  return camera_points;
}


vMatrix<long> vHandShape::WorldToPixel(float roll, float pitch, float yaw, 
                                         long size, long bbox_size)
{
  vMatrix<float> camera_points = CameraPoints(roll, pitch,  yaw);
  vMatrix<long> pixels = CameraToPixel(camera_points, size, bbox_size);
  return pixels;
}


static vMatrix<long> S_CameraToPixel(vMatrix<float> camera_points, long size,
                                      long bbox_size, float min_x, float max_x,
                                      float min_y, float max_y)
{
  float world_width = max_x - min_x;
  float world_height = max_y - min_y;
  float world_length = Max(world_width, world_height);
  float scale = ((float) bbox_size) / world_length;
  float center_x = (max_x + min_x) / (float) 2.0;
  float center_y = (max_y + min_y) / (float) 2.0;
  float translation = ((float) size) / (float) 2.0;

  long number_of_points = camera_points.Cols();
  vMatrix<long> pixels_matrix(number_of_points, 2);
  vArray2(long) pixels = pixels_matrix.Matrix2();

  vArray2(float) cp = camera_points.Matrix2();

  // Debug line:
  long i;
  for (i = 0; i < number_of_points; i++)
  {
    pixels[i][0] = size - vRound((cp[1][i] - center_y) * scale + translation);
    pixels[i][1] = vRound((cp[0][i] - center_x) * scale + translation);
  }  
  return pixels_matrix;
}


vMatrix<long> vHandShape::CameraToPixel(vMatrix<float> camera_points, 
                                           long size, long bbox_size)
{
  long number_of_points = camera_points.Cols();

  // Find bounding box in 3D
  float min_x, min_y, max_x, max_y;
  min_x = max_x = camera_points(0, 0);
  min_y = max_y = camera_points(1, 0);

  vArray2(float) cp = camera_points.Matrix2();

  long i;
  for (i = 0; i < number_of_points; i++)
  {
    if (cp[0][i] < min_x) min_x = cp[0][i];
    if (cp[0][i] > max_x) max_x = cp[0][i];
    if (cp[1][i] < min_y) min_y = cp[1][i];
    if (cp[1][i] > max_y) max_y = cp[1][i];
  }

  vMatrix<long> pixels_matrix = S_CameraToPixel(camera_points, size, bbox_size,
                                                 min_x, max_x, min_y, max_y);
  return pixels_matrix;
}


color_image * vHandShape::ColorLinks(grayscale_image * link_ids)
{
  long rows = link_ids->Rows();
  long cols = link_ids->Cols();
  long size = rows * cols;
  vArray(uchar) ids = link_ids->Matrix();

  color_image * result = new color_image(rows, cols);
  vArray(uchar) R = result->GetR();
  vArray(uchar) G = result->GetG();
  vArray(uchar) B = result->GetB();

  long r, g, b;
  r = g = b = 0;

  long i;
  for (i = 0; i < size; i++)
  {
    long link_id = ids[i];
    color_map->Color(link_id, &r, &g, &b);
    R[i] = (uchar) r;
    G[i] = (uchar) g;
    B[i] = (uchar) b;
  }

  return result;
}

  
color_image * vHandShape::DrawPoints2(long size, long bbox_size)
{
  return DrawPoints5(size, bbox_size, 0, 0, 0);
}


color_image * vHandShape::DrawPoints5(long size, long bbox_size,
                                   float roll, float pitch, float yaw)
{
  if (is_valid <= 0) return 0;

  vMatrix<long> pixel_matrix = WorldToPixel(roll, pitch, yaw, size, bbox_size);
  vArray2(long) pixels = pixel_matrix.Matrix2();

  grayscale_image link_ids(size, size);
  vWriteValue(&link_ids, (uchar) 0);
  vArray2(uchar) ids = link_ids.Matrix2(0);
  vArray(uchar) vids = vertex_ids.Matrix();
  long number_of_points = pixel_matrix.Rows();

  long i;
  for (i = 0; i < number_of_points; i++)
  {
    long id = vids[i];
    long row = pixels[i][0];
    long col = pixels[i][1];
    ids[row][col] = (uchar) id;
  }
  color_image * result = ColorLinks(&link_ids);

  return result;
}


color_image * vHandShape::DrawPolygons5(long size, long bbox_size,
                                     float roll, float pitch, float yaw)
{
  if (is_valid <= 0) return 0;

  vMatrix<long> pixel_matrix = WorldToPixel(roll, pitch, yaw, size, bbox_size);
  vArray2(long) pixels = pixel_matrix.Matrix2();

  // Now, draw polygon lines.
  grayscale_image link_ids(size, size);
  vWriteValue(&link_ids, (uchar) 0);

  v3dMatrix<long> polygons_matrix = model->polygons;
  vArray2(long) polygons = polygons_matrix.Matrix2(0);
  long number_of_corners = polygons_matrix.Rows();
  long index = 0;
  while(index < number_of_corners)
  {
    long link_id = polygons[index][0];
    long polygon_id = polygons[index][1];
    long corners = polygons[index][2];
    long first_vertex_index = polygons[index][3];
    long end_index = index + corners - 1;
    long j;
    for (j = index; j < end_index; j++)
    {
      long vertex_index1 = polygons[j][3];
      long vertex_index2 = polygons[j+1][3];
      long row1 = pixels[vertex_index1][0];
      long col1 = pixels[vertex_index1][1];
      long row2 = pixels[vertex_index2][0];
      long col2 = pixels[vertex_index2][1];
      vDrawLine1(&link_ids, row1, col1, row2, col2, (uchar) link_id);      
    }

    long vertex_index1 = polygons[j][3];
    long vertex_index2 = first_vertex_index;
    long row1 = pixels[vertex_index1][0];
    long col1 = pixels[vertex_index1][1];
    long row2 = pixels[vertex_index2][0];
    long col2 = pixels[vertex_index2][1];
    vDrawLine1(&link_ids, row1, col1, row2, col2, (uchar) link_id);      
    index = index + corners;
  }

  color_image * result = ColorLinks(&link_ids);
  return result;
}


// Last working version
//color_image * vHandShape::DrawLinks5(long size, long bbox_size,
color_image * vHandShape::DrawLinksAlt5(long size, long bbox_size,
                                  float roll, float pitch, float yaw)
{
  if (is_valid <= 0) return 0;
  vMatrix<float> camera_points = CameraPoints(roll, pitch, yaw);
  vMatrix<long> pixels_matrix = CameraToPixel(camera_points, size, bbox_size);
  vArray2(long) pixels = pixels_matrix.Matrix2();
  
  // Now, draw the links
  grayscale_image link_id_matrix(size, size);
  vWriteValue(&link_id_matrix, (uchar) 0);
  v3dMatrix<float> depth_matrix(size, size, 1);
  vWriteValue(&depth_matrix, (float) -1000000);
  vArray2(uchar) link_ids = link_id_matrix.Matrix2(0);
  vArray2(float) depths = depth_matrix.Matrix2(0);

  v3dMatrix<long> polygons_matrix = model->polygons;
  vArray2(long) polygons = polygons_matrix.Matrix2(0);
  long number_of_corners = polygons_matrix.Rows();
  long index = 0;

  class_pointer (long) corner_rows = vnew (long, 4);
  class_pointer (long) corner_cols = vnew (long, 4);

  while(index < number_of_corners)
  {
    long link_id = polygons[index][0];
    long polygon_id = polygons[index][1];
    long corners = polygons[index][2];
    float z_sum = 0;
    long i;
    for (i = 0; i < corners; i++)
    {
      long v = polygons[index+i][3];
      corner_rows[i] = pixels[v][0];
      corner_cols[i] = pixels[v][1];
      z_sum += camera_points(2, v);
    }

    float z = z_sum / (float) corners;
    if (corners <= 3)
    {
      for (i = corners; i < 4; i++)
      {
        corner_rows[i] = corner_rows[i-1];
        corner_cols[i] = corner_cols[i-1];
      }
    }
    vector<vPoint> quad_pixels;
    vQuadPixels(corner_rows, corner_cols, &quad_pixels);
    long number = quad_pixels.size();
    for (i = 0; i < number; i++)
    {
      vPoint & point = quad_pixels[i];
      long row = point.row;
      long col = point.col;
      if (z > depths[row][col])
      {
        depths[row][col] = z;
        link_ids[row][col] = (uchar) link_id;
      }
    }

    index = index + corners;
  }

  color_image * result = ColorLinks(&link_id_matrix);

  delete_pointer (corner_rows);
  delete_pointer (corner_cols);
  return result;
}


// Experimental version
color_image * vHandShape::DrawLinks5(long size, long bbox_size,
//color_image * vHandShape::DrawLinksAlt5(long size, long bbox_size,
                                      float roll, float pitch, float yaw)
{
  if (is_valid <= 0) return 0;
  static vector<vPoint> triangle_pixels(100000);
  static vector<vPoint> triangle_pixels2(100000);
  vMatrix<float> camera_points = CameraPoints(roll, pitch, yaw);
  vMatrix<long> pixels_matrix = CameraToPixel(camera_points, size, bbox_size);
  vArray2(long) pixels = pixels_matrix.Matrix2();
  
  // Now, draw the links
  grayscale_image link_id_matrix(size, size);
  vWriteValue(&link_id_matrix, (uchar) 0);
  v3dMatrix<float> depth_matrix(size, size, 1);
  vWriteValue(&depth_matrix, (float) -1000000);
  vArray2(uchar) link_ids = link_id_matrix.Matrix2(0);
  vArray2(float) depths = depth_matrix.Matrix2(0);

  v3dMatrix<long> polygons_matrix = model->polygons;
  vArray2(long) polygons = polygons_matrix.Matrix2(0);
  long number_of_corners = polygons_matrix.Rows();
  long index = 0;
  class_pointer (long) corner_rows = vnew (long, 4);
  class_pointer (long) corner_cols = vnew (long, 4);

//  vector<long> stats;

  while(index < number_of_corners)
  {
    long link_id = polygons[index][0];
    long corners = polygons[index][2];
    float z_sum = 0;
    long i;
    for (i = 0; i < corners; i++)
    {
      long v = polygons[index+i][3];
      corner_rows[i] = pixels[v][0];
      corner_cols[i] = pixels[v][1];
      z_sum += camera_points(2, v);
    }

    float z = z_sum / (float) corners;

    // Note that vTrianglePixel only reads the first three
    // vertices.
    long number = vTrianglePixels(corner_rows, corner_cols, &triangle_pixels);
    long number2 = 0;
    if (corners == 4)
    {
      corner_rows[1] = corner_rows[3];
      corner_cols[1] = corner_cols[3];
      number2 = vTrianglePixels(corner_rows, corner_cols, &triangle_pixels2);
    }

    for (i = 0; i < number; i++)
    {
      vPoint & point = triangle_pixels[i];
      long row = point.row;
      long col = point.col;
      if (z > depths[row][col])
      {
        depths[row][col] = z;
        link_ids[row][col] = (uchar) link_id;
      }
    }

    for (i = 0; i < number2; i++)
    {
      vPoint & point = triangle_pixels2[i];
      long row = point.row;
      long col = point.col;
      if (z > depths[row][col])
      {
        depths[row][col] = z;
        link_ids[row][col] = (uchar) link_id;
      }
    }

    index = index + corners;
  }

  color_image * result = ColorLinks(&link_id_matrix);

//  std::sort(stats.begin(), stats.end(), less<double>());
//  long i;
//  for (i = 0; i < stats.size(); i += 100)
//  {
//    vPrint("%5li: %li\n", i, stats[i]);
//  }

//  for (i = i-100; i < stats.size(); i++)
//  {
//    vPrint("%5li: %li\n", i, stats[i]);
//  }
  delete_pointer (corner_rows);
  delete_pointer (corner_cols);

  return result;
}


long vHandShape::SetScale(float x, float y, float z)
{
  if (is_valid == 0) return 0;
  long number_of_vertices = world_vertices.Cols();
  if (displacements.Rows() != number_of_vertices) return 0;
  
  vArray2(float) d = displacements.Matrix2();
  vArray2(float) normal = world_vertices.Matrix2();
  vArray2(float) scaled = scaled_vertices.Matrix2();
  float x_factor = x - (float) 1.0;
  float y_factor = y - (float) 1.0;
  float z_factor = z - (float) 1.0;
  
  long i;
  for (i = 0; i < number_of_vertices; i++)
  {
    float x = normal[0][i];
    float y = normal[1][i];
    float z = normal[2][i];

    scaled[0][i] = x + d[i][0] * x_factor + d[i][3] * y_factor + d[i][6] * z_factor;
    scaled[1][i] = y + d[i][1] * x_factor + d[i][4] * y_factor + d[i][7] * z_factor;
    scaled[2][i] = z + d[i][2] * x_factor + d[i][5] * y_factor + d[i][8] * z_factor;
  }

  return 1;
}


// This function is used for debugging. For every link it prints out
// its world matrix, orientation matrix, and 
// world matrix * inv(orientation matrix), which is what Poser 
// considers as world matrix.
long vHandShape::PrintMatrices()
{
  vPrint("\n");
  if (is_valid <= 0) 
  {
    vPrint("Invalid shape\n");
    return 0;
  }

  long number = NumberOfLinks();
  long i;
  for (i = 0; i < number; i++)
  {
    PrintLinkMatrices(i);
  }

  return 1;
}


long vHandShape::PrintLinkMatrices(long link_id)
{
  if ((is_valid <= 0) || (link_id < 0) || (link_id >= NumberOfLinks()))
  {
    return 0;
  }

  vPrint("%2li, %s:\n", link_id, model->link_names[link_id]);
  vMatrix<float> world_matrix = world_matrices[link_id];
  vMatrix<float> orientation_matrix = orientation_matrices[link_id];
  vMatrix<float> orientation = orientations[link_id];
  vMatrix<float> poser_world = world_matrix * orientation_matrix.Inv();

  orientation.Print("orientation");
  orientation_matrix.Print("orientation_matrix");
  poser_world.Print("poser_world");
  world_matrix.Print("world_matrix");
      
  return 1;
}


