#ifndef VASSILIS_PROJECTIONS_H
#define VASSILIS_PROJECTIONS_H


#include "basics/algebra.h"
#include "basics/drawing.h"


// result = vLinkCameraParams(camera_params, world_matrix, orientation, viewpoints)
// arguments:
// camera_params: current camera parameters in poser.
// world_matrix: current world matrix of the link that the camera is pointing to.
// orientation: current orientation of the link that the camera is pointing to.
// viewpoints: a list of 3D orientations that we want to see the link under.
//
// result: a list of camera parameters (roll, pitch, yaw) that will make
// the link seen under the desired 3D orientations
vMatrix<double> vLinkCameraParams(vMatrix<double> world_matrix, 
                                    vMatrix<double> orientation,
                                    vMatrix<double> viewpoints);


// vOrientationMatrix(orientation) returns a matrix 
// that describes the rotation that must be performed
// to link coordinates before we apply the link world
// matrix to them. The argument "orientation"
// is a tuple describing rotations around the x axis,
// y axis, and z axis. The result is the matrix corresponding to 
// the product of these rotations. The order of the multiplications
// is not specified in Poser, I've been trying to figure that out.
vMatrix<double> vOrientationMatrix(vMatrix<double> orientation);


// returns the 4x4 projective matrix representing rotating by angle 
// theta around the 3D axis (x,y,z)
vMatrix<double> vRotationMatrix4(double theta, double x, double y, double z);


// returns the 3x3 matrix representing rotation by theta radians
// around axis going from origin to (x, y, z)
vMatrix<double> vRotationMatrix(double theta, double x, double y, double z);


// result = vCameraMatrix(roll, pitch, yaw, canonical_matrix, center)
// arguments:
// roll, pitch, yaw, are the camera parameters as given by Poser.
// canonical_matrix is the camera matrix (as given by Poser) corresponding
// to roll = 0, pitch = 0, yaw = 0. All angles are passed in degrees.
// center is the center of the viewing sphere. We assume that the camera
// is always located on the viewing sphere, on the only position that
// allows it to point at the center while attaining the three orientation
// parameters (roll, pitch and yaw).
// roll rotates along the image plane, without affecting 3D camera position.
// roll does not affect the camera x axis.
// yaw rotates along the world y axis, and it affects the camera x axis.
// pitch rotates along the camera x axis
vMatrix<double> vCameraMatrix(double roll, double pitch, double yaw, 
                                vMatrix<double> canonical_matrix, 
                                vMatrix<double> center);


// result = vPitchMatrix(matrix, pitch_diff, center) assumes that matrix
// is the camera matrix corresponding to roll = 0, pitch = 0, yaw = 0.
// center is the center of the viewing sphere for the camera. It returns the
// matrix that corresponds to roll = 0, pitch = pitch, yaw = 0.
// center is four-dimensional (homogeneous coordinates).
vMatrix<double> vPitchMatrix(vMatrix<double> canonical, double pitch, 
                               vMatrix<double> center);


// result = vAdjustYaw(matrix, yaw, center) assumes that matrix
// is the camera matrix corresponding to roll = 0, pitch = x, yaw = 0.
// center is the center of the viewing sphere for the camera. It returns the
// matrix that corresponds to roll = 0, pitch = pitch, yaw = yaw.
// center is four-dimensional (homogeneous coordinates).
vMatrix<double> vAdjustYaw(vMatrix<double> canonical, double yaw, 
                             vMatrix<double>center);


// returns the roll, pitch, yaw parameters, in degrees, that correspond to 
// the 3x3 rotation matrix that is the top left of link_matrix (after
// dividing by the scale of the matrix.
vMatrix<double> vLinkOrientation(vMatrix<double> link_matrix);


// Determinant of a 4x4 matrix
double vDeterminant4(vMatrix<double> matrix4);


// projection is a 4x4 matrix projecting from link coordinates to world
// coordinates.
// returns the scale part of the projection. We assume that the projection 
// consists of rotating, scaling, and translating (in that order).
double vProjectionScale(vMatrix<double> projection);


// translation is given in four dimensions (homogeneous coordinates),
// and represents a 3D displacement. We return the 4x4 matrix
// that applies that translation.
vMatrix<double> vTranslationMatrix(vMatrix<double> translation);


long vTestProjectionFunctions();


// vHandModel is a class that contains information about a hand model.
// For the time being, my intention is to use this class to store
// information exported from Poser, related to Poser's hand model. I 
// should be able to use this information to render different viewpoints
// of different joint angle configurations at different scales.
class vHandModel
{
public:
  long number_of_links;
  vArray(char *) link_names;
  v3dMatrix<double> * vertices;
  v3dMatrix<double> * weld_vertices;
  v3dMatrix<long> * polygons;
  long is_valid;

//  v3dMatrix<double> * remeshed_vertices;
//  v3dMatrix<double> * remeshed_polygons;

public:
  vHandModel();
  ~vHandModel();

  inline long IsValid()
  {
    return is_valid;
  }

  inline long NumberOfLinks()
  {
    return 18;
  }

  // PreprocessPolygons takes in a matrix describing all the polygons
  // in the hand model. The matrix has, at each row, information
  // describing a vertex of a polygon, in the following format:
  // (link_id, polygon_id, number_of_vertices, x, y, z)
  // where (x,y,z) are the coordinates of the vertex in the local
  // coordinate system of the link. Then, a new matrix is created,
  // where each of the vertices is mapped to an index into the vertices
  // matrix, so that we get rows of the form 
  // (link_id, polygon_id, number_of_vertices, index). This way, we
  // can easily look up the corresponding vertices in a particular
  // hand shape (vHandShapeSlow object), where those vertices are represented
  // in world coordinates.
  long PreprocessPolygons();

  // File where the polygon information is stored.
  char * PolygonFile();

  // File where the raw polygon information is stored (from which
  // the regular polygon matrix is built, using PreprocessPolygons);
  char * RawPolygonFile();

  // Returns a default directory where information about the hand
  // model is stored.
  char * ModelDirectory();

  // This function is used to obtain some statistics that would be useful
  // for remeshing.
  long SideLengthStats();

//  long Remesh(double max_distance);
};


class vHandShapeSlow
{
public:
  long is_valid;
  char * name;
  vHandModel * model;

  // world_vertices store unscaled coordinates. These coordinates are fixed
  // for a given joint angle configuration, and are read from disk.
  v3dMatrix<double> * world_vertices;
  vArray(vMatrix<double>) world_matrices;
  vArray(vMatrix<double>) orientation_matrices;
  vArray(vMatrix<double>) orientations;
  vArray(vMatrix<double>) scale_matrices;
  
  // displacements describe how a vertex moves (i.e. what to add
  // to its canonical scale coordinates) when we scale the x, y or z
  // axis of every link in the body.
  vMatrix<double> displacements;

  // One band for each link. In each band, fourth row is world coordinates
  // of link origin, first row is orientation of x axis, second row
  // is orientation of y axis, and third row is orientation of z axis.
  // Orientations are represented as unit vectors.
  v3dMatrix<double> * link_axes;

  // scaled_vertices store scaled coordinates, depending on the last call
  // to Scale(x, y, z);
  v3dMatrix<double> * scaled_vertices;
  vColorMap * color_map;

  // Stores the parent of each link
  vMatrix<long> parents;
  
  // Stores the number of vertices in each link.
  vMatrix<long> num_vertices;

  // Stores the starting index of the first vertex for each link
  vMatrix<long> starting_index;
  
  // Each row of shared_vertices has the form link_id1, index1, index2,
  // and the meaning is that vertex of index index1 belongs to link_id1
  // and it is identical to vertex with index index2, which belongs
  // to the parent of link_id1.
  vMatrix<long> shared_vertices;

  long Zero();
  long LoadVertices();
  long LoadMatrices();
  char * ShapeDirectory();
  char * BinFile(const char * prefix);
  char * TextFile(const char * prefix);
  char * InfoFile(const char * prefix, const char * suffix);

  char * VertexFile();
  char * MatrixFile();
  char * TextVertexFile();
  char * TextMatrixFile();

  // An auxiliary function for LoadMatrices, it reads one text
  // matrix.
  long ReadTextMatrix(FILE * fp, vMatrix<double> * matrix,
                      char letter1, char letter2);

  // This function is used to experiment with different orderings
  // in producing the orientation matrix. Each argument should be
  // 'x', 'y', or 'z', or all arguments should be '0' (in which case
  // the identity is returned).
  vMatrix<double> AltOrientationMatrix(long link_id, char first, 
                                        char second, char third);

  // Stores information into displacements.
  long SetDisplacements();

  // Stores information into parents.
  long SetParents();

  // Stores information into num_vertices and starting_index.
  long SetNumberOfVertices();

  // Stores information into shared_vertices
  long FindSharedVertices();

public:
  vHandShapeSlow();
  vHandShapeSlow(const char * name);
  ~vHandShapeSlow();

  inline long IsValid()
  {
    return is_valid;
  }

  inline long NumberOfLinks()
  {
    return 18;
  }

  vMatrix<double> CameraPoints(double roll, double pitch, double yaw);
  vMatrix<long> WorldToPixel(double roll, double pitch, double yaw, 
                              long size, long bbox_size);
  vMatrix<long> CameraToPixel(vMatrix<double> camera_points, 
                               long size, long bbox_size);
  color_image * ColorLinks(grayscale_image * link_ids);

  color_image * DrawPoints2(long size, long bbox_size);
  color_image * DrawPoints5(long size, long bbox_size,
                        double roll, double pitch, double yaw);
  color_image * DrawPolygons5(long size, long bbox_size,
                          double roll, double pitch, double yaw);
  color_image * DrawLinks5(long size, long bbox_size,
                       double roll, double pitch, double yaw);

  long ScaleX(double scale);
  long Trash();

  // This function is used to experiment with different orderings
  // in producing the orientation matrix. The arguments are passed
  // to AltOrientationMatrix to get orientation matrices for each link.
  // Then, the world matrices are adjusted accordingly.
  long AltWorldMatrices(char first, char second, char third);

  // These functions will test an alternative way of rendering that
  // applies world matrices to local vertices (as opposed to reading
  // world vertices from the disk).
  vMatrix<double> AltCameraPoints(double roll, double pitch, double yaw);
  vMatrix<long> AltWorldToPixel(double roll, double pitch, double yaw, 
                              long size, long bbox_size);
  color_image * AltDrawPoints5(long size, long bbox_size,
                        double roll, double pitch, double yaw);
  color_image * AltDrawPolygons5(long size, long bbox_size,
                          double roll, double pitch, double yaw);
  color_image * AltDrawLinks5(long size, long bbox_size,
                          double roll, double pitch, double yaw);
  long SetScale(double x, double y, double z);

  // These functions are used for debugging. For every link they print out
  // its world matrix, orientation matrix, and 
  // world matrix * inv(orientation matrix), which is what Poser 
  // considers as world matrix.
  long PrintMatrices();
  long PrintLinkMatrices(long link_id);
};


// Optimized version of vHandShapeSlow. The goal is to improve rendering
// speed, at least for extracting edge images of hands.
class vHandShape
{
public:
  long is_valid;
  char * name;
  vHandModel * model;

  // vertex_ids is a 1 x number_of_vertices matrix, that contains
  // at the i-th position the ID of the link to which the i-th
  // vertex (i.e. i-th column in world_vertices and scaled_vertices)
  // belongs.
  vMatrix<uchar> vertex_ids;

  // world_vertices store unscaled coordinates. These coordinates are fixed
  // for a given joint angle configuration and given scales, and are read 
  // from disk. Each column stores homogeneous coordinates of a vertex, 
  // in (x, y, z, 1) order.
  vMatrix<float> world_vertices;

  // displacements describe how a vertex moves (i.e. what to add
  // to its canonical scale coordinates) when we scale the x, y or z
  // axis of every link in the body.
  vMatrix<float> displacements;

  // scaled_vertices store scaled coordinates, depending on the last call
  // to Scale(x, y, z). Each column stores homogeneous coordinates of a 
  // vertex, in (x, y, z, 1) order.
  vMatrix<float> scaled_vertices;

  // These matrices are useful to determine the 3D orientation of a link.
  vArray(vMatrix<float>) world_matrices;
  vArray(vMatrix<float>) orientation_matrices;
  vArray(vMatrix<float>) orientations;

  // max distance between two points in world_matrices.
  double diameter;

  // max and min scaling parameters that we are allowed to pass
  // to SetScale.
  double max_scale, min_scale;
  
  // One band for each link. In each band, fourth row is world coordinates
  // of link origin, first row is orientation of x axis, second row
  // is orientation of y axis, and third row is orientation of z axis.
  // Orientations are represented as unit vectors.
  v3dMatrix<float> * link_axes;

  vColorMap * color_map;

  // Stores the parent of each link
  vMatrix<long> parents;
  
  // Stores the number of vertices in each link.
  vMatrix<long> num_vertices;

  // Stores the starting index of the first vertex for each link
  vMatrix<long> starting_index;
  
  // Each row of shared_vertices has the form link_id1, index1, index2,
  // and the meaning is that vertex of index index1 belongs to link_id1
  // and it is identical to vertex with index index2, which belongs
  // to the parent of link_id1.
  vMatrix<long> shared_vertices;

  // dense_points is a denser set of 3D points, that can be used to quickly
  // render a hand shape so that each pixel receives a 3D point, and no 
  // holes are left. I am not using it in practice, it turned out I need
  // about ten million points to get a mesh that is dense enough.
  vMatrix<float> dense_points;
  vMatrix<uchar> dense_ids;

  long LoadVertices();
  long LoadMatrices();
  char * ShapeDirectory();
  char * BinFile(const char * prefix);
  char * TextFile(const char * prefix);
  char * InfoFile(const char * prefix, const char * suffix);

  char * VertexFile();
  char * MatrixFile();
  char * TextVertexFile();
  char * TextMatrixFile();

  // An auxiliary function for LoadMatrices, it reads one text
  // matrix.
  long ReadTextMatrix(FILE * fp, vMatrix<float> * matrix,
                      char letter1, char letter2);

  // This function is used to experiment with different orderings
  // in producing the orientation matrix. Each argument should be
  // 'x', 'y', or 'z', or all arguments should be '0' (in which case
  // the identity is returned).
  vMatrix<double> AltOrientationMatrix(long link_id, char first, 
                                        char second, char third);

  // Stores information into displacements.
  long SetDisplacements();

  // Stores information into parents.
  long SetParents();

  // Stores information into num_vertices and starting_index.
  long SetNumberOfVertices();

  // Stores information into shared_vertices
  long FindSharedVertices();

  // Stores information into dense_points, dense_ids, and 
  // scaled_dense_points. Not used in practice, because
  // it is too slow and creates too many points.
  long CreateDensePoints(long pixel_diameter);

  inline float Diameter()
  {
    if (diameter < 0) CalculateDiameter();
    return (float) diameter;
  }

  long CalculateDiameter();

  // Given that the rendered image of the hand shape will
  // have at most image_diameter diameter, calculate and
  // return the max distance between a 3D point and its nearest
  // neighbor in dense_points that would guarantee that
  // rendering dense_points doesn't leave any holes in the image.
  float MaxPixelDistance(float image_diameter);
  long Zero();

public:
  vHandShape();
  vHandShape(const char * name);
  ~vHandShape();

  inline long IsValid()
  {
    return is_valid;
  }

  inline long NumberOfLinks()
  {
    return 18;
  }

  vMatrix<float> CameraPoints(float roll, float pitch, float yaw);
  vMatrix<long> WorldToPixel(float roll, float pitch, float yaw, 
                              long size, long bbox_size);
  vMatrix<long> CameraToPixel(vMatrix<float> camera_points, 
                               long size, long bbox_size);
  color_image * ColorLinks(grayscale_image * link_ids);

  color_image * DrawPoints2(long size, long bbox_size);
  color_image * DrawPoints5(long size, long bbox_size,
                        float roll, float pitch, float yaw);
  color_image * DrawPolygons5(long size, long bbox_size,
                          float roll, float pitch, float yaw);
  color_image * DrawLinks5(long size, long bbox_size,
                       float roll, float pitch, float yaw);
  color_image * DrawLinksAlt5(long size, long bbox_size,
                          float roll, float pitch, float yaw);

  long SetScale(float x, float y, float z);

  // These functions are used for debugging. For every link they print out
  // its world matrix, orientation matrix, and 
  // world matrix * inv(orientation matrix), which is what Poser 
  // considers as world matrix.
  long PrintMatrices();
  long PrintLinkMatrices(long link_id);
};







#endif // VASSILIS_PROJECTIONS_H
