#include "stdafx.h"

#include <math.h>
#include "quaternion.h"
#include "algo_templates.h"
#include "basics/precomputed.h"
#include "basics/drawing_temp.h"

#include "basics/definitions.h"


// These are some global variables that are used to optimize some
// quaternion computations
static const double rad_factor = M_PI / 180;
static double pi_90 = M_PI / 90.0;
static double pi_180 = M_PI / 180.0;
static double pi_360 = M_PI / 360.0;
static double two_pi = 2 * M_PI;
static double junk1 = vPrecomputedSin::PrecomputeValues();
static double junk2 = vPrecomputedCos::PrecomputeValues();
static double junk3 = vPrecomputedAcos::PrecomputeValues();

static const float * sin_values = vPrecomputedSin::Values();
static const float * cos_values = vPrecomputedCos::Values();
static const float * acos_values = vPrecomputedAcos::Values();
static double sin_factor = vPrecomputedSin::Factor();
static double acos_factor = vPrecomputedAcos::Factor();


vQuaternion::vQuaternion() : vMatrix<double>(1, 4)
{
  vWriteValue(this, (double) 0);
}


vQuaternion::~vQuaternion()
{
}


// vAngleQuaternion(theta, x, y, z) returns the quaternion that
// corresponds to rotating around axis connecting 0 to x, y, z by 
// angle theta.
vQuaternion vAngleQuaternion(double theta, double x, double y, double z)
{
  vQuaternion q;
  vArray(double) q_matrix = q.Matrix();
  double half_theta = theta / 2;
  q_matrix[0] = cos(half_theta);
  double half_theta_sin = sin(half_theta);
  double axis_norm = sqrt(x*x + y*y + z*z);
  if (axis_norm == 0)
  {
    exit_error("Zero axis passed to vAngleQuaternion.\n");
  }
  q_matrix[1] = x / axis_norm * half_theta_sin;
  q_matrix[2] = y / axis_norm * half_theta_sin;
  q_matrix[3] = z / axis_norm * half_theta_sin;
  return q;
}


// result = vQuatAngle(q) returns the angle theta such that q represents
// counterclockwise rotation around some axis by angle theta. The 
// result is in degrees
double vQuatAngle(vQuaternion & q)
{
  double q_0 = q(0);
  double half_theta = acos(q(0));
  double theta = 2 * half_theta;
  double result = theta * 180 / M_PI;
  if (result >= 360.0) result -= 360.0;
  return result;
}


// result = quat_axis(q)
// Returns the axis of rotation for the rotation that q represents
vMatrix<double> vQuatAxis(vQuaternion & q)
{
  vMatrix<double> result_matrix(1, 3);
  vArray(double) result = result_matrix.Matrix();
  double half_theta = acos(q(0));
  double half_theta_sin = sin(half_theta);

  // Take care of numerical errors
  if (vAbs(half_theta_sin) < 0.000001)
  {
    vWriteValue(&result_matrix, (double) 0);
    return result_matrix;
  }

  result[0] = q(1) / half_theta_sin;
  result[1] = q(2) / half_theta_sin;
  result[2] = q(3) / half_theta_sin;
  return result_matrix;
}

// Returns the inverse of q
vQuaternion vQuatInverse(vQuaternion & q)
{
  vQuaternion result;
  result(0) = q(0);
  result(1) = -q(1);
  result(2) = -q(2);
  result(3) = -q(3);
  double norm = q(0) * q(0) + q(1) * q(1) + q(2) * q(2) + q(3) * q(3);
  if (norm == 0)
  {
    exit_error("Inverse called on the zero quaternion\n");
  }
  result(0) /= norm;
  result(1) /= norm;
  result(2) /= norm;
  result(3) /= norm;
  return result;
}


// Returns the quaternion multiplication of q1 and q2
vQuaternion vQuatMultiply(vQuaternion & q1, vQuaternion & q2)
{
  double s1 = q1(0);
  double s2 = q2(0);
  vMatrix<double> v1(1, 3);
  vMatrix<double> v2(1, 3);
  // v1(1:3) = q1(2:4);
  q1.Copy(&v1, 0, 0, 1, 3, 0, 0);
  // v2(1:3) = q2(2:4);
  q2.Copy(&v2, 0, 0, 1, 3, 0, 0);
  vQuaternion q;
  double q_0 = q1(0) * q2(0) - vDot(v1, v2);
  // Adjust for numerical errors.
//  if (q_0 > 1) q_0 = 1;
//  if (q_0 < -1) q_0 = -1;
  q(0) = q_0;
  
  //result(2:4) = s1 * v2 + s2 * v1 + cross(v1, v2);
  vMatrix<double> v = s1 * v2 + s2 * v1 + vCross(v1, v2);
  q(1) = v(0);
  q(2) = v(1);
  q(3) = v(2);

  return q;
}


// Rotates the point represented by quaternion point by applying the 
// rotation represented by quaternion q. The 3D point (x, y, z) 
// is represented as the quaternion (0, x, y, z);
vQuaternion vQuatRotate(vQuaternion & point, vQuaternion & q)
{
  //  Matlab code:
  //  q_inverse = quat_inverse(q);
  //  result = quat_multiply(q, quat_multiply(point, q_inverse));
  vQuaternion q_inverse = vQuatInverse(q);
  vQuaternion result = vQuatMultiply(q, vQuatMultiply(point, q_inverse));
  return result;
}


// result = vViewDifference(x1, y1, z1, x2, y2, z2)
// (x1, y1, z1) and (x2, y2, z2) express camera view parameters as
// used in my HandRendering program (x is latitude, y is longitude, z is 
// image plane rotation). The result returned here is the quaternion
// representing the rotation to go from (x1, y1, z1) to (x2, y2, z2)
vQuaternion vOldViewDifference(double x1, double y1, double z1,
                                 double x2, double y2, double z2)
{
  //  Matlab code:
  //  q1 = view_quaternion(x1, y1, z1);
  //  q2 = view_quaternion(x2, y2, z2);
  //  result = quat_multiply(q2, quat_inverse(q1));
  vQuaternion q1 = vOldViewQuaternion(x1, y1, z1);
  vQuaternion q2 = vOldViewQuaternion(x2, y2, z2);
  vQuaternion result = vQuatMultiply(q2, vQuatInverse(q1));
  return result;
}


// result = vViewDistance(x1, y1, z1, x2, y2, z2) computes the angle
// distance between viewpoints (x1, y1, z1) and (x2, y2, z2) (as
// used in my HandRendering program). The arguments and result are 
// in degrees. 
double vOldViewDistance(double x1, double y1, double z1,
                         double x2, double y2, double z2)
{
  //  Matlab code
  //  q1 = view_quaternion(x1, y1, z1);
  //  q2 = view_quaternion(x2, y2, z2);
  //  q1_to_q2 = quat_multiply(q2, quat_inverse(q1));
  //  result1 = quat_angle(q1_to_q2);
  //  result2 = 360 - result1;
  //  result = min(result1, result2);
  vQuaternion q1 = vOldViewQuaternion(x1, y1, z1);
  vQuaternion q2 = vOldViewQuaternion(x2, y2, z2);
  vQuaternion q1_to_q2 = vQuatMultiply(q2, vQuatInverse(q1));
  double result1 = vAbs(vQuatAngle(q1_to_q2));
  double result2 = 360 - result1;
  double result = Min(result1, result2);
  return result;
}


// result = vViewQuaternion(x_angle, y_angle, orientation) returns 
// the quaternion that performs the rotation that moves the camera viewpoint
// and orientation from the north pole (0, 0, 0) to latitude x_angle,
// longitude y_angle and image plane rotation orientation. This is
// applicable to the camera viewpoint used by my HandRendering program.
// Angle values are expressed in degrees.
vQuaternion vOldViewQuaternion(double x_angle, double y_angle, 
                                 double orientation)
{
  // convert from degrees to rads
  double x_rad = x_angle / 180.0 * M_PI;
  double y_rad = y_angle / 180.0 * M_PI;
  double z_rad = -orientation / 180.0 * M_PI;

  // First, rotate around y axis
  // Get quaternion for rotation around y axis
  vQuaternion q_y = vAngleQuaternion(y_rad, 0, 1, 0);
  // rotate x axis accordingly
  vQuaternion x_axis;
  x_axis(1) = 1.0;
  vQuaternion x_axis_rotated = vQuatRotate(x_axis, q_y);
  // Get the quaternion that rotates around the new x axis by x_rad
  vQuaternion q_x = vAngleQuaternion(x_rad, x_axis_rotated(1), x_axis_rotated(2), 
                                       x_axis_rotated(3));
  // Now, move from north pole to new viewpoint by rotating first around
  // y axis and then about the new x axis
  vQuaternion north_pole;
  north_pole(2) = 1.0;
  vQuaternion viewpoint = vQuatRotate(north_pole, q_x);

  vQuaternion image_plane_rotation = vAngleQuaternion(z_rad, viewpoint(1), viewpoint(2), 
                                                        viewpoint(3));
  vQuaternion result = vQuatMultiply(image_plane_rotation, vQuatMultiply(q_x, q_y));
  return result;
}


// The vViewDistanceTimer functions compute lots of quaternion distances
// in different ways, and can be used to evaluate the efficiency of each way.
double vOldViewDistanceTimer1(long times)
{
//  long long_x1 = vRandomLong2(0, 90);
//  long long_y1 = vRandomLong2(0, 360);
//  long long_z1 = vRandomLong2(0, 360);

//  long long_x2 = vRandomLong2(0, 90);
//  long long_y2 = vRandomLong2(0, 360);
//  long long_z2 = vRandomLong2(0, 360);

//  double x1 = long_x1;
//  double y1 = long_y1;
//  double z1 = long_z1;

//  double x2 = long_x2;
//  double y2 = long_y2;
//  double z2 = long_z2;

  double x1 = 50;
  double y1 = 120;
  double z1 = 310;

  double x2 = 10;
  double y2 = 30;
  double z2 = 163;

  long i;
  double distance;
  for (i = 0; i < times; i++)
  {
    distance = vOldViewDistance(x1, y1, z1, x2, y2, z2);
  }

  return distance;
}


// S_OldViewQuaternion3 is not used, but is useful as the prototype
// for S_OldViewQuaternion2 (which is an optimized version of this).
// vOldViewQuaternion is the prototype for S_OldViewQuaternion3
vQuaternion S_OldViewQuaternion3(double x, double y, double z)
{
  // general variables
  double q1_0, q1_1, q1_2, q1_3;
  double q2_0, q2_1, q2_2, q2_3;
  double q3_0, q3_1, q3_2, q3_3;
  double q4_0, q4_1, q4_2, q4_3;
  double norm;

  // variables used in multiplying
  double dot, cross_0, cross_1, cross_2;

  // variables used in vAngleQuaternion
  double half_theta, axis_norm, half_theta_sin;

  // convert from degrees to rads
  double x_rad = x / 180.0 * M_PI;
  double y_rad = y / 180.0 * M_PI;
  double z_rad = -z / 180.0 * M_PI;

  // First, rotate around y axis
  // Get quaternion for rotation around y axis
  // vQuaternion q_y = vAngleQuaternion(y_rad, 0, 1, 0);
  q1_0 = 0;
  q1_1 = 1;
  q1_2 = 0;

  half_theta = y_rad / 2;
  double q_y_0 = cos(half_theta);
  axis_norm = sqrt(q1_0*q1_0 + q1_1*q1_1 + q1_2*q1_2);
  half_theta_sin = sin(half_theta);
  if (axis_norm == 0)
  {
    exit_error("Zero axis passed to vAngleQuaternion.\n");
  }

  double q_y_1 = q1_0 / axis_norm * half_theta_sin;
  double q_y_2 = q1_1 / axis_norm * half_theta_sin;
  double q_y_3 = q1_2 / axis_norm * half_theta_sin;
  // End of vAngleQuaternion


  // rotate x axis accordingly
  //vQuaternion x_axis_rotated = vQuatRotate(x_axis, q_y);
  // First invert q_y
  q3_0 = q_y_0;
  q3_1 = q_y_1;
  q3_2 = q_y_2;
  q3_3 = q_y_3;
//  q3_0 = q_y(0);
 // q3_1 = q_y(1);
 // q3_2 = q_y(2);
 // q3_3 = q_y(3);
  norm = q3_0 * q3_0 + q3_1 * q3_1 + q3_2 * q3_2 + q3_3 * q3_3;
  q2_0 = q3_0 / norm;
  q2_1 = -q3_1 / norm;
  q2_2 = -q3_2 / norm;
  q2_3 = -q3_3 / norm;

  // Now multiply the x_axis with q_x inverse
  q1_0 = 0;
  q1_1 = 1;
  q1_2 = 0;
  q1_3 = 0;

  dot = q1_1 * q2_1 + q1_2 * q2_2 + q1_3 * q2_3;
  cross_0 = q1_2 * q2_3 - q1_3 * q2_2;
  cross_1 = q1_3 * q2_1 - q1_1 * q2_3;
  cross_2 = q1_1 * q2_2 - q1_2 * q2_1;

  q4_0 = q1_0 * q2_0 - dot;
  q4_1 = q1_0 * q2_1 + q2_0 * q1_1 + cross_0;
  q4_2 = q1_0 * q2_2 + q2_0 * q1_2 + cross_1;
  q4_3 = q1_0 * q2_3 + q2_0 * q1_3 + cross_2;

  // Now multiply q_x by the previous result

  dot = q3_1 * q4_1 + q3_2 * q4_2 + q3_3 * q4_3;
  cross_0 = q3_2 * q4_3 - q3_3 * q4_2;
  cross_1 = q3_3 * q4_1 - q3_1 * q4_3;
  cross_2 = q3_1 * q4_2 - q3_2 * q4_1;

  double x_axis_rotated_0 = q3_0 * q4_0 - dot;
  double x_axis_rotated_1 = q3_0 * q4_1 + q4_0 * q3_1 + cross_0;
  double x_axis_rotated_2 = q3_0 * q4_2 + q4_0 * q3_2 + cross_1;
  double x_axis_rotated_3 = q3_0 * q4_3 + q4_0 * q3_3 + cross_2;

  // End of multiplying

  // Get the quaternion that rotates around the new x axis by x_rad
  //vQuaternion q_x = vAngleQuaternion(x_rad, x_axis_rotated(1), x_axis_rotated(2), 
  //                                     x_axis_rotated(3));

  q1_0 = x_axis_rotated_1;
  q1_1 = x_axis_rotated_2;
  q1_2 = x_axis_rotated_3;

  half_theta = x_rad / 2;
  double q_x_0 = cos(half_theta);
  axis_norm = sqrt(q1_0*q1_0 + q1_1*q1_1 + q1_2*q1_2);
  half_theta_sin = sin(half_theta);
  if (axis_norm == 0)
  {
    exit_error("Zero axis passed to vAngleQuaternion.\n");
  }

  double q_x_1 = q1_0 / axis_norm * half_theta_sin;
  double q_x_2 = q1_1 / axis_norm * half_theta_sin;
  double q_x_3 = q1_2 / axis_norm * half_theta_sin;
  // End of vAngleQuaternion
  
  // vQuaternion viewpoint = vQuatRotate(north_pole, q_x);
  
  // First invert q_x
  q3_0 = q_x_0;
  q3_1 = q_x_1;
  q3_2 = q_x_2;
  q3_3 = q_x_3;
  norm = q3_0 * q3_0 + q3_1 * q3_1 + q3_2 * q3_2 + q3_3 * q3_3;
  q2_0 = q3_0 / norm;
  q2_1 = -q3_1 / norm;
  q2_2 = -q3_2 / norm;
  q2_3 = -q3_3 / norm;

  // Now multiply the north_pole with q_x inverse
  q1_0 = 0;
  q1_1 = 0;
  q1_2 = 1;
  q1_3 = 0;

  dot = q1_1 * q2_1 + q1_2 * q2_2 + q1_3 * q2_3;
  cross_0 = q1_2 * q2_3 - q1_3 * q2_2;
  cross_1 = q1_3 * q2_1 - q1_1 * q2_3;
  cross_2 = q1_1 * q2_2 - q1_2 * q2_1;

  q4_0 = q1_0 * q2_0 - dot;
  q4_1 = q1_0 * q2_1 + q2_0 * q1_1 + cross_0;
  q4_2 = q1_0 * q2_2 + q2_0 * q1_2 + cross_1;
  q4_3 = q1_0 * q2_3 + q2_0 * q1_3 + cross_2;

  // Now multiply q_x by the previous result

  dot = q3_1 * q4_1 + q3_2 * q4_2 + q3_3 * q4_3;
  cross_0 = q3_2 * q4_3 - q3_3 * q4_2;
  cross_1 = q3_3 * q4_1 - q3_1 * q4_3;
  cross_2 = q3_1 * q4_2 - q3_2 * q4_1;

  double viewpoint_0 = q3_0 * q4_0 - dot;
  double viewpoint_1 = q3_0 * q4_1 + q4_0 * q3_1 + cross_0;
  double viewpoint_2 = q3_0 * q4_2 + q4_0 * q3_2 + cross_1;
  double viewpoint_3 = q3_0 * q4_3 + q4_0 * q3_3 + cross_2;

  // End of multiplying
  // End of vQuatRotate

//  vQuaternion q_inverse = vQuatInverse(q);
//  vQuaternion result = vQuatMultiply(q, vQuatMultiply(point, q_inverse));


  // setting image_plane_rotation = vAngleQuaternion(z_rad, viewpoint(1), viewpoint(2), 
  //                                                  viewpoint(3))
  q1_0 = viewpoint_1;
  q1_1 = viewpoint_2;
  q1_2 = viewpoint_3;

  half_theta = z_rad / 2;
  double image_plane_rotation_0 = cos(half_theta);
  axis_norm = sqrt(q1_0*q1_0 + q1_1*q1_1 + q1_2*q1_2);
  half_theta_sin = sin(half_theta);
  if (axis_norm == 0)
  {
    exit_error("Zero axis passed to vAngleQuaternion.\n");
  }
  double image_plane_rotation_1 = q1_0 / axis_norm * half_theta_sin;
  double image_plane_rotation_2 = q1_1 / axis_norm * half_theta_sin;
  double image_plane_rotation_3 = q1_2 / axis_norm * half_theta_sin;


  // Multiplying q_x * q_y
  q1_0 = q_x_0;
  q1_1 = q_x_1;
  q1_2 = q_x_2;
  q1_3 = q_x_3;

  q2_0 = q_y_0;
  q2_1 = q_y_1;
  q2_2 = q_y_2;
  q2_3 = q_y_3;

  dot = q1_1 * q2_1 + q1_2 * q2_2 + q1_3 * q2_3;
  cross_0 = q1_2 * q2_3 - q1_3 * q2_2;
  cross_1 = q1_3 * q2_1 - q1_1 * q2_3;
  cross_2 = q1_1 * q2_2 - q1_2 * q2_1;

  q3_0 = q1_0 * q2_0 - dot;
  q3_1 = q1_0 * q2_1 + q2_0 * q1_1 + cross_0;
  q3_2 = q1_0 * q2_2 + q2_0 * q1_2 + cross_1;
  q3_3 = q1_0 * q2_3 + q2_0 * q1_3 + cross_2;
  // End of multiplying

  // multiplying image_plane_rotation1 by q3
  q1_0 = image_plane_rotation_0;
  q1_1 = image_plane_rotation_1;
  q1_2 = image_plane_rotation_2;
  q1_3 = image_plane_rotation_3;

  q2_0 = q3_0;
  q2_1 = q3_1;
  q2_2 = q3_2;
  q2_3 = q3_3;

  dot = q1_1 * q2_1 + q1_2 * q2_2 + q1_3 * q2_3;
  cross_0 = q1_2 * q2_3 - q1_3 * q2_2;
  cross_1 = q1_3 * q2_1 - q1_1 * q2_3;
  cross_2 = q1_1 * q2_2 - q1_2 * q2_1;

  q3_0 = q1_0 * q2_0 - dot;
  q3_1 = q1_0 * q2_1 + q2_0 * q1_1 + cross_0;
  q3_2 = q1_0 * q2_2 + q2_0 * q1_2 + cross_1;
  q3_3 = q1_0 * q2_3 + q2_0 * q1_3 + cross_2;
  // end of multiplying

  vQuaternion result;
  result(0) = q3_0;
  result(1) = q3_1;
  result(2) = q3_2;
  result(3) = q3_3;

  return result;
}


// The caller must insure that x, y and z are between 0 and 360 degrees
// This version is highly optimized and because of that it skips many
// tests to validate its inputs. The program can crash if x, y or z
// are outside the range of 0-360 degrees
void S_OldViewQuaternion2(double x, double y, double z, double * q)
{
  // general variables
  double q1_0, q1_1, q1_2, q1_3;

  // variables used in multiplying
  double dot, cross_0, cross_1, cross_2;

  // variables used in vAngleQuaternion
  double half_theta_sin;

  // variables used to get precomputed sin and cos values
  double weight1, weight2, double_index;
  long index, inc_index;

  double half_x_rad = x * pi_360;
  double half_y_rad = y * pi_360;
  double half_z_rad = M_PI - z * pi_360;

  // First, rotate around y axis
  // Get quaternion for rotation around y axis
  // vQuaternion q_y = vAngleQuaternion(y_rad, 0, 1, 0);
  //double q_y_0 = vPrecomputedCos::Cos(half_theta);
  //double q_y_2 = vPrecomputedSin::Sin(half_theta);
  double_index = half_y_rad  * sin_factor;
  index = (long) double_index;
  inc_index = index+1;
  weight1 = inc_index-double_index;
  weight2 = double_index-index;
  double q_y_2 = weight1 * sin_values[index] + weight2 * sin_values[inc_index];
  double q_y_0 = weight1 * cos_values[index] + weight2 * cos_values[inc_index];

  // End of vAngleQuaternion


  // rotate x axis accordingly
  //vQuaternion x_axis_rotated = vQuatRotate(x_axis, q_y);
  // multiply the x_axis with q_x inverse
  double x_axis_rotated_1 = 2 * q_y_0 * q_y_0 - 1;
  double x_axis_rotated_3 = - 2 * q_y_0 * q_y_2;
  // End of multiplying

  // Get the quaternion that rotates around the new x axis by x_rad
  //vQuaternion q_x = vAngleQuaternion(x_rad, x_axis_rotated(1), x_axis_rotated(2), 
  //                                     x_axis_rotated(3));
  //double q_x_0 = vPrecomputedCos::Cos(half_theta);
  //half_theta_sin = vPrecomputedSin::Sin(half_theta);
  double_index = half_x_rad * sin_factor;
  index = (long) double_index;
  inc_index = index+1;
  weight1 = inc_index-double_index;
  weight2 = double_index-index;
  half_theta_sin = weight1 * sin_values[index] + weight2 * sin_values[inc_index];
  double q_x_0 = weight1 * cos_values[index] + weight2 * cos_values[inc_index];

  double q_x_1 = x_axis_rotated_1 * half_theta_sin;
  double q_x_3 = x_axis_rotated_3 * half_theta_sin;
  // End of vAngleQuaternion
  
  // vQuaternion viewpoint = vQuatRotate(north_pole, q_x);
  // multiply q_x by (north_pole * q_x inverse)
  cross_1 = -q_x_3 * q_x_3 - q_x_1 * q_x_1;
  double viewpoint_1 = -2 * q_x_0 * q_x_3;
  double viewpoint_2 = q_x_0 * q_x_0 + cross_1;
  double viewpoint_3 = 2 * q_x_0 * q_x_1;
  // End of multiplying
  // End of vQuatRotate

  // setting image_plane_rotation = vAngleQuaternion(z_rad, viewpoint(1), viewpoint(2), 
  //                                                  viewpoint(3))
  //half_theta = z_rad / 2;
  //double image_plane_rotation_0 = vPrecomputedCos::Cos(half_theta);
  //half_theta_sin = vPrecomputedSin::Sin(half_theta);
  double_index = half_z_rad * sin_factor;
  index = (long) double_index;
  inc_index = index+1;
  weight1 = inc_index-double_index;
  weight2 = double_index-index;
  half_theta_sin = weight1 * sin_values[index] + weight2 * sin_values[inc_index];
  double image_plane_rotation_0 = weight1 * cos_values[index] + weight2 * cos_values[inc_index];

  double image_plane_rotation_1 = viewpoint_1 * half_theta_sin;
  double image_plane_rotation_2 = viewpoint_2 * half_theta_sin;
  double image_plane_rotation_3 = viewpoint_3 * half_theta_sin;


  // Multiplying q_x * q_y
  cross_0 = -q_x_3 * q_y_2;
  cross_2 = q_x_1 * q_y_2;

  q1_0 = q_x_0 * q_y_0;
  q1_1 = q_y_0 * q_x_1 + cross_0;
  q1_2 = q_x_0 * q_y_2;
  q1_3 = q_y_0 * q_x_3 + cross_2;

  // End of multiplying

  // multiplying image_plane_rotation1 by q3
  dot = image_plane_rotation_1 * q1_1 + image_plane_rotation_2 * q1_2 + image_plane_rotation_3 * q1_3;
  cross_0 = image_plane_rotation_2 * q1_3 - image_plane_rotation_3 * q1_2;
  cross_1 = image_plane_rotation_3 * q1_1 - image_plane_rotation_1 * q1_3;
  cross_2 = image_plane_rotation_1 * q1_2 - image_plane_rotation_2 * q1_1;

  q[0] = image_plane_rotation_0 * q1_0 - dot;
  q[1] = image_plane_rotation_0 * q1_1 + q1_0 * image_plane_rotation_1 + cross_0;
  q[2] = image_plane_rotation_0 * q1_2 + q1_0 * image_plane_rotation_2 + cross_1;
  q[3] = image_plane_rotation_0 * q1_3 + q1_0 * image_plane_rotation_3 + cross_2;
  // end of multiplying
}


// An optimized version of vOldViewDistance
double vOldViewDistance2(double x1, double y1, double z1, 
                          double x2, double y2, double z2)
{
  double q1[4], q2[4];

  S_OldViewQuaternion2(x1, y1, z1, q1);
  S_OldViewQuaternion2(x2, y2, z2, q2);

  double q3_0 = q1[0] * q2[0] + q1[1] * q2[1] + q1[2] * q2[2] + q1[3] * q2[3];

  float double_index = ((float) q3_0 + (float) 1) * (float) acos_factor;
  long index = (long) double_index;
  long inc_index = index+1;
  float weight1 = inc_index-double_index;
  float weight2 = double_index-index;
  double inv_cosine = weight1 * acos_values[index] + weight2 * acos_values[inc_index];
  double result = inv_cosine / pi_360;
  if (result > 180) result = 360 - result;
  return result;
}


double vOldViewDistanceTimer2(long times)
{
//  long long_x1 = vRandomLong2(0, 90);
//  long long_y1 = vRandomLong2(0, 360);
//  long long_z1 = vRandomLong2(0, 360);

//  long long_x2 = vRandomLong2(0, 90);
//  long long_y2 = vRandomLong2(0, 360);
//  long long_z2 = vRandomLong2(0, 360);

//  double x1 = long_x1;
//  double y1 = long_y1;
//  double z1 = long_z1;

//  double x2 = long_x2;
//  double y2 = long_y2;
//  double z2 = long_z2;

  double x1 = 50;
  double y1 = 120;
  double z1 = 310;

  double x2 = 10;
  double y2 = 30;
  double z2 = 163;

  long i;
  double distance;
  for (i = 0; i < times; i++)
  {
    distance = vOldViewDistance2(x1, y1, z1, x2, y2, z2);
  }

  return distance;
}


static double S_Empty(double x1, double y1, double z1, 
                      double x2, double y2, double z2)
{
  return log(0.2);
}


double vOldViewDistanceTimer3(long times)
{
  long long_x1 = vRandomLong2(0, 90);
  long long_y1 = vRandomLong2(0, 360);
  long long_z1 = vRandomLong2(0, 360);

  long long_x2 = vRandomLong2(0, 90);
  long long_y2 = vRandomLong2(0, 360);
  long long_z2 = vRandomLong2(0, 360);

  double x1 = long_x1;
  double y1 = long_y1;
  double z1 = long_z1;

  double x2 = long_x2;
  double y2 = long_y2;
  double z2 = long_z2;

  long i;
  double distance;
  for (i = 0; i < times; i++)
  {
    distance = S_Empty(x1, y1, z1, x2, y2, z2);
  }

  return distance;
}


// This function computes the 3D position of a point on the surface
// of the unit sphere specified by x_angle and y_angle, i.e. 
// x_angle degrees south of the north pole and y_angle degrees east
// of the x=0 plane. In the 3D coordinates, x increases towards the 
// right, y increases upwards, and z increases in the direction 
// from the screen towards me.
vPoint3 vOldSphereSurfaceTo3D(double x_angle, double y_angle)
{
  static const double junk_value1 = vPrecomputedSin::PrecomputeValues();
  static const double junk_value2 = vPrecomputedCos::PrecomputeValues();
  x_angle = vAdjustAngle(x_angle);
  y_angle = vAdjustAngle(y_angle);
	double rad_x = x_angle * rad_factor;
	double rad_y = y_angle * rad_factor;

  vPoint3 result;
  result.y = vPrecomputedCos::Cos(rad_x);
	double non_y = sqrt(1 - result.y * result.y);
  result.x = non_y * vPrecomputedSin::Sin(rad_y);
  result.z = non_y * vPrecomputedCos::Cos(rad_y);
  return result;
}


// This is the inverse of vSphereSurfacePoint.
vPoint3 vOld3DToSphereSurface(double camera_x, double camera_y, 
                                double camera_z)
{
  double temp = camera_x * camera_x + camera_y * camera_y + camera_z * camera_z;
  double distance_from_origin = sqrt(temp);

  if (distance_from_origin == 0) 
  {
    exit_error("Error: v3DToSphereSurface called with 0 distance from origin\n");
  }

  double temp_x, temp_y, temp_z, x_angle, y_angle;
  temp_x = camera_x / distance_from_origin;
  temp_y = camera_y / distance_from_origin;
  temp_z = camera_z / distance_from_origin;

	double non_y = sqrt(1 - temp_y * temp_y);
  if (non_y == 0)
  {
    x_angle = acos(temp_y) * 360 / (2 * M_PI);
    return vPoint3(x_angle, 0, 0);
  }

  double sin_y, cos_y, cos_x, rad_x, rad_y;
  cos_x = temp_y;
  sin_y = temp_x / non_y;
  cos_y = temp_z / non_y;
  
  // We have to take care of certain numerical errors, to make sure 
  // everything is well defined. At some point I got 
  // cos_y = -1 - 2.22e-16.

  if (cos_x > 1) cos_x = 1;
  if (cos_x < -1) cos_x = -1;
  if (sin_y > 1) sin_y = 1;
  if (sin_y < -1 ) sin_y = -1;
  if (cos_y > 1) cos_y = 1;
  if (cos_y < -1) cos_y = -1;

  rad_x = acos(cos_x);
  rad_y = acos(cos_y);
  assert(rad_y >= 0);
  if (sin_y < 0) rad_y = 2 * M_PI - rad_y;
  x_angle = rad_x * 360 / (2 * M_PI);
  y_angle = rad_y * 360 / (2 * M_PI);
  return vPoint3(x_angle, y_angle, 0);
}


// vQuaternionDistance returns the absolute value of the angle
// of the smallest rotation that converts q1 into q2.
double vQuaternionDistance(vQuaternion & q1, vQuaternion & q2)
{
  vQuaternion difference = vQuatMultiply(q2, vQuatInverse(q1));
  double angle1 = vAbs(vQuatAngle(difference));
  double angle2 = 360 - angle1;
  double result = Min(angle1, angle2);
  return result;
}


// Computes the quaternion that rotates by theta degrees 
// around the axis defined
// by connecting the center of the unit sphere to the surface point
// on the sphere with coordinates (x,y) (x degrees south of the north 
// pole, y degrees east)
vQuaternion vOldRotationQuaternion(double theta, double x, double y)
{
  vPoint3 point = vOldSphereSurfaceTo3D(x, y);
  vQuaternion result = vAngleQuaternion(theta, point.x, point.y, point.z);
  return result;
}


// Computes the distance between the quaternions defined by
// these two sets of parameters.
double vOldRotationDistance(double theta1, double x1, double y1,
                             double theta2, double x2, double y2)
{
  double rad1 = theta1 * rad_factor;
  double rad2 = theta2 * rad_factor;
  vQuaternion q1 = vOldRotationQuaternion(theta1, x1, y1);
  vQuaternion q2 = vOldRotationQuaternion(theta2, x2, y2);
  double distance = vQuaternionDistance(q1, q2);
  return distance;
}


// The inverse of vViewQuaternion. It puts in x_angle_p,
// y_angle_p, z_angle_p, the camera viewpoint and orientation
// corresponding to the quaternion q. Angles are returned
// in degrees.
long vOldQuaternionView(vQuaternion & q, double * x_angle_p,
                      double * y_angle_p, double * z_angle_p)
{
  vQuaternion north_pole;
  north_pole(2) = 1.0;

  // We want to figure out the camera viewpoint and orientation
  // that corresponds to q. The camera viewpoint can be obtained
  // by rotating the north pole using q.
  vQuaternion viewpoint = vQuatRotate(north_pole, q);
  double camera_x = viewpoint(1);
  double camera_y = viewpoint(2);
  double camera_z = viewpoint(3);

  double temp = camera_x * camera_x + camera_y * camera_y + camera_z * camera_z;
  double distance_from_origin = sqrt(temp);

  if (distance_from_origin == 0) return 0;

  double temp_x, temp_y, temp_z;
  temp_x = camera_x / distance_from_origin;
  temp_y = camera_y / distance_from_origin;
  temp_z = camera_z / distance_from_origin;

  double x_angle, y_angle, z_angle;
    
  double non_y = sqrt(1 - temp_y * temp_y);
  if (non_y == 0)
  {
    x_angle = acos(temp_y) * 180.0 / M_PI;
    y_angle = 0;
  }
 
  else
  {
    double sin_y, cos_y, cos_x, rad_x, rad_y;
    cos_x = temp_y;
    sin_y = temp_x / non_y;
    cos_y = temp_z / non_y;
  
    // We have to take care of certain numerical errors, to make sure 
    // everything is well defined. At some point I got 
    // cos_y = -1 - 2.22e-16.

    if (cos_x > 1) cos_x = 1;
    if (cos_x < -1) cos_x = -1;
    if (sin_y > 1) sin_y = 1;
    if (sin_y < -1 ) sin_y = -1;
    if (cos_y > 1) cos_y = 1;
    if (cos_y < -1) cos_y = -1;

    rad_x = acos(cos_x);
    rad_y = acos(cos_y);
    assert(rad_y >= 0);
    if (sin_y < 0) rad_y = 2 * M_PI - rad_y;
    x_angle = rad_x * 180 / M_PI;
    y_angle = rad_y * 180 / M_PI;
  }

  vQuaternion viewpoint_quat = vOldViewQuaternion(x_angle, y_angle, 0);
  vQuaternion viewpoint_to_q = vQuatMultiply(q, vQuatInverse(viewpoint_quat));

  double axis_dot_product = 0;
  long i;
  for (i = 1; i <= 3; i++)
  {
    axis_dot_product += viewpoint(i) * viewpoint_to_q(i);
  }

  z_angle = vQuatAngle(viewpoint_to_q);
  if (axis_dot_product >= 0) z_angle = 360 - z_angle;
  if (z_angle >= 360) z_angle -= 360;

  *x_angle_p = x_angle;
  *y_angle_p = y_angle;
  *z_angle_p = z_angle;
  return 1;
}


////////////////////////////////////////////////////////////////
//                                                            //
//       Beginning of implementation of vQuaternion2         //
//                                                            //
////////////////////////////////////////////////////////////////


vQuaternion2::vQuaternion2(double theta, vPoint3 & point)
{
  static long junk1 = vPrecomputeValues();

  double half_theta = theta / 2;
  cos_half_theta = vPrecomputedCos::Cos(half_theta);
  double half_theta_sin = vPrecomputedSin::Sin(half_theta);
  double axis_norm = sqrt(x*x + y*y + z*z);
  if (axis_norm == 0)
  {
    exit_error("Zero axis passed to vAngleQuaternion.\n");
  }

  double factor = half_theta_sin / axis_norm;
  x = point.x * factor;
  y= point.y * factor;
  z = point.z * factor;
}


vQuaternion2::vQuaternion2(double theta, double x_axis, double y_axis, double z_axis)
{
  static long junk1 = vPrecomputeValues();

  double half_theta = theta / 2;
  cos_half_theta = vPrecomputedCos::Cos(half_theta);
  double half_theta_sin = vPrecomputedSin::Sin(half_theta);
  double axis_norm = sqrt(x_axis*x_axis + y_axis*y_axis + z_axis*z_axis);
  if (axis_norm < 0.0000001)
  {
    x = 0;
    y = 0;
    z = 0;
  }
  else
  {
    double factor = half_theta_sin / axis_norm;
    x = x_axis * factor;
    y= y_axis * factor;
    z = z_axis * factor;
  }
}



vQuaternion2::vQuaternion2(double x_angle, double y_angle, double orientation)
{
  static long junk1 = vPrecomputeValues();
  double q[4];
  S_OldViewQuaternion2(x_angle, y_angle, orientation, q);
  cos_half_theta = q[0];
  x = q[1];
  y = q[2];
  z = q[3];
}


vQuaternion2::~vQuaternion2()
{
}


// Convert to the class vQuaternion
vQuaternion vQuaternion2::ToQuaternion()
{
  vQuaternion result;
  vArray(double) entries = result.Matrix();
  entries[0] = cos_half_theta;
  entries[1] = x;
  entries[2] = y;
  entries[3] = z;
  return result;
}


long vQuaternion2::Print()
{
  return Print("quaternion2");
}


long vQuaternion2::Print(const char * name)
{
  ToQuaternion().Print(name);
  return 1;
}


// Returns the quaternion multiplication of q1 and q2
vQuaternion2 vQuatMultiply2(vQuaternion2 & q1, vQuaternion2 & q2)
{
//  dot = q1_1 * q2_1 + q1_2 * q2_2 + q1_3 * q2_3;
//  cross_0 = q1_2 * q2_3 - q1_3 * q2_2;
//  cross_1 = q1_3 * q2_1 - q1_1 * q2_3;
//  cross_2 = q1_1 * q2_2 - q1_2 * q2_1;

//  q4_0 = q1_0 * q2_0 - dot;
//  q4_1 = q1_0 * q2_1 + q2_0 * q1_1 + cross_0;
//  q4_2 = q1_0 * q2_2 + q2_0 * q1_2 + cross_1;
//  q4_3 = q1_0 * q2_3 + q2_0 * q1_3 + cross_2;

  double s1 = q1.cos_half_theta;
  double s2 = q2.cos_half_theta;

  vQuaternion2 q;
  double q_0 = s1 * s2 - q1.x * q2.x - q1.y * q2.y - q1.z * q2.z;

  // Adjust for numerical errors.
  // This is really a bug. For example when the
  // multiplication happens with a point whose norm is over 1,
  // then q_0 should go outside the [-1,1] range, and this
  // adjustment will lead to a correct result. I need to adjust
  // for that. As an intermediate solution, I will force the program
  // to exit, so that I can see exactly when it is caused.
  if (q_0 > 1) 
  {
    vPrint("correcting q_0 = %lf for numerical errors\n");
    q1.Print("q1 = ");
    q2.Print("q2 = ");
    exit_error("Exiting the program\n");
    q_0 = 1;
  }

  if (q_0 < -1) 
  {
    vPrint("correcting q_0 = %lf for numerical errors\n");
    q1.Print("q1 = ");
    q2.Print("q2 = ");
    exit_error("Exiting the program\n");
    q_0 = -1;
  }

  q.cos_half_theta = q_0;
  
  double cross_0 = q1.y * q2.z - q1.z * q2.y;
  double cross_1 = q1.z * q2.x - q1.x * q2.z;
  double cross_2 = q1.x * q2.y - q1.y * q2.x;

  q.x = s1 * q2.x + s2 * q1.x + cross_0;
  q.y = s1 * q2.y + s2 * q1.y + cross_1;
  q.z = s1 * q2.z + s2 * q1.z + cross_2;
 
  
//  result(2:4) = s1 * v2 + s2 * v1 + cross(v1, v2);
//  vMatrix<double> v = s1 * v2 + s2 * v1 + vCross(v1, v2);
//  vArray(double) ventries = v.Matrix();
//  q.x = ventries[0];
//  q.y = ventries[1];
//  q.z = ventries[2];

  return q;
}


// The inverse of vViewQuaternion. It puts in x_angle_p,
// y_angle_p, z_angle_p, the camera viewpoint and orientation
// corresponding to the quaternion q. Angles are returned
// in degrees.
long vOldQuaternionView2(vQuaternion2 & q, double * x_angle_p,
                          double * y_angle_p, double * z_angle_p)
{
  static long junk = vPrecomputeValues();

  vQuaternion2 north_pole;
  north_pole.cos_half_theta = 0;
  north_pole.x = 0;
  north_pole.y = 1;
  north_pole.z = 0;

  // We want to figure out the camera viewpoint and orientation
  // that corresponds to q. The camera viewpoint can be obtained
  // by rotating the north pole using q.
  vQuaternion2 viewpoint = vQuatRotate2(north_pole, q);
  double camera_x = viewpoint.x;
  double camera_y = viewpoint.y;
  double camera_z = viewpoint.z;

  double temp = camera_x * camera_x + camera_y * camera_y + camera_z * camera_z;
  double distance_from_origin = sqrt(temp);

  if (distance_from_origin == 0) return 0;

  double temp_x, temp_y, temp_z;
  temp_x = camera_x / distance_from_origin;
  temp_y = camera_y / distance_from_origin;
  temp_z = camera_z / distance_from_origin;

  double x_angle, y_angle, z_angle;
    
  double non_y = sqrt(1 - temp_y * temp_y);
  if (non_y == 0)
  {
    x_angle = vPrecomputedAcos::Acos(temp_y) * 180.0 / M_PI;
    y_angle = 0;
  }
 
  else
  {
    double sin_y, cos_y, cos_x, rad_x, rad_y;
    cos_x = temp_y;
    sin_y = temp_x / non_y;
    cos_y = temp_z / non_y;
  
    // We have to take care of certain numerical errors, to make sure 
    // everything is well defined. At some point I got 
    // cos_y = -1 - 2.22e-16.

    if (cos_x > 1) cos_x = 1;
    if (cos_x < -1) cos_x = -1;
    if (sin_y > 1) sin_y = 1;
    if (sin_y < -1 ) sin_y = -1;
    if (cos_y > 1) cos_y = 1;
    if (cos_y < -1) cos_y = -1;

    rad_x = vPrecomputedAcos::Acos(cos_x);
    rad_y = vPrecomputedAcos::Acos(cos_y);
    assert(rad_y >= 0);
    if (sin_y < 0) rad_y = 2 * M_PI - rad_y;
    x_angle = rad_x * 180 / M_PI;
    y_angle = rad_y * 180 / M_PI;
  }

  vQuaternion2 viewpoint_quat(x_angle, y_angle, 0);
  vQuaternion2 viewpoint_to_q = vQuatMultiply2(q, vQuatInverse2(viewpoint_quat));

  double axis_dot_product = viewpoint.x * viewpoint_to_q.x;
  axis_dot_product += viewpoint.y * viewpoint_to_q.y;
  axis_dot_product += viewpoint.z * viewpoint_to_q.z;

  z_angle = viewpoint_to_q.Angle();
  if (axis_dot_product >= 0) z_angle = 360 - z_angle;
  if (z_angle >= 360) z_angle -= 360;

  *x_angle_p = x_angle;
  *y_angle_p = y_angle;
  *z_angle_p = z_angle;
  return 1;
}


// Rotates the point represented by quaternion point by applying the 
// rotation represented by quaternion q. The 3D point (x, y, z) 
// is represented as the quaternion (0, x, y, z);
vQuaternion2 vQuatRotate2(vQuaternion2 & point, vQuaternion2 & q)
{
  //  Matlab code:
  //  q_inverse = quat_inverse(q);
  //  result = quat_multiply(q, quat_multiply(point, q_inverse));
  vQuaternion2 q_inverse = vQuatInverse2(q);
  vQuaternion2 result = vQuatMultiply2(q, vQuatMultiply2(point, q_inverse));
  return result;
}


// Returns the inverse of q
vQuaternion2 vQuatInverse2(vQuaternion2 & q)
{
  vQuaternion2 result;
  result.cos_half_theta = q.cos_half_theta;
  result.x = -q.x;
  result.y = -q.y;
  result.z = -q.z;
  double norm = q.cos_half_theta * q.cos_half_theta + q.x * q.x + q.y * q.y + q.z * q.z;
  if (norm == 0)
  {
    exit_error("Inverse called on the zero quaternion\n");
  }
  result.cos_half_theta /= norm;
  result.x /= norm;
  result.y /= norm;
  result.z /= norm;
  return result;
}


// result = vViewDifference2(x1, y1, z1, x2, y2, z2)
// (x1, y1, z1) and (x2, y2, z2) express camera view parameters as
// used in my HandRendering program (x is latitude, y is longitude, z is 
// image plane rotation). The result returned here is the quaternion
// representing the rotation to go from (x1, y1, z1) to (x2, y2, z2)
vQuaternion2 vOldViewDifference2(double x1, double y1, double z1,
                                   double x2, double y2, double z2)
{
  //  Matlab code:
  //  q1 = view_quaternion(x1, y1, z1);
  //  q2 = view_quaternion(x2, y2, z2);
  //  result = quat_multiply(q2, quat_inverse(q1));
  vQuaternion2 q1(x1, y1, z1);
  vQuaternion2 q2(x2, y2, z2);
  vQuaternion2 result = vQuatMultiply2(q2, vQuatInverse2(q1));
  return result;
}


// This matrix represents the 3x3 matrix M that performs the same 3D rotation
// as the quaternion. Basically, 
// M * Transpose(x,y,z) = vQuatRotate((0,x,y,z), q).
vMatrix<double> vQuaternionMatrix(vQuaternion & q)
{
  double a = q(0);
  double b = q(1);
  double c = q(2);
  double d = q(3);

  double r11 = a*a + b*b - c*c - d*d;
  double r12 = 2*b*c - 2*a*d;
  double r13 = 2*a*c + 2*b*d;

  double r21 = 2*b*c + 2*a*d;
  double r22 = a*a - b*b + c*c - d*d;
  double r23 = 2*c*d - 2*a*b;
  
  double r31 = 2*b*d - 2*a*c;
  double r32 = 2*a*b + 2*c*d;
  double r33 = a*a - b*b - c*c + d*d;

  vMatrix<double> result(3,3);
  result.WriteN(9, r11, r12, r13, r21, r22, r23, r31, r32, r33);
  return result;
}


// This matrix represents the 3x3 projection matrix M that corresponds to the
// viewpoint specified by quaternion q, assuming orthographic projection. The
// viewpoint corresponding to frontal view (3D point 0,0, 1, facing towards
// increasing decreasing z) corresponds to the 
// identity projection matrix. On the other
// hand, the "identity" quaternion corresponds to the north pole viewpoint,
// as described elsewhere in the code.
vMatrix<double> vOldProjectionMatrix(vQuaternion & q)
{
  vMatrix<double> backprojection = vOldBackprojectionMatrix(q);
  vMatrix<double> result = backprojection.Inv();
  return result;
}


// This matrix represents the 3x3 backprojection matrix M that corresponds to 
// the viewpoint specified by quaternion q. M is used to backproject image pixels
// into 3D lines.
vMatrix<double> vOldBackprojectionMatrix(vQuaternion & q)
{
  // This matrix corresponds to the rotation that must be applied to 
  // the (90,0,0) viewpoint in order to make it equal to the viewpoint
  // represented by q.
  vQuaternion frontal = vOldViewQuaternion(90, 0, 0);
  vQuaternion inverse_frontal = vQuatInverse(frontal);
  vQuaternion result_quat = vQuatMultiply(q, inverse_frontal);
  vMatrix<double> result = vQuaternionMatrix(result_quat);
  return result;
}


// In vOldLineOrientation, the arguments describe the appearance of the 3D
// line from two viewpoints. The appearance is described as an angle,
// in radians. It is the angle that the line makes with the image x axis.
vQuaternion vOldLineOrientation(vQuaternion & viewpoint1, double angle1,
                               vQuaternion & viewpoint2, double angle2)
{
  ushort error_flag = 0;

  vMatrix<double> rotation1 = vOldBackprojectionMatrix(viewpoint1);
  vMatrix<double> rotation2 = vOldBackprojectionMatrix(viewpoint2);
  rotation1.Print("rotation 1:");
  rotation2.Print("rotation 2:");

  v3D_Plane plane1 = vProjectionPlane(rotation1, angle1);
  plane1.Print("plane 1:\n");

  v3D_Plane plane2 = vProjectionPlane(rotation2, angle2);
  plane2.Print("plane 2:\n");

  v3D_Line line = vPlaneIntersection(plane1, plane2, &error_flag);
  vPrint("error_flag = %li\n", error_flag);
  line.Print("line: ");

  vPoint3 viewpoint = vUnitSphereIntersection(line);
  vPoint3 angles = vOld3DToSphereSurface(viewpoint.x, viewpoint.y, viewpoint.z);
  vQuaternion result = vOldViewQuaternion(angles.x, angles.y, angles.z);
  return result;
}


// Returns the plane that projects onto a line going through the
// origin with orientation angle1 under the given viewoint. Note
// that p is simply the rotation matrix that represents the 
// transformation that is needed to move the camera from the
// canonical (0,0,0) viewpoint to its current viewpoint. The 
// matrix p is the inverse of the projection matrix that should 
// be applied to points in the 3D world in order to project them
// to the corresponding image plane. Therefore, p is the right
// matrix to multiply with if we want to backproject image points
// into 3D lines, or image lines into 3D planes.
v3D_Plane vProjectionPlane(vMatrix<double> p, double angle1)
{
  double a = p(0,0);
  double b = p(0,1);
  double c = p(0,2);
  double d = p(1,0);
  double e = p(1,1);
  double f = p(1,2);
  double g = p(2,0);
  double h = p(2,1);
  double i = p(2,2);

  v3D_Plane plane;

  // Get the slope of the angle.
  double y_slope = vPrecomputedTan::Tan(angle1);

  // We must avoid the pathological case where angle1 = (2k+1) * pi.
  if (vAbs(y_slope) <= 1)
  {
    // Mathematica gives the following equation for the plane that projects
    // to the line with angle1, where sx is the slope.
    // {{a x + b sx x + c z}, 
    //  {d x + e sx x + f z}, 
    //  {g x + h sx x + i z}}
    plane.xs_factor = a + b * y_slope;
    plane.ys_factor = d + e * y_slope;
    plane.zs_factor = g + h * y_slope;
    plane.xt_factor = c;
    plane.yt_factor = f;
    plane.zt_factor = i;
  }

  else
  {
    double x_coord = vPrecomputedCos::Cos(angle1);
    double y_coord = vPrecomputedSin::Sin(angle1);
    double x_slope = x_coord / y_coord;

    plane.xs_factor = a * x_slope + b;
    plane.ys_factor = d * x_slope + e;
    plane.zs_factor = g * x_slope + h;
    plane.xt_factor = c;
    plane.yt_factor = f;
    plane.zt_factor = i;
  }

  return plane;
}


// Returns the intersection between two 3D planes, which, of course, 
// is a line (or empty). The solutions (and exceptions to the solutions)
// were found using Mathematica. error_flag, if non-zero, receives
// a value that indicates whether the function was successful in 
// solving the equation. The function can fail if the intersection
// is not a line (for example if the two planes are identical), or
// if certain numerical conditions are not satisfied (the formulas
// involve divisions, and I haven't handled all cases where the divisor 
// may be 0). If the function succeeds, error_flag gets the value 0.
v3D_Line vPlaneIntersection(v3D_Plane plane1, 
                              v3D_Plane plane2,
                              ushort * error_flag)
{
  double xs1 = plane1.xs_factor;
  double xt1 = plane1.xt_factor;
  double ys1 = plane1.ys_factor;
  double yt1 = plane1.yt_factor;
  double zs1 = plane1.zs_factor;
  double zt1 = plane1.zt_factor;

  double xs2 = plane2.xs_factor;
  double xt2 = plane2.xt_factor;
  double ys2 = plane2.ys_factor;
  double yt2 = plane2.yt_factor;
  double zs2 = plane2.zs_factor;
  double zt2 = plane2.zt_factor;

  v3D_Line result = v3D_Line();
  double numerator, denom, term1, term2;
  if (error_flag != 0) *error_flag = 0;

  // trying to parameterize based on s1:
  denom = xt2*yt1*zs2 - xt1*yt2*zs2 - xt2*ys2*zt1 + xs2*yt2*zt1 + xt1*ys2*zt2 - xs2*yt1*zt2;
  if (vAbs(denom) >= 0.01)
  {
    // Mathematica gives for x_factor the numerator 
    // -(xt1 (xt2 ys2 zs1 - xs2 yt2 zs1 - xt2 ys1 zs2 + xs2 ys1 zt2) + 
    //   xs1 (xt2 yt1 zs2 - xt2 ys2 zt1 + xs2 yt2 zt1 - xs2 yt1 zt2))
    term1 = -xt1 * (xt2*ys2*zs1 - xs2*yt2*zs1 - xt2*ys1*zs2 + xs2*ys1*zt2);
    term2 = -xs1 * (xt2*yt1*zs2 - xt2*ys2*zt1 + xs2*yt2*zt1 - xs2*yt1*zt2);
    numerator = term1 + term2;
    result.x_factor = numerator / -denom;

    // Mathematica gives for y_factor the numerator 
    // xs2 yt2 (yt1 zs1 - ys1 zt1) + xt2 ys2 (-(yt1 zs1) + ys1 zt1) + 
    // (xt1 ys1 - xs1 yt1) (yt2 zs2 - ys2 zt2)
    term1 = xs2*yt2 * (yt1*zs1 - ys1*zt1) + xt2*ys2*(-(yt1*zs1) + ys1*zt1);
    term2 = (xt1*ys1 - xs1*yt1) * (yt2*zs2 - ys2*zt2);
    numerator = term1 + term2;
    result.y_factor = numerator / -denom;

    // Mathematica gives for z_factor the numerator 
    // -(xt1 yt2 zs1 zs2) + xs1 yt2 zs2 zt1 + xt2 zs2 (yt1 zs1 - ys1 zt1) + 
    // xt1 ys2 zs1 zt2 - xs2 yt1 zs1 zt2 + xs2 ys1 zt1 zt2 - xs1 ys2 zt1 zt2
    term1 = -(xt1*yt2*zs1*zs2) + xs1*yt2*zs2*zt1 + xt2*zs2 * (yt1*zs1 - ys1*zt1);
    term2 = xt1*ys2*zs1*zt2 - xs2*yt1*zs1*zt2 + xs2*ys1*zt1*zt2 - xs1*ys2*zt1*zt2;
    numerator = term1 + term2;
    result.z_factor = numerator / denom;
    result.Normalize();
    return result;  
  }

  // trying to parameterize based on t1:
  denom = xt2*ys2*zs1 - xs2*yt2*zs1 - xt2*ys1*zs2 + xs1*yt2*zs2 + xs2*ys1*zt2 - xs1*ys2*zt2;
  if (vAbs(denom) >= 0.01)
  {
    // Mathematica gives for x_factor the numerator 
    // xt1 (xt2 ys2 zs1 - xs2 yt2 zs1 - xt2 ys1 zs2 + xs2 ys1 zt2) + 
    // xs1 (xt2 yt1 zs2 - xt2 ys2 zt1 + xs2 yt2 zt1 - xs2 yt1 zt2)   
    term1 = xt1 * (xt2*ys2*zs1 - xs2*yt2*zs1 - xt2*ys1*zs2 + xs2*ys1*zt2) ;
    term2 = xs1 * (xt2*yt1*zs2 - xt2*ys2*zt1 + xs2*yt2*zt1 - xs2*yt1*zt2);
    numerator = term1 + term2;
    result.x_factor = numerator / denom;

    // Mathematica gives for y_factor the numerator 
    // xt2 ys2 (yt1 zs1 - ys1 zt1) + xs2 (-(yt1 yt2 zs1) + ys1 yt2 zt1) - 
    // (xt1 ys1 - xs1 yt1) (yt2 zs2 - ys2 zt2)   
    term1 = xt2*ys2 * (yt1*zs1 - ys1*zt1) + xs2 * (-(yt1*yt2*zs1) + ys1*yt2*zt1);
    term2 = -(xt1*ys1 - xs1*yt1) * (yt2*zs2 - ys2*zt2);
    numerator = term1 + term2;
    result.y_factor = numerator / denom;

    // Mathematica gives for z_factor the numerator 
    // -(xt1 yt2 zs1 zs2) + xs1 yt2 zs2 zt1 + xt2 zs2 (yt1 zs1 - ys1 zt1) 
    // + xt1 ys2 zs1 zt2 - xs2 yt1 zs1 zt2 + xs2 ys1 zt1 zt2 - xs1 ys2 zt1 zt2 
    term1 = -(xt1*yt2*zs1*zs2) + xs1*yt2*zs2*zt1 + xt2*zs2 * (yt1*zs1 - ys1*zt1);
    term2 = xt1*ys2*zs1*zt2 - xs2*yt1*zs1*zt2 + xs2*ys1*zt1*zt2 - xs1*ys2*zt1*zt2;
    numerator = term1 + term2;
    result.z_factor = numerator / denom;
    result.Normalize();
    return result;  
  }

  // If we get here, it means we failed to find a valid intersection.
  if (error_flag != 0) *error_flag = 1;
  return v3D_Line();
}


// Returns the intersection between two 3D planes, which, of course, 
// is a line (or empty).
// An older version. The newer version handles more cases.
v3D_Line vPlaneIntersection2(v3D_Plane plane1, 
                               v3D_Plane plane2, 
                               ushort * error_flag)
{
  // The result from mathematica is:
  // (-(t xt2 yt1 zs2) + t xt1 yt2 zs2 + t xt2 ys2 zt1 - t xs2 yt2 zt1 - 
  //    t xt1 ys2 zt2 + t xs2 yt1 zt2) / 
  // (-(xt2 ys2 zs1) + xs2 yt2 zs1 + xt2 ys1 zs2 - xs1 yt2 zs2 - 
  //    xs2 ys1 zt2 + xs1 ys2 zt2)  
  double xs1 = plane1.xs_factor;
  double xt1 = plane1.xt_factor;
  double ys1 = plane1.ys_factor;
  double yt1 = plane1.yt_factor;
  double zs1 = plane1.zs_factor;
  double zt1 = plane1.zt_factor;

  double xs2 = plane2.xs_factor;
  double xt2 = plane2.xt_factor;
  double ys2 = plane2.ys_factor;
  double yt2 = plane2.yt_factor;
  double zs2 = plane2.zs_factor;
  double zt2 = plane2.zt_factor;

  v3D_Line result = v3D_Line();
  double s_numerator = -xt2*yt1*zs2 + xt1*yt2*zs2 + xt2*ys2*zt1 - xs2*yt2*zt1 - xt1*ys2*zt2 + xs2*yt1*zt2;
  double s_denominator = -xt2*ys2*zs1 + xs2*yt2*zs1 + xt2*ys1*zs2 - xs1*yt2*zs2 - xs2*ys1*zt2 + xs1*ys2*zt2;
  if (vAbs(s_denominator) < 0.01)
  {
    if (error_flag != 0) *error_flag = 1;
    return v3D_Line();
  }

  double s_term = s_numerator / s_denominator;
  result.x_factor = xs1 * s_term + xt1;
  result.y_factor = ys1 * s_term + yt1;
  result.z_factor = zs1 * s_term + zt1;
  if (error_flag != 0) *error_flag = 0;

  result.Normalize();
  return result;  
}


// Returns the point at which a line intersects the unit sphere.
// It is assumed that the line goes through the origin.
vPoint3 vUnitSphereIntersection(v3D_Line line)
{
  return vPoint3(1, 1, 1);
}


long v3D_Line::Print(const char * str)
{
  vPrint(str);
  vPrint("(x, y, z) = (%.3lf t, %.3lf t, %.3lf t)\n", 
          x_factor, y_factor, z_factor);

  return 1;
}


void v3D_Line::Normalize()
{
  double square_norm = x_factor * x_factor + y_factor * y_factor + z_factor * z_factor;
  double norm = sqrt(square_norm);
  x_factor = x_factor / norm;
  y_factor = y_factor / norm;
  z_factor = z_factor / norm;
}


long v3D_Plane::Print(const char * str)
{
  vPrint(str);
  vPrint("x = %.3lf s + %.3lf t\n", xs_factor, xt_factor);
  vPrint("y = %.3lf s + %.3lf t\n", ys_factor, yt_factor);
  vPrint("z = %.3lf s + %.3lf t\n", zs_factor, zt_factor);

  return 1;
}


// vVerifyPlaneIntersection is used to verify that 
// vPlaneIntersection is doing the right thing.
long vOldVerifyPlaneIntersection(vQuaternion viewpoint1, double angle1,
                                  vQuaternion viewpoint2, double angle2)
{
  ushort error_flag;

  vMatrix<double> rotation1 = vOldBackprojectionMatrix(viewpoint1);
  v3D_Plane plane1 = vProjectionPlane(rotation1, angle1);
  plane1.Print("plane1:\n");

  vMatrix<double> rotation2 = vOldBackprojectionMatrix(viewpoint2);
  v3D_Plane plane2 = vProjectionPlane(rotation2, angle2);
  plane2.Print("plane2:\n");

  v3D_Line intersection = vPlaneIntersection(plane1, plane2, &error_flag);
  vPrint("error_flag = %li\n", error_flag);
  intersection.Print("intersection:");

  // Compute the forward projection matrices:
  vMatrix<double> projection1 = rotation1.Inv();
  vMatrix<double> projection2 = rotation2.Inv();

  // Ask the user to input a point on the intersection, by 
  // specifying its distance from the origin. Then, we will
  // compute the 3D coordinates of that point, its projections 
  // on the two images, and the slopes corresponding to those
  // projections.
  while(1)
  {
    vPrint("\nInput a distance from origin, -1 to exit:\n");
    double distance = 0;
    vScan("%lf", &distance);
    if (distance == -1) break;
    vPoint3 point = intersection.Point(distance);
    point.Print("3D point = ");
    // Convert the point to a matrix representation.
    vMatrix<double> point_matrix(3,1);
    point_matrix.WriteN(3, point.x, point.y, point.z);
    vMatrix<double> pixel1 = projection1 * point_matrix;
    vMatrix<double> pixel2 = projection2 * point_matrix;
    pixel1.Print("pixel1: ");
    pixel2.Print("pixel2: ");

    // Figure out the slopes corresponding to the two pixels.
    double pixel_slope1, pixel_slope2;
    if (pixel1(0) == 0.0) pixel_slope1 = 0;
    else pixel_slope1 = pixel1(1) / pixel1(0);
    if (pixel2(0) == 0.0) pixel_slope2 = 0;
    else pixel_slope2 = pixel2(1) / pixel2(0);
    vPrint("pixel_slope1 = %.3lf, pixel_slope2 = %.3lf\n",
            pixel_slope1, pixel_slope2);

    // Figure out the angles corresponding to the two slopes.
    double pixel_angle1 = vAngle(pixel1(0), pixel1(1)) * 180 / M_PI;
    double pixel_angle2 = vAngle(pixel2(0), pixel2(1)) * 180 / M_PI;
    vPrint("pixel_angle1 = %.3lf, pixel_angle2 = %.3lf\n",
            pixel_angle1, pixel_angle2);
  }

  return 1;
}


// vVerifyPlaneIntersection2 is used to verify that 
// vPlaneIntersection2 (an obsolete version of vPlaneIntersection)
// is doing the right thing, and to compare its results with those
// of vVerifyPlaneIntersection.
long vOldVerifyPlaneIntersection2(vQuaternion viewpoint1, double angle1,
                                   vQuaternion viewpoint2, double angle2)
{
  ushort error_flag;

  vMatrix<double> rotation1 = vOldBackprojectionMatrix(viewpoint1);
  v3D_Plane plane1 = vProjectionPlane(rotation1, angle1);
  plane1.Print("plane1:\n");

  vMatrix<double> rotation2 = vOldBackprojectionMatrix(viewpoint2);
  v3D_Plane plane2 = vProjectionPlane(rotation2, angle2);
  plane2.Print("plane2:\n");

  v3D_Line intersection = vPlaneIntersection2(plane1, plane2, &error_flag);
  vPrint("error_flag = %li\n", error_flag);
  intersection.Print("intersection:");

  // Compute the forward projection matrices:
  vMatrix<double> projection1 = rotation1.Inv();
  vMatrix<double> projection2 = rotation2.Inv();

  // Ask the user to input a point on the intersection, by 
  // specifying its distance from the origin. Then, we will
  // compute the 3D coordinates of that point, its projections 
  // on the two images, and the slopes corresponding to those
  // projections.
  while(1)
  {
    vPrint("\nInput a distance from origin, -1 to exit:\n");
    double distance = 0;
    vScan("%lf", &distance);
    if (distance == -1) break;
    vPoint3 point = intersection.Point(distance);
    point.Print("3D point = ");
    // Convert the point to a matrix representation.
    vMatrix<double> point_matrix(3,1);
    point_matrix.WriteN(3, point.x, point.y, point.z);
    vMatrix<double> pixel1 = projection1 * point_matrix;
    vMatrix<double> pixel2 = projection2 * point_matrix;
    pixel1.Print("pixel1: ");
    pixel2.Print("pixel2: ");

    // Figure out the slopes corresponding to the two pixels.
    double pixel_slope1, pixel_slope2;
    if (pixel1(0) == 0.0) pixel_slope1 = 0;
    else pixel_slope1 = pixel1(1) / pixel1(0);
    if (pixel2(0) == 0.0) pixel_slope2 = 0;
    else pixel_slope2 = pixel2(1) / pixel2(0);
    vPrint("pixel_slope1 = %.3lf, pixel_slope2 = %.3lf\n",
            pixel_slope1, pixel_slope2);

    // Figure out the angles corresponding to the two slopes.
    double pixel_angle1 = vAngle(pixel1(0), pixel1(1)) * 180 / M_PI;
    double pixel_angle2 = vAngle(pixel2(0), pixel2(1)) * 180 / M_PI;
    vPrint("pixel_angle1 = %.3lf, pixel_angle2 = %.3lf\n",
            pixel_angle1, pixel_angle2);
  }

  return 1;
}
