#ifndef VASSILIS_QUATERNION_H
#define VASSILIS_QUATERNION_H

#include "basics/algebra.h"
#include "gestures/algorithms.h"
#include "basics/precomputed.h"

class vQuaternion : public vMatrix<double>
{
public:
  vQuaternion ();
  ~vQuaternion();
};


// class vQuaternion2 is used when we care more for speed and 
// less for elegance of code. It has the same functionality
// as vQuaternion. Of course, functions that work with
// vQuaternion have to be rewritten for vQuaternion2
//
// A quaternion represents a rotation about the axis connecting the 
// origin to (x, y, z) by an amount of theta.
class vQuaternion2 : public vPoint3
{
public:
  double cos_half_theta; // cos(theta/2)

public:
  vQuaternion2() 
  {
    static long junk = vPrecomputeValues();
  }

  vQuaternion2(double theta, double x_axis, double y_axis, double z_axis);
  vQuaternion2(double theta, vPoint3 & point);

  // Same as vViewQuaternion
  vQuaternion2(double x_angle, double y_angle, double orientation);
  ~vQuaternion2();

  inline double CosHalfTheta()
  {
    return cos_half_theta;
  }

  inline double X()
  {
    return x;
  }

  inline double Y()
  {
    return y;
  }

  inline double Z()
  {
    return z;
  }

  inline double Angle()
  {
    double half_theta = vPrecomputedAcos::Acos(cos_half_theta);
    double theta = 2 * half_theta;
    double result = theta * 180 / M_PI;
    if (result >= 360.0) result -= 360.0;
    return result;
  }

  inline vPoint3 Axis()
  {
    vPoint3 result;
    double half_theta = vPrecomputedAcos::Acos(cos_half_theta);
    double half_theta_sin = vPrecomputedSin::Sin(half_theta);

    // Take care of numerical errors
    if (vAbs(half_theta_sin) < 0.000001)
    {
      result.x = 0;
      result.y = 0;
      result.z = 0;
      return result;
    }

    result.x = x / half_theta_sin;
    result.y = y / half_theta_sin;
    result.z = z / half_theta_sin;
    return result;
  }

  // Convert to the class vQuaternion
  vQuaternion ToQuaternion();

  long Print();
  long Print(const char * name);
};



// vAngleQuaternion(theta, x, y, z) returns the quaternion that
// corresponds to rotating around axis connecting 0 to x, y, z by 
// angle theta.
vQuaternion vAngleQuaternion(double theta, double x, double y, double z);

// result = vQuatAngle(q) returns the angle theta such that q represents
// counterclockwise rotation around some axis by angle theta. The 
// result is in degrees
double vQuatAngle(vQuaternion & q);

// result = quat_axis(q)
// Returns the axis of rotation for the rotation that q represents
vMatrix<double> vQuatAxis(vQuaternion & q);

// Returns the inverse of q
vQuaternion vQuatInverse(vQuaternion & q);

// Returns the quaternion multiplication of q1 and q2
vQuaternion vQuatMultiply(vQuaternion & q1, vQuaternion & q2);

// Rotates the point represented by quaternion point by applying the 
// rotation represented by quaternion q. The 3D point (x, y, z) 
// is represented as the quaternion (0, x, y, z);
vQuaternion vQuatRotate(vQuaternion & point, vQuaternion & q);

// result = vViewDifference(x1, y1, z1, x2, y2, z2)
// (x1, y1, z1) and (x2, y2, z2) express camera view parameters as
// used in my HandRendering program (x is latitude, y is longitude, z is 
// image plane rotation). The result returned here is the quaternion
// representing the rotation to go from (x1, y1, z1) to (x2, y2, z2)
vQuaternion vOldViewDifference(double x1, double y1, double z1,
                                 double x2, double y2, double z2);

// result = vViewDistance(x1, y1, z1, x2, y2, z2) computes the angle
// distance between viewpoints (x1, y1, z1) and (x2, y2, z2) (as
// used in my HandRendering program). The arguments and result are 
// in degrees. 
double vOldViewDistance(double x1, double y1, double z1,
                      double x2, double y2, double z2);

// A heavily optimized version of vViewDistance, to the point that it's
// very hard to read.
double vOldViewDistance2(double x1, double y1, double z1,
                       double x2, double y2, double z2);

// result = vViewQuaternion(x_angle, y_angle, orientation) returns 
// the quaternion that performs the rotation that moves the camera viewpoint
// and orientation from the north pole (0, 0, 0) to latitude x_angle,
// longitude y_angle and image plane rotation orientation. This is
// applicable to the camera viewpoint used by my HandRendering program.
// Angle values are expressed in degrees.
vQuaternion vOldViewQuaternion(double x_angle, double y_angle, 
                              double orientation);

// Make degrees to be between 0 and 360
inline double vAdjustAngle(double degrees)
{
  if (degrees < 0) 
  {
    long times = (long) (-degrees/360 + 1);
    degrees += times * 360;
  }
  else if (degrees >= 360)
  {
    long times = (long) (degrees/360);
    degrees -= times * 360;
  }
  return degrees;
}


// The vViewDistanceTimer functions compute lots of quaternion distances
// in different ways, and can be used to evaluate the efficiency of each way.
double vOldViewDistanceTimer1(long times);
double vOldViewDistanceTimer2(long times);
double vOldViewDistanceTimer3(long times);

// This function computes the 3D position of a point on the surface
// of the unit sphere specified by x_angle and y_angle, i.e. 
// x_angle degrees south of the north pole and y_angle degrees east
// of the x=0 plane.
vPoint3 vOldSphereSurfaceTo3D(double x_angle, double y_angle);

// This is the inverse of vSphereSurfacePoint.
vPoint3 vOld3DToSphereSurface(double camera_x, double camera_y, 
                             double camera_z);

// vQuaternionDistance returns the absolute value of the angle
// of the smallest rotation that converts q1 into q2.
double vQuaternionDistance(vQuaternion & q1, vQuaternion & q2);

// Computes the quaternion that rotates by theta degrees 
// around the axis defined
// by connecting the center of the unit sphere to the surface point
// on the sphere with coordinates (x,y) (x degrees south of the north 
// pole, y degrees east)
vQuaternion vOldRotationQuaternion(double theta, double x, double y);

// Computes the distance between the quaternions defined by
// these two sets of parameters.
double vOldRotationDistance(double theta1, double x1, double y1,
                          double theta2, double x2, double y2);

// The inverse of vViewQuaternion. It puts in x_angle_p,
// y_angle_p, z_angle_p, the camera viewpoint and orientation
// corresponding to the quaternion q. Angles are returned
// in degrees.
long vOldQuaternionView(vQuaternion & q, double * x_angle_p,
                      double * y_angle_p, double * z_angle_p);

// Returns the quaternion multiplication of q1 and q2
vQuaternion2 vQuatMultiply2(vQuaternion2 & q1, vQuaternion2 & q2);

// The inverse of vViewQuaternion. It puts in x_angle_p,
// y_angle_p, z_angle_p, the camera viewpoint and orientation
// corresponding to the quaternion q. Angles are returned
// in degrees.
long vOldQuaternionView2(vQuaternion2 & q, double * x_angle_p,
                       double * y_angle_p, double * z_angle_p);

// Rotates the point represented by quaternion point by applying the 
// rotation represented by quaternion q. The 3D point (x, y, z) 
// is represented as the quaternion (0, x, y, z);
vQuaternion2 vQuatRotate2(vQuaternion2 & point, vQuaternion2 & q);

// Returns the inverse of q
vQuaternion2 vQuatInverse2(vQuaternion2 & q);

// result = vViewDifference2(x1, y1, z1, x2, y2, z2)
// (x1, y1, z1) and (x2, y2, z2) express camera view parameters as
// used in my HandRendering program (x is latitude, y is longitude, z is 
// image plane rotation). The result returned here is the quaternion
// representing the rotation to go from (x1, y1, z1) to (x2, y2, z2)
vQuaternion2 vOldViewDifference2(double x1, double y1, double z1,
                                double x2, double y2, double z2);

// This matrix represents the 3x3 matrix M that performs the same 3D rotation
// as the quaternion. Basically, 
// M * Transpose(x,y,z) = vQuatRotate((0,x,y,z), q).
vMatrix<double> vQuaternionMatrix(vQuaternion & q);

// This matrix represents the 3x3 projection matrix M that corresponds to the
// viewpoint specified by quaternion q, assuming orthographic projection. The
// viewpoint corresponding to frontal view (3D point 0,0, -1, facing towards
// increasing z) corresponds to the identity projection matrix. On the other
// hand, the "identity" quaternion corresponds to the north pole viewpoint,
// as described elsewhere in the code.
vMatrix<double> vOldProjectionMatrix(vQuaternion & q);

// This matrix represents the 3x3 backprojection matrix M that corresponds to 
// the viewpoint specified by quaternion q. M is used to backproject image pixels
// into 3D lines.
vMatrix<double> vOldBackprojectionMatrix(vQuaternion & q);


// The next few functions deal with the problem of identifying the 
// 3D orientation of an object, i.e. the camera rotation that
// will change the appearance of the object from a "canonical view"
// to what we actually observe. I want to solve this problem for
// the particular case of hand links, as observed using Poser. In
// that case, it is easy to observe the 2D projection of the x, y 
// and z axis of a link from different viewponts. The "canonical"
// view for a link has its x axis parallel to the world (and camera) x axis,
// and the y axis parallel to the world (and camera) y axis, whereas the
// z axis of the object projects to a single point in the image. Note
// that we assume orthographic projection.

// A subproblem of determining 3D orientation for an object is 
// determining 3D orientation for a line, from two calibrated
// views of that line, assuming orthographic projection. The 3D
// orientation of a line is defined to be the rotation (or, equivalently,
// the camera viewpoint) that makes the world x axis parallel to that line.

// In vOldLineOrientation, the arguments describe the appearance of the 3D
// line from two viewpoints. The appearance is described as an angle,
// in radians.
vQuaternion vOldLineOrientation(vQuaternion & viewpoint1, double angle1,
                                  vQuaternion & viewpoint2, double angle2);

// Represents the equation of a line in 3D. That equation can be given 
// by specifying each of the coordinates as a function of a parameter t.
// It is assumed that the line goes through the origin.
class v3D_Line
{
public:
  double x_factor, y_factor, z_factor;

  v3D_Line()
  {}

  // Get the point at position t on the line.
  inline vPoint3 Point(double t)
  {
    return vPoint3(x_factor * t, y_factor * t, z_factor * t);
  }

  long Print(const char * str);

  // Normalize the factors, so that t corresponds to distance from 
  // origin.
  void Normalize();
};


// Represents the equation of a plane in 3D. The equation can be iven
// by specifying each of the coordinates as a function of two variables s 
// and t. It is assumed that the plane goes through the origin.
class v3D_Plane
{
public:
  double xs_factor, xt_factor;
  double ys_factor, yt_factor;
  double zs_factor, zt_factor;

  v3D_Plane() 
  {}

  // Returns the point on the plane that is parameterized by s and t.
  inline vPoint3 Point(double s, double t)
  {
    double x = xs_factor * s + xt_factor * t;
    double y = ys_factor * s + yt_factor * t;
    double z = zs_factor * s + zt_factor * t;
  }

  long Print(const char * str);
};


// Returns the plane that projects onto a line going through the
// origin with orientation angle1 under the given viewoint.
v3D_Plane vProjectionPlane(vMatrix<double> viewpoint1, double angle1);

// Returns the intersection between two 3D planes, which, of course, 
// is a line (or empty).
v3D_Line vPlaneIntersection(v3D_Plane plane1, 
                              v3D_Plane plane2,
                              ushort * error_flag);

// Returns the point at which a line intersects the unit sphere.
// It is assumed that the line goes through the origin.
vPoint3 vUnitSphereIntersection(v3D_Line line);


// vVerifyPlaneIntersection is used to verify that 
// vPlaneIntersection is doing the right thing.
long vOldVerifyPlaneIntersection(vQuaternion viewpoint1, double angle1,
                                  vQuaternion viewpoint2, double angle2);

// vVerifyPlaneIntersection2 is used to verify that 
// vPlaneIntersection2 (an obsolete version of vPlaneIntersection)
// is doing the right thing, and to compare its results with those
// of vVerifyPlaneIntersection.
long vOldVerifyPlaneIntersection2(vQuaternion viewpoint1, double angle1,
                                   vQuaternion viewpoint2, double angle2);


#endif // VASSILIS_QUATERNION_H
