#include "stdafx.h"

#include "ranking.h"
#include "algo_templates.h"
#include "basics/definitions.h"
#include "basics/video_common.h"
#include "gesture_data.h"
#include "virtual_hand/FrameViewFinder.h"
#include "basics/drawing_temp.h"
#include "opencv_utils.h"
#include "quaternion.h"

#include "basics/definitions.h"

using namespace std;


// The format of the ground truth file should be:
// 0 LINE
// 1 LINE
// ...
// N LINE
//
// where LINE is
// training_pose_index x_angle y_angle orientation
//
// training_pose is long. x_angle, y_angle and orientation
// are double (and measured in degrees).
vector<vViewGroundTruth> * vViewGroundTruth::Load(const char * filename)
{
  FILE * fp = fopen(filename, vFOPEN_READ);
  if (fp == 0) 
  {
    vPrint("No ground truth file %s found\n", filename);
    return 0;
  }

  ushort success = 1;
  long frame = 0;
  vector<vViewGroundTruth> * entries = new vector<vViewGroundTruth>;
  while(1)
  {
    vViewGroundTruth entry;
    long items = fscanf(fp, "%li %li %lf %lf %lf\n",
                        &(entry.test_frame_index),
                        &(entry.training_pose_index),
                        &(entry.x_angle), &(entry.y_angle),
                        &(entry.orientation));
    // If items <= 0 then we have probably reached the end of file, we
    // have been successful in reading the file to the best of 
    // our knowledge (I could make more rigorous tests but it is not
    // worth it).
    if (items <= 0)
    {
      break;
    }

    else if (items != 5)
    {
      vPrint("Error in reading ground truth file %s\n", filename);
      success = 0;
      break;
    }

    if (entry.test_frame_index != frame)
    {
      vPrint("Error, apparently bad frame index in ground truth file %s\n",
              filename);
      success = 0;
      break;
    }
    frame++;
    entries->push_back(entry);
  }

  fclose(fp);
  if (success == 0)
  {
    return 0;
  }
  else
  {
    return entries;
  }
}


// For each ground truth in the vector it saves its distances
// from each of the about 4000 possible camera view parameters.
long vViewGroundTruth::SaveGroundTruthDistances(vector<vViewGroundTruth> * gts,
                                                 const char * ground_truth_file)
{
  char * output_file = vReplaceEnding2(ground_truth_file, ".gt_distances");
  vFrameViewFinder finder;
  long number_of_views = finder.FramesPerPose();
  long scores_per_frame = finder.ScoresPerPose();
  long number_of_gts = gts->size();
  long view;
  long test;
  vMatrix<ushort> distances_matrix(number_of_gts, scores_per_frame);
  vArray2(ushort) distances = distances_matrix.Matrix2();
  for (test = 0; test < number_of_gts; test++)
  {
    vViewGroundTruth & gt = (*gts)[test];
    double x = gt.x_angle;
    double y = gt.y_angle;
    double z = gt.orientation;

    for (view = 0; view < number_of_views; view++)
    {
      long rotation;
      for (rotation = 0; rotation < 4; rotation++)
      {
 //   double distance = 
 //     finder.Distance(training_frame, 4 - rotation,
 //                                     x_angle, y_angle, orientation);
        double distance = finder.Distance(view, rotation, x, y, z);
        distances[test][view*4+rotation] = (ushort) (distance * 10);
      }
    }
    vPrint("Processed frame %li of %li\r", test, number_of_gts);
  }
  vPrint("\n");
  long result = distances_matrix.Write(output_file);
  if (result <= 0)
  {
    vPrint("Failed to save distances to %s.\n", output_file);
  }
  vdelete2(output_file);
  return result;
}

  
vMatchScore::vMatchScore()
{
  is_valid = 0;
  ground_distance = (double) -1;
  score = -1;
  frame = -1;
  view = -1;
  rank = -1;
}


vMatchScore::vMatchScore(double in_score, long in_frame, long in_view)
{
  score = (float) in_score;
  frame = in_frame;
  view = in_view;
  is_valid = 1;
  ground_distance = -1;
}


void vMatchScore::Print()
{
  vPrint("score = %.12lf, distance = %.1lf, frame = %li, view = %li, rank = %li\r\n",
          score, ground_distance, frame, view, rank);
}


string vMatchScore::PrintToString()
{
  char buffer[400];
  
  sprintf(buffer, 
          "score = %.3lf, distance = %.1lf, frame = %li, view = %li, rank = %li\r\n",
          score, ground_distance, frame, view, rank);
  string result = buffer;
  return result;
}

// The reason I have a second print function is that, when I print results 
// from sequence_match_stats and I save them and view them using Emacs, 
// I want a quick way to search for correct matches with given ranks and
// scores. If correct matches and top matches are printed the same, and
// I search, say, for "rank = 0", it will find all top matches of rank 0,
// and it will take a while to find a correct match of rank 0. This way,
// I can print correct matches with PrintToString() and top matches with 
// PrintToString2().
// If I search for "rank = 0" I only find correct matches of rank 0,
// and if I search for "score: 0." I only find top match scores that are
// below 1.
string vMatchScore::PrintToString2()
{
  char buffer[400];
  
  sprintf(buffer, 
          "score: %.3lf, distance: %.1lf, frame: %li, view: %li, rank: %li\r\n",
          score, ground_distance, frame, view, rank);
  string result = buffer;
  return result;
}

ushort vMatchScore::Fprint(FILE * fp)
{
  if (fp == stdout)
  {
    Print();
    return 1;
  }

  long items = fprintf(fp, 
                       "score = %.3lf, distance = %.1lf, frame = %li, view = %li, rank = %li\r\n",
                       score, ground_distance, frame, view, rank);
  if (items < 0) 
  {
    return 0;
  }
  else return 1;
}


long vMatchScore::Save(FILE * fp)
{
  long items = 0;
  double d_score = score;
  items = vWritePCDoubles(fp, &d_score, 1);
  items += vWritePCLongs(fp, &frame, 1);
  items += vWritePCLongs(fp, &view, 1);
  double d_ground_distance = ground_distance;
  items += vWritePCDoubles(fp, &d_ground_distance, 1);
  items += vWritePCLongs(fp, &rank, 1);

  if (items != 5) return 0;
  else return 1;
}


long vMatchScore::Load(FILE * fp)
{
  long items = 0;
  double d_score = 0;
  items = vReadPCDoubles(fp, &d_score, 1);
  score = (float) d_score;
  items += vReadPCLongs(fp, &frame, 1);
  items += vReadPCLongs(fp, &view, 1);
  double d_ground_distance = 0;
  items += vReadPCDoubles(fp, &d_ground_distance, 1);
  ground_distance = (float) d_ground_distance;
  items += vReadPCLongs(fp, &rank, 1);

  if (items != 5) return 0;
  else 
  {
    is_valid = 1;
    return 1;
  }
}


class ScoreMore
{
public:
  ushort operator() (const vMatchScore & score1, const vMatchScore & score2);
};


class ScoreLess
{
public:
  ushort operator() (const vMatchScore & score1, const vMatchScore & score2);
};


class ScoreDistanceLess
{
public:
  ushort operator() (const vMatchScore & score1, const vMatchScore & score2);
};

class ScoreFrameLess
{
public:
  ushort operator() (const vMatchScore & score1, const vMatchScore & score2);
};



ushort ScoreMore::operator() (const vMatchScore & vscore1, const vMatchScore & vscore2)
{
  return (vscore1.score > vscore2.score);
}


ushort ScoreLess::operator() (const vMatchScore & vscore1, const vMatchScore & vscore2)
{
  return (vscore1.score < vscore2.score);
}


ushort ScoreDistanceLess::operator () (const vMatchScore & vscore1, 
                                       const vMatchScore & vscore2)
{
  if (vscore1.ground_distance == vscore2.ground_distance)
  {
    return (vscore1.rank < vscore2.rank);
  }
  else 
  {
    return (vscore1.ground_distance < vscore2.ground_distance);
  }
}


ushort ScoreFrameLess::operator() (const vMatchScore & vscore1, const vMatchScore & vscore2)
{
  if (vscore1.frame != vscore2.frame)
  {
    return (vscore1.frame < vscore2.frame);
  }
  else
  {
    return (vscore1.view < vscore2.view);
  }
}



class ScorepMore
{
public:
  ushort operator() (const void * score1, const void * score2);
};



ushort ScorepMore::operator() (const void * vscore1, const void * vscore2)
{
  const vMatchScore * score1 = (const vMatchScore *) vscore1;
  const vMatchScore * score2 = (const vMatchScore *) vscore2;
  return (score1->score > score2->score);
}


///////////////////////////////////////////////////////////////
//
// Beginning of implementation of class vMatchScores
//
///////////////////////////////////////////////////////////////


long vMatchScores::Initialize()
{
  is_valid = 0;
  descending = 0;
  scores = vZero(vMatchScore);

  frame_sorted = vZero(vMatchScore);
  distance_sorted = vZero(vMatchScore);
  score_sorted = vZero(vMatchScore);

  input_filename = 0;
  test_filename = 0;
  test_frame = 0;
  ground_truth = 0;

  fp = 0;

  // Give a default value for ground_truth_distance;
  ground_truth_distance = (double) 30.0;
  ground_distances = 0;
  frame_matches = 0;
  return 1;
}


vMatchScores::vMatchScores()
{
  Initialize();
}



vMatchScores::vMatchScores(const char * in_filename)
{
  Initialize();

  // We try to open a file called filename, or filename.bin, or filename.txt
  char * temp_name = vCopyString(in_filename);
  fp = fopen(temp_name, vFOPEN_READ);
  if (fp == 0) 
  {
    vdelete2(temp_name);
    temp_name = vMergeStrings2(in_filename, ".bin");
    fp = fopen(temp_name, vFOPEN_READ);
    if (fp == 0)
    {
      vdelete2(temp_name);
      temp_name = vMergeStrings2(in_filename, ".txt");
      fp = fopen(temp_name, vFOPEN_READ);
      if (fp == 0)
      {
        vdelete2(temp_name);
        return;
      }
    }
  }
  
  input_filename = vCopyString(temp_name);
  vdelete2(temp_name);
  
  // Scores can be saved in a text or a binary format.
  if (vStringCaseEndsIn(input_filename, ".txt"))
  {
    ReadTextFile();
  }
  else if (vStringCaseEndsIn(input_filename, ".bin"))
  {
    ReadBinaryFile();
  }
}


// This constructor generates scores given a matrix that already contains
// the distance between the test frame and each training frame.
vMatchScores::vMatchScores(v3dMatrix<double> * distances_matrix)
{
  Initialize();
  
  number_of_scores = distances_matrix->Size();
  scores = vnew(vMatchScore, number_of_scores+1);
  vArray(double) distances = distances_matrix->Matrix();
  vector<vMatchScore> scores_vector(number_of_scores);

  long i;
  for (i = 0; i < number_of_scores; i++)
  {
    double distance = distances[i];
    vMatchScore & score = scores_vector[i];
    score.score = (float) distance;
  }

  SetRanks(&scores_vector);
}


// This constructor generates scores given the following: A 
// matrix "features" that, in each column, has the numerical values of
// all training frames for a particular feature. "column", which
// specifies the column that is of interest to us. "test_feature"
// is the numerical value of the particular feature specified
// by column for the test frame. The score for each training
// frame is the absolute difference between the feature for that
// frame and the feature for the test frame.
vMatchScores::vMatchScores(v3dMatrix<double> * features_matrix, long column, 
                             double test_feature)
{
  Initialize();

  number_of_scores = features_matrix->Rows();
  scores = vnew(vMatchScore, number_of_scores+1);
  vArray2(double) features = features_matrix->Matrix2(0);
  vector<vMatchScore> scores_vector(number_of_scores);

  long i;
  for (i = 0; i < number_of_scores; i++)
  {
    double training_feature = features[i][column];
    vMatchScore & score = scores_vector[i];
    score.score = (float) vAbs(training_feature - test_feature);
  }

  SetRanks(&scores_vector);
}


// Similar to previous, but features is a matrix of floats.
vMatchScores::vMatchScores(v3dMatrix<float> * features_matrix, long column, 
                             double test_feature)
{
  Initialize();

  number_of_scores = features_matrix->Rows();
  scores = vnew(vMatchScore, number_of_scores+1);
  vArray2(float) features = features_matrix->Matrix2(0);
  vector<vMatchScore> scores_vector(number_of_scores);

  long i;
  for (i = 0; i < number_of_scores; i++)
  {
    double training_feature = features[i][column];
    vMatchScore & score = scores_vector[i];
    score.score = (float) vAbs(training_feature - test_feature);
  }

  SetRanks(&scores_vector);
}

  
// This constructor generates scores given the following: A 
// matrix "features" that is the result of applying PCA to the
// feature vectors of the training frames (each row is a vector). 
// "test_vector"
// is the feature_vector for the test frame (already converted using
// PCA). "pca" is the object
// that knows how to convert a vector using PCA and how
// to compute Mahalanobis distances between two vectors (after they
// have been converted using PCA). 
vMatchScores::vMatchScores(vMatrix<double> * features_matrix, 
                             vMatrix<double> * test_features, vPCA * pca)
{
  Initialize();

  number_of_scores = features_matrix->Rows();
  scores = vnew(vMatchScore, number_of_scores+1);
  vector<vMatchScore> scores_vector(number_of_scores);
  long dimensions = pca->UsedDimensions();
  vArray2(double) training_features = features_matrix->Matrix2();
  vArray(double) test_data = test_features->Matrix();
  vMatrix<double> eigenvalues_matrix = pca->Eigenvalues();
  vArray(double) eigenvalues = eigenvalues_matrix.Matrix();

  long i;
  for (i = 0; i < number_of_scores; i++)
  {
    double mahalanobis_distance = 0;
    vArray(double) current_training = training_features[i];
    for (long j = 0; j < dimensions; j++)
    {
      double temp_distance = current_training[j] - test_data[j];
      mahalanobis_distance += temp_distance * temp_distance / eigenvalues[j];
    }

    vMatchScore & score = scores_vector[i];
    score.score = (float) mahalanobis_distance;
  }

  SetRanks(&scores_vector);
}

  
// This constructor generates scores given the following: A 
// matrix "features" contains  feature vectors of the training frames 
// (each row is a vector). 
// "test_vector"
// is the feature_vector for the test frame. dimensions is the number
// of dimensions to use for the feature vectors.
vMatchScores::vMatchScores(v3dMatrix<double> * features_matrix, 
                             v3dMatrix<double> * test_features, long dimensions)
{
  Initialize();

  number_of_scores = features_matrix->Rows();
  scores = vnew(vMatchScore, number_of_scores+1);
  vector<vMatchScore> scores_vector(number_of_scores);
  vArray2(double) training_features = features_matrix->Matrix2(0);
  vArray(double) test_data = test_features->Matrix();

  long i;
  for (i = 0; i < number_of_scores; i++)
  {
    double distance = 0;
    vArray(double) current_training = training_features[i];
    for (long j = 0; j < dimensions; j++)
    {
      double temp_distance = current_training[j] - test_data[j];
      //distance += temp_distance * temp_distance;
      distance += vAbs(temp_distance);
      //distance += log(vAbs(temp_distance)+1);
    }

    vMatchScore & score = scores_vector[i];
    score.score = (float) distance;
  }

  SetRanks(&scores_vector);
}


// NOTE: It is not just the arguments that are different between the next 
// constructor and the previous. Here we compute L2 distances, before we 
// were computing L1 distances.
// Builds scores based on the L2 distance between test features and every 
// row of features_matrix. The "dimensions" argument specifies how many dimensions
// to use (dimensions from 0 to dimensions-1 are used).
vMatchScores::vMatchScores(v3dMatrix<float> * features_matrix, 
                             v3dMatrix<float> * test_features, long dimensions)
{
  Initialize();

  number_of_scores = features_matrix->Rows();
  scores = vnew(vMatchScore, number_of_scores+1);
  vector<vMatchScore> scores_vector(number_of_scores);
  vArray2(float) training_features = features_matrix->Matrix2(0);
  vArray(float) test_data = test_features->Matrix();

  long i;
  for (i = 0; i < number_of_scores; i++)
  {
    double distance = 0;
    vArray(float) current_training = training_features[i];
    for (long j = 0; j < dimensions; j++)
    {
      float temp_distance = current_training[j] - test_data[j];
      //distance += temp_distance * temp_distance;
      distance += temp_distance * temp_distance;
      //distance += log(vAbs(temp_distance)+1);
    }

    distance = sqrt(distance);
    vMatchScore & score = scores_vector[i];
    score.score = (float) distance;
  }

  SetRanks(&scores_vector);
}




// Builds scores based on the L1 distance between test features and every 
// row of features_matrix. The dimensions of features_matrix used are the
// dimensions from start_dimension to start_dimension+dimensions-1.
vMatchScores::vMatchScores(v3dMatrix<float> * features_matrix, 
                             v3dMatrix<double> * test_features, 
                             long start_dimension, long dimensions)
{
  Initialize();

  number_of_scores = features_matrix->Rows();
  long number_of_dimensions = features_matrix->Cols();
  long test_dimensions = test_features->Cols();
  if (dimensions > test_dimensions) return;
  if ((start_dimension < 0) || (dimensions < 0)) return;
  if (start_dimension + dimensions > number_of_dimensions) return;

  scores = vnew(vMatchScore, number_of_scores+1);
  vector<vMatchScore> scores_vector(number_of_scores);
  vArray2(float) training_features = features_matrix->Matrix2(0);
  vArray(double) test_data = test_features->Matrix();
  long end_dimension = start_dimension + dimensions - 1;

  long i;
  for (i = 0; i < number_of_scores; i++)
  {
    double distance = 0;
    vArray(float) current_training = training_features[i];
    for (long j = start_dimension; j <= end_dimension; j++)
    {
      double temp_distance = current_training[j] - test_data[j-start_dimension];
      //distance += temp_distance * temp_distance;
      distance += vAbs(temp_distance);
      //distance += log(vAbs(temp_distance)+1);
    }

    vMatchScore & score = scores_vector[i];
    score.score = (float) distance;
  }

  SetRanks(&scores_vector);
}

  
vMatchScores::vMatchScores(v3dMatrix<float> * features_matrix, 
                             v3dMatrix<float> * test_features, 
                             long start_dimension, long dimensions)
{
  Initialize();

  number_of_scores = features_matrix->Rows();
  long number_of_dimensions = features_matrix->Cols();
  long test_dimensions = test_features->Cols();
  if (dimensions > test_dimensions) return;
  if ((start_dimension < 0) || (dimensions < 0)) return;
  if (start_dimension + dimensions > number_of_dimensions) return;

  scores = vnew(vMatchScore, number_of_scores+1);
  vector<vMatchScore> scores_vector(number_of_scores);
  vArray2(float) training_features = features_matrix->Matrix2(0);
  vArray(float) test_data = test_features->Matrix();
  long end_dimension = start_dimension + dimensions - 1;

  long i;
  for (i = 0; i < number_of_scores; i++)
  {
    double distance = 0;
    vArray(float) current_training = training_features[i];
    for (long j = start_dimension; j <= end_dimension; j++)
    {
      double temp_distance = current_training[j] - test_data[j-start_dimension];
      //distance += temp_distance * temp_distance;
      distance += vAbs(temp_distance);
      //distance += log(vAbs(temp_distance)+1);
    }

    vMatchScore & score = scores_vector[i];
    score.score = (float) distance;
  }

  SetRanks(&scores_vector);
}

  
// Builds scores based on the L1 distance between test features and every 
// row of features_matrix. The dimensions of features_matrix used are the
// dimensions from start_dimension to start_dimension+dimensions-1.
// reference_images is a matrix which is useful only for its third column,
// which contains weights (it is the type of matrix that is used in
// interface to define the embedding).
vMatchScores::vMatchScores(v3dMatrix<float> * features_matrix, 
                             v3dMatrix<double> * test_features, 
                             long start_dimension, long dimensions,
                             v3dMatrix<float> * reference_images)
{
  Initialize();
  vArray2(float) weights = reference_images->Matrix2(0);

  number_of_scores = features_matrix->Rows();
  long number_of_dimensions = features_matrix->Cols();
  long test_dimensions = test_features->Cols();
  if (dimensions > test_dimensions) return;
  if ((start_dimension < 0) || (dimensions < 0)) return;
  if (start_dimension + dimensions > number_of_dimensions) return;

  scores = vnew(vMatchScore, number_of_scores+1);
  vector<vMatchScore> scores_vector(number_of_scores);
  vArray2(float) training_features = features_matrix->Matrix2(0);
  vArray(double) test_data = test_features->Matrix();
  long end_dimension = start_dimension + dimensions - 1;

  long i;
  for (i = 0; i < number_of_scores; i++)
  {
    double distance = 0;
    vArray(float) current_training = training_features[i];
    for (long j = start_dimension; j <= end_dimension; j++)
    {
      double temp_distance = current_training[j] - test_data[j-start_dimension];
      float weight = weights[j][2];
      //distance += temp_distance * temp_distance;
      distance += weight * vAbs(temp_distance);
      //distance += log(vAbs(temp_distance)+1);
    }

    vMatchScore & score = scores_vector[i];
    score.score = (float) distance;
  }

  SetRanks(&scores_vector);
}


vMatchScores::vMatchScores(v3dMatrix<float> * features_matrix, 
                             v3dMatrix<float> * test_features, 
                             long start_dimension, long dimensions,
                             v3dMatrix<float> * reference_images)
{
  Initialize();
  vArray2(float) weights = reference_images->Matrix2(0);

  number_of_scores = features_matrix->Rows();
  long number_of_dimensions = features_matrix->Cols();
  long test_dimensions = test_features->Cols();
  if (dimensions > test_dimensions) return;
  if ((start_dimension < 0) || (dimensions < 0)) return;
  if (start_dimension + dimensions > number_of_dimensions) return;

  scores = vnew(vMatchScore, number_of_scores+1);
  vector<vMatchScore> scores_vector(number_of_scores);
  vArray2(float) training_features = features_matrix->Matrix2(0);
  vArray(float) test_data = test_features->Matrix();
  long end_dimension = start_dimension + dimensions - 1;

  long i;
  for (i = 0; i < number_of_scores; i++)
  {
    double distance = 0;
    vArray(float) current_training = training_features[i];
    for (long j = start_dimension; j <= end_dimension; j++)
    {
      double temp_distance = current_training[j] - test_data[j-start_dimension];
      float weight = weights[j][2];
      //distance += temp_distance * temp_distance;
      distance += weight * vAbs(temp_distance);
      //distance += log(vAbs(temp_distance)+1);
    }

    vMatchScore & score = scores_vector[i];
    score.score = (float) distance;
  }

  SetRanks(&scores_vector);
}


// This constructor is useful if we only calculate scores for some
// of the frames in the database, but not for the rest. Then,
// the rest of the frames get scores that are higher than the
// highest score for the frames that were evaluated.
vMatchScores::vMatchScores(vArray(vMatchScore) in_scores, 
                             long in_number_of_scores)
{
  Initialize();

  number_of_scores = in_number_of_scores;
  scores = in_scores;

  vector<vMatchScore> scores_vector(number_of_scores);

  long i;
  for (i = 0; i < number_of_scores; i++)
  {
    scores_vector[i] = scores[i];
  }

  SetRanks2(&scores_vector);;
}

  
vMatchScores::vMatchScores(vMatchScores * in_scores)
{
  Initialize();

  number_of_scores = in_scores->NumberOfScores();
  scores = vnew(vMatchScore, number_of_scores);

  vector<vMatchScore> scores_vector(number_of_scores);
  vArray(vMatchScore) scores_array = in_scores->GetScores();

  long i;
  for (i = 0; i < number_of_scores; i++)
  {
    scores_vector[i] = scores_array[i];
  }

  SetRanks2(&scores_vector);;
}

  
// This constructor is useful for creating match scores out of scores
// computed using the Poser1 database.
vMatchScores::vMatchScores(v3dMatrix<float> * in_scores_matrix, vPoseSet poses)
{
  Initialize();
  vFrameViewFinder finder;

  number_of_scores = in_scores_matrix->Size();
  vArray(float) in_scores = in_scores_matrix->Matrix();
  scores = vnew(vMatchScore, number_of_scores);

  vector<vMatchScore> scores_vector(number_of_scores);

  long i;
  for (i = 0; i < number_of_scores; i++)
  {
    vDatabasePose & poser1_pose = poses.Pose(i);

    vMatchScore & score = scores[i];
    score.score = in_scores[i];
    finder.ConvertPoser1Pose(poser1_pose.frame, poser1_pose.rotation,
                             &(score.frame), &(score.view));
    score.is_valid = 1;
    scores_vector[i] = scores[i];
  }

  SetRanks2(&scores_vector);;
}



// This can read a text file, which (we assume) has scores only for one
// test frame. Therefore, number_of_frames is 1.
// Each line in the text file is like 
// FRAME VIEW SCORE
// We read all the lines, create the scores, and sort them.
ushort vMatchScores::ReadTextFile()
{
  double score = 0;
  long frame = 0;
  long view = 0;
  vector<vMatchScore> scores_vector;

  // Read each line of text and create the corresponding match score
  while(1)
  {
    long items_read = fscanf(fp, "%li %li %lf", &frame, &view, &score);
    if (items_read != 3) break;
    vMatchScore match_score = vMatchScore(score, frame, view);
    scores_vector.push_back(match_score);
  }
  number_of_scores = scores_vector.size();

  // sort the scores.
  if (number_of_scores < 1) return 0;
  std::sort(scores_vector.begin(), scores_vector.end(), ScoreMore());
  descending = 1;
  scores = vnew(vMatchScore, number_of_scores+1);

  SetRanks(&scores_vector);
  return 1;
}


// Use the score objects in scores_vector to fill in
// the scores in the scores array and the frame_sorted
// array, and set the rank of those scores correctly
long vMatchScores::SetRanks(vector<vMatchScore> * scores_vector)
{
  long i;
  for (i = 0; i < number_of_scores; i++)
  {
    vMatchScore & score = (*scores_vector)[i];
    score.frame = i / 4;
    score.view = i % 4;
    score.is_valid = 1;
  }

  // sort the scores.
  std::sort(scores_vector->begin(), scores_vector->end(), ScoreLess());
  descending = 0;
  
  for (i = 0; i < number_of_scores; i++)
  {
    scores[i] = (*scores_vector)[i];
    scores[i].rank = i;
  }
  
  EqualizeRanks();
  StoreFrameSorted();

  is_valid = 1;
  is_text = 0;
  current_frame = 0;
  current_score = 0;
  number_of_frames = 1;
  return 1;
}



// Like SetRanks, but here it handles the case 
// where scores_vector only contains some of the possible
// scores
long vMatchScores::SetRanks2(vector<vMatchScore> * scores_vector)
{
  long i;
  for (i = 0; i < number_of_scores; i++)
  {
    vMatchScore & score = (*scores_vector)[i];
    score.is_valid = 1;
  }

  // sort the scores.
  std::sort(scores_vector->begin(), scores_vector->end(), ScoreLess());
  descending = 0;
  
  for (i = 0; i < number_of_scores; i++)
  {
    scores[i] = (*scores_vector)[i];
    scores[i].rank = i;
  }
  
  EqualizeRanks();
  StoreFrameSorted2();
  
  is_valid = 1;
  is_text = 0;
  current_frame = 0;
  current_score = 0;
  number_of_frames = 1;
  return 1;
}


// Some times a whole sequence of scores can actually have the same
// score. In that case, their ranks should also be adjusted to
// be the same. What rank should that be? Here I choose the minimum
// of the ranks of all those equal scores.
long vMatchScores::EqualizeRanks()
{
  if (number_of_scores == 0) return 1;
  long i;
  double previous_score = scores[0].score;
  long previous_start = 0;
  for (i = 1; i < number_of_scores; i++)
  {
    double score = scores[i].score;
    if (score != previous_score)
    {
      if (i > previous_start + 1)
      {
        long rank = (i-1 + previous_start) / 2;
        long j;
        for (j = previous_start; j < i; j++)
        {
          scores[j].rank = previous_start;
        }
      }
      previous_start = i;
      previous_score = score;
    }

  }

  return 1;
}

  
// A binary file has the following format:
// long number_of_frames (number of test frames for which scores have been stored)
// long number_of_scores (number of scores stored for each frame)
// FRAME_SCORE frame_scores[number_of_frames]
//
// where FRAME_SCORE is like that:
// long frame_numbers[number_of_scores]
// long views[number_of_scores]
// double scores[number_of_scores]
//
// vMatchScores stores in memory only scores for one frame, every time we 
// want to get scores for another frame, it just loads it up from the file.
// Currently (6/20/2001) I only generate binary files containing scores for
// one frame. This is certainly an allowable case given the preceding format
// definition.
ushort vMatchScores::ReadBinaryFile()
{
  // Read the header (number_of_frames and number_of_scores)
  long items = vReadPCLong(fp, &number_of_frames);
  if (items != 1) return 0;
  items = vReadPCLong(fp, &number_of_scores);
  if (items != 1) return 0;

  if (number_of_frames < 1) return 0;
  if (number_of_scores < 1) return 0;

  initial_offset = ftell(fp);
  frame_size = 16 * number_of_scores;
  
  // Load up scores for frame 0
  SetCurrentFrame(0);

  SetCurrentScore(0);
  is_text = 0;
  return 1;
}


// For format specs, see comments before ReadBinaryFile. This function
// saves the current scores that are loaded into memory into a one-frame
// binary file
ushort vMatchScores::WriteBinary()
{
  // We don't convert to binary if we read from a binary file anyway.
  if (vStringEndsIn(input_filename, ".txt"))
  {
    if (!is_text) return 0;
  }
  if (input_filename == 0) return 0;
  char * output_file = vReplaceEnding2(input_filename, ".bin");
  // We don't want to accidentally write over an existing file.
//  if (vFileExists(output_file))
//  {
//    vPrint("%s already exists\n", output_file);
//    vdelete2(output_file);
//   return 0;
//  }
  FILE * fp = fopen(output_file, vFOPEN_WRITE);
  if (fp == 0)
  {
    vPrint("Can't open %s\n", output_file);
    vdelete2(output_file);
    return 0;
  }

  // Write the header (number_of_frames and number_of_scores)
  assert(number_of_frames == 1);
  long items = vWritePCLong(fp, number_of_frames);
  if (items != 1) return 0;
  items = vWritePCLong(fp, number_of_scores);
  if (items != 1) return 0;

  if (number_of_frames < 1) return 0;
  if (number_of_scores < 1) return 0;

  ushort success = 1;
  double * score_values = new double[number_of_scores];
  long * frames = new long[number_of_scores];
  long * views = new long[number_of_scores];

  // Create the actual score objects and store the values there.
  long i;
  for (i = 0; i < number_of_scores; i++)
  {
    score_values[i] = scores[i].score;
    frames[i] = scores[i].frame;
    views[i] = scores[i].view;
  }

  // Reading the frame numbers for each score
  items = vWritePCLongs(fp, frames, number_of_scores);
  if (items != number_of_scores) success = 0;

  // Reading the views for each score (4 views, for 0, 90, 270, 360 degree rotations)
  items = vWritePCLongs(fp, views, number_of_scores);
  if (items != number_of_scores) success = 0;
  // Reading the actual scores, which are assumed to be sorted.
  items = vWritePCDoubles(fp, score_values, number_of_scores);
  if (items != number_of_scores) success = 0;
  fclose(fp);
 
  vdelete2(score_values);
  vdelete2(frames);
  vdelete2(views);
  vdelete2(output_file);

// The next two lines I think were a bug.
//  vdelete2(distance_sorted);
//  vdelete2(score_sorted);
  return success;
}


vMatchScores::~vMatchScores()
{
  vdelete2(scores);
  vdelete2(frame_sorted);
  if (fp != 0)
  {
    fclose(fp);
  }
  vdelete2(input_filename);

  vdelete2(test_filename);
  vdelete(ground_truth);
  vdelete2(score_sorted);
  vdelete2(distance_sorted);
}


ushort vMatchScores::SetOrder(ushort in_descending)
{
  if ((in_descending != 0) && (in_descending != 1))
  {
    return 0;
  }
  if (in_descending != descending)
  {
    Reverse();
  }
  return 1;
}


long vMatchScores::StoreFrameSorted()
{
  vdelete2(frame_sorted);
  frame_sorted = vnew(vMatchScore, number_of_scores + 1);
  long i;
  for (i = 0; i < number_of_scores; i++)
  {
    vMatchScore score = scores[i];
    long frame = score.frame;
    long view = score.view;
    long index = frame * 4 + view;
    frame_sorted[index] = score;
  }
  return 1;
}

  
long vMatchScores::StoreFrameSorted2()
{
  vdelete2(frame_sorted);
  frame_sorted = vnew(vMatchScore, number_of_scores + 1);
  vector<vMatchScore> temp_scores(number_of_scores);
  long i;
  for (i = 0; i < number_of_scores; i++)
  {
    temp_scores[i] = scores[i];
  }

  std::sort(temp_scores.begin(), temp_scores.end(), ScoreFrameLess());
  for (i = 0; i < number_of_scores; i++)
  {
    frame_sorted[i] = temp_scores[i];
  }

  return 1;
}

  
long vMatchScores::CurrentFrame()
{
  return current_frame;
}

// Loads in the scores for the given frame from the file
ushort vMatchScores::SetCurrentFrame(long frame)
{
  if (frame < 0) return 0;
  if (frame >= number_of_frames) return 0;

  current_frame = frame;
  long file_position = initial_offset + frame_size * current_frame;
  int result = fseek(fp, file_position, SEEK_SET);
  if (result != 0) return 0;

  ushort success = 1;
  double * buffer1 = new double[number_of_scores];
  long * buffer2 = new long[number_of_scores];
  long * buffer3 = new long[number_of_scores];

  long items;
  // Reading the frame numbers for each score
  items = vReadPCLongs(fp, buffer2, number_of_scores);
  if (items != number_of_scores) success = 0;
  // Reading the views for each score (4 views, for 0, 90, 270, 360 degree rotations)
  items = vReadPCLongs(fp, buffer3, number_of_scores);
  if (items != number_of_scores) success = 0;
  // Reading the actual scores, which are assumed to be sorted.
  items = vReadPCDoubles(fp, buffer1, number_of_scores);
  if (items != number_of_scores) success = 0;

  if (success != 0)
  {
    scores = vnew(vMatchScore, number_of_scores+1);
    long i;
    for (i = 0; i < number_of_scores; i++)
    {
      scores[i].score = (float) buffer1[i];
      scores[i].frame = buffer2[i];
      scores[i].view = buffer3[i];
      scores[i].is_valid = 1;
      scores[i].rank = i;
    }
  }

  StoreFrameSorted();

  // If we are reading from a binary file, we assume that
  // the scores have already been saved there in sorted order.
  // We just need to figure out if it was ascending or descending
  // order.
  if (scores[0].score > scores[number_of_scores-1].score) descending = 1;
  else descending = 0;

  // Create the actual score objects and store the values there.
  vdelete2(buffer1);
  vdelete2(buffer2);
  vdelete2(buffer3);
  if (success != 0) is_valid = 1;
  return success;
}


ushort vMatchScores::PreviousFrame()
{
  return SetCurrentFrame(current_frame - 1);
}


ushort vMatchScores::NextFrame()
{
  return SetCurrentFrame(current_frame + 1);
}


ushort vMatchScores::FirstFrame()
{
  return SetCurrentFrame(0);
}


ushort vMatchScores::LastFrame()
{
  return SetCurrentFrame(number_of_frames - 1);
}


ushort vMatchScores::SetCurrentScore(long index)
{
  if (index < 0) return 0;
  if (index >= number_of_scores) return 0;
  current_score = index;
  return 1;
}


vMatchScore vMatchScores::GetScore(long frame, long rank)
{
  ushort success = SetCurrentFrame(frame);
  if (success == 0) return vMatchScore();
  success = SetCurrentScore(rank);
  if (success == 0) return vMatchScore();
  else return CurrentScore();
}


vMatchScore vMatchScores::GetScore(long rank)
{
  ushort success = SetCurrentScore(rank);
  if (success == 0) return vMatchScore();
  else return CurrentScore();
}


vMatchScore vMatchScores::CurrentScore()
{
  return scores[current_score];
}


vMatchScore vMatchScores::PreviousScore()
{
  return GetScore(current_score - 1);
}


vMatchScore vMatchScores::NextScore()
{
  return GetScore(current_score + 1);
}


vMatchScore vMatchScores::FirstScore()
{
  return GetScore(0);
}


vMatchScore vMatchScores::LastScore()
{
  return GetScore(number_of_scores - 1);
}


long vMatchScores::FrameSortedIndex(long frame, long view)
{
  if (number_of_scores == 0) return -1;
  long range_min = 0;
  long range_max = number_of_scores - 1;
  long current = range_max / 2;
  vMatchScore temp_score;
  temp_score.frame = frame;
  temp_score.view = view;
  // We need a ScoreFrameLess object to call its operator () that
  // compares two match scores in terms of order of their frames and views.
  ScoreFrameLess junk_object;
  while(range_min < range_max)
  {
    vMatchScore & score = frame_sorted[current];
    if (junk_object.operator () (temp_score, score) == 1)
    {
      range_max = current - 1;
    }
    else
    {
      range_min = current;
    }
    current = (range_min + range_max + 1) / 2;
  }
  vMatchScore & found_score = frame_sorted[range_min];
  if ((found_score.frame == frame) && (found_score.view == view))
  {
    return range_min;
  }
  else 
  {
    return -1;
  }
}

  
// This works also when we only have scores for some frames of the database.
// GetRank requires that we have scores for the whole database.
long vMatchScores::GetRank2(long frame, long view)
{
  long i;
  for (i = 0; i < number_of_scores; i++)
  {
    vMatchScore & score = frame_sorted[i];
    if ((score.frame == frame) && (score.view == view))
    {
      return score.rank;
    }
  }
  return -1;
}


long vMatchScores::NumberOfScores()
{
  return number_of_scores;
}

// Reverse the order in which the scores appear.
ushort vMatchScores::Reverse()
{
  descending = 1 - descending;
  long number = NumberOfScores();
  long i;
  long index2 = number - 1;
  for (i = 0; i < number / 2; i++)
  {
    vMatchScore temp_score = scores[i];
    scores[i] = scores[index2];
    scores[index2] = temp_score;
    scores[i].rank = i;
    scores[index2].rank = index2;
    index2--;
  }

  // Resort score_sorted.
  SetTestFrame(test_frame);

  return 1;
}


ushort vMatchScores::IsValid()
{
  return is_valid;
}


// Removes two endings from the filename to get the part of the filename 
// that is common among all files containing scores for the same 
// matching function and same test video.
char * vMatchScores::GetTextBase()
{
  char * chopped1 = vReplaceEnding2(input_filename, "");
  char * result = vReplaceEnding2(chopped1, "");
  vdelete2(chopped1);
  return result;
}


// Sets the variable input_filename (which is slightly misnamed, since
// it is also used for output, when we call WriteBinary).
long vMatchScores::SetFilename(const char * new_filename)
{
  vdelete2(input_filename);
  input_filename = vCopyString(new_filename);
  return 1;
}


  
// Returns text_base.frame
char * vMatchScores::FrameFilename(const char * text_base, long frame)
{
  char buffer[80];
  sprintf(buffer, ".%li", frame);
  return vMergeStrings2(text_base, buffer);
}


// SetTestFilename sets the test_filename, thus enabling the program
// to look for corresponding auxiliary files (like a .gt ground truth
// file).
ushort vMatchScores::SetTestFilename(const char * in_test_filename)
{
  vdelete2(test_filename);
  test_filename = vCopyString(in_test_filename);
  vdelete(ground_truth);
  // Load the ground truth file that corresponds to the 
  // test video file (if any)l
  char * ground_truth_file = vReplaceEnding2(test_filename, ".gt");
  // Note that vViewGroundTruth::Load will return NULL if
  // the file is not there or if its format is wrong.
  ground_truth = vViewGroundTruth::Load(ground_truth_file);
  vdelete2(ground_truth_file);
  return 1;
}

// Provided a ground truth file has been loaded, GroundMatch will
// return the score between the current test frame and the training
// view that is the rank-th closest (in terms of ground truth) to it.
// Views for which the underlying pose does not match the test frame's
// pose are not considered at all.
vMatchScore vMatchScores::GroundMatch(long rank)
{
  vMatchScore score;
  if (vIsZero(ground_truth)) return score;
  if (vIsZero(score_sorted)) return score;
  if (rank < 0) return score;
  if (rank >= frame_matches) return score;

  return distance_sorted[rank];
}

// Provided a ground truth file has been loaded, CorrectMatch will
// return the rank-th best score between the current test frame and 
// a corresponding training pose whose view/orientation parameters 
// are within 
// ground_truth_distance from those of the test frame. As in GroundMatch,
// views for which the underlying pose does not match the test frame's
// pose are not considered at all. The distance doesn't have to be 
// based on the L2 metric, I may try different ones.
//
// In total_matches we return the total number of correct matches that
// were found.
vMatchScore vMatchScores::CorrectMatch(long rank, long * total_matches)
{
  vMatchScore score;
  score.rank = number_of_scores;
  if (vIsZero(ground_truth)) return score;
  if (vIsZero(distance_sorted)) return score;
  if (rank < 0) return score;
  if (rank >= frame_matches) return score;

  long match;
  long result_index = -1;
  long counter = 0;
  // Go through the scores in sorted order, and stop when we find
  // the rank-th (starting with 0-th) best score whose distance
  // is no bigger than ground_truth_distance; counter will count
  // how many "correct" scores we have found so far.
  for (match = 0; match < frame_matches; match++)
  {
    vMatchScore score = score_sorted[match];
    if (score.ground_distance <= ground_truth_distance)
    {
      if (counter == rank) 
      {
        result_index = match;
        if (total_matches == 0) break;
      }
      counter++;
    }
  }

  // If we didn't find enough "correct" matches, we return
  // an uninitialized (invalid) score.
  if (result_index == -1) return score;
  else 
  {
    if (total_matches != 0)
    {
      *total_matches = counter;
    }
    return score_sorted[result_index];
  }
}

// Sets the variable ground_truth_distance, which is used in 
// CorrectMatch();
ushort vMatchScores::SetGroundTruthDistance(double in_ground_truth_distance)
{
  ground_truth_distance = in_ground_truth_distance;
  return 1;
}


// If ground truth is available it returns 1 and sets x, y and z
// contents to the camera view contained in the ground truth.
// If ground truth is not available, it returns 0.
long vMatchScores::GroundTruthView(double * x, double * y, double * z)
{
  if (ground_truth == 0) return 0;
  vViewGroundTruth params = (*ground_truth)[test_frame];
  *x = params.x_angle;
  *y = params.y_angle;
  *z = params.orientation;
  return 1;
}


// in_test_frame is a frame from the test sequence for which 
// the scores have been generated. We set test_frame equal
// to it, and we also store information about the results
// for that frame in the distance_sorted and score_sorted
// arrays, if ground truth information is available. 

/* // This version doesn't work in cases where we don't 
   // have scores for each frame, but only for a few frames.
ushort vMatchScores::SetTestFrame(long in_test_frame)
{
  if (in_test_frame < 0) return 0;
  if (ground_truth == 0) return 0;
  if (in_test_frame >= ground_truth->size()) return 0;
  if (ground_distances == 0) return 0;
  test_frame = in_test_frame;

  // Store information in the distance_sorted and score_sorted
  // arrays. 
  vViewGroundTruth params = (*ground_truth)[test_frame];
  vector<long> training_frames;
  vector<long> rotations;
  vector<double> distances;

  long i;

  // Now find the scores for the correct training pose
  // and put them into a vector;
  long first = finder.FirstPoseView(params.training_pose_index);
  long last = finder.LastPoseView(params.training_pose_index);
  vector<vMatchScore> frame_scores;
  for (i = 0; i < number_of_scores; i++)
  {
    if ((scores[i].frame >= first) && (scores[i].frame <= last))
    {
      frame_scores.push_back(scores[i]);
    }
  }

  assert(frame_scores.size() == finder.ScoresPerPose());

  long number_of_matches = frame_scores.size();
  double x_angle = params.x_angle;
  double y_angle = params.y_angle;
  double orientation = params.orientation;
  long frames_per_pose = finder.FramesPerPose();
  vArray2(ushort) dist_matrix = ground_distances->Matrix2(0);

  for (i = 0; i < number_of_matches; i++)
  {
    vMatchScore & score = frame_scores[i];
    long training_frame = score.frame;
    long rotation = score.view;
 //   double distance = 
 //     finder.Distance(training_frame, 4 - rotation,
 //                                     x_angle, y_angle, orientation);
 
    long distance_index = 4 * (training_frame % frames_per_pose) + rotation;
    double temp_distance = dist_matrix[test_frame][distance_index];
    double distance = temp_distance / 10;
    score.ground_distance = distance;
    scores[score.rank].ground_distance = distance;
    long index = score.frame * 4 + score.view;
    frame_sorted[index].ground_distance = distance;
  }


  // If scores are actually in descending order, sort in descending order
  // of score value
  if (descending == 0)
  {
    std::sort(frame_scores.begin(), frame_scores.end(), ScoreLess());
  }

  else
  {
    std::sort(frame_scores.begin(), frame_scores.end(), ScoreMore());
  }
   

  // Store the sorted scores into the score_sorted array
  if (vIsZero(score_sorted)) 
  {
    score_sorted = vnew(vMatchScore, number_of_matches);
  }

  for (i = 0; i < number_of_matches; i++)
  {
    score_sorted[i] = frame_scores[i];
  }

  // Sort the scores based on their ground_distance value
  std::sort(frame_scores.begin(), frame_scores.end(), ScoreDistanceLess());

  // Store the sorted scores into the distance_sorted array
  if (vIsZero(distance_sorted)) 
  {
    distance_sorted = vnew(vMatchScore, number_of_matches);
  }
  for (i = 0; i < number_of_matches; i++)
  {
    distance_sorted[i] = frame_scores[i];
  }

  return 1;
}
*/


ushort vMatchScores::SetTestFrame(long in_test_frame)
{
  if (in_test_frame < 0) return 0;
  if (ground_truth == 0) return 0;
  if (in_test_frame >= (long) ground_truth->size()) return 0;
  if (ground_distances == 0) return 0;
  test_frame = in_test_frame;

  // Store information in the distance_sorted and score_sorted
  // arrays. 
  vViewGroundTruth params = (*ground_truth)[test_frame];
  vector<long> training_frames;
  vector<long> rotations;
  vector<double> distances;

  long i;

  // Now find the scores for the correct training pose
  // and put them into a vector;
  long first = finder.FirstPoseView(params.training_pose_index);
  long last = finder.LastPoseView(params.training_pose_index);
  vector<vMatchScore> frame_scores;
  for (i = 0; i < number_of_scores; i++)
  {
    if ((scores[i].frame >= first) && (scores[i].frame <= last))
    {
      frame_scores.push_back(scores[i]);
    }
  }

  frame_matches = frame_scores.size();
  if (frame_matches == 0) 
  {
    vdelete2(score_sorted);
    score_sorted = vZero(vMatchScore);
    vdelete2(distance_sorted);
    distance_sorted = vZero(vMatchScore);
    return 1;
  }

  long frames_per_pose = finder.FramesPerPose();
  vArray2(ushort) dist_matrix = ground_distances->Matrix2(0);

  for (i = 0; i < frame_matches; i++)
  {
    vMatchScore & score = frame_scores[i];
    long training_frame = score.frame;
    long rotation = score.view;
 //   double distance = 
 //     finder.Distance(training_frame, 4 - rotation,
 //                                     x_angle, y_angle, orientation);
 
    long distance_index = 4 * (training_frame % frames_per_pose) + rotation;
    double temp_distance = dist_matrix[test_frame][distance_index];
    double distance = temp_distance / 10;
    score.ground_distance = (float) distance;
    scores[score.rank].ground_distance = (float) distance;
    long index = FrameSortedIndex(training_frame, rotation);
    frame_sorted[index].ground_distance = (float) distance;
  }

  // If scores are actually in descending order, sort in descending order
  // of score value
  if (descending == 0)
  {
    std::sort(frame_scores.begin(), frame_scores.end(), ScoreLess());
  }

  else
  {
    std::sort(frame_scores.begin(), frame_scores.end(), ScoreMore());
  }
   

  // Store the sorted scores into the score_sorted array
  if (vIsZero(score_sorted)) 
  {
    score_sorted = vnew(vMatchScore, frame_matches);
  }

  for (i = 0; i < frame_matches; i++)
  {
    score_sorted[i] = frame_scores[i];
  }

  // Sort the scores based on their ground_distance value
  std::sort(frame_scores.begin(), frame_scores.end(), ScoreDistanceLess());

  // Store the sorted scores into the distance_sorted array
  if (vIsZero(distance_sorted)) 
  {
    distance_sorted = vnew(vMatchScore, frame_matches);
  }
  for (i = 0; i < frame_matches; i++)
  {
    distance_sorted[i] = frame_scores[i];
  }

  return 1;
}


// Sets the ground truth to the given parameters. This is used
// by GroundMatch and CorrectMatch
long vMatchScores::SetGroundTruth(long training_pose, double x_angle, 
                                   double y_angle, double z_angle)
{
  if (ground_distances == 0) return 0;
  vector<long> training_frames;
  vector<long> rotations;
  vector<double> distances;

  long i;

  // Now find the scores for the correct training pose
  // and put them into a vector;
  long first = finder.FirstPoseView(training_pose);
  long last = finder.LastPoseView(training_pose);
  vector<vMatchScore> frame_scores;
  for (i = 0; i < number_of_scores; i++)
  {
    if ((scores[i].frame >= first) && (scores[i].frame <= last))
    {
      frame_scores.push_back(scores[i]);
    }
  }

  frame_matches = frame_scores.size();
  if (frame_matches == 0) 
  {
    vdelete2(score_sorted);
    score_sorted = vZero(vMatchScore);
    vdelete2(distance_sorted);
    distance_sorted = vZero(vMatchScore);
    return 1;
  }

  long frames_per_pose = finder.FramesPerPose();
  vArray2(ushort) dist_matrix = ground_distances->Matrix2(0);

  for (i = 0; i < frame_matches; i++)
  {
    vMatchScore & score = frame_scores[i];
    long training_frame = score.frame;
    long rotation = score.view;
    double distance = 
       finder.Distance(training_frame, rotation,
                       x_angle, y_angle, z_angle);
 
    //long distance_index = 4 * (training_frame % frames_per_pose) + rotation;
    //double temp_distance = dist_matrix[test_frame][distance_index];
    //double distance = temp_distance / 10;
    score.ground_distance = (float) distance;
    scores[score.rank].ground_distance = (float) distance;
    long index = FrameSortedIndex(training_frame, rotation);
    frame_sorted[index].ground_distance = (float) distance;
  }

  // If scores are actually in descending order, sort in descending order
  // of score value
  if (descending == 0)
  {
    std::sort(frame_scores.begin(), frame_scores.end(), ScoreLess());
  }

  else
  {
    std::sort(frame_scores.begin(), frame_scores.end(), ScoreMore());
  }
   

  // Store the sorted scores into the score_sorted array
  if (vIsZero(score_sorted)) 
  {
    score_sorted = vnew(vMatchScore, frame_matches);
  }

  for (i = 0; i < frame_matches; i++)
  {
    score_sorted[i] = frame_scores[i];
  }

  // Sort the scores based on their ground_distance value
  std::sort(frame_scores.begin(), frame_scores.end(), ScoreDistanceLess());

  // Store the sorted scores into the distance_sorted array
  if (vIsZero(distance_sorted)) 
  {
    distance_sorted = vnew(vMatchScore, frame_matches);
  }
  for (i = 0; i < frame_matches; i++)
  {
    distance_sorted[i] = frame_scores[i];
  }

  return 1;
}


// Sets the ground truth parameters to be those of the database image
// identified by the given frame and view
long vMatchScores::SetGroundTruthView(long frame, long view)
{
  vFrameViewFinder finder;
  long pose;
  double x_angle, y_angle, z_angle;
  pose = 0;
  x_angle = y_angle = z_angle = (double) 0;
  finder.GetViewParams(frame, view, &pose, &x_angle, &y_angle, &z_angle);
  vPrint("x_angle = %lf, y_angle = %lf, z_angle = %lf\n", x_angle, y_angle, z_angle);
  return SetGroundTruth(pose, x_angle, y_angle, z_angle);
}


  
// Reads in scores from files text_base.0.txt, text_base.1.txt ... text_base.limit.txt
// and creates a binary file that contains, for each text file, the number_of_scores
// highest and lowest scores from that file. 
//
// Notice that this function is not static. We call it with an object. This allows 
// us to use some of the information that the object holds. For example, the 
// text base can be obtained from the input file by removing the two endings (.txt
// and the .number before it). Based on the value of descending we know if we 
// should reverse the scores before saving them or not.
//
// Actually, number_of_scores is not supported yet.

ushort vMatchScores::ToBinary(long limit)
{
  // To get the text base we need to chop off the two endings (something
  // like ".1.txt")
  char * text_base = GetTextBase();
  long base_length = strlen(text_base);
  long text_length = base_length + 20;
  ushort success = 1;

  char * text_name = new char[text_length];
  long i = 0;
  while(1)
  {
    if (i > limit) break;
    sprintf(text_name, "%s.%li.txt", text_base, i);
    i++;
    char * bin_name = vReplaceEnding2(text_name, ".bin");
    ushort file_exists = vFileExists(bin_name);
    vdelete2(bin_name);
    if (file_exists)
    {
      vPrint("%s is already converted\n", text_name);
      continue;
    }

    vMatchScores in_scores = vMatchScores(text_name);
    if (in_scores.IsValid() == 0) 
    {
      vPrint("%s score file not found\n", text_name);
      continue;
    }

    in_scores.SetOrder(descending);

    ushort temp_success = in_scores.WriteBinary();
    if (temp_success == 0)
    {
      vPrint("Error in converting %s to binary\n", text_name);
      success = 0;
    }
    else
    {
      vPrint("Converted score file %s to binary\n", text_name);
    }
  }

  vdelete2(text_name);
  vdelete2(text_base);
  return 1;
}


// Deprecated (9/21/2001)
// GroundTruthStatistics() prints useful information about the matching rank
// of the views that we consider "correct", for a sequence of test frames. We
// assume (as in ToBinary()) that the scores for that sequence have been saved
// in files of the form base.i.end where base is any string, i is the frame number
// and end is txt or bin.
ushort vMatchScores::GroundTruthStatistics(long limit)
{
  assert(0);
  return GroundTruthStatistics3(stdout, limit);
}

// Deprecated (9/21/2001)
// Like GroundTruthStatistics, but it prints the results to the given file
ushort vMatchScores::GroundTruthStatistics2(const char * filename, 
                                             long limit)
{
  assert(0);
  if (vFileExists(filename)) 
  {
    vPrint("GroundTruthStatistics2 error: File %s exists\n", filename);
    return 0;
  }

  FILE * fp = fopen(filename, vFOPEN_WRITE);
  if (fp == 0) 
  {
    vPrint("GroundTruthStatistics2 error: Could not open %s\n", filename);
    return 0;
  }
  ushort success = GroundTruthStatistics3(fp, limit);
  fclose(fp);
  return success;
}

// Deprecated
// LikeGroundTruthStatistics2, but the file is given as an 
// open file pointer, instead of as a filename.
ushort vMatchScores::GroundTruthStatistics3(FILE * fp, long limit)
{
  assert(0);
  // To get the text base we need to chop off the two endings (something
  // like ".1.txt")
  char * text_base = GetTextBase();
  long base_length = strlen(text_base);
  long text_length = base_length + 20;
  ushort success = 1;

  long i = 0;
  while(1)
  {
    if (i > limit) break;

    char * text_name = FrameFilename(text_base, i);
    vMatchScores in_scores = vMatchScores(text_name);
    if (in_scores.IsValid() == 0) 
    {
      vPrint("%s score file not found\n", text_name);
      vdelete2(text_name);
      i++;
      continue;
    }

    in_scores.SetOrder(descending);

    in_scores.SetGroundTruthDistance(ground_truth_distance);
    in_scores.SetTestFilename(test_filename);
    in_scores.SetTestFrame(i);

    if (fp == stdout)
    {
      vPrint("\r\nFrame %li\r\n\r\n", i);
    }
  
    else
    {
      fprintf(fp, "\r\nFrame %li\r\n\r\n", i);
    }

    ushort temp_success = in_scores.FrameStatistics(fp);

    if (temp_success == 0)
    {
      vPrint("Error in getting statistics from %li\n", i);
      success = 0;
    }

    i++;
    vdelete2(text_name);
 }

  vdelete2(text_base);
  return 1;


  return 0;
}


// Like GroundTruthStatistics3, but only prints statistics for this object.
ushort vMatchScores::FrameStatistics(FILE * fp)
{
  vMatchScore score;

  ushort result = 1;
  long top_scores = 5;
  long total_matches;
  long i;
  ushort temp_success;
  
  for (i = 0; i < top_scores; i++)
  {
    score = CorrectMatch(i, &total_matches);
    if (score.is_valid != 0)
    {
      if (fp == stdout)
      {
        vPrint("%9li: ", i);
      }
      else
      {
        fprintf(fp, "%9li: ", i);
      }
      temp_success = score.Fprint(fp);
      if (temp_success == 0) result = 0;
    }
  }
  
  if (fp == stdout)
  {
    vPrint("\r\n%li correct matches were found.\r\n", total_matches);
  }
  else
  {
    fprintf(fp, "\r\n%li correct matches were found.\r\n", total_matches);
  }
  if (total_matches > top_scores)
  {
    score = CorrectMatch(total_matches - 1, 0);
    assert(score.is_valid != 0);
    if (fp == stdout)
    {
      vPrint("%9li: ", total_matches - 1);
    }
    else
    {
      fprintf(fp, "%9li: ", total_matches - 1);
    }
    temp_success = score.Fprint(fp);
    if (temp_success == 0) result = 0;
  }

  if (fp == stdout)
  {
    vPrint("\r\n--------------------------------------------\r\n");
  }
  else
  {
    fprintf(fp, "\r\n--------------------------------------------\r\n");
  }

  return result;
}


// Filter uses another vMatchScores object to discard scores in this 
// object that correspond to frames that score really bad (rank > threshold)
// in the other scores object.
long  vMatchScores::Filter(vMatchScores * other_scores, long threshold)
{
  if (threshold < 0) return 0;
  if (threshold >= number_of_scores) return 0;
  if (number_of_scores != other_scores->NumberOfScores()) return 0;

  double score_threshold = scores[threshold].score;
  vArray(vMatchScore) other_array = other_scores->frame_sorted;

  long i;
  for (i = 0; i < number_of_scores; i++)
  {
    long frame = scores[i].frame;
    long view = scores[i].view;
    long index = frame * 4 + view;
    long other_score_rank = other_array[index].rank;
    if (other_score_rank < threshold) continue;
    scores[i].score = (float) score_threshold;
  }

  vMatchScore * scores_array = vData(scores);
  if (descending != 0)
  {
    std::sort(&(scores_array[0]), &(scores_array[number_of_scores]), 
              ScoreMore());
  }
  else
  {
    std::sort(&(scores_array[0]), &(scores_array[number_of_scores]), 
              ScoreLess());
  }
  for (i = 0; i < number_of_scores; i++)
  {
    scores[i].rank = i;
  }

  // Redo ground_truth calculations.
  SetTestFrame(test_frame);

  return 1;
}


// KeepHighest returns a new vMatchScores object, that contains scores only
// for matches_to_keep frames. The highest-ranking matches are kept.
// For example, if matches_to_keep is 100, a new vMatchScores object
// containing scores for the 100 highest matches (as described in
// this object) is returned.
vMatchScores * vMatchScores::KeepHighest(long matches_to_keep)
{
  if (matches_to_keep > number_of_scores) return 0;
  vArray(vMatchScore) new_array = vnew(vMatchScore, 
                                         matches_to_keep);
  long i;
  for (i = 0; i < matches_to_keep; i++)
  {
    new_array[i] = scores[i];
  }
  vMatchScores * result = new vMatchScores(new_array, matches_to_keep);
  return result;
}


// Select creates a new vMatchScores object containing scores
// (as described in this object) for all frames for which
// scores exist in scores2. This function verifies that
// for all scores in scores2 there is a score with identical
// frame and view in frame_sorted. If not, 0 is returned.
vMatchScores * vMatchScores::Select(vMatchScores * scores2)
{
  vArray(vMatchScore) frame_sorted2 = scores2->GetFrameSorted();
  long new_number = scores2->NumberOfScores();
  vArray(vMatchScore) new_array = vnew(vMatchScore, new_number);
  long index1, index2;
  index1 = 0;
  long success = 1;
  for (index2 = 0; index2 < new_number; index2++)
  {
    vMatchScore score2 = frame_sorted2[index2];
    long frame2 = score2.frame;
    long view2 = score2.view;
    vMatchScore score1;
    while(1)
    {
      score1 = frame_sorted[index1];
      index1++;
      if ((score1.frame == frame2) &&
          (score1.view == view2))
      {
        break;
      }
      if ((score1.frame > frame2) || (index1 >= number_of_scores))
      {
        success = 0;
        break;
      }
    }
    if (success == 0) break;
    new_array[index2] = score1;
  }
  if (success == 0)
  {
    vdelete2(new_array);
    return 0;
  }
  else
  {
    vMatchScores * result = new vMatchScores(new_array, new_number);
    return result;
  }
}

  
// The function Combine combines the ranks of the scores in *this with the
// ranks of the scores in scores2, and uses the combined log ranks to
// create a new vMatchScores object. The weight of this is weight, and 
  // the weight of scores2 is 1-weight.
vMatchScores * vMatchScores::Combine1(vMatchScores * scores2, double weight1)
{
  vMatrix<double> log_matrix = vPrecalculateLogs3(number_of_scores);
  vArray(double) logs = log_matrix.Matrix();

  vArray(vMatchScore) frame_sorted2 = scores2->frame_sorted;
  v3dMatrix<double> ranks_matrix(1, number_of_scores, 1);
  vArray(double) ranks = ranks_matrix.Matrix();
  double weight2 = 1 - weight1;
  long i;
  for (i = 0; i < number_of_scores; i++)
  {
    // We add 1 to the ranks to avoid trying to calculate log(0) (for the 
    // match score whose rank is 0).
    long rank1 = frame_sorted[i].rank + 1;
    long rank2 = frame_sorted2[i].rank + 1;
    ranks[i] = weight1 * logs[rank1] + weight2 * logs[rank2];
  }

  vMatchScores * result = new vMatchScores(&ranks_matrix);
  return result;
}


// This function is used to combine many different vMatchScores 
// objects. number tells how many vMatchScores objects 
// we are passing. After number we pass scores and weights. The
// argument list looks like 
// (number, scores1, weight1, scores2, weight2, ..., scores_n, weight_n)
// It is assumed (the function verifies that) that the frames and views
// stored in the i-th score in the frame_sorted array of all scores
// are identical. If not, 0 is returned.
vMatchScores * vMatchScores::CombineLogs(long number, ...)
{
  if (number <= 0) return 0;
  // Create the scores and weights arrays.
  vArray(vMatchScores *) scores_array = vnew(vMatchScores *, number);
  vArray(double) weights_array = vnew(double, number);

  // Parse the arguments, store them into the appropriate arrays.
  va_list arguments;
  long i;
  va_start(arguments, number);
  for(i = 0; i < number; i++)
  {
    scores_array[i] = va_arg(arguments, vMatchScores *);
    weights_array[i] = va_arg(arguments, double);
  }
  va_end(arguments);

  // Make sure all vMatchScores objects have the same number of scores.
  vMatchScores * result = CombineLogs2(number, scores_array, weights_array);
  vdelete2(scores_array);
  vdelete2(weights_array);
  return result;
}


vMatchScores * 
vMatchScores::CombineLogs2(long number, vArray(vMatchScores *) scores_array,
                            vArray(double) weights_array)
{
  if (number <= 0) return 0;
  // Create the scores and weights arrays.

  // Make sure all vMatchScores objects have the same number of scores.
  long matches_to_keep = scores_array[0]->NumberOfScores();
  long i;
  for (i = 0; i < number; i++)
  {
    long current_number = scores_array[i]->NumberOfScores();
    if (current_number != matches_to_keep)
    {
      vdelete2(scores_array);
      vdelete2(weights_array);
      return 0;
    }
  }

  vArray2(vMatchScore) scores = vnew(vArray(vMatchScore), number);
  for (i = 0; i < number; i++)
  {
    scores[i] = scores_array[i]->GetFrameSorted();
  }

    
  // Create the new array of scores and precompute logs.
  vMatrix<double> logs_matrix = vPrecalculateLogs3(matches_to_keep+1);
  vArray(double) logs = logs_matrix.Matrix();
  vArray(vMatchScore) new_array = vnew(vMatchScore, matches_to_keep+1);
  long success = 1;
  for (i = 0; i < matches_to_keep; i++)
  {
    vMatchScore score = scores[0][i];
    double log_sum = 0;
    for (long j = 0; j < number; j++)
    {
      vMatchScore current_score = scores[j][i];
      if (current_score.frame != score.frame) success = 0;
      if (current_score.view != score.view) success = 0;
      log_sum += weights_array[j] * logs[current_score.rank + 1];
    }
    score.score = (float) log_sum;
    new_array[i] = score;
  }

  vdelete2(scores);
  if (success == 0)
  {
    vdelete2(new_array);
    return 0;
  }
  else
  {
    vMatchScores * result = new vMatchScores(new_array, matches_to_keep);
    return result;
  }
}

  
vMatchScores * vMatchScores::CombineSums(long number, ...)
{
  if (number <= 0) return 0;
  // Create the scores and weights arrays.
  vArray(vMatchScores *) scores_array = vnew(vMatchScores *, number);
  vArray(double) weights_array = vnew(double, number);
  vArray2(vMatchScore) scores = vnew(vArray(vMatchScore), number);

  // Parse the arguments, store them into the appropriate arrays.
  va_list arguments;
  long i;
  va_start(arguments, number);
  for(i = 0; i < number; i++)
  {
    scores_array[i] = va_arg(arguments, vMatchScores *);
    weights_array[i] = va_arg(arguments, double);
    scores[i] = scores_array[i]->GetFrameSorted();
  }
  va_end(arguments);

  // Make sure all vMatchScores objects have the same number of scores.
  long matches_to_keep = scores_array[0]->NumberOfScores();
  for (i = 0; i < number; i++)
  {
    long current_number = scores_array[i]->NumberOfScores();
    if (current_number != matches_to_keep)
    {
      vdelete2(scores_array);
      vdelete2(weights_array);
      vdelete2(scores);
      return 0;
    }
  }

  vArray(vMatchScore) new_array = vnew(vMatchScore, matches_to_keep+1);
  vPrint("matches_to_keep = %li\n", matches_to_keep);
  long success = 1;
  for (i = 0; i < matches_to_keep; i++)
  {
    vPrint("class %li: ", i);
    vMatchScore score = scores[0][i];
    double sum = 0;
    for (long j = 0; j < number; j++)
    {
      vMatchScore current_score = scores[j][i];
      if (current_score.frame != score.frame) success = 0;
      if (current_score.view != score.view) success = 0;
      sum += weights_array[j] * current_score.rank;
      vPrint("(%.1lf, %li) ", weights_array[j], current_score.rank);
    }
    vPrint("\n");
    score.score = (float) sum;
    new_array[i] = score;
  }

  vdelete2(scores_array);
  vdelete2(weights_array);
  vdelete2(scores);
  if (success == 0)
  {
    vdelete2(new_array);
    return 0;
  }
  else
  {
    vMatchScores * result = new vMatchScores(new_array, matches_to_keep);
    return result;
  }
}


vMatchScores * vMatchScores::PETS_Filter1()
{
  vMatchScores * result = new vMatchScores();
  result->Initialize();
  
  result->number_of_scores = 10;
  result->scores = vnew(vMatchScore, result->number_of_scores+1);

  long index = 0;
  result->scores[index].frame = 431;
  result->scores[index].view = 0;
  index++;
  
  result->scores[index].frame = 21060;
  result->scores[index].view = 3;
  index++;
  
  result->scores[index].frame = 23746;
  result->scores[index].view = 0;
  index++;
  
  result->scores[index].frame = 3097;
  result->scores[index].view = 3;
  index++;
  
  result->scores[index].frame = 6623;
  result->scores[index].view = 3;
  index++;
  
  result->scores[index].frame = 9719;
  result->scores[index].view = 3;
  index++;
  
  result->scores[index].frame = 5591;
  result->scores[index].view = 0;
  index++;
  
  result->scores[index].frame = 7655;
  result->scores[index].view = 0;
  index++;
  
  result->scores[index].frame = 10740;
  result->scores[index].view = 3;
  index++;
  
  result->scores[index].frame = 8676;
  result->scores[index].view = 3;
  
  long i;
  for (i = 0; i < result->number_of_scores; i++)
  {
    result->scores[i].is_valid = 1;
  }

  result->StoreFrameSorted2();

  return result;
}


// Sometimes we have scores for a set of frames and another vMatchScores
// object has more accurate scores for a subset of those frames. 
// Merge() replaces the scores in scores2 with scores in the
// scores1 (that is passed in as an argument) when such scores exist.
vMatchScores * vMatchScores::Merge(vMatchScores * scores1, 
                                     vMatchScores * scores2)
{
  vArray(vMatchScore) frame_sorted1 = scores1->frame_sorted;
  long number1 = scores1->NumberOfScores();
  vArray(vMatchScore) frame_sorted2 = scores2->frame_sorted;
  long number2 = scores2->NumberOfScores();
  vArray(vMatchScore) new_scores = vnew(vMatchScore, number1+number2);
  long number_of_scores = 0;

  // Figure out the worst score (max score) in scores1, 
  // and the min score in scores2, so that
  // we can make sure all scores in scores2 are adjusted
  // to be worse than the scores in scores1.
  double worst_score1 = scores1->GetScore(number1 - 1).score;
  double best_score2 = scores2->GetScore(0).score;
  double adjustment = worst_score1 - best_score2 + 1;
  long i1 = 0;
  long i2 = 0;
  vMatchScore & score1 = frame_sorted1[i1];
  vMatchScore & score2 = frame_sorted2[i2];
  long frame1 = score1.frame;
  long frame2 = score2.frame;
  long view1 = score1.view;
  long view2 = score2.view;
  // Go through the two frame_sorted arrays, and put scores in 
  // a frame_sorted way into new_scores
  while(1)
  {
    if ((frame1 < frame2) || ((frame1 == frame2) && (view1 < view2)))
    {
      new_scores[number_of_scores] = score1;
      i1++;
      if (i1 >= number1) break;
      score1 = frame_sorted1[i1];
      frame1 = score1.frame;
      view1 = score1.view;
    }
    else if ((frame1 == frame2) && (view1 == view2))
    {
      new_scores[number_of_scores] = score1;
      i1++;
      i2++;
      if (i1 >= number1) break;
      if (i2 >= number2) break;
      score1 = frame_sorted1[i1];
      frame1 = score1.frame;
      view1 = score1.view;

      score2 = frame_sorted2[i2];
      frame2 = score2.frame;
      view2 = score2.view;
    }
    else
    {
      new_scores[number_of_scores] = score2;
      new_scores[number_of_scores].score += (float) adjustment;
      i2++;
      score2 = frame_sorted2[i2];
      frame2 = score2.frame;
      view2 = score2.view;
    }
    number_of_scores++;
  }

  number_of_scores++;
  while(i1 < number1)
  {
    new_scores[number_of_scores] = frame_sorted1[i1];
    i1++;
    number_of_scores++;
  }

  while(i2 < number2)
  {
    new_scores[number_of_scores] = frame_sorted2[i2];
    new_scores[number_of_scores].score += (float) adjustment;
    i2++;
    number_of_scores++;
  }

  vMatchScores * result = new vMatchScores(new_scores, number_of_scores);
  return result;
}



///////////////////////////////////////////////////////////////
//
// End of implementation of class vMatchScores
//
///////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////
//
// Beginning of implementation of class vFrameMatchStats
//
///////////////////////////////////////////////////////////////


vFrameMatchStats::vFrameMatchStats()
{  
  ground_distance = -1;
}



// correct_limit specifies the number of highest ranking correct
// matches that we want to preserve. ground_limit is the number
// of matches closest to the ground truth that we want to preserve.
vFrameMatchStats::vFrameMatchStats(vMatchScores * scores, 
                                     long in_correct_limit, 
                                     long in_ground_limit)
{
  is_valid = 0;
  ground_distance = -1;
  if (scores->IsValid() <= 0) return;
  if (scores->GroundTruthKnown() <= 0) return;

  // Get the number of correct matches;
  vMatchScore temp = scores->CorrectMatch(0, &number_of_correct_matches);

  long i;
  // Get the correct matches
  for (i = 0; i < in_correct_limit; i++)
  {
    vMatchScore score = scores->CorrectMatch(i);
    //if (score.is_valid <= 0) break;
    // some times the score can be invalid, if
    // there were no correct matches found at all.
    // We make those scores valid here.
    score.is_valid = 1;
    correct_matches.push_back(score);
  }

  // Get the ground matches
  for (i = 0; i < in_ground_limit; i++)
  {
    vMatchScore score = scores->GroundMatch(i);
    //if (score.is_valid <= 0) break;
    score.is_valid = 1;
    ground_matches.push_back(score);
  }
  long success = scores->GroundTruthView(&x, &y, &z);
  correct_to_print = correct_matches.size();
  ground_to_print = ground_matches.size();
  if (success <= 0) return;
  is_valid = 1;
}


vFrameMatchStats::vFrameMatchStats(vMatchScores * scores, long in_correct_limit, 
                                     long in_ground_limit, long keep_top)
{
  is_valid = 0;
  ground_distance = -1;
  if (scores->IsValid() <= 0) return;
  if (scores->GroundTruthKnown() <= 0) return;

  // Get the number of correct matches;
  vMatchScore temp = scores->CorrectMatch(0, &number_of_correct_matches);

  long i;
  // Get the correct matches
  for (i = 0; i < in_correct_limit; i++)
  {
    vMatchScore score = scores->CorrectMatch(i);
    //if (score.is_valid <= 0) break;
    // some times the score can be invalid, if
    // there were no correct matches found at all.
    // We make those scores valid here.
    score.is_valid = 1;
    correct_matches.push_back(score);
  }

  // Get the ground matches
  for (i = 0; i < in_ground_limit; i++)
  {
    vMatchScore score = scores->GroundMatch(i);
    //if (score.is_valid <= 0) break;
    score.is_valid = 1;
    ground_matches.push_back(score);
  }

  for (i = 0; i < keep_top; i++)
  {
    vMatchScore score = scores->GetScore(i);
    if (score.is_valid <= 0) break;
    top_scores.push_back(score);
  }

  long success = scores->GroundTruthView(&x, &y, &z);
  correct_to_print = correct_matches.size();
  ground_to_print = ground_matches.size();
  top_to_print = 5;
  if (success <= 0) return;
  is_valid = 1;
}


// This constructor basically collects the scores of the views stored
// in other_stats. We assume that other_stats
// was created using some other scores object, and now we just want
// to see how those selected views do given these scores.
vFrameMatchStats::vFrameMatchStats(vMatchScores * scores, 
                                     vFrameMatchStats * other_stats)
{
  ground_distance = -1;
  long i;

  // Get correct matches;
  i = 0;
  while(1)
  {
    vMatchScore score = other_stats->CorrectMatch(i);
    //if (score.is_valid <= 0) break;
    score.is_valid = 1;
    i++;
    vMatchScore other_score = scores->GetRank(score.frame, score.view);
    correct_matches.push_back(other_score);
  }

  // Get ground matches;
  i = 0;
  while(1)
  {
    vMatchScore score = other_stats->GroundMatch(i);
    if (score.is_valid <= 0) break;
    i++;
    vMatchScore other_score = scores->GetRank(score.frame, score.view);
    ground_matches.push_back(other_score);
  }

  // Get top scores;
  i = 0;
  while(1)
  {
    vMatchScore score = other_stats->TopScore(i);
    if (score.is_valid <= 0) break;
    i++;
    vMatchScore top_score = scores->GetRank(score.frame, score.view);
    top_scores.push_back(top_score);
  }
  long success = scores->GroundTruthView(&x, &y, &z);
  correct_to_print = correct_matches.size();
  ground_to_print = ground_matches.size();
  if (success <= 0) return;
  is_valid = 1;
}


vFrameMatchStats::~vFrameMatchStats()
{
}


long vFrameMatchStats::SetGroundDistance(long in_ground_distance)
{
  if (in_ground_distance < 0) return 0;
  if (in_ground_distance > 180) return 0;
  ground_distance = in_ground_distance;
  number_of_correct_matches = 0;
  long correct_size = correct_matches.size();
  long i;
  for (i = 0; i < correct_size; i++)
  {
    if (correct_matches[i].ground_distance <= ground_distance)
    {
      number_of_correct_matches++;
    }
  }

  return 1;
}

  
vMatchScore vFrameMatchStats::CorrectMatch(long index)
{
  long correct_size = correct_matches.size();
  long i;
  long counter = -1;
  for (i = 0; i < (long) correct_matches.size(); i++)
  {
    vMatchScore & current = correct_matches[i];
    if (current.ground_distance <= ground_distance) counter++;
    if (counter == index)
    {
      return current;
    }
  }

//  if ((index < 0) || (index >= correct_matches.size())) return vMatchScore();
//  return correct_matches[index];
  return vMatchScore();
}



string vFrameMatchStats::PrintToString()
{
  string result;
  char buffer[400];
  if (correct_matches.size() > 0)
  {
    sprintf(buffer, "%li correct matches: \r\n", number_of_correct_matches);
    result.append(buffer);
    long i;
    for (i = 0; i < correct_to_print; i++)
    {
      char * number_string = vLongToString(i);
      result.append(number_string);
      vdelete2(number_string);
      result.append(": ");
      result.append(CorrectMatch(i).PrintToString());
    }
  }

  if (ground_matches.size() > 0)
  {
    result.append("Ground matches: \r\n");
    long number = ground_matches.size();
    long i;
    for (i = 0; i < ground_to_print; i++)
    {
      char * number_string = vLongToString(i);
      result.append(number_string);
      vdelete2(number_string);
      result.append(": ");
      result.append(ground_matches[i].PrintToString());
    }
  }

  if (top_scores.size() > 0)
  {
    result.append("Top scores: \r\n");
    long number = top_scores.size();
    number = Min(number, top_to_print);
    long i;
    for (i = 0; i < number; i++)
    {
      char * number_string = vLongToString(i);
      result.append(number_string);
      vdelete2(number_string);
      result.append(": ");
      result.append(top_scores[i].PrintToString2());
    }
  }
  result.append("\r\n");
  return result;
}


long vFrameMatchStats::Print()
{
  string result = PrintToString();
  vPrint("%s", result.c_str());
  return 1;
}


long vFrameMatchStats::PrintToFile(const char * filename)
{
  FILE * fp = fopen(filename, vFOPEN_WRITE);
  if (fp == 0) return 0;
  long result = PrintToFile(fp);
  fclose(fp);
  return result;
}


long vFrameMatchStats::PrintToFile(FILE * fp)
{
  string text = PrintToString();
  long bytes = fprintf(fp, "%s", text.c_str());
  if (bytes >= 0) return 1;
  else return bytes;
}


long vFrameMatchStats::Save(FILE * fp)
{
  long correct_number = correct_matches.size();
  long ground_number = ground_matches.size();
  long top_number = top_scores.size();

  long items = 0;
  items += vWritePCLongs(fp, &correct_number, 1);
  items += vWritePCLongs(fp, &ground_number, 1);
  items += vWritePCLongs(fp, &top_number, 1);
  items += vWritePCLongs(fp, &number_of_correct_matches, 1);
  items += vWritePCLongs(fp, &is_valid, 1);
  items += vWritePCLongs(fp, &correct_to_print, 1);
  items += vWritePCLongs(fp, &ground_to_print, 1);

  items += vWritePCDoubles(fp, &x, 1);
  items += vWritePCDoubles(fp, &y, 1);
  items += vWritePCDoubles(fp, &z, 1);
  
  if (items != 10) return 0;
  long i, success;
  for (i = 0; i < correct_number; i++)
  {
    success = correct_matches[i].Save(fp);
    if (success == 0) return 0;
  }

  for (i = 0; i < ground_number; i++)
  {
    success = ground_matches[i].Save(fp);
    if (success == 0) return 0;
  }

  for (i = 0; i < top_number; i++)
  {
    success = top_scores[i].Save(fp);
    if (success == 0) return 0;
  }

  return 1;
}


long vFrameMatchStats::Load(FILE * fp)
{
  long correct_number, ground_number, top_number;
  number_of_correct_matches = correct_number = ground_number = top_number = 0;

  long items = 0;
  items += vReadPCLongs(fp, &correct_number, 1);
  items += vReadPCLongs(fp, &ground_number, 1);
  items += vReadPCLongs(fp, &top_number, 1);
  items += vReadPCLongs(fp, &number_of_correct_matches, 1);
  items += vReadPCLongs(fp, &is_valid, 1);
  items += vReadPCLongs(fp, &correct_to_print, 1);
  items += vReadPCLongs(fp, &ground_to_print, 1);

  items += vReadPCDoubles(fp, &x, 1);
  items += vReadPCDoubles(fp, &y, 1);
  items += vReadPCDoubles(fp, &z, 1);

  // is_vaild will be set to 1 if it was 1 to begin with and the 
  // object was read successfully.
  long previous_is_valid = 1;
  is_valid = 0;
  if (items != 10) return 0;
  
  correct_matches = vector<vMatchScore>(correct_number);
  ground_matches = vector<vMatchScore>(ground_number);
  top_scores = vector<vMatchScore>(top_number);

  long i, success;
  for (i = 0; i < correct_number; i++)
  {
    success = correct_matches[i].Load(fp);
    if (success == 0) return 0;
  }

  for (i = 0; i < ground_number; i++)
  {
    success = ground_matches[i].Load(fp);
    if (success == 0) return 0;
  }

  for (i = 0; i < top_number; i++)
  {
    success = top_scores[i].Load(fp);
    if (success == 0) return 0;
  }

  is_valid = 1;
  return 1;
}

///////////////////////////////////////////////////////////////
//
// End of implementation of class vFrameMatchStats
//
///////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////
//
// Beginning of implementation of class vSequenceMatchStats
//
///////////////////////////////////////////////////////////////


vSequenceMatchStats::vSequenceMatchStats()
{
  top_to_print = 5;
}


vSequenceMatchStats::~vSequenceMatchStats()
{
  long frames = frame_stats.size();
  long i;
  for (i = 0; i < frames; i++)
  {
    vdelete(frame_stats[i]);
  }
}


long vSequenceMatchStats::AddFrameStats(vFrameMatchStats * new_frame_stats,
                                         long frame_number)
{
  frame_stats.push_back(new_frame_stats);
  frame_numbers.push_back(frame_number);
  selected.push_back(1);
  return 1;
}


// mean rank of the index-th highest ranking correct matches
// across frames. -1 if a correct match of such a rank was
// not found in every single frame, or if there are no frames.
double vSequenceMatchStats::CorrectMean(long index)
{
  long number = frame_numbers.size();
  if (number == 0) return -1;
  long frame;
  double sum = (double) 0;
  double counter = 0;
  for (frame = 0; frame < number; frame++)
  {
    if (selected[frame] != 1) continue;
    counter++;
    vMatchScore score = frame_stats[frame]->CorrectMatch(index);
    if (score.is_valid <= 0)
    {
      return -1;
    }

    sum += score.rank;
  }

  double result = -1;
  if (counter > 0)
  {
    result = sum / counter;
  }
  return result;
}


// median rank of the index-th highest ranking correct matches
// across frames
long vSequenceMatchStats::CorrectMedian(long index)
{
  long number = frame_numbers.size();
  if (number == 0) return -1;
  long frame;
  vector<long> ranks;
  long counter = 0;
  for (frame = 0; frame < number; frame++)
  {
    if (selected[frame] != 1) continue;
    vMatchScore score = frame_stats[frame]->CorrectMatch(index);
    if (score.is_valid <= 0)
    {
      return -1;
    }
    else
    {
      ranks.push_back(score.rank);
    }
    counter++;
  }

  if (counter == 0) return -1;
  std::sort(ranks.begin(), ranks.end(), less<long>());
  return ranks[counter / 2];
}


// Number of correct matches of the given index whose rank 
// was below the given rank. Returns -1 if, for one frame,
// a correct match of the given index was not found
long vSequenceMatchStats::CorrectMatchesUnder(long index, long rank)
{
  long number = frame_numbers.size();
  if (number == 0) return -1;
  long frame;
  long counter = 0;
  for (frame = 0; frame < number; frame++)
  {
    if (selected[frame] != 1) continue;
    vMatchScore score = frame_stats[frame]->CorrectMatch(index);
    if (score.is_valid <= 0)
    {
      return -1;
    }
    if (score.rank < rank) counter++;  
  }

  return counter;
}


// Fraction of correct matches of the given index whose rank 
// was below the given rank
double vSequenceMatchStats::PcorrectMatchesUnder(long index, long rank)
{
  long number = frame_numbers.size();
  if (number == 0) return -1;
  long frame;
  double counter = 0;
  double selected_counter = 0;
  for (frame = 0; frame < number; frame++)
  {
    if (selected[frame] != 1) continue;
    selected_counter++;
    vMatchScore score = frame_stats[frame]->CorrectMatch(index);
    if (score.is_valid <= 0)
    {
      return -1;
    }
    if (score.rank < rank) counter++;  
  }

  if (selected_counter == 0) return -1;
  return counter/selected_counter;
}


// Number of correct matches of the given index that fell between
// low rank and high rank and whose ground_truth distance was
// between low_distance and high_distance.
long vSequenceMatchStats::CorrectMatchesWithin(long index, long low_rank, long high_rank,
                                                long low_distance, long high_distance)
{
  long number = frame_numbers.size();
  if (number == 0) return -1;
  long frame;
  long counter = 0;
  for (frame = 0; frame < number; frame++)
  {
    if (selected[frame] != 1) continue;
    vMatchScore score = frame_stats[frame]->CorrectMatch(index);
    if (score.is_valid <= 0)
    {
      return -1;
    }
    if ((score.rank >= low_rank) && (score.rank <= high_rank) &&
        (score.ground_distance >= low_distance) && 
        (score.ground_distance <= high_distance))
    {
      counter++;  
    }
  }

  return counter;
}

  
// mean rank of the index-th highest ranking ground matches
// across frames. -1 if a correct match of such a rank was
// not found in every single frame, or if there are no frames.
double vSequenceMatchStats::GroundMean(long index)
{
  long number = frame_numbers.size();
  if (number == 0) return -1;
  long frame;
  double sum = (double) 0;
  double counter = 0;
  for (frame = 0; frame < number; frame++)
  {
    if (selected[frame] != 1) continue;
    counter++;
    vMatchScore score = frame_stats[frame]->GroundMatch(index);
    if (score.is_valid <= 0)
    {
      return -1;
    }

    sum += score.rank;
  }

  double result = -1;
  if (counter > 0)
  {
    result = sum / counter;
  }
  return result;
}


// mean rank of the index-th highest ranking correct matches
// across frames. Returns -1 if, for one of the frames, a 
// ground match for the given index could not be found.
long vSequenceMatchStats::GroundMedian(long index)
{
  long number = frame_numbers.size();
  if (number == 0) return -1;
  long frame;
  vector<long> ranks;
  long counter = 0;
  for (frame = 0; frame < number; frame++)
  {
    if (selected[frame] != 1) continue;
    vMatchScore score = frame_stats[frame]->GroundMatch(index);
    if (score.is_valid <= 0)
    {
      return -1;
    }
    else
    {
      ranks.push_back(score.rank);
    }
    counter++;
  }

  if (counter == 0) return -1;
  std::sort(ranks.begin(), ranks.end(), less<long>());
  return ranks[counter / 2];
}

// Number of ground matches of the given index whose rank 
// was below the given rank
long vSequenceMatchStats::GroundMatchesUnder(long index, long rank)
{
  long number = frame_numbers.size();
  if (number == 0) return -1;
  long frame;
  long counter = 0;
  for (frame = 0; frame < number; frame++)
  {
    if (selected[frame] != 1) continue;
    vMatchScore score = frame_stats[frame]->GroundMatch(index);
    if (score.is_valid <= 0)
    {
      return -1;
    }
    if (score.rank < rank) counter++;  
  }

  return counter;
}


long vSequenceMatchStats::SelectAll()
{
  long number_of_frames = frame_stats.size();
  selected = vector<long>(number_of_frames);
  long i;
  for (i = 0; i < number_of_frames; i++)
  {
    selected[i] = 1;
  }
  return 1;
}

// Selects frames whose ground truth viewpoint is within
// tolerance of the viewpoint specified by x and y.
long vSequenceMatchStats::SelectViewpoint(double x, double y, double tolerance)
{
  long number_of_frames = frame_stats.size();
  selected = vector<long>(number_of_frames);
  long i;
  for (i = 0; i < number_of_frames; i++)
  {
    double frame_x, frame_y, frame_z;
    frame_x = frame_y = frame_z = 0;
    frame_stats[i]->GroundTruthView(&frame_x, &frame_y, &frame_z);
    double distance1 = vOldViewDistance(x, y, 0, frame_x, frame_y, 0);
    double distance2 = vOldViewDistance(x, y+180, 0, frame_x, frame_y, 0);
    if (Min(distance1, distance2) <= tolerance)
    {
      selected[i] = 1;
    }
    else
    {
      selected[i] = 0;
    }
  }
  return 1;
}


// Selects frame whose ground truth viewpoint is at
// distance from frontal view, +- tolerance.
long vSequenceMatchStats::SelectFrontDistance(double distance, double tolerance)
{
  long number_of_frames = frame_stats.size();
  selected = vector<long>(number_of_frames);
  long i;
  for (i = 0; i < number_of_frames; i++)
  {
    double frame_x, frame_y, frame_z;
    frame_x = frame_y = frame_z = 0;
    frame_stats[i]->GroundTruthView(&frame_x, &frame_y, &frame_z);
    double distance1 = vOldViewDistance(90, 0, 0, frame_x, frame_y, 0);
    double distance2 = vOldViewDistance(90, 180, 0, frame_x, frame_y, 0);
    double front_distance = Min(distance1, distance2);
    if (vAbs(front_distance - distance) <= tolerance)
    {
      selected[i] = 1;
    }
    else
    {
      selected[i] = 0;
    }
  }
  return 1;
}


// Selects frame whose ground truth viewpoint is at
// distance from frontal view, +- tolerance.
double vSequenceMatchStats::GetFrontDistance(long frame)
{
  if (frame < 0) return -1;
  long number_of_frames = frame_stats.size();
  if (frame >= number_of_frames) return -1;
  double frame_x, frame_y, frame_z;
  frame_x = frame_y = frame_z = 0;
  frame_stats[frame]->GroundTruthView(&frame_x, &frame_y, &frame_z);
  double distance1 = vOldViewDistance(90, 0, 0, frame_x, frame_y, 0);
  double distance2 = vOldViewDistance(90, 180, 0, frame_x, frame_y, 0);
  double front_distance = Min(distance1, distance2);
  return front_distance;
}


long vSequenceMatchStats::SetCorrectToPrint(long in_correct_to_print)
{
  long frames = frame_stats.size();
  long i;
  for (i = 0; i < frames; i++)
  {
    frame_stats[i]->SetCorrectToPrint(in_correct_to_print);
  }
  return 1;
}


// Selects frames between start_frame and end_frame.
long vSequenceMatchStats::SelectRange(long start_frame, long end_frame)
{
  long i;
  long number_of_frames = frame_stats.size();
  if (start_frame < 0) start_frame = 0;
  if (end_frame >= number_of_frames) end_frame = number_of_frames - 1;
  for (i = 0; i < number_of_frames; i++)
  {
    selected[i] = 0;
  }

  for (i = start_frame; i <= end_frame; i++)
  {
    selected[i] = 1;
  }
  return 1;
}


// Selects frames between start and end frame, and frontal distance within 
// tolerance of distance. A combiation of SelectRange and SelectFrontDistance.
long vSequenceMatchStats::SelectRangeFrontDistance(long start_frame, long end_frame,
                                                    double distance, double tolerance)
{
  SelectRange(start_frame, end_frame);
  long i;
  long number_of_frames = frame_stats.size();
  for (i = 0; i < number_of_frames; i++)
  {
    if (selected[i] == 0) continue;
    double frame_x, frame_y, frame_z;
    frame_x = frame_y = frame_z = 0;
    frame_stats[i]->GroundTruthView(&frame_x, &frame_y, &frame_z);
    double distance1 = vOldViewDistance(90, 0, 0, frame_x, frame_y, 0);
    double distance2 = vOldViewDistance(90, 180, 0, frame_x, frame_y, 0);
    double front_distance = Min(distance1, distance2);
    if (vAbs(front_distance - distance) <= tolerance)
    {
      selected[i] = 1;
    }
    else
    {
      selected[i] = 0;
    }
  }
  return 1;
}


  
long vSequenceMatchStats::Print()
{
  long frames = frame_stats.size();
  long i;
  double correct_numbers = 0;
  long counter = 0;
  for (i = 0; i < frames; i++)
  {
    if (selected[i] != 1) continue;
    counter++;
    long frame_number = frame_numbers[i];
    vPrint("frame %li: \r\n", frame_number);
    frame_stats[i]->SetTopToPrint(top_to_print);
    frame_stats[i]->Print();
    correct_numbers += frame_stats[i]->NumberOfCorrectMatches();
  }

  vPrint("\r\n");
  vPrint("%li frames were evaluated\n", counter);
  vPrint("Average number of correct matches: %.2f\r\n", correct_numbers / (double) counter);
  vPrint("Correct mean: %lf\r\n", CorrectMean(0));
  vPrint("Correct median: %li\r\n", CorrectMedian(0));
  vPrint("Ground mean: %lf\r\n", GroundMean(0));
  vPrint("Ground median: %li\r\n\n", GroundMedian(0));

  long correct_matches1 = CorrectMatchesUnder(0, 1);
  double correct_fraction1 = 100 * ((double) correct_matches1) / (double) counter;
  long correct_matches10 = CorrectMatchesUnder(0, 10);
  double correct_fraction10 = 100 * ((double) correct_matches10) / (double) counter;
  long correct_matches100 = CorrectMatchesUnder(0, 100);
  double correct_fraction100 = 100 * ((double) correct_matches100) / (double) counter;
  long correct_matches1000 = CorrectMatchesUnder(0, 1000);
  double correct_fraction1000 = 100 * ((double) correct_matches1000) / (double) counter;
  long correct_matches10000 = CorrectMatchesUnder(0, 10000);
  double correct_fraction10000 = 100 * ((double) correct_matches10000) / (double) counter;

  long ground_matches10 = GroundMatchesUnder(0, 10);
  double ground_fraction10 = 100 * ((double) ground_matches10) / (double) counter;
  long ground_matches100 = GroundMatchesUnder(0, 100);
  double ground_fraction100 = 100 * ((double) ground_matches100) / (double) counter;
  long ground_matches1000 = GroundMatchesUnder(0, 1000);
  double ground_fraction1000 = 100 * ((double) ground_matches1000) / (double) counter;
  long ground_matches10000 = GroundMatchesUnder(0, 10000);
  double ground_fraction10000 = 100 * ((double) ground_matches10000) / (double) counter;

  vPrint("Correct matches under 1     : %.1lf%%, %li\n", 
          correct_fraction1, correct_matches1);
  vPrint("Correct matches under 10    : %.1lf%%, %li (%li)\n", 
          correct_fraction10, correct_matches10 - correct_matches1,
          correct_matches10);
  vPrint("Correct matches under 100   : %.1lf%%, %li (%li)\n", 
          correct_fraction100, correct_matches100 - correct_matches10,
          correct_matches100);
  vPrint("Correct matches under 1000  : %.1lf%%, %li (%li)\n", 
           correct_fraction1000, correct_matches1000 - correct_matches100,
           correct_matches1000);
  vPrint("Correct matches under 10000 : %.1lf%%, %li (%li)\n", 
          correct_fraction10000, correct_matches10000 - correct_matches1000,
          correct_matches10000);
  vPrint("Correct matches over 10000  : %.1lf%%, %li\n\n",
          100 - correct_fraction10000, counter - correct_matches10000);

  vPrint("Ground matches under 10     : %.1lf%%, %li\n", 
          ground_fraction10, ground_matches10);
  vPrint("Ground matches under 100    : %.1lf%%, %li (%li)\n", 
          ground_fraction100, ground_matches100 - ground_matches10,
          ground_matches100);
  vPrint("Ground matches under 1000   : %.1lf%%, %li (%li)\n", 
          ground_fraction1000, ground_matches1000 - ground_matches100,
          ground_matches1000);
  vPrint("Ground matches under 10000  : %.1lf%%, %li (%li)\n", 
          ground_fraction10000, ground_matches10000 - ground_matches1000,
          ground_matches10000);
  vPrint("Ground matches over 10000   : %.1lf%%, %li\n\n",
          100 - ground_fraction10000, counter - ground_matches10000);

  vPrint("10-correct matches under 1000: %li\n",
           CorrectMatchesUnder(9, 1000));
  vPrint("15-correct matches under 1000: %li\n",
           CorrectMatchesUnder(14, 1000));
  vPrint("20-correct matches under 1000: %li\n",
           CorrectMatchesUnder(19, 1000));
  vPrint("10-correct median: %li\n", CorrectMedian(9));
  vPrint("15-correct median: %li\n", CorrectMedian(14));
  vPrint("20-correct median: %li\n", CorrectMedian(19));

  return 1;
}


string vSequenceMatchStats::PrintToString()
{
  string result;
  long frames = frame_stats.size();
  char buffer[400];
  double correct_numbers = 0;
  long i;
  long counter = 0;
  for (i = 0; i < frames; i++)
  {
    if (selected[i] != 1) continue;
    counter++;
    long frame_number = frame_numbers[i];
    sprintf(buffer, "frame %li: \r\n", frame_number);
    result.append(buffer);
    result.append(frame_stats[i]->PrintToString());
    correct_numbers += frame_stats[i]->NumberOfCorrectMatches();
  }

  sprintf(buffer, "\r\n");
  result.append(buffer);
  sprintf(buffer, "%li frames were evaluated\n", counter);
  result.append(buffer);
  sprintf(buffer, "Average number of correct matches: %.2f\r\n", 
          correct_numbers / (double) counter);
  result.append(buffer);
  sprintf(buffer, "Correct mean: %lf\r\n", CorrectMean(0));
  result.append(buffer);
  sprintf(buffer, "Correct median: %li\r\n", CorrectMedian(0));
  result.append(buffer);
  sprintf(buffer, "Ground mean: %lf\r\n", GroundMean(0));
  result.append(buffer);
  sprintf(buffer, "Ground median: %li\r\n\n", GroundMedian(0));
  result.append(buffer);

  long correct_matches10 = CorrectMatchesUnder(0, 10);
  long correct_matches100 = CorrectMatchesUnder(0, 100);
  long correct_matches1000 = CorrectMatchesUnder(0, 1000);
  long correct_matches10000 = CorrectMatchesUnder(0, 10000);

  long ground_matches10 = GroundMatchesUnder(0, 10);
  long ground_matches100 = GroundMatchesUnder(0, 100);
  long ground_matches1000 = GroundMatchesUnder(0, 1000);
  long ground_matches10000 = GroundMatchesUnder(0, 10000);

  sprintf(buffer, "Correct matches under 10    : %li\n", 
          correct_matches10);
  result.append(buffer);
  sprintf(buffer, "Correct matches under 100   : %li\n", 
          correct_matches100 - correct_matches10);
  result.append(buffer);
  sprintf(buffer, "Correct matches under 1000  : %li\n", 
          correct_matches1000 - correct_matches100);
  result.append(buffer);
  sprintf(buffer, "Correct matches under 10000 : %li\n", 
          correct_matches10000 - correct_matches1000);
  result.append(buffer);
  sprintf(buffer, "Correct matches over 10000  : %li\n\n",
          counter - correct_matches10000);
  result.append(buffer);

  sprintf(buffer, "Ground matches under 10     : %li\n", 
          ground_matches10);
  result.append(buffer);
  sprintf(buffer, "Ground matches under 100    : %li\n", 
          ground_matches100 - ground_matches10);
  result.append(buffer);
  sprintf(buffer, "Ground matches under 1000   : %li\n", 
          ground_matches1000 - ground_matches100);
  result.append(buffer);
  sprintf(buffer, "Ground matches under 10000  : %li\n", 
          ground_matches10000 - ground_matches1000);
  result.append(buffer);
  sprintf(buffer, "Ground matches over 10000   : %li\n\n",
          counter - ground_matches10000);
  result.append(buffer);

  return result;
}


long vSequenceMatchStats::PrintToFile(const char * filename)
{
  FILE * fp = fopen(filename, vFOPEN_WRITE);
  if (fp == 0) return 0;
  long result = PrintToFile(fp);
  fclose(fp);
  return result;
}

  
long vSequenceMatchStats::PrintToFile(FILE * fp)
{
  string text = PrintToString();
  long bytes = fprintf(fp, "%s", text.c_str());
  if (bytes >= 0) return 1;
  else return bytes;
}


long vSequenceMatchStats::SetGroundDistance(long ground_distance)
{
  if (ground_distance < 0) return 0;
  if (ground_distance > 180) return 0;
  long frame_stats_size = frame_stats.size();
  long i;
  for (i = 0; i < frame_stats_size; i++)
  {
    frame_stats[i]->SetGroundDistance(ground_distance);
  }
  return 1;
}


long vSequenceMatchStats::Save(const char * filename)
{
  // First make sure the file doesn't exist. This function is not allowed
  // to overwrite existing files
  FILE * fp = fopen(filename, vFOPEN_READ);
  if (fp != 0)
  {
    fclose(fp);
    vPrint("%s already exists\n", filename);
    return 0;
  }

  fp = fopen(filename, vFOPEN_WRITE);
  long success = Save(fp);
  fclose(fp);
  return success;
}


long vSequenceMatchStats::Save(FILE * fp)
{
  long frame_stats_size = frame_stats.size();
  long frame_numbers_size = frame_numbers.size();
  long selected_size = selected.size();
  
  long items = 0;
  items += vWritePCLongs(fp, &frame_stats_size, 1);
  items += vWritePCLongs(fp, &frame_numbers_size, 1);
  items += vWritePCLongs(fp, &selected_size, 1);
  if (items != 3) return 0;

  long i;
  for (i = 0; i < frame_stats_size; i++)
  {
    long success = frame_stats[i]->Save(fp);
    if (success <= 0) return success;
  }

  for (i = 0; i < frame_numbers_size; i++)
  {
    items = vWritePCLongs(fp, &(frame_numbers[i]), 1);
    if (items != 1) return 0;
  }

  for (i = 0; i < selected_size; i++)
  {
    items = vWritePCLongs(fp, &(selected[i]), 1);
    if (items != 1) return 0;
  }

  return 1;
}


long vSequenceMatchStats::Load(const char * filename)
{
  FILE * fp = fopen(filename, vFOPEN_READ);
  long success = Load(fp);
  if (fp == 0) return 0;
  fclose(fp);
  return success;
}


long vSequenceMatchStats::Load(FILE * fp)
{
  if (fp == 0) return 0;
  long frame_stats_size, frame_numbers_size, selected_size;
  frame_stats_size = frame_numbers_size = selected_size = 0;
  long items = 0;
  items += vReadPCLongs(fp, &frame_stats_size, 1);
  items += vReadPCLongs(fp, &frame_numbers_size, 1);
  items += vReadPCLongs(fp, &selected_size, 1);
  if (items != 3) return 0;

  frame_stats = vector<vFrameMatchStats *>(frame_stats_size);
  frame_numbers = vector<long>(frame_numbers_size);
  selected = vector<long>(selected_size);

  long i;
  for (i = 0; i < frame_stats_size; i++)
  {
    frame_stats[i] = new vFrameMatchStats();
    long success = frame_stats[i]->Load(fp);
    if (success <= 0) return success;
  }

  for (i = 0; i < frame_numbers_size; i++)
  {
    items = vReadPCLongs(fp, &(frame_numbers[i]), 1);
    if (items != 1) return 0;
  }

  for (i = 0; i < selected_size; i++)
  {
    items = vReadPCLongs(fp, &(selected[i]), 1);
    if (items != 1) return 0;
  }

  return 1;
}


long vSequenceMatchStats::NumberOfFrames()
{
  long number1 = frame_stats.size();
  long number2 = frame_numbers.size();
  if (number1 != number2)
  {
    exit_error("Error: number1 != number2\n, %li %li",
                    number1, number2);
  }
  return number1;
}


// returns the frame stats for which the corresponding
// entry in frame_numbers is frame.
vFrameMatchStats * vSequenceMatchStats::FrameStats(long frame)
{
  long number = frame_numbers.size();
  long i;
  long index = -1;
  for (i = 0; i < number; i++)
  {
    if (frame == frame_numbers[i])
    {
      index = i;
      break;
    }
  }
  if (index == -1)
  {
    return 0;
  }
  else
  {
    return frame_stats[i];
  }
}

  
// returns frame_stats[index].
vFrameMatchStats * vSequenceMatchStats::FrameStats2(long index)
{
  long number = frame_numbers.size();
  if (index < 0) return 0;
  if (index >= number) return 0;
  return frame_stats[index];
}

  
// Removes all vFrameMatchStats objects from the vectors,
// without calling their destructor. This is useful when
// we move those objects from one vSequenceMatchStats object
// to another.
long vSequenceMatchStats::RemoveAllStats()
{
  frame_stats.erase(frame_stats.begin(), frame_stats.end());
  frame_numbers.erase(frame_numbers.begin(), frame_numbers.end());
  selected.erase(selected.begin(), selected.end());
  return 1;
}

  
///////////////////////////////////////////////////////////////
//                                                           //
// End of implementation of class vSequenceMatchStats       //
//                                                           //
///////////////////////////////////////////////////////////////


vCombinedScores::vCombinedScores()
{
}


vCombinedScores::vCombinedScores(long in_number_of_scores, long in_test_frames, 
                                   long in_training_frames)
{
  scores = in_number_of_scores;
  test_frames = in_test_frames;
  training_frames = in_training_frames;
  ranks = v3dMatrix<long>(test_frames, training_frames, scores+1);
  vWriteValue(&ranks, (long) 0);
 // distances = v3dMatrix<float>(number_of_scores, test_frames, training_frames);
}


vCombinedScores::~vCombinedScores()
{
}


long vCombinedScores::EnterScores(vMatchScores * match_scores, long score_index, long test_index)
{
  vArray(vMatchScore) frame_sorted = match_scores->GetFrameSorted();
  vArray(long) ranks1 = ranks.Matrix3()[score_index][test_index];
  //vArray(float) distances1 = distances.Matrix3()[score][test_index];
  long i;
  for (i = 0; i < training_frames; i++)
  {
    vMatchScore & score = frame_sorted[i];
    ranks1[i] = score.rank;
    //distances1[i] = (float) score.ground_distance);
  }
  return 1;
}


long vCombinedScores::Load(const char * filename)
{
  FILE * fp = fopen(filename, vFOPEN_READ);
  long items = vReadPCLongs(fp, &scores, 1);
  items += vReadPCLongs(fp, &test_frames, 1);
  items += vReadPCLongs(fp, &training_frames, 1);

  if (items != 3)
  {
    vPrint("Error in reading %s\n", filename);
    fclose(fp);
    return 0;
  }

  // We must create a band_order vector to pass to ReadBandwise, because we 
  // have one extra band for which there is no data.
  vector<long> band_order(scores);
  long i;
  for (i = 0; i < scores; i++)
  {
    band_order[i] = i;
  }

  long success1 = ranks.ReadBandwise(fp, &band_order);
  long success2 = 1;
//  long success2 = distances.ReadBandwise(fp);
  long result = 1;
  if ((success1 <= 0) || (success2 <= 0))
  {
    vPrint("Error, couldn't read info from %s\n");
    result = 0;
  }
  fclose(fp);
  return result;

}


long vCombinedScores::Save(const char * filename)
{
  FILE * fp = fopen(filename, vFOPEN_WRITE);
  long items = vWritePCLongs(fp, &scores, 1);
  items += vWritePCLongs(fp, &test_frames, 1);
  items += vWritePCLongs(fp, &training_frames, 1);

  if (items != 3)
  {
    vPrint("Error in writing %s\n", filename);
    fclose(fp);
    return 0;
  }

  long success1 = ranks.WriteBandwise(fp);
  long success2 = 1;
  //long success2 = distances.WriteBandwise(fp);
  long result = 1;
  if ((success1 <= 0) || (success2 <= 0))
  {
    vPrint("Error, couldn't write info to %s\n");
    result = 0;
  }
  fclose(fp);
  return result;
}


vMatchScores * vCombinedScores::CombineSums(vMatrix<double> * weights_matrix, 
                                              long test_frame)
{
  if (weights_matrix->Size() != scores) return 0;
  vArray(double) weights = weights_matrix->Matrix();

  vMatrix<double> combinations_matrix(1, training_frames);
  vArray(double) combinations = combinations_matrix.Matrix();
  vArray3(long) ranks3 = ranks.Matrix3();
  long i;
  for (i = 0; i < training_frames; i++)
  {
    double rank_sum = 0;
    for (long j = 0; j < scores; j++)
    {
      rank_sum += weights[j] * ranks3[j][test_frame][i];
    }
    combinations[i] = rank_sum;
  }

  vMatchScores * result = new vMatchScores(&combinations_matrix);
  return result;
}


vMatchScores * vCombinedScores::CombineRoots(vMatrix<double> * weights_matrix,
                                               long test_frame)
{
  static vMatrix<double> root_matrix = vPrecalculateRoots2(training_frames);
  static vArray(double) roots = root_matrix.Matrix();
  if (root_matrix.Size() < training_frames)
  {
    exit_error("Error in vCombinedScores::CombineRoots\n");
  }

  if (weights_matrix->Size() != scores) return 0;
  vArray(double) weights = weights_matrix->Matrix();

  vMatrix<double> combinations_matrix(1, training_frames);
  vArray(double) combinations = combinations_matrix.Matrix();
  vArray3(long) ranks3 = ranks.Matrix3();
  long i;
  for (i = 0; i < training_frames; i++)
  {
    double rank_sum = 0;
    for (long j = 0; j < scores; j++)
    {
      long rank = ranks3[j][test_frame][i];
      double root = roots[rank];
      rank_sum += weights[j] * root;
    }
    combinations[i] = rank_sum;
  }

  vMatchScores * result = new vMatchScores(&combinations_matrix);
  return result;
}


vMatchScores * vCombinedScores::CombineLogs(vMatrix<double> * weights_matrix, 
                                              long test_frame)
{
  static vMatrix<double> log_matrix = vPrecalculateLogs2(training_frames);
  static vArray(double) logs = log_matrix.Matrix();
  if (log_matrix.Size() < training_frames)
  {
    exit_error("Error in vCombinedScores::Combinelogs\n");
  }

  if (weights_matrix->Size() > ranks.Bands()) return 0;
  vArray(double) weights = weights_matrix->Matrix();

  vMatrix<double> combinations_matrix(1, training_frames);
  vArray(double) combinations = combinations_matrix.Matrix();
  vArray3(long) ranks3 = ranks.Matrix3();
  long i;
  for (i = 0; i < training_frames; i++)
  {
    double rank_sum = 0;
    for (long j = 0; j < weights_matrix->Size(); j++)
    {
      long rank = ranks3[j][test_frame][i];
      double log1 = logs[rank+1];
      rank_sum += weights[j] * log1;
    }
    combinations[i] = rank_sum;
  }

  vMatchScores * result = new vMatchScores(&combinations_matrix);
  return result;
}


vMatchScores * vCombinedScores::CombineLogs2(vMatrix<double> * weights_matrix, 
                                               long test_frame, long add)
{
  static vMatrix<double> log_matrix = vPrecalculateLogs2(training_frames+100000);
  static vArray(double) logs = log_matrix.Matrix();
  if (log_matrix.Size() < training_frames)
  {
    exit_error("Error in vCombinedScores::Combinelogs\n");
  }

  if (weights_matrix->Size() != scores) return 0;
  vArray(double) weights = weights_matrix->Matrix();

  vMatrix<double> combinations_matrix(1, training_frames);
  vArray(double) combinations = combinations_matrix.Matrix();
  vArray3(long) ranks3 = ranks.Matrix3();
  long i;
  for (i = 0; i < training_frames; i++)
  {
    double rank_sum = 0;
    for (long j = 0; j < scores; j++)
    {
      long rank = ranks3[j][test_frame][i];
      double log1 = logs[rank+1+add];
      rank_sum += weights[j] * log1;
    }
    combinations[i] = rank_sum;
  }

  vMatchScores * result = new vMatchScores(&combinations_matrix);
  return result;
}


vMatchScores * vCombinedScores::GetScores(long index, long test_frame)
{
  vMatrix<double> combinations_matrix(1, training_frames);
  vArray(double) combinations = combinations_matrix.Matrix();
  vArray3(long) ranks3 = ranks.Matrix3();
  long i;
  for (i = 0; i < training_frames; i++)
  {
    combinations[i] = ranks3[index][test_frame][i];
  }

  vMatchScores * result = new vMatchScores(&combinations_matrix);
  return result;
}

