#ifndef VASSILIS_RANKING_H
#define VASSILIS_RANKING_H

#include <vector>

#include "basics/auxiliaries.h"
#include "virtual_hand/FrameViewFinder.h"
#include "basics/algebra.h"
#include "algorithms.h"
#include "hand_database.h"

using namespace std;

class vViewGroundTruth
{
public:
  long test_frame_index;
  long training_pose_index;
  double x_angle;
  double y_angle;
  double orientation;

  vViewGroundTruth()
  {
  }

  ~vViewGroundTruth()
  {
  }

  static vector<vViewGroundTruth> * Load(const char * filename);
  
  // For each ground truth in the vector it saves its distances
  // from each of the about 4000 possible camera view parameters.
  static long SaveGroundTruthDistances(vector<vViewGroundTruth> * gts,
                                       const char * ground_truth_file);
};


class vMatchScore
{
public:
  float score;
  long frame;
  long view;
  ushort is_valid;

  // ground_distance is not read from the file where
  // score, frame and view are saved. It is loaded later,
  // using ground truth data if available. It will be 
  // set to the distance of the view/orientation parameters
  // of the frame from which the score was obtained to
  // the view/orientation parameters of the test frame.
  float ground_distance;

  // rank is also not saved. It will be the rank of the score
  // (0 will be for the best match). It will be equal to
  // the index of the score in the scores array of vMatchScores.
  long rank;

  vMatchScore();
  vMatchScore(double in_score, long in_frame, long in_view);
  void Print();  
  ushort Fprint(FILE * fp);  
  std::string PrintToString();
  std::string PrintToString2();

  long Save(FILE * fp);
  long Load(FILE * fp);
};


// This class is used in the following: Suppose we have a test frame of
// a hand and we want to match it with lots of training frames, for which
// we know the ground truth. We do the matching, and we get a matching 
// score between the test frame and each of the training frames. 
// vMatchScores is used to store all those matching scores, and answer
// queries like: Give me the training frame that is the n-th similar 
// frame to the test frame (based on the scores), or give me the 
// frame number and similarity rank of training frames that are 
// "correct" matches.
class vMatchScores
{
private:
  // The "scores" array stores the scores of all test frames, sorted in 
  // ascending or descending order
  vArray(vMatchScore) scores;

  // This array also stores the scores of all test frames, but it
  // stores them in the same order in which they were saved (i.e.
  // sorted by frame number and view number).
  vArray(vMatchScore) frame_sorted;
  // frame matches is the number of database frames that come from
  // the same category of hand shape as the test input.
  long frame_matches;

  long number_of_frames; // Number of test images for which we have scores.
  long number_of_scores; // Number of scores for each test image.
  ushort is_valid;
  long current_frame;
  long current_score;
  ushort is_text;
  long frame_size;
  long initial_offset;
  FILE * fp;
  char * input_filename;

  // indicates whether the scores are in descending order (if 1) or
  // in ascending order (if 0).
  ushort descending;

  // Even if test_filename is not set to any meaningful value, vMatchScores
  // can still answer queries about scores of given rank. However, if 
  // the test_filename is set to the test video file from which the test frames
  // for the scores came, then vMatchScores can look for auxiliary files 
  // sharing a name with test_filename (but with different extensions) that
  // give information that is necessary to answering additional queries. The only
  // such auxiliary file I am implementing today (6/21/2001) is a .gt ground
  // truth file, which for every test frame will contain: its index in the
  // test video file, the index of the correct frame in the training poses,
  // and the x angle, y angle and orientation that best align the test
  // frame with the training pose.
  char * test_filename;

  // test_frame will indicate what particular test frame we are interested in.
  // Its value affects the behavior of GroundMatch() and CorrectMatch();
  long test_frame;

  // ground_truth will contain the information stored in the .gt file (as 
  // described in the comments for test_filename). If it is 0, certain
  // queries will not be answerable. It will be 0 if SetTestFilename is
  // never called, or if the .gt file for the test_filename does not
  // exist.
  
  vector<vViewGroundTruth> * ground_truth;

  // ground_distances[test][view] is 10 * the distance 
  // between the ground truth viewing parameters and the viewing
  // parameters for the given view.
  v3dMatrix<ushort> * ground_distances;

  // vFrameViewFinder knows the correspondences between training view
  // numbers on the one hand and training pose, x angle, y angle and
  // orientation on the other hand.
  vFrameViewFinder finder;

  // ground_truth_distance is used in the function CorrectMatch. It
  // is measured in degrees (not radians).
  double ground_truth_distance;

  // distance_sorted will be an array of scores corresponding to
  // a given training pose (all different renderings of it). It
  // will be sorted by distance of the the view/orientation 
  // parameters of the frame of each score from the view/orientation
  // parameters of the ground truth for the particular test frame. 
  // This way it will be easy to 
  vArray(vMatchScore) distance_sorted;

  // ground_sorted will also be an array of scores corresponding to
  // a given training pose. It will be sorted by the actual score 
  // value.
  vArray(vMatchScore) score_sorted;

  // Use the score objects in scores_vector to fill in
  // the scores in the scores array and the frame_sorted
  // array, and set the rank of those scores correctly
  long SetRanks(vector<vMatchScore> * scores_vector);

  // Like SetRanks, but here it handles the case 
  // where scores_vector only contains some of the possible
  // scores
  long SetRanks2(vector<vMatchScore> * scores_vector);

  // Some times a whole sequence of scores can actually have the same
  // score. In that case, their ranks should also be adjusted to
  // be the same. What rank should that be? Here I choose the minimum
  // of the ranks of all those equal scores.
  long EqualizeRanks();

  ushort ReadTextFile();
  ushort ReadBinaryFile();

  // Resort scores in the reverse order than the one in which
  // they are currently sorted.
  ushort Reverse();


public:
  // There are different ways to get match scores between a test frame 
  // and a database of training frames. Each constructor handles a
  // different way.

  // This constructor creates a vMatchScores object with no scores.
  vMatchScores();

  // This constructor reads in scores from a file. The scores can
  // be saved in text or binary format.
  vMatchScores(const char * in_filename);

  // This constructor generates scores given a matrix that already contains
  // the distance between the test frame and each training frame.
  vMatchScores(v3dMatrix<double> * distances);

  // This constructor generates scores given the following: A 
  // matrix "features" that, in each column, has the numerical values of
  // all training frames for a particular feature. "column", which
  // specifies the column that is of interest to us. "test_feature"
  // is the numerical value of the particular feature specified
  // by column for the test frame. The score for each training
  // frame is the absolute difference between the feature for that
  // frame and the feature for the test frame.
  vMatchScores(v3dMatrix<double> * features, long column, 
                double test_feature);
  
  // Similar to previous, but features is a matrix of floats.
  vMatchScores(v3dMatrix<float> * features, long column, 
                double test_feature);
  
  // Builds scores based on the distance between the test features and each row of 
  // features matrix. The distance is the Mahalanobis distance. The covariance and
  // number of dimensions to be used are specified in the pca object.
  vMatchScores(vMatrix<double> * features_matrix, 
                vMatrix<double> * test_features, vPCA * pca);

  // Builds scores based on the L1 distance between test features and every 
  // row of features_matrix. The "dimensions" argument specifies how many dimensions
  // to use (dimensions from 0 to dimensions-1 are used).
  vMatchScores(v3dMatrix<double> * features_matrix, 
                v3dMatrix<double> * test_features, long dimensions);

  // NOTE: It is not just the arguments that are different between the next 
  // constructor and the previous. Here we compute L2 distances, before we 
  // were computing L1 distances.
  // Builds scores based on the L2 distance between test features and every 
  // row of features_matrix. The "dimensions" argument specifies how many dimensions
  // to use (dimensions from 0 to dimensions-1 are used).
  vMatchScores(v3dMatrix<float> * features_matrix, 
                v3dMatrix<float> * test_features, long dimensions);


  // Builds scores based on the L1 distance between test features and every 
  // row of features_matrix. The dimensions of features_matrix used are the
  // dimensions from start_dimension to start_dimension+dimensions-1.
  vMatchScores(v3dMatrix<float> * features_matrix, 
                v3dMatrix<double> * test_features, 
                long start_dimension, long dimensions);

  vMatchScores(v3dMatrix<float> * features_matrix, 
                v3dMatrix<float> * test_features, 
                long start_dimension, long dimensions);

  // Builds scores based on the L1 distance between test features and every 
  // row of features_matrix. The dimensions of features_matrix used are the
  // dimensions from start_dimension to start_dimension+dimensions-1.
  // reference_images is a matrix which is useful only for its third column,
  // which contains weights (it is the type of matrix that is used in
  // interface to define the embedding).
  vMatchScores(v3dMatrix<float> * features_matrix, 
                v3dMatrix<double> * test_features, 
                long start_dimension, long dimensions,
                v3dMatrix<float> * reference_images);

  vMatchScores(v3dMatrix<float> * features_matrix, 
                v3dMatrix<float> * test_features, 
                long start_dimension, long dimensions,
                v3dMatrix<float> * reference_images);


  // This constructor is useful if we only calculate scores for some
  // of the frames in the database, but not for the rest. Then,
  // the rest of the frames get scores that are higher than the
  // highest score for the frames that were evaluated.
  vMatchScores(vArray(vMatchScore) scores, long number_of_scores);

  // We return a copy of the passed scores.
  vMatchScores(vMatchScores * scores);

  // This constructor is useful for creating match scores out of scores
  // computed using the Poser1 database.
  vMatchScores(v3dMatrix<float> * scores, vPoseSet poses);

  long Initialize();
  
  ~vMatchScores();

  inline long GroundTruthKnown()
  {
    return (ground_truth != 0);
  }

  // If in_descending is 1, scores will be sorted in descending order.
  // If it is 0, they will be sorted in ascending order. Otherwise,
  // 0 is returned.
  ushort SetOrder(ushort in_descending);

  // StoreFrameSorted is efficient when we have scores for every
  // frame and view of the database.
  long StoreFrameSorted();
  // StoreFrameSorted2 should be used when we have scores for some,
  // but not all views of the database.
  long StoreFrameSorted2();

  long CurrentFrame();
  ushort SetCurrentFrame(long frame);
  ushort PreviousFrame();
  ushort NextFrame();
  ushort FirstFrame();
  ushort LastFrame();

  ushort SetCurrentScore(long index);
  vMatchScore GetScore(long frame, long rank);
  vMatchScore GetScore(long rank);
  vMatchScore CurrentScore();
  vMatchScore PreviousScore();
  vMatchScore NextScore();
  vMatchScore FirstScore();
  vMatchScore LastScore();

  // Returns the index of the score corresponding to given frame and rank,
  // (in the frame_sorted array).
  // or -1 if that score was not found. Finding the score has to be done
  // using binary search, because in certain cases we don't have a score
  // for each database frame.
  long FrameSortedIndex(long frame, long view);

  inline vMatchScore GetRank(long frame, long view)
  {
    long index = frame * 4 + view;
    if (index < 0) return vMatchScore();
    if (index >= number_of_scores) return vMatchScore();
    return frame_sorted[index];
  }


  // This works also when we only have scores for some frames of the database.
  // GetRank requires that we have scores for the whole database.
  long GetRank2(long frame, long view);
  long NumberOfScores();
  ushort IsValid();

  // Removes two endings from the filename to get the part of the filename 
  // that is common among all files containing scores for the same 
  // matching function and same test video.
  char * GetTextBase();

  // Sets the variable input_filename (which is slightly misnamed, since
  // it is also used for output, when we call WriteBinary).
  long SetFilename(const char * new_filename);

  // Returns text_base.frame
  char * FrameFilename(const char * text_base, long frame);

  // SetTestFilename sets the test_filename, thus enabling the program
  // to look for corresponding auxiliary files (like a .gt ground truth
  // file).
  ushort SetTestFilename(const char * in_test_filename);

  // SetTestFrame() sets the current test frame number, that will be
  // used by GroundMatch and CorrectMatch();
  ushort SetTestFrame(long in_test_frame);

  // Sets the ground truth to the given parameters. This is used
  // by GroundMatch and CorrectMatch
  long SetGroundTruth(long training_pose, double x_angle, 
                      double y_angle, double z_angle);

  // Sets the ground truth parameters to be those of the database image
  // identified by the given frame and view
  long SetGroundTruthView(long frame, long view);

  inline long SetGroundDistances(v3dMatrix<ushort> * in_ground_distances)
  {
    ground_distances = in_ground_distances;
    return 1;
  }

  inline vArray(vMatchScore) GetFrameSorted()
  {
    return frame_sorted;
  }

  inline vArray(vMatchScore) GetScores()
  {
    return scores;
  }

  // Provided a ground truth file has been loaded, GroundMatch will
  // return the score between the current test frame and the training
  // view that is the rank-th closest (in terms of ground truth) to it.
  // Views for which the underlying pose does not match the test frame's
  // pose are not considered at all.
  vMatchScore GroundMatch(long rank);

  // Provided a ground truth file has been loaded, CorrectMatch will
  // return the rank-th best score between the current test frame and 
  // a corresponding training pose whose view/orientation parameters 
  // are within 
  // ground_truth_distance from those of the test frame. As in GroundMatch,
  // views for which the underlying pose does not match the test frame's
  // pose are not considered at all. The distance doesn't have to be 
  // based on the L2 metric, I may try different ones.
  vMatchScore CorrectMatch(long rank, long * total_matches = 0);

  // Sets the variable ground_truth_distance, which is used in 
  // CorrectMatch();
  ushort SetGroundTruthDistance(double in_ground_truth_distance);

  // If ground truth is available it returns 1 and sets x, y and z
  // contents to the camera view contained in the ground truth.
  // If ground truth is not available, it returns 0.
  long GroundTruthView(double * x, double * y, double * z);

  // Writes the scores to a binary file (its name comes from 
  // input_filename);
  ushort WriteBinary();

  // Reads in scores from files text_base.0.txt, text_base.1.txt ... text_base.limit.txt
  // and creates a binary file that contains, for each text file, the number_of_scores
  // highest and lowest scores from that file.
  //static ushort ToBinary(const char * text_base, long limit, long number_of_scores);

  // Supporting a number_of_scores parameter is not implemented yet.
  ushort ToBinary(long limit);
  
  // Filter uses another vMatchScores object to discard scores in this 
  // object that correspond to frames that score really bad (rank > threshold)
  // in the other scores object.
  long Filter(vMatchScores * other_scores, long threshold);

  // As of 9/22/2001, the "GroundTruthStatistics" and FrameStatistics
  // functions are obsolete. I will delete them when I make sure that 
  // the replacements are ready.
  // The problem with them is that they assume that the match scores come 
  // from a file. Now, with scores based on moments already implemented
  // and scores based on edge histograms and protrusions coming up, there
  // are more complicated ways in which scores may get generated (even by
  // combining a set of other scores). So, instead, there will be a
  // vFrameMatchStats class and a vSequenceMatchStats, and the interpreter
  // class will be in charge of putting frame stats together in a sequence 
  // stats object.

  // GroundTruthStatistics() prints useful information about the matching rank
  // of the views that we consider "correct", for a sequence of test frames. We
  // assume (as in ToBinary()) that the scores for that sequence have been saved
  // in files of the form base.i.end where base is any string, i is the frame number
  // and end is txt or bin.
  ushort GroundTruthStatistics(long limit);
  
  // Like GroundTruthStatistics, but it prints the results to the given file
  ushort GroundTruthStatistics2(const char * filename, long limit);

  // LikeGroundTruthStatistics2, but the file is given as an 
  // open file pointer, instead of as a filename.
  ushort GroundTruthStatistics3(FILE * fp, long limit);

  // Like GroundTruthStatistics3, but only prints statistics for this object.
  ushort FrameStatistics(FILE * fp);

  // KeepHighest returns a new vMatchScores object, that contains scores only
  // for matches_to_keep frames. The highest-ranking matches are kept.
  // For example, if matches_to_keep is 100, a new vMatchScores object
  // containing scores for the 100 highest matches (as described in
  // this object) is returned.
  vMatchScores * KeepHighest(long matches_to_keep);

  // Select creates a new vMatchScores object containing scores
  // (as described in this object) for all frames for which
  // scores exist in scores2. This function verifies that
  // for all scores in scores2 there is a score with identical
  // frame and view in frame_sorted. If not, 0 is returned.
  vMatchScores * Select(vMatchScores * scores2);

  // The function Combine combines the ranks of the scores in *this with the
  // ranks of the scores in scores2, and uses the combined log ranks to
  // create a new vMatchScores object. The weight of this is weight, and 
  // the weight of scores2 is 1-weight.
  vMatchScores * Combine1(vMatchScores * scores2, double weight);

  // These function is used to combine many different vMatchScores 
  // objects. number tells how many vMatchScores objects 
  // we are passing. After number we pass scores and weights. The
  // argument list looks like 
  // (number, scores1, weight1, scores2, weight2, ..., scores_n, weight_n)
  // It is assumed (the function verifies that) that the frames and views
  // stored in the i-th score in the frame_sorted array of all scores
  // are identical. If not, 0 is returned.
  static vMatchScores * CombineLogs(long number, ...);
  static vMatchScores * CombineSums(long number, ...);
  static vMatchScores * CombineLogs2(long number, vArray(vMatchScores *) scores,
                                      vArray(double) weights);

  static vMatchScores * PETS_Filter1();

  // Sometimes we have scores for a set of frames and another vMatchScores
  // object has more accurate scores for a subset of those frames. 
  // Merge() replaces the scores in scorse2 with scores in the
  // scores1 (that is passed in as an argument) when such scores exist.
  static vMatchScores * Merge(vMatchScores * scores1, vMatchScores * scores2);
};


// class vFrameMatchStats is in charge of collecting statistics that 
// summarize a vMatchScores object, given ground truth for a particular
// frame. We want to know the highest ranking correct scores, but also
// the rank of the matches closest to the ground truth. We may also want
// to collect all scores in ranks start, start + k, start + 2k, ..., end.
// We want to
// be able to print the results in different ways and to file, a string,
// or to standard output.
class vFrameMatchStats
{
private:
  vector<vMatchScore> ground_matches;
  vector<vMatchScore> correct_matches;

  // number_of_correct_matches is the total number of correct matches
  // within ground_distance.
  long number_of_correct_matches;
  vector<vMatchScore> top_scores;
  // View parameters of the test frame for which the statistics
  // are compiled (they are contained in the ground truth that
  // is part of the match scores).
  double x, y, z;

  long is_valid;
  
  // When we print, we may want to print only a subset of the 
  // correct matches and ground truth matches. correct_to_print
  // and ground_to_print specify how many of each should be
  // printed (starting from rank 0)
  long correct_to_print, ground_to_print, top_to_print;
  
  // ground_distance can be set to any value desired, and then 
  // CorrectMatch() will ignore correct_matches farther away
  // than ground_distance;
  double ground_distance;

public:
  vFrameMatchStats();

  // correct_limit specifies the number of highest ranking correct
  // matches that we want to preserve. ground_limit is the number
  // of matches closest to the ground truth that we want to preserve.
  vFrameMatchStats(vMatchScores * scores, long in_correct_limit, 
                    long in_ground_limit);

  vFrameMatchStats(vMatchScores * scores, long start, long end,
                    long period);

  // This constructor basically collects the scores of the views stored
  // in the other_scores of other_stats. We assume that other_stats
  // was created using some other scores object, and now we just want
  // to see how those selected views do given these scores.
  vFrameMatchStats(vMatchScores * scores, vFrameMatchStats * other_stats);

  ~vFrameMatchStats();

  inline long GetGroundDistance()
  {
    return (long) ground_distance;
  }

  long SetGroundDistance(long in_ground_distance);

  inline vMatchScore GroundMatch(long index)
  {
    if ((index < 0) || (index >= (long) ground_matches.size())) 
	{
	  return vMatchScore();
	}
    return ground_matches[index];
  }

  vMatchScore CorrectMatch(long index);

  inline long NumberOfCorrectMatches()
  {
    return number_of_correct_matches;
  }

  inline vMatchScore TopScore(long index)
  {
    if ((index < 0) || (index >= (long) top_scores.size())) return vMatchScore();
    return top_scores[index];
  }

  inline long IsValid()
  {
    return is_valid;
  }

  inline long GroundTruthView(double * out_x, double * out_y, double * out_z)
  {
    *out_x = x;
    *out_y = y;
    *out_z = z;
    return 1;
  }

  inline long SetCorrectToPrint(long in_correct_to_print)
  {
    if (in_correct_to_print < 0) return 0;
    if (in_correct_to_print >= (long) correct_matches.size()) return 0;
    correct_to_print = in_correct_to_print;
    return 1;
  }

  inline long SetGroundToPrint(long in_ground_to_print)
  {
    if (in_ground_to_print < 0) return 0;
    if (in_ground_to_print >= (long) ground_matches.size()) return 0;
    ground_to_print = in_ground_to_print;
    return 1;
  }

  inline long SetTopToPrint(long in_top_to_print)
  {
    if (in_top_to_print < 0) return 0;
    top_to_print = in_top_to_print;
    return 1;
  }

  inline long GetCorrectToPrint()
  {
    return correct_to_print;
  }

  inline long GetGroundToPrint()
  {
    return ground_to_print;
  }

  std::string PrintToString();

  long Print();

  long PrintToFile(const char * filename);

  long PrintToFile(FILE * fp);

  long Save(FILE * fp);
  long Load(FILE * fp);
};


// This class will put together the information from several different 
// vFrameMatchStats objects, so that it can answer queries like: What
// is the median/mean/log mean... of the best correct matches or ground
// matches in those frames? How many best ground or correct matches
// have ranks smaller than a given threshold? It will also know how
// to print stuff to the standard output, to a string and to a file
class vSequenceMatchStats
{
private:
  vector<vFrameMatchStats *> frame_stats;
  vector<long> frame_numbers;
  // In compiling the sequence statistics, we consider a frame only
  // if selected[frame] = 1.
  vector<long> selected;
  long top_to_print;

public:
  vSequenceMatchStats();
  ~vSequenceMatchStats();

  long AddFrameStats(vFrameMatchStats * new_frame_stats, long frame);
  
  // mean rank of the index-th highest ranking correct matches
  // across frames
  double CorrectMean(long index);

  // median rank of the index-th highest ranking correct matches
  // across frames
  long CorrectMedian(long index);

  // Number of correct matches of the given index whose rank 
  // was below the given rank
  long CorrectMatchesUnder(long index, long rank);

  // Fraction of correct matches of the given index whose rank 
  // was below the given rank
  double PcorrectMatchesUnder(long index, long rank);

  // Number of correct matches of the given index that fell between
  // low rank and high rank and whose ground_truth distance was
  // between low_distance and high_distance.
  long CorrectMatchesWithin(long index, long low_rank, long high_rank,
                            long low_distance, long high_distance);

  // median rank of the index-th highest ranking groud matches
  // across frames
  double GroundMean(long index);

  // mean rank of the index-th highest ranking correct matches
  // across frames
  long GroundMedian(long index);

  // Number of ground matches of the given index whose rank 
  // was below the given rank
  long GroundMatchesUnder(long index, long rank);

  long SelectAll();

  // Selects frames whose ground truth viewpoint is within
  // tolerance of the viewpoint specified by x and y.
  long SelectViewpoint(double x, double y, double tolerance);

  // Selects frame whose ground truth viewpoint is at
  // distance from frontal view, +- tolerance.
  long SelectFrontDistance(double distance, double tolerance);
  double GetFrontDistance(long frame);

  long SetCorrectToPrint(long in_correct_to_print);

  // Selects frames between start_frame and end_frame.
  long SelectRange(long start_frame, long end_frame);

  // Selects frames between start and end frame, and frontal distance within 
  // tolerance of distance. A combiation of SelectRange and SelectFrontDistance.
  long SelectRangeFrontDistance(long start_frame, long end_frame,
                                double distance, double tolerance);

  long Print();
  std::string PrintToString();
  long PrintToFile(const char * filename);
  long PrintToFile(FILE * fp);

  long SetGroundDistance(long ground_distance);
  long Save(const char * filename);
  long Save(FILE * fp);
  long Load(const char * filename);
  long Load(FILE * fp);

  inline long SetTopToPrint(long number)
  {
    top_to_print = number;
    return 1;
  }

  long NumberOfFrames();
  // returns the frame stats for which the corresponding
  // entry in frame_numbers is frame.
  vFrameMatchStats * FrameStats(long frame);
  // returns frame_stats[index].
  vFrameMatchStats * FrameStats2(long index);

  // Removes all vFrameMatchStats objects from the vectors,
  // without calling their destructor. This is useful when
  // we move those objects from one vSequenceMatchStats object
  // to another.
  long RemoveAllStats();
};


class vCombinedScores
{
public:
  v3dMatrix<long> ranks;
//  v3dMatrix<float> distances;

  long scores;
  long test_frames;
  long training_frames;

public:
  vCombinedScores();
  vCombinedScores(long number_of_scores, long test_frames, long training_frames);
  ~vCombinedScores();

  long EnterScores(vMatchScores * scores, long score_index, long test_index);
  long Load(const char * filename);
  long Save(const char * filename);

  vMatchScores * CombineSums(vMatrix<double> * weights, long test_frame);
  vMatchScores * CombineRoots(vMatrix<double> * weights, long test_frame);
  vMatchScores * CombineLogs(vMatrix<double> * weights, long test_frame);
  vMatchScores * CombineLogs2(vMatrix<double> * weights, long test_frame, long add);
  vMatchScores * GetScores(long index, long test_frame);
};






#endif // VASSILIS_RANKING_H
