#ifndef VASSILIS_THESIS_MODULE_H
#define VASSILIS_THESIS_MODULE_H


#include "basics/graphical_module.h"
#include "gestures/algo_templates.h"
#include "gestures/chamfer_distances.h"

class thesis_module : public class_module
{
public:
  // the next two lines are mandatory for any module
  typedef long (thesis_module::*type_procedure) ();
  std::vector<type_procedure> procedures;

  graphical_module * graphics_module;

  // Data related to the skin detector
  vArray(vSkinTableBase *) skin_tables;
  vSkinTableBase * skin_table;
  ushort skin_data_initialized;
  ulong current_table_index;
  const char * skin_directory;
  const char * non_skin_directory;

  class_hand_database * database;

  // this matrix stores the last computation of chamfer distances
  // between the query and the database.
  float_matrix query_distances;

public:

  //////////////////////////////////////////////////////////
  //                                                      //
  // start of obligatory member functions for any module  //
  //                                                      //
  //////////////////////////////////////////////////////////

  thesis_module(class_interpreter * interpreter);
  virtual ~thesis_module();
  
  virtual long initialize();
  virtual long clean_up();
  
  long enter_command(char * command, type_procedure procedure);
  virtual long process_command(const char * command);
  long sort_commands ();

  //////////////////////////////////////////////////////////
  //                                                      //
  //  end of obligatory member functions for any module   //
  //                                                      //
  //////////////////////////////////////////////////////////


  //////////////////////////////////////////////////////////
  //                                                      //
  //  start of functions accessing the graphical module   //
  //                                                      //
  //////////////////////////////////////////////////////////


  long show_image(general_image * image);
  long default_input(long vertical, long horizontal);
  long check_input();
  general_image * current_image ();
  class_sequence_information * current_sequence();



  //////////////////////////////////////////////////////////
  //                                                      //
  //   end of functions accessing the graphical module    //
  //                                                      //
  //////////////////////////////////////////////////////////

  // start of functions about skin detection
  long DeleteSkinData();
  long InitializeSkinData();
  long PickSkinTables();
  long DetectSkin();
  // end of functions about skin detection

  // Start of section using the IPL library
  long CannyDetector0a();
  long CannyDetector0b();
  long CannyDetector(double low);
  long CannyDetector3(double low, double high, int aperture);

  long DistanceTransformSlow();
  long DistanceTransform();

  long enclosing_circle();
  long enclosing_circle_normalize();
  long enclosing_circle_normalize3();

  // End of section using the IPL library


  // hacky/heuristic detection of the hand bounding box 
  long SkinHand();
  long KeepPalm2();

  // this are one-shot procedures, that are used once to combine
  // the BMP outputs of poser (in cartoon format and in color format)
  // into AVI sequences, and to then generate the edge images
  long poser_cartoon_sequence();
  long poser_color_sequence();
  long poser_first_edges(); // here the edge images are not normalized
  long poser_normalized_edges(); // here the edge images are normalized

  // chamfer distance between two frames
  long chamfer_distance0();
  long chamfer_distance(long input1, long input2);
  long chamfer_distance_fast();

  long database_create();
  long database_destroy();
  long database_edges();
  
  // like database_edges, but here the user only provides
  // one number to specify the pose, as opposed to two numbers in
  // database_edges
  long database_pose(); 
  long database_pose(long pose); 

  long database_load_distances();
  long database_clear_distances();
  long database_distances();

  long database_save_pixels();
  long database_load_pixels();
  long database_pixels();

  long database_rotation();
  long database_print();
  long database_chamfer(); // we specify poses
  long database_chamfer2(); // we specify frames and orientations
  
  // treats the current image as a query image and normalizes
  // that image as required by the database.
  long database_normalize(); 
  
  // creates chancery features for the current query image
  // and displays a specific orientation.
  long chamfer_feature_test();

  // computes the chamfer distance between the specified orientation
  // of the query image and the specified database frame,
  // by first creating chamfer feature objects for both the query
  // and the database
  long chamfer_feature_distance();

  // normalizes the current query image and
  // computes the chamfer distances between the query
  // and every pose represented in the database.
  long database_chamfer_distances();

  // displays information about the rank-th nearest neighbor
  // of the query, starting with the first neighbor.
  long chamfer_rank();
  long chamfer_rank(long rank);
  
  // save a matrix of distances between every frame of the 
  // current sequence and every pose represented in the database.
  long sequence_chamfer_distances();
  
  // for specified frame and orientation of the current input
  // we load from a file the distances from the frame to all database poses
  // and we adjust those distances based on the orientation
  long load_chamfer_distances();

  long save_full_distances();

  long convert();
};













#endif // VASSILIS_THESIS_MODULE_H
