#include "stdafx.h"

#include "vplatform.h"

#include <assert.h>
#include <stdio.h>

#include "basics/algebra.h"
#include "basics/read_avi.h"
#include "volumetric.h"
#include "basics/matmath.h"
#include "basics/drawing_temp.h"

#include "basics/definitions.h"

using namespace std;


vBackgroundModel::vBackgroundModel(vSequenceStatistics<uchar> * stats)
{
  is_valid = 0;
  background_mean = vCopyImage(stats->GetMeans());
  background_std = vCopyImage(stats->GetDeviations());
  // We only handle color backgrounds (we need three bands for them).
  if (background_mean->Bands() != 3) return;
  if (background_std->Bands() != 3) return;
  if (background_mean->Rows() != background_std->Rows()) return;
  if (background_mean->Cols() != background_std->Cols()) return;
  is_valid = 1;

  // Take square root of background std;
  long size = background_std->AllBandSize();
  vArray(double) stds = background_std->Matrix();
  long i;
  for (i = 0; i < size; i++)
  {
    stds[i] = sqrt(stds[i]);
  }
}


vBackgroundModel::~vBackgroundModel()
{
  vdelete(background_mean);
  vdelete(background_std);
}


v3dMatrix<double> * vBackgroundModel::BackgroundMean()
{
  if (is_valid) return background_mean;
  else return 0;
}


v3dMatrix<double> * vBackgroundModel::BackgroundStd()
{
  if (is_valid) return background_std;
  else return 0;
}


v3dMatrix<double> * 
vBackgroundModel::EstimateForeground(v3dMatrix<uchar> * the_image)
{
  if (!is_valid) return 0;
  long rows = the_image->Rows();
  long cols = the_image->Cols();
  long bands = the_image->Bands();
  if (rows > background_mean->Rows()) return 0;
  if (cols > background_mean->Cols()) return 0;
  if (bands != background_mean->Bands()) return 0;
  
  v3dMatrix<double> * result = new v3dMatrix<double>(rows, cols, bands);
  vArray2(uchar) source_matrix = the_image->Matrix2();
  vArray2(double) result_matrix = result->Matrix2();
  vArray2(double) mean_matrix = background_mean->Matrix2();
  vArray2(double) std_matrix = background_std->Matrix2();

  long i, band;
  long size = rows * cols;
  for (band = 0; band < bands; band++)
  {
    for (i = 0; i < size; i++)
    {
      double current_std = std_matrix[band][i];
      if (current_std == 0.0) current_std = 0.1;
      double value = source_matrix[band][i] - mean_matrix[band][i];
      if (value < 0) value = -value;
      result_matrix[band][i] = value / current_std;
    }
  }
  return result;
}


color_image * vBackgroundModel::SubtractBackground(color_image * the_image, double threshold)
{
  if (is_valid <= 0) return 0;
  long rows = the_image->Rows();
  long cols = the_image->Cols();
  if (rows != background_mean->Rows()) return 0;
  if (cols != background_mean->Cols()) return 0;
  vArray2(uchar) source = the_image->Matrix2();
  color_image * result_image = new color_image(rows, cols);
  vArray2(uchar) result = result_image->Matrix2();
  long size = the_image->Size();

  v3dMatrix<double> * foreground_estimate = EstimateForeground(the_image);
  v3dMatrix<uchar> * binary_image = foreground_estimate->threshold(threshold);
  vArray(uchar) binary = binary_image->Matrix();
  vdelete(foreground_estimate);

  long i;
  for (i = 0; i < size; i++)
  {
    if (binary[i] == 0)
    {
      result[0][i] = result[1][i] = result[2][i] = 0;
    }
    else
    {
      result[0][i] = source[0][i];
      result[1][i] = source[1][i];
      result[2][i] = source[2][i];
    }
  }
  vdelete(binary_image);
  return result_image;
}


ushort vBackgroundModel::IsValid()
{
  return (ushort)is_valid;
}


vCalibrationInfo::vCalibrationInfo(const char * filename)
{
  matrix(&R, 3, 3);
  matrix(&T, 3, 1);
  matrix(&P, 3, 4);
  matrix(&RT, 3, 4);
  matrix(&A, 3, 3);
  valid = 0;

  FILE * fp = fopen(filename, vFOPEN_READ);
  if (fp == 0) 
  {
    return;
  }
  
  long items_read;
  items_read = fscanf(fp, "A:%lf, S:%lf, B:%lf, U:%lf , V:%lf\n\n\n", 
		      &a, &s, &b, &u, &v);
  if (items_read != 5) goto calibration_info_exit;

  items_read = fscanf(fp, "Kappa1:%lf , Kappa2:%lf\n\n\n", &kappa1, &kappa2);
  if (items_read != 2) goto calibration_info_exit;

  double temp;
  items_read = fscanf(fp, "Plane: pict_%lf-avi\n\nRotation Matrix:\n", &temp);
  if (items_read != 1) goto calibration_info_exit;

  long i, j;
  for (i = 1; i <= 3; i++)
    for (j = 1; j <= 3; j++)
    {
      items_read = fscanf(fp, "%lf", &R[i][j]);
      if (items_read != 1) goto calibration_info_exit;
    }

  items_read = fscanf(fp, "\n\nTranslation Vector:\n%lf %lf %lf", 
		      &T[1][1], &T[2][1], &T[3][1]);
  if (items_read != 3) goto calibration_info_exit;

  valid = 1;
//  mat_copy(R, RT, 3, 3, 1, 1);
//  mat_copy(T, RT, 3, 1, 1, 4);

  A[1][1] = a;
  A[1][2] = s;
  A[1][3] = u;

  A[2][1] = 0;
  A[2][2] = b;
  A[2][3] = v;

  A[3][1] = 0;
  A[3][2] = 0;
  A[3][3] = 1;

//  mat_mult(A, RT, P, 3, 3, 4);
  valid = 1;

 calibration_info_exit:
  fclose(fp);
}


vCalibrationInfo::~vCalibrationInfo()
{
//  free_matrix(R, 3, 3);
//  free_matrix(T, 3, 1);
//  free_matrix(P, 3, 4);
//  free_matrix(RT, 3, 4);
//  free_matrix(A, 3, 3);
}


void vCalibrationInfo::Print()
{
  if (!valid) 
  {
    printf("Not valid\n");
    return;
  }

  printf("a = %lf;\n", a);
  printf("s = %lf;\n", s);
  printf("b = %lf;\n", b);
  printf("u = %lf;\n", u);
  printf("v = %lf;\n", v);
  printf("kappa1 = %lf;\n", kappa1);
  printf("kappa2 = %lf;\n", kappa2);

//  print_matrix("Rotation matrix R", R, 3, 3);
//  print_matrix("Translation T", T, 3, 1);
//  print_matrix("RT", RT, 3, 4);
//  print_matrix("A", A, 3, 3);
//  print_matrix("Projection matrix P", P, 3, 4);
}


v3dMatrix<uchar> * 
vCalibrationInfo::Undistort(v3dMatrix<uchar> * distorted)
{
  long rows = distorted->Rows();
  long cols = distorted->Cols();
  long bands = distorted->Bands();
  vArray3(uchar) matrix1 = distorted->Matrix3();
  
  v3dMatrix<uchar> * result = new v3dMatrix<uchar>(rows, cols, bands);
  vArray3(uchar) matrix2 = result->Matrix3();
//  vWriteValue(result, (uchar) 0);

  long row, col, band;
  for (row = 0; row < rows; row++)
  {
    for (col = 0; col < cols; col++)
    {
      double u1 = col;
      double v1 = row;
      double x_d = (u1 - u) / a;
      double y_d = (v1 - v) / b;
      double r2 = x_d * x_d + y_d * y_d;
      double r4 = r2 * r2;
      double coeff = 1 + kappa1 * r2 + kappa2 * r4;
      double x_u = x_d * coeff;
      double y_u = y_d * coeff;
      double u2 = a * x_u + u;
      double v2 = b * y_u + v;
      long row2 = (long) v2;
      long col2 = (long) u2;
      if ((row2 >= 0) && (row2 < rows) && (col2 >= 0) && (col2 < cols))
      {
	for (band = 0; band < bands; band++)
	{
	  matrix2[band][row2][col2] = matrix1[band][row][col];
	}
      }
    }
  }

  return result;
}


void vCalibrationInfo::UndistortedCoords(double in_row, double in_col, 
					  double & out_row, double & out_col)
{
  double u1 = in_col;
  double v1 = in_row;
  double x_d = (u1 - u) / a;
  double y_d = (v1 - v) / b;
  double r2 = x_d * x_d + y_d * y_d;
  double r4 = r2 * r2;
  double coeff = 1 + kappa1 * r2 + kappa2 * r4;
  double x_u = x_d * coeff;
  double y_u = y_d * coeff;
  double u2 = a * x_u + u;
  double v2 = b * y_u + v;
  out_col = u2;
  out_row = v2;
}


void vCalibrationInfo::UndistortedCoords2(double in_row, double in_col, 
					   double & out_row, double & out_col)
{
  double xd = in_col;
  double yd = in_row;
  
  xd = (xd - u)/a;
  yd = (yd - v)/b;
  
  double rad = xd*xd+yd*yd;
  
  xd =  xd + xd*(kappa1*rad + kappa2*rad*rad);
  yd =  yd + yd*(kappa1*rad + kappa2*rad*rad);
  
  out_col = a*xd+u;
  out_row = b*yd+v;
}


void vCalibrationInfo::DistortedCoords(double in_row, double in_col, 
					double & out_row, double & out_col)
{
}


ushort vCalibrationInfo::IsValid()
{
  return valid;
}
