#ifndef VASSILIS_AVI_FILE_H
#define VASSILIS_AVI_FILE_H

#include <vfw.h>
#include "basics/image.h"
#include "basics/video_file.h"
#include "basics/read_avi.h"

// For the time being, this code only handles uncompressed video
// streams.

class vAVI_File_Init
{
public:
  vAVI_File_Init();
  ~vAVI_File_Init();
};

class vAVI_File
{
private:
  // This variable is just used so that when it is 
  // created its constructor calls AVIFileInit() and 
  // when it is destroyed
  // (at program's end) its destructor calls AVIFileExit();

  static vAVI_File_Init avi_file_init;
  
  unsigned short file_open;
  unsigned short write_flag;
  PAVIFILE file;
  char * filename;

public:
  vAVI_File();
  ~vAVI_File();

  static vAVI_File * OpenWrite(const char * filename);
  static vAVI_File * OpenRead(const char * filename);
  void CloseFile();
  unsigned short WriteFlag();
  char * GetFilename();
  PAVIFILE * GetHandle();
};



class vAVI_Stream : public vVideoFile
{
private:
  unsigned short exclusive_file_flag;
  vAVI_File * file;
  PAVISTREAM stream;
  long color_bits;
  long bytes_per_pixel;
  long bytes_per_frame;

  unsigned char * buffer;
  BITMAPINFO * bitmap_info;
  BITMAPINFOHEADER * bitmap_header;
  RGBQUAD * palette;
  
  long bitmap_info_size;
  long bitmap_header_size;
  long palette_size;
  unsigned char * frame_data;
  long current_frame;

  void Initialize(vAVI_File * avi_file);
  void Initialize(vAVI_File * avi_file, 
                  long rate_p, long scale_p,
                  long rows_p, long cols_p,
                  long color_bytes_p);

public:
  vAVI_Stream(vAVI_File * avi_file);
  vAVI_Stream(const char * filename);
  vAVI_Stream(const char * filename, long rows_p, long cols_p, 
               long color_bytes_p);
  vAVI_Stream(vAVI_File * avi_file, long rows_p, long cols_p,
               long rate_p, long scale_p, long color_bytes_p);
  vAVI_Stream(const char * filename, long rows_p, long cols_p,
               long rate_p, long scale_p, long color_bytes_p);
  ~vAVI_Stream();

  // Beginning of functions required by vVideoSequence
  grayscale_image * ReadNextFrameGray();
  grayscale_image * ReadFrameGray(long number);
  color_image * ReadFrame(long number);
  color_image * ReadNextFrame();


  ushort InitializeWriting();
  ushort write_frame(color_image * image);
  ushort write_frameGray(grayscale_image * gray);
  ushort FinishWriting();
  // End of functions required by vVideoSequence

  unsigned short ReadFrame();
  unsigned short write_frame();
  unsigned short write_frame(unsigned char * data);
  unsigned char * GetFrame();
  char * GetFilename();
  long CurrentFrameID();
  long GetRows();
  long GetCols();
  long GetWidth();
  long GetHeight();
  long GetFrameBytes();
  long GetColorBytes();
  long GetRate();
  long GetScale();
  RGBQUAD * GetPalette();
  unsigned short IsValid();

  color_image * To_color_image();
  color_image * To_color_image(long left, long right, long top, long bottom);
	v3dMatrix<uchar> * ToGrayImage();
  v3dMatrix<uchar> * ToGrayImage(long left, long right, long top, long bottom);

  void From_color_image(color_image * image);
  void From_color_image(color_image * image, long left, long right, 
                    long top, long bottom);
  void FromGrayImage(v3dMatrix<uchar> * image);
  void FromGrayImage(v3dMatrix<uchar> * image, long left, long right, 
                    long top, long bottom);

  void Print(char * filename);
  virtual long BytesPerPixel()
  {
    return bytes_per_pixel;
  }
};


class vAVI_Text_Stream
{
private:
  unsigned short exclusive_file_flag;
  unsigned short valid_stream;
  unsigned short write_flag;
  vAVI_File * file;
  PAVISTREAM stream;
  long left, right, top, bottom;
  long scale, rate;
  long frame_size;

  unsigned char * frame_data;
  long current_frame;
  void Initialize(vAVI_File * avi_file);
  void Initialize(vAVI_File * avi_file, 
                  long scale_p, long rate_p, long size,
                  long left_p, long right_p, 
                  long top_p, long bottom_p);


public:
  vAVI_Text_Stream(vAVI_File * avi_file);
  vAVI_Text_Stream(char * filename);
  vAVI_Text_Stream(vAVI_File * avi_file, 
                    long scale, long rate,
                    long size,
                    long left_p, long right_p, 
                    long top_p, long bottom_p);
  vAVI_Text_Stream(char * filename, 
                    long scale_p, long rate_p, long size,
                    long left_p, long right_p, 
                    long top_p, long bottom_p);
  ~vAVI_Text_Stream();

  unsigned short ReadFrame();
  unsigned short write_frame();
  unsigned char * GetFrame();
  char * GetFilename();
  long CurrentFrameID();
  long GetLeft();
  long GetRight();
  long GetTop();
  long GetBottom();
  unsigned short IsValid();
};


// The class vMFC_AVI_Set gives me a way to work around the 2GB limitation
// on the file size of an AVI file. This way, I can split a sequence into
// many AVI files, each of which is around 2GB large (or smaller, if, for
// example, I want each subfile to fit on a CD). There has to be a header file,
// ending in .aviset, which contains information about the sequence.
class vMFC_AVI_Set : public vVideoFile
{
protected:
  long frames_per_file;
  vAVI_Stream * current_avi;
  vReadAVI * current_read_avi;
  long current_number;
  char * base_filename;
  long bytes_per_pixel;

  // Returns the maximum number of frames that we can fit in 
  // a single AVI file with given frame size and color bytes,
  // before exceeding the 2GB limit. This is a slightly
  // conservative estimate, since I don't have the specs of the
  // AVI format, and therefore I can't get a closed form answer.
  inline long MaxFramesPerFile(long in_rows, long in_cols, 
                               long in_color_bytes)
  {
    long max_frame_size = in_rows * in_cols * in_color_bytes + 4096 * 8;
    long max_frames = 2000000000 / max_frame_size;
    // I don't expect max_frames to ever be less than 10, and if it is,
    // then I should think about some more appropriate format.
    if (max_frames < 10) return -1;
    else return max_frames;
  }
    
  // Returns the number of the AVI file containing the given frame
  inline long FrameFile(long frame)
  {
    if (frame < 0) return -1;
    return (frame / frames_per_file);
  }

  // Returns the actual frame number of the given frame in the 
  // AVI file where it is stored.
  inline long FrameFileIndex(long frame)
  {
    return (frame % frames_per_file);
  }

  // Returns 1 if the frame can be loaded from current_avi,
  // 0 otherwise
  inline long IsLoaded(long frame)
  {
    long number = FrameFile(frame);
    if (number == current_number) return 1;
    else return 0;
  }

  // Returns the number of frames that the n-th file should have
  inline long FileFrames(long n)
  {
    if (n < 0) return -1;
    if (number_of_frames <= 0) return -1;
    if (n > (number_of_frames - 1) / frames_per_file + 1) return -1;
    if (n <= (number_of_frames - 1) / frames_per_file) return frames_per_file;
    else 
    {
      long result = number_of_frames % frames_per_file;
      if (result == 0) result = frames_per_file;
      return result;
    }
  }


  // Returns the filename of the n-th AVI file in the set of
  // AVI files that makes up this sequence.
  char * NumberedFile(long n);

  // Opens the n-th AVI file, returns a result > 0 if successful.
  long OpenNumber(long n);

  long Initialize();
  long CleanUp();
  char * BaseFilename(const char * filename);
  char * Directory();
  long MakeDirectory();

public:
  vMFC_AVI_Set(const char * filename);
  vMFC_AVI_Set(const char * filename, long rows_p, long cols_p, 
               long color_bytes_p, long frames_per_file);
  vMFC_AVI_Set(const char * filename, long rows_p, long cols_p,
               long rate_p, long scale_p, long color_bytes_p);
  vMFC_AVI_Set(const char * filename, long rows_p, long cols_p,
               long rate_p, long scale_p, long color_bytes_p, 
               long frames_per_file);
  ~vMFC_AVI_Set();

  inline long BytesPerPixel()
  {
    return bytes_per_pixel;
  }

  virtual long SetBounds(long in_top, long in_left, long in_height, long in_width);

  // ReadFrameGray reads in the specified frame and converts it
  // (if necessary) to a grayscale_image.
  virtual grayscale_image * ReadNextFrameGray();
  virtual grayscale_image * ReadFrameGray(long number);

  // ReadFrame reads in the specified frame and converts it
  // (if necessary) to a color_image.
  virtual color_image * ReadNextFrame();
  virtual color_image * ReadFrame(long number);
  
  virtual ushort InitializeWriting();
  virtual ushort write_frame(color_image * image);
  virtual ushort write_frameGray(grayscale_image * gray);
  virtual ushort FinishWriting();
};


#endif VASSILIS_AVI_FILE_H
