#include "stdafx.h"

#include "opencv_utils.h"
#include "basics/video_common.h"
#include "basics/rle_encoding.h"
#include "basics/image_algorithms.h"
#include "basics/drawing_temp.h"
#include "basics/image_templates.h"

// If the OpenCV library has been installed, use the regular 
// definitions of the functions
//
// Note on OpenCV installation: In addition to the OpenCV libraries
// and DLLs, we need the IPL (image processing library) DLLs. I 
// just copy them into the same directory as the OpenCV DLLs 
// (Intel/OpenCV/lib).
#ifdef VASSILIS_OPENCV_LIBRARY

// Creates a 3-band (3-channel) image with the given dimensions
IplImage * IplColorImage(long rows, long cols)
{
  if (rows <= 0) return 0;
  if (cols <= 0) return 0;

  CvSize size;
  size.width = cols;
  size.height = rows;
  IplImage * ipl_image = cvCreateImage(size, IPL_DEPTH_8U, 3);
  ipl_image->origin = IPL_ORIGIN_TL;

  assert(memcmp(ipl_image->colorModel, "RGB", 4) == 0);
  assert(memcmp(ipl_image->channelSeq, "BGR", 4) == 0);
  return ipl_image;
}

// Creates a 1-band (grayscale) image with the given dimensions
IplImage * IplGrayImage(long rows, long cols)
{
  if (rows <= 0) return 0;
  if (cols <= 0) return 0;

  CvSize size;
  size.width = cols;
  size.height = rows;
  IplImage * ipl_image = cvCreateImage(size, IPL_DEPTH_8U, 1);
  ipl_image->origin = IPL_ORIGIN_TL;

  return ipl_image;
}


// Creates a 1-band (grayscale) image with the given dimensions
IplImage * IplInt32Image(long rows, long cols)
{
  if (rows <= 0) return 0;
  if (cols <= 0) return 0;

  CvSize size;
  size.width = cols;
  size.height = rows;
  IplImage * ipl_image = cvCreateImage(size, IPL_DEPTH_32S, 1);
  ipl_image->origin = IPL_ORIGIN_TL;

  return ipl_image;
}


// Creates a 1-band (grayscale) image with the given dimensions
IplImage * IplFloatImage(long rows, long cols)
{
  if (rows <= 0) return 0;
  if (cols <= 0) return 0;

  CvSize size;
  size.width = cols;
  size.height = rows;
  IplImage * ipl_image = cvCreateImage(size, IPL_DEPTH_32F, 1);
  ipl_image->origin = IPL_ORIGIN_TL;

  return ipl_image;
}


color_image * vIplTocolor_image(IplImage * ipl_image)
{
  if (ipl_image == 0) return 0;
  
  // First make sure that ipl_image is in a format that my code
  // is ready to handle. I don't take care of all possibilities.

  // OpenCV requires that the pixel values are interleaved, and not 
  // stored in separate planes.
  if (ipl_image->dataOrder != IPL_DATA_ORDER_PIXEL) return 0;
  // maskROI and tileInfo must be 0;
  if (ipl_image->maskROI != 0) return 0;
  if (ipl_image->tileInfo != 0) return 0;
  long pixel_depth = ipl_image->depth;
  if (pixel_depth != IPL_DEPTH_8U) return 0;

  // The OpenCV manual says that the colorModel and channelSeq fields are 
  // ignored by OpenCV. I would still rather verify that they are what I
  // expect ("RGB" for colorModel, "BGR" for channelSeq).
  char * color_model = ipl_image->colorModel;
  char * band_order = ipl_image->channelSeq;
  if (strcmp(color_model, "RGB") != 0) return 0;
  if (strcmp(band_order, "BGR") != 0) return 0;

  // Now we know that we can handle ipl_image. We read in the fields
  // that will give us the necessary information to convert it.
  long width_step = ipl_image->widthStep;
  long origin = ipl_image->origin;
  long rows = ipl_image->height;
  long cols = ipl_image->width;
  long bands = ipl_image->nChannels;
  uchar * ipl_data = 0;
  cvGetImageRawData(ipl_image, &ipl_data, 0, 0);

  // Now we create the result image and write its pixel values;
  color_image * the_image = new color_image(rows, cols);
  vArray(uchar) R = the_image->GetR();
  vArray(uchar) G = the_image->GetG();
  vArray(uchar) B = the_image->GetB();

  long row, col;
  long v_index = 0;
  long ipl_index;
  long ipl_row, ipl_row_increment;
  
  // If the origin is top left, then, as we read ipl_data from
  // left to right we are reading the pixels of ipl_image from
  // top to bottom.
  if (origin == IPL_ORIGIN_TL)
  {
    ipl_row = 0;
    ipl_row_increment = 1;
  }
  else // In this case ipl_data has the pixels in order from bottom to top.
  {
    assert(origin == IPL_ORIGIN_BL);
    ipl_row = rows - 1;
    ipl_row_increment = -1;
  }
  
  for(row = 0; row < rows; row++)
  {
    ipl_index = ipl_row * width_step;
    for (col = 0; col < cols; col++)
    {
      B[v_index] = ipl_data[ipl_index++];
      G[v_index] = ipl_data[ipl_index++];
      R[v_index] = ipl_data[ipl_index++];
      v_index++;
    }
    ipl_row += ipl_row_increment;
  }

  return the_image;
}


IplImage * vvImageToIpl(color_image * the_image)
{
  if (the_image == 0) return 0;

  CvSize size;
  long rows = the_image->Rows();
  long cols = the_image->Cols();
  size.width = cols;
  size.height = rows;
  long pixels = the_image->Size();
  IplImage * ipl_image = cvCreateImage(size, IPL_DEPTH_8U, 3);
  ipl_image->origin = IPL_ORIGIN_TL;

  // widthStep is the number of bytes between points on the same column
  // in successive rows.
  long width_step = ipl_image->widthStep;
  assert(memcmp(ipl_image->colorModel, "RGB", 4) == 0);
  assert(memcmp(ipl_image->channelSeq, "BGR", 4) == 0);
  vArray(uchar) R = the_image->GetR();
  vArray(uchar) G = the_image->GetG();
  vArray(uchar) B = the_image->GetB();

  // ipl_data will contain the raw data of ipl_image. I was wondering
  // if I should use ipl_image->imageData or ipl_image->imageDataOrigin
  // I am using cvGetImageRawData now, but I'm not sure if it is 
  // the right thing.
  uchar * ipl_data = 0;
  cvGetImageRawData(ipl_image, &ipl_data, 0, 0);
  long row, col;
  long v_index = 0;
  long ipl_index;

  for (row = 0; row < rows; row++)
  {
    ipl_index = row * width_step;
    for (col = 0; col < cols; col++)
    {
      ipl_data[ipl_index++] = B[v_index];
      ipl_data[ipl_index++] = G[v_index];
      ipl_data[ipl_index++] = R[v_index];
      v_index++;
    }
  }

  return ipl_image;
}


// Converts the given ipl image to a grayscale_image, assuming that the ipl image
// is a 1-channel 8-bits/pixel/band image
grayscale_image * vIplToGray(IplImage * ipl_image)
{
  if (ipl_image == 0) return 0;
  // If ipl_image is not a one-band image return 0;
  if (ipl_image->nChannels != 1) return 0;
  long rows = ipl_image->height;
  long cols = ipl_image->width;
  grayscale_image * result = new grayscale_image(rows, cols);
  ushort success = vIplImageToBand(ipl_image, result, 0);
  if (success == 0)
  {
    vdelete(result);
    return(0);
  }
  else
  {
    return(result);
  }
}


// Converts the given ipl image to a grayscale_image, assuming that the ipl image
// is a 1-channel 8-bits/pixel/band image
v3dMatrix<float> * vIplToFloat(IplImage * ipl_image)
{
  if (ipl_image == 0) return 0;
  // If ipl_image is not a one-band image return 0;
  if (ipl_image->nChannels != 1) return 0;
  long rows = ipl_image->height;
  long cols = ipl_image->width;
  v3dMatrix<float> * result = new v3dMatrix<float>(rows, cols, 1);
  ushort success = vIplFloatImageToBand(ipl_image, result, 0);
  if (success == 0)
  {
    vdelete(result);
    return(0);
  }
  else
  {
    return(result);
  }
}


// Converts from a grayscale_image to an IplImage
IplImage * vGrayToIpl(grayscale_image * gray)
{
  IplImage * result = vBandToIplImage(gray, 0);
  return result;
}


// Creates a 1-band ipl image containing the specified band 
// of color_image
IplImage * vBandToIplImage(v3dMatrix<uchar> * the_image, long band)
{
  if (the_image == 0) return 0;
  long rows = the_image->Rows();
  long cols = the_image->Cols();

  // Create the result image
  IplImage * result = IplGrayImage(rows, cols);
  if (result == 0) return 0;
  if (the_image == 0) return 0;
  if ((band < 0) || (band >= the_image->Bands())) return 0;
  // Get the input band from the input image
  vArray(uchar) matrix = the_image->Matrix(band);

  // ipl_data will contain the raw data of ipl_image. I was wondering
  // if I should use ipl_image->imageData or ipl_image->imageDataOrigin
  // I am using cvGetImageRawData now, but I'm not sure if it is 
  // the right thing.
  uchar * ipl_data = 0;
  cvGetImageRawData(result, &ipl_data, 0, 0);
  long width_step = result->widthStep;
  long row, col;
  long v_index = 0;
  long ipl_index;

  // Copy the input band into the ipl image
  for (row = 0; row < rows; row++)
  {
    // Figure out the start of the pixels of the current row in the ipl image
    ipl_index = row * width_step;
    for (col = 0; col < cols; col++)
    {
      ipl_data[ipl_index++] = matrix[v_index];
      v_index++;
    }
  }
  return result;
}

// Copies over the pixel values from the source image to the
// specified band of the target image ,
// assuming that the source is a one-band 8 bits/pixel image.
ushort vIplImageToBand(IplImage * source, v3dMatrix<uchar> * target, long band)
{
  if (source == 0) return 0;
  if (target == 0) return 0;
  if ((band < 0) || (band >= target->Bands())) return 0;
  long bands = source->nChannels;
  if (bands != 1) return 0;
  
  // First make sure that ipl_image is in a format that my code
  // is ready to handle. I don't take care of all possibilities.

  // OpenCV requires that the pixel values are interleaved, and not 
  // stored in separate planes.
  if (source->dataOrder != IPL_DATA_ORDER_PIXEL) return 0;
  // maskROI and tileInfo must be 0;
  if (source->maskROI != 0) return 0;
  if (source->tileInfo != 0) return 0;
  long pixel_depth = source->depth;
  if (pixel_depth != IPL_DEPTH_8U) return 0;

  // The OpenCV manual says that the colorModel and channelSeq fields are 
  // ignored by OpenCV. I would still rather verify that they are what I
  // expect ("RGB" for colorModel, "BGR" for channelSeq).
  char * color_model = source->colorModel;
  char * band_order = source->channelSeq;

  // Now we know that we can handle ipl_image. We read in the fields
  // that will give us the necessary information to convert it.
  long width_step = source->widthStep;
  long origin = source->origin;
  long rows = source->height;
  long cols = source->width;
  uchar * ipl_data = 0;
  cvGetImageRawData(source, &ipl_data, 0, 0);

  // Now we create the result image and write its pixel values;
  vArray(uchar) matrix = target->Matrix(band);

  long row, col;
  long v_index = 0;
  long ipl_index;
  long ipl_row, ipl_row_increment;
  
  // If the origin is top left, then, as we read ipl_data from
  // left to right we are reading the pixels of ipl_image from
  // top to bottom.
  if (origin == IPL_ORIGIN_TL)
  {
    ipl_row = 0;
    ipl_row_increment = 1;
  }
  else // In this case ipl_data has the pixels in order from bottom to top.
  {
    assert(origin == IPL_ORIGIN_BL);
    ipl_row = rows - 1;
    ipl_row_increment = -1;
  }
  
  for(row = 0; row < rows; row++)
  {
    ipl_index = ipl_row * width_step;
    for (col = 0; col < cols; col++)
    {
      matrix[v_index] = ipl_data[ipl_index++];
      v_index++;
    }
    ipl_row += ipl_row_increment;
  }

  return 1;
}


// Copies over the pixel values from the source image to the
// specified band of the target image ,
// assuming that the source is a one-band 32 bits/pixel float image.
ushort vIplFloatImageToBand(IplImage * source, v3dMatrix<float> * target, long band)
{
  if (source == 0) return 0;
  if (target == 0) return 0;
  if ((band < 0) || (band >= target->Bands())) return 0;
  long bands = source->nChannels;
  if (bands != 1) return 0;
  
  // First make sure that ipl_image is in a format that my code
  // is ready to handle. I don't take care of all possibilities.

  // OpenCV requires that the pixel values are interleaved, and not 
  // stored in separate planes.
  if (source->dataOrder != IPL_DATA_ORDER_PIXEL) return 0;
  // maskROI and tileInfo must be 0;
  if (source->maskROI != 0) return 0;
  if (source->tileInfo != 0) return 0;
  long pixel_depth = source->depth;
  if (pixel_depth != IPL_DEPTH_32F) return 0;

  // The OpenCV manual says that the colorModel and channelSeq fields are 
  // ignored by OpenCV. I would still rather verify that they are what I
  // expect ("RGB" for colorModel, "BGR" for channelSeq).
  char * color_model = source->colorModel;
  char * band_order = source->channelSeq;

  // Now we know that we can handle ipl_image. We read in the fields
  // that will give us the necessary information to convert it.
  long width_step = source->widthStep / sizeof(float);
  long origin = source->origin;
  long rows = source->height;
  long cols = source->width;
  uchar * temp_ipl_data = 0;
  cvGetImageRawData(source, &temp_ipl_data, 0, 0);
  float * ipl_data = (float *) temp_ipl_data;

  // Now we create the result image and write its pixel values;
  vArray(float) matrix = target->Matrix(band);

  long row, col;
  long v_index = 0;
  long ipl_index;
  long ipl_row, ipl_row_increment;
  
  // If the origin is top left, then, as we read ipl_data from
  // left to right we are reading the pixels of ipl_image from
  // top to bottom.
  if (origin == IPL_ORIGIN_TL)
  {
    ipl_row = 0;
    ipl_row_increment = 1;
  }
  else // In this case ipl_data has the pixels in order from bottom to top.
  {
    assert(origin == IPL_ORIGIN_BL);
    ipl_row = rows - 1;
    ipl_row_increment = -1;
  }
  
  for(row = 0; row < rows; row++)
  {
    ipl_index = ipl_row * width_step;
    for (col = 0; col < cols; col++)
    {
      matrix[v_index] = ipl_data[ipl_index++];
      v_index++;
    }
    ipl_row += ipl_row_increment;
  }

  return 1;
}


// Returns the results of Canny edge detection 
grayscale_image * vCannyDetector(grayscale_image * source, uchar low)
{
  uchar high;
  if (low < 128) high = 2 * low;
  else high = 255;
  return vCannyDetector(source, low, high, 3);
}

                              
// Returns the results of Canny edge detection 
grayscale_image * vCannyDetector(grayscale_image * source, uchar low, 
                              uchar high, long aperture)
{
  // The cvCanny function will crash the program if the following conditions
  // are violated
  if (low < 0) return 0;
  if (high < 0) return 0;
  if (low > 255) return 0;
  if (high > 255) return 0;
  if (aperture < 1) return 0;
  if ((aperture % 2) != 1) return 0;

  // Convert input to the IPL format
  IplImage * ipl_input = vGrayToIpl(source);
  if (ipl_input == 0) return 0;
  // Create output IPL image
  IplImage * ipl_output = IplGrayImage(source->Rows(), source->Cols());
  // Call the Canny edge detector function
  cvCanny(ipl_input, ipl_output, low, high, aperture);
  // Convert the result to grayscale_image
  grayscale_image * result = vIplToGray(ipl_output);
  cvReleaseImage(&ipl_input);
  cvReleaseImage(&ipl_output);
  return result;
}


ushort vSequenceCannyEdges(const char * input_file, const char * arg_output_file,
                            uchar low, uchar high, long aperture)
{
  vVideoFile * in_video = vOpenReadVideoFile(input_file);
  if (in_video == 0)
  {
    vPrint("Can't open %s\n", input_file);
    return 0;
  }
  long rows = in_video->Rows();
  long cols = in_video->Cols();
  char * output_file = vReplaceEnding2(arg_output_file, ".rle");
  vRLE_Sequence * out_video = new vRLE_Sequence(output_file, rows, cols);
  if (out_video == 0)
  {
    vPrint("Can't open %s\n", output_file);
    vdelete2(output_file);
    vCloseVideoFile(in_video);
    return 0;
  }
  long frame = 0;
  ushort result = 1;
  while(1)
  {
    grayscale_image * in_frame = in_video->ReadFrameGray(frame);
    if (in_frame == 0) break;
    grayscale_image * edges = vCannyDetector(in_frame, low, high, aperture);
    vdelete(in_frame);
    if (edges == 0)
    {
      vPrint("Error, couldn't get edges\n");
      result = 0;
      break;
    }
     
    ushort success = out_video->write_frameGray2(edges, 0, 0);
    vdelete(edges);
    if (success == 0)
    {
      vPrint("Failed to write frame %li as binary\n", frame);
      result = 0;
      break;
    }
    else
    {
      vPrint("Processed frame %li\r", frame);
    }
  
    frame++;
  }
  vPrint("\n");
  vdelete2(output_file);
  vCloseVideoFile(in_video);
  out_video->FinishWriting();
  vdelete(out_video);
  return result;
}


// Returns the results of distance transform using the Borgefors algorithm.
// I use non-zero values for the feature points, but CV expects that
// the feature points are marked with 0. So, I need to convert from
// my format to the format that CV wants.
v3dMatrix<float> * vCVDistanceTransform(general_image * input_base)
{
  grayscale_image * input = (grayscale_image *) input_base->threshold(1);
  vInverseImage(input);
  IplImage * ipl_input = vGrayToIpl(input);

  long rows = input->Rows();
  long cols = input->Cols();
  IplImage * ipl_result = IplFloatImage(rows, cols);
  cvDistTransform(ipl_input, ipl_result, CV_DIST_L2, CV_DIST_MASK_5, 0);
//  grayscale_image * result = vIplToGray(ipl_result);
  v3dMatrix<float> * result = vIplToFloat(ipl_result);
  vdelete(input);
  cvReleaseImage(&ipl_input);
  cvReleaseImage(&ipl_result);
  return result;
}


// Returns the 7 normalized central moments and the 7 Hu moments for 
// the image. Puts the number 14 in the argument "number".
double * vImageMoments(general_image * input, long * number)
{
  grayscale_image * gray = input->ToGray();
  IplImage * ipl_input = vGrayToIpl(gray);

  CvMoments moments;
  // The third argument, 1, says "treat the input image as binary".
  cvMoments(ipl_input, &moments, 1);
  if (gray != input) vdelete(gray);
  cvReleaseImage(&ipl_input);

  CvHuMoments hu_moments;
  cvGetHuMoments(&moments, &hu_moments);

  double * result = new double[14];
  long index = 0;
  result[index++] = cvGetNormalizedCentralMoment(&moments, 1, 1);
  result[index++] = cvGetNormalizedCentralMoment(&moments, 2, 0);
  result[index++] = cvGetNormalizedCentralMoment(&moments, 2, 1);
  result[index++] = cvGetNormalizedCentralMoment(&moments, 3, 0);
  result[index++] = cvGetNormalizedCentralMoment(&moments, 0, 2);
  result[index++] = cvGetNormalizedCentralMoment(&moments, 1, 2);
  result[index++] = cvGetNormalizedCentralMoment(&moments, 0, 3);

  result[index++] = hu_moments.hu1;
  result[index++] = hu_moments.hu2;
  result[index++] = hu_moments.hu3;
  result[index++] = hu_moments.hu4;
  result[index++] = hu_moments.hu5;
  result[index++] = hu_moments.hu6;
  result[index++] = hu_moments.hu7;
  assert(index == 14);
  *number = index;

  return result;
}


long vConvexHull(vArray(long) rows, vArray(long) cols, int number_of_points, 
                  vArray(long) * hull_rowsp, vArray(long) * hull_colsp, 
                  long * hull_sizep)
{
  if (number_of_points < 1)
  {
    *hull_sizep = 0;
    return 0;
  }
  CvPoint * cv_points = new CvPoint[number_of_points];
  long i;
  for (i = 0; i < number_of_points; i++)
  {
    cv_points[i].x = cols[i];
    cv_points[i].y = rows[i];
  }

  int * cv_hull = new int[number_of_points];
  int cv_hull_size = 0;
  cvConvexHull(cv_points, number_of_points, 0, CV_CLOCKWISE,
               cv_hull, &cv_hull_size);

  long hull_size = cv_hull_size;
  vArray(long) hull_rows = vnew(long, hull_size);
  vArray(long) hull_cols = vnew(long, hull_size);
  for (i = 0; i < hull_size; i++)
  {
    long index = cv_hull[i];
    hull_rows[i] = rows[index];
    hull_cols[i] = cols[index];
  }
  
  delete [] cv_points;
  delete [] cv_hull;
  *hull_rowsp = hull_rows;
  *hull_colsp = hull_cols;
  *hull_sizep = hull_size;
  return 1;
}


// Input parameters: color_image, quality, min_distance.
// Output parameters: locations.
long vGoodFeaturesToTrack(grayscale_image * the_image, vector<vPoint2> * locations,
                           double quality, double min_distance)
{
  const long max_features = 10000;

  IplImage * ipl_image = vGrayToIpl(the_image);

  long rows = the_image->Rows();
  long cols = the_image->Cols();

  IplImage * temp1 = IplFloatImage(rows, cols);
  IplImage * temp2 = IplFloatImage(rows, cols);
  vArray(CvPoint2D32f) cv_locations = vnew(CvPoint2D32f, max_features);
  int count = max_features;

  cvGoodFeaturesToTrack(ipl_image, temp1, temp2, vData(cv_locations),
                        &count, quality, min_distance);

  long i;
  for (i = 0; i < count; i++)
  {
    double row = cv_locations[i].y;
    double col = cv_locations[i].x;
    locations->push_back(vPoint2(row, col));
  }


  vdelete2(cv_locations);
  cvReleaseImage(&temp1);
  cvReleaseImage(&temp2);
  cvReleaseImage(&ipl_image);
  return count;
}


long vOpticalFlowPyrLK(grayscale_image * the_image1, grayscale_image * the_image2,
                        vector<vPoint2> * locations1, vector<vPoint2> * locations2,
                        long window_size, long pyramids,
                        vector<long> * status, vector<double> * error)
{
  long rows = the_image1->Rows();
  long cols = the_image1->Cols();
  if ((rows != the_image2->Rows()) || (cols != the_image2->Cols()))
  {
    return 0;
  }

  // Make sure that the input images have the right size, so that
  // calling the OpenCV function will not cause the program to crash.
  long padded_rows = rows;
  long padded_cols = cols;
  long factor = 2 << pyramids;
  vPrint("factor = %li\n", factor);
  long rows_remainder = rows % factor;
  long cols_remainder = cols % factor;
  if (rows_remainder != 0)
  {
    padded_rows = rows - rows_remainder + factor;
  }
  if (cols_remainder != 0)
  {
    padded_cols = cols - cols_remainder + factor;
  }

  grayscale_image * padded1, * padded2;
  if ((rows_remainder == 0) && (cols_remainder == 0))
  {
    padded1 = the_image1;
    padded2 = the_image2;
  }
  else
  {
    padded1 = (grayscale_image *) vPad(the_image1, padded_rows, padded_cols);
    padded2 = (grayscale_image *) vPad(the_image2, padded_rows, padded_cols);
  }



  IplImage * ipl_image1 = vGrayToIpl(padded1);
  IplImage * ipl_image2 = vGrayToIpl(padded2);

  IplImage * pyramid1 = IplGrayImage(padded_rows, padded_cols+8);
  IplImage * pyramid2 = IplGrayImage(padded_rows, padded_cols+8);

  long count = locations1->size();
  vArray(CvPoint2D32f) cv_locations1 = vCvPoints2D32f(locations1);
  vArray(CvPoint2D32f) cv_locations2 = vnew(CvPoint2D32f, count);

  vArray(char) status_array = vnew(char, count);
  vArray(float) error_array = vnew(float, count);

  vPrint("window size = %li, pyramids = %li\n", window_size, pyramids);

  cvCalcOpticalFlowPyrLK(ipl_image1, ipl_image2, pyramid1, pyramid2,
                         vData(cv_locations1), vData(cv_locations2), count,
                         cvSize(window_size, window_size), pyramids, 
                         vData(status_array), vData(error_array),
                         cvTermCriteria(CV_TERMCRIT_ITER|CV_TERMCRIT_EPS,20,0.03), 0);

  long i;
  for (i = 0; i < count; i++)
  {
    double row = cv_locations2[i].y;
    double col = cv_locations2[i].x;
    if (row >= rows-1) row = rows-1;
    else if (row < 0) row = 0;
    if (col >= cols-1) col = cols-1;
    else if (col < 0) col = 0;

    locations2->push_back(vPoint2(row, col));
    status->push_back(status_array[i]);
    error->push_back(error_array[i]);
  }

  vdelete2(status_array);
  vdelete2(error_array);
  vdelete2(cv_locations2);
  vdelete2(cv_locations1);
  cvReleaseImage(&pyramid2);
  cvReleaseImage(&pyramid1);
  cvReleaseImage(&ipl_image2);
  cvReleaseImage(&ipl_image1);
  if (padded1 != the_image1) vdelete(padded1);
  if (padded2 != the_image2) vdelete(padded2);

  return 1;
}


// Input parameters: the_image1, the_image2, locations1, window_size, pyramids
// Output parameters: locations2, status, error.
// 
// The function cvCalcOpticalFlowPyrLK will crash if the rows or cols
// of the input images are not multiples of 2 to the power of pyramids.
// Because of that, we pad the images with zeros, and at the end we
// adjust the locations to make sure they are within bounds.
vArray(CvPoint2D32f) vCvPoints2D32f(vector<vPoint2> * points)
{
  long count = points->size();
  vArray(CvPoint2D32f) result = vnew(CvPoint2D32f, count);

  long i;
  for (i = 0; i < count; i++)
  {
    result[i].x = (float) (*points)[i].col;
    result[i].y = (float) (*points)[i].row;
  }
  return result;
}

// This is just included as a sample complete program
//static int vMain(int argc, char ** argv)
//{
//  vReadAVI avi("G:\\users\\athitsos\\sequences\\large\\vassilis_gestures_851_slave2_color.avi");
//  color_image * the_image = avi.ReadFrame(0);
//  IplImage * ipl_image = vvImageToIpl(color_image);
//  IplImage * ipl_image2 = vvImageToIpl(color_image);
//  cvCanny(ipl_image, ipl_image2, 10, 250, 3);
//  color_image * the_image2 = vIplTocolor_image(ipl_image2);
//  the_image2->Save("G:\\users\\athitsos\\sequences\\851_edges.0.ppm");
//
//  v3dMatrix<double> * difference = vImageDifference(color_image, the_image2);
//  double max = function_image_maximum(difference);
//  printf("max = %li\n", max);
//
//  vdelete(color_image);
//  vdelete(the_image2);
//  cvReleaseImage(&ipl_image);
//
//  return 0;
//}

// If the OpenCV library has not been installed (VASSSILIS_OPENCV_LIBRARY
// not defined) then use empty definitions for the functions.
#else

// Returns the results of Canny edge detection 
grayscale_image * vCannyDetector(grayscale_image * source, uchar low)
{
  vPrint("Can't do canny detection. OpenCV not installed\n");
  return 0;
}


// Returns the results of Canny edge detection 
grayscale_image * vCannyDetector(grayscale_image * source, uchar low, 
                              uchar high, long aperture)
{
  vPrint("Can't do canny detection. OpenCV not installed\n");
  return 0;
}


ushort vSequenceCannyEdges(const char * input_file, const char * arg_output_file,
                            uchar low, uchar high, long aperture)
{
  vPrint("Can't do canny detection. OpenCV not installed\n");
  return 0; 
}


// Returns the 7 normalized central moments and the 7 Hue moments for 
// the image. Puts the number 14 in the argument "number".
double * vImageMoments(general_image * input, long * number)
{
  vPrint("Can't get image moments. OpenCV not installed\n");
  return 0;
}


v3dMatrix<float> * vCVDistanceTransform(general_image * input_base)
{
  vPrint("Can't compute distance transform. OpenCV not installed\n");
  return 0;
}


long vConvexHull(vArray(long) rows, vArray(long) cols, int number_of_points, 
                  vArray(long) * hull_rowsp, vArray(long) * hull_colsp, 
                  long * hull_sizep)
{
  vPrint("Can't compute vConvexHull. OpenCV not installed\n");
  return 0;
}


// Input parameters: color_image, quality, min_distance.
// Output parameters: locations.
long vGoodFeaturesToTrack(grayscale_image * the_image, vector<vPoint2> * locations,
                           double quality, double min_distance)
{
  vPrint("Can't compute vGoodFeaturesToTrack. OpenCV not installed\n");
  return 0;
}


long vOpticalFlowPyrLK(grayscale_image * the_image1, grayscale_image * the_image2,
                        vector<vPoint2> * locations1, vector<vPoint2> * locations2,
                        long window_size, long pyramids,
                        vector<long> * status, vector<double> * error)
{
  vPrint("Can't compute vOpticalFlowPyrLK. OpenCV not installed\n");
  return 0;
}


vArray(CvPoint2D32f) vCvPoints2D32f(vector<vPoint2> * points)
{
  vPrint("Can't compute vCvPoints2D32f. OpenCV not installed\n");
  return 0;
}

color_image * vIplTocolor_image(IplImage * ipl_image)
{
  vPrint("Can't compute vIplTocolor_image. OpenCV not installed\n");
  return 0;
}

#endif // VASSILIS_OPENCV_LIBRARY


// Saves the 14 moments (7 normalized central moments, 7 Hu moments)
// of the given sequence into the given file. Returns a number > 0
// if successful. The difference with vSequenceImageMoments is that
// here, for every frame, we also rotate it by 90, 180 and 270 degrees,
// and save the moments of those frames, too.
short vSequenceImageMoments(const char * sequence_file, 
                             const char * moments_file)
{
  short result = 1;
  vVideoFile * sequence = vOpenReadVideoFile(sequence_file);
  if (sequence->IsValid() == 0) 
  {
    vCloseVideoFile(sequence);
    return 0;
  }

  FILE * out_fp = fopen(moments_file, vFOPEN_WRITE);
  if (out_fp == 0)
  {
    vCloseVideoFile(sequence);
    return -1;
  }

  long frame = 0;
  vPrint("\n");
  while(1)
  {
    grayscale_image * gray;
    long number, items;
    double * moments;

    // Read frame from video sequence
    gray = sequence->ReadFrameGray(frame);
    if (gray == 0) break;

    // Save moments for the frame
    moments = vImageMoments(gray, &number);
    items = fwrite(moments, sizeof(double), number, out_fp);
    vdelete2(moments);
    if (items != number) 
    {
      result = -3;
      vdelete(gray);
      break;
    }
  
    vdelete(gray);
    vPrint("\rProcessed frame %li", frame);
    frame++;
  }

  vPrint("\n\n");
  fclose(out_fp);
  vCloseVideoFile(sequence);
  return result;
}


// Saves the 14 moments (7 normalized central moments, 7 Hu moments)
// of the given sequence into the given file. Returns a number > 0
// if successful. The difference with vSequenceImageMoments is that
// here, for every frame, we also rotate it by 90, 180 and 270 degrees,
// and save the moments of those frames, too.
short vSequenceImageMoments_b(const char * sequence_file, 
                             const char * moments_file)
{
  short result = 1;
  vVideoFile * sequence = vOpenReadVideoFile(sequence_file);
  if (sequence->IsValid() == 0) 
  {
    vCloseVideoFile(sequence);
    return 0;
  }

  FILE * out_fp = fopen(moments_file, vFOPEN_WRITE);
  if (out_fp == 0)
  {
    vCloseVideoFile(sequence);
    return -1;
  }

  long frame = 0;
  vPrint("\n");
  while(1)
  {
    grayscale_image * gray, * temp;
    long number, items;
    double * moments;

    // Read frame from video sequence
    gray = sequence->ReadFrameGray(frame);
    if (gray == 0) break;

    // Save moments for the frame
    moments = vImageMoments(gray, &number);
    items = fwrite(moments, sizeof(double), number, out_fp);
    vdelete2(moments);
    if (items != number) 
    {
      result = -3;
      vdelete(gray);
      break;
    }
  
    // Save moments for the frame rotated 90 degrees to the right.
    temp = gray;
    gray = (grayscale_image *) vRotateRight(temp);
    vdelete(temp);

    moments = vImageMoments(gray, &number);
    items = fwrite(moments, sizeof(double), number, out_fp);
    vdelete2(moments);
    if (items != number) 
    {
      result = -3;
      vdelete(gray);
      break;
    }


    // Save moments for the frame rotated 180 degrees to the right.
    temp = gray;
    gray = (grayscale_image *) vRotateRight(temp);
    vdelete(temp);

    moments = vImageMoments(gray, &number);
    items = fwrite(moments, sizeof(double), number, out_fp);
    vdelete2(moments);
    if (items != number) 
    {
      result = -3;
      vdelete(gray);
      break;
    }


    // Save moments for the frame rotated 270 degrees to the right.
    temp = gray;
    gray = (grayscale_image *) vRotateRight(temp);
    vdelete(temp);

    moments = vImageMoments(gray, &number);
    items = fwrite(moments, sizeof(double), number, out_fp);
    vdelete2(moments);
    if (items != number) 
    {
      result = -3;
      vdelete(gray);
      break;
    }

    vdelete(gray);
    vPrint("\rProcessed frame %li", frame);
    frame++;
  }

  vPrint("\n\n");
  fclose(out_fp);
  vCloseVideoFile(sequence);
  return result;
}



///////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////

