/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#ifndef __CVERROR_INT_H_
#define __CVERROR_INT_H_

#include "cverror.h"

/*
// Note:
// Before using macros CV_ERROR, CV_CHECK, CV_CALL, and CV_ASSERT in your
// function, you must define two object in function: string constant FuncName
// and label exit. You can define string constant by macro CV_FUNCNAME or
// manually. If you use features from header "_cvExit.h", label exit will be
// defined by it's macros, otherwise you must define this label explicitly.
*/

/*
// You can define string constant FuncName (required by CV_ERROR) by macro
// CV_FUNCNAME, for example:
//     CV_FUNCNAME( "cvDoSomething" );
*/
#define CV_FUNCNAME( Name )                                            \
    static char FuncName[] = Name



typedef struct _CvContext
{
    CVStatus        CVLastStatus;
    int             CVErrMode;
    CVErrorCallBack CVErrorFunc;
    CvStackRecord*  CVStack;
    int             CVStackSize;
    int             CVStackCapacity;
    
} CvContext;

OPENCVAPI void icvInitContext(CvContext* context);
OPENCVAPI void icvDestroyContext(CvContext* context);
OPENCVAPI CvContext* icvGetContext();

/*
// This macro unconditionally raises error with passed code and message.
// After raising error, control will be transferred to the exit label. For
// example:
//     if ( a == 0 ) CV_ERROR( IPL_StsBadArg, "Bad argument \"a\"." );
*/
#define CV_ERROR( Code, Msg )                                                  \
{                                                                              \
     cvError( (Code), (FuncName), Msg, __FILE__, __LINE__ );                   \
     EXIT;                                                                     \
}


#define CV_ERR_STATUS  cvGetErrStatus


/*
// You can use macro CV_CHECK for checking error status after CV (or IPL)
// function call. If error detected, control will be transferred to the exit
// label. For exapmle:
//     cvDoSomething( 1, 2, 3 );
//     CV_CHECK();
*/
#define CV_CHECK()                                                     \
{                                                                      \
    if( CV_ERR_STATUS() < 0 )                                          \
        CV_ERROR( CV_StsBackTrace, "Inner function failed." );        \
}

/*
// Use this macro to call CV (or IPL) procedure (i. e. function with void
// return valiu) and check error status after calling. If called function
// failed, control will be transferred to the exit label.
//     CV_CALL( cvDoSomething( 1, 2, 3 ) );
*/
#define CV_CALL( Func )                                                \
{                                                                      \
    /* start profile */                                                \
    CV_START_CALL( #Func, __FILE__, __LINE__ );                        \
    Func;                                                              \
    CV_END_CALL( __FILE__, __LINE__ );                                 \
    CV_CHECK();                                                        \
}

#define CV_ERROR_FROM_STATUS( result )                \
    CV_ERROR( icvErrorFromStatus( result ), "OpenCV function failed" );

/*
// This macro asserts given condition. If condition evaliated as false (i. e.
// zero), error will be raised with code IPL_StsInternal, and control will be
// transferred to the exit label. If condition evaluated as true, normal
// execution will continue (control will flow to the next statement). For
// example:
//     CV_ASSERT( i != 0 );
*/
#define CV_ASSERT( Condition )                                           \
{                                                                        \
    if( !(Condition) ) CV_ERROR( CV_StsInternal, "Assertion failed." ); \
}


#define IPPI_CALL( Func )                                              \
{                                                                      \
      CvStatus  ippi_call_result;                                      \
      CV_START_CALL( #Func, __FILE__, __LINE__ );                      \
      ippi_call_result = Func;                                         \
      CV_END_CALL( __FILE__, __LINE__ );                               \
                                                                       \
      if( ippi_call_result < 0 )                                       \
            CV_ERROR_FROM_STATUS( (ippi_call_result));                 \
}

#endif /* __CVERROR_INT_H_ */

/* End of file. */

