/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*//*M///////////////////////////////////////////////////////////////////////////////////////
   //
   //                      INTEL CORPORATION PROPRIETARY INFORMATION
   //         This software is supplied under the terms of a license agreement or
   //         nondisclosure agreement with Intel Corporation and may not be copied
   //         or disclosed except in accordance with the terms of that agreement.
   //               Copyright (c) 2000 Intel Corporation. All Rights Reserved.
   //
   //    RCS:
   //       $Source:   icvNorm_Inf.cpp$
   //       $Revision: 00.00.05$
   //      Purpose:
   //      Contents:
   //      Authors:
   //        Dmitry Abrosimov
   //M */

#include "_cv.h"

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Names:   icvNorm_Inf_8u_C1MR,  icvNorm_Inf_8s_C1MR,  icvNorm_Inf_32f_C1MR, 
//             icvNorm_Inf_8u_C3CMR, icvNorm_Inf_8s_C3CMR, icvNorm_Inf_32f_C3CMR, 
//
//    Purpose: Calculate infinite norm of an image with mask
//
//    Parameters:
//         pSrc, pMask       - pointers to source image and mask array,
//         srcStep, maskStep - their steps,
//         roiSize           - their size or ROI size,
//         coi               - COI index (1..3) (if 3-channel image).
//
//    Returns:
//         Cv status
//
//    Notes:
//F*/

/* ///////////////////////////////// 8uC1 flavor ////////////////////////////////////// */
IPCVAPI_IMPL( CvStatus, icvNorm_Inf_8u_C1MR, (const uchar * pSrc,
                                             int srcStep,
                                             const uchar * pMask,
                                             int maskStep, CvSize roiSize, double *norm) )
{
    int i, j, s = 0;

/* parameters check */
    if( !pSrc )
        return CV_NULLPTR_ERR;
    if( !pMask )
        return CV_NULLPTR_ERR;
    if( srcStep < 1 )
        return CV_BADSTEP_ERR;
    if( maskStep < 1 )
        return CV_BADSTEP_ERR;
    if( roiSize.width < 1 || roiSize.height < 1 || roiSize.width > srcStep ||
        roiSize.width > maskStep )
        return CV_BADSIZE_ERR;

    if( roiSize.width == srcStep && roiSize.width == maskStep )
    {
        roiSize.width *= roiSize.height;
        roiSize.height = 1;
    }

    for( i = 0; i < roiSize.height; i++, pSrc += srcStep, pMask += maskStep )
    {
        for( j = 0; j < roiSize.width - 3; j += 4 )
        {
            int m = (pMask[j] == 0) - 1;
            int v = m & pSrc[j];

            if( s < v )
                s = v;
            m = (pMask[j + 1] == 0) - 1;
            v = m & pSrc[j + 1];
            if( s < v )
                s = v;
            m = (pMask[j + 2] == 0) - 1;
            v = m & pSrc[j + 2];
            if( s < v )
                s = v;
            m = (pMask[j + 3] == 0) - 1;
            v = m & pSrc[j + 3];
            if( s < v )
                s = v;
        }
        for( ; j < roiSize.width; j++ )
        {
            int m = (pMask[j] == 0) - 1;
            int v = m & pSrc[j];

            if( s < v )
                s = v;
        }
    }
    *norm = s;

    return CV_NO_ERR;
}

/* ///////////////////////////////// 8uC3 flavor ////////////////////////////////////// */
IPCVAPI_IMPL( CvStatus, icvNorm_Inf_8u_C3CMR, (const uchar * pSrc,
                                              int srcStep,
                                              const uchar * pMask,
                                              int maskStep,
                                              CvSize roiSize, int coi, double *norm) )
{
    int i, j, s = 0;

/* parameters check */
    if( !pSrc )
        return CV_NULLPTR_ERR;
    if( !pMask )
        return CV_NULLPTR_ERR;
    if( srcStep < 1 )
        return CV_BADSTEP_ERR;
    if( maskStep < 1 )
        return CV_BADSTEP_ERR;
    if( coi < 1 || coi > 3 )
        return CV_UNSUPPORTED_COI_ERR;
    if( roiSize.width < 1 || roiSize.height < 1 || roiSize.width > srcStep ||
        roiSize.width > maskStep )
        return CV_BADSIZE_ERR;

    if( roiSize.width == srcStep && roiSize.width == maskStep )
    {
        roiSize.width *= roiSize.height;
        roiSize.height = 1;
    }

    coi--;
    pSrc += coi;

    for( i = 0; i < roiSize.height; i++, pSrc += srcStep, pMask += maskStep )
    {
        for( j = 0; j < roiSize.width - 3; j += 4 )
        {
            int m = (pMask[j] == 0) - 1;
            int v = m & pSrc[3 * j];

            if( s < v )
                s = v;
            m = (pMask[j + 1] == 0) - 1;
            v = m & pSrc[3 * j + 3];
            if( s < v )
                s = v;
            m = (pMask[j + 2] == 0) - 1;
            v = m & pSrc[3 * j + 6];
            if( s < v )
                s = v;
            m = (pMask[j + 3] == 0) - 1;
            v = m & pSrc[3 * j + 9];
            if( s < v )
                s = v;
        }
        for( ; j < roiSize.width; j++ )
        {
            int m = (pMask[j] == 0) - 1;
            int v = m & pSrc[3 * j];

            if( s < v )
                s = v;
        }
    }
    *norm = s;

    return CV_NO_ERR;
}

/* //////////////////////////////////////////////////////////////////////////////////// */
/* //////////////////////////////////////////////////////////////////////////////////// */
/* //////////////////////////////////////////////////////////////////////////////////// */
/* ////////////////////////////////// 8sC1 flavor ///////////////////////////////////// */

IPCVAPI_IMPL( CvStatus, icvNorm_Inf_8s_C1MR, (const char *pSrc,
                                             int srcStep,
                                             const uchar * pMask,
                                             int maskStep, CvSize roiSize, double *norm) )
{
    int i, j, s = 0;

/* parameters check */
    if( !pSrc )
        return CV_NULLPTR_ERR;
    if( !pMask )
        return CV_NULLPTR_ERR;
    if( srcStep < 1 )
        return CV_BADSTEP_ERR;
    if( maskStep < 1 )
        return CV_BADSTEP_ERR;
    if( roiSize.width < 1 || roiSize.height < 1 || roiSize.width > srcStep ||
        roiSize.width > maskStep )
        return CV_BADSIZE_ERR;

    if( roiSize.width == srcStep && roiSize.width == maskStep )
    {
        roiSize.width *= roiSize.height;
        roiSize.height = 1;
    }

    for( i = 0; i < roiSize.height; i++, pSrc += srcStep, pMask += maskStep )
    {
        for( j = 0; j < roiSize.width - 3; j += 4 )
        {
            int m = (pMask[j] == 0) - 1;
            int v = pSrc[j] & m;

            v = _CV_ABS( v );
            if( s < v )
                s = v;
            m = (pMask[j + 1] == 0) - 1;
            v = pSrc[j + 1] & m;
            v = _CV_ABS( v );
            if( s < v )
                s = v;
            m = (pMask[j + 2] == 0) - 1;
            v = pSrc[j + 2] & m;
            v = _CV_ABS( v );
            if( s < v )
                s = v;
            m = (pMask[j + 3] == 0) - 1;
            v = pSrc[j + 3] & m;
            v = _CV_ABS( v );
            if( s < v )
                s = v;
        }
        for( ; j < roiSize.width; j++ )
        {
            int m = (pMask[j] == 0) - 1;
            int v = pSrc[j] & m;

            v = _CV_ABS( v );
            if( s < v )
                s = v;
        }
    }
    *norm = s;

    return CV_NO_ERR;
}


/* ///////////////////////////////// 8sC3 flavor ////////////////////////////////////// */
IPCVAPI_IMPL( CvStatus, icvNorm_Inf_8s_C3CMR, (const char *pSrc,
                                              int srcStep,
                                              const uchar * pMask,
                                              int maskStep,
                                              CvSize roiSize, int coi, double *norm) )
{
    int i, j, s = 0;

/* parameters check */
    if( !pSrc )
        return CV_NULLPTR_ERR;
    if( !pMask )
        return CV_NULLPTR_ERR;
    if( srcStep < 1 )
        return CV_BADSTEP_ERR;
    if( maskStep < 1 )
        return CV_BADSTEP_ERR;
    if( coi < 1 || coi > 3 )
        return CV_UNSUPPORTED_COI_ERR;
    if( roiSize.width < 1 || roiSize.height < 1 || roiSize.width > srcStep ||
        roiSize.width > maskStep )
        return CV_BADSIZE_ERR;

    if( roiSize.width == srcStep && roiSize.width == maskStep )
    {
        roiSize.width *= roiSize.height;
        roiSize.height = 1;
    }

    coi--;
    pSrc += coi;

    for( i = 0; i < roiSize.height; i++, pSrc += srcStep, pMask += maskStep )
    {
        for( j = 0; j < roiSize.width - 3; j += 4 )
        {
            int m = (pMask[j] == 0) - 1;
            int v = pSrc[3 * j] & m;

            v = _CV_ABS( v );
            if( s < v )
                s = v;
            m = (pMask[j + 1] == 0) - 1;
            v = pSrc[3 * j + 3] & m;
            v = _CV_ABS( v );
            if( s < v )
                s = v;
            m = (pMask[j + 2] == 0) - 1;
            v = pSrc[3 * j + 6] & m;
            v = _CV_ABS( v );
            if( s < v )
                s = v;
            m = (pMask[j + 3] == 0) - 1;
            v = pSrc[3 * j + 9] & m;
            v = _CV_ABS( v );
            if( s < v )
                s = v;
        }
        for( ; j < roiSize.width; j++ )
        {
            int m = (pMask[j] == 0) - 1;
            int v = pSrc[3 * j] & m;

            v = _CV_ABS( v );
            if( s < v )
                s = v;
        }
    }
    *norm = s;

    return CV_NO_ERR;
}

/* //////////////////////////////////////////////////////////////////////////////////// */
/* //////////////////////////////////////////////////////////////////////////////////// */
/* //////////////////////////////////////////////////////////////////////////////////// */
/* ///////////////////////////////// 32fC1 flavor ///////////////////////////////////// */

IPCVAPI_IMPL( CvStatus, icvNorm_Inf_32f_C1MR, (const float *pSrc,
                                              int srcStep,
                                              const uchar * pMask,
                                              int maskStep, CvSize roiSize, double *norm) )
{
    int i, j, a0 = 0x7fffffff;
    float f = 0.f;

    srcStep /= 4;

/* parameters check */
    if( !pSrc )
        return CV_NULLPTR_ERR;
    if( !pMask )
        return CV_NULLPTR_ERR;
    if( srcStep < 1 )
        return CV_BADSTEP_ERR;
    if( maskStep < 1 )
        return CV_BADSTEP_ERR;
    if( roiSize.width < 1 || roiSize.height < 1 || roiSize.width > srcStep ||
        roiSize.width > maskStep )
        return CV_BADSIZE_ERR;

    if( roiSize.width == srcStep && roiSize.width == maskStep )
    {
        roiSize.width *= roiSize.height;
        roiSize.height = 1;
    }

    for( i = 0; i < roiSize.height; i++, pSrc += srcStep, pMask += maskStep )
    {
        for( j = 0; j < roiSize.width - 3; j += 4 )
        {
            int m = (pMask[j] == 0) - 1;
            int v = ((int *) pSrc)[j] & a0;
            int g = v & m;

            m = (pMask[j + 1] == 0) - 1;
            v = ((int *) pSrc)[j + 1] & a0;
            v = v & m;
            if( g < v )
                g = v;
            m = (pMask[j + 2] == 0) - 1;
            v = ((int *) pSrc)[j + 2] & a0;
            v = v & m;
            if( g < v )
                g = v;
            m = (pMask[j + 3] == 0) - 1;
            v = ((int *) pSrc)[j + 3] & a0;
            v = v & m;
            if( g < v )
                g = v;
            if( f < *(float *) &g )
                f = *(float *) &g;
        }
        for( ; j < roiSize.width; j++ )
        {
            int m = (pMask[j] == 0) - 1;
            int v = (((int *) pSrc)[j] & a0) & m;

            if( f < *(float *) &v )
                f = *(float *) &v;
        }
    }
    *norm = (double) f;

    return CV_NO_ERR;
}

/* ///////////////////////////////// 32fC3 flavor ///////////////////////////////////// */

IPCVAPI_IMPL( CvStatus, icvNorm_Inf_32f_C3CMR, (const float *pSrc,
                                               int srcStep,
                                               const uchar * pMask,
                                               int maskStep,
                                               CvSize roiSize, int coi, double *norm) )
{
    int i, j, a0 = 0x7fffffff;
    float f = 0.f;

    srcStep /= 4;

/* parameters check */
    if( !pSrc )
        return CV_NULLPTR_ERR;
    if( !pMask )
        return CV_NULLPTR_ERR;
    if( srcStep < 1 )
        return CV_BADSTEP_ERR;
    if( maskStep < 1 )
        return CV_BADSTEP_ERR;
    if( coi < 1 || coi > 3 )
        return CV_UNSUPPORTED_COI_ERR;
    if( roiSize.width < 1 || roiSize.height < 1 || roiSize.width > srcStep ||
        roiSize.width > maskStep )
        return CV_BADSIZE_ERR;

    if( roiSize.width == srcStep && roiSize.width == maskStep )
    {
        roiSize.width *= roiSize.height;
        roiSize.height = 1;
    }

    coi--;
    pSrc += coi;

    for( i = 0; i < roiSize.height; i++, pSrc += srcStep, pMask += maskStep )
    {
        for( j = 0; j < roiSize.width - 3; j += 4 )
        {
            int m = (pMask[j] == 0) - 1;
            int v = ((int *) pSrc)[3 * j] & a0;
            int g = v & m;

            m = (pMask[j + 1] == 0) - 1;
            v = ((int *) pSrc)[3 * j + 3] & a0;
            v = v & m;
            if( g < v )
                g = v;
            m = (pMask[j + 2] == 0) - 1;
            v = ((int *) pSrc)[3 * j + 6] & a0;
            v = v & m;
            if( g < v )
                g = v;
            m = (pMask[j + 3] == 0) - 1;
            v = ((int *) pSrc)[3 * j + 9] & a0;
            v = v & m;
            if( g < v )
                g = v;
            if( f < *(float *) &g )
                f = *(float *) &g;
        }
        for( ; j < roiSize.width; j++ )
        {
            int m = (pMask[j] == 0) - 1;
            int v = (((int *) pSrc)[3 * j] & a0) & m;

            if( f < *(float *) &v )
                f = *(float *) &v;
        }
    }
    *norm = (double) f;

    return CV_NO_ERR;
}

/* =====  END OF FILE  ===== */
