/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include "_cv.h"

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Names:   icvNormDiff_L2_8u_C1MR,  icvNormDiff_L2_8s_C1MR,  icvNormDiff_L2_32f_C1MR, 
//             icvNormDiff_L2_8u_C3CMR, icvNormDiff_L2_8s_C3CMR, icvNormDiff_L2_32f_C3CMR, 
//
//    Purpose: Calculate 2nd power difference norm of two images with mask
//
//    Parameters:
//         pSrc1,    pSrc2,    pMask    - pointers to source images and mask array,
//         src1Step, src2Step, maskStep - their steps,
//         roiSize                      - their size or ROI size,
//         coi                          - COI index (1..3) (if 3-channel images).
//
//    Returns:
//         Cv status
//
//    Notes:
//F*/

/* ///////////////////////////////// 8uC1 flavor ////////////////////////////////////// */
IPCVAPI_IMPL( CvStatus, icvNormDiff_L2_8u_C1MR, (const uchar * pSrc1,
                                                int src1Step,
                                                const uchar * pSrc2,
                                                int src2Step,
                                                const uchar * pMask,
                                                int maskStep, CvSize roiSize, double *norm) )
{
    int i, j;
    int64 norm1 = 0;

/* parameters check */
    if( !pSrc1 || !pSrc2 )
        return CV_NULLPTR_ERR;
    if( !pMask )
        return CV_NULLPTR_ERR;
    if( src1Step < 1 || src2Step < 1 )
        return CV_BADSTEP_ERR;
    if( maskStep < 1 )
        return CV_BADSTEP_ERR;
    if( roiSize.width < 1 || roiSize.height < 1 || roiSize.width > src1Step ||
        roiSize.width > src2Step || roiSize.width > maskStep )
        return CV_BADSIZE_ERR;

    if( roiSize.width == src1Step && roiSize.width == maskStep )
    {
        roiSize.width *= roiSize.height;
        roiSize.height = 1;
    }

    for( i = 0; i < roiSize.height; i++, pSrc1 += src1Step, pSrc2 += src2Step,
         pMask += maskStep )
    {
        for( j = 0; j < roiSize.width - 3; j += 4 )
        {
            int m = (pMask[j] == 0) - 1;
            int v = (pSrc1[j] - pSrc2[j]) & m;
            int e = v * v;

            m = (pMask[j + 1] == 0) - 1;
            v = (pSrc1[j + 1] - pSrc2[j + 1]) & m;
            e += v * v;
            m = (pMask[j + 2] == 0) - 1;
            v = (pSrc1[j + 2] - pSrc2[j + 2]) & m;
            e += v * v;
            m = (pMask[j + 3] == 0) - 1;
            v = (pSrc1[j + 3] - pSrc2[j + 3]) & m;
            e += v * v;
            norm1 += e;
        }
        for( ; j < roiSize.width; j++ )
        {
            int m = (pMask[j] == 0) - 1;
            int v = (pSrc1[j] - pSrc2[j]) & m;

            norm1 += v * v;
        }
    }

    *norm = sqrt( (double) norm1 );

    return CV_NO_ERR;
}

/* ///////////////////////////////// 8uC3 flavor ////////////////////////////////////// */
IPCVAPI_IMPL( CvStatus, icvNormDiff_L2_8u_C3CMR, (const uchar * pSrc1,
                                                 int src1Step,
                                                 const uchar * pSrc2,
                                                 int src2Step,
                                                 const uchar * pMask,
                                                 int maskStep,
                                                 CvSize roiSize, int coi, double *norm) )
{
    int i, j;
    int64 norm1 = 0;

/* parameters check */
    if( !pSrc1 || !pSrc2 )
        return CV_NULLPTR_ERR;
    if( !pMask )
        return CV_NULLPTR_ERR;
    if( src1Step < 1 || src2Step < 1 )
        return CV_BADSTEP_ERR;
    if( maskStep < 1 )
        return CV_BADSTEP_ERR;
    if( coi < 1 || coi > 3 )
        return CV_UNSUPPORTED_COI_ERR;
    if( roiSize.width < 1 || roiSize.height < 1 || roiSize.width > src1Step ||
        roiSize.width > src2Step || roiSize.width > maskStep )
        return CV_BADSIZE_ERR;

    if( roiSize.width == src1Step && roiSize.width == maskStep )
    {
        roiSize.width *= roiSize.height;
        roiSize.height = 1;
    }

    coi--;
    pSrc1 += coi;
    pSrc2 += coi;

    for( i = 0; i < roiSize.height; i++, pSrc1 += src1Step, pSrc2 += src2Step,
         pMask += maskStep )
    {
        for( j = 0; j < roiSize.width - 3; j += 4 )
        {
            int m = (pMask[j] == 0) - 1;
            int v = (pSrc1[3 * j] - pSrc2[3 * j]) & m;
            int e = v * v;

            m = (pMask[j + 1] == 0) - 1;
            v = (pSrc1[3 * j + 3] - pSrc2[3 * j + 3]) & m;
            e += v * v;
            m = (pMask[j + 2] == 0) - 1;
            v = (pSrc1[3 * j + 6] - pSrc2[3 * j + 6]) & m;
            e += v * v;
            m = (pMask[j + 3] == 0) - 1;
            v = (pSrc1[3 * j + 9] - pSrc2[3 * j + 9]) & m;
            e += v * v;
            norm1 += e;
        }
        for( ; j < roiSize.width; j++ )
        {
            int m = (pMask[j] == 0) - 1;
            int v = (pSrc1[3 * j] - pSrc2[3 * j]) & m;

            norm1 += v * v;
        }
    }

    *norm = sqrt( (double) norm1 );

    return CV_NO_ERR;
}

/* //////////////////////////////////////////////////////////////////////////////////// */
/* //////////////////////////////////////////////////////////////////////////////////// */
/* //////////////////////////////////////////////////////////////////////////////////// */
/* ////////////////////////////////// 8sC1 flavor ///////////////////////////////////// */

IPCVAPI_IMPL( CvStatus, icvNormDiff_L2_8s_C1MR, (const char *pSrc1,
                                                int src1Step,
                                                const char *pSrc2,
                                                int src2Step,
                                                const uchar * pMask,
                                                int maskStep, CvSize roiSize, double *norm) )
{
    int i, j;
    int64 norm1 = 0;

/* parameters check */
    if( !pSrc1 || !pSrc2 )
        return CV_NULLPTR_ERR;
    if( !pMask )
        return CV_NULLPTR_ERR;
    if( src1Step < 1 || src2Step < 1 )
        return CV_BADSTEP_ERR;
    if( maskStep < 1 )
        return CV_BADSTEP_ERR;
    if( roiSize.width < 1 || roiSize.height < 1 || roiSize.width > src1Step ||
        roiSize.width > src2Step || roiSize.width > maskStep )
        return CV_BADSIZE_ERR;

    if( roiSize.width == src1Step && roiSize.width == maskStep )
    {
        roiSize.width *= roiSize.height;
        roiSize.height = 1;
    }

    for( i = 0; i < roiSize.height; i++, pSrc1 += src1Step, pSrc2 += src2Step,
         pMask += maskStep )
    {
        for( j = 0; j < roiSize.width - 3; j += 4 )
        {
            int m = (pMask[j] == 0) - 1;
            int v = (pSrc1[j] - pSrc2[j]) & m;
            int e = v * v;

            m = (pMask[j + 1] == 0) - 1;
            v = (pSrc1[j + 1] - pSrc2[j + 1]) & m;
            e += v * v;
            m = (pMask[j + 2] == 0) - 1;
            v = (pSrc1[j + 2] - pSrc2[j + 2]) & m;
            e += v * v;
            m = (pMask[j + 3] == 0) - 1;
            v = (pSrc1[j + 3] - pSrc2[j + 3]) & m;
            e += v * v;
            norm1 += e;
        }
        for( ; j < roiSize.width; j++ )
        {
            int m = (pMask[j] == 0) - 1;
            int v = (pSrc1[j] - pSrc2[j]) & m;

            norm1 += v * v;
        }
    }

    *norm = sqrt( (double) norm1 );

    return CV_NO_ERR;
}


/* ///////////////////////////////// 8sC3 flavor ////////////////////////////////////// */
IPCVAPI_IMPL( CvStatus, icvNormDiff_L2_8s_C3CMR, (const char *pSrc1,
                                                 int src1Step,
                                                 const char *pSrc2,
                                                 int src2Step,
                                                 const uchar * pMask,
                                                 int maskStep,
                                                 CvSize roiSize, int coi, double *norm) )
{
    int i, j;
    int64 norm1 = 0;

/* parameters check */
    if( !pSrc1 || !pSrc2 )
        return CV_NULLPTR_ERR;
    if( !pMask )
        return CV_NULLPTR_ERR;
    if( src1Step < 1 || src2Step < 1 )
        return CV_BADSTEP_ERR;
    if( maskStep < 1 )
        return CV_BADSTEP_ERR;
    if( coi < 1 || coi > 3 )
        return CV_UNSUPPORTED_COI_ERR;
    if( roiSize.width < 1 || roiSize.height < 1 || roiSize.width > src1Step ||
        roiSize.width > src2Step || roiSize.width > maskStep )
        return CV_BADSIZE_ERR;

    if( roiSize.width == src1Step && roiSize.width == maskStep )
    {
        roiSize.width *= roiSize.height;
        roiSize.height = 1;
    }

    coi--;
    pSrc1 += coi;
    pSrc2 += coi;

    for( i = 0; i < roiSize.height; i++, pSrc1 += src1Step, pSrc2 += src2Step,
         pMask += maskStep )
    {
        for( j = 0; j < roiSize.width - 3; j += 4 )
        {
            int m = (pMask[j] == 0) - 1;
            int v = (pSrc1[3 * j] - pSrc2[3 * j]) & m;
            int e = v * v;

            m = (pMask[j + 1] == 0) - 1;
            v = (pSrc1[3 * j + 3] - pSrc2[3 * j + 3]) & m;
            e += v * v;
            m = (pMask[j + 2] == 0) - 1;
            v = (pSrc1[3 * j + 6] - pSrc2[3 * j + 6]) & m;
            e += v * v;
            m = (pMask[j + 3] == 0) - 1;
            v = (pSrc1[3 * j + 9] - pSrc2[3 * j + 9]) & m;
            e += v * v;
            norm1 += e;
        }
        for( ; j < roiSize.width; j++ )
        {
            int m = (pMask[j] == 0) - 1;
            int v = (pSrc1[3 * j] - pSrc2[3 * j]) & m;

            norm1 += v * v;
        }
    }

    *norm = sqrt( (double) norm1 );

    return CV_NO_ERR;
}

/* //////////////////////////////////////////////////////////////////////////////////// */
/* //////////////////////////////////////////////////////////////////////////////////// */
/* //////////////////////////////////////////////////////////////////////////////////// */
/* ///////////////////////////////// 32fC1 flavor ///////////////////////////////////// */

IPCVAPI_IMPL( CvStatus, icvNormDiff_L2_32f_C1MR, (const float *pSrc1,
                                                 int src1Step,
                                                 const float *pSrc2,
                                                 int src2Step,
                                                 const uchar * pMask,
                                                 int maskStep, CvSize roiSize, double *norm) )
{
    int i, j;

    src1Step /= 4;
    src2Step /= 4;

/* parameters check */
    if( !pSrc1 || !pSrc2 )
        return CV_NULLPTR_ERR;
    if( !pMask )
        return CV_NULLPTR_ERR;
    if( src1Step < 1 || src2Step < 1 )
        return CV_BADSTEP_ERR;
    if( maskStep < 1 )
        return CV_BADSTEP_ERR;
    if( roiSize.width < 1 || roiSize.height < 1 || roiSize.width > src1Step ||
        roiSize.width > src2Step || roiSize.width > maskStep )
        return CV_BADSIZE_ERR;

    if( roiSize.width == src1Step && roiSize.width == maskStep )
    {
        roiSize.width *= roiSize.height;
        roiSize.height = 1;
    }

    *norm = 0;

    for( i = 0; i < roiSize.height; i++, pSrc1 += src1Step, pSrc2 += src2Step,
         pMask += maskStep )
    {
        for( j = 0; j < roiSize.width - 3; j += 4 )
        {
            double v = pSrc1[j] - pSrc2[j];
            int64 m = (pMask[j] == 0) - 1;
            int64 d = (*(int64 *) & v) & m;
            double g = *(double *) &d;
            double e = g * g;

            v = pSrc1[j + 1] - pSrc2[j + 1];
            m = (pMask[j + 1] == 0) - 1;
            d = (*(int64 *) & v) & m;
            g = *(double *) &d;
            e += g * g;
            v = pSrc1[j + 2] - pSrc2[j + 2];
            m = (pMask[j + 2] == 0) - 1;
            d = (*(int64 *) & v) & m;
            g = *(double *) &d;
            e += g * g;
            v = pSrc1[j + 3] - pSrc2[j + 3];
            m = (pMask[j + 3] == 0) - 1;
            d = (*(int64 *) & v) & m;
            g = *(double *) &d;
            e += g * g;
            *norm += e;
        }
        for( ; j < roiSize.width; j++ )
        {
            double v = pSrc1[j] - pSrc2[j];
            int64 m = (pMask[j] == 0) - 1;
            int64 d = (*(int64 *) & v) & m;
            double g = *(double *) &d;

            *norm += g * g;
        }
    }

    *norm = sqrt( *norm );

    return CV_NO_ERR;
}

/* ///////////////////////////////// 32fC3 flavor ///////////////////////////////////// */

IPCVAPI_IMPL( CvStatus, icvNormDiff_L2_32f_C3CMR, (const float *pSrc1,
                                                  int src1Step,
                                                  const float *pSrc2,
                                                  int src2Step,
                                                  const uchar * pMask,
                                                  int maskStep,
                                                  CvSize roiSize, int coi, double *norm) )
{
    int i, j;

    src1Step /= 4;
    src2Step /= 4;

/* parameters check */
    if( !pSrc1 || !pSrc2 )
        return CV_NULLPTR_ERR;
    if( !pMask )
        return CV_NULLPTR_ERR;
    if( src1Step < 1 || src2Step < 1 )
        return CV_BADSTEP_ERR;
    if( maskStep < 1 )
        return CV_BADSTEP_ERR;
    if( coi < 1 || coi > 3 )
        return CV_UNSUPPORTED_COI_ERR;
    if( roiSize.width < 1 || roiSize.height < 1 || roiSize.width > src1Step ||
        roiSize.width > src2Step || roiSize.width > maskStep )
        return CV_BADSIZE_ERR;

    if( roiSize.width == src1Step && roiSize.width == maskStep )
    {
        roiSize.width *= roiSize.height;
        roiSize.height = 1;
    }

    coi--;
    pSrc1 += coi;
    pSrc2 += coi;
    *norm = 0;

    for( i = 0; i < roiSize.height; i++, pSrc1 += src1Step, pSrc2 += src2Step,
         pMask += maskStep )
    {
        for( j = 0; j < roiSize.width - 3; j += 4 )
        {
            double v = pSrc1[3 * j] - pSrc2[3 * j];
            int64 m = (pMask[j] == 0) - 1;
            int64 d = (*(int64 *) & v) & m;
            double g = *(double *) &d;
            double e = g * g;

            v = pSrc1[3 * j + 3] - pSrc2[3 * j + 3];
            m = (pMask[j + 1] == 0) - 1;
            d = (*(int64 *) & v) & m;
            g = *(double *) &d;
            e += g * g;
            v = pSrc1[3 * j + 6] - pSrc2[3 * j + 6];
            m = (pMask[j + 2] == 0) - 1;
            d = (*(int64 *) & v) & m;
            g = *(double *) &d;
            e += g * g;
            v = pSrc1[3 * j + 9] - pSrc2[3 * j + 9];
            m = (pMask[j + 3] == 0) - 1;
            d = (*(int64 *) & v) & m;
            g = *(double *) &d;
            e += g * g;
            *norm += e;
        }
        for( ; j < roiSize.width; j++ )
        {
            double v = pSrc1[3 * j] - pSrc2[3 * j];
            int64 m = (pMask[j] == 0) - 1;
            int64 d = (*(int64 *) & v) & m;
            double g = *(double *) &d;

            *norm += g * g;
        }
    }

    *norm = sqrt( *norm );

    return CV_NO_ERR;
}

/* =====  END OF FILE  ===== */
