/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include <assert.h>
#include "_cv.h"
#include "_cvimgstat.h"

IPCVAPI_IMPL( CvStatus, icvAbsDiff_8u_C1R, (uchar * src1, int src1_step,
                                            uchar * src2, int src2_step,
                                            uchar * dst, int dst_step, CvSize size) )
{
    int x, y;

    if( !src1 || !src2 || !dst )
        return CV_NULLPTR_ERR;
    if( size.width <= 0 || size.height <= 0 ||
        src1_step < size.width || src2_step < size.width || dst_step < size.width )
        return CV_BADSIZE_ERR;

    if( size.width == src1_step && size.width == src2_step && size.width == dst_step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( y = 0; y < size.height; y++, src1 += src1_step, src2 += src1_step, dst += dst_step )
    {
        for( x = 0; x <= size.width - 4; x += 4 )
        {
            int t0 = src1[x] - src2[x];
            int t1 = src1[x + 1] - src2[x + 1];
            int t2 = src1[x + 2] - src2[x + 2];
            int t3 = src1[x + 3] - src2[x + 3];

            t0 = _CV_ABS( t0 );
            t1 = _CV_ABS( t1 );
            t2 = _CV_ABS( t2 );
            t3 = _CV_ABS( t3 );

            dst[x] = (uchar) t0;
            dst[x + 1] = (uchar) t1;
            dst[x + 2] = (uchar) t2;
            dst[x + 3] = (uchar) t3;
        }

        for( ; x < size.width; x++ )
        {
            int t0 = src1[x] - src2[x];

            t0 = _CV_ABS( t0 );
            dst[x] = (uchar) t0;
        }
    }

    return CV_OK;
}


IPCVAPI_IMPL( CvStatus, icvAbsDiff_32f_C1R, (float *src1, int src1_step,
                                             float *src2, int src2_step,
                                             float *dst, int dst_step, CvSize size) )
{
    int x, y;

    if( !src1 || !src2 || !dst )
        return CV_NULLPTR_ERR;
    if( size.width <= 0 || size.height <= 0 ||
        src1_step < size.width * 4 || src2_step < size.width * 4 ||
        dst_step < size.width * 4 || ((src1_step | src2_step | dst_step) & 3) != 0 )
        return CV_BADSIZE_ERR;

    src1_step /= 4;
    src2_step /= 4;
    dst_step /= 4;

    if( size.width == src1_step && size.width == src2_step && size.width == dst_step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( y = 0; y < size.height; y++, src1 += src1_step, src2 += src2_step, dst += dst_step )
    {
        for( x = 0; x <= size.width - 4; x += 4 )
        {
            double t0 = src1[x] - src2[x];
            double t1 = src1[x + 1] - src2[x + 1];

            t0 = fabs( t0 );
            t1 = fabs( t1 );

            dst[x] = (float) t0;
            dst[x + 1] = (float) t1;

            t0 = src1[x + 2] - src2[x + 2];
            t1 = src1[x + 3] - src2[x + 3];

            t0 = fabs( t0 );
            t1 = fabs( t1 );

            dst[x + 2] = (float) t0;
            dst[x + 3] = (float) t1;
        }

        for( ; x < size.width; x++ )
        {
            double t0 = src1[x] - src2[x];

            t0 = fabs( t0 );
            dst[x] = (float) t0;
        }
    }

    return CV_OK;
}

/*************************** icvAbsDiff *****************************/

IPCVAPI_IMPL( CvStatus, icvAbsDiffC_8u_C1R, (uchar * src, int src_step,
                                             uchar * dst, int dst_step,
                                             CvSize size, int value) )
{
    int x, y;
    int ival = value;

    if( !src || !dst )
        return CV_NULLPTR_ERR;

    if( size.width <= 0 || size.height <= 0 || src_step < size.width || dst_step < size.width )
        return CV_BADSIZE_ERR;

    if( size.width == src_step && size.width == dst_step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    if( ival < 0 )
        ival = 0;
    if( ival > 255 )
        ival = 255;

    for( y = 0; y < size.height; y++, src += src_step, dst += dst_step )
    {
        for( x = 0; x <= size.width - 4; x += 4 )
        {
            int t0 = src[x] - ival;
            int t1 = src[x + 1] - ival;
            int t2 = src[x + 2] - ival;
            int t3 = src[x + 3] - ival;

            t0 = _CV_ABS( t0 );
            t1 = _CV_ABS( t1 );
            t2 = _CV_ABS( t2 );
            t3 = _CV_ABS( t3 );

            dst[x] = (uchar) t0;
            dst[x + 1] = (uchar) t1;
            dst[x + 2] = (uchar) t2;
            dst[x + 3] = (uchar) t3;
        }

        for( ; x < size.width; x++ )
        {
            int t0 = src[x] - ival;

            t0 = _CV_ABS( t0 );
            dst[x] = (uchar) t0;
        }
    }

    return CV_OK;
}

IPCVAPI_IMPL( CvStatus, icvAbsDiffC_32f_C1R, (float *src, int src_step,
                                              float *dst, int dst_step,
                                              CvSize size, float value) )
{
    int x, y;

    if( !src || !dst )
        return CV_NULLPTR_ERR;

    if( size.width <= 0 || size.height <= 0 ||
        src_step < size.width * 4 || dst_step < size.width * 4 ||
        ((src_step | dst_step) & 3) != 0 )
        return CV_BADSIZE_ERR;

    src_step /= 4;
    dst_step /= 4;

    if( size.width == src_step && size.width == dst_step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( y = 0; y < size.height; y++, src += src_step, dst += dst_step )
    {
        for( x = 0; x <= size.width - 4; x += 4 )
        {
            double t0 = src[x] - value;
            double t1 = src[x + 1] - value;

            t0 = fabs( t0 );
            t1 = fabs( t1 );

            dst[x] = (float) t0;
            dst[x + 1] = (float) t1;

            t0 = src[x + 2] - value;
            t1 = src[x + 3] - value;

            t0 = fabs( t0 );
            t1 = fabs( t1 );

            dst[x + 2] = (float) t0;
            dst[x + 3] = (float) t1;
        }

        for( ; x < size.width; x++ )
        {
            double t0 = src[x] - value;

            t0 = fabs( t0 );
            dst[x] = (float) t0;
        }
    }

    return CV_OK;
}


/* 
   Finds L1 norm between two blocks.
*/
int
icvCmpBlocksL1_8u_C1( const uchar * vec1, const uchar * vec2, int len )
{
    int i, sum = 0;

    for( i = 0; i <= len - 4; i += 4 )
    {
        int t0 = vec1[i] - vec2[i];
        int t1 = vec1[i + 1] - vec2[i + 1];
        int t2 = vec1[i + 2] - vec2[i + 2];
        int t3 = vec1[i + 3] - vec2[i + 3];

        sum += _CV_ABS( t0 ) + _CV_ABS( t1 ) + _CV_ABS( t2 ) + _CV_ABS( t3 );
    }

    for( ; i < len; i++ )
    {
        int t0 = vec1[i] - vec2[i];

        sum += _CV_ABS( t0 );
    }

    return sum;
}


int64
icvCmpBlocksL2_8u_C1( const uchar * vec1, const uchar * vec2, int len )
{
    int i, s = 0;
    int64 sum = 0;

    for( i = 0; i <= len - 4; i += 4 )
    {
        int v = vec1[i] - vec2[i];
        int e = v * v;

        v = vec1[i + 1] - vec2[i + 1];
        e += v * v;
        v = vec1[i + 2] - vec2[i + 2];
        e += v * v;
        v = vec1[i + 3] - vec2[i + 3];
        e += v * v;
        sum += e;
    }

    for( ; i < len; i++ )
    {
        int v = vec1[i] - vec2[i];

        s += v * v;
    }

    return sum + s;
}


int64
icvCmpBlocksL2_8s_C1( const char *vec1, const char *vec2, int len )
{
    int i, s = 0;
    int64 sum = 0;

    for( i = 0; i <= len - 4; i += 4 )
    {
        int v = vec1[i] - vec2[i];
        int e = v * v;

        v = vec1[i + 1] - vec2[i + 1];
        e += v * v;
        v = vec1[i + 2] - vec2[i + 2];
        e += v * v;
        v = vec1[i + 3] - vec2[i + 3];
        e += v * v;
        sum += e;
    }

    for( ; i < len; i++ )
    {
        int v = vec1[i] - vec2[i];

        s += v * v;
    }

    return sum + s;
}


double
icvCmpBlocksL2_32f_C1( const float *vec1, const float *vec2, int len )
{
    double sum = 0;
    int i;

    for( i = 0; i <= len - 4; i += 4 )
    {
        double v0 = vec1[i] - vec2[i];
        double v1 = vec1[i + 1] - vec2[i + 1];
        double v2 = vec1[i + 2] - vec2[i + 2];
        double v3 = vec1[i + 3] - vec2[i + 3];

        sum += v0 * v0 + v1 * v1 + v2 * v2 + v3 * v3;
    }
    for( ; i < len; i++ )
    {
        double v = vec1[i] - vec2[i];

        sum += v * v;
    }
    return sum;
}


/* 
   Calculates cross correlation for two blocks.
*/
int64
icvCrossCorr_8u_C1( const uchar * vec1, const uchar * vec2, int len )
{
    int i, s = 0;
    int64 sum = 0;

    for( i = 0; i <= len - 4; i += 4 )
    {
        int e = vec1[i] * vec2[i];
        int v = vec1[i + 1] * vec2[i + 1];

        e += v;
        v = vec1[i + 2] * vec2[i + 2];
        e += v;
        v = vec1[i + 3] * vec2[i + 3];
        e += v;
        sum += e;
    }

    for( ; i < len; i++ )
    {
        s += vec1[i] * vec2[i];
    }

    return sum + s;
}


int64
icvCrossCorr_8s_C1( const char *vec1, const char *vec2, int len )
{
    int i, s = 0;
    int64 sum = 0;

    for( i = 0; i <= len - 4; i += 4 )
    {
        int e = vec1[i] * vec2[i];
        int v = vec1[i + 1] * vec2[i + 1];

        e += v;
        v = vec1[i + 2] * vec2[i + 2];
        e += v;
        v = vec1[i + 3] * vec2[i + 3];
        e += v;
        sum += e;
    }

    for( ; i < len; i++ )
    {
        s += vec1[i] * vec2[i];
    }

    return sum + s;
}


double
icvCrossCorr_32f_C1( const float *vec1, const float *vec2, int len )
{
    double sum = 0;
    int i;

    for( i = 0; i <= len - 4; i += 4 )
    {
        double v0 = vec1[i] * vec2[i];
        double v1 = vec1[i + 1] * vec2[i + 1];
        double v2 = vec1[i + 2] * vec2[i + 2];
        double v3 = vec1[i + 3] * vec2[i + 3];

        sum += v0 + v1 + v2 + v3;
    }
    for( ; i < len; i++ )
    {
        double v = vec1[i] * vec2[i];

        sum += v;
    }
    return sum;
}


/* 
   Calculates cross correlation for two blocks.
*/
int64
icvSumPixels_8u_C1( const uchar * vec, int len )
{
    int i, s = 0;
    int64 sum = 0;

    for( i = 0; i <= len - 4; i += 4 )
    {
        sum += vec[i] + vec[i + 1] + vec[i + 2] + vec[i + 3];
    }

    for( ; i < len; i++ )
    {
        s += vec[i];
    }

    return sum + s;
}


int64
icvSumPixels_8s_C1( const char *vec, int len )
{
    int i, s = 0;
    int64 sum = 0;

    for( i = 0; i <= len - 4; i += 4 )
    {
        sum += vec[i] + vec[i + 1] + vec[i + 2] + vec[i + 3];
    }

    for( ; i < len; i++ )
    {
        s += vec[i];
    }

    return sum + s;
}


double
icvSumPixels_32f_C1( const float *vec, int len )
{
    double sum = 0;
    int i;

    for( i = 0; i <= len - 4; i += 4 )
    {
        sum += vec[i] + vec[i + 1] + vec[i + 2] + vec[i + 3];
    }

    for( ; i < len; i++ )
    {
        sum += vec[i];
    }
    return sum;
}


/* End of file. */
