/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include "_cv.h"

/*F-------------------------------------------------------------------------------------*/
/*    Name: CVAcc8u32f, CVAcc8s32f, CVAcc32f32f                                      */
/*    Purpose: Accumulating the mean: pDst(x,y)=pSrc(x,y)+pDst(x,y)                     */
/*    Context:                                                                          */
/*    Parameters:                                                                       */
/*              pSrc - source image                                                     */
/*              pDst - mean statistic destination image                                 */
/*              len - length of pixel array                                             */
/*    Returns:                                                                          */
/*              CV_NO_ERR if all ok or error code                                          */
/*    Notes:                                                                            */
/*F-------------------------------------------------------------------------------------*/
IPCVAPI_IMPL( CvStatus, icvAdd_8u32f_C1IR, (const unsigned char *pSrc, int srcStep,
                                            float *pSrcDst, int srcDstStep, CvSize roiSize) )
{
    int x, y;

    srcDstStep /= sizeof( float );

    for( y = 0; y < roiSize.height; y++, pSrc += srcStep, pSrcDst += srcDstStep )
    {
        for( x = 0; x <= roiSize.width - 4; x += 4 )
        {
            float t0 = pSrcDst[x] + pSrc[x];
            float t1 = pSrcDst[x + 1] + pSrc[x + 1];

            pSrcDst[x] = t0;
            pSrcDst[x + 1] = t1;

            t0 = pSrcDst[x + 2] + pSrc[x + 2];
            t1 = pSrcDst[x + 3] + pSrc[x + 3];

            pSrcDst[x + 2] = t0;
            pSrcDst[x + 3] = t1;
        }

        for( ; x < roiSize.width; x++ )
        {
            float t0 = pSrcDst[x] + pSrc[x];

            pSrcDst[x] = t0;
        }
    }

    return CV_NO_ERR;
}


IPCVAPI_IMPL( CvStatus, icvAdd_8s32f_C1IR, (const char *pSrc, int srcStep,
                                            float *pSrcDst, int srcDstStep, CvSize roiSize) )
{
    int x, y;

    srcDstStep /= sizeof( float );

    for( y = 0; y < roiSize.height; y++, pSrc += srcStep, pSrcDst += srcDstStep )
    {
        for( x = 0; x <= roiSize.width - 4; x += 4 )
        {
            float t0 = pSrcDst[x] + pSrc[x];
            float t1 = pSrcDst[x + 1] + pSrc[x + 1];

            pSrcDst[x] = t0;
            pSrcDst[x + 1] = t1;

            t0 = pSrcDst[x + 2] + pSrc[x + 2];
            t1 = pSrcDst[x + 3] + pSrc[x + 3];

            pSrcDst[x + 2] = t0;
            pSrcDst[x + 3] = t1;
        }

        for( ; x < roiSize.width; x++ )
        {
            float t0 = pSrcDst[x] + pSrc[x];

            pSrcDst[x] = t0;
        }
    }

    return CV_NO_ERR;
}


IPCVAPI_IMPL( CvStatus, icvAdd_32f_C1IR, (const float *pSrc, int srcStep,
                                          float *pSrcDst, int srcDstStep, CvSize roiSize) )
{
    int x, y;

    srcDstStep /= sizeof( float );
    srcStep /= sizeof( float );

    for( y = 0; y < roiSize.height; y++, pSrc += srcStep, pSrcDst += srcDstStep )
    {
        for( x = 0; x <= roiSize.width - 4; x += 4 )
        {
            float t0 = pSrcDst[x] + pSrc[x];
            float t1 = pSrcDst[x + 1] + pSrc[x + 1];

            pSrcDst[x] = t0;
            pSrcDst[x + 1] = t1;

            t0 = pSrcDst[x + 2] + pSrc[x + 2];
            t1 = pSrcDst[x + 3] + pSrc[x + 3];

            pSrcDst[x + 2] = t0;
            pSrcDst[x + 3] = t1;
        }

        for( ; x < roiSize.width; x++ )
        {
            float t0 = pSrcDst[x] + pSrc[x];

            pSrcDst[x] = t0;
        }
    }

    return CV_NO_ERR;
}

/*F-------------------------------------------------------------------------------------*/
/*    Name: CVSquareAcc8u32f, CVSquareAcc8s32f, CVSquareAcc32f32f                       */
/*    Purpose: Accumulating the mean square: pDst(x,y)=pSrc(x,y)*pSrc(x,y)+pDst(x,y)    */
/*    Context:                                                                          */
/*    Parameters:                                                                       */
/*              pSrc - source image                                                     */
/*              pDst - mean square statistic destination image                          */
/*              len - length of pixel array                                             */
/*    Returns:                                                                          */
/*              CV_NO_ERR if all ok or error code                                       */
/*    Notes:                                                                            */
/*F-------------------------------------------------------------------------------------*/
IPCVAPI_IMPL( CvStatus, icvAddSquare_8u32f_C1IR, (const unsigned char *pSrc, int srcStep,
                                                  float *pSrcDst, int srcDstStep,
                                                  CvSize roiSize) )
{
    int x, y;
    srcDstStep /= sizeof( float );

    for( y = 0; y < roiSize.height; y++, pSrc += srcStep, pSrcDst += srcDstStep )
    {
        for( x = 0; x <= roiSize.width - 4; x += 4 )
        {
            float t0 = pSrcDst[x] + pSrc[x] * pSrc[x];
            float t1 = pSrcDst[x + 1] + pSrc[x + 1] * pSrc[x + 1];

            pSrcDst[x] = t0;
            pSrcDst[x + 1] = t1;

            t0 = pSrcDst[x + 2] + pSrc[x + 2] * pSrc[x + 2];
            t1 = pSrcDst[x + 3] + pSrc[x + 3] * pSrc[x + 3];

            pSrcDst[x + 2] = t0;
            pSrcDst[x + 3] = t1;
        }

        for( ; x < roiSize.width; x++ )
        {
            float t0 = pSrcDst[x] + pSrc[x] * pSrc[x];

            pSrcDst[x] = t0;
        }
    }

    return CV_NO_ERR;
}

IPCVAPI_IMPL( CvStatus, icvAddSquare_8s32f_C1IR, (const char *pSrc, int srcStep,
                                                  float *pSrcDst, int srcDstStep,
                                                  CvSize roiSize) )
{
    int x, y;

    srcDstStep /= sizeof( float );

    for( y = 0; y < roiSize.height; y++, pSrc += srcStep, pSrcDst += srcDstStep )
    {
        for( x = 0; x <= roiSize.width - 4; x += 4 )
        {
            float t0 = pSrcDst[x] + pSrc[x] * pSrc[x];
            float t1 = pSrcDst[x + 1] + pSrc[x + 1] * pSrc[x + 1];

            pSrcDst[x] = t0;
            pSrcDst[x + 1] = t1;

            t0 = pSrcDst[x + 2] + pSrc[x + 2] * pSrc[x + 2];
            t1 = pSrcDst[x + 3] + pSrc[x + 3] * pSrc[x + 3];

            pSrcDst[x + 2] = t0;
            pSrcDst[x + 3] = t1;
        }

        for( ; x < roiSize.width; x++ )
        {
            float t0 = pSrcDst[x] + pSrc[x] * pSrc[x];

            pSrcDst[x] = t0;
        }
    }

    return CV_NO_ERR;
}

IPCVAPI_IMPL( CvStatus, icvAddSquare_32f_C1IR, (const float *pSrc, int srcStep,
                                                float *pSrcDst, int srcDstStep,
                                                CvSize roiSize) )
{
    int x, y;

    srcDstStep /= sizeof( float );
    srcStep /= sizeof( float );

    for( y = 0; y < roiSize.height; y++, pSrc += srcStep, pSrcDst += srcDstStep )
    {
        for( x = 0; x <= roiSize.width - 4; x += 4 )
        {
            float t0 = pSrcDst[x] + pSrc[x] * pSrc[x];
            float t1 = pSrcDst[x + 1] + pSrc[x + 1] * pSrc[x + 1];

            pSrcDst[x] = t0;
            pSrcDst[x + 1] = t1;

            t0 = pSrcDst[x + 2] + pSrc[x + 2] * pSrc[x + 2];
            t1 = pSrcDst[x + 3] + pSrc[x + 3] * pSrc[x + 3];

            pSrcDst[x + 2] = t0;
            pSrcDst[x + 3] = t1;
        }

        for( ; x < roiSize.width; x++ )
        {
            float t0 = pSrcDst[x] + pSrc[x] * pSrc[x];

            pSrcDst[x] = t0;
        }
    }

    return CV_NO_ERR;
}


/*F-------------------------------------------------------------------------------------*/
/*    Name: icvMultiplyAcc8u32f, icvMultiplyAcc8s32f, icvMultiplyAcc32f32f              */
/*    Purpose: Accumulating the images covariance:                                      */
/*                        pDst(x,y)=pSrcA(x,y)*pSrcA(x,y)+pDst(x,y)                     */
/*    Context:                                                                          */
/*    Parameters:                                                                       */
/*              pSrcA, pSrcB - source images                                            */
/*              pDst - covariance statistic destination image                           */
/*              len - length of pixel array                                             */
/*    Returns:                                                                          */
/*              CV_NO_ERR if all ok or error code                                       */
/*    Notes:                                                                            */
/*F-------------------------------------------------------------------------------------*/
IPCVAPI_IMPL( CvStatus, icvAddProduct_8u32f_C1IR, (const unsigned char *pSrc1,
                                                   int src1Step,
                                                   const unsigned char *pSrc2,
                                                   int src2Step,
                                                   float *pSrcDst, int srcDstStep,
                                                   CvSize roiSize) )
{
    int x, y;

    srcDstStep /= sizeof( float );

    for( y = 0; y < roiSize.height; y++, pSrc1 += src1Step, pSrc2 += src2Step,
         pSrcDst += srcDstStep )
    {
        for( x = 0; x <= roiSize.width - 4; x += 4 )
        {
            float t0 = pSrcDst[x] + pSrc1[x] * pSrc2[x];
            float t1 = pSrcDst[x + 1] + pSrc1[x + 1] * pSrc2[x + 1];

            pSrcDst[x] = t0;
            pSrcDst[x + 1] = t1;

            t0 = pSrcDst[x + 2] + pSrc1[x + 2] * pSrc2[x + 2];
            t1 = pSrcDst[x + 3] + pSrc1[x + 3] * pSrc2[x + 3];

            pSrcDst[x + 2] = t0;
            pSrcDst[x + 3] = t1;
        }

        for( ; x < roiSize.width; x++ )
        {
            float t0 = pSrcDst[x] + pSrc1[x] * pSrc2[x];

            pSrcDst[x] = t0;
        }
    }

    return CV_NO_ERR;
}


IPCVAPI_IMPL( CvStatus, icvAddProduct_8s32f_C1IR, (const char *pSrc1, int src1Step,
                                                   const char *pSrc2, int src2Step,
                                                   float *pSrcDst, int srcDstStep,
                                                   CvSize roiSize) )
{
    int x, y;

    srcDstStep /= sizeof( float );

    for( y = 0; y < roiSize.height; y++, pSrc1 += src1Step, pSrc2 += src2Step,
         pSrcDst += srcDstStep )
    {
        for( x = 0; x <= roiSize.width - 4; x += 4 )
        {
            float t0 = pSrcDst[x] + pSrc1[x] * pSrc2[x];
            float t1 = pSrcDst[x + 1] + pSrc1[x + 1] * pSrc2[x + 1];

            pSrcDst[x] = t0;
            pSrcDst[x + 1] = t1;

            t0 = pSrcDst[x + 2] + pSrc1[x + 2] * pSrc2[x + 2];
            t1 = pSrcDst[x + 3] + pSrc1[x + 3] * pSrc2[x + 3];

            pSrcDst[x + 2] = t0;
            pSrcDst[x + 3] = t1;
        }

        for( ; x < roiSize.width; x++ )
        {
            float t0 = pSrcDst[x] + pSrc1[x] * pSrc2[x];

            pSrcDst[x] = t0;
        }
    }

    return CV_NO_ERR;
}


IPCVAPI_IMPL( CvStatus, icvAddProduct_32f_C1IR, (const float *pSrc1, int src1Step,
                                                 const float *pSrc2, int src2Step,
                                                 float *pSrcDst, int srcDstStep,
                                                 CvSize roiSize) )
{
    int x, y;

    srcDstStep /= sizeof( float );
    src1Step /= sizeof( float );
    src2Step /= sizeof( float );

    for( y = 0; y < roiSize.height; y++, pSrc1 += src1Step, pSrc2 += src2Step,
         pSrcDst += srcDstStep )
    {
        for( x = 0; x <= roiSize.width - 4; x += 4 )
        {
            float t0 = pSrcDst[x] + pSrc1[x] * pSrc2[x];
            float t1 = pSrcDst[x + 1] + pSrc1[x + 1] * pSrc2[x + 1];

            pSrcDst[x] = t0;
            pSrcDst[x + 1] = t1;

            t0 = pSrcDst[x + 2] + pSrc1[x + 2] * pSrc2[x + 2];
            t1 = pSrcDst[x + 3] + pSrc1[x + 3] * pSrc2[x + 3];

            pSrcDst[x + 2] = t0;
            pSrcDst[x + 3] = t1;
        }

        for( ; x < roiSize.width; x++ )
        {
            float t0 = pSrcDst[x] + pSrc1[x] * pSrc2[x];

            pSrcDst[x] = t0;
        }
    }

    return CV_NO_ERR;
}


/*F-------------------------------------------------------------------------------------*/
/*    Name: icvMaskRunAvg8u32f, icvMaskRunAvg8s32f, icvMaskRunAvg32f32f                 */
/*    Purpose: Calculating linear combination of Dst and Src:                           */
/*                                      pDst(x,y)=alpha*pSrc(x,y)+(1-alpha)*pDst(x,y)   */
/*    Context:                                                                          */
/*    Parameters:                                                                       */
/*              pSrc - source image                                                     */
/*              pDst - destination image                                                */
/*              len - length of pixel array                                             */
/*              alpha - coefficient                                                     */
/*    Returns:                                                                          */
/*              CV_NO_ERR if all ok or error code                                       */
/*    Notes:                                                                            */
/*F-------------------------------------------------------------------------------------*/
IPCVAPI_IMPL( CvStatus, icvAddWeighted_8u32f_C1IR, (const unsigned char *pSrc,
                                                    int srcStep,
                                                    float *pSrcDst, int srcDstStep,
                                                    CvSize roiSize, float alpha) )
{
    int x, y;

    srcDstStep /= sizeof( float );

    /* pSrcDst[x]*(1-alpha) + pSrc[x]*alpha = pSrcDst[x] + alpha*(pSrc[x] - pSrcDst[x]) */
    for( y = 0; y < roiSize.height; y++, pSrc += srcStep, pSrcDst += srcDstStep )
    {
        for( x = 0; x <= roiSize.width - 4; x += 4 )
        {
            float t0 = pSrcDst[x] + alpha * (pSrc[x] - pSrcDst[x]);
            float t1 = pSrcDst[x + 1] + alpha * (pSrc[x + 1] - pSrcDst[x + 1]);

            pSrcDst[x] = t0;
            pSrcDst[x + 1] = t1;

            t0 = pSrcDst[x + 2] + alpha * (pSrc[x + 2] - pSrcDst[x + 2]);
            t1 = pSrcDst[x + 3] + alpha * (pSrc[x + 3] - pSrcDst[x + 3]);

            pSrcDst[x + 2] = t0;
            pSrcDst[x + 3] = t1;
        }

        for( ; x < roiSize.width; x++ )
        {
            float t0 = pSrcDst[x] + alpha * (pSrc[x] - pSrcDst[x]);

            pSrcDst[x] = t0;
        }
    }

    return CV_NO_ERR;
}



IPCVAPI_IMPL( CvStatus, icvAddWeighted_8s32f_C1IR, (const char *pSrc, int srcStep,
                                                    float *pSrcDst, int srcDstStep,
                                                    CvSize roiSize, float alpha) )
{
    int x, y;

    srcDstStep /= sizeof( float );

    for( y = 0; y < roiSize.height; y++, pSrc += srcStep, pSrcDst += srcDstStep )
    {
        for( x = 0; x <= roiSize.width - 4; x += 4 )
        {
            float t0 = pSrcDst[x] + alpha * (pSrc[x] - pSrcDst[x]);
            float t1 = pSrcDst[x + 1] + alpha * (pSrc[x + 1] - pSrcDst[x + 1]);

            pSrcDst[x] = t0;
            pSrcDst[x + 1] = t1;

            t0 = pSrcDst[x + 2] + alpha * (pSrc[x + 2] - pSrcDst[x + 2]);
            t1 = pSrcDst[x + 3] + alpha * (pSrc[x + 3] - pSrcDst[x + 3]);

            pSrcDst[x + 2] = t0;
            pSrcDst[x + 3] = t1;
        }

        for( ; x < roiSize.width; x++ )
        {
            float t0 = pSrcDst[x] + alpha * (pSrc[x] - pSrcDst[x]);

            pSrcDst[x] = t0;
        }
    }

    return CV_NO_ERR;
}


IPCVAPI_IMPL( CvStatus, icvAddWeighted_32f_C1IR, (const float *pSrc, int srcStep,
                                                  float *pSrcDst, int srcDstStep,
                                                  CvSize roiSize, float alpha) )
{
    int x, y;

    srcDstStep /= sizeof( float );
    srcStep /= sizeof( float );

    for( y = 0; y < roiSize.height; y++, pSrc += srcStep, pSrcDst += srcDstStep )
    {
        for( x = 0; x <= roiSize.width - 4; x += 4 )
        {
            float t0 = pSrcDst[x] + alpha * (pSrc[x] - pSrcDst[x]);
            float t1 = pSrcDst[x + 1] + alpha * (pSrc[x + 1] - pSrcDst[x + 1]);

            pSrcDst[x] = t0;
            pSrcDst[x + 1] = t1;

            t0 = pSrcDst[x + 2] + alpha * (pSrc[x + 2] - pSrcDst[x + 2]);
            t1 = pSrcDst[x + 3] + alpha * (pSrc[x + 3] - pSrcDst[x + 3]);

            pSrcDst[x + 2] = t0;
            pSrcDst[x + 3] = t1;
        }

        for( ; x < roiSize.width; x++ )
        {
            float t0 = pSrcDst[x] + alpha * (pSrc[x] - pSrcDst[x]);

            pSrcDst[x] = t0;
        }
    }

    return CV_NO_ERR;
}
