/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include "_cv.h"
#include "_cvwrap.h"
#include "_cvarray.h"


//double coef = 0.001;
double coef = 0.000;
/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    icvAdaptiveThreshold8uC1R
//    Purpose: Adaptive Thresholding the source array
//    Context:
//    Parameters:
//      Src     - source array
//      roi     - size of picture in elements
//      srcStep - lenght of string
//      type    - thresholding type, must be one of
//                  CV_THRESH_BINARY       - val = (val > Thresh ? MAX    : 0)
//                  CV_THRESH_BINARY_INV   - val = (val > Thresh ? 0      : MAX)
//                  CV_THRESH_TOZERO       - val = (val > Thresh ? val    : 0)
//                  CV_THRESH_TOZERO_INV   - val = (val > Thresh ? 0      : val)
//      size     - size of thresholding neighborhood  ( 1 - 3x3 , 2 - 5x5, 3 - 7x7, 4 - 9x9 )
//    Returns:
//    Notes:
//      The MAX constant for uchar is 255, for char is 127
//F*/
#define __OPTIMAZED__FINDINITIALTHESHOLD__


IPCVAPI_IMPL( CvStatus, icvAdaptiveThreshold8uC1R, (uchar * src,
                                                    int src_step,
                                                    uchar * dst,
                                                    int dst_step,
                                                    CvSize roi,
                                                    uchar maxValue,
                                                    CvThreshType type,
                                                    int size, uchar minDisp) )
{
    //     Some variables 
    static int i, j, s, t, tt, jj;
    int thresh;
    int mean, disp, minDisp2;
    uchar *srct, *dstt;

    //  uchar *tresh_buf;  

    //  Check for bad arguments 
    if( !src || !dst )
        return CV_NULLPTR_ERR;
    if( roi.width <= 0 || roi.height <= 0 )
        return CV_BADSIZE_ERR;
    if( roi.width > src_step )
        return CV_BADSIZE_ERR;
    if( roi.width > dst_step )
        return CV_BADSIZE_ERR;
    if( type != CV_THRESH_BINARY &&
        type != CV_THRESH_BINARY_INV &&
        type != CV_THRESH_TOZERO && type != CV_THRESH_TOZERO_INV )
    {
        return CV_BADFACTOR_ERR;
    }

    if( size != 1 && size != 2 && size != 3 && size != 4 )
    {
        return CV_BADFACTOR_ERR;
    }

    //    Calculating 
    minDisp2 = (int) minDisp *minDisp;

    thresh = -1;

    srct = src;
    dstt = dst;

    switch (type)
    {
    case CV_THRESH_BINARY:

        icvFindInitialThreshold( srct, src_step, dstt, dst_step, roi, size, minDisp2,
                                 &thresh );
        srct = src;
        dstt = dst;

        for( i = 0; i < roi.height; i++, srct += src_step, dstt += dst_step )
        {
            for( j = 0; j < roi.width; j++ )
            {
                mean = 0;
                disp = 0;
                tt = 0;
                for( s = -size; s < size; s++ )
                {
                    if( i + s >= 0 && i + s < roi.height )
                    {
                        for( t = -size; t < size; t++ )
                        {
                            int jt = j + t;

                            if( jt >= 0 && jt < roi.width )
                            {
                                int jj = jt + s * src_step;

                                mean += srct[jj];
                                disp += srct[jj] * srct[jj];
                                tt++;
                            }
                        }
                    }
                }
                mean = mean / tt;
                disp = (int) (disp / tt - mean * mean);

                if( disp > minDisp2 )
                    thresh = mean + (int) (coef * disp);

                dstt[j] = (uchar) (((thresh - srct[j]) >> 8) & maxValue);

            }
        }
        break;

    case CV_THRESH_BINARY_INV:

        icvFindInitialThreshold( srct, src_step, dstt, dst_step, roi, size, minDisp2,
                                 &thresh );
        srct = src;
        dstt = dst;
        for( i = 0; i < roi.height; i++, srct += src_step, dstt += dst_step )
            for( j = 0; j < roi.width; j++ )
            {
                mean = 0;
                disp = 0;
                tt = 0;
                for( s = -size; s < size; s++ )
                {
                    if( i + s >= 0 && i + s < roi.height )
                    {
                        for( t = -size; t < size; t++ )
                        {
                            int jt = j + t;

                            if( jt >= 0 && jt < roi.width )
                            {
                                jj = jt + s * src_step;
                                mean += srct[jj];
                                disp += srct[jj] * srct[jj];
                                tt += 1;
                            }
                        }
                    }
                }
                mean = mean / tt;
                disp = (int) (disp / tt - mean * mean);

                if( disp > minDisp2 )
                    thresh = mean + (int) (coef * disp);

                dstt[j] = (uchar) (((srct[j] - thresh - 1) >> 8) & maxValue);
            }
        break;

    case CV_THRESH_TOZERO:
        icvFindInitialThreshold( srct, src_step, dstt, dst_step, roi, size, minDisp2,
                                 &thresh );
        srct = src;
        dstt = dst;
        for( i = 0; i < roi.height; i++, srct += src_step, dstt += dst_step )
            for( j = 0; j < roi.width; j++ )
            {

                mean = 0;
                disp = 0;
                tt = 0;
                for( s = -size; s < size; s++ )
                {
                    if( i + s >= 0 && i + s < roi.height )
                    {
                        for( t = -size; t < size; t++ )
                        {
                            int jt = j + t;

                            if( jt >= 0 && jt < roi.width )
                            {
                                jj = jt + s * src_step;
                                mean += srct[jj];
                                disp += srct[jj] * srct[jj];
                                tt += 1;
                            }
                        }
                    }
                }
                mean = mean / tt;
                disp = (int) (disp / tt - mean * mean);

                if( disp > minDisp2 )
                    thresh = mean + (int) (coef * disp);

                long temp = srct[j];

                dstt[j] = (uchar) (((thresh - temp) >> 31) & temp);
            }
        break;

    case CV_THRESH_TOZERO_INV:
        icvFindInitialThreshold( srct, src_step, dstt, dst_step, roi, size, minDisp2,
                                 &thresh );
        srct = src;
        dstt = dst;
        for( i = 0; i < roi.height; i++, srct += src_step, dstt += dst_step )
            for( j = 0; j < roi.width; j++ )
            {
                mean = 0;
                disp = 0;
                tt = 0;
                for( s = -size; s < size; s++ )
                {
                    if( i + s >= 0 && i + s < roi.height )
                    {
                        for( t = -size; t < size; t++ )
                        {
                            int jt = j + t;

                            if( jt >= 0 && jt < roi.width )
                            {
                                int jj = jt + s * src_step;

                                mean += srct[jj];
                                disp += srct[jj] * srct[jj];
                                tt += 1;
                            }
                        }
                    }
                }
                mean = mean / tt;
                disp = (int) (disp / tt - mean * mean);

                if( disp > minDisp2 )
                    thresh = mean + (int) (coef * disp);

                long temp = srct[j];

                dstt[j] = (uchar) ((((int) temp - thresh - 1) >> 31) & temp);
            }
        break;
    default:
        assert( NULL );
    }
    return CV_NO_ERR;
}                               // icvAdaptThreshold8uC1R 


  /*F///////////////////////////////////////////////////////////////////////////////////////
     //    Name:    icvFindInitialThreshold
     //    Purpose: 
     //    Context:
     //    Parameters:
     //    Returns:
     //    Notes:
     //F */
IPCVAPI_IMPL( CvStatus, icvFindInitialThreshold, (uchar * src, int src_step,
                                                  uchar * dst, int dst_step,
                                                  CvSize roi, int size,
                                                  int minDisp, int *thresh) )
{
    static int i, j, s, t, tt, jj;
    int mean, disp;

    for( i = 0; i < roi.height; i++, src += src_step, dst += dst_step )
    {
        for( j = 0; j < roi.width; j++ )
        {
            mean = 0;
            disp = 0;
            tt = 0;
            for( s = -size; s < size; s++ )
            {
                if( i + s >= 0 && i + s < roi.height )
                {
                    for( t = -size; t < size; t++ )
                    {
                        int jt = j + t;

                        if( jt >= 0 && jt < roi.width )
                        {
                            int jj = jt + s * src_step;

                            mean += src[jj];
                            disp += src[jj] * src[jj];
                            tt++;
                        }
                    }
                }
            }
            mean = mean / tt;
            disp = (int) (disp / tt - mean * mean);

            if( disp > minDisp )
            {
                *thresh = mean + (int) (coef * disp);
                return CV_NO_ERR;
            }

        }
    }
    return CV_NO_ERR;
}                               // icvFindInitialThreshold


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    FindInitiThreshold
//    Purpose: 
//    Context:
//    Parameters:
//    Returns:
//    Notes:
//F*/
CvStatus
FindInitiThreshold( uchar * src, int src_step, CvSize roi,
                    int size, int minDisp, int *thresh )
{
    //////////////////////// Variables ////////////////////////
    ///////////
    int height = roi.height;
    int width = roi.width;
    int y, x;                   // curent coordinates
    int s, t;                   // curent coordinates in scan window
    int Limit_s_min, Limit_s_max;       // coordinates of scan window
    int Limit_t_min, Limit_t_max;
    int Delta;                  // amount of pixels in scan window
    int mean, disp;             // average, dispersion
    uchar Val;                  // pixel
    uchar *src_s;               // curent line in scan window

    for( y = 0; y < height; y++ )
    {
        Limit_s_min = (0 > (y - size)) ? 0 : (y - size);
        Limit_s_max = (height < (y + size)) ? height : (y + size);

        for( x = 0; x < width; x++ )
        {
            mean = 0;
            disp = 0;

            Limit_t_min = (0 > (x - size)) ? 0 : (x - size);
            Limit_t_max = (width < (x + size)) ? width : (x + size);

            src_s = src + Limit_s_min * src_step;
            for( s = Limit_s_min; s < Limit_s_max; s++, src_s += src_step )
            {
                for( t = Limit_t_min; t < Limit_t_max; t++ )
                {

                    Val = src_s[t];
                    mean += Val;
                    disp += Val * Val;
                }
            }

            Delta = (Limit_t_max - Limit_t_min) * (Limit_t_max - Limit_t_min);

            mean = mean / Delta;

            disp = (int) (disp / Delta - mean * mean);

            if( disp > minDisp )
            {
                *thresh = mean + (int) (coef * disp);
                return CV_NO_ERR;
            }
        }
    }
    return CV_NO_ERR;
}                               // icvFindInitialThreshold


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    _FindInitiThreshold
//    Purpose: Find initial threshold
//    Context:
//    Parameters:
//      src - pointer to source image data
//      size - size of scan window
//      MeanLine - buffer for keeping average of pixels in scan line
//      DispLine - buffer for keeping power of 2 of pixels in scan line
//
//    Returns:
//    Notes:
//F*/
void
_Read_Lines( uchar * src, int *MeanLine, int *DispLine, int width, int size )
{
    //////////// some variables //////////////
    int Disp;                   // dispersion
    int Mean;                   // mesan
    uchar *Beg;                 // pointer to first pixel on scan line
    uchar *End;                 // pointer to the pixel after last pixel of scan line
    uchar Val_1, Val_2;         // temp values ( value of depth in pixel )
    register int x;

    /////////// Init ///////////////
    Disp = Mean = 0;
    End = Beg = src;
    for( ; Beg < src + size; )
    {
        Val_1 = *Beg;
        Mean += Val_1;
        Disp += Val_1 * Val_1;

        Beg++;
    }                           // for(  ; Beg < src + size; )

    /////////////// calculating //////////////////////
    for(  /*register int */ x = 0; x < size; x++ )
    {

        Val_1 = *Beg;
        Mean += Val_1;
        Disp += Val_1 * Val_1;

        MeanLine[x] = Mean;
        DispLine[x] = Disp;

        Beg++;
    }                           // for( x = 0 ; x < size; x++ )

    for( ; x < (width - size); x++ )
    {
        Val_1 = *Beg;
        Val_2 = *End;

        Mean += (Val_1 - Val_2);
        Disp += (Val_1 * Val_1 - Val_2 * Val_2);

        MeanLine[x] = Mean;
        DispLine[x] = Disp;

        Beg++;
        End++;
    }                           // for(  ; x < ( width - size ); x++ )

    for( ; x < width; x++ )
    {
        Val_1 = *End;
        Mean -= Val_1;
        Disp -= Val_1 * Val_1;

        MeanLine[x] = Mean;
        DispLine[x] = Disp;

        End++;
    }                           // for( ; x < width; x++ )

}                               // CvStatus _Read_Lines( uchar* src, int* MeanLine, int* DispLine ,int y, int width , int size )

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    _FindInitiThreshold
//    Purpose: Find initial threshold
//    Context:
//    Parameters:
//      src - pointer to source image data
//      src_step - line step of src data
//      roi - image sizes
//      size - size of scan window
//      minDisp - min dispersion
//      thresh - pointer to thresholds
//
//    Returns:
//    Notes:
//F*/
CvStatus
_FindInitiThreshold( uchar * src,
                     int src_step, CvSize roi, int size, int minDisp, int *thresh )
{
    //////////////////////// Variables ////////////////////////
    uchar *cur_src_line;        // cyrent line of src image

    ///////////
    int height = roi.height;
    int width = roi.width;
    int y, x;                   // curent coordinates

    /////////////
    int Disp = 0;               // dispersion
    int Mean = 0;               // average
    int t1, t2, t3;             // temporaly variables

    /////////////
    int *Lines = NULL;          // pointer to allocated memory
    int *MeanLines;             // 2*size+1 lines buffer of intermediate result of average culculating
    int *DispLines;             // 2*size+1 lines buffer of intermediate result of dispersion culculating
    int *MeanLines_NextLine;    // pointer to second line in MeanLines buffer
    int *DispLines_NextLine;    // pointer to second line in DispLines buffer
    int *MeanLines_LastLine;    // pointer to last line in  MeanLines buffer
    int *DispLines_LastLine;    // pointer to last line in  DispLines buffer

    ///////////
    int *Means = NULL;          // buffer of calculating means ( for recoverage )
    int *Disps = NULL;          // buffer of calculating means ( for recoverage )

    ///////////
    int *p1, *p2;               // temporaly poiters

    ///////////
    int Y_Amount;               // amount pixels in y - cat of scan window

    /////////// Sizes
    int S_W;
    int S_W_W = (S_W = size * width) + width;
    int MemSize = 2 * (2 * S_W_W + width);
    int SizeOfLines = 2 * S_W_W;
    int LastLine;
    int SizeOfMove = (LastLine = 2 * S_W + width) * sizeof( int );



    /////////////// Allocating memory for lines
    Lines = new int[MemSize];

    if( Lines == NULL )
    {
        return CV_OUTOFMEM_ERR;

    }

    MeanLines = Lines;
    DispLines = Lines + SizeOfLines;

    MeanLines_NextLine = MeanLines + width;
    DispLines_NextLine = DispLines + width;

    MeanLines_LastLine = MeanLines + LastLine;
    DispLines_LastLine = DispLines + LastLine;

    Means = DispLines + SizeOfLines;
    Disps = Means + width;

    ////////////// Initialising ////////////////
    memset( Lines, 0, MemSize * sizeof( int ));

    cur_src_line = src;
    for( y = 0; y < S_W; y += width, cur_src_line += src_step )
    {
        _Read_Lines( cur_src_line, MeanLines + y, DispLines + y, width, size );
    }

    for( x = 0; x < width; x++ )
    {
        Mean = 0;
        Disp = 0;
        for( t1 = x; t1 < (x + S_W); t1 += width )
        {
            Mean += MeanLines[t1];
            Disp += DispLines[t1];
        }
        Means[x] = Mean;
        Disps[x] = Disp;
    }


    ////////////////// calculating /////////////////////////
    Y_Amount = size + 1;
    cur_src_line = src + size * src_step;
    t2 = height - size;
    t3 = width - size;
    for( y = 0; y < height; y++, cur_src_line += src_step )
    {
        /////////////// Move lines //////////////
        memmove( MeanLines_NextLine, MeanLines, SizeOfMove );
        memmove( DispLines_NextLine, DispLines, SizeOfMove );

        ////////////// Fill lines ///////////////
        if( y < t2 )
        {
            _Read_Lines( cur_src_line, MeanLines, DispLines, width, size );
        }                       // if( y < t2 )

        ///////////// calculating ///////////// 
        t1 = size * Y_Amount;
        p1 = Means;
        p2 = Disps;
        for( x = 0; x < width; x++, p1++, p2++ )
        {

            Mean = *p1;
            Disp = *p2;

            if( y < t2 )
            {
                Mean += MeanLines[x];
                Disp += DispLines[x];
            }                   // if( y < t2 )
            if( y > size )
            {

                Mean -= MeanLines_LastLine[x];
                Disp -= DispLines_LastLine[x];
            }                   // if( y > size )

            *p1 = Mean;
            *p2 = Disp;

            Mean = Mean / t1;
            Disp = (Disp / t1 - Mean * Mean);

            if( Disp > minDisp )
            {
                *thresh = Mean + (int) (coef * Disp);
                delete[]Lines;
                return CV_NO_ERR;
            }                   // if( Disp > minDisp) 

            if( x < size )
            {
                t1 += Y_Amount;
            }                   // if( x < size )
            else if( x >= t3 )
            {
                t1 += Y_Amount;
            }                   //else if( x >=  t3  )

        }                       // for( x = 0 ; x < width ; x++ , p1++, p2++)

        if( y < size )
        {
            Y_Amount++;
        }                       // if( y < size )
        else if( y >= t2 )
        {
            Y_Amount--;
        }                       // else if( y >= t2 )

    }                           // for( y = 0; y < height; y++ , cur_src_line += src_step )

    delete[]Lines;
    return CV_NO_ERR;

}                               // CvStatus _FindInitiThreshold


CvStatus
_CalcDispMean( IplImage * srcIm, IplImage * dispIm, IplImage * meanIm, int size )
{
    //////////// check input parametrs ///////////////////
    if( (srcIm == NULL) || (dispIm == NULL) || (meanIm == NULL) )
    {
        return CV_NULLPTR_ERR;
    }                           // if

    //////////// variables ///////////////
    uchar *src;
    int step_srcIm;
    int number = 2 * size + 1;
    CvSize roi_srcIm;

    IplImage*     AverageIm_f;
    IplImage*     Temp1_f;
    IplImage*     Temp2_f;
    _CvConvState* convState;

    /////////////// initialising /////////////////////////
    cvGetImageRawData( srcIm, &src, &step_srcIm, &roi_srcIm );

    //////////// creating images ///////////////////////
    Temp2_f = cvCreateImage( roi_srcIm, IPL_DEPTH_32F, 1 );
    AverageIm_f = cvCreateImage( roi_srcIm, IPL_DEPTH_32F, 1 );
    Temp1_f = cvCreateImage( roi_srcIm, IPL_DEPTH_32F, 1 );
    icvBlurInitAlloc( roi_srcIm.width, cv32f, number, &convState );

    if( (Temp2_f == NULL) || (AverageIm_f == NULL) || (Temp1_f == NULL) || !convState )
    {
        cvReleaseImage( &Temp2_f );
        cvReleaseImage( &AverageIm_f );
        cvReleaseImage( &Temp1_f );
        icvConvolFree( &convState );
        return CV_OUTOFMEM_ERR;
    }                           // if

    ////////////// calculating //////////////////////
    cvConvertScale( srcIm, Temp2_f, 1, 0 );
    icvBlur_32f_C1R( (float*)Temp2_f->imageData,     Temp2_f->widthStep,
                     (float*)AverageIm_f->imageData, AverageIm_f->widthStep,
                     &roi_srcIm, convState, 0 );
    icvSquare( Temp2_f, Temp1_f );      // square
    icvBlur_32f_C1R( (float*)Temp1_f->imageData, Temp1_f->widthStep,
                     (float*)Temp2_f->imageData, Temp2_f->widthStep,
                     &roi_srcIm, convState, 0 );

    icvSubtract( Temp2_f, AverageIm_f, Temp1_f );       // dispersion powered 2

    cvConvertScale( Temp2_f, dispIm, 1.0 / (number * number), 0 );    // convert to IPL_DEPTH_8U
    cvConvertScale( AverageIm_f, meanIm, 1.0 / (number * number), 0 );

    ///////////// relesing memory ///////////////////////////
    cvReleaseImage( &Temp1_f );
    cvReleaseImage( &Temp2_f );
    cvReleaseImage( &AverageIm_f );
    icvConvolFree( &convState );

    return CV_NO_ERR;

} // CvStatus _CalcDispMean( IplImage* srcIm, IplImage* dispIm, IplImage* meanIm , int size )

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    AdaptThreshold_8u_C1R
//    Purpose: Adaptive Thresholding the source array
//    Context:
//    Parameters:
//      SrcIm   - source image
//      src     - source array
//      dst     - destination array
//      roi     - size of picture in elements
//      dst_step,src_step - lenght of string
//      type    - thresholding type, must be one of
//                  CV_THRESH_BINARY       - val = (val > Thresh ? MAX    : 0)
//                  CV_THRESH_BINARY_INV   - val = (val > Thresh ? 0      : MAX)
//                  CV_THRESH_TOZERO       - val = (val > Thresh ? val    : 0)
//                  CV_THRESH_TOZERO_INV   - val = (val > Thresh ? 0      : val)
//      size     - size of thresholding neighborhood  ( 1 - 3x3 , 2 - 5x5, 3 - 7x7, 4 - 9x9 )
//    Returns:
//    Notes:
//      The MAX constant for uchar is 255, for char is 127
//F*/
CvStatus
AdaptThreshold_8u_C1R( IplImage * SrcIm, uchar * src, int src_step,
                       uchar * dst, int dst_step,
                       CvSize roi, uchar maxValue, CvThreshType type,
                       int size, uchar minDisp )
{
    //////////////// Some variables  /////////////////////

    IplImage *AverageIm;
    IplImage *DispersionOfPower2;
    IplImage *Image;

    CvSize roi_Image;
    CvSize roi_Disp;
    CvSize roi_Aver;

    uchar disp;

    uchar *image;
    uchar *Disp;
    uchar *Aver;
    uchar *srct;
    uchar *dstt;
    uchar *Dispt;
    uchar *Avert;

    int step_Image;
    int step_Disp;
    int step_Aver;

    int thresh;
    int minDisp2;

    int i, j;
    int temp2;
    long temp1;

    CvStatus Ret;

    //////////////// Check for bad arguments ////////////////
    if( !src || !dst )
        return CV_NULLPTR_ERR;
    if( roi.width <= 0 || roi.height <= 0 )
        return CV_BADSIZE_ERR;
    if( roi.width > src_step )
        return CV_BADSIZE_ERR;
    if( roi.width > dst_step )
        return CV_BADSIZE_ERR;
    if( type != CV_THRESH_BINARY &&
        type != CV_THRESH_BINARY_INV &&
        type != CV_THRESH_TOZERO && type != CV_THRESH_TOZERO_INV )
        return CV_BADFACTOR_ERR;

    if( size != 1 && size != 2 && size != 3 && size != 4 )
        return CV_BADFACTOR_ERR;

    //////////////// creating images ////////////////

    AverageIm = cvCreateImage( roi, IPL_DEPTH_8U, 1 );
    DispersionOfPower2 = cvCreateImage( roi, IPL_DEPTH_8U, 1 );

    if( (AverageIm == NULL) || (DispersionOfPower2 == NULL) )
    {
        return CV_OUTOFMEM_ERR;
    }                           // if

    //////////////// calculating dispersion ////////////////

    Ret = _CalcDispMean( SrcIm, AverageIm, DispersionOfPower2, size );
    if( CV_NO_ERR != Ret )
    {
        cvReleaseImage( &AverageIm );
        cvReleaseImage( &DispersionOfPower2 );
        return Ret;
    }

    ///////////////// initialising variables ///////////////////

    cvGetImageRawData( DispersionOfPower2, &Disp, &step_Disp, &roi_Disp );
    cvGetImageRawData( AverageIm, &Aver, &step_Aver, &roi_Aver );

    minDisp2 = (int) minDisp *(int) minDisp;

    srct = src;
    dstt = dst;
    Dispt = Disp;
    Avert = Aver;

    //////////////////// Calculating ////////////////////

    _FindInitiThreshold( src, src_step, roi, size, minDisp2, &thresh );

    switch (type)
    {
        ////////////////// CV_THRESH_BINARY ///////////////////////
    case CV_THRESH_BINARY:
        for( i = 0; i < roi.height; i++, srct += src_step,
             dstt += dst_step, Dispt += step_Disp, Avert += step_Aver )
        {
            for( j = 0; j < roi.width; j++ )
            {
                disp = Dispt[j];
                if( disp > minDisp2 )
                {
                    thresh = Avert[j] + (int) (coef * disp);
                }               // if(disp > minDisp2) 

                dstt[j] = (uchar) (((thresh - srct[j]) >> 8) & maxValue);
            }                   // for( j = 0; j < roi.width; j++ )
        }                       // for( i
        break;

        ///////////////////// CV_THRESH_BINARY_INV ////////////////////////
    case CV_THRESH_BINARY_INV:
        Image = cvCreateImage( roi, IPL_DEPTH_8U, 1 );
        if( Image == NULL )
        {
            cvReleaseImage( &AverageIm );
            cvReleaseImage( &DispersionOfPower2 );
            return CV_OUTOFMEM_ERR;
        }                       // if( Image == NULL )

        icvSubtractS( SrcIm, Image, 1, false );

        cvGetImageRawData( Image, &image, &step_Image, &roi_Image );

        for( i = 0; i < roi_Image.height; i++, image += step_Image,
             dstt += dst_step, Dispt += step_Disp, Avert += step_Aver )
        {
            for( j = 0; j < roi_Image.width; j++ )
            {
                disp = Dispt[j];
                if( disp > minDisp2 )
                {
                    thresh = Avert[j] + (int) (coef * disp);
                }               // if(disp > minDisp2) 
                dstt[j] = (uchar) (((image[j] - thresh) >> 8) & maxValue);
            }                   // for( j = 0; j < roi.width; j++ )
        }                       // for( i

        cvReleaseImage( &Image );
        break;

        ////////////////////// CV_THRESH_TOZERO ///////////////////////
    case CV_THRESH_TOZERO:
        for( i = 0; i < roi.height; i++, srct += src_step,
             dstt += dst_step, Dispt += step_Disp, Avert += step_Aver )
        {
            for( j = 0; j < roi.width; j++ )
            {
                disp = Dispt[j];
                if( disp > minDisp2 )
                {
                    thresh = Avert[j] + (int) (coef * disp);
                }               // if(disp > minDisp2) 

                temp1 = srct[j];
                dstt[j] = (uchar) (((thresh - temp1) >> 31) & temp1);
            }                   // for( j = 0; j < roi.width; j++ )
        }                       // for( i
        break;

        /////////////////////// CV_THRESH_TOZERO_INV ////////////////////////////////
    case CV_THRESH_TOZERO_INV:
        Image = cvCreateImage( roi, IPL_DEPTH_8U, 1 );
        if( Image == NULL )
        {

            cvReleaseImage( &AverageIm );
            cvReleaseImage( &DispersionOfPower2 );
            return CV_OUTOFMEM_ERR;
        }                       // if( Image == NULL )

        icvSubtractS( SrcIm, Image, 1, false );

        cvGetImageRawData( Image, &image, &step_Image, &roi_Image );

        for( i = 0; i < roi_Image.height; i++, image += step_Image,
             dstt += dst_step, Dispt += step_Disp, Avert += step_Aver )
        {
            for( j = 0; j < roi_Image.width; j++ )
            {

                disp = Dispt[j];

                if( disp > minDisp2 )
                {
                    thresh = Avert[j] + (int) (coef * disp);
                }               // if(disp > minDisp2) 

                temp2 = (int) image[j];
                dstt[j] = (uchar) (((temp2 - thresh) >> 31) & temp2);
            }                   // for( j = 0; j < roi.width; j++ )
        }                       // for( i

        cvReleaseImage( &Image );
        break;

    default:
        assert( NULL );
    }                           // switch( type )


    /////////////// releasing of images ///////////////
    cvReleaseImage( &AverageIm );
    cvReleaseImage( &DispersionOfPower2 );     

    return CV_NO_ERR;

}                               /* AdaptThreshold_8u_C1R */



/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    cvAdaptiveThreshold
//    Purpose: Adaptive Thresholding the source image
//    Context:
//    Parameters:
//      srcIm     - source image
//      dstIm     - result thresholding image
//      maxValue       - the maximum value of the image pixel
//      method    - method for the adaptive threshold calculation 
                    (now CV_STDDEF_ADAPTIVE_THRESH only)  
//      type      - thresholding type, must be one of
//                  CV_THRESH_BINARY       - val = (val > Thresh ? MAX    : 0)
//                  CV_THRESH_BINARY_INV   - val = (val > Thresh ? 0      : MAX)
//                  CV_THRESH_TOZERO       - val = (val > Thresh ? val    : 0)
//                  CV_THRESH_TOZERO_INV   - val = (val > Thresh ? 0      : val)
//      parameters - pointer to the input parameters (for the
//                   CV_STDDEF_ADAPTIVE_THRESH method parameters[0] is size of
//                   the neighborhood thresholding, (one of the 1-(3x3),2-(5x5),or
//                   3-(7x7)), parameters[1] is the value of the minimum variance
//    Returns:
//    Notes:
//F*/
CV_IMPL void
cvAdaptiveThreshold( IplImage *srcIm, IplImage *dstIm,
                     double maxValue, CvAdaptiveThreshMethod method,
                     CvThreshType type, double *parameters )
{
    ///////////////// Some  variables /////////////////
    CvSize roi_src;
    CvSize roi_dst;

    int step_src;
    int step_dst;

    uchar *src_data = 0;
    uchar *dst_data = 0;


    CV_FUNCNAME( "cvAdaptiveThreshold" );

    __BEGIN__;

    ///////////////// Checking /////////////////
    CV_CALL( CV_CHECK_IMAGE( srcIm ));
    CV_CALL( CV_CHECK_IMAGE( dstIm ));

    if( srcIm->depth != dstIm->depth )
    {
        CV_ERROR( IPL_BadDepth, "bad destination depth" );
    }                           // if( srcIm->depth != dstIm->depth )

    if( srcIm->depth != IPL_DEPTH_8U )
    {
        CV_ERROR( IPL_BadDepth, "bad destination depth" );
    }                           // if(srcIm->depth != IPL_DEPTH_8U)  

    if( srcIm->nChannels != dstIm->nChannels || srcIm->nChannels != 1 )
    {
        CV_ERROR( IPL_BadNumChannels, "bad channels numbers" );
    }  // if( srcIm->nChannels != dstIm->nChannels || srcIm->nChannels != 1 )


    ///////////////// Init variables /////////////////
    cvGetImageRawData( srcIm, &src_data, &step_src, &roi_src );
    cvGetImageRawData( dstIm, &dst_data, &step_dst, &roi_dst );

    ///////////////// Checking /////////////////
    if( roi_src.width != roi_dst.width || roi_src.height != roi_dst.height )
    {
        CV_ERROR( IPL_BadImageSize, "not equal roi" );
    }  // if(roi_src.width != roi_dst.width || roi_src.height != roi_dst.height)

    ///////////////// Call methods /////////////////
    switch (srcIm->depth)
    {
        ///////////////// IPL_DEPTH_8U /////////////////
    case IPL_DEPTH_8U:
        switch (method)
        {
            ///////////////// CV_STDDEV_ADAPTIVE_THRESH /////////////////
        case CV_STDDEV_ADAPTIVE_THRESH:
/*            
          IPPI_CALL( icvAdaptiveThreshold8uC1R( src_data, step_src,
                                              dst_data, step_dst, roi_src,
                                              (uchar)cvRound(maxValue), type, 
                                              cvRound(parameters[0]), // size
                                              (uchar)cvRound(parameters[1]))); // minDisp
*/
            AdaptThreshold_8u_C1R( srcIm, src_data, step_src,
                                   dst_data, step_dst,
                                   roi_src, (uchar) cvRound( maxValue ),
                                   type, cvRound( parameters[0] ),
                                   (uchar) cvRound( parameters[1] ));   // minDisp

            break;

        default:
            CV_ERROR_FROM_STATUS( CV_BADCOEF_ERR );
        }                       // switch(method)
        break;

    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }                           // switch(srcIm->depth)

    __CLEANUP__;
    __END__;
}                               // cvAdaptiveThreshold




/* End of file. */
