/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include "_cv.h"
#include "_cvarray.h"
#include "cv.h"

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    icvSet
//    Purpose:
//      Fill image by value
//    Context:
//    Parameters:
//      img  - source array
//      value - fill value
//    Returns:
//    Notes:
//F*/
void
icvSet( IplImage * img, float value )
{
    assert( img );
    if( !img )
        return;

    assert( img->imageData );
    if( !img->imageData )
        return;

    uchar *data;
    int step;
    CvSize size;
    int x, y;

    int lenght = img->width * img->nChannels;

    cvGetImageRawData( img, &data, &step, &size );

    short *data16 = (short *) data;
    int *data32 = (int *) data;

    char val8 = (char) value;
    short val16 = (short) value;
    int val32 = (int) value;

    switch (img->depth)
    {
    case IPL_DEPTH_8U:
        val8 = (uchar) value;
    case IPL_DEPTH_8S:
        for( y = 0; y < size.height; y++, data += img->widthStep )
            for( x = 0; x < lenght; x++ )
                data[x] = (char) val8;
        return;
    case IPL_DEPTH_16U:
    case IPL_DEPTH_16S:
        for( y = 0; y < size.height;
             y++, data16 = (short *) ((char *) data16 + img->widthStep) )
            for( x = 0; x < lenght; x++ )
                data16[x] = val16;
        return;
    case IPL_DEPTH_32S:
        for( y = 0; y < size.height; y++, data32 = (int *) ((char *) data32 + img->widthStep) )
            for( x = 0; x < lenght; x++ )
                data32[x] = val32;
        return;
    case IPL_DEPTH_32F:
        val32 = *(int *) (&value);
        for( y = 0; y < size.height; y++, data32 = (int *) ((char *) data32 + img->widthStep) )
            for( x = 0; x < lenght; x++ )
                data32[x] = val32;
        return;
    }
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    icvCopy
//    Purpose:
//      Copying from one image to another
//    Context:
//    Parameters:
//      src - source
//      dst - destination
//    Returns:
//    Notes:
//F*/
void
icvCopy( IplImage * src, IplImage * dst )
{
    assert( src && dst );
    if( !src || !dst )
        return;
    assert( src->imageData && dst->imageData );
    if( !src->imageData || !dst->imageData )
        return;

    uchar *src_data;
    uchar *dst_data;
    int src_step;
    int dst_step;
    CvSize src_size;
    CvSize dst_size;

    cvGetImageRawData( src, &src_data, &src_step, &src_size );
    cvGetImageRawData( dst, &dst_data, &dst_step, &dst_size );

    if( src_size.width != dst_size.width || src_size.height != dst_size.height ||
        src->nChannels != dst->nChannels || src->depth != dst->depth )
    {
        assert( 0 );
        return;
    }

    int len = src_size.width * src->nChannels;

    switch (src->depth)
    {
    case IPL_DEPTH_16U:
    case IPL_DEPTH_16S:
        len *= 2;
        break;
    case IPL_DEPTH_32S:
    case IPL_DEPTH_32F:
        len *= 4;
        break;
    }

    for( int y = 0; y < src_size.height; y++, src_data += src_step, dst_data += dst_step )
        memcpy( dst_data, src_data, len );
}                               /* icvCopy */


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    icv1DCopy
//    Purpose:
//      Copying 1D array from source to destination array
//    Context:
//    Parameters:
//      pSrc - source array
//      pDst - destination array
//      lLen - lenght of arrays
//    Returns:
//    Notes:
//F*/
void
icv1DCopy( void *pSrc, void *pDst, int lLen )
{
    long i;

    for( i = 0; i < lLen; i++ )
        ((char *) pDst)[i] = ((char *) pSrc)[i];
}                               /* icv1DCopy */


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    icv1ucSet
//    Purpose:
//      Setting unsigned char value to every element of destination value
//    Context:
//    Parameters:
//      ucValue - source value
//      pDst    - destination array
//      lLen    - lenght of destination array
//    Returns:
//    Notes:
//F*/
void
icv1ucSet( unsigned char ucValue, void *pDst, long lLen )
{
    long i;

    for( i = 0; i < lLen; i++ )
        ((unsigned char *) pDst)[i] = ucValue;
}                               /* icv1ucSet */


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    icvCopy8uC1R
//    Purpose:
//      Copies content of ROI of source image into destination image
//    Context:
//    Parameters:
//      src      - source image
//      src_step - full width of source image row in bytes
//      dst      - destination image
//      dst_step - full width of destination image row in bytes
//      size     - width and height of copied ROI in pixels
//    Returns:
//    Notes:
//F*/
void
icvCopy8uC1R( const uchar * src, int src_step, uchar * dst, int dst_step, CvSize size )
{
    int y;

    assert( src && dst );
    assert( size.width > 0 && size.height > 0 &&
            src_step >= size.width && dst_step >= size.width );

    for( y = 0; y < size.height; y++, src += src_step, dst += dst_step )
    {
        memcpy( dst, src, size.width );
    }
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    icvCopyToBuf8uC1R
//    Purpose:
//      Copies content of ROI of source image into destination buffer
//    Context:
//    Parameters:
//      src      - source image
//      src_step - full width of source image row in bytes
//      dst      - destination buffer
//      size     - width and height of copied ROI in pixels
//    Returns:
//    Notes:
//F*/
void
icvCopyToBuf8uC1R( const uchar * src, int src_step, uchar * dst, CvSize size )
{
    int y;

    assert( src && dst );
    assert( size.width > 0 && size.height > 0 && src_step >= size.width );

    for( y = 0; y < size.height; y++, src += src_step, dst += size.width )
    {
        memcpy( dst, src, size.width );
    }
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    icvSet32f
//    Purpose:
//      Setting unsigned char value to every element of destination value
//    Context:
//    Parameters:
//      ucValue - source value
//      pDst    - destination array
//      lLen    - lenght of destination array
//    Returns:
//    Notes:
//F*/
void
icvSet32f( float Value, float *pDst, long lLen )
{
    long i;

    for( i = 0; i < lLen; i++ )
        pDst[i] = Value;
}                               /* icvSet32f */


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    icvSubtract
//    Purpose:
//      Subs one image from another and puts result to destination image dst = srcA - srcB
//    Context:
//    Parameters:
//      srcA - first source
//      srcB - second source
//      dst  - destination
//    Returns:
//    Notes:
//F*/
void
icvSubtract( IplImage * srcA, IplImage * srcB, IplImage * dst )
{
    if( !srcA || !srcB || !dst )
    {
        assert( 0 );
        return;
    }

    if( !srcA->imageData || !srcB->imageData || !dst->imageData )
    {
        assert( 0 );
        return;
    }

    uchar *srcA_data;
    uchar *srcB_data;
    uchar *dst_data;
    int srcA_step;
    int srcB_step;
    int dst_step;
    CvSize srcA_size;
    CvSize srcB_size;
    CvSize dst_size;

    cvGetImageRawData( srcA, &srcA_data, &srcA_step, &srcA_size );
    cvGetImageRawData( srcB, &srcB_data, &srcB_step, &srcB_size );
    cvGetImageRawData( dst, &dst_data, &dst_step, &dst_size );

    if( srcA_size.width != dst_size.width || srcA_size.height != dst_size.height ||
        srcB_size.width != dst_size.width || srcB_size.height != dst_size.height ||
        srcA->nChannels != dst->nChannels || srcB->nChannels != dst->nChannels ||
        srcA->depth != dst->depth || srcB->depth != dst->depth )
    {
        assert( 0 );
        return;
    }

    switch (srcA->depth)
    {
    case IPL_DEPTH_8U:
        {
            uchar *s1 = srcA_data;
            uchar *s2 = srcB_data;
            uchar *d = dst_data;

            for( int y = 0; y < srcA_size.height; y++, s1 += srcA->widthStep,
                 s2 += srcB->widthStep, d += dst->widthStep )
                for( int x = 0; x < srcA_size.width; x++ )
                    d[x] = (uchar) (s1[x] - s2[x]);
        }
        break;
    case IPL_DEPTH_8S:
        {
            char *s1 = (char *) srcA_data;
            char *s2 = (char *) srcB_data;
            char *d = (char *) dst_data;

            for( int y = 0; y < srcA_size.height; y++, s1 += srcA->widthStep,
                 s2 += srcB->widthStep, d += dst->widthStep )
                for( int x = 0; x < srcA_size.width; x++ )
                    d[x] = (char) (s1[x] - s2[x]);
        }
        break;
    case IPL_DEPTH_32S:
        {
            int *s1 = (int *) srcA_data;
            int *s2 = (int *) srcB_data;
            int *d = (int *) dst_data;

            for( int y = 0; y < srcA_size.height;
                 y++, s1 = (int *) ((char *) s1 + srcA->widthStep), s2 =
                 (int *) ((char *) s2 + srcB->widthStep), d =
                 (int *) ((char *) d + dst->widthStep) )
                for( int x = 0; x < srcA_size.width; x++ )
                    d[x] = s1[x] - s2[x];
        }
        break;
    case IPL_DEPTH_32F:
        {
            float *s1 = (float *) srcA_data;
            float *s2 = (float *) srcB_data;
            float *d = (float *) dst_data;

            for( int y = 0; y < srcA_size.height;
                 y++, s1 = (float *) ((char *) s1 + srcA->widthStep), s2 =
                 (float *) ((char *) s2 + srcB->widthStep), d =
                 (float *) ((char *) d + dst->widthStep) )
                for( int x = 0; x < srcA_size.width; x++ )
                    d[x] = s1[x] - s2[x];
        }
        break;
    }
}                               /* icvSubtract */


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    icvSubtractS
//    Purpose:
//      The functions change the image intensity as follows:
//      If flip is false, the value is subtracted from the image pixel values.
//      If flip is true, the image pixel values are subtracted from the value.
//    Context:
//    Parameters:
//      src - source
//      dst  - destination
//    Returns:
//    Notes:
//F*/
void
icvSubtractS( IplImage * src, IplImage * dst, float value, int flip )
{
    if( !src || !dst )
    {
        assert( 0 );
        return;
    }

    if( !src->imageData || !dst->imageData )
    {
        assert( 0 );
        return;
    }

    uchar *src_data;
    uchar *dst_data;
    int src_step;
    int dst_step;
    CvSize src_size;
    CvSize dst_size;

    cvGetImageRawData( src, &src_data, &src_step, &src_size );
    cvGetImageRawData( dst, &dst_data, &dst_step, &dst_size );

    if( src_size.width != dst_size.width || src_size.height != dst_size.height ||
        src->nChannels != dst->nChannels || src->depth != dst->depth )
    {
        assert( 0 );
        return;
    }

    switch (src->depth)
    {
    case IPL_DEPTH_8U:
        {
            uchar *s1 = src_data;
            uchar *d = dst_data;
            uchar v = (uchar) value;

            if( !flip )
                for( int y = 0; y < src_size.height; y++, s1 += src->widthStep,
                     d += dst->widthStep )
                    for( int x = 0; x < src_size.width; x++ )
                        d[x] = (uchar) (s1[x] - v);
            else
                for( int y = 0; y < src_size.height; y++, s1 += src->widthStep,
                     d += dst->widthStep )
                    for( int x = 0; x < src_size.width; x++ )
                        d[x] = (uchar) (v - s1[x]);
        }
        break;
    case IPL_DEPTH_8S:
        {
            char *s1 = (char *) src_data;
            char *d = (char *) dst_data;
            char v = (char) value;

            if( !flip )
                for( int y = 0; y < src_size.height; y++, s1 += src->widthStep,
                     d += dst->widthStep )
                    for( int x = 0; x < src_size.width; x++ )
                        d[x] = (char) (s1[x] - v);
            else
                for( int y = 0; y < src_size.height; y++, s1 += src->widthStep,
                     d += dst->widthStep )
                    for( int x = 0; x < src_size.width; x++ )
                        d[x] = (char) (v - s1[x]);
        }
        break;
    case IPL_DEPTH_32S:
        {
            int *s1 = (int *) src_data;
            int *d = (int *) dst_data;
            int v = (int) value;

            if( !flip )
                for( int y = 0; y < src_size.height;
                     y++, s1 = (int *) ((char *) s1 + src->widthStep), d =
                     (int *) ((char *) d + dst->widthStep) )
                    for( int x = 0; x < src_size.width; x++ )
                        d[x] = s1[x] - v;
            else
                for( int y = 0; y < src_size.height;
                     y++, s1 = (int *) ((char *) s1 + src->widthStep), d =
                     (int *) ((char *) d + dst->widthStep) )
                    for( int x = 0; x < src_size.width; x++ )
                        d[x] = s1[x] - v;
        }
        break;
    case IPL_DEPTH_32F:
        {
            float *s1 = (float *) src_data;
            float *d = (float *) dst_data;
            float v = (uchar) value;

            if( !flip )
                for( int y = 0; y < src_size.height;
                     y++, s1 = (float *) ((char *) s1 + src->widthStep), d =
                     (float *) ((char *) d + dst->widthStep) )
                    for( int x = 0; x < src_size.width; x++ )
                        d[x] = s1[x] - v;
            else
                for( int y = 0; y < src_size.height;
                     y++, s1 = (float *) ((char *) s1 + src->widthStep), d =
                     (float *) ((char *) d + dst->widthStep) )
                    for( int x = 0; x < src_size.width; x++ )
                        d[x] = s1[x] - v;
        }
        break;
    }
}                               /* icvSubtractS */


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    icvSquare
//    Purpose:
//      The function increases the intensity of an image by squaring each pixel value.
//    Context:
//    Parameters:
//      src - first source
//      dst - destination
//    Returns:
//    Notes:
//F*/
void
icvSquare( IplImage * src, IplImage * dst )
{
    if( !src || !dst )
    {
        assert( 0 );
        return;
    }

    if( !src->imageData || !dst->imageData )
    {
        assert( 0 );
        return;
    }

    uchar *src_data;
    uchar *dst_data;
    int src_step;
    int dst_step;
    CvSize src_size;
    CvSize dst_size;

    cvGetImageRawData( src, &src_data, &src_step, &src_size );
    cvGetImageRawData( dst, &dst_data, &dst_step, &dst_size );

    if( src_size.width != dst_size.width || src_size.height != dst_size.height ||
        src->nChannels != dst->nChannels || src->depth != dst->depth )
    {
        assert( 0 );
        return;
    }

    switch (src->depth)
    {
    case IPL_DEPTH_8U:
        {
            uchar *s1 = src_data;
            uchar *d = dst_data;

            for( int y = 0; y < src_size.height; y++, s1 += src->widthStep,
                 d += dst->widthStep )
                for( int x = 0; x < src_size.width; x++ )
                    d[x] = (uchar) (s1[x] * s1[x]);
        }
        break;
    case IPL_DEPTH_8S:
        {
            char *s1 = (char *) src_data;
            char *d = (char *) dst_data;

            for( int y = 0; y < src_size.height; y++, s1 += src->widthStep,
                 d += dst->widthStep )
                for( int x = 0; x < src_size.width; x++ )
                    d[x] = (char) (s1[x] * s1[x]);
        }
        break;
    case IPL_DEPTH_32S:
        {
            int *s1 = (int *) src_data;
            int *d = (int *) dst_data;

            for( int y = 0; y < src_size.height;
                 y++, s1 = (int *) ((char *) s1 + src->widthStep), d =
                 (int *) ((char *) d + dst->widthStep) )
                for( int x = 0; x < src_size.width; x++ )
                    d[x] = s1[x] * s1[x];
        }
        break;
    case IPL_DEPTH_32F:
        {
            float *s1 = (float *) src_data;
            float *d = (float *) dst_data;

            for( int y = 0; y < src_size.height;
                 y++, s1 = (float *) ((char *) s1 + src->widthStep), d =
                 (float *) ((char *) d + dst->widthStep) )
                for( int x = 0; x < src_size.width; x++ )
                    d[x] = s1[x] * s1[x];
        }
        break;
    }
}                               /* icvSquare */

/* End of file. */
