/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/
#include "_cv.h"
#include "_cvwrap.h"


CV_IMPL void
cvAcc( IplImage * img, IplImage * sum, IplImage * mask )
{
    uchar *img_data = 0;
    uchar *sum_data = 0;
    uchar *mask_data = 0;
    int img_step = 0;
    int sum_step = 0;
    int mask_step = 0;
    CvSize img_size;
    CvSize sum_size;
    CvSize mask_size;
    int imgBtPix;

    CV_FUNCNAME( "cvAcc" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( img ));
    CV_CALL( CV_CHECK_IMAGE( sum ));
    if( mask )
        CV_CALL( CV_CHECK_MASK_IMAGE( mask ));

    if( sum->depth != IPL_DEPTH_32F )
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );

    if( img->nChannels != sum->nChannels )
        CV_ERROR( IPL_BadNumChannels, "images have different number of channels" );

    cvGetImageRawData( img, &img_data, &img_step, &img_size );
    cvGetImageRawData( sum, &sum_data, &sum_step, &sum_size );

    if( mask )
        cvGetImageRawData( mask, &mask_data, &mask_step, &mask_size );
    else
        mask_size = sum_size;

    if( img_size != sum_size || mask_size != sum_size )
        CV_ERROR( IPL_StsBadArg, "images' ROIs have different size" );

    img_size.width *= img->nChannels;
    imgBtPix = icvGetBtPix( img );

    if( img_step == img_size.width*imgBtPix &&
        sum_step == sum_size.width*sizeof_float &&
        (!mask || mask_step == mask_size.width ))
    {
        img_size.width *= img_size.height;
        img_step *= img_size.height;
        mask_step *= img_size.height;
        sum_step *= img_size.height;

        img_size.height = 1;
    }

	img_size.width *= img->nChannels;

    switch (img->depth)
    {
    case IPL_DEPTH_8U:
        if( mask )
        {
            IPPI_CALL( icvAdd_8u32f_C1IMR( img_data, img_step,
                                           mask_data, mask_step,
                                           (float*)sum_data, sum_step,
                                           img_size ));
        }
        else
        {
            IPPI_CALL( icvAdd_8u32f_C1IR( img_data, img_step,
                                          (float*)sum_data, sum_step,
                                          img_size ));
        }
        break;
    case IPL_DEPTH_8S:
        if( mask )
        {
            IPPI_CALL( icvAdd_8s32f_C1IMR( (char *) img_data, img_step,
                                           mask_data, mask_step,
                                           (float *) sum_data, sum_step,
                                           img_size ));
        }
        else
        {
            IPPI_CALL( icvAdd_8s32f_C1IR( (char *) img_data, img_step,
                                          (float *) sum_data, sum_step,
                                          img_size ));
        }
        break;
    case IPL_DEPTH_32F:
        if( mask )
        {
            IPPI_CALL( icvAdd_32f_C1IMR( (float *) img_data, img_step,
                                         mask_data, mask_step,
                                         (float *) sum_data, sum_step,
                                         img_size ));
        }
        else
        {
            IPPI_CALL( icvAdd_32f_C1IR( (float *) img_data, img_step,
                                        (float *) sum_data, sum_step,
                                        img_size ));
        }
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }
    __CLEANUP__;
    __END__;
}


CV_IMPL void
cvSquareAcc( IplImage * img, IplImage * sq_sum, IplImage * mask )
{
    uchar *img_data = 0;
    uchar *sum_data = 0;
    uchar *mask_data = 0;
    int img_step = 0;
    int sum_step = 0;
    int mask_step = 0;
    CvSize img_size;
    CvSize sum_size;
    CvSize mask_size;
    int imgBtPix;

    CV_FUNCNAME( "cvSquareAcc" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( img ));
    CV_CALL( CV_CHECK_IMAGE( sq_sum ));
    if( mask )
        CV_CALL( CV_CHECK_MASK_IMAGE( mask ));

    if( sq_sum->depth != IPL_DEPTH_32F )
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );

    if( img->nChannels != sq_sum->nChannels )
        CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );

    if(mask)
        if( img->nChannels != mask->nChannels )
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );

    cvGetImageRawData( img, &img_data, &img_step, &img_size );
    cvGetImageRawData( sq_sum, &sum_data, &sum_step, &sum_size );
    
    if( mask )
        cvGetImageRawData( mask, &mask_data, &mask_step, &mask_size );
    else
        mask_size = sum_size;

    if( img_size != sum_size || mask_size != sum_size )
        CV_ERROR( IPL_StsBadArg, "images' ROIs have different size" );

    imgBtPix = icvGetBtPix( img );

    if( img_step == img_size.width*imgBtPix &&
        sum_step == sum_size.width*sizeof_float &&
        (!mask || mask_step == mask_size.width ))
    {
        img_size.width *= img_size.height;
        img_step *= img_size.height;
        mask_step *= img_size.height;
        sum_step *= img_size.height;

        img_size.height = 1;
    }

	img_size.width *= img->nChannels;

    switch (img->depth)
    {
    case IPL_DEPTH_8U:
        if( mask )
        {
            IPPI_CALL( icvAddSquare_8u32f_C1IMR( img_data, img_step,
                                           mask_data, mask_step,
                                           (float*)sum_data, sum_step,
                                           img_size ));
        }
        else
        {
            IPPI_CALL( icvAddSquare_8u32f_C1IR( img_data, img_step,
                                          (float*)sum_data, sum_step,
                                          img_size ));
        }
        break;
    case IPL_DEPTH_8S:
        if( mask )
        {
            IPPI_CALL( icvAddSquare_8s32f_C1IMR((char *) img_data, img_step,
                                               mask_data, mask_step,
                                               (float *) sum_data, sum_step,
                                               img_size ));
        }
        else
        {
            IPPI_CALL( icvAddSquare_8s32f_C1IR((char *) img_data, img_step,
                                              (float *) sum_data, sum_step,
                                              img_size ));
        }
        break;
    case IPL_DEPTH_32F:
        if( mask )
        {
            IPPI_CALL( icvAddSquare_32f_C1IMR((float *) img_data, img_step,
                                               mask_data, mask_step,
                                              (float *) sum_data, sum_step,
                                              img_size ));
        }
        else
        {
            IPPI_CALL( icvAddSquare_32f_C1IR( (float *) img_data, img_step,
                                              (float *) sum_data, sum_step,
                                              img_size ));
        }
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }
    __CLEANUP__;
    __END__;
}


CV_IMPL void
cvMultiplyAcc( IplImage * imgA, IplImage * imgB, IplImage * acc, IplImage * mask )
{
    uchar *imgA_data = 0;
    uchar *imgB_data = 0;
    uchar *acc_data = 0;
    uchar *mask_data = 0;
    int imgA_step = 0;
    int imgB_step = 0;
    int acc_step = 0;
    int mask_step = 0;
    CvSize imgA_size;
    CvSize imgB_size;
    CvSize acc_size;
    CvSize mask_size;
    int imgBtPix;

    CV_FUNCNAME( "cvMultiplyAcc" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( imgA ));
    CV_CALL( CV_CHECK_IMAGE( imgB ));
    CV_CALL( CV_CHECK_IMAGE( acc ));
    if( mask )
        CV_CALL( CV_CHECK_MASK_IMAGE( mask ));

    if( acc->depth != IPL_DEPTH_32F )
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );

    if( imgA->depth != imgB->depth )
        CV_ERROR( IPL_BadDepth, "input images have different depths" );

    if( imgA->nChannels != acc->nChannels || imgB->nChannels != acc->nChannels )
        CV_ERROR( IPL_BadNumChannels, "images have different number of channels" );

    cvGetImageRawData( imgA, &imgA_data, &imgA_step, &imgA_size );
    cvGetImageRawData( imgB, &imgB_data, &imgB_step, &imgB_size );
    cvGetImageRawData( acc, &acc_data, &acc_step, &acc_size );

    if( mask )
        cvGetImageRawData( mask, &mask_data, &mask_step, &mask_size );
    else
        mask_size = acc_size;

    if( imgA_size != acc_size || imgA_size != acc_size || mask_size != acc_size )
        CV_ERROR( IPL_StsBadArg, "input images' ROIs have different size" );

    imgA_size.width *= imgA->nChannels;

    imgBtPix = icvGetBtPix( imgA );

    if( imgA_step == imgA_size.width*imgBtPix &&
        imgB_step == imgB_size.width*imgBtPix &&
        acc_step == acc_size.width*sizeof_float &&
        (!mask || mask_step == mask_size.width ))
    {
        imgA_size.width *= imgA_size.height;
        imgA_step *= imgA_size.height;
        imgB_step *= imgA_size.height;
        mask_step *= imgA_size.height;
        acc_step *= imgA_size.height;

        imgA_size.height = 1;
    }

	imgA_size.width *= imgA->nChannels;

    switch (imgA->depth)
    {
    case IPL_DEPTH_8U:
        if( mask )
        {
            IPPI_CALL( icvAddProduct_8u32f_C1IMR( imgA_data, imgA_step,
                                                   imgB_data, imgB_step,
                                                   mask_data, mask_step,
                                                   (float *) acc_data, acc_step,
                                                   imgA_size ));
        }
        else
        {
            IPPI_CALL( icvAddProduct_8u32f_C1IR( imgA_data, imgA_step,
                                                 imgB_data, imgB_step,
                                                 (float *) acc_data, acc_step,
                                                 imgA_size ));
        }
        break;
    case IPL_DEPTH_8S:
        if( mask )
        {
            IPPI_CALL( icvAddProduct_8s32f_C1IMR( (char *) imgA_data, imgA_step,
                                                   (char *) imgB_data, imgB_step,
                                                   mask_data, mask_step,
                                                   (float *) acc_data, acc_step,
                                                   imgA_size ));
        }
        else
        {
            IPPI_CALL( icvAddProduct_8s32f_C1IR( (char *) imgA_data, imgA_step,
                                                 (char *) imgB_data, imgB_step,
                                                 (float *) acc_data, acc_step,
                                                 imgA_size ));
        }
        break;
    case IPL_DEPTH_32F:
        if( mask )
        {
            IPPI_CALL( icvAddProduct_32f_C1IMR( (float *) imgA_data, imgA_step,
                                                 (float *) imgB_data, imgB_step,
                                                 mask_data, mask_step,
                                                 (float *) acc_data, acc_step, imgA_size ));
        }
        else
        {
            IPPI_CALL( icvAddProduct_32f_C1IR( (float *) imgA_data, imgA_step,
                                               (float *) imgB_data, imgB_step,
                                               (float *) acc_data, acc_step, imgA_size ));
        }
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }
    __CLEANUP__;
    __END__;
}


CV_IMPL void
cvRunningAvg( IplImage * imgY, IplImage * imgU, double alpha, IplImage * mask )
{
    uchar *imgY_data = 0;
    uchar *imgU_data = 0;
    uchar *mask_data = 0;
    int imgY_step = 0;
    int imgU_step = 0;
    int mask_step = 0;
    CvSize imgY_size;
    CvSize imgU_size;
    CvSize mask_size;
    int imgBtPix;

    CV_FUNCNAME( "cvRunningAvg" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( imgY ));
    CV_CALL( CV_CHECK_IMAGE( imgU ));

    if( mask )
        CV_CALL( CV_CHECK_MASK_IMAGE( mask ));

    if( imgU->depth != IPL_DEPTH_32F )
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );

    if( imgU->nChannels != imgY->nChannels ||
        (mask && imgU->nChannels != mask->nChannels) )
        CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );

    cvGetImageRawData( imgU, &imgU_data, &imgU_step, &imgU_size );
    cvGetImageRawData( imgY, &imgY_data, &imgY_step, &imgY_size );
    
    if( mask )
        cvGetImageRawData( mask, &mask_data, &mask_step, &mask_size );
    else
        mask_size = imgU_size;

    if( imgU_size != imgY_size || imgU_size != mask_size )
        CV_ERROR( IPL_StsBadArg, "input images' ROIs have different size" );

    imgBtPix = icvGetBtPix( imgY );

    if( imgY_step == imgY_size.width*imgBtPix &&
        imgU_step == imgU_size.width*sizeof_float &&
        (!mask || mask_step == mask_size.width ))
    {
        imgU_size.width *= imgU_size.height;
        imgU_step *= imgU_size.height;
        imgY_step *= imgY_size.height;
        imgU_size.height = 1;
        if( mask )
            mask_step *= mask_size.height;
    }

	imgU_size.width *= imgY->nChannels;

    switch (imgY->depth)
    {
    case IPL_DEPTH_8U:
        if( mask )
        {
            IPPI_CALL( icvAddWeighted_8u32f_C1IMR( imgY_data, imgY_step,
                                                   mask_data, mask_step,
                                                   (float *) imgU_data, imgU_step,
                                                   imgU_size, (float) alpha ));
        }
        else
        {
            IPPI_CALL( icvAddWeighted_8u32f_C1IR( imgY_data, imgY_step,
                                                  (float *) imgU_data, imgU_step,
                                                  imgU_size, (float) alpha ));
        }
        break;
    case IPL_DEPTH_8S:
        if( mask )
        {
            IPPI_CALL( icvAddWeighted_8s32f_C1IMR( (char *) imgY_data, imgY_step,
                                                    mask_data, mask_step,
                                                    (float *) imgU_data, imgU_step,
                                                    imgU_size, (float) alpha ));
        }
        else
        {
            IPPI_CALL( icvAddWeighted_8s32f_C1IR( (char *) imgY_data, imgY_step,
                                                  (float *) imgU_data, imgU_step,
                                                  imgU_size, (float) alpha ));
        }
        break;
    case IPL_DEPTH_32F:
        if( mask )
        {
            IPPI_CALL( icvAddWeighted_32f_C1IMR( (float *) imgY_data, imgY_step,
                                                  mask_data, mask_step,
                                                  (float *) imgU_data, imgU_step,
                                                  imgU_size, (float) alpha ));
        }
        else
        {
            IPPI_CALL( icvAddWeighted_32f_C1IR( (float *) imgY_data, imgY_step,
                                                (float *) imgU_data, imgU_step,
                                                imgU_size, (float) alpha ));
        }
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }
    __CLEANUP__;
    __END__;
}

/* End of file */
