/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/
#include "_cv.h"
#include "_cvwrap.h"
#include "_cvutils.h"

/*F//////////////////////////////////////////////////////////////////////////////////////
//    Name: cvUnDistortOnce
//    Purpose: The function corrects radial and tangential image distortion using known
//             matrix of the camera intrinsic parameters and distortion coefficients
//    Context:
//    Parameters:  srcImage    - source (distorted) image
//                 dstImage    - output (undistorted) image
//                 intrMatrix  - matrix of the camera intrinsic parameters
//                 distCoeffs  - vector of the distortion coefficients (k1, k2, p1 and p2)
//                 interpolate - interpolation toggle (optional parameter)
//
//    Notes:   1. If interpolate = 0, interpolation disabled;
//                else (default) bilinear interpolation is used.
//             2. If p1 = p2 = 0, function acts faster.
//F*/
CV_IMPL void
cvUnDistortOnce( IplImage* srcImage,
                 IplImage* dstImage,
                 float*    intrMatrix,
                 float*    distCoeffs,
                 int       interpolate )
{
    uchar *src_data;
    uchar *dst_data;
    int src_step = 0;
    int dst_step = 0;
    CvSize src_size;
    CvSize dst_size;

    CV_FUNCNAME( "cvUnDistortOnce" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( srcImage ));
    cvGetImageRawData( srcImage, &src_data, &src_step, &src_size );

    if( srcImage->depth != IPL_DEPTH_8U )
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    if( srcImage->nChannels != 1 && srcImage->nChannels != 3 )
        CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );

    CV_CALL( CV_CHECK_IMAGE( dstImage ));
    cvGetImageRawData( dstImage, &dst_data, &dst_step, &dst_size );

    if( dstImage->depth != IPL_DEPTH_8U )
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    if( dstImage->nChannels != 1 && dstImage->nChannels != 3 )
        CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );

    if( src_size != dst_size )
        CV_ERROR( IPL_StsBadArg, "Different sizes of images" );

    if( src_step != dst_step )
        CV_ERROR( IPL_StsBadArg, "Different steps of images" );

    if( srcImage->depth != dstImage->depth )
        CV_ERROR( IPL_StsBadArg, "Different depths of images" );

    if( srcImage->nChannels != dstImage->nChannels )
        CV_ERROR( IPL_StsBadArg, "Different channel numbers of images" );

    if( srcImage->nChannels == 1 )
    {
        if( distCoeffs[2] == 0 && distCoeffs[3] == 0 )
        {
            CV_CALL( icvUnDistort1_8u_C1R( src_data, dst_data, src_step, src_size,
                                            intrMatrix, distCoeffs, interpolate ));
        }
        else
        {
            CV_CALL( icvUnDistortEx_8u_C1R( src_data, dst_data, src_step, src_size,
                                             intrMatrix, distCoeffs, interpolate ));
        }
    }
    else
    {
        if( distCoeffs[2] == 0 && distCoeffs[3] == 0 )
        {
            CV_CALL( icvUnDistort1_8u_C3R( src_data, dst_data, src_step, src_size,
                                            intrMatrix, distCoeffs, interpolate ));
        }
        else
        {
            CV_CALL( icvUnDistortEx_8u_C3R( src_data, dst_data, src_step, src_size,
                                             intrMatrix, distCoeffs, interpolate ));
        }
    }

    __CLEANUP__;
    __END__;
}

/*F//////////////////////////////////////////////////////////////////////////////////////
//    Name: cvUnDistortInit
//    Purpose: The function calculates arrays of distorted points indices and
//             interpolation coefficients for cvUnDistort function using known
//             matrix of the camera intrinsic parameters and distortion coefficients
//    Context:
//    Parameters:  srcImage    - source (distorted) image
//                 intrMatrix  - matrix of the camera intrinsic parameters
//                 distCoeffs  - vector of the distortion coefficients (k1, k2, p1 and p2)
//                 data        - distortion data array
//                 interpolate - interpolation toggle (optional parameter)
//
//    Notes:  1. If interpolate=0, interpolation disabled;
//               else (default) bilinear interpolation is used.
//            2. Array data must be allocated before. If interpolate = 0, its length
//               must be size.width*size.height elements; else 3*size.width*size.height.
//F*/
CV_IMPL void
cvUnDistortInit( IplImage* srcImage,
                 float*    intrMatrix,
                 float*    distCoeffs,
                 int*      data,
                 int       interpolate )
{
    uchar *src_data;
    int src_step = 0;
    CvSize src_size;

    CV_FUNCNAME( "cvUnDistortInit" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( srcImage ));
    cvGetImageRawData( srcImage, &src_data, &src_step, &src_size );

    if( srcImage->depth != IPL_DEPTH_8U )
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    if( srcImage->nChannels != 1 && srcImage->nChannels != 3 )
        CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );

    if( srcImage->nChannels == 1 )
    {
        CV_CALL( icvUnDistortInit_8u_C1R( src_size, src_step, intrMatrix, distCoeffs,
                                           interpolate, data ));
    }
    else
    {
        CV_CALL( icvUnDistortInit_8u_C3R( src_size, src_step, intrMatrix, distCoeffs,
                                           interpolate, data ));
    }

    __CLEANUP__;
    __END__;
}

/*F//////////////////////////////////////////////////////////////////////////////////////
//    Name: cvUnDistort
//    Purpose: The function corrects radial and tangential distortion in the frame
//             using previousely calculated arrays of distorted points indices and
//             undistortion coefficients
//    Context:
//    Parameters:  srcImage    - source (distorted) image
//                 dstImage    - output (undistorted) image
//                 data        - distortion data array
//                 interpolate - interpolation toggle (optional parameter)
//
//    Notes:  1. If interpolate=0, interpolation disabled;
//               else (default) bilinear interpolation is used.
//            2. Array data must be allocated and calculated by the cvUnDistortInit
//               function before. If interpolate = 0, its length must be
//               size.width*size.height elements; else 3*size.width*size.height.
//F*/
CV_IMPL void
cvUnDistort( IplImage* srcImage,
             IplImage* dstImage,
             int*      data,
             int       interpolate )
{
    uchar *src_data;
    uchar *dst_data;
    int src_step = 0;
    int dst_step = 0;
    CvSize src_size;
    CvSize dst_size;

    CV_FUNCNAME( "cvUnDistort" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( srcImage ));
    cvGetImageRawData( srcImage, &src_data, &src_step, &src_size );

    if( srcImage->depth != IPL_DEPTH_8U )
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    if( srcImage->nChannels != 1 && srcImage->nChannels != 3 )
        CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );

    CV_CALL( CV_CHECK_IMAGE( dstImage ));
    cvGetImageRawData( dstImage, &dst_data, &dst_step, &dst_size );

    if( dstImage->depth != IPL_DEPTH_8U )
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    if( dstImage->nChannels != 1 && dstImage->nChannels != 3 )
        CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );

    if( src_size != dst_size )
        CV_ERROR( IPL_StsBadArg, "Different sizes of images" );

    if( src_step != dst_step )
        CV_ERROR( IPL_StsBadArg, "Different steps of images" );

    if( srcImage->depth != dstImage->depth )
        CV_ERROR( IPL_StsBadArg, "Different depths of images" );

    if( srcImage->nChannels != dstImage->nChannels )
        CV_ERROR( IPL_StsBadArg, "Different channel numbers of images" );


    if( srcImage->nChannels == 1 )
    {
        CV_CALL( icvUnDistort_8u_C1R( src_data, dst_data, src_step, src_size,
                                       interpolate, data ));
    }
    else
    {
        CV_CALL( icvUnDistort_8u_C3R( src_data, dst_data, src_step, src_size,
                                       interpolate, data ));
    }

    __CLEANUP__;
    __END__;
}

/*======================================================================================*/

void
cvRodrigues( CvMatr32f       rotMatr,
             CvVect32f       rotVect,
             CvMatr32f       Jacobian,
             CvRodriguesType convType )
{
    CV_FUNCNAME( "cvRodrigues" );

    __BEGIN__;

    IPPI_CALL( icvRodrigues( rotMatr, rotVect, Jacobian, convType ));

    __CLEANUP__;
    __END__;
}

void
cvRodrigues_64d( CvMatr64d rotMatr,
                 CvVect64d rotVect, CvMatr64d Jacobian, CvRodriguesType convType )
{
    CV_FUNCNAME( "cvRodrigues_64d" );

    __BEGIN__;

    IPPI_CALL( icvRodrigues_64d( rotMatr, rotVect, Jacobian, convType ));

    __CLEANUP__;
    __END__;
}


void
cvProjectPoints( int numPoints,
                 CvPoint3D64d * objectPoints,
                 CvVect64d rotVect,
                 CvVect64d transVect,
                 CvVect64d focalLength,
                 CvPoint2D64d principalPoint,
                 CvVect64d distortion,
                 CvPoint2D64d * imagePoints,
                 CvVect64d derivPointsRot,
                 CvVect64d derivPointsTrans,
                 CvVect64d derivPointsFocal,
                 CvVect64d derivPointsPrincipal, CvVect64d derivPointsDistort )
{
    CV_FUNCNAME( "cvProjectPoints" );

    __BEGIN__;

    IPPI_CALL( icvProjectPoints( numPoints, objectPoints, rotVect,
                                 transVect, focalLength, principalPoint,
                                 distortion, imagePoints, derivPointsRot,
                                 derivPointsTrans, derivPointsFocal,
                                 derivPointsPrincipal, derivPointsDistort ));

    __CLEANUP__;
    __END__;
}

/* End of file. */
