/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/
#include "_cv.h"
#include "_cvwrap.h"
#include "_cvimgstat.h"


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    cvMeanShift
//    Purpose: MeanShift algorithm
//    Context:
//    Parameters:
//      imgProb     - 2D object probability distribution
//      windowIn    - CvRect of CAMSHIFT Window intial size
//      numIters    - If CAMSHIFT iterates this many times, stop
//      windowOut   - Location, height and width of converged CAMSHIFT window
//      len         - If != NULL, return equivalent len
//      width       - If != NULL, return equivalent width
//      itersUsed   - Returns number of iterations CAMSHIFT took to converge
//    Returns:
//      The function itself returns the area found
//    Notes:
//F*/
CV_IMPL int
cvMeanShift( IplImage* imgProb, CvRect windowIn,
             CvTermCriteria criteria, CvConnectedComp* comp )
{
    CvSize roi;
    IplROI *saved_roi = 0;
    IplROI new_roi;
    double m00 = 0, m10, m01, inv_m00;
    CvMoments moments;

    int wi2 = windowIn.width / 2;
    int hi2 = windowIn.height / 2;
    int xc = windowIn.x + wi2;
    int yc = windowIn.y + hi2;
    int i = 0, eps;

    CV_FUNCNAME( "cvMeanShift" );

    if( imgProb )
        saved_roi = imgProb->roi;
    if( comp )
        comp->rect = windowIn;

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( imgProb ));
    cvGetImageRawData( imgProb, 0, 0, &roi );

    if( windowIn.height <= 0 || windowIn.width <= 0 )
        CV_ERROR( IPL_StsBadArg, "Input window has non-positive sizes" );

    if( !icvIsRectInRect( windowIn, cvRect( 0, 0, roi.width, roi.height )))
        CV_ERROR( IPL_StsBadArg, "Input window is not inside image ROI" );

    CV_CALL( CV_CHECK_TERM_CRIT( criteria ));

    if( imgProb->roi )
    {
        new_roi.xOffset = imgProb->roi->xOffset;
        new_roi.yOffset = imgProb->roi->yOffset;
    }
    else
    {
        new_roi.xOffset = new_roi.yOffset = 0;
    }

    new_roi = icvMakeROI( new_roi.xOffset + windowIn.x,
                          new_roi.yOffset + windowIn.y,
                          windowIn.width, windowIn.height, icvGetImageCOI( imgProb ));

    imgProb->roi = &new_roi;

    switch (criteria.type)
    {
    case CV_TERMCRIT_ITER:
        criteria.epsilon = 1.f;
        break;
    case CV_TERMCRIT_EPS:
        criteria.maxIter = 100;
        break;
    }

    eps = cvRound( criteria.epsilon * criteria.epsilon );

    for( i = 0; i < criteria.maxIter; i++ )
    {
        int _xc, _yc;

        CV_CALL( cvMoments( imgProb, &moments, 0 ));

        m00 = cvGetSpatialMoment( &moments, 0, 0 );
        m10 = cvGetSpatialMoment( &moments, 1, 0 );
        m01 = cvGetSpatialMoment( &moments, 0, 1 );

        /* Calculating center of mass */
        if( m00 == 0 )
            break;

        inv_m00 = 1. / m00;
        _xc = cvRound( m10 * inv_m00 + xc - wi2 );
        _yc = cvRound( m01 * inv_m00 + yc - hi2 );

        if( _xc < wi2 )
            _xc = wi2;
        else if( _xc + windowIn.width - wi2 > roi.width )
            _xc = roi.width - (windowIn.width - wi2);

        if( _yc < hi2 )
            _yc = hi2;
        else if( _yc + windowIn.height - hi2 > roi.height )
            _yc = roi.height - (windowIn.height - hi2);

        /* Check for coverage centers mass & window */
        if( (xc - _xc) * (xc - _xc) + (yc - _yc) * (yc - _yc) < eps )
            break;

        new_roi.xOffset += _xc - xc;
        new_roi.yOffset += _yc - yc;

        xc = _xc;
        yc = _yc;
    }

    __CLEANUP__;
    __END__;

    if( imgProb )
        imgProb->roi = saved_roi;
    if( comp )
    {
        comp->rect = cvRect( xc - wi2, yc - hi2, windowIn.width, windowIn.height );
        comp->area = (float) m00;
    }

    return i;
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    cvCamShift
//    Purpose: CAMSHIFT algorithm
//    Context:
//    Parameters:
//      imgProb     - 2D object probability distribution
//      windowIn    - CvRect of CAMSHIFT Window intial size
//      criteria    - criteria of stop finding window
//      windowOut   - Location, height and width of converged CAMSHIFT window
//      orientation - If != NULL, return distribution orientation
//      len         - If != NULL, return equivalent len
//      width       - If != NULL, return equivalent width
//      area        - sum of all elements in result window
//      itersUsed   - Returns number of iterations CAMSHIFT took to converge
//    Returns:
//      The function itself returns the area found
//    Notes:
//F*/
CV_IMPL int
cvCamShift( IplImage* imgProb, CvRect windowIn,
            CvTermCriteria criteria,
            CvConnectedComp* _comp,
            CvBox2D* box )
{
    const int TOLERANCE = 10;
    CvMoments moments;
    CvSize roi;
    IplROI *saved_roi = 0;
    IplROI new_roi;
    double m00 = 0, m10, m01, m20, m11, m02, inv_m00;
    double a, b, c, xc, yc;
    double rotate_a, rotate_c;
    double theta = 0, square;
    double cs, sn;
    double length = 0, width = 0;
    int itersUsed = 0;
    CvConnectedComp comp;

    CV_FUNCNAME( "cvCamShift" );

    if( imgProb )
        saved_roi = imgProb->roi;

    comp.rect = windowIn;

    __BEGIN__;

    CV_CALL( itersUsed = cvMeanShift( imgProb, windowIn, criteria, &comp ));
    windowIn = comp.rect;

    cvGetImageRawData( imgProb, 0, 0, &roi );

    windowIn.x -= TOLERANCE;
    if( windowIn.x < 0 )
        windowIn.x = 0;

    windowIn.y -= TOLERANCE;
    if( windowIn.y < 0 )
        windowIn.y = 0;

    windowIn.width += 2 * TOLERANCE;
    if( windowIn.x + windowIn.width > roi.width )
        windowIn.width = roi.width - windowIn.x;

    windowIn.height += 2 * TOLERANCE;
    if( windowIn.y + windowIn.height > roi.height )
        windowIn.height = roi.height - windowIn.y;

    if( imgProb->roi )
    {
        new_roi.xOffset = imgProb->roi->xOffset;
        new_roi.yOffset = imgProb->roi->yOffset;
    }
    else
    {
        new_roi.xOffset = new_roi.yOffset = 0;
    }

    new_roi = icvMakeROI( new_roi.xOffset + windowIn.x,
                          new_roi.yOffset + windowIn.y,
                          windowIn.width, windowIn.height, icvGetImageCOI( imgProb ));

    imgProb->roi = &new_roi;

    /* Calculating moments in new center mass */
    CV_CALL( cvMoments( imgProb, &moments, 0 ));
    imgProb->roi = saved_roi;

    m00 = cvGetSpatialMoment( &moments, 0, 0 );
    m10 = cvGetSpatialMoment( &moments, 1, 0 );
    m01 = cvGetSpatialMoment( &moments, 0, 1 );
    m11 = cvGetCentralMoment( &moments, 1, 1 );
    m20 = cvGetCentralMoment( &moments, 2, 0 );
    m02 = cvGetCentralMoment( &moments, 0, 2 );

    if( m00 == 0 )
        EXIT;

    inv_m00 = 1. / m00;
    xc = cvRound( m10 * inv_m00 + windowIn.x );
    yc = cvRound( m01 * inv_m00 + windowIn.y );
    a = m20 * inv_m00;
    b = m11 * inv_m00;
    c = m02 * inv_m00;

    /* Calculating width & height */
    square = sqrt( 4 * b * b + (a - c) * (a - c) );

    /* Calculating orientation */
    theta = atan2( 2 * b, a - c + square );

    /* Calculating width & lenght of figure */
    cs = cos( theta );
    sn = sin( theta );

    rotate_a = cs * cs * m20 + 2 * cs * sn * m11 + sn * sn * m02;
    rotate_c = sn * sn * m20 - 2 * cs * sn * m11 + cs * cs * m02;
    length = sqrt( rotate_a * inv_m00 ) * 4;
    width = sqrt( rotate_c * inv_m00 ) * 4;

    /* In case, when tetta is 0 or 1.57... the Lenght & Width may be exchanged */
    if( length < width )
    {
        CV_SWAP( length, width, m11 );
        CV_SWAP( cs, sn, m11 );
        theta = IPL_PI_2 - theta;
    }

    /* Saving results */
    if( _comp || box )
    {
        int t0, t1;
        int _xc = cvRound( xc );
        int _yc = cvRound( yc );

        t0 = cvRound( fabs( length * cs ));
        t1 = cvRound( fabs( width * sn ));

        t0 = MAX( t0, t1 ) + 2;
        comp.rect.width = MIN( t0, (roi.width - _xc) * 2 );

        t0 = cvRound( fabs( length * sn ));
        t1 = cvRound( fabs( width * cs ));

        t0 = MAX( t0, t1 ) + 2;
        comp.rect.height = MIN( t0, (roi.height - _yc) * 2 );

        comp.rect.x = MAX( 0, _xc - comp.rect.width / 2 );
        comp.rect.y = MAX( 0, _yc - comp.rect.height / 2 );

        comp.rect.width = MIN( roi.width - comp.rect.x, comp.rect.width );
        comp.rect.height = MIN( roi.height - comp.rect.y, comp.rect.height );
        comp.area = (float) m00;
    }

    __CLEANUP__;
    __END__;

    if( _comp )
        *_comp = comp;
    
    if( box )
    {
        box->size.height = (float) length;
        box->size.width = (float) width;
        box->angle = (float) theta;
        box->center = cvPoint2D32f( comp.rect.x + comp.rect.width*0.5f,
                                    comp.rect.y + comp.rect.height*0.5f);
    }

    return itersUsed;
}

/* End of file. */
