/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include "_cv.h"
#include "_cvconv.h"
#define TG22  0.4142f
#define PUSH2(i,j) {stack[stack_count++] = i;\
                    stack[stack_count++] = j;}
#define POP2(i,j)  {j = stack[--stack_count] ;\
                    i  =stack[--stack_count];}
inline int
icvSuppress( int a, int b, int c )
{
    /* a can be < 0 , but we must use its absolute value */

    /* compare c with b */
    int tmp = (b >= c) ? (b) : (-b);

    /* compare result with a */
    return ( tmp < 0 ) ? tmp : ((tmp >= abs( a )) ? (tmp) : (-tmp));
}


IPCVAPI_IMPL( CvStatus, icvCannyGetSize, (CvSize roi, int *bufferSize) )
{
    if( (roi.width <= 0) && (roi.height <= 0) )
        return CV_BADSIZE_ERR;
    if( !bufferSize )
        return CV_NULLPTR_ERR;
    (*bufferSize) = (roi.height + 2) * (roi.width + 2) * 12;
    return CV_NO_ERR;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:      ippiCanny8uC1R     
//    Purpose:   finds edges on image using J.Canny algorithm
//    Context:   
//    Parameters:
//               src - source image,
//               srcStep - its step,
//               dst - destination binary image with edges,
//               dstStep - its step,
//               roi - size of ROI,
//               opSize - size of Sobel operator aperture, 
//               lowThreshold, 
//               highThreshold - tresholds for hysteresis thresholding  
//                 
//    Returns:   
//    Notes: image gradient magnitude has scale factor 2^(2*opSize-1)
//           so user must choose appropriate lowThreshold and highThreshold 
//           i.e. if real gradient magnitude is 1, then 3x3 Sobel used in this function 
//           will compute 8 //opSize is 2//
//F*/

IPCVAPI_IMPL( CvStatus, icvCanny_16s8u_C1R, (short *pDX,
                                             int dxStep,
                                             short *pDY,
                                             int dyStep,
                                             unsigned char *dst,
                                             int dstStep,
                                             CvSize roi,
                                             float lowThreshold,
                                             float highThreshold, void *Buffer) )
{
    /* lines of dynamic array */
    int mag_width = roi.width + 2;
    int mag_height = roi.height + 2;
    unsigned char *sector = (unsigned char *) Buffer;
    float *magnitude = (float *) Buffer + roi.height * roi.width;
    int *stack = (int *) magnitude + mag_height * mag_width;
    int stack_count = 0;
    unsigned char *tmpdst = dst;
    int i, j;
    float low2 = lowThreshold;
    float high2 = highThreshold;
    int xstep = dxStep / 2;
    int ystep = dyStep / 2;

    /* magnitude image will contain one pixel border */


    /* magnitude and angle images */
    float *local_magnitude = magnitude + mag_width + 1;


    /* sector numbers 

       Top-Left

       2    3     4
       *   *   *
       * * *
       1 ********* 1
       * * *
       *  *   *
       4    3    2

     */


    unsigned char *local_sector = sector;

    /* Some Variables */

    int shift[4];





    /* Check Bad Arguments */
    if( (pDX == NULL) || (pDY == NULL) )
        return CV_NULLPTR_ERR;
    if( (dxStep <= 0) || (dstStep <= 0) )
        return CV_BADSIZE_ERR;
    if( (roi.width <= 0) || (roi.height <= 0) )
        return CV_BADSIZE_ERR;


    /* shift array init */
    shift[0] = 1;
    shift[1] = mag_width + 1;
    shift[2] = mag_width;
    shift[3] = mag_width - 1;


    /* computing orientation and magnitude */
    for( i = 0; i < roi.height; i++ )
    {
        for( j = 0; j < roi.width; j++ )
        {

            int intx = pDX[i * xstep + j];
            int inty = pDY[i * ystep + j];

            float x = (float) (abs( intx ));
            float y = (float) (abs( inty ));

            float mag = x + y;
            unsigned char sec = 0;

            /* estimating sector and magnitude */
            if( mag > low2 )
            {
                int x2 = abs( intx ) << 1;
                float tg22x = (x * TG22);
                float tg67x = (float) x2 + tg22x;
                CvPoint point;

                point.x = j;
                point.y = i;

                if( y > tg67x )
                    sec = 3;
                else if( y < tg22x )
                    sec = 1;
                else
                    sec = (unsigned char) (((intx >> 31) == (inty >> 31)) ? 2 : 4);

                local_magnitude[j] = mag;
                local_sector[j] = sec;
            }
            else
                local_magnitude[j] = 0;

        }
        local_sector += roi.width;
        local_magnitude += mag_width;
    }


/****************************************************************************************\
*                           Nonmaxima suppression                                        *
\****************************************************************************************/
    /* set border magnitude to zero */
    memset( magnitude, 0, mag_width * 4 );
    memset( &magnitude[(mag_height - 1) * mag_width], 0, mag_width * 4 );

    for( i = 0; i < mag_height; i++ )
    {
        magnitude[i * mag_width] = magnitude[i * mag_width + mag_width - 1] = 0;
    }

    /* internal part of image */

    for( i = 0; i < roi.height; i++ )
    {
        /*int last = (i==last_line)?last_element:(line_size-1); */
        for( j = 0; j < roi.width; j++ )
        {
            if( magnitude[(i + 1) * mag_width + j + 1] )
            {
                unsigned char sec = sector[i * roi.width + j];

                float *center = &magnitude[(i + 1) * mag_width + j + 1];
                float val = *center;

                /* shift to unprocessed */
                float b1 = *(center + shift[sec - 1]);

                /* shift to processed */
                float b2 = *(center - shift[sec - 1]);

                if( (val > b1) && (val > fabs( b2 )))
                {
                    if( val > high2 )
                        PUSH2( i, j );
                }
                else
                {
                    *center = -val;
                }
            }
        }
        memset( tmpdst, 0, roi.width );
        tmpdst += dstStep;
    }


/****************************************************************************************\
*                           Hysteresis thresholding                                      *
\****************************************************************************************/

    while( stack_count )
    {
        POP2( i, j );
        if( !dst[i * dstStep + j] )
        {
            if( (magnitude[i * mag_width + j] > low2) && (!dst[(i - 1) * dstStep + j - 1]) )
                PUSH2( i - 1, j - 1 );
            if( (magnitude[i * mag_width + j + 1] > low2) && (!dst[(i - 1) * dstStep + j]) )
                PUSH2( i - 1, j );
            if( (magnitude[i * mag_width + j + 2] > low2) &&
                (!dst[(i - 1) * dstStep + j + 1]) ) PUSH2( i - 1, j + 1 );
            if( (magnitude[(i + 1) * mag_width + j] > low2) && (!dst[i * dstStep + j - 1]) )
                PUSH2( i, j - 1 );
            if( (magnitude[(i + 1) * mag_width + j + 2] > low2) &&
                (!dst[i * dstStep + j + 1]) ) PUSH2( i, j + 1 );
            if( (magnitude[(i + 2) * mag_width + j] > low2) &&
                (!dst[(i + 1) * dstStep + j - 1]) ) PUSH2( i + 1, j - 1 );
            if( (magnitude[(i + 2) * mag_width + j + 1] > low2) &&
                (!dst[(i + 1) * dstStep + j]) ) PUSH2( i + 1, j );
            if( (magnitude[(i + 2) * mag_width + j + 2] > low2) &&
                (!dst[(i + 1) * dstStep + j + 1]) )
                PUSH2( i + 1, j + 1 );
            dst[i * dstStep + j] = 255;
        }
    }

    return CV_NO_ERR;
}

/* End of file. */
