/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/
#include "_cv.h"
#include "_cvwrap.h"
#include "_cvconv.h"
#include "_cvconvert.h"

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    icvFindCornerSubPix8uC1R
//    Purpose:
//      Finds corners on the image with sub-pixel accuracy given 
//      initial guesses for those corners.
//    Context:
//    Parameters:
//      img        - image pointer
//      step       - full width of the image (in bytes)
//      size       - width & height of the image (in pixels)
//      corners    - initial coordinates of corners
//      count      - number of corner points
//      win        - search window size for each corner.
//                   actually, for each corner (x,y), the window
//                   (x - win.width .. x + win.width,y - win.height .. y + win_height)
//                   is used. (window  moves with the point after every iteration)
//      zeroZone   - size of zero zone in the middle of the mask.
//      criteria   - This parameter specifies, how many times iterate and what precision
//                   is required.
//    Returns:
//      Nothing       
//    Notes:
//      Size of destination ROI is not passed into the function, because
//      it assumes dst ROI size:
//      = (src_size.width - 2, src_size.height - 2) if both kernels are used
//      = (src_size.width - 2, src_size.height)     if horizontal kernel != 0 only.
//      = (src_size.width, src_size.height - 2)     if vertical kernel != 0 only.
F*/
CvStatus
icvFindCornerSubPix8uC1R( uchar * img, int step, CvSize size,
                          CvPoint2D32f * corners, int count,
                          CvSize win, CvSize zeroZone, CvTermCriteria criteria )
{
    const int MAX_ITERS = 100;

    static const float drv_x[] = { -1.f, 0.f, 1.f }, drv_y[] =
    {
    0.f, 0.5f, 0.f};

    float local_buffer[1 << 12];
    float *buffer;
    float *maskX;
    float *maskY;
    float *mask;
    float *src_buffer;
    float *gx_buffer;
    float *gy_buffer;
    int win_w = win.width * 2 + 1, win_h = win.height * 2 + 1;
    int win_rect_size = (win_w + 4) * (win_h + 4);
    int buffer_size;
    double coeff;
    CvSize src_buf_size;
    int i, j, k, pt_i;
    int max_iters;
    double eps;

    if( !img || !corners )
        return CV_NULLPTR_ERR;
    if( count <= 0 )
        return CV_BADSIZE_ERR;
    if( win.width <= 0 || win.height <= 0 )
        return CV_BADSIZE_ERR;
    if( size.width < win_w + 4 || size.height < win_h + 4 )
        return CV_BADSIZE_ERR;

    /* initialize variables, controlling loop termination */
    switch (criteria.type)
    {
    case CV_TERMCRIT_ITER:
        eps = 0.f;
        max_iters = criteria.maxIter;
        break;
    case CV_TERMCRIT_EPS:
        eps = criteria.epsilon;
        max_iters = MAX_ITERS;
        break;
    case CV_TERMCRIT_ITER | CV_TERMCRIT_EPS:
        eps = criteria.epsilon;
        max_iters = criteria.maxIter;
        break;
    default:
        assert( 0 );
        return CV_BADFLAG_ERR;
    }

    eps = MAX( eps, 0 );
    eps *= eps;                 /* use square of error in comparsion operations. */

    max_iters = MAX( max_iters, 1 );
    max_iters = MIN( max_iters, MAX_ITERS );

    /* setup buffer */
    buffer_size = (win_rect_size * 5 + win_w + win_h + 32) * sizeof( float );

    buffer = local_buffer;

    if( buffer_size > sizeof( local_buffer ))
    {
        buffer = (float *) icvAlloc( buffer_size );
        if( !buffer )
            return CV_OUTOFMEM_ERR;
    }

    /* assign pointers */
    maskX = buffer;
    maskY = maskX + win_w + 4;
    mask = maskY + win_h + 4;
    src_buffer = mask + win_w * win_h;
    gx_buffer = src_buffer + win_rect_size;
    gy_buffer = gx_buffer + win_rect_size;

    coeff = 1. / (win.width * win.width);

    /* calculate mask */
    for( i = -win.width, k = 0; i <= win.width; i++, k++ )
    {
        maskX[k] = (float) exp( -i * i * coeff );
    }

    if( win.width == win.height )
    {
        maskY = maskX;
    }
    else
    {
        coeff = 1. / (win.height * win.height);
        for( i = -win.height, k = 0; i <= win.height; i++, k++ )
        {
            maskY[k] = (float) exp( -i * i * coeff );
        }
    }

    for( i = 0; i < win_h; i++ )
    {
        for( j = 0; j < win_w; j++ )
        {
            mask[i * win_w + j] = maskX[j] * maskY[i];
        }
    }


    /* make zero_zone */
    if( zeroZone.width >= 0 && zeroZone.height >= 0 &&
        zeroZone.width * 2 + 1 < win_w && zeroZone.height * 2 + 1 < win_h )
    {
        for( i = win.height - zeroZone.height; i <= win.height + zeroZone.height; i++ )
        {
            for( j = win.width - zeroZone.width; j <= win.width + zeroZone.width; j++ )
            {
                mask[i * win_w + j] = 0;
            }
        }
    }

    /* set sizes of image rectangles, used in convolutions */
    src_buf_size.width = win_w + 2;
    src_buf_size.height = win_h + 2;

    /* do optimization loop for all the points */
    for( pt_i = 0; pt_i < count; pt_i++ )
    {
        CvPoint2D32f cT = corners[pt_i], cI = cT;
        int iter = 0;
        double err;

        do
        {
            CvPoint2D32f cI2;
            double a, b, c, bb1, bb2, inv_dt;

            icvGetRectSubPix_8u32f_C1R( img, step, size,
                                         src_buffer, (win_w + 2) * sizeof( src_buffer[0] ),
                                         cvSize( win_w + 2, win_h + 2 ), cI );

            /* calc derivatives */
            icvSepConvSmall3_32f( src_buffer, src_buf_size.width * 4,
                                  gx_buffer, win_w * 4, src_buf_size, drv_x, drv_y, buffer );

            icvSepConvSmall3_32f( src_buffer, src_buf_size.width * 4,
                                  gy_buffer, win_w * 4, src_buf_size, drv_y, drv_x, buffer );

            a = b = c = bb1 = bb2 = 0;

            /* process gradient */
            for( i = 0, k = 0; i < win_h; i++ )
            {
                double py = cI.y + (i - win.height);

                for( j = 0; j < win_w; j++, k++ )
                {
                    double m = mask[k];
                    double tgx = gx_buffer[k];
                    double tgy = gy_buffer[k];
                    double gxx = tgx * tgx * m;
                    double gxy = tgx * tgy * m;
                    double gyy = tgy * tgy * m;
                    double px = cI.x + (j - win.width);

                    a += gxx;
                    b += gxy;
                    c += gyy;

                    bb1 += gxx * px + gxy * py;
                    bb2 += gxy * px + gyy * py;
                }
            }

            inv_dt = 1. / (a * c - b * b);

            assert( a >= 0 && inv_dt >= 0 );

            /* find new approximation */
            cI2.x = (float) ((c * bb1 - b * bb2) * inv_dt);
            cI2.y = (float) ((a * bb2 - b * bb1) * inv_dt);

            /*if( line_feat ) */
            {
                float A[4];
                float V[4];
                float E[4];

                A[0] = (float) a;
                A[1] = A[2] = (float) b;
                A[3] = (float) c;
                icvSVDecomposition_32f( A, 2, 2, E, V );
                assert( E[0] >= E[3] && E[3] >= 0 );

                if( E[0] > E[3] * 30 )
                {
                    double t = (cI.x - cI2.x) * V[2] + (cI.y - cI2.y) * V[3];

                    cI2.x = (float) (cI2.x + t * V[2]);
                    cI2.y = (float) (cI2.y + t * V[3]);
                }
            }

            err = (cI2.x - cI.x) * (cI2.x - cI.x) + (cI2.y - cI.y) * (cI2.y - cI.y);
            cI = cI2;
        }
        while( ++iter < max_iters && err > eps );

        /* if new point is too far from initial, it means poor convergence.
           leave initial point as the result */
        if( fabs( cI.x - cT.x ) > win.width || fabs( cI.y - cT.y ) > win.height )
        {
            cI = cT;
        }

        corners[pt_i] = cI;     /* store result */
    }

    if( buffer != local_buffer )
        icvFree( &buffer );
    return CV_NO_ERR;
}


CV_IMPL void
cvFindCornerSubPix( IplImage * src, CvPoint2D32f * corners,
                    int count, CvSize win, CvSize zeroZone, CvTermCriteria criteria )
{
    uchar *src_data = 0;
    int src_step = 0;
    CvSize src_size;

    CV_FUNCNAME( "cvFindCornerSubPix" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( src ));
    cvGetImageRawData( src, &src_data, &src_step, &src_size );

    if( src->nChannels != 1 )
        CV_ERROR( IPL_StsBadArg, "Function doesn't support multichannel image" );

    switch (src->depth)
    {
    case IPL_DEPTH_8U:
        IPPI_CALL( icvFindCornerSubPix8uC1R( src_data, src_step, src_size,
                                             corners, count, win, zeroZone, criteria ));
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }

    __CLEANUP__;
    __END__;

    return;
}

/* End of file. */
