/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include "_cv.h"
#include "_cvimgstat.h"

/****************************************************************************************\
*                                  CountNonZero                                          *
\****************************************************************************************/

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name: icvCountNonZero
//    Purpose:  Counts non-zero pixels in the image
//    Context:
//    Parameters:  
//      img  - source image
//      step - its step
//      size - its size (width and height)
//      coi  - index of channel (if color image)
//    Returns:
//      number of non-zero pixels or negative value (if error)
//    Notes:
//F*/

/* /////////////// 8uC1 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvCountNonZero_8u_C1R, (const uchar * img,
                                                 int step, CvSize size, int *count) )
{
    int i, j, sum = 0;

    if( !img || !count )
        return CV_NULLPTR_ERR;
    if( size.width < 0 || size.height < 0 || step < size.width )
        return CV_BADSIZE_ERR;

    if( size.width == step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( i = 0; i < size.height; i++, img += step )
    {
        for( j = 0; j < size.width - 3; j += 4 )
            sum += (img[j] != 0) + (img[j + 1] != 0) + (img[j + 2] != 0) + (img[j + 3] != 0);
        for( ; j < size.width; j++ )
            sum += (img[j] != 0);
    }

    *count = sum;
    return CV_OK;
}


/* /////////////// 8uC3 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvCountNonZero_8u_C3CR, (const uchar * img,
                                                  int step, CvSize size, int coi, int *count) )
{
    int i, j, sum = 0;

    size.width *= 3;
    coi--;

    if( !img || !count )
        return CV_NULLPTR_ERR;
    if( size.width < 0 || size.height < 0 || step < size.width )
        return CV_BADSIZE_ERR;
    if( (unsigned) coi > 2 )
        return CV_BADRANGE_ERR;

    if( size.width == step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    img += coi;
    for( i = 0; i < size.height; i++, img += step )
    {
        for( j = 0; j < size.width - 9; j += 12 )
            sum += (img[j] != 0) + (img[j + 3] != 0) + (img[j + 6] != 0) + (img[j + 9] != 0);
        for( ; j < size.width; j += 3 )
            sum += (img[j] != 0);
    }

    *count = sum;
    return CV_OK;
}


/* /////////////// 32fC1 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvCountNonZero_32f_C1R, (const float *img,
                                                  int step, CvSize size, int *count) )
{
    int i, j, sum = 0;
    int *imgi = (int *) img;

    if( !img || !count )
        return CV_NULLPTR_ERR;
    if( size.width < 0 || size.height < 0 ||
        step < size.width * sizeof_float || (step & (sizeof_float - 1)) != 0 )
        return CV_BADSIZE_ERR;

    step /= sizeof_float;

    if( size.width == step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( i = 0; i < size.height; i++, imgi += step )
    {
        for( j = 0; j < size.width - 3; j += 4 )
            sum += (imgi[j] * 2 != 0) + (imgi[j + 1] * 2 != 0) +
                (imgi[j + 2] * 2 != 0) + (imgi[j + 3] * 2 != 0);
        for( ; j < size.width; j++ )
            sum += (imgi[j] * 2 != 0);
    }

    *count = sum;
    return CV_OK;
}

/* /////////////// 32fC3 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvCountNonZero_32f_C3CR, (const float *img,
                                                   int step,
                                                   CvSize size, int coi, int *count) )
{
    int i, j, sum = 0;
    int *imgi = (int *) img;

    size.width *= 3;
    coi--;

    if( !img || !count )
        return CV_NULLPTR_ERR;
    if( size.width < 0 || size.height < 0 ||
        step < size.width * sizeof_float || (step & (sizeof_float - 1)) != 0 )
        return CV_BADSIZE_ERR;
    if( (unsigned) coi > 2 )
        return CV_BADRANGE_ERR;

    step /= sizeof_float;
    imgi += coi;

    if( size.width == step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( i = 0; i < size.height; i++, imgi += step )
    {
        for( j = 0; j < size.width - 9; j += 12 )
            sum += (imgi[j] * 2 != 0) + (imgi[j + 3] * 2 != 0) +
                (imgi[j + 6] * 2 != 0) + (imgi[j + 9] * 2 != 0);
        for( ; j < size.width; j += 3 )
            sum += (imgi[j] * 2 != 0);
    }

    *count = sum;
    return CV_OK;
}

/*  End of file  */
