/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/
#include "_cv.h"
#include "_cvwrap.h"

CV_IMPL void
cvCanny( IplImage * src, IplImage * dst,
         double low_thresh, double high_thresh, int aperture_size )
{
    uchar *src_data = 0;
    int src_step = 0;
    uchar *dst_data = 0;
    int dst_step = 0;
    _CvConvState *pX;
    _CvConvState *pY;
    short *pDx = NULL;
    short *pDy = NULL;
    void *Buffer;
    int buf_size;
    CvSize src_size;
    CvSize dst_size;


    CV_FUNCNAME( "cvCanny" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( src ));
    cvGetImageRawData( src, &src_data, &src_step, &src_size );

    CV_CALL( CV_CHECK_IMAGE( dst ));
    cvGetImageRawData( dst, &dst_data, &dst_step, &dst_size );

    if( dst->depth != IPL_DEPTH_8U )
        CV_ERROR( IPL_StsBadArg, "dst  has not appropriate formats" );

    if( (src->nChannels != 1) || (dst->nChannels != 1) )
        CV_ERROR( IPL_StsBadArg, "Function doesn't support multichannel image" );

    if( src_size != dst_size )
        CV_ERROR( IPL_StsBadArg, "src and dst have different sizes" );


    switch (src->depth)
    {
    case IPL_DEPTH_8U:
        pDx = (short *) icvAlloc( src_size.height * src_size.width * sizeof( short ));
        pDy = (short *) icvAlloc( src_size.height * src_size.width * sizeof( short ));

        IPPI_CALL( icvCannyGetSize( src_size, &buf_size ));
        Buffer = icvAlloc( buf_size );

        IPPI_CALL( icvSobelInitAlloc
                   ( src_size.width, cv8u, aperture_size, src->origin, 1, 0, &pX ));
        IPPI_CALL( icvSobelInitAlloc
                   ( src_size.width, cv8u, aperture_size, src->origin, 0, 1, &pY ));

        IPPI_CALL( icvSobel_8u16s_C1R
                   ( src_data, src_step, pDx, src_size.width * sizeof( short ), &src_size, pX,

                     0 ));
        IPPI_CALL( icvSobel_8u16s_C1R
                   ( src_data, src_step, pDy, src_size.width * sizeof( short ), &src_size, pY,

                     0 ));
        IPPI_CALL( icvConvolFree( &pX ));
        IPPI_CALL( icvConvolFree( &pY ));


        IPPI_CALL( icvCanny_16s8u_C1R( pDx, src_size.width * sizeof( short ),
                                       pDy, src_size.width * sizeof( short ),
                                       (uchar *) dst_data, dst_step,
                                       src_size,(float) low_thresh, (float) high_thresh, Buffer ));
        icvFree( &pDx );
        icvFree( &pDy );
        icvFree( &Buffer );
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }

    __CLEANUP__;
    __END__;
    return;
}


CV_IMPL void
cvCornerMinEigenVal( IplImage * src, IplImage * eigenv, int block_size, int aperture_size )
{
    uchar *src_data = 0;
    uchar *eigenv_data = 0;
    int src_step = 0;
    int eigenv_step = 0;
    CvSize src_size;
    CvSize eigenv_size;
    int buf_size;
    void *buffer = 0;

    CV_FUNCNAME( "cvCornerMinEigenVal" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( src ));
    cvGetImageRawData( src, &src_data, &src_step, &src_size );
    CV_CALL( CV_CHECK_IMAGE( eigenv ));
    cvGetImageRawData( eigenv, &eigenv_data, &eigenv_step, &eigenv_size );

    if( src->nChannels != 1 )
        CV_ERROR( IPL_StsBadArg, "Function doesn't support multichannel source" );

    if( eigenv->nChannels != 1 )
        CV_ERROR( IPL_BadNumChannels, "Dst has more than 1 channel" );

    if( eigenv->depth != IPL_DEPTH_32F )
        CV_ERROR( IPL_BadDepth, "Dst has not appropriate depth (IPL_DEPTH_32F)" );

    IPPI_CALL( icvMinEigenValGetSize( src_size.width, aperture_size, block_size, &buf_size ));
    buffer = icvAlloc( buf_size );

    switch (src->depth)
    {
    case IPL_DEPTH_8U:
        IPPI_CALL( icvMinEigenVal_8u32f_C1R( src_data, src_step,
                                             (float*)eigenv_data, eigenv_step,
                                             src_size, aperture_size,
                                             block_size, buffer ));
        break;

    case IPL_DEPTH_8S:
        IPPI_CALL( icvMinEigenVal_8s32f_C1R( (char*)src_data, src_step,
                                             (float*)eigenv_data, eigenv_step,
                                             src_size, aperture_size,
                                             block_size, buffer ));
        break;

    case IPL_DEPTH_32F:
        IPPI_CALL( icvMinEigenVal_32f_C1R( (float*)src_data, src_step,
                                           (float*)eigenv_data, eigenv_step,
                                           src_size, aperture_size,
                                           block_size, buffer ));
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }

    __CLEANUP__;
    __END__;

    icvFree( &buffer );
}


CV_IMPL void
cvCornerEigenValsAndVecs( IplImage * src,
                          IplImage * eigenvv, int block_size, int aperture_size )
{
    uchar *src_data = 0;
    uchar *eigenvv_data = 0;
    int src_step = 0;
    int eigenvv_step = 0;
    CvSize src_size;
    CvSize eigenvv_size;
    int buf_size;
    void *buffer = 0;

    CV_FUNCNAME( "cvCornerEigenValsAndVecs" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( src ));
    cvGetImageRawData( src, &src_data, &src_step, &src_size );
    CV_CALL( CV_CHECK_IMAGE( eigenvv ));
    cvGetImageRawData( eigenvv, &eigenvv_data, &eigenvv_step, &eigenvv_size );

    if( src->nChannels != 1 )
        CV_ERROR( IPL_StsBadArg, "Function doesn't support multichannel source image" );
    if( (eigenvv->nChannels != 1) && (eigenvv->nChannels != 6) )
        CV_ERROR( IPL_StsBadArg, "Dst has not appropriate number of channels (1 or 6)" );
    if( eigenvv->depth != IPL_DEPTH_32F )
        CV_ERROR( IPL_StsBadArg, "Dst has not appropriate depth (IPL_DEPTH_32F)" );

    IPPI_CALL( icvEigenValsVecsGetSize( src_size.width, aperture_size, block_size, &buf_size ));
    buffer = icvAlloc( buf_size );

    switch (src->depth)
    {
    case IPL_DEPTH_8U:
        IPPI_CALL( icvEigenValsVecs_8u32f_C1R( src_data, src_step,
                                               (float*)eigenvv_data, eigenvv_step,
                                               src_size, aperture_size,
                                               block_size, buffer ));
        break;

    case IPL_DEPTH_8S:
        IPPI_CALL( icvEigenValsVecs_8s32f_C1R( (char*)src_data, src_step,
                                               (float*)eigenvv_data, eigenvv_step,
                                               src_size, aperture_size,
                                               block_size, buffer ));
        break;

    case IPL_DEPTH_32F:
        IPPI_CALL( icvEigenValsVecs_32f_C1R( (float*)src_data, src_step,
                                             (float*)eigenvv_data, eigenvv_step,
                                             src_size, aperture_size,
                                             block_size, buffer ));
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }

    __CLEANUP__;
    __END__;

    icvFree( &buffer );
}

CV_IMPL void
cvPreCornerDetect( IplImage * src, IplImage * dst, int aperture_size )
{
    uchar *src_data = 0;
    int src_step = 0;
    uchar *dst_data = 0;
    int dst_step = 0;
    CvSize src_size;
    CvSize dst_size;

    CV_FUNCNAME( "cvPreCornerDetect" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( src ));
    cvGetImageRawData( src, &src_data, &src_step, &src_size );

    CV_CALL( CV_CHECK_IMAGE( dst ));
    cvGetImageRawData( dst, &dst_data, &dst_step, &dst_size );

    if( dst->depth != IPL_DEPTH_32F )
        CV_ERROR( IPL_StsBadArg, "dst  has not appropriate formats" );

    if( (src->nChannels != 1) || (dst->nChannels != 1) )
        CV_ERROR( IPL_StsBadArg, "Function doesn't support multichannel image" );

    if( src_size != dst_size )
        CV_ERROR( IPL_StsBadArg, "src and dst have different sizes" );


    switch (src->depth)
    {
    case IPL_DEPTH_8U:
        IPPI_CALL( icvPreCornerDetect8uC1R( src_data, src_step,
                                             (float *) dst_data, dst_step, src_size,
                                             aperture_size ));
        break;
    case IPL_DEPTH_8S:
        IPPI_CALL( icvPreCornerDetect8sC1R( (char *) src_data, src_step,
                                             (float *) dst_data, dst_step, src_size,
                                             aperture_size ));
        break;
    case IPL_DEPTH_32F:
        IPPI_CALL( icvPreCornerDetect32fC1R( (float *) src_data, src_step,
                                              (float *) dst_data, dst_step, src_size,
                                              aperture_size ));
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }

    __CLEANUP__;
    __END__;

    return;
}
