/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/
#include "_cv.h"
#include "_cvwrap.h"
#include "_cvutils.h"


#define  cmp_features( f1, f2 )  (*(f1) > *(f2))

_CV_IMPLEMENT_QSORT( icvSortFeatures, int *, cmp_features );

void
cvGoodFeaturesToTrack( IplImage * image,
                       IplImage * eigImage,
                       IplImage * tempImage,
                       CvPoint2D32f * corners,
                       int *corner_count, double quality_level, double min_distance )
{
    double max_val = 0;
    int max_count = 0;
    int count = 0;
    int x, y, i, k = 0;
    int min_dist;
    CvSize sz = { 0, 0 };

    /* when selecting points, use integer coordinates */
    CvPoint *ptr = (CvPoint *) corners;

    /* process floating-point images using integer arithmetics */
    int *eig_data = 0;
    int *eig_data0 = 0;
    int *tmp_data = 0;
    int **ptr_data = 0;
    int eig_step = 0, tmp_step = 0;
    uchar *mask_data = 0;
    int mask_step = 0;

    IplROI *eig_roi = 0, *tmp_roi = 0;
    IplROI roi;
    IplImage min_img, max_img;

    CV_FUNCNAME( "cvGoodFeaturesToTrack" );

    if( eigImage && tempImage )
    {
        eig_roi = eigImage->roi;
        tmp_roi = tempImage->roi;
    }

    __BEGIN__;

    CV_CALL( CV_CHECK_MASK_IMAGE( image ));
    CV_CALL( CV_CHECK_IMAGE( tempImage ));
    CV_CALL( CV_CHECK_IMAGE( eigImage ));

    if( tempImage->nChannels != 1 || eigImage->nChannels != 1 )
        CV_ERROR( IPL_BadNumChannels, "unsupported number of channels" );

    if( tempImage->depth != IPL_DEPTH_32F || eigImage->depth != IPL_DEPTH_32F )
        CV_ERROR( IPL_BadDepth, "unsupported depth" );

    if( corner_count )
    {
        max_count = *corner_count;
        *corner_count = 0;
    }

    if( !corners || !corner_count )
        CV_ERROR( IPL_StsBadArg, "corner or corner_count pointers are NULL" );

    if( max_count <= 0 )
        CV_ERROR( IPL_StsBadArg, "corner count is non positive" );

    if( quality_level <= 0 || min_distance < 0 )
        CV_ERROR( IPL_StsBadArg, "quality level or min distance are non positive" );

    cvGetImageRawData( image, 0, 0, &sz );

    /* setup ROIs */
    roi.coi = roi.xOffset = roi.yOffset = 0;
    roi.width = sz.width;
    roi.height = sz.height;

    eigImage->roi = tempImage->roi = &roi;

    cvGetImageRawData( eigImage, (uchar **) & eig_data, &eig_step, 0 );
    cvGetImageRawData( tempImage, (uchar **) & tmp_data, &tmp_step, 0 );

    CV_CALL( cvInitImageHeader( &min_img, sz, image->depth, 1, image->origin, 4, 1 ));
    CV_CALL( cvInitImageHeader( &max_img, sz, image->depth, 1, image->origin, 4, 1 ));

    cvSetImageData( &min_img, tempImage->imageData, (sz.width + 3) & -4 );
    cvSetImageData( &max_img, min_img.imageData + min_img.imageSize, (sz.width + 3) & -4 );

    cvGetImageRawData( &max_img, &mask_data, &mask_step, 0 );

    /* determine min & max neighbor for every pixel and difference between them */
    /*CV_CALL( cvErode( image, &min_img, 0, 1 ));
       CV_CALL( cvDilate( image, &max_img, 0, 1 ));
       CV_CALL( iplSubtract( &max_img, &min_img, &max_img )); */

    CV_CALL( cvCornerMinEigenVal( image, eigImage, 3, 3 ))
        CV_CALL( cvMinMaxLoc( eigImage, 0, &max_val, 0, 0 ))
        CV_CALL( cvThreshold( eigImage, eigImage, max_val * quality_level,
                              0, CV_THRESH_TOZERO ))
        CV_CALL( cvDilate( eigImage, tempImage, 0, 1 ));

    min_dist = cvRound( min_distance * min_distance );

    eig_step /= 4;
    tmp_step /= 4;

    eig_data0 = eig_data;
    ptr_data = (int **) tmp_data;

    /* collect list of pointers to features - put them into temporary image */
    for( y = 1, k = 0; y < sz.height - 1; y++ )
    {
        eig_data += eig_step;
        tmp_data += tmp_step;
        mask_data += mask_step;

        for( x = 1; x < sz.width - 1; x++ )
        {
            int val = eig_data[x];

            if( val != 0 && val == tmp_data[x] )
                ptr_data[k++] = eig_data + x;
        }
    }

    icvSortFeatures( ptr_data, k, 0 );

    /* select the most strong features */
    for( i = 0; i < k; i++ )
    {
        int j = count, ofs = ptr_data[i] - eig_data0;
        int y = ofs / eig_step;
        int x = ofs - y * eig_step;

        if( min_dist != 0 )
        {
            for( j = 0; j < count; j++ )
            {
                int dx = x - ptr[j].x;
                int dy = y - ptr[j].y;
                int dist = dx * dx + dy * dy;

                if( dist < min_dist )
                    break;
            }
        }

        if( j == count )
        {
            ptr[count].x = x;
            ptr[count].y = y;
            if( ++count >= max_count )
                break;
        }

        if( count >= max_count )
            break;
    }


    /* convert points to floating-point format */
    for( i = 0; i < count; i++ )
    {
        corners[i].x = (float) ptr[i].x;
        corners[i].y = (float) ptr[i].y;
    }

    *corner_count = count;

    __CLEANUP__;

    if( eigImage && tempImage )
    {
        eigImage->roi = eig_roi;
        tempImage->roi = tmp_roi;
    }

    __END__;
}

/* End of file. */
