/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/
#include "_cv.h"
#include "_cvwrap.h"

CV_IMPL void
cvSobel( IplImage * src, IplImage * dst, int dx, int dy, int aperture_size )
{
    uchar *src_data = 0;
    int src_step = 0;
    uchar *dst_data = 0;
    int dst_step = 0;
    CvSize src_size;
    CvSize dst_size;
    _CvConvState *ConS;

    CV_FUNCNAME( "cvSobel" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( src ));
    cvGetImageRawData( src, &src_data, &src_step, &src_size );

    CV_CALL( CV_CHECK_IMAGE( dst ));
    cvGetImageRawData( dst, &dst_data, &dst_step, &dst_size );

    if( ((src->depth == IPL_DEPTH_8U) || (src->depth == IPL_DEPTH_8S)) &&
        (dst->depth != IPL_DEPTH_16S) )
        CV_ERROR( IPL_StsBadArg, "dst  has not appropriate formats" );
    if( (src->depth == IPL_DEPTH_32F) && (dst->depth != IPL_DEPTH_32F) )
        CV_ERROR( IPL_StsBadArg, "dst  has not appropriate formats" );

    if( (src->nChannels != 1) || (dst->nChannels != 1) )
        CV_ERROR( IPL_StsBadArg, "Function doesn't support multichannel image" );

    if( src_size != dst_size )
        CV_ERROR( IPL_StsBadArg, "src and dst have different sizes" );


    switch (src->depth)
    {
    case IPL_DEPTH_8U:
        IPPI_CALL( icvSobelInitAlloc
                   ( src_size.width, cv8u, aperture_size, src->origin, dx, dy, &ConS ));
        IPPI_CALL( icvSobel_8u16s_C1R
                   ( src_data, src_step, (short *) dst_data, dst_step, &src_size, ConS, 0 ));
        IPPI_CALL( icvConvolFree( &ConS ));
        break;

    case IPL_DEPTH_8S:
        IPPI_CALL( icvSobelInitAlloc
                   ( src_size.width, cv8s, aperture_size, src->origin, dx, dy, &ConS ));
        IPPI_CALL( icvSobel_8s16s_C1R
                   ( (char *) src_data, src_step, (short *) dst_data, dst_step, &src_size,
                     ConS, 0 ));
        IPPI_CALL( icvConvolFree( &ConS ));
        break;

    case IPL_DEPTH_32F:
        IPPI_CALL( icvSobelInitAlloc
                   ( src_size.width, cv32f, aperture_size, src->origin, dx, dy, &ConS ));
        IPPI_CALL( icvSobel_32f_C1R
                   ( (float *) src_data, src_step, (float *) dst_data, dst_step, &src_size,
                     ConS, 0 ));
        IPPI_CALL( icvConvolFree( &ConS ));
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }

    __CLEANUP__;
    __END__;


}

CV_IMPL void
cvLaplace( IplImage * src, IplImage * dst, int aperture_size )
{
    uchar *src_data = 0;
    int src_step = 0;
    uchar *dst_data = 0;
    int dst_step = 0;
    _CvConvState *ConS;
    CvSize src_size;
    CvSize dst_size;

    CV_FUNCNAME( "cvLaplace" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( src ));
    cvGetImageRawData( src, &src_data, &src_step, &src_size );

    CV_CALL( CV_CHECK_IMAGE( dst ));
    cvGetImageRawData( dst, &dst_data, &dst_step, &dst_size );

    if( ((src->depth == IPL_DEPTH_8U) || (src->depth == IPL_DEPTH_8S)) &&
        (dst->depth != IPL_DEPTH_16S) )
        CV_ERROR( IPL_StsBadArg, "dst  has not appropriate formats" );
    if( (src->depth == IPL_DEPTH_32F) && (dst->depth != IPL_DEPTH_32F) )
        CV_ERROR( IPL_StsBadArg, "dst  has not appropriate formats" );

    if( (src->nChannels != 1) || (dst->nChannels != 1) )
        CV_ERROR( IPL_StsBadArg, "Function doesn't support multichannel image" );

    if( src_size != dst_size )
        CV_ERROR( IPL_StsBadArg, "src and dst have different sizes" );


    switch (src->depth)
    {
    case IPL_DEPTH_8U:
        IPPI_CALL( icvLaplaceInitAlloc( src_size.width, cv8u, aperture_size, &ConS ));
        IPPI_CALL( icvLaplace_8u16s_C1R( src_data, src_step, (short *) dst_data, dst_step,
                                         &src_size, ConS, 0 ));
        IPPI_CALL( icvConvolFree( &ConS ));
        break;

    case IPL_DEPTH_8S:
        IPPI_CALL( icvLaplaceInitAlloc( src_size.width, cv8u, aperture_size, &ConS ));
        IPPI_CALL( icvLaplace_8s16s_C1R
                   ( (char *) src_data, src_step, (short *) dst_data, dst_step, &src_size,
                     ConS, 0 ));
        IPPI_CALL( icvConvolFree( &ConS ));
        break;

    case IPL_DEPTH_32F:
        IPPI_CALL( icvLaplaceInitAlloc( src_size.width, cv32f, aperture_size, &ConS ));
        IPPI_CALL( icvLaplace_32f_C1R
                   ( (float *) src_data, src_step, (float *) dst_data, dst_step, &src_size,
                     ConS, 0 ));
        IPPI_CALL( icvConvolFree( &ConS ));
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }

    __CLEANUP__;
    __END__;

    return;
}
