/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/
#include "_cv.h"
#include "_cvwrap.h"

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name: cvFloodFill, cvFloodFill_8
//    Purpose: The functions fill the seed pixel environs inside which all pixel
//             values are not far from each other. 
//    Context:
//    Parameters:  img        - initial image (in the beginning)
//                              which is "repainted" during the function action,
//                 initPoint  - coordinates of the seed point inside image ROI,
//                 nv         - new value of repainted area pixels,
//                 lo_diff, up_diff - maximal lower and upper differences of the values of
//                              appurtenant to repainted area pixel and one of its
//                              neighbour,
//                 comp       - pointer to connected component structure of the
//                              repainted area
//
//    Notes:   The 1st function looks for 4-connected environs, the 2nd one - 8-connected.
//F*/

CV_IMPL void
cvFloodFill( IplImage *img, CvPoint seed_point,
             double nv, double lo_diff, double up_diff,
             CvConnectedComp *comp, int connectivity )
{
    uchar *img_data = 0;
    int img_step = 0;
    CvSize img_size;
    
    int bufSize;
    void *buffer = 0;
    bool isSimple;

    CV_FUNCNAME( "cvFloodFill" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( img ));
    cvGetImageRawData( img, &img_data, &img_step, &img_size );
    if( img->nChannels > 1 )
        CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );

    if( connectivity != 4 && connectivity != 8 )
        CV_ERROR_FROM_STATUS( CV_BADFLAG_ERR );

    if( lo_diff == 0 && up_diff == 0 )
    {
        isSimple = true;
        IPPI_CALL( icvFloodFillGetSize( img_size, &bufSize ));
    }
    else
    {
        isSimple = false;
        IPPI_CALL( icvFloodFillGetSize_Grad( img_size, &bufSize ));
    }
    
    buffer = icvAlloc( bufSize );
    if( !buffer )
        CV_ERROR_FROM_STATUS( CV_OUTOFMEM_ERR );

    switch (img->depth)
    {
    case IPL_DEPTH_8U:
        if( isSimple )
        {
            if( connectivity == 4 )
            {
                IPPI_CALL( icvFloodFill_4Con_8u_C1IR( img_data, img_step, img_size,
                                                    seed_point, (int)nv, comp, buffer ));
            }
            else
            {
                IPPI_CALL( icvFloodFill_8Con_8u_C1IR( img_data, img_step, img_size,
                                                    seed_point, (int)nv, comp, buffer ));
            }
        }
        else
        {
            if( connectivity == 4 )
            {
                IPPI_CALL( icvFloodFill_Grad4Con_8u_C1IR( img_data, img_step, img_size,
                                                        seed_point, (int)nv,
                                                        (int)lo_diff, (int)up_diff,
                                                        comp, buffer ));
            }
            else
            {
                IPPI_CALL( icvFloodFill_Grad8Con_8u_C1IR( img_data, img_step, img_size,
                                                        seed_point, (int)nv,
                                                        (int)lo_diff, (int)up_diff,
                                                        comp, buffer ));
            }
        }
        break;
    case IPL_DEPTH_32F:
        if( isSimple )
        {
            if( connectivity == 4 )
            {
                IPPI_CALL( icvFloodFill_4Con_32f_C1IR((float*)img_data, img_step, img_size,
                                                    seed_point, (float)nv, comp, buffer));
            }
            else
            {
                IPPI_CALL( icvFloodFill_8Con_32f_C1IR((float*)img_data, img_step, img_size,
                                                    seed_point, (float)nv, comp, buffer));
            }
        }
        else
        {
            if( connectivity == 4 )
            {
                IPPI_CALL( icvFloodFill_Grad4Con_32f_C1IR((float*)img_data, img_step,
                                                        img_size, seed_point, (float)nv,
                                                        (float)lo_diff, (float)up_diff,
                                                        comp, buffer ));
            }
            else
            {
                IPPI_CALL( icvFloodFill_Grad8Con_32f_C1IR((float*)img_data, img_step,
                                                        img_size, seed_point, (float)nv,
                                                        (float)lo_diff, (float)up_diff,
                                                        comp, buffer ));
            }
        }
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }


    __CLEANUP__;
    icvFree( &buffer );
    __END__;
}

/*  END  OF   FILE  */
