/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/
#include "_cv.h"

/*F///////////////////////////////////////////////////////////////////////////////////////
//   Name:    icvCopyVector_64d
//   Purpose:
//     Copies one vector to the another
//   Context:
//   Parameters:
//     Source - source vector 
//     length - length of the vector in elements
//     Dest   - destination vector   
//   Returns:
//     CV_NO_ERR or error code
//   Notes:
//F*/
IPCVAPI_IMPL( CvStatus, icvCopyVector_64d, (CvMatr64d Source, int length, CvMatr64d Dest) )
{
    int i;

    if( !Source )
        return CV_NULLPTR_ERR;
    if( length <= 0 )
        return CV_BADFACTOR_ERR;


    for( i = 0; i < length; i++ )
    {
        Dest[i] = Source[i];
    }
    return CV_NO_ERR;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//   Name:    icvCopyMatrix_32f
//   Purpose:
//     Copies one matrix to the another
//   Context:
//   Parameters:
//     Source - Soudce matrix 
//     width - number of columns in  matrix
//     height - number of rows in new matrix
//     Dest -   destination Matrix   
//   Returns:
//     CV_NO_ERR or error code
//   Notes:
//F*/
IPCVAPI_IMPL( CvStatus, icvCopyMatrix_64d, (CvMatr64d Source,
                                            int width, int height, CvMatr64d Dest) )
{
    int i;

    if( !Source )
        return CV_NULLPTR_ERR;
    if( width * height <= 0 )
        return CV_BADFACTOR_ERR;


    for( i = 0; i < width * height; i++ )
    {
        Dest[i] = Source[i];
    }
    return CV_NO_ERR;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//   Name:    icvSetElement_64d
//   Purpose:
//     Sets the element of the matrix
//   Context:
//   Parameters:
//     Matrix - matrix  
//     width - number of columns in  matrix
//     height - number of rows in  matrix
//     x      - column number
//     y      - row number
//     Value  - value to set   
//     Source - pointer to data   
//   Returns:
//     CV_NO_ERR or error code
//   Notes:
//F*/
IPCVAPI_IMPL( CvStatus, icvSetElement_64d, (CvMatr64d Matrix,
                                            int width,
                                            int height, int x, int y, double Value) )
{
    if( !Matrix )
        return CV_NULLPTR_ERR;
    if( (width <= 0) || (height <= 0) )
        return CV_BADFACTOR_ERR;
    if( (x < 0) || (y < 0) )
        return CV_BADFACTOR_ERR;
    Matrix[y * width + x] = Value;
    return CV_NO_ERR;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//   Name:    icvGetElement_64d
//   Purpose:
//     Returns the element of the matrix
//   Context:
//   Parameters:
//     Matrix - matrix  
//     width - number of columns in  matrix
//     height - number of rows in  matrix
//     x      - column number
//     y      - row number
//     Value  - value to set   
//     Source - pointer to data   
//   Returns:
//     CV_NO_ERR or error code
//   Notes:
//F*/
IPCVAPI_IMPL( CvStatus, icvGetElement_64d, (CvMatr64d Matrix,
                                            int width,
                                            int height, int x, int y, double *Value) )
{
    if( !Matrix )
        return CV_NULLPTR_ERR;
    if( (width <= 0) || (height <= 0) )
        return CV_BADFACTOR_ERR;
    if( (x < 0) || (y < 0) )
        return CV_BADFACTOR_ERR;
    *Value = Matrix[y * width + x];
    return CV_NO_ERR;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//   Name:    cvSetZero_64d
//   Purpose:
//     Sets all elements of the Matrix to 0
//   Context:
//   Parameters:
//     Matr  - Matrix to Set 
//     width - number of columns in  matrix
//     height - number of rows in new matrix
//     
//   Returns:
//     CV_NO_ERR or error code
//   Notes:
//F*/
IPCVAPI_IMPL( CvStatus, icvSetZero_64d, (CvMatr64d Matr, int width, int height) )
{
    if( !Matr )
        return CV_NULLPTR_ERR;
    if( (width <= 0) || (height <= 0) )
        return CV_BADFACTOR_ERR;

    memset( Matr, 0, width * height * sizeof( double ));

    return CV_NO_ERR;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//   Name:    cvSetIdentity_64d
//   Purpose:
//     Sets matrix to Identity operator
//   Context:
//   Parameters:
//     Matr  - Matrix to Set 
//     width - number of columns in  matrix
//     height - number of rows in new matrix
//     
//   Returns:
//     CV_NO_ERR or error code
//   Notes:
//F*/
IPCVAPI_IMPL( CvStatus, icvSetIdentity_64d, (CvMatr64d Matr, int width, int height) )
{
    int i, j;
    double *dst = Matr;
    int Step = width;

    if( !Matr )
        return CV_NULLPTR_ERR;
    if( (width <= 0) || (height <= 0) )
        return CV_BADFACTOR_ERR;
    for( j = 0; j < height; j++ )
    {
        for( i = 0; i < width; i++ )
        {
            dst[i] = (i == j) ? 1.0 : 0.0;
        }
        dst += Step;
    }

    return CV_NO_ERR;
}

/* End of File */
