/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/
#include "_cv.h"
#include "_cvwrap.h"
#include <float.h>

CV_IMPL double
cvSumPixels( IplImage * img )
{
    uchar *img_data = 0;
    int img_step = 0;
    CvSize img_size;
    int coi;
    int channels;
    double sum = -DBL_MAX;

    CV_FUNCNAME( "cvSum_" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( img ));
    cvGetImageRawData( img, &img_data, &img_step, &img_size );

    coi = icvGetImageCOI( img );
    channels = img->nChannels;

    if( channels != 1 && (channels != 3 || coi == 0) )
        CV_ERROR( IPL_BadCOI, icvUnsupportedFormat );

    switch (img->depth)
    {
    case IPL_DEPTH_8U:
        switch (channels)
        {
        case 1:
            IPPI_CALL( icvSumPixels_8u_C1R( img_data, img_step, img_size, &sum ));
            break;
        case 3:
            IPPI_CALL( icvSumPixels_8u_C3CR( img_data, img_step, img_size, coi, &sum ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    case IPL_DEPTH_8S:
        switch (channels)
        {
        case 1:
            IPPI_CALL( icvSumPixels_8s_C1R( (char *) img_data, img_step, img_size, &sum ));
            break;
        case 3:
            IPPI_CALL( icvSumPixels_8s_C3CR( (char *) img_data, img_step,
                                              img_size, coi, &sum ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    case IPL_DEPTH_32F:
        switch (channels)
        {
        case 1:
            IPPI_CALL( icvSumPixels_32f_C1R( (float *) img_data, img_step, img_size, &sum ));
            break;
        case 3:
            IPPI_CALL( icvSumPixels_32f_C3CR( (float *) img_data, img_step,
                                               img_size, coi, &sum ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }

    __CLEANUP__;
    __END__;

    return sum;
}


CV_IMPL int
cvCountNonZero( IplImage * img )
{
    uchar *img_data = 0;
    int img_step = 0;
    CvSize img_size;
    int coi;
    int channels;
    int count = -1;

    CV_FUNCNAME( "cvCountNonZero_" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( img ));
    cvGetImageRawData( img, &img_data, &img_step, &img_size );

    coi = icvGetImageCOI( img );
    channels = img->nChannels;

    if( channels != 1 && (channels != 3 || coi == 0) )
        CV_ERROR( IPL_BadCOI, icvUnsupportedFormat );

    switch (img->depth)
    {
    case IPL_DEPTH_8U:
    case IPL_DEPTH_8S:
        switch (channels)
        {
        case 1:
            IPPI_CALL( icvCountNonZero_8u_C1R( img_data, img_step, img_size, &count ));
            break;
        case 3:
            IPPI_CALL( icvCountNonZero_8u_C3CR( img_data, img_step, img_size, coi, &count ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    case IPL_DEPTH_32F:
        switch (channels)
        {
        case 1:
            IPPI_CALL( icvCountNonZero_32f_C1R
                       ( (float *) img_data, img_step, img_size, &count ));
            break;
        case 3:
            IPPI_CALL( icvCountNonZero_32f_C3CR
                       ( (float *) img_data, img_step, img_size, coi, &count ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }

    __CLEANUP__;
    __END__;

    return count;
}


static double
icvMeanMask( IplImage * img, IplImage * mask )
{
    uchar *img_data = 0;
    uchar *mask_data = 0;
    int img_step = 0;
    int mask_step = 0;
    CvSize img_size;
    CvSize mask_size;
    int coi;
    int channels;
    double mean = -DBL_MAX;

    CV_FUNCNAME( "ocvMeanMask" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( img ));
    CV_CALL( CV_CHECK_MASK_IMAGE( mask ));

    cvGetImageRawData( img, &img_data, &img_step, &img_size );
    cvGetImageRawData( mask, &mask_data, &mask_step, &mask_size );

    coi = icvGetImageCOI( img );
    channels = img->nChannels;

    if( channels != 1 && (channels != 3 || coi == 0) )
        CV_ERROR( IPL_BadCOI, icvUnsupportedFormat );

    switch (img->depth)
    {
    case IPL_DEPTH_8U:
        switch (channels)
        {
        case 1:
            IPPI_CALL( icvMean_8u_C1MR( (const uchar *) img_data, img_step,
                                         (const uchar *) mask_data, mask_step,
                                         img_size, &mean ));
            break;
        case 3:
            IPPI_CALL( icvMean_8u_C3CMR( (const uchar *) img_data, img_step,
                                          (const uchar *) mask_data, mask_step,
                                          img_size, coi, &mean ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    case IPL_DEPTH_8S:
        switch (channels)
        {
        case 1:
            IPPI_CALL( icvMean_8s_C1MR( (const char *) img_data, img_step,
                                         (const uchar *) mask_data, mask_step,
                                         img_size, &mean ));
            break;
        case 3:
            IPPI_CALL( icvMean_8s_C3CMR( (const char *) img_data, img_step,
                                          (const uchar *) mask_data, mask_step,
                                          img_size, coi, &mean ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    case IPL_DEPTH_32F:
        switch (channels)
        {
        case 1:
            IPPI_CALL( icvMean_32f_C1MR( (const float *) img_data, img_step,
                                          (const uchar *) mask_data, mask_step,
                                          img_size, &mean ));
            break;
        case 3:
            IPPI_CALL( icvMean_32f_C3CMR( (const float *) img_data, img_step,
                                           (const uchar *) mask_data, mask_step,
                                           img_size, coi, &mean ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }

    __CLEANUP__;
    __END__;

    return mean;
}


CV_IMPL double
cvMean( IplImage * img, IplImage * mask )
{
    if( mask )
        return icvMeanMask( img, mask );

    CvSize img_size;
    double mean = 0;

    CV_FUNCNAME( "cvMean" );

    __BEGIN__;

    CV_CALL( mean = cvSumPixels( img ));
    cvGetImageRawData( img, 0, 0, &img_size );

    if( img_size.width != 0 && img_size.height != 0 )
    {
        mean /= img_size.width * img_size.height;
    }

    __CLEANUP__;
    __END__;

    return mean;
}


static void
icvMean_StdDevMask( IplImage * img, IplImage * mask, double *mean, double *stddev )
{
    uchar *img_data = 0;
    uchar *mask_data = 0;

    int img_step = 0;
    int mask_step = 0;

    CvSize img_size;
    CvSize mask_size;

    int coi;
    int channels;

    CV_FUNCNAME( "_cvMean_StdDevMask" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( img ));
    CV_CALL( CV_CHECK_MASK_IMAGE( mask ));

    cvGetImageRawData( img, &img_data, &img_step, &img_size );
    cvGetImageRawData( mask, &mask_data, &mask_step, &mask_size );

    coi = icvGetImageCOI( img );
    channels = img->nChannels;

    if( channels != 1 && (channels != 3 || coi == 0) )
        CV_ERROR( IPL_BadCOI, icvUnsupportedFormat );

    switch (img->depth)
    {
    case IPL_DEPTH_8U:
        switch (channels)
        {
        case 1:
            IPPI_CALL( icvMean_StdDev_8u_C1MR( (const uchar *) img_data, img_step,
                                                (const uchar *) mask_data, mask_step,
                                                img_size, mean, stddev ));
            break;
        case 3:
            IPPI_CALL( icvMean_StdDev_8u_C3CMR( (const uchar *) img_data, img_step,
                                                 (const uchar *) mask_data, mask_step,
                                                 img_size, coi, mean, stddev ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    case IPL_DEPTH_8S:
        switch (channels)
        {
        case 1:
            IPPI_CALL( icvMean_StdDev_8s_C1MR( (const char *) img_data, img_step,
                                                (const uchar *) mask_data, mask_step,
                                                img_size, mean, stddev ));
            break;
        case 3:
            IPPI_CALL( icvMean_StdDev_8s_C3CMR( (const char *) img_data, img_step,
                                                 (const uchar *) mask_data, mask_step,
                                                 img_size, coi, mean, stddev ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    case IPL_DEPTH_32F:
        switch (channels)
        {
        case 1:
            IPPI_CALL( icvMean_StdDev_32f_C1MR( (const float *) img_data, img_step,
                                                 (const uchar *) mask_data, mask_step,
                                                 img_size, mean, stddev ));
            break;
        case 3:
            IPPI_CALL( icvMean_StdDev_32f_C3CMR( (const float *) img_data, img_step,
                                                  (const uchar *) mask_data, mask_step,
                                                  img_size, coi, mean, stddev ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }

    __CLEANUP__;
    __END__;
}


CV_IMPL void
cvMean_StdDev( IplImage * img, double *mean, double *stddev, IplImage * mask )
{
    if( mask )
    {
        icvMean_StdDevMask( img, mask, mean, stddev );
        return;
    }

    uchar *img_data = 0;
    int img_step = 0;
    CvSize img_size;
    int coi;
    int channels;

    CV_FUNCNAME( "cvMean_StdDev" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( img ));
    cvGetImageRawData( img, &img_data, &img_step, &img_size );

    coi = icvGetImageCOI( img );
    channels = img->nChannels;

    if( channels != 1 && (channels != 3 || coi == 0) )
        CV_ERROR( IPL_BadCOI, icvUnsupportedFormat );

    switch (img->depth)
    {
    case IPL_DEPTH_8U:
        switch (channels)
        {
        case 1:
            IPPI_CALL( icvMean_StdDev_8u_C1R( (const uchar *) img_data, img_step, img_size,
                                               mean, stddev ));
            break;
        case 3:
            IPPI_CALL( icvMean_StdDev_8u_C3CR( (const uchar *) img_data, img_step, img_size,
                                                coi, mean, stddev ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    case IPL_DEPTH_8S:
        switch (channels)
        {
        case 1:
            IPPI_CALL( icvMean_StdDev_8s_C1R( (const char *) img_data, img_step, img_size,
                                               mean, stddev ));
            break;
        case 3:
            IPPI_CALL( icvMean_StdDev_8s_C3CR( (const char *) img_data, img_step, img_size,
                                                coi, mean, stddev ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    case IPL_DEPTH_32F:
        switch (channels)
        {
        case 1:
            IPPI_CALL( icvMean_StdDev_32f_C1R( (const float *) img_data, img_step, img_size,
                                                mean, stddev ));
            break;
        case 3:
            IPPI_CALL( icvMean_StdDev_32f_C3CR( (const float *) img_data, img_step, img_size,
                                                 coi, mean, stddev ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }

    __CLEANUP__;
    __END__;
}



static void
icvMinMaxLocMask( IplImage * img, IplImage * mask,
                  double *_min_val, double *_max_val, CvPoint * min_loc, CvPoint * max_loc )
{
    uchar *img_data = 0;
    uchar *mask_data = 0;

    int img_step = 0;
    int mask_step = 0;

    CvSize img_size;
    CvSize mask_size;

    int coi;
    int channels;
    float min_val = 0, max_val = 0;

    CV_FUNCNAME( "_cvMinMaxLocMask" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( img ));
    CV_CALL( CV_CHECK_MASK_IMAGE( mask ));

    cvGetImageRawData( img, &img_data, &img_step, &img_size );
    cvGetImageRawData( mask, &mask_data, &mask_step, &mask_size );

    coi = icvGetImageCOI( img );
    channels = img->nChannels;

    if( channels != 1 && (channels != 3 || coi == 0) )
        CV_ERROR( IPL_BadCOI, icvUnsupportedFormat );

    switch (img->depth)
    {
    case IPL_DEPTH_8U:
        switch (channels)
        {
        case 1:
            IPPI_CALL( icvMinMaxIndx_8u_C1MR( (const uchar *) img_data, img_step,
                                               (const uchar *) mask_data, mask_step,
                                               img_size, &min_val, &max_val,
                                               min_loc, max_loc ));
            break;
        case 3:
            IPPI_CALL( icvMinMaxIndx_8u_C3CMR( (const uchar *) img_data, img_step,
                                                (const uchar *) mask_data, mask_step,
                                                img_size, coi, &min_val, &max_val,
                                                min_loc, max_loc ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    case IPL_DEPTH_8S:
        switch (channels)
        {
        case 1:
            IPPI_CALL( icvMinMaxIndx_8s_C1MR( (const char *) img_data, img_step,
                                               (const uchar *) mask_data, mask_step,
                                               img_size, &min_val, &max_val,
                                               min_loc, max_loc ));
            break;
        case 3:
            IPPI_CALL( icvMinMaxIndx_8s_C3CMR( (const char *) img_data, img_step,
                                                (const uchar *) mask_data, mask_step,
                                                img_size, coi, &min_val, &max_val,
                                                min_loc, max_loc ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    case IPL_DEPTH_32F:
        switch (channels)
        {
        case 1:
            IPPI_CALL( icvMinMaxIndx_32f_C1MR( (const float *) img_data, img_step,
                                                (const uchar *) mask_data, mask_step,
                                                img_size, &min_val, &max_val,
                                                min_loc, max_loc ));
            break;
        case 3:
            IPPI_CALL( icvMinMaxIndx_32f_C3CMR( (const float *) img_data, img_step,
                                                 (const uchar *) mask_data, mask_step,
                                                 img_size, coi, &min_val, &max_val,
                                                 min_loc, max_loc ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }

    __CLEANUP__;
    __END__;

    if( _min_val )
        *_min_val = (float) min_val;
    if( _max_val )
        *_max_val = (float) max_val;
}


CV_IMPL void
cvMinMaxLoc( IplImage * img, double *_min_val, double *_max_val,
             CvPoint * min_loc, CvPoint * max_loc, IplImage * mask )
{
    if( mask )
    {
        icvMinMaxLocMask( img, mask, _min_val, _max_val, min_loc, max_loc );
        return;
    }

    uchar *img_data = 0;
    int img_step = 0;
    CvSize img_size;
    int coi;
    int channels;
    float min_val = 0, max_val = 0;

    CV_FUNCNAME( "cvMinMaxLoc" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( img ));
    cvGetImageRawData( img, &img_data, &img_step, &img_size );

    coi = icvGetImageCOI( img );
    channels = img->nChannels;

    if( channels != 1 && (channels != 3 || coi == 0) )
        CV_ERROR( IPL_BadCOI, icvUnsupportedFormat );

    switch (img->depth)
    {
    case IPL_DEPTH_8U:
        switch (channels)
        {
        case 1:
            IPPI_CALL( icvMinMaxIndx_8u_C1R( (const uchar *) img_data, img_step, img_size,
                                              &min_val, &max_val, min_loc, max_loc ));
            break;
        case 3:
            IPPI_CALL( icvMinMaxIndx_8u_C3CR( (const uchar *) img_data, img_step, img_size,
                                               coi, &min_val, &max_val, min_loc, max_loc ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    case IPL_DEPTH_8S:
        switch (channels)
        {
        case 1:
            IPPI_CALL( icvMinMaxIndx_8s_C1R( (const char *) img_data, img_step, img_size,
                                              &min_val, &max_val, min_loc, max_loc ));
            break;
        case 3:
            IPPI_CALL( icvMinMaxIndx_8s_C3CR( (const char *) img_data, img_step, img_size,
                                               coi, &min_val, &max_val, min_loc, max_loc ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    case IPL_DEPTH_16S:
        switch (channels)
        {
        case 1:
            IPPI_CALL( icvMinMaxIndx_16s_C1R( (const short *) img_data, img_step, img_size,
                                               &min_val, &max_val, min_loc, max_loc ));
            break;
        case 3:
            IPPI_CALL( icvMinMaxIndx_16s_C3CR( (const short *) img_data, img_step, img_size,
                                                coi, &min_val, &max_val, min_loc, max_loc ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    case IPL_DEPTH_32F:
        switch (channels)
        {
        case 1:
            IPPI_CALL( icvMinMaxIndx_32f_C1R( (const float *) img_data, img_step, img_size,
                                               &min_val, &max_val, min_loc, max_loc ));
            break;
        case 3:
            IPPI_CALL( icvMinMaxIndx_32f_C3CR( (const float *) img_data, img_step, img_size,
                                                coi, &min_val, &max_val, min_loc, max_loc ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }

    __CLEANUP__;
    __END__;

    if( _min_val )
        *_min_val = (float) min_val;
    if( _max_val )
        *_max_val = (float) max_val;
}


CV_IMPL void
cvMoments( IplImage * img, CvMoments * moments, int binary )
{
    uchar *img_data = 0;
    int img_step = 0;
    CvSize img_size;
    int coi;
    int channels;

    CV_FUNCNAME( "cvMoments" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( img ));
    cvGetImageRawData( img, &img_data, &img_step, &img_size );

    coi = icvGetImageCOI( img );
    channels = img->nChannels;

    if( channels != 1 && (channels != 3 || coi == 0) )
        CV_ERROR( IPL_BadCOI, icvUnsupportedFormat );

    switch (img->depth)
    {
    case IPL_DEPTH_8U:
        switch (channels)
        {
        case 1:
            if( !binary )
            {
                IPPI_CALL( icvMomentsGray_8u_C1R( (const uchar *) img_data, img_step,
                                                   img_size, moments ));
            }
            else
            {
                IPPI_CALL( icvMomentsBinary_8u_C1R( (const uchar *) img_data, img_step,
                                                     img_size, moments ));
            }
            break;
        case 3:
            if( !binary )
            {
                IPPI_CALL( icvMomentsGray_8u_C3CR( (const uchar *) img_data, img_step,
                                                    img_size, coi, moments ));
            }
            else
            {
                IPPI_CALL( icvMomentsBinary_8u_C3CR( (const uchar *) img_data, img_step,
                                                      img_size, coi, moments ));
            }
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    case IPL_DEPTH_8S:
        switch (channels)
        {
        case 1:
            if( !binary )
            {
                IPPI_CALL( icvMomentsGray_8s_C1R( (const char *) img_data, img_step,
                                                   img_size, moments ));
            }
            else
            {
                IPPI_CALL( icvMomentsBinary_8u_C1R( (const uchar *) img_data, img_step,
                                                     img_size, moments ));
            }
            break;
        case 3:
            if( !binary )
            {
                IPPI_CALL( icvMomentsGray_8s_C3CR( (const char *) img_data, img_step,
                                                    img_size, coi, moments ));
            }
            else
            {
                IPPI_CALL( icvMomentsBinary_8u_C3CR( (const uchar *) img_data, img_step,
                                                      img_size, coi, moments ));
            }
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    case IPL_DEPTH_32F:
        switch (channels)
        {
        case 1:
            if( !binary )
            {
                IPPI_CALL( icvMomentsGray_32f_C1R( (const float *) img_data, img_step,
                                                    img_size, moments ));
            }
            else
            {
                IPPI_CALL( icvMomentsBinary_32f_C1R( (const float *) img_data, img_step,
                                                      img_size, moments ));
            }
            break;
        case 3:
            if( !binary )
            {
                IPPI_CALL( icvMomentsGray_32f_C3CR( (const float *) img_data, img_step,
                                                     img_size, coi, moments ));
            }
            else
            {
                IPPI_CALL( icvMomentsBinary_32f_C3CR( (const float *) img_data, img_step,
                                                       img_size, coi, moments ));
            }
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }

    __CLEANUP__;
    __END__;
}


static double
icvNormMask( IplImage * imgA, IplImage * imgB, IplImage * mask, int norm_type )
{
    uchar *imgA_data = 0;
    int imgA_step = 0;
    uchar *imgB_data = 0;
    int imgB_step = 0;
    uchar *mask_data = 0;
    int mask_step = 0;
    CvStatus result;

    CvSize imgA_size;
    CvSize imgB_size;
    CvSize mask_size;
    int coi;
    int channels;
    double norm = -DBL_MAX;

    CV_FUNCNAME( "_cvNormMask" );

    __BEGIN__;

    CV_CALL( CV_CHECK_MASK_IMAGE( mask ));
    cvGetImageRawData( mask, &mask_data, &mask_step, &mask_size );

    CV_CALL( CV_CHECK_IMAGE( imgA ));
    cvGetImageRawData( imgA, &imgA_data, &imgA_step, &imgA_size );
    coi = icvGetImageCOI( imgA );

    if( imgB )
    {
        CV_CALL( CV_CHECK_IMAGE( imgB ));
        cvGetImageRawData( imgB, &imgB_data, &imgB_step, &imgB_size );

        if( imgA->depth != imgB->depth || imgA->nChannels != imgB->nChannels )
            CV_ERROR( IPL_StsBadArg, "imgA and imgB have different formats" );
        if( coi != icvGetImageCOI( imgB ))
            CV_ERROR( IPL_StsBadArg, "imgA and imgB have different COIs" );

        if( imgA_size != imgB_size )
            CV_ERROR( IPL_StsBadArg, "imgA and imgB have different sizes" );
    }

    if( imgA_size != mask_size )
        CV_ERROR( IPL_StsBadArg, "imgA and mask have different sizes" );

    channels = imgA->nChannels;

    if( coi > 0 && channels != 3 )
        CV_ERROR( IPL_BadCOI, icvUnsupportedFormat );

    if( coi == 0 )
    {
        imgA_size.width *= channels;
        imgA_step *= channels;
        imgB_step *= channels;
        channels = 1;
    }

    switch (imgA->depth)
    {
    case IPL_DEPTH_8U:
        switch (channels)
        {
        case 1:
            switch (norm_type)
            {
            case CV_C:
                result = icvNorm_Inf_8u_C1MR( imgA_data, imgA_step,
                                              mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_DIFF_C:
                result = icvNormDiff_Inf_8u_C1MR( imgA_data, imgA_step,
                                                  imgB_data, imgA_step,
                                                  mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_RELATIVE_C:
                result = icvNormRel_Inf_8u_C1MR( imgA_data, imgA_step,
                                                 imgB_data, imgA_step,
                                                 mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_L1:
                result = icvNorm_L1_8u_C1MR( imgA_data, imgA_step,
                                             mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_DIFF_L1:
                result = icvNormDiff_L1_8u_C1MR( imgA_data, imgA_step,
                                                 imgB_data, imgA_step,
                                                 mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_RELATIVE_L1:
                result = icvNormRel_L1_8u_C1MR( imgA_data, imgA_step,
                                                imgB_data, imgA_step,
                                                mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_L2:
                result = icvNorm_L2_8u_C1MR( imgA_data, imgA_step,
                                             mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_DIFF_L2:
                result = icvNormDiff_L2_8u_C1MR( imgA_data, imgA_step,
                                                 imgB_data, imgA_step,
                                                 mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_RELATIVE_L2:
                result = icvNormRel_L2_8u_C1MR( imgA_data, imgA_step,
                                                imgB_data, imgA_step,
                                                mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            default:
                CV_ERROR( IPL_StsInternal, "Unsupported &norm type" );
            }
            break;
        case 3:
            switch (norm_type)
            {
            case CV_C:
                result = icvNorm_Inf_8u_C3CMR( imgA_data, imgA_step,
                                               mask_data, mask_step, imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_DIFF_C:
                result = icvNormDiff_Inf_8u_C3CMR( imgA_data, imgA_step,
                                                   imgB_data, imgA_step,
                                                   mask_data, mask_step,
                                                   imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_RELATIVE_C:
                result = icvNormRel_Inf_8u_C3CMR( imgA_data, imgA_step,
                                                  imgB_data, imgA_step,
                                                  mask_data, mask_step,
                                                  imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_L1:
                result = icvNorm_L1_8u_C3CMR( imgA_data, imgA_step,
                                              mask_data, mask_step, imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_DIFF_L1:
                result = icvNormDiff_L1_8u_C3CMR( imgA_data, imgA_step,
                                                  imgB_data, imgA_step,
                                                  mask_data, mask_step,
                                                  imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_RELATIVE_L1:
                result = icvNormRel_L1_8u_C3CMR( imgA_data, imgA_step,
                                                 imgB_data, imgA_step,
                                                 mask_data, mask_step, imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_L2:
                result = icvNorm_L2_8u_C3CMR( imgA_data, imgA_step,
                                              mask_data, mask_step, imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_DIFF_L2:
                result = icvNormDiff_L2_8u_C3CMR( imgA_data, imgA_step,
                                                  imgB_data, imgA_step,
                                                  mask_data, mask_step,
                                                  imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_RELATIVE_L2:
                result = icvNormRel_L2_8u_C3CMR( imgA_data, imgA_step,
                                                 imgB_data, imgA_step,
                                                 mask_data, mask_step, imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            default:
                CV_ERROR( IPL_StsInternal, "Unsupported &norm type" );
            }
            break;

        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );

        }
        break;
    case IPL_DEPTH_8S:
        switch (channels)
        {
        case 1:
            switch (norm_type)
            {
            case CV_C:
                result = icvNorm_Inf_8s_C1MR( (char *) imgA_data, imgA_step,
                                              mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_DIFF_C:
                result = icvNormDiff_Inf_8s_C1MR( (char *) imgA_data, imgA_step,
                                                  (char *) imgB_data, imgA_step,
                                                  mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_RELATIVE_C:
                result = icvNormRel_Inf_8s_C1MR( (char *) imgA_data, imgA_step,
                                                 (char *) imgB_data, imgA_step,
                                                 mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_L1:
                result = icvNorm_L1_8s_C1MR( (char *) imgA_data, imgA_step,
                                             mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_DIFF_L1:
                result = icvNormDiff_L1_8s_C1MR( (char *) imgA_data, imgA_step,
                                                 (char *) imgB_data, imgA_step,
                                                 mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_RELATIVE_L1:
                result = icvNormRel_L1_8s_C1MR( (char *) imgA_data, imgA_step,
                                                (char *) imgB_data, imgA_step,
                                                mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_L2:
                result = icvNorm_L2_8s_C1MR( (char *) imgA_data, imgA_step,
                                             mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_DIFF_L2:
                result = icvNormDiff_L2_8s_C1MR( (char *) imgA_data, imgA_step,
                                                 (char *) imgB_data, imgA_step,
                                                 mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_RELATIVE_L2:
                result = icvNormRel_L2_8s_C1MR( (char *) imgA_data, imgA_step,
                                                (char *) imgB_data, imgA_step,
                                                mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            default:
                CV_ERROR( IPL_StsInternal, "Unsupported &norm type" );
            }
            break;

        case 3:

            switch (norm_type)
            {
            case CV_C:
                result = icvNorm_Inf_8s_C3CMR( (char *) imgA_data, imgA_step,
                                               mask_data, mask_step, imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_DIFF_C:
                result = icvNormDiff_Inf_8s_C3CMR( (char *) imgA_data, imgA_step,
                                                   (char *) imgB_data, imgA_step,
                                                   mask_data, mask_step,
                                                   imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_RELATIVE_C:
                result = icvNormRel_Inf_8s_C3CMR( (char *) imgA_data, imgA_step,
                                                  (char *) imgB_data, imgA_step,
                                                  mask_data, mask_step,
                                                  imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_L1:
                result = icvNorm_L1_8s_C3CMR( (char *) imgA_data, imgA_step,
                                              mask_data, mask_step, imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_DIFF_L1:
                result = icvNormDiff_L1_8s_C3CMR( (char *) imgA_data, imgA_step,
                                                  (char *) imgB_data, imgA_step,
                                                  mask_data, mask_step,
                                                  imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_RELATIVE_L1:
                result = icvNormRel_L1_8s_C3CMR( (char *) imgA_data, imgA_step,
                                                 (char *) imgB_data, imgA_step,
                                                 mask_data, mask_step, imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_L2:
                result = icvNorm_L2_8s_C3CMR( (char *) imgA_data, imgA_step,
                                              mask_data, mask_step, imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_DIFF_L2:
                result = icvNormDiff_L2_8s_C3CMR( (char *) imgA_data, imgA_step,
                                                  (char *) imgB_data, imgA_step,
                                                  mask_data, mask_step,
                                                  imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_RELATIVE_L2:
                result = icvNormRel_L2_8s_C3CMR( (char *) imgA_data, imgA_step,
                                                 (char *) imgB_data, imgA_step,
                                                 mask_data, mask_step, imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            default:
                CV_ERROR( IPL_StsInternal, "Unsupported &norm type" );
            }
            break;

        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );

        }
        break;
    case IPL_DEPTH_32F:
        switch (channels)
        {
        case 1:
            switch (norm_type)
            {
            case CV_C:
                result = icvNorm_Inf_32f_C1MR( (float *) imgA_data, imgA_step,
                                               mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_DIFF_C:
                result = icvNormDiff_Inf_32f_C1MR( (float *) imgA_data, imgA_step,
                                                   (float *) imgB_data, imgA_step,
                                                   mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_RELATIVE_C:
                result = icvNormRel_Inf_32f_C1MR( (float *) imgA_data, imgA_step,
                                                  (float *) imgB_data, imgA_step,
                                                  mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_L1:
                result = icvNorm_L1_32f_C1MR( (float *) imgA_data, imgA_step,
                                              mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_DIFF_L1:
                result = icvNormDiff_L1_32f_C1MR( (float *) imgA_data, imgA_step,
                                                  (float *) imgB_data, imgA_step,
                                                  mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_RELATIVE_L1:
                result = icvNormRel_L1_32f_C1MR( (float *) imgA_data, imgA_step,
                                                 (float *) imgB_data, imgA_step,
                                                 mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_L2:
                result = icvNorm_L2_32f_C1MR( (float *) imgA_data, imgA_step,
                                              mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_DIFF_L2:
                result = icvNormDiff_L2_32f_C1MR( (float *) imgA_data, imgA_step,
                                                  (float *) imgB_data, imgA_step,
                                                  mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_RELATIVE_L2:
                result = icvNormRel_L2_32f_C1MR( (float *) imgA_data, imgA_step,
                                                 (float *) imgB_data, imgA_step,
                                                 mask_data, mask_step, imgA_size, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            default:
                CV_ERROR( IPL_StsInternal, "Unsupported &norm type" );
            }
            break;

        case 3:

            switch (norm_type)
            {
            case CV_C:
                result = icvNorm_Inf_32f_C3CMR( (float *) imgA_data, imgA_step,
                                                mask_data, mask_step, imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_DIFF_C:
                result = icvNormDiff_Inf_32f_C3CMR( (float *) imgA_data, imgA_step,
                                                    (float *) imgB_data, imgA_step,
                                                    mask_data, mask_step,
                                                    imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_RELATIVE_C:
                result = icvNormRel_Inf_32f_C3CMR( (float *) imgA_data, imgA_step,
                                                   (float *) imgB_data, imgA_step,
                                                   mask_data, mask_step,
                                                   imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_L1:
                result = icvNorm_L1_32f_C3CMR( (float *) imgA_data, imgA_step,
                                               mask_data, mask_step, imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_DIFF_L1:
                result = icvNormDiff_L1_32f_C3CMR( (float *) imgA_data, imgA_step,
                                                   (float *) imgB_data, imgA_step,
                                                   mask_data, mask_step,
                                                   imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_RELATIVE_L1:
                result = icvNormRel_L1_32f_C3CMR( (float *) imgA_data, imgA_step,
                                                  (float *) imgB_data, imgA_step,
                                                  mask_data, mask_step,
                                                  imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_L2:
                result = icvNorm_L2_32f_C3CMR( (float *) imgA_data, imgA_step,
                                               mask_data, mask_step, imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_DIFF_L2:
                result = icvNormDiff_L2_32f_C3CMR( (float *) imgA_data, imgA_step,
                                                   (float *) imgB_data, imgA_step,
                                                   mask_data, mask_step,
                                                   imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            case CV_RELATIVE_L2:
                result = icvNormRel_L2_32f_C3CMR( (float *) imgA_data, imgA_step,
                                                  (float *) imgB_data, imgA_step,
                                                  mask_data, mask_step,
                                                  imgA_size, coi, &norm );
                if( result )
                    CV_ERROR( IPL_StsInternal, icvCvErrMsg );
                break;
            default:
                CV_ERROR( IPL_StsInternal, "Unsupported &norm type" );
            }
            break;

        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );

        }
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }

    __CLEANUP__;
    __END__;

    return norm;
}


CV_IMPL double
cvNorm( IplImage * imgA, IplImage * imgB, int norm_type, IplImage * mask )
{
    if( mask )
        return icvNormMask( imgA, imgB, mask, norm_type );

    uchar *imgA_data = 0;
    int imgA_step = 0;
    uchar *imgB_data = 0;
    int imgB_step = 0;
    CvSize imgA_size;
    CvSize imgB_size;
    int coi;
    int channels;
    double norm = -DBL_MAX;

    CV_FUNCNAME( "cvNorm" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( imgA ));
    cvGetImageRawData( imgA, &imgA_data, &imgA_step, &imgA_size );
    coi = icvGetImageCOI( imgA );

    if( imgB )
    {
        CV_CALL( CV_CHECK_IMAGE( imgB ));
        cvGetImageRawData( imgB, &imgB_data, &imgB_step, &imgB_size );

        if( imgA->depth != imgB->depth || imgA->nChannels != imgB->nChannels )
            CV_ERROR( IPL_StsBadArg, "imgA and imgB have different formats" );
        if( coi != icvGetImageCOI( imgB ))
            CV_ERROR( IPL_StsBadArg, "imgA and imgB have different COIs" );

        if( imgA_size != imgB_size )
            CV_ERROR( IPL_StsBadArg, "imgA and imgB have different sizes" );
    }

    channels = imgA->nChannels;

    if( coi > 0 && channels != 3 )
        CV_ERROR( IPL_BadCOI, icvUnsupportedFormat );

    if( coi == 0 )
    {
        imgA_size.width *= channels;
        imgA_step *= channels;
        imgB_step *= channels;
        channels = 1;
    }

    switch (imgA->depth)
    {
    case IPL_DEPTH_8U:
        switch (channels)
        {
        case 1:
            IPPI_CALL( icvNorm_8u_C1R( (const uchar *) imgA_data, imgA_step,
                                        (const uchar *) imgB_data, imgB_step,
                                        imgA_size, norm_type, &norm ));
            break;
        case 3:
            IPPI_CALL( icvNorm_8u_C3CR( (const uchar *) imgA_data, imgA_step,
                                         (const uchar *) imgB_data, imgB_step,
                                         imgA_size, coi, norm_type, &norm ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    case IPL_DEPTH_8S:
        switch (channels)
        {
        case 1:
            IPPI_CALL( icvNorm_8s_C1R( (const char *) imgA_data, imgA_step,
                                        (const char *) imgB_data, imgB_step,
                                        imgA_size, norm_type, &norm ));
            break;
        case 3:
            IPPI_CALL( icvNorm_8s_C3CR( (const char *) imgA_data, imgA_step,
                                         (const char *) imgB_data, imgB_step,
                                         imgA_size, coi, norm_type, &norm ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    case IPL_DEPTH_32F:
        switch (channels)
        {
        case 1:
            IPPI_CALL( icvNorm_32f_C1R( (const float *) imgA_data, imgA_step,
                                         (const float *) imgB_data, imgB_step,
                                         imgA_size, norm_type, &norm ));
            break;
        case 3:
            IPPI_CALL( icvNorm_32f_C3CR( (const float *) imgA_data, imgA_step,
                                          (const float *) imgB_data, imgB_step,
                                          imgA_size, coi, norm_type, &norm ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }

    __CLEANUP__;
    __END__;

    return norm;
}


/* End of file. */
