/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/
#include "_cv.h"
#include "_cvwrap.h"

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    cvCreateKalman
//    Purpose: Creating CvKalman structure and allocating memory for it
//    Context:
//    Parameters:
//      Kalman        - double pointer to CvKalman structure
//      DynParams     - dimension of the dynamical vector
//      MeasureParams - dimension of the measurement vector
//    Returns:
//    Notes:
//      
//F*/
CV_IMPL CvKalman*
cvCreateKalman( int DynamParams, int MeasureParams )
{
    CvKalman *Kalm = 0;

    CV_FUNCNAME( "cvCreateKalman" );
    __BEGIN__;
    /* allocating memory for the structure */
    Kalm = (CvKalman *) icvAlloc( sizeof( CvKalman ));
    /* Setting struture fields */
    Kalm->DP = DynamParams;
    Kalm->MP = MeasureParams;
    /* allocating memory for the structure fields */
    Kalm->DynamMatr = (float *) icvAlloc( sizeof( float ) * DynamParams * DynamParams );
    Kalm->MeasurementMatr =

        (float *) icvAlloc( sizeof( float ) * MeasureParams * DynamParams );
    Kalm->MNCovariance = (float *) icvAlloc( sizeof( float ) * MeasureParams * DynamParams );
    Kalm->PNCovariance = (float *) icvAlloc( sizeof( float ) * DynamParams * DynamParams );
    Kalm->PriorState = (float *) icvAlloc( sizeof( float ) * DynamParams );
    Kalm->PosterState = (float *) icvAlloc( sizeof( float ) * DynamParams );
    Kalm->PosterErrorCovariance =

        (float *) icvAlloc( sizeof( float ) * DynamParams * DynamParams );
    Kalm->PriorErrorCovariance =

        (float *) icvAlloc( sizeof( float ) * DynamParams * DynamParams );
    Kalm->KalmGainMatr = (float *) icvAlloc( sizeof( float ) * DynamParams * MeasureParams );
    Kalm->Temp1 = (float *) icvAlloc( sizeof( float ) * DynamParams * DynamParams );
    Kalm->Temp2 = (float *) icvAlloc( sizeof( float ) * DynamParams * DynamParams );

    /* Returning created structure */
    __CLEANUP__;
    __END__;

    return Kalm;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    cvReleaseKalman
//    Purpose: Releases CvKalman structure and frees memory allocated for it
//    Context:
//    Parameters:
//      Kalman        - double pointer to CvKalman structure
//    Returns:
//    Notes:
//      
//F*/
CV_IMPL void
cvReleaseKalman( CvKalman ** Kalman )
{
    CvKalman *Kalm;

    CV_FUNCNAME( "cvReleaseKalman" );
    __BEGIN__;
    Kalm = *Kalman;
    /* freeing the memory */
    icvFree( &Kalm->PriorState );
    icvFree( &Kalm->PosterState );
    icvFree( &Kalm->DynamMatr );
    icvFree( &Kalm->MeasurementMatr );
    icvFree( &Kalm->PNCovariance );
    icvFree( &Kalm->MNCovariance );
    icvFree( &Kalm->PosterErrorCovariance );
    icvFree( &Kalm->PriorErrorCovariance );
    icvFree( &Kalm->KalmGainMatr );
    icvFree( &Kalm->Temp1 );
    icvFree( &Kalm->Temp2 );

    /* deallocating the structure */
    icvFree( Kalman );
    Kalman = NULL;
    __CLEANUP__;
    __END__;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    cvKalmanUpdateByTime
//    Purpose: Performing Time update routine for Kalman Filter
//    Context:
//    Parameters:
//      Kalman  - pointer to CvKalman structure
//    Returns:
//    Notes:
//      
//F*/

CV_IMPL void
cvKalmanUpdateByTime( CvKalman * Kalman )
{
    CV_FUNCNAME( "cvCreateKalman" );
    __BEGIN__;
    /*Updating the state */
    icvTransformVector_32f( Kalman->DynamMatr,
                             Kalman->PosterState, Kalman->PriorState, Kalman->DP, Kalman->DP );

    /*Updating the Error Covariances */
    icvMulTransposed_32f( Kalman->PosterErrorCovariance,
                          Kalman->DP,
                          Kalman->DP,
                          Kalman->DynamMatr, Kalman->DP, Kalman->DP, Kalman->Temp1 );
    icvMulMatrix_32f( Kalman->DynamMatr,
                      Kalman->DP,
                      Kalman->DP,
                      Kalman->Temp1, Kalman->DP, Kalman->DP, Kalman->PriorErrorCovariance );
    icvAddMatrix_32f( Kalman->PriorErrorCovariance,
                      Kalman->PNCovariance,
                      Kalman->PriorErrorCovariance, Kalman->DP, Kalman->DP );
    __CLEANUP__;
    __END__;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    cvKalmanUpdateByTMeasurement
//    Purpose: Performing Measurement update routine for Kalman Filter
//    Context:
//    Parameters:
//    Kalman  - pointer to CvKalman structure
//    Measurement - Current Measurement vector 
//    Returns:
//    Notes:
//      
//F*/
CV_IMPL void
cvKalmanUpdateByMeasurement( CvKalman * Kalman, CvMat * Measurement )
{
    float *Measure;

    CV_FUNCNAME( "cvCreateKalman" );
    __BEGIN__;
    if( (int) Measurement->type > 63 )
        CV_ERROR( IPL_StsBadArg, "source  has not appropriate format" );
    if( (Measurement->cols != 1) || (Measurement->rows != Kalman->MP) )
        CV_ERROR( IPL_StsBadArg, "source  has not appropriate size" );
    Measure = Measurement->data.fl;

    /* Calculating Kalman Gain Matrix */
    icvMulTransposed_32f( Kalman->PriorErrorCovariance,
                          Kalman->DP,
                          Kalman->DP,
                          Kalman->MeasurementMatr,
                          Kalman->DP,
                          Kalman->MP,
                          Kalman->Temp1 );

    icvMulMatrix_32f( Kalman->MeasurementMatr,
                      Kalman->DP,
                      Kalman->MP,
                      Kalman->Temp1, Kalman->MP, Kalman->DP, Kalman->KalmGainMatr );

    icvAddMatrix_32f( Kalman->KalmGainMatr,
                      Kalman->MNCovariance, Kalman->KalmGainMatr, Kalman->MP, Kalman->MP );

    icvInvertMatrix_32f( Kalman->KalmGainMatr, Kalman->MP, Kalman->Temp2 );

    icvMulMatrix_32f( Kalman->Temp1,
                      Kalman->MP,
                      Kalman->DP,
                      Kalman->Temp2, Kalman->MP, Kalman->MP, Kalman->KalmGainMatr );

    /* Update the estimation */
    icvTransformVector_32f( Kalman->MeasurementMatr,
                             Kalman->PriorState, Kalman->Temp1, Kalman->DP, Kalman->MP );
    icvSubVector_32f( Measure, Kalman->Temp1, Kalman->Temp2, Kalman->MP );
    icvTransformVector_32f( Kalman->KalmGainMatr,
                             Kalman->Temp2, Kalman->Temp1, Kalman->MP, Kalman->DP );
    icvAddVector_32f( Kalman->PriorState, Kalman->Temp1, Kalman->PosterState, Kalman->DP );

    /* Update the Error Covariance */
    icvMulMatrix_32f( Kalman->KalmGainMatr,
                      Kalman->MP,
                      Kalman->DP,
                      Kalman->MeasurementMatr,
                      Kalman->DP, Kalman->MP, Kalman->PosterErrorCovariance );
    icvSetIdentity_32f( Kalman->Temp1, Kalman->DP, Kalman->DP );
    icvSubMatrix_32f( Kalman->Temp1,
                      Kalman->PosterErrorCovariance, Kalman->Temp1, Kalman->DP, Kalman->DP );
    icvMulMatrix_32f( Kalman->Temp1,
                      Kalman->DP,
                      Kalman->DP,
                      Kalman->PriorErrorCovariance,
                      Kalman->DP, Kalman->DP, Kalman->PosterErrorCovariance );
    __CLEANUP__;
    __END__;
}
