/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/


#include <math.h>
#include "_cv.h"

#define EPSILON             1.e-4
#define REAL_ZERO(x)        ( (x) < EPSILON && (x) > -EPSILON)

/*=====================================================================================*/

int icvGaussMxN( double *A, double *B, int M, int N, double **solutions );
double icvDet( double *M );
double icvMinor( double *M, int x, int y );

CvStatus
icvClear4( CvMatrix4 * matrix )
{
    if( matrix == 0 )
        return CV_NULLPTR_ERR;

    matrix->m[0][0] = 0;
    matrix->m[0][1] = 0;
    matrix->m[0][2] = 0;
    matrix->m[0][3] = 0;

    matrix->m[1][0] = 0;
    matrix->m[1][1] = 0;
    matrix->m[1][2] = 0;
    matrix->m[1][3] = 0;

    matrix->m[2][0] = 0;
    matrix->m[2][1] = 0;
    matrix->m[2][2] = 0;
    matrix->m[2][3] = 0;

    matrix->m[3][0] = 0;
    matrix->m[3][1] = 0;
    matrix->m[3][2] = 0;
    matrix->m[3][3] = 0;

    return CV_NO_ERR;

}                               /* icvClear4 */


/*=====================================================================================*/

CvStatus
icvClear3( CvMatrix3 * matrix )
{
    if( matrix == 0 )
        return CV_NULLPTR_ERR;

    matrix->m[0][0] = 0;
    matrix->m[0][1] = 0;
    matrix->m[0][2] = 0;
    matrix->m[1][0] = 0;
    matrix->m[1][1] = 0;
    matrix->m[1][2] = 0;
    matrix->m[2][0] = 0;
    matrix->m[2][1] = 0;
    matrix->m[2][2] = 0;

    return CV_NO_ERR;

}                               /* icv_Clear3 */


/*=====================================================================================*/

CvStatus
icvCopyMatrix4( CvMatrix4 * src, CvMatrix4 * dst )
{
    int i;

    if( src == 0 || dst == 0 )
        return CV_NULLPTR_ERR;

    for( i = 0; i < 4; i++ )
    {
        dst->m[0][i] = src->m[0][i];
        dst->m[1][i] = src->m[1][i];
        dst->m[2][i] = src->m[2][i];
        dst->m[3][i] = src->m[3][i];
    }                           /* for */

    return CV_NO_ERR;

}                               /* icvCopyMatrix4 */


/*=====================================================================================*/

CvStatus
icvCopyMatrix3( CvMatrix3 * src, CvMatrix3 * dst )
{
    int i;

    if( src == 0 || dst == 0 )
        return CV_NULLPTR_ERR;

    for( i = 0; i < 3; i++ )
    {
        dst->m[0][i] = src->m[0][i];
        dst->m[1][i] = src->m[1][i];
        dst->m[2][i] = src->m[2][i];
    }                           /* for */

    return CV_NO_ERR;

}                               /* icvCopyMatrix3 */


/*=====================================================================================*/

CvStatus
icvInverseMatrix4( CvMatrix4 * src, CvMatrix4 * dst )
{
    float d;

    /* it would be better to calculate temporary local matrix and then copy it into "dst"...? */
    float w23_01, w23_02, w23_03, w23_12, w23_13, w23_23;
    float w01_01, w01_02, w01_03, w01_12, w01_13, w01_23;

    if( src == 0 || dst == 0 )
        return CV_NULLPTR_ERR;

    w23_12 = src->m[2][1] * src->m[3][2] - src->m[2][2] * src->m[3][1];
    w23_13 = src->m[2][1] * src->m[3][3] - src->m[2][3] * src->m[3][1];
    w23_23 = src->m[2][2] * src->m[3][3] - src->m[2][3] * src->m[3][2];

    w01_12 = src->m[0][1] * src->m[1][2] - src->m[0][2] * src->m[1][1];
    w01_13 = src->m[0][1] * src->m[1][3] - src->m[0][3] * src->m[1][1];
    w01_23 = src->m[0][2] * src->m[1][3] - src->m[0][3] * src->m[1][2];

    dst->m[0][0] = src->m[1][1] * w23_23 - src->m[1][2] * w23_13 + src->m[1][3] * w23_12;

    dst->m[0][1] = src->m[0][2] * w23_13 - src->m[0][1] * w23_23 - src->m[0][3] * w23_12;

    dst->m[0][2] = src->m[3][1] * w01_23 - src->m[3][2] * w01_13 + src->m[3][3] * w01_12;

    dst->m[0][3] = src->m[2][2] * w01_13 - src->m[2][1] * w01_23 - src->m[2][3] * w01_12;

    d = src->m[0][0] * dst->m[0][0] +
        src->m[1][0] * dst->m[0][1] +
        src->m[2][0] * dst->m[0][2] + src->m[3][0] * dst->m[0][3];

    if( d > -EPSILON && d < EPSILON )
    {
        icvCopyMatrix4( src, dst );     /* no invers */
        return CV_BADFACTOR_ERR;        /* error */
    }                           /* if */

    d = 1.0f / d;
    w23_01 = src->m[2][0] * src->m[3][1] - src->m[2][1] * src->m[3][0];
    w23_02 = src->m[2][0] * src->m[3][2] - src->m[2][2] * src->m[3][0];
    w23_03 = src->m[2][0] * src->m[3][3] - src->m[2][3] * src->m[3][0];
    w01_01 = src->m[0][0] * src->m[1][1] - src->m[0][1] * src->m[1][0];
    w01_02 = src->m[0][0] * src->m[1][2] - src->m[0][2] * src->m[1][0];
    w01_03 = src->m[0][0] * src->m[1][3] - src->m[0][3] * src->m[1][0];

    dst->m[0][0] *= d;
    dst->m[0][1] *= d;
    dst->m[0][2] *= d;
    dst->m[0][3] *= d;

    dst->m[1][0] = (src->m[1][2] * w23_03 - src->m[1][0] * w23_23 - src->m[1][3] * w23_02) * d;

    dst->m[1][1] = (src->m[0][0] * w23_23 - src->m[0][2] * w23_03 + src->m[0][3] * w23_02) * d;

    dst->m[1][2] = (src->m[3][2] * w01_03 - src->m[3][0] * w01_23 - src->m[3][3] * w01_02) * d;

    dst->m[1][3] = (src->m[2][0] * w01_23 - src->m[2][2] * w01_03 + src->m[2][3] * w01_02) * d;

    /* for Microsoft Visual 5.0 */
    dst->m[2][0] = (src->m[1][0] * w23_13 -
                    (src->m[1][1] * w23_03) + (src->m[1][3] * w23_01)) * d;
    /* end */

    dst->m[2][1] = (src->m[0][1] * w23_03 - src->m[0][0] * w23_13 - src->m[0][3] * w23_01) * d;

    dst->m[2][2] = (src->m[3][0] * w01_13 - src->m[3][1] * w01_03 + src->m[3][3] * w01_01) * d;

    dst->m[2][3] = (src->m[2][1] * w01_03 - src->m[2][0] * w01_13 - src->m[2][3] * w01_01) * d;

    dst->m[3][0] = (src->m[1][1] * w23_02 - src->m[1][0] * w23_12 - src->m[1][2] * w23_01) * d;

    dst->m[3][1] = (src->m[0][0] * w23_12 - src->m[0][1] * w23_02 + src->m[0][2] * w23_01) * d;

    dst->m[3][2] = (src->m[3][1] * w01_02 - src->m[3][0] * w01_12 - src->m[3][2] * w01_01) * d;

    dst->m[3][3] = (src->m[2][0] * w01_12 - src->m[2][1] * w01_02 + src->m[2][2] * w01_01) * d;

    return CV_NO_ERR;

}                               /* icvInverseMatrix4 */


/*=====================================================================================*/

CvStatus
icvGetNormalVector3( CvMatrix3 * Matrix, float *v )
{
/*  return vector v that is any 3-vector perpendicular
    to all the row vectors of Matrix */

    double *solutions = 0;
    double M[3 * 3];
    double B[3] = { 0.f, 0.f, 0.f };
    int i, j, res;

    if( Matrix == 0 || v == 0 )
        return CV_NULLPTR_ERR;

    for( i = 0; i < 3; i++ )
    {
        for( j = 0; j < 3; j++ )
            M[i * 3 + j] = (double) (Matrix->m[i][j]);
    }                           /* for */

    res = icvGaussMxN( M, B, 3, 3, &solutions );

    if( res == -1 )
        return CV_BADFACTOR_ERR;

    if( res > 0 && solutions )
    {
        v[0] = (float) solutions[0];
        v[1] = (float) solutions[1];
        v[2] = (float) solutions[2];
        res = 0;
    }
    else
        res = 1;

    if( solutions )
        icvFree( &solutions );

    if( res )
        return CV_BADFACTOR_ERR;
    else
        return CV_NO_ERR;

}                               /* icvgetNormalVector3 */


/*=====================================================================================*/

CvStatus
icvGetNormalVector4( CvMatrix4 * Matrix, float *v )
{
/*  return vector v that is any 4-vector perpendicular
    to all the row vectors of Matrix */

    double *solutions = 0;
    double M[3 * 4];
    double B[3] = { 0.f, 0.f, 0.f };
    int i, j, res;

    if( Matrix == 0 || v == 0 )
        return CV_NULLPTR_ERR;

    for( i = 0; i < 3; i++ )
    {
        for( j = 0; j < 4; j++ )
            M[i * 4 + j] = (double) (Matrix->m[i][j]);
    }                           /* for */

    res = icvGaussMxN( M, B, 3, 4, &solutions );

    if( res == -1 )
        return CV_BADFACTOR_ERR;

    if( res > 0 && solutions )
    {
        v[0] = (float) solutions[0];
        v[1] = (float) solutions[1];
        v[2] = (float) solutions[2];
        v[3] = (float) solutions[3];
        res = 0;
    }
    else
        res = 1;

    if( solutions )
        icvFree( &solutions );

    if( res )
        return CV_BADFACTOR_ERR;
    else
        return CV_NO_ERR;

}                               /* icvgetNormalVector4 */


/*=====================================================================================*/

CvStatus
icvMultMatrixVector4( CvMatrix4 * m, float *src, float *dst )
{
    float x, y, z, w;
    float a00, a01, a02, a03;
    float a10, a11, a12, a13;
    float a20, a21, a22, a23;
    float a30, a31, a32, a33;

    if( m == 0 || src == 0 || dst == 0 )
        return CV_NULLPTR_ERR;

    a00 = m->m[0][0];
    a01 = m->m[0][1];
    a02 = m->m[0][2];
    a03 = m->m[0][3];
    a10 = m->m[1][0];
    a11 = m->m[1][1];
    a12 = m->m[1][2];
    a13 = m->m[1][3];
    a20 = m->m[2][0];
    a21 = m->m[2][1];
    a22 = m->m[2][2];
    a23 = m->m[2][3];
    a30 = m->m[3][0];
    a31 = m->m[3][1];
    a32 = m->m[3][2];
    a33 = m->m[3][3];

    x = src[0];
    y = src[1];
    z = src[2];
    w = src[3];

    dst[0] = a00 * x + a01 * y + a02 * z + a03 * w;
    dst[1] = a10 * x + a11 * y + a12 * z + a13 * w;
    dst[2] = a20 * x + a21 * y + a22 * z + a23 * w;
    dst[3] = a30 * x + a31 * y + a32 * z + a33 * w;

    return CV_NO_ERR;

}                               /* icvMultMatrixVector4 */


/*=====================================================================================*/

CvStatus
icvMultMatrixVector3( CvMatrix3 * m, float *src, float *dst )
{
    if( m == 0 || src == 0 || dst == 0 )
        return CV_NULLPTR_ERR;

    dst[0] = m->m[0][0] * src[0] + m->m[0][1] * src[1] + m->m[0][2] * src[2];
    dst[1] = m->m[1][0] * src[0] + m->m[1][1] * src[1] + m->m[1][2] * src[2];
    dst[2] = m->m[2][0] * src[0] + m->m[2][1] * src[1] + m->m[2][2] * src[2];

    return CV_NO_ERR;

}                               /* icvMultMatrixVector3 */


/*=====================================================================================*/

CvStatus
icvMultMatrixTVector3( CvMatrix3 * m, float *src, float *dst )
{
    if( m == 0 || src == 0 || dst == 0 )
        return CV_NULLPTR_ERR;

    dst[0] = m->m[0][0] * src[0] + m->m[1][0] * src[1] + m->m[2][0] * src[2];
    dst[1] = m->m[0][1] * src[0] + m->m[1][1] * src[1] + m->m[2][1] * src[2];
    dst[2] = m->m[0][2] * src[0] + m->m[1][2] * src[1] + m->m[2][2] * src[2];

    return CV_NO_ERR;

}                               /* icvMultMatrixTVector3 */


/*=====================================================================================*/

CvStatus
icvNormalizeVector( float *vector, int dim )
{
    float len = 0.f;
    int i, res;

    if( vector == 0 )
        return CV_NULLPTR_ERR;

    if( dim <= 0 )
        return CV_BADFACTOR_ERR;

    for( i = 0; i < dim; i++ )
        len += vector[i] * vector[i];

    res = (int) REAL_ZERO( len );

    if( !res )
    {

        len = 1.f / (float) sqrt( len );
        for( i = 0; i < dim; i++ )
            vector[i] *= len;
    }                           /* if */

    if( res )
        return CV_BADFACTOR_ERR;
    else
        return CV_NO_ERR;

}                               /* icvnormalizeVector */


/*=====================================================================================*/

CvStatus
icvxVectorOperator( float *vector, CvMatrix3 * dst )
{
    float x, y, z;


    if( vector == 0 && dst == 0 )
        return CV_NULLPTR_ERR;

    x = vector[0];
    y = vector[1];
    z = vector[2];

    dst->m[0][0] = dst->m[1][1] = dst->m[2][2] = 0.f;
    dst->m[0][1] = -z;
    dst->m[0][2] = y;
    dst->m[1][0] = z;
    dst->m[1][2] = -x;
    dst->m[2][0] = -y;
    dst->m[2][1] = x;
    return CV_NO_ERR;

}                               /* icvxVectorOperator */


/*=====================================================================================*/

CvStatus
icvMultMatrix4( CvMatrix4 * src1, CvMatrix4 * src2, CvMatrix4 * dst )
{
    /* Do dst = src1 * src2 */
    int i, j;

    if( src1 == 0 && src2 == 0 && dst == 0 )
        return CV_NULLPTR_ERR;

    for( i = 0; i < 4; i++ )
    {
        for( j = 0; j < 4; j++ )
        {
            dst->m[i][j] = src1->m[i][0] * src2->m[0][j] +
                src1->m[i][1] * src2->m[1][j] +
                src1->m[i][2] * src2->m[2][j] + src1->m[i][3] * src2->m[3][j];
        }                       /* for */
    }                           /* for */

    return CV_NO_ERR;
}                               /* icvMultMatrix4 */


/*=====================================================================================*/

CvStatus
icvMultMatrix3( CvMatrix3 * src1, CvMatrix3 * src2, CvMatrix3 * dst )
{
    /* Do dst = src1 * src2 */
    int i, j;

    if( src1 == 0 && src2 == 0 && dst == 0 )
        return CV_NULLPTR_ERR;

    for( i = 0; i < 3; i++ )
    {
        for( j = 0; j < 3; j++ )
        {
            dst->m[i][j] = src1->m[i][0] * src2->m[0][j] +
                src1->m[i][1] * src2->m[1][j] + src1->m[i][2] * src2->m[2][j];
        }                       /* for */
    }                           /* for */

    return CV_NO_ERR;

}                               /* icvMultMatrix3 */


/*=====================================================================================*/

CvStatus
icvTransposeMatrix4( CvMatrix4 * src )
{
    float tmp;

    if( src == 0 )
        return CV_NULLPTR_ERR;

    tmp = src->m[0][1];
    src->m[0][1] = src->m[1][0];
    src->m[1][0] = tmp;
    tmp = src->m[0][2];
    src->m[0][2] = src->m[2][0];
    src->m[2][0] = tmp;
    tmp = src->m[0][3];
    src->m[0][3] = src->m[3][0];
    src->m[3][0] = tmp;
    tmp = src->m[1][2];
    src->m[1][2] = src->m[2][1];
    src->m[2][1] = tmp;
    tmp = src->m[1][3];
    src->m[1][3] = src->m[3][1];
    src->m[3][1] = tmp;
    tmp = src->m[2][3];
    src->m[2][3] = src->m[3][2];
    src->m[3][2] = tmp;

    return CV_NO_ERR;

}                               /* icvTransposeMatrix4 */


/*=====================================================================================*/

CvStatus
icvTransposeMatrix3( CvMatrix3 * src )
{
    float tmp;

    if( src == 0 )
        return CV_NULLPTR_ERR;

    tmp = src->m[0][1];
    src->m[0][1] = src->m[1][0];
    src->m[1][0] = tmp;
    tmp = src->m[0][2];
    src->m[0][2] = src->m[2][0];
    src->m[2][0] = tmp;
    tmp = src->m[1][2];
    src->m[1][2] = src->m[2][1];
    src->m[2][1] = tmp;

    return CV_NO_ERR;

}                               /* icvTransposeMatrix3 */


/*=====================================================================================*/

CvStatus
icvInverseMatrix3( CvMatrix3 * src, CvMatrix3 * dst )
{
    float d;
    int i;

    if( src == 0 && dst == 0 )
        return CV_NULLPTR_ERR;

    /* it would be better to calculate temporary
       local matrix and then copy it into "dst"...? */

    dst->m[0][0] = src->m[1][1] * src->m[2][2] - src->m[1][2] * src->m[2][1];
    dst->m[1][0] = -src->m[1][0] * src->m[2][2] + src->m[1][2] * src->m[2][0];
    dst->m[2][0] = src->m[1][0] * src->m[2][1] - src->m[1][1] * src->m[2][0];

    d = src->m[0][0] * dst->m[0][0] +
        src->m[0][1] * dst->m[1][0] + src->m[0][2] * dst->m[2][0];

    if( REAL_ZERO( d ))
    {

        for( i = 0; i < 3; i++ )
        {

            dst->m[i][0] = src->m[i][0];
            dst->m[i][1] = src->m[i][1];
            dst->m[i][2] = src->m[i][2];
        }                       /* for */

        return CV_BADFACTOR_ERR;
    }                           /* if */

    d = 1.0f / d;

    dst->m[0][0] *= d;
    dst->m[1][0] *= d;
    dst->m[2][0] *= d;

    dst->m[0][1] = (-src->m[0][1] * src->m[2][2] + src->m[0][2] * src->m[2][1]) * d;

    dst->m[1][1] = (src->m[0][0] * src->m[2][2] - src->m[0][2] * src->m[2][0]) * d;

    dst->m[2][1] = (-src->m[0][0] * src->m[2][1] + src->m[0][1] * src->m[2][0]) * d;

    dst->m[0][2] = (src->m[0][1] * src->m[1][2] - src->m[0][2] * src->m[1][1]) * d;

    dst->m[1][2] = (-src->m[0][0] * src->m[1][2] + src->m[0][2] * src->m[1][0]) * d;

    dst->m[2][2] = (src->m[0][0] * src->m[1][1] - src->m[0][1] * src->m[1][0]) * d;

    return CV_NO_ERR;
}                               /* icvInverseMatrix3 */


/*=====================================================================================*/

CvStatus
icvCrossLines( float *line1, float *line2, float *cross_point )
{
    float delta;

    if( line1 == 0 && line2 == 0 && cross_point == 0 )
        return CV_NULLPTR_ERR;

    delta = line1[0] * line2[1] - line1[1] * line2[0];

    if( REAL_ZERO( delta ))
        return CV_BADFACTOR_ERR;

    cross_point[0] = (-line1[2] * line2[1] + line1[1] * line2[2]) / delta;
    cross_point[1] = (-line1[0] * line2[2] + line1[2] * line2[0]) / delta;
    cross_point[2] = 1;

    return CV_NO_ERR;
}                               /* icvCrossLines */


/*=====================================================================================*/

CvStatus
icvRotate( CvMatrix4 * T, float angle, float x, float y, float z )
{
    CvMatrix4 M, C;
    float x0, y0, z0, mod_u;
    float u_ut[3][3], I_u_ut[3][3], S[3][3];
    float angle_rad;
    int i, j;

    if( T == 0 )
        return CV_NULLPTR_ERR;

    angle_rad = angle * 0.01745329222222f;

    mod_u = x * x + y * y + z * z;

    if( REAL_ZERO( mod_u ))
        return CV_BADFACTOR_ERR;

    mod_u = 1 / (float) sqrt( mod_u );
    x0 = x * mod_u;
    y0 = y * mod_u;
    z0 = z * mod_u;

    S[0][0] = S[1][1] = S[2][2] = 0;

    S[0][1] = -z0;
    S[0][2] = y0;
    S[1][0] = z0;
    S[1][2] = -x0;
    S[2][0] = -y0;
    S[2][1] = x0;


    u_ut[0][0] = x0 * x0;
    u_ut[0][1] = x0 * y0;
    u_ut[0][2] = x0 * z0;
    u_ut[1][0] = y0 * x0;
    u_ut[1][1] = y0 * y0;
    u_ut[1][2] = y0 * z0;
    u_ut[2][0] = z0 * x0;
    u_ut[2][1] = z0 * y0;
    u_ut[2][2] = z0 * z0;

    I_u_ut[0][0] = 1 - x0 * x0;
    I_u_ut[1][1] = 1 - y0 * y0;
    I_u_ut[2][2] = 1 - z0 * z0;

    I_u_ut[0][1] = -y0 * x0;
    I_u_ut[0][2] = -x0 * z0;
    I_u_ut[1][0] = -y0 * x0;
    I_u_ut[1][2] = -y0 * z0;
    I_u_ut[2][0] = -z0 * x0;
    I_u_ut[2][1] = -y0 * z0;

    M.m[3][3] = 1.f;

    M.m[3][0] = M.m[3][1] = M.m[3][2] = 0;
    M.m[0][3] = M.m[1][3] = M.m[2][3] = 0;

    for( i = 0; i < 3; i++ )
    {
        for( j = 0; j < 3; j++ )
        {
            M.m[j][i] = u_ut[i][j] +
                (float) cos( angle_rad ) * I_u_ut[i][j] + (float) sin( angle_rad ) * S[i][j];
        }                       /* for */
    }                           /* for */

    for( i = 0; i < 4; i++ )
    {
        for( j = 0; j < 4; j++ )
            C.m[i][j] = T->m[i][j];
    }                           /* for */

    icvMultMatrix4( &M, &C, T );

    return CV_NO_ERR;

}                               /* icvRotate */


/*=====================================================================================*/

CvStatus
icvTranslate( CvMatrix4 * T, float x, float y, float z )
{
    CvMatrix4 M, C;
    int i, j;

    if( T == 0 )
        return CV_NULLPTR_ERR;

    M.m[0][0] = 1;
    M.m[0][1] = 0;
    M.m[0][2] = 0;
    M.m[0][3] = 0;
    M.m[1][0] = 0;
    M.m[1][1] = 1;
    M.m[1][2] = 0;
    M.m[1][3] = 0;
    M.m[2][0] = 0;
    M.m[2][1] = 0;
    M.m[2][2] = 1;
    M.m[2][3] = 0;
    M.m[3][0] = x;
    M.m[3][1] = y;
    M.m[3][2] = z;
    M.m[3][3] = 1;

    for( i = 0; i < 4; i++ )
    {
        for( j = 0; j < 4; j++ )
            C.m[i][j] = T->m[i][j];
    }                           /* for */

    icvMultMatrix4( &M, &C, T );

    return CV_NO_ERR;

}                               /* icvTranslate */


/*=====================================================================================*/

int
icvGaussMxN( double *A, double *B, int M, int N, double **solutions )
{
    int *variables;
    int row, swapi, i, i_best = 0, j, j_best = 0, t;
    double swapd, ratio, bigest;

    if( !A || !B || !M || !N )
        return -1;

    variables = (int *) icvAlloc( (long) N * sizeof( int ));

    if( variables == 0 )
        return -1;

    for( i = 0; i < N; i++ )
    {
        variables[i] = i;
    }                           /* for */

    /* -----  Direct way  ----- */

    for( row = 0; row < M; row++ )
    {

        bigest = 0;

        for( j = row; j < M; j++ )
        {                       /* search non null element */
            for( i = row; i < N; i++ )
            {

                if( fabs( A[j * N + i] ) > fabs( bigest ))
                {
                    bigest = A[j * N + i];
                    i_best = i;
                    j_best = j;
                }               /* if */
            }                   /* for */
        }                       /* for */

        if( REAL_ZERO( bigest ))
            break;              /* if all shank elements are null */

        if( j_best - row )
        {

            for( t = 0; t < N; t++ )
            {                   /* swap a rows */

                swapd = A[row * N + t];
                A[row * N + t] = A[j_best * N + t];
                A[j_best * N + t] = swapd;
            }                   /* for */

            swapd = B[row];
            B[row] = B[j_best];
            B[j_best] = swapd;
        }                       /* if */

        if( i_best - row )
        {

            for( t = 0; t < M; t++ )
            {                   /* swap a columns  */

                swapd = A[t * N + i_best];
                A[t * N + i_best] = A[t * N + row];
                A[t * N + row] = swapd;
            }                   /* for */

            swapi = variables[row];
            variables[row] = variables[i_best];
            variables[i_best] = swapi;
        }                       /* if */

        for( i = row + 1; i < M; i++ )
        {                       /* recounting A and B */

            ratio = -A[i * N + row] / A[row * N + row];
            B[i] += B[row] * ratio;

            for( j = N - 1; j >= row; j-- )
            {

                A[i * N + j] += A[row * N + j] * ratio;
            }                   /* for */
        }                       /* for */
    }                           /* for */

    if( row < M )
    {                           /* if rank(A)<M */

        for( j = row; j < M; j++ )
        {
            if( !REAL_ZERO( B[j] ))
            {

                icvFree( &variables );
                return -1;      /* if system is antithetic */
            }                   /* if */
        }                       /* for */

        M = row;                /* decreasing size of the task */
    }                           /* if */

    /* ----- Reverse way ----- */

    if( M < N )
    {                           /* if solution are not exclusive */

        *solutions = (double *) icvAlloc( ((N - M + 1) * N) * sizeof( double ));

        if( *solutions == 0 )
        {
            icvFree( &variables );
            return -1;
        }


        for( t = M; t <= N; t++ )
        {
            for( j = M; j < N; j++ )
            {

                (*solutions)[(t - M) * N + variables[j]] = (double) (t == j);
            }                   /* for */

            for( i = M - 1; i >= 0; i-- )
            {                   /* finding component of solution */

                if( t < N )
                {
                    (*solutions)[(t - M) * N + variables[i]] = 0;
                }
                else
                {
                    (*solutions)[(t - M) * N + variables[i]] = B[i] / A[i * N + i];
                }               /* if */

                for( j = i + 1; j < N; j++ )
                {

                    (*solutions)[(t - M) * N + variables[i]] -=
                        (*solutions)[(t - M) * N + variables[j]] * A[i * N + j] / A[i * N + i];
                }               /* for */
            }                   /* for */
        }                       /* for */

        icvFree( &variables );
        return N - M;
    }                           /* if */

    *solutions = (double *) icvAlloc( (N) * sizeof( double ));

    if( solutions == 0 )
        return -1;

    for( i = N - 1; i >= 0; i-- )
    {                           /* finding exclusive solution */

        (*solutions)[variables[i]] = B[i] / A[i * N + i];

        for( j = i + 1; j < N; j++ )
        {

            (*solutions)[variables[i]] -=
                (*solutions)[variables[j]] * A[i * N + j] / A[i * N + i];
        }                       /* for */
    }                           /* for */

    icvFree( &variables );
    return 0;

}                               /* icvGaussMxN */

/*=====================================================================================*/

CvStatus
icvGetCoof( double *f1, double *f2, double *a2, double *a1, double *a0 )
{
    double G[9], a3;
    int i;

    if( !f1 || !f2 || !a0 || !a1 || !a2 )
        return CV_BADFACTOR_ERR;

    for( i = 0; i < 9; i++ )
    {

        G[i] = f1[i] - f2[i];
    }                           /* for */

    a3 = icvDet( G );

    if( REAL_ZERO( a3 ))
        return CV_BADFACTOR_ERR;

    *a2 = 0;
    *a1 = 0;
    *a0 = icvDet( f2 );

    for( i = 0; i < 9; i++ )
    {

        *a2 += f2[i] * icvMinor( G, (int) (i % 3), (int) (i / 3) );
        *a1 += G[i] * icvMinor( f2, (int) (i % 3), (int) (i / 3) );
    }                           /* for */

    *a0 /= a3;
    *a1 /= a3;
    *a2 /= a3;

    return CV_NO_ERR;

}                               /* icvGetCoof */
