/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

/* The following three lines - only for debugging A6 configuration in PII processor */

#include "_cv.h"
#include <float.h>

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       cvsAddMatrix32f   
//    Purpose:    Sum of 2 matrix
//    Context:   
//    Parameters: srcMatr1 - first vector
//                srcMatr2 - second vector
//                destMatr - destination vector
//                matrWidth
//                matrHeight - sizes of matrix       
//               
//    Returns:   
//    Notes:     
//F*/
IPCVAPI_IMPL( CvStatus, icvAddMatrix_32f, (CvMatr32f srcMatr1,
                                           CvMatr32f srcMatr2,
                                           CvMatr32f destMatr, int matrWidth, int matrHeight) )
{
    int step, i, j;
    float *src1;
    float *src2;
    float *dst;

    /* check bad arguments */
    if( srcMatr1 == NULL )
        return CV_NULLPTR_ERR;
    if( destMatr == NULL )
        return CV_NULLPTR_ERR;
    if( srcMatr2 == NULL )
        return CV_NULLPTR_ERR;
    if( matrWidth <= 0 )
        return CV_BADSIZE_ERR;
    if( matrHeight <= 0 )
        return CV_BADSIZE_ERR;

    step = matrWidth;

    src1 = srcMatr1;
    src2 = srcMatr2;
    dst = destMatr;

    for( i = 0; i < matrHeight; i++ )
    {
        for( j = 0; j < matrWidth; j++ )
            dst[j] = src1[j] + src2[j];
        dst += step;
        src2 += step;
        src1 += step;
    }
    return CV_NO_ERR;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       cvSubMatrix_32f   
//    Purpose:    Difference of 2 matrix
//    Context:   
//    Parameters: srcMatr1 - first matrix
//                srcMatr2 - second matrix(subtrahend)
//                destMatr - destination matrix
//                matrWidth
//                matrHeight - sizes of matrix       
//               
//    Returns:   
//    Notes:     
//F*/
IPCVAPI_IMPL( CvStatus, icvSubMatrix_32f, (CvMatr32f srcMatr1,
                                           CvMatr32f srcMatr2,
                                           CvMatr32f destMatr, int matrWidth, int matrHeight) )
{
    int step, i, j;
    float *src1;
    float *src2;
    float *dst;

    /* check bad arguments */
    if( srcMatr1 == NULL )
        return CV_NULLPTR_ERR;
    if( destMatr == NULL )
        return CV_NULLPTR_ERR;
    if( srcMatr2 == NULL )
        return CV_NULLPTR_ERR;
    if( matrWidth <= 0 )
        return CV_BADSIZE_ERR;
    if( matrHeight <= 0 )
        return CV_BADSIZE_ERR;

    step = matrWidth;

    src1 = srcMatr1;
    src2 = srcMatr2;
    dst = destMatr;

    for( i = 0; i < matrHeight; i++ )
    {
        for( j = 0; j < matrWidth; j++ )
        {
            dst[j] = src1[j] - src2[j];
        }
        dst += step;
        src2 += step;
        src1 += step;
    }
    return CV_NO_ERR;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       cvScaleMatrix_32f   
//    Purpose:    Multiplies the Matrix by a value
//    Context:   
//    Parameters: srcMatr -  source  matrix
//                destMatr - destination matrix
//                matrWidth
//                matrHeight - sizes of matrix
//                value - factor       
//               
//    Returns:   
//    Notes:     
//F*/
IPCVAPI_IMPL( CvStatus, icvScaleMatrix_32f, (CvMatr32f srcMatr,
                                             CvMatr32f destMatr,
                                             int matrWidth, int matrHeight, float value) )
{
    int step, i, j;
    float *src;
    float *dst;

    /* check bad arguments */
    if( srcMatr == NULL )
        return CV_NULLPTR_ERR;
    if( destMatr == NULL )
        return CV_NULLPTR_ERR;
    if( matrWidth <= 0 )
        return CV_BADSIZE_ERR;
    if( matrHeight <= 0 )
        return CV_BADSIZE_ERR;

    step = matrWidth;

    src = srcMatr;
    dst = destMatr;

    for( i = 0; i < matrHeight; i++ )
    {
        for( j = 0; j < matrWidth; j++ )
        {
            dst[j] = value * src[j];
        }
        dst += step;
        src += step;
    }
    return CV_NO_ERR;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       cvMulMatrix_32f   
//    Purpose:    Product of 2 matrix
//    Context:   
//    Parameters: srcMatr1 - first matrix
//                matrWidth1
//                matrHeight1 - sizes of srcMatr1
//                srcMatr2 - second matrix
//                matrWidth2
//                matrHeight2 - sizes of srcMatr2       
//
//                destMatr - destination matrix
//               
//    Returns:
//    CV_NO_ERROR or error code     
//    Notes:
//    matrWidth1 mus be equal to matrHeight2      
//F*/
IPCVAPI_IMPL( CvStatus, icvMulMatrix_32f, (CvMatr32f srcMatr1,
                                           int matrWidth1,
                                           int matrHeight1,
                                           CvMatr32f srcMatr2,
                                           int matrWidth2,
                                           int matrHeight2, CvMatr32f destMatr) )
{
    int i, j, k;
    float sum = 0;
    float *first = srcMatr1;
    float *second = srcMatr2;
    float *dest = destMatr;

    /* check bad arguments */
    if( srcMatr1 == NULL )
        return CV_NULLPTR_ERR;
    if( destMatr == NULL )
        return CV_NULLPTR_ERR;
    if( srcMatr2 == NULL )
        return CV_NULLPTR_ERR;
    if( matrWidth1 <= 0 )
        return CV_BADSIZE_ERR;
    if( matrHeight1 <= 0 )
        return CV_BADSIZE_ERR;
    if( matrWidth2 <= 0 )
        return CV_BADSIZE_ERR;
    if( matrHeight2 <= 0 )
        return CV_BADSIZE_ERR;
    if( matrHeight2 != matrWidth1 )
        return CV_BADSIZE_ERR;
    for( j = 0; j < matrHeight1; j++ )
    {
        for( i = 0; i < matrWidth2; i++ )
        {
            sum = 0;
            second = srcMatr2 + i;
            for( k = 0; k < matrWidth1; k++ )
            {
                sum += first[k] * (*second);
                second += matrWidth2;
            }
            dest[i] = sum;
        }
        first += matrWidth1;
        dest += matrWidth2;
    }
    return CV_NO_ERR;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       ippmMulTransposed_32f   
//    Purpose:    Product of 2 matrix
//    Context:   
//    Parameters: srcMatr1 - first matrix
//                matrWidth1
//                matrHeight1 - sizes of srcMatr1
//                srcMatr2 - second matrix
//                matrWidth2
//                matrHeight2 - sizes of srcMatr2       
//
//                destMatr - destination matrix
//               
//    Returns:
//    IPP_NO_ERROR or error code     
//    Notes:
//    matrWidth1 mus be equal to matrHeight2      
//F*/
IPCVAPI_IMPL( CvStatus, icvMulTransposed_32f, (CvMatr32f srcMatr1,
                                               int matrWidth1,
                                               int matrHeight1,
                                               CvMatr32f srcMatr2,
                                               int matrWidth2,
                                               int matrHeight2, CvMatr32f destMatr) )
{
    int i, j;

    for( i = 0; i < matrHeight1; i++ )
    {
        for( j = 0; j < matrHeight2; j++ )
        {
            icvDotProduct_32f( srcMatr1 + i * matrWidth1,
                                srcMatr2 + j * matrWidth2,
                                matrWidth1, destMatr + i * matrHeight2 + j );
        }
    }
    return CV_NO_ERR;
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       cvMulTransMatrixL_32f, cvMulTransMatrixR_32f
//    Purpose:    Multiply a matrix by transposed itself
//    Context:   
//    Parameters: srcMatr (matrHeight, matrWidth) - source matrix
//                                (matrHeight - rows number, matrWidth - columns number)
//                destMatr(matrHeight, matrWidth) - destination matrix
//    Returns:
//    CV_NO_ERROR or error code     
//    Notes: 1) The function cvMulTransMatrixL_32f calculates product A At,
//              where A - source matrix, At - its transposition.
//              The function cvMulTransMatrixR_32f calculates product At A.
//           2) Performs more quickly then ordinary multiplication.
//F*/
IPCVAPI_IMPL( CvStatus, icvMulTransMatrixL_32f, (CvMatr32f srcMatr,
                                                 int matrWidth,
                                                 int matrHeight, CvMatr32f destMatr) )
{
    float temp;
    int i, j;

    if( srcMatr == NULL || destMatr == NULL || srcMatr == destMatr )
        return CV_NULLPTR_ERR;
    if( matrHeight <= 0 || matrWidth <= 0 )
        return CV_BADSIZE_ERR;


    for( i = 0; i < matrHeight; i++ )
    {
        for( j = 0; j <= i; j++ )
        {
            icvDotProduct_32f( srcMatr + j * matrWidth,
                                srcMatr + i * matrWidth, matrWidth, &temp );
            destMatr[i * matrHeight + j] = temp;
            destMatr[j * matrHeight + i] = temp;
        }

    }
    return CV_NO_ERR;
}

/*--------------------------------------------------------------------------------------*/
IPCVAPI_IMPL( CvStatus, icvMulTransMatrixR_32f, (CvMatr32f srcMatr,
                                                 int matrWidth,
                                                 int matrHeight, CvMatr32f destMatr) )
{
    int i, j;
    float *C = destMatr, *col_i;

    if( srcMatr == NULL || C == NULL || srcMatr == C )
        return CV_NULLPTR_ERR;
    if( matrHeight <= 0 || matrWidth <= 0 )
        return CV_BADSIZE_ERR;

    col_i = (float *) icvAlloc( 4 * matrHeight );

    for( i = 0; i < matrWidth; i++, C += matrWidth )
    {
        for( j = 0; j < matrHeight; j++ )
            col_i[j] = srcMatr[j * matrWidth + i];
        for( j = i; j < matrWidth; j++ )
        {
            int st = matrWidth, k = 0;
            float s = 0.0f, *A = srcMatr;

            for( ; k < matrHeight; k++, A += st )
                s += col_i[k] * A[j];
            C[j] = s;
        }
    }
    icvFree( (void **) &col_i );
    for( i = 1; i < matrWidth; i++ )
        for( j = 0; j < i; j++ )
            destMatr[j + i * matrWidth] = destMatr[i + j * matrWidth];
    return CV_NO_ERR;

}

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       cvTransposeMatrix_32f   
//    Purpose:    Transposes matrix
//    Context:   
//    Parameters: srcMatr - source matrix
//                matrWidth
//                matrHeight - sizes of srcMatr
//                destMatr - destination matrix
//               
//    Returns:
//    CV_NO_ERROR or error code     
//    Notes:
//    matrWidth1 mus be equal to matrHeight2      
//F*/
IPCVAPI_IMPL( CvStatus, icvTransposeMatrix_32f, (CvMatr32f srcMatr,
                                                 int matrWidth,
                                                 int matrHeight, CvMatr32f destMatr) )
{
    int i, j;
    float *dst = destMatr;
    float *src = srcMatr;
    int Decrem = matrWidth * matrHeight - 1;

    if( srcMatr == NULL )
        return CV_NULLPTR_ERR;
    if( destMatr == NULL )
        return CV_NULLPTR_ERR;
    if( matrWidth <= 0 )
        return CV_BADSIZE_ERR;
    if( matrHeight <= 0 )
        return CV_BADSIZE_ERR;
    for( j = 0; j < matrHeight; j++ )
    {
        for( i = 0; i < matrWidth; i++ )
        {
            *dst = src[i];
            dst += matrHeight;
        }
        dst -= Decrem;
        src += matrWidth;
    }
    return CV_NO_ERR;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       cvInvertMatrix_32f   
//    Purpose:    Invertes matrix
//    Context:   
//    Parameters: srcMatr - source matrix
//                matrSize - sizes of srcMatr
//                destMatr - destination matrix
//               
//    Returns:
//    CV_NO_ERROR or error code     
//    Notes:
//F*/
IPCVAPI_IMPL( CvStatus, icvInvertMatrix_32f, (CvMatr32f srcMatr,
                                              int matrSize, CvMatr32f destMatr) )
{
    CvMatr32f Temp;
    CvVect32f tmpVect;
    CvVect32f tmpLineVect;
    CvVect32f tmpDstLineVect;
    int i, j;
    float Mult;
    int change, st;
    int change_sign;

    change_sign = 0;
    Temp = icvCreateMatrix_32f( matrSize, matrSize );
    tmpVect = icvCreateVector_32f( matrSize );
    tmpLineVect = icvCreateVector_32f( matrSize );
    tmpDstLineVect = icvCreateVector_32f( matrSize );

    icvCopyMatrix_32f( srcMatr, matrSize, matrSize, Temp );
    icvSetIdentity_32f( destMatr, matrSize, matrSize );

    for( i = 0; i < matrSize; i++ )
    {

        /* Sort line for non zero element */

        change = i;


        /* !!! In change when test !=0 if all values == 0 then DET=0 and INV not exist */

        for( st = i; st < matrSize; st++ )
        {

            if( Temp[st * matrSize + i] != 0 )
            {

                change = st;
                break;
            }
        }

        if( change != i )
        {
            icvCopyVector_32f( &Temp[(i) * matrSize], matrSize, tmpVect );

            icvCopyVector_32f( &Temp[st * matrSize], matrSize, &Temp[(i) * matrSize] );

            icvCopyVector_32f( tmpVect, matrSize, &Temp[st * matrSize] );

            icvCopyVector_32f( &destMatr[(i) * matrSize], matrSize, tmpVect );

            icvCopyVector_32f( &destMatr[st * matrSize],
                                matrSize, &destMatr[(i) * matrSize] );

            icvCopyVector_32f( tmpVect, matrSize, &destMatr[st * matrSize] );
            change_sign++;

        }
        else
        {
            if( Temp[i * matrSize + i] == 0 )
            {
                /* Inv matrix does not exist det == 0 */
                return CV_BADCOEF_ERR;
            }
        }

        Mult = 1.0f / Temp[i * matrSize + i];

        Temp[i * matrSize + i] = 1.0f;

        /* Test !!! */

        icvScaleVector_32f( &Temp[i * matrSize + i + 1],
                             &Temp[i * matrSize + i + 1], matrSize - i - 1, Mult );

        icvScaleVector_32f( &destMatr[i * matrSize],
                             &destMatr[i * matrSize], matrSize, Mult );

        /* Subtructing */

        icvCopyVector_32f( &Temp[i * matrSize + i + 1], matrSize - i - 1, tmpLineVect );

        icvCopyVector_32f( &destMatr[i * matrSize], matrSize, tmpDstLineVect );

        for( j = i + 1; j < matrSize; j++ )
        {

            Mult = Temp[j * matrSize + i];

            Temp[j * matrSize + i] = 0.0f;

            icvScaleVector_32f( tmpLineVect, tmpVect, matrSize - i - 1, Mult );

            icvSubVector_32f( &Temp[j * matrSize + i + 1],
                               tmpVect, &Temp[j * matrSize + i + 1], matrSize - i - 1 );


            icvScaleVector_32f( tmpDstLineVect, tmpVect, matrSize, Mult );

            icvSubVector_32f( &destMatr[j * matrSize],
                               tmpVect, &destMatr[j * matrSize], matrSize );

        }
    }

    /* Down to top direct */

    for( i = matrSize - 1; i > 0; i-- )
    {

        /* Subtructing */

        icvCopyVector_32f( &Temp[i * matrSize + i + 1], matrSize - i - 1, tmpLineVect );

        icvCopyVector_32f( &destMatr[i * matrSize], matrSize, tmpDstLineVect );

        for( j = i - 1; j >= 0; j-- )
        {

            Mult = Temp[j * matrSize + i];

            Temp[j * matrSize + i] = 0.0f;

            icvScaleVector_32f( tmpLineVect, tmpVect, matrSize - i - 1, Mult );

            icvSubVector_32f( &Temp[j * matrSize + i + 1],
                               tmpVect, &Temp[j * matrSize + i + 1], matrSize - i - 1 );


            icvScaleVector_32f( tmpDstLineVect, tmpVect, matrSize, Mult );

            icvSubVector_32f( &destMatr[j * matrSize],
                               tmpVect, &destMatr[j * matrSize], matrSize );


        }
    }

    if( change_sign & 1 == 1 )
    {
        icvScaleMatrix_32f( destMatr, destMatr, matrSize, matrSize, -1.0f );
    }


    icvDeleteMatrix( Temp );
    icvDeleteVector( tmpVect );
    icvDeleteVector( tmpLineVect );
    icvDeleteVector( tmpDstLineVect );

    return CV_NO_ERR;
}                               /* icvInvertMatrix_32f */

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       icvDetMatrix_32f   
//    Purpose:    Compute determinate of matrix
//    Context:   
//    Parameters: matr - source matrix
//                matrSize - sizes of srcMatr
//                det - result
//               
//    Returns:
//    CV_NO_ERROR or error code     
//    Notes:
//F*/
IPCVAPI_IMPL( CvStatus, icvDetMatrix_32f, (CvMatr32f matr, int matrSize, double* det) )
{
    CvMatr32f matrWork;
    CvVect32f tmpLine;
    int i, j, t;
    int change_sign;
    float coef;
    float mul;
    double norm = 0, eps, eps0 = 1.0e-15;

    *det = 0;

    if( matr == NULL || det == NULL ) return CV_NULLPTR_ERR;
    if( matrSize < 1 ) return CV_BADFACTOR_ERR;

    if( matrSize == 1 )
    {
        *det = (double)(*matr);
        return CV_NO_ERR;
    }

    if( matrSize == 2 )
    {
        *det = matr[0]*matr[3] - matr[1]*matr[2];
        return CV_NO_ERR;
    }

    if( matrSize == 3 )
    {
        *det = matr[0] * ( matr[4]*matr[8] - matr[5]*matr[7] )
             + matr[3] * ( matr[2]*matr[7] - matr[1]*matr[8] )
             + matr[6] * ( matr[1]*matr[5] - matr[2]*matr[4] );
        return CV_NO_ERR;
    }

    if( matrSize == 4 )
    {
        *det = matr[0] * (matr[5] * (matr[10]*matr[15] - matr[11]*matr[14])
                       +  matr[6] * (matr[11]*matr[13] - matr[9] *matr[15])
                       +  matr[7] * (matr[9] *matr[14] - matr[10]*matr[13]) );

        *det+= matr[1] * (matr[6] * (matr[8] *matr[15] - matr[11]*matr[12])
                       + matr[7]  * (matr[10]*matr[12] - matr[8] *matr[14])
                       + matr[4]  * (matr[11]*matr[14] - matr[10]*matr[15]) );

        *det+= matr[2] * (matr[7] * (matr[8] *matr[13] - matr[9] *matr[12])
                       + matr[4]  * (matr[9] *matr[15] - matr[11]*matr[13])
                       + matr[5]  * (matr[11]*matr[12] - matr[8] *matr[15]) );

        *det+= matr[3] * (matr[4] * (matr[10]*matr[13] - matr[9] *matr[14])
                       + matr[5]  * (matr[8] *matr[14] - matr[10]*matr[12])
                       + matr[6]  * (matr[9] *matr[12] - matr[8] *matr[13]) );
        return CV_NO_ERR;
    }

    if( matrSize > 6 ) /* Large matrices must be developed with double presicion */
    {
        CvMatr64d matrWork64 = icvCreateMatrix_64d( matrSize, matrSize );
        CvStatus result;

        for( i=0; i<matrSize*matrSize; i++ ) matrWork64[i] = (double)matr[i];
        result = icvDetMatrix_64d( matrWork64, matrSize, det);
        icvDeleteMatrix( matrWork64 );

        return result;
    }

    else
    {
    matrWork = icvCreateMatrix_32f( matrSize, matrSize );
    tmpLine  = icvCreateVector_32f( matrSize );

    change_sign = 0;

    icvCopyMatrix_32f( matr, matrSize, matrSize, matrWork );

    for( i=0; i<matrSize*matrSize; i++ )
        norm += fabs( matr[i] );

    norm /= matrSize*matrSize;
    eps = eps0 * norm;

    /* Make matrix is up triangular */
    for( i = 0; i < matrSize - 1; i++ )
    {
        int     in = i * matrSize;
        double  mx = fabs( matrWork[in+i] );
        int change = i;

        /* find maximal element */

        for( t = i+1; t < matrSize; t++ )
        {
            double w = fabs( matrWork[t*matrSize + i] );
            if( w > mx )
            {
                mx     = w;
                change = t;
            }
        }

        if( mx < eps ) /* Element != 0 not found;  det = 0 */
        {
            *det = 0.0;
            icvDeleteMatrix( matrWork );
            icvDeleteVector( tmpLine );
            return CV_NO_ERR;
        }

        if( change != i ) /* change lines */
        {
            int nchange = change * matrSize;

            icvCopyVector_32f( matrWork + nchange, matrSize,  tmpLine );
            icvCopyVector_32f( matrWork + in,      matrSize,  matrWork + nchange );
            icvCopyVector_32f( tmpLine,            matrSize,  matrWork + in );

            change_sign++;
        }

        /* Virtual set elements to zero */

        coef = 1.0f / matrWork[in+i];

        for( j = i+1; j < matrSize; j++ )
        {
            int jn = j * matrSize;

            if( matrWork[jn + i] )
            {
                mul = coef * matrWork[jn + i];

                //icvScaleVector_32f( matrWork+in+i+1, tmpLine, matrSize-i-1, mul );
                //icvSubVector_32f  ( matrWork+jn+i+1, tmpLine, matrWork+jn+i+1,
                //                                                   matrSize-i-1 );
                /* more precision, but slower :  */
                 for( t=i+1; t<matrSize; t++ ) matrWork[jn+t] -= mul*matrWork[in+t]; 
            }
        }

    }  /* i */

    *det = ( (change_sign & 1) == 1 ) ? -1.0 : 1.0;

    for( t=0, j=0; t < matrSize; t++, j += matrSize+1 )
        *det *= (double)matrWork[j];

    icvDeleteMatrix( matrWork );
    icvDeleteVector( tmpLine );
    return CV_NO_ERR;

    }  /* else */
}                               /* icvDetMatrix_32f */

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       icvTrace_32f   
//    Purpose:    Compute the trace of matrix, sum of diagonal elements
//    Context:   
//    Parameters: matr - source matrix
//                width
//                height - sizes of matr
//                trace - result
//               
//    Returns:
//    CV_NO_ERROR or error code     
//    Notes:
//F*/
IPCVAPI_IMPL( CvStatus, icvTrace_32f, (CvMatr32f matr, int width, int height, float *trace) )
{
    int t;
    int size;
    float tr;

    size = min( width, height );
    tr = 0;
    for( t = 0; t < size; t++ )
    {
        tr += matr[t * width + t];
    }
    *trace = tr;

    return CV_NO_ERR;
}                               /* icvTrace_32f */


/* End of file */
