/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include "_cv.h"
#include "_cvarray.h"

#include <float.h>

float
icvsDotProduct( float *src1, float *src2, int length )
{
    int i;
    float sum = 0.f;

    for( i = 0; i < length; i++ )
    {
        sum += src1[i] * src2[i];
    }
    return sum;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       icvTransformVector_32f
//    Purpose:    Calculates linear transform of a vector by a matrix.
//    Context:   
//    Parameters: matr - matrix of transformation
//                srcVect - vector
//                destVect - result of matrix-vector product
//                matrWidth
//                matrHeight - sizes of matrix               
//    Returns:   
//    Notes:    
//F*/

IPCVAPI_IMPL( CvStatus, icvTransformVector_32f, (CvMatr32f matr,
                                                 CvVect32f srcVec,
                                                 CvVect32f destVec,
                                                 int matrWidth, int matrHeight) )
{
    int i;
    float *tmp_vect;

    /* check bad arguments */
    if( srcVec == NULL )
        return CV_NULLPTR_ERR;
    if( matr == NULL )
        return CV_NULLPTR_ERR;
    if( destVec == NULL )
        return CV_NULLPTR_ERR;
    if( matrWidth <= 0 )
        return CV_BADSIZE_ERR;
    if( matrHeight <= 0 )
        return CV_BADSIZE_ERR;

    if( srcVec == destVec )
    {
        /* create temp vector for inplace version */
        tmp_vect = (float *) icvAlloc( matrHeight * sizeof( float ));

        if( tmp_vect == NULL )
            return CV_OUTOFMEM_ERR;
    }
    else
        tmp_vect = destVec;


    /* computing */
    for( i = 0; i < matrHeight; i++ )
    {
        tmp_vect[i] = icvsDotProduct( matr, srcVec, matrWidth );
        matr += matrWidth;
    }

    /* if inplace - copy tmp vector to destination */
    if( srcVec == destVec )
    {
        icv1DCopy( tmp_vect, destVec, matrWidth /* or step */  );
        icvFree( &tmp_vect );
    }

    return CV_NO_ERR;
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       icvTransformVector_64d
//    Purpose:    Calculates linear transform of a vector by a matrix.
//    Context:   
//    Parameters: matr - matrix of transformation
//                srcVect - vector
//                destVect - result of matrix-vector product
//                matrWidth
//                matrHeight - sizes of matrix               
//    Returns:   
//    Notes:    
//F*/
IPCVAPI_IMPL( CvStatus, icvTransformVector_64d, (CvMatr64d matr,
                                                 CvVect64d srcVec,
                                                 CvVect64d destVec,
                                                 int matrWidth, int matrHeight) )
{
    int i, j;
    double Sum;

    if( srcVec == NULL )
        return CV_NULLPTR_ERR;
    if( matr == NULL )
        return CV_NULLPTR_ERR;
    if( destVec == NULL )
        return CV_NULLPTR_ERR;
    if( matrWidth <= 0 )
        return CV_BADSIZE_ERR;
    if( matrHeight <= 0 )
        return CV_BADSIZE_ERR;
    for( i = 0; i < matrHeight; i++ )
    {
        Sum = 0;
        for( j = 0; j < matrWidth; j++ )
        {
            Sum += matr[i * matrWidth + j] * srcVec[j];
        }
        destVec[i] = Sum;
    }
    return CV_NO_ERR;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       icvTransformVector4x1_32f
//    Purpose:    Calculates linear transform of a vector 4x1 by a matrix 4x4.
//    Context:   
//    Parameters: matr - matrix of transformation
//                srcVect - vector
//                destVect - result of matrix-vector product
//        
//    Returns:   
//    Notes:    
//F*/
IPCVAPI_IMPL( CvStatus, icvTransformVector4x1_32f, (CvMatr32f matr,
                                                    CvVect32f srcVec, CvVect32f destVec) )
{
    if( srcVec == NULL )
        return CV_NULLPTR_ERR;
    if( matr == NULL )
        return CV_NULLPTR_ERR;
    if( destVec == NULL )
        return CV_NULLPTR_ERR;
    destVec[0] = matr[0] * srcVec[0] +
        matr[1] * srcVec[1] + matr[2] * srcVec[2] + matr[3] * srcVec[3];
    destVec[1] = matr[4] * srcVec[0] +
        matr[5] * srcVec[1] + matr[6] * srcVec[2] + matr[7] * srcVec[3];
    destVec[2] = matr[8] * srcVec[0] +
        matr[9] * srcVec[1] + matr[10] * srcVec[2] + matr[11] * srcVec[3];
    destVec[3] = matr[12] * srcVec[0] +
        matr[13] * srcVec[1] + matr[14] * srcVec[2] + matr[15] * srcVec[3];
    return CV_NO_ERR;
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       icvTransformVector3x1_32f
//    Purpose:    Calculates linear transform of a vector 3x1 by a matrix 3x3.
//    Context:   
//    Parameters: matr - matrix of transformation
//                srcVect - vector
//                destVect - result of matrix-vector product
        
//    Returns:   
//    Notes:    
//F*/
IPCVAPI_IMPL( CvStatus, icvTransformVector3x1_32f, (CvMatr32f matr,
                                                    CvVect32f srcVec, CvVect32f destVec) )
{
    if( srcVec == NULL )
        return CV_NULLPTR_ERR;
    if( matr == NULL )
        return CV_NULLPTR_ERR;
    if( destVec == NULL )
        return CV_NULLPTR_ERR;
    destVec[0] = matr[0] * srcVec[0] + matr[1] * srcVec[1] + matr[2] * srcVec[2];
    destVec[1] = matr[3] * srcVec[0] + matr[4] * srcVec[1] + matr[5] * srcVec[2];
    destVec[2] = matr[6] * srcVec[0] + matr[7] * srcVec[1] + matr[8] * srcVec[2];
    return CV_NO_ERR;
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       icvsCalcQuadraticForm
//    Purpose:    Calculates the quadratic form on the two vectors defined by matrix.
//    Context:   
//    Parameters: matr - matrix of quadratic form
//                srcVec1 - left vector
//                srcVec2 - right vector
//                matrWidth
//                matrHeight - sizes of matrix               
//                mean - result of vector-matrix-vector product
//    Returns:   
//    Notes:     
//F*/
IPCVAPI_IMPL( CvStatus, icvCalcQuadraticForm_32f, (CvVect32f srcVec1,
                                                   CvVect32f srcVec2,
                                                   CvMatr32f matr,
                                                   int matrWidth,
                                                   int matrHeight, float *mean) )
{
    CvStatus result;
    float *tmp_vect;

    /* check bad arguments */
    if( srcVec1 == NULL )
        return CV_NULLPTR_ERR;
    if( matr == NULL )
        return CV_NULLPTR_ERR;
    if( srcVec2 == NULL )
        return CV_NULLPTR_ERR;
    if( matrWidth <= 0 )
        return CV_BADSIZE_ERR;
    if( matrHeight <= 0 )
        return CV_BADSIZE_ERR;
    if( mean == NULL )
        return CV_NULLPTR_ERR;

    /* create temp vector */
    tmp_vect = (float *) icvAlloc( matrHeight * sizeof( float ));

    if( tmp_vect == NULL )
        return CV_OUTOFMEM_ERR;

    result = icvTransformVector_32f( matr, srcVec2, tmp_vect, matrWidth, matrHeight );
    if( result == CV_NO_ERR )
    {
        *mean = icvsDotProduct( tmp_vect, srcVec1, matrHeight );
    }

    icvFree( tmp_vect );

    return result;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       icvsMahalanobis_32f
//    Purpose:    Calculates mahalanobis distance. 
//    Context:   
//    Parameters: matr - covariance matrix
//                srcVec1 - vector
//                srcVec2 - vector
//                matrSize - sizes of matrix               
//                distance - mahalanobis distance
//    Returns:   
//    Notes:     
//F*/
IPCVAPI_IMPL( CvStatus, icvMahalanobis_32f, (CvVect32f srcVec1,
                                             CvVect32f srcVec2,
                                             CvMatr32f matr, int matrSize, float *distance) )
{
    CvStatus result;
    CvVect32f tmp_vect;
    CvVect32f tmp_vect2;

    if( srcVec1 == NULL )
        return CV_NULLPTR_ERR;
    if( matr == NULL )
        return CV_NULLPTR_ERR;
    if( srcVec2 == NULL )
        return CV_NULLPTR_ERR;
    if( matrSize <= 0 )
        return CV_BADSIZE_ERR;
    if( distance == NULL )
        return CV_NULLPTR_ERR;

    tmp_vect = icvCreateVector_32f( matrSize );
    tmp_vect2 = icvCreateVector_32f( matrSize );
    if( (!tmp_vect) || (!tmp_vect2) )
        return CV_OUTOFMEM_ERR;

    icvSubVector_32f( srcVec1, srcVec2, tmp_vect, matrSize );

    result = icvTransformVector_32f( matr, tmp_vect, tmp_vect2, matrSize, matrSize );

    if( result == CV_NO_ERR )
    {
        float d2 = icvsDotProduct( tmp_vect, tmp_vect2, matrSize );

        *distance = cvSqrt( d2 );
    }

    icvDeleteVector( tmp_vect );
    icvDeleteVector( tmp_vect2 );

    return result;
}

void
icvbTransformVector3D_32f( CvMatr32f matr, CvVect32f srcVec, int count, CvVect32f destVec )
{
    int i;
    float a1 = matr[0], b1 = matr[1], c1 = matr[2],
        a2 = matr[4], b2 = matr[5], c2 = matr[6],
        a3 = matr[8], b3 = matr[9], c3 = matr[10], t1 = matr[3], t2 = matr[7], t3 = matr[11];

    for( i = 0; i < 3 * count; i += 3 )
    {
        float x = srcVec[i], y = srcVec[i + 1], z = srcVec[i + 2];

        destVec[i] = a1 * x + b1 * y + c1 * z + t1;
        destVec[i + 1] = a2 * x + b2 * y + c2 * z + t2;
        destVec[i + 2] = a3 * x + b3 * y + c3 * z + t3;
    }
}

void
icvbPerspectiveVector3D_32f( CvMatr32f matr, CvVect32f srcVec, int count, CvVect32f destVec )
{
    int i;
    float a1 = matr[0], b1 = matr[1], c1 = matr[2],
        a2 = matr[4], b2 = matr[5], c2 = matr[6],
        a3 = matr[8], b3 = matr[9], c3 = matr[10],
        t1 = matr[3], t2 = matr[7], t3 = matr[11],
        a4 = matr[12], b4 = matr[13], c4 = matr[14], t4 = matr[15];

    for( i = 0; i < 3 * count; i += 3 )
    {
        float x = srcVec[i], y = srcVec[i + 1], z = srcVec[i + 2];
        float w = a4 * x + b4 * y + c4 * z + t4;

        if( fabs( w ) > FLT_MIN )
        {
            destVec[i] = (a1 * x + b1 * y + c1 * z + t1) / w;
            destVec[i + 1] = (a2 * x + b2 * y + c2 * z + t2) / w;
            destVec[i + 2] = (a3 * x + b3 * y + c3 * z + t3) / w;
        }
        else
        {
            destVec[i] = 0;
            destVec[i + 1] = 0;
            destVec[i + 2] = 0;
        }
    }
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       icvbTransformVector3x1P_32f
//    Purpose:    Calculates a product of a matrix 4x4 and an array of vectors (x, y, z, 1)
//                  and then divides the first three components of the result by the fourth.
//                  If the fourth component is zero then the output is (0, 0, 0, 0).
//    Context:   
//    Parameters: matr - matrix of transformation (4x4)
//                srcVect - input array of vectors (x, y, z)
//                count - number of input vectors
//                destVect - array of vectors (x, y, z)
//    Returns:   
//    Notes:    
//F*/
IPCVAPI_IMPL( CvStatus, icvbTransformVector3x1P_32f, (CvMatr32f matr,
                                                      CvVect32f srcVec,
                                                      int count, CvVect32f destVec) )
{
    if( matr == NULL || srcVec == NULL || destVec == NULL )
    {
        return CV_NULLPTR_ERR;
    }

    if( count <= 0 )
    {
        return CV_BADFACTOR_ERR;
    }

    if( fabs( matr[12] ) < FLT_MIN && fabs( matr[13] ) < FLT_MIN &&
        fabs( matr[14] ) < FLT_MIN && fabs( matr[15] - 1.0f ) < FLT_MIN )
    {
        icvbTransformVector3D_32f( matr, srcVec, count, destVec );
    }
    else
    {
        icvbPerspectiveVector3D_32f( matr, srcVec, count, destVec );
    }

    return CV_NO_ERR;
}

/* End of file */
