/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include "_cv.h"
#include "_cvimgstat.h"
#include <float.h>

/****************************************************************************************\
*                              Mean calculation with mask                                *
\****************************************************************************************/

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name: icvMeanMask
//    Purpose:  Calculates mean value over the pixels 
//              in the selected (by mask) region
//    Context:
//    Parameters:  
//      img  - source image
//      step - its step
//      mask - mask image ( non-zero values indicate region )
//      mask_step - its step
//      size - its size (width and height)
//      coi  - index of channel (if color image)
//    Returns:
//      mean value
//    Notes:
//F*/

/* /////////////// 8uC1 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMean_8u_C1MR, (const uchar * img, int step,
                                          const uchar * mask, int mask_step,
                                          CvSize size, double *mean) )
{
    int i, j, sum = 0, pix = 0;

    if( !img || !mask || !mean )
        return CV_NULLPTR_ERR;
    if( size.width <= 0 || size.height <= 0 || step < size.width || mask_step < size.width )
        return CV_BADSIZE_ERR;

    if( size.width == step && size.width == mask_step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( i = 0; i < size.height; i++, img += step, mask += mask_step )
    {
        for( j = 0; j < size.width - 3; j += 4 )
        {
            int m = (mask[j] == 0) - 1;

            sum += m & img[j];
            pix -= m;

            m = (mask[j + 1] == 0) - 1;
            sum += m & img[j + 1];
            pix -= m;

            m = (mask[j + 2] == 0) - 1;
            sum += m & img[j + 2];
            pix -= m;

            m = (mask[j + 3] == 0) - 1;
            sum += m & img[j + 3];
            pix -= m;
        }
        for( ; j < size.width; j++ )
        {
            int m = (mask[j] == 0) - 1;

            sum += img[j] & m;
            pix -= m;
        }
    }

    *mean = pix ? ((double) sum) / pix : 0;
    return CV_OK;
}


/* /////////////// 8uC3 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMean_8u_C3CMR, (const uchar * img, int step,
                                           const uchar * mask, int mask_step,
                                           CvSize size, int coi, double *mean) )
{
    int i, j, sum = 0, pix = 0;

    coi--;

    if( !img || !mask || !mean )
        return CV_NULLPTR_ERR;
    if( size.width <= 0 || size.height <= 0 ||
        step < size.width * 3 || mask_step < size.width )
        return CV_BADSIZE_ERR;
    if( (unsigned) coi > 2 )
        return CV_BADRANGE_ERR;


    if( size.width * 3 == step && size.width == mask_step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    img += coi;
    for( i = 0; i < size.height; i++, img += step, mask += mask_step )
    {
        for( j = 0; j < size.width - 3; j += 4 )
        {
            int m = (mask[j] == 0) - 1;

            sum += m & img[j * 3];
            pix -= m;

            m = (mask[j + 1] == 0) - 1;
            sum += m & img[j * 3 + 3];
            pix -= m;

            m = (mask[j + 2] == 0) - 1;
            sum += m & img[j * 3 + 6];
            pix -= m;

            m = (mask[j + 3] == 0) - 1;
            sum += m & img[j * 3 + 9];
            pix -= m;
        }
        for( ; j < size.width; j++ )
        {
            int m = (mask[j] == 0) - 1;

            sum += img[j * 3] & m;
            pix -= m;
        }
    }

    *mean = pix ? ((double) sum) / pix : 0;
    return CV_OK;
}


/* /////////////// 8sC1 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMean_8s_C1MR, (const char *img, int step,
                                          const uchar * mask, int mask_step,
                                          CvSize size, double *mean) )
{
    int i, j, sum = 0, pix = 0;

    if( !img || !mask || !mean )
        return CV_NULLPTR_ERR;
    if( size.width <= 0 || size.height <= 0 || step < size.width || mask_step < size.width )
        return CV_BADSIZE_ERR;

    if( size.width == step && size.width == mask_step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( i = 0; i < size.height; i++, img += step, mask += mask_step )
    {
        for( j = 0; j < size.width - 3; j += 4 )
        {
            int m = (mask[j] == 0) - 1;

            sum += m & img[j];
            pix -= m;

            m = (mask[j + 1] == 0) - 1;
            sum += m & img[j + 1];
            pix -= m;

            m = (mask[j + 2] == 0) - 1;
            sum += m & img[j + 2];
            pix -= m;

            m = (mask[j + 3] == 0) - 1;
            sum += m & img[j + 3];
            pix -= m;
        }
        for( ; j < size.width; j++ )
        {
            int m = (mask[j] == 0) - 1;

            sum += img[j] & m;
            pix -= m;
        }
    }

    *mean = pix ? ((double) sum) / pix : 0;
    return CV_OK;
}


/* /////////////// 8sC3 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMean_8s_C3CMR, (const char *img, int step,
                                           const uchar * mask, int mask_step,
                                           CvSize size, int coi, double *mean) )
{
    int i, j, sum = 0, pix = 0;

    coi--;

    if( !img || !mask || !mean )
        return CV_NULLPTR_ERR;
    if( size.width <= 0 || size.height <= 0 ||
        step < size.width * 3 || mask_step < size.width )
        return CV_BADSIZE_ERR;
    if( (unsigned) coi > 2 )
        return CV_BADRANGE_ERR;


    if( size.width * 3 == step && size.width == mask_step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    img += coi;
    for( i = 0; i < size.height; i++, img += step, mask += mask_step )
    {
        for( j = 0; j < size.width - 3; j += 4 )
        {
            int m = (mask[j] == 0) - 1;

            sum += m & img[j * 3];
            pix -= m;

            m = (mask[j + 1] == 0) - 1;
            sum += m & img[j * 3 + 3];
            pix -= m;

            m = (mask[j + 2] == 0) - 1;
            sum += m & img[j * 3 + 6];
            pix -= m;

            m = (mask[j + 3] == 0) - 1;
            sum += m & img[j * 3 + 9];
            pix -= m;
        }
        for( ; j < size.width; j++ )
        {
            int m = (mask[j] == 0) - 1;

            sum += img[j * 3] & m;
            pix -= m;
        }
    }

    *mean = pix ? ((double) sum) / pix : 0;
    return CV_OK;
}


/* /////////////// 32fC1 flavor //////////////////// */

IPCVAPI_IMPL( CvStatus, icvMean_32f_C1MR, (const float *img, int step,
                                           const uchar * mask, int mask_step,
                                           CvSize size, double *mean) )
{
    int i, j, pix = 0;
    double sum = 0;
    float coeff[2] = { 0.f, 1.f };

    if( !img || !mask || !mean )
        return CV_NULLPTR_ERR;
    if( size.width <= 0 || size.height <= 0 ||
        step < size.width || mask_step < size.width || step % sizeof_float != 0 )
        return CV_BADSIZE_ERR;

    if( size.width == step && size.width == mask_step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    step /= sizeof_float;

    for( i = 0; i < size.height; i++, img += step, mask += mask_step )
    {
        for( j = 0; j < size.width - 3; j += 4 )
        {
            int m = mask[j] != 0;

            sum += img[j] * coeff[m];
            pix += m;

            m = mask[j + 1] != 0;
            sum += img[j + 1] * coeff[m];
            pix += m;

            m = mask[j + 2] != 0;
            sum += img[j + 2] * coeff[m];
            pix += m;

            m = mask[j + 3] != 0;
            sum += img[j + 3] * coeff[m];
            pix += m;
        }
        for( ; j < size.width; j++ )
        {
            int m = mask[j] != 0;

            sum += img[j] * coeff[m];
            pix += m;
        }
    }

    *mean = pix ? ((double) sum) / pix : 0;
    return CV_OK;
}


/* /////////////// 32fC3 flavor //////////////////// */

IPCVAPI_IMPL( CvStatus, icvMean_32f_C3CMR, (const float *img, int step,
                                            const uchar * mask, int mask_step,
                                            CvSize size, int coi, double *mean) )
{
    int i, j, pix = 0;
    double sum = 0;
    float coeff[2] = { 0.f, 1.f };

    coi--;

    if( !img || !mask || !mean )
        return CV_NULLPTR_ERR;
    if( size.width <= 0 || size.height <= 0 ||
        step < size.width * 3 || mask_step < size.width || step % sizeof_float != 0 )
        return CV_BADSIZE_ERR;
    if( (unsigned) coi > 2 )
        return CV_BADRANGE_ERR;


    if( size.width * 3 == step && size.width == mask_step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    img += coi;
    step /= sizeof_float;

    for( i = 0; i < size.height; i++, img += step, mask += mask_step )
    {
        for( j = 0; j < size.width - 3; j += 4 )
        {
            int m = mask[j] != 0;

            sum += img[j * 3] * coeff[m];
            pix += m;

            m = mask[j + 1] != 0;
            sum += img[j * 3 + 3] * coeff[m];
            pix += m;

            m = mask[j + 2] != 0;
            sum += img[j * 3 + 6] * coeff[m];
            pix += m;

            m = mask[j + 3] != 0;
            sum += img[j * 3 + 9] * coeff[m];
            pix += m;
        }
        for( ; j < size.width; j++ )
        {
            int m = mask[j] != 0;

            sum += img[j * 3] * coeff[m];
            pix += m;
        }
    }

    *mean = pix ? ((double) sum) / pix : 0;
    return CV_OK;
}

/*  End of file  */
