/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include "_cv.h"
#include "_cvimgstat.h"

/****************************************************************************************\
*                             Mean and StdDev calculation                                *
\****************************************************************************************/

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name: icvMean_StdDev
//    Purpose:  Calculates mean value over the pixels in the ROI
//    Context:
//    Parameters:  
//      img  - source image
//      step - its step
//      size - its size (width and height)
//      coi  - index of channel (if color image)
//      mean - pointer to mean value
//      stddev - pointer to standard deviation
//    Returns:
//      CV_NO_ERR or error code
//    Notes:
//F*/

/* /////////////// 8uC1 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMean_StdDev_8u_C1R, (const uchar * img, int step, CvSize size,
                                                double *mean, double *stddev) )
{
    int i, j;
    int sum = 0;
    int64 sum2 = 0;

    if( !img || !mean || !stddev )
        return CV_NULLPTR_ERR;
    if( size.width <= 0 || size.height <= 0 || step < size.width )
        return CV_BADSIZE_ERR;

    if( size.width == step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( i = 0; i < size.height; i++, img += step )
    {
        for( j = 0; j < size.width - 3; j += 4 )
        {
            int v = img[j];
            int s = v;
            int s2 = v * v;

            v = img[j + 1];
            s += v;
            s2 += v * v;
            v = img[j + 2];
            s += v;
            s2 += v * v;
            v = img[j + 3];
            s += v;
            s2 += v * v;
            sum += s;
            sum2 += s2;
        }
        for( ; j < size.width; j++ )
        {
            int v = img[j];

            sum += v;
            sum2 += v * v;
        }
    }

    {
        double t = 1. / (size.width * size.height);
        double mean_val = sum * t;

        *mean = mean_val;
        *stddev = sqrt( sum2 * t - mean_val * mean_val );
    }

    return CV_NO_ERR;
}


/* /////////////// 8uC3 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMean_StdDev_8u_C3CR, (const uchar * img, int step, CvSize size,
                                                 int coi, double *mean, double *stddev) )
{
    int i, j;
    int sum = 0;
    int64 sum2 = 0;

    coi--;
    size.width *= 3;

    if( !img || (!mean && !stddev) )
        return CV_NULLPTR_ERR;
    if( size.width <= 0 || size.height <= 0 || step < size.width )
        return CV_BADSIZE_ERR;

    if( (unsigned) coi > 2 )
        return CV_BADRANGE_ERR;

    if( size.width == step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    img += coi;

    for( i = 0; i < size.height; i++, img += step )
    {
        for( j = 0; j < size.width - 9; j += 12 )
        {
            int v = img[j];
            int s = v;
            int s2 = v * v;

            v = img[j + 3];
            s += v;
            s2 += v * v;
            v = img[j + 6];
            s += v;
            s2 += v * v;
            v = img[j + 9];
            s += v;
            s2 += v * v;
            sum += s;
            sum2 += s2;
        }
        for( ; j < size.width; j += 3 )
        {
            int v = img[j];

            sum += v;
            sum2 += v * v;
        }
    }

    {
        double t = 3. / (size.width * size.height);
        double mean_val = sum * t;

        *mean = mean_val;
        *stddev = sqrt( sum2 * t - mean_val * mean_val );
    }

    return CV_NO_ERR;
}

/* /////////////// 8sC1 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMean_StdDev_8s_C1R, (const char *img, int step, CvSize size,
                                                double *mean, double *stddev) )
{
    int i, j;
    int sum = 0;
    int64 sum2 = 0;

    if( !img || (!mean && !stddev) )
        return CV_NULLPTR_ERR;
    if( size.width <= 0 || size.height <= 0 || step < size.width )
        return CV_BADSIZE_ERR;

    if( size.width == step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( i = 0; i < size.height; i++, img += step )
    {
        for( j = 0; j < size.width - 3; j += 4 )
        {
            int v = img[j];
            int s = v;
            int s2 = v * v;

            v = img[j + 1];
            s += v;
            s2 += v * v;
            v = img[j + 2];
            s += v;
            s2 += v * v;
            v = img[j + 3];
            s += v;
            s2 += v * v;
            sum += s;
            sum2 += s2;
        }
        for( ; j < size.width; j++ )
        {
            int v = img[j];

            sum += v;
            sum2 += v * v;
        }
    }

    {
        double t = 1. / (size.width * size.height);
        double mean_val = sum * t;

        *mean = mean_val;
        *stddev = sqrt( sum2 * t - mean_val * mean_val );
    }

    return CV_NO_ERR;
}

/* /////////////// 8uC3 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMean_StdDev_8s_C3CR, (const char *img, int step, CvSize size,
                                                 int coi, double *mean, double *stddev) )
{
    int i, j;
    int sum = 0;
    int64 sum2 = 0;

    coi--;
    size.width *= 3;

    if( !img || (!mean && !stddev) )
        return CV_NULLPTR_ERR;
    if( size.width <= 0 || size.height <= 0 || step < size.width )
        return CV_BADSIZE_ERR;

    if( (unsigned) coi > 2 )
        return CV_BADRANGE_ERR;

    if( size.width == step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    img += coi;

    for( i = 0; i < size.height; i++, img += step )
    {
        for( j = 0; j < size.width - 9; j += 12 )
        {
            int v = img[j];
            int s = v;
            int s2 = v * v;

            v = img[j + 3];
            s += v;
            s2 += v * v;
            v = img[j + 6];
            s += v;
            s2 += v * v;
            v = img[j + 9];
            s += v;
            s2 += v * v;
            sum += s;
            sum2 += s2;
        }
        for( ; j < size.width; j += 3 )
        {
            int v = img[j];

            sum += v;
            sum2 += v * v;
        }
    }

    {
        double t = 3. / (size.width * size.height);
        double mean_val = sum * t;

        *mean = mean_val;
        *stddev = sqrt( sum2 * t - mean_val * mean_val );
    }

    return CV_NO_ERR;
}


/* /////////////// 32fC1 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMean_StdDev_32f_C1R, (const float *img, int step, CvSize size,
                                                 double *mean, double *stddev) )
{
    int i, j;
    double sum = 0, sum2 = 0;

    if( !img || (!mean && !stddev) )
        return CV_NULLPTR_ERR;
    if( size.width <= 0 || size.height <= 0 || step < size.width * 4 || (step & 3) != 0 )
        return CV_BADSIZE_ERR;

    step /= 4;

    if( size.width == step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( i = 0; i < size.height; i++, img += step )
    {
        for( j = 0; j <= size.width - 8; j += 8 )
        {
            {
                double v0 = img[j];
                double v1 = img[j + 1];
                double v2 = img[j + 2];
                double v3 = img[j + 3];

                sum2 += v0 * v0 + v1 * v1 + v2 * v2 + v3 * v3;
                sum += v0 + v1 + v2 + v3;

                v0 = img[j + 4];
                v1 = img[j + 5];
                v2 = img[j + 6];
                v3 = img[j + 7];

                sum2 += v0 * v0 + v1 * v1 + v2 * v2 + v3 * v3;
                sum += v0 + v1 + v2 + v3;
            }
        }
        for( ; j < size.width; j++ )
        {
            double v = img[j];

            sum += v;
            sum2 += v * v;
        }
    }

    {
        double t = 1. / (size.width * size.height);
        double mean_val = sum * t;

        *mean = mean_val;
        *stddev = sqrt( sum2 * t - mean_val * mean_val );
    }

    return CV_NO_ERR;
}

/* /////////////// 32fC3 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMean_StdDev_32f_C3CR, (const float *img, int step, CvSize size,
                                                  int coi, double *mean, double *stddev) )
{
    int i, j;
    double sum = 0, sum2 = 0;

    coi--;
    size.width *= 3;

    if( !img || (!mean && !stddev) )
        return CV_NULLPTR_ERR;
    if( size.width <= 0 || size.height <= 0 || step < size.width * 4 || (step & 3) != 0 )
        return CV_BADSIZE_ERR;

    if( (unsigned) coi > 2 )
        return CV_BADRANGE_ERR;
    img += coi;
    step /= 4;

    if( size.width == step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( i = 0; i < size.height; i++, img += step )
    {
        for( j = 0; j < size.width - 9; j += 12 )
        {
            double v0 = img[j];
            double v1 = img[j + 3];
            double v2 = img[j + 6];
            double v3 = img[j + 9];

            sum2 += v0 * v0 + v1 * v1 + v2 * v2 + v3 * v3;
            sum += v0 + v1 + v2 + v3;
        }
        for( ; j < size.width; j += 3 )
        {
            double v = img[j];

            sum += v;
            sum2 += v * v;
        }
    }

    {
        double t = 3. / (size.width * size.height);
        double mean_val = sum * t;

        *mean = mean_val;
        *stddev = sqrt( sum2 * t - mean_val * mean_val );
    }

    return CV_NO_ERR;
}


/****************************************************************************************\
*                      Mean and StdDev calculation with mask                             *
\****************************************************************************************/

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name: icvMean_StdDev_
//    Purpose:  Calculates mean value over the pixels in the selected region
//    Context:
//    Parameters:  
//      img  - source image
//      step - its step
//      mask - mask image ( non-zero values indicate region )
//      mask_step - its step
//      size - its size (width and height)
//      coi  - index of channel (if color image)
//      mean - pointer to mean value
//      stddev - pointer to standard deviation
//    Returns:
//      CV_NO_ERR or error code
//    Notes:
//F*/

/* /////////////// 8uC1 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMean_StdDev_8u_C1MR, (const uchar * img, int step,
                                                 const uchar * mask, int mask_step,
                                                 CvSize size, double *mean, double *stddev) )
{
    int i, j, pix = 0;
    int sum = 0;
    int64 sum2 = 0;

    if( !img || !mask || (!mean && !stddev) )
        return CV_NULLPTR_ERR;
    if( size.width <= 0 || size.height <= 0 || step < size.width || mask_step < size.width )
        return CV_BADSIZE_ERR;

    if( size.width == step && size.width == mask_step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( i = 0; i < size.height; i++, img += step, mask += mask_step )
    {
        for( j = 0; j < size.width - 3; j += 4 )
        {
            int m0 = (mask[j] == 0) - 1;
            int m1 = (mask[j + 1] == 0) - 1;
            int m2 = (mask[j + 2] == 0) - 1;
            int m3 = (mask[j + 3] == 0) - 1;

            pix -= m0 + m1 + m2 + m3;
            m0 &= img[j];
            m1 &= img[j + 1];
            m2 &= img[j + 2];
            m3 &= img[j + 3];

            sum2 += m0 * m0 + m1 * m1 + m2 * m2 + m3 * m3;
            sum += m0 + m1 + m2 + m3;
        }

        for( ; j < size.width; j++ )
        {
            int m = (mask[j] == 0) - 1;

            pix -= m;
            m &= img[j];
            sum += m;
            sum2 += m * m;
        }
    }

    {
        double t = pix ? 1. / pix : 0;
        double mean_val = sum * t;

        *mean = mean_val;
        *stddev = sqrt( sum2 * t - mean_val * mean_val );
    }

    return CV_NO_ERR;
}


/* /////////////// 8uC3 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMean_StdDev_8u_C3CMR, (const uchar * img, int step,
                                                  const uchar * mask, int mask_step,
                                                  CvSize size, int coi,
                                                  double *mean, double *stddev) )
{
    int i, j, pix = 0;
    int sum = 0;
    int64 sum2 = 0;

    coi--;

    if( !img || !mask || (!mean && !stddev) )
        return CV_NULLPTR_ERR;
    if( size.width <= 0 || size.height <= 0 ||
        step < size.width * 3 || mask_step < size.width )
        return CV_BADSIZE_ERR;

    if( (unsigned) coi > 2 )
        return CV_BADRANGE_ERR;
    img += coi;

    if( size.width * 3 == step && size.width == mask_step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( i = 0; i < size.height; i++, img += step, mask += mask_step )
    {
        for( j = 0; j < size.width - 3; j += 4 )
        {
            int m0 = (mask[j] == 0) - 1;
            int m1 = (mask[j + 1] == 0) - 1;
            int m2 = (mask[j + 2] == 0) - 1;
            int m3 = (mask[j + 3] == 0) - 1;

            pix -= m0 + m1 + m2 + m3;
            m0 &= img[j * 3];
            m1 &= img[j * 3 + 3];
            m2 &= img[j * 3 + 6];
            m3 &= img[j * 3 + 9];

            sum += m0 + m1 + m2 + m3;
            sum2 += m0 * m0 + m1 * m1 + m2 * m2 + m3 * m3;
        }

        for( ; j < size.width; j++ )
        {
            int m = (mask[j] == 0) - 1;

            pix -= m;
            m &= img[j * 3];
            sum += m;
            sum2 += m * m;
        }
    }

    {
        double t = pix ? 1. / pix : 0;
        double mean_val = sum * t;

        *mean = mean_val;
        *stddev = sqrt( sum2 * t - mean_val * mean_val );
    }

    return CV_NO_ERR;
}


/* /////////////// 8sC1 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMean_StdDev_8s_C1MR, (const char *img, int step,
                                                 const uchar * mask, int mask_step,
                                                 CvSize size, double *mean, double *stddev) )
{
    int i, j, pix = 0;
    int sum = 0;
    int64 sum2 = 0;

    if( !img || !mask || (!mean && !stddev) )
        return CV_NULLPTR_ERR;
    if( size.width <= 0 || size.height <= 0 || step < size.width || mask_step < size.width )
        return CV_BADSIZE_ERR;

    if( size.width == step && size.width == mask_step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( i = 0; i < size.height; i++, img += step, mask += mask_step )
    {
        for( j = 0; j < size.width - 3; j += 4 )
        {
            int m0 = (mask[j] == 0) - 1;
            int m1 = (mask[j + 1] == 0) - 1;
            int m2 = (mask[j + 2] == 0) - 1;
            int m3 = (mask[j + 3] == 0) - 1;

            pix -= m0 + m1 + m2 + m3;
            m0 &= img[j];
            m1 &= img[j + 1];
            m2 &= img[j + 2];
            m3 &= img[j + 3];

            sum2 += m0 * m0 + m1 * m1 + m2 * m2 + m3 * m3;
            sum += m0 + m1 + m2 + m3;
        }

        for( ; j < size.width; j++ )
        {
            int m = (mask[j] == 0) - 1;

            pix -= m;
            m &= img[j];
            sum += m;
            sum2 += m * m;
        }
    }

    {
        double t = pix ? 1. / pix : 0;
        double mean_val = sum * t;

        *mean = mean_val;
        *stddev = sqrt( sum2 * t - mean_val * mean_val );
    }

    return CV_NO_ERR;
}


/* /////////////// 8sC3 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMean_StdDev_8s_C3CMR, (const char *img, int step,
                                                  const uchar * mask, int mask_step,
                                                  CvSize size, int coi,
                                                  double *mean, double *stddev) )
{
    int i, j, pix = 0;
    int sum = 0;
    int64 sum2 = 0;

    coi--;

    if( !img || !mask || (!mean && !stddev) )
        return CV_NULLPTR_ERR;
    if( size.width <= 0 || size.height <= 0 ||
        step < size.width * 3 || mask_step < size.width )
        return CV_BADSIZE_ERR;

    if( (unsigned) coi > 2 )
        return CV_BADRANGE_ERR;
    img += coi;

    if( size.width * 3 == step && size.width == mask_step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( i = 0; i < size.height; i++, img += step, mask += mask_step )
    {
        for( j = 0; j < size.width - 3; j += 4 )
        {
            int m0 = (mask[j] == 0) - 1;
            int m1 = (mask[j + 1] == 0) - 1;
            int m2 = (mask[j + 2] == 0) - 1;
            int m3 = (mask[j + 3] == 0) - 1;

            pix -= m0 + m1 + m2 + m3;
            m0 &= img[j * 3];
            m1 &= img[j * 3 + 3];
            m2 &= img[j * 3 + 6];
            m3 &= img[j * 3 + 9];

            sum2 += m0 * m0 + m1 * m1 + m2 * m2 + m3 * m3;
            sum += m0 + m1 + m2 + m3;
        }

        for( ; j < size.width; j++ )
        {
            int m = (mask[j] == 0) - 1;

            pix -= m;
            m &= img[j * 3];
            sum += m;
            sum2 += m * m;
        }
    }

    {
        double t = pix ? 1. / pix : 0;
        double mean_val = sum * t;

        *mean = mean_val;
        *stddev = sqrt( sum2 * t - mean_val * mean_val );
    }

    return CV_NO_ERR;
}


/* /////////////// 32fC1 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMean_StdDev_32f_C1MR, (const float *img, int step,
                                                  const uchar * mask, int mask_step,
                                                  CvSize size, double *mean, double *stddev) )
{
    int i, j, pix = 0;
    double sum = 0, sum2 = 0;

    if( !img || !mask || (!mean && !stddev) )
        return CV_NULLPTR_ERR;
    if( size.width <= 0 || size.height <= 0 ||
        step < size.width * 4 || mask_step < size.width || (step & 3) != 0 )
        return CV_BADSIZE_ERR;

    step /= 4;

    if( size.width == step && size.width == mask_step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( i = 0; i < size.height; i++, img += step, mask += mask_step )
    {
        for( j = 0; j <= size.width - 8; j += 8 )
        {
            {
                float f0, f1, f2, f3;

                int m = (mask[j] == 0) - 1;

                pix -= m;
                *((int *) &f0) = ((int *) img)[j] & m;

                m = (mask[j + 1] == 0) - 1;
                pix -= m;
                *((int *) &f1) = ((int *) img)[j + 1] & m;

                m = (mask[j + 2] == 0) - 1;
                pix -= m;
                *((int *) &f2) = ((int *) img)[j + 2] & m;

                m = (mask[j + 3] == 0) - 1;
                pix -= m;
                *((int *) &f3) = ((int *) img)[j + 3] & m;

                sum2 += f0 * f0 + f1 * f1 + f2 * f2 + f3 * f3;
                sum += f0 + f1 + f2 + f3;

                m = (mask[j + 4] == 0) - 1;
                pix -= m;
                *((int *) &f0) = ((int *) img)[j + 4] & m;

                m = (mask[j + 5] == 0) - 1;
                pix -= m;
                *((int *) &f1) = ((int *) img)[j + 5] & m;

                m = (mask[j + 6] == 0) - 1;
                pix -= m;
                *((int *) &f2) = ((int *) img)[j + 6] & m;

                m = (mask[j + 7] == 0) - 1;
                pix -= m;
                *((int *) &f3) = ((int *) img)[j + 7] & m;

                sum2 += f0 * f0 + f1 * f1 + f2 * f2 + f3 * f3;
                sum += f0 + f1 + f2 + f3;
            }
        }

        for( ; j < size.width; j++ )
        {
            int m = (mask[j] == 0) - 1;
            double f;

            pix -= m;
            m &= ((int *) img)[j];
            f = *((float *) &m);

            sum += f;
            sum2 += f * f;
        }
    }

    {
        double t = pix ? 1. / pix : 0;
        double mean_val = sum * t;

        *mean = mean_val;
        *stddev = sqrt( sum2 * t - mean_val * mean_val );
    }

    return CV_NO_ERR;
}

/* /////////////// 32fC3 flavor //////////////////// */

IPCVAPI_IMPL( CvStatus, icvMean_StdDev_32f_C3CMR, (const float *img, int step,
                                                   const uchar * mask, int mask_step,
                                                   CvSize size, int coi,
                                                   double *mean, double *stddev) )
{
    int i, j, pix = 0;
    double sum = 0, sum2 = 0;

    coi--;

    if( !img || !mask || (!mean && !stddev) )
        return CV_NULLPTR_ERR;
    if( size.width <= 0 || size.height <= 0 ||
        step < size.width * 12 || mask_step < size.width || (step & 3) != 0 )
        return CV_BADSIZE_ERR;

    if( (unsigned) coi > 2 )
        return CV_BADRANGE_ERR;

    img += coi;
    step /= 4;

    if( size.width * 3 == step && size.width == mask_step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( i = 0; i < size.height; i++, img += step, mask += mask_step )
    {
        for( j = 0; j < size.width - 3; j += 4 )
        {
            float f0, f1, f2, f3;

            int m = (mask[j] == 0) - 1;

            pix -= m;
            *((int *) &f0) = ((int *) img)[j * 3] & m;

            m = (mask[j + 1] == 0) - 1;
            pix -= m;
            *((int *) &f1) = ((int *) img)[j * 3 + 3] & m;

            m = (mask[j + 2] == 0) - 1;
            pix -= m;
            *((int *) &f2) = ((int *) img)[j * 3 + 6] & m;

            m = (mask[j + 3] == 0) - 1;
            pix -= m;
            *((int *) &f3) = ((int *) img)[j * 3 + 9] & m;

            sum2 += f0 * f0 + f1 * f1 + f2 * f2 + f3 * f3;
            sum += f0 + f1 + f2 + f3;
        }

        for( ; j < size.width; j++ )
        {
            int m = (mask[j] == 0) - 1;
            double f;

            pix -= m;
            m &= ((int *) img)[j * 3];
            f = *((float *) &m);

            sum += f;
            sum2 += f * f;
        }
    }

    {
        double t = pix ? 1. / pix : 0;
        double mean_val = sum * t;

        *mean = mean_val;
        *stddev = sqrt( sum2 * t - mean_val * mean_val );
    }

    return CV_NO_ERR;
}

/*  End of file  */
