/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include "_cv.h"
#include <limits.h>
#include <float.h>
#include "_cvimgstat.h"

/****************************************************************************************\
*                                  MinMaxIndx_                                             *
\****************************************************************************************/

static CvStatus
icvWriteMinMaxResults( float min_val, float max_val,
                       int min_loc, int max_loc,
                       int width, float *minVal, float *maxVal,
                       CvPoint * minLoc, CvPoint * maxLoc )
{
    if( (min_loc | max_loc) < 0 )
    {
        min_loc = max_loc = 0;
        min_val = max_val = 0.f;
    }

    if( minVal )
        *minVal = min_val;
    if( maxVal )
        *maxVal = max_val;

    if( minLoc )
    {
        minLoc->y = min_loc / width;
        minLoc->x = min_loc - minLoc->y * width;
    }

    if( maxLoc )
    {
        maxLoc->y = max_loc / width;
        maxLoc->x = max_loc - maxLoc->y * width;
    }
    return CV_NO_ERR;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name: icvMinMaxIndx_
//    Purpose:  Finds minimum and maximum values in the image and their coordinates
//    Context:
//    Parameters:
//      img  - source image
//      step - its step
//      size - its size (width and height)
//      coi  - index of channel (if color image)
//      minVal - pointer to minimum value
//      maxVal - pointer to maximum value
//      minLoc - minimum coordinates
//      maxLoc - maximum coordinates
//    Returns:
//      CV_NO_ERR if success or error code
//    Notes:
//F*/

/* ////////////// 8uC1 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMinMaxIndx_8u_C1R, (const uchar * img, int step, CvSize size,
                                               float *minVal, float *maxVal,
                                               CvPoint * minLoc, CvPoint * maxLoc) )
{
    int min_val = 256, max_val = -256, min_loc = -1, max_loc = -1;
    int real_width = size.width;
    int i, j, k;

    if( !img || (!minVal && !maxVal && !minLoc && !maxLoc) )
        return CV_NULLPTR_ERR;

    if( size.width <= 0 || size.height <= 0 || step < size.width )
        return CV_BADSIZE_ERR;

    if( size.width == step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( i = 0, k = 0; i < size.height; i++, img += step )
    {
        for( j = 0; j < size.width; j++, k++ )
        {
            int val = img[j];

            if( min_val > val )
            {
                min_val = val;
                min_loc = k;
            }
            if( max_val < val )
            {
                max_val = val;
                max_loc = k;
            }
        }
    }

    return icvWriteMinMaxResults( (float) min_val, (float) max_val,
                                  min_loc, max_loc, real_width, minVal, maxVal, minLoc,
                                  maxLoc );
}


/* ////////////// 8uC3 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMinMaxIndx_8u_C3CR,
              (const uchar * img, int step, CvSize size, int coi, float *minVal, float *maxVal,
               CvPoint * minLoc, CvPoint * maxLoc) )
{
    int min_val = 256, max_val = -256, min_loc = -1, max_loc = -1;
    int real_width = size.width;
    int i, j, k;

    coi--;
    size.width *= 3;

    if( !img || (!minVal && !maxVal && !minLoc && !maxLoc) )
        return CV_NULLPTR_ERR;

    if( size.width <= 0 || size.height <= 0 || step < size.width )
        return CV_BADSIZE_ERR;

    if( (unsigned) coi > 2 )
        return CV_BADRANGE_ERR;

    if( size.width == step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    img += coi;
    for( i = 0, k = 0; i < size.height; i++, img += step )
    {
        for( j = 0; j < size.width; j += 3, k++ )
        {
            int val = img[j];

            if( min_val > val )
            {
                min_val = val;
                min_loc = k;
            }
            if( max_val < val )
            {
                max_val = val;
                max_loc = k;
            }
        }
    }

    return icvWriteMinMaxResults( (float) min_val, (float) max_val,
                                  min_loc, max_loc, real_width, minVal, maxVal, minLoc,
                                  maxLoc );
}


/* /////////////// 8sC1 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMinMaxIndx_8s_C1R, (const char *img, int step, CvSize size,
                                               float *minVal, float *maxVal,
                                               CvPoint * minLoc, CvPoint * maxLoc) )
{
    int min_val = 256, max_val = -256, min_loc = -1, max_loc = -1;
    int real_width = size.width;
    int i, j, k;

    if( !img || (!minVal && !maxVal && !minLoc && !maxLoc) )
        return CV_NULLPTR_ERR;

    if( size.width <= 0 || size.height <= 0 || step < size.width )
        return CV_BADSIZE_ERR;

    if( size.width == step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( i = 0, k = 0; i < size.height; i++, img += step )
    {
        for( j = 0; j < size.width; j++, k++ )
        {
            int val = img[j];

            if( min_val > val )
            {
                min_val = val;
                min_loc = k;
            }
            if( max_val < val )
            {
                max_val = val;
                max_loc = k;
            }
        }
    }

    return icvWriteMinMaxResults( (float) min_val, (float) max_val,
                                  min_loc, max_loc, real_width, minVal, maxVal, minLoc,
                                  maxLoc );
}


/* /////////////// 8sC3 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMinMaxIndx_8s_C3CR,
              (const char *img, int step, CvSize size, int coi, float *minVal, float *maxVal,
               CvPoint * minLoc, CvPoint * maxLoc) )
{
    int min_val = 256, max_val = -256, min_loc = -1, max_loc = -1;
    int real_width = size.width;
    int i, j, k;

    coi--;
    size.width *= 3;

    if( !img || (!minVal && !maxVal && !minLoc && !maxLoc) )
        return CV_NULLPTR_ERR;

    if( size.width <= 0 || size.height <= 0 || step < size.width )
        return CV_BADSIZE_ERR;

    if( (unsigned) coi > 2 )
        return CV_BADRANGE_ERR;

    if( size.width == step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    img += coi;
    for( i = 0, k = 0; i < size.height; i++, img += step )
    {
        for( j = 0; j < size.width; j += 3, k++ )
        {
            int val = img[j];

            if( min_val > val )
            {
                min_val = val;
                min_loc = k;
            }
            if( max_val < val )
            {
                max_val = val;
                max_loc = k;
            }
        }
    }

    return icvWriteMinMaxResults( (float) min_val, (float) max_val,
                                  min_loc, max_loc, real_width, minVal, maxVal, minLoc,
                                  maxLoc );
}


/* /////////////// 16sC1 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMinMaxIndx_16s_C1R, (const short *img, int step, CvSize size,
                                                float *minVal, float *maxVal,
                                                CvPoint * minLoc, CvPoint * maxLoc) )
{
    int min_val = INT_MAX, max_val = INT_MIN, min_loc = -1, max_loc = -1;
    int real_width = size.width;
    int i, j, k;

    if( !img || (!minVal && !maxVal && !minLoc && !maxLoc) )
        return CV_NULLPTR_ERR;

    if( size.width <= 0 || size.height <= 0 ||
        step < size.width || (step & (sizeof_short - 1)) != 0 )
        return CV_BADSIZE_ERR;

    step /= sizeof_short;

    if( size.width == step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( i = 0, k = 0; i < size.height; i++, img += step )
    {
        for( j = 0; j < size.width; j++, k++ )
        {
            int val = img[j];

            if( min_val > val )
            {
                min_val = val;
                min_loc = k;
            }
            if( max_val < val )
            {
                max_val = val;
                max_loc = k;
            }
        }
    }

    return icvWriteMinMaxResults( (float) min_val, (float) max_val,
                                  min_loc, max_loc, real_width, minVal, maxVal, minLoc,
                                  maxLoc );
}


/* /////////////// 16sC3 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMinMaxIndx_16s_C3CR,
              (const short *img, int step, CvSize size, int coi, float *minVal, float *maxVal,
               CvPoint * minLoc, CvPoint * maxLoc) )
{
    int min_val = INT_MAX, max_val = INT_MIN, min_loc = -1, max_loc = -1;
    int real_width = size.width;
    int i, j, k;

    coi--;
    size.width *= 3;

    if( !img || (!minVal && !maxVal && !minLoc && !maxLoc) )
        return CV_NULLPTR_ERR;

    if( size.width <= 0 || size.height <= 0 ||
        step < size.width || (step & (sizeof_short - 1)) != 0 )
        return CV_BADSIZE_ERR;

    if( (unsigned) coi > 2 )
        return CV_BADRANGE_ERR;

    step /= sizeof_short;

    if( size.width == step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    img += coi;
    for( i = 0, k = 0; i < size.height; i++, img += step )
    {
        for( j = 0; j < size.width; j += 3, k++ )
        {
            int val = img[j];

            if( min_val > val )
            {
                min_val = val;
                min_loc = k;
            }
            if( max_val < val )
            {
                max_val = val;
                max_loc = k;
            }
        }
    }

    return icvWriteMinMaxResults( (float) min_val, (float) max_val,
                                  min_loc, max_loc, real_width, minVal, maxVal, minLoc,
                                  maxLoc );
}



/* //////////// 32fC1 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMinMaxIndx_32f_C1R, (const float *img, int step, CvSize size,
                                                float *minVal, float *maxVal,
                                                CvPoint * minLoc, CvPoint * maxLoc) )
{
    int min_val = _CV_POS_INF, max_val = _CV_NEG_INF, min_loc = -1, max_loc = -1;
    int real_width = size.width;
    int i, j, k;

    if( !img || (!minVal && !maxVal && !minLoc && !maxLoc) )
        return CV_NULLPTR_ERR;

    if( size.width <= 0 || size.height <= 0 || step < size.width * 4 || (step & 3) != 0 )
        return CV_BADSIZE_ERR;

    step /= 4;

    if( size.width == step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( i = 0, k = 0; i < size.height; i++, img += step )
    {
        for( j = 0; j < size.width; j++, k++ )
        {
            int val = ((int *) img)[j];

            val = _CV_TOGGLE_FLT( val );
            if( val > max_val )
            {
                max_val = val;
                max_loc = k;
            }
            if( val < min_val )
            {
                min_val = val;
                min_loc = k;
            }
        }
    }

    min_val = _CV_TOGGLE_FLT( min_val );
    max_val = _CV_TOGGLE_FLT( max_val );

    return icvWriteMinMaxResults( *((float *) &min_val), *((float *) &max_val),
                                  min_loc, max_loc, real_width, minVal, maxVal, minLoc,
                                  maxLoc );
}

/* ///////////// 32fC3 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMinMaxIndx_32f_C3CR, (const float *img, int step,
                                                 CvSize size, int coi,
                                                 float *minVal, float *maxVal,
                                                 CvPoint * minLoc, CvPoint * maxLoc) )
{
    int min_val = _CV_POS_INF, max_val = _CV_NEG_INF, min_loc = -1, max_loc = -1;
    int real_width = size.width;
    int i, j, k;

    coi--;
    size.width *= 3;

    if( !img || (!minVal && !maxVal && !minLoc && !maxLoc) )
        return CV_NULLPTR_ERR;

    if( size.width <= 0 || size.height <= 0 || step < size.width * 4 || (step & 3) != 0 )
        return CV_BADSIZE_ERR;

    if( (unsigned) coi > 2 )
        return CV_BADRANGE_ERR;

    step /= 4;

    if( size.width == step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    img += coi;

    for( i = 0, k = 0; i < size.height; i++, img += step )
    {
        for( j = 0; j < size.width; j += 3, k++ )
        {
            int val = ((int *) img)[j];

            val = _CV_TOGGLE_FLT( val );
            if( val > max_val )
            {
                max_val = val;
                max_loc = k;
            }
            if( val < min_val )
            {
                min_val = val;
                min_loc = k;
            }
        }
    }

    min_val = _CV_TOGGLE_FLT( min_val );
    max_val = _CV_TOGGLE_FLT( max_val );

    return icvWriteMinMaxResults( *((float *) &min_val), *((float *) &max_val),
                                  min_loc, max_loc, real_width, minVal, maxVal, minLoc,
                                  maxLoc );
}

/****************************************************************************************\
*                              MinMaxIndx_ with mask                                       *
\****************************************************************************************/

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name: icvMinMaxIndx_
//    Purpose:  Finds minimum and maximum values in the selected 
//              region of the image and their coordinates
//    Context:
//    Parameters:
//      img  - source image
//      step - its step
//      mask - mask image ( non-zero values indicates region )
//      mask_step - its step
//      size - image size (width and height)
//      coi  - index of channel (if color image)
//      minVal - pointer to minimum value
//      maxVal - pointer to maximum value
//      minLoc - minimum coordinates
//      maxLoc - maximum coordinates
//    Returns:
//      CV_NO_ERR if success or error code
//    Notes:
//F*/

/* ///////////// 8uC1 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMinMaxIndx_8u_C1MR, (const uchar * img, int step,
                                                const uchar * mask, int mask_step,
                                                CvSize size, float *minVal, float *maxVal,
                                                CvPoint * minLoc, CvPoint * maxLoc) )
{
    int min_val = 256, max_val = -256, min_loc = -1, max_loc = -1;
    int real_width = size.width;
    int i, j, k;

    if( !img || !mask || (!minVal && !maxVal && !minLoc && !maxLoc) )
        return CV_NULLPTR_ERR;

    if( size.width <= 0 || size.height <= 0 || step < size.width || mask_step < size.width )
        return CV_BADSIZE_ERR;

    if( size.width == step && size.width == mask_step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( i = 0, k = 0; i < size.height; i++, img += step, mask += mask_step )
    {
        for( j = 0; j < size.width; j++, k++ )
        {
            int val = img[j], val1;
            int mask_mask = (mask[j] != 0) - 1;

            val1 = val + (1024 & mask_mask);
            val -= (1024 & mask_mask);

            if( val1 < min_val )
            {
                min_val = val1;
                min_loc = k;
            }

            if( val > max_val )
            {
                max_val = val;
                max_loc = k;
            }
        }
    }

    return icvWriteMinMaxResults( (float) min_val, (float) max_val,
                                  min_loc, max_loc, real_width, minVal, maxVal, minLoc,
                                  maxLoc );
}


/* ///////////// 8uC3 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMinMaxIndx_8u_C3CMR, (const uchar * img, int step,
                                                 const uchar * mask, int mask_step,
                                                 CvSize size, int coi,
                                                 float *minVal, float *maxVal,
                                                 CvPoint * minLoc, CvPoint * maxLoc) )
{
    int min_val = 256, max_val = -256, min_loc = -1, max_loc = -1;
    int real_width = size.width;
    int i, j, k;

    coi--;

    if( !img || !mask || (!minVal && !maxVal && !minLoc && !maxLoc) )
        return CV_NULLPTR_ERR;

    if( size.width <= 0 || size.height <= 0 ||
        step < size.width * 3 || mask_step < size.width )
        return CV_BADSIZE_ERR;

    if( (unsigned) coi > 2 )
        return CV_BADRANGE_ERR;

    if( size.width * 3 == step && size.width == mask_step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    img += coi;

    for( i = 0, k = 0; i < size.height; i++, img += step, mask += mask_step )
    {
        for( j = 0; j < size.width; j++, k++ )
        {
            int val = img[j * 3], val1;
            int mask_mask = (mask[j] != 0) - 1;

            val1 = val + (1024 & mask_mask);
            val -= (1024 & mask_mask);

            if( val1 < min_val )
            {
                min_val = val1;
                min_loc = k;
            }

            if( val > max_val )
            {
                max_val = val;
                max_loc = k;
            }
        }
    }

    return icvWriteMinMaxResults( (float) min_val, (float) max_val,
                                  min_loc, max_loc, real_width, minVal, maxVal, minLoc,
                                  maxLoc );
}


/* ///////////// 8sC1 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMinMaxIndx_8s_C1MR, (const char *img, int step,
                                                const uchar * mask, int mask_step,
                                                CvSize size, float *minVal, float *maxVal,
                                                CvPoint * minLoc, CvPoint * maxLoc) )
{
    int min_val = 256, max_val = -256, min_loc = -1, max_loc = -1;
    int real_width = size.width;
    int i, j, k;

    if( !img || !mask || (!minVal && !maxVal && !minLoc && !maxLoc) )
        return CV_NULLPTR_ERR;

    if( size.width <= 0 || size.height <= 0 || step < size.width )
        return CV_BADSIZE_ERR;

    if( size.width == step && size.width == mask_step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( i = 0, k = 0; i < size.height; i++, img += step, mask += mask_step )
    {
        for( j = 0; j < size.width; j++, k++ )
        {
            int val = img[j], val1;
            int mask_mask = (mask[j] != 0) - 1;

            val1 = val + (1024 & mask_mask);
            val -= (1024 & mask_mask);

            if( val1 < min_val )
            {
                min_val = val1;
                min_loc = k;
            }

            if( val > max_val )
            {
                max_val = val;
                max_loc = k;
            }
        }
    }

    return icvWriteMinMaxResults( (float) min_val, (float) max_val,
                                  min_loc, max_loc, real_width, minVal, maxVal, minLoc,
                                  maxLoc );
}


/* ///////////// 8sC3 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMinMaxIndx_8s_C3CMR, (const char *img, int step,
                                                 const uchar * mask, int mask_step,
                                                 CvSize size, int coi, float *minVal,
                                                 float *maxVal, CvPoint * minLoc,
                                                 CvPoint * maxLoc) )
{
    int min_val = 256, max_val = -256, min_loc = -1, max_loc = -1;
    int real_width = size.width;
    int i, j, k;

    coi--;

    if( !img || !mask || (!minVal && !maxVal && !minLoc && !maxLoc) )
        return CV_NULLPTR_ERR;

    if( size.width <= 0 || size.height <= 0 ||
        step < size.width * 3 || mask_step < size.width )
        return CV_BADSIZE_ERR;

    if( (unsigned) coi > 2 )
        return CV_BADRANGE_ERR;

    if( size.width * 3 == step && size.width == mask_step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    img += coi;
    for( i = 0, k = 0; i < size.height; i++, img += step, mask += mask_step )
    {
        for( j = 0; j < size.width; j++, k++ )
        {
            int val = img[j * 3], val1;
            int mask_mask = (mask[j] != 0) - 1;

            val1 = val + (1024 & mask_mask);
            val -= (1024 & mask_mask);

            if( val1 < min_val )
            {
                min_val = val1;
                min_loc = k;
            }

            if( val > max_val )
            {
                max_val = val;
                max_loc = k;
            }
        }
    }

    return icvWriteMinMaxResults( (float) min_val, (float) max_val,
                                  min_loc, max_loc, real_width, minVal, maxVal, minLoc,
                                  maxLoc );
}


/* ///////////// 32fC1 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMinMaxIndx_32f_C1MR, (const float *img, int step,
                                                 const uchar * mask, int mask_step,
                                                 CvSize size, float *minVal, float *maxVal,
                                                 CvPoint * minLoc, CvPoint * maxLoc) )
{
    int min_val = INT_MAX, max_val = INT_MIN, min_loc = -1, max_loc = -1;
    int real_width = size.width;
    int i, j, k;

    if( !img || !mask || (!minVal && !maxVal && !minLoc && !maxLoc) )
        return CV_NULLPTR_ERR;

    if( size.width <= 0 || size.height <= 0 || step < size.width * 4 || (step & 3) != 0 )
        return CV_BADSIZE_ERR;

    step /= 4;

    if( size.width * 4 == step && size.width == mask_step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( i = 0, k = 0; i < size.height; i++, img += step, mask += mask_step )
    {
        for( j = 0; j < size.width; j++, k++ )
        {
            int val = ((int *) img)[j], val1;
            int mask_mask = (mask[j] != 0) - 1;

            val = _CV_TOGGLE_FLT( val );
            val1 = val + ((INT_MAX - val) & mask_mask);
            val += (INT_MIN - val) & mask_mask;

            if( val1 < min_val )
            {
                min_val = val1;
                min_loc = k;
            }

            if( val > max_val )
            {
                max_val = val;
                max_loc = k;
            }
        }
    }

    min_val = _CV_TOGGLE_FLT( min_val );
    max_val = _CV_TOGGLE_FLT( max_val );

    return icvWriteMinMaxResults( *((float *) &min_val), *((float *) &max_val),
                                  min_loc, max_loc, real_width, minVal, maxVal, minLoc,
                                  maxLoc );
}


/* ///////////// 32fC3 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMinMaxIndx_32f_C3CMR, (const float *img, int step,
                                                  const uchar * mask, int mask_step,
                                                  CvSize size, int coi, float *minVal,
                                                  float *maxVal, CvPoint * minLoc,
                                                  CvPoint * maxLoc) )
{
    int min_val = INT_MAX, max_val = INT_MIN, min_loc = -1, max_loc = -1;
    int real_width = size.width;
    int i, j, k;

    coi--;

    if( !img || !mask || (!minVal && !maxVal && !minLoc && !maxLoc) )
        return CV_NULLPTR_ERR;

    if( size.width <= 0 || size.height <= 0 ||
        step < size.width * 12 || (step & 3) != 0 || mask_step < size.width )
        return CV_BADSIZE_ERR;

    if( (unsigned) coi > 2 )
        return CV_BADRANGE_ERR;

    step /= 4;
    img += coi;

    if( size.width * 12 == step && size.width == mask_step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( i = 0, k = 0; i < size.height; i++, img += step, mask += mask_step )
    {
        for( j = 0; j < size.width; j++, k++ )
        {
            int val = ((int *) img)[j * 3], val1;
            int mask_mask = (mask[j] != 0) - 1;

            val = _CV_TOGGLE_FLT( val );
            val1 = val + ((INT_MAX - val) & mask_mask);
            val += (INT_MIN - val) & mask_mask;

            if( val1 < min_val )
            {
                min_val = val1;
                min_loc = k;
            }

            if( val > max_val )
            {
                max_val = val;
                max_loc = k;
            }
        }
    }

    min_val = _CV_TOGGLE_FLT( min_val );
    max_val = _CV_TOGGLE_FLT( max_val );

    return icvWriteMinMaxResults( *((float *) &min_val), *((float *) &max_val),
                                  min_loc, max_loc, real_width, minVal, maxVal, minLoc,
                                  maxLoc );
}

/*  End of file  */
