/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/
#include "_cv.h"
#include "_cvwrap.h"
#include "_cvimgstat.h"
#include <float.h>

/* The function calculates center of gravity and central second order moments */
CvStatus
icvCompleteMomentState( CvMoments * moments )
{
    double cx = 0, cy = 0;
    double mu20, mu11, mu02;

    assert( moments != 0 );
    moments->inv_sqrt_m00 = 0;

    if( moments->m00 != 0 )
    {
        double inv_m00 = 1. / moments->m00;

        cx = moments->m10 * inv_m00;
        cy = moments->m01 * inv_m00;
        moments->inv_sqrt_m00 = sqrt( inv_m00 );
    }

    /* mu20 = m20 - m10*cx */
    mu20 = moments->m20 - moments->m10 * cx;
    /* mu11 = m11 - m10*cy */
    mu11 = moments->m11 - moments->m10 * cy;
    /* mu02 = m02 - m01*cy */
    mu02 = moments->m02 - moments->m01 * cy;

    moments->mu20 = mu20;
    moments->mu11 = mu11;
    moments->mu02 = mu02;

    /* mu30 = m30 - cx*(3*mu20 + cx*m10) */
    moments->mu30 = moments->m30 - cx * (3 * mu20 + cx * moments->m10);
    mu11 += mu11;
    /* mu21 = m21 - cx*(2*mu11 + cx*m01) - cy*mu20 */
    moments->mu21 = moments->m21 - cx * (mu11 + cx * moments->m01) - cy * mu20;
    /* mu12 = m12 - cy*(2*mu11 + cy*m10) - cx*mu02 */
    moments->mu12 = moments->m12 - cy * (mu11 + cy * moments->m10) - cx * mu02;
    /* mu03 = m03 - cy*(3*mu02 + cy*m01) */
    moments->mu03 = moments->m03 - cy * (3 * mu02 + cy * moments->m01);

    return CV_NO_ERR;
}


/* summarizes moment values for all tiles */
static CvStatus
icvCompleteMomentState_int( int *tiles, CvSize size, int tile_size, CvMoments * moments )
{
    int x, y;
    memset( &(moments->m00), 0, 10 * sizeof( double ));

    for( y = 0; y < size.height; y += tile_size )
    {
        for( x = 0; x < size.width; x += tile_size, tiles += 10 )
        {
            double dx = x, dy = y;
            double dxm = dx * tiles[0], dym = dy * tiles[0];

            /* + m00 ( = m00' ) */
            moments->m00 += tiles[0];

            /* + m10 ( = m10' + dx*m00' ) */
            moments->m10 += tiles[1] + dxm;

            /* + m01 ( = m01' + dy*m00' ) */
            moments->m01 += tiles[2] + dym;

            /* + m20 ( = m20' + 2*dx*m10' + dx*dx*m00' ) */
            moments->m20 += tiles[3] + dx * (tiles[1] * 2 + dxm);

            /* + m11 ( = m11' + dx*m01' + dy*m10' + dx*dy*m00' ) */
            moments->m11 += tiles[4] + dx * (tiles[2] + dym) + dy * tiles[1];

            /* + m02 ( = m02' + 2*dy*m01' + dy*dy*m00' ) */
            moments->m02 += tiles[5] + dy * (tiles[2] * 2 + dym);

            /* + m30 ( = m30' + 3*dx*m20' + 3*dx*dx*m10' + dx*dx*dx*m00' ) */
            moments->m30 += tiles[6] + dx * (3. * tiles[3] + dx * (3. * tiles[1] + dxm));

            /* + m21 (= m21' + dx*(2*m11' + 2*dy*m10' + dx*m01' + dx*dy*m00') + dy*m20') */
            moments->m21 += tiles[7] + dx * (2 * (tiles[4] + dy * tiles[1]) +
                                             dx * (tiles[2] + dym)) + dy * tiles[3];

            /* + m12 (= m12' + dy*(2*m11' + 2*dx*m01' + dy*m10' + dx*dy*m00') + dx*m02') */
            moments->m12 += tiles[8] + dy * (2 * (tiles[4] + dx * tiles[2]) +
                                             dy * (tiles[1] + dxm)) + dx * tiles[5];

            /* + m03 ( = m03' + 3*dy*m02' + 3*dy*dy*m01' + dy*dy*dy*m00' ) */
            moments->m03 += tiles[9] + dy * (3. * tiles[5] + dy * (3. * tiles[2] + dym));
        }
    }

    return icvCompleteMomentState( moments );
}


/****************************************************************************************\
*                             Spatial Moments caluclating                                *
\****************************************************************************************/

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name: icvMoments
//    Purpose:  Find all neccessary values for calculating spatial/central/normalized
//              moments up to third order
//    Context:
//    Parameters:
//      img  - source image
//      step - its step
//      size - its size (width and height)
//      coi  - index of channel (if color image)
//      moments - structure containing resultant values.
//                The structure is input for icvGetSpatialMoment, 
//                icvGetCentralMoment and icvGetNormalizedCentralMoment functions
//    Returns:
//      CV_NO_ERR if success or error code
//    Notes:
//F*/

static void
icvCalcMomentsInTile8uC1( uchar * img, int step, CvSize size, int *mom )
{
    int x, y, sx_init = (size.width & -4) * (size.width & -4), sy = 0;

    assert( img && step >= size.width && size.width && (size.width | size.height) >= 0 );
    memset( mom, 0, 10 * sizeof( mom[0] ));

    for( y = 0; y < size.height; sy += 2 * y + 1, y++, img += step )
    {
        int x0 = 0;
        int x1 = 0;
        int x2 = 0;
        int x3 = 0;
        int sx = sx_init;

        for( x = 0; x < size.width - 3; x += 4 )
        {
            int p0 = img[x], p1 = img[x + 1], p2 = img[x + 2], p3 = img[x + 3];
            int t = p1;
            int a, b, c;

            p0 += p1 + p2 + p3; /* p0 + p1 + p2 + p3 */
            p1 += 2 * p2 + 3 * p3;      /* p1 + p2*2 + p3*3 */
            p2 = p1 + 2 * p2 + 6 * p3;  /* p1 + p2*4 + p3*9 */
            p3 = 2 * p2 - t + 9 * p3;   /* p1 + p2*8 + p3*27 */

            a = x * p0 + p1;    /* x*p0 + (x+1)*p1 + (x+2)*p2 + (x+3)*p3 */
            b = x * p1 + p2;    /* (x+1)*p1 + 2*(x+2)*p2 + 3*(x+3)*p3 */
            c = x * p2 + p3;    /* (x+1)*p1 + 4*(x+2)*p2 + 9*(x+3)*p3 */

            x0 += p0;
            x1 += a;
            a = a * x + b;      /*(x^2)*p0+((x+1)^2)*p1+((x+2)^2)*p2+((x+3)^2)*p3 */
            x2 += a;
            x3 += (a + b) * x + c;      /*x3 += (x^3)*p0+((x+1)^3)*p1+((x+2)^3)*p2+((x+3)^3)*p3 */
        }

        /* process the rest */
        for( ; x < size.width; sx += 2 * x + 1, x++ )
        {
            int p = img[x];
            int xp = x * p;

            x0 += p;
            x1 += xp;
            x2 += sx * p;
            x3 += sx * xp;
        }

        {
            int py = y * x0;

            mom[9] += py * sy;  /* m03 */
            mom[8] += x1 * sy;  /* m12 */
            mom[7] += x2 * y;   /* m21 */
            mom[6] += x3;       /* m30 */
            mom[5] += x0 * sy;  /* m02 */
            mom[4] += x1 * y;   /* m11 */
            mom[3] += x2;       /* m20 */
            mom[2] += py;       /* m01 */
            mom[1] += x1;       /* m10 */
            mom[0] += x0;       /* m00 */
        }
    }
}


static void
icvCalcMomentsInTile8sC1( char *img, int step, CvSize size, int *mom )
{
    int x, y, sx_init = (size.width & -4) * (size.width & -4), sy = 0;

    assert( img && step >= size.width && size.width && (size.width | size.height) >= 0 );
    memset( mom, 0, 10 * sizeof( mom[0] ));

    for( y = 0; y < size.height; sy += 2 * y + 1, y++, img += step )
    {
        int x0 = 0;
        int x1 = 0;
        int x2 = 0;
        int x3 = 0;
        int sx = sx_init;

        for( x = 0; x < size.width - 3; x += 4 )
        {
            int p0 = img[x], p1 = img[x + 1], p2 = img[x + 2], p3 = img[x + 3];
            int t = p1;
            int a, b, c;

            p0 += p1 + p2 + p3; /* p0 + p1 + p2 + p3 */
            p1 += 2 * p2 + 3 * p3;      /* p1 + p2*2 + p3*3 */
            p2 = p1 + 2 * p2 + 6 * p3;  /* p1 + p2*4 + p3*9 */
            p3 = 2 * p2 - t + 9 * p3;   /* p1 + p2*8 + p3*27 */

            a = x * p0 + p1;    /* x*p0 + (x+1)*p1 + (x+2)*p2 + (x+3)*p3 */
            b = x * p1 + p2;    /* (x+1)*p1 + 2*(x+2)*p2 + 3*(x+3)*p3 */
            c = x * p2 + p3;    /* (x+1)*p1 + 4*(x+2)*p2 + 9*(x+3)*p3 */

            x0 += p0;
            x1 += a;
            a = a * x + b;      /* (x^2)*p0+((x+1)^2)*p1+((x+2)^2)*p2+((x+3)^2)*p3 */
            x2 += a;
            x3 += (a + b) * x + c;      /*x3 += (x^3)*p0+((x+1)^3)*p1+((x+2)^3)*p2+((x+3)^3)*p3 */
        }

        /* process the rest */
        for( ; x < size.width; sx += 2 * x + 1, x++ )
        {
            int p = img[x];
            int xp = x * p;

            x0 += p;
            x1 += xp;
            x2 += sx * p;
            x3 += sx * xp;
        }

        {
            int py = y * x0;

            mom[9] += py * sy;  /* m03 */
            mom[8] += x1 * sy;  /* m12 */
            mom[7] += x2 * y;   /* m21 */
            mom[6] += x3;       /* m30 */
            mom[5] += x0 * sy;  /* m02 */
            mom[4] += x1 * y;   /* m11 */
            mom[3] += x2;       /* m20 */
            mom[2] += py;       /* m01 */
            mom[1] += x1;       /* m10 */
            mom[0] += x0;       /* m00 */
        }
    }
}


/* ////////////// 8uC1 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMomentsGray_8u_C1R, (const uchar * img,
                                                int step, CvSize size, CvMoments * moments) )
{
    const int tile_size = 32;
    int x, y, k;
    int local_buffer[8192];
    int *tiles = local_buffer;
    int tiles_buf_size;
    CvStatus status;

    if( !img || !moments )
        return CV_NULLPTR_ERR;
    if( (size.width | size.height) < 0 || step < size.width )
        return CV_BADSIZE_ERR;

    memset( moments, 0, sizeof( *moments ));

    tiles_buf_size = ((size.width + tile_size - 1) / tile_size) *
        ((size.height + tile_size - 1) / tile_size) * 10 * sizeof( int );

    if( tiles_buf_size > sizeof( local_buffer ))
    {
        tiles = (int *) icvAlloc( tiles_buf_size );
        if( !tiles )
            return CV_OUTOFMEM_ERR;
    }

    for( y = 0, k = 0; y < size.height; y += tile_size, img += tile_size * step )
    {
        for( x = 0; x < size.width; x += tile_size, k += 10 )
        {
            CvSize sz;

            sz.width = size.width - x;
            sz.height = size.height - y;

            sz.width = _CV_MIN2( sz.width, tile_size );
            sz.height = _CV_MIN2( sz.height, tile_size );

            icvCalcMomentsInTile8uC1( (uchar *) (img + x), step, sz, tiles + k );
        }
    }

    status = icvCompleteMomentState_int( tiles, size, tile_size, moments );
    if( tiles != local_buffer )
        icvFree( &tiles );

    return status;
}


/* /////////////// 8uC3 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMomentsGray_8u_C3CR, (const uchar * img, int step, CvSize size,
                                                 int coi, CvMoments * moments) )
{
    int x, y, sx_init = (size.width & -4) * (size.width & -4), sy = 0, sum = 0;
    double mom[9];

    coi--;

    if( !img || !moments )
        return CV_NULLPTR_ERR;
    if( (size.width | size.height) < 0 || step < size.width * 3 )
        return CV_BADSIZE_ERR;
    if( (unsigned) coi > 2 )
        return CV_BADRANGE_ERR;

    memset( &mom, 0, sizeof( mom ));
    img += coi;

    for( y = 0; y < size.height; sy += 2 * y + 1, y++, img += step )
    {
        double x2 = 0;
        double x3 = 0;
        int x0 = 0;
        int x1 = 0;
        int sx = sx_init;

        for( x = 0; x < size.width - 3; x += 4 )
        {
            int p0 = img[x * 3], p1 = img[x * 3 + 3], p2 = img[x * 3 + 6], p3 = img[x * 3 + 9];
            int t = p1;
            int _a;
            double a, b, c;

            p0 += p1 + p2 + p3; /* p0 + p1 + p2 + p3 */
            p1 += 2 * p2 + 3 * p3;      /* p1 + p2*2 + p3*3 */
            p2 = p1 + 2 * p2 + 6 * p3;  /* p1 + p2*4 + p3*9 */
            p3 = 2 * p2 - t + 9 * p3;   /* p1 + p2*8 + p3*27 */

            _a = x * p0 + p1;   /* x*p0 + (x+1)*p1 + (x+2)*p2 + (x+3)*p3 */
            b = x * p1 + p2;    /* (x+1)*p1 + 2*(x+2)*p2 + 3*(x+3)*p3 */
            c = x * p2 + p3;    /* (x+1)*p1 + 4*(x+2)*p2 + 9*(x+3)*p3 */

            x0 += p0;
            x1 += _a;
            a = ((double) _a) * x + b;  /* (x^2)*p0+((x+1)^2)*p1+((x+2)^2)*p2+((x+3)^2)*p3 */
            x2 += a;
            x3 += (a + b) * x + c;      /* x3 += (x^3)*p0+((x+1)^3)*p1+((x+2)^3)*p2+((x+3)^3)*p3 */
        }

        /* process the rest */
        for( ; x < size.width; sx += 2 * x + 1, x++ )
        {
            int p = img[x * 3];
            int xp = x * p;
            double _x2 = sx;

            x0 += p;
            x1 += xp;
            x2 += _x2 * p;
            x3 += _x2 * xp;
        }

        {
            double _y = y;
            double _py = _y * x0;
            double _y2 = sy;

            sum += x0;

            mom[8] += _py * _y2;        /* m03 */
            mom[7] += x1 * _y2; /* m12 */
            mom[6] += x2 * _y;  /* m21 */
            mom[5] += x3;       /* m30 */
            mom[4] += x0 * _y2; /* m02 */
            mom[3] += x1 * _y;  /* m11 */
            mom[2] += x2;       /* m20 */
            mom[1] += _py;      /* m01 */
            mom[0] += x1;       /* m10 */
        }
    }

    moments->m00 = sum;
    memcpy( &(moments->m00) + 1, mom, 9 * sizeof( double ));

    return icvCompleteMomentState( moments );
}

/* /////////////// 8sC1 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMomentsGray_8s_C1R, (const char *img,
                                                int step, CvSize size, CvMoments * moments) )
{
    const int tile_size = 32;
    int x, y, k;
    int local_buffer[8192];
    int *tiles = local_buffer;
    int tiles_buf_size;
    CvStatus status;

    if( !img || !moments )
        return CV_NULLPTR_ERR;
    if( (size.width | size.height) < 0 || step < size.width )
        return CV_BADSIZE_ERR;

    memset( moments, 0, sizeof( *moments ));

    tiles_buf_size = ((size.width + tile_size - 1) / tile_size) *
        ((size.height + tile_size - 1) / tile_size) * 10 * sizeof( int );

    if( tiles_buf_size > sizeof( local_buffer ))
    {
        tiles = (int *) icvAlloc( tiles_buf_size );
        if( !tiles )
            return CV_OUTOFMEM_ERR;
    }

    for( y = 0, k = 0; y < size.height; y += tile_size, img += tile_size * step )
    {
        for( x = 0; x < size.width; x += tile_size, k += 10 )
        {
            CvSize sz;

            sz.width = size.width - x;
            sz.height = size.height - y;

            sz.width = _CV_MIN2( sz.width, tile_size );
            sz.height = _CV_MIN2( sz.height, tile_size );

            icvCalcMomentsInTile8sC1( (char *) (img + x), step, sz, tiles + k );
        }
    }

    status = icvCompleteMomentState_int( tiles, size, tile_size, moments );
    if( tiles != local_buffer )
        icvFree( &tiles );

    return status;
}


/* /////////////// 8sC3 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMomentsGray_8s_C3CR, (const char *img, int step, CvSize size,
                                                 int coi, CvMoments * moments) )
{
    int x, y, sx_init = (size.width & -4) * (size.width & -4), sy = 0, sum = 0;
    double mom[9];

    coi--;

    if( !img || !moments )
        return CV_NULLPTR_ERR;
    if( (size.width | size.height) < 0 || step < size.width * 3 )
        return CV_BADSIZE_ERR;
    if( (unsigned) coi > 2 )
        return CV_BADRANGE_ERR;

    memset( &mom, 0, sizeof( mom ));
    img += coi;

    for( y = 0; y < size.height; sy += 2 * y + 1, y++, img += step )
    {
        double x2 = 0;
        double x3 = 0;
        int x0 = 0;
        int x1 = 0;
        int sx = sx_init;

        for( x = 0; x < size.width - 3; x += 4 )
        {
            int p0 = img[x * 3], p1 = img[x * 3 + 3], p2 = img[x * 3 + 6], p3 = img[x * 3 + 9];
            int t = p1;
            int _a;
            double a, b, c;

            p0 += p1 + p2 + p3; /* p0 + p1 + p2 + p3 */
            p1 += 2 * p2 + 3 * p3;      /* p1 + p2*2 + p3*3 */
            p2 = p1 + 2 * p2 + 6 * p3;  /* p1 + p2*4 + p3*9 */
            p3 = 2 * p2 - t + 9 * p3;   /* p1 + p2*8 + p3*27 */

            _a = x * p0 + p1;   /* x*p0 + (x+1)*p1 + (x+2)*p2 + (x+3)*p3 */
            b = x * p1 + p2;    /* (x+1)*p1 + 2*(x+2)*p2 + 3*(x+3)*p3 */
            c = x * p2 + p3;    /* (x+1)*p1 + 4*(x+2)*p2 + 9*(x+3)*p3 */

            x0 += p0;
            x1 += _a;
            a = ((double) _a) * x + b;  /* (x^2)*p0+((x+1)^2)*p1+((x+2)^2)*p2+((x+3)^2)*p3 */
            x2 += a;
            x3 += (a + b) * x + c;      /* x3 += (x^3)*p0+((x+1)^3)*p1+((x+2)^3)*p2+((x+3)^3)*p3 */
        }

        /* process the rest */
        for( ; x < size.width; sx += 2 * x + 1, x++ )
        {
            int p = img[x * 3];
            int xp = x * p;
            double _x2 = sx;

            x0 += p;
            x1 += xp;
            x2 += _x2 * p;
            x3 += _x2 * xp;
        }

        {
            double _y = y;
            double _py = _y * x0;
            double _y2 = sy;

            sum += x0;

            mom[8] += _py * _y2;        /* m03 */
            mom[7] += x1 * _y2; /* m12 */
            mom[6] += x2 * _y;  /* m21 */
            mom[5] += x3;       /* m30 */
            mom[4] += x0 * _y2; /* m02 */
            mom[3] += x1 * _y;  /* m11 */
            mom[2] += x2;       /* m20 */
            mom[1] += _py;      /* m01 */
            mom[0] += x1;       /* m10 */
        }
    }

    moments->m00 = sum;
    memcpy( &(moments->m00) + 1, mom, 9 * sizeof( double ));

    return icvCompleteMomentState( moments );
}

/* /////////////// 32fC1 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMomentsGray_32f_C1R, (const float *img,
                                                 int step, CvSize size, CvMoments * moments) )
{
    int x, y, sx_init = (size.width & -4) * (size.width & -4), sy = 0;
    double mom[10];

    if( !img || !moments )
        return CV_NULLPTR_ERR;
    if( (size.width | size.height) < 0 || step < size.width * 4 || (step & 3) != 0 )
        return CV_BADSIZE_ERR;

    memset( &mom, 0, sizeof( mom ));
    step /= 4;

    for( y = 0; y < size.height; sy += 2 * y + 1, y++, img += step )
    {
        double x2 = 0;
        double x3 = 0;
        double x1 = 0;
        double x0 = 0;
        int sx = sx_init;

        for( x = 0; x < size.width - 3; x += 4 )
        {
            double a, b, c, d;

            a = img[x] + img[x + 1] + img[x + 2] + img[x + 3];  /* p0 + p1 + p2 + p3 */
            b = img[x + 1] + 2 * img[x + 2] + 3 * img[x + 3];   /* p1 + p2*2 + p3*3 */
            c = img[x + 1] + 4 * img[x + 2] + 9 * img[x + 3];   /* p1 + p2*4 + p3*9 */
            d = img[x + 1] + 8 * img[x + 2] + 27 * img[x + 3];  /* p1 + p2*8 + p3*27 */

            x0 += a;

            a = a * x + b;      /* x*p0 + (x+1)*p1 + (x+2)*p2 + (x+3)*p3 */
            b = b * x + c;      /* (x+1)*p1 + 2*(x+2)*p2 + 3*(x+3)*p3 */
            c = c * x + d;      /* (x+1)*p1 + 4*(x+2)*p2 + 9*(x+3)*p3 */

            x1 += a;
            a = a * x + b;      /* (x^2)*p0+((x+1)^2)*p1+((x+2)^2)*p2+((x+3)^2)*p3 */
            x2 += a;
            x3 += (a + b) * x + c;      /* x3 += (x^3)*p0+((x+1)^3)*p1+((x+2)^3)*p2+((x+3)^3)*p3 */
        }

        /* process the rest */
        for( ; x < size.width; sx += 2 * x + 1, x++ )
        {
            double p = img[x];
            double xp = x * p;
            double _x2 = sx;

            x0 += p;
            x1 += xp;
            x2 += _x2 * p;
            x3 += _x2 * xp;
        }

        {
            double _y = y;
            double _py = _y * x0;
            double _y2 = sy;

            mom[9] += _py * _y2;        /* m03 */
            mom[8] += x1 * _y2; /* m12 */
            mom[7] += x2 * _y;  /* m21 */
            mom[6] += x3;       /* m30 */
            mom[5] += x0 * _y2; /* m02 */
            mom[4] += x1 * _y;  /* m11 */
            mom[3] += x2;       /* m20 */
            mom[2] += _py;      /* m01 */
            mom[1] += x1;       /* m10 */
            mom[0] += x0;       /* m00 */
        }
    }

    memcpy( &(moments->m00), mom, 10 * sizeof( double ));

    return icvCompleteMomentState( moments );
}


/* /////////////// 32fC3 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMomentsGray_32f_C3CR, (const float *img, int step, CvSize size,
                                                  int coi, CvMoments * moments) )
{
    int x, y, sx_init = (size.width & -4) * (size.width & -4), sy = 0;
    double mom[10];

    coi--;

    if( !img || !moments )
        return CV_NULLPTR_ERR;
    if( (size.width | size.height) < 0 || step < size.width * 12 || (step & 3) != 0 )
        return CV_BADSIZE_ERR;
    if( (unsigned) coi > 2 )
        return CV_BADRANGE_ERR;

    memset( &mom, 0, sizeof( mom ));
    step /= 4;
    img += coi;

    for( y = 0; y < size.height; sy += 2 * y + 1, y++, img += step )
    {
        double x2 = 0;
        double x3 = 0;
        double x1 = 0;
        double x0 = 0;
        int sx = sx_init;

        for( x = 0; x < size.width - 3; x += 4 )
        {
            double a, b, c, d;

            a = img[x * 3] + img[x * 3 + 3] + img[x * 3 + 6] + img[x * 3 + 9];  /* p0 + p1 + p2 + p3 */
            b = img[x * 3 + 3] + 2 * img[x * 3 + 6] + 3 * img[x * 3 + 9];       /* p1 + p2*2 + p3*3 */
            c = img[x * 3 + 3] + 4 * img[x * 3 + 6] + 9 * img[x * 3 + 9];       /* p1 + p2*4 + p3*9 */
            d = img[x * 3 + 3] + 8 * img[x * 3 + 6] + 27 * img[x * 3 + 9];      /* p1 + p2*8 + p3*27 */

            x0 += a;

            a = x * a + b;      /* x*p0 + (x+1)*p1 + (x+2)*p2 + (x+3)*p3 */
            b = x * b + c;      /* (x+1)*p1 + 2*(x+2)*p2 + 3*(x+3)*p3 */
            c = x * c + d;      /* (x+1)*p1 + 4*(x+2)*p2 + 9*(x+3)*p3 */

            x1 += a;
            a = a * x + b;      /* (x^2)*p0+((x+1)^2)*p1+((x+2)^2)*p2+((x+3)^2)*p3 */
            x2 += a;
            x3 += (a + b) * x + c;      /* x3 += (x^3)*p0+((x+1)^3)*p1+((x+2)^3)*p2+((x+3)^3)*p3 */
        }

        /* process the rest */
        for( ; x < size.width; sx += 2 * x + 1, x++ )
        {
            double p = img[x * 3];
            double xp = x * p;
            double _x2 = sx;

            x0 += p;
            x1 += xp;
            x2 += _x2 * p;
            x3 += _x2 * xp;
        }

        {
            double _y = y;
            double _py = _y * x0;
            double _y2 = sy;

            mom[9] += _py * _y2;        /* m03 */
            mom[8] += x1 * _y2; /* m12 */
            mom[7] += x2 * _y;  /* m21 */
            mom[6] += x3;       /* m30 */
            mom[5] += x0 * _y2; /* m02 */
            mom[4] += x1 * _y;  /* m11 */
            mom[3] += x2;       /* m20 */
            mom[2] += _py;      /* m01 */
            mom[1] += x1;       /* m10 */
            mom[0] += x0;       /* m00 */
        }
    }

    memcpy( &(moments->m00), mom, 10 * sizeof( double ));

    return icvCompleteMomentState( moments );
}


/****************************************************************************************\
*                              Binary Moments caluclating                                *
\****************************************************************************************/
/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name: icvMomentsBinary
//    Purpose:  Find all neccessary values for calculating spatial/central/normalized
//              moments up to third order. Input image is considered as binary image
//              that is each non-zero pixel is treated as "1" and 
//              each zero pixel is treated as "0".
//    Context:
//    Parameters:
//      img  - source image
//      step - its step
//      size - its size (width and height)
//      coi  - index of channel (if color image)
//      moments - structure containing resultant values.
//                The structure is input for icvGetSpatialMoment, 
//                icvGetCentralMoment and icvGetNormalizedCentralMoment functions
//    Returns:
//      CV_NO_ERR if success or error code
//    Notes:
//F*/

/* /////////////// 8uC1 flavor //////////////////// */

static const int mbt[][4] = { {0, 0, 0, 0}, {1, 0, 0, 0}, {1, 1, 1, 1}, {2, 1, 1, 1},
{1, 2, 4, 8}, {2, 2, 4, 8}, {2, 3, 5, 9}, {3, 3, 5, 9},
{1, 3, 9, 27}, {2, 3, 9, 27}, {2, 4, 10, 28}, {3, 4, 10, 28},
{2, 5, 13, 35}, {3, 5, 13, 35}, {3, 6, 14, 36}, {4, 6, 14, 36}
};

static void
icvCalcBinaryMomentsInTile8uC1( uchar * img, int step, CvSize size, int *mom )
{
    int x, y, sx_init = (size.width & -4) * (size.width & -4), sy = 0;

    assert( img && mom );
    memset( mom, 0, 10 * sizeof( mom[0] ));

    for( y = 0; y < size.height; sy += 2 * y + 1, y++, img += step )
    {
        int x3 = 0;
        int x2 = 0;
        int x1 = 0;
        int x0 = 0;
        int sx = sx_init;

        for( x = 0; x < size.width - 3; x += 4 )
        {
            int idx = (img[x] != 0) + (img[x + 1] != 0) * 2 +
                (img[x + 2] != 0) * 4 + (img[x + 3] != 0) * 8;
            int a, b, c;

            a = x * mbt[idx][0] + mbt[idx][1];  /* x*p0+(x+1)*p1+(x+2)*p2+(x+3)*p3 */
            b = x * mbt[idx][1] + mbt[idx][2];  /* (x+1)*p1+2*(x+2)*p2+3*(x+3)*p3 */
            c = x * mbt[idx][2] + mbt[idx][3];  /* (x+1)*p1+4*(x+2)*p2+9*(x+3)*p3 */

            x0 += mbt[idx][0];
            x1 += a;
            a = a * x + b;      /* (x^2)*p0+((x+1)^2)*p1+((x+2)^2)*p2+((x+3)^2)*p3 */
            x2 += a;
            x3 += (a + b) * x + c;      /* (x^3)*p0+((x+1)^3)*p1+ */
            /* ((x+2)^3)*p2+((x+3)^3)*p3 */
        }

        /* process the rest */
        for( ; x < size.width; sx += 2 * x + 1, x++ )
        {
            int p = (img[x] == 0) - 1;
            int xp = x & p;

            x0 -= p;
            x1 += xp;
            x2 += sx & p;
            x3 += sx * xp;
        }

        {
            int py = y * x0;

            mom[9] += py * sy;  /* m03 */
            mom[8] += x1 * sy;  /* m12 */
            mom[7] += x2 * y;   /* m21 */
            mom[6] += x3;       /* m30 */
            mom[5] += x0 * sy;  /* m02 */
            mom[4] += x1 * y;   /* m11 */
            mom[3] += x2;       /* m20 */
            mom[2] += py;       /* m01 */
            mom[1] += x1;       /* m10 */
            mom[0] += x0;       /* m00 */
        }
    }
}


static void
icvCalcBinaryMomentsInTile32fC1( int *img, int step, CvSize size, int *mom )
{
    int x, y, sx_init = (size.width & -4) * (size.width & -4), sy = 0;

    assert( img && mom );
    memset( mom, 0, 10 * sizeof( mom[0] ));

    for( y = 0; y < size.height; sy += 2 * y + 1, y++, img += step )
    {
        int x3 = 0;
        int x2 = 0;
        int x1 = 0;
        int x0 = 0;
        int sx = sx_init;

        for( x = 0; x < size.width - 3; x += 4 )
        {
            int idx = (img[x] * 2 != 0) + (img[x + 1] * 2 != 0) * 2 +
                (img[x + 2] * 2 != 0) * 4 + (img[x + 3] * 2 != 0) * 8;
            int a, b, c;

            a = x * mbt[idx][0] + mbt[idx][1];  /* x*p0+(x+1)*p1+(x+2)*p2+(x+3)*p3 */
            b = x * mbt[idx][1] + mbt[idx][2];  /* (x+1)*p1+2*(x+2)*p2+3*(x+3)*p3 */
            c = x * mbt[idx][2] + mbt[idx][3];  /* (x+1)*p1+4*(x+2)*p2+9*(x+3)*p3 */

            x0 += mbt[idx][0];
            x1 += a;
            a = a * x + b;      /* (x^2)*p0+((x+1)^2)*p1+((x+2)^2)*p2+((x+3)^2)*p3 */
            x2 += a;
            x3 += (a + b) * x + c;      /* (x^3)*p0+((x+1)^3)*p1+ */
            /* ((x+2)^3)*p2+((x+3)^3)*p3 */
        }

        /* process the rest */
        for( ; x < size.width; sx += 2 * x + 1, x++ )
        {
            int p = (img[x] == 0) - 1;
            int xp = x & p;

            x0 -= p;
            x1 += xp;
            x2 += sx & p;
            x3 += sx * xp;
        }

        {
            int py = y * x0;

            mom[9] += py * sy;  /* m03 */
            mom[8] += x1 * sy;  /* m12 */
            mom[7] += x2 * y;   /* m21 */
            mom[6] += x3;       /* m30 */
            mom[5] += x0 * sy;  /* m02 */
            mom[4] += x1 * y;   /* m11 */
            mom[3] += x2;       /* m20 */
            mom[2] += py;       /* m01 */
            mom[1] += x1;       /* m10 */
            mom[0] += x0;       /* m00 */
        }
    }
}


IPCVAPI_IMPL( CvStatus, icvMomentsBinary_8u_C1R, (const uchar * img, int step, CvSize size,
                                                  CvMoments * moments) )
{
    const int tile_size = 64;

    int x, y, k;
    int local_buffer[8192];
    int *tiles = local_buffer;
    int tiles_buf_size;
    CvStatus status;

    if( !img || !moments )
        return CV_NULLPTR_ERR;
    if( (size.width | size.height) < 0 || step < size.width )
        return CV_BADSIZE_ERR;

    memset( moments, 0, sizeof( *moments ));

    tiles_buf_size = ((size.width + tile_size - 1) / tile_size) *
        ((size.height + tile_size - 1) / tile_size) * 10 * sizeof( int );

    if( tiles_buf_size > sizeof( local_buffer ))
    {
        tiles = (int *) icvAlloc( tiles_buf_size );
        if( !tiles )
            return CV_OUTOFMEM_ERR;
    }

    /* put table into cache */
    for( y = 0; y < 16; y++ )
        moments->m00 = mbt[y][0];

    for( y = 0, k = 0; y < size.height; y += tile_size, img += tile_size * step )
    {
        for( x = 0; x < size.width; x += tile_size, k += 10 )
        {
            CvSize sz;

            sz.width = size.width - x;
            sz.height = size.height - y;

            sz.width = _CV_MIN2( sz.width, tile_size );
            sz.height = _CV_MIN2( sz.height, tile_size );

            icvCalcBinaryMomentsInTile8uC1( (uchar *) (img + x), step, sz, tiles + k );
        }
    }

    status = icvCompleteMomentState_int( tiles, size, tile_size, moments );
    if( tiles != local_buffer )
        icvFree( &tiles );

    return status;
}


/* ////////////// 8uC3 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMomentsBinary_8u_C3CR, (const uchar * img, int step, CvSize size,
                                                   int coi, CvMoments * moments) )
{
    int x, y, sx_init = (size.width & -4) * (size.width & -4), sy = 0, sum = 0;
    double mom[9];

    coi--;

    if( !img || !moments )
        return CV_NULLPTR_ERR;
    if( (size.width | size.height) < 0 || step < size.width * 3 )
        return CV_BADSIZE_ERR;
    if( (unsigned) coi > 2 )
        return CV_BADRANGE_ERR;

    memset( mom, 0, sizeof( mom ));
    img += coi;

    /* put table into cache */
    for( y = 0; y < 16; y++ )
        moments->m00 = mbt[y][0];

    for( y = 0; y < size.height; sy += 2 * y + 1, y++, img += step )
    {
        double x3 = 0;
        int x0 = 0;
        int x1 = 0;
        int x2 = 0;
        int sx = sx_init;

        for( x = 0; x < size.width - 3; x += 4 )
        {
            int idx = (img[x * 3] != 0) + (img[x * 3 + 3] != 0) * 2 +
                (img[x * 3 + 6] != 0) * 4 + (img[x * 3 + 9] != 0) * 8;
            int _a, _b, _c;

            _a = x * mbt[idx][0] + mbt[idx][1]; /* x*p0+(x+1)*p1+(x+2)*p2+(x+3)*p3 */
            _b = x * mbt[idx][1] + mbt[idx][2]; /* (x+1)*p1+2*(x+2)*p2+3*(x+3)*p3 */
            _c = x * mbt[idx][2] + mbt[idx][3]; /* (x+1)*p1+4*(x+2)*p2+9*(x+3)*p3 */

            x0 += mbt[idx][0];
            x1 += _a;
            _a = _a * x + _b;   /* (x^2)*p0+((x+1)^2)*p1+((x+2)^2)*p2+((x+3)^2)*p3 */
            x2 += _a;
            x3 += ((double) (_a + _b)) * x + _c;        /* (x^3)*p0+((x+1)^3)*p1+ */
            /* ((x+2)^3)*p2+((x+3)^3)*p3 */
        }

        /* process the rest */
        for( ; x < size.width; sx += 2 * x + 1, x++ )
        {
            int p = (img[x * 3] == 0) - 1;
            int xp = x & p;

            x0 -= p;
            x1 += xp;
            x2 += sx & p;
            x3 += ((double) sx) * xp;
        }

        {
            double _y = y;
            double _py = _y * x0;
            double _y2 = sy;

            sum += x0;

            mom[8] += _py * _y2;        /* m03 */
            mom[7] += x1 * _y2; /* m12 */
            mom[6] += x2 * _y;  /* m21 */
            mom[5] += x3;       /* m30 */
            mom[4] += x0 * _y2; /* m02 */
            mom[3] += x1 * _y;  /* m11 */
            mom[2] += x2;       /* m20 */
            mom[1] += _py;      /* m01 */
            mom[0] += x1;       /* m10 */
        }
    }

    moments->m00 = sum;
    memcpy( &(moments->m00) + 1, mom, 9 * sizeof( double ));

    return icvCompleteMomentState( moments );
}

/* /////////////// 8sC1 flavor //////////////////// */

IPCVAPI_IMPL( CvStatus, icvMomentsBinary_8s_C1R, (const char *img, int step, CvSize size,
                                                  CvMoments * moments) )
{
    return icvMomentsBinary_8u_C1R( (uchar *) img, step, size, moments );
}

/* /////////////// 8sC3 flavor //////////////////// */

IPCVAPI_IMPL( CvStatus, icvMomentsBinary_8s_C3CR, (const char *img, int step, CvSize size,
                                                   int coi, CvMoments * moments) )
{
    return icvMomentsBinary_8u_C3CR( (const uchar *) img, step, size, coi, moments );
}


/* /////////////// 32fC1 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMomentsBinary_32f_C1R, (const float *img,
                                                   int step,
                                                   CvSize size, CvMoments * moments) )
{
    const int tile_size = 64;
    int x, y, k;
    int local_buffer[8192];
    int *tiles = local_buffer;
    int tiles_buf_size;
    CvStatus status;

    if( !img || !moments )
        return CV_NULLPTR_ERR;
    if( (size.width | size.height) < 0 || step < size.width * 4 || (step & 3) != 0 )
        return CV_BADSIZE_ERR;

    memset( moments, 0, sizeof( *moments ));
    step /= 4;

    tiles_buf_size = ((size.width + tile_size - 1) / tile_size) *
        ((size.height + tile_size - 1) / tile_size) * 10 * sizeof( int );

    if( tiles_buf_size > sizeof( local_buffer ))
    {
        tiles = (int *) icvAlloc( tiles_buf_size );
        if( !tiles )
            return CV_OUTOFMEM_ERR;
    }

    /* put table into cache */
    for( y = 0; y < 16; y++ )
        moments->m00 = mbt[y][0];

    for( y = 0, k = 0; y < size.height; y += tile_size, img += tile_size * step )
    {
        for( x = 0; x < size.width; x += tile_size, k += 10 )
        {
            CvSize sz;

            sz.width = size.width - x;
            sz.height = size.height - y;

            sz.width = _CV_MIN2( sz.width, tile_size );
            sz.height = _CV_MIN2( sz.height, tile_size );

            icvCalcBinaryMomentsInTile32fC1( (int *) img + x, step, sz, tiles + k );
        }
    }

    status = icvCompleteMomentState_int( tiles, size, tile_size, moments );
    if( tiles != local_buffer )
        icvFree( &tiles );

    return status;
}


/* /////////////// 32fC3 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvMomentsBinary_32f_C3CR, (const float *imgf,
                                                    int step,
                                                    CvSize size,
                                                    int coi, CvMoments * moments) )
{
    int x, y, sx_init = (size.width & -4) * (size.width & -4), sy = 0, sum = 0;
    int *img = (int *) imgf;
    double mom[9];

    coi--;

    if( !img || !moments )
        return CV_NULLPTR_ERR;
    if( (size.width | size.height) < 0 || step < size.width * 12 || (step & 3) != 0 )
        return CV_BADSIZE_ERR;
    if( (unsigned) coi > 2 )
        return CV_BADRANGE_ERR;

    memset( mom, 0, sizeof( mom ));
    step /= 4;
    img += coi;

    /* put table into cache */
    for( y = 0; y < 16; y++ )
        moments->m00 = mbt[y][0];

    for( y = 0; y < size.height; sy += 2 * y + 1, y++, img += step )
    {
        double x3 = 0;
        int x0 = 0;
        int x1 = 0;
        int x2 = 0;
        int sx = sx_init;

        for( x = 0; x < size.width - 3; x += 4 )
        {
            int idx = (img[x * 3] * 2 != 0) + (img[x * 3 + 3] * 2 != 0) * 2 +
                (img[x * 3 + 6] * 2 != 0) * 4 + (img[x * 3 + 9] * 2 != 0) * 8;
            int _a, _b, _c;

            _a = x * mbt[idx][0] + mbt[idx][1]; /* x*p0+(x+1)*p1+(x+2)*p2+(x+3)*p3 */
            _b = x * mbt[idx][1] + mbt[idx][2]; /* (x+1)*p1+2*(x+2)*p2+3*(x+3)*p3 */
            _c = x * mbt[idx][2] + mbt[idx][3]; /* (x+1)*p1+4*(x+2)*p2+9*(x+3)*p3 */

            x0 += mbt[idx][0];
            x1 += _a;
            _a = _a * x + _b;   /* (x^2)*p0+((x+1)^2)*p1+((x+2)^2)*p2+((x+3)^2)*p3 */
            x2 += _a;
            x3 += ((double) (_a + _b)) * x + _c;        /* (x^3)*p0+((x+1)^3)*p1+ */
            /* ((x+2)^3)*p2+((x+3)^3)*p3 */
        }

        /* process the rest */
        for( ; x < size.width; sx += 2 * x + 1, x++ )
        {
            int p = (img[x * 3] * 2 == 0) - 1;
            int xp = x & p;

            x0 -= p;
            x1 += xp;
            x2 += sx & p;
            x3 += ((double) sx) * xp;
        }

        {
            double _y = y;
            double _py = _y * x0;
            double _y2 = sy;

            sum += x0;

            mom[8] += _py * _y2;        /* m03 */
            mom[7] += x1 * _y2; /* m12 */
            mom[6] += x2 * _y;  /* m21 */
            mom[5] += x3;       /* m30 */
            mom[4] += x0 * _y2; /* m02 */
            mom[3] += x1 * _y;  /* m11 */
            mom[2] += x2;       /* m20 */
            mom[1] += _py;      /* m01 */
            mom[0] += x1;       /* m10 */
        }
    }

    moments->m00 = sum;
    memcpy( &(moments->m00) + 1, mom, 9 * sizeof( double ));

    return icvCompleteMomentState( moments );
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name: cvGetHuMoments
//    Purpose: Returns Hu moments
//    Context:
//    Parameters:
//      mState  - moment structure filled by one of the icvMoments[Binary]*** function
//      HuState - pointer to output structure containing seven Hu moments
//    Returns:
//      CV_NO_ERR if success or error code
//    Notes:
//F*/
CV_IMPL void
cvGetHuMoments( CvMoments * mState, CvHuMoments * HuState )
{
    CV_FUNCNAME( "cvGetHuMoments" );

    __BEGIN__;

    if( !mState || !HuState )
        CV_ERROR_FROM_STATUS( CV_NULLPTR_ERR );

    {
        double m00s = mState->inv_sqrt_m00, m00 = m00s * m00s, s2 = m00 * m00, s3 = s2 * m00s;

        double nu20 = mState->mu20 * s2,
            nu11 = mState->mu11 * s2,
            nu02 = mState->mu02 * s2,
            nu30 = mState->mu30 * s3,
            nu21 = mState->mu21 * s3, nu12 = mState->mu12 * s3, nu03 = mState->mu03 * s3;

        double t0 = nu30 + nu12;
        double t1 = nu21 + nu03;

        double q0 = t0 * t0, q1 = t1 * t1;

        double n4 = 4 * nu11;
        double s = nu20 + nu02;
        double d = nu20 - nu02;

        HuState->hu1 = s;
        HuState->hu2 = d * d + n4 * nu11;
        HuState->hu4 = q0 + q1;
        HuState->hu6 = d * (q0 - q1) + n4 * t0 * t1;

        t0 *= q0 - 3 * q1;
        t1 *= 3 * q0 - q1;

        q0 = nu30 - 3 * nu12;
        q1 = 3 * nu21 - nu03;

        HuState->hu3 = q0 * q0 + q1 * q1;
        HuState->hu5 = q0 * t0 + q1 * t1;
        HuState->hu7 = q1 * t0 - q0 * t1;
    }

    __CLEANUP__;
    __END__;
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name: cvGetSpatialMoment
//    Purpose:  Returns spatial moment(x_order, y_order) which is determined as:
//              m(x_o,y_o) = sum (x ^ x_o)*(y ^ y_o)*I(x,y)
//              0 <= x_o, y_o; x_o + y_o <= 3
//    Context:
//    Parameters:
//      mom  - moment structure filled by one of the icvMoments[Binary]*** function
//      x_order - x order of the moment
//      y_order - y order of the moment
//    Returns:
//      moment value or large negative number (-DBL_MAX) if error
//    Notes:
//F*/
CV_IMPL double
cvGetSpatialMoment( CvMoments * moments, int x_order, int y_order )
{
    int order = x_order + y_order;
    double moment = -DBL_MAX;

    CV_FUNCNAME( "cvGetSpatialMoment" );

    __BEGIN__;

    if( !moments )
        CV_ERROR_FROM_STATUS( CV_NULLPTR_ERR );
    if( (x_order | y_order) < 0 || order > 3 )
        CV_ERROR_FROM_STATUS( CV_BADRANGE_ERR );

    moment = (&(moments->m00))[order + (order >> 1) + (order > 2) * 2 + y_order];

    __CLEANUP__;
    __END__;

    return moment;
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name: cvGetCentralMoment
//    Purpose:  Returns central moment(x_order, y_order) which is determined as:
//              mu(x_o,y_o) = sum ((x - xc)^ x_o)*((y - yc) ^ y_o)*I(x,y)
//              0 <= x_o, y_o; x_o + y_o <= 3,
//              (xc, yc) = (m10/m00,m01/m00) - center of gravity 
//    Context:
//    Parameters:
//      mom  - moment structure filled by one of the icvMoments[Binary]*** function
//      x_order - x order of the moment
//      y_order - y order of the moment
//    Returns:
//      moment value or large negative number (-DBL_MAX) if error
//    Notes:
//F*/
CV_IMPL double
cvGetCentralMoment( CvMoments * moments, int x_order, int y_order )
{
    int order = x_order + y_order;
    double mu = 0;

    CV_FUNCNAME( "cvGetCentralMoment" );

    __BEGIN__;

    if( !moments )
        CV_ERROR_FROM_STATUS( CV_NULLPTR_ERR );
    if( (x_order | y_order) < 0 || order > 3 )
        CV_ERROR_FROM_STATUS( CV_BADRANGE_ERR );

    if( order >= 2 )
    {
        mu = (&(moments->m00))[4 + order * 3 + y_order];
    }
    else if( order == 0 )
        mu = moments->m00;

    __CLEANUP__;
    __END__;

    return mu;
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name: cvGetNormalizedCentralMoment
//    Purpose: Returns normalized central moment(x_order,y_order) which is determined as:
//             nu(x_o,y_o) = mu(x_o, y_o)/(m00 ^ (((x_o + y_o)/2) + 1))
//             0 <= x_o, y_o; x_o + y_o <= 3,
//             (xc, yc) = (m10/m00,m01/m00) - center of gravity 
//    Context:
//    Parameters:
//      mom  - moment structure filled by one of the icvMoments[Binary]*** function
//      x_order - x order of the moment
//      y_order - y order of the moment
//    Returns:
//      moment value or large negative number (-DBL_MAX) if error
//    Notes:
//F*/
CV_IMPL double
cvGetNormalizedCentralMoment( CvMoments * moments, int x_order, int y_order )
{
    int order = x_order + y_order;
    double mu = 0;
    double m00s, m00;

    CV_FUNCNAME( "cvGetCentralNormalizedMoment" );

    __BEGIN__;

    mu = cvGetCentralMoment( moments, x_order, y_order );
    CV_CHECK();

    m00s = moments->inv_sqrt_m00;
    m00 = m00s * m00s;

    while( --order >= 0 )
        m00 *= m00s;
    mu *= m00;

    __CLEANUP__;
    __END__;

    return mu;
}


/* End of file. */
