/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/
#include "_cv.h"
#include "_cvwrap.h"
#include "_cvarray.h"

CV_IMPL IplConvKernel *
cvCreateStructuringElementEx( int cols, int rows,
                              int anchorX, int anchorY, CvElementShape shape, int *values )
{
    IplConvKernel *element = 0;
    int i, size = rows * cols;
    int *vals;
    int element_size = sizeof( *element ) + size * sizeof( element->values[0] );

    CV_FUNCNAME( "cvCreateStructuringElementEx" );

    __BEGIN__;

    if( !values && shape == CV_SHAPE_CUSTOM )
        CV_ERROR_FROM_STATUS( CV_NULLPTR_ERR );

    if( cols <= 0 || rows <= 0 ||
        (unsigned) anchorX >= (unsigned) cols || (unsigned) anchorY >= (unsigned) rows )
        CV_ERROR_FROM_STATUS( CV_BADSIZE_ERR );

    element_size = (element_size + 31) & -32;
    element = (IplConvKernel *) icvAlloc( element_size );
    if( !element )
        CV_ERROR_FROM_STATUS( CV_OUTOFMEM_ERR );

    element->nCols = cols;
    element->nRows = rows;
    element->anchorX = anchorX;
    element->anchorY = anchorY;
    element->nShiftR = shape < CV_SHAPE_ELLIPSE ? shape : CV_SHAPE_CUSTOM;
    element->values = vals = (int *) (element + 1);

    switch (shape)
    {
    case CV_SHAPE_RECT:
        memset( vals, -1, sizeof( vals[0] ) * cols * rows );
        break;
    case CV_SHAPE_CROSS:
        memset( vals, 0, sizeof( vals[0] ) * cols * rows );
        for( i = 0; i < cols; i++ )
        {
            vals[i + anchorY * cols] = -1;
        }

        for( i = 0; i < rows; i++ )
        {
            vals[anchorX + i * cols] = -1;
        }
        break;
    case CV_SHAPE_ELLIPSE:
        {
            int r = (rows + 1) / 2;
            int c = cols / 2;
            double inv_r2 = 1. / (((double) (r)) * (r));

            memset( vals, 0, sizeof( vals[0] ) * cols * rows );

            for( i = 0; i < r; i++ )
            {
                int y = r - i;
                int dx = cvRound( c * sqrt( (r * r - y * y) * inv_r2 ));
                int x1 = c - dx;
                int x2 = c + dx;

                if( x1 < 0 )
                    x1 = 0;
                if( x2 >= cols )
                    x2 = cols;
                x2 -= x1 - 1;
                memset( vals + i * cols + x1, -1, x2 * sizeof( int ));
                memset( vals + (rows - i - 1) * cols + x1, -1, x2 * sizeof( int ));
            }
        }
        break;
    case CV_SHAPE_CUSTOM:
        for( i = 0; i < size; i++ )
        {
            vals[i] = !values || values[i] ? -1 : 0;
        }
        break;
    default:
        icvFree( &element );
        CV_ERROR_FROM_STATUS( CV_BADFLAG_ERR );
    }

    __CLEANUP__;
    __END__;

    return element;
}


CV_IMPL void
cvReleaseStructuringElement( IplConvKernel ** element )
{
    CV_FUNCNAME( "cvReleaseStructuringElement" );

    __BEGIN__;

    if( !element )
        CV_ERROR_FROM_STATUS( CV_NULLPTR_ERR );
    icvFree( element );

    __CLEANUP__;
    __END__;
}


typedef CvStatus( CV_STDCALL * CvMorphFunc ) (uchar * src, int src_step,
                                              uchar * dst, int dst_step,
                                              CvSize * size, struct CvMorphState * state,
                                              int stage);

CV_IMPL void
icvMorphOp( IplImage * src, IplImage * dst, IplConvKernel * element, int iterations, int mop )
{
    static CvMorphFunc funcs[36];

    CvMorphFunc func = 0;
    CvMorphState *state = 0;
    CvStatus status = CV_OK;
    uchar *src_data = 0;
    uchar *dst_data = 0;

    int src_step = 0;
    int dst_step = 0;
    int i;

    CvSize src_size;
    CvSize dst_size;
    CvElementShape shape;

    CV_FUNCNAME( "cvErode/cvDilate" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( src ));
    cvGetImageRawData( src, &src_data, &src_step, &src_size );

    if( dst != src )
    {
        CV_CALL( CV_CHECK_IMAGE( dst ));

        if( src->depth != dst->depth || src->nChannels != dst->nChannels )
            CV_ERROR( IPL_StsBadArg, "src and dst have different formats" );
        cvGetImageRawData( dst, &dst_data, &dst_step, &dst_size );

        if( src_size != dst_size )
            CV_ERROR( IPL_StsBadArg, "src and dst have different sizes" );
    }
    else
    {
        dst_data = src_data;
        dst_step = src_step;
    }

    if( src->depth != IPL_DEPTH_8U && src->depth != IPL_DEPTH_32F )
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );

    if( src->nChannels != 1 && src->nChannels != 3 && src->nChannels != 4 )
        CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );

    if( element )
    {
        status =
            icvMorphologyInitAlloc( src_size.width, src->depth == IPL_DEPTH_8U ? cv8u : cv32f,
                                     src->nChannels, cvSize( element->nCols, element->nRows ),
                                     cvPoint( element->anchorX, element->anchorY ),
                                     (CvElementShape) (element->nShiftR), element->values,
                                     &state );
        shape = (CvElementShape) (element->nShiftR);
        shape = shape < CV_SHAPE_ELLIPSE ? shape : CV_SHAPE_CUSTOM;
    }
    else
    {
        status =
            icvMorphologyInitAlloc( src_size.width, src->depth == IPL_DEPTH_8U ? cv8u : cv32f,
                                     src->nChannels, cvSize( 3, 3 ), cvPoint( 1, 1 ),
                                     CV_SHAPE_RECT, 0, &state );
        shape = CV_SHAPE_RECT;
    }
    if( status < 0 )
        CV_ERROR_FROM_STATUS( status );

    if( !funcs[0] )
    {
        funcs[1] = (CvMorphFunc) icvErodeStrip_Rect_8u_C3R;
        funcs[2] = (CvMorphFunc) icvErodeStrip_Rect_8u_C4R;
        funcs[3] = (CvMorphFunc) icvErodeStrip_Rect_32f_C1R;
        funcs[4] = (CvMorphFunc) icvErodeStrip_Rect_32f_C3R;
        funcs[5] = (CvMorphFunc) icvErodeStrip_Rect_32f_C4R;

        funcs[6] = (CvMorphFunc) icvErodeStrip_Cross_8u_C1R;
        funcs[7] = (CvMorphFunc) icvErodeStrip_Cross_8u_C3R;
        funcs[8] = (CvMorphFunc) icvErodeStrip_Cross_8u_C4R;
        funcs[9] = (CvMorphFunc) icvErodeStrip_Cross_32f_C1R;
        funcs[10] = (CvMorphFunc) icvErodeStrip_Cross_32f_C3R;
        funcs[11] = (CvMorphFunc) icvErodeStrip_Cross_32f_C4R;

        funcs[12] = (CvMorphFunc) icvErodeStrip_8u_C1R;
        funcs[13] = (CvMorphFunc) icvErodeStrip_8u_C3R;
        funcs[14] = (CvMorphFunc) icvErodeStrip_8u_C4R;
        funcs[15] = (CvMorphFunc) icvErodeStrip_32f_C1R;
        funcs[16] = (CvMorphFunc) icvErodeStrip_32f_C3R;
        funcs[17] = (CvMorphFunc) icvErodeStrip_32f_C4R;

        funcs[18] = (CvMorphFunc) icvDilateStrip_Rect_8u_C1R;
        funcs[19] = (CvMorphFunc) icvDilateStrip_Rect_8u_C3R;
        funcs[20] = (CvMorphFunc) icvDilateStrip_Rect_8u_C4R;
        funcs[21] = (CvMorphFunc) icvDilateStrip_Rect_32f_C1R;
        funcs[22] = (CvMorphFunc) icvDilateStrip_Rect_32f_C3R;
        funcs[23] = (CvMorphFunc) icvDilateStrip_Rect_32f_C4R;

        funcs[24] = (CvMorphFunc) icvDilateStrip_Cross_8u_C1R;
        funcs[25] = (CvMorphFunc) icvDilateStrip_Cross_8u_C3R;
        funcs[26] = (CvMorphFunc) icvDilateStrip_Cross_8u_C4R;
        funcs[27] = (CvMorphFunc) icvDilateStrip_Cross_32f_C1R;
        funcs[28] = (CvMorphFunc) icvDilateStrip_Cross_32f_C3R;
        funcs[29] = (CvMorphFunc) icvDilateStrip_Cross_32f_C4R;

        funcs[30] = (CvMorphFunc) icvDilateStrip_8u_C1R;
        funcs[31] = (CvMorphFunc) icvDilateStrip_8u_C3R;
        funcs[32] = (CvMorphFunc) icvDilateStrip_8u_C4R;
        funcs[33] = (CvMorphFunc) icvDilateStrip_32f_C1R;
        funcs[34] = (CvMorphFunc) icvDilateStrip_32f_C3R;
        funcs[35] = (CvMorphFunc) icvDilateStrip_32f_C4R;

        funcs[0] = (CvMorphFunc) icvErodeStrip_Rect_8u_C1R;
    }

    func = funcs[(src->nChannels >> 1) + (src->depth == IPL_DEPTH_32F) * 3 +
                 (shape == CV_SHAPE_RECT ? 0 : shape == CV_SHAPE_CROSS ? 1 : 2) * 6 +
                 mop * 18];

    for( i = 0; i < iterations; i++ )
    {
        IPPI_CALL( func( src_data, src_step, dst_data, dst_step, &src_size, state, 0 ));
        src_data = dst_data;
        src_step = dst_step;
    }

    __CLEANUP__;
    __END__;

    icvMorphologyFree( &state );
}



CV_IMPL void
cvErode( IplImage * src, IplImage * dst, IplConvKernel * element, int iterations )
{
    icvMorphOp( src, dst, element, iterations, 0 );
}


CV_IMPL void
cvDilate( IplImage * src, IplImage * dst, IplConvKernel * element, int iterations )
{
    icvMorphOp( src, dst, element, iterations, 1 );
}


CV_IMPL void
cvMorphologyEx( IplImage * src, IplImage * dst,
                IplImage * temp, IplConvKernel * element, CvMorphOp op, int iterations )
{
    CV_FUNCNAME( "cvMorhologyEx" );

    __BEGIN__;

    if( (op == CV_MOP_GRADIENT ||
         (op == CV_MOP_TOPHAT || op == CV_MOP_BLACKHAT) && src == dst) && temp == 0 )
        CV_ERROR( IPL_HeaderIsNull, "temp image required" );

    if( temp == src || temp == dst )
        CV_ERROR( IPL_HeaderIsNull, "temp image is equal to src or dst" );

    switch (op)
    {
    case CV_MOP_OPEN:
        CV_CALL( cvErode( src, dst, element, iterations ));
        CV_CALL( cvDilate( dst, dst, element, iterations ));
        break;
    case CV_MOP_CLOSE:
        CV_CALL( cvDilate( src, dst, element, iterations ));
        CV_CALL( cvErode( dst, dst, element, iterations ));
        break;
    case CV_MOP_GRADIENT:
        CV_CALL( cvErode( src, temp, element, iterations ));
        CV_CALL( cvDilate( src, dst, element, iterations ));
        CV_CALL( icvSubtract( dst, temp, dst ));
        break;
    case CV_MOP_TOPHAT:
        CV_CALL( cvErode( src, src != dst ? dst : temp, element, iterations ));
        CV_CALL( icvSubtract( src, src != dst ? dst : temp, dst ));
        break;
    case CV_MOP_BLACKHAT:
        CV_CALL( cvDilate( src, src != dst ? dst : temp, element, iterations ));
        CV_CALL( icvSubtract( src != dst ? dst : temp, src, dst ));
        break;
    default:
        CV_ERROR( IPL_StsBadArg, "unknown morphological operation" );
    }

    __CLEANUP__;
    __END__;
}


/* End of file. */
