/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include "_cv.h"
#include "_cvimgstat.h"
#include <float.h>

/****************************************************************************************\
*                                  N o r m                                               *
\****************************************************************************************/

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name: cvNorm
//    Purpose: Calculates norm, difference norm and relative difference norm of images
//    Context:
//    Parameters:
//      imgA,  imgB  - source images,
//      stepA, stepB - their steps,
//      size or roi  - their size or ROI size,
//      coi          - channel index (1..3) (if color image)
//      normType     - norm type (see notes).
//    Returns:
//      value of norm or large negative number (if error)
//    Notes:
//      If imgB is NULL, norm of the image A is calculated, and norm type may be
//      _CV_C, _CV_L1 or _CV_L2. If imgB is not NULL, difference norm or relative
//      difference norm is calculated. In this event norm type may be also
//      _CV_RELATIVE_C, _CV_RELATIVE_L1 and _CV_RELATIVE_L2.
//F*/


/* //////////////////////////////// 8uC1 flavor /////////////////////////////////////// */
IPCVAPI_IMPL( CvStatus, icvNorm_8u_C1R, (const uchar * imgA, int stepA,
                                         const uchar * imgB, int stepB,
                                         CvSize roi, int normType, double *_norm) )
{
    int i, j, s = 0, g = 0;
    int64 norm1 = 0, norm2 = 0;
    double norm = -DBL_MAX, nrm, epsilon = 1.0e-30;

    if( !imgA )
        return CV_NULLPTR_ERR;
    if( (roi.width | roi.height) < 0 || stepA < roi.width )
        return CV_BADSIZE_ERR;
    if( imgB && (stepB < roi.width) )
        return CV_BADSIZE_ERR;

    if( imgB )
    {
        if( normType < _CV_RELATIVE_C )
            normType |= _CV_DIFF;
    }
    else
        normType &= _CV_RELATIVE - 1;

    switch (normType)
    {
    case _CV_C:                /* ||A|| C  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA )
        {
            for( j = 0; j < roi.width - 3; j += 4 )
            {
                int v = imgA[j];

                if( s < v )
                    s = v;
                v = imgA[j + 1];
                if( s < v )
                    s = v;
                v = imgA[j + 2];
                if( s < v )
                    s = v;
                v = imgA[j + 3];
                if( s < v )
                    s = v;
            }
            for( ; j < roi.width; j++ )
            {
                int v = imgA[j];

                if( s < v )
                    s = v;
            }
        }
        norm = s;
        break;
    case _CV_L1:               /* ||A|| L1  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA )
            for( j = 0; j < roi.width; j++ )
                norm1 += imgA[j];
        norm = (double) norm1;
        break;
    case _CV_L2:               /* ||A|| L2  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA )
        {
            for( j = 0; j < roi.width - 3; j += 4 )
            {
                int v = imgA[j];
                int e = v * v;

                v = imgA[j + 1];
                e += v * v;
                v = imgA[j + 2];
                e += v * v;
                v = imgA[j + 3];
                e += v * v;
                norm1 += e;
            }
            for( ; j < roi.width; j++ )
            {
                int v = imgA[j];

                norm1 += v * v;
            }
        }
        norm = sqrt( (double) norm1 );
        break;
    case _CV_DIFF_C:           /* ||A-B|| C  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 3; j += 4 )
            {
                int v = imgA[j] - imgB[j];

                v = _CV_ABS( v );
                if( s < v )
                    s = v;
                v = imgA[j + 1] - imgB[j + 1];
                v = _CV_ABS( v );
                if( s < v )
                    s = v;
                v = imgA[j + 2] - imgB[j + 2];
                v = _CV_ABS( v );
                if( s < v )
                    s = v;
                v = imgA[j + 3] - imgB[j + 3];
                v = _CV_ABS( v );
                if( s < v )
                    s = v;
            }
            for( ; j < roi.width; j++ )
            {
                int v = imgA[j] - imgB[j];

                v = _CV_ABS( v );
                if( s < v )
                    s = v;
            }
        }
        norm = s;
        break;
    case _CV_DIFF_L1:          /* ||A-B|| L1  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 3; j += 4 )
            {
                int v = imgA[j] - imgB[j];

                v = _CV_ABS( v );
                s += v;
                v = imgA[j + 1] - imgB[j + 1];
                v = _CV_ABS( v );
                s += v;
                v = imgA[j + 2] - imgB[j + 2];
                v = _CV_ABS( v );
                s += v;
                v = imgA[j + 3] - imgB[j + 3];
                v = _CV_ABS( v );
                s += v;
            }
            for( ; j < roi.width; j++ )
                s += _CV_ABS( imgA[j] - imgB[j] );
        }
        norm = s;
        break;
    case _CV_DIFF_L2:          /* ||A-B|| L2  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 3; j += 4 )
            {
                int v = imgA[j] - imgB[j];
                int e = v * v;

                v = imgA[j + 1] - imgB[j + 1];
                e += v * v;
                v = imgA[j + 2] - imgB[j + 2];
                e += v * v;
                v = imgA[j + 3] - imgB[j + 3];
                e += v * v;
                norm1 += e;
            }
            for( ; j < roi.width; j++ )
            {
                int v = imgA[j] - imgB[j];

                norm1 += v * v;
            }
        }
        norm = sqrt( (double) norm1 );
        break;
    case _CV_RELATIVE_C:       /* ||A-B|| / ||B||  C */
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 3; j += 4 )
            {
                int e = imgB[j];
                int v = imgA[j] - imgB[j];

                v = _CV_ABS( v );
                if( g < e )
                    g = e;
                if( s < v )
                    s = v;
                e = imgB[j + 1];
                v = imgA[j + 1] - imgB[j + 1];
                v = _CV_ABS( v );
                if( g < e )
                    g = e;
                if( s < v )
                    s = v;
                e = imgB[j + 2];
                v = imgA[j + 2] - imgB[j + 2];
                v = _CV_ABS( v );
                if( g < e )
                    g = e;
                if( s < v )
                    s = v;
                e = imgB[j + 3];
                v = imgA[j + 3] - imgB[j + 3];
                v = _CV_ABS( v );
                if( g < e )
                    g = e;
                if( s < v )
                    s = v;
            }
            for( ; j < roi.width; j++ )
            {
                int e = imgB[j];
                int v = imgA[j] - imgB[j];

                v = _CV_ABS( v );
                if( g < e )
                    g = e;
                if( s < v )
                    s = v;
            }
        }
        nrm = (g) ? (double) g : epsilon;
        norm = s / nrm;
        break;
    case _CV_RELATIVE_L1:      /* ||A-B|| / ||B||  L1 */
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 3; j += 4 )
            {
                int v = imgA[j] - imgB[j];

                v = _CV_ABS( v );
                s += v;
                v = imgA[j + 1] - imgB[j + 1];
                v = _CV_ABS( v );
                s += v;
                v = imgA[j + 2] - imgB[j + 2];
                v = _CV_ABS( v );
                s += v;
                v = imgA[j + 3] - imgB[j + 3];
                v = _CV_ABS( v );
                s += v;
                g += imgB[j] + imgB[j + 1] + imgB[j + 2] + imgB[j + 3];
            }
            for( ; j < roi.width; j++ )
            {
                int v = imgA[j] - imgB[j];

                v = _CV_ABS( v );
                s += v;
                g += imgB[j];
            }
        }
        nrm = (g) ? (double) g : epsilon;
        norm = s / nrm;
        break;
    case _CV_RELATIVE_L2:      /* ||A-B|| / ||B||  L2 */
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 3; j += 4 )
            {
                int e = imgB[j];
                int v = imgA[j] - imgB[j];
                int e2 = e * e;
                int v2 = v * v;

                e = imgB[j + 1];
                v = imgA[j + 1] - imgB[j + 1];
                e2 += e * e;
                v2 += v * v;
                e = imgB[j + 2];
                v = imgA[j + 2] - imgB[j + 2];
                e2 += e * e;
                v2 += v * v;
                e = imgB[j + 3];
                v = imgA[j + 3] - imgB[j + 3];
                e2 += e * e;
                v2 += v * v;
                norm1 += v2;
                norm2 += e2;
            }
            for( ; j < roi.width; j++ )
            {
                int e = imgB[j];
                int v = imgA[j] - imgB[j];

                norm1 += v * v;
                norm2 += e * e;
            }
        }
        nrm = (norm2) ? (double) norm2 : epsilon;
        norm = sqrt( (double) norm1 / nrm );
        break;
    default:
        return CV_BADRANGE_ERR;
    }                           /* end of switch */

    *_norm = norm;
    return CV_OK;
} /*-------- End of icvNorm8uC1R --------------------------------------------------------- */

/* ///////////////////////////////// 8uC3 flavor ///////////////////////////////////////// */
IPCVAPI_IMPL( CvStatus, icvNorm_8u_C3CR, (const uchar * imgA, int stepA,
                                          const uchar * imgB, int stepB,
                                          CvSize roi, int coi, int normType, double *_norm) )
{
    int i, j, s = 0, g = 0;

#ifndef __STDC__
    int64 norm1 = 0, norm2 = 0;
#else
    double norm1 = 0, norm2 = 0;
#endif
    double norm = 0, nrm, epsilon = 1.0e-40;

    roi.width *= 3;

    if( coi < 1 || coi > 3 )
        return CV_BADRANGE_ERR;
    if( !imgA )
        return CV_NULLPTR_ERR;
    if( (roi.width | roi.height) < 0 || stepA < roi.width )
        return CV_BADSIZE_ERR;
    if( imgB && (stepB < roi.width) )
        return CV_BADSIZE_ERR;

    coi--;
    imgA += coi;

    if( imgB )
        imgB += coi;

    if( imgB )
    {
        if( normType < _CV_RELATIVE_C )
            normType |= _CV_DIFF;
    }
    else
        normType &= _CV_RELATIVE - 1;

    switch (normType)
    {
    case _CV_C:                /* ||A|| C  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA )
        {
            for( j = 0; j < roi.width - 9; j += 12 )
            {
                int v = imgA[j];

                if( s < v )
                    s = v;
                v = imgA[j + 3];
                if( s < v )
                    s = v;
                v = imgA[j + 6];
                if( s < v )
                    s = v;
                v = imgA[j + 9];
                if( s < v )
                    s = v;
            }
            for( ; j < roi.width; j += 3 )
            {
                int v = imgA[j];

                if( s < v )
                    s = v;
            }
        }
        norm = s;
        break;
    case _CV_L1:               /* ||A|| L1  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA )
            for( j = 0; j < roi.width; j += 3 )
                norm += imgA[j];
        break;
    case _CV_L2:               /* ||A|| L2  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA )
        {
            for( j = 0; j < roi.width - 9; j += 12 )
            {
                int v = imgA[j];
                int e = v * v;

                v = imgA[j + 3];
                e += v * v;
                v = imgA[j + 6];
                e += v * v;
                v = imgA[j + 9];
                e += v * v;
                norm1 += e;
            }
            for( ; j < roi.width; j += 3 )
            {
                int v = imgA[j];

                norm1 += v * v;
            }
        }
        norm = sqrt( (double) norm1 );
        break;
    case _CV_DIFF_C:           /* ||A-B|| C  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 9; j += 12 )
            {
                int v = imgA[j] - imgB[j];

                v = _CV_ABS( v );
                if( s < v )
                    s = v;
                v = imgA[j + 3] - imgB[j + 3];
                v = _CV_ABS( v );
                if( s < v )
                    s = v;
                v = imgA[j + 6] - imgB[j + 6];
                v = _CV_ABS( v );
                if( s < v )
                    s = v;
                v = imgA[j + 9] - imgB[j + 9];
                v = _CV_ABS( v );
                if( s < v )
                    s = v;
            }
            for( ; j < roi.width; j += 3 )
            {
                int v = imgA[j] - imgB[j];

                v = _CV_ABS( v );
                if( s < v )
                    s = v;
            }
        }
        norm = s;
        break;
    case _CV_DIFF_L1:          /* ||A-B|| L1  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 9; j += 12 )
            {
                int v = imgA[j] - imgB[j];

                v = _CV_ABS( v );
                s += v;
                v = imgA[j + 3] - imgB[j + 3];
                v = _CV_ABS( v );
                s += v;
                v = imgA[j + 6] - imgB[j + 6];
                v = _CV_ABS( v );
                s += v;
                v = imgA[j + 9] - imgB[j + 9];
                v = _CV_ABS( v );
                s += v;
            }
            for( ; j < roi.width; j += 3 )
                s += _CV_ABS( imgA[j] - imgB[j] );
        }
        norm = s;
        break;
    case _CV_DIFF_L2:          /* ||A-B|| L2  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 9; j += 12 )
            {
                int v = imgA[j] - imgB[j];
                int e = v * v;

                v = imgA[j + 3] - imgB[j + 3];
                e += v * v;
                v = imgA[j + 6] - imgB[j + 6];
                e += v * v;
                v = imgA[j + 9] - imgB[j + 9];
                e += v * v;
                norm1 += e;
            }
            for( ; j < roi.width; j += 3 )
            {
                int v = imgA[j] - imgB[j];

                norm1 += v * v;
            }
        }
        norm = sqrt( (double) norm1 );
        break;
    case _CV_RELATIVE_C:       /* ||A-B|| / ||B||  C */
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 9; j += 12 )
            {
                int e = imgB[j];
                int v = imgA[j] - imgB[j];

                v = _CV_ABS( v );
                if( g < e )
                    g = e;
                if( s < v )
                    s = v;
                e = imgB[j + 3];
                v = imgA[j + 3] - imgB[j + 3];
                v = _CV_ABS( v );
                if( g < e )
                    g = e;
                if( s < v )
                    s = v;
                e = imgB[j + 6];
                v = imgA[j + 6] - imgB[j + 6];
                v = _CV_ABS( v );
                if( g < e )
                    g = e;
                if( s < v )
                    s = v;
                e = imgB[j + 9];
                v = imgA[j + 9] - imgB[j + 9];
                v = _CV_ABS( v );
                if( g < e )
                    g = e;
                if( s < v )
                    s = v;
            }
            for( ; j < roi.width; j += 3 )
            {
                int e = imgB[j];
                int v = imgA[j] - imgB[j];

                v = _CV_ABS( v );
                if( g < e )
                    g = e;
                if( s < v )
                    s = v;
            }
        }
        nrm = (g) ? (double) g : epsilon;
        norm = s / nrm;
        break;
    case _CV_RELATIVE_L1:      /* ||A-B|| / ||B||  L1 */
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 9; j += 12 )
            {
                int v = imgA[j] - imgB[j];

                v = _CV_ABS( v );
                s += v;
                v = imgA[j + 3] - imgB[j + 3];
                v = _CV_ABS( v );
                s += v;
                v = imgA[j + 6] - imgB[j + 6];
                v = _CV_ABS( v );
                s += v;
                v = imgA[j + 9] - imgB[j + 9];
                v = _CV_ABS( v );
                s += v;
                g += imgB[j] + imgB[j + 3] + imgB[j + 6] + imgB[j + 9];
            }
            for( ; j < roi.width; j += 3 )
            {
                int v = imgA[j] - imgB[j];

                v = _CV_ABS( v );
                s += v;
                g += imgB[j];
            }
        }
        nrm = (g) ? (double) g : epsilon;
        norm = s / nrm;
        break;
    case _CV_RELATIVE_L2:      /* ||A-B|| / ||B||  L2 */
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 9; j += 12 )
            {
                int e = imgB[j];
                int v = imgA[j] - imgB[j];
                int e2 = e * e;
                int v2 = v * v;

                e = imgB[j + 3];
                v = imgA[j + 3] - imgB[j + 3];
                e2 += e * e;
                v2 += v * v;
                e = imgB[j + 6];
                v = imgA[j + 6] - imgB[j + 6];
                e2 += e * e;
                v2 += v * v;
                e = imgB[j + 9];
                v = imgA[j + 9] - imgB[j + 9];
                e2 += e * e;
                v2 += v * v;
                norm1 += v2;
                norm2 += e2;
            }
            for( ; j < roi.width; j += 3 )
            {
                int e = imgB[j];
                int v = imgA[j] - imgB[j];

                norm1 += v * v;
                norm2 += e * e;
            }
        }
        nrm = (norm2) ? (double) norm2 : epsilon;
        norm = sqrt( (double) norm1 / nrm );
        break;
    default:
        return CV_BADRANGE_ERR;
    }                           /* end of switch */

    *_norm = norm;
    return CV_OK;
}/*-------- End of icvNorm8uC3  --------------------------------------------------------- */

/* //////////////////////////////////////////////////////////////////////////////////// */
/* //////////////////////////////////////////////////////////////////////////////////// */
/* //////////////////////////////////////////////////////////////////////////////////// */

/* ////////////////////////////////// 8sC1 flavor ///////////////////////////////////// */
IPCVAPI_IMPL( CvStatus, icvNorm_8s_C1R, (const char *imgA, int stepA,
                                         const char *imgB, int stepB,
                                         CvSize roi, int normType, double *_norm) )
{
    int i, j, s = 0, g = 0;
    int64 norm1 = 0, norm2 = 0;
    double norm = -DBL_MAX, nrm, epsilon = 1.0e-30;

    if( !imgA )
        return CV_NULLPTR_ERR;
    if( (roi.width | roi.height) < 0 || stepA < roi.width )
        return CV_BADSIZE_ERR;
    if( imgB && (stepB < roi.width) )
        return CV_BADSIZE_ERR;

    if( imgB )
    {
        if( normType < _CV_RELATIVE_C )
            normType |= _CV_DIFF;
    }
    else
        normType &= _CV_RELATIVE - 1;

    switch (normType)
    {
    case _CV_C:                /* ||A|| C  norm */
        {
            int g = 0;

            for( i = 0; i < roi.height; i++, imgA += stepA )
            {
                for( j = 0; j < roi.width - 3; j += 4 )
                {
                    int v = imgA[j];

                    v = _CV_ABS( v );
                    if( g < v )
                        g = v;
                    v = imgA[j + 1];
                    v = _CV_ABS( v );
                    if( g < v )
                        g = v;
                    v = imgA[j + 2];
                    v = _CV_ABS( v );
                    if( g < v )
                        g = v;
                    v = imgA[j + 3];
                    v = _CV_ABS( v );
                    if( g < v )
                        g = v;
                }
                for( ; j < roi.width; j++ )
                {
                    int v = imgA[j];

                    v = _CV_ABS( v );
                    if( g < v )
                        g = v;
                }
            }
            norm = g;
        }
        break;
    case _CV_L1:               /* ||A|| L1  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA )
        {
            for( j = 0; j < roi.width - 3; j += 4 )
            {
                int v = imgA[j];

                s += _CV_ABS( v );
                v = imgA[j + 1];
                s += _CV_ABS( v );
                v = imgA[j + 2];
                s += _CV_ABS( v );
                v = imgA[j + 3];
                s += _CV_ABS( v );
            }
            for( ; j < roi.width; j++ )
                s += _CV_ABS( imgA[j] );
        }
        norm = s;
        break;
    case _CV_L2:               /* ||A|| L2  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA )
        {
            for( j = 0; j < roi.width - 3; j += 4 )
            {
                int v = imgA[j];
                int e = v * v;

                v = imgA[j + 1];
                e += v * v;
                v = imgA[j + 2];
                e += v * v;
                v = imgA[j + 3];
                e += v * v;
                norm1 += e;
            }
            for( ; j < roi.width; j++ )
            {
                int v = imgA[j];

                norm1 += v * v;
            }
        }
        norm = sqrt( (double) norm1 );
        break;
    case _CV_DIFF_C:           /* ||A-B|| C  norm */
        s = _CV_MAX2( s, _CV_ABS( *imgA - *imgB ));
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 3; j += 4 )
            {
                int v = imgA[j] - imgB[j];

                v = _CV_ABS( v );
                if( s < v )
                    s = v;
                v = imgA[j + 1] - imgB[j + 1];
                v = _CV_ABS( v );
                if( s < v )
                    s = v;
                v = imgA[j + 2] - imgB[j + 2];
                v = _CV_ABS( v );
                if( s < v )
                    s = v;
                v = imgA[j + 3] - imgB[j + 3];
                v = _CV_ABS( v );
                if( s < v )
                    s = v;
            }
            for( ; j < roi.width; j++ )
            {
                int v = imgA[j] - imgB[j];

                v = _CV_ABS( v );
                if( s < v )
                    s = v;
            }
        }
        norm = s;
        break;
    case _CV_DIFF_L1:          /* ||A-B|| L1  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 3; j += 4 )
            {
                int v = imgA[j] - imgB[j];

                s += _CV_ABS( v );
                v = imgA[j + 1] - imgB[j + 1];
                s += _CV_ABS( v );
                v = imgA[j + 2] - imgB[j + 2];
                s += _CV_ABS( v );
                v = imgA[j + 3] - imgB[j + 3];
                s += _CV_ABS( v );
            }
            for( ; j < roi.width; j++ )
                s += _CV_ABS( imgA[j] - imgB[j] );
        }
        norm = s;
        break;
    case _CV_DIFF_L2:          /* ||A-B|| L2  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 3; j += 4 )
            {
                int v = imgA[j] - imgB[j];
                int e = v * v;

                v = imgA[j + 1] - imgB[j + 1];
                e += v * v;
                v = imgA[j + 2] - imgB[j + 2];
                e += v * v;
                v = imgA[j + 3] - imgB[j + 3];
                e += v * v;
                norm1 += e;
            }
            for( ; j < roi.width; j++ )
            {
                int v = imgA[j] - imgB[j];

                norm1 += v * v;
            }
        }
        norm = sqrt( (double) norm1 );
        break;
    case _CV_RELATIVE_C:       /* ||A-B|| / ||B||  C */
        s = _CV_MAX2( s, _CV_ABS( *imgA - *imgB ));
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 3; j += 4 )
            {
                int e = imgB[j];
                int v = imgA[j] - imgB[j];

                e = _CV_ABS( e );
                v = _CV_ABS( v );
                if( g < e )
                    g = e;
                if( s < v )
                    s = v;
                e = imgB[j + 1];
                e = _CV_ABS( e );
                v = imgA[j + 1] - imgB[j + 1];
                v = _CV_ABS( v );
                if( g < e )
                    g = e;
                if( s < v )
                    s = v;
                e = imgB[j + 2];
                e = _CV_ABS( e );
                v = imgA[j + 2] - imgB[j + 2];
                v = _CV_ABS( v );
                if( g < e )
                    g = e;
                if( s < v )
                    s = v;
                e = imgB[j + 3];
                e = _CV_ABS( e );
                v = imgA[j + 3] - imgB[j + 3];
                v = _CV_ABS( v );
                if( g < e )
                    g = e;
                if( s < v )
                    s = v;
            }
            for( ; j < roi.width; j++ )
            {
                int e = imgB[j];
                int v = imgA[j] - imgB[j];

                e = _CV_ABS( e );
                v = _CV_ABS( v );
                if( g < e )
                    g = e;
                if( s < v )
                    s = v;
            }
        }
        nrm = (g) ? (double) g : epsilon;
        norm = s / nrm;
        break;
    case _CV_RELATIVE_L1:      /* ||A-B|| / ||B||  L1 */
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 3; j += 4 )
            {
                int v = imgA[j] - imgB[j];
                int e = imgB[j];

                s += _CV_ABS( v );
                g += _CV_ABS( e );
                v = imgA[j + 1] - imgB[j + 1];
                e = imgB[j + 1];
                s += _CV_ABS( v );
                g += _CV_ABS( e );
                v = imgA[j + 2] - imgB[j + 2];
                e = imgB[j + 2];
                s += _CV_ABS( v );
                g += _CV_ABS( e );
                v = imgA[j + 3] - imgB[j + 3];
                e = imgB[j + 3];
                s += _CV_ABS( v );
                g += _CV_ABS( e );
            }
            for( ; j < roi.width; j++ )
            {
                int v = imgA[j] - imgB[j];
                int e = imgB[j];

                s += _CV_ABS( v );
                g += _CV_ABS( e );
            }
        }
        nrm = (g) ? (double) g : epsilon;
        norm = s / nrm;
        break;
    case _CV_RELATIVE_L2:      /* ||A-B|| / ||B||  L2 */
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 3; j += 4 )
            {
                int e = imgB[j];
                int v = imgA[j] - imgB[j];
                int e2 = e * e;
                int v2 = v * v;

                e = imgB[j + 1];
                v = imgA[j + 1] - imgB[j + 1];
                e2 += e * e;
                v2 += v * v;
                e = imgB[j + 2];
                v = imgA[j + 2] - imgB[j + 2];
                e2 += e * e;
                v2 += v * v;
                e = imgB[j + 3];
                v = imgA[j + 3] - imgB[j + 3];
                e2 += e * e;
                v2 += v * v;
                norm1 += v2;
                norm2 += e2;
            }
            for( ; j < roi.width; j++ )
            {
                int e = imgB[j];
                int v = imgA[j] - imgB[j];

                norm1 += v * v;
                norm2 += e * e;
            }
        }
        nrm = (norm2) ? (double) norm2 : epsilon;
        norm = sqrt( (double) norm1 / nrm );
        break;
    default:
        return CV_BADRANGE_ERR;
    }                           /* end of switch */

    *_norm = norm;
    return CV_OK;
}/*-------- End of icvNorm8sC1R ------------------------------------------------------- */

/* ///////////////////////////////// 8sC3 flavor ////////////////////////////////////// */
IPCVAPI_IMPL( CvStatus, icvNorm_8s_C3CR, (const char *imgA, int stepA,
                                          const char *imgB, int stepB,
                                          CvSize roi, int coi, int normType, double *_norm) )
{
    int i, j, s = 0, g = 0;

#ifndef __STDC__
    int64 norm1 = 0, norm2 = 0;
#else
    double norm1 = 0, norm2 = 0;
#endif
    double norm = 0, nrm, epsilon = 1.0e-40;

    roi.width *= 3;

    if( coi < 1 || coi > 3 )
        return CV_BADRANGE_ERR;
    if( !imgA )
        return CV_NULLPTR_ERR;
    if( (roi.width | roi.height) < 0 || stepA < roi.width )
        return CV_BADSIZE_ERR;
    if( imgB && (stepB < roi.width) )
        return CV_BADSIZE_ERR;

    coi--;
    imgA += coi;

    if( imgB )
        imgB += coi;

    if( imgB )
    {
        if( normType < _CV_RELATIVE_C )
            normType |= _CV_DIFF;
    }
    else
        normType &= _CV_RELATIVE - 1;

    switch (normType)
    {
    case _CV_C:                /* ||A|| C  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA )
        {
            for( j = 0; j < roi.width - 9; j += 12 )
            {
                int v = imgA[j];

                v = _CV_ABS( v );
                if( s < v )
                    s = v;
                v = imgA[j + 3];
                v = _CV_ABS( v );
                if( s < v )
                    s = v;
                v = imgA[j + 6];
                v = _CV_ABS( v );
                if( s < v )
                    s = v;
                v = imgA[j + 9];
                v = _CV_ABS( v );
                if( s < v )
                    s = v;
            }
            for( ; j < roi.width; j += 3 )
            {
                int v = imgA[j];

                v = _CV_ABS( v );
                if( s < v )
                    s = v;
            }
        }
        norm = s;
        break;
    case _CV_L1:               /* ||A|| L1  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA )
        {
            for( j = 0; j < roi.width - 9; j += 12 )
            {
                int v = imgA[j];

                s += _CV_ABS( v );
                v = imgA[j + 3];
                s += _CV_ABS( v );
                v = imgA[j + 6];
                s += _CV_ABS( v );
                v = imgA[j + 9];
                s += _CV_ABS( v );
            }
            for( ; j < roi.width; j += 3 )
                s += _CV_ABS( imgA[j] );
        }
        norm = s;
        break;
    case _CV_L2:               /* ||A|| L2  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA )
        {
            for( j = 0; j < roi.width - 9; j += 12 )
            {
                int v = imgA[j];
                int e = v * v;

                v = imgA[j + 3];
                e += v * v;
                v = imgA[j + 6];
                e += v * v;
                v = imgA[j + 9];
                e += v * v;
                norm1 += e;
            }
            for( ; j < roi.width; j += 3 )
            {
                int v = imgA[j];

                norm1 += v * v;
            }
        }
        norm = sqrt( (double) norm1 );
        break;
    case _CV_DIFF_C:           /* ||A-B|| C  norm */
        s = _CV_MAX2( s, _CV_ABS( *imgA - *imgB ));
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 9; j += 12 )
            {
                int v = imgA[j] - imgB[j];

                v = _CV_ABS( v );
                if( s < v )
                    s = v;
                v = imgA[j + 3] - imgB[j + 3];
                v = _CV_ABS( v );
                if( s < v )
                    s = v;
                v = imgA[j + 6] - imgB[j + 6];
                v = _CV_ABS( v );
                if( s < v )
                    s = v;
                v = imgA[j + 9] - imgB[j + 9];
                v = _CV_ABS( v );
                if( s < v )
                    s = v;
            }
            for( ; j < roi.width; j += 3 )
            {
                int v = imgA[j] - imgB[j];

                v = _CV_ABS( v );
                if( s < v )
                    s = v;
            }
        }
        norm = s;
        break;
    case _CV_DIFF_L1:          /* ||A-B|| L1  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 9; j += 12 )
            {
                int v = imgA[j] - imgB[j];

                s += _CV_ABS( v );
                v = imgA[j + 3] - imgB[j + 3];
                s += _CV_ABS( v );
                v = imgA[j + 6] - imgB[j + 6];
                s += _CV_ABS( v );
                v = imgA[j + 9] - imgB[j + 9];
                s += _CV_ABS( v );
            }
            for( ; j < roi.width; j += 3 )
                s += _CV_ABS( imgA[j] - imgB[j] );
        }
        norm = s;
        break;
    case _CV_DIFF_L2:          /* ||A-B|| L2  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 9; j += 12 )
            {
                int v = imgA[j] - imgB[j];
                int e = v * v;

                v = imgA[j + 3] - imgB[j + 3];
                e += v * v;
                v = imgA[j + 6] - imgB[j + 6];
                e += v * v;
                v = imgA[j + 9] - imgB[j + 9];
                e += v * v;
                norm1 += e;
            }
            for( ; j < roi.width; j += 3 )
            {
                int v = imgA[j] - imgB[j];

                norm1 += v * v;
            }
        }
        norm = sqrt( (double) norm1 );
        break;
    case _CV_RELATIVE_C:       /* ||A-B|| / ||B||  C */
        s = _CV_MAX2( s, _CV_ABS( *imgA - *imgB ));
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 9; j += 12 )
            {
                int e = imgB[j];
                int v = imgA[j] - imgB[j];

                e = _CV_ABS( e );
                v = _CV_ABS( v );
                if( g < e )
                    g = e;
                if( s < v )
                    s = v;
                e = imgB[j + 3];
                e = _CV_ABS( e );
                v = imgA[j + 3] - imgB[j + 3];
                v = _CV_ABS( v );
                if( g < e )
                    g = e;
                if( s < v )
                    s = v;
                e = imgB[j + 6];
                e = _CV_ABS( e );
                v = imgA[j + 6] - imgB[j + 6];
                v = _CV_ABS( v );
                if( g < e )
                    g = e;
                if( s < v )
                    s = v;
                e = imgB[j + 9];
                e = _CV_ABS( e );
                v = imgA[j + 9] - imgB[j + 9];
                v = _CV_ABS( v );
                if( g < e )
                    g = e;
                if( s < v )
                    s = v;
            }
            for( ; j < roi.width; j += 3 )
            {
                int e = imgB[j];
                int v = imgA[j] - imgB[j];

                e = _CV_ABS( e );
                v = _CV_ABS( v );
                if( g < e )
                    g = e;
                if( s < v )
                    s = v;
            }
        }
        nrm = (g) ? (double) g : epsilon;
        norm = s / nrm;
        break;
    case _CV_RELATIVE_L1:      /* ||A-B|| / ||B||  L1 */
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 9; j += 12 )
            {
                int v = imgA[j] - imgB[j];
                int e = imgB[j];

                s += _CV_ABS( v );
                g += _CV_ABS( e );
                v = imgA[j + 3] - imgB[j + 3];
                e = imgB[j + 3];
                s += _CV_ABS( v );
                g += _CV_ABS( e );
                v = imgA[j + 6] - imgB[j + 6];
                e = imgB[j + 6];
                s += _CV_ABS( v );
                g += _CV_ABS( e );
                v = imgA[j + 9] - imgB[j + 9];
                e = imgB[j + 9];
                s += _CV_ABS( v );
                g += _CV_ABS( e );
            }
            for( ; j < roi.width; j += 3 )
            {
                int v = imgA[j] - imgB[j];
                int e = imgB[j];

                s += _CV_ABS( v );
                g += _CV_ABS( e );
            }
        }
        nrm = (g) ? (double) g : epsilon;
        norm = s / nrm;
        break;
    case _CV_RELATIVE_L2:      /* ||A-B|| / ||B||  L2 */
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 9; j += 12 )
            {
                int e = imgB[j];
                int v = imgA[j] - imgB[j];
                int e2 = e * e;
                int v2 = v * v;

                e = imgB[j + 3];
                v = imgA[j + 3] - imgB[j + 3];
                e2 += e * e;
                v2 += v * v;
                e = imgB[j + 6];
                v = imgA[j + 6] - imgB[j + 6];
                e2 += e * e;
                v2 += v * v;
                e = imgB[j + 9];
                v = imgA[j + 9] - imgB[j + 9];
                e2 += e * e;
                v2 += v * v;
                norm1 += v2;
                norm2 += e2;
            }
            for( ; j < roi.width; j += 3 )
            {
                int e = imgB[j];
                int v = imgA[j] - imgB[j];

                norm1 += v * v;
                norm2 += e * e;
            }
        }
        nrm = (norm2) ? (double) norm2 : epsilon;
        norm = sqrt( (double) norm1 / nrm );
        break;
    default:
        return CV_BADRANGE_ERR;
    }                           /* end of switch */

    *_norm = norm;
    return CV_OK;
}

/*------------------------------------------------------------------------------------- */

/* //////////////////////////////////////////////////////////////////////////////////// */
/* //////////////////////////////////////////////////////////////////////////////////// */
/* //////////////////////////////////////////////////////////////////////////////////// */

/* ///////////////////////////////// 32fC1 flavor ///////////////////////////////////// */
IPCVAPI_IMPL( CvStatus, icvNorm_32f_C1R, (const float *imgA, int stepA,
                                          const float *imgB, int stepB,
                                          CvSize roi, int normType, double *_norm) )
{
    int i, j, a0 = 0x7fffffff;
    double norm = 0.e0, epsilon = 1.0e-40;

    if( !imgA )
        return CV_NULLPTR_ERR;
    if( (roi.width | roi.height) < 0 || stepA < roi.width * sizeof_float ||
        stepA % sizeof_float != 0 )
        return CV_BADSIZE_ERR;
    if( imgB && (stepB < roi.width * sizeof_float || stepB % sizeof_float != 0) )
        return CV_BADSIZE_ERR;

    stepA /= sizeof_float;
    stepB /= sizeof_float;

    if( imgB )
    {
        if( normType < _CV_RELATIVE_C )
            normType |= _CV_DIFF;
    }
    else
        normType &= _CV_RELATIVE - 1;

    switch (normType)
    {
    case _CV_C:                /* ||A|| C  norm */
        {
            int g = 0;

            for( i = 0; i < roi.height; i++, imgA += stepA )
            {
                for( j = 0; j < roi.width - 3; j += 4 )
                {
                    int v0 = ((int *) imgA)[j] & a0;
                    int v1 = ((int *) imgA)[j + 1] & a0;
                    int v2 = ((int *) imgA)[j + 2] & a0;
                    int v3 = ((int *) imgA)[j + 3] & a0;

                    if( g < v0 )
                        g = v0;
                    if( g < v1 )
                        g = v1;
                    if( g < v2 )
                        g = v2;
                    if( g < v3 )
                        g = v3;
                }
                for( ; j < roi.width; j++ )
                {
                    int v = ((int *) imgA)[j] & a0;

                    if( g < v )
                        g = v;
                }
            }
            norm = (double) (*(float *) &g);
        }
        break;
    case _CV_L1:               /* ||A|| L1  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA )
        {
            for( j = 0; j < roi.width - 3; j += 4 )
            {
                norm +=
                    fabs( imgA[j] ) + fabs( imgA[j + 1] ) + fabs( imgA[j + 2] ) +
                    fabs( imgA[j + 3] );
            }
            for( ; j < roi.width; j++ )
            {
                int v = ((int *) imgA)[j] & a0;

                norm += *(float *) &v;
            }
        }
        break;
    case _CV_L2:               /* ||A|| L2  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA )
        {
            for( j = 0; j < roi.width - 3; j += 4 )
            {
                double v0 = imgA[j];
                double v1 = imgA[j + 1];
                double v2 = imgA[j + 2];
                double v3 = imgA[j + 3];

                norm += v0 * v0 + v1 * v1 + v2 * v2 + v3 * v3;
            }
            for( ; j < roi.width; j++ )
            {
                double v = imgA[j];

                norm += v * v;
            }
        }
        norm = sqrt( norm );
        break;
    case _CV_DIFF_C:           /* ||A-B|| C  norm */
        {
            double g = 0;

            for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
            {
                for( j = 0; j < roi.width - 3; j += 4 )
                {
                    double v0 = fabs( imgA[j] - imgB[j] );
                    double v1 = fabs( imgA[j + 1] - imgB[j + 1] );
                    double v2 = fabs( imgA[j + 2] - imgB[j + 2] );
                    double v3 = fabs( imgA[j + 3] - imgB[j + 3] );

                    if( g < v0 )
                        g = v0;
                    if( g < v1 )
                        g = v1;
                    if( g < v2 )
                        g = v2;
                    if( g < v3 )
                        g = v3;
                }
                for( ; j < roi.width; j++ )
                {
                    double v = fabs( imgA[j] - imgB[j] );

                    if( g < v )
                        g = v;
                }
            }
            norm = g;
        }
        break;
    case _CV_DIFF_L1:          /* ||A-B|| L1  norm */
        {
            for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
            {
                for( j = 0; j < roi.width - 3; j += 4 )
                {
                    double c0 = fabs( imgA[j] - imgB[j] );
                    double c1 = fabs( imgA[j + 1] - imgB[j + 1] );
                    double c2 = fabs( imgA[j + 2] - imgB[j + 2] );
                    double c3 = fabs( imgA[j + 3] - imgB[j + 3] );

                    norm += c0 + c1 + c2 + c3;
                }
                for( ; j < roi.width; j++ )
                {
                    double v = imgA[j] - imgB[j];

                    norm += v < 0 ? -v : v;
                }
            }
        }
        break;
    case _CV_DIFF_L2:          /* ||A-B|| L2  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 3; j += 4 )
            {
                double v0 = imgA[j] - imgB[j];
                double v1 = imgA[j + 1] - imgB[j + 1];
                double v2 = imgA[j + 2] - imgB[j + 2];
                double v3 = imgA[j + 3] - imgB[j + 3];

                norm += v0 * v0 + v1 * v1 + v2 * v2 + v3 * v3;
            }
            for( ; j < roi.width; j++ )
            {
                double v = imgA[j] - imgB[j];

                norm += v * v;
            }
        }
        norm = sqrt( norm );
        break;
    case _CV_RELATIVE_C:       /* ||A-B||/||B|| C  norm */
        {
            double bb = 0, cc = 0;

            for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
            {
                for( j = 0; j < roi.width - 3; j += 4 )
                {
                    double b0 = imgB[j];
                    double b1 = imgB[j + 1];
                    double c0 = fabs( imgA[j] - b0 );
                    double c1 = fabs( imgA[j + 1] - b1 );

                    b0 = fabs( b0 );
                    b1 = fabs( b1 );
                    if( bb < b0 )
                        bb = b0;
                    if( bb < b1 )
                        bb = b1;
                    if( cc < c0 )
                        cc = c0;
                    if( cc < c1 )
                        cc = c1;
                    b0 = imgB[j + 2];
                    b1 = imgB[j + 3];
                    c0 = fabs( imgA[j + 2] - b0 );
                    c1 = fabs( imgA[j + 3] - b1 );
                    b0 = fabs( b0 );
                    b1 = fabs( b1 );
                    if( bb < b0 )
                        bb = b0;
                    if( bb < b1 )
                        bb = b1;
                    if( cc < c0 )
                        cc = c0;
                    if( cc < c1 )
                        cc = c1;
                }
                for( ; j < roi.width; j++ )
                {
                    double b0 = imgB[j];
                    double c0 = fabs( imgA[j] - b0 );

                    b0 = fabs( b0 );
                    if( bb < b0 )
                        bb = b0;
                    if( cc < c0 )
                        cc = c0;
                }
            }
            norm = cc / ((bb) ? bb : epsilon);
        }
        break;
    case _CV_RELATIVE_L1:      /* ||A-B||/||B|| L1  norm */
        {
            double bb = 0, cc = 0;

            for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
            {
                for( j = 0; j < roi.width - 3; j += 4 )
                {
                    double b0 = imgB[j];
                    double b1 = imgB[j + 1];
                    double c0 = fabs( imgA[j] - b0 );
                    double c1 = fabs( imgA[j + 1] - b1 );

                    bb += fabs( b0 ) + fabs( b1 );
                    cc += c0 + c1;
                    b0 = imgB[j + 2];
                    b1 = imgB[j + 3];
                    c0 = fabs( imgA[j + 2] - b0 );
                    c1 = fabs( imgA[j + 3] - b1 );
                    bb += fabs( b0 ) + fabs( b1 );
                    cc += c0 + c1;
                }
                for( ; j < roi.width; j++ )
                {
                    double b0 = imgB[j];
                    double c0 = fabs( imgA[j] - b0 );

                    bb += fabs( b0 );
                    cc += c0;
                }
            }
            norm = cc / ((bb) ? bb : epsilon);
        }
        break;
    case _CV_RELATIVE_L2:      /* ||A-B||/||B|| L2  norm */
        {
            double bb = 0, cc = 0;

            for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
            {
                for( j = 0; j < roi.width - 3; j += 4 )
                {
                    double b0 = imgB[j];
                    double b1 = imgB[j + 1];
                    double c0 = imgA[j] - b0;
                    double c1 = imgA[j + 1] - b1;

                    bb += b0 * b0 + b1 * b1;
                    cc += c0 * c0 + c1 * c1;
                    b0 = imgB[j + 2];
                    b1 = imgB[j + 3];
                    c0 = imgA[j + 2] - b0;
                    c1 = imgA[j + 3] - b1;
                    bb += b0 * b0 + b1 * b1;
                    cc += c0 * c0 + c1 * c1;
                }
                for( ; j < roi.width; j++ )
                {
                    double b0 = imgB[j];
                    double c0 = imgA[j] - b0;

                    bb += b0 * b0;
                    cc += c0 * c0;
                }
            }
            norm = sqrt( cc / ((bb) ? bb : epsilon) );
        }
        break;
    default:
        return CV_BADRANGE_ERR;
    }                           /* end of switch */

    *_norm = norm;
    return CV_OK;
}/*-------- End of icvNorm32fC1R ----------------------------------------------------- */

/* /////////////////////////////////// 32fC3 flavor /////////////////////////////////// */

IPCVAPI_IMPL( CvStatus, icvNorm_32f_C3CR, (const float *imgA, int stepA,
                                           const float *imgB, int stepB,
                                           CvSize roi, int coi, int normType, double *_norm) )
{
    int i, j, a0 = 0x7fffffff;
    double norm = 0.e0, epsilon = 1.0e-12;

    roi.width *= 3;

    if( !imgA )
        return CV_NULLPTR_ERR;
    if( (roi.width | roi.height) < 0 || stepA < roi.width * sizeof_float ||
        stepA % sizeof_float != 0 )
        return CV_BADSIZE_ERR;
    if( imgB && (stepB < roi.width * sizeof_float || stepB % sizeof_float != 0) )
        return CV_BADSIZE_ERR;

    stepA /= sizeof_float;
    stepB /= sizeof_float;
    coi--;
    imgA += coi;
    if( imgB )
        imgB += coi;

    if( imgB )
    {
        if( normType < _CV_RELATIVE_C )
            normType |= _CV_DIFF;
    }
    else
        normType &= _CV_RELATIVE - 1;

    switch (normType)
    {
    case _CV_C:                /* ||A|| C  norm */
        {
            int g = 0;

            for( i = 0; i < roi.height; i++, imgA += stepA )
            {
                for( j = 0; j < roi.width - 11; j += 12 )
                {
                    int v0 = ((int *) imgA)[j] & a0;
                    int v1 = ((int *) imgA)[j + 3] & a0;
                    int v2 = ((int *) imgA)[j + 6] & a0;
                    int v3 = ((int *) imgA)[j + 9] & a0;

                    if( g < v0 )
                        g = v0;
                    if( g < v1 )
                        g = v1;
                    if( g < v2 )
                        g = v2;
                    if( g < v3 )
                        g = v3;
                }
                for( ; j < roi.width; j += 3 )
                {
                    int v = ((int *) imgA)[j] & a0;

                    if( g < v )
                        g = v;
                }
            }
            norm = (double) (*(float *) &g);
        }
        break;
    case _CV_L1:               /* ||A|| L1  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA )
        {
            for( j = 0; j < roi.width - 11; j += 12 )
                norm +=
                    fabs( imgA[j] ) + fabs( imgA[j + 3] ) + fabs( imgA[j + 6] ) +
                    fabs( imgA[j + 9] );
            for( ; j < roi.width; j += 3 )
            {
                int v = ((int *) imgA)[j] & a0;

                norm += *(float *) &v;
            }
        }
        break;
    case _CV_L2:               /* ||A|| L2  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA )
        {
            for( j = 0; j < roi.width - 11; j += 12 )
            {
                double v0 = imgA[j];
                double v1 = imgA[j + 3];
                double v2 = imgA[j + 6];
                double v3 = imgA[j + 9];

                norm += v0 * v0 + v1 * v1 + v2 * v2 + v3 * v3;
            }
            for( ; j < roi.width; j += 3 )
            {
                double v = imgA[j];

                norm += v * v;
            }
        }
        norm = sqrt( norm );
        break;
    case _CV_DIFF_C:           /* ||A-B|| C  norm */
        {
            double g = 0;

            for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
            {
                for( j = 0; j < roi.width - 11; j += 12 )
                {
                    double v0 = fabs( imgA[j] - imgB[j] );
                    double v1 = fabs( imgA[j + 3] - imgB[j + 3] );
                    double v2 = fabs( imgA[j + 6] - imgB[j + 6] );
                    double v3 = fabs( imgA[j + 9] - imgB[j + 9] );

                    if( g < v0 )
                        g = v0;
                    if( g < v1 )
                        g = v1;
                    if( g < v2 )
                        g = v2;
                    if( g < v3 )
                        g = v3;
                }
                for( ; j < roi.width; j += 3 )
                {
                    double v = fabs( imgA[j] - imgB[j] );

                    if( g < v )
                        g = v;
                }
            }
            norm = g;
        }
        break;
    case _CV_DIFF_L1:          /* ||A-B|| L1  norm */
        {
            for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
            {
                for( j = 0; j < roi.width - 11; j += 12 )
                {
                    double c0 = fabs( imgA[j] - imgB[j] );
                    double c1 = fabs( imgA[j + 3] - imgB[j + 3] );
                    double c2 = fabs( imgA[j + 6] - imgB[j + 6] );
                    double c3 = fabs( imgA[j + 9] - imgB[j + 9] );

                    norm += c0 + c1 + c2 + c3;
                }
                for( ; j < roi.width; j += 3 )
                {
                    double v = imgA[j] - imgB[j];

                    norm += v < 0 ? -v : v;
                }
            }
        }
        break;
    case _CV_DIFF_L2:          /* ||A-B|| L2  norm */
        for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
        {
            for( j = 0; j < roi.width - 11; j += 12 )
            {
                double v0 = imgA[j] - imgB[j];
                double v1 = imgA[j + 3] - imgB[j + 3];
                double v2 = imgA[j + 6] - imgB[j + 6];
                double v3 = imgA[j + 9] - imgB[j + 9];

                norm += v0 * v0 + v1 * v1 + v2 * v2 + v3 * v3;
            }
            for( ; j < roi.width; j += 3 )
            {
                double v = imgA[j] - imgB[j];

                norm += v * v;
            }
        }
        norm = sqrt( norm );
        break;
    case _CV_RELATIVE_C:       /* ||A-B||/||B|| C  norm */
        {
            double bb = 0, cc = 0;

            for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
            {
                for( j = 0; j < roi.width - 11; j += 12 )
                {
                    double b0 = imgB[j];
                    double b1 = imgB[j + 3];
                    double c0 = fabs( imgA[j] - b0 );
                    double c1 = fabs( imgA[j + 3] - b1 );

                    b0 = fabs( b0 );
                    b1 = fabs( b1 );
                    if( bb < b0 )
                        bb = b0;
                    if( bb < b1 )
                        bb = b1;
                    if( cc < c0 )
                        cc = c0;
                    if( cc < c1 )
                        cc = c1;
                    b0 = imgB[j + 6];
                    b1 = imgB[j + 9];
                    c0 = fabs( imgA[j + 6] - b0 );
                    c1 = fabs( imgA[j + 9] - b1 );
                    b0 = fabs( b0 );
                    b1 = fabs( b1 );
                    if( bb < b0 )
                        bb = b0;
                    if( bb < b1 )
                        bb = b1;
                    if( cc < c0 )
                        cc = c0;
                    if( cc < c1 )
                        cc = c1;
                }
                for( ; j < roi.width; j += 3 )
                {
                    double b0 = imgB[j];
                    double c0 = fabs( imgA[j] - b0 );

                    b0 = fabs( b0 );
                    if( bb < b0 )
                        bb = b0;
                    if( cc < c0 )
                        cc = c0;
                }
            }
            norm = cc / ((bb) ? bb : epsilon);
        }
        break;
    case _CV_RELATIVE_L1:      /* ||A-B||/||B|| L1  norm */
        {
            double bb = 0, cc = 0;

            for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
            {
                for( j = 0; j < roi.width - 11; j += 12 )
                {
                    double b0 = imgB[j];
                    double b1 = imgB[j + 3];
                    double c0 = fabs( imgA[j] - b0 );
                    double c1 = fabs( imgA[j + 3] - b1 );

                    bb += fabs( b0 ) + fabs( b1 );
                    cc += c0 + c1;
                    b0 = imgB[j + 6];
                    b1 = imgB[j + 9];
                    c0 = fabs( imgA[j + 6] - b0 );
                    c1 = fabs( imgA[j + 9] - b1 );
                    bb += fabs( b0 ) + fabs( b1 );
                    cc += c0 + c1;
                }
                for( ; j < roi.width; j += 3 )
                {
                    double b0 = imgB[j];
                    double c0 = fabs( imgA[j] - b0 );

                    bb += fabs( b0 );
                    cc += c0;
                }
            }
            norm = cc / ((bb) ? bb : epsilon);
        }
        break;
    case _CV_RELATIVE_L2:      /* ||A-B||/||B|| L2  norm */
        {
            double bb = 0, cc = 0;

            for( i = 0; i < roi.height; i++, imgA += stepA, imgB += stepB )
            {
                for( j = 0; j < roi.width - 11; j += 12 )
                {
                    double b0 = imgB[j];
                    double b1 = imgB[j + 3];
                    double c0 = imgA[j] - b0;
                    double c1 = imgA[j + 3] - b1;

                    bb += b0 * b0 + b1 * b1;
                    cc += c0 * c0 + c1 * c1;
                    b0 = imgB[j + 6];
                    b1 = imgB[j + 9];
                    c0 = imgA[j + 6] - b0;
                    c1 = imgA[j + 9] - b1;
                    bb += b0 * b0 + b1 * b1;
                    cc += c0 * c0 + c1 * c1;
                }
                for( ; j < roi.width; j += 3 )
                {
                    double b0 = imgB[j];
                    double c0 = imgA[j] - b0;

                    bb += b0 * b0;
                    cc += c0 * c0;
                }
            }
            norm = sqrt( cc / ((bb) ? bb : epsilon) );
        }
        break;
    default:
        return CV_BADRANGE_ERR;
    }                           /* end of switch */

    *_norm = norm;
    return CV_OK;
}/*-------- End of icvNorm32fC3R ----------------------------------------------------- */

/* =====  END OF FILE  ===== */
