/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/
#include "_cv.h"
#include "_cvwrap.h"

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    cvCalcOpticalFlowBM
//    Purpose: Optical flow implementation
//    Context: 
//    Parameters:
//             srcA, srcB - source image
//             velx, vely - destination image
//    Returns:
//
//    Notes:
//F*/
CV_IMPL void
cvCalcOpticalFlowBM( IplImage * srcA, IplImage * srcB,
                     CvSize blockSize, CvSize shiftSize,
                     CvSize maxRange, int usePrevious, IplImage * velx, IplImage * vely )
{
    CV_FUNCNAME( "cvCalcOpticalFlowBM" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( srcA ));
    CV_CALL( CV_CHECK_IMAGE( srcB ));
    CV_CALL( CV_CHECK_IMAGE( velx ));
    CV_CALL( CV_CHECK_IMAGE( vely ));

    if( srcA->depth != IPL_DEPTH_8U || srcB->depth != IPL_DEPTH_8U ||
        velx->depth != IPL_DEPTH_32F || vely->depth != IPL_DEPTH_32F )
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );

    if( srcA->nChannels != 1 || srcA->nChannels != 1 ||
        velx->nChannels != 1 || vely->nChannels != 1 )
        CV_ERROR( IPL_BadNumChannels, "input image has more than one channel" );

    CvSize roi_srcA;
    CvSize roi_srcB;
    uchar *data_srcA;
    uchar *data_srcB;

    CvSize roi_velx;
    CvSize roi_vely;
    uchar *data_velx;
    uchar *data_vely;

    int step_src;
    int step_vel;

    cvGetImageRawData( srcA, &data_srcA, &step_src, &roi_srcA );
    cvGetImageRawData( srcB, &data_srcB, &step_src, &roi_srcB );
    cvGetImageRawData( velx, &data_velx, &step_vel, &roi_velx );
    cvGetImageRawData( vely, &data_vely, &step_vel, &roi_vely );

    if( roi_srcA.height != roi_srcB.height || roi_srcA.width != roi_srcB.width ||
        roi_velx.height != roi_vely.height || roi_velx.width != roi_vely.width )
        CV_ERROR( IPL_BadROISize, "source and destination images have different ROI" );

    if( srcA->widthStep != srcB->widthStep || velx->widthStep != vely->widthStep )
        CV_ERROR( IPL_BadStep, "source and destination images have different step" );

    IPPI_CALL( icvCalcOpticalFlowBM_8u32fR( data_srcA,
                                             data_srcB,
                                             step_src,
                                             roi_srcA,
                                             blockSize, shiftSize, maxRange, usePrevious,
                                             (float *) data_velx,
                                             (float *) data_vely, step_vel ));
    __CLEANUP__;
    __END__;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    cvCalcOpticalFlowLK
//    Purpose: Optical flow implementation
//    Context: 
//    Parameters:
//             srcA, srcB - source image
//             velx, vely - destination image
//    Returns:
//
//    Notes:
//F*/
CV_IMPL void
cvCalcOpticalFlowLK( IplImage * srcA, IplImage * srcB, CvSize winSize,
                     IplImage * velx, IplImage * vely )
{
    CV_FUNCNAME( "cvCalcOpticalFlowLK" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( srcA ));
    CV_CALL( CV_CHECK_IMAGE( srcB ));
    CV_CALL( CV_CHECK_IMAGE( velx ));
    CV_CALL( CV_CHECK_IMAGE( vely ));

    if( srcA->depth != IPL_DEPTH_8U || srcB->depth != IPL_DEPTH_8U ||
        velx->depth != IPL_DEPTH_32F || vely->depth != IPL_DEPTH_32F )
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );

    if( srcA->nChannels != 1 || srcA->nChannels != 1 ||
        velx->nChannels != 1 || vely->nChannels != 1 )
        CV_ERROR( IPL_BadNumChannels, "input image has more than one channel" );

    CvSize roi_srcA;
    CvSize roi_srcB;
    uchar *data_srcA;
    uchar *data_srcB;

    CvSize roi_velx;
    CvSize roi_vely;
    uchar *data_velx;
    uchar *data_vely;

    int step_src;
    int step_vel;

    cvGetImageRawData( srcA, &data_srcA, &step_src, &roi_srcA );
    cvGetImageRawData( srcB, &data_srcB, &step_src, &roi_srcB );
    cvGetImageRawData( velx, &data_velx, &step_vel, &roi_velx );
    cvGetImageRawData( vely, &data_vely, &step_vel, &roi_vely );

    if( roi_srcA.height != roi_srcB.height || roi_srcA.width != roi_srcB.width ||
        roi_velx.height != roi_vely.height || roi_velx.width != roi_vely.width )
        CV_ERROR( IPL_BadROISize, "source and destination images have different ROI" );

    if( srcA->widthStep != srcB->widthStep || velx->widthStep != vely->widthStep )
        CV_ERROR( IPL_BadStep, "source and destination images have different step" );

    IPPI_CALL( icvCalcOpticalFlowLK_8u32fR( data_srcA,
                                             data_srcB,
                                             step_src,
                                             roi_srcA,
                                             winSize,
                                             (float *) data_velx,
                                             (float *) data_vely, step_vel ));
    __CLEANUP__;
    __END__;
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    cvCalcOpticalFlowHS
//    Purpose: Optical flow implementation
//    Context: 
//    Parameters:
//             srcA, srcB - source image
//             velx, vely - destination image
//    Returns:
//
//    Notes:
//F*/
CV_IMPL void
cvCalcOpticalFlowHS( IplImage * srcA, IplImage * srcB, int usePrevious,
                     IplImage * velx, IplImage * vely, double lambda, CvTermCriteria criteria )
{
    CV_FUNCNAME( "cvCalcOpticalFlowHS" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( srcA ));
    CV_CALL( CV_CHECK_IMAGE( srcB ));
    CV_CALL( CV_CHECK_IMAGE( velx ));
    CV_CALL( CV_CHECK_IMAGE( vely ));

    if( srcA->depth != IPL_DEPTH_8U || srcB->depth != IPL_DEPTH_8U ||
        velx->depth != IPL_DEPTH_32F || vely->depth != IPL_DEPTH_32F )
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );

    if( srcA->nChannels != 1 || srcA->nChannels != 1 ||
        velx->nChannels != 1 || vely->nChannels != 1 )
        CV_ERROR( IPL_BadNumChannels, "input image has more than one channel" );

    CvSize roi_srcA;
    CvSize roi_srcB;
    uchar *data_srcA;
    uchar *data_srcB;

    CvSize roi_velx;
    CvSize roi_vely;
    uchar *data_velx;
    uchar *data_vely;

    int step_src;
    int step_vel;

    cvGetImageRawData( srcA, &data_srcA, &step_src, &roi_srcA );
    cvGetImageRawData( srcB, &data_srcB, &step_src, &roi_srcB );
    cvGetImageRawData( velx, &data_velx, &step_vel, &roi_velx );
    cvGetImageRawData( vely, &data_vely, &step_vel, &roi_vely );

    if( roi_srcA.height != roi_srcB.height || roi_srcA.width != roi_srcB.width ||
        roi_velx.height != roi_vely.height || roi_velx.width != roi_vely.width )
        CV_ERROR( IPL_BadROISize, "source and destination images have different ROI" );

    if( srcA->widthStep != srcB->widthStep || velx->widthStep != vely->widthStep )
        CV_ERROR( IPL_BadStep, "source and destination images have different step" );

    IPPI_CALL( icvCalcOpticalFlowHS_8u32fR( data_srcA,
                                             data_srcB,
                                             step_src,
                                             roi_srcA,
                                             usePrevious,
                                             (float *) data_velx,
                                             (float *) data_vely,
                                             step_vel, (float) lambda, criteria ));
    __CLEANUP__;
    __END__;
}


/* End of file. */
