/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include "_cv.h"

#define _CV_ALIGN  8

/* macro for calculating cyclic buffer step */
#define _CV_PYR_BUF_STEP_8U( width, Cs ) \
   ((((width)+_CV_ALIGN*2)*Cs*sizeof(int) + _CV_ALIGN*2 - 1) &-_CV_ALIGN*2)


/****************************************************************************************\

                         Down-sampling pyramids core functions

\****************************************************************************************/

/****************************************************************************************\
                               Filtering macros (integer)
\****************************************************************************************/

/* COMMON CASE */
/* 1/16[1    4    6    4    1]       */
/* ...| x0 | x1 | x2 | x3 | x4 |...  */
#define  PD_FILTER( x0, x1, x2, x3, x4 ) ((x2)*6+((x1)+(x3))*4+(x0)+(x4))

/* MACROS FOR BORDERS */

/* | b I a | b | reflection used ("I" denotes the image boundary) */

/* LEFT/TOP */
/* 1/16[1    4    6    4    1]       */
/*    | x2 | x1 I x0 | x1 | x2 |...  */
#define  PD_LT(x0,x1,x2)                 ((x0)*6 + (x1)*8 + (x2)*2)

/* RIGHT/BOTTOM */
/* 1/16[1    4    6    4    1]       */
/* ...| x0 | x1 | x2 | x3 I x2 |     */
#define  PD_RB(x0,x1,x2,x3)              ((x0) + ((x1) + (x3))*4 + (x2)*7)

/* SINGULAR CASE ( width == 2 || height == 2 ) */
/* 1/16[1    4    6    4    1]       */
/*    | x0 | x1 I x0 | x1 I x0 |     */
#define  PD_SINGULAR(x0,x1)              (((x0) + (x1))*8)

#define  PD_SCALE(x)                     (((x) + (1<<7)) >> 8)

#define  PD_SZ  5

static CvStatus
icvPyrDownG5x5_8u( const uchar * src, int srcStep,
                   uchar * dst, int dstStep, CvSize srcRoi, void *buf, int Cs )
{
    int *buffer = (int *) buf;  /* pointer to temporary buffer */
    int *rows[PD_SZ];           /* array of rows pointers. dim(rows) is PD_SZ */
    int y, top_row = 0;
    int W = srcRoi.width;
    int H = srcRoi.height;
    int Wd = W / 2, Wdn = Wd * Cs;
    int buffer_step = _CV_PYR_BUF_STEP_8U( W, Cs ) / sizeof( rows[0][0] );

    assert( Cs == 1 || Cs == 3 );

    /* main loop */
    for( y = 0; y < H; y += 2, dst += dstStep )
    {
        /* set first and last indices of buffer rows which are need to be filled */
        int fst = y == 0 ? 0 : PD_SZ - 2,
            lst = y == 0 ? (H <= PD_SZ / 2 ? H : PD_SZ / 2 + 1) :
            y < H - PD_SZ / 2 ? PD_SZ : H + PD_SZ / 2 - y;
        int x, y1, k = top_row;
        int x1 = buffer_step;
        int *row01, *row23, *row4;

        /* assign rows pointers */
        for( y1 = 0; y1 < PD_SZ; y1++ )
        {
            rows[y1] = buffer + k * buffer_step;
            k++;
            k &= k <= PD_SZ ? -1 : 0;
        }

        row01 = rows[0];
        row23 = rows[2];
        row4 = rows[4];

        /* fill new buffer rows with filtered source (horizontal conv) */
        if( Cs == 1 )
        {
            if( W > PD_SZ / 2 )
                for( y1 = fst; y1 < lst; y1++, src += srcStep )
                {
                    int *row = rows[y1];

                    /* process left & right bounds */
                    row[0] = PD_LT( src[0], src[1], src[2] );
                    row[Wd - 1] = PD_RB( src[Wd * 2 - 4], src[Wd * 2 - 3],
                                         src[Wd * 2 - 2], src[Wd * 2 - 1] );
                    /* other points (even) */
                    for( x = 1; x < Wd - 1; x++ )
                    {
                        row[x] = PD_FILTER( src[2 * x - 2], src[2 * x - 1], src[2 * x],
                                            src[2 * x + 1], src[2 * x + 2] );
                    }
                }
            else
                for( y1 = fst; y1 < lst; y1++, src += srcStep )
                {
                    rows[y1][0] = PD_SINGULAR( src[0], src[1] );
                }
        }
        else                    /* Cs == 3 */
        {
            for( y1 = fst; y1 < lst; y1++, src += srcStep )
            {
                int *row = rows[y1];

                if( W > PD_SZ / 2 )
                {
                    int c;

                    for( c = 0; c < 3; c++ )
                    {
                        /* process left & right bounds  */
                        row[c] = PD_LT( src[c], src[3 + c], src[6 + c] );
                        row[Wdn - 3 + c] = PD_RB( src[Wdn * 2 - 12 + c], src[Wdn * 2 - 9 + c],
                                                  src[Wdn * 2 - 6 + c], src[Wdn * 2 - 3 + c] );
                    }
                    /* other points (even) */
                    for( x = 3; x < Wdn - 3; x += 3 )
                    {
                        row[x] = PD_FILTER( src[2 * x - 6], src[2 * x - 3], src[2 * x],
                                            src[2 * x + 3], src[2 * x + 6] );
                        row[x + 1] = PD_FILTER( src[2 * x - 5], src[2 * x - 2], src[2 * x + 1],
                                                src[2 * x + 4], src[2 * x + 7] );
                        row[x + 2] = PD_FILTER( src[2 * x - 4], src[2 * x - 1], src[2 * x + 2],
                                                src[2 * x + 5], src[2 * x + 8] );
                    }
                }
                else            /* W <= PD_SZ/2 */
                {
                    row[0] = PD_SINGULAR( src[0], src[3] );
                    row[1] = PD_SINGULAR( src[1], src[4] );
                    row[2] = PD_SINGULAR( src[2], src[5] );
                }
            }
        }

        /* second pass. Do vertical conv and write results do destination image */
        if( y > 0 )
        {
            if( y < H - PD_SZ / 2 )
            {
                for( x = 0; x < Wdn; x++, x1++ )
                {
                    dst[x] = (uchar) PD_SCALE( PD_FILTER( row01[x], row01[x1], row23[x],
                                                          row23[x1], row4[x] ));
                }
                top_row += 2;
                top_row &= top_row <= PD_SZ ? -1 : 0;
            }
            else                /* bottom */
                for( x = 0; x < Wdn; x++, x1++ )
                    dst[x] = (uchar) PD_SCALE( PD_RB( row01[x], row01[x1],
                                                      row23[x], row23[x1] ));
        }
        else
        {
            if( H > PD_SZ / 2 ) /* top */
                for( x = 0; x < Wdn; x++, x1++ )
                    dst[x] = (uchar) PD_SCALE( PD_LT( row01[x], row01[x1], row23[x] ));
            else                /* H <= PD_SZ/2 */
                for( x = 0; x < Wdn; x++, x1++ )
                    dst[x] = (uchar) PD_SCALE( PD_SINGULAR( row01[x], row01[x1] ));
        }
    }

    return CV_OK;
}


/****************************************************************************************\
*                                  8s version                                            *
\****************************************************************************************/

static CvStatus
icvPyrDownG5x5_8s( const char *src, int srcStep,
                   char *dst, int dstStep, CvSize srcRoi, void *buf, int Cs )
{
    int *buffer = (int *) buf;  /* pointer to temporary buffer */
    int *rows[PD_SZ];           /* array of rows pointers. dim(rows) is PD_SZ */
    int y, top_row = 0;
    int W = srcRoi.width;
    int H = srcRoi.height;
    int Wd = W / 2, Wdn = Wd * Cs;
    int buffer_step = _CV_PYR_BUF_STEP_8U( W, Cs ) / sizeof( rows[0][0] );

    assert( Cs == 1 || Cs == 3 );

    /* main loop */
    for( y = 0; y < H; y += 2, dst += dstStep )
    {
        /* set first and last indices of buffer rows which are need to be filled */
        int fst = y == 0 ? 0 : PD_SZ - 2,
            lst = y == 0 ? (H <= PD_SZ / 2 ? H : PD_SZ / 2 + 1) :
            y < H - PD_SZ / 2 ? PD_SZ : H + PD_SZ / 2 - y;
        int x, y1, k = top_row;
        int x1 = buffer_step;
        int *row01, *row23, *row4;

        /* assign rows pointers */
        for( y1 = 0; y1 < PD_SZ; y1++ )
        {
            rows[y1] = buffer + k * buffer_step;
            k++;
            k &= k <= PD_SZ ? -1 : 0;
        }

        row01 = rows[0];
        row23 = rows[2];
        row4 = rows[4];

        /* fill new buffer rows with filtered source (horizontal conv) */
        if( Cs == 1 )
        {
            if( W > PD_SZ / 2 )
                for( y1 = fst; y1 < lst; y1++, src += srcStep )
                {
                    int *row = rows[y1];

                    /* process left & right bounds */
                    row[0] = PD_LT( src[0], src[1], src[2] );
                    row[Wd - 1] = PD_RB( src[Wd * 2 - 4], src[Wd * 2 - 3],
                                         src[Wd * 2 - 2], src[Wd * 2 - 1] );
                    /* other points (even) */
                    for( x = 1; x < Wd - 1; x++ )
                    {
                        row[x] = PD_FILTER( src[2 * x - 2], src[2 * x - 1], src[2 * x],
                                            src[2 * x + 1], src[2 * x + 2] );
                    }
                }
            else
                for( y1 = fst; y1 < lst; y1++, src += srcStep )
                {
                    rows[y1][0] = PD_SINGULAR( src[0], src[1] );
                }
        }
        else                    /* Cs == 3 */
        {
            for( y1 = fst; y1 < lst; y1++, src += srcStep )
            {
                int *row = rows[y1];

                if( W > PD_SZ / 2 )
                {
                    int c;

                    for( c = 0; c < 3; c++ )
                    {
                        /* process left & right bounds  */
                        row[c] = PD_LT( src[c], src[3 + c], src[6 + c] );
                        row[Wdn - 3 + c] = PD_RB( src[Wdn * 2 - 12 + c], src[Wdn * 2 - 9 + c],
                                                  src[Wdn * 2 - 6 + c], src[Wdn * 2 - 3 + c] );
                    }
                    /* other points (even) */
                    for( x = 3; x < Wdn - 3; x += 3 )
                    {
                        row[x] = PD_FILTER( src[2 * x - 6], src[2 * x - 3], src[2 * x],
                                            src[2 * x + 3], src[2 * x + 6] );
                        row[x + 1] = PD_FILTER( src[2 * x - 5], src[2 * x - 2], src[2 * x + 1],
                                                src[2 * x + 4], src[2 * x + 7] );
                        row[x + 2] = PD_FILTER( src[2 * x - 4], src[2 * x - 1], src[2 * x + 2],
                                                src[2 * x + 5], src[2 * x + 8] );
                    }
                }
                else            /* W <= PD_SZ/2 */
                {
                    row[0] = PD_SINGULAR( src[0], src[3] );
                    row[1] = PD_SINGULAR( src[1], src[4] );
                    row[2] = PD_SINGULAR( src[2], src[5] );
                }
            }
        }

        /* second pass. Do vertical conv and write results do destination image */
        if( y > 0 )
        {
            if( y < H - PD_SZ / 2 )
            {
                for( x = 0; x < Wdn; x++, x1++ )
                {
                    dst[x] = (char) PD_SCALE( PD_FILTER( row01[x], row01[x1], row23[x],
                                                         row23[x1], row4[x] ));
                }
                top_row += 2;
                top_row &= top_row <= PD_SZ ? -1 : 0;
            }
            else                /* bottom */
                for( x = 0; x < Wdn; x++, x1++ )
                    dst[x] = (char) PD_SCALE( PD_RB( row01[x], row01[x1],
                                                     row23[x], row23[x1] ));
        }
        else
        {
            if( H > PD_SZ / 2 ) /* top */
                for( x = 0; x < Wdn; x++, x1++ )
                    dst[x] = (char) PD_SCALE( PD_LT( row01[x], row01[x1], row23[x] ));
            else                /* H <= PD_SZ/2 */
                for( x = 0; x < Wdn; x++, x1++ )
                    dst[x] = (char) PD_SCALE( PD_SINGULAR( row01[x], row01[x1] ));
        }
    }

    return CV_OK;
}


/****************************************************************************************\

                           Up-sampling pyramids core functions

\****************************************************************************************/

/****************************************************************************************\
                                   Filtering macros
\****************************************************************************************/

/* COMMON CASE: NON ZERO */
/* 1/16[1    4   6    4   1]       */
/* ...| x0 | 0 | x1 | 0 | x2 |...  */
#define  PU_FILTER( x0, x1, x2 )         ((x1)*6 + (x0) + (x2))

/* ZERO POINT AT CENTER */
/* 1/16[1   4    6   4    1]      */
/* ...| 0 | x0 | 0 | x1 | 0 |...  */
#define  PU_FILTER_ZI( x0, x1 )          (((x0) + (x1))*4)

/* MACROS FOR BORDERS */

/* | b I a | b | reflection */

/* LEFT/TOP */
/* 1/16[1    4   6    4   1]       */
/*    | x1 | 0 I x0 | 0 | x1 |...  */
#define  PU_LT( x0, x1 )                 ((x0)*6 + (x1)*2)

/* 1/16[1   4    6   4    1]       */
/*    | 0 I x0 | 0 | x1 | 0 |...   */
#define  PU_LT_ZI( x0, x1 )              PU_FILTER_ZI((x0),(x1))

/* RIGHT/BOTTOM: NON ZERO */
/* 1/16[1    4   6    4   1]       */
/* ...| x0 | 0 | x1 | 0 I x1 |     */
#define  PU_RB( x0, x1 )                 ((x0) + (x1)*7)

/* RIGHT/BOTTOM: ZERO POINT AT CENTER */
/* 1/16[1   4    6   4    1]       */
/* ...| 0 | x0 | 0 I x0 | 0 |      */
#define  PU_RB_ZI( x0 )                  ((x0)*8)

/* SINGULAR CASE */
/* 1/16[1    4   6    4   1]       */
/*    | x0 | 0 I x0 | 0 I x0 |     */
#define  PU_SINGULAR( x0 )               PU_RB_ZI((x0)) /* <--| the same formulas */
#define  PU_SINGULAR_ZI( x0 )            PU_RB_ZI((x0)) /* <--| */

#define  PU_SCALE(x)                     (((x) + (1<<5)) >> 6)

#define  PU_SZ  3

static CvStatus
icvPyrUpG5x5_8u( const uchar * src, int srcStep,
                 uchar * dst, int dstStep, CvSize srcRoi, void *buf, int Cs )
{
    int *buffer = (int *) buf;
    int *rows[PU_SZ];
    int y, top_row = 0;
    int W = srcRoi.width;
    int H = srcRoi.height;
    int Wd = W * 2, Wdn = Wd * Cs, Wn = W * Cs;
    int buffer_step = _CV_PYR_BUF_STEP_8U( Wd, Cs ) / sizeof( rows[0][0] );

    assert( Cs == 1 || Cs == 3 );

    /* main loop */
    for( y = 0; y < H; y++, dst += 2 * dstStep )
    {
        int fst = y == 0 ? 0 : PU_SZ - 1,
            lst = y == 0 ? (H <= PU_SZ / 2 ? H : PU_SZ / 2 + 1) :
            y < H - PU_SZ / 2 ? PU_SZ : H + PU_SZ / 2 - y;
        int x, y1, k = top_row;
        int *row0, *row1, *row2;
        uchar *dst1;

        /* assign rows pointers */
        for( y1 = 0; y1 < PU_SZ; y1++ )
        {
            rows[y1] = buffer + k * buffer_step;
            k++;
            k &= k < PU_SZ ? -1 : 0;
        }

        row0 = rows[0];
        row1 = rows[1];
        row2 = rows[2];
        dst1 = dst + dstStep;

        /* fill new buffer rows with filtered source (horizontal conv) */
        if( Cs == 1 )
            if( W > PU_SZ / 2 )
                for( y1 = fst; y1 < lst; y1++, src += srcStep )
                {
                    int *row = rows[y1];

                    /* process left & right bounds */
                    row[0] = PU_LT( src[0], src[1] );
                    row[1] = PU_LT_ZI( src[0], src[1] );
                    row[W * 2 - 2] = PU_RB( src[W - 2], src[W - 1] );
                    row[W * 2 - 1] = PU_RB_ZI( src[W - 1] );
                    /* other points */
                    for( x = 1; x < W - 1; x++ )
                    {
                        row[2 * x] = PU_FILTER( src[x - 1], src[x], src[x + 1] );
                        row[2 * x + 1] = PU_FILTER_ZI( src[x], src[x + 1] );
                    }
                }
            else                /* W <= PU_SZ/2 */
                for( y1 = fst; y1 < lst; y1++, src += srcStep )
                {
                    int *row = rows[y1];
                    int val = src[0];

                    row[0] = PU_SINGULAR( val );
                    row[1] = PU_SINGULAR_ZI( val );
                }
        else                    /* Cs == 3 */
            for( y1 = fst; y1 < lst; y1++, src += srcStep )
            {
                int *row = rows[y1];

                if( W > PU_SZ / 2 )
                {
                    int c;

                    for( c = 0; c < 3; c++ )
                    {
                        /* process left & right bounds  */
                        row[c] = PU_LT( src[c], src[3 + c] );
                        row[3 + c] = PU_LT_ZI( src[c], src[3 + c] );
                        row[Wn * 2 - 6 + c] = PU_RB( src[Wn - 6 + c], src[Wn - 3 + c] );
                        row[Wn * 2 - 3 + c] = PU_RB_ZI( src[Wn - 3 + c] );
                    }
                    /* other points */
                    for( x = 3; x < Wn - 3; x += 3 )
                    {
                        row[2 * x] = PU_FILTER( src[x - 3], src[x], src[x + 3] );
                        row[2 * x + 3] = PU_FILTER_ZI( src[x], src[x + 3] );

                        row[2 * x + 1] = PU_FILTER( src[x - 2], src[x + 1], src[x + 4] );
                        row[2 * x + 4] = PU_FILTER_ZI( src[x + 1], src[x + 4] );

                        row[2 * x + 2] = PU_FILTER( src[x - 1], src[x + 2], src[x + 5] );
                        row[2 * x + 5] = PU_FILTER_ZI( src[x + 2], src[x + 5] );
                    }
                }
                else            /* W <= PU_SZ/2 */
                {
                    int c;

                    for( c = 0; c < 3; c++ )
                    {
                        row[c] = PU_SINGULAR( src[c] );
                        row[3 + c] = PU_SINGULAR_ZI( src[c] );
                    }
                }
            }

        /* second pass. Do vertical conv and write results do destination image */
        if( y > 0 )
        {
            if( y < H - PU_SZ / 2 )
            {
                for( x = 0; x < Wdn; x++ )
                {
                    dst[x] = (uchar) PU_SCALE( PU_FILTER( row0[x], row1[x], row2[x] ));
                    dst1[x] = (uchar) PU_SCALE( PU_FILTER_ZI( row1[x], row2[x] ));
                }
                top_row++;
                top_row &= top_row < PU_SZ ? -1 : 0;
            }
            else                /* bottom */
                for( x = 0; x < Wdn; x++ )
                {
                    dst[x] = (uchar) PU_SCALE( PU_RB( row0[x], row1[x] ));
                    dst1[x] = (uchar) PU_SCALE( PU_RB_ZI( row1[x] ));
                }
        }
        else
        {
            if( H > PU_SZ / 2 ) /* top */
                for( x = 0; x < Wdn; x++ )
                {
                    dst[x] = (uchar) PU_SCALE( PU_LT( row0[x], row1[x] ));
                    dst1[x] = (uchar) PU_SCALE( PU_LT_ZI( row0[x], row1[x] ));
                }
            else                /* H <= PU_SZ/2 */
                for( x = 0; x < Wdn; x++ )
                {
                    dst[x] = (uchar) PU_SCALE( PU_SINGULAR( row0[x] ));
                    dst1[x] = (uchar) PU_SCALE( PU_SINGULAR_ZI( row0[x] ));
                }
        }
    }

    return CV_OK;
}


static CvStatus
icvPyrUpG5x5_8s( const char *src, int srcStep,
                 char *dst, int dstStep, CvSize srcRoi, void *buf, int Cs )
{
    int *buffer = (int *) buf;
    int *rows[PU_SZ];
    int y, top_row = 0;
    int W = srcRoi.width;
    int H = srcRoi.height;
    int Wd = W * 2, Wdn = Wd * Cs, Wn = W * Cs;
    int buffer_step = _CV_PYR_BUF_STEP_8U( Wd, Cs ) / sizeof( rows[0][0] );

    assert( Cs == 1 || Cs == 3 );

    /* main loop */
    for( y = 0; y < H; y++, dst += 2 * dstStep )
    {
        int fst = y == 0 ? 0 : PU_SZ - 1,
            lst = y == 0 ? (H <= PU_SZ / 2 ? H : PU_SZ / 2 + 1) :
            y < H - PU_SZ / 2 ? PU_SZ : H + PU_SZ / 2 - y;
        int x, y1, k = top_row;
        int *row0, *row1, *row2;
        char *dst1;

        /* assign rows pointers */
        for( y1 = 0; y1 < PU_SZ; y1++ )
        {
            rows[y1] = buffer + k * buffer_step;
            k++;
            k &= k < PU_SZ ? -1 : 0;
        }

        row0 = rows[0];
        row1 = rows[1];
        row2 = rows[2];
        dst1 = dst + dstStep;

        /* fill new buffer rows with filtered source (horizontal conv) */
        if( Cs == 1 )
            if( W > PU_SZ / 2 )
                for( y1 = fst; y1 < lst; y1++, src += srcStep )
                {
                    int *row = rows[y1];

                    /* process left & right bounds */
                    row[0] = PU_LT( src[0], src[1] );
                    row[1] = PU_LT_ZI( src[0], src[1] );
                    row[W * 2 - 2] = PU_RB( src[W - 2], src[W - 1] );
                    row[W * 2 - 1] = PU_RB_ZI( src[W - 1] );
                    /* other points */
                    for( x = 1; x < W - 1; x++ )
                    {
                        row[2 * x] = PU_FILTER( src[x - 1], src[x], src[x + 1] );
                        row[2 * x + 1] = PU_FILTER_ZI( src[x], src[x + 1] );
                    }
                }
            else                /* W <= PU_SZ/2 */
                for( y1 = fst; y1 < lst; y1++, src += srcStep )
                {
                    int *row = rows[y1];
                    int val = src[0];

                    row[0] = PU_SINGULAR( val );
                    row[1] = PU_SINGULAR_ZI( val );
                }
        else                    /* Cs == 3 */
            for( y1 = fst; y1 < lst; y1++, src += srcStep )
            {
                int *row = rows[y1];

                if( W > PU_SZ / 2 )
                {
                    int c;

                    for( c = 0; c < 3; c++ )
                    {
                        /* process left & right bounds  */
                        row[c] = PU_LT( src[c], src[3 + c] );
                        row[3 + c] = PU_LT_ZI( src[c], src[3 + c] );
                        row[Wn * 2 - 6 + c] = PU_RB( src[Wn - 6 + c], src[Wn - 3 + c] );
                        row[Wn * 2 - 3 + c] = PU_RB_ZI( src[Wn - 3 + c] );
                    }
                    /* other points */
                    for( x = 3; x < Wn - 3; x += 3 )
                    {
                        row[2 * x] = PU_FILTER( src[x - 3], src[x], src[x + 3] );
                        row[2 * x + 3] = PU_FILTER_ZI( src[x], src[x + 3] );

                        row[2 * x + 1] = PU_FILTER( src[x - 2], src[x + 1], src[x + 4] );
                        row[2 * x + 4] = PU_FILTER_ZI( src[x + 1], src[x + 4] );

                        row[2 * x + 2] = PU_FILTER( src[x - 1], src[x + 2], src[x + 5] );
                        row[2 * x + 5] = PU_FILTER_ZI( src[x + 2], src[x + 5] );
                    }
                }
                else            /* W <= PU_SZ/2 */
                {
                    int c;

                    for( c = 0; c < 3; c++ )
                    {
                        row[c] = PU_SINGULAR( src[c] );
                        row[3 + c] = PU_SINGULAR_ZI( src[c] );
                    }
                }
            }

        /* second pass. Do vertical conv and write results do destination image */
        if( y > 0 )
        {
            if( y < H - PU_SZ / 2 )
            {
                for( x = 0; x < Wdn; x++ )
                {
                    dst[x] = (char) PU_SCALE( PU_FILTER( row0[x], row1[x], row2[x] ));
                    dst1[x] = (char) PU_SCALE( PU_FILTER_ZI( row1[x], row2[x] ));
                }
                top_row++;
                top_row &= top_row < PU_SZ ? -1 : 0;
            }
            else                /* bottom */
                for( x = 0; x < Wdn; x++ )
                {
                    dst[x] = (char) PU_SCALE( PU_RB( row0[x], row1[x] ));
                    dst1[x] = (char) PU_SCALE( PU_RB_ZI( row1[x] ));
                }
        }
        else
        {
            if( H > PU_SZ / 2 ) /* top */
                for( x = 0; x < Wdn; x++ )
                {
                    dst[x] = (char) PU_SCALE( PU_LT( row0[x], row1[x] ));
                    dst1[x] = (char) PU_SCALE( PU_LT_ZI( row0[x], row1[x] ));
                }
            else                /* H <= PU_SZ/2 */
                for( x = 0; x < Wdn; x++ )
                {
                    dst[x] = (char) PU_SCALE( PU_SINGULAR( row0[x] ));
                    dst1[x] = (char) PU_SCALE( PU_SINGULAR_ZI( row0[x] ));
                }
        }
    }

    return CV_OK;
}


/****************************************************************************************\

                                     External functions

\****************************************************************************************/

/* ///////////////////////////////////////////////////////////////////////////////////////
//  Name:       icvPyrDown_Gauss5x5_8u_C1R, icvPyrDown_Gauss5x5_8u_C3R,
//              icvPyrDown_Gauss5x5_8s_C1R, icvPyrDown_Gauss5x5_8s_C3R,
//              icvPyrDown_Gauss5x5_32f_C1R, icvPyrDown_Gauss5x5_32f_C3R,
//
//              icvPyrUp_Gauss5x5_8u_C1R, icvPyrUp_Gauss5x5_8u_C3R,
//              icvPyrUp_Gauss5x5_8s_C1R, icvPyrUp_Gauss5x5_8s_C3R,
//              icvPyrUp_Gauss5x5_32f_C1R, icvPyrUp_Gauss5x5_32f_C3R,
//
//  Purpose:    Perform downsampling/upsampling of the image with 5x5 gaussian.
//
//  Return:
//    ippStsNoErr              Ok
//    CV_NULLPTR_ERR         One of pointers is NULL
//    CV_BADSIZE_ERR            The width or height of images is less or equal zero.
//                             For icvPyrDown*** if width or height are odd
//                             (non-divisible by 2).
//    ippStsStepErr            Step is too small to fit image.
//
//  Arguments:
//    pSrc                     pointer to source image
//    srcStep                  step in source image
//    pDst                     pointer to destination image
//    dstStep                  step in source image
//    roiSize                      source image ROI size. Destination image width and
//                             height will be twice large (PyrUp)
//                             or twice smaller (PyrDown)
//    pBuffer                  preallocated temporary cyclic buffer,
//                             which size is calculated using
//                             ippPyrUpGetSize_Gauss_5x5 or ippPyrDownGetSize_Gauss_5x5
*/
IPCVAPI_IMPL( CvStatus, icvPyrUp_Gauss5x5_8u_C1R, (const uchar * pSrc, int srcStep,
                                                   uchar * pDst, int dstStep,
                                                   CvSize roiSize, void *pBuffer) )
{
    if( !pSrc || !pDst )
        return CV_NULLPTR_ERR;

    if( roiSize.width < 0 || roiSize.height < 0 ||
        srcStep < roiSize.width || dstStep < roiSize.width * 2 )
        return CV_BADSIZE_ERR;

    return icvPyrUpG5x5_8u( pSrc, srcStep, pDst, dstStep, roiSize, pBuffer, 1 );
}


IPCVAPI_IMPL( CvStatus, icvPyrUp_Gauss5x5_8u_C3R, (const uchar * pSrc, int srcStep,
                                                   uchar * pDst, int dstStep,
                                                   CvSize roiSize, void *pBuffer) )
{
    if( !pSrc || !pDst )
        return CV_NULLPTR_ERR;

    if( roiSize.width < 0 || roiSize.height < 0 ||
        srcStep < roiSize.width * 3 || dstStep < roiSize.width * 6 )
        return CV_BADSIZE_ERR;

    return icvPyrUpG5x5_8u( pSrc, srcStep, pDst, dstStep, roiSize, pBuffer, 3 );
}


IPCVAPI_IMPL( CvStatus, icvPyrUp_Gauss5x5_8s_C1R, (const char *pSrc, int srcStep,
                                                   char *pDst, int dstStep,
                                                   CvSize roiSize, void *pBuffer) )
{
    if( !pSrc || !pDst )
        return CV_NULLPTR_ERR;

    if( roiSize.width < 0 || roiSize.height < 0 ||
        srcStep < roiSize.width || dstStep < roiSize.width * 2 )
        return CV_BADSIZE_ERR;

    return icvPyrUpG5x5_8s( pSrc, srcStep, pDst, dstStep, roiSize, pBuffer, 1 );
}


IPCVAPI_IMPL( CvStatus, icvPyrUp_Gauss5x5_8s_C3R, (const char *pSrc, int srcStep,
                                                   char *pDst, int dstStep,
                                                   CvSize roiSize, void *pBuffer) )
{
    if( !pSrc || !pDst )
        return CV_NULLPTR_ERR;

    if( roiSize.width < 0 || roiSize.height < 0 ||
        srcStep < roiSize.width * 3 || dstStep < roiSize.width * 6 )
        return CV_BADSIZE_ERR;

    return icvPyrUpG5x5_8s( pSrc, srcStep, pDst, dstStep, roiSize, pBuffer, 3 );
}

/* down_sampling */
IPCVAPI_IMPL( CvStatus, icvPyrDown_Gauss5x5_8u_C1R, (const uchar * pSrc, int srcStep,
                                                     uchar * pDst, int dstStep,
                                                     CvSize roiSize, void *pBuffer) )
{
    if( !pSrc || !pDst )
        return CV_NULLPTR_ERR;

    if( roiSize.width < 0 || roiSize.height < 0 ||
        srcStep < roiSize.width || dstStep < (roiSize.width >> 1) ||
        (roiSize.width & 1) != 0 || (roiSize.height & 1) != 0 )
        return CV_BADSIZE_ERR;

    return icvPyrDownG5x5_8u( pSrc, srcStep, pDst, dstStep, roiSize, pBuffer, 1 );
}


IPCVAPI_IMPL( CvStatus, icvPyrDown_Gauss5x5_8u_C3R, (const uchar * pSrc, int srcStep,
                                                     uchar * pDst, int dstStep,
                                                     CvSize roiSize, void *pBuffer) )
{
    if( !pSrc || !pDst )
        return CV_NULLPTR_ERR;

    if( roiSize.width < 0 || roiSize.height < 0 ||
        srcStep < roiSize.width * 3 || dstStep < (roiSize.width >> 1) * 3 ||
        (roiSize.width & 1) != 0 || (roiSize.height & 1) != 0 )
        return CV_BADSIZE_ERR;

    return icvPyrDownG5x5_8u( pSrc, srcStep, pDst, dstStep, roiSize, pBuffer, 3 );
}


IPCVAPI_IMPL( CvStatus, icvPyrDown_Gauss5x5_8s_C1R, (const char *pSrc, int srcStep,
                                                     char *pDst, int dstStep,
                                                     CvSize roiSize, void *pBuffer) )
{
    if( !pSrc || !pDst )
        return CV_NULLPTR_ERR;

    if( roiSize.width < 0 || roiSize.height < 0 ||
        srcStep < roiSize.width || dstStep < (roiSize.width >> 1) ||
        (roiSize.width & 1) != 0 || (roiSize.height & 1) != 0 )
        return CV_BADSIZE_ERR;

    return icvPyrDownG5x5_8s( pSrc, srcStep, pDst, dstStep, roiSize, pBuffer, 1 );
}


IPCVAPI_IMPL( CvStatus, icvPyrDown_Gauss5x5_8s_C3R, (const char *pSrc, int srcStep,
                                                     char *pDst, int dstStep,
                                                     CvSize roiSize, void *pBuffer) )
{
    if( !pSrc || !pDst )
        return CV_NULLPTR_ERR;

    if( roiSize.width < 0 || roiSize.height < 0 ||
        srcStep < roiSize.width * 3 || dstStep < (roiSize.width >> 1) * 3 ||
        (roiSize.width & 1) != 0 || (roiSize.height & 1) != 0 )
        return CV_BADSIZE_ERR;

    return icvPyrDownG5x5_8s( pSrc, srcStep, pDst, dstStep, roiSize, pBuffer, 3 );
}


/* ///////////////////////////////////////////////////////////////////////////////////////
//  Name:       icvPyrUpGetSize_Gauss5x5, icvPyrDownGetSize_Gauss5x5
//  Purpose:    Calculates cyclic buffer size for pyramids.
//  Return:
//    ippStsNoErr              Ok
//    ippStsDepthErr           Specified depth is invalid or unsupported
//    ippStsNumChannensErr     Specified number of channels is invalid or unsupported
//    CV_BADSIZE_ERR            The width or height of images is less or equal zero
//                             For icvPyrDown*** if width or height are odd
//                             (non-divisible by 2).
//    ippStsStepErr            The steps in images are less or equal zero
//
//  Arguments:
//    roiWidth                 width of image ROI in pixels
//    depth                    image depth
//    channels                 number of image channels
//    bufSize                  output parameter. Size of cyclic buffer for pyramids.
*/
IPCVAPI_IMPL( CvStatus, icvPyrUpGetBufSize_Gauss5x5, (int roiWidth, CvDataType /*dataType */,
                                                      int channels, int *bufSize) )
{
    int bufStep = 0;

    if( !bufSize )
        return CV_NULLPTR_ERR;
    *bufSize = 0;

    if( roiWidth < 0 )
        return CV_BADSIZE_ERR;
    if( channels != 1 && channels != 3 )
        return CV_UNSUPPORTED_CHANNELS_ERR;

    bufStep = _CV_PYR_BUF_STEP_8U( roiWidth * 2, channels );

    *bufSize = bufStep * PU_SZ;
    return CV_OK;
}


IPCVAPI_IMPL( CvStatus, icvPyrDownGetBufSize_Gauss5x5, (int roiWidth, CvDataType /*dataType */ ,
                                                        int channels, int *bufSize) )
{
    int bufStep = 0;

    if( !bufSize )
        return CV_NULLPTR_ERR;
    *bufSize = 0;

    if( roiWidth < 0 || (roiWidth & 1) != 0 )
        return CV_BADSIZE_ERR;
    if( channels != 1 && channels != 3 )
        return CV_UNSUPPORTED_CHANNELS_ERR;

    bufStep = _CV_PYR_BUF_STEP_8U( roiWidth, channels );

    *bufSize = bufStep * (PD_SZ + 1);
    return CV_OK;
}

/* End of file. */
