/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/
#include "_cv.h"
#include "_cvwrap.h"

#define CV_ALIGN  8

/* macro for calculating cyclic buffer step */
#define CV_PYR_BUF_STEP_8U( width, Cs ) \
   ((((width)+CV_ALIGN*2)*Cs*sizeof(int) + CV_ALIGN*2 - 1) &-CV_ALIGN*2)

#define  PU_SZ  3
#define  PD_SZ  5

CV_IMPL void
cvPyrUp( IplImage * src, IplImage * dst, IplFilter _filter )
{
    uchar *src_data = 0;
    uchar *dst_data = 0;

    int src_step = 0;
    int dst_step = 0;

    void *buffer = 0;
    int buffer_size = 0;

    CvSize src_size;
    CvSize dst_size;
    CvFilter filter = (CvFilter) _filter;

    CV_FUNCNAME( "cvPyrUp" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( src ));
    CV_CALL( CV_CHECK_IMAGE( dst ));

    if( src->depth != dst->depth || src->nChannels != dst->nChannels )
        CV_ERROR( IPL_StsBadArg, "src and dst have different formats" );

    if( filter != CV_GAUSSIAN_5x5 )
        CV_ERROR( IPL_BadOrder, "this filter type not supported" );

    cvGetImageRawData( src, &src_data, &src_step, &src_size );
    cvGetImageRawData( dst, &dst_data, &dst_step, &dst_size );

    if( src_size.width * 2 != dst_size.width || src_size.height * 2 != dst_size.height )
        CV_ERROR( IPL_StsBadArg, "src and dst have uncoordinated sizes" );

    IPPI_CALL( icvPyrUpGetBufSize_Gauss5x5( src_size.width,
                                             src->depth == IPL_DEPTH_8U ? cv8u : src->depth ==
                                             IPL_DEPTH_8S ? cv8s : cv32f, src->nChannels,
                                             &buffer_size ));
    buffer = icvAlloc( buffer_size );
    if( !buffer )
        CV_ERROR( CV_OUTOFMEM_ERR, "not enough memory" );

    switch (src->depth)
    {
    case IPL_DEPTH_8U:
        switch (src->nChannels)
        {
        case 1:
            IPPI_CALL( icvPyrUp_Gauss5x5_8u_C1R( (const uchar *) src_data, src_step,
                                                  dst_data, dst_step, src_size, buffer ));
            break;
        case 3:
            IPPI_CALL( icvPyrUp_Gauss5x5_8u_C3R( (const uchar *) src_data, src_step,
                                                  dst_data, dst_step, src_size, buffer ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    case IPL_DEPTH_8S:
        switch (src->nChannels)
        {
        case 1:
            IPPI_CALL( icvPyrUp_Gauss5x5_8s_C1R( (const char *) src_data, src_step,
                                                  (char *) dst_data, dst_step,
                                                  src_size, buffer ));
            break;
        case 3:
            IPPI_CALL( icvPyrUp_Gauss5x5_8s_C3R( (const char *) src_data, src_step,
                                                  (char *) dst_data, dst_step,
                                                  src_size, buffer ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    case IPL_DEPTH_32F:
        switch (src->nChannels)
        {
        case 1:
            IPPI_CALL( icvPyrUp_Gauss5x5_32f_C1R( (const float *) src_data, src_step,
                                                   (float *) dst_data, dst_step,
                                                   src_size, buffer ));
            break;
        case 3:
            IPPI_CALL( icvPyrUp_Gauss5x5_32f_C3R( (const float *) src_data, src_step,
                                                   (float *) dst_data, dst_step,
                                                   src_size, buffer ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }

    __CLEANUP__;
    if( buffer )
        icvFree( &buffer );
    __END__;
}


CV_IMPL void
cvPyrDown( IplImage * src, IplImage * dst, IplFilter _filter )
{
    uchar *src_data = 0;
    uchar *dst_data = 0;

    int src_step = 0;
    int dst_step = 0;

    void *buffer = 0;
    int buffer_size = 0;

    CvSize src_size;
    CvSize dst_size;
    CvFilter filter = (CvFilter) _filter;

    CV_FUNCNAME( "cvPyrDown" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( src ));
    CV_CALL( CV_CHECK_IMAGE( dst ));

    if( src->depth != dst->depth || src->nChannels != dst->nChannels )
        CV_ERROR( IPL_StsBadArg, "src and dst have different formats" );

    cvGetImageRawData( src, &src_data, &src_step, &src_size );
    cvGetImageRawData( dst, &dst_data, &dst_step, &dst_size );

    if( src_size.width != dst_size.width * 2 || src_size.height != dst_size.height * 2 )
        CV_ERROR( IPL_StsBadArg, "src and dst have uncoordinated sizes" );

    if( filter != CV_GAUSSIAN_5x5 )
        CV_ERROR( IPL_BadOrder, "this filter type not supported" );

    IPPI_CALL( icvPyrDownGetBufSize_Gauss5x5( src_size.width,
                                               src->depth ==
                                               IPL_DEPTH_8U ? cv8u : src->depth ==
                                               IPL_DEPTH_8S ? cv8s : cv32f, src->nChannels,
                                               &buffer_size ));
    buffer = icvAlloc( buffer_size );
    if( !buffer )
        CV_ERROR( CV_OUTOFMEM_ERR, "not enough memory" );

    switch (src->depth)
    {
    case IPL_DEPTH_8U:
        switch (src->nChannels)
        {
        case 1:
            IPPI_CALL( icvPyrDown_Gauss5x5_8u_C1R( (const uchar *) src_data, src_step,
                                                    dst_data, dst_step, src_size, buffer ));
            break;
        case 3:
            IPPI_CALL( icvPyrDown_Gauss5x5_8u_C3R( (const uchar *) src_data, src_step,
                                                    dst_data, dst_step, src_size, buffer ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    case IPL_DEPTH_8S:
        switch (src->nChannels)
        {
        case 1:
            IPPI_CALL( icvPyrDown_Gauss5x5_8s_C1R( (const char *) src_data, src_step,
                                                    (char *) dst_data, dst_step,
                                                    src_size, buffer ));
            break;
        case 3:
            IPPI_CALL( icvPyrDown_Gauss5x5_8s_C3R( (const char *) src_data, src_step,
                                                    (char *) dst_data, dst_step,
                                                    src_size, buffer ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    case IPL_DEPTH_32F:
        switch (src->nChannels)
        {
        case 1:
            IPPI_CALL( icvPyrDown_Gauss5x5_32f_C1R( (const float *) src_data, src_step,
                                                     (float *) dst_data, dst_step,
                                                     src_size, buffer ));
            break;
        case 3:
            IPPI_CALL( icvPyrDown_Gauss5x5_32f_C3R( (const float *) src_data, src_step,
                                                     (float *) dst_data, dst_step,
                                                     src_size, buffer ));
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }

    __CLEANUP__;
    if( buffer )
        icvFree( &buffer );
    __END__;
}

/* End of file. */
