/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/
#include "_cv.h"
#include "_cvwrap.h"

#define _CV_GET_BT_PIX( type )  (((int)type & 255) >> 3)

/* 
   Initializes line iterator.
   Returns number of points on the line or negative number if error.
*/
int
icvInitLineIterator( const void *img, int step, CvSize size, int bt_pix,
                     CvPoint pt1, CvPoint pt2, CvLineIterator * iterator )
{
    int dx, dy, s;

    assert( bt_pix == 1 || bt_pix == 3 || bt_pix == 4 || bt_pix == 12 );

    if( !img || !iterator )
        return CV_NULLPTR_ERR;

    if( size.width <= 0 || size.height <= 0 || step < size.width * bt_pix )
        return CV_BADSIZE_ERR;

    if( (unsigned) pt1.x >= (unsigned) size.width ||
        (unsigned) pt2.x >= (unsigned) size.width ||
        (unsigned) pt1.y >= (unsigned) size.height ||
        (unsigned) pt2.y >= (unsigned) size.height )
        return CV_BADPOINT_ERR;

    iterator->ptr = (uchar *) img + pt1.y * step + pt1.x * bt_pix;

    dx = pt2.x - pt1.x;
    dy = pt2.y - pt1.y;

    s = dx >> 31;
    dx = (dx ^ s) - s;
    bt_pix = (bt_pix ^ s) - s;

    s = dy >> 31;
    dy = (dy ^ s) - s;
    step = (step ^ s) - s;

    assert( dx >= 0 && dy >= 0 );

    if( dx > dy )
    {
        iterator->err = dx;
        iterator->plus_delta = dx + dx;
        iterator->minus_delta = dy + dy;
        iterator->plus_step = step;
        iterator->minus_step = bt_pix;
    }
    else
    {
        iterator->err = dy;
        iterator->plus_delta = dy + dy;
        iterator->minus_delta = dx + dx;
        iterator->plus_step = bt_pix;
        iterator->minus_step = step;
    }

    return iterator->err + 1;
}

/**************************************************************************************\
*                                   line samplers                                      *
\**************************************************************************************/


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    icvSampleLine...
//    Purpose:
//      Scans the 8-connected straight line,
//      connecting two specified points and stores the points on the
//      line to the buffer.
//    Context:
//    Parameters:
//      img  - pointer to top-left corner of image ROI
//      step - full width of the image (in bytes)
//      size - width and height of the ROI
//      pt1  - starting point
//      pt2  - ending point
//      buffer - where to store the line points.
//               it must have size enough to store
//               max(abs(pt1.x - pt2.x)+1,abs(pt1.y - pt2.y)+1) pixels.
//    Returns:
//      number of points stored or error code
//    Notes:
F*/
IPCVAPI_IMPL( int, icvSampleLine_8u_C1R, (const uchar * img, int step, CvSize size,
                                          uchar * buffer, CvPoint pt1, CvPoint pt2) )
{
    int i, count;
    CvLineIterator iterator;

    if( !buffer )
        return CV_NULLPTR_ERR;

    count = icvInitLineIterator( img, step, size, 1, pt1, pt2, &iterator );
    if( count < 0 )
        return count;
    for( i = 0; i < count; i++ )
    {
        buffer[i] = iterator.ptr[0];
        CV_NEXT_LINE_POINT( iterator );
    }

    return count;
}


IPCVAPI_IMPL( int, icvSampleLine_8u_C3R, (const uchar * img, int step, CvSize size,
                                          uchar * buffer, CvPoint pt1, CvPoint pt2) )
{
    int i, count;
    CvLineIterator iterator;

    if( !buffer )
        return CV_NULLPTR_ERR;

    count = icvInitLineIterator( img, step, size, 3, pt1, pt2, &iterator );
    if( count < 0 )
        return count;
    for( i = 0; i < count * 3; i += 3 )
    {
        buffer[i] = iterator.ptr[0];
        buffer[i + 1] = iterator.ptr[1];
        buffer[i + 2] = iterator.ptr[2];
        CV_NEXT_LINE_POINT( iterator );
    }

    return count;
}


IPCVAPI_IMPL( int, icvSampleLine_32f_C1R, (const float *img, int step, CvSize size,
                                           float *buffer, CvPoint pt1, CvPoint pt2) )
{
    int i, count;
    CvLineIterator iterator;

    if( !buffer )
        return CV_NULLPTR_ERR;

    count = icvInitLineIterator( img, step, size, 4, pt1, pt2, &iterator );
    if( count < 0 )
        return count;
    for( i = 0; i < count; i++ )
    {
        ((int *) buffer)[i] = ((int *) iterator.ptr)[0];
        CV_NEXT_LINE_POINT( iterator );
    }

    return count;
}


IPCVAPI_IMPL( int, icvSampleLine_32f_C3R, (const float *img, int step, CvSize size,
                                           float *buffer, CvPoint pt1, CvPoint pt2) )
{
    int i, count;
    CvLineIterator iterator;

    if( !buffer )
        return CV_NULLPTR_ERR;

    count = icvInitLineIterator( img, step, size, 12, pt1, pt2, &iterator );
    if( count < 0 )
        return count;

    for( i = 0; i < count * 3; i += 3 )
    {
        ((int *) buffer)[i] = ((int *) iterator.ptr)[0];
        ((int *) buffer)[i + 1] = ((int *) iterator.ptr)[1];
        ((int *) buffer)[i + 2] = ((int *) iterator.ptr)[2];
        CV_NEXT_LINE_POINT( iterator );
    }

    return count;
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    icvGetRectSubPix8u32fC1R
//    Purpose:
//      Extracts rectangle from the byte-depth image with sub-pixel resolution
//      (it is done with bilinear interpolation) and writes the rectangle into
//      floating point buffer.
//    Context:
//    Parameters:
//      src      - source image
//      src_step - full width of source image row in bytes
//      src_size - width and height of copied ROI in pixels
//      p        - center of rectangle
//      win_size - the following rectangle is extracted:
//                     (p.x - win_size.width .. p.x + win_size.width,
//                      p.y - win_size.height .. p.y + win_size.height)
//      buffer   - destination buffer, which is able to fit resultant rectangle
//                 (win_size.width*2 + 1) by (win_size.height*2 + 1) pixels.
//    Returns:
//    Notes:
//      If the extracted rectangle is partially outside the image
//      then the missing pixels are taken from corresponding boundary pixels
//      (that is first and last row and first and last column are repeated (virtually)
//      beyond the image.
//F*/
IPCVAPI_IMPL( CvStatus, icvGetRectSubPix_8u32f_C1R, (const uchar * src, int src_step,
                                                     CvSize src_size,
                                                     float *dst, int dst_step,
                                                     CvSize win_size, CvPoint2D32f center) )
{
    CvPoint ip;
    float a, b;
    int i, j;

    if( !src || !dst )
        return CV_NULLPTR_ERR;
    if( src_size.width <= 1 || src_size.height <= 1 ||
        src_step < src_size.width ||
        win_size.width < 0 || win_size.height < 0 ||
        (win_size.width & win_size.height & 1) == 0 ||
        dst_step < win_size.width * sizeof_float ||
        (dst_step & (sizeof_float - 1)) != 0 )
        return CV_BADSIZE_ERR;

    ip.x = cvFloor( center.x );
    ip.y = cvFloor( center.y );

    a = center.x - ip.x;
    b = center.y - ip.y;

    ip.x -= win_size.width >> 1;
    ip.y -= win_size.height >> 1;

    dst_step /= sizeof_float;

    if( 0 <= ip.x && ip.x + win_size.width < src_size.width &&
        0 <= ip.y && ip.y + win_size.height < src_size.height )
    {
        /* extracted rectangle is totally inside the image */
        src += ip.y * src_step + ip.x;

        for( i = 0; i < win_size.height; i++, src += src_step, dst += dst_step )
        {
            for( j = 0; j < win_size.width; j++ )
            {
                float s0 = CV_8TO32F(src[j]);
                float s1 = CV_8TO32F(src[j + src_step]);

                s0 += a * (CV_8TO32F(src[j + 1]) - s0);
                s1 += a * (CV_8TO32F(src[j + src_step + 1]) - s1);

                dst[j] = s0 + b * (s1 - s0);
            }
        }
    }
    else
    {
        int min_i, max_i;
        int min_j, max_j;

        if( ip.x >= 0 )
        {
            src += ip.x;
            min_j = 0;
        }
        else
        {
            min_j = -ip.x;
            if( min_j > win_size.width )
                min_j = win_size.width;
        }

        if( ip.x + win_size.width < src_size.width )
            max_j = win_size.width;
        else
        {
            max_j = src_size.width - ip.x - 1;
            if( max_j < 0 )
            {
                src += max_j;
                max_j = 0;
            }
            assert( max_j <= win_size.width );
        }

        if( ip.y >= 0 )
        {
            src += ip.y * src_step;
            min_i = 0;
        }
        else
            min_i = -ip.y;

        if( ip.y + win_size.height < src_size.height )
            max_i = win_size.height;
        else
        {
            max_i = src_size.height - ip.y - 1;
            if( max_i < 0 )
            {
                src += max_i*src_step;
                max_i = 0;
            }
        }

        src -= min_j;

        for( i = 0; i < win_size.height; i++, dst += dst_step )
        {
            const uchar *src2 = src + src_step;

            if( i < min_i || max_i <= i )
                src2 -= src_step;

            for( j = 0; j < min_j; j++ )
            {
                float s0 = CV_8TO32F(src[min_j]);
                float s1 = CV_8TO32F(src2[min_j]);

                dst[j] = s0 + b * (s1 - s0);
            }

            for( ; j < max_j; j++ )
            {
                float s0 = CV_8TO32F(src[j]);
                float s1 = CV_8TO32F(src2[j]);

                s0 += a * (CV_8TO32F(src[j + 1]) - s0);
                s1 += a * (CV_8TO32F(src2[j + 1]) - s1);

                dst[j] = s0 + b * (s1 - s0);
            }

            for( ; j < win_size.width; j++ )
            {
                float s0 = CV_8TO32F(src[max_j]);
                float s1 = CV_8TO32F(src2[max_j]);

                dst[j] = s0 + b * (s1 - s0);
            }

            if( i >= min_i )
                src = src2;
        }
    }

    return CV_OK;
}


IPCVAPI_IMPL( CvStatus, icvGetRectSubPix_8s32f_C1R, (const char *src, int src_step,
                                                     CvSize src_size,
                                                     float *dst, int dst_step,
                                                     CvSize win_size, CvPoint2D32f center) )
{
    CvPoint ip;
    float a, b;
    int i, j;

    if( !src || !dst )
        return CV_NULLPTR_ERR;
    if( src_size.width <= 1 || src_size.height <= 1 ||
        src_step < src_size.width ||
        win_size.width < 0 || win_size.height < 0 ||
        (win_size.width & win_size.height & 1) == 0 ||
        dst_step < win_size.width * sizeof_float ||
        (dst_step & (sizeof_float - 1)) != 0 )
        return CV_BADSIZE_ERR;

    ip.x = cvFloor( center.x );
    ip.y = cvFloor( center.y );

    a = center.x - ip.x;
    b = center.y - ip.y;

    ip.x -= win_size.width >> 1;
    ip.y -= win_size.height >> 1;

    dst_step /= sizeof_float;

    if( 0 <= ip.x && ip.x + win_size.width < src_size.width &&
        0 <= ip.y && ip.y + win_size.height < src_size.height )
    {
        /* extracted rectangle is totally inside the image */
        src += ip.y * src_step + ip.x;

        for( i = 0; i < win_size.height; i++, src += src_step, dst += dst_step )
        {
            for( j = 0; j < win_size.width; j++ )
            {
                float s0 = CV_8TO32F(src[j]);
                float s1 = CV_8TO32F(src[j + src_step]);

                s0 += a * (CV_8TO32F(src[j + 1]) - s0);
                s1 += a * (CV_8TO32F(src[j + src_step + 1]) - s1);

                dst[j] = s0 + b * (s1 - s0);
            }
        }
    }
    else
    {
        int min_i, max_i;
        int min_j, max_j;

        if( ip.x >= 0 )
        {
            src += ip.x;
            min_j = 0;
        }
        else
        {
            min_j = -ip.x;
            if( min_j > win_size.width )
                min_j = win_size.width;
        }

        if( ip.x + win_size.width < src_size.width )
            max_j = win_size.width;
        else
        {
            max_j = src_size.width - ip.x - 1;
            if( max_j < 0 )
            {
                src += max_j;
                max_j = 0;
            }
            assert( max_j <= win_size.width );
        }

        if( ip.y >= 0 )
        {
            src += ip.y * src_step;
            min_i = 0;
        }
        else
            min_i = -ip.y;

        if( ip.y + win_size.height < src_size.height )
            max_i = win_size.height;
        else
        {
            max_i = src_size.height - ip.y - 1;
            if( max_i < 0 )
            {
                src += max_i*src_step;
                max_i = 0;
            }
        }

        src -= min_j;

        for( i = 0; i < win_size.height; i++, dst += dst_step )
        {
            const char *src2 = src + src_step;

            if( i < min_i || max_i <= i )
                src2 -= src_step;

            for( j = 0; j < min_j; j++ )
            {
                float s0 = CV_8TO32F(src[min_j]);
                float s1 = CV_8TO32F(src2[min_j]);

                dst[j] = s0 + b * (s1 - s0);
            }

            for( ; j < max_j; j++ )
            {
                float s0 = CV_8TO32F(src[j]);
                float s1 = CV_8TO32F(src2[j]);

                s0 += a * (CV_8TO32F(src[j + 1]) - s0);
                s1 += a * (CV_8TO32F(src2[j + 1]) - s1);

                dst[j] = s0 + b * (s1 - s0);
            }

            for( ; j < win_size.width; j++ )
            {
                float s0 = CV_8TO32F(src[max_j]);
                float s1 = CV_8TO32F(src2[max_j]);

                dst[j] = s0 + b * (s1 - s0);
            }

            if( i >= min_i )
                src = src2;
        }
    }

    return CV_OK;
}


IPCVAPI_IMPL( CvStatus, icvGetRectSubPix_32f_C1R, (const float *src, int src_step,
                                                   CvSize src_size,
                                                   float *dst, int dst_step,
                                                   CvSize win_size, CvPoint2D32f center) )
{
    CvPoint ip;
    float a, b;
    int i, j;

    if( !src || !dst )
        return CV_NULLPTR_ERR;
    if( src_size.width <= 1 || src_size.height <= 1 ||
        src_step < src_size.width*sizeof_float ||
        (src_step & (sizeof_float - 1)) != 0 ||
        win_size.width < 0 || win_size.height < 0 ||
        (win_size.width & win_size.height & 1) == 0 ||
        dst_step < win_size.width * sizeof_float ||
        (dst_step & (sizeof_float - 1)) != 0 )
        return CV_BADSIZE_ERR;

    ip.x = cvFloor( center.x );
    ip.y = cvFloor( center.y );

    a = center.x - ip.x;
    b = center.y - ip.y;

    ip.x -= win_size.width >> 1;
    ip.y -= win_size.height >> 1;

    src_step /= sizeof_float;
    dst_step /= sizeof_float;

    if( 0 <= ip.x && ip.x + win_size.width < src_size.width &&
        0 <= ip.y && ip.y + win_size.height < src_size.height )
    {
        /* extracted rectangle is totally inside the image */
        src += ip.y * src_step + ip.x;

        for( i = 0; i < win_size.height; i++, src += src_step, dst += dst_step )
        {
            for( j = 0; j < win_size.width; j++ )
            {
                float s0 = src[j];
                float s1 = src[j + src_step];

                s0 += a * (src[j + 1] - s0);
                s1 += a * (src[j + src_step + 1] - s1);

                dst[j] = s0 + b * (s1 - s0);
            }
        }
    }
    else
    {
        int min_i, max_i;
        int min_j, max_j;

        if( ip.x >= 0 )
        {
            src += ip.x;
            min_j = 0;
        }
        else
        {
            min_j = -ip.x;
            if( min_j > win_size.width )
                min_j = win_size.width;
        }

        if( ip.x + win_size.width < src_size.width )
            max_j = win_size.width;
        else
        {
            max_j = src_size.width - ip.x - 1;
            if( max_j < 0 )
            {
                src += max_j;
                max_j = 0;
            }
            assert( max_j <= win_size.width );
        }

        if( ip.y >= 0 )
        {
            src += ip.y * src_step;
            min_i = 0;
        }
        else
            min_i = -ip.y;

        if( ip.y + win_size.height < src_size.height )
            max_i = win_size.height;
        else
        {
            max_i = src_size.height - ip.y - 1;
            if( max_i < 0 )
            {
                src += max_i*src_step;
                max_i = 0;
            }
        }

        src -= min_j;

        for( i = 0; i < win_size.height; i++, dst += dst_step )
        {
            const float *src2 = src + src_step;

            if( i < min_i || max_i <= i )
                src2 -= src_step;

            for( j = 0; j < min_j; j++ )
            {
                float s0 = src[min_j];
                float s1 = src2[min_j];

                dst[j] = s0 + b * (s1 - s0);
            }

            for( ; j < max_j; j++ )
            {
                float s0 = src[j];
                float s1 = src2[j];

                s0 += a * (src[j + 1] - s0);
                s1 += a * (src2[j + 1] - s1);

                dst[j] = s0 + b * (s1 - s0);
            }

            for( ; j < win_size.width; j++ )
            {
                float s0 = src[max_j];
                float s1 = src2[max_j];

                dst[j] = s0 + b * (s1 - s0);
            }

            if( i >= min_i )
                src = src2;
        }
    }

    return CV_OK;
}


IPCVAPI_IMPL( CvStatus, icvGetQuadrangleSubPix_8u32f_C1R,
              (const uchar * src, int srcStep, CvSize src_size,
               float *dst, int dstStep, CvSize win_size, CvPoint2D32f center, const float *A) )
{
    int x, y, x0, y0, x1, y1, x2, y2, x3, y3;

    if( !src || !dst || !A )
        return CV_NULLPTR_ERR;

    if( (win_size.width & win_size.height & 1) == 0 )
        return CV_BADSIZE_ERR;

    if( src_size.width <= 0 || src_size.height <= 0 ||
        (unsigned) srcStep < (unsigned) src_size.width ||
        (unsigned) dstStep < (unsigned) win_size.width * 4 || (dstStep & 3) != 0 )
        return CV_BADSIZE_ERR;

    if( (unsigned) cvRound( center.x ) >= (unsigned) src_size.width ||
        (unsigned) cvRound( center.y ) >= (unsigned) src_size.height )
        return CV_BADRANGE_ERR;

    dstStep /= 4;

    win_size.width >>= 1;
    win_size.height >>= 1;

#define GET_X(X,Y) ((X)*A[0] + (Y)*A[1] + center.x)
#define GET_Y(X,Y) ((X)*A[2] + (Y)*A[3] + center.y)
#define PTR_AT(X,Y) (src + (Y)*srcStep + (X))

    x0 = cvFloor( GET_X( -win_size.width, -win_size.height ));
    y0 = cvFloor( GET_Y( -win_size.width, -win_size.height ));
    x1 = cvFloor( GET_X( win_size.width, -win_size.height ));
    y1 = cvFloor( GET_Y( win_size.width, -win_size.height ));
    x2 = cvFloor( GET_X( -win_size.width, win_size.height ));
    y2 = cvFloor( GET_Y( -win_size.width, win_size.height ));
    x3 = cvFloor( GET_X( win_size.width, win_size.height ));
    y3 = cvFloor( GET_Y( win_size.width, win_size.height ));

    dst += win_size.width;

    /* check if the quadrangle is entirely in the image */
#if 0
    if( (unsigned) x0 < (unsigned) (src_size.width - 1) &&
        (unsigned) y0 < (unsigned) (src_size.height - 1) &&
        (unsigned) x1 < (unsigned) (src_size.width - 1) &&
        (unsigned) y1 < (unsigned) (src_size.height - 1) &&
        (unsigned) x2 < (unsigned) (src_size.width - 1) &&
        (unsigned) y2 < (unsigned) (src_size.height - 1) &&
        (unsigned) x3 < (unsigned) (src_size.width - 1) &&
        (unsigned) y3 < (unsigned) (src_size.height - 1) )
    {
        // faster version
        for( y = -win_size.height; y <= win_size.height; y++, dst += dstStep )
        {
            for( x = -win_size.width; x <= win_size.width; x++ )
            {
                float xs = GET_X( x, y );
                float ys = GET_Y( x, y );
                float p0, p1;

                int ixs = cvFloor( xs );
                int iys = cvFloor( ys );
                const uchar *ptr = PTR_AT( ixs, iys );

                xs -= ixs;
                ys -= iys;

                p0 = ptr[0] + xs * (ptr[1] - ptr[0]);
                p1 = ptr[srcStep] + xs * (ptr[srcStep + 1] - ptr[srcStep]);
                dst[x] = p0 + ys * (p1 - p0);
            }
        }
    }
    else
#endif
    {
        // slower version - when the quadrangle tangents or intersect image bounds
        for( y = -win_size.height; y <= win_size.height; y++, dst += dstStep )
        {
            for( x = -win_size.width; x <= win_size.width; x++ )
            {
                float xs = GET_X( x, y );
                float ys = GET_Y( x, y );
                float p0, p1;

                int ixs = cvFloor( xs );
                int iys = cvFloor( ys );

                xs -= ixs;
                ys -= iys;

#if 0
                if( (unsigned) ixs < (unsigned) (src_size.width - 1) &&
                    (unsigned) iys < (unsigned) (src_size.height - 1) )
                {
                    const uchar *ptr = PTR_AT( ixs, iys );

                    p0 = ptr[0] + xs * (ptr[1] - ptr[0]);
                    p1 = ptr[srcStep] + xs * (ptr[srcStep + 1] - ptr[srcStep]);
                    dst[x] = p0 + ys * (p1 - p0);
                }
                else
#endif
                {
                    // the slowest branch

                    int t0, t1;

#define CLIP_X(x) (unsigned)(x) < (unsigned)src_size.width ?  \
                         (x) : (x) < 0 ? 0 : src_size.width - 1

#define CLIP_Y(y) (unsigned)(y) < (unsigned)src_size.height ? \
                         (y) : (y) < 0 ? 0 : src_size.height - 1

                    x0 = CLIP_X( ixs );
                    y0 = CLIP_Y( iys );
                    x1 = CLIP_X( ixs + 1 );
                    y1 = CLIP_Y( iys + 1 );

                    t0 = *PTR_AT( x0, y0 );
                    t1 = *PTR_AT( x1, y0 );
                    p0 = t0 + xs * (t1 - t0);

                    t0 = *PTR_AT( x0, y1 );
                    t1 = *PTR_AT( x1, y1 );
                    p1 = t0 + xs * (t1 - t0);
                    p0 = p0 + ys * (p1 - p0);
                    assert( 0.f <= p0 && p0 <= 255.f );
                    dst[x] = p0;

#undef CLIP_X
#undef CLIP_Y
                }
            }
        }
    }

#undef GET_X
#undef GET_Y
#undef PTR_AT

    return CV_OK;
}


IPCVAPI_IMPL( CvStatus, icvGetQuadrangleSubPix_8s32f_C1R,
              (const char *src, int srcStep, CvSize src_size,
               float *dst, int dstStep, CvSize win_size, CvPoint2D32f center, const float *A) )
{
    int x, y, x0, y0, x1, y1, x2, y2, x3, y3;

    if( !src || !dst || !A )
        return CV_NULLPTR_ERR;

    if( (win_size.width & win_size.height & 1) == 0 )
        return CV_BADSIZE_ERR;

    if( src_size.width <= 0 || src_size.height <= 0 ||
        (unsigned) srcStep < (unsigned) src_size.width ||
        (unsigned) dstStep < (unsigned) win_size.width * 4 || (dstStep & 3) != 0 )
        return CV_BADSIZE_ERR;

    dstStep /= 4;

    win_size.width >>= 1;
    win_size.height >>= 1;

#define GET_X(X,Y) ((X)*A[0] + (Y)*A[1] + center.x)
#define GET_Y(X,Y) ((X)*A[2] + (Y)*A[3] + center.y)
#define PTR_AT(X,Y) (src + (Y)*srcStep + (X))

    x0 = cvFloor( GET_X( -win_size.width, -win_size.height ));
    y0 = cvFloor( GET_Y( -win_size.width, -win_size.height ));
    x1 = cvFloor( GET_X( win_size.width, -win_size.height ));
    y1 = cvFloor( GET_Y( win_size.width, -win_size.height ));
    x2 = cvFloor( GET_X( -win_size.width, win_size.height ));
    y2 = cvFloor( GET_Y( -win_size.width, win_size.height ));
    x3 = cvFloor( GET_X( win_size.width, win_size.height ));
    y3 = cvFloor( GET_Y( win_size.width, win_size.height ));

    dst += win_size.width;

    /* check if the quadrangle is entirely in the image */
    if( (unsigned) x0 < (unsigned) (src_size.width - 1) &&
        (unsigned) y0 < (unsigned) (src_size.height - 1) &&
        (unsigned) x1 < (unsigned) (src_size.width - 1) &&
        (unsigned) y1 < (unsigned) (src_size.height - 1) &&
        (unsigned) x2 < (unsigned) (src_size.width - 1) &&
        (unsigned) y2 < (unsigned) (src_size.height - 1) &&
        (unsigned) x3 < (unsigned) (src_size.width - 1) &&
        (unsigned) y3 < (unsigned) (src_size.height - 1) )
    {
        /* faster version */
        for( y = -win_size.height; y <= win_size.height; y++, dst += dstStep )
        {
            for( x = -win_size.width; x <= win_size.width; x++ )
            {
                float xs = GET_X( x, y );
                float ys = GET_Y( x, y );
                float p0, p1;

                int ixs = cvFloor( xs );
                int iys = cvFloor( ys );
                const char *ptr = PTR_AT( ixs, iys );

                xs -= ixs;
                ys -= iys;

                p0 = ptr[0] + xs * (ptr[1] - ptr[0]);
                p1 = ptr[srcStep] + xs * (ptr[srcStep + 1] - ptr[srcStep]);
                dst[x] = p0 + ys * (p1 - p0);
            }
        }
    }
    else
    {
        /* slower version - when the quadrangle tangents or intersect image bounds */
        for( y = -win_size.height; y <= win_size.height; y++, dst += dstStep )
        {
            for( x = -win_size.width; x <= win_size.width; x++ )
            {
                float xs = GET_X( x, y );
                float ys = GET_Y( x, y );
                float p0, p1;

                int ixs = cvFloor( xs );
                int iys = cvFloor( ys );

                xs -= ixs;
                ys -= iys;

                if( (unsigned) ixs < (unsigned) (src_size.width - 1) &&
                    (unsigned) iys < (unsigned) (src_size.height - 1) )
                {
                    const char *ptr = PTR_AT( ixs, iys );

                    p0 = ptr[0] + xs * (ptr[1] - ptr[0]);
                    p1 = ptr[srcStep] + xs * (ptr[srcStep + 1] - ptr[srcStep]);
                    dst[x] = p0 + ys * (p1 - p0);
                }
                else
                {
                    /* the slowest branch */

                    int t0, t1;

#define CLIP_X(x) (unsigned)(x) < (unsigned)src_size.width ?  \
                         (x) : (x) < 0 ? 0 : src_size.width - 1

#define CLIP_Y(y) (unsigned)(y) < (unsigned)src_size.height ? \
                         (y) : (y) < 0 ? 0 : src_size.width - 1

                    x0 = CLIP_X( ixs );
                    y0 = CLIP_X( iys );
                    x1 = CLIP_X( ixs + 1 );
                    y1 = CLIP_X( iys + 1 );

                    t0 = *PTR_AT( x0, y0 );
                    t1 = *PTR_AT( x1, y0 );
                    p0 = t0 + xs * (t1 - t0);

                    t0 = *PTR_AT( x0, y1 );
                    t1 = *PTR_AT( x1, y1 );
                    p1 = t0 + xs * (t1 - t0);
                    dst[x] = p0 + ys * (p1 - p0);

#undef CLIP_X
#undef CLIP_Y
                }
            }
        }
    }

#undef GET_X
#undef GET_Y
#undef PTR_AT

    return CV_OK;
}


IPCVAPI_IMPL( CvStatus, icvGetQuadrangleSubPix_32f_C1R,
              (const float *src, int srcStep, CvSize src_size,
               float *dst, int dstStep, CvSize win_size, CvPoint2D32f center, const float *A) )
{
    int x, y, x0, y0, x1, y1, x2, y2, x3, y3;

    if( !src || !dst || !A )
        return CV_NULLPTR_ERR;

    if( (win_size.width & win_size.height & 1) == 0 )
        return CV_BADSIZE_ERR;

    if( src_size.width <= 0 || src_size.height <= 0 ||
        (unsigned) srcStep < (unsigned) src_size.width * 4 ||
        (unsigned) dstStep < (unsigned) win_size.width * 4 || ((dstStep | srcStep) & 3) != 0 )
        return CV_BADSIZE_ERR;

    srcStep /= 4;
    dstStep /= 4;

    win_size.width >>= 1;
    win_size.height >>= 1;

#define GET_X(X,Y) ((X)*A[0] + (Y)*A[1] + center.x)
#define GET_Y(X,Y) ((X)*A[2] + (Y)*A[3] + center.y)
#define PTR_AT(X,Y) (src + (Y)*srcStep + (X))

    x0 = cvFloor( GET_X( -win_size.width, -win_size.height ));
    y0 = cvFloor( GET_Y( -win_size.width, -win_size.height ));
    x1 = cvFloor( GET_X( win_size.width, -win_size.height ));
    y1 = cvFloor( GET_Y( win_size.width, -win_size.height ));
    x2 = cvFloor( GET_X( -win_size.width, win_size.height ));
    y2 = cvFloor( GET_Y( -win_size.width, win_size.height ));
    x3 = cvFloor( GET_X( win_size.width, win_size.height ));
    y3 = cvFloor( GET_Y( win_size.width, win_size.height ));

    dst += win_size.width;

    /* check if the quadrangle is entirely in the image */
    if( (unsigned) x0 < (unsigned) (src_size.width - 1) &&
        (unsigned) y0 < (unsigned) (src_size.height - 1) &&
        (unsigned) x1 < (unsigned) (src_size.width - 1) &&
        (unsigned) y1 < (unsigned) (src_size.height - 1) &&
        (unsigned) x2 < (unsigned) (src_size.width - 1) &&
        (unsigned) y2 < (unsigned) (src_size.height - 1) &&
        (unsigned) x3 < (unsigned) (src_size.width - 1) &&
        (unsigned) y3 < (unsigned) (src_size.height - 1) )
    {
        /* faster version */
        for( y = -win_size.height; y <= win_size.height; y++, dst += dstStep )
        {
            for( x = -win_size.width; x <= win_size.width; x++ )
            {
                float xs = GET_X( x, y );
                float ys = GET_Y( x, y );
                float p0, p1;

                int ixs = cvFloor( xs );
                int iys = cvFloor( ys );
                const float *ptr = PTR_AT( ixs, iys );

                xs -= ixs;
                ys -= iys;

                p0 = ptr[0] + xs * (ptr[1] - ptr[0]);
                p1 = ptr[srcStep] + xs * (ptr[srcStep + 1] - ptr[srcStep]);
                dst[x] = p0 + ys * (p1 - p0);
            }
        }
    }
    else
    {
        /* slower version - when the quadrangle tangents or intersect image bounds */
        for( y = -win_size.height; y <= win_size.height; y++, dst += dstStep )
        {
            for( x = -win_size.width; x <= win_size.width; x++ )
            {
                float xs = GET_X( x, y );
                float ys = GET_Y( x, y );
                float p0, p1;

                int ixs = cvFloor( xs );
                int iys = cvFloor( ys );

                xs -= ixs;
                ys -= iys;

                if( (unsigned) ixs < (unsigned) (src_size.width - 1) &&
                    (unsigned) iys < (unsigned) (src_size.height - 1) )
                {
                    const float *ptr = PTR_AT( ixs, iys );

                    p0 = ptr[0] + xs * (ptr[1] - ptr[0]);
                    p1 = ptr[srcStep] + xs * (ptr[srcStep + 1] - ptr[srcStep]);
                    dst[x] = p0 + ys * (p1 - p0);
                }
                else
                {
                    /* the slowest branch */
                    float t0, t1;

#define CLIP_X(x) (unsigned)(x) < (unsigned)src_size.width ?  \
                         (x) : (x) < 0 ? 0 : src_size.width - 1

#define CLIP_Y(y) (unsigned)(y) < (unsigned)src_size.height ? \
                         (y) : (y) < 0 ? 0 : src_size.width - 1

                    x0 = CLIP_X( ixs );
                    y0 = CLIP_X( iys );
                    x1 = CLIP_X( ixs + 1 );
                    y1 = CLIP_X( iys + 1 );

                    t0 = *PTR_AT( x0, y0 );
                    t1 = *PTR_AT( x1, y0 );
                    p0 = t0 + xs * (t1 - t0);

                    t0 = *PTR_AT( x0, y1 );
                    t1 = *PTR_AT( x1, y1 );
                    p1 = t0 + xs * (t1 - t0);
                    dst[x] = p0 + ys * (p1 - p0);

#undef CLIP_X
#undef CLIP_Y
                }
            }
        }
    }

#undef GET_X
#undef GET_Y
#undef PTR_AT

    return CV_OK;
}


/* 
   Initializes line iterator.
   Returns number of points on the line or negative number if error.
*/
CV_IMPL int
cvInitLineIterator( IplImage * img, CvPoint pt1, CvPoint pt2, CvLineIterator * iterator )
{
    CvSize img_size;
    uchar *img_data;
    int img_step = 0;
    int count = -1;

    CV_FUNCNAME( "cvInitLineIterator" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( img ));
    cvGetImageRawData( img, &img_data, &img_step, &img_size );

    switch (img->depth)
    {
    case IPL_DEPTH_8U:
    case IPL_DEPTH_8S:
    case IPL_DEPTH_32F:
        switch (img->nChannels)
        {
        case 1:
        case 3:
            count = icvInitLineIterator( img_data, img_step, img_size,
                                         icvGetBtPix( img ), pt1, pt2, iterator );
            if( count < 0 )
                CV_ERROR( IPL_StsInternal, icvCvErrMsg );
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }

    __CLEANUP__;
    __END__;

    return count;
}


/* End of file. */
