/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include "_cv.h"
#include "_cvimgstat.h"
#include <float.h>

/****************************************************************************************\
*                             Find sum of pixels in the ROI                              *
\****************************************************************************************/

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name: icvSumPixels_
//    Purpose: Calculates sum of all pixels
//    Context:
//    Parameters:
//      img  - source image
//      step - its step
//      size - its size
//      coi  - channel index (1..3) (if color image)
//    Returns:
//      sum of pixels values or large negative number (if error)
//    Notes:
//F*/

/* ///////////// 8uC1 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvSumPixels_8u_C1R, (const uchar * img,
                                              int step, CvSize size, double *sum) )
{
    int i, j, s = 0;

    if( !img || !sum )
        return CV_NULLPTR_ERR;
    if( (size.width | size.height) < 0 || step < size.width )
        return CV_BADSIZE_ERR;

    if( size.width == step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( i = 0; i < size.height; i++, img += step )
    {
        for( j = 0; j < size.width - 3; j += 4 )
            s += img[j] + img[j + 1] + img[j + 2] + img[j + 3];
        for( ; j < size.width; j++ )
            s += img[j];
    }

    *sum = s;
    return CV_OK;
}


/* ///////////// 8uC3 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvSumPixels_8u_C3CR, (const uchar * img,
                                               int step, CvSize size, int coi, double *sum) )
{
    int i, j, s = 0;

    size.width *= 3;
    coi--;

    if( !img || !sum )
        return CV_NULLPTR_ERR;
    if( (size.width | size.height) < 0 || step < size.width )
        return CV_BADSIZE_ERR;

    if( size.width == step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    img += coi;

    for( i = 0; i < size.height; i++, img += step )
    {
        for( j = 0; j < size.width - 9; j += 12 )
            s += img[j] + img[j + 3] + img[j + 6] + img[j + 9];
        for( ; j < size.width; j += 3 )
            s += img[j];
    }

    *sum = s;
    return CV_OK;
}


/* ///////////// 8sC1 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvSumPixels_8s_C1R, (const char *img,
                                              int step, CvSize size, double *sum) )
{
    int i, j, s = 0;

    if( !img || !sum )
        return CV_NULLPTR_ERR;
    if( (size.width | size.height) < 0 || step < size.width )
        return CV_BADSIZE_ERR;

    if( size.width == step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    for( i = 0; i < size.height; i++, img += step )
    {
        for( j = 0; j < size.width - 3; j += 4 )
            s += img[j] + img[j + 1] + img[j + 2] + img[j + 3];
        for( ; j < size.width; j++ )
            s += img[j];
    }

    *sum = s;
    return CV_OK;
}


/* ///////////// 8sC3 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvSumPixels_8s_C3CR, (const char *img,
                                               int step, CvSize size, int coi, double *sum) )
{
    int i, j, s = 0;

    size.width *= 3;
    coi--;

    if( !img || !sum )
        return CV_NULLPTR_ERR;
    if( (size.width | size.height) < 0 || step < size.width )
        return CV_BADSIZE_ERR;

    if( size.width == step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    img += coi;

    for( i = 0; i < size.height; i++, img += step )
    {
        for( j = 0; j < size.width - 9; j += 12 )
            s += img[j] + img[j + 3] + img[j + 6] + img[j + 9];
        for( ; j < size.width; j += 3 )
            s += img[j];
    }

    *sum = s;
    return CV_OK;
}


/* ///////////// 32fC1 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvSumPixels_32f_C1R, (const float *img,
                                               int step, CvSize size, double *sum) )
{
    int i, j;
    double s = 0;

    if( !img || !sum )
        return CV_NULLPTR_ERR;
    if( (size.width | size.height) < 0 || step < size.width * sizeof_float ||
        step % sizeof_float != 0 )
        return CV_BADSIZE_ERR;

    if( size.width == step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    step /= sizeof_float;

    for( i = 0; i < size.height; i++, img += step )
    {
        for( j = 0; j < size.width - 3; j += 4 )
            s += img[j] + img[j + 1] + img[j + 2] + img[j + 3];
        for( ; j < size.width; j++ )
            s += img[j];
    }

    *sum = s;
    return CV_OK;
}


/* ///////////// 32fC3 flavor //////////////////// */
IPCVAPI_IMPL( CvStatus, icvSumPixels_32f_C3CR, (const float *img,
                                                int step, CvSize size, int coi, double *sum) )
{
    int i, j;
    double s = 0;

    size.width *= 3;
    coi--;

    if( !img || !sum )
        return CV_NULLPTR_ERR;
    if( (size.width | size.height) < 0 || step < size.width * sizeof_float ||
        step % sizeof_float != 0 )
        return CV_BADSIZE_ERR;

    if( size.width == step )
    {
        size.width *= size.height;
        size.height = 1;
    }

    img += coi;
    step /= sizeof_float;

    for( i = 0; i < size.height; i++, img += step )
    {
        for( j = 0; j < size.width - 9; j += 12 )
            s += img[j] + img[j + 3] + img[j + 6] + img[j + 9];
        for( ; j < size.width; j += 3 )
            s += img[j];
    }

    *sum = s;
    return CV_OK;
}


/* End of file. */
