/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include "_cv.h"
#include <float.h>

void icvDAXPY( int n, double *x, double a, double *y );
double icvDDOT( int n, double *x, double *y );
void icvGivens( int n, double *x, double *y, double c, double s );

CvStatus icvSVDInitialWV_32f( CvMatr32f A,      /* matrix size n x m */
                              int n,    /* number of columns, width  */
                              int m,    /* numbrt of rows, height */
                              CvMatr32f w,      /* matrix size n x m */
                              CvMatr32f v,      /* matrix size n x m */
                              CvVect32f rv1 );  /*  vector n */

CvStatus icvSVDBiDiag_32f( CvMatr32f A, int n,  /* number of colomns, width, */
                           int m,       /* number of rows, height */
                           CvMatr32f w, /* matrix  n x n */
                           CvMatr32f rv1,       /* vector n */
                           float *anorm );

float icvPythag_32f( float a, float b );        /*const */

CvStatus icvSVDSym_32f( CvMatr32f LtL, CvVect32f resV, int num );

float square_32f( float x );

/*========================================================================*/
CvStatus
icvSVDInitialWV_32f( CvMatr32f A,       /* matrix size n x m */
                     int n,     /* number of columns, width  */
                     int m,     /* numbrt of rows, height */
                     CvMatr32f w,       /* matrix size n x m */
                     CvMatr32f v,       /* matrix size n x m */
                     CvVect32f rv1 )    /*  vector n */
{
    int /*m = nrows(), n = ncols(), */ i, j, k, l;
    float f, g = 0.0, s;

    for( i = (n - 1); i >= 0; i-- )
    {
        l = i + 1;
        if( i < (n - 1) )
        {
            if( g )
            {
                for( j = l; j < n; j++ )
                    v[j * n + i] = (A[i * n + j] / A[i * n + l]) / g;
                /* double division to reduce underflow */
                for( j = l; j < n; j++ )
                {
                    for( s = 0.0, k = l; k < n; k++ )
                        s += A[i * n + k] * v[k * n + j];
                    for( k = l; k < n; k++ )
                        v[k * n + j] += s * v[k * n + i];
                }
            }
            for( j = l; j < n; j++ )
                v[i * n + j] = v[j * n + i] = 0.0;
        }
        v[i * n + i] = 1.0;
        g = rv1[i];
    }
    for( i = (n - 1); i >= 0; i-- )
    {
        l = i + 1;
        g = w[i * n + i];
        if( i < (n - 1) )
        {
            for( j = l; j < n; j++ )
                A[i * n + j] = 0.0;
        }
        if( g )
        {
            g = 1.0f / g;
            if( i != (n - 1) )
            {
                for( j = l; j < n; j++ )
                {
                    for( s = 0.0, k = l; k < m; k++ )
                        s += A[k * n + i] * A[k * n + j];
                    f = (s / A[i * n + i]) * g;
                    for( k = i; k < m; k++ )
                        A[k * n + j] += f * A[k * n + i];
                }
            }
            for( j = i; j < m; j++ )
                A[j * n + i] *= g;
        }
        else
        {
            for( j = i; j < m; j++ )
                A[j * n + i] = 0.0;
        }
        A[i * n + i] += 1.0;
    }
    return CV_NO_ERR;
}

/*========================================================================*/


CvStatus
icvSVDBiDiag_32f( CvMatr32f A, int n,   /* number of colomns, width, */
                  int m,        /* number of rows, height */
                  CvMatr32f w,  /* matrix  n x n */
                  CvMatr32f rv1,        /* vector n */
                  float *anorm )
{
    float f, g = 0.0, h, r, s = 0.0, scale = 0.0;
    int /*m = nrows(), n = ncols(), */ i, j, k, l;

    *anorm = 0.0;
    for( i = 0; i < n; i++ )
    {
        l = i + 1;
        rv1[i] = scale * g;
        g = s = scale = 0.0;
        if( i < m )
        {
            for( k = i; k < m; k++ )
                scale += (float) fabs( A[k * n + i] );
            if( scale )
            {
                for( k = i; k < m; k++ )
                {
                    A[k * n + i] /= scale;
                    s += A[k * n + i] * A[k * n + i];
                }
                f = A[i * n + i];
                g = (float) sqrt( s );
                if( f >= 0.0 )
                    g = -g;
                h = f * g - s;
                A[i * n + i] = f - g;
                if( i != (n - 1) )
                {
                    for( j = l; j < n; j++ )
                    {
                        for( s = 0.0, k = i; k < m; k++ )
                            s += A[k * n + i] * A[k * n + j];
                        f = s / h;
                        for( k = i; k < m; k++ )
                            A[k * n + j] += f * A[k * n + i];
                    }
                }
                for( k = i; k < m; k++ )
                    A[k * n + i] *= scale;
            }
        }
        w[i * n + i] = scale * g;
        g = s = scale = 0.0;
        if( i < m && i != (n - 1) )
        {
            for( k = l; k < n; k++ )
                scale += (float) fabs( A[i * n + k] );
            if( scale )
            {
                for( k = l; k < n; k++ )
                {
                    A[i * n + k] /= scale;
                    s += A[i * n + k] * A[i * n + k];
                }
                f = A[i * n + l];
                g = (float) sqrt( s );
                if( f >= 0.0 )
                    g = -g;
                h = 1.0f / (f * g - s);
                A[i * n + l] = f - g;
                for( k = l; k < n; k++ )
                    rv1[k] = A[i * n + k] * h;
                if( i != (m - 1) )
                {
                    for( j = l; j < m; j++ )
                    {
                        for( s = 0.0, k = l; k < n; k++ )
                            s += A[j * n + k] * A[i * n + k];
                        for( k = l; k < n; k++ )
                            A[j * n + k] += s * rv1[k];
                    }
                }
                for( k = l; k < n; k++ )
                    A[i * n + k] *= scale;
            }
        }
        r = ((float) fabs( w[i * n + i] ) + (float) fabs( rv1[i] ));
        if( r > *anorm )
            *anorm = r;
    }
    return CV_NO_ERR;
}

/*========================================================================*/

float
icvPythag_32f( float a, float b )       /*const */
{
    /* computes (a^2 + b^2)^1/2 without destructive underflow or overflow */
    float at, bt, ct;

    at = (float) fabs( a );
    bt = (float) fabs( b );
    if( at > bt )
    {
        ct = bt / at;
        return ( float ) (at * sqrt( 1.0 + ct * ct ));
    }
    else
    {
        if( bt )
        {
            ct = at / bt;
            return ( float ) (bt * sqrt( 1.0 + ct * ct ));
        }
        else
        {
            return 0.0;
        }
    }
}

/*========================================================================*/

IPCVAPI_IMPL( CvStatus, icvSVDecomposition_32f, (CvMatr32f A, int n,    /* number of columns */
                                                 int m, /* number of rows    */
                                                 CvMatr32f w,   /*  singular values size n x n */
                                                 CvMatr32f v) ) /*  size n x n */
{
    /* Singular value decomposition.  A = [a][w][vT].
       Input A matrix is this.  Output orthogonal a matrix overwrites this. 
       Output n x n w matrix is diagonal.  Output n x n v matrix is orthogonal.
       Ainverse = [v][diag(1/w)][aT].  Inverses of a and v equal their transposes.
       Scratch 1 x n rv1 vector is used by this function.
       w~=0 elements may have 1/w set to 0 to solve singular system. */

    int flag, i, its, j, jj, k, l, nm = 0;
    float c, f, h, s, x, y, z;
    float anorm = 0.0, g = 0.0, r;
    CvVect32f rv1;
    CvMatr32f tmpA;
    int min_num;
    float tmpf;

    tmpA = icvCreateMatrix_32f( n, m );
    rv1 = icvCreateVector_32f( n );

    icvCopyMatrix_32f( A, n, m, tmpA );
    icvSetZero_32f( rv1, 1, n );

    if( m < n )
        return CV_BADCOEF_ERR;  /* a error: rows < cols */
    icvSetZero_32f( w, n, n );
    icvSetZero_32f( v, n, n );

    icvSVDBiDiag_32f( A, n, m, w, rv1, &anorm );
    icvSVDInitialWV_32f( A, n, m, w, v, rv1 );
    
    for( k = (n - 1); k >= 0; k-- )
    {
        for( its = 1; its <= 30; its++ )
        {
            flag = 1;
            for( l = k; l >= 0; l-- )
            {
                nm = l - 1;
                if( (fabs( rv1[l] ) + anorm) == anorm )
                {
                    flag = 0;
                    break;
                }
                if( (fabs( w[nm * n + nm] ) + anorm) == anorm )
                    break;
            }
            if( flag )
            {
                c = 0.0;
                s = 1.0;
                for( i = l; i <= k; i++ )
                {
                    f = s * rv1[i];
                    if( (fabs( f ) + anorm) != anorm )
                    {
                        g = w[i * n + i];
                        h = icvPythag_32f( f, g );
                        w[i * n + i] = h;
                        h = 1.0f / h;
                        c = g * h;
                        s = (-f * h);
                        for( j = 0; j < m; j++ )
                        {
                            y = A[j * n + nm];
                            z = A[j * n + i];
                            A[j * n + nm] = y * c + z * s;
                            A[j * n + i] = z * c - y * s;
                        }
                    }
                }
            }
            z = w[k * n + k];
            if( l == k )
            {
                if( z < 0.0 )
                {
                    w[k * n + k] = -z;
                    for( j = 0; j < n; j++ )
                        v[j * n + k] = (-v[j * n + k]);
                }
                break;
            }
            if( its == 30 )
            {
                return CV_BADCOEF_ERR;  /* error: no convergence */
            }
            x = w[l * n + l];
            nm = k - 1;
            y = w[nm * n + nm];
            g = rv1[nm];
            h = rv1[k];
            f = ((y - z) * (y + z) + (g - h) * (g + h)) / (2.0f * h * y);
            g = icvPythag_32f( f, 1.0 );
            r = (f >= 0.0 ? g : -g);
            f = ((x - z) * (x + z) + h * ((y / (f + r)) - h)) / x;
            c = s = 1.0;
            for( j = l; j <= nm; j++ )
            {
                i = j + 1;
                g = rv1[i];
                y = w[i * n + i];
                h = s * g;
                g = c * g;
                z = icvPythag_32f( f, h );
                rv1[j] = z;
                c = f / z;
                s = h / z;
                f = x * c + g * s;
                g = g * c - x * s;
                h = y * s;
                y = y * c;
                for( jj = 0; jj < n; jj++ )
                {
                    x = v[jj * n + j];
                    z = v[jj * n + i];
                    v[jj * n + j] = x * c + z * s;
                    v[jj * n + i] = z * c - x * s;
                }
                z = icvPythag_32f( f, h );
                w[j * n + j] = z;
                if( z )
                {
                    z = 1.0f / z;
                    c = f * z;
                    s = h * z;
                }
                f = (c * g) + (s * y);
                x = (c * y) - (s * g);
                for( jj = 0; jj < m; jj++ )
                {
                    y = A[jj * n + j];
                    z = A[jj * n + i];
                    A[jj * n + j] = y * c + z * s;
                    A[jj * n + i] = z * c - y * s;
                }
            }
            rv1[l] = 0.0;
            rv1[k] = f;
            w[k * n + k] = x;
        }
    }


    /* Sort values */
    for( i = n - 1; i >= 0; i-- )
    {
        min_num = i;
        for( j = 0; j < i; j++ )
        {                       /* find minimal mean */
            if( w[j * n + j] < w[i * n + i] )
            {
                min_num = j;
            }
        }

        if( i != min_num )
        {
            tmpf = w[min_num * n + min_num];
            w[min_num * n + min_num] = w[i * n + i];
            w[i * n + i] = tmpf;

            for( k = 0; k < n; k++ )
            {
                tmpf = A[k * n + i];
                A[k * n + i] = A[k * n + min_num];
                A[k * n + min_num] = tmpf;
            }

            for( k = 0; k < n; k++ )
            {
                tmpf = v[k * n + i];
                v[k * n + i] = v[k * n + min_num];
                v[k * n + min_num] = tmpf;
            }

        }
    }

    {
        CvMatr32f tmp_matrV;
        CvMatr32f tmp_matrS;
        CvVect32f tmp_f1;
        CvVect32f tmp_f2;
        int i, j, t;
        float dot;

        tmp_matrV = icvCreateMatrix_32f( n, n );
        tmp_matrS = icvCreateMatrix_32f( n, n );
        tmp_f1 = icvCreateVector_32f( n );
        tmp_f2 = icvCreateVector_32f( n );

        /* Find eigen vectors */


        icvSetZero_32f( tmp_matrS, n, n );

        icvJacobiEigens_32f( tmpA, tmp_matrV, tmp_matrS,       /* this matrix uses as a vector(3) */
                              n, 0.0f );

        /* Change signs if need */
        for( i = 0; i < n; i++ )
        {                       /* for each column */

            for( t = 0; t < n; t++ )
            {
                tmp_f1[t] = tmp_matrV[t * n + i];
                tmp_f2[t] = A[t * n + i];
            }

            icvDotProduct_32f( tmp_f1, tmp_f2, n, &dot );
            if( dot < 0 )
            {
                for( j = 0; j < n; j++ )
                {               /* for each row */
                    A[j * n + i] = -A[j * n + i];
                    v[j * n + i] = -v[j * n + i];
                }
            }
        }

        icvDeleteMatrix( tmp_matrV );
        icvDeleteMatrix( tmp_matrS );
        icvDeleteVector( tmp_f1 );
        icvDeleteVector( tmp_f2 );

    }

    icvDeleteMatrix( tmpA );
    icvDeleteVector( rv1 );
    return CV_NO_ERR;
}



/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////


CvStatus
icvSVDInitialWV_64d( CvMatr64d A,       /* matrix size n x m */
                     int n,     /* number of columns, width  */
                     int m,     /* numbrt of rows, height */
                     CvMatr64d w,       /* matrix size n x m */
                     CvMatr64d v,       /* matrix size n x m */
                     CvVect64d rv1 )    /*  vector n */
{
    int /*m = nrows(), n = ncols(), */ i, j, k, l;
    double f, g = 0.0, s;

    for( i = (n - 1); i >= 0; i-- )
    {
        l = i + 1;
        if( i < (n - 1) )
        {
            if( g )
            {
                for( j = l; j < n; j++ )
                    v[j * n + i] = (A[i * n + j] / A[i * n + l]) / g;
                /* double division to reduce underflow */
                for( j = l; j < n; j++ )
                {
                    for( s = 0.0, k = l; k < n; k++ )
                        s += A[i * n + k] * v[k * n + j];
                    for( k = l; k < n; k++ )
                        v[k * n + j] += s * v[k * n + i];
                }
            }
            for( j = l; j < n; j++ )
                v[i * n + j] = v[j * n + i] = 0.0;
        }
        v[i * n + i] = 1.0;
        g = rv1[i];
    }
    for( i = (n - 1); i >= 0; i-- )
    {
        l = i + 1;
        g = w[i * n + i];
        if( i < (n - 1) )
        {
            for( j = l; j < n; j++ )
                A[i * n + j] = 0.0;
        }
        if( g )
        {
            g = 1.0f / g;
            if( i != (n - 1) )
            {
                for( j = l; j < n; j++ )
                {
                    for( s = 0.0, k = l; k < m; k++ )
                        s += A[k * n + i] * A[k * n + j];
                    f = (s / A[i * n + i]) * g;
                    for( k = i; k < m; k++ )
                        A[k * n + j] += f * A[k * n + i];
                }
            }
            for( j = i; j < m; j++ )
                A[j * n + i] *= g;
        }
        else
        {
            for( j = i; j < m; j++ )
                A[j * n + i] = 0.0;
        }
        A[i * n + i] += 1.0;
    }
    return CV_NO_ERR;
}

/*========================================================================*/


CvStatus
icvSVDBiDiag_64d( CvMatr64d A, int n,   /* number of colomns, width, */
                  int m,        /* number of rows, height */
                  CvMatr64d w,  /* matrix  n x n */
                  CvMatr64d rv1,        /* vector n */
                  double *anorm )
{
    double f, g = 0.0, h, r, s = 0.0, scale = 0.0;
    int /*m = nrows(), n = ncols(), */ i, j, k, l;

    *anorm = 0.0;
    for( i = 0; i < n; i++ )
    {
        l = i + 1;
        rv1[i] = scale * g;
        g = s = scale = 0.0;
        if( i < m )
        {
            for( k = i; k < m; k++ )
                scale += fabs( A[k * n + i] );
            if( scale )
            {
                for( k = i; k < m; k++ )
                {
                    A[k * n + i] /= scale;
                    s += A[k * n + i] * A[k * n + i];
                }
                f = A[i * n + i];
                g = sqrt( s );
                if( f >= 0.0 )
                    g = -g;
                h = f * g - s;
                A[i * n + i] = f - g;
                if( i != (n - 1) )
                {
                    for( j = l; j < n; j++ )
                    {
                        for( s = 0.0, k = i; k < m; k++ )
                            s += A[k * n + i] * A[k * n + j];
                        f = s / h;
                        for( k = i; k < m; k++ )
                            A[k * n + j] += f * A[k * n + i];
                    }
                }
                for( k = i; k < m; k++ )
                    A[k * n + i] *= scale;
            }
        }
        w[i * n + i] = scale * g;
        g = s = scale = 0.0;
        if( i < m && i != (n - 1) )
        {
            for( k = l; k < n; k++ )
                scale += fabs( A[i * n + k] );
            if( scale )
            {
                for( k = l; k < n; k++ )
                {
                    A[i * n + k] /= scale;
                    s += A[i * n + k] * A[i * n + k];
                }
                f = A[i * n + l];
                g = sqrt( s );
                if( f >= 0.0 )
                    g = -g;
                h = 1.0f / (f * g - s);
                A[i * n + l] = f - g;
                for( k = l; k < n; k++ )
                    rv1[k] = A[i * n + k] * h;
                if( i != (m - 1) )
                {
                    for( j = l; j < m; j++ )
                    {
                        for( s = 0.0, k = l; k < n; k++ )
                            s += A[j * n + k] * A[i * n + k];
                        for( k = l; k < n; k++ )
                            A[j * n + k] += s * rv1[k];
                    }
                }
                for( k = l; k < n; k++ )
                    A[i * n + k] *= scale;
            }
        }
        r = (fabs( w[i * n + i] ) + fabs( rv1[i] ));
        if( r > *anorm )
            *anorm = r;
    }
    return CV_NO_ERR;
}

/*========================================================================*/

double
icvPythag_64d( double a, double b )     /*const */
{
    /* computes (a^2 + b^2)^1/2 without destructive underflow or overflow */
    double at, bt, ct;

    at = fabs( a );
    bt = fabs( b );
    if( at > bt )
    {
        ct = bt / at;
        return ( at * sqrt( 1.0 + ct * ct ));
    }
    else
    {
        if( bt )
        {
            ct = at / bt;
            return ( bt * sqrt( 1.0 + ct * ct ));
        }
        else
        {
            return 0.0;
        }
    }
}

/*========================================================================*/

IPCVAPI_IMPL( CvStatus, icvSVDecomposition_64d, (CvMatr64d A, int n,    /* number of columns */
                                                 int m, /* number of rows    */
                                                 CvMatr64d w,   /*  singular values size n x n */
                                                 CvMatr64d v) ) /*  size n x n */
{
    /* Singular value decomposition.  A = [a][w][vT].
       Input A matrix is this.  Output orthogonal a matrix overwrites this. 
       Output n x n w matrix is diagonal.  Output n x n v matrix is orthogonal.
       Ainverse = [v][diag(1/w)][aT].  Inverses of a and v equal their transposes.
       Scratch 1 x n rv1 vector is used by this function.
       w~=0 elements may have 1/w set to 0 to solve singular system. */

    int flag, i, its, j, jj, k, l, nm = 0;
    double c, f, h, s, x, y, z;
    double anorm = 0.0, g = 0.0, r;
    CvVect64d rv1;
    CvMatr64d tmpA;
    int min_num;
    double tmpf;

    tmpA = icvCreateMatrix_64d( n, m );
    rv1 = icvCreateVector_64d( n );

    icvCopyMatrix_64d( A, n, m, tmpA );
    icvSetZero_64d( rv1, 1, n );

    if( m < n )
        return CV_BADCOEF_ERR;  /* a error: rows < cols */
    icvSetZero_64d( w, n, n );
    icvSetZero_64d( v, n, n );


    icvSVDBiDiag_64d( A, n, m, w, rv1, &anorm );
    icvSVDInitialWV_64d( A, n, m, w, v, rv1 );

    for( k = (n - 1); k >= 0; k-- )
    {
        for( its = 1; its <= 30; its++ )
        {
            flag = 1;
            for( l = k; l >= 0; l-- )
            {
                nm = l - 1;
                if( (fabs( rv1[l] ) + anorm) == anorm )
                {
                    flag = 0;
                    break;
                }
                if( (fabs( w[nm * n + nm] ) + anorm) == anorm )
                    break;
            }
            if( flag )
            {
                c = 0.0;
                s = 1.0;
                for( i = l; i <= k; i++ )
                {
                    f = s * rv1[i];
                    if( (fabs( f ) + anorm) != anorm )
                    {
                        g = w[i * n + i];
                        h = icvPythag_64d( f, g );
                        w[i * n + i] = h;
                        h = 1.0f / h;
                        c = g * h;
                        s = (-f * h);
                        for( j = 0; j < m; j++ )
                        {
                            y = A[j * n + nm];
                            z = A[j * n + i];
                            A[j * n + nm] = y * c + z * s;
                            A[j * n + i] = z * c - y * s;
                        }
                    }
                }
            }
            z = w[k * n + k];
            if( l == k )
            {
                if( z < 0.0 )
                {
                    w[k * n + k] = -z;
                    for( j = 0; j < n; j++ )
                        v[j * n + k] = (-v[j * n + k]);
                }
                break;
            }
            if( its == 30 )
            {
                return CV_BADCOEF_ERR;  /* error: no convergence */
            }
            x = w[l * n + l];
            nm = k - 1;
            y = w[nm * n + nm];
            g = rv1[nm];
            h = rv1[k];
            f = ((y - z) * (y + z) + (g - h) * (g + h)) / (2.0f * h * y);
            g = icvPythag_64d( f, 1.0 );
            r = (f >= 0.0 ? g : -g);
            f = ((x - z) * (x + z) + h * ((y / (f + r)) - h)) / x;
            c = s = 1.0;
            for( j = l; j <= nm; j++ )
            {
                i = j + 1;
                g = rv1[i];
                y = w[i * n + i];
                h = s * g;
                g = c * g;
                z = icvPythag_64d( f, h );
                rv1[j] = z;
                c = f / z;
                s = h / z;
                f = x * c + g * s;
                g = g * c - x * s;
                h = y * s;
                y = y * c;
                for( jj = 0; jj < n; jj++ )
                {
                    x = v[jj * n + j];
                    z = v[jj * n + i];
                    v[jj * n + j] = x * c + z * s;
                    v[jj * n + i] = z * c - x * s;
                }
                z = icvPythag_64d( f, h );
                w[j * n + j] = z;
                if( z )
                {
                    z = 1.0f / z;
                    c = f * z;
                    s = h * z;
                }
                f = (c * g) + (s * y);
                x = (c * y) - (s * g);
                for( jj = 0; jj < m; jj++ )
                {
                    y = A[jj * n + j];
                    z = A[jj * n + i];
                    A[jj * n + j] = y * c + z * s;
                    A[jj * n + i] = z * c - y * s;
                }
            }
            rv1[l] = 0.0;
            rv1[k] = f;
            w[k * n + k] = x;
        }
    }


    /* Sort values */
    for( i = n - 1; i >= 0; i-- )
    {
        min_num = i;
        for( j = 0; j < i; j++ )
        {                       /* find minimal mean */
            if( w[j * n + j] < w[i * n + i] )
            {
                min_num = j;
            }
        }

        if( i != min_num )
        {
            tmpf = w[min_num * n + min_num];
            w[min_num * n + min_num] = w[i * n + i];
            w[i * n + i] = tmpf;

            for( k = 0; k < n; k++ )
            {
                tmpf = A[k * n + i];
                A[k * n + i] = A[k * n + min_num];
                A[k * n + min_num] = tmpf;
            }

            for( k = 0; k < n; k++ )
            {
                tmpf = v[k * n + i];
                v[k * n + i] = v[k * n + min_num];
                v[k * n + min_num] = tmpf;
            }

        }
    }

    {
        CvMatr64d tmp_matrV;
        CvMatr64d tmp_matrS;
        CvVect64d tmp_f1;
        CvVect64d tmp_f2;
        int i, j, t;
        double dot;

        tmp_matrV = icvCreateMatrix_64d( n, n );
        tmp_matrS = icvCreateMatrix_64d( n, n );
        tmp_f1 = icvCreateVector_64d( n );
        tmp_f2 = icvCreateVector_64d( n );

        /* Find eigen vectors */


        icvSetZero_64d( tmp_matrS, n, n );

        icvJacobiEigens_64d( tmpA, tmp_matrV, tmp_matrS,       /* this matrix uses as a vector(3) */
                              n, 0.0f );

        /* Change signs if need */
        for( i = 0; i < n; i++ )
        {                       /* for each column */

            for( t = 0; t < n; t++ )
            {
                tmp_f1[t] = tmp_matrV[t * n + i];
                tmp_f2[t] = A[t * n + i];
            }

            icvDotProduct_64d( tmp_f1, tmp_f2, n, &dot );
            if( dot < 0 )
            {
                for( j = 0; j < n; j++ )
                {               /* for each row */
                    A[j * n + i] = -A[j * n + i];
                    v[j * n + i] = -v[j * n + i];
                }
            }
        }

        icvDeleteMatrix( tmp_matrV );
        icvDeleteMatrix( tmp_matrS );
        icvDeleteVector( tmp_f1 );
        icvDeleteVector( tmp_f2 );

    }

    icvDeleteMatrix( tmpA );
    icvDeleteVector( rv1 );
    return CV_NO_ERR;
}






/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////


static double
pythag( double a, double b )
{
    a = fabs( a );
    b = fabs( b );
    if( a > b )
    {
        b /= a;
        a *= sqrt( 1 + b * b );
    }
    else if( b != 0 )
    {
        a /= b;
        a = b * sqrt( 1 + a * a );
    }

    return a;
}


/****************************************************************************************/
/****************************************************************************************/

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name: icvSVD_64f
//    Purpose: Calculates singular value decomposition for rectangular
//             matrix a (m rows x n columns), such that u'*a*v = w,
//             u and v are ortogonal, w is diagonal and only the diagonal elements
//             are stored.
//    Context:
//    Parameters:  
//      m    - number of rows in matrix a
//      n    - number of columns in matrix a
//      a    - source matrix
//      lda  - step of the matrix a (full number of columns)
//      w    - singular values of the matrix a. array of min(m,n) elements.
//             it is ordered in decreasing order.
//      u    - left singular vectors - ortogonal matrix (m x m). Can be 0,
//             if is not required (function returns transposed matrix).
//      ldu  - step of the matrix u (full number of columns)
//      v    - right singular vectors - ortogonal matrix (n x n). Can be 0,
//             if is not required (function returns transposed matrix).
//      ldv  - step of the matrix v (full number of columns)
//    Returns:
//      CV_NO_ERR or error code
//    Notes:
//      Function damages source matrix a
//F*/
IPCVAPI_IMPL( CvStatus, icvSVD_64f, (int m, int n, double *a, int lda,
                                     double *w, double *u, int ldu, double *v, int ldv) )
{
#define MAX_SIZE  450
#define MAX_ITERS  30
    double e[MAX_SIZE];
    double temp[MAX_SIZE];
    double *w1, *e1;
    double *hv;
    double ku0 = 0, kv0 = 0;
    double anorm = 0;
    double *a1 = a, *u0 = u, *v0 = v;
    double *u1, *v1;
    int ldu1, ldv1;
    double scale, h;
    int i, j, k, l;
    int nm, m1, n1;
    int iters = 0;

    if( m <= 0 || MAX_SIZE < m || n <= 0 || MAX_SIZE < n )
        return CV_BADRANGE_ERR;

    if( !a || !w )
        return CV_NULLPTR_ERR;
    if( lda < n )
        return CV_BADSIZE_ERR;

    if( u && ldu < m )
        return CV_BADSIZE_ERR;
    if( v && ldv < n )
        return CV_BADSIZE_ERR;

    if( a == u || a == v || (u == v && u != 0) )
        return CV_INPLACE_NOT_SUPPORTED_ERR;

    if( m >= n )
    {
        w1 = w;
        e1 = e + 1;
        nm = n;
    }
    else
    {
        w1 = e + 1;
        e1 = w;
        nm = m;
    }

    memset( w, 0, nm * sizeof( double ));
    memset( e, 0, nm * sizeof( double ));

    m1 = m;
    n1 = n;

    if( m < n )
        goto row_transform;

    for( ;; )
    {
        if( m1 == 0 )
            break;

        scale = h = 0;

        a = a1;
        hv = u ? u : w1;

        for( j = 0; j < m1; j++ )
        {
            double t = a[0];

            hv[j] = t;
            scale += fabs( t );
            a += lda;
        }

        if( scale != 0 )
        {
            double f, g, s = 0;

            f = 1. / scale;

            for( j = 0; j < m1; j++ )
            {
                double t = (hv[j] *= f);

                s += t * t;
            }

            g = sqrt( s );
            f = hv[0];
            if( f >= 0 )
                g = -g;
            hv[0] = f - g;
            h = 1. / (f * g - s);

            memset( temp, 0, n1 * sizeof( temp[0] ));

            a = a1;

            /* calc temp[0:n-i] = a[i:m,i:n]'*hv[0:m-i] */
            for( j = 0; j < m1; j++, a += lda )
            {
                s = hv[j];
                icvDAXPY( n1 - 1, a + 1, s, temp + 1 );
            }

            /* scale temp: temp[0:n-i] = temp[0:n-i] * (-2/(hv[0:m-i]'*hv[0:m-i])); */
            for( k = 1; k < n1; k++ )
                temp[k] *= h;

            a = a1;

            /* modify a: a[i:m,i:n] = a[i:m,i:n] + hv[0:m-i]*temp[0:n-i]' */
            for( j = 0; j < m1; j++, a += lda )
            {
                s = hv[j];
                icvDAXPY( n1 - 1, temp + 1, s, a + 1 );
            }
            *w1++ = g * scale;
        }
        /* store -2/(hv'*hv) */
        if( u )
        {
            if( m1 == m )
                ku0 = h;
            else
                hv[-1] = h;
        }

        a1++;
        n1--;
        if( v )
            v += ldv + 1;

      row_transform:

        if( n1 == 0 )
            break;

        scale = h = 0;

        a = a1;
        hv = v ? v : e1;

        for( j = 0; j < n1; j++ )
        {
            double t = a[j];

            hv[j] = t;
            scale += fabs( t );
        }

        if( scale != 0 )
        {
            double f, g, s = 0;

            f = 1. / scale;

            for( j = 0; j < n1; j++ )
            {
                double t = (hv[j] *= f);

                s += t * t;
            }

            g = sqrt( s );
            f = hv[0];
            if( f >= 0 )
                g = -g;
            hv[0] = f - g;
            h = 1. / (f * g - s);

            /* update a[i:m:i+1:n] = a[i:m,i+1:n] + (a[i:m,i+1:n]*hv[0:m-i])*... */
            for( j = 1; j < m1; j++ )
            {
                a += lda;
                s = icvDDOT( n1, hv, a ) * h;

                icvDAXPY( n1, hv, s, a );
            }
            *e1++ = g * scale;
        }

        /* store -2/(hv'*hv) */
        if( v )
        {
            if( n1 == n )
                kv0 = h;
            else
                hv[-1] = h;
        }

        a1 += lda;
        m1--;
        if( u )
            u += ldu + 1;
    }

    m1 -= m1 != 0;
    n1 -= n1 != 0;

    /* accumulate left transformations */
    if( u )
    {
        m1 = m - m1;
        u = u0 + m1 * ldu;
        for( i = m1; i < m; i++, u += ldu )
        {
            memset( u + m1, 0, (m - m1) * sizeof( u[0] ));
            u[i] = 1.;
        }

        for( i = m1 - 1; i >= 0; i-- )
        {
            double h, s;

            l = m - i;
            hv = u0 + (ldu + 1) * i;
            h = i == 0 ? ku0 : hv[-1];

            assert( h <= 0 );

            if( h != 0 )
            {
                u = hv;
                for( j = 1; j < l; j++ )
                {
                    u += ldu;
                    s = icvDDOT( l - 1, hv + 1, u + 1 ) * h;

                    u[0] = s * hv[0];
                    icvDAXPY( l - 1, hv + 1, s, u + 1 );
                }

                s = hv[0] * h;
                for( k = 0; k < l; k++ )
                    hv[k] *= s;
                hv[0] += 1;
            }
            else
            {
                for( j = 1; j < l; j++ )
                    hv[j] = hv[j * ldu] = 0;
                hv[0] = 1;
            }
        }
        u = u0;
    }


    /* accumulate right transformations */
    if( v )
    {
        n1 = n - n1;
        v = v0 + n1 * ldv;
        for( i = n1; i < n; i++, v += ldv )
        {
            memset( v + n1, 0, (n - n1) * sizeof( v[0] ));
            v[i] = 1.;
        }

        for( i = n1 - 1; i >= 0; i-- )
        {
            double h, s;

            l = n - i;
            hv = v0 + (ldv + 1) * i;
            h = i == 0 ? kv0 : hv[-1];

            assert( h <= 0 );

            if( h != 0 )
            {
                v = hv;
                for( j = 1; j < l; j++ )
                {
                    v += ldv;
                    s = icvDDOT( l - 1, hv + 1, v + 1 ) * h;

                    v[0] = s * hv[0];
                    icvDAXPY( l - 1, hv + 1, s, v + 1 );
                }

                s = hv[0] * h;
                for( k = 0; k < l; k++ )
                    hv[k] *= s;
                hv[0] += 1;
            }
            else
            {
                for( j = 1; j < l; j++ )
                    hv[j] = hv[j * ldv] = 0;
                hv[0] = 1;
            }
        }
        v = v0;
    }

    for( i = 0; i < nm; i++ )
    {
        double tnorm = fabs( w[i] ) + fabs( e[i] );

        if( anorm < tnorm )
            anorm = tnorm;
    }

    if( m >= n )
    {
        m1 = m;
        n1 = n;
        u1 = u;
        ldu1 = ldu;
        v1 = v;
        ldv1 = ldv;
    }
    else
    {
        m1 = n;
        n1 = m;
        u1 = v;
        ldu1 = ldv;
        v1 = u;
        ldv1 = ldu;
    }

    /* diagonalization of the bidiagonal form */
    for( k = nm - 1; k >= 0; k-- )
    {
        double z = 0;

        iters = 0;

        for( ;; )               /* do iterations */
        {
            double c, s, f, g, h, x, y;
            int flag = 0;

            /* test for splitting */
            for( l = k; l >= 0; l-- )
            {
                if( anorm + fabs( e[l] ) == anorm )
                {
                    flag = 1;
                    break;
                }
                assert( l > 0 );
                if( anorm + fabs( w[l - 1] ) == anorm )
                    break;
            }

            if( !flag )
            {
                c = 0;
                s = 1;

                for( i = l; i <= k; i++ )
                {
                    double f = s * e[i];

                    e[i] *= c;

                    if( anorm + fabs( f ) == anorm )
                        break;

                    g = w[i];
                    h = pythag( f, g );
                    w[i] = h;
                    c = g / h;
                    s = -f / h;

                    if( u1 )
                    {
                        icvGivens( m1, u1 + ldu1 * (i - 1), u1 + ldu1 * i, c, s );
                    }
                }
            }

            z = w[k];
            if( l == k || iters++ == MAX_ITERS )
                break;

            /* shift from bottom 2x2 minor */
            x = w[l];
            y = w[k - 1];
            g = e[k - 1];
            h = e[k];
            f = 0.5 * (((g + z) / h) * ((g - z) / y) + y / h - h / y);
            g = pythag( f, 1 );
            if( f < 0 )
                g = -g;
            f = x - (z / x) * z + (h / x) * (y / (f + g) - h);
            /* next QR transformation */
            c = s = 1;

            for( i = l + 1; i <= k; i++ )
            {
                g = e[i];
                y = w[i];
                h = s * g;
                g *= c;
                z = pythag( f, h );
                e[i - 1] = z;
                c = f / z;
                s = h / z;
                f = x * c + g * s;
                g = -x * s + g * c;
                h = y * s;
                y *= c;

                if( v1 )
                {
                    icvGivens( n1, v1 + ldv1 * (i - 1), v1 + ldv1 * i, c, s );
                }

                z = pythag( f, h );
                w[i - 1] = z;

                /* rotation can be arbitrary if z == 0 */
                if( z != 0 )
                {
                    c = f / z;
                    s = h / z;
                }
                f = c * g + s * y;
                x = -s * g + c * y;

                if( u1 )
                {
                    icvGivens( m1, u1 + ldu1 * (i - 1), u1 + ldu1 * i, c, s );
                }
            }

            e[l] = 0;
            e[k] = f;
            w[k] = x;
        }                       /* end of iteration loop */

        if( iters > MAX_ITERS )
            break;

        if( z < 0 )
        {
            w[k] = -z;
            if( v )
            {
                for( j = 0; j < n; j++ )
                    v[j + k * ldv] = -v[j + k * ldv];
            }
        }
    }                           /* end of diagonalization loop */

    /* sort singular values */
    for( i = 0; i < nm; i++ )
    {
        k = i;
        for( j = i + 1; j < nm; j++ )
            if( w[k] < w[j] )
                k = j;

        if( k != i )
        {
            /* swap i & k values */
            double t = w[k];

            w[k] = w[i];
            w[i] = t;

            if( v )
            {
                for( j = 0; j < n; j++ )
                {
                    t = v[j + ldv * k];
                    v[j + ldv * k] = v[j + ldv * i];
                    v[j + ldv * i] = t;
                }
            }

            if( u )
            {
                for( j = 0; j < m; j++ )
                {
                    t = u[j + k * ldu];
                    u[j + ldu * k] = u[j + i * ldu];
                    u[j + ldu * i] = t;
                }
            }
        }
    }

#if 0
    memset( a0, 0, lda * m * sizeof( a0[0] ));

    k = m < n ? m : n;

    for( i = 0; i < k; i++ )
        a0[lda * i + i] = w[i];

    a0 += m >= n ? 1 : lda;

    for( i = 0; i < k - 1; i++ )
        a0[lda * i + i] = e[i + 1];
#endif

    return CV_NO_ERR;
}


#if 0
#include <stdio.h>

static void
icvSaveMatrix( FILE * f, const char *matrixName, const double *a, int lda, int m, int n )
{
    if( f )
    {
        int i, j;

        fprintf( f, "%s = [", matrixName );

        for( i = 0; i < m; i++ )
        {
            for( j = 0; j < n; j++ )
            {
                fprintf( f, "%30.20e%s", a[lda * i + j],
                         j < n - 1 ? ", " : i < m - 1 ? ";\n" : "];\n" );
            }
        }
    }
}
#else
#define icvSaveMatrix(f,matrixName,a,lda,m,n)
#endif


IPCVAPI_IMPL( CvStatus, icvPseudoInverse_64f, (int m, int n, double *a, int lda,
                                               double eps, double *temp_buf) )
{
    int i, j, min_mn = MIN( m, n );
    double *w = temp_buf;
    double *v = w + min_mn;
    double *u = v + n * n;
    double *test_a = 0;
    double *e = 0;

    //FILE* f = fopen("e:\\user\\vp\\log.txt", "wt");

    if( !a || !temp_buf )
        return CV_NULLPTR_ERR;
    if( m < 2 || n < 2 )
        return CV_BADSIZE_ERR;

    test_a = (double *) icvAlloc( (m * n + m * m) * sizeof( a[0] ));
    memcpy( test_a, a, m * n * sizeof( a[0] ));
    e = test_a + m * n;

    icvSaveMatrix( f, "a", a, lda, m, n );

    /* calculate SVD for a: u', v and w */
    icvSVD_64f( m, n, a, lda, w, u, m, v, n );

    icvSaveMatrix( f, "u", u, m, m, m );
    icvSaveMatrix( f, "v", v, n, n, n );
    icvSaveMatrix( f, "w", w, 1, n, 1 );

    //icvTransposeMatrix_64d( v, n, n, v );
    for( i = 0; i < n; i++ )
    {
        for( j = 0; j < i; j++ )
        {
            double t;

            CV_SWAP( v[i * n + j], v[j * n + i], t );
        }
    }

    icvSaveMatrix( f, "v2", v, n, n, n );

    /* calculate w^-1 */
    for( i = 0; i < min_mn; i++ )
    {
        double t = w[i];

        assert( t >= 0 );

        w[i] = t > eps ? 1. / t : 0;
    }

    /* calculate a^-1 = v*(w^-1)*u' */
    for( i = 0; i < min_mn; i++ )
    {
        icvScaleVector_64d( u + i * m, u + i * m, m, w[i] );
    }

    icvSetZero_64d( u + min_mn * m, m, n - min_mn );

    icvSaveMatrix( f, "b", u, m, n, m );

    icvMulMatrix_64d( v, n, n, u, m, n, a );

    icvSaveMatrix( f, "c", a, m, n, m );
    //fclose(f);

    /*
       icvMulMatrix_64d( test_a, n, m, a, m, n, e );
       {
       double s = 0;
       for( i = 0; i < m; i++ )
       {
       for( j = 0; j < m; j++ )
       {
       double t = fabs(e[i*m+j] - (i == j));
       s = MAX( s, t );
       }
       }
       }
     */

    icvFree( &test_a );

    return CV_OK;
}
