/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include "_cv.h"

typedef struct _CvUnDistortData
{
    int ind;
    ushort a0;
    ushort a1;
    ushort a2;
    ushort a3;
}
CvUnDistortData;

/*F//////////////////////////////////////////////////////////////////////////////////////
//    Names: icvUnDistortInit_8uC1R, icvUnDistortInit_8uC3R, 
//    Purpose: The functions calculate arrays of distorted points indices and
//             interpolation coefficients for cvUnDistort function
//    Context:
//    Parameters:  size        - size of each image or of its ROI
//                 step        - full width of each image in bytes
//                 intrMatrix  - matrix of the camera intrinsic parameters
//                 distCoeffs  - vector of the distortion coefficients (k1, k2, p1 and p2)
//                 interToggle - interpolation toggle
//                 data        - distortion data array
//
//    Returns: CV_NO_ERR or error code
//
//    Notes:  1. If interToggle=0, interpolation disabled;
//               else bilinear interpolation is used.
//            2. Array data must be allocated before. If interToggle = 0, its length
//               must be size.width*size.height elements; else 3*size.width*size.height.
//F*/
/*______________________________________________________________________________________*/

IPCVAPI_IMPL( CvStatus, icvUnDistortInit_8u_C1R, (CvSize size,
                                                  int step,
                                                  float *intrMatrix,
                                                  float *distCoeffs,
                                                  int interToggle, int *data) )
{
    const float a1 = 1.f / intrMatrix[0], b1 = 1.f / intrMatrix[4],
        u0 = intrMatrix[2], v0 = intrMatrix[5],
        k1 = distCoeffs[0], k2 = distCoeffs[1], p1 = distCoeffs[2], p2 = distCoeffs[3];
    int u, v;
    float p22 = 2.f * p2;

    if( size.width < 1 || size.height < 1 )
        return CV_BADSIZE_ERR;
    if( size.width > step )
        return CV_BADSIZE_ERR;
    if( intrMatrix == NULL || distCoeffs == NULL || data == NULL )
        return CV_NULLPTR_ERR;

    if( !interToggle )
    {
        for( v = 0; v < size.height; v++, data += size.width )
        {
            float dv = v - v0;
            float y = b1 * dv;
            float y1 = p1 / y;
            float y2 = y * y;
            float y3 = 2.f * p1 * y;

            for( u = 0; u < size.width; u++ )
            {
                float du = u - u0;
                float x = a1 * du;
                float x1 = p2 / x;
                float x2 = x * x;
                float x3 = p22 * x;
                float r2 = x2 + y2;
                float bx = r2 * (k1 + r2 * k2) + x3 + y3;
                float by = bx + r2 * y1;
                int ud = u, vd = v;

                bx += r2 * x1;
                ud += cvRound( bx * du );
                vd += cvRound( by * dv );
                data[u] = ud < 0 || ud >= size.width || vd < 0 || vd >= size.height ?
                    0 : vd * step + ud;
            }
        }
    }
    else                        /* interpolation */
    {
        const int sizex = size.width - 2, sizey = size.height - 2;
        CvUnDistortData *uData = (CvUnDistortData *) data;

#ifndef _CVL_A6
        const float s15 = 65536.f;
        const int bmax = 65535;
#else
        const float s15 = 32768.f;
        const int bmax = 32767;
#endif

        for( v = 0; v < size.height; v++, uData += size.width )
        {
            float dv = v - v0;
            float y = b1 * dv;
            float y1 = p1 / y;
            float y2 = y * y;
            float y3 = 2.f * p1 * y;

            for( u = 0; u < size.width; u++ )
            {
                float du = u - u0;
                float x = a1 * du;
                float x1 = p2 / x;
                float x2 = x * x;
                float x3 = p22 * x;
                float r2 = x2 + y2;
                float bx = r2 * (k1 + r2 * k2) + x3 + y3;
                float by = bx + r2 * y1;
                float uf = (float) u, vf = (float) v;
                int ud, vd;

                bx += r2 * x1;
                uf += bx * du;
                vf += by * dv;
                ud = cvFloor( uf );
                vd = cvFloor( vf );
                if( ud < 0 || ud > sizex || vd < 0 || vd > sizey )
                {
                    (uData + u)->ind = 0;
                    
                        (uData + u)->a0 = (uData + u)->a1 = (uData + u)->a2 = (uData + u)->a3 =
                        (ushort) 0;
                }
                else
                {
                    float uf1, vf1;
                    int b0, b1, b2, b3;

                    (uData + u)->ind = vd * step + ud;
                    uf -= (float) ud;
                    vf -= (float) vd;
                    uf1 = 1.f - uf;
                    vf1 = 1.f - vf;
                    b0 = (int) (s15 * uf1 * vf1);
                    b1 = (int) (s15 * uf * vf1);
                    b2 = (int) (s15 * uf1 * vf);
                    b3 = (int) (s15 * uf * vf);
                    if( b0 < 0 )
                        b0 = 0;
                    if( b1 < 0 )
                        b1 = 0;
                    if( b2 < 0 )
                        b2 = 0;
                    if( b3 < 0 )
                        b3 = 0;
                    if( b0 > bmax )
                        b0 = bmax;
                    if( b1 > bmax )
                        b1 = bmax;
                    if( b2 > bmax )
                        b2 = bmax;
                    if( b3 > bmax )
                        b3 = bmax;
                    (uData + u)->a0 = (ushort) b0;
                    (uData + u)->a1 = (ushort) b1;
                    (uData + u)->a2 = (ushort) b2;
                    (uData + u)->a3 = (ushort) b3;
                }
            }                   /* u */
        }                       /* v */
    }                           /* else */

    return CV_NO_ERR;
}
/*______________________________________________________________________________________*/

IPCVAPI_IMPL( CvStatus, icvUnDistortInit_8u_C3R, (CvSize size,
                                                  int step,
                                                  float *intrMatrix,
                                                  float *distCoeffs,
                                                  int interToggle, int *data) )
{
    const float a1 = 1.f / intrMatrix[0], b1 = 1.f / intrMatrix[4],
        u0 = intrMatrix[2], v0 = intrMatrix[5],
        k1 = distCoeffs[0], k2 = distCoeffs[1], p1 = distCoeffs[2], p2 = distCoeffs[3];
    int u, v;
    float p22 = 2.f * p2;

    if( size.width < 1 || size.height < 1 )
        return CV_BADSIZE_ERR;
    if( size.width > step )
        return CV_BADSIZE_ERR;
    if( intrMatrix == NULL || distCoeffs == NULL || data == NULL )
        return CV_NULLPTR_ERR;

    if( !interToggle )
    {
        for( v = 0; v < size.height; v++, data += size.width )
        {
            float dv = v - v0;
            float y = b1 * dv;
            float y1 = p1 / y;
            float y2 = y * y;
            float y3 = 2.f * p1 * y;

            for( u = 0; u < size.width; u++ )
            {
                float du = u - u0;
                float x = a1 * du;
                float x1 = p2 / x;
                float x2 = x * x;
                float x3 = p22 * x;
                float r2 = x2 + y2;
                float bx = r2 * (k1 + r2 * k2) + x3 + y3;
                float by = bx + r2 * y1;
                int ud = u, vd = v;

                bx += r2 * x1;
                ud += cvRound( bx * du );
                vd += cvRound( by * dv );
                data[u] = ud < 0 || ud >= size.width || vd < 0 || vd >= size.height ?
                    0 : vd * step + 3 * ud;
            }
        }
    }
    else                        /* interpolation */
    {
        const int sizex = size.width - 2, sizey = size.height - 2;
        CvUnDistortData *uData = (CvUnDistortData *) data;

#ifndef _CVL_A6
        const float s15 = 65536.f;
        const int bmax = 65535;
#else
        const float s15 = 32768.f;
        const int bmax = 32767;
#endif

        for( v = 0; v < size.height; v++, uData += size.width )
        {
            float dv = v - v0;
            float y = b1 * dv;
            float y1 = p1 / y;
            float y2 = y * y;
            float y3 = 2.f * p1 * y;

            for( u = 0; u < size.width; u++ )
            {
                float du = u - u0;
                float x = a1 * du;
                float x1 = p2 / x;
                float x2 = x * x;
                float x3 = p22 * x;
                float r2 = x2 + y2;
                float bx = r2 * (k1 + r2 * k2) + x3 + y3;
                float by = bx + r2 * y1;
                float uf = (float) u, vf = (float) v;
                int ud, vd;

                bx += r2 * x1;
                uf += bx * du;
                vf += by * dv;
                ud = cvFloor( uf );
                vd = cvFloor( vf );
                if( ud < 0 || ud > sizex || vd < 0 || vd > sizey )
                {
                    (uData + u)->ind = 0;
                    
                        (uData + u)->a0 = (uData + u)->a1 = (uData + u)->a2 = (uData + u)->a3 =
                        (ushort) 0;
                }
                else
                {
                    float uf1, vf1;
                    int b0, b1, b2, b3;

                    (uData + u)->ind = vd * step + 3 * ud;
                    uf -= (float) ud;
                    vf -= (float) vd;
                    uf1 = 1.f - uf;
                    vf1 = 1.f - vf;
                    b0 = (int) (s15 * uf1 * vf1);
                    b1 = (int) (s15 * uf * vf1);
                    b2 = (int) (s15 * uf1 * vf);
                    b3 = (int) (s15 * uf * vf);
                    if( b0 < 0 )
                        b0 = 0;
                    if( b1 < 0 )
                        b1 = 0;
                    if( b2 < 0 )
                        b2 = 0;
                    if( b3 < 0 )
                        b3 = 0;
                    if( b0 > bmax )
                        b0 = bmax;
                    if( b1 > bmax )
                        b1 = bmax;
                    if( b2 > bmax )
                        b2 = bmax;
                    if( b3 > bmax )
                        b3 = bmax;
                    (uData + u)->a0 = (ushort) b0;
                    (uData + u)->a1 = (ushort) b1;
                    (uData + u)->a2 = (ushort) b2;
                    (uData + u)->a3 = (ushort) b3;
                }
            }                   /* u */
        }                       /* v */
    }                           /* else */

    return CV_NO_ERR;
}

/*======================================================================================*/

/*F//////////////////////////////////////////////////////////////////////////////////////
//    Names: icvUnDistort_8u_C1R, icvUnDistort_8u_C3R
//    Purpose: The functions correct radial and tangential distortion in the frame
//             using previousely calculated arrays of distorted points indices and
//             undistortion coefficients
//    Context:
//    Parameters:  srcImage    - source (distorted) image
//                 dstImage    - output (undistorted) image
//                 step        - full width of each image in bytes
//                 size        - size of each image or of its ROI
//                 interToggle - interpolation toggle
//                 data        - distortion data array
//
//    Returns: CV_NO_ERR or error code
//
//    Notes:  1. Either icvUnDistortInit_8u_C1R or icvUnDistortInit_8u_C3R function
//               must be used previously
//            2. See Notes to the icvUnDistortInit_8u_C1R, icvUnDistortInit_8u_C3R
//               functions
//F*/
/*______________________________________________________________________________________*/

IPCVAPI_IMPL( CvStatus, icvUnDistort_8u_C1R, (uchar * srcImage,
                                              uchar * dstImage,
                                              int step,
                                              CvSize size, int interToggle, int *data) )
{
    int u, v;
    uchar buf = *srcImage;

    if( step < 1 )
        return CV_BADFACTOR_ERR;
    if( size.width > step || size.width < 1 || size.height < 1 )
        return CV_BADSIZE_ERR;
    if( srcImage == NULL || dstImage == NULL || data == NULL )
        return CV_NULLPTR_ERR;

    *srcImage = (uchar) 0;

    if( step == size.width )
    {
        size.width *= size.height;
        size.height = 1;
    }

    if( !interToggle )          /* Interpolation disabled */

        for( v = 0; v < size.height; v++, dstImage += step, data += size.width )
        {
            for( u = 0; u < size.width - 15; u += 16 )
            {
                dstImage[u] = srcImage[data[u]];
                dstImage[u + 1] = srcImage[data[u + 1]];
                dstImage[u + 2] = srcImage[data[u + 2]];
                dstImage[u + 3] = srcImage[data[u + 3]];
                dstImage[u + 4] = srcImage[data[u + 4]];
                dstImage[u + 5] = srcImage[data[u + 5]];
                dstImage[u + 6] = srcImage[data[u + 6]];
                dstImage[u + 7] = srcImage[data[u + 7]];
                dstImage[u + 8] = srcImage[data[u + 8]];
                dstImage[u + 9] = srcImage[data[u + 9]];
                dstImage[u + 10] = srcImage[data[u + 10]];
                dstImage[u + 11] = srcImage[data[u + 11]];
                dstImage[u + 12] = srcImage[data[u + 12]];
                dstImage[u + 13] = srcImage[data[u + 13]];
                dstImage[u + 14] = srcImage[data[u + 14]];
                dstImage[u + 15] = srcImage[data[u + 15]];
            }

            for( ; u < size.width; u++ )
            {
                dstImage[u] = srcImage[data[u]];
            }
        }

    else
        /* Interpolation enabled */
    {
        CvUnDistortData *uData = (CvUnDistortData *) data;

        for( v = 0; v < size.height; v++, dstImage += step, uData += size.width )
        {
            for( u = 0; u < size.width - 3; u += 4 )
            {
                CvUnDistortData d0 = uData[u];
                CvUnDistortData d1 = uData[u + 1];
                CvUnDistortData d2 = uData[u + 2];
                CvUnDistortData d3 = uData[u + 3];
                int v0 = srcImage[d0.ind] * d0.a0;
                int v1 = srcImage[d1.ind] * d1.a0;
                int v2 = srcImage[d2.ind] * d2.a0;
                int v3 = srcImage[d3.ind] * d3.a0;
                uchar w[4];

                v0 += srcImage[d0.ind + 1] * d0.a1;
                v1 += srcImage[d1.ind + 1] * d1.a1;
                v2 += srcImage[d2.ind + 1] * d2.a1;
                v3 += srcImage[d3.ind + 1] * d3.a1;

                v0 += srcImage[d0.ind + step] * d0.a2;
                v1 += srcImage[d1.ind + step] * d1.a2;
                v2 += srcImage[d2.ind + step] * d2.a2;
                v3 += srcImage[d3.ind + step] * d3.a2;

                v0 += srcImage[d0.ind + step + 1] * d0.a3;
                v1 += srcImage[d1.ind + step + 1] * d1.a3;
                v2 += srcImage[d2.ind + step + 1] * d2.a3;
                v3 += srcImage[d3.ind + step + 1] * d3.a3;

                w[0] = (uchar) (v0 >> 16);
                w[1] = (uchar) (v1 >> 16);
                w[2] = (uchar) (v2 >> 16);
                w[3] = (uchar) (v3 >> 16);

                *(int *) (dstImage + u) = *(int *) w;
            }

            for( ; u < size.width; u++ )
            {
                CvUnDistortData d = uData[u];
                int v0 = srcImage[d.ind];
                int v1 = srcImage[d.ind + 1];
                int v2 = srcImage[d.ind + step];
                int v3 = srcImage[d.ind + step + 1];

                dstImage[u] = (uchar) ((v0 * d.a0 + v1 * d.a1 + v2 * d.a2 + v3 * d.a3) >> 16);
            }
        }
    }

    *srcImage = buf;
    return CV_NO_ERR;
}
/*______________________________________________________ 3-CHANNEL IMAGES ______________*/

IPCVAPI_IMPL( CvStatus, icvUnDistort_8u_C3R, (uchar * srcImage,
                                              uchar * dstImage,
                                              int step,
                                              CvSize size, int interToggle, int *data) )
{
    int u, u3, v;
    uchar buf1 = *srcImage, buf2 = srcImage[1], buf3 = srcImage[2];

    if( step < 1 )
        return CV_BADFACTOR_ERR;
    if( size.width > step || size.width < 1 || size.height < 1 )
        return CV_BADSIZE_ERR;
    if( srcImage == NULL || dstImage == NULL || data == NULL )
        return CV_NULLPTR_ERR;

    *srcImage = (uchar) 0;
    srcImage[1] = (uchar) 0;
    srcImage[2] = (uchar) 0;

    if( step == 3 * size.width )
    {
        size.width *= size.height;
        size.height = 1;
    }

    if( !interToggle )          /* Interpolation disabled */

        for( v = 0; v < size.height; v++, dstImage += step, data += size.width )
        {

            if( v == size.height - 1 )
                size.width -= 8;

            for( u = 0, u3 = 0; u < size.width - 7; u += 8, u3 += 24 )
            {
                *(int *) (dstImage + u3) = *(int *) (srcImage + data[u]);
                *(int *) (dstImage + u3 + 3) = *(int *) (srcImage + data[u + 1]);
                *(int *) (dstImage + u3 + 6) = *(int *) (srcImage + data[u + 2]);
                *(int *) (dstImage + u3 + 9) = *(int *) (srcImage + data[u + 3]);
                *(int *) (dstImage + u3 + 12) = *(int *) (srcImage + data[u + 4]);
                *(int *) (dstImage + u3 + 15) = *(int *) (srcImage + data[u + 5]);
                *(int *) (dstImage + u3 + 18) = *(int *) (srcImage + data[u + 6]);
                *(int *) (dstImage + u3 + 21) = *(int *) (srcImage + data[u + 7]);

            }

            if( v == size.height - 1 )
                size.width += 8;

            for( ; u < size.width; u++, u3 += 3 )
            {
                int m = data[u];

                dstImage[u3] = srcImage[m];
                dstImage[u3 + 1] = srcImage[m + 1];
                dstImage[u3 + 2] = srcImage[m + 2];
            }
        }

    else
        /* Interpolation enabled */
    {
        CvUnDistortData *uData = (CvUnDistortData *) data;

        for( v = 0; v < size.height; v++, dstImage += step, uData += size.width )
        {
            if( v == size.height - 1 )
                size.width -= 4;

            for( u = 0, u3 = 0; u < size.width - 3; u += 4, u3 += 12 )
            {
                CvUnDistortData d0 = uData[u];
                CvUnDistortData d1 = uData[u + 1];
                CvUnDistortData d2 = uData[u + 2];
                CvUnDistortData d3 = uData[u + 3];
                int v0 = srcImage[d0.ind] * d0.a0;
                int v1 = srcImage[d0.ind + 1] * d0.a0;
                int v2 = srcImage[d0.ind + 2] * d0.a0;
                uchar w[4];

                v0 += srcImage[d0.ind + 3] * d0.a1;
                v1 += srcImage[d0.ind + 4] * d0.a1;
                v2 += srcImage[d0.ind + 5] * d0.a1;

                v0 += srcImage[d0.ind + step] * d0.a2;
                v1 += srcImage[d0.ind + step + 1] * d0.a2;
                v2 += srcImage[d0.ind + step + 2] * d0.a2;

                v0 += srcImage[d0.ind + step + 3] * d0.a3;
                v1 += srcImage[d0.ind + step + 4] * d0.a3;
                v2 += srcImage[d0.ind + step + 5] * d0.a3;

                w[0] = (uchar) (v0 >> 16);
                w[1] = (uchar) (v1 >> 16);
                w[2] = (uchar) (v2 >> 16);

                *(int *) (dstImage + u3) = *(int *) w;

                v0 = srcImage[d1.ind] * d1.a0;
                v1 = srcImage[d1.ind + 1] * d1.a0;
                v2 = srcImage[d1.ind + 2] * d1.a0;

                v0 += srcImage[d1.ind + 3] * d1.a1;
                v1 += srcImage[d1.ind + 4] * d1.a1;
                v2 += srcImage[d1.ind + 5] * d1.a1;

                v0 += srcImage[d1.ind + step] * d1.a2;
                v1 += srcImage[d1.ind + step + 1] * d1.a2;
                v2 += srcImage[d1.ind + step + 2] * d1.a2;

                v0 += srcImage[d1.ind + step + 3] * d1.a3;
                v1 += srcImage[d1.ind + step + 4] * d1.a3;
                v2 += srcImage[d1.ind + step + 5] * d1.a3;

                w[0] = (uchar) (v0 >> 16);
                w[1] = (uchar) (v1 >> 16);
                w[2] = (uchar) (v2 >> 16);

                *(int *) (dstImage + u3 + 3) = *(int *) w;

                v0 = srcImage[d2.ind] * d2.a0;
                v1 = srcImage[d2.ind + 1] * d2.a0;
                v2 = srcImage[d2.ind + 2] * d2.a0;

                v0 += srcImage[d2.ind + 3] * d2.a1;
                v1 += srcImage[d2.ind + 4] * d2.a1;
                v2 += srcImage[d2.ind + 5] * d2.a1;

                v0 += srcImage[d2.ind + step] * d2.a2;
                v1 += srcImage[d2.ind + step + 1] * d2.a2;
                v2 += srcImage[d2.ind + step + 2] * d2.a2;

                v0 += srcImage[d2.ind + step + 3] * d2.a3;
                v1 += srcImage[d2.ind + step + 4] * d2.a3;
                v2 += srcImage[d2.ind + step + 5] * d2.a3;

                w[0] = (uchar) (v0 >> 16);
                w[1] = (uchar) (v1 >> 16);
                w[2] = (uchar) (v2 >> 16);

                *(int *) (dstImage + u3 + 6) = *(int *) w;

                v0 = srcImage[d3.ind] * d3.a0;
                v1 = srcImage[d3.ind + 1] * d3.a0;
                v2 = srcImage[d3.ind + 2] * d3.a0;

                v0 += srcImage[d3.ind + 3] * d3.a1;
                v1 += srcImage[d3.ind + 4] * d3.a1;
                v2 += srcImage[d3.ind + 5] * d3.a1;

                v0 += srcImage[d3.ind + step] * d3.a2;
                v1 += srcImage[d3.ind + step + 1] * d3.a2;
                v2 += srcImage[d3.ind + step + 2] * d3.a2;

                v0 += srcImage[d3.ind + step + 3] * d3.a3;
                v1 += srcImage[d3.ind + step + 4] * d3.a3;
                v2 += srcImage[d3.ind + step + 5] * d3.a3;

                w[0] = (uchar) (v0 >> 16);
                w[1] = (uchar) (v1 >> 16);
                w[2] = (uchar) (v2 >> 16);

                *(int *) (dstImage + u3 + 9) = *(int *) w;

            }

            if( v == size.height - 1 )
                size.width += 4;

            for( ; u < size.width; u++, u3 += 3 )
            {
                CvUnDistortData d = uData[u];
                int v0 = srcImage[d.ind];
                int v1 = srcImage[d.ind + 3];
                int v2 = srcImage[d.ind + step];
                int v3 = srcImage[d.ind + step + 3];

                dstImage[u3] = (uchar) ((v0 * d.a0 + v1 * d.a1 + v2 * d.a2 + v3 * d.a3) >> 16);
                v0 = srcImage[d.ind + 1];
                v1 = srcImage[d.ind + 4];
                v2 = srcImage[d.ind + step + 1];
                v3 = srcImage[d.ind + step + 4];
                dstImage[u3 + 1] =
                    (uchar) ((v0 * d.a0 + v1 * d.a1 + v2 * d.a2 + v3 * d.a3) >> 16);
                v0 = srcImage[d.ind + 2];
                v1 = srcImage[d.ind + 5];
                v2 = srcImage[d.ind + step + 2];
                v3 = srcImage[d.ind + step + 5];
                dstImage[u3 + 2] =
                    (uchar) ((v0 * d.a0 + v1 * d.a1 + v2 * d.a2 + v3 * d.a3) >> 16);
            }
        }
    }

    *srcImage = buf1;
    srcImage[1] = buf2;
    srcImage[2] = buf3;
    return CV_NO_ERR;
}

/*======================================================================================*/

/*F//////////////////////////////////////////////////////////////////////////////////////
//    Names: icvUnDistort1_8uC1R, icvUnDistort1_8uC3R
//    Purpose: The functions correct radial image distortion using known matrix of the
//             camera intrinsic parameters and distortion coefficients
//    Context:
//    Parameters:  srcImage    - source (distorted) image
//                 dstImage    - output (undistorted) image
//                 step        - full width of each image in bytes
//                 size        - ROI size of each image
//                 intrMatrix  - matrix of the camera intrinsic parameters
//                 distCoeffs  - vector of two distortion coefficients (k1 and k2)
//                 interToggle - interpolation toggle
//
//    Returns: CV_NO_ERR or error code
//
//    Notes:   If interToggle=0, interpolation disabled;
//             else bilinear interpolation is used.
//F*/
/*______________________________________________________________________________________*/

#define S  22
#define S2 11
#define FM (float)0x400000

IPCVAPI_IMPL( CvStatus, icvUnDistort1_8u_C1R, (uchar * srcImage,
                                               uchar * dstImage,
                                               int step,
                                               CvSize size,
                                               float *intrMatrix,
                                               float *distCoeffs, int interToggle) )
{
    const float fm = FM;
    const float a1 = 1.f / intrMatrix[0], b1 = 1.f / intrMatrix[4],
        u0 = intrMatrix[2], v0 = intrMatrix[5], k1 = distCoeffs[0], k2 = distCoeffs[1];
    float *x2;
    float *du;
    int u, v;
    uchar buf = *srcImage;

    if( step < 1 )
        return CV_BADFACTOR_ERR;
    if( size.width > step || size.width < 1 || size.height < 1 )
        return CV_BADSIZE_ERR;
    if( srcImage == NULL || dstImage == NULL || intrMatrix == NULL || distCoeffs == NULL )
        return CV_NULLPTR_ERR;

    x2 = (float *) icvAlloc( sizeof( float ) * size.width );

    if( x2 == NULL )
        return CV_OUTOFMEM_ERR;
    du = (float *) icvAlloc( sizeof( float ) * size.width );

    if( du == NULL )
    {
        icvFree( (void **) &x2 );
        return CV_OUTOFMEM_ERR;
    }

    *srcImage = (uchar) 0;

    if( !interToggle )
    {
        for( u = 0; u < size.width; u++ )
        {
            float w = u - u0;
            float x = a1 * w;

            x2[u] = x * x;
            du[u] = w;
        }
        for( v = 0; v < size.height; v++, dstImage += step )
        {
            float y = b1 * (v - v0);
            float y2 = y * y;
            float dv = v - v0;

            for( u = 0; u < size.width; u++ )
            {
                float r2 = x2[u] + y2;
                float dist = r2 * (k1 + r2 * k2);
                int ud = u + cvRound( du[u] * dist );
                int vd = v + cvRound( dv * dist );

                dstImage[u] = (uchar) (ud < 0 || ud >= size.width || vd < 0 ||
                                       vd >= size.height ? 0 : srcImage[vd * step + ud]);
            }
        }
    }
    else
    {
        int sizex = size.width - 2, sizey = size.height - 2;

        for( u = 0; u < size.width; u++ )
        {
            float w = u - u0;
            float x = a1 * w;

            x2[u] = x * x;
            du[u] = fm * w;
        }
        for( v = 0; v < size.height; v++, dstImage += step )
        {
            float y = b1 * (v - v0);
            float y2 = y * y;
            float dv = fm * (v - v0);

            for( u = 0; u < size.width; u++ )
            {
                float r2 = x2[u] + y2;
                float dist = r2 * (k1 + r2 * k2);
                int iu = cvRound( du[u] * dist );
                int iv = cvRound( dv * dist );
                int ud = iu >> S;
                int vd = iv >> S;

                iu -= ud << S;
                iv -= vd << S;
                ud += u;
                vd += v;

                if( ud < 0 || ud > sizex || vd < 0 || vd > sizey )
                    dstImage[u] = 0;
                else
                {
                    int iuv = (iu >> S2) * (iv >> S2);
                    int uv = vd * step + ud;
                    int a0 = srcImage[uv];
                    int a = srcImage[uv + 1] - a0;
                    int b = srcImage[uv + step] - a0;
                    int c = srcImage[uv + step + 1] - a0 - a - b;
                    int d = ((a * iu + b * iv + c * iuv) >> S) + a0;

                    d = !(d & ~255) ? d : d < 0 ? 0 : 255;
                    dstImage[u] = (uchar) d;
                }
            }
        }
    }

    *srcImage = buf;
    icvFree( (void **) &x2 );
    icvFree( (void **) &du );
    return CV_NO_ERR;
}
/*______________________________________________________ 3-CHANNEL IMAGES ______________*/

IPCVAPI_IMPL( CvStatus, icvUnDistort1_8u_C3R, (uchar * srcImage,
                                               uchar * dstImage,
                                               int step,
                                               CvSize size,
                                               float *intrMatrix,
                                               float *distCoeffs, int interToggle) )
{
    const float fm = FM;
    const float a1 = 1.f / intrMatrix[0], b1 = 1.f / intrMatrix[4],
        u0 = intrMatrix[2], v0 = intrMatrix[5], k1 = distCoeffs[0], k2 = distCoeffs[1];
    const int p = 0xFFFFFF;
    float *x2;
    float *du;
    int u, v;
    uchar buf0 = *srcImage, buf1 = srcImage[1], buf2 = srcImage[2];

    if( step < 1 )
        return CV_BADFACTOR_ERR;
    if( size.width > step || size.width < 1 || size.height < 1 )
        return CV_BADSIZE_ERR;
    if( srcImage == NULL || dstImage == NULL || intrMatrix == NULL || distCoeffs == NULL )
        return CV_NULLPTR_ERR;

    x2 = (float *) icvAlloc( sizeof( float ) * size.width );

    if( x2 == NULL )
        return CV_OUTOFMEM_ERR;
    du = (float *) icvAlloc( sizeof( float ) * size.width );

    if( du == NULL )
    {
        icvFree( (void **) &x2 );
        return CV_OUTOFMEM_ERR;
    }

    srcImage[0] = srcImage[1] = srcImage[2] = 0;

    if( !interToggle )
    {
        for( u = 0; u < size.width; u++ )
        {
            float w = u - u0;
            float x = a1 * w;

            x2[u] = x * x;
            du[u] = w;
        }

        for( v = 0; v < size.height; v++, dstImage += step )
        {
            float y = b1 * (v - v0);
            float y2 = y * y;
            float dv = v - v0;

            if( v == size.height - 1 )
                size.width--;
            for( u = 0; u < size.width; u++ )
            {
                float r2 = x2[u] + y2;
                float dist = r2 * (k1 + r2 * k2);
                int ud = u + cvRound( du[u] * dist );
                int vd = v + cvRound( dv * dist );
                int u3 = u + u + u;

                if( ud < 0 || ud >= size.width || vd < 0 || vd >= size.height )
                    *(int *) (dstImage + u3) = 0;
                else
                {
                    int uv = vd * step + ud + ud + ud;

                    *(int *) (dstImage + u3) = *(int *) (srcImage + uv) & p;
                }
            }
            if( v == size.height - 1 )
            {
                float r2 = x2[u] + y2;
                float dist = r2 * (k1 + r2 * k2);
                int ud = u + cvRound( du[u] * dist );
                int vd = v + cvRound( dv * dist );
                int u3 = u + u + u;

                size.width++;
                if( ud < 0 || ud >= size.width || vd < 0 || vd >= size.height )
                    dstImage[u3] = dstImage[u3 + 1] = dstImage[u3 + 2] = 0;
                else
                {
                    int uv = vd * step + ud + ud + ud;

                    dstImage[u3] = srcImage[uv];
                    dstImage[u3 + 1] = srcImage[uv + 1];
                    dstImage[u3 + 2] = srcImage[uv + 2];
                }
            }
        }
    }
    else                        /* interpolation */
    {
        int sizex = size.width - 2, sizey = size.height - 2;

        for( u = 0; u < size.width; u++ )
        {
            float w = u - u0;
            float x = a1 * w;

            x2[u] = x * x;
            du[u] = fm * w;
        }

        for( v = 0; v < size.height; v++, dstImage += step )
        {
            float y = b1 * (v - v0);
            float y2 = y * y;
            float dv = fm * (v - v0);

            if( v == size.height - 1 )
                size.width--;
            for( u = 0; u < size.width; u++ )
            {
                float r2 = x2[u] + y2;
                float dist = r2 * (k1 + r2 * k2);
                int iu = cvRound( du[u] * dist );
                int iv = cvRound( dv * dist );
                int ud = iu >> S;
                int vd = iv >> S;
                int u3 = u + u + u;

                iu -= ud << S;
                iv -= vd << S;
                ud += u;
                vd += v;

                if( ud < 0 || ud > sizex || vd < 0 || vd > sizey )
                    *(int *) (dstImage + u3) = 0;
                else
                {
                    int iuv = (iu >> S2) * (iv >> S2);
                    int uv = vd * step + ud + ud + ud;
                    int uvs = uv + step;
                    int a01 = srcImage[uv];
                    int a02 = srcImage[uv + 1];
                    int a03 = srcImage[uv + 2];
                    int a1 = srcImage[uv + 3];
                    int a2 = srcImage[uv + 4];
                    int a3 = srcImage[uv + 5];
                    int b1 = srcImage[uvs];
                    int b2 = srcImage[uvs + 1];
                    int b3 = srcImage[uvs + 2];
                    int c1 = srcImage[uvs + 3];
                    int c2 = srcImage[uvs + 4];
                    int c3 = srcImage[uvs + 5];
                    int d1, d2, d3;

                    a1 -= a01;
                    a2 -= a02;
                    a3 -= a03;
                    b1 -= a01;
                    b2 -= a02;
                    b3 -= a03;
                    c1 -= a01 + a1 + b1;
                    c2 -= a02 + a2 + b2;
                    c3 -= a03 + a3 + b3;
                    d1 = ((a1 * iu + b1 * iv + c1 * iuv) >> S) + a01;
                    d2 = ((a2 * iu + b2 * iv + c2 * iuv) >> S) + a02;
                    d3 = ((a3 * iu + b3 * iv + c3 * iuv) >> S) + a03;
                    d1 = !(d1 & ~255) ? d1 : d1 < 0 ? 0 : 255;
                    d2 = !(d2 & ~255) ? d2 : d2 < 0 ? 0 : 255;
                    d3 = !(d3 & ~255) ? d3 : d3 < 0 ? 0 : 255;
                    d1 |= (d2 << 8) | (d3 << 16);
                    *(int *) (dstImage + u3) = d1;
                }
            }
            if( v == size.height - 1 )
            {
                float r2 = x2[u] + y2;
                float dist = r2 * (k1 + r2 * k2);
                int iu = cvRound( du[u] * dist );
                int iv = cvRound( dv * dist );
                int ud = iu >> S;
                int vd = iv >> S;
                int u3 = u + u + u;

                size.width++;
                iu -= ud << S;
                iv -= vd << S;
                ud += u;
                vd += v;

                if( ud < 0 || ud > sizex || vd < 0 || vd > sizey )
                    dstImage[u3] = dstImage[u3 + 1] = dstImage[u3 + 2] = 0;
                else
                {
                    int iuv = (iu >> S2) * (iv >> S2);
                    int uv = vd * step + ud + ud + ud;
                    int uvs = uv + step;
                    int a01 = srcImage[uv];
                    int a02 = srcImage[uv + 1];
                    int a03 = srcImage[uv + 2];
                    int a1 = srcImage[uv + 3];
                    int a2 = srcImage[uv + 4];
                    int a3 = srcImage[uv + 5];
                    int b1 = srcImage[uvs];
                    int b2 = srcImage[uvs + 1];
                    int b3 = srcImage[uvs + 2];
                    int c1 = srcImage[uvs + 3];
                    int c2 = srcImage[uvs + 4];
                    int c3 = srcImage[uvs + 5];
                    int d1, d2, d3;

                    a1 -= a01;
                    a2 -= a02;
                    a3 -= a03;
                    b1 -= a01;
                    b2 -= a02;
                    b3 -= a03;
                    c1 -= a01 + a1 + b1;
                    c2 -= a02 + a2 + b2;
                    c3 -= a03 + a3 + b3;
                    d1 = ((a1 * iu + b1 * iv + c1 * iuv) >> S) + a01;
                    d2 = ((a2 * iu + b2 * iv + c2 * iuv) >> S) + a02;
                    d3 = ((a3 * iu + b3 * iv + c3 * iuv) >> S) + a03;
                    dstImage[u3] = (uchar) d1;
                    dstImage[u3 + 1] = (uchar) d2;
                    dstImage[u3 + 2] = (uchar) d3;
                }
            }
        }
    }

    *srcImage   = buf0;
    srcImage[1] = buf1;
    srcImage[2] = buf2;
    icvFree( (void **) &x2 );
    icvFree( (void **) &du );
    return CV_NO_ERR;
}

/*======================================================================================*/

/*F//////////////////////////////////////////////////////////////////////////////////////
//    Names: icvUnDistortEx_8uC1R, icvUnDistortEx_8uC3R
//    Purpose: The functions correct radial and tangential image distortions using known
//             matrix of the camera intrinsic parameters and distortion coefficients
//    Context:
//    Parameters:  srcImage    - source (distorted) image
//                 dstImage    - output (undistorted) image
//                 step        - full width of each image in bytes
//                 size        - ROI size of each image
//                 intrMatrix  - matrix of the camera intrinsic parameters
//                 distCoeffs  - vector of the distortion coefficients
//                               (k1, k2, p1, p2)
//                 interToggle - interpolation toggle
//
//    Returns: CV_NO_ERR or error code
//
//    Notes:   If interToggle=0, interpolation disabled;
//             else bilinear interpolation is used.
//F*/
/*______________________________________________________________________________________*/

IPCVAPI_IMPL( CvStatus, icvUnDistortEx_8u_C1R, (uchar * srcImage,
                                                uchar * dstImage,
                                                int step,
                                                CvSize size,
                                                float *intrMatrix,
                                                float *distCoeffs, int interToggle) )
{
    const float fm = FM;
    const float a1 = 1.f / intrMatrix[0], b1 = 1.f / intrMatrix[4],
        u0 = intrMatrix[2], v0 = intrMatrix[5],
        k1 = distCoeffs[0], k2 = distCoeffs[1], p1 = distCoeffs[2], p2 = distCoeffs[3];
    float *x1;
    float *x2;
    float *x3;
    float *du;
    int u, v;
    uchar buf = *srcImage;

    if( step < 1 )
        return CV_BADFACTOR_ERR;
    if( size.width > step || size.width < 1 || size.height < 1 )
        return CV_BADSIZE_ERR;
    if( srcImage == NULL || dstImage == NULL || intrMatrix == NULL || distCoeffs == NULL )
        return CV_NULLPTR_ERR;

    /*if ( !p1 && !p2 ) return icvUnDistort1_8uC1R ( srcImage, dstImage, step, size,
       intrMatrix, distCoeffs, interToggle ); */

    x1 = (float *) icvAlloc( sizeof( float ) * size.width );
    x2 = (float *) icvAlloc( sizeof( float ) * size.width );
    x3 = (float *) icvAlloc( sizeof( float ) * size.width );
    du = (float *) icvAlloc( sizeof( float ) * size.width );

    if( x1 == NULL || x2 == NULL || x3 == NULL || du == NULL )
    {
        if( x1 )
            icvFree( (void **) &x1 );
        if( x2 )
            icvFree( (void **) &x2 );
        if( x3 )
            icvFree( (void **) &x3 );
        if( du )
            icvFree( (void **) &du );
        return CV_OUTOFMEM_ERR;
    }

    *srcImage = (uchar) 1;

    if( !interToggle )
    {
        for( u = 0; u < size.width; u++ )
        {
            float w = u - u0;
            float x = a1 * w;

            x1[u] = p2 / x;
            x2[u] = x * x;
            x3[u] = 2.f * p2 * x;
            du[u] = w;
        }

        for( v = 0; v < size.height; v++, dstImage += step )
        {
            float dv = v - v0;
            float y = b1 * dv;
            float y1 = p1 / y;
            float y2 = y * y;
            float y3 = 2.f * p1 * y;

            for( u = 0; u < size.width; u++ )
            {
                float r2 = x2[u] + y2;
                float bx = r2 * (k1 + r2 * k2) + x3[u] + y3;
                float by = bx + r2 * y1;
                int ud = u, vd = v;

                bx += r2 * x1[u];
                ud += cvRound( bx * du[u] );
                vd += cvRound( by * dv );
                dstImage[u] = (uchar) (ud < 0 || ud >= size.width || vd < 0 ||
                                       vd >= size.height ? 0 : srcImage[vd * step + ud]);
            }
        }
    }
    else                        /* interpolation */
    {
        int sizex = size.width - 2, sizey = size.height - 2;

        for( u = 0; u < size.width; u++ )
        {
            float w = u - u0;
            float x = a1 * w;

            x1[u] = p2 / x;
            x2[u] = x * x;
            x3[u] = 2.f * p2 * x;
            du[u] = fm * w;
        }

        for( v = 0; v < size.height; v++, dstImage += step )
        {
            float dv = v - v0;
            float y = b1 * dv;
            float y1 = p1 / y;
            float y2 = y * y;
            float y3 = 2.f * p1 * y;

            dv *= fm;

            for( u = 0; u < size.width; u++ )
            {
                float r2 = x2[u] + y2;
                float bx = r2 * (k1 + r2 * k2) + x3[u] + y3;
                float by = bx + r2 * y1;
                int iu, iv, ud, vd;

                bx += r2 * x1[u];
                iu = cvRound( bx * du[u] );
                iv = cvRound( by * dv );
                ud = iu >> S;
                vd = iv >> S;
                iu -= ud << S;
                iv -= vd << S;
                ud += u;
                vd += v;

                if( ud < 0 || ud > sizex || vd < 0 || vd > sizey )
                    dstImage[u] = 0;
                else
                {
                    int iuv = (iu >> S2) * (iv >> S2);
                    int uv = vd * step + ud;
                    int a0 = srcImage[uv];
                    int a = srcImage[uv + 1] - a0;
                    int b = srcImage[uv + step] - a0;
                    int c = srcImage[uv + step + 1] - a0 - a - b;
                    int d = ((a * iu + b * iv + c * iuv) >> S) + a0;

                    d = !(d & ~255) ? d : d < 0 ? 0 : 255;
                    dstImage[u] = (uchar) d;
                }
            }
        }
    }

    *srcImage = buf;
    icvFree( (void **) &x1 );
    icvFree( (void **) &x2 );
    icvFree( (void **) &x3 );
    icvFree( (void **) &du );
    return CV_NO_ERR;
}
/*______________________________________________________ 3-CHANNEL IMAGES ______________*/

IPCVAPI_IMPL( CvStatus, icvUnDistortEx_8u_C3R, (uchar * srcImage,
                                                uchar * dstImage,
                                                int step,
                                                CvSize size,
                                                float *intrMatrix,
                                                float *distCoeffs, int interToggle) )
{
    const float fm = FM;
    const float a1 = 1.f / intrMatrix[0], b1 = 1.f / intrMatrix[4],
        u0 = intrMatrix[2], v0 = intrMatrix[5],
        k1 = distCoeffs[0], k2 = distCoeffs[1], p1 = distCoeffs[2], p2 = distCoeffs[3];
    const int p = 0xFFFFFF;
    float *x1;
    float *x2;
    float *x3;
    float *du;
    int u, v;
    uchar buf0 = *srcImage, buf1 = srcImage[1], buf2 = srcImage[2];

    if( step < 1 )
        return CV_BADFACTOR_ERR;
    if( size.width > step || size.width < 1 || size.height < 1 )
        return CV_BADSIZE_ERR;
    if( srcImage == NULL || dstImage == NULL || intrMatrix == NULL || distCoeffs == NULL )
        return CV_NULLPTR_ERR;

    /*if ( !p1 && !p2 ) return icvUnDistort1_8uC3R ( srcImage, dstImage, step, size,
       intrMatrix, distCoeffs, interToggle ); */

    x1 = (float *) icvAlloc( sizeof( float ) * size.width );
    x2 = (float *) icvAlloc( sizeof( float ) * size.width );
    x3 = (float *) icvAlloc( sizeof( float ) * size.width );
    du = (float *) icvAlloc( sizeof( float ) * size.width );

    if( x1 == NULL || x2 == NULL || x3 == NULL || du == NULL )
    {
        if( x1 )
            icvFree( (void **) &x1 );
        if( x2 )
            icvFree( (void **) &x2 );
        if( x3 )
            icvFree( (void **) &x3 );
        if( du )
            icvFree( (void **) &du );
        return CV_OUTOFMEM_ERR;
    }

    srcImage[0] = srcImage[1] = srcImage[2] = 0;

    if( !interToggle )
    {
        for( u = 0; u < size.width; u++ )
        {
            float w = u - u0;
            float x = a1 * w;

            x1[u] = p2 / x;
            x2[u] = x * x;
            x3[u] = 2.f * p2 * x;
            du[u] = w;
        }

        for( v = 0; v < size.height; v++, dstImage += step )
        {
            float dv = v - v0;
            float y = b1 * dv;
            float y1 = p1 / y;
            float y2 = y * y;
            float y3 = 2.f * p1 * y;

            if( v == size.height - 1 )
                size.width--;

            for( u = 0; u < size.width; u++ )
            {
                float r2 = x2[u] + y2;
                float bx = r2 * (k1 + r2 * k2) + x3[u] + y3;
                float by = bx + r2 * y1;
                int ud = u, vd = v, u3 = u + u + u;

                bx += r2 * x1[u];
                ud += cvRound( bx * du[u] );
                vd += cvRound( by * dv );
                if( ud < 0 || ud >= size.width || vd < 0 || vd >= size.height )
                    *(int *) (dstImage + u3) = 0;
                else
                {
                    int uv = vd * step + ud + ud + ud;

                    *(int *) (dstImage + u3) = *(int *) (srcImage + uv) & p;
                }
            }

            if( v == size.height - 1 )
            {
                float r2 = x2[u] + y2;
                float bx = r2 * (k1 + r2 * k2) + x3[u] + y3;
                float by = bx + r2 * y1;
                int ud = u, vd = v, u3 = u + u + u;

                bx += r2 * x1[u];
                ud += cvRound( bx * du[u] );
                vd += cvRound( by * dv );
                size.width++;
                if( ud < 0 || ud >= size.width || vd < 0 || vd >= size.height )
                    dstImage[u3] = dstImage[u3 + 1] = dstImage[u3 + 2] = 0;
                else
                {
                    int uv = vd * step + ud + ud + ud;

                    dstImage[u3] = srcImage[uv];
                    dstImage[u3 + 1] = srcImage[uv + 1];
                    dstImage[u3 + 2] = srcImage[uv + 2];
                }
            }
        }
    }
    else                        /* interpolation */
    {
        int sizex = size.width - 2, sizey = size.height - 2;

        for( u = 0; u < size.width; u++ )
        {
            float w = u - u0;
            float x = a1 * w;

            x1[u] = p2 / x;
            x2[u] = x * x;
            x3[u] = 2.f * p2 * x;
            du[u] = fm * w;
        }

        for( v = 0; v < size.height; v++, dstImage += step )
        {
            float dv = v - v0;
            float y = b1 * dv;
            float y1 = p1 / y;
            float y2 = y * y;
            float y3 = 2.f * p1 * y;

            dv *= fm;

            if( v == size.height - 1 )
                size.width--;

            for( u = 0; u < size.width; u++ )
            {
                float r2 = x2[u] + y2;
                float bx = r2 * (k1 + r2 * k2) + x3[u] + y3;
                float by = bx + r2 * y1;
                int iu, iv, ud, vd, u3 = u + u + u;

                bx += r2 * x1[u];
                iu = cvRound( bx * du[u] );
                iv = cvRound( by * dv );
                ud = iu >> S;
                vd = iv >> S;
                iu -= ud << S;
                iv -= vd << S;
                ud += u;
                vd += v;

                if( ud < 0 || ud > sizex || vd < 0 || vd > sizey )
                    *(int *) (dstImage + u3) = 0;
                else
                {
                    int iuv = (iu >> S2) * (iv >> S2);
                    int uv = vd * step + ud + ud + ud;
                    int uvs = uv + step;
                    int a01 = srcImage[uv];
                    int a02 = srcImage[uv + 1];
                    int a03 = srcImage[uv + 2];
                    int a1 = srcImage[uv + 3];
                    int a2 = srcImage[uv + 4];
                    int a3 = srcImage[uv + 5];
                    int b1 = srcImage[uvs];
                    int b2 = srcImage[uvs + 1];
                    int b3 = srcImage[uvs + 2];
                    int c1 = srcImage[uvs + 3];
                    int c2 = srcImage[uvs + 4];
                    int c3 = srcImage[uvs + 5];
                    int d1, d2, d3;

                    a1 -= a01;
                    a2 -= a02;
                    a3 -= a03;
                    b1 -= a01;
                    b2 -= a02;
                    b3 -= a03;
                    c1 -= a01 + a1 + b1;
                    c2 -= a02 + a2 + b2;
                    c3 -= a03 + a3 + b3;
                    d1 = ((a1 * iu + b1 * iv + c1 * iuv) >> S) + a01;
                    d2 = ((a2 * iu + b2 * iv + c2 * iuv) >> S) + a02;
                    d3 = ((a3 * iu + b3 * iv + c3 * iuv) >> S) + a03;
                    d1 = !(d1 & ~255) ? d1 : d1 < 0 ? 0 : 255;
                    d2 = !(d2 & ~255) ? d2 : d2 < 0 ? 0 : 255;
                    d3 = !(d3 & ~255) ? d3 : d3 < 0 ? 0 : 255;
                    d1 |= (d2 << 8) | (d3 << 16);
                    *(int *) (dstImage + u3) = d1;
                }
            }

            if( v == size.height - 1 )
            {
                float r2 = x2[u] + y2;
                float bx = r2 * (k1 + r2 * k2) + x3[u] + y3;
                float by = bx + r2 * y1;
                int iu, iv, ud, vd, u3 = u + u + u;

                bx += r2 * x1[u];
                iu = cvRound( bx * du[u] );
                iv = cvRound( by * dv );
                ud = iu >> S;
                vd = iv >> S;
                iu -= ud << S;
                iv -= vd << S;
                ud += u;
                vd += v;
                size.width++;

                if( ud < 0 || ud > sizex || vd < 0 || vd > sizey )
                    dstImage[u3] = dstImage[u3 + 1] = dstImage[u3 + 2] = 0;
                else
                {
                    int iuv = (iu >> S2) * (iv >> S2);
                    int uv = vd * step + ud + ud + ud;
                    int uvs = uv + step;
                    int a01 = srcImage[uv];
                    int a02 = srcImage[uv + 1];
                    int a03 = srcImage[uv + 2];
                    int a1 = srcImage[uv + 3];
                    int a2 = srcImage[uv + 4];
                    int a3 = srcImage[uv + 5];
                    int b1 = srcImage[uvs];
                    int b2 = srcImage[uvs + 1];
                    int b3 = srcImage[uvs + 2];
                    int c1 = srcImage[uvs + 3];
                    int c2 = srcImage[uvs + 4];
                    int c3 = srcImage[uvs + 5];
                    int d1, d2, d3;

                    a1 -= a01;
                    a2 -= a02;
                    a3 -= a03;
                    b1 -= a01;
                    b2 -= a02;
                    b3 -= a03;
                    c1 -= a01 + a1 + b1;
                    c2 -= a02 + a2 + b2;
                    c3 -= a03 + a3 + b3;
                    d1 = ((a1 * iu + b1 * iv + c1 * iuv) >> S) + a01;
                    d2 = ((a2 * iu + b2 * iv + c2 * iuv) >> S) + a02;
                    d3 = ((a3 * iu + b3 * iv + c3 * iuv) >> S) + a03;
                    dstImage[u3] = (uchar) d1;
                    dstImage[u3 + 1] = (uchar) d2;
                    dstImage[u3 + 2] = (uchar) d3;
                }
            }
        }
    }

    *srcImage   = buf0;
    srcImage[1] = buf1;
    srcImage[2] = buf2;
    icvFree( (void **) &x1 );
    icvFree( (void **) &x2 );
    icvFree( (void **) &x3 );
    icvFree( (void **) &du );
    return CV_NO_ERR;
}
/*______________________________________________________________________________________*/

/*  End of file  */
