/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/
#include "_cv.h"
#include "_cvwrap.h"
#include "_cvutils.h"
#include "_cvarray.h"
#include "_cvconvert.h"
#include <limits.h>

/****************************************************************************************\
*                                      Samplers                                          *
\****************************************************************************************/

CV_IMPL int
cvSampleLine( IplImage * img, CvPoint pt1, CvPoint pt2, void *buffer )
{
    CvSize img_size;
    uchar *img_data;
    int img_step = 0;
    int count = -1;

    CV_FUNCNAME( "cvSampleLine" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( img ));
    cvGetImageRawData( img, &img_data, &img_step, &img_size );

    switch (img->depth)
    {
    case IPL_DEPTH_8U:
    case IPL_DEPTH_8S:
        switch (img->nChannels)
        {
        case 1:
            count = icvSampleLine_8u_C1R( img_data, img_step, img_size,
                                           (uchar *) buffer, pt1, pt2 );
            if( count < 0 )
                CV_ERROR( IPL_StsInternal, icvCvErrMsg );
            break;
        case 3:
            count = icvSampleLine_8u_C3R( img_data, img_step, img_size,
                                           (uchar *) buffer, pt1, pt2 );
            if( count < 0 )
                CV_ERROR( IPL_StsInternal, icvCvErrMsg );
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    case IPL_DEPTH_32F:
        switch (img->nChannels)
        {
        case 1:
            count = icvSampleLine_32f_C1R( (float *) img_data, img_step, img_size,
                                            (float *) buffer, pt1, pt2 );
            if( count < 0 )
                CV_ERROR( IPL_StsInternal, icvCvErrMsg );
            break;
        case 3:
            count = icvSampleLine_32f_C3R( (float *) img_data, img_step, img_size,
                                            (float *) buffer, pt1, pt2 );
            if( count < 0 )
                CV_ERROR( IPL_StsInternal, icvCvErrMsg );
            break;
        default:
            CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );
        }
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }

    __CLEANUP__;
    __END__;

    return count;
}


CV_IMPL void
cvGetRectSubPix( IplImage * src, IplImage * dst, CvPoint2D32f center )
{
    CvSize src_size, dst_size;
    uchar *src_data;
    float *dst_data;
    int src_step = 0, dst_step = 0;

    CV_FUNCNAME( "cvGetRectSubPix" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( src ));
    CV_CALL( CV_CHECK_IMAGE( dst ));

    cvGetImageRawData( src, &src_data, &src_step, &src_size );
    cvGetImageRawData( dst, (uchar **) & dst_data, &dst_step, &dst_size );

    if( dst_size.width > src_size.width || dst_size.height > src_size.height )
        CV_ERROR( IPL_StsBadArg, "dst ROI is larger than src ROI" );

    if( src->nChannels != 1 || src->nChannels != dst->nChannels )
        CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );

    if( dst->depth != IPL_DEPTH_32F )
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );

    switch (src->depth)
    {
    case IPL_DEPTH_8U:
        IPPI_CALL( icvGetRectSubPix_8u32f_C1R( (const uchar *) src_data, src_step, src_size,
                                                (float *) dst_data, dst_step, dst_size,
                                                center ));
        break;
    case IPL_DEPTH_8S:
        IPPI_CALL( icvGetRectSubPix_8s32f_C1R( (const char *) src_data, src_step, src_size,
                                                (float *) dst_data, dst_step, dst_size,
                                                center ));
        break;
    case IPL_DEPTH_32F:
        IPPI_CALL( icvGetRectSubPix_32f_C1R( (const float *) src_data, src_step, src_size,
                                              (float *) dst_data, dst_step, dst_size,
                                              center ));
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }

    __CLEANUP__;
    __END__;
}


/****************************************************************************************\
*                                      Conversion                                        *
\****************************************************************************************/

CV_IMPL void
cvCvtPixToPlane( IplImage * src, IplImage * dst0,
                 IplImage * dst1, IplImage * dst2, IplImage * dst3 )
{
    uchar *src_data = 0, *dst_data[4];
    int src_step = 0, dst_step, dst0_step = 0, dst1_step = 0, dst2_step = 0, dst3_step = 0;
    CvSize src_size, dst0_size, dst1_size, dst2_size, dst3_size;

    CV_FUNCNAME( "cvCvtPixToPlane" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( src ));
    if( dst0 )
        CV_CALL( CV_CHECK_IMAGE( dst0 ));
    if( dst1 )
        CV_CALL( CV_CHECK_IMAGE( dst1 ));
    if( dst2 )
        CV_CALL( CV_CHECK_IMAGE( dst2 ));

    memset( dst_data, 0, sizeof( dst_data ));

    if( src->nChannels == 4 )
    {
        if( dst3 )
        {
            CV_CALL( CV_CHECK_IMAGE( dst3 ));

            if( dst3->depth != src->depth )
                CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
            if( dst3->nChannels != 1 )
                CV_ERROR( IPL_BadNumChannels, "output image have wrong number of channels" );
        }
    }
    else
    {
        if( dst3 )
            CV_ERROR( IPL_StsBadArg, "dst3 is not NULL while src has three channels" );
    }

    if( src->depth != IPL_DEPTH_8U && src->depth != IPL_DEPTH_32F )
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );

    if( src->nChannels != 3 && src->nChannels != 4 )
        CV_ERROR( IPL_BadNumChannels, "input image have wrong number of channels" );
    if( (dst0 && dst0->nChannels != 1) ||
        (dst1 && dst1->nChannels != 1) || (dst2 && dst2->nChannels != 1) )
        CV_ERROR( IPL_BadNumChannels, "output image have wrong number of channels" );

    if( (dst0 && dst0->depth != src->depth) ||
        (dst1 && dst1->depth != src->depth) || (dst2 && dst2->depth != src->depth) )
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );

    cvGetImageRawData( src, &src_data, &src_step, &src_size );
    if( dst0 )
        cvGetImageRawData( dst0, dst_data, &dst0_step, &dst0_size );
    if( dst1 )
        cvGetImageRawData( dst1, dst_data + 1, &dst1_step, &dst1_size );
    if( dst2 )
        cvGetImageRawData( dst2, dst_data + 2, &dst2_step, &dst2_size );

    if( (src->nChannels == 4) && dst3 )
        cvGetImageRawData( dst3, dst_data + 3, &dst3_step, &dst3_size );

    dst_step = MAX( dst0_step, dst1_step );
    dst_step = MAX( dst_step, dst2_step );
    dst_step = MAX( dst_step, dst3_step );

    if( dst_step == 0 ||
        dst0_step != dst_step && dst0_step != 0 ||
        dst1_step != dst_step && dst1_step != 0 ||
        dst2_step != dst_step && dst2_step != 0 || dst3_step != dst_step && dst3_step != 0 )
        CV_ERROR( IPL_StsBadArg, "strides of output images are different" );

    if( src->nChannels == 3 )
    {
        switch (src->depth)
        {
        case IPL_DEPTH_8U:
            IPPI_CALL( icvCopy_8u_C3P3R( src_data, src_step, dst_data, dst_step, src_size ));
            break;
        case IPL_DEPTH_32F:
            IPPI_CALL( icvCopy_32f_C3P3R( (const float *) src_data, src_step,
                                           (float **) dst_data, dst_step, src_size ));
            break;
        default:
            CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
            break;
        }
    }
    else
    {
        switch (src->depth)
        {
        case IPL_DEPTH_8U:
            IPPI_CALL( icvCopy_8u_C4P4R( src_data, src_step, dst_data, dst_step, src_size ));
            break;
        case IPL_DEPTH_32F:
            IPPI_CALL( icvCopy_32f_C4P4R( (float *) src_data, src_step,
                                           (float **) dst_data, dst_step, src_size ));
            break;
        default:
            CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
            break;
        }
    }

    __CLEANUP__;
    __END__;
}



CV_IMPL void
cvCvtPlaneToPix( IplImage * src0, IplImage * src1,
                 IplImage * src2, IplImage * src3, IplImage * dst )
{
    uchar *dst_data = 0, *src_data[4];
    int dst_step = 0, src0_step = 0, src1_step = 0, src2_step = 0, src3_step = 0;
    CvSize dst_size, src0_size, src1_size, src2_size, src3_size;

    CV_FUNCNAME( "cvCvtPixToPlane" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( dst ));
    if( src0 )
        CV_CALL( CV_CHECK_IMAGE( src0 ));
    if( src1 )
        CV_CALL( CV_CHECK_IMAGE( src1 ));
    if( src2 )
        CV_CALL( CV_CHECK_IMAGE( src2 ));

    if( dst->nChannels == 4 )
    {
        if( src3 )
        {
            CV_CALL( CV_CHECK_IMAGE( src3 ));

            if( src3->depth != dst->depth )
                CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
            if( src3->nChannels != 1 )
                CV_ERROR( IPL_BadNumChannels, "output image have wrong number of channels" );
        }
    }
    else
    {
        if( src3 )
            CV_ERROR( IPL_StsBadArg, "src3 is not NULL while dst has three channels" );
    }

    if( dst->depth != IPL_DEPTH_8U && dst->depth != IPL_DEPTH_32F )
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );

    if( dst->nChannels != 3 && dst->nChannels != 4 )
        CV_ERROR( IPL_BadNumChannels, "input image have wrong number of channels" );
    if( (src0 && src0->nChannels != 1) ||
        (src1 && src1->nChannels != 1) || (src2 && src2->nChannels != 1) )
        CV_ERROR( IPL_BadNumChannels, "output image have wrong number of channels" );

    if( (src0 && src0->depth != dst->depth) ||
        (src1 && src1->depth != dst->depth) || (src2 && src2->depth != dst->depth) )
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );

    cvGetImageRawData( dst, &dst_data, &dst_step, &dst_size );
    if( src0 )
        cvGetImageRawData( src0, src_data, &src0_step, &src0_size );
    if( src1 )
        cvGetImageRawData( src1, src_data + 1, &src1_step, &src1_size );
    if( src2 )
        cvGetImageRawData( src2, src_data + 2, &src2_step, &src2_size );

    if( (dst->nChannels == 4) && src3 )
        cvGetImageRawData( src3, src_data + 3, &src3_step, &src3_size );

    if( src0_step != src1_step || src1_step != src2_step ||
        src2_step != src3_step && src3_step != 0 )
        CV_ERROR( IPL_StsBadArg, "strides of output images are different" );

    if( dst->nChannels == 3 )
    {
        switch (dst->depth)
        {
        case IPL_DEPTH_8U:
            IPPI_CALL( icvCopy_8u_P3C3R( (const uchar **) src_data, src0_step,
                                          dst_data, dst_step, dst_size ));
            break;
        case IPL_DEPTH_32F:
            IPPI_CALL( icvCopy_32f_P3C3R( (const float **) src_data, src0_step,
                                           (float *) dst_data, dst_step, dst_size ));
            break;
        default:
            CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
            break;
        }
    }
    else
    {
        switch (dst->depth)
        {
        case IPL_DEPTH_8U:
            IPPI_CALL( icvCopy_8u_P4C4R( (const uchar **) src_data, src0_step,
                                          dst_data, dst_step, dst_size ));
            break;
        case IPL_DEPTH_32F:
            IPPI_CALL( icvCopy_32f_P4C4R( (const float **) src_data, src0_step,
                                           (float *) dst_data, dst_step, dst_size ));
            break;
        default:
            CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
            break;
        }
    }

    __CLEANUP__;
    __END__;
}


typedef CvStatus ( CV_STDCALL *CvConvertFunc)( void* src, int src_step,
                                               void* dst, int dst_step,
                                               CvSize size, double scale, double shift );


CV_IMPL void
cvConvertScale( IplImage * src, IplImage * dst, double scale, double shift )
{
    static CvConvertFunc funcs[25];
    
    uchar *src_data = 0, *dst_data = 0;
    int src_step = 0, dst_step = 0;
    CvSize src_size, dst_size;
    int src_idx, dst_idx, idx;

    CV_FUNCNAME( "cvConvertScale" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( src ));
    CV_CALL( CV_CHECK_IMAGE( dst ));

    if( src->nChannels != dst->nChannels )
        CV_ERROR( IPL_BadNumChannels, "src and dst have different number of channels" );

    if( icvGetImageCOI( src ) != 0 || icvGetImageCOI( dst ) != 0 )
        CV_ERROR( IPL_BadCOI, "src or dst have channel of interest selected" );

    cvGetImageRawData( src, &src_data, &src_step, &src_size );
    cvGetImageRawData( dst, &dst_data, &dst_step, &dst_size );

    if( src_size != dst_size )
        CV_ERROR( IPL_StsBadArg, "src and dst have different ROI sizes" );

    if( !funcs[0] )
    {
        funcs[0] = 0;
        funcs[1] = 0;
        funcs[2] = (CvConvertFunc)icvCvt_8u16s_C1R;
        funcs[3] = 0;
        funcs[4] = (CvConvertFunc)icvCvt_8u32f_C1R;

        funcs[5] = 0;
        funcs[6] = 0;
        funcs[7] = (CvConvertFunc)icvCvt_8s16s_C1R;
        funcs[8] = 0;
        funcs[9] = (CvConvertFunc)icvCvt_8s32f_C1R;

        funcs[10] = (CvConvertFunc)icvCvt_16s8u_C1R;
        funcs[11] = (CvConvertFunc)icvCvt_16s8s_C1R;
        funcs[12] = 0;
        funcs[13] = 0;
        funcs[14] = (CvConvertFunc)icvCvt_16s32f_C1R;

        funcs[15] = 0;
        funcs[16] = 0;
        funcs[17] = 0;
        funcs[18] = 0;
        funcs[19] = (CvConvertFunc)icvCvt_32s32f_C1R;

        funcs[20] = (CvConvertFunc)icvCvt_32f8u_C1R;
        funcs[21] = (CvConvertFunc)icvCvt_32f8s_C1R;
        funcs[22] = (CvConvertFunc)icvCvt_32f16s_C1R;
        funcs[23] = (CvConvertFunc)icvCvt_32f32s_C1R;
        funcs[24] = 0;
    }

    src_idx = src->depth == IPL_DEPTH_8U ?  0 : src->depth == IPL_DEPTH_8S ? 1 :
              src->depth == IPL_DEPTH_16S ? 2 : src->depth == IPL_DEPTH_32S ? 3 : 
              src->depth == IPL_DEPTH_32F ? 4 : -1;

    dst_idx = dst->depth == IPL_DEPTH_8U ?  0 : dst->depth == IPL_DEPTH_8S ? 1 :
              dst->depth == IPL_DEPTH_16S ? 2 : dst->depth == IPL_DEPTH_32S ? 3 : 
              dst->depth == IPL_DEPTH_32F ? 4 : -1;

    idx = src_idx * 5 + dst_idx;

    if( src_idx < 0 || dst_idx < 0 || funcs[idx] == 0 )
        CV_ERROR( IPL_StsBadArg, "src and dst have unsupported type combination" );

    src_size.width *= src->nChannels;

    if( src_size.width*icvGetBtPix(src) == src_step &&
        src_size.width*icvGetBtPix(dst) == dst_step )
    {
        src_size.width *= src_size.height;
        src_step *= src_size.height;
        dst_step *= src_size.height;
        src_size.height = 1;
    }

    IPPI_CALL( funcs[idx]( src_data, src_step, dst_data, dst_step, src_size, scale, shift ));

    __CLEANUP__;
    __END__;
}


/****************************************************************************************\
*                                Comparison/Matching                                     *
\****************************************************************************************/

CV_IMPL void
cvAbsDiff( IplImage * srcA, IplImage * srcB, IplImage * dst )
{
    CvSize srcA_size, srcB_size, dst_size;
    uchar *srcA_data = 0;
    uchar *srcB_data = 0;
    uchar *dst_data = 0;
    int srcA_step = 0, srcB_step = 0, dst_step = 0;

    CV_FUNCNAME( "cvAbsDiff" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( srcA ));
    CV_CALL( CV_CHECK_IMAGE( srcB ));
    CV_CALL( CV_CHECK_IMAGE( dst ));

    cvGetImageRawData( srcA, &srcA_data, &srcA_step, &srcA_size );
    cvGetImageRawData( srcB, &srcB_data, &srcB_step, &srcB_size );
    cvGetImageRawData( dst, &dst_data, &dst_step, &dst_size );

    if( srcA->nChannels != srcB->nChannels || srcA->nChannels != dst->nChannels )
        CV_ERROR( IPL_BadNumChannels, "images have not the same number of channels" );

    if( srcA->depth != srcB->depth || srcA->depth != dst->depth )
        CV_ERROR( IPL_BadDepth, "images have not the same depth" );

    if( icvGetImageCOI( srcA ) != 0 ||
        icvGetImageCOI( srcB ) != 0 || icvGetImageCOI( dst ) != 0 )
        CV_ERROR( IPL_BadCOI, "srcA or srcB or dst have channel of interest selected" );

    if( srcA_size != srcB_size || srcA_size != dst_size )
        CV_ERROR( IPL_StsBadArg, "dst ROI is larger than src ROI" );

    srcA_size.width *= srcA->nChannels;

    switch (srcA->depth)
    {
    case IPL_DEPTH_8U:
        IPPI_CALL( icvAbsDiff_8u_C1R( srcA_data, srcA_step, srcB_data, srcB_step,
                                       dst_data, dst_step, srcA_size ));
        break;
    case IPL_DEPTH_32F:
        IPPI_CALL( icvAbsDiff_32f_C1R( (float *) srcA_data, srcA_step,
                                        (float *) srcB_data, srcB_step,
                                        (float *) dst_data, dst_step, srcA_size ));
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }

    __CLEANUP__;
    __END__;
}


CV_IMPL void
cvAbsDiffS( IplImage * src, IplImage * dst, double value )
{
    CvSize src_size, dst_size;
    uchar *src_data = 0;
    uchar *dst_data = 0;
    int src_step = 0, dst_step = 0;

    CV_FUNCNAME( "cvAbsDiffS" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( src ));
    CV_CALL( CV_CHECK_IMAGE( dst ));

    cvGetImageRawData( src, &src_data, &src_step, &src_size );
    cvGetImageRawData( dst, &dst_data, &dst_step, &dst_size );

    if( src->nChannels != dst->nChannels )
        CV_ERROR( IPL_BadNumChannels, "images have not the same number of channels" );

    if( src->depth != dst->depth )
        CV_ERROR( IPL_BadDepth, "images have not the same depth" );

    if( icvGetImageCOI( src ) != 0 || icvGetImageCOI( dst ) != 0 )
        CV_ERROR( IPL_BadCOI, "src or srcB or dst have channel of interest selected" );

    if( src_size != dst_size )
        CV_ERROR( IPL_StsBadArg, "dst ROI is larger than src ROI" );

    src_size.width *= src->nChannels;

    switch (src->depth)
    {
    case IPL_DEPTH_8U:
        IPPI_CALL( icvAbsDiffC_8u_C1R( src_data, src_step,
                                        dst_data, dst_step, src_size, cvRound( value )));
        break;
    case IPL_DEPTH_32F:
        IPPI_CALL( icvAbsDiffC_32f_C1R( (float *) src_data, src_step,
                                         (float *) dst_data, dst_step,
                                         src_size, (float) value ));
        break;
    default:
        CV_ERROR( IPL_BadDepth, icvUnsupportedFormat );
    }

    __CLEANUP__;
    __END__;
}



typedef CvStatus( CV_STDCALL * CvMatchBufSizeFunc ) (CvSize roiSize,
                                                      CvSize templSize,
                                                      CvDataType dataType, int *bufferSize);

typedef CvStatus( CV_STDCALL * CvMatchTemplFunc ) (const void *pImage,
                                                    int imageStep, CvSize roiSize,
                                                    const void *pTemplate,
                                                    int templStep, CvSize templSize,
                                                    void *pResult, int resultStep,

                                                    void *pBuffer);

CV_IMPL void
cvMatchTemplate( IplImage * img,
                 IplImage * templ, IplImage * result, CvTemplMatchMethod method )
{
    static CvMatchBufSizeFunc bufSizeFuncs[6];
    static CvMatchTemplFunc funcs[18];

    CvSize imgSize, templSize, resultSize;
    CvDataType dataType = cv16s;
    int dataOffset = -1;
    uchar *imgData;
    int imgStep = 0;
    uchar *templData;
    int templStep = 0;
    uchar *resultData;
    int resultStep = 0;
    int bufferSize;
    int imethod = (int) method;
    void *buffer = 0;

    CV_FUNCNAME( "cvMatchTemplate" );

    __BEGIN__;

    CV_CALL( CV_CHECK_IMAGE( img ));
    CV_CALL( CV_CHECK_IMAGE( templ ));
    CV_CALL( CV_CHECK_IMAGE( result ));

    cvGetImageRawData( img, &imgData, &imgStep, &imgSize );
    cvGetImageRawData( templ, &templData, &templStep, &templSize );
    cvGetImageRawData( result, &resultData, &resultStep, &resultSize );

    if( img->nChannels != 1 || img->nChannels != templ->nChannels ||
        img->nChannels != result->nChannels )
        CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );

    if( img->depth != templ->depth || result->depth != IPL_DEPTH_32F )
        CV_ERROR( IPL_BadNumChannels, icvUnsupportedFormat );

    if( resultSize.width != imgSize.width - templSize.width + 1 ||
        resultSize.height != imgSize.height - templSize.height + 1 )
        CV_ERROR( IPL_StsBadArg, "output image ROI size is incorrect" );

    if( method < CV_TM_SQDIFF || method > CV_TM_CCOEFF_NORMED )
        CV_ERROR( IPL_StsBadArg, "unknown comparison method" );

    if( !bufSizeFuncs[0] )
    {
        bufSizeFuncs[1] = (CvMatchBufSizeFunc) icvMatchTemplateGetBufSize_SqDiffNormed;
        bufSizeFuncs[2] = (CvMatchBufSizeFunc) icvMatchTemplateGetBufSize_Corr;
        bufSizeFuncs[3] = (CvMatchBufSizeFunc) icvMatchTemplateGetBufSize_CorrNormed;
        bufSizeFuncs[4] = (CvMatchBufSizeFunc) icvMatchTemplateGetBufSize_Coeff;
        bufSizeFuncs[5] = (CvMatchBufSizeFunc) icvMatchTemplateGetBufSize_CoeffNormed;

        bufSizeFuncs[0] = (CvMatchBufSizeFunc) icvMatchTemplateGetBufSize_SqDiff;
    }

    if( !funcs[0] )
    {
        funcs[1] = (CvMatchTemplFunc) icvMatchTemplate_SqDiffNormed_8u32f_C1R;
        funcs[2] = (CvMatchTemplFunc) icvMatchTemplate_Corr_8u32f_C1R;
        funcs[3] = (CvMatchTemplFunc) icvMatchTemplate_CorrNormed_8u32f_C1R;
        funcs[4] = (CvMatchTemplFunc) icvMatchTemplate_Coeff_8u32f_C1R;
        funcs[5] = (CvMatchTemplFunc) icvMatchTemplate_CoeffNormed_8u32f_C1R;

        funcs[6] = (CvMatchTemplFunc) icvMatchTemplate_SqDiff_8s32f_C1R;
        funcs[7] = (CvMatchTemplFunc) icvMatchTemplate_SqDiffNormed_8s32f_C1R;
        funcs[8] = (CvMatchTemplFunc) icvMatchTemplate_Corr_8s32f_C1R;
        funcs[9] = (CvMatchTemplFunc) icvMatchTemplate_CorrNormed_8s32f_C1R;
        funcs[10] = (CvMatchTemplFunc) icvMatchTemplate_Coeff_8s32f_C1R;
        funcs[11] = (CvMatchTemplFunc) icvMatchTemplate_CoeffNormed_8s32f_C1R;

        funcs[12] = (CvMatchTemplFunc) icvMatchTemplate_SqDiff_32f_C1R;
        funcs[13] = (CvMatchTemplFunc) icvMatchTemplate_SqDiffNormed_32f_C1R;
        funcs[14] = (CvMatchTemplFunc) icvMatchTemplate_Corr_32f_C1R;
        funcs[15] = (CvMatchTemplFunc) icvMatchTemplate_CorrNormed_32f_C1R;
        funcs[16] = (CvMatchTemplFunc) icvMatchTemplate_Coeff_32f_C1R;
        funcs[17] = (CvMatchTemplFunc) icvMatchTemplate_CoeffNormed_32f_C1R;

        funcs[0] = (CvMatchTemplFunc) icvMatchTemplate_SqDiff_8u32f_C1R;
    }

    switch (img->depth)
    {
    case IPL_DEPTH_8U:
        dataType = cv8u;
        dataOffset = 0;
        break;
    case IPL_DEPTH_8S:
        dataType = cv8s;
        dataOffset = 1;
        break;
    case IPL_DEPTH_32F:
        dataType = cv32f;
        dataOffset = 2;
        break;
    default:
        CV_ERROR( IPL_StsBadArg, icvUnsupportedFormat );
    }

    IPPI_CALL( bufSizeFuncs[imethod] (imgSize, templSize, dataType, &bufferSize) );

    buffer = icvAlloc( bufferSize );
    if( !buffer )
        CV_ERROR_FROM_STATUS( CV_OUTOFMEM_ERR );

    IPPI_CALL( funcs[imethod + dataOffset * 6] (imgData, imgStep, imgSize,
                                                templData, templStep, templSize,
                                                resultData, resultStep, buffer) );

    __CLEANUP__;
    __END__;

    icvFree( &buffer );
}


CV_IMPL void
cvFillImage( IplImage * image, double val )
{
    CV_FUNCNAME( "cvFillImage" );

    __BEGIN__;

    icvSet( image, (float)cvRound( val ));

    __CLEANUP__;
    __END__;
}


/****************************************************************************************\
*                                Mathematical functions                                  *
\****************************************************************************************/

CV_IMPL void
cvbFastArctan( const float *y, const float *x, float *angle, int len )
{
    CV_FUNCNAME( "cvbFastArctan" );

    __BEGIN__;

    IPPI_CALL( icvbFastArctan_32f( y, x, angle, len ));

    __CLEANUP__;
    __END__;
}


CV_IMPL void
cvbCartToPolar( const float *y, const float *x, float *mag, float *angle, int len )
{
    CV_FUNCNAME( "cvbCartToPolar" );

    __BEGIN__;

    IPPI_CALL( icvbCartToPolar32f( (float *) y, (float *) x, mag, angle, len ));

    __CLEANUP__;
    __END__;
}

CV_IMPL void
cvbSqrt( const float *x, float *sqrt_x, int len )
{
    CV_FUNCNAME( "cvbSqrt" );

    __BEGIN__;

    IPPI_CALL( icvbSqrt_32f( x, sqrt_x, len ));

    __CLEANUP__;
    __END__;
}


CV_IMPL void
cvbInvSqrt( const float *x, float *inv_sqrt_x, int len )
{
    CV_FUNCNAME( "cvbInvSqrt" );

    __BEGIN__;

    IPPI_CALL( icvbInvSqrt_32f( x, inv_sqrt_x, len ));

    __CLEANUP__;
    __END__;
}


CV_IMPL void
cvbReciprocal( const float *x, float *inv_x, int len )
{
    CV_FUNCNAME( "cvbReciprocal" );

    __BEGIN__;

    IPPI_CALL( icvbRecip_32f( x, inv_x, len ));

    __CLEANUP__;
    __END__;
}


CV_IMPL void
cvbFastExp( const float *x, double *exp_x, int len )
{
    CV_FUNCNAME( "cvbFastExp" );

    __BEGIN__;

    IPPI_CALL( icvbExp_32f64f( x, exp_x, len ));

    __CLEANUP__;
    __END__;
}


CV_IMPL void
cvbFastLog( const double *x, float *log_x, int len )
{
    CV_FUNCNAME( "cvbFastLog" );

    __BEGIN__;

    IPPI_CALL( icvbLog_64f32f( x, log_x, len ));

    __CLEANUP__;
    __END__;
}


CV_IMPL void
cvRandInit( CvRandState * state, double lower, double upper, int seed )
{
    CV_FUNCNAME( "cvRandInit" );

    __BEGIN__;

    if( !state )
        CV_ERROR_FROM_STATUS( CV_NULLPTR_ERR );
    if( lower > upper )
        CV_ERROR_FROM_STATUS( CV_BADRANGE_ERR );

    state->factor = (float)(upper - lower);
    state->delta = (float)(lower - state->factor);

    state->seed = seed ? seed : UINT_MAX;
    state->carry = 0;

    __CLEANUP__;
    __END__;
}


CV_IMPL void
cvRandSetRange( CvRandState * state, double lower, double upper )
{
    CV_FUNCNAME( "cvRandSetRange" );

    __BEGIN__;

    if( !state )
        CV_ERROR_FROM_STATUS( CV_NULLPTR_ERR );
    if( lower > upper )
        CV_ERROR_FROM_STATUS( CV_BADRANGE_ERR );

    state->factor = (float)(upper - lower);
    state->delta = (float)(lower - state->factor);

    __CLEANUP__;
    __END__;
}


CV_IMPL void
cvbRand( CvRandState * state, float *dst, int len )
{
    CV_FUNCNAME( "cvbRand" );

    __BEGIN__;

    IPPI_CALL( icvbRand32f( state, dst, len ));

    __CLEANUP__;
    __END__;
}


CV_IMPL unsigned
cvRandNext( CvRandState * state )
{
    unsigned seed = 0;
    uint64 temp;

    CV_FUNCNAME( "cvRandNext" );

    __BEGIN__;

    if( !state )
        CV_ERROR_FROM_STATUS( CV_NULLPTR_ERR );

    seed = state->seed;
    temp = ((uint64) seed) * _CV_MWC_A + state->carry;
    state->seed = seed = (unsigned) temp;
    state->carry = (unsigned) (temp >> 32);

    __CLEANUP__;
    __END__;

    return seed;
}


/* End of file. */
