/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include "_cv.h"

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       icvAddVector_32f   
//    Purpose:    Sum of 2 vectors
//    Context:   
//    Parameters: src1 - first vector
//                src2 - second vector
//                dest - destination vector
//                length - size of vectors
//               
//    Returns:   
//    Notes:     
//F*/
IPCVAPI_IMPL( CvStatus, icvAddVector_32f, (CvVect32f src1,
                                           CvVect32f src2, CvVect32f dest, int length) )
{
    int i;

    /* check bad arguments */
    if( src1 == NULL )
        return CV_NULLPTR_ERR;
    if( src2 == NULL )
        return CV_NULLPTR_ERR;
    if( dest == NULL )
        return CV_NULLPTR_ERR;
    if( length <= 0 )
        return CV_BADSIZE_ERR;


    for( i = 0; i < length; i++ )
    {
        dest[i] = src1[i] + src2[i];
    }

    return CV_NO_ERR;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       icvSubVector_32f 
//    Purpose:    Subtract one vector from other
//    Context:   
//    Parameters: src1 - first vector
//                src2 - second vector
//                dest - destination vector
//                length - size of vectors
//               
//    Returns:   
//    Notes:     
//F*/
IPCVAPI_IMPL( CvStatus, icvSubVector_32f, (CvVect32f src1,
                                           CvVect32f src2, CvVect32f dest, int length) )
{
    int i;

    /* check bad arguments */
    if( src1 == NULL )
        return CV_NULLPTR_ERR;
    if( src2 == NULL )
        return CV_NULLPTR_ERR;
    if( dest == NULL )
        return CV_NULLPTR_ERR;
    if( length <= 0 )
        return CV_BADSIZE_ERR;

    for( i = 0; i < length; i++ )
    {
        dest[i] = src1[i] - src2[i];
    }

    return CV_NO_ERR;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       icvDotProduct32f   
//    Purpose:    Computes dot product of 2 vectors
//    Context:   
//    Parameters: src1 - first vector
//                src2 - second vector
//                length - size of vectors
//                value  - result of dot product (output parameter)
//    Returns:   
//    Notes:     
//F*/
IPCVAPI_IMPL( CvStatus, icvDotProduct_32f, (CvVect32f src1,
                                            CvVect32f src2, int length, float *value) )
{
    int i;
    float sum = 0.f;

    /* check bad arguments */
    if( src1 == NULL )
        return CV_NULLPTR_ERR;
    if( src2 == NULL )
        return CV_NULLPTR_ERR;
    if( length <= 0 )
        return CV_BADSIZE_ERR;

    for( i = 0; i < length; i++ )
    {
        sum += src1[i] * src2[i];
    }

    *value = sum;
    return CV_NO_ERR;
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       icvCrossProduct2L_32f  
//    Purpose:    Computes cross product of two 3D vectors
//                if A and B - input vectors, then
//                C = A x B is cross product. Its components are
//
//                Cx = AyBz - AzBy 
//                Cy = AzBx - AxBz 
//                Cz = AxBy - AyBx 
//
//    Context:   
//    Parameters: src1 - first vector
//                src2 - second vector
//                dest - cross product 
//    Returns:   
//    Notes:     
//F*/
IPCVAPI_IMPL( CvStatus, icvCrossProduct2L_32f, (CvVect32f src1,
                                                CvVect32f src2, CvVect32f dest) )
{
    /* check bad arguments */
    if( src1 == NULL )
        return CV_NULLPTR_ERR;
    if( src2 == NULL )
        return CV_NULLPTR_ERR;
    if( dest == NULL )
        return CV_NULLPTR_ERR;

    dest[2] = src1[0] * src2[1] - src1[1] * src2[0];
    dest[0] = src1[1] * src2[2] - src1[2] * src2[1];
    dest[1] = src1[2] * src2[0] - src1[0] * src2[2];
    /*
       dest[2] = src1[0]*src2[1]-src1[1]*src2[0];
       dest[1] = src1[2]*src2[0]-src1[0]*src2[2];
       dest[0] = src1[1]*src2[2]-src1[2]*src2[1];
     */

    return CV_NO_ERR;
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       icvCrossProduct3P32f  
//    Purpose:    Computes cross product of three 4D vectors
//                if A,B,C - input vectors, then
//                D = A x B x C is cross product. Its components are...
//    Context:   
//    Parameters: src1 - first vector
//                src2 - second vector
//                dest - cross product 
//    Returns:   
//    Notes:     
//F*/
IPCVAPI_IMPL( CvStatus, icvCrossProduct3P_32f, (CvVect32f src1,
                                                CvVect32f src2,
                                                CvVect32f src3, CvVect32f dest) )
{
    float *a = src1;
    float *b = src2;
    float *c = src3;
    float *d = dest;

    float b0c2;
    float b2c0;

    float b1c2;
    float b2c1;

    float b0c1;
    float b1c0;

    float b1c3;
    float b3c1;

    float b2c3;
    float b3c2;

    float b0c3;
    float b3c0;

    /* check bad arguments */
    if( src1 == NULL )
        return CV_NULLPTR_ERR;
    if( src2 == NULL )
        return CV_NULLPTR_ERR;
    if( src3 == NULL )
        return CV_NULLPTR_ERR;
    if( dest == NULL )
        return CV_NULLPTR_ERR;

    /* ( A0, A1, A2, A3 )
       ( B0, B1, B2, B3 )
       ( C0, C1, C2, C3 )
     */
    b0c2 = b[0] * c[2];
    b2c0 = c[0] * b[2];

    b1c2 = b[1] * c[2];
    b2c1 = c[1] * b[2];

    b0c1 = b[0] * c[1];
    b1c0 = c[0] * b[1];

    d[3] = -(a[0] * (b1c2 - b2c1) - a[1] * (b0c2 - b2c0) + a[2] * (b0c1 - b1c0));

    b1c3 = b[1] * c[3];
    b3c1 = c[1] * b[3];

    b2c3 = b[2] * c[3];
    b3c2 = c[2] * b[3];

    d[0] = a[1] * (b2c3 - b3c2) - a[2] * (b1c3 - b3c1) + a[3] * (b1c2 - b2c1);

    b0c3 = b[0] * c[3];
    b3c0 = c[0] * b[3];


    d[1] = -(a[0] * (b2c3 - b3c2) - a[2] * (b0c3 - b3c0) + a[3] * (b0c2 - b2c0));


    d[2] = a[0] * (b1c3 - b3c1) - a[1] * (b0c3 - b3c0) + a[3] * (b0c1 - b1c0);

    return CV_NO_ERR;

#if 0
    /* multiply by elements ( A0, A1, A2, A3 ) & ( B1, B2, B3, B0 ) */
    ab1[0] = a[0] * b[1];
    ab1[1] = a[1] * b[2];
    ab1[2] = a[2] * b[3];
    ab1[3] = a[3] * b[0];

    /* multiply by elements ( A0, A1, A2, A3 ) & ( B2, B3, B0, B1 ) */
    ab2[0] = a[0] * b[2];
    ab2[1] = a[1] * b[3];
    ab2[2] = a[2] * b[0];
    ab2[3] = a[3] * b[1];

    /* multiply by elements ( A0, A1, A2, A3 ) & ( B3, B0, B1, B2 ) */
    ab3[0] = a[0] * b[3];
    ab3[1] = a[1] * b[0];
    ab3[2] = a[2] * b[1];
    ab3[3] = a[3] * b[2];

    /* multiply by elements ( A0, A1, A2, A3 ) & ( C1, C2, C3, C0 ) */
    ac1[0] = a[0] * c[1];
    ac1[1] = a[1] * c[2];
    ac1[2] = a[2] * c[3];
    ac1[3] = a[3] * c[0];

    /* multiply by elements ( A0, A1, A2, A3 ) & ( C2, C3, C0, C1 ) */
    ac2[0] = a[0] * c[2];
    ac2[1] = a[1] * c[3];
    ac2[2] = a[2] * c[0];
    ac2[3] = a[3] * c[1];

    /* multiply by elements ( A0, A1, A2, A3 ) & ( C3, C0, C1, C2 ) */
    ac3[0] = a[0] * c[3];
    ac3[1] = a[1] * c[0];
    ac3[2] = a[2] * c[1];
    ac3[3] = a[3] * c[2];
#endif
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       icvScaleVect_32f   
//    Purpose:    Multiplies  the Vector by a value.
//    Context:   
//    Parameters: src - source vector
//                dest - destination vector
//                length - size of vectors
//                value  - multiplier
//    Returns:   
//    Notes:     
//F*/
IPCVAPI_IMPL( CvStatus, icvScaleVector_32f, (CvVect32f src,
                                             CvVect32f dest, int length, float value) )
{
    int i;
    float val = value;

    if( src == NULL )
        return CV_NULLPTR_ERR;
    if( dest == NULL )
        return CV_NULLPTR_ERR;
    if( length <= 0 )
        return CV_BADSIZE_ERR;

    for( i = 0; i < length; i++ )
    {
        dest[i] = src[i] * val;
    }

    return CV_NO_ERR;
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       icvNormVector_32f   
//    Purpose:    Find norm of vector.
//    Context:   
//    Parameters: vect - vector
//                length - size of vectors
//                norm  - norm of vector
//    Returns:   
//    Notes:     
//F*/
IPCVAPI_IMPL( CvStatus, icvNormVector_32f, (CvVect32f vect, int length, float *norm) )
{
    int t;
    float sum;

    sum = 0.0f;

    for( t = 0; t < length; t++ )
    {
        sum += vect[t] * vect[t];
    }

    *norm = (float) sqrt( sum );

    return CV_NO_ERR;
}                               /* icvNormVector_32f */


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       icvMulVectors_32f   
//    Purpose:    Multiplicate each components of vectors
//    Context:   
//    Parameters: src1   - first  source vector 
//                src2   - second source vector 
//                dest   - destination result vector 
//                length - length of vectors
//    Returns:   
//    Notes:     
//F*/
IPCVAPI_IMPL( CvStatus, icvMulVectors_32f, (CvVect32f src1,
                                            CvVect32f src2, CvVect32f dest, int length) )
{
    int t;

    for( t = 0; t < length; t++ )
    {
        dest[t] = src1[t] * src2[t];
    }

    return CV_NO_ERR;
}                               /* icvMulVectors_32f */


/* End of file. */
