/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include "_cv.h"

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       icvAddVector_64d   
//    Purpose:    Sum of 2 vectors
//    Context:   
//    Parameters: src1 - first vector
//                src2 - second vector
//                dest - destination vector
//                length - size of vectors
//               
//    Returns:   
//    Notes:     
//F*/
IPCVAPI_IMPL( CvStatus, icvAddVector_64d, (CvVect64d src1,
                                           CvVect64d src2, CvVect64d dest, int length) )
{
    int i;

    /* check bad arguments */
    if( src1 == 0 )
        return CV_NULLPTR_ERR;
    if( src2 == 0 )
        return CV_NULLPTR_ERR;
    if( dest == 0 )
        return CV_NULLPTR_ERR;
    if( length <= 0 )
        return CV_BADSIZE_ERR;


    for( i = 0; i < length; i++ )
    {
        dest[i] = src1[i] + src2[i];
    }

    return CV_NO_ERR;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       icvSubVector_64d 
//    Purpose:    Subtract one vector from other
//    Context:   
//    Parameters: src1 - first vector
//                src2 - second vector
//                dest - destination vector
//                length - size of vectors
//               
//    Returns:   
//    Notes:     
//F*/
IPCVAPI_IMPL( CvStatus, icvSubVector_64d, (CvVect64d src1,
                                           CvVect64d src2, CvVect64d dest, int length) )
{
    int i;

    /* check bad arguments */
    if( src1 == 0 )
        return CV_NULLPTR_ERR;
    if( src2 == 0 )
        return CV_NULLPTR_ERR;
    if( dest == 0 )
        return CV_NULLPTR_ERR;
    if( length <= 0 )
        return CV_BADSIZE_ERR;

    for( i = 0; i < length; i++ )
    {
        dest[i] = src1[i] - src2[i];
    }

    return CV_NO_ERR;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       icvDotProduct_64d   
//    Purpose:    Computes dot product of 2 vectors
//    Context:   
//    Parameters: src1 - first vector
//                src2 - second vector
//                length - size of vectors
//                value  - result of dot product (output parameter)
//    Returns:   
//    Notes:     
//F*/
IPCVAPI_IMPL( CvStatus, icvDotProduct_64d, (CvVect64d src1,
                                            CvVect64d src2, int length, double *value) )
{
    int i;
    double sum = 0.f;

    /* check bad arguments */
    if( src1 == 0 )
        return CV_NULLPTR_ERR;
    if( src2 == 0 )
        return CV_NULLPTR_ERR;
    if( length <= 0 )
        return CV_BADSIZE_ERR;

    for( i = 0; i < length; i++ )
    {
        sum += src1[i] * src2[i];
    }

    *value = sum;
    return CV_NO_ERR;
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       icvScaleVect_64d   
//    Purpose:    Multiplies  the Vector by a value.
//    Context:   
//    Parameters: src - source vector
//                dest - destination vector
//                length - size of vectors
//                value  - multiplier
//    Returns:   
//    Notes:     
//F*/
/*=====================================================================================*/
IPCVAPI_IMPL( CvStatus, icvScaleVector_64d, (CvVect64d src,
                                             CvVect64d dest, int length, double value) )
{
    int i;
    double val = value;

    if( src == 0 )
        return CV_NULLPTR_ERR;
    if( dest == 0 )
        return CV_NULLPTR_ERR;
    if( length <= 0 )
        return CV_BADSIZE_ERR;

    for( i = 0; i < length; i++ )
    {
        dest[i] = src[i] * val;
    }

    return CV_NO_ERR;
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       icvNormVector_64d   
//    Purpose:    Find norm of vector.
//    Context:   
//    Parameters: vect - vector
//                length - size of vectors
//                norm  - norm of vector
//    Returns:   
//    Notes:     
//F*/
IPCVAPI_IMPL( CvStatus, icvNormVector_64d, (double *vect, int length, double *norm) )
{
    int t;
    double sum;

    sum = 0.0f;

    for( t = 0; t < length; t++ )
    {
        sum += vect[t] * vect[t];
    }

    *norm = (double) sqrt( sum );

    return CV_NO_ERR;
}                               /* icvNormVector_64d */


void
icvDAXPY( int n, double *x, double a, double *y )
{
    int i;

    for( i = 0; i <= n - 4; i += 4 )
    {
        double t0 = y[i] + a * x[i];
        double t1 = y[i + 1] + a * x[i + 1];

        y[i] = t0;
        y[i + 1] = t1;
        t0 = y[i + 2] + a * x[i + 2];
        t1 = y[i + 3] + a * x[i + 3];
        y[i + 2] = t0;
        y[i + 3] = t1;
    }

    for( ; i < n; i++ )
        y[i] += a * x[i];
}


double
icvDDOT( int n, double *x, double *y )
{
    int i;
    double s = 0;

    for( i = 0; i <= n - 4; i += 4 )
    {
        s += x[i] * y[i] + x[i + 1] * y[i + 1] + x[i + 2] * y[i + 2] + x[i + 3] * y[i + 3];
    }

    for( ; i < n; i++ )
        s += x[i] * y[i];

    return s;
}


void
icvGivens( int n, double *x, double *y, double c, double s )
{
    int i;
    double cpd = c - s;
    double cmd = c + s;

    for( i = 0; i < n; i++ )
    {
        double t0 = x[i];
        double t1 = y[i];
        double t2 = c * (t0 + t1);

        x[i] = t2 - t1 * cpd;
        y[i] = t2 - t0 * cmd;
    }
}


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:       icvMulVectors_64d   
//    Purpose:    Multiplicate each components of vectors
//    Context:   
//    Parameters: src1   - first  source vector 
//                src2   - second source vector 
//                dest   - destination result vector 
//                length - length of vectors
//    Returns:   
//    Notes:     
//F*/
IPCVAPI_IMPL( CvStatus, icvMulVectors_64d, (CvVect64d src1,
                                            CvVect64d src2, CvVect64d dest, int length) )
{
    int t;

    for( t = 0; t < length; t++ )
    {
        dest[t] = src1[t] * src2[t];
    }

    return CV_NO_ERR;
}                               /* icvMulVectors_64d */


/* End of file. */
