/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/
#include "_cv.h"
#include "_cvwrap.h"

const char *icvUnsupportedFormat = "Unsupported format";
const char *icvCvErrMsg = "icv function failed";

/*
#ifdef WIN32
static CRITICAL_SECTION init_section(CRITICAL_SECTION* sect)
{
    InitializeCriticalSection(sect);
    return *sect;
}

static CRITICAL_SECTION memory_allocating = init_section(&memory_allocating);

#define ENTER_SECTION EnterCriticalSection(&memory_allocating)
#define LEAVE_SECTION LeaveCriticalSection(&memory_allocating)

#else

#define ENTER_SECTION
#define LEAVE_SECTION

#endif
*/

/*======================== C O M M O N   F U N C T I O N S =============================*/

static IplROI icv_iplROI( int coi, int xOffset, int yOffset, int width, int height )
{
    IplROI roi;
    roi.coi     = coi;
    roi.xOffset = xOffset;
    roi.yOffset = yOffset;
    roi.width   = width;
    roi.height  = height;
    return roi;
}


CV_IMPL IplROI *
icvCreateROI( int coi, int xOffset, int yOffset, int width, int height )
{
    IplROI *roi = (IplROI *) icvAlloc( sizeof( *roi ));

    *roi = icv_iplROI( coi, xOffset, yOffset, width, height );
    return roi;
}


CV_IMPL IplImage *
cvCreateImageHeader( CvSize size, int depth, int channels )
{
    IplImage *img = 0;

    CV_FUNCNAME( "cvCreateImageHeader" );

    char *colorModel;
    char *channelSeq;

    switch (channels)
    {
    case 1:
        colorModel = "GRAY";
        channelSeq = "GRAY";
        break;

    case 3:
        colorModel = "RGB";
        channelSeq = "BGR";
        break;

    case 4:
        colorModel = "RGB";
        channelSeq = "BGRA";
        break;

    default:
        CV_ERROR( IPL_BadNumChannels, "Bad number of channels." );
    }

   __BEGIN__

#ifndef WITH_IPL
    CV_CALL( img = (IplImage *) icvAlloc( sizeof( *img )));
    cvInitImageHeader( img, size, depth, channels, IPL_ORIGIN_TL, 4, 1 );
#else
    img = iplCreateImageHeader( channels, 0, depth, colorModel, channelSeq,
                     IPL_DATA_ORDER_PIXEL, IPL_ORIGIN_TL, 4,
                     size.width, size.height, 0, 0, 0, 0 );
#endif
    __CLEANUP__
    if( cvGetErrStatus() < 0 )
    {
        if( img )
        {
            img->roi = 0;
            cvReleaseImageHeader( &img );
        }
    }
    __END__

    return img;
}


CV_IMPL IplImage *
cvCreateImage( CvSize size, int depth, int channels )
{
    IplImage *img = 0;

    CV_FUNCNAME( "cvCreateImage" );

    __BEGIN__

    CV_CALL( img = cvCreateImageHeader( size, depth, channels ));
#ifndef WITH_IPL
    CV_CALL( img->imageData = (char *) icvAlloc( img->height * img->widthStep ));
#else
    switch(depth)
    {
    case IPL_DEPTH_32F:
        iplAllocateImageFP( img, 0, 0 );
        break;
    default:
        iplAllocateImage( img, 0, 0 );
        break;
    }
#endif

    __CLEANUP__

    if( cvGetErrStatus() < 0 )
        cvReleaseImage( &img );

    __END__

    return img;
}

CV_IMPL void
cvReleaseImageHeader( IplImage ** image )
{
    CV_FUNCNAME( "cvReleaseImageHeader" );
    __BEGIN__
    if( image && *image )
    {
#ifndef WITH_IPL
        if( (*image)->roi )
            icvFree( &(*image)->roi );
        icvFree( image );
#else
        iplDeallocate( *image, IPL_IMAGE_HEADER | IPL_IMAGE_ROI );
        *image = 0;
#endif
    }
    __CLEANUP__
    __END__
}


CV_IMPL void
cvReleaseImage( IplImage ** image )
{
    CV_FUNCNAME( "cvReleaseImage" );
    __BEGIN__
    if( image && *image )
    {
#ifndef WITH_IPL
        if( (*image)->imageData )
            icvFree( &(*image)->imageData );
        if( (*image)->roi )
            icvFree( &(*image)->roi );
        icvFree( image );
#else
        iplDeallocate(*image, IPL_IMAGE_ALL);
        *image = 0;
#endif
    }
    __CLEANUP__
    __END__
}


void
icvCheckImageHeader( IplImage * img, const char *image_name )
{
    CV_FUNCNAME( "_cvCheckImageHeader" );

    __BEGIN__

    if( !img )
        CV_ERROR( IPL_HeaderIsNull, image_name );

    if( img->dataOrder != IPL_DATA_ORDER_PIXEL && img->nChannels > 1 )
        CV_ERROR( IPL_BadOrder, image_name );

    if( img->maskROI != 0 || img->tileInfo != 0 )
        CV_ERROR( IPL_StsBadArg, image_name );

    if( img->imageData == 0 )
        CV_ERROR( IPL_BadDataPtr, image_name );

    if( img->roi &&
        ((img->roi->xOffset | img->roi->yOffset | img->roi->width | img->roi->height) < 0 ||
         img->roi->xOffset + img->roi->width > img->width ||
         img->roi->yOffset + img->roi->height > img->height ||
         (unsigned) (img->roi->coi) > (unsigned) (img->nChannels)))
        CV_ERROR( IPL_BadROISize, image_name );

    __CLEANUP__
    __END__
}


void
icvCheckMaskImageHeader( IplImage * img, const char *image_name )
{
    CV_FUNCNAME( "cvCheckMaskImageHeader" );

    __BEGIN__

    CV_CALL( icvCheckImageHeader( img, image_name ));

    if( img->depth != IPL_DEPTH_8U && img->depth != IPL_DEPTH_8S )
        CV_ERROR( IPL_BadDepth, image_name );

    if( img->nChannels != 1 )
        CV_ERROR( IPL_BadNumChannels, image_name );

    __CLEANUP__
    __END__
}


CV_IMPL void
cvCreateImageData( IplImage * image )
{
    CV_FUNCNAME( "cvCreateImageData" );

    __BEGIN__

#ifndef WITH_IPL
    CV_CALL( image->imageData = (char *) icvAlloc( image->height * image->widthStep ));
    image->imageDataOrigin = image->imageData;
#else
    if( image->depth == IPL_DEPTH_32F )
        iplAllocateImageFP( image, 0, 0 );
    else
        iplAllocateImage( image, 0, 0 );
#endif

    __CLEANUP__
    __END__
}

CV_IMPL void
cvReleaseImageData( IplImage * image )
{
    CV_FUNCNAME( "cvReleaseImageData" );
    __BEGIN__

#ifndef WITH_IPL
    icvFree( &image->imageData );
    image->imageDataOrigin = 0;
#else
    iplDeallocate(image, IPL_IMAGE_DATA);
#endif

    __CLEANUP__
    __END__
}

CV_IMPL void
cvCopyImage( IplImage * src, IplImage * dst )
{
    CV_FUNCNAME( "cvCopyImage" );

    __BEGIN__
#ifndef WITH_IPL
    int src_size = src->widthStep * src->height;
    int dst_size = dst->widthStep * dst->height;
    char *data = dst->imageData;
    IplROI *dst_roi = dst->roi;

    if( src_size != dst_size )
    {
        cvReleaseImageData( dst );
        data = 0;
    }

    *dst = *src;
    dst->roi = dst_roi;
    if( !data )
    {
        dst->imageData = dst->imageDataOrigin = 0;
        cvCreateImageData( dst );
    }
    else
    {
        dst->imageData = dst->imageDataOrigin = data;
    }

    if( src->roi )
        cvSetImageROI( dst, cvRect( src->roi->xOffset, src->roi->yOffset,
                                    src->roi->width, src->roi->height ));
    else
        cvSetImageROI( dst, cvRect( 0, 0, src->width, src->height ));

    memcpy( dst->imageData, src->imageData, src_size );
#else
    iplCopy( src, dst );
#endif
    __CLEANUP__
    __END__
}


CV_IMPL void
cvSetImageData( IplImage * image, void *data, int step )
{
    int bt_pix;

    CV_FUNCNAME( "cvSetImageData" );

    __BEGIN__
    if( !(image && (data == 0 && step == 0 || data != 0 && step > 0)))
        CV_ERROR( IPL_StsBadArg,
                  "!(image && (data == 0 && step == 0 || data != 0 && step > 0))" );

    bt_pix = icvGetBtPix( image );

    image->imageData = image->imageDataOrigin = (char *) data;
    image->widthStep = step;
    image->imageSize = step * image->height;
    if( image->roi && image->roi->xOffset + image->roi->width > image->width )
    {
        image->roi->width = image->width - image->roi->xOffset;
    }

    if( (((int) data | step) & 7) == 0 &&
        ((image->align * bt_pix + 3) & -4) < step &&
        ((image->align * bt_pix + 7) & -8) == step )
    {
        image->align = 8;
    }
    else
    {
        image->align = 4;
    }
    __CLEANUP__
    __END__
}


CV_IMPL void
cvSetImageROI( IplImage * image, CvRect rect )
{
    CV_FUNCNAME( "cvSetImageROI" );

    __BEGIN__
    if( !image )
        CV_ERROR( IPL_HeaderIsNull, 0 );

    if( rect.x > image->width || rect.y > image->height )
        CV_ERROR( IPL_BadROISize, 0 );

    if( rect.x + rect.width < 0 || rect.y + rect.height < 0 )
        CV_ERROR( IPL_BadROISize, 0 );

    if( rect.x < 0 )
    {
        rect.width += rect.x;
        rect.x = 0;
    }

    if( rect.y < 0 )
    {
        rect.height += rect.y;
        rect.y = 0;
    }

    if( rect.x + rect.width > image->width )
        rect.width = image->width - rect.x;

    if( rect.y + rect.height > image->height )
        rect.height = image->height - rect.y;

    if( image->roi )
    {
        image->roi->xOffset = rect.x;
        image->roi->yOffset = rect.y;
        image->roi->width = rect.width;
        image->roi->height = rect.height;
    }
    else
    {
#ifndef WITH_IPL
        IplROI *roi = 0;

        CV_CALL( roi = icvCreateROI( 0, rect.x, rect.y, rect.width, rect.height ));
        image->roi = roi;
#else
        image->roi = iplCreateROI( 0, rect.x, rect.y, rect.width, rect.height );
#endif
    }

    __CLEANUP__
    __END__
}


CV_IMPL void
cvSetImageCOI( IplImage * image, int coi )
{
    CV_FUNCNAME( "cvSetImageCOI" );

    __BEGIN__
    if( !image )
        CV_ERROR( IPL_HeaderIsNull, 0 );

    if( (unsigned) coi > (unsigned) (image->nChannels) )
        CV_ERROR( IPL_BadCOI, 0 );

    if( image->roi || coi != 0 )
    {
        if( image->roi )
            image->roi->coi = coi;
        else
        {
#ifndef WITH_IPL
        IplROI *roi = 0;

            CV_CALL( roi = icvCreateROI( coi, 0, 0, image->width, image->height ));
            image->roi = roi;
#else
            image->roi = iplCreateROI( coi, 0, 0, image->width, image->height );
#endif
        }
    }

    __CLEANUP__
    __END__
}

/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    cvGetImageRawData
//    Purpose: returns pointer to the top-left corner of the image ROI,
//             step of the image and size of the ROI
//    Context:
//    Parameters:
//
//    Returns:
//             Error code
//    Notes:
//
//F*/
CV_IMPL void
cvGetImageRawData( const IplImage * image, uchar ** data, int *step, CvSize * roi_size )
{
    CV_FUNCNAME( "cvGetImageRawData" );

    __BEGIN__
    if( step )
        *step = image->widthStep;

    if( image->roi )
    {
        if( roi_size )
        {
            roi_size->width = image->roi->width;
            roi_size->height = image->roi->height;
        }

        if( data )
        {
            *data = (uchar *) (image->imageData + image->roi->yOffset * image->widthStep +
                               image->roi->xOffset * icvGetBtPix( image ));
        }
    }
    else
    {
        if( roi_size )
        {
            roi_size->width = image->width;
            roi_size->height = image->height;
        }

        if( data )
        {
            *data = (uchar *) (image->imageData);
        }
    }
    __CLEANUP__
    __END__
}


#define IMAGE_STEP(image,align)                           \
    (((image->width * image->nChannels * (image->depth & ~IPL_DEPTH_SIGN) + 7)/ 8) \
    + align - 1) & (~(align - 1))


/*F///////////////////////////////////////////////////////////////////////////////////////
//    Name:    icvInitImageHeader
//    Purpose: Filling image header (not allocate memory)
//    Context: 
//    Parameters:
//             
//    Returns:
//             pointer to the image or NULL (if error)
//    Notes:
//             
//F*/
CV_IMPL void
cvInitImageHeader( IplImage * image,
                   CvSize size, int depth, int channels, int origin, int align, int clear )
{
    CV_FUNCNAME( "cvInitImageHeader" );

    __BEGIN__
    if( !image )
        CV_ERROR( IPL_HeaderIsNull, "null pointer to header" );

    if( clear )
    {
        memset( image, 0, sizeof( *image ));
        image->imageSize = sizeof( *image );
        if( channels == 1 )
        {
            strncpy( image->colorModel, "GRAY", 4 );
            strncpy( image->channelSeq, "GRAY", 4 );
        }
        else if( channels == 3 )
        {
            strncpy( image->colorModel, "RGB", 4 );
            strncpy( image->channelSeq, "BGR", 4 );
        }
        else if( channels == 4 )
        {
            strncpy( image->colorModel, "RGB", 4 );
            strncpy( image->channelSeq, "BGRA", 4 );
        }
    }

    if( size.width < 0 || size.height < 0 )
        CV_ERROR( IPL_BadROISize, "Bad input roi" );

    if( (depth != IPL_DEPTH_1U &&
         depth != IPL_DEPTH_8U && depth != IPL_DEPTH_8S &&
         depth != IPL_DEPTH_16U && depth != IPL_DEPTH_16S &&
         depth != IPL_DEPTH_32S && depth != IPL_DEPTH_32F) || channels < 0 )
        CV_ERROR( IPL_BadDepth, "Unsupported format" );
    if( origin != CV_ORIGIN_BL && origin != CV_ORIGIN_TL )
        CV_ERROR( IPL_BadOrigin, "Bad input origin" );

    if( align != 4 && align != 8 )
        CV_ERROR( IPL_BadAlign, "Bad input align" );

    image->width = size.width;
    image->height = size.height;

    if( image->roi )
    {
        image->roi->coi = 0;
        image->roi->xOffset = image->roi->yOffset = 0;
        image->roi->width = size.width;
        image->roi->height = size.height;
    }

    image->nChannels = MAX( channels, 1 );
    image->depth = depth;
    image->align = align;
    image->widthStep = IMAGE_STEP( image, align );
    image->origin = origin;

    __CLEANUP__
    __END__
}


void
icvCheckTermCriteria( CvTermCriteria* criteria, const char* name )
{
    CV_FUNCNAME( "_cvCheckTermCriteria" );

    name;

    __BEGIN__

    if( !criteria )
        CV_ERROR( IPL_StsNullPtr, "Internal error" );

    if( (criteria->type & ~(CV_TERMCRIT_EPS | CV_TERMCRIT_ITER)) != 0 )
        CV_ERROR( IPL_StsBadArg,
                  "Unknown type of term criteria" );

    if( (criteria->type & CV_TERMCRIT_ITER) != 0 && criteria->maxIter <= 0 )
        CV_ERROR( IPL_StsBadArg,
                  "Iterations flag is set and maximum number of iterations is <= 0" );

    if( (criteria->type & CV_TERMCRIT_EPS) != 0 && criteria->epsilon < 0 )
        CV_ERROR( IPL_StsBadArg, "Accuracy flag is set and epsilon is < 0" );

    if( (criteria->type & (CV_TERMCRIT_EPS | CV_TERMCRIT_ITER)) == 0 )
        CV_ERROR( IPL_StsBadArg,
                  "Neither accuracy nor maximum iterations "
                  "number flags are set in criteria type" );

    __CLEANUP__
    __END__
}


CV_IMPL CvRect
cvGetImageROI( IplImage* img )
{
    CV_FUNCNAME( "cvGetImageROI" );

    __BEGIN__

    if( !img )
        CV_ERROR( IPL_StsNullPtr, "Null pointer to image" );

    if(img->roi)
        return cvRect( img->roi->xOffset, img->roi->yOffset,
                       img->roi->width, img->roi->height );
    else
        return cvRect( 0, 0, img->width, img->height );

    __CLEANUP__
    __END__
}


CV_IMPL IplImage*
cvCloneImage( IplImage* src )
{
    CV_FUNCNAME( "cvCloneImage" );

    IplImage* dst = 0;

    __BEGIN__

    CV_CALL( CV_CHECK_IMAGE( src ));
#ifndef WITH_IPL
    dst = cvCreateImageHeader( cvSize(src->width, src->height),
                               src->depth, src->nChannels );

    if( src->roi )
    {
        dst->roi = (IplROI*)icvAlloc( sizeof(IplROI) );
        *dst->roi = *src->roi;
    }

    if( src->imageData )
    {
        cvSetImageData( dst, icvAlloc( src->widthStep * src->height ), src->widthStep );
        memcpy( dst->imageData, src->imageData, dst->widthStep * dst->height );
    }

    if( src->maskROI )
        dst->maskROI = cvCloneImage( src->maskROI );
#else
    dst = iplCloneImage( src );
#endif

    __CLEANUP__
    __END__

    return dst;
}


/* End of file. */
