/*M///////////////////////////////////////////////////////////////////////////////////////
//
//  IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
//
//  By downloading, copying, installing or using the software you agree to this license.
//  If you do not agree to this license, do not download, install,
//  copy or use the software.
//
//
//                        Intel License Agreement
//                For Open Source Computer Vision Library
//
// Copyright (C) 2000, Intel Corporation, all rights reserved.
// Third party copyrights are property of their respective owners.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//
//   * Redistribution's of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//
//   * Redistribution's in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//
//   * The name of Intel Corporation may not be used to endorse or promote products
//     derived from this software without specific prior written permission.
//
// This software is provided by the copyright holders and contributors "as is" and
// any express or implied warranties, including, but not limited to, the implied
// warranties of merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the Intel Corporation or contributors be liable for any direct,
// indirect, incidental, special, exemplary, or consequential damages
// (including, but not limited to, procurement of substitute goods or services;
// loss of use, data, or profits; or business interruption) however caused
// and on any theory of liability, whether in contract, strict liability,
// or tort (including negligence or otherwise) arising in any way out of
// the use of this software, even if advised of the possibility of such damage.
//
//M*/

#include "precomp.h"
#include "_cvaux.h"
#include <math.h>
#include "_cvutils.h"
#include "_cvwrap.h"

inline double weight(double x)
{
	return 1/(x*x + 1);
};

void icvBilateralFiltering8u(IplImage* in, IplImage* out, int thresh_space, int thresh_color);
void icvBilateralFiltering16s(IplImage* in, IplImage* out, int thresh_space, int thresh_color);

CV_IMPL void icvBilateralFiltering(IplImage* in, IplImage* out, int thresh_space, int thresh_color)
{
	switch(in->depth)
	{
	case IPL_DEPTH_8U:
		icvBilateralFiltering8u(in, out, thresh_space, thresh_color);
		break;

	case IPL_DEPTH_16S:
		icvBilateralFiltering16s(in, out, thresh_space, thresh_color);
		break;

	default:
		assert(0);
		return;
	}
}


void icvBilateralFiltering8u(IplImage* in, IplImage* out, int thresh_space, int thresh_color)
{
	uchar* idata;
	uchar* odata;
	uchar* _idata;
	uchar* _odata;
	int step;
	CvSize size;

	double sigma_color = thresh_color; 
	double sigma_space = thresh_space;

	double i2sigma_color = 1/(sigma_color*sigma_color);
	double i2sigma_space = 1/(sigma_space*sigma_space);	

	cvGetImageRawData(in, &_idata, &step, &size);
	cvGetImageRawData(out, &_odata, &step, &size);
	idata = _idata;
	odata = _odata;

	double mean1 = 0;
	double mean0 = 0;
	double w;

	int i, j;

#define INIT\
			color = idata[0];\
			mean0 = 0;mean1 = 0;

#define KERNEL_ELEMENT(m, n)\
			temp_color = idata[(m)*step + (n)];\
			w = (double)weight(-((m)*(m) + (n)*(n))*i2sigma_space - (temp_color - color)*(temp_color - color)*i2sigma_color);\
			mean0 += w;\
			mean1 += idata[(m)*step + (n)]*w;\

#define UPDATE_OUTPUT\
			*odata = (char)floor(mean1/mean0);

	int color, temp_color;

	// j = 0, i = 0;
	INIT;
	KERNEL_ELEMENT(0, 0);
	KERNEL_ELEMENT(0, 1);
	KERNEL_ELEMENT(1, 0);
	UPDATE_OUTPUT;
	idata++;
	odata++;

	for(i = 1; i < size.width - 1; i++, idata += 1, odata += 1)
	{
		INIT;
		KERNEL_ELEMENT(0, 0);
		KERNEL_ELEMENT(0, 1);
		KERNEL_ELEMENT(0, -1);
		KERNEL_ELEMENT(1, 0);
		KERNEL_ELEMENT(1, 1);
		KERNEL_ELEMENT(1, -1);
		UPDATE_OUTPUT;
	}


	// j = 0, i = size.width - 1
	INIT;
	KERNEL_ELEMENT(0, 0);
	KERNEL_ELEMENT(0, -1);
	KERNEL_ELEMENT(1, 0);
	UPDATE_OUTPUT;
	idata = _idata + step;
	odata = _odata + step;

	for(j = 1; j < size.height - 1; j++, _idata += step, idata = _idata, _odata += step, odata = _odata)
	{
		// i = 0;
		INIT;
		KERNEL_ELEMENT(0, 0);
		KERNEL_ELEMENT(0, 1);
		KERNEL_ELEMENT(-1, 0);
		KERNEL_ELEMENT(-1, 1);
		KERNEL_ELEMENT(1, 0);
		KERNEL_ELEMENT(1, 1);
		UPDATE_OUTPUT;

		for(int i = 1; i < size.width - 1; i++, idata += 1, odata += 1)
		{
			INIT;
			KERNEL_ELEMENT(0, 0);
			KERNEL_ELEMENT(1, 0);
			KERNEL_ELEMENT(-1, 0);
			KERNEL_ELEMENT(-1, 1);
			KERNEL_ELEMENT(0, 1);
			KERNEL_ELEMENT(1, 1);
			KERNEL_ELEMENT(-1, -1);
			KERNEL_ELEMENT(0, -1);
			KERNEL_ELEMENT(1, -1);
			UPDATE_OUTPUT;
		}

		// i = size.width - 1;
		INIT;
		KERNEL_ELEMENT(0, 0);
		KERNEL_ELEMENT(0, -1);
		KERNEL_ELEMENT(-1, 0);
		KERNEL_ELEMENT(-1, -1);
		KERNEL_ELEMENT(1, 0);
		KERNEL_ELEMENT(1, -1);
		UPDATE_OUTPUT;
	}

	// j = size.height - 1, i = 0;
	INIT;
	KERNEL_ELEMENT(0, 0);
	KERNEL_ELEMENT(0, 1);
	KERNEL_ELEMENT(-1, 0);
	UPDATE_OUTPUT;
	idata++;
	odata++;

	for(i = 1; i < size.width - 1; i++, idata += 1, odata += 1)
	{
		INIT;
		KERNEL_ELEMENT(0, 0);
		KERNEL_ELEMENT(0, 1);
		KERNEL_ELEMENT(0, -1);
		KERNEL_ELEMENT(-1, 0);
		KERNEL_ELEMENT(-1, 1);
		KERNEL_ELEMENT(-1, -1);
		UPDATE_OUTPUT;
	}
	// j = size.height - 1, i = size.width - 1
	INIT;
	KERNEL_ELEMENT(0, 0);
	KERNEL_ELEMENT(0, -1);
	KERNEL_ELEMENT(1, 0);
	UPDATE_OUTPUT;


#undef INIT
#undef KERNEL_ELEMENT
#undef UPDATE_OUTPUT
}

void icvBilateralFiltering16s(IplImage* in, IplImage* out, int thresh_space, int thresh_color)
{
	short* idata;
	short* odata;
	uchar* _idata;
	uchar* _odata;
	int step;
	CvSize size;

	double sigma_color = thresh_color; 
	double sigma_space = thresh_space;

	double i2sigma_color = 1/(sigma_color*sigma_color);
	double i2sigma_space = 1/(sigma_space*sigma_space);	

	cvGetImageRawData(in, &_idata, &step, &size);
	cvGetImageRawData(out, &_odata, &step, &size);
	idata = (short*)_idata;
	odata = (short*)_odata;

	double mean1 = 0;
	double mean0 = 0;
	double w;

	int i, j;

#define INIT\
			color = idata[0];\
			mean0 = 0;mean1 = 0;

#define KERNEL_ELEMENT(m, n)\
			temp_color = idata[(m)*step + (n)];\
			w = (double)weight(-((m)*(m) + (n)*(n))*i2sigma_space - (temp_color - color)*(temp_color - color)*i2sigma_color);\
			mean0 += w;\
			mean1 += idata[(m)*step + (n)]*w;\

#define UPDATE_OUTPUT\
			*odata = (short)floor(mean1/mean0);

	int color, temp_color;

	// j = 0, i = 0;
	INIT;
	KERNEL_ELEMENT(0, 0);
	KERNEL_ELEMENT(0, 1);
	KERNEL_ELEMENT(1, 0);
	UPDATE_OUTPUT;
	idata++;
	odata++;

	for(i = 1; i < size.width - 1; i++, idata += 1, odata += 1)
	{
		INIT;
		KERNEL_ELEMENT(0, 0);
		KERNEL_ELEMENT(0, 1);
		KERNEL_ELEMENT(0, -1);
		KERNEL_ELEMENT(1, 0);
		KERNEL_ELEMENT(1, 1);
		KERNEL_ELEMENT(1, -1);
		UPDATE_OUTPUT;
	}


	// j = 0, i = size.width - 1
	INIT;
	KERNEL_ELEMENT(0, 0);
	KERNEL_ELEMENT(0, -1);
	KERNEL_ELEMENT(1, 0);
	UPDATE_OUTPUT;
	idata = (short*)(_idata + step);
	odata = (short*)(_odata + step);

	for(j = 1; j < size.height - 1; j++, _idata += step, idata = (short*)_idata, _odata += step, odata = (short*)_odata)
	{
		// i = 0;
		INIT;
		KERNEL_ELEMENT(0, 0);
		KERNEL_ELEMENT(0, 1);
		KERNEL_ELEMENT(-1, 0);
		KERNEL_ELEMENT(-1, 1);
		KERNEL_ELEMENT(1, 0);
		KERNEL_ELEMENT(1, 1);
		UPDATE_OUTPUT;

		for(int i = 1; i < size.width - 1; i++, idata += 1, odata += 1)
		{
			INIT;
			KERNEL_ELEMENT(0, 0);
			KERNEL_ELEMENT(1, 0);
			KERNEL_ELEMENT(-1, 0);
			KERNEL_ELEMENT(-1, 1);
			KERNEL_ELEMENT(0, 1);
			KERNEL_ELEMENT(1, 1);
			KERNEL_ELEMENT(-1, -1);
			KERNEL_ELEMENT(0, -1);
			KERNEL_ELEMENT(1, -1);
			UPDATE_OUTPUT;
		}

		// i = size.width - 1;
		INIT;
		KERNEL_ELEMENT(0, 0);
		KERNEL_ELEMENT(0, -1);
		KERNEL_ELEMENT(-1, 0);
		KERNEL_ELEMENT(-1, -1);
		KERNEL_ELEMENT(1, 0);
		KERNEL_ELEMENT(1, -1);
		UPDATE_OUTPUT;
	}

	// j = size.height - 1, i = 0;
	INIT;
	KERNEL_ELEMENT(0, 0);
	KERNEL_ELEMENT(0, 1);
	KERNEL_ELEMENT(-1, 0);
	UPDATE_OUTPUT;
	idata++;
	odata++;

	for(i = 1; i < size.width - 1; i++, idata += 1, odata += 1)
	{
		INIT;
		KERNEL_ELEMENT(0, 0);
		KERNEL_ELEMENT(0, 1);
		KERNEL_ELEMENT(0, -1);
		KERNEL_ELEMENT(-1, 0);
		KERNEL_ELEMENT(-1, 1);
		KERNEL_ELEMENT(-1, -1);
		UPDATE_OUTPUT;
	}
	// j = size.height - 1, i = size.width - 1
	INIT;
	KERNEL_ELEMENT(0, 0);
	KERNEL_ELEMENT(0, -1);
	KERNEL_ELEMENT(1, 0);
	UPDATE_OUTPUT;

#undef INIT
#undef KERNEL_ELEMENT
#undef UPDATE_OUTPUT
	
}