#include "stdafx.h"

#include "fingers.h"
#include "basics/image_algorithms.h"
#include "basics/image_templates.h"
#include "algo_templates.h"
#include "opencv_utils.h"
#include "opencv_utils2.h"
#include "contour.h"
#include "basics/drawing_temp.h"
#include "hausdorff_matching.h"

#include "basics/definitions.h"


// Input arguments:
// color_image: a dilated image of the contour of a particular hand link,
// as detected using vLinkBoundaries, for a database image.
// point_rows, point_cols: the points on the boundary of that hand link.
// result: it is an output parameter, to which we store, at the end of 
// the functio, two line segments, both obtained by fitting a line
// to the boundary points in two different ways.
// First, we find the combination of center location (candidates are
// all points in point_rows and point cols) and orientation (among
// a uniformly-spaced fixed set of orientations) for which the line 
// segment with length 100 receives the most votes. We call that the
// "max line".
// The first line segment stored in result is the largest
// component of pixels along the max line.
// The second line segment is defined by the first and last edge pixels
// along the max line.
long vFingerLines(grayscale_image * the_image, vector<long> * point_rows, 
                   vector<long> * point_cols, vector<vLineSegment> * result)
{
  // For every point in the boundary and every orientation (from a 
  // uniformly-spaced set of orientations) compute the corresponding
  // line votes.
  v3dMatrix<double> * line_votes = vContourLineVotes(the_image,point_rows, point_cols);
  v3dMatrix<double> * pure_votes = vCopyBand2(line_votes, 0);

  long max_point_index, max_orientation_index;
  max_point_index = max_orientation_index = 0;
  // Find the center and orientation of the line segment with the highest
  // number of votes.
  // First find the corresponding indices in the pure_votes array.
  double max = function_image_maximum3(pure_votes, &max_point_index, &max_orientation_index);
  vArray3(double) votes3 = line_votes->Matrix3();
  // Now read the orientation and center location based on those indices.
  double max_orientation = votes3[1][max_point_index][max_orientation_index];
  double max_row = votes3[2][max_point_index][max_orientation_index];
  double max_col = votes3[3][max_point_index][max_orientation_index];
  vdelete(pure_votes);
  vArray2(uchar) edges = the_image->Matrix2(0);

  double row1, col1, row2, col2, row_step, col_step;
  long junk;
  row1 = col1 = row2 = col2 = row_step = col_step = 0;
  // Get the endpoints of the line segment centered at the 
  // max votes location and with max orientation, and length 100.
  vLineEnds(max_row, max_col, max_orientation, 100, 
             &row1, &col1, &row2, &col2, &row_step, &col_step, &junk);

  vPrint("(max_row, max_col) = (%lf, %lf)\n", max_row, max_col);
  long first_one = -1;
  long last_one = -1;
  long current_length = 0;
  long max_length = -1;
  long current_start = -1;
  long max_start = -1;
  long steps = 101;

  long step;
  // Find the largest connected component of points on the line segment,
  // and the first and last non-zero pixels on the line segment.
  for (step = 0; step < steps; step++)
  {
    // Get the current point on the segment.
    long row = vRound(row1 + step * row_step);
    long col = vRound(col1 + step * col_step);
    if (the_image->check_bounds(row, col) == 0) continue;
    vPrint("%li %li\n", row, col);
    // If the point is not an edge pixels, we should check if we 
    // have just reached the endpoint of the largest connected 
    // component of edge pixels seen so far along this line.
    if (edges[row][col] == 0)
    {
      // current start was set to be the last non-edge pixel seen
      // before the current pixel. We update it so it either 
      // points to the current pixel (and then length is 0) or
      // to the beginning of the first edge pixel we saw after
      // the last non-edge pixel we saw.
      current_start++;
      // This is the length (possibly zero) of the connected
      // component of pixels along the line that end with the 
      // previous pixel.
      current_length = step - current_start - 1;
      // If the current length is the largest seen so far, we 
      // store that info.
      if (current_length >= max_length)
      {
        max_length = current_length;
        max_start = current_start;
      }
      // This is the last non-edge pixel we see, till the next one.
      current_start = step;
    }
    else
    {
      // If the current pixel is an edge pixel, check if it is 
      // the first one we have seen so far. At any rate it is 
      // definitely the last one we have seen so far.
      if (first_one == -1)
      {
        first_one = step;
      }
      last_one = step;
    }
  }

  vLineSegment connected, longest;
  // Store the info about the maximum connected component along
  // the line with max votes.
  connected.row1 = vRound(row1 + max_start * row_step);
  connected.col1 = vRound(col1 + max_start * col_step);
  connected.row2 = vRound(row1 + (max_start + max_length) * row_step);
  connected.col2 = vRound(col1 + (max_start + max_length) * col_step);

  // Store the info about the line segment between the first and the last
  // edge pixel found along the line with max votes.
  longest.row1 = vRound(row1 + first_one * row_step);
  longest.col1 = vRound(col1 + first_one * col_step);
  longest.row2 = vRound(row1 + last_one * row_step);
  longest.col2 = vRound(col1 + last_one * col_step);
  
  result->push_back(connected);
  result->push_back(longest);

  vdelete(line_votes);

  return 1;
}


// Input arguments:
// color_image: a dilated image of the contour of a particular hand link,
// as detected using vLinkBoundaries, for a database image.
// point_rows, point_cols: the points on the boundary of that hand link.
// result: a 3D matrix, whose rows are equal to the number of points described
// by point_rows and point_cols, and whose columns are equal to the
// number of orientations we want to consider.
// It contains the following information:
// result[i][j][0]: the amount of pixel votes received by the line
// passing from point i at orientation indexed by j.
// result[i][j][1]: the orientation indexed by j.
// result[i][j][2]: the row of the i-th point: (*point_rows)[i].
// result[i][j][3]: the col of the i-th point: (*point_cols)[i].
v3dMatrix<double> * vContourLineVotes(grayscale_image * the_image, 
                                         vector<long> * point_rows,
                                         vector<long> * point_cols)
{
  // Specify how many orientations we want to consider.
  const double number_of_orientations = 48;
  long number_of_points = point_rows->size();
  
  // Create the result matrix and initialize entries to 0.
  v3dMatrix<double> * result_matrix = 
    new v3dMatrix<double>(number_of_points, (long) number_of_orientations, 4);
  vWriteValue(result_matrix, (double) 0);
  vArray3(double) result = result_matrix->Matrix3();

  long rows = the_image->Rows();
  long cols = the_image->Cols();

  // Create a votes matrix, that will be useful as an argument to 
  // vOrientedLines7b.
  v3dMatrix<long> * votes_matrix = new v3dMatrix<long>(rows, cols, 1);
  vArray2(long) votes = votes_matrix->Matrix2(0);

  long orientation_index;
  const double orientation_step = M_PI / number_of_orientations;
  // Specify how far we should look in counting votes for each line.
  const long length = 100;
  
  // Set number of votes to zero for pixels corresponding to
  // boundary points.
  long i;
  for (i = 0; i < number_of_points; i++)
  {
    long row = (*point_rows)[i];
    long col = (*point_cols)[i];
    votes[row][col] = 0;
  }
  
  for (orientation_index = 0; 
       orientation_index < number_of_orientations; 
       orientation_index++)
  {
    double orientation = orientation_index * orientation_step;

    // Count the votes
    vOrientedLines7b(the_image,point_rows, point_cols, votes_matrix, orientation, length);

    // Record the votes and reset the votes_matrix counters to 0
    for (i = 0; i < number_of_points; i++)
    {
      long row = (*point_rows)[i];
      long col = (*point_cols)[i];
      result[0][i][orientation_index] = votes[row][col];
      result[1][i][orientation_index] = orientation;
      result[2][i][orientation_index] = row;
      result[3][i][orientation_index] = col;
      // Reset the votes entry to 0, so that we can reuse the votes matrix
      // in calls to vOrientedLines7b with the same row, col, but 
      // different orientation.
      votes[row][col] = 0;
    }
  }

  vdelete(votes_matrix);
  return result_matrix;
}


// Input arguments:
// color_image: a database image of a hand.
// edge_image: a database edge image corresponding to color_image.
// Output arguments:
// point_rows_p, point_cols_p: each of them will receive
// an array of vectors. These two arrays of vectors, at 
// index [i][j], will contain the location of the j-th
// boundary point on the i-th link. Links have IDs 1 to 16.
// (the background has ID 0).
long vLinkBoundaries(color_image * the_image, grayscale_image * edge_image,
                      vArray(vector<long>) * point_rows_p, 
                      vArray(vector<long>) * point_cols_p)
{
  // max_link is 1 + the maximum number that we use to encode
  // the id of a link in an image.
  static const long max_link = 17;
  // The R band contains enough information to determine the
  // link id at every pixel.
  vArray2(uchar) R = the_image->GetR2();
  // We dilate the image, because for each boundary there are
  // at least two pixels that we need to find here: one pixel for
  // each link that ends at that boundary. The edge image only
  // has to encode one of those pixels (just because that was what
  // I chose for creating database edge images).
  grayscale_image * dilated = vDilateImage(edge_image, 1, 255);
  // Create arrays of vectors that will hold the coordinates
  // of each boundary point of each link.
  vArray(vector<long>) point_rows = vnew(vector<long>, max_link);
  vArray(vector<long>) point_cols = vnew(vector<long>, max_link);
  long rows = the_image->Rows();
  long cols = the_image->Cols();

  // near_rows[i] and near_cols[i], for i between 0 and 3, 
  // encode the relative displacement of each of the four neighbors
  // of a pixel with respect to that pixel.
  static long near_rows[4] = {-1, 1, 0, 0};
  static long near_cols[4] = {0, 0, -1, 1};

  vector<long> candidate_rows, candidate_cols;
  // Candidate boundary points are all non-zero points of the 
  // dilated image.
  vGetNonZeroPoints(dilated, &candidate_rows, &candidate_cols);
  long number = candidate_rows.size();

  long i, j;
  // Go through all candidate points.
  for (i = 0; i < number; i++)
  {
    // Read the current pixel location
    long row = candidate_rows[i];
    long col = candidate_cols[i];

    // Get the link id at the current pixel. If background, continue.
    long value = R[row][col];
    if (value == 0) continue;
    // Check if the pixel is actually a boundary pixel, i.e. if one of
    // its four neighbors that belong to a different link, or to the
    // background. Flag will be 1 if the pixel is a boundary pixel.
    long flag = 0;
    for (j = 0; j < 4; j++)
    {
      long row2 = row + near_rows[j];
      long col2 = col + near_cols[j];
      long value2 = R[row2][col2];
      if (value != value2) 
      {
        flag = 1;
        break;
      }
    }

    if (flag == 0) continue;
    // label is the link id of the pixel. It is value % 70
    // because that is the way I chose to encode link id
    // information in the database images (with the program
    // save_views).
    long label = value % 70;
    point_rows[label].push_back(row);
    point_cols[label].push_back(col);
  }

  // Save the arrays of vectors into the output arguments.
  *point_rows_p = point_rows;
  *point_cols_p = point_cols;
  vdelete(dilated);
  return 1;
}


// inputs: 
// color_image: it is assumed to be a database image of 
// a hand, with links colored in the way I color them in
// save_views (each link one color, link IDs encoded in the 
// red band).
// point_rows, point_cols: The i-th entries of them are
// vectors specifying the boundary points of the i-th link.
// Result:
// An image, where each boundary point is non-zero and all other
// points are zero. Each non-boundary point receives the color
// it had in the original image, except that its blue band
// is the sum of the blue band of color_image plus the link ID of
// the pixel. I don't know why I did it this way, I looked at
// the way the function is used and I did not find any particular
// use for that manipulation of the blue band.
color_image * vLinkBoundariesImage(color_image * the_image, vArray(vector<long>) point_rows, 
                                vArray(vector<long>) point_cols)
{
  long rows = the_image->Rows();
  long cols = the_image->Cols();
  color_image * result = new color_image(rows, cols);
  vArray2(uchar) in_R = the_image->GetR2();
  vArray2(uchar) in_G = the_image->GetG2();
  vArray2(uchar) in_B = the_image->GetB2();

  vArray2(uchar) out_R = result->GetR2();
  vArray2(uchar) out_G = result->GetG2();
  vArray2(uchar) out_B = result->GetB2();

  vWriteValue(result, (uchar) 0);

  static const long max_link = 17;
  long i;
  // Go through the boundary of each link.
  for (i = 1; i < max_link; i++)
  {
    // Go through all the points in the boundary of the current link.
    long size = point_rows[i].size();
    long j;
    for (j = 0; j < size; j++)
    {
      // Get the current point, and put the appropriate color for that 
      // point into the result image.
      long row = point_rows[i][j];
      long col = point_cols[i][j];
      out_R[row][col] = in_R[row][col];
      out_G[row][col] = in_G[row][col];
      out_B[row][col] = in_B[row][col] + (uchar) i;
    }
  }

  return result;
}




// A database hand image typically has every link colored in a different way,
// and the link id for a pixel is the remainder of dividing the red value
// of that pixel by 70. Here we add that id into the other bands, to make
// it easier to call vSelectRange to select one link.
long vLinkIDs(color_image * the_image)
{
  vArray2(uchar) values = the_image->Matrix2();
  long i;
  long size = the_image->Size();
  for (i = 0; i < size; i++)
  {
    uchar id = values[0][i] % 70;
    values[1][i] += id;
    values[2][i] += id;
  }

  return 1;
}


// See comments for vLinkIDs. Here we return an image that, at 
// each pixel, contains the link ID for that pixel.
grayscale_image * vLinkIDsB(color_image * the_image)
{
  static const long number_of_links = 17;
  grayscale_image * link_ids = (grayscale_image *) vCopyBand2(the_image,0);
  long size = link_ids->Size();
  vArray(uchar) ids = link_ids->Matrix();
  long i;
  // Extract the link ids.
  for (i = 0; i < size; i++)
  {
    ids[i] = ids[i] % 70;
    if (i == 484511)
    {
      vPrint("trash\n");
    }
  }

  return link_ids;
}


// Inputs: 
// link_ids: an image of a hand that contains, for each pixel,
// the link ID of the link that it belongs to.
// areas: an array that (we assume) is long enough to hold
// a number corresponding to each of the links. The number that
// is stored is the number of pixels that belong to that link.
long vLinkAreas(grayscale_image * link_ids, long * areas)
{
  const long number_of_links = 17;
  vArray(uchar) ids = link_ids->Matrix();
  long size = link_ids->Size();
  long i;
  // Zero out the counters.
  for (i = 0; i < number_of_links; i++)
  {
    areas[i] = 0;
  }

  // Count the area of each link.
  for (i = 0; i < size; i++)
  {
    long id = ids[i];
    areas[id]++;
  }
  return 1;
}


// See comments for vLinkAreas. 
// Inputs:
// link_ids: as in vLinkAreas.
// ellipse_params: An array that (we assume) is long enough to hold
// an ellipse for each link. That ellipse is the one that best
// fits the particular link.
// Note: This function could be much more efficient. Here we compute
// an entire image for each link ID, we could instead simply extract
// the points belonging to each link.
long vLinkEllipseParams(grayscale_image * link_ids, vEllipse * ellipse_params)
{
  const long number_of_links = 17;
  long i;
  for (i = 1; i < number_of_links; i++)
  {
    // Extract the pixels corresponding to the current link i.
    grayscale_image * selected = (grayscale_image *) vSelectRange(link_ids, 0, (uchar) i, (uchar) i);
    // Fit an ellipse to those pixels.
    ellipse_params[i] = vFitEllipse1(selected);
    vdelete(selected);
  }

  return 1;
}


// For each link we extract the contour. Multiple contours can
// be extracted for each finger, but each of the contours has
// to be connected.
// Inputs:
// color_image: a database image of a hand, where each link is 
// colored with a different color, that encodes the link's ID.
// point_rows, point_cols: they are as extracted using
// vLinkBoundaries. The i-th entries are vectors of boundary points
// of the i-th link.
// The result contains all the contours that we found.
// This function is a bit hard to follow, and it seems it is not
// used to get any other information, it is only called by
// interpreter2.cpp for visualization.
vector<vFingerContour> vFingerContours(color_image * the_image,
                                         vArray(vector<long>) point_rows, 
                                         vArray(vector<long>) point_cols)
{
  long rows = the_image->Rows();
  long cols = the_image->Cols();
  grayscale_image * scratch_image = new grayscale_image(rows, cols);
  vWriteValue(scratch_image, (uchar) 15);
  vArray2(uchar) scratch = scratch_image->Matrix2(0);

  // Create a gray scale image that can be passed to vContour::BorderContour
  // More simply (it seems to me),
  // grayscale_image * labels_image = vLinkIDsB(the_image).
  grayscale_image * labels_image = new grayscale_image(rows, cols);
  vArray(uchar) labels = labels_image->Matrix();
  vArray(uchar) red = the_image->GetR();
  long i;
  long size = rows * cols;
  for (i = 0; i < size; i++)
  {
    labels[i] = red[i] % 70;
  }

  static const long max_link = 17;
  vector<vFingerContour> result;

  for (i = 2; i < max_link; i++)
  {
    long size = point_rows[i].size();
    long j;

    // Mark the contour points of link i as available for use
    // in BorderContour
    for (j = 0; j < size; j++)
    {
      long row = point_rows[i][j];
      long col = point_cols[i][j];
      scratch[row][col] = 0;
    }

    // Get all the contours for link i
    for (j = 0; j < size; j++)
    {
      long row = point_rows[i][j];
      long col = point_cols[i][j];

      if (scratch[row][col] != 0) continue;
      vContour * contour = new vContour();
      contour->SetForegroundRange((uchar) i, (uchar) i);
      contour->BorderContour(labels_image, row, col, scratch_image);
      result.push_back(vFingerContour(i, contour));
    }
  }
  vdelete(scratch_image);
  vdelete(labels_image);

  return result;
}


// This function is not used anywhere else in the code, as of
// 3/14/2003.
// Overall, color_image is any binary image. We get the contour
// of the non-zero region, and then for every point on the 
// contour and every one of a fixed set of orientations we
// count votes for the segment centered at that point and 
// with that orientation. The "best contour line" is the line
// segment of the line with orientation equal, more or less,
// to the shape orientation of the binary region, that receives the 
// most votes, and that starts
// and ends at the first and last non-zero pixels of that line.
vLineSegment vBestContourLine(grayscale_image * the_image)
{
  // Get the orientation of the binary shape.
  double orientation = vShapeOrientation(the_image);
  // Get the contour.
  vContour contour;
  contour.OuterContour(the_image);
  vector<long> * rows = contour.GetYsLong();
  vector<long> * cols = contour.GetXsLong();
  // Get the votes for all center/orientation combinations.
  v3dMatrix<double> * votes = vContourLineVotes(the_image,contour.GetYsLong(), contour.GetXsLong());
  vdelete(rows);
  vdelete(cols);

  v3dMatrix<double> * pure_votes = vCopyBand2(votes, 0);
  long max_point_index, max_orientation_index;
  max_point_index = max_orientation_index = 0;
  long number_of_orientations = votes->Cols();
  // Compute the orientation index corresponding to the shape
  // orientation.
  long orientation_index = vRound(orientation * number_of_orientations / M_PI);
  // Find the line with most votes and that orientation.
  double max = color_imageColMax3(pure_votes, orientation_index, &max_point_index);
  long max_row = vRound(votes->Matrix3()[2][max_point_index][orientation_index]);
  long max_col = vRound(votes->Matrix3()[3][max_point_index][orientation_index]);
  vLineSegment line = vProjectionEndpoints(the_image,max_row, max_col, orientation);
  return line;
}


// Input: the_image,a color image belonging to the hand database. The id
// of the link that a pixel belongs to is the remainder of dividing its
// red band by 70.
// Goal: To find the links (or finger/subfinger groups of links) that 
// look straight and are likely to give nice, easily detectable straight lines
// in input images.
// Output: A vector of such links (or link groupings), described by
// vStraightLinks objects.
// Note: There is an alternative function, vGoodLinksForLines2, that
// I think is doing a better job.
long vGoodLinksForLines(color_image * the_image, vector<vStraightLinks> * links)
{
  static const long number_of_links = 17;
  // Will hold the number of pixels for every link
  long areas[number_of_links];
  vEllipse ellipses[number_of_links];

  // Convert the color image to a gray scale image where every pixel has
  // a value equal to the id of the link it belongs to.
  grayscale_image * link_ids = vLinkIDsB(the_image);
  // Count the areas
  vLinkAreas(link_ids, areas);
  // Get width, length, elongation, orientation, for every link
  vLinkEllipseParams(link_ids, ellipses);

  long i;
  // Print out the gathered info
  for (i = 2; i < number_of_links; i++)
  {
    vPrint("\nLink %li: ", i);
    vPrint("Area = %li\n", areas[i]);
    if ((ellipses[i].width == 0) || (ellipses[i].length == 0)) continue;
    vPrint("Length = %lf\nWidth = %lf\nElongation = %lf\nlength2 = %lf\nwidth2 = %lf\n",
            ellipses[i].length, ellipses[i].width, ellipses[i].length/ellipses[i].width,
            ((double) areas[i]) / ellipses[i].width, ((double) areas[i]) / ellipses[i].length);
    vPrint("Orientation = %lf\n", ellipses[i].angle);
    vPrint("\n");
  }

  // We have some heuristics for what links to keep and what links to reject.
  // One heuristic says that the width of a straight link should not be less 
  // than a given fraction of the maximum width of a straight link 
  // found in this image. Similarly, the area of a link should not be less
  // than a given fraction of the area of the largest straight link found in
  // this picture. Because of that, we need to find the maximum width
  // and maximum area among all links.
  double max_width, max_area;
  max_width = max_area = 0;
  for (i = 3; i < number_of_links; i++)
  {
    if (areas[i] < 100) continue;
    if (ellipses[i].width < 5) continue;
    // Must not be too short (so that we know which axis is the major axis)
    // or too long (that would mean that it is partially occluded)
    double elongation = ellipses[i].length / ellipses[i].width;
    if ((elongation  < 1.5) || (elongation > 4)) continue;
    // Because of model inaccuracies or occlusion, sometimes a link shows
    // as two connected components. In that case the width of the fitted 
    // ellipse is much less than pixel area/length of fitted ellipse.
    // At any rate, we should reject such links.
    if (ellipses[i].width >= (areas[i] / ellipses[i].length)) continue;
    if (ellipses[i].length >= (areas[i] / ellipses[i].width)) continue;
    if (ellipses[i].width > max_width) max_width = ellipses[i].width;
    if (areas[i] > max_area) max_area = areas[i];
  }

  // Now, for every link, pass it through a number of heuristic tests, and 
  // keep it if it passes all tests.
  for (i = 3; i < number_of_links; i++)
  {
    // Must exceed a minimum area and width
    if (areas[i] < 100) continue;
    if (ellipses[i].width < 5) continue;
    // Must not be too short (so that we know which axis is the major axis)
    // or too long (that would mean that it is partially occluded)
    double elongation = ellipses[i].length / ellipses[i].width;
    if ((elongation  < 1.5) || (elongation > 4)) continue;
    // Because of model inaccuracies or occlusion, sometimes a link shows
    // as two connected components. In that case the width of the fitted 
    // ellipse is much less than pixel area/length of fitted ellipse.
    // At any rate, we should reject such links.
    if (ellipses[i].width >= (areas[i] / ellipses[i].length)) continue;
    if (ellipses[i].length >= (areas[i] / ellipses[i].width)) continue;
    // The area and width must be more than a given fraction of the
    // max area and max width found among straight links.
    if (areas[i] < max_area/5) continue;
    if (ellipses[i].width < max_width/2) continue;
    // For the outer links in fingers we require a lower fraction of
    // max width and max area
    if ((ellipses[i].width < max_width / 1.5) && (i > 4) && (i%3 != 1))
    {
      continue;
    }
    if ((areas[i] < max_area / 2.5) && ((i == 3) || (i == 5) || (i == 8) || (i == 11) || (i == 14)))
    {
      continue;
    }
    vStraightLinks link;
    link.link1 = i;
    link.link2 = i;
    link.link3 = i;
    link.quality = 1;
//    vPrint("Checking out link %li-%li\n", link.link1, link.link3);
    long success = vEstimateFingerLines(link_ids, &link);
    if (success > 0)
    {
      links->push_back(link);
    }
  }
  
  vPrint("max_area = %lf, max_width = %lf\n", max_area, max_width);
  vdelete(link_ids);
  return 1;
}


// Input: A gray scale image, where every pixel is labeled with the id
// of the hand link that it belongs to, according to the conventions 
// followed in the hand database. Also, a vStraightLinks pointer,
// that describes the link or group of links that we are interested in.
// Output: two line segments (line1 and line2 in the vStraightLinks object) 
// that are close to parallel to the major axis of the straight link, 
// and that more or less go along the contour of the link.
// Note: there is an alternative function, called vEstimateFingerLines2,
// which I believe is more appropriate for this task, because it also
// takes as input an edge image. Some contour points of the link should
// not be considered, because they do not really correspond to edges, they
// just connect the link to its adjacent link.
long vEstimateFingerLines(grayscale_image * link_ids, vStraightLinks * finger)
{
  // Select the straight link out of the image
  grayscale_image * selected = 
    (grayscale_image *) vSelectRange(link_ids, 0, (uchar) finger->link1, 
                                  (uchar) finger->link3);
  // Fit an ellipse
  vEllipse ellipse = vFitEllipse1(selected);
  vector<Label> labels;
  // Find the connected components of the link
  v3dMatrix<long> * components = selected->ConnectedComponents(&labels);
  vdelete(selected);
  if (labels.size() < 2)
  {
    vdelete(components);
    return 0;
  }

  // Choose the largest component
  std::sort(labels.begin(), labels.end(), LabelMore());
  grayscale_image * largest_component = vSelectValue(components, (long) labels[1].id);
  vdelete(components);
  // Get the contour of the largest component
  vContour contour;
  contour.OuterContour(largest_component);
  vdelete(largest_component);
  // Choose from the contour pixels whose orientations are close to parallel
  // to the major axis of the fitted ellipse, and divide those pixels into
  // two groups, one on the left half plane of the major axis, and one on
  // the right half plane of the major axis.
  vector<double> rows1, cols1, rows2, cols2;
  long success = vFindOuterLines(&contour, ellipse.center_row, 
                                  ellipse.center_col, ellipse.angle, 
                                  &rows1, &cols1, &rows2, &cols2);
  if (success == 0) return 0;
  // Fit a line segment to each group of pixels
  finger->line1 = vFitLineSegment(&rows1, &cols1);
  finger->line2 = vFitLineSegment(&rows2, &cols2);

  return 1;
}


// Given a line of slope=angle, that goes through row and col, choose
// from the contour points whose orientations (expressed as first 
// derivatives) are more or less parallel to the line (within 
// some given tolerance). Furthermore, divide those pixels into
// two groups, one for each half plane that the line defines.
long vFindOuterLines(vContour * contour, double row, double col, double angle,
                      vector<double> * rows1, vector<double> * cols1,
                      vector<double> * rows2, vector<double> * cols2)
{
  // tolerance (in radians) within which the orientation of the pixels
  // has to be with respect to the angle parameter.
  static const double tolerance = .6;
  // Specifies how many points away we should look to compute
  // the first derivative.
  static const long di = 3;

  vector<double> * rows =  contour->GetYs();
  vector<double> * cols = contour->GetXs();
  long number = rows->size();
  if (number < 20) return 0;

  long i;
  for (i = 0; i < number; i++)
  {
    // if, for some reason, we should not use the current point, it will have
    // values -row, -col for row and col.
    if ((*rows)[i] < 0) continue;

    // Get the indices of the neighbors of the current point that we should 
    // use to calculate the derivative (i.e. the orientation of the pixel).
    long i1 = (i+number-di) % number;
    long i2 = (i+di) % number;
    double row1 = vAbs((*rows)[i1]);
    double col1 = vAbs((*cols)[i1]);
    double row2 = vAbs((*rows)[i2]);
    double col2 = vAbs((*cols)[i2]);
    // Get the orientation of the current point.
    double orientation = vAngle(col2-col1, row2-row1);
    // Check if the orientation is acceptable.
    if (vCircularDistance(angle, orientation, M_PI) > tolerance) continue;

    // Check which side of the middle line the pixel is on.
    long plane = vFindHalfPlane(col, row, angle, (*cols)[i], (*rows)[i]);
    // Store the pixel in the appropriate set.
    if (plane == -1) 
    {
      rows1->push_back((*rows)[i]);
      cols1->push_back((*cols)[i]);
    }
    else if (plane == 1)
    {
      rows2->push_back((*rows)[i]);
      cols2->push_back((*cols)[i]);
    }
  }
  return 1;
}


// This is a wrapper around vFitLine2D, which itself is a 
// wrapper around the corresponding OpenCV function. 
// vFitLine2D gives us a line, but we actually want a 
// line segment, with specific endpoints. Those endpoints
// are simply the first and last point on the fitted line
// where points specified in rows and cols are projected.
// This function fits a line segment into the points specified
// by rows and cols.
vLineSegment vFitLineSegment(vector<double> * rows, vector<double> * cols)
{
  double row, col, angle;
  row = col = angle = 0;
  long max_index = 0;
  double max_row = 0;
  double max_col = 0;
  // Get the maximum values for coordinates occurring in rows and cols.
  function_vector_maximum(rows, &max_index, &max_row);
  function_vector_maximum(cols, &max_index, &max_col);

  // Convert coordinates into integers.
  vector<long> long_rows, long_cols;
  vDoubleToLongVector(&long_rows, rows);
  vDoubleToLongVector(&long_cols, cols);
  // Create an image large enough to fit all the points. Mark all
  // those points as non-zero in the image.
  grayscale_image * the_image = new grayscale_image((long) max_row+3, (long) max_col+3);
  vWriteValue(the_image,(uchar) 0);
  vWriteValue2(the_image,&long_rows, &long_cols, (uchar) 255);

  // Fit a line to the points.
  long success = vFitLine2D(rows, cols, &row, &col, &angle, CV_DIST_WELSCH);
  if (success <= 0) 
  {
    vLineSegment failed;
    failed.row1 = -1;
    vdelete(the_image);
    return failed;
  }
  // Figure out the endpoints of the fitted line.
  vLineSegment result = vProjectionEndpoints(the_image,(long) row, (long) col, angle);
  vdelete(the_image);

  return result;
}


// Here we also pass in as input an edge image. A pixel in color_image should
// not be considered an edge pixel unless it is within distance 1 of
// an edge pixel in edges. Also, vGoodLinksForLines doesn't try
// to group links into larger groups. Here we try to do that.
long vGoodLinksForLines2(color_image * the_image, grayscale_image * edges, 
                          vector<vStraightLinks> * links)
{
  static const long number_of_links = 17;
  // Will hold the number of pixels for every link
  long areas[number_of_links];
  vEllipse ellipses[number_of_links];

  // Convert the color image to a gray scale image where every pixel has
  // a value equal to the id of the link it belongs to.
  grayscale_image * link_ids = vLinkIDsB(the_image);
  // Count the areas
  vLinkAreas(link_ids, areas);
  // Get width, length, elongation, orientation, for every link
  vLinkEllipseParams(link_ids, ellipses);

  long i;
  // Print out the gathered info
  for (i = 2; i < number_of_links; i++)
  {
    vPrint("\nLink %li: ", i);
    vPrint("Area = %li\n", areas[i]);
    if ((ellipses[i].width == 0) || (ellipses[i].length == 0) ||
        (areas[i] == 0)) continue;
    vPrint("Length = %lf\nWidth = %lf\nElongation = %lf\nlength2 = %lf\nwidth2 = %lf\n",
            ellipses[i].length, ellipses[i].width, ellipses[i].length/ellipses[i].width,
            ((double) areas[i]) / ellipses[i].width, ((double) areas[i]) / ellipses[i].length);
    vPrint("Orientation = %lf\n", ellipses[i].angle);
    vPrint("\n");
  }

  // We have some heuristics for what links to keep and what links to reject.
  // One heuristic says that the width of a straight link should not be less 
  // than a given fraction of the maximum width of a straight link 
  // found in this image. Similarly, the area of a link should not be less
  // than a given fraction of the area of the largest straight link found in
  // this picture. Because of that, we need to find the maximum width
  // and maximum area among all links.
  double max_width, max_area;
  max_width = max_area = 0;
  for (i = 3; i < number_of_links; i++)
  {
    if (areas[i] < 100) continue;
    if (ellipses[i].width < 5) continue;
    // Must not be too short (so that we know which axis is the major axis)
    // or too long (that would mean that it is partially occluded)
    double elongation = ellipses[i].length / ellipses[i].width;
    if ((elongation  < 1.5) || (elongation > 4)) continue;
    // Because of model inaccuracies or occlusion, sometimes a link shows
    // as two connected components. In that case the width of the fitted 
    // ellipse is much less than pixel area/length of fitted ellipse.
    // At any rate, we should reject such links.
    if (ellipses[i].width >= (areas[i] / ellipses[i].length)) continue;
    if (ellipses[i].length >= (areas[i] / ellipses[i].width)) continue;
    if (ellipses[i].width > max_width) max_width = ellipses[i].width;
    if (areas[i] > max_area) max_area = areas[i];
  }

  // Now, for every link, pass it through a number of heuristic tests, and 
  // keep it if it passes all tests.
  for (i = 3; i < number_of_links; i++)
  {
    // Must exceed a minimum area and width
    if (areas[i] < 100) continue;
    if (ellipses[i].width < 5) continue;
    // Must not be too short (so that we know which axis is the major axis)
    // or too long (that would mean that it is partially occluded)
    double elongation = ellipses[i].length / ellipses[i].width;
    if ((elongation  < 1.5) || (elongation > 4)) continue;
    // Because of model inaccuracies or occlusion, sometimes a link shows
    // as two connected components. In that case the width of the fitted 
    // ellipse is much less than pixel area/length of fitted ellipse.
    // At any rate, we should reject such links.
    if (ellipses[i].width >= (areas[i] / ellipses[i].length)) continue;
    if (ellipses[i].length >= (areas[i] / ellipses[i].width)) continue;
    // The area and width must be more than a given fraction of the
    // max area and max width found among straight links.
    if (areas[i] < max_area/5) continue;
    if (ellipses[i].width < max_width/2) continue;
    // For the outer links in fingers we require a lower fraction of
    // max width and max area
    if ((ellipses[i].width < max_width / 1.5) && (i > 4) && (i%3 != 1))
    {
      continue;
    }
    if ((areas[i] < max_area / 2.5) && ((i == 3) || (i == 5) || (i == 8) || (i == 11) || (i == 14)))
    {
      continue;
    }
    vStraightLinks link;
    link.link1 = i;
    link.link2 = i;
    link.link3 = i;
    link.ellipse = ellipses[i];
    link.quality = 1;
//    vPrint("Checking out link %li-%li\n", link.link1, link.link3);
    long success = vEstimateFingerLines2(link_ids, edges, &link);
    if (success > 0)
    {
      links->push_back(link);
    }
  }

  vMergeLinks(link_ids, edges, links, areas, ellipses);
  // vGroupUnusedLinks is just a placeholder right now, it still needs
  // to be implemented.
  vGroupUnusedLinks(link_ids, edges, links);

  vPrint("max_area = %lf, max_width = %lf\n", max_area, max_width);
  vdelete(link_ids);
  return 1;
}


// Like vEstimateFingerLines, but here we also take an edge image
// as input (see comments for vGoodLinksForLines2).
long vEstimateFingerLines2(grayscale_image * link_ids, grayscale_image * edges,
                            vStraightLinks * finger)
{
  // Select the straight link out of the image
  grayscale_image * selected = 
    (grayscale_image *) vSelectRange(link_ids, 0, (uchar) finger->link1, 
                                  (uchar) finger->link3);
  // Fit an ellipse
  vEllipse ellipse = vFitEllipse1(selected);
  vector<Label> labels;
  // Find the connected components of the link
  v3dMatrix<long> * components = selected->ConnectedComponents(&labels);
  vdelete(selected);
  if (labels.size() < 2)
  {
    vdelete(components);
    return 0;
  }

  // Choose the largest component
  std::sort(labels.begin(), labels.end(), LabelMore());
  grayscale_image * largest_component = vSelectValue(components, (long) labels[1].id);
  vdelete(components);
  // Get the contour of the largest component
  vContour contour;
  contour.OuterContour(largest_component);

  // Go through the contour and overwrite points that are zero in the edge image
  vector<double> * xs = contour.GetXs();
  vector<double> * ys = contour.GetYs();
  long number = xs->size();
  long i;
  vArray2(uchar) edges2 = edges->Matrix2(0);
  for (i = 0; i < number; i++)
  {
    long row = vRound((*ys)[i]);
    long col = vRound((*xs)[i]);
    if (edges2[row][col] == 0)
    {
      (*ys)[i] = -row;
      (*xs)[i] = -col;
    }
  }

  vdelete(largest_component);
  // Choose from the contour pixels whose orientations are close to parallel
  // to the major axis of the fitted ellipse, and divide those pixels into
  // two groups, one on the left half plane of the major axis, and one on
  // the right half plane of the major axis.
  vector<double> rows1, cols1, rows2, cols2;
  long success = vFindOuterLines(&contour, ellipse.center_row, 
                                  ellipse.center_col, ellipse.angle, 
                                  &rows1, &cols1, &rows2, &cols2);
  if (success == 0) return 0;
  // Fit a line segment to each group of pixels
  finger->line1 = vFitLineSegment(&rows1, &cols1);
  finger->line2 = vFitLineSegment(&rows2, &cols2);

  return 1;
}


// Decide if finger links corresponding to the same finger should be merged.
// The result is:
// 0 if no merging should be done.
// 1 if the inner links should be merged
// 2 if the outer links should be merged
// 3 if all three links should be merged.
long vMergeLinkTest(grayscale_image * link_ids, grayscale_image * edges,
                     vector<vStraightLinks> * links, long index,
                     long * areas, vEllipse * ellipses)
{
  long merge_inner = vMergeLinkTest2(link_ids, edges, links, index-1, index,
                                      areas, ellipses);
  long merge_outer = vMergeLinkTest2(link_ids, edges, links, index, index+1,
                                      areas, ellipses);
  return 2 * merge_outer + merge_inner;
}


// Decide if two consecutive links should be merged.
long vMergeLinkTest2(grayscale_image * link_ids, grayscale_image * edges,
                      vector<vStraightLinks> * links, long index1, long index2,
                      long * areas, vEllipse * ellipses)
{
//  vPrint("MergeLinkTest2(%li, %li)\n", index1, index2);

  // Check which links are present in the argument links.
  // At the same time, find the max area among all links
  // and the max width among all links.
  long link_index1, link_index2;
  link_index1 = link_index2 = -1;
  long number = links->size();
  double max_area, max_width;
  max_area = max_width = 0;
  long i;
  for (i = 0; i < number; i++)
  {
    long link_index = (*links)[i].link1;
    if (link_index == index1)
    {
      link_index1 = i;
    }
    else if (link_index == index2)
    {
      link_index2 = i;
    }
    if (areas[link_index] >= max_area)
    {
      max_area = areas[link_index];
    }
    if (ellipses[link_index].width < max_width)
    {
      max_width = ellipses[link_index].width;
    }
  }

  long merge_flag = 0;

  // If neither link was identified, we don't merge them.
  // I'm not sure if this is the right thing to do. Maybe
  // we should pass the unified link as input to 
  // vGoodLinksForLines2.
  if ((link_index1 == -1) && (link_index2 == -1))
  {
    return 0;
  }

  // If link1 is there but link2 is missing:
  else if ((link_index1 >= 0) && (link_index2 < 0))
  {
    vStraightLinks link  = (*links)[link_index1];
    // Check if link2 satisfies some minimum width and area requirements.
    long link_test = vLinkTest1(index2, areas, ellipses, max_area, max_width);
    if (link_test > 0)
    {
      vStraightLinks combined = vGetStraightLinks(link_ids, edges, index1, index2);
      // Check if the combined link has lines that are close enough to parallel 
      // to the lines of the first link.
      if (vParallelTest1(link, combined))
      {
        merge_flag = 1;
      }
    }
  }

  // Same thing if link1 is not there and link2 is there.
  else if ((link_index1 < 0) && (link_index2 >= 0))
  {
    vStraightLinks link = (*links)[link_index2];
    long link_test = vLinkTest1(index1, areas, ellipses, max_area, max_width);
    if (link_test > 0)
    {
      vStraightLinks combined = vGetStraightLinks(link_ids, edges, index1, index2);
      if (vParallelTest1(link, combined))
      {
        merge_flag = 1;
      }
    }
  }

  // Here check if the lines of the two links are parallel
  // to each other.
  else if ((link_index1 >= 0) && (link_index2 >= 0))
  {
    vStraightLinks link1 = (*links)[link_index1];
    vStraightLinks link2 = (*links)[link_index2];
    if (vParallelTest2(link1, link2))
    {
        merge_flag = 1;
    }
  }

  return merge_flag;
}


// Determine which links should be merged.
// links is both an input and an output argument. At the 
// end of the function it contains links that were not 
// merged and the result of links that were merged.
long vMergeLinks(grayscale_image * link_ids, grayscale_image * edges, 
                  vector<vStraightLinks> * links, 
                  long * areas, vEllipse * ellipses)
{
  long i;
  vector<vStraightLinks> new_links;
  static const long number_of_links = 17;
  long number1 = links->size();
  // For every finger, check if some and which of its links
  // should be merged.
  for (i = 3; i < number_of_links; i+=3)
  {
    long merge_flag = vMergeLinkTest(link_ids, edges, links, i,
                                      areas, ellipses);
    long low = i;
    long high = i;
    if (merge_flag == 0) continue;
    // FIgure out which links should be merged.
    else if (merge_flag == 1)
    {
      low = i-1;
    }
    else if (merge_flag == 2)
    {
      high = i+1;
    }
    else if (merge_flag == 3)
    {
      low = i-1;
      high = i+1;
    }
    vPrint("Merging links %li %li\n",  low, high);
    vStraightLinks merged = vGetStraightLinks(link_ids, edges, low, high);
    new_links.push_back(merged);
  }

  // Combine links and new links.
  vector<vStraightLinks> result_links;
  long number2 = new_links.size();
  for (i = 3; i < 17; i++)
  {
    long index1 = -1;
    long index2 = -1;
    long j;
    // Look for the position of link i in links.
    for (j = 0; j < number1; j++)
    {
      if ((*links)[j].link1 == i)
      {
        index1 = j;
        break;
      }
    }

    // Look for the position of link i in new_links.
    for (j = 0; j < number2; j++)
    {
      if ((i >= new_links[j].link1) && (i <= new_links[j].link3))
      {
        index2 = j;
        break;
      }
    }

    // If the link is in neither links nor new_links, do nothing.
    if ((index1 == -1) && (index2 == -1)) continue;
    // If the link is only in new_links, include it, but make sure
    // we don't include it twice (and check that the lines are
    // well-defined).
    else if ((index1 == -1) && (index2 >= 0))
    {
      if (new_links[index2].link1 != i) continue;
      if ((new_links[index2].line1.row1 < 0) && (new_links[index2].line2.row1 < 0))
      {
        continue;
      }
      result_links.push_back(new_links[index2]);
    }
    // If the link is only in links, include it as long as the 
    // lines are well defined.
    else if ((index1 >= 0) && (index2 == -1))
    {
      if (((*links)[index1].line1.row1 < 0) && ((*links)[index1].line2.row1 < 0))
      {
        continue;
      }
      result_links.push_back((*links)[index1]);
    }
    // If the link is in both links and new_links, make sure it is not added
    // twice, and that the that the new link lines are well defined. If yes,
    // add the new link. If the lines are not well defined, add the old link,
    // as long as the lines of the long link are well defined.
    else if ((index1 >= 0) && (index2 >= 0))
    {     
      long bad1 = ((*links)[index1].line1.row1 < 0) && ((*links)[index1].line2.row1 < 0);
      long bad2 = (new_links[index2].line1.row1 < 0) && (new_links[index2].line2.row1 < 0);
      if ((bad1 != 0) && (bad2 != 0)) continue;
      else if (bad2 == 0)
      {
        if (new_links[index2].link1 == i)
        {
          result_links.push_back(new_links[index2]);
        }
      }
      else
      {
        result_links.push_back((*links)[index1]);
      }
    }
  }
  links->erase(links->begin(), links->end());
  links->insert(links->begin(), result_links.begin(), result_links.end());
  return 1;
}


long vGroupUnusedLinks(grayscale_image * link_ids, grayscale_image * edges, 
                        vector<vStraightLinks> * links)
{
  return 0;
}


// Determine if the link is good enough to be a part of a
// straight links group (doesn't need special elongation, 
// but needs min width requirements).
long vLinkTest1(long i, long * areas, vEllipse * ellipses, 
                 double max_area, double max_width)
{
  static const long number_of_links = 17;
  if (i < 3) return 0;
  if (i >= number_of_links) return 0;
  // Must exceed a minimum area and width
  if (areas[i] < 100) return 0;
  if (ellipses[i].width < 5) return 0;
  // Must not be too short (so that we know which axis is the major axis)
  // or too long (that would mean that it is partially occluded)
  // The area and width must be more than a given fraction of the
  // max area and max width found among straight links.
  if (areas[i] < max_area/10) return 0;
  if (ellipses[i].width < max_width/2) return 0;
  // For the outer links in fingers we require a lower fraction of
  // max width and max area. For the inner links the criteria are
  // more stringent.
  if ((ellipses[i].width < max_width / 1.5) && (i > 4) && (i%3 != 1))
  {
    return 0;
  }
  if ((areas[i] < max_area / 4) && ((i == 3) || (i == 5) || (i == 8) || (i == 11) || (i == 14)))
  {
    return 0;
  }
  return 1;
}


// Create the vStraightLinks object that best describes the group
// of links with ids from low to high.
vStraightLinks vGetStraightLinks(grayscale_image * link_ids, grayscale_image * edges,
                                   long low, long high)
{
  vStraightLinks link;
  link.link1 = low;
  link.link2 = (low+high)/2;
  link.link3 = high;
//  vPrint("Checking out link %li-%li\n", link.link1, link.link3);
  link.quality = vEstimateFingerLines2(link_ids, edges, &link);
  return link;
}


// Determine if two straight link groups, one consisting of a single link,
// and the second consisting of that same link plus one or two more, are
// such that the second is a natural extension of the first. This is what
// we use to decide if a single link group should be replaced by an extended
// group of two or three links that includes the original link.
long vParallelTest1(vStraightLinks link1, vStraightLinks link2)
{
  vPrint("ParallelTest1(%li-%li, %li-%li)\n", 
          link1.link1, link1.link3, link2.link1, link2.link3);

  // For each pair of line segments such that one comes from link1 and the
  // other comes from link2, verify that they have similar orientations and
  // one endpoint of one is close to an endpoint of the other one.
  long result = 0;
  if (vLineSegmentTest1(link1.line1, link2.line1) > 0)
  {
    return 1;
  }
  if (vLineSegmentTest1(link1.line1, link2.line2) > 0)
  {
    return 1;
  }
  if (vLineSegmentTest1(link1.line2, link2.line1) > 0)
  {
    return 1;
  }
  if (vLineSegmentTest1(link1.line2, link2.line2) > 0)
  {
    return 1;
  }
  return result;
}

    
// Determine if two disjoint straight link groups should be merged
// into one.    
long vParallelTest2(vStraightLinks link1, vStraightLinks link2)
{
  vPrint("ParallelTest2(%li-%li, %li-%li)\n\n", 
           link1.link1, link1.link3, link2.link1, link2.link3);
  long result = 0;
  if (vLineSegmentTest2(link1.line1, link2.line1) > 0)
  {
    return 1;
  }
  if (vLineSegmentTest2(link1.line1, link2.line2) > 0)
  {
    return 1;
  }
  if (vLineSegmentTest2(link1.line2, link2.line1) > 0)
  {
    return 1;
  }
  if (vLineSegmentTest2(link1.line2, link2.line2) > 0)
  {
    return 1;
  }
  return result;
}
    

// Return true if the two lines are close to parallel, one endpoint
// of line1 is close to one endpoint of line2 and the other endpoint
// of line1 is relatively far from the other endpoint of line2.
long vLineSegmentTest1(vLineSegment line1, vLineSegment line2)
{
  if (line1.row1 < 0) return 0;
  if (line2.row1 < 0) return 0;

  double angle1 = vAngle(line1.col2 - line1.col1, line1.row2 - line1.row1);
  double angle2 = vAngle(line2.col2 - line2.col1, line2.row2 - line2.row1);
  static const double angle_diff_limit = 0.3;
  double distance = vCircularDistance(angle1, angle2, (double) M_PI);

  if (vAbs(distance) > angle_diff_limit) return 0;

  if (vEndpointTest1(line1.row1, line1.col1, line2.row1, line2.col1,
                      line1.row2, line1.col2, line2.row2, line2.col2) > 0)
  {
    return 1;
  }
  if (vEndpointTest1(line1.row1, line1.col1, line2.row2, line2.col2,
                      line1.row2, line1.col2, line2.row1, line2.col1) > 0)
  {
    return 1;
  }
  if (vEndpointTest1(line1.row2, line1.col2, line2.row1, line2.col1,
                      line1.row1, line1.col1, line2.row2, line2.col2) > 0)
  {
    return 1;
  }
  if (vEndpointTest1(line1.row2, line1.col2, line2.row2, line2.col2,
                      line1.row1, line1.col1, line2.row1, line2.col1) > 0)
  {
    return 1;
  }
  return 0;
}


// Return true if the two lines are close to parallel, their start points
// are close to each other and the lines go to opposite directions
// from the start points.
long vLineSegmentTest2(vLineSegment line1, vLineSegment line2)
{
  if (line1.row1 < 0) return 0;
  if (line2.row1 < 0) return 0;

  double angle1 = vAngle(line1.col2 - line1.col1, line1.row2 - line1.row1);
  double angle2 = vAngle(line2.col2 - line2.col1, line2.row2 - line2.row1);
  static const double angle_diff_limit = 0.3;
  double distance = vCircularDistance(angle1, angle2, (double) M_PI);

  if (vAbs(distance) > angle_diff_limit) return 0;

  if (vEndpointTest2(line1.row1, line1.col1, line2.row1, line2.col1,
                      line1.row2, line1.col2, line2.row2, line2.col2) > 0)
  {
    return 1;
  }
  if (vEndpointTest2(line1.row1, line1.col1, line2.row2, line2.col2,
                      line1.row2, line1.col2, line2.row1, line2.col1) > 0)
  {
    return 1;
  }
  if (vEndpointTest2(line1.row2, line1.col2, line2.row1, line2.col1,
                      line1.row1, line1.col1, line2.row2, line2.col2) > 0)
  {
    return 1;
  }
  if (vEndpointTest2(line1.row2, line1.col2, line2.row2, line2.col2,
                      line1.row1, line1.col1, line2.row1, line2.col1) > 0)
  {
    return 1;
  }
  return 0;
}


// Returns 1 if (row1a, col1a) is within 10 pixels of (row2a, col2a),
// and (row1b, col1b) has a distance from (row2b, col2b) that is more than
// two thirds of the difference in lengths between the first and the second segment.
// 5/12/2003: I don't quite understand why the second test is there (i.e. the
// test comparing (distance-b to length2-length1) / 1.5.
long vEndpointTest1(double row1a, double col1a, double row2a, double col2a,
                     double row1b, double col1b, double row2b, double col2b)
{
  double row_diff1 = row1a - row1b;
  double col_diff1 = col1a - col1b;
  double row_diff2 = row2a - row2b;
  double col_diff2 = col2a - col2b;

  double length1 = sqrt(row_diff1 * row_diff1 + col_diff1 * col_diff1);
  double length2 = sqrt(row_diff2 * row_diff2 + col_diff2 * col_diff2);

  double row_diff_a = row1a - row2a;
  double col_diff_a = col1a - col2a;
  double row_diff_b = row1b - row2b;
  double col_diff_b = col1b - col2b;

  double distance_a = sqrt(row_diff_a * row_diff_a + col_diff_a * col_diff_a);
  double distance_b = sqrt(row_diff_b * row_diff_b + col_diff_b * col_diff_b);
  if (distance_a > 10) return 0;
  if (distance_b < ((length2 - length1) / 1.5)) return 0;
  return 1;
}


// Returns true if (row1a, col1a) is close to (row2a, col2a), i.e. within
// 10 pixels, and if (row2a, col2a) is relatively far from (row2b, col2b),
// where relatively far is defined to be more than two-thirds of the sum
// of the lengths of link1 and link2.
long vEndpointTest2(double row1a, double col1a, double row2a, double col2a,
                     double row1b, double col1b, double row2b, double col2b)
{
  double row_diff1 = row1a - row1b;
  double col_diff1 = col1a - col1b;
  double row_diff2 = row2a - row2b;
  double col_diff2 = col2a - col2b;

  double length1 = sqrt(row_diff1 * row_diff1 + col_diff1 * col_diff1);
  double length2 = sqrt(row_diff2 * row_diff2 + col_diff2 * col_diff2);

  double row_diff_a = row1a - row2a;
  double col_diff_a = col1a - col2a;
  double row_diff_b = row1b - row2b;
  double col_diff_b = col1b - col2b;

  double distance_a = sqrt(row_diff_a * row_diff_a + col_diff_a * col_diff_a);
  double distance_b = sqrt(row_diff_b * row_diff_b + col_diff_b * col_diff_b);

  if (distance_a > 10) return 0;
  if (distance_b < ((length1 + length2) / 1.5)) return 0;
  return 1;
}


// This class is used to sort line segment indices based on frequencies,
// for example in VS_Histogram::FrequencyRanking.
class vIndexFrequencyPair
{
public:
  long index;
  double frequency;

  vIndexFrequencyPair()
  {
  }

  vIndexFrequencyPair(long in_index, double in_frequency)
  {
    index = in_index;
    frequency = in_frequency;
  }
};


class vFrequencyLess
{
public:
  ushort operator() (const vIndexFrequencyPair & pair1, 
                     const vIndexFrequencyPair & pair2)
  {
    return (pair1.frequency < pair2.frequency);
  }
};



/////////////////////////////////////////////////////////////////////
//                                                                 //
//      Beginning of implementation of class vDOLVS_Histogram     //
//                                                                 //
/////////////////////////////////////////////////////////////////////


static long S_InverseArray(long * indices, long * inverse, long values, long length);

static long S_ReverseArray(long * array, long length);

// Input: inverse is an increasing array, of given length, whose values are 
// between 0 and the argument "values".
// Output: "indices" is an output argument. indices[i] is set to j iff 
// inverse[j-1] <= i and inverse[j] > i. This gives a quick way of 
// generating monotonic maps from some continuum of values to some indices, 
// so that different ranges of values are sampled at different density. For
// example, if we call S_InverseArray(indices, {1, 3, 7}, 10, 3), at the 
// end indices has the values {0, 1, 1, 2, 2, 2, 2, 3, 3, 3}. Then,
// indices[4] = 2, meaning that inverse[1] <= 4 and inverse[2] > 4. This is 
// true, since inverse[1] = 3 and inverse[2] = 7. We assume that inverse[-1] 
// has a virtual value of 0.
static long S_InverseArray(long * indices, long * inverse, long values, long length)
{
  long i, j;
  long start, end;
  if (values == 0) return 0;
  if (length == 0) return 0;
  start = inverse[0];
  // Store the value 0 in indices[0...start-1]
  for (i = 0; i < start; i++)
  {
    indices[i] = 0;
  }

  for(j = 0; j < length-1; j++)
  {
    start = inverse[j];
    end = inverse[j+1];

    // Some bug catchers:
    if ((end <= start) || (start < 0) || (end >= values))
    {
      exit_error("S_InverseArray bug: %li %li %li %li\n",
                      values, length, start, end);
    }
    
    // Store the value j in indices[start...end-1]
    for (i = start; i < end; i++)
    {
      indices[i] = j;
    }
  }

  // Store the value length-1 in indices[end...values-1]
  for (i = end; i < values; i++)
  {
    indices[i] = length-1;
  }

  return 1;
}


// replace each entry a in the array with b-a, where
// b = max(array).
static long S_RevertArray(long * array, long length)
{
  if (length == 0) return 0;
  long max = array[0];
  long i;
  for (i = 1; i < length; i++)
  {
    if (array[i] > max) max = array[i];
  }

  // Store the reverted array into the original array
  for (i = 0; i < length; i++)
  {
    array[i] = max - array[i];
  }

  return 1;
}


vDOLVS_Histogram::vDOLVS_Histogram()
{
  D_length = 25;
  O_length = 10;
  L_length = 10;
  vlength = 10;
  S_length = 10;

  D_values = 400;
  O_values = 360;
  L_values = 500;
  vvalues = 1000;
  S_values = 2000;
  
  // Power seriers of a=1.18432437049485, minus 1.
  // a = 160 ^ (1/30).
  // First we create temporary arrays, because this way we can enter all
  // array values in one line using the curly-brace notation.
  long D_inverse_temp[] = 
  {0,1,2,3,4,5,7,8,10,12,14,17,20,24,28,34,40,48,57,68,80,95,113,134,159};
  memcpy(D_inverse, D_inverse_temp, D_length * sizeof(long));

  long O_inverse_temp[] = {0, 8, 15, 23, 30, 38, 45, 53, 60, 75};
  memcpy(O_inverse, O_inverse_temp, O_length * sizeof(long));

  long L_inverse_temp[] = {0,5,10,15,20,30,40,50,65,80};
  memcpy(L_inverse, L_inverse_temp, L_length * sizeof(long));

  long vinverse_temp[] = {0,10,20,30,40,50,60,70,80,90};
  memcpy(vinverse, vinverse_temp, vlength * sizeof(long));

  long S_inverse_temp[] = {0,30,60,90,120,150,180,210,240,270};
  memcpy(S_inverse, S_inverse_temp, S_length * sizeof(long));

  // Store in the indices arrays the indices corresponding to
  // each D,O,L,V, and S value.
  S_InverseArray(D_indices, D_inverse, D_values, D_length);
  S_InverseArray(O_indices, O_inverse, O_values, O_length);
  S_InverseArray(L_indices, L_inverse, L_values, L_length);
  S_InverseArray(vindices, vinverse, vvalues, vlength);
  S_InverseArray(S_indices, S_inverse, S_values, S_length);

  // We need to revert the indices for Vote density and Strength, 
  // since, for those values, higher values indicate better matches.
  S_RevertArray(vindices, vvalues);
  S_RevertArray(S_indices, S_values);

  input_summary = vZero(long);
}


vDOLVS_Histogram::~vDOLVS_Histogram()
{
  vdelete2(input_summary);
}


// Store the value 0 in all bins.
long vDOLVS_Histogram::Zero()
{
  long i,j,k,l,m;
  for (i = 0; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        DOL[i][j][k] = 0;
        for (l = 0; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            histogram[i][j][k][l][m] = 0;
          }
        }
      }
    }
  }

  return 1;
}


// Add to each bin of this histogram the corresponding bin of h.
long vDOLVS_Histogram::Add(vDOLVS_Histogram * h)
{
  long i,j,k,l,m;
  for (i = 0; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            histogram[i][j][k][l][m] += h->histogram[i][j][k][l][m];
          }
        }
      }
    }
  }

  return 1;
}


// Return a new histogram that is a copy of the current one.
vDOLVS_Histogram * vDOLVS_Histogram::Copy()
{
  vDOLVS_Histogram * result = new vDOLVS_Histogram();
  
  long i,j,k,l,m;
  for (i = 0; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            result->histogram[i][j][k][l][m] = histogram[i][j][k][l][m];
          }
        }
      }
    }
  }

  return result;
}


// If this histogram is equal to histogram2 in every bin, we 
// return 1. Else we return 0 minus the number of bins where
// the two histograms are different.
long vDOLVS_Histogram::Equal(vDOLVS_Histogram * histogram2)
{ 
  long i,j,k,l,m;
  long differences = 0;
  for (i = 0; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            if (histogram2->histogram[i][j][k][l][m] != histogram[i][j][k][l][m])
            {
              differences++;
            }
          }
        }
      }
    }
  }

  if (differences == 0) return 1;
  return -differences;
}


// Replace the value in every bin in the histogram with the sum of
// the values of all bins "less than or equal" to that bin in the 
// histogram.
long vDOLVS_Histogram::MakeCumulative()
{
  // Add along Distance
  long i,j,k,l,m;
  for (i = 1; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            histogram[i][j][k][l][m] += histogram[i-1][j][k][l][m];
          }
        }
      }
    }
  }

  // Add along Orientation
  for (i = 0; i < D_length; i++)
  {
    for (j = 1; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            histogram[i][j][k][l][m] += histogram[i][j-1][k][l][m];
          }
        }
      }
    }
  }

  // Add along Length
  for (i = 0; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 1; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            histogram[i][j][k][l][m] += histogram[i][j][k-1][l][m];
          }
        }
      }
    }
  }

  // Add along Vote Density
  for (i = 0; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 1; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            histogram[i][j][k][l][m] += histogram[i][j][k][l-1][m];
          }
        }
      }
    }
  }

  // Add along Strength
  for (i = 0; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 1; m < S_length; m++)
          {
            histogram[i][j][k][l][m] += histogram[i][j][k][l][m-1];
          }
        }
      }
    }
  }

  return 1;
}


// Similar to MakeCumulative but for the DOL histogram, which,
// as of 5/12/2003, I don't know what it's used for.
long vDOLVS_Histogram::MakeDOLCumulative()
{
  // Add along Distance
  long i,j,k;
  for (i = 1; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        DOL[i][j][k] += DOL[i-1][j][k];
      }
    }
  }

  // Add along Orientation
  for (i = 0; i < D_length; i++)
  {
    for (j = 1; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        DOL[i][j][k] += DOL[i][j-1][k];
      }
    }
  }

  // Add along Length
  for (i = 0; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 1; k < L_length; k++)
      {
        DOL[i][j][k] += DOL[i][j][k-1];
      }
    }
  }

  return 1;
}

  
// An alternative version of MakeCumulative, where, instead 
// of summing up all values in the "<=" bins, we simply
// put a 1 if one of those bins is non-zero, 0 otherwise.
long vDOLVS_Histogram::MakeExistentialCumulative2()
{
  // Go along Distance
  long i,j,k,l,m;
  for (i = 1; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            long previous = histogram[i-1][j][k][l][m]; 
            long current = Min(histogram[i][j][k][l][m], (long) 1);
            histogram[i][j][k][l][m] = Max(previous, current);
          }
        }
      }
    }
  }

  // Add along Orientation
  for (i = 0; i < D_length; i++)
  {
    for (j = 1; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            long previous = histogram[i][j-1][k][l][m]; 
            long current = Min(histogram[i][j][k][l][m], (long) 1);
            histogram[i][j][k][l][m] = Max(previous, current);
          }
        }
      }
    }
  }

  // Add along Length
  for (i = 0; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 1; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            long previous = histogram[i][j][k-1][l][m]; 
            long current = Min(histogram[i][j][k][l][m], (long) 1);
            histogram[i][j][k][l][m] = Max(previous, current);
          }
        }
      }
    }
  }

  // Add along Vote Density
  for (i = 0; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 1; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            long previous = histogram[i][j][k][l-1][m]; 
            long current = Min(histogram[i][j][k][l][m], (long) 1);
            histogram[i][j][k][l][m] = Max(previous, current);
          }
        }
      }
    }
  }

  // Add along Strength
  for (i = 0; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 1; m < S_length; m++)
          {
            long previous = histogram[i][j][k][l][m-1]; 
            long current = Min(histogram[i][j][k][l][m], (long) 1);
            histogram[i][j][k][l][m] = Max(previous, current);
          }
        }
      }
    }
  }

  return 1;
}


// Like MakeExistentialCumulative2, but it operates on each
// bit in parallel. Essentially, it treates the histogram
// as if it is 32 separate DOLVS histograms, each consisting
// of the i-th bit of each DOLVS bin in this histogram. It
// Basically converts all those 32 histograms into
// cumulative in parallel, and then sums up the number of
// ones in each bin (i.e. in the binary form of the long
// number stored in that bin).
long vDOLVS_Histogram::MakeExistentialCumulative()
{
  static const long junk1 = vPrecomputedBitSums::PrecomputeValues();
  const long * bit_sums = vPrecomputedBitSums::Values();

  // Go along Distance
  long i,j,k,l,m;
  for (i = 1; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            long previous = histogram[i-1][j][k][l][m]; 
            histogram[i][j][k][l][m] |= previous;
          }
        }
      }
    }
  }

  // Add along Orientation
  for (i = 0; i < D_length; i++)
  {
    for (j = 1; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            long previous = histogram[i][j-1][k][l][m]; 
            histogram[i][j][k][l][m] |= previous;
          }
        }
      }
    }
  }

  // Add along Length
  for (i = 0; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 1; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            long previous = histogram[i][j][k-1][l][m]; 
            histogram[i][j][k][l][m] |= previous;
          }
        }
      }
    }
  }

  // Add along Vote Density
  for (i = 0; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 1; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            long previous = histogram[i][j][k][l-1][m]; 
            histogram[i][j][k][l][m] |= previous;
          }
        }
      }
    }
  }

  // Add along Strength
  for (i = 0; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 1; m < S_length; m++)
          {
            long previous = histogram[i][j][k][l][m-1]; 
            histogram[i][j][k][l][m] |= previous;
          }
        }
      }
    }
  }

  // now convert bit patterns to numbers
  for (i = 0; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            ulong entry = histogram[i][j][k][l][m];
            ulong entry1 = entry & 65535;
            ulong entry2 = entry >> 16;
            histogram[i][j][k][l][m] = bit_sums[entry1] + bit_sums[entry2];
          }
        }
      }
    }
  }

  return 1;
}


// Yet another alternative version of MakeCumulative, where
// at each bin we store the max value occurring in the "<=" bins.
long vDOLVS_Histogram::Propagate()
{
  // Add along Distance
  long i,j,k,l,m;
  for (i = 1; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            long previous = histogram[i-1][j][k][l][m]; 
            long current = histogram[i][j][k][l][m];
            histogram[i][j][k][l][m] = Max(previous, current);
          }
        }
      }
    }
  }

  // Add along Orientation
  for (i = 0; i < D_length; i++)
  {
    for (j = 1; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            long previous = histogram[i][j-1][k][l][m]; 
            long current = histogram[i][j][k][l][m];
            histogram[i][j][k][l][m] = Max(previous, current);
          }
        }
      }
    }
  }

  // Add along Length
  for (i = 0; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 1; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            long previous = histogram[i][j][k-1][l][m]; 
            long current = histogram[i][j][k][l][m];
            histogram[i][j][k][l][m] = Max(previous, current);
          }
        }
      }
    }
  }

  // Add along Vote Density
  for (i = 0; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 1; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            long previous = histogram[i][j][k][l-1][m]; 
            long current = histogram[i][j][k][l][m];
            histogram[i][j][k][l][m] = Max(previous, current);
          }
        }
      }
    }
  }

  // Add along Strength
  for (i = 0; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 1; m < S_length; m++)
          {
            long previous = histogram[i][j][k][l][m-1]; 
            long current = histogram[i][j][k][l][m];
            histogram[i][j][k][l][m] = Max(previous, current);
          }
        }
      }
    }
  }

  return 1;
}


// Save the entries of the histogram into a file.
long vDOLVS_Histogram::Save(const char * filename)
{
  FILE * fp = fopen(filename, vFOPEN_WRITE);
  if (fp == 0) return 0;
  long i,j,k,l,m;
  long index = 0;
  long total_size = D_length * O_length * L_length * vlength * S_length;
  vArray(long) buffer = vnew(long, total_size);

  for (i = 0; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            buffer[index++] = histogram[i][j][k][l][m];
          }
        }
      }
    }
  }

  long success = vWritePCLongs(fp, vData(buffer), total_size);
  fclose(fp);
  vdelete2(buffer);
  if (success == total_size)
  {
    return 1;
  }
  else
  {
    return 0;
  }
}


// Load the entries of the histogram from a file.
long vDOLVS_Histogram::Load(const char * filename)
{
  FILE * fp = fopen(filename, vFOPEN_READ);
  if (fp == 0) return 0;
  long i,j,k,l,m;
  long index = 0;
  long total_size = D_length * O_length * L_length * vlength * S_length;
  vArray(long) buffer = vnew(long, total_size);

  long success = vReadPCLongs(fp, vData(buffer), total_size);
  fclose(fp);
  for (i = 0; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            histogram[i][j][k][l][m] = buffer[index++];
          }
        }
      }
    }
  }

  vdelete2(buffer);
  if (success == total_size)
  {
    return 1;
  }
  else
  {
    return 0;
  }
}


// Here we calculate, for each match between the model link_segment
// and one of the input segments, the DOLVS values describing the match.
// This is an older and slower version, that I'm just keeping 
// around for the time being. 
// This version doesn't use input_summary, and it's slow because
// it is precomputing a lot of things over and over again.
long vDOLVS_Histogram::ProcessLinkSegment2(vLineSegment link_segment, 
                                            vector<vLineSegmentMatch> * input_segments)
{
  static long junk = vPrecomputedLengths::PrecomputeValues(256, 256);
  static const double rad_to_deg = 180.0 / M_PI;

  // Precompute information about the model line.
  long center_row = (long) (link_segment.row1 + link_segment.row2) / 2;
  long center_col = (long) (link_segment.col1 + link_segment.col2) / 2;
  long drow = (long) (link_segment.row1 - link_segment.row2);
  long dcol = (long) (link_segment.col1 - link_segment.col2);
  long abs_drow = vAbs(drow);
  long abs_dcol = vAbs(dcol);
  long length = vPrecomputedLengths::Length(abs_drow, abs_dcol);
  float orientation = (float) vAngle(dcol, drow);

  long number = input_segments->size();
  long i;
  for (i = 0; i < number; i++)
  {
    vLineSegmentMatch & segment_match = (*input_segments)[i];

    // Get dolvs values, and save them on the segment match.
    long drow = vAbs(segment_match.center_row - center_row);
    long dcol = vAbs(segment_match.center_col - center_col);

    long d = vPrecomputedLengths::LengthR(drow, dcol);
    long d_orientation = (long) vCircularDistance(segment_match.orientation, orientation, (float) M_PI);
    long o = vAbs(d_orientation);

    // For speed test
//    long d = 0;
//    long o = 0;

    long l = vAbs(vRound(segment_match.length - length));
    long v = vAbs(vRound(segment_match.vote_density * 100));
    long s = vRound(segment_match.strength);
    segment_match.d = d;
    segment_match.o = o;
    segment_match.l = l;
    segment_match.v = v;
    segment_match.s = s;
    
  }

  return 1;
}


// Here we calculate, for each match between the model link_segment
// and one of the input segments, the DOLVS values describing the match.
// Newer and faster version of ProcessLinkSegment2, that uses input_summary.
long vDOLVS_Histogram::ProcessLinkSegment(vLineSegment link_segment)
{
  static long junk = vPrecomputedLengths::PrecomputeValues(256, 256);
  static const double rad_to_deg = 180.0 / M_PI;
  // Precompute some information about the model line.
  long center_row = vRound((link_segment.row1 + link_segment.row2) / 2);
  long center_col = vRound((link_segment.col1 + link_segment.col2) / 2);
  long drow = vRound(link_segment.row1 - link_segment.row2);
  long dcol = vRound(link_segment.col1 - link_segment.col2);
  long abs_drow = vAbs(drow);
  long abs_dcol = vAbs(dcol);
  long length = vPrecomputedLengths::Length(abs_drow, abs_dcol);
  long orientation = vRound(vAngle(dcol, drow) * rad_to_deg);
//  vPrint("model orientation = %li\n", orientation);

  long i;
  long index = 0;
  vArray2(long) lengths = vPrecomputedLengths::values;
  for (i = 0; i < input_length; i++)
  {
    index = items_per_segment*i;
    // Get dol values (V and S are independent of the model line,
    // and they have been precomputed).
    long drow = vAbs(center_row - input_summary[index]);
    long dcol = vAbs(center_col - input_summary[index+1]);

    input_summary[index+6] = lengths[drow][dcol];
    long input_orientation = input_summary[index+2];
    long o = vAbs(input_orientation - orientation);
    if (o > 90) o = 180 - o;
    input_summary[index+7] = o;
    input_summary[index+8] = vAbs(input_summary[index+3] - length);
  }

  return 1;
}


// A faster implementation that combines ProcessLinkSegment2 and MarkEntries2.
// For every match between link segment and one of the input segments, 
// we increment the value in corresponding DOLVS bin.
long vDOLVS_Histogram::ProcessMarkSegment2(vLineSegment link_segment, 
                                            vArray(long) input_segments, long number)
{
  static long junk = vPrecomputedLengths::PrecomputeValues(256, 256);
  static const double rad_to_deg = 180.0 / M_PI;

  // Precompute some info (position, length, orientation) about the 
  // model line.
  long center_row = vRound((link_segment.row1 + link_segment.row2) / 2);
  long center_col = vRound((link_segment.col1 + link_segment.col2) / 2);
  long drow = vRound(link_segment.row1 - link_segment.row2);
  long dcol = vRound(link_segment.col1 - link_segment.col2);
  long abs_drow = vAbs(drow);
  long abs_dcol = vAbs(dcol);
  long length = vPrecomputedLengths::Length(abs_drow, abs_dcol);
  long orientation = vRound(vAngle(dcol, drow) * rad_to_deg);

  long i;
  long index = 0;
  vArray2(long) lengths = vPrecomputedLengths::values;
  // Go through the input line segments.
  for (i = 0; i < number; i++)
  {
    index = items_per_segment*i;
    // Get dol values (vs values have been precomputed).
    long drow = vAbs(center_row - input_segments[index]);
    long dcol = vAbs(center_col - input_segments[index+1]);

    long d = lengths[drow][dcol];
    long input_orientation = input_segments[index+2];
    long o = vAbs(input_orientation - orientation);
    if (o > 90) o = 180 - o;
    long l = vAbs(input_segments[index+3] - length);

    // Compute the indices defining the DOLVS bin corresponding to the match.
    long id = D_indices[d];
    long io = O_indices[o];
    long il = L_indices[l];
    long iv = vindices[input_segments[index+4]];
    long is = S_indices[input_segments[index+5]];
    
    // Increment the value in the appropriate DOLVS bin.
    histogram[id][io][il][iv][is]++;
  }

  return 1;
}


// A faster implementation of ProcessMarkSegment2. This is the version
// that was actually used in the CVPR 2003 experiments. Here we don't 
// need the input_segments and number arguments of ProcessMarkSegment2,
// because we assume that the member variables "input_summary" and
// "input_length" already contain the information that we need. Another 
// thing that speeds up the overall implementation is that here we treat
// the DOLVS histogram as 32 different binary histograms. The "counter" 
// argument specifies which of those 32 histograms we will mark. The
// advantage of this is that then we can call MakeExistentialCumulative,
// which processes all 32 histograms at once. This way we have to 
// go through all bins of the histogram only once per 32 model lines.
// With the previous implementation, that ProcessMarkSegment2 is a part of,
// we would have to go through all bins of the histogram once per model
// line.
long vDOLVS_Histogram::ProcessMarkSegment3(vLineSegment link_segment,
                                            long counter)
{
  static long junk = vPrecomputedLengths::PrecomputeValues(256, 256);
  static const double rad_to_deg = 180.0 / M_PI;

  // Precompute the POL of the model line.
  long center_row = vRound((link_segment.row1 + link_segment.row2) / 2);
  long center_col = vRound((link_segment.col1 + link_segment.col2) / 2);
  long drow = vRound(link_segment.row1 - link_segment.row2);
  long dcol = vRound(link_segment.col1 - link_segment.col2);
  long abs_drow = vAbs(drow);
  long abs_dcol = vAbs(dcol);
  long length = vPrecomputedLengths::Length(abs_drow, abs_dcol);
  long orientation = vRound(vAngle(dcol, drow) * rad_to_deg);

  long i;
  long index = 0;
  // Compute the mask that can be used to access the bits corresponding
  // to the appropriate binary histogram.
  long bit = counter % 32;
  long mask = 1 << bit;
  vArray2(long) lengths = vPrecomputedLengths::values;
  for (i = 0; i < input_length; i++)
  {
    index = items_per_segment*i;
    // Get dol values for the match.
    long drow = vAbs(center_row - input_summary[index]);
    long dcol = vAbs(center_col - input_summary[index+1]);

    long d = lengths[drow][dcol];
    long input_orientation = input_summary[index+2];
    long o = vAbs(input_orientation - orientation);
    if (o > 90) o = 180 - o;
    long l = vAbs(input_summary[index+3] - length);

    // Compute the DOLVS indices.
    long id = D_indices[d];
    long io = O_indices[o];
    long il = L_indices[l];
    long iv = vindices[input_summary[index+4]];
    long is = S_indices[input_summary[index+5]];
    
    // Set to 1 the appropriate bit in the appropriate bin of the 
    // histogram.
    histogram[id][io][il][iv][is] |= mask;
  }

  return 1;
}


// Here we use the dolvs entries in each of the input segments
// (as computed, for example, by ProcessLinkSegment) and we mark
// them on to the histogram (each bin receives as many votes
// as the number of segments whose DOLVS entries fall into that bin).
// Newer and faster version of MarkEntries2, uses input_summary.
// It assumes that for input line i, entries 9*i + 6,7,8 in the 
// input_summary are dd, do, dl (differences in DOL between 
// line i and some fixed model line).
// This function is not used in the actual experiments anymore, it
// has been replaced by ProcessMarkSegment3.
long vDOLVS_Histogram::MarkEntries()
{
  Zero();
  long i;
  long id, io, il, iv, is;
  id = io = il = iv = is = 0;
  for (i = 0; i < input_length; i++)
  {
    long index = items_per_segment * i;
    long & d = input_summary[index+6];
    long & o = input_summary[index+7];
    long & l = input_summary[index+8];
    long & v = input_summary[index+4];
    long & s = input_summary[index+5];

    // Get dolvs indices
    GetIndices(d, o, l, v, s, id, io, il, iv, is);
    histogram[id][io][il][iv][is]++;
  }

  return 1;
}


// Older and slower version of SmallestMatchFrequency, that is not 
// used in the actual CVPR 2003 experiments. We assume that the 
// DOLVS histogram already contains the appropriate values. Then, given
// a model line and a set of input lines, we return the frequency
// of the least frequent match.
long vDOLVS_Histogram::SmallestMatchFrequency2(vLineSegment link_segment, 
                                                vector<vLineSegmentMatch> * input_segments)
{
  static long junk = vPrecomputedLengths::PrecomputeValues(256, 256);
  static const double rad_to_deg = 180.0 / M_PI;
  
  // Compute POL for the model line.
  long center_row = (long) (link_segment.row1 + link_segment.row2) / 2;
  long center_col = (long) (link_segment.col1 + link_segment.col2) / 2;
  long drow = vRound(link_segment.row1 - link_segment.row2);
  long dcol = vRound(link_segment.col1 - link_segment.col2);
  long abs_drow = vAbs(drow);
  long abs_dcol = vAbs(dcol);
  long length = vPrecomputedLengths::Length(abs_drow, abs_dcol);
  float orientation = (float) vAngle(dcol, drow);

  long number = input_segments->size();
  long i;
  long smallest_frequency = histogram[D_length-1][O_length-1][L_length-1][vlength-1][S_length-1];

  // Go through all input lines.
  for (i = 0; i < number; i++)
  {
    vLineSegmentMatch & segment_match = (*input_segments)[i];

    // Get dolvs values
    long drow = vAbs(segment_match.center_row - center_row);
    long dcol = vAbs(segment_match.center_col - center_col);

    long d = vPrecomputedLengths::LengthR(drow, dcol);
    float d_orientation = vCircularDistance(segment_match.orientation, orientation, (float) M_PI);
    long o = vAbs(vRound(d_orientation * rad_to_deg));

    long l = vAbs(vRound(segment_match.length - length));
    long v = vAbs(vRound(segment_match.vote_density * 100));
    long s = vRound(segment_match.strength);
    segment_match.d = d;
    segment_match.o = o;
    segment_match.l = l;
    segment_match.v = v;
    segment_match.s = s;

    // Get the frequency of that match, check if it is the smallest
    // seen so far.
    long frequency = DOLVS(d, o, l, v, s);
    smallest_frequency = Min(smallest_frequency, frequency);
  }

  return smallest_frequency;
}


// This is a version of SmallestMatchFrequency2 that is useful 
// when we have a list of candidate line segments that
// we should consider (as opposed to considering all input segments).
// Those candidates can be obtained, for example, using a line grid.
// Also, this function doesn't require ProcessLinkSegment to be called
// first, it does the job of that function as well, but only for the 
// given candidates, so that it saves time. We do assume that the 
// function MakeInputSummary has already been called.
long vDOLVS_Histogram::SmallestMatchFrequency3(vLineSegment link_segment, 
                                                vArray(long) candidates,
                                                long candidate_number)
{
  static long junk = vPrecomputedLengths::PrecomputeValues(256, 256);
  static const double rad_to_deg = 180.0 / M_PI;

  // Compute POL for the model line.
  long center_row = vRound((link_segment.row1 + link_segment.row2) / 2);
  long center_col = vRound((link_segment.col1 + link_segment.col2) / 2);
  long drow = vRound(link_segment.row1 - link_segment.row2);
  long dcol = vRound(link_segment.col1 - link_segment.col2);
  long abs_drow = vAbs(drow);
  long abs_dcol = vAbs(dcol);
  long length = vPrecomputedLengths::Length(abs_drow, abs_dcol);
  long orientation = vRound(vAngle(dcol, drow) * rad_to_deg);
  vArray2(long) lengths = vPrecomputedLengths::values;

  long smallest_frequency = histogram[D_length-1][O_length-1][L_length-1][vlength-1][S_length-1] + 1;
  long min_index = -1;
  long i;
  // Go through all candidate input lines.
  for (i = 0; i < candidate_number; i++)
  {
    long candidate_index = candidates[i];
    long index = items_per_segment*candidate_index;
    // Get dol values for the current match.
    long drow = vAbs(center_row - input_summary[index]);
    long dcol = vAbs(center_col - input_summary[index+1]);

    long d = lengths[drow][dcol];
    long input_orientation = input_summary[index+2];
    long o = vAbs(input_orientation - orientation);
    if (o > 90) o = 180 - o;
    long l = vAbs(input_summary[index+3] - length);

    // Get the indices for the appropriate histogram bin.
    long id = D_indices[d];
    long io = O_indices[o];
    long il = L_indices[l];
    long iv = vindices[input_summary[index+4]];
    long is = S_indices[input_summary[index+5]];

    // Get the frequency of the match, check if it is the smallest
    // frequency seen so far.
    long frequency = histogram[id][io][il][iv][is];
    if (frequency < smallest_frequency)
    {
      smallest_frequency = frequency;
      min_index = candidate_index;
    }
  }

  return smallest_frequency;
}



// This function is called when we have already built the existential
// DOLVS, and each bin contains either frequencies, or probabilities.
// It finds, for each match of link_segment, the corresponding entry,
// and it returns the smallest among those entries. It uses input_summary,
// so it is a faster version of SmallestMatchFrequency2, that computes
// the same thing.
long vDOLVS_Histogram::SmallestMatchFrequency(vLineSegment link_segment)
{
  return SmallestMatchFrequency(link_segment, 0);
}


// See comments for SmallestMatchFrequency(vLineSegment).
// Here, in addition to returning the smallest frequency, we store
// in best_match_index the index of the line segment that attained
// that smallest frequency.
long vDOLVS_Histogram::SmallestMatchFrequency(vLineSegment link_segment,
                                               long * best_match_index)
{
  static long junk = vPrecomputedLengths::PrecomputeValues(256, 256);
  static const double rad_to_deg = 180.0 / M_PI;

  // Compute POL for the model line.
  long center_row = vRound((link_segment.row1 + link_segment.row2) / 2);
  long center_col = vRound((link_segment.col1 + link_segment.col2) / 2);
  long drow = (long) (link_segment.row1 - link_segment.row2);
  long dcol = (long) (link_segment.col1 - link_segment.col2);
  long abs_drow = vAbs(drow);
  long abs_dcol = vAbs(dcol);
  long length = vPrecomputedLengths::Length(abs_drow, abs_dcol);
  long orientation = vRound(vAngle(dcol, drow) * rad_to_deg);
  vArray2(long) lengths = vPrecomputedLengths::values;

  long number = input_length;
  long i;
  long smallest_frequency = histogram[D_length-1][O_length-1][L_length-1][vlength-1][S_length-1] + 1;
  long min_index = -1;
  for (i = 0; i < number; i++)
  {
    long index = items_per_segment * i;
    // Get dol values
    long drow = vAbs(input_summary[index] - center_row);
    long dcol = vAbs(input_summary[index+1] - center_col);

    long d = lengths[drow][dcol];
    long o = vAbs(input_summary[index+2] - orientation);
    if (o > 90) o = 180 - o;

    long l = vAbs(input_summary[index+3] - length);

    // Get indices for the histogram ibn.
    long id = D_indices[d];
    long io = O_indices[o];
    long il = L_indices[l];
    long iv = vindices[input_summary[index+4]];
    long is = S_indices[input_summary[index+5]];

    // Look up the frequency.
    long frequency = histogram[id][io][il][iv][is];
    if (frequency < smallest_frequency)
    {
      smallest_frequency = frequency;
      min_index = i;
    }
  }

  if (min_index == -1)
  {
    exit_error("Error, no best match found\n");
  }

  if (best_match_index != 0)
  {
    *best_match_index = min_index;
  }

  return smallest_frequency;
}


// Return all input segments, ranked in order of similarity to 
// the model line specified by the argument. It calls 
// FrequencyRanking.
vArray(long) vDOLVS_Histogram::FrequencyRanking1(vLineSegment model_line)
{
  long center_row = vRound(model_line.CenterRow());
  long center_col = vRound(model_line.CenterCol());
  double rads = model_line.Orientation();
  long orientation = vRound(model_line.OrientationDegrees());
  long length = vRound(model_line.Length());

  return FrequencyRanking4(center_row, center_col, orientation, length);
}

  
// Return all input segments, ranked in order of similarity to 
// the model line specified by the argument. It calls FrequencyRanking5.
vArray(long) vDOLVS_Histogram::FrequencyRanking2(vLineSegment model_line, 
                                                   vector<long> * frequencies)
{
  long center_row = vRound(model_line.CenterRow());
  long center_col = vRound(model_line.CenterCol());
  double rads = model_line.Orientation();
  long orientation = vRound(model_line.OrientationDegrees());
  long length = vRound(model_line.Length());

  return FrequencyRanking5(center_row, center_col, orientation, length,
                           frequencies);
}


// Returns an array of indices of line segments, sorted by frequency of the
// match of each line segment with the argument. The argument is a line
// segment with the given position, orientation and length.
vArray(long) vDOLVS_Histogram::FrequencyRanking4(long center_row, long center_col, 
                                                   long orientation, long length)
{
  vector<long> frequencies;
  return FrequencyRanking5(center_row, center_col, orientation, length, &frequencies);
}

                                                  
// Return all input segments, ranked in order of similarity to 
// the model line specified by the arguments. It puts the 
// frequency corresponding to each input segment into the
// frequencies array.
vArray(long) vDOLVS_Histogram::FrequencyRanking5(long center_row, long center_col, 
                                                   long orientation, long length,
                                                   vector<long> * frequency_vector)
{
  static long junk = vPrecomputedLengths::PrecomputeValues(256, 256);
  static const double rad_to_deg = 180.0 / M_PI;
  vArray2(long) lengths = vPrecomputedLengths::values;

  // Prepare for count sort. We need to initialize counters to 0.
  long counters[100];
  frequency_vector->erase(frequency_vector->begin(), frequency_vector->end());
  frequency_vector->reserve(input_length);
  vArray(long) result = vnew(long, input_length);
  vArray(long) frequencies = vnew(long, input_length);
  long i;
  for (i = 0; i < 100; i++)
  {
    counters[i] = 0;
  }

  long number = input_length;
  // Count the occurrence of each frequency. For efficiency, save
  // the frequency into the frequencies array.
  for (i = 0; i < number; i++)
  {
    long index = items_per_segment * i;
    // Get dolvs values
    long drow = vAbs(input_summary[index] - center_row);
    long dcol = vAbs(input_summary[index+1] - center_col);

    long d = lengths[drow][dcol];
    long o = vAbs(input_summary[index+2] - orientation);
    if (o > 90) o = 180 - o;

    long l = vAbs(input_summary[index+3] - length);

    long id = D_indices[d];
    long io = O_indices[o];
    long il = L_indices[l];
    long iv = vindices[input_summary[index+4]];
    long is = S_indices[input_summary[index+5]];

    long frequency = histogram[id][io][il][iv][is];
    if (frequency < 0) exit_error("Negative frequency in DOLVS\n");
    if (frequency > 101) exit_error("Frequency over 100 in DOLVS\n");
    if (frequency >= 100) frequency = 99;
    frequencies[i] = frequency;
    counters[frequency]++;
  }

  // Starting indices for each frequency.
  vArray(long) indices = vnew(long, 100);
  indices[0] = 0;
  for (i = 1; i < 100; i++)
  {
    indices[i] = indices[i-1] + counters[i-1];
  }

  // Put the sorted line segment indices into result
  for (i = 0; i < input_length; i++)
  {
    long frequency = frequencies[i];
    long index = indices[frequency];
    indices[frequency]++;
    result[index] = i;
    (*frequency_vector)[index] = frequency;
  }

  vdelete2(frequencies);
  vdelete2(indices);
  return result;
}


// Here we use the dolvs entries in each of the input segments
// (as computed, for example, by ProcessLinkSegment) and we mark
// them on to the histogram (each bin receives as many votes
// as the number of segments whose DOLVS entries fall into that bin).
// Older and slower version of MarkEntries, this version doesn't use
// input_summary.
long vDOLVS_Histogram::MarkEntries2(vector<vLineSegmentMatch> * input_segments)
{
  Zero();
  long number = input_segments->size();
  long i;
  long id, io, il, iv, is;
  id = io = il = iv = is = 0;
  for (i = 0; i < number; i++)
  {
    vLineSegmentMatch & segment_match = (*input_segments)[i];
    long & d = segment_match.d;
    long & o = segment_match.o;
    long & l = segment_match.l;
    long & v = segment_match.v;
    long & s = segment_match.s;

    // Get dolvs indices
    GetIndices(d, o, l, v, s, id, io, il, iv, is);
    histogram[id][io][il][iv][is]++;
  }

  return 1;
}


// In this function, we assume that the DOLVS of each segment with respect
// to a given model line has already
// been computed (and stored in that segment's dolvs variables). This function
// gives as score to each segment the corresponding entry in the histogram,
// and sorts the segments in ascending order.
long vDOLVS_Histogram::RankInputSegments(vector<vLineSegmentMatch> * segments)
{
  long number = segments->size();
  if (number != input_length)
  {
    return 0;
  }
  long i;
  long id, io, il, iv, is;
  id = io = il = iv = is = 0;
  long flag = 0;
  for (i = 0; i < number; i++)
  {
    vLineSegmentMatch & match = (*segments)[i];
    long index = items_per_segment * i;
    match.d = input_summary[index+6];
    match.o = input_summary[index+7];
    match.l = input_summary[index+8];
    match.v = input_summary[index+4];
    match.s = input_summary[index+5];
    GetIndices(match.d, match.o, match.l, match.v, match.s, 
               id, io, il, iv, is);
    match.score = (float) histogram[id][io][il][iv][is];
  }

  std::sort(segments->begin(), segments->end(), vLineSegmentMatchLess());
  return 1;
}


// What we compute here is the following: For each DOLVS bin, how common
// is it for a random model line to have among the input lines a match
// that is as good or better than that DOLVS bin. Model lines
// are sampled from model_lines_matrix uniformly, as specified by
// the arguments "start" and "period".
// Older and slower version of ExistentialFrequencies, creates a 
// cumulative DOLVS histogram for each model line.
long vDOLVS_Histogram::ExistentialFrequencies2(v3dMatrix<uchar> * model_lines_matrix,
                                                vector<vLineSegmentMatch> * input_lines,
                                                long start, long period)
{
  vector<vLineSegment> model_lines;
  long number_of_model_lines = vSampleLineSegments(model_lines_matrix, 
                                                    start, period, &model_lines);
  Zero();
  long i;
  vPrint("\n");
  for (i = 0; i < number_of_model_lines; i++)
  {
    vPrint("processing model line %li of %li\r", i+1, number_of_model_lines);
    vDOLVS_Histogram histogram;
    histogram.ProcessLinkSegment2(model_lines[i], input_lines);
    histogram.MarkEntries2(input_lines);
    histogram.MakeExistentialCumulative2();
    Add(&histogram);
  }
  vPrint("\n");

  return 1;
}

// Commented out on 5/12/2003
// For each segment, save center_row, center_col, o, l, v, s, and also keep room
// for do, dl, ds (that depend on each model line).
//vArray(long) S_SegmentSummaryInfo(vector<vLineSegmentMatch> * input_lines)
//{
//  long size = input_lines->size();
//  static const long items_per_segment = 9;
//  static const double rad_to_deg = 180 / M_PI;
//  vArray(long) result = vnew(long, items_per_segment * size);
//  long i;
//  long index = 0;
//  for (i = 0; i < size; i++)
//  {
//    vLineSegmentMatch & match = (*input_lines)[i];
//    result[index++] = match.center_row;
//    result[index++] = match.center_col;
//    result[index++] = vRound(match.orientation * rad_to_deg);
//    result[index++] = vRound(match.length);
//    result[index++] = vRound(match.vote_density * 100);
//    result[index++] = vRound(match.strength);
//    // Make room for d,o,l (that depend on the difference with each model line).
//    index += 3;
//  }
//  return result;
//}



// In preparation for computing the existential DOLVS and the PDF,
// here we create an array of longs (input_summary) that has,
// precomputed, all the information that we need from input_lines.
// input_summary has 9 (items_per_segment) entries per input line,
// they occur in the following order:
// center_row, center_col, orientation (degrees), length, votes, strength
// The last 3 of the nine entries are not used right now.
long  vDOLVS_Histogram::MakeInputSummary(vector<vLineSegmentMatch> * input_lines)
{
  long size = input_lines->size();
  input_length = size;
  items_per_segment = 9;
  static const double rad_to_deg = 180 / M_PI;
  delete_pointer(input_summary);
  input_summary = vnew(long, items_per_segment * size);
  long i;
  long index = 0;
  for (i = 0; i < size; i++)
  {
    vLineSegmentMatch & match = (*input_lines)[i];
    input_summary[index++] = match.center_row;
    input_summary[index++] = match.center_col;
    input_summary[index++] = vRound(match.orientation * rad_to_deg);
    input_summary[index++] = vRound(match.length);
    input_summary[index++] = vRound(match.vote_density * 100);
    input_summary[index++] = vRound(match.strength);
    // Make room for d,o,l (that depend on the difference with each model line).
    index += 3;
  }
  return 1;
}



// What we compute here is the following: For each DOLVS bin, how common
// is it for a random model line to have among the input lines a match
// that is as good or better than that DOLVS bin. Model lines
// are sampled from model_lines_matrix uniformly, as specified by
// the arguments "start" and "period".
// This is a newer and faster version of ExistentialFrequencies2. Instead
// of creating, filling in and converting to cumulative a DOLVS histogram
// for each model line, we instead create a single histogram to which
// we mark entries for 32 model lines before we accumulate. This cuts down
// the time spent for accumulation (that has to go through all DOLVS bins)
// by a factor of 32. The trick is that when we mark entries for each
// of 32 model lines, we simply set a bit in the corresponding bin, that
// depends on the number of the line (0-31). Then, we can accumulate in
// two steps. In the first step we go from lower to higher bins taking
// logical ORs between the previous and next bin, so that each bit
// gets propagated separately. Then, for each bin, we replace its entry
// with the number of ones in the binary number representing that entry.
long vDOLVS_Histogram::ExistentialFrequencies(v3dMatrix<uchar> * model_lines_matrix,
                                               vector<vLineSegmentMatch> * input_lines,
                                               long start, long period)
{
  // Sample some model lines, to use for building the histogram.
  vector<vLineSegment> model_lines;
  long number_of_model_lines = vSampleLineSegments(model_lines_matrix, 
                                                    start, period, &model_lines);
  // Store zeros on all bins.
  Zero();  
  // Preprocess input line segments, to extract the information that we need.
  MakeInputSummary(input_lines);

  // Create an auxiliary histogram, that will be used to mark
  // the matches of individual model lines.
  vDOLVS_Histogram * histogram = new vDOLVS_Histogram();
  histogram->input_summary = input_summary;
  histogram->input_length = input_length;
  histogram->items_per_segment = 9;
  histogram->Zero();

  long i;
  vPrint("\n");  
  // Go through all sampled model lines.
  for (i = 0; i < number_of_model_lines; i++)
  {
    vPrint("processing model line %li of %li\r", i+1, number_of_model_lines);

    // Mark in the histogram all matches between the current model line
    // and all input lines.
    histogram->ProcessMarkSegment3(model_lines[i], i);
    // We treat histogram as 32 separate binary histograms, that can 
    // hold information about the matches of 32 model lines at the same
    // time. So, every time we process 32 model lines, we need to 
    // make the histogram cumulative and add its information to this histogram.
    if (i % 32 == 31)
    {
      histogram->MakeExistentialCumulative();
      Add(histogram);
      histogram->Zero();
    }
  }

  vPrint("\n");
  // Process, if needed, the information about the last few model lines.
  if (number_of_model_lines % 32 != 0)
  {
    histogram->MakeExistentialCumulative();
    Add(histogram);
  }

  // Delete the auxiliary histogram.
  histogram->input_summary = vZero(long);
  vdelete(histogram);

  return 1;
}


// Here we assume that this object is an existential DOLVS. We return
// a vector, which, at entry i, indicates the number of random model
// lines for which the best DOLVS match had existential frequency less than or 
// equal to i. 
// This version is not used in the experiments, PDF() is used instead.
vMatrix<long> vDOLVS_Histogram::PDF2(v3dMatrix<uchar> * model_lines_matrix,
                                       vector<vLineSegmentMatch> * input_lines,
                                       long start, long period)
{
  vector<vLineSegment> model_lines;
  long number_of_model_lines = vSampleLineSegments(model_lines_matrix, 
                                                    start, period, &model_lines);
  long i;
  long max_frequency_inc = MaxValue() + 1;
  vMatrix<long> result_matrix(1, max_frequency_inc);
  vArray(long) result = result_matrix.Matrix();
  for (i = 0; i < max_frequency_inc; i++)
  {
    result[i] = 0;
  }

  vPrint("\n");
  for (i = 0; i < number_of_model_lines; i++)
  {
    vPrint("processing model line %li of %li\r", i+1, number_of_model_lines);
    long smallest_frequency = SmallestMatchFrequency2(model_lines[i], input_lines);
    result[smallest_frequency]++;
  }

  vPrint("\n");
  
  for (i = 1; i < max_frequency_inc; i++)
  {
    result[i] += result[i-1];
  }

  return result_matrix;
}


// Here we assume that this object is an existential DOLVS. We return
// a vector, which, at entry i, indicates the number of random model
// lines for which the best DOLVS match had existential frequency less than or 
// equal to i. 
// Same functionality as for PDF2, but here we call SmallestMatchFrequency,
// as opposed to SmallestMatchFrequency2. This is the version actually used
// in the experiments.
vMatrix<long> vDOLVS_Histogram::PDF(v3dMatrix<uchar> * model_lines_matrix,
                                      long start, long period)
{
  // Sample model lines.
  vector<vLineSegment> model_lines;
  long number_of_model_lines = vSampleLineSegments(model_lines_matrix, 
                                                    start, period, &model_lines);
  long i;
  // Get the largest frequency.
  long max_frequency_inc = MaxValue() + 1;
  // Create the matrix that maps frequencies to probabilities. They are not 
  // quite probabilities actually, they are more like frequencies of frequencies,
  // at least when divided by the total number of lines.
  vMatrix<long> result_matrix(1, max_frequency_inc);
  vArray(long) result = result_matrix.Matrix();
  for (i = 0; i < max_frequency_inc; i++)
  {
    result[i] = 0;
  }

  vPrint("\n");
  // Go through all model lines
  for (i = 0; i < number_of_model_lines; i++)
  {
    // Find the smallest frequency for the current model line.
    vPrint("processing model line %li of %li\r", i+1, number_of_model_lines);
    long smallest_frequency = SmallestMatchFrequency(model_lines[i]);
    result[smallest_frequency]++;
  }

  vPrint("\n");
  
  for (i = 1; i < max_frequency_inc; i++)
  {
    result[i] += result[i-1];
  }

  return result_matrix;
}


// This function is used for debugging. It counts the number
// of bins that have values equal to the argument "value".
long vDOLVS_Histogram::CountValue(long value)
{
  long i,j,k,l,m;
  long count = 0;
  for (i = 0; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            if (histogram[i][j][k][l][m] == value)
            {
              count++;
            }
          }
        }
      }
    }
  }

  return count;
}


// This function is used for debugging. It counts the number
// of bins that have values greater than or equal to the argument "value".
long vDOLVS_Histogram::CountGequal(long value)
{
  long i,j,k,l,m;
  long count = 0;
  for (i = 0; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            if (histogram[i][j][k][l][m] >= value)
            {
              count++;
            }
          }
        }
      }
    }
  }

  return count;
}


// This function is used for debugging. It counts the number
// of bins that have values less than or equal to the argument "value".
long vDOLVS_Histogram::CountLequal(long value)
{
  long i,j,k,l,m;
  long count = 0;
  for (i = 0; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            if (histogram[i][j][k][l][m] <= value)
            {
              count++;
            }
          }
        }
      }
    }
  }

  return count;
}


// Returns the max value stored in the histogram.
long vDOLVS_Histogram::MaxValue()
{
  long i,j,k,l,m;
  long max_value = histogram[0][0][0][0][0];
  for (i = 0; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            max_value = Max(max_value, histogram[i][j][k][l][m]);
          }
        }
      }
    }
  }

  return max_value;
}


// Converts each entry in the DOLVS histogram by looking
// up the corresponding probability of that entry at 
// dolvs_pdf_matrix.
// Important note: The entries are converted to integers
// between 0 and 100. It could turn out in the future that 
// double numbers are more useful.
long vDOLVS_Histogram::ToPDF(vMatrix<long> * dolvs_pdf_matrix)
{
  long i,j,k,l,m;
  vArray(long) pdf_values = dolvs_pdf_matrix->Matrix();
  long max_value = MaxValue();
  if (max_value >= dolvs_pdf_matrix->Cols()) return 0;

  double factor = 100.0 / (double) max_value;

  for (i = 0; i < D_length; i++)
  {
    for (j = 0; j < O_length; j++)
    {
      for (k = 0; k < L_length; k++)
      {
        for (l = 0; l < vlength; l++)
        {
          for (m = 0; m < S_length; m++)
          {
            long current_entry = histogram[i][j][k][l][m];
            histogram[i][j][k][l][m] = vRound(pdf_values[current_entry] * factor);
          }
        }
      }
    }
  }

  return 1;
}
  

// segment1 and segment2 are two model line segments. We want to
// find the pair of input line segments that best match the pair
// of model lines. indices1[i] and indices2[i] are indices of the i-th 
// best pair of input lines. 
long vDOLVS_Histogram::LinePairMatches(vLineSegment segment1, 
                                        vLineSegment segment2,
                                        vector<long> * indices1, 
                                        vector<long> * indices2,
                                        vector<float> * scores)
{
  // Ideally we want to return a ranking of all pairs of input lines.
  // However, if, say, we have 50000 input lines, we would run out
  // of memory. So, we have to put a limit on the number of pairs 
  // we can rank. We impose such a limit by specifying max_lines_to_use,
  // which specifies how many input lines to use in choosing each segment
  // of a matching pair. We only use the top max_line_to_use matches
  // for each model line in generating pairs.
  static const long max_lines_to_use = 2000;
  long lines_to_use;
  if (input_length <  max_lines_to_use) lines_to_use = input_length;
  else lines_to_use = max_lines_to_use;

  long number_of_pairs = lines_to_use * lines_to_use;

  // Resize the two vectors.
  indices1->erase(indices1->begin(), indices1->end());
  indices1->reserve(number_of_pairs);
  indices2->erase(indices2->begin(), indices2->end());
  indices2->reserve(number_of_pairs);
  scores->erase(scores->begin(), scores->end());
  scores->reserve(number_of_pairs);

  vector<vIndexFrequencyPair> pair_scores(number_of_pairs);

  // Get the ranking of matches for the two model line segments.
  vector<long> frequencies1;
  vector<long> frequencies2;
  vArray(long) matches1 = FrequencyRanking2(segment1, &frequencies1);
  vArray(long) matches2 = FrequencyRanking2(segment2, &frequencies2);

  float model_row_diff = (float) (segment2.CenterRow() - segment1.CenterRow());
  float model_col_diff = (float) (segment2.CenterCol() - segment1.CenterCol());
  float model_orientation_diff = (float) vCircularDistance(segment2.OrientationDegrees(), 
                                                    segment1.OrientationDegrees(), 
                                                    (double) 180);
  float model_length_diff = (float) segment2.Length() - (float) segment1.Length();

  long i, j, index1, index2;
  for (i = 0; i < lines_to_use; i++)
  {
    index1 = i * items_per_segment;
    float frequency1 = (float) (frequencies1[i] + 1);
    long center_row1 = input_summary[index1++];
    long center_col1 = input_summary[index1++];
    long orientation1 = input_summary[index1++];
    long length1 = input_summary[index1];

    for (j = 0; j < lines_to_use; j++)
    {
      index2 = j * items_per_segment;
      float frequency2 = (float) (frequencies2[j] + 1);
      long center_row2 = input_summary[index2++];
      long center_col2 = input_summary[index2++];
      long orientation2 = input_summary[index2++];
      long length2 = input_summary[index2];

      long pair_index = i * lines_to_use + j;
      float row_diff = (float) (center_row2 - center_row1);
      float col_diff = (float) (center_col2 - center_col1);
      float orientation_diff = (float) vCircularDistance(orientation2, orientation1, 
                                                  (long) 180);
      float length_diff = (float) (length2 - length1);

      float row_factor = vAbs(row_diff - model_row_diff);
      float col_factor = vAbs(col_diff - model_col_diff);
      float location = sqrt(row_factor * row_factor + col_factor * col_factor) + 1;
      float orientation = vAbs(orientation_diff - model_orientation_diff) + 1;
      float length = vAbs(length_diff - model_length_diff) + 1;

      double frequency = frequency1 * frequency2 * location * orientation * length;

      pair_scores[pair_index].index = pair_index;
      pair_scores[pair_index].frequency = frequency;
    }
  }

  // Rank all input pairs in order of decreasing similarity to the model 
  // line pair.
  std::sort(pair_scores.begin(), pair_scores.end(), vFrequencyLess());

  // Store the sorted indices and frequencies into the result arguments
  for (i = 0; i < number_of_pairs; i++)
  {
    long index = pair_scores[i].index;
    index1 = index / lines_to_use;
    index2 = index % lines_to_use;

    (*indices1)[i] = index1;
    (*indices2)[i] = index2;
    (*scores)[i] = (float) pair_scores[i].frequency;
  }

  return 1;
}



/////////////////////////////////////////////////////////////////////
//                                                                 //
//         End of implementation of class vDOLVS_Histogram        //
//                                                                 //
/////////////////////////////////////////////////////////////////////


// model_links is supposed to be all the straight links extracted from
// a single model image.
vLineMatchFrameStats::vLineMatchFrameStats(vDOLVS_Histogram * histogram,
                                             vector<vStraightLinks> * model_links)
{
  Initialize(histogram, model_links, -1, -1);
}


// model_links is supposed to be all the straight links extracted from
// a single model image.
vLineMatchFrameStats::vLineMatchFrameStats(vDOLVS_Histogram * histogram,
                                             vector<vStraightLinks> * model_links,
                                             long in_frame, long in_view)
{
  Initialize(histogram, model_links, in_frame, in_view);
}

                                             
// model_links is supposed to be all the straight links extracted from
// a single model image.
vLineMatchFrameStats::vLineMatchFrameStats(vDOLVS_Histogram * histogram,
                                             vMatrix<long> * pdf_matrix,
                                             vector<vStraightLinks> * model_links,
                                             vector<vLineSegmentMatch> * input_segments)
{
  Initialize2(histogram, pdf_matrix, model_links, input_segments, -1, -1);
}


// model_links is supposed to be all the straight links extracted from
// a single model image.
vLineMatchFrameStats::vLineMatchFrameStats(vDOLVS_Histogram * histogram,
                                             vMatrix<long> * pdf_matrix,
                                             vector<vStraightLinks> * model_links,
                                             vector<vLineSegmentMatch> * input_segments,
                                             long frame, long view)
{
  Initialize2(histogram, pdf_matrix, model_links, input_segments, frame, view);
}


// model_links is supposed to be all the straight links extracted from
// a single model image.
vLineMatchFrameStats::vLineMatchFrameStats(vDOLVS_Histogram * histogram,
                                             vLineGrid * grid,
                                             vector<vStraightLinks> * model_links,
                                             double fraction)
{
  Initialize3(histogram, grid, model_links, fraction, -1, -1);
}
  

// model_links is supposed to be all the straight links extracted from
// a single model image.
vLineMatchFrameStats::vLineMatchFrameStats(vDOLVS_Histogram * histogram,
                                             vLineGrid * grid,
                                             vector<vStraightLinks> * model_links,
                                             double fraction, long frame, long view)
{
  Initialize3(histogram, grid, model_links, fraction, frame, view);
}
  

  
// model_links is supposed to be all the straight links extracted from
// a single model image.
long vLineMatchFrameStats::Initialize(vDOLVS_Histogram * histogram,
                                       vector<vStraightLinks> * model_links,
                                       long in_frame, long in_view)
{
  frame = in_frame;
  view = in_view;
  number_of_lines = 2 * model_links->size();
  scores = vnew(long, number_of_lines);
  average = 0;
  long link;
  // Go through all the links
  for (link = 0; link < (long) model_links->size(); link++)
  {
    vLineSegment line1 = (*model_links)[link].line1;
    histogram->ProcessLinkSegment(line1);
    double score = histogram->SmallestMatchFrequency(line1);
    scores[2*link] = (long) score;
    average += score;

    vLineSegment line2 = (*model_links)[link].line2;
    histogram->ProcessLinkSegment(line2);
    score = histogram->SmallestMatchFrequency(line2);
    scores[2*link+1] = (long) score;
    average += score;
  }

  if (number_of_lines != 0)
  {
    average /= (double) number_of_lines;
  }
  return 1;
}


// model_links is supposed to be all the straight links extracted from
// a single model image.
long vLineMatchFrameStats::Initialize2(vDOLVS_Histogram * histogram,
                                        vMatrix<long> * pdf_matrix,
                                        vector<vStraightLinks> * model_links,
                                        vector<vLineSegmentMatch> * input_segments,
                                        long in_frame, long in_view)
{
  long pdf_limit = pdf_matrix->Cols() - 1;
  if (pdf_limit <= 0) return 0;
  frame = in_frame;
  view = in_view;
  number_of_lines = 2 * model_links->size();
  scores = vnew(long, number_of_lines);
  average = 0;
  long link;
  vArray(long) probabilities = pdf_matrix->Matrix(0);
  double factor = 100.0 / pdf_limit;

  // Go through all the links.
  for (link = 0; link < (long) model_links->size(); link++)
  {
    vLineSegment line1 = (*model_links)[link].line1;
    histogram->ProcessLinkSegment(line1);
    histogram->RankInputSegments(input_segments);
    double score = (*input_segments)[0].score;
    scores[2*link] = (long) (score * factor);
    average += scores[2*link];

    vLineSegment line2 = (*model_links)[link].line2;
    histogram->ProcessLinkSegment(line2);
    histogram->RankInputSegments(input_segments);
    score = (*input_segments)[0].score;
    scores[2*link+1] = (long) (score * factor);
    average += scores[2*link+1];
  }

  if (number_of_lines != 0)
  {
    average /= (double) number_of_lines;
  }
  return 1;
}


// model_links is supposed to be all the straight links extracted from
// a single model image.
long vLineMatchFrameStats::Initialize3(vDOLVS_Histogram * dolvs,
                                        vLineGrid * grid,
                                        vector<vStraightLinks> * model_links,
                                        double fraction, long in_frame, long in_view)
{
  frame = in_frame;
  view = in_view;
  number_of_lines = 2 * model_links->size();
  scores = vnew(long, number_of_lines);
  average = 0;
  long link;

  // Go through all the links.
  for (link = 0; link < (long) model_links->size(); link++)
  {
    vLineSegment line1 = (*model_links)[link].line1;
    double score = grid->SmallestFrequency(dolvs, line1, fraction);
    scores[2*link] = (long) score;
    average += score;

    vLineSegment line2 = (*model_links)[link].line2;
    score = grid->SmallestFrequency(dolvs, line2, fraction);
    scores[2*link+1] = (long) score;
    average += score;
  }

  if (number_of_lines != 0)
  {
    average /= (double) number_of_lines;
  }
  return 1;
}


long vLineMatchFrameStats::Print()
{
  long i;
  vPrint("%li lines, image %li,%li\n", number_of_lines, frame, view);
  for (i = 0; i < number_of_lines; i+=2)
  {
    vPrint("%li-%li, ", scores[i], scores[i+1]);
  }

  const double factors[] = {1, 0.5561, 0.5561, 0.5561, 0.7346, 0.7346, 0.8809, 0.8809, 1};
  double new_score;

  if (number_of_lines >= 8)
  {
    new_score = average;
  }
  else if (number_of_lines == 0)
  {
    new_score = 50;
  }
  else
  {
    new_score = (average - 50) * factors[number_of_lines] + 50;
  }
  vPrint("\navg = %.1lf, adjusted = %.1lf\n\n", average, new_score);
  return 1;
}


// Return the filename where the stats should be saved for the given frame, 
// with given start and period. path_name is the name of a video sequence.
char * vLineMatchStats::Filename(const char * path_name, long frame,
                                  long start, long period)
{
  char ending[100];
  sprintf(ending, ".dolvs_match_%li_%li_%li", frame, start, period);
  char * output_file = vReplaceEnding2(path_name, ending);
  return output_file;
}


// Load the stats for the given frame, start and period. 
// path_name is the name of a video sequence.
long vLineMatchStats::Load(const char * path_name, long frame,
                            long start, long period)
{
  char * filename = Filename(path_name, frame, start, period);
  long result = Load(filename);
  vdelete2(filename);
  return result;
}

  
long vLineMatchStats::Load(const char * filename)
{
  // Open input file
  FILE * fp = fopen(filename, vFOPEN_READ);
  if (fp == 0)
  {
    vPrint("Can't open %s\n", filename);
    return 0;
  }

  long items = 0;

  // Read number of frames;
  long number_of_frames = 0;
  items += vReadPCLongs(fp, &number_of_frames, 1);
  if (items != 1)
  {
    fclose(fp);
    vPrint("Can't open %s\n", filename);
    return 0;
  }

  // Read in the rest of the information
  vArray(long) frames = vnew(long, number_of_frames);
  vArray(long) views = vnew(long, number_of_frames);
  vArray(long) numbers_of_lines = vnew(long, number_of_frames);
  vArray2(long) scores = vnew(vArray(long), number_of_frames);

  items += vReadPCLongs(fp, vData(frames), number_of_frames);
  items += vReadPCLongs(fp, vData(views), number_of_frames);
  items += vReadPCLongs(fp, vData(numbers_of_lines), number_of_frames);

  long total_items = 3*number_of_frames + 1;
  long i;
  for (i = 0; i < number_of_frames; i++)
  {
    scores[i] = vnew(long, numbers_of_lines[i]);
    items += vReadPCLongs(fp, vData(scores[i]), numbers_of_lines[i]);
    total_items += numbers_of_lines[i];
  }
  fclose(fp);
  long success = 1;
  if (total_items != items)
  {
    vPrint("Error, total_items = %li, items = %li in vSaveLineMatchFrameStats\n",
            total_items, items);
    success = 0;
  }
  fclose(fp);

  // Package up the information into vLineMatchFrameStats objects
  frame_stats.erase(frame_stats.begin(), frame_stats.end());
  frame_stats = vector<vLineMatchFrameStats>(number_of_frames);
  for (i = 0; i < number_of_frames; i++)
  {
    vLineMatchFrameStats & stats = frame_stats[i];
    stats.frame = frames[i];
    stats.view = views[i];
    stats.number_of_lines = numbers_of_lines[i];
    stats.scores = scores[i];
    long j;
    double average = 0;
    for (j = 0; j < stats.number_of_lines; j++)
    {
      average += scores[i][j];
    }
    if (stats.number_of_lines != 0)
    {
      average /= stats.number_of_lines;
    }
    stats.average = average;
  }

  vdelete2(frames);
  vdelete2(views);
  vdelete2(numbers_of_lines);
  vdelete2(scores);
  return success;
}


// Save the stats for the given frame, start and period. 
// path_name is the name of a video sequence.
long vLineMatchStats::Save(const char * path_name, long frame,
                            long start, long period)
{
  char * filename = Filename(path_name, frame, start, period);
  long result = Save(filename);
  vdelete2(filename);
  return result;
}

  
long vLineMatchStats::Save(const char * filename)
{
  // Compute the information
  long number_of_frames = frame_stats.size();
  vArray(long) frames = vnew(long, number_of_frames);
  vArray(long) views = vnew(long, number_of_frames);
  vArray(long) numbers_of_lines = vnew(long, number_of_frames);
  vArray2(long) scores = vnew(vArray(long), number_of_frames);

  // Package up the information into a couple of big arrays
  long i;
  for (i = 0; i < number_of_frames; i++)
  {
    vLineMatchFrameStats & stats = frame_stats[i];
    frames[i] = stats.frame;
    views[i] = stats.view;
    long number_of_lines = stats.number_of_lines;
    numbers_of_lines[i] = number_of_lines;
    scores[i] = stats.scores;
  }

  // Write the information to disk
  FILE * fp = fopen(filename, vFOPEN_WRITE);
  if (fp == 0)
  {
    vPrint("Can't open %s\n", filename);
  }
  long items = 0;
  items += vWritePCLongs(fp, &number_of_frames, 1);
  items += vWritePCLongs(fp, vData(frames), number_of_frames);
  items += vWritePCLongs(fp, vData(views), number_of_frames);
  items += vWritePCLongs(fp, vData(numbers_of_lines), number_of_frames);

  long total_items = 3*number_of_frames + 1;
  for (i = 0; i < number_of_frames; i++)
  {
    items += vWritePCLongs(fp, vData(scores[i]), numbers_of_lines[i]);
    total_items += numbers_of_lines[i];
  }
  fclose(fp);
  long success = 1;
  if (total_items != items)
  {
    vPrint("Error, total_items = %li, items = %li in vSaveLineMatchFrameStats\n",
            total_items, items);
    success = 0;
  }
  vdelete2(frames);
  vdelete2(views);
  vdelete2(numbers_of_lines);
  vdelete2(scores);
  return success;
}


// Tries to see if there is a vLineMatchFrameStats object
// in frame_stats that is equal to frame_stats2 in frame, view,
// number_of_lines and scores. If there is no object equal
// in frame and view it returns -1. If there is an object that
// is equal in frame and view but differs on the rest it returns 
// 0. Otherwise it returns 1.
// This is useful for debugging.
long vLineMatchStats::Compare(vLineMatchFrameStats * frame_stats2)
{
  long number_of_frames = frame_stats.size();
  long i;
  // found will be set to 1 if and only if we find an object
  // in frame_stats that has frame and view equal to those of frame_stats2.
  long found = 0;
  for (i = 0; i < number_of_frames; i++)
  {
    vLineMatchFrameStats & stats = frame_stats[i];
    if ((stats.frame != frame_stats2->frame) 
        || (stats.view != frame_stats2->view)) continue;

    found = 1;
    stats.Print();
    if (stats.number_of_lines != frame_stats2->number_of_lines) continue;
    long j;
    long equal = 1;
    for (j = 0; j < stats.number_of_lines; j++)
    {
      if (stats.scores[j] != frame_stats2->scores[j]) equal = 0;
    }

    // We only get here if we found a vLineMatchFrameStats object
    // that is equal in content to frame_stats2.
    if (equal == 1) return 1;
  }

  // We only get here if we didn't find a vLineMatchFrameStats object
  // that is equal in content to frame_stats2.
  if (found == 0)
  {
    // Here we didn't find any object with frame and view equal to 
    // those of frame_stats2.
    return -1;
  }
  else
  {
    return 0;
  }

}


vMatchScores * vLineMatchStats::MatchScores()
{
  long number_of_frames = frame_stats.size();
  // The factors are used to adjust the scores of database views with a small number of
  // lines (2, 4, or 6). Obviously, it is much easier for a view with two lines to attain
  // a score of 10 than for a view with 10 lines. Each score for those lines is adjusted
  // as follows: score = (score - 50) * factor + 50. The factor is the standard deviation
  // of the scores for 8 lines divided by the standard deviation for the given number of lines.
  const double factors[] = {1, 0.5561, 0.5561, 0.5561, 0.7346, 0.7346, 0.8809, 0.8809, 1};
  if (number_of_frames == 0) return 0;
  vArray(vMatchScore) match_score_array = vnew(vMatchScore, number_of_frames);
  long i;
  for (i = 0; i < number_of_frames; i++)
  {
    vLineMatchFrameStats & stats = frame_stats[i];
    vMatchScore & match_score = match_score_array[i];
    match_score.frame = stats.frame;
    match_score.view = stats.view;
    
    // Maybe a more principled way would be to add everywhere
    // the average probability of lines.
    const long limit = 8;
    long  number_of_lines = stats.number_of_lines;
    if (number_of_lines >= limit)
    {
      match_score.score = (float) stats.average;
    }
    else if (number_of_lines == 0)
    {
      match_score.score = 50;
    }
    else
    {
      match_score.score = (float) (stats.average - 50) * (float) factors[number_of_lines] + (float) 50;
    }
    match_score.is_valid = 1;
  }
  vMatchScores * result = new vMatchScores(match_score_array, number_of_frames);
  return result;
}



// Reorder the line segments in segments so that the ones that
// are better (or worse) than d and o and l and v and s are placed first.
// direction specifies if they should be better (positive values)
// or worse (negative values or 0). We return the number of matches that
// are better (or worse). If one of dolvs is negative, we don't use it.
// A value of 0 or negative chooses bigger DOL and smaller VS.
// A value of 1 chooses smaller DOL and larger VS. A value of 2 chooses
// smaller DO and larger LVS.
long vCountLineSegments(vector<vLineSegmentMatch> * segments, long score,
                         long d, long o, long l, long v, long s, long direction)
{
  // vector<vLineSegmentMatch> category1, category2;
  long number = segments->size();
  vector<long> category1(number), category2(number);

  long counter1 = 0;
  long counter2 = 0;
  long i;
  vPrint("\n");
  for (i = 0; i < number; i++)
  {
    vLineSegmentMatch & segment = (*segments)[i];
//    vPrint("i = %li\r", i);
    long sd = segment.d;
    long so = segment.o;
    long sl = segment.l;
    long sv = segment.v;
    long ss = segment.s;

    long category = 1;
    if (direction <= 0)
    {
      if ((score >= 0) && (segment.score < score)) category = 2;
      if ((d >= 0) && (sd < d)) category = 2;
      if ((o >= 0) && (so < o)) category = 2;
      if ((l >= 0) && (sl < l)) category = 2;
      if ((v >= 0) && (sv > v)) category = 2;
      if ((s >= 0) && (ss > s)) category = 2;
    }

    else if (direction == 1)
    {
      if ((score >= 0) && (segment.score > score)) category = 2;
      if ((d >= 0) && (sd > d)) category = 2;
      if ((o >= 0) && (so > o)) category = 2;
      if ((l >= 0) && (sl > l)) category = 2;
      if ((v >= 0) && (sv < v)) category = 2;
      if ((s >= 0) && (ss < s)) category = 2;
    }

    else if (direction == 2)
    {
      if ((score >= 0) && (segment.score > score)) category = 2;
      if ((d >= 0) && (sd > d)) category = 2;
      if ((o >= 0) && (so > o)) category = 2;
      if ((l >= 0) && (sl < l)) category = 2;
      if ((v >= 0) && (sv < v)) category = 2;
      if ((s >= 0) && (ss < s)) category = 2;
    }

  
//    if (category == 1) category1.push_back(segment);
//    else (category2.push_back(segment));
    if (category == 1)
    {
      category1[counter1++] = i;
    }
    else
    {
      category2[counter2++] = i;
    }
  }

  if (counter1 == 0) return 1;
  vPrint("\n");
  i = counter1-1;
  long j = 0;
  long index1 = category1[i];
  long index2 = category2[j];
  while(index1 > index2)
  {
    // Swap the two segments.
    vLineSegmentMatch segment1 = (*segments)[index1];
    (*segments)[index1] = (*segments)[index2];
    (*segments)[index2] = segment1;
    i--;
    j++;
    if (i < 0) break;
    if (j == counter2) break;
    index1 = category1[i];
    index2 = category2[j];
  }

  return counter1;
}



// Erase all line segments except for the ones that
// are better (or worse) than the given d and o and l and v and s.
// direction specifies if they should be better (positive values)
// or worse (negative values or 0). We return the number of matches that
// are better (or worse). If one of dolvs is negative, we don't use it.
// A value of 0 or negative chooses bigger DOL and smaller VS.
// A value of 1 chooses smaller DOL and larger VS. A value of 2 chooses
// smaller DO and larger LVS.
long vSelectLineSegments(vector<vLineSegmentMatch> * segments, long score,
                          long d, long o, long l, long v, long s, long direction)
{
  long result = vCountLineSegments(segments, score, d, o, l, v, s, direction);
  if (result > 0)
  {
    segments->erase(segments->begin() + result, segments->end());
  }
  
  return result;
}


// If the database has 107328 images, the result is of size 107328x30x4.
// band corresponds to database image. row corresponds to a line for a 
// link corresponding to that row.
// cols are the coordinates of the two endpoints of that line 
// (in the order row1, col1, row2, col2). We use some conventions.
// A line that covers more than one link is saved at the position
// corresponding to the link with the lowest id. The entry at another link
// will be (1, i, 0, 0) where 1 signifies that the link has a line, that is shared.
// i is the row (0 to 29) where the line is stored.
// For some links no lines are found. For those links we have entries
// (0,0,0,0) in the corresponding rows.
v3dMatrix<uchar> * vAllModelLines(vVideoFile * database, 
                                     vVideoFile * edges)
{
  static const long views_per_frame = 4;
  long frames = database->NumberOfFrames();
  if (edges->NumberOfFrames() != frames)
  {
    vPrint("edges has different number of frames than database\n");
    return 0;
  }
  long images = frames * views_per_frame;

  long bands = frames * views_per_frame;
  static const long rows = 30;
  static const long cols = 4;
  v3dMatrix<uchar> * model_lines_matrix = new v3dMatrix<uchar>(rows, cols, bands);
  vWriteValue(model_lines_matrix, (uchar) 0);

  long frame, view;
  long errors = 0;
  color_image * model_image = 0;
  grayscale_image * model_edges = 0;
  color_image * rotated_model = 0;
  grayscale_image * rotated_edges = 0;
  vPrint("\n");
  for (frame = 0; frame < frames; frame++)
  {
    vPrint("\n\n!!!!!!!!!!!!!!!!!!!!!!!frame = %li!!!!!!!!!!!!!!!!!!!!\n\n", frame);
    color_image * model_image = database->ReadFrame(frame);
    if (model_image == 0)
    {
      vPrint("Couldn't read frame %li from database\n", frame);
      errors++;
      break;
    }
    grayscale_image * model_edges = edges->ReadFrameGray(frame);
    if (model_edges == 0)
    {
      vPrint("Couldn't read frame %li from database\n", frame);
      errors++;
      break;
    }

    grayscale_image * dilated = vDilateImage(model_edges, 1, 255);
    for (view = 0; view < views_per_frame; view++)
    {
      long angle = view * 90;
      rotated_model = (color_image *) vRotateRight(model_image, angle);
      rotated_edges = (grayscale_image *) vRotateRight(dilated, angle);
      vector<vStraightLinks> link_lines;
      vGoodLinksForLines2(rotated_model, rotated_edges, &link_lines);
      vWriteModelLines(model_lines_matrix, &link_lines, 
                        frame * views_per_frame + view);
      vdelete(rotated_model);
      vdelete(rotated_edges);
    }
    vdelete(model_image);
    vdelete(model_edges);
    vdelete(dilated);
    model_image = 0;
    model_edges = 0;
    if (frame % 1000 == 0)
    {
      char * base = "g:\\users\\athitsos\\sequences\\model_lines";
      char * number_string = vLongToString(frame);
      char * filename = vMergeStrings3(base, ".", number_string);

      model_lines_matrix->Write(filename);
      vdelete2(number_string);
      vdelete2(filename);
    }
  }
  vdelete(model_image);
  vdelete(model_edges);
  vPrint("\n");

  return model_lines_matrix;
}


// model_lines_matrix is a matrix of the kind that we create and return
// at vAllModelLines. This function writes information for a particular
// database image, to band image_id, using the information about lines
// that is held in link_lines. image_id is a number specifying a 
// database image.
long vWriteModelLines(v3dMatrix<uchar> * model_lines_matrix, 
                       vector<vStraightLinks> * link_lines, 
                       long image_id)
{
  static const long adjust = 2;
  vArray2(uchar) image_lines = model_lines_matrix->Matrix2(image_id);
  long number_of_lines = link_lines->size();
  long i;
  for (i = 0; i < number_of_lines; i++)
  {
    vStraightLinks & links = (*link_lines)[i];
    long id1 = links.link1 - adjust;
    long id3 = links.link3 - adjust;
    if (links.line1.row1 == -1) continue;
    if (links.line2.row1 == -1) continue;
    long row1 = 2*id1;
    image_lines[row1][0] = (uchar) links.line1.row1;
    image_lines[row1][1] = (uchar) links.line1.col1;
    image_lines[row1][2] = (uchar) links.line1.row2;
    image_lines[row1][3] = (uchar) links.line1.col2;
    long row2 = row1+1;
    image_lines[row2][0] = (uchar) links.line2.row1;
    image_lines[row2][1] = (uchar) links.line2.col1;
    image_lines[row2][2] = (uchar) links.line2.row2;
    image_lines[row2][3] = (uchar) links.line2.col2;

    long id;
    for (id = id1 + 1; id <= id3; id++)
    {
      long row3 = 2*id;
      image_lines[row3][0] = 1;
      image_lines[row3][1] = (uchar) row1;
      long row4 = row3 + 1;
      image_lines[row4][0] = 1;
      image_lines[row4][1] = (uchar) row2;
    }
  }
    
  return 1;
}


vector<vStraightLinks> * vReadModelLines3(v3dMatrix<uchar> * model_lines_matrix,
                                            long model_frame, long model_view)
{
  long image_id = model_frame * 4 + model_view;
  return vReadModelLines(model_lines_matrix, image_id);
}


// vReadModelLines is the inverse of vWriteModelLines: it creates
// vStraightLinks objects based on the entries in the 
// model_line_matrix.
vector<vStraightLinks> * vReadModelLines(v3dMatrix<uchar> * model_lines_matrix,
                                           long image_id)
{
  static const long adjust = 2;
  vector<vStraightLinks> * result = new vector<vStraightLinks>;
  if ((image_id < 0) || (image_id >= model_lines_matrix->Bands()))
  {
    return 0;
  }

  vArray2(uchar) image_lines = model_lines_matrix->Matrix2(image_id);
  long rows = model_lines_matrix->Rows();
  long row;
  for (row = 0; row < rows; row = row+2)
  {
    if (image_lines[row][0] == 0) continue;
    if (image_lines[row][0] == 1) continue;

    vStraightLinks links;
    links.link1 = row/2 + adjust;
    
    long row1 = row;
    links.line1.row1 = image_lines[row1][0];
    links.line1.col1 = image_lines[row1][1];
    links.line1.row2 = image_lines[row1][2];
    links.line1.col2 = image_lines[row1][3];

    long row2 = row1+1;
    links.line2.row1 = image_lines[row2][0];
    links.line2.col1 = image_lines[row2][1];
    links.line2.row2 = image_lines[row2][2];
    links.line2.col2 = image_lines[row2][3];
    
    long row3 = row1+2;
    long row4 = row1+4;
    if ((row3 < rows) && (image_lines[row3][0] == 1))
    {
      if ((row4 < rows) && (image_lines[row4][0] == 1))
      {
        links.link2 = links.link1+1;
        links.link3 = links.link1+2;
      }
      else
      {
        links.link2 = links.link1+1;
        links.link3 = links.link1+1;
      }
    }
    else
    {
      links.link2 = links.link1;
      links.link3 = links.link1;
    }

    result->push_back(links);
  }
    
  return result;
}


// We compute all model lines calling vAllModelLines, and we save
// them in a file that shares name with database, but has the 
// extension .model_lines.
v3dMatrix<uchar> * vSaveAllModelLines(vVideoFile * database,
                                         vVideoFile * edges)
{
  v3dMatrix<uchar> * model_lines = vAllModelLines(database, edges);
  const char * database_filename = database->GetFilename();
  char * model_filename = vReplaceEnding2(database_filename, ".model_lines");
  long result = model_lines->Write(model_filename);
  if (result <= 0)
  {
    vPrint("Error in saving model lines to %s\n", model_filename);
  }
  vdelete2(model_filename);
  return model_lines;
}


// We compute all model lines calling vAllModelLines, and we save
// them in a file that shares name with database, but has the 
// extension .model_lines.
v3dMatrix<uchar> * vLoadAllModelLines(vVideoFile * database)
{
  const char * database_filename = database->GetFilename();
  char * model_filename = vReplaceEnding2(database_filename, ".model_lines");
  v3dMatrix<uchar> * result = v3dMatrix<uchar>::Read(model_filename);
  
  if (result == 0)
  {
    vPrint("Error in reading model lines from %s\n", model_filename);
  }
  vdelete2(model_filename);
  return result;
}


// Return the line that is saved at the corresponding index in
// the model_lines_matrix. index -> (band, row), where
// band = index/30 and row = index % 30. If there is 
// no line saved at that position, we return a segment
// with -1 for row1.
vLineSegment vReadLineSegment(v3dMatrix<uchar> * model_lines_matrix,
                                long index)
{
  vLineSegment result;
  result.row1 = -1;
  if (index < 0) return result;
  long bands = model_lines_matrix->Bands();
  long rows = model_lines_matrix->Rows();
  long total = bands * rows;
  if (index >= total) return result;
  long band = index / rows;
  long row = index % rows;
  vArray3(uchar) model_lines = model_lines_matrix->Matrix3();
  if (model_lines[band][row][0] <= 1)
  {
    return result;
  }
  result.row1 = model_lines[band][row][0];
  result.col1 = model_lines[band][row][1];
  result.row2 = model_lines[band][row][2];
  result.col2 = model_lines[band][row][3];
  
  return result;
}


// For every index from {start, start + period, start+2*period, ..., bands * rows-1}
// we read the line segment that is described there (if any) and we save it 
// at segments. 
long vSampleLineSegments(v3dMatrix<uchar> * model_lines_matrix,
                          long start, long period,
                          vector<vLineSegment> * segments)
{
  if (period <= 0) return 0;
  long bands = model_lines_matrix->Bands();
  long rows = model_lines_matrix->Rows();
  long total = bands * rows;

  vArray3(uchar) model_lines = model_lines_matrix->Matrix3();
  vLineSegment segment;
  
  long index;
  for (index = start; index < total; index += period)
  {
    long band = index / rows;
    long row = index % rows;
    if (model_lines[band][row][0] <= 1)
    {
      continue;
    }

    segment.row1 = model_lines[band][row][0];
    segment.col1 = model_lines[band][row][1];
    segment.row2 = model_lines[band][row][2];
    segment.col2 = model_lines[band][row][3];
    segments->push_back(segment);
  }  

  return segments->size();
}


// Compute line match stats (i.e. the information about how well
// each model line matched the input) for the 
// hand pose images specified by database_frames and database_views.
// This is like vComputeLineMatchStats3, but is slower and more exact, 
// because it does not use a grid.
vLineMatchStats * vComputeLineMatchStats(vDOLVS_Histogram * histogram,
                                           v3dMatrix<uchar> * all_model_lines,
                                           vector<long> * database_frames,
                                           vector<long> * database_views)
{
  // Compute the information
  long number_of_frames = database_frames->size();
  vLineMatchStats * line_match_stats = new vLineMatchStats();

  long i;
  vPrint("\n");
  for (i = 0; i < number_of_frames; i++)
  {
    vPrint("Processing match %li of %li\r", i+1, number_of_frames);
    long frame = (*database_frames)[i];
    long view = (*database_views)[i];
    long index = frame * 4 + view;
    vector<vStraightLinks> * model_lines = vReadModelLines(all_model_lines, index);
    vLineMatchFrameStats frame_stats(histogram, model_lines,
                                      frame, view);
    vdelete(model_lines);
    line_match_stats->frame_stats.push_back(frame_stats);
  }

  vPrint("\n");
  return line_match_stats;
}


// Computes line match stats for all hand pose images in the database.
// It basically sets up the right arguments and calls 
// vComputeLineMatchStats3.
vLineMatchStats * vComputeLineMatchStats2(vDOLVS_Histogram * histogram,
                                            vLineGrid * grid,
                                            v3dMatrix<uchar> * all_model_lines,
                                            double fraction)
{
  const long frames = 107328;
  vector<long> database_frames(frames);
  vector<long> database_views(frames);
  long i;
  for (i = 0; i < frames; i++)
  {
    long frame = i / 4;
    long view = i % 4;
    database_frames[i] = frame;
    database_views[i] = view;
  }
  return vComputeLineMatchStats3(histogram, grid, all_model_lines, 
                                  &database_frames, &database_views, fraction);
}


// This is like vComputeLineMatchStats, but it is faster and a bit 
// inaccurate, because it uses a grid.
vLineMatchStats * vComputeLineMatchStats3(vDOLVS_Histogram * histogram,
                                            vLineGrid * grid,
                                            v3dMatrix<uchar> * all_model_lines,
                                            vector<long> * database_frames,
                                            vector<long> * database_views,
                                            double fraction)
{
  // Compute the information
  long number_of_frames = database_frames->size();
  vLineMatchStats * line_match_stats = new vLineMatchStats();

  long i;
  vPrint("\n");
  for (i = 0; i < number_of_frames; i++)
  {
    if (i % 100 == 0)
    {
      vPrint("Processing match %li of %li\r", i+1, number_of_frames);
    }
    long frame = (*database_frames)[i];
    long view = (*database_views)[i];
    long index = frame * 4 + view;
    vector<vStraightLinks> * model_lines = vReadModelLines(all_model_lines, index);
    vLineMatchFrameStats frame_stats(histogram, grid, model_lines, fraction,
                                      frame, view);
    vdelete(model_lines);
    line_match_stats->frame_stats.push_back(frame_stats);
  }

  vPrint("\n");
  return line_match_stats;
}


long vSaveLineMatchFrameStats(vDOLVS_Histogram * histogram,
                               v3dMatrix<uchar> * all_model_lines,
                               long start, long period, long frame,
                               vector<long> * database_frames,
                               vector<long> * database_views,
                               vector<vLineSegmentMatch> * input_segments,
                               const char * input_filename)
{
  // Compute the information
  long number_of_frames = database_frames->size();
  vLineMatchStats line_match_stats;

  long i;
  for (i = 0; i < number_of_frames; i++)
  {
    vPrint("Processing match %li\n", i);
    long frame = (*database_frames)[i];
    long view = (*database_views)[i];
    long index = frame * 4 + view;
    vector<vStraightLinks> * model_lines = vReadModelLines(all_model_lines, index);
    vLineMatchFrameStats frame_stats(histogram, model_lines,
                                      frame, view);
    frame_stats.Print();
    vdelete(model_lines);
    line_match_stats.frame_stats.push_back(frame_stats);
  }

  long success = line_match_stats.Save(input_filename, frame, start, period); 
  return success;
}


// Returns a filename to which we should save the existential_DOLVS
// histogram of the given frame from video file "path_name", if
// the existential_DOLVS histogram was computed with a sampling
// of model lines described by start and period.
char * vDOLVSFrameFile(const char * path_name, long frame, 
                        long start, long period)
{
  char dolvs_ending[100];
  sprintf(dolvs_ending, ".dolvs_%li_%li_%li", frame, start, period);
  char * dolvs_filename = vReplaceEnding2(path_name, dolvs_ending);
  return dolvs_filename;
}


char * vDOLVS_PDF_File(const char * path_name, long frame, 
                        long start, long period)
{
  char ending[100];
  sprintf(ending, ".dolvs_pdf_%li_%li_%li", frame, start, period);
  char * pdf_filename = vReplaceEnding2(path_name, ending);
  return pdf_filename;
}


long vFrequencyToPDF(vMatrix<long> & dolvs_pdf, 
                      vector<vLineSegmentMatch> * line_segments)
{
  if (dolvs_pdf.Bands() != 1) return 0;
  if (dolvs_pdf.Rows() != 1) return 0;
  long number = line_segments->size();
  vArray(long) pdf = dolvs_pdf.Matrix();
  long pdf_limit = dolvs_pdf.Cols() - 1;
  double factor = 100.0 / pdf_limit;

  long success = 1;
  long i;
  for (i = 0; i < number; i++)
  {
    vLineSegmentMatch & match = (*line_segments)[i];
    long frequency = vRound(match.score);
    if ((frequency < 0) || (frequency > pdf_limit))
    {
      success = 0;
      break;
    }
    match.score = (float) pdf[frequency] * (float) factor;
  }
  return success;
}


vArray3(uchar) S_InitScratchTable(vector<vLineSegmentMatch> * segments)
{
  static const long orientation_bins = 24;
  static const long length_bins = 20;
  static const long length_per_pixel = orientation_bins * length_bins;
  static const long rows = 256;
  static const long cols = 256;

  // Count number of distinct center row, center_col locations
  vArray2(uchar) positions;
  vHyperArray2(&positions, rows, cols);
  long number = segments->size();

  // First, put zeros in their corresponding bins
  long i;
  for (i = 0; i < number; i++)
  {
    vLineSegmentMatch & match = (*segments)[i];
    long row = match.center_row;
    long col = match.center_col;
    positions[row][col] = 0;
  }

  // Now count them
  long counter = 0;
  for (i = 0; i < number; i++)
  {
    vLineSegmentMatch & match = (*segments)[i];
    long row = match.center_row;
    long col = match.center_col;
    if (positions[row][col] != 0) continue;
    positions[row][col] = 1;
    counter++;
  }

  // Allocate memory for biggest possible number of segments
  long length = counter * length_per_pixel;
  vArray(uchar) big_array = vnew(uchar, length);
  bzero(vData(big_array), length);

  // Create the result array.
  vArray3(uchar) result;
  vHyperArray2(&result, rows, cols);

  // Set all indices to zero.
  long j;
  for (i = 0; i < rows; i++)
  {
    for (j = 0; j < cols; j++)
    {
      result[i][j] = vZero(uchar);
    }
  }

  // Now set each result[row][col] to the appropriate offset.
  long offset = 0;
  for (i = 0; i < number; i++)
  {
    vLineSegmentMatch & match = (*segments)[i];
    long row = match.center_row;
    long col = match.center_col;
    if (!predicate_zero(result[row][col])) continue;
    result[row][col] = big_array + offset;
    offset += length_per_pixel;
  }

  vDelete2(positions, rows);
  return result;
}


long S_DeleteScratchTable(vArray3(uchar) table, vector<vLineSegmentMatch> * segments)
{
  // Find the first non-zero entry in table[row][col] and delete it.
  // Then delete the table.
  if (segments->size() == 0) return 0;
  vLineSegmentMatch & match = (*segments)[0];
  long row = match.center_row;
  long col = match.center_col;
  vdelete2(table[row][col]);

  vDelete2(table, 256);
  return 1;
}


// Creates clusters of the line segments in line_segments, and puts
// the best (meaning highest ranked) segment of each cluster into
// samples. A cluster is defined kind of loosely. We scan line segments,
// in order of their appearance in line_segments, and for every segment S
// that we find that doesn't have a cluster yet, we create a cluster
// centered around it, we put the segment S into samples, and we mark
// the area around the segment S as already belonging to a cluster. How
// large an area we mark is specified using location_diff, orientation_diff
// length_diff. Any segment whose DOL difference to S is less than or equal
// to these three parameters will be represented by the segment S.
long vLineSegmentSamples(vector<vLineSegmentMatch> * line_segments,
                          vector<vLineSegmentMatch> * samples, 
                          long location_diff, long orientation_diff,
                          long length_diff)
{
  static const long orientation_bins = 24;
  static const long length_bins = 20;
  static const long max_length = 100;
  const long length_offset = 5;
  static const double orientation_factor = ((double) orientation_bins) / M_PI;
  static const double length_factor = ((double) length_bins) / (double) max_length;
  static const long row_limit = 255;
  static const long col_limit = 255;
  static const long length_limit = length_bins - 1;
  static const long orientation_limit = orientation_bins - 1;

  if (location_diff < 0) return 0;
  if (orientation_diff < 0) return 0;
  if (length_diff < 0) return 0;
  if (line_segments->size() == 0) return 0;

  vArray3(uchar) scratch = S_InitScratchTable(line_segments);

  long number = line_segments->size();
  long i;
  for (i = 0; i < number; i++)
  {
    vLineSegmentMatch & match = (*line_segments)[i];
    long center_row = match.center_row;
    long center_col = match.center_col;
    long orientation = vRound(match.orientation * orientation_factor);
    long length = vRound((match.length - length_offset) * length_factor);
    long index = orientation * length_bins + length;
    if (scratch[center_row][center_col][index] != 0) continue;

    samples->push_back(match);
    long row_start = Max(center_row - location_diff, (long) 0);
    long col_start = Max(center_col - location_diff, (long) 0);
    long orientation_start = (orientation + orientation_bins - orientation_diff) % orientation_bins;
    while(orientation_start < 0)
    {
      orientation_start += orientation_bins;
    }
    long length_start = Max(length - length_diff, (long) 0);

    long row_end = Min(center_row + location_diff, row_limit);
    long col_end = Min(center_col + location_diff, col_limit);
    long orientation_end = (orientation + orientation_diff + 1) % orientation_bins;
    long length_end = Min(length + length_diff, length_bins-1);

    long row, col, o, l;
    for (row = row_start; row <= row_end; row++)
    {
      for (col = col_start; col <= col_end; col++)
      {
        if ((row < 0) || (row > row_limit) || 
            (col < 0) || (col > col_limit))
        {
          exit_error("error\n");
        }
        // scratch[row][col] == 0 if there are no segments for that row and col.
        if (predicate_zero (scratch[row][col])) continue;
        for (o = orientation_start; o != orientation_end; o = (o+1)%orientation_bins)
        {
          for (l = length_start; l <= length_end; l++)
          {
            index = o*length_bins + l;
            if ((row < 0) || (row > row_limit) || 
                (col < 0) || (col > col_limit) ||
                ( index < 0) || (index >= 480))
            {
              exit_error("error\n");
            }
            scratch[row][col][o*length_bins+l] = 1;
          }
        }
      }
    }
  }

  long number_of_samples = samples->size();
  vPrint("%li samples chosen\n", number_of_samples);
  S_DeleteScratchTable(scratch, line_segments);
  return 1;
}


static inline long S_SegmentGequal(vLineSegmentMatch & match1, 
                                   vLineSegmentMatch & match2)
{
  if ((match1.d <= match2.d) && (match1.o <= match2.o) &&
      (match1.l <= match2.l) && (match1.v >= match2.v) &&
      (match1.s >= match2.s))
  {
    return 1;
  }
  else
  {
    return 0;
  }
}


// Pick all segments S in line segments that are maximal elements (if 
// we have identical segments pick one of them only), and put them into 
// best segments.
long vSelectBestSegments(vector<vLineSegmentMatch> * line_segments,
                          vector<vLineSegmentMatch> * best_segments)
{
  // if line_segments is empty, we are done
  long size1 = line_segments->size();
  if (size1 == 0) return 0;  

  vector<vLineSegmentMatch *> current_segments(size1);

  // Copy segments from line_segments to current_segments;
  long i;
  for (i = 0; i < size1; i++)
  {
    current_segments[i] = &((*line_segments)[i]);
  }

  long current_size = size1;
  long iterations = 0;
  vPrint("iterations = %li, segments = %li\n", iterations, current_size);
  while (current_size != 0)
  {
    // Find a maximal element
    vLineSegmentMatch current_max = *(current_segments[0]);
    for (i = 1; i < current_size; i++)
    {
      vLineSegmentMatch & current = *(current_segments)[i];
      if (S_SegmentGequal(current, current_max))
      {
        current_max = current;
      }
    }

    // Add maximal element to best segments
    best_segments->push_back(current_max);

    long next_size = 0;
    // Put to the beginnign of current_segments the segments 
    // that are not less than the maximal element, and count the 
    // number of those segments.
    for (i = 1; i < current_size; i++)
    {
      vLineSegmentMatch * current = current_segments[i];
      if (S_SegmentGequal(current_max, *current) == 0)
      {
        current_segments[next_size] = current;
        next_size++;
      }
    }
  
    current_size = next_size;
    iterations++;
    vPrint("iterations = %li, segments = %li\n", iterations, current_size);
  }

  return 1;
}


vMatchScores * vLineChamferScores(grayscale_image * the_image, vector<long> * frames,
                                    vector<long> * views, 
                                    vLineChamferInfo * info)
{
  if (info->input_points == 0)
  {
    // Get the location of non-zero points (edge pixels) in the image.
    vGetNonZeroPoints4(the_image,&(info->input_rows), 
                        &(info->input_cols), &(info->input_points));
  }

  // If the distance transform is not part of the info, compute it.
  if (info->dt == 0)
  {
    info->dt = vCVDistanceTransform(the_image);
  }

  long number_of_frames = frames->size();
  vArray(vMatchScore) scores = vnew(vMatchScore, number_of_frames);
  long i;
  for (i = 0; i < number_of_frames; i++)
  {
    vMatchScore & score = scores[i];
    score.frame = (*frames)[i];
    score.view = (*views)[i];
    
    info->pairings = vFindLinePairings(score.frame, score.view, 
                                        info->all_model_lines, 
                                        info->input_lines, info->dolvs);
    score.score = (float) vLineChamferDistance(info, score.frame, score.view);
  }

  vMatchScores * match_scores = new vMatchScores(scores, number_of_frames);
  return match_scores;
}


vector<vLinePairing> * 
vFindLinePairings(long model_frame, long model_view,
                   v3dMatrix<uchar> * all_model_lines,
                   vector<vLineSegmentMatch> * input_lines,
                   vDOLVS_Histogram * dolvs)
{
  vector<vStraightLinks> * model_lines = 
    vReadModelLines3(all_model_lines, model_frame, model_view);
  long number_of_model_lines = model_lines->size();
  
  vector<vLinePairing> * result = new vector<vLinePairing>;
  long i;
  for (i = 0; i < number_of_model_lines; i++)
  {
    vLinePairing pairing;
    vStraightLinks & link_lines = (*model_lines)[i];
    vLineSegment & model_line1 = link_lines.line1;
    vLineSegment & model_line2 = link_lines.line2;
    long best_index = 0;
    if (model_line1.row1 >= 0)
    {
      long frequency = dolvs->SmallestMatchFrequency(model_line1, &best_index);
      pairing.model_line = model_line1;
      pairing.input_line = (*input_lines)[best_index];
      result->push_back(pairing);
    }

    if (model_line2.row1 >= 0)
    {
      long frequency = dolvs->SmallestMatchFrequency(model_line2, &best_index);
      pairing.model_line = model_line1;
      pairing.input_line = (*input_lines)[best_index];
      result->push_back(pairing);
    }
  }

  return result;
}


double vLineChamferDistance(vLineChamferInfo * info, long frame, long view)
{

  vArray(long) input_leftover_rows;
  vArray(long) input_leftover_cols;
  long input_leftover_number = 0;

  vArray(long) model_rows;
  vArray(long) model_cols;
  long model_points = info->point_sets->GetPoints(frame, view, 
                                                  &model_rows, &model_cols);
  long model_leftover_number = info->point_sets->LeftoverLength(frame);

  vLeftoverPoints2(info->input_rows, info->input_cols, info->input_points, 
                    info->pairings, 2,
                    &input_leftover_rows, &input_leftover_cols, 
                    &input_leftover_number);
                 
  double score1 = vOptimizedChamfer(model_rows, model_cols, 
                                       model_leftover_number,
                                       info->input_rows, info->input_cols,
                                       info->input_points, -1);
  double score2 = vModelToImagePairingScore(info->pairings);

  double score3 = vOptimizedChamfer(input_leftover_rows, input_leftover_cols, 
                                       input_leftover_number,
                                       model_rows, model_cols, model_points, -1);
  double score4 = vImageToModelPairingScore(info->pairings);
  
  long model_line_points = model_points - model_leftover_number;
  long input_line_points = info->input_points - input_leftover_number;

  double model_to_input_score = ((double) model_leftover_number) * score1;
  model_to_input_score += ((double) model_line_points) * score2;
  if (model_points != 0)
  {
    model_to_input_score /= ((double) model_points);
  }

  double input_to_model_score = ((double) input_leftover_number) * score1;
  input_to_model_score += ((double) input_line_points) * score2;
  if (info->input_points != 0)
  {
    input_to_model_score /= ((double) info->input_points);
  }

  double matching_score = model_to_input_score + input_to_model_score;
  vdelete2(model_rows);
  vdelete2(model_cols);
  vdelete2(input_leftover_rows);
  vdelete2(input_leftover_cols);

  return matching_score;
}


// This is just a wrapper around vLeftoverPoints, where instead of passing
// in a vector of line segments, we pass in a vector of line segment pairs.
// Each of those pairs contains a line from the model and its best match
// in the image. Based on the value model_or_image (1 for model, 2 for image)
// we select the appropriate line segments from pairings, and then call
// vLeftoverPoints.
long vLeftoverPoints2(vArray(long) point_rows, vArray(long) point_cols, 
                       long number_of_points, vector<vLinePairing> * pairings, 
                       long model_or_image,
                       vArray(long) * leftover_rowsp, 
                       vArray(long) * leftover_colsp, 
                       long * leftover_numberp)
{
  // The only valid values for model_or_image are 1 (for model)
  // and 2 (for image)
  if ((model_or_image != 1) && (model_or_image != 2))
  {
    return 0;
  }

  vector<vLineSegmentLong> line_segments;
  long number = pairings->size();
  long i;
  if (model_or_image == 1)
  {
    for (i = 0; i < number; i++)
    {
      line_segments.push_back(vLineSegmentLong((*pairings)[i].model_line));
    }
  }
  else if (model_or_image == 2)
  {
    for (i = 0; i < number; i++)
    {
      line_segments.push_back((*pairings)[i].input_line.segment);
    }
  }
  
  return vLeftoverPoints(point_rows, point_cols, number_of_points,
                          &line_segments, leftover_rowsp,
                          leftover_colsp, leftover_numberp);
}


// We have a set "number_of_points" points, whose coordinates are stored
// in point_rows and point_cols. We want to figure out which of those
// points do NOT belong to the line segments stored in lines, and store
// the coordinates of these leftover points into leftover_rowsp and
// leftover_colsp (which will contain arrays that we create in this
// function). We put the number of those leftover points into 
// leftover_numberp.
// This function is actually not useful for what I originally wrote it, because
// we actually have to exclude all pixels that were used to fit this function.
// I just keep it here in case it turns out to be useful for something later.
long vLeftoverPointsTemp(vArray(long) point_rows, vArray(long) point_cols, 
                          long number_of_points, vector<vLineSegmentLong> * lines, 
                          vArray(long) * leftover_rowsp, 
                          vArray(long) * leftover_colsp, 
                          long * leftover_numberp)
{
  // This function doesn't do the right thing. We need to exclude not only the 
  // points marked by those lines, but the points in an entire search region.
  exit_error("error, this function is incorrect and should not be used.\n");

  // We assume that all coordinates have values between 0 and max_dimension-1.
  // If not, we will catch that and exit.
  const long max_dimension = 256;
  static v3dMatrix<uchar> scratch_matrix(max_dimension, max_dimension, 1);

  // Create result arrays;
  vArray(long) leftover_rows = vnew(long, number_of_points);
  vArray(long) leftover_cols = vnew(long, number_of_points);

  // We need to figure out for each point in point_rows and point_cols if
  // it belongs to one of the segments. 
  // The strategy is as follows: First, mark the pixels in scratch that correspond
  // to points in point_rows/point_cols to 0. Then, mark the pixels in scratch
  // that correspond to line segments with a non-zero value. Finally, go over
  // the points in point_rows/point_cols again, and choose the ones for which
  // the entry in scratch is still 0.

  // So, first, mark all scratch pixels that belong to point_rows/point_cols with 0.
  vArray2(uchar) scratch = scratch_matrix.Matrix2(0);
  long i;
  for (i = 0; i < number_of_points; i++)
  {
    long row = point_rows[i];
    long col = point_cols[i];
    scratch[row][col] = 0;
  }
  
  // Now mark all pixels that belong to one of the line segments with a 
  // non-zero value.
  long number_of_lines = lines->size();
  for (i = 0; i < number_of_lines; i++)
  {
    vLineSegmentLong & segment = (*lines)[i];
    vDrawLine1(&scratch_matrix, segment.row1, segment.col1,
                segment.row2, segment.col2, (uchar) 255);
  }

  // Now choose all points in point_rows/point_cols whose entry in scratch
  // is still 0.

  long counter = 0;
  for (i = 0; i < number_of_points; i++)
  {
    long row = point_rows[i];
    long col = point_cols[i];
    if (scratch[row][col] != 0) continue;
    leftover_rows[counter] = row;
    leftover_cols[counter] = col;
    counter++;
  }

  // Store the results into the argument pointers
  *leftover_rowsp = leftover_rows;
  *leftover_colsp = leftover_cols;
  *leftover_numberp = counter;
  return 1;
}


// We have a set "number_of_points" points, whose coordinates are stored
// in point_rows and point_cols. We want to figure out which of those
// points do NOT belong to the line segments stored in lines, and store
// the coordinates of these leftover points into leftover_rowsp and
// leftover_colsp (which will contain arrays that we create in this
// function). We put the number of those leftover points into 
// leftover_numberp.
// In general, a point belongs to a line segment if its distance from 
// the center of the segment is <= half the length of the segment,
// and the angle the point makes with the orientation of the segment
// is less than 0.1 radians.
long vLeftoverPoints(vArray(long) point_rows, vArray(long) point_cols, 
                      long number_of_points, vector<vLineSegmentLong> * lines, 
                      vArray(long) * leftover_rowsp, 
                      vArray(long) * leftover_colsp, 
                      long * leftover_numberp)
{
  // Make sure certain values are precomputed.
  static const long junk1 = vPrecomputedLengths::PrecomputeValues(256, 256);
  static const long junk2 = vPrecomputedAngles::PrecomputeValues(256, 256);
  vArray2(long) lengths = vPrecomputedLengths::values;
  vArray2(float) angles = vPrecomputedAngles::values;

  // Create result arrays;
  vArray(long) leftover_rows = vnew(long, number_of_points);
  vArray(long) leftover_cols = vnew(long, number_of_points);
  long number_of_lines = lines->size();
  // Precompute for each segment in lines its center and orientation
  vArray(long) center_rows = vnew(long, number_of_lines);
  vArray(long) center_cols = vnew(long, number_of_lines);
  vArray(double) orientations = vnew(double, number_of_lines);
  vArray(long) half_lengths = vnew(long, number_of_lines);
  const double orientation_threshold = 0.1;

  long i, j;
  for (i = 0; i < number_of_lines; i++)
  {
    vLineSegmentLong & segment = (*lines)[i];
    center_rows[i] = (segment.row1 + segment.row2) / 2;
    center_cols[i] = (segment.col1 + segment.col2) / 2;
    long drow = segment.row1 - segment.row2;
    long dcol = segment.col1 - segment.col2;
    long abs_drow = vAbs(drow);
    long abs_dcol = vAbs(dcol);
    half_lengths[i] = lengths[abs_drow][abs_dcol] / 2;
    // This can be optimized by precomputing angles.
    orientations[i] = angles[drow][dcol];
  }

  const float half_pi = (float) (M_PI/2);

  long counter = 0;
  for (i = 0; i < number_of_points; i++)
  {
    long row = point_rows[i];
    long col = point_cols[i];
    // Figure out if the point belongs to any line.
    long found = 0;
    for (j = 0; j < number_of_lines; j++)
    {
      long drow = row - center_rows[j];
      long dcol = col - center_cols[j];
      long abs_drow = vAbs(drow);
      long abs_dcol = vAbs(dcol);
      long distance = lengths[abs_drow][abs_dcol];
      if (distance > half_lengths[j]) continue;
      float angle = angles[drow+256][dcol+256];
      float angle_diff = (float) vAbs(angle - orientations[j]);
      if (angle_diff > half_pi) angle_diff = (float) M_PI - angle_diff;
      if (angle_diff > orientation_threshold) continue;
      found = 1;
    }

    if (found == 0)
    {
      leftover_rows[counter] = row;
      leftover_cols[counter] = col;
      counter++;
    }
  }

  // Store the results into the argument pointers
  *leftover_rowsp = leftover_rows;
  *leftover_colsp = leftover_cols;
  *leftover_numberp = counter;
  return 1;
}


// In vEstimateFingerLines2 we fit lines into some finger edges. Here
// we want to figure out which points are used to fit those lines, because
// we want to build the set of model image points that are NOT used to 
// fit those lines (also called model leftover points elsewhere in the code).
// Those leftover points are important, because they are not described
// by the extracted straight lines and they need to also be matched in order
// to have a more accurate similarity measure.
// The line finding code from vEstimateFingerLines2 is repeated here.
// The coordinates of the points that are used to fit the lines are stored
// in out_rows and out_cols.
long vPointsCoveredByLines(grayscale_image * link_ids, grayscale_image * dilated,
                            vStraightLinks * finger,
                            vector<long> * out_rows, vector<long> * out_cols)
{
  // Select the straight link out of the image
  grayscale_image * selected = 
    (grayscale_image *) vSelectRange(link_ids, 0, (uchar) finger->link1, 
                                  (uchar) finger->link3);
  // Fit an ellipse
  vEllipse ellipse = vFitEllipse1(selected);
  vector<Label> labels;
  // Find the connected components of the link
  v3dMatrix<long> * components = selected->ConnectedComponents(&labels);
  vdelete(selected);
  if (labels.size() < 2)
  {
    vdelete(components);
    return 0;
  }

  // Choose the largest component
  std::sort(labels.begin(), labels.end(), LabelMore());
  grayscale_image * largest_component = vSelectValue(components, (long) labels[1].id);
  vdelete(components);
  // Get the contour of the largest component
  vContour contour;
  contour.OuterContour(largest_component);

  // Go through the contour and overwrite points that are zero in the edge image
  vector<double> * xs = contour.GetXs();
  vector<double> * ys = contour.GetYs();
  long number = xs->size();
  long i;
  vArray2(uchar) dilated2 = dilated->Matrix2(0);
  for (i = 0; i < number; i++)
  {
    long row = vRound((*ys)[i]);
    long col = vRound((*xs)[i]);
    if (dilated2[row][col] == 0)
    {
      (*ys)[i] = -row;
      (*xs)[i] = -col;
    }
  }

  vdelete(largest_component);
  // Choose from the contour pixels whose orientations are close to parallel
  // to the major axis of the fitted ellipse, and divide those pixels into
  // two groups, one on the left half plane of the major axis, and one on
  // the right half plane of the major axis.
  vector<double> rows1, cols1, rows2, cols2;
  long success = vFindOuterLines(&contour, ellipse.center_row, 
                                  ellipse.center_col, ellipse.angle, 
                                  &rows1, &cols1, &rows2, &cols2);
  if (success == 0) return 0;
  
  long size1 = rows1.size();
  long size2 = rows2.size();
  for (i = 0; i < size1; i++)
  {
    out_rows->push_back((long) rows1[i]);
    out_cols->push_back((long) cols1[i]);
  }

  for (i = 0; i < size2; i++)
  {
    out_rows->push_back((long) rows2[i]);
    out_cols->push_back((long) cols2[i]);
  }

  return 1;
}


// We compute here the coordinates of the model edge points that were NOT 
// used to create the lines contained in the vStraightLink objects.
long vLeftoverModelPoints(grayscale_image * link_ids, grayscale_image * edges,
                           grayscale_image * dilated,
                           vector<vStraightLinks> * finger_lines,
                           vector<long> * out_rows, vector<long> * out_cols)
{
  vector<long> all_rows, all_cols;
  vGetNonZeroPoints(edges, &all_rows, &all_cols);
  long rows = link_ids->Rows();
  long cols = link_ids->Cols();

  vector<long> used_rows, used_cols;
  long number_of_fingers = finger_lines->size();
  long i;
  // Put all points that were used to generate straight link lines
  // into used_rows/used_cols. In principle some of these points may
  // even be repeated, we don't care. It is the complement of those
  // points that we do care about.
  for (i = 0; i < number_of_fingers; i++)
  {
    vStraightLinks & finger = (*finger_lines)[i];
    vPointsCoveredByLines(link_ids, edges, &finger, &used_rows, &used_cols);
  }

  v3dMatrix<uchar> scratch_matrix(rows, cols, 1);
  vArray2(uchar) scratch = scratch_matrix.Matrix2(0);
  // First mark all_rows/all_cols on scratch as 0, then
  // mark used_rows/used_cols as non-zero, and then put into
  // out_rows/out_cols the pixels that remain zero.

  long all_points = all_rows.size();
  long used_points = used_rows.size();

  // Mark all points with 0
  for (i = 0; i < all_points; i++)
  {
    long row = all_rows[i];
    long col = all_cols[i];
    scratch[row][col] = 0;
  }

  // Mark used points with non-zero value (255).
  for (i = 0; i < used_points; i++)
  {
    long row = used_rows[i];
    long col = used_cols[i];
    scratch[row][col] = 1;
  }

  // Now put all points that are still 0 into the result
  for (i = 0; i < all_points; i++)
  {
    long row = all_rows[i];
    long col = all_cols[i];
    if (scratch[row][col] != 0) continue;
    out_rows->push_back(row);
    out_cols->push_back(col);
  }

  return 1;
}


static inline double vLinePairingScore(vLineSegmentLong & segment1,
                                        vLineSegmentLong & segment2, long * lengthp)
{
  static const long junk1 = vPrecomputedLengths::PrecomputeValues(256, 256);
  vArray2(long) lengths = vPrecomputedLengths::values;

  // Compute the length of the segment1
  long dr = vAbs(segment1.row1 - segment1.row2);
  long dc = vAbs(segment1.col1 - segment1.col2);
  long length1 = lengths[dr][dc];
  *lengthp = length1;

  long row_diff1a = vAbs(segment1.row1 - segment2.row1);
  long col_diff1a = vAbs(segment1.col1 - segment2.col1);
  long row_diff2a = vAbs(segment1.row2 - segment2.row2);
  long col_diff2a = vAbs(segment1.col2 - segment2.col2);
  double distance1 = lengths[row_diff1a][col_diff1a] + lengths[row_diff2a][col_diff2a];

  long row_diff1b = vAbs(segment1.row1 - segment2.row2);
  long col_diff1b = vAbs(segment1.col1 - segment2.col2);
  long row_diff2b = vAbs(segment1.row2 - segment2.row1);
  long col_diff2b = vAbs(segment1.col2 - segment2.col1);
  double distance2 = lengths[row_diff1b][col_diff1b] + lengths[row_diff2b][col_diff2b];

  if (distance1 < distance2)
  {
    return distance1 / 2;
  }
  else
  {
    return distance2 / 2;
  }
}


double vModelToImagePairingScore(vector<vLinePairing> * pairings)
{
  long length = 0;
  double total_lengths = 0;
  double total_distances = 0;
  long number_of_lines = pairings->size();
  long i;
  for (i = 0; i < number_of_lines; i++)
  {
    vLinePairing & pairing = (*pairings)[i];
    vLineSegmentLong segment1 = vLineSegmentLong(pairing.model_line);
    vLineSegmentLong & segment2 = pairing.input_line.segment;
    double distance = vLinePairingScore(segment1, segment2, &length);
    total_lengths += length;
    total_distances += distance * (double) length;
  }

  if (total_lengths == 0)
  {
    return 0;
  }
  else
  {
    return total_distances / total_lengths;
  }
}


double vImageToModelPairingScore(vector<vLinePairing> * pairings)
{
  long length = 0;
  double total_lengths = 0;
  double total_distances = 0;
  long number_of_lines = pairings->size();
  long i;
  for (i = 0; i < number_of_lines; i++)
  {
    vLinePairing & pairing = (*pairings)[i];
    vLineSegmentLong & segment1 = vLineSegmentLong(pairing.model_line);
    vLineSegmentLong & segment2 = pairing.input_line.segment;
    double distance = vLinePairingScore(segment2, segment1, &length);
    total_lengths += length;
    total_distances += distance * (double) length;
  }

  if (total_lengths == 0)
  {
    return 0;
  }
  else
  {
    return total_distances / total_lengths;
  }
}



// This function is used as follows: suppose I want to know how model
// images that have 8 lines tend to score using line scores. Then,
// I call this function, giving it a dolvs for the current input
// image, specifying line_number = 8, and asking it to find a number
// of model images with 8 lines and collect statistics on them. 
long vLineNumberStatistics(v3dMatrix<uchar> * all_model_lines,
                            vDOLVS_Histogram * dolvs,
                            long number)
{
  const long line_limit = 9;
  
  // First, collect all the model images that have the specified number
  // of lines.
  long model_views = all_model_lines->Bands();
  vector<long> view_indices;

  long i;

  vector<double> scores[line_limit];
  long numbers_found[line_limit];
  long j;
  for (j = 0; j < line_limit; j++)
  {
    scores[j] = vector<double>(number);
    long link_number = j;
    long number_of_lines = 2*j;
    view_indices.erase(view_indices.begin(), view_indices.end());
    for (i = 0; i < model_views; i++)
    {
      vector<vStraightLinks> * links = vReadModelLines(all_model_lines, i);
      if (links->size() == link_number)
      {
        view_indices.push_back(i);
      }
      vdelete(links);
    }

    numbers_found[j] = view_indices.size();
    vPrint("There are %li images with %li lines\n", numbers_found[j], number_of_lines);

    // Now, collect scores from a sample of those frames.
    for (i = 0; i < number; i++)
    {
      long random_index = vRandomLong2(0, numbers_found[j] - 1);
      long index = view_indices[random_index];
      vector<vStraightLinks> * links = vReadModelLines(all_model_lines, index);
      vLineMatchFrameStats stats(dolvs, links);
      scores[j][i] = stats.Average();
      vdelete(links);
    }

    std::sort(scores[j].begin(), scores[j].end(), less<double>());
  }

  for (i = 0; i < number; i++)
  {
    vPrint("%4li ", i);
    for (j = 0; j < line_limit; j++)
    {
      vPrint("%3li ", (long) vRound(scores[j][i]));
    }
    vPrint("\n");
  }

  return 1;
}


vMatchScores * vPalmCutoffScores(v3dMatrix<uchar> * cutoffs_matrix,
                                   vLineSegmentLong * cutoff)
{
  const static long junk = vPrecomputedLengths::PrecomputeValues(256, 256);
  vArray2(long) lengths = vPrecomputedLengths::values;
  long number_of_views = cutoffs_matrix->Rows();
  vArray2(uchar) cutoffs = cutoffs_matrix->Matrix2(0);
  long row1 = cutoff->row1;
  long col1 = cutoff->col1;
  long row2 = cutoff->row2;
  long col2 = cutoff->col2;

  v3dMatrix<double> distances_matrix(1, number_of_views, 1);
  vArray(double) distances = distances_matrix.Matrix();
  long i;
  for (i = 0; i < number_of_views; i++)
  {
    long crow1 = cutoffs[i][0];
    long ccol1 = cutoffs[i][1];
    long crow2 = cutoffs[i][2];
    long ccol2 = cutoffs[i][3];

    long row_diff11 = vAbs(crow1 - row1);
    long col_diff11= vAbs(ccol1 - col1);

    long row_diff12 = vAbs(crow1 - row2);
    long col_diff12= vAbs(ccol1 - col2);

    long row_diff21 = vAbs(crow2 - row1);
    long col_diff21= vAbs(ccol2 - col1);

    long row_diff22 = vAbs(crow2 - row2);
    long col_diff22= vAbs(ccol2 - col2);

    if ((row_diff11 >= 256) || (col_diff11 >= 256) ||
        (row_diff11 >= 256) || (col_diff11 >= 256) ||
        (row_diff11 >= 256) || (col_diff11 >= 256) ||
        (row_diff11 >= 256) || (col_diff11 >= 256))
    {
      exit_error("Error in vPalmCutoffScores\n");
    }

    long length11 = lengths[row_diff11][col_diff11];
    long length12 = lengths[row_diff12][col_diff12];
    long length21 = lengths[row_diff21][col_diff21];
    long length22 = lengths[row_diff22][col_diff22];

    long min_length = Min(length11 + length22, length12 + length21);
    distances[i] = min_length;
  }

  vMatchScores * scores = new vMatchScores(&distances_matrix);
  return scores;
}


vLineGrid::vLineGrid(long in_d_samples, long in_o_samples, long in_l_samples)
{
  d_samples = in_d_samples;
  o_samples = in_o_samples;
  l_samples = in_l_samples;

  // Sloppy code, 192, 180 and 100 should be member variables.
  d_values = vUniformSpacing(192, d_samples);
  o_values = vCircularUniformSpacing(180, o_samples);
  l_values = vUniformSpacing(100, l_samples);

  // Sloppy code, 300 should be a constant
  d_indices = vnew(long, 300);
  o_indices = vnew(long, 300);
  l_indices = vnew(long, 300);

  number_of_segments = 0;

  double d_step = ((double) 192) / d_samples;
  double o_step = ((double) 180) / o_samples;
  double l_step = ((double) 100) / l_samples;
  long i;

  for (i = 0; i < 300; i++)
  {
    // Sloppy code, 192 should be a member variable, same 
    // in the next few lines).
    double d_index = ((double) i) / d_step;
    if (d_index >= d_samples) d_index = d_samples - 1;
    d_indices[i] = (long) d_index;

    double o_index = ((double) i) / o_step;
    if (o_index >= o_samples) o_index = o_samples - 1;
    o_indices[i] = (long) o_index;
    
    double l_index = ((double) i) / l_step;
    if (l_index >= l_samples) l_index = l_samples - 1;
    l_indices[i] = (long) l_index;
  }

  vHyperArray4(&rankings, d_samples, d_samples, o_samples, l_samples);
  long j, k, l;
  for (i = 0; i < d_samples; i++)
  {
    for (j = 0; j < d_samples; j++)
    {
      for (k = 0; k < o_samples; k++)
      {
        for (l = 0; l < l_samples; l++)
        {
          rankings[i][j][k][l] = vZero(long);
        }
      }
    }
  }
}


vLineGrid::~vLineGrid()
{
  vdelete2(d_values);
  vdelete2(o_values);
  vdelete2(l_values);
  vdelete2(d_indices);
  vdelete2(o_indices);
  vdelete2(l_indices);
  
  long i, j, k, l;
  for (i = 0; i < d_samples; i++)
  {
    for (j = 0; j < d_samples; j++)
    {
      for (k = 0; k < o_samples; k++)
      {
        for (l = 0; l < l_samples; l++)
        {
          vdelete2(rankings[i][j][k][l]);
        }
        vdelete2(rankings[i][j][k]);
      }
    }
  }
  vDelete3(rankings, i, j);
}


// Prints information that is useful for debugging, or to just verify
// that the grid is constructed correctly.
long vLineGrid::PrintValues()
{
  vPrint("d_samples = %li, o_samples = %li, l_samples = %li\n", 
          d_samples, o_samples, l_samples);
  vPrint("\n      d_values, o_values, l_values\n");
  long i;
  long max_value = Max(d_samples, Max(o_samples, l_samples));
  for (i = 0; i < max_value; i++)
  {
    vPrint("%4li  ", i);
    if (i < d_samples)
    {
      vPrint("%9li ", d_values[i]);
    }
    else vPrint("%10s", " ");
    if (i < o_samples)
    {
      vPrint("%9li ", o_values[i]);
    }
    else vPrint("%10s", " ");
    if (i < l_samples)
    {
      vPrint("%9li ", l_values[i]);
    }
    else vPrint("%10s", " ");
    vPrint("\n");
  }

  vPrint("\n      d_indices, o_indices, l_indices\n");

  // Sloppy code: Assumes 300 indices for all D, O, L.
  for (i = 0; i < 300; i++)
  {
    vPrint("%4li  ", i);
    vPrint("%10li ", d_indices[i]);
    vPrint("%10li ", o_indices[i]);
    vPrint("%10li ", l_indices[i]);
    vPrint("\n");
  }
  return 1;
}


long vLineGrid::ComputeRankings(vDOLVS_Histogram * dolvs)
{
  long i, j, k, l;
  number_of_segments =  dolvs->InputLength();
  if (number_of_segments <= 0) return -1;
  for (i = 0; i < d_samples; i++)
  {
    for (j = 0; j < d_samples; j++)
    {
      for (k = 0; k < o_samples; k++)
      {
        for (l = 0; l < l_samples; l++)
        {
          vdelete2(rankings[i][j][k][l]);
          long row = d_values[i];
          long col = d_values[j];
          long o = o_values[k];
          long len = l_values[l];
          rankings[i][j][k][l] = dolvs->FrequencyRanking4(row, col, o, len);
        }
      }
    }
  }
  return 1;
}


long vLineGrid::SmallestFrequency(vDOLVS_Histogram * dolvs, 
                                   vLineSegment segment,
                                   double fraction)
{
  if (number_of_segments <= 0) return -1;

  // Figure out how many candidates we should use.
  long candidate_number;
  if (fraction <= 0) candidate_number = 10;
  else if (fraction >= 1) candidate_number = number_of_segments;
  else
  {
    candidate_number = vRound(fraction * (double) number_of_segments);
  }
  if (candidate_number > number_of_segments) candidate_number = number_of_segments;

  // Figure out the nearest neighbor in the grid.
  static long junk = vPrecomputedLengths::PrecomputeValues(256, 256);
  static const double rad_to_deg = 180.0 / M_PI;
  long center_row = (long) (segment.row1 + segment.row2) / 2;
  long center_col = (long) (segment.col1 + segment.col2) / 2;
  long drow = (long) (segment.row1 - segment.row2);
  long dcol = (long) (segment.col1 - segment.col2);
  long abs_drow = vAbs(drow);
  long abs_dcol = vAbs(dcol);
  long length = vPrecomputedLengths::Length(abs_drow, abs_dcol);
  float orientation = (float) vAngle(dcol, drow);
  long o = vRound(orientation * rad_to_deg);
  
  long row_index = d_indices[center_row];
  long col_index = d_indices[center_col];
  long o_index = o_indices[o];
  long l_index = l_indices[length];
  vArray(long) candidates = rankings[row_index][col_index][o_index][l_index];

  long result = dolvs->SmallestMatchFrequency3(segment, candidates, 
                                               candidate_number);
  return result;
}


long vLineGrid::ComputeRankings2(VS_Histogram * vs_histogram, 
                                  vector<vLineSegmentMatch> * input_lines)
{
  long i, j, k, l;
  number_of_segments =  input_lines->size();
  if (number_of_segments <= 0) return -1;
  for (i = 0; i < d_samples; i++)
  {
    for (j = 0; j < d_samples; j++)
    {
      for (k = 0; k < o_samples; k++)
      {
        for (l = 0; l < l_samples; l++)
        {
          vdelete2(rankings[i][j][k][l]);
          long row = d_values[i];
          long col = d_values[j];
          long o = o_values[k];
          long len = l_values[l];
          rankings[i][j][k][l] = vs_histogram->FrequencyRanking(input_lines, row, col, o, len);
        }
      }
    }
  }
  return 1;
}


static const long vs_histogram_max_v = 100;
static const long vs_histogram_max_s = 255;


VS_Histogram::VS_Histogram(vector<vLineSegmentMatch> * input_lines) :
vMatrix<double>(vs_histogram_max_v, vs_histogram_max_s)
{
  max_v = vs_histogram_max_v;
  max_s = vs_histogram_max_s;
  vWriteValue(this, (double) 0);
  ProcessInputLines(input_lines);
  MakeCumulative();
}


VS_Histogram::~VS_Histogram()
{
  DecrementReferences();
}


long VS_Histogram::ProcessInputLines(vector<vLineSegmentMatch> * input_lines)
{
  // For each input line, increment the corresponding VS bin,
  number_of_lines = input_lines->size();
  long i;
  for (i = 0; i < number_of_lines; i++)
  {
    vLineSegmentMatch & input_line = (*input_lines)[i];
    long v = vRound(input_line.vote_density * max_v);
    if (v >= max_v) v = max_v - 1;
    long s = input_line.s;

    // Catch possible bugs.
    if (check_bounds(v, s) <= 0)
    {
      exit_error("Illegal (v,s): (%li, %li)\n", v, s);
    }
    matrix3[0][v][s]++;
  }

  return 1;
}


// We store, in every VS bin, the sum of all bins that are
// "better than or equal" to that bin.
long VS_Histogram::MakeCumulative()
{
  long v, s;

  // Add along V.
  for (v = max_v - 2; v >= 0; v--)
  {
    for (s = max_s - 1; s >= 0; s--)
    {
      matrix3[0][v][s] += matrix3[0][v+1][s];
    }
  }

  // Add along S.
  for (v = max_v - 1; v >= 0; v--)
  {
    for (s = max_s - 2; s >= 0; s--)
    {
      matrix3[0][v][s] += matrix3[0][v][s+1];
    }
  }

  return 1;
}


// row, col, orientation and len define a model line. We want
// to rank input line segments in decreasing order of how well they
// match the model line.
vArray(long) 
VS_Histogram::FrequencyRanking(vector<vLineSegmentMatch> * input_lines, 
                               long row, long col, long orientation, 
                               long len)
{
  static long junk = vPrecomputedLengths::PrecomputeValues(256, 256);
  static const double rad_to_deg = 180.0 / M_PI;
  vArray2(long) lengths = vPrecomputedLengths::values;

  // Prepare for count sort. We need to initialize counters to 0.
  long input_length = input_lines->size();
  vector<vIndexFrequencyPair> frequencies(input_length);
  vArray(long) result = vnew(long, input_length);
  long i;

  // Measure the rareness (or frequency) of each match.
  for (i = 0; i < input_length; i++)
  {
    vLineSegmentMatch & input_line = (*input_lines)[i];
    
    // Get dolvs values
    long drow = vAbs(input_line.center_row - row);
    long dcol = vAbs(input_line.center_col - col);

    long d = lengths[drow][dcol];
    long o = vAbs(vRound(input_line.orientation * rad_to_deg - orientation));
    if (o > 90) o = 180 - o;
    long l = (long) vAbs(input_line.length - len);
    long v = vRound(input_line.vote_density * max_v);
    if (v >= max_v) v = max_v - 1;
    long s = input_line.s;

    long vs = (long) matrix3[0][v][s];

    double frequency = ((double) vs) * (d+1) * (d+1) * (o+1) * ((double) l+1);
    frequencies[i] = vIndexFrequencyPair(i, frequency);
    input_line.score = (float) frequency;
    input_line.d = d;
    input_line.o = o;
    input_line.l = l;
  }

  // sort based on frequency.
  std::sort(frequencies.begin(), frequencies.end(), vFrequencyLess());

  // Put the sorted line segment indices into result
  for (i = 0; i < input_length; i++)
  {
    result[i] = frequencies[i].index;
  }

  return result;
}


/////////////////////////////////////////////////////////////////////
//                                                                 //
//   Beginning of implementation of class vPointSetOrientation    //
//                                                                 //
/////////////////////////////////////////////////////////////////////

vPointSetOrientation::vPointSetOrientation(vVideoFile * sequence,
                                             vPointSetSequence * locations)
{
  rows = sequence->Rows();
  cols = sequence->Cols();
  if ((rows > 256) || (cols > 256))
  {
    exit_error("Coordinates exceeding 256 in vPointSetOrientation\n");
  }

  number_of_frames = sequence->NumberOfFrames();
  lengths = vnew(long, number_of_frames);
  indices = vnew(long, number_of_frames);
  total_length = locations->TotalLength();
  point_orientations = vnew(uchar, total_length);

  long i;
  vPrint("\n");
  for (i = 0; i < number_of_frames; i++)
  //for (i = 0; i < 10; i++)
  {
    color_image * database_frame = sequence->ReadFrame(i);
    if (database_frame == 0)
    {
      exit_error("error in reading frame %li\n");
    }

    v3dMatrix<double> * blurred1 = vBlur(database_frame);
    v3dMatrix<double> * blurred2 = vBlur(blurred1);
    v3dMatrix<double> * blurred3 = vBlur(blurred2);
    grayscale_image * gray = blurred3->ToGrayCopy();
    v3dMatrix<double> * orientations = vEdgeOrientations(gray);
    vArray2(double) orientations2 = orientations->Matrix2(0);
    vdelete(database_frame);
    vdelete(blurred1);
    vdelete(blurred2);
    vdelete(blurred3);
    vdelete(gray);
    long length = locations->Length(i);
    lengths[i] = length;
    if (i == 0)
    {
      indices[i] = 0;
    }
    else 
    {
      indices[i] = indices[i-1] + lengths[i-1];
    }

    long j;
    long index = indices[i];

    vArray(uchar) point_rows = locations->PointRows(i);
    vArray(uchar) point_cols = locations->PointCols(i);
    for (j = 0; j < length; j++)
    {
      long row = point_rows[j];
      long col = point_cols[j];
      point_orientations[index] = OrientationToUchar((float) orientations2[row][col]);
      index++;
    }
    vdelete(orientations);
    if (i % 100 == 0)
    {
      vPrint("processed frame %li of %li\r", i, number_of_frames);
    }
  }
  vPrint("\n");
}


vPointSetOrientation::~vPointSetOrientation()
{
  vdelete2(lengths);
  vdelete2(indices);
  vdelete2(point_orientations);
}


// Assuming that filename specifies a database, we return a string 
// that is the same as the filename, except that the ending has been
// adjusted to the default ending for files where vPointSetOrientation
// objects are saved.
char * vPointSetOrientation::AdjustEnding(const char * filename)
{
  return vReplaceEnding2(filename, ".point_orientations");
}


vPointSetOrientation * vPointSetOrientation::Load(const char * filename)
{
  char * input_name = AdjustEnding(filename);
  FILE * fp = fopen(input_name, vFOPEN_READ);
  vdelete2(input_name);
  if (fp == 0) return 0;
  vPointSetOrientation * result = new vPointSetOrientation();
  long items = 0;
  items += vReadPCLongs(fp, &(result->rows), 1);
  items += vReadPCLongs(fp, &(result->cols), 1);
  items += vReadPCLongs(fp, &(result->number_of_frames), 1);
  items += vReadPCLongs(fp, &(result->total_length), 1);

  if ((items != 4) || (result->rows < 0) || (result->cols < 0) ||
      (result->rows > 256) || (result->cols > 256))
  {
    vPrint("Error in loading point set orientation\n");
    vdelete(result);
    fclose(fp);
    return 0;
  }

  result->lengths = vnew(long, result->number_of_frames);
  result->indices = vnew(long, result->number_of_frames);
  result->point_orientations = vnew(uchar, result->total_length);

  items = vReadPCLongs(fp, vData(result->lengths), result->number_of_frames);
  items += vReadPCLongs(fp, vData(result->indices), result->number_of_frames);
  items += fread(vData(result->point_orientations), sizeof(uchar), result->total_length, fp);
  fclose(fp);
  if (items != 2 * result->number_of_frames + result->total_length)
  {
    vPrint("Error in reading point set orientation\n");
    vdelete(result);
    return 0;
  }
  return result;
}


long vPointSetOrientation::Save(const char * filename)
{
  char * input_name = AdjustEnding(filename);
  FILE * fp = fopen(input_name, vFOPEN_WRITE);
  vdelete2(input_name);
  if (fp == 0) return 0;
  long items = 0;
  items += vWritePCLongs(fp, &rows, 1);
  items += vWritePCLongs(fp, &cols, 1);
  items += vWritePCLongs(fp, &number_of_frames, 1);
  items += vWritePCLongs(fp, &total_length, 1);

  if (items != 4)
  {
    vPrint("Error in saving point set sequence\n");
    fclose(fp);
    return 0;
  }

  items = vWritePCLongs(fp, vData(lengths), number_of_frames);
  items += vWritePCLongs(fp, vData(indices), number_of_frames);
  items += fwrite(vData(point_orientations), sizeof(uchar), total_length, fp);
  fclose(fp);
  if (items != 2 * number_of_frames + total_length)
  {
    vPrint("Error in writing point set sequence\n");
    return 0;
  }
  return 1;
}


// returns the number of points.
// In orientationsp we put store the orientations of
// each point in the set of points corresponding
// to the given frame and given view
long vPointSetOrientation::GetOrientations(long frame, long view, 
                                            vArray(float) * orientationsp)
{
  long number = Length(frame);
  // Get the pixel orientations for given frame and view 0.
  vArray(uchar) frame_orientations = PointOrientations(frame);
  
  // Copy over the orientations into the result array.
  vArray(float) result_orientations = vnew(float, number);
  long i;
  for (i = 0; i < number; i++)
  {
    result_orientations[i] = UcharToOrientation(frame_orientations[i]);
  }

  // Rotate points by rotation corresponding to given view.
  float adjustment = (float) view * (float) M_PI / (float) 2.0;
  for (i = 0; i < number; i++)
  {
    result_orientations[i] = AddOrientations(result_orientations[i], 
                                             adjustment);
  }

  *orientationsp = result_orientations;
  return number;
}


/////////////////////////////////////////////////////////////////////
//                                                                 //
//     End of implementation of class vPointSetOrientation        //
//                                                                 //
/////////////////////////////////////////////////////////////////////




vFingerCurve::vFingerCurve(long in_number_of_scales, 
                             long in_number_of_orientations)
{
  height = 50;
  width = 50;
  number_of_scales = in_number_of_scales;
  number_of_orientations = in_number_of_orientations;
  number_of_templates = number_of_scales * number_of_orientations;

  prototype = MakePrototype();
  all_templates = vnew(grayscale_image *, number_of_templates);
  scales = vnew(double, number_of_templates);
  orientations = vnew(double, number_of_templates);
  all_point_rows = vnew(vArray(long), number_of_templates);
  all_point_cols = vnew(vArray(long), number_of_templates);
  all_lengths = vnew(long, number_of_templates);
}


vFingerCurve::~vFingerCurve()
{
  vdelete(prototype);

  long i;
  for (i = 0; i < number_of_templates; i++)
  {
    vdelete(all_templates[i]);
  }
  vdelete2(all_templates);

  vdelete2(scales);
  vdelete2(orientations);

  for (i = 0; i < number_of_templates; i++)
  {
    vdelete2(all_point_rows[i]);
  }
  vdelete2(all_point_rows);

    for (i = 0; i < number_of_templates; i++)
  {
    vdelete2(all_point_cols[i]);
  }
  vdelete2(all_point_cols);

  vdelete2(all_lengths);
}

grayscale_image * vFingerCurve::MakePrototype()
{
  grayscale_image * result = new grayscale_image(height, width);
  vWriteValue(result, (uchar) 0);

  return result;
}


//v3dMatrix<double> * vFingerCurve::Match(grayscale_image * the_image);
//v3dMatrix<double> * vFingerCurve::Match(v3dMatrix<float> * dt);


// result[row][col] is the (rounded) distance between location (row, col) and
// the nearest location in the input image that has a value higher than
// or equal to that of location (row, col). For the pixel in the image
// with max magnitude, the value is equal to the length of the diagonal
// of the image plus 1.
v3dMatrix<long> * vDominanceTransform(general_image * input_image)
{
  if (input_image->Bands() != 1) return 0;
  
  v3dMatrix<float> input_matrix(input_image);
  vArray2(float) input = input_matrix.Matrix2(0);
  long rows = input_matrix.Rows();
  long cols = input_matrix.Cols();
  vPrecomputedLengths::PrecomputeValues(rows, cols);
  vArray2(long) lengths = vPrecomputedLengths::values;

  v3dMatrix<long> * result_matrix = new v3dMatrix<long>(rows, cols, 1);
  vArray2(long) result = result_matrix->Matrix2(0);

  list<vPoint> points;
  // Put into the list all points;
  long row, col;
  for (row = 0; row < rows; row++)
  {
    for (col = 0; col < cols; col++)
    {
      points.push_back(vPoint(row, col));
    }
  }

  long distance = 1;
  long area = (distance * 2 + 1) * (distance * 2 + 1);
  while(area < (long) points.size())
  {
    long number = points.size();
    vPrint("distance = %li, points = %li\n", distance, number);
    S_WindowBasedDominance(input, result, &points, distance, lengths, rows, cols);
    distance++;
    area = (distance * 2 + 1) * (distance * 2 + 1);
  }

  long number = points.size();
  vPrint("entering list phase, %li points\n", number);
  S_ListBasedDominance(input, result, &points, lengths, rows, cols);
  return result_matrix;
}


// An auxiliary function for vDominanceTransform. For every point in
// points, it checks whether there exists a point in input, within
// the specified distance, that has a greater value. If the answer
// is yes, then we store the value distance in the corresponding location
// in output. This can be done in more efficient ways, but initially I 
// will go for ease of implementation. This function is more efficient
// than S_ListBasedDominance when we have lots of points and a small
// distance.
long S_WindowBasedDominance(vArray2(float) input, vArray2(long) result, 
                            std::list<vPoint> * points, long distance, 
                            vArray2(long) lengths, long rows, long cols)
{
  // Go through all points.
  list<vPoint>::iterator iter;
  for (iter = points->begin(); iter != points->end();)
  {
    list<vPoint>::iterator next_iter = ++iter;
    iter--;
    vPoint & point = *iter;
    long center_row = point.row;
    long center_col = point.col;
    long top = Max(center_row - distance, (long) 0);
    long left = Max(center_col - distance, (long) 0);
    long bottom = Min(center_row + distance, (long) (rows-1));
    long right = Min(center_col + distance, (long) (cols-1));

    float center_value = input[center_row][center_col];
    // We decrease the value of the current point, so that we don't
    // detect it as we go through the square.
    input[center_row][center_col] = center_value - (float) 1.0;

    // Go through pixels in the square.
    long row, col;
    long break_flag = 0;
    for (row = top; row <= bottom; row++)
    {
      long row_diff = vAbs(row - center_row);
      for (col = left; col <= right; col++)
      {
        long col_diff = vAbs(col - center_col);
        long current_distance = lengths[row_diff][col_diff];
        // This check is necessary since there are pixels in the square whose
        // distance from the center is greater than distance.
        if (current_distance > distance) continue;
        if (input[row][col] >= center_value)
        {
          result[center_row][center_col] = distance;
          points->erase(iter);
          break_flag = 1;
          break;
        }
      }
      if (break_flag != 0) break;
    }
    // Restore the value at the center.
    input[center_row][center_col] = center_value;
    iter = next_iter;
  }

  return 1;
}


// An auxiliary function for vDominanceTransform. For every point p1 in
// points, it finds the point p2 in points that is the farthest from p1,
// among all points in points that have a greater value (in input) than
// p1. This function is more efficient than S_WindowBasedDominance when
// the number of points is sufficiently small and the min distance
// between those points is sufficiently large.
long S_ListBasedDominance(vArray2(float) input, vArray2(long) result, 
                            std::list<vPoint> * points, 
                            vArray2(long) lengths, long rows, long cols)
{
  // 1 + max possible distance between points.
  long max_distance = lengths[rows-1][cols-1] + 1;

  // Go through all points.
  list<vPoint>::iterator iter;
  for (iter = points->begin(); iter != points->end(); iter++)
  {
    vPoint & center_point = *iter;
    long center_row = center_point.row;
    long center_col = center_point.col;

    float center_value = input[center_row][center_col];
    // We decrease the value of the current point, so that we don't
    // detect it as we go through the list again.
    input[center_row][center_col] = center_value - (float) 1.0;

    // Go through all points.
    list<vPoint>::iterator iter2;
    long min_distance = max_distance;
    for (iter2 = points->begin(); iter2 != points->end(); iter2++)
    {
      vPoint & point = *iter2;
      long row = point.row;
      long col = point.col;
      if (input[row][col] >= center_value)
      {
        long row_diff = vAbs(row - center_row);
        long col_diff = vAbs(col - center_col);
        long current_distance = lengths[row_diff][col_diff];
        min_distance = Min(current_distance, min_distance);
      }
    }

    result[center_row][center_col] = min_distance;
    // Restore the value at the center.
    input[center_row][center_col] = center_value;
  }

  return 1;
}


// This function is built for domt arguments that are a dominance transforms.
// For the top number pixels, we draw the circle of their dominance on output.
long vDrawTopCircles(v3dMatrix<long> * domt, color_image * output, long number)
{
  vector<vPoint> top_locations;
//  vTopLocations(domt, number, &top_locations);
  vArray2(long) radii = domt->Matrix2(0);

  long i;
  for (i = 0; i < number; i++)
  {
    long row = top_locations[i].row;
    long col = top_locations[i].col;
    vDrawCircle1(output, row, col, radii[row][col], (uchar) 255);
  }
  return 1;
}


// corners gives the three corners of a triangle.
// We store into points a sampling 
// of the 3D surface of the triangle that guarantees that any point
// on the surface of the triangle is within max_distance distance
// of one of the points stored in points.
long vTrianglePixels3D(vPoint3f * corners, 
                        vector<vPoint3f> * points, float max_distance)
{
  // compute the step at which we must sample points.
  float step = sqrt((float) 2) * max_distance;

  // Start from the first corner.
  float start_x = corners[0].x;
  float start_y = corners[0].y;
  float start_z = corners[0].z;

  // Compute direction of line leading from first to second corner.
  float ab_dx = corners[1].x - start_x;
  float ab_dy = corners[1].y - start_y;
  float ab_dz = corners[1].z - start_z;

  float ab_length = sqrt(ab_dx * ab_dx + ab_dy * ab_dy + ab_dz * ab_dz);
  if (ab_length == 0) return 0;
  float ab_factor = step / ab_length;
  float x_step1 = ab_dx * ab_factor;
  float y_step1 = ab_dy * ab_factor;
  float z_step1 = ab_dz * ab_factor;

  // Compute direction of line leading from first to third corner.
  float ac_dx = corners[2].x - start_x;
  float ac_dy = corners[2].y - start_y;
  float ac_dz = corners[2].z - start_z;

  float ac_length = sqrt(ac_dx * ac_dx + ac_dy * ac_dy + ac_dz * ac_dz);
  if (ac_length == 0) return 0;
  float dot_product = vAbs(ac_dx * x_step1 + ac_dy * y_step1 + ac_dz * z_step1);
  float cosine = dot_product / (ab_length * ac_length);
  float sine = sqrt(1 - cosine * cosine);
  if (sine == 0) return 0;

  float ac_factor = step / (sine * ac_length);

  // Compute how many steps along first-to-third line we need.
  long ac_steps = vRound(((float) 1.0) / ac_factor + 0.5);
  float ac_step = ac_length / (float) ac_steps;
  float length_step = ab_length / ((float) ac_steps);

  float f_ac_steps = (float) ac_steps;
  float x_step2 = ac_dx / f_ac_steps;
  float y_step2 = ac_dy / f_ac_steps;
  float z_step2 = ac_dz / f_ac_steps;

  long points_before = points->size();

  // Go along AC line
  long i, j;
  for (i = 0; i < ac_steps; i++)
  {
    float f_i = (float) i;
    float ac_x = start_x + f_i * x_step2;
    float ac_y = start_y + f_i * y_step2;
    float ac_z = start_z + f_i * z_step2;

    float length = ab_length - (f_i * length_step);
    long ab_steps = vRound(length / step + 0.5);
    float ab_factor = length / (ab_length * (float) ab_steps);

    float ab_x_step = ab_dx * ab_factor;
    float ab_y_step = ab_dy * ab_factor;
    float ab_z_step = ab_dz * ab_factor;
    for (j = 0; j <= ab_steps; j++)
    {
      float f_j = (float) j;
      float x = ac_x + f_j * ab_x_step;
      float y = ac_y + f_j * ab_y_step;
      float z = ac_z + f_j * ab_z_step;
      points->push_back(vPoint3f(x, y, z));
    }
  }

  points->push_back(vPoint3f(corners[2].x, corners[2].y, corners[2].z));

  long points_after = points->size();
  long result = points_after - points_before;
  return result;
}


// result(i, j) will be the mean of all entries in band i, column j.
vMatrix<float> vColumnMeans(v3dMatrix<float> * values_matrix)
{
  long rows = values_matrix->Rows();
  if (rows == 0) return vMatrix<float>();
  long cols = values_matrix->Cols();
  long bands = values_matrix->Bands();
  vMatrix<float> result_matrix(bands, cols);
  vArray3(float) values = values_matrix->Matrix3();
  vArray2(float) result = result_matrix.Matrix2();

  long row, col, band;
  for (band = 0; band < bands; band++)
  {
    for (col = 0; col < cols; col++)
    {
      float sum = 0;
      for (row = 0; row < rows; row++)
      {
        sum = sum + values[band][row][col];
      }
      result[band][col] = sum / (float) rows;
    }
  }

  return result_matrix;
}


// result(i, j) will be the median of all entries in band i, column j.
vMatrix<float> vColumnMedians(v3dMatrix<float> * values_matrix)
{
  long rows = values_matrix->Rows();
  if (rows == 0) return vMatrix<float>();
  long cols = values_matrix->Cols();
  long bands = values_matrix->Bands();
  vMatrix<float> result_matrix(bands, cols);
  vArray3(float) values = values_matrix->Matrix3();
  vArray2(float) result = result_matrix.Matrix2();

  long row, col, band;
  long index = 0;
  long median_rank = rows / 2 + 1;
  for (band = 0; band < bands; band++)
  {
    for (col = 0; col < cols; col++)
    {
      vector<float> row_values(rows);
      for (row = 0; row < rows; row++)
      {
        row_values[row] = values[band][row][col];
      }
      result[band][col] = kth_smallest_ca(median_rank, &row_values, &index);
    }
  }

  return result_matrix;
}


// result(i, j) will be the k-th largest value (starting at 1) of all entries in 
// band i, column j.
vMatrix<float> vColumnKth(v3dMatrix<float> * values_matrix, long k)
{
  long rows = values_matrix->Rows();
  if (rows == 0) return vMatrix<float>();
  if (k <= 0) return vMatrix<float>();
  if (k > rows) return vMatrix<float>();
  long cols = values_matrix->Cols();
  long bands = values_matrix->Bands();
  vMatrix<float> result_matrix(bands, cols);
  vArray3(float) values = values_matrix->Matrix3();
  vArray2(float) result = result_matrix.Matrix2();

  long row, col, band;
  long index = 0;
  for (band = 0; band < bands; band++)
  {
    for (col = 0; col < cols; col++)
    {
      vector<float> row_values(rows);
      for (row = 0; row < rows; row++)
      {
        row_values[row] = values[band][row][col];
      }
      result[band][col] = kth_smallest_ca(k, &row_values, &index);
    }
  }

  return result_matrix;
}


